/*	Copyright (c) 1987 Motorola Inc
	All Rights Reserved
		@(#)autoload.c	35.1
*/
/*
                          autoload.c

    run this from ".profile" to automatically download a gt terminal
    if it needs it

*/

#include <fcntl.h>
#include <stdio.h>
#include <termio.h>
#include <sys/signal.h>


#define TIMELIMIT 5          /* wait this many seconds for gt reply */
#define NREPLY  512          /* accept this many bytes in reply */
#define DLTEST  "\033[=0b"   /* send this to gt to get its status */
#define DLTLEN  5            /* length of DLTEST string */
#define ESCAPE '\033'

extern int  errno;

int  ifd,      /* input file descriptor for terminal */
     ofd,      /* output file descriptor for terminal */
     ccnt;     /* count of chracters read */

char  ibuf[NREPLY];       /* text of reply held here */

struct termio  omodes;    /* original terminal modes stored here */


main(argc, argv)
char **argv;
{
  int  rs,          /* status of "read" */
       stat,        /* exit status of command executed by "system" */
       i,           /* g.p. loop counter */
       timeout();   /* function called when we get SIGALRM */


  char  c,              /* last char read */
        *ibp = ibuf;    /* ptr to current slot in buffer */

  struct termio  nmodes;    /* new modes set up here */

  unsigned char  *cpp;		/* for resetting terminal modes */

  static char cmdstr[50] = 	/* command string for "system" */
		{"tdl /usr/lib/dl.term/" };

  ofd = 1;


  /*  get another path to stdin, set it for "no delay"  */


  if( (ifd = open("/dev/tty", O_RDONLY|O_NDELAY))  ==  -1 ) {
    exit(errno);
  }

  /*  get current terminal modes  */

  if( ioctl(ifd, TCGETA, &omodes)  ==  -1 )  {
    exit(errno);
  }

  /* now reset modes to turn off echo, etc. */

  nmodes = omodes;
  nmodes.c_lflag  &=  ~(ISIG|ICANON|ECHO);
  cpp = nmodes.c_cc;

  for( i=0; i<NCC; i++ )
     *cpp++ = '\0';        /* MIN=0, TIME=0, no special chars */

  if( ioctl(ifd, TCSETAF, &nmodes)  ==  -1 ) {
    quit(errno);
  }

 
  /*  set trap for SIGALRM, then set up the alarm itself  */

  if( signal(SIGALRM, timeout)  ==  -1 ) {
    quit(errno);
  }

  alarm((unsigned)TIMELIMIT);

  
  /*  write id string to terminal  */

  ioctl(ifd, TCFLSH, 0);    /* first flush the input queue */

  if( write(ofd, DLTEST, (unsigned)DLTLEN)  ==  -1 ) {
    quit(errno);
  }


  /*  read chars from terminal until:

         error in input; or

         character read is 'b'  (response terminator); or

         NREPLY chars have been read

         timeout (TIMELIMIT seconds without input)                 */


  do {

     c = '\0';

     if( (rs = read(ifd, ibp, (unsigned)1))  ==  -1 )  {
       printf("\n ERROR %d in read\n", errno);
       quit(errno);
     }

     if( rs )  {
       c = *ibp++;
       if( ++ccnt >= NREPLY ) {  /* too many bytes in reply */
         quit(6);
       }
     }


  } while( c  !=  'b' );

  /*  we have complete reply now. if second char of reply is '='
      gt needs to be downloaded.
  */

  alarm(0);

  /*  find the reply in the buffer. it starts with an "escape"  
      anything before the escape is typeahead by an overanxious user
      all typeahed characters will be LOST                          */

  for( i=0; i<NREPLY; i++ )  
     if( ibuf[i]  ==  ESCAPE )
       break;

  if( i >= NREPLY )  {  /* escape was not found - must not be a gt */
    quit(0);
  }

  if( ibuf[i+2]  !=  '>' )  {
    ioctl(ifd, TCSETA, &omodes);
	if (argc != 2 || strlen (argv[1]) > (sizeof (cmdstr) - 5))
		quit(1);
	strcat (cmdstr, argv[1]);
    stat = system ( cmdstr );
    quit(stat);
  }

  quit(0);
}




/*  call "quit" instead of "exit" if terminal modes have been changed  */


quit(code)
int  code;            /* exit status */
{

   alarm(0);

   ioctl(ifd, TCSETA, &omodes);

   close(ifd);
   close(ofd);

   exit(code);
}





/*  function called on SIGALRM - read loop timed out  */

timeout()
{
   
  quit(25);
}
