/*		@(#)alts.c	35.1			*/	
/*
 *	Copyright (c) 1987 Motorola Inc.  
 *		All Rights Reserved.
 */
#include "dinit.h"

/* assign alternate tracks to the bad tracks */

assignalt()
{
	int	comp_dist();
	int	distance, d, i, j, k, dups;
	struct alt_info *p1, *p2, *p3; /* ptr to alt_info array 	  */
	unsigned *tmp_alt_lst, *tp1 = NULL;  /* temp list of alternates */

	if (bad_cnt == 0) {		/* no bad spots at all! */
		VID.vid_btl = 0;
		return;
	}

	/* The "alts_used" count is used to keep track of how many 	*/
	/* alternates have been assigned. (Zeroed when declared)	*/
	/* This may be much less than	bad_cnt in the case of a */
	/* controller with sector slip.			*/
	/* controller with sector slip.			*/

	/* remove any bad alternate tracks 				*/
	/*								*/
	/* NOTE:							*/
	/* 	this "search" for bad alternates MUST occur before the  */
	/*	alternates and the bad tracks get "scrambled" i.e. they */
	/*	MUST be in ascending order.				*/

	for(p1= p2= sv_ptr, i=0, k=0; i < bad_cnt && k < altcnt; ) {
		if (( j= p1->dk_eqv - p2->dk_avalt) < 0 ) {
			p1++;		/* step "track equiv" index */
			i++;		/* inc. "bads checked" */
		}
		else if ( j > 0 ) {
			p2++;		/* step "alt index" */
			k++;		/* inc. "alts checked" */
		}
		else {			/* equal, a bad one */
			p2->dk_avalt = -1;	/* don't use this alt */
			p1++;			/* step "track equiv" index */
			p2++;			/* step "alt index" */
			i++;			/* inc. "bads checked" */
			k++;			/* inc. "alts checked" */
			alts_used++;		/* another one bites the dust */
		}
	}

	/* compute heuristic for each track: distance to closest alternate */

	for (p1 = sv_ptr, i=0; i < bad_cnt; p1++, i++) {
		distance = BIG;
		for (p2 = sv_ptr, j=0; j < altcnt; p2++, distance=d, j++ ) {
			if ((d = p1->dk_eqv - p2->dk_avalt) < 0)
				d = -d;
			if (d >= distance)
				break; /* distance increasing, can quit */
		}
		/* save distance of this bad spot to nearest alternate */
		p1->dk_dist = distance;
	}

	/* save alternate list (it may have been changed if there was a */
	/* bad alternate */

	if(( tmp_alt_lst = (unsigned *) malloc( altcnt*sizeof(unsigned))) == NULL ) {
		fprintf( stderr, "dinit: out of space\n" );
		exit(ENOSPC);
	}

	for( i=0, tp1 = tmp_alt_lst, p2 = sv_ptr; i < altcnt; i++, p2++ )
		tp1[i] = p2->dk_avalt; /* copy alt. list info to tmp */

	/* sort alt_info array by "distance" to determine order of */
	/* bad spot assignments					   */

	p1= sv_ptr;

	qsort(p1, bad_cnt, sizeof(struct alt_info ), comp_dist);

	/* after sort, (alt list in alt_info is  "scrambled"), */
	/* restore alt list (to ascending order) */

	for( i=0, tp1 = tmp_alt_lst, p2 = sv_ptr; i < altcnt; i++, p2++ )
		p2->dk_avalt = tp1[i] ; /* restore alt. list info */

	free(tmp_alt_lst);

	/* check for bad diagnostic tracks */

	if (hasdiag) {
		int	tpc = IOS.ios_hds;
		int	head;

		for (i = 0, p1 = sv_ptr; i < bad_cnt; i++) {
		   /* if bad spot in diagnostic cylinders */
		   if ((( p1->dk_eqv >= diaglo) && p1->dk_eqv < diaglo +tpc) ||
		       ((p1->dk_eqv >= diaghi) && (p1->dk_eqv < diaghi +tpc))) {
				distance = BIG;
				head = p1->dk_eqv % tpc;  /* head of bad spot */
				for (k = 0, p2 = sv_ptr; k < altcnt; k++) {
					if ( p2->dk_avalt == -1 )     /* used */
						continue;
					/* assign some alternate */
					if ((d = p1->dk_eqv - p2->dk_avalt) < 0)
						d = -d;
					if((d < distance) && (p2->dk_avalt % tpc == head)) {
						/* assign alt on same head */
						distance = d;
						p3=p2; /* save alt just found */
						/* and keep looking for */
						/* closer alts */
					}
				}
				for (k = 0; k < tpc; k++)
				/* update diagnostic dir to point to alt */
					if (diagdir[k].d_track1 == p1->dk_eqv )
						diagdir[k].d_track1 = p3->dk_avalt;
					else if (diagdir[k].d_track2 == p1->dk_eqv)
						diagdir[k].d_track2 = p3->dk_avalt;
				/* actually assign the alt */
				p1->dk_alt = p3->dk_avalt;
				p3->dk_avalt = -1;	/* mark "used" */
				alts_used++;	/* ...and another one ... */
			}
		}
	}
		
	/* inititialize "dups" entries */
	for (i=0, p1 = sv_ptr; i < info_size; i++, p1++ )
		p1->dk_dups = -1;
	
	/* check for dup. track equivalents */
	for (i=1, p1 = sv_ptr; i <= bad_cnt; i++, p1++ ) { 

		/* if entry already found, skip checks */
		if ( p1->dk_dups != -1 )
			continue;

		/* new entry, mark as "unique" */
		p1->dk_dups = i;

		/* search for other entries for equivalent tracks */
		for (j=0, p2 = p1 +1; j < (bad_cnt -i); j++, p2++)
			 if (p1->dk_eqv == p2->dk_eqv)
				/* mark same as "unique" entry */
				p2->dk_dups = i;
	}
	
	/* count duplicates */
	for (j=0, p1 = sv_ptr; j < bad_cnt; j++, p1++) {

		/* don't match if "-1" it's already processed */
		if (p1->dk_dups == -1)
			continue;

		/* always check complete list */
		for ( dups=0, k=0, p2 = sv_ptr; k < bad_cnt; k++, p2++) {
			if ( p1->dk_dups == p2->dk_dups)
				dups++;	/* found a match */
		}

		/* at this point, dups >= 1 */

		/* now that we may have duplicates,
		 * see if they exceed "sector slip"
		 */

		if ((dups <= IOS.ios_ssc ) && (p1->dk_sector != LFLAG) &&
			(p1->dk_dist > 0)) {
			/* handle with sector slip unless we don't know where */
			/* the defect actually is (denoted by dk_sector ==    */
			/* LFLAG) and not an alternate, dist == 0 means */
			/* it's an alt, don't assign an alt to itself */
			/* only flag from current point */
	   		for (k=0, p2 = p1 +1; k < (bad_cnt -j); k++) {
		   		if (( p1->dk_dups == p2->dk_dups) && (p2->dk_sector !=LFLAG)) 
			   		p2->dk_dups = -1; /* needs no alt */
	   		}
			p1->dk_dups = -1; /* do NOT assign an alt */
		}
	}

	
	/* allocate in heuristic order the closest available alternate */

	for (p1 = sv_ptr, i=0; i < bad_cnt; p1++, i++) {

		/* if no alt is required, skip this entry 		 */
		/* the reasons are: 1) it may be handled by sector slip	 */
		/* 		    2) it may be on the same track and	 */
		/*		       may have been previously assigned */
		/*		       as a "duplicate"			 */

		if (p1->dk_dups == -1) /* dups == -1 means no alt needed */
			continue;

		distance = BIG;
		for (p2 = sv_ptr, j=0; j < altcnt; p2++, j++ ) {
			/* if alt is already used, skip this entry */
			if ( p2->dk_avalt == -1 )
				continue; 
			if ((d = p1->dk_eqv - p2->dk_avalt) < 0)
				d = -d;
			if (d >= distance) {
				break;
			}
			distance=d;
			p3=p2;	/* save ptr to "current" alt */
		} 

		if (alts_used == altcnt) { 
			fprintf(stderr, "dinit: not enough alternate tracks\n");
			exit(2);
		}

		p1->dk_alt = p3->dk_avalt;	/* assign alternate */
		p3->dk_avalt = -1;		/* mark alt as "used" */
		alts_used++;			/* ...and another one ... */

		/* check for dup. values and if found, assign the same alt */

		for (p2 = sv_ptr, j=0; j < bad_cnt; j++, p2++ ) {
			if( p1 == p2 ) /* same entry */
				continue;
			if( p2->dk_dups == p1->dk_dups ) {  /* match ? */
				p2->dk_alt = p1->dk_alt;  /* assign same alt */
				p2->dk_dups= -1;       /* mark as "assigned" */
			        /* to prevent accidental reassignment of alt */
			}
		}

		/* prevent accidental reassignment of alt */
		p1->dk_dups= -1;       /* mark as "assigned" */
	}
}

/* altscyl: check for alternates on the given cylinder */
altscyl(cyl)
register cyl;
{	register unsigned *alst;
	register i;

	if(( alst = alt_lst ) == NULL )
		return(0);	/* no alternates */

	if( IOS.ios_hds == 0 )
		return(0);	/* invalid head count */

	/*
	 *	search through alternates list
	 */

	for( i=0; i<altcnt; i++, alst++ )
		if( *alst/IOS.ios_hds == cyl )
			return(1);
	return(0);
}

/* assign alternate for new bad spot */

assignnew()
{
	int	comp_dist();
	int	distance, d, i, j, k, dups, done;
	struct dkbadlst *bl;
	struct alt_info *p1, *p2, *p3, *p_newspot, *p_badalt;
	unsigned *tmp_alt_lst, *tp1 = NULL;  /* temp list of alternates */
	int badalt, newspot;

	if (bad_cnt == 0) {		/* no bad spots at all! */
		VID.vid_btl = 0;
		return(-1);
	}

	p_newspot = &sv_ptr[bad_cnt-1];	/* address of "new bad spot " */ 
	p_badalt = NULL;
	newspot = p_newspot->dk_eqv;		/* track equivalent of the new bad spot */

	/* check for bad spot already on the list */ 
	/* NOTE: this applies to 321 & 360 types */

	done = 1;
	for (p1 = sv_ptr, i=0; i < bad_cnt; p1++, i++) {
		if(p1 == p_newspot)
			continue;
		if(p_newspot->dk_eqv == p1->dk_eqv) {	/* already a bad spot */
			if(p1->dk_alt !=0) {				/* already has alt	*/
				badalt = p1->dk_alt;			/* save bad alt */
				p1->dk_alt = 0;				/* assign a NEW alt  (later) */
				if(bad_cnt == maxbad) {
					fprintf(stderr,"dinit: too many bad spots\n");
					exit(-1);
				}
				if(done) {
					p_badalt = &sv_ptr[bad_cnt++];	/* place to put bad alt */
					p_badalt->dk_eqv = badalt;
					if (( VID.vid_strat == (DKSOFTWARE & 0xff))
						|| (VID.vid_strat == DKHARDWARE )) {
						p_badalt->dk_cyl = badalt;
						p_badalt->dk_head = p_badalt->dk_sector = 0;
					}else {
						p_badalt->dk_sector = LFLAG;
						p_badalt->dk_head = badalt % IOS.ios_hds;
						p_badalt->dk_cyl = badalt / IOS.ios_hds;
					}
					done =0;		/* only put bad alt on list once */
				}
			}
		}
	}

	/* check for bad spot on an alternate already in use */ 

	for (p1 = sv_ptr, i=0; i < bad_cnt; p1++, i++) {
		if(p_newspot->dk_eqv == p1->dk_alt) {	/* alt already in use */
				badalt = p1->dk_alt;			/* save bad alt */
				p1->dk_alt = 0;				/* assign a NEW alt  (later) */
				if(done)	{	/* don't do this if you have done for 321 */
					p_badalt = p1;					/* dummy up to assign alt */
					newspot = p1->dk_eqv;		/* spot to read/restore data */
				}
				if((bad_cnt +1) == maxbad) { /* bad_cnt +1 cuz we need 2 alts */
					fprintf(stderr,"dinit: too many bad spots\n");
					exit(-1);
				}
			}
		}
	
	/* check for bad diagnostic tracks */

	if (hasdiag) {
		int	tpc = IOS.ios_hds;
		int	head;

		for (i = 0, p1 = sv_ptr; i < bad_cnt; i++) {
		   /* if bad spot in diagnostic cylinders */
		   if ((( p1->dk_eqv >= diaglo) && p1->dk_eqv < diaglo +tpc) ||
		       ((p1->dk_eqv >= diaghi) && (p1->dk_eqv < diaghi +tpc))) {
				distance = BIG;
				head = p1->dk_eqv % tpc;  /* head of bad spot */
				for (k = 0, p2 = sv_ptr; k < altcnt; k++) {
					if ( p2->dk_avalt == -1 )     /* used */
						continue;
					/* assign some alternate */
					if ((d = p1->dk_eqv - p2->dk_avalt) < 0)
						d = -d;
					if((d < distance) && (p2->dk_avalt % tpc == head)) {
						/* assign alt on same head */
						distance = d;
						p3=p2; /* save alt just found */
						/* and keep looking for */
						/* closer alts */
					}
				}
				for (k = 0; k < tpc; k++)
				/* update diagnostic dir to point to alt */
					if (diagdir[k].d_track1 == p1->dk_eqv )
						diagdir[k].d_track1 = p3->dk_avalt;
					else if (diagdir[k].d_track2 == p1->dk_eqv)
						diagdir[k].d_track2 = p3->dk_avalt;
				/* actually assign the alt */
				p1->dk_alt = p3->dk_avalt;
				p3->dk_avalt = -1;	/* mark "used" */
				alts_used++;	/* ...and another one ... */
			}
		}
	}
		
	/* inititialize "dups" entries */
	for (i=0, p1 = sv_ptr; i < info_size; i++, p1++ )
		p1->dk_dups = -1;
	
	/* check for dup. track equivalents */
	for (i=1, p1 = sv_ptr; i <= bad_cnt; i++, p1++ ) {

		/* if entry already found, skip checks */
		if ( p1->dk_dups != -1 )
			continue;

		/* new entry, mark as "unique" */
		p1->dk_dups = i;

		/* search for other entries for equivalent tracks */
		for (j=0, p2 = p1 +1; j < (bad_cnt -i); j++, p2++)
			 if (p1->dk_eqv == p2->dk_eqv)
				/* mark same as "unique" entry */
				p2->dk_dups = i;
	}

	/* allocate alternate(s) for the new bad spot (and bad alt, if any) */

	p1 = p_newspot;

doit:
		distance = BIG;
		for (p2 = sv_ptr, j=0; j < altcnt; p2++, j++ ) {
			/* if alt is already used, skip this entry */
			if ( p2->dk_avalt == -1 )
				continue; 
			if ((d = p1->dk_eqv - p2->dk_avalt) < 0)
				d = -d;
			if (d >= distance) {
				break;
			}
			distance=d;
			p3=p2;	/* save ptr to "current" alt */
		} 

		if (alts_used == altcnt) { 
			fprintf(stderr, "dinit: not enough alternate tracks\n");
			exit(2);
		}

		p1->dk_alt = p3->dk_avalt;	/* assign alternate */
		p3->dk_avalt = -1;		/* mark alt as "used" */
		alts_used++;			/* ...and another one ... */

	/* now see if any "old bad spots" match the new entry */

		if(p1 == p_newspot) {	
		/* only need to check for dups if spot was entered */
		/* a bad (used) alternate will only appear in the list once */
	   	for (p2 = sv_ptr; p2 < p_newspot; p2++) {
		  		if((p1->dk_dups == p2->dk_dups) && (p1->dk_alt != 0)) {
					p2->dk_alt = p1->dk_alt;	/* assign same alt */
					p2->dk_dups = -1; /* needs no alt */
				}
			}
		}

	/* prevent accidental reassignment of alt */
	p1->dk_dups= -1;       /* mark as "assigned" */

	if(p_badalt != NULL) { /* loop if an alternate went bad */
		p1 = p_badalt;
		p_badalt = NULL;
		goto doit;
	}
	return(newspot);
}

/*
 *	savedata - Try to save the data or at least as much of it as possible
 *	   for restoration after the alternate is assigned and the
 *	   controller is notified
 */

savedata(trk)
register unsigned trk;
{
	register i;
	register  fid;
	register lss = IOS.ios_rec; /* logical sector size */
	int spt = IOS.ios_spt; /* sectors per track */
	register boff = trk*spt*lss; /* byte offset of trk */
	register bpt = (spt*lss)/BSIZE; /* blocks per track */
	register char *block;
	register char **tp;

#ifdef CMDEBUG
	printf("CMDEBUG: saving data on trk %d,block %d(%x)\n",trk,boff,boff);
#endif
	/* allocate space for pointers to block read */

	if((trksave = (char **) malloc(sizeof(char *)*bpt)) == NULL ) {
		fprintf( stderr, "dinit: out of space\n" );
		exit(ENOSPC);
	}
	if((fid = open(rdev, O_RDONLY ) )== -1) {
		fprintf(stderr,
			"dinit: cannot open raw device `%s'\n", rdev);
		exit(ENXIO);
	}
	if (lseek(fid, boff, 0) == -1) {/* seek to bad track */
		fprintf(stderr, "dinit:  seek error on track %d\n",trk);
		exit(1);
	}

	for (i = 0,block = NULL,tp=trksave; i < bpt; tp++,i++) {
		*tp = NULL; /* clear pointer */
		if(block == NULL)
			/* allocate space for block */
			if((block = (char *) malloc( BSIZE )) == NULL ) {
				fprintf( stderr, "dinit: out of space\n" );
				exit(ENOSPC);
			}
		if(read( fid,block,BSIZE) != BSIZE) { /* bad sector */
#ifdef CMDEBUG
			printf("CMDEBUG: found bad block %d\n"
				,(boff+i*BSIZE)/BSIZE);
#endif
			if (lseek(fid, boff+((i+1)*BSIZE), 0) == -1) {
				fprintf(stderr, "dinit:  seek error on track %d\n",trk);
				exit(1);
			}
			
		} else {
			*tp = block;
			block = NULL;
		}
		
	}
	close(fid);
}

/*
 *	restdata - restore the data to the new trk
 */

restdata(trk)
register unsigned trk;
{
	register i;
	register fid;
	register lss = IOS.ios_rec; /* logical sector size */
	register spt = IOS.ios_spt; /* sectors per track */
	register bpt = (spt*lss)/BSIZE; /* blocks per track */
	register boff = trk*spt*lss; /* byte offset of trk */
	register char **tp;

#ifdef CMDEBUG
	printf("CMDEBUG: restoring data on track %d ,block %d(%x)\n",trk,boff,boff);
#endif
	if((fid = open(rdev, O_WRONLY ) )== -1) {
		fprintf(stderr,
			"dinit: cannot open raw device `%s'\n", rdev);
		exit(ENXIO);
	}
	if (lseek(fid, boff, 0) == -1) {/* seek to new track */
		fprintf(stderr, "dinit:  seek error on track %d\n",trk);
		exit(1);
	}

	for (i = 0,tp = trksave; i < bpt; tp++,i++) {
		if(*tp == NULL) { /* this was a bad sector, no data saved */
			/* seek to next sector */
			if (lseek(fid, boff+((i+1)*BSIZE), 0) == -1) {
				fprintf(stderr,
				"dinit:  seek error on track %d\n",trk);
				exit(1);
			}
			continue;
		}
		if(write( fid, *tp, BSIZE) != BSIZE) { /* bad sector */
			fprintf(stderr,
			"dinit:  write error on track %d\n",trk);
			exit(1);
		}
	}
}
