/*		@(#)diag.c	35.1			*/	
/*
 *	Copyright (c) 1987 Motorola Inc.  
 *		All Rights Reserved.
 */
#include "dinit.h"

/* initialize diagnostic test area directory */

initdiag()
{
	int	tracksz = IOS.ios_psm *IOS.ios_spt;
	int	cylsz = tracksz *IOS.ios_hds;
	int	i, len, loc, diag_sz;
	int tmp;
	register struct diagent *p;

	lss = IOS.ios_rec;

	loc = lss * 2;		/* skip over volume id and config entries */

/* this is NOT device independent code !!! */
/* and should be modified to handle all permutations of blocksize/sectorsize */
/* also watch for similar code in writhead.c */

	/* compute size of diagnostic directory, if any */
	diag_sz = (hasdiag != 0) ? lss : 0;

	if ((VID.vid_strat == (DKSOFTWARE & 0xff)) || (VID.vid_strat == DKHARDWARE)) {
		loc += lss;	/* skip hash table */
		tmp = (maxbad * sizeof(struct badent) + diag_sz);
		loc = ((loc + tmp + BSIZE - 1)/BSIZE) * BSIZE; /* skip bad spot info and diag. dir. */
	} else if( VID.vid_strat != DKPERFECT ) {
		/* skip hardware bad spot information */
		loc += ((maxbad * sizeof(struct dkbadlst) + diag_sz + BSIZE - 1)/BSIZE) * BSIZE; /* skip bad spot info and diag. dir. */
	}

	/* if there is a slice table, bump "loc" to allow for it */

	if (IOS.ios_slccnt != 0) {  
		/* compute slice table location, allow room for a boot loader */

		IOS.ios_slcblk = (loc + MAXBOOT + (BSIZE-1))/BSIZE; 
		/* force to block boundary*/

		loc += (((sizeof(struct slice)*IOS.ios_slccnt)+BSIZE-1)/BSIZE)
			 *BSIZE;
	}

	if (!hasdiag)	/* we have set up slice block ptr (in config block) */
			/* no need to continue if no diagnostic tracks      */
		return;

	/*
	 *	compute diagnostic information 
	 */

	diaglo = (loc + MAXBOOT + (cylsz - 1)) / cylsz;	/* cyl # */
	while( altscyl(diaglo) )	/* alternates on cylinder? */
		diaglo++;		/* try next */
	diaglo *= IOS.ios_hds;		/* convert to track number */

	diaghi = (IOS.ios_trk - 1);	/* cyl # */
	while( altscyl(diaghi) )	/* alternates on cylinder? */
		diaghi--;		/* try prev */
	diaghi *= IOS.ios_hds;		/* convert to track number */

	diagdir = (struct diagent *) malloc((IOS.ios_hds + 1)
					*  sizeof(struct diagent));

	if (diagdir == NULL) {
		fprintf(stderr, "dinit: cannot allocate diagnostic dir.\n");
		exit(2);
	}
	len = tracksz / lss;

	/*
	 *	initialize diagnostic track data
	 */

	for (i = 0, p = diagdir; i < IOS.ios_hds; i++, p++) {
		p->d_track1 = diaglo + i;
		p->d_track2 = diaghi + i;
		p->d_len1 = p->d_len2 = len;
	}
	p->d_track1 = p->d_track2 = p->d_len1 = p->d_len2 = 0;
}


/* copy the diagnostic tracks to the disk */

copydiag(fid)
int	fid;
{
	int	tracksz = IOS.ios_psm *IOS.ios_spt;
	char	*buffer;
	register struct diagent *p;
	int	lspt, diag_sz;
	register int	i;

	lss = IOS.ios_rec;

	loc = lss * 2;		/* skip over volume id and config entries */

	if ((VID.vid_strat == (DKSOFTWARE & 0xff)) || (VID.vid_strat == DKHARDWARE)) {
		loc += lss;	/* skip hash table */

		loc += (((maxbad * sizeof(struct badent))+ lss - 1)/lss) * lss; /* skip bad spot info and diag. dir. */
	} else if (VID.vid_strat != DKPERFECT) {
		/* skip bad spot info */
		loc += ((maxbad * sizeof(struct dkbadlst) + lss - 1)/lss) * lss;
	}

	/* switch from track # to logical sector number */

	lspt = tracksz / lss;
	for (i = 0, p = diagdir; i < IOS.ios_hds; i++, p++) {
		p->d_track1 *= lspt;
		p->d_track2 *= lspt;
	}

	lseek(fid, loc, 0);
	write(fid, diagdir, (IOS.ios_hds + 1) * sizeof(struct diagent ));
	VID.vid_dta = loc / lss;

	buffer = malloc(tracksz);
	if (buffer == NULL) {
		fprintf(stderr, "dinit: cannot allocate track buffer\n");
		exit(2);
	}

	diagtrk(buffer, tracksz);

	for (i = 0, p = diagdir; i < IOS.ios_hds; i++, p++) {
		lseek(fid,  p->d_track1 * lss, 0);
		write(fid, buffer, tracksz);
		lseek(fid,  p->d_track2 * lss, 0);
		write(fid, buffer, tracksz);
	}
	free(buffer);
}


/* generate diagnostic disk track pattern (based on VERSAdos assembly
   routine RANDOM in INIT utility but is hopefully machine independent now) */

static unsigned int	seed = 0x00092855;

diagtrk(buffer, tracksize)
register int	tracksize;
register char	*buffer;
{
	register int	i;
	int	r;

	for (i = 0; i < tracksize - 3; i += 4) {
		r = random();
		*buffer++ = r >> 24;
		*buffer++ = r >> 16;
		*buffer++ = r >>  8;
		*buffer++ = r;
	}
}


/* a (not very) random function for producing diagnostic data */

static int	random()
{
	register int	xbit;

	if (seed == 0)
		seed = 0xff;
	xbit = ((((seed & 0xff) >> 3) ^ (seed & 0xff)) >> 1) & 1;
	seed = (seed >> 1) | xbit << 31;
	if (seed == 0xffffffff)
		seed++;
	return (seed);
}

