/*	@(#)dinit.c	35.1		*/
/*
 *	Copyright (c) 1987 Motorola Inc.  
 *		All Rights Reserved.
 */

/* Disk formatting and initialization utility */
/*
 *	options:
 *		-a	- if "r option", print alternates read too
 *		-b file - insert contents of file (a.out format) as bootloader
 *		-d desc	- put "desc" into vid_vd (volume description) field
 *		-e	- substitute "EXORMACS" for "MOTOROLA" in vid_mac
 *		-f	- format the disk
 *		-h file - read manufacturers defect list from drive
 *				and create defect list file
 *		-i	- inquire for gap,inodes... for "mkfs"
 *		-n	- add a new bad track/spot
 *		-m	- use ddefs info to "mkfs" (only after format)
 *		-o	- [obsolete] override previous information on disk
 *		-r	- read bad spot list from disk and print it on stdout
 *		-s	- skip save and restore data when adding new bad spot
 *		-t file	- take bad track numbers from file (instead of user)
 *		-x	- if "r option", print in hex
 *		-T 	- take bad spot info in track format
 *				(default is head/track)
 */


/*
 *	global variables
 */

#include "dinit.h"

char	*bootloader;		/* bootloader file name 	*/
char	*badname=NULL;		/* bad spots file name 		*/
char	*type;			/* disk type name 		*/
char	*rdev;			/* raw device name (slice 7)	*/
char	bdev[BUFSIZ];		/* block device name (slice 7)	*/
char	formatter[BUFSIZ];	/* disk formatter name 		*/
char	rdprog[BUFSIZ];		/* manufacturer defect list program */
char	*mfrfile; 		/* name of file to create	*/
char	motname[8];		/* 'MOTOROLA' or 'EXORMACS' 	*/
char	volname[20];		/* volume name			*/
int	newformat = 0;		/* new format request		*/
int	bad_cnt = 0;		/* number of bad spots 	*/
int	loc;			/*				*/
int	lss;			/*				*/
int	updown;			/* tells compdk to sort low to high */
struct dkbadlst *bad_lst;	/*				*/
struct dkbadlst *new_spot;	/* cyl/head/logical sector, alt for new bad spot*/
unsigned *alt_lst = NULL;	/* ptr to list of alternates 	*/
struct alt_info *alt_ptr, *sv_ptr; /* ptr to alt_info array 	*/
int	altcnt = 0;		/* number of alternates  	*/
int info_size;			/* size of "info" array  	*/
unsigned char	def_vid_strat;	/* bad spot strategy (from ddefs file) */
unsigned short	def_vid_btl;	/* Max. number of bad spots (from ddefs file) */
unsigned char	def_ios_hds;	/* # of heads from ddefs file */
unsigned short	def_ios_trk;	/* # of cylinders from ddefs file */
int	alts_used= 0;		/* number of alternate track used */
char **trksave;			/* list of saved sectors for recovery */

/*
 *	option flags
 */

char	doformat = NO;		/* format the disk */
char	makefs = NO;		/* do NOT build file systems and slice table  */
char	iopt = NO;		/* do not ask for gap size, inodes for mkfs   */
char	eopt = NO;		/* flag for "-e" option  */
char	dopt = NO;		/* flag for "-d" option  */
char	nopt = NO;		/* flag for "-n" option  */
char	ropt = NO;		/* flag for "-r" option  */
char	Ropt = NO;		/* flag for "-R" option  */
char	aopt = NO;		/* flag for "-a" option  */
char	xopt = NO;		/* flag for "-x" option  */
char	topt = NO;		/* flag for "-T" option  */
char	sopt = NO;		/* flag for "-s" option  */

/*
 *	configuration information
 */

int	fs0_size = 0;		/* #blocks in slice 0 for 'mkfs(1M)' */
int	fs1_size = 0;		/* #blocks in slice 1 for 'mkfs(1M)' */
int	fs1_slice = 0;		/* slice to use for filesystem 1 (/usr) */
int	sw_size = 0;		/* #blocks in swap area */
int	sw_slice = 0;		/* slice to use for swap area */
int	maxbad = 0;		/* maximum number of bad tracks per disk */
struct badent *bttable;		/* bad track table */
char	hasboot = NO; 		/* copy in the bootloader */
int	reserved = 0;		/* number of reserved blocks at end */
struct slice *slctab = NULL;	/* slice table pointer */

/*
 *	diagnostic test area support
 */

char	hasdiag = NO; 		/* create diagnostic tracks */

struct diagent *diagdir;	/* diagnostic test area directory */

int	diaglo;		/* the starting track of the first diagnostic area */
int	diaghi;		/* the starting track of the second diagnostic area */

/*
 *	volume id and config blocks
 */

struct dkblk0 dkblk0;
unsigned char	diagpat[] = {	/* diagnostic pattern for vid */
	0x0f, 	0x1e, 	0x2d, 	0x3c, 	0x4b, 	0x5a, 	0x69, 	0x78,
	0x87, 	0x96, 	0xa5, 	0xb4, 	0xc3, 	0xd2, 	0xe1, 	0xf0, 	
	0x0f, 	0x1e, 	0x2d, 	0x3c, 	0x4b, 	0x5a, 	0x69, 	0x78,
	0x87, 	0x96, 	0xa5, 	0xb4, 	0xc3, 	0xd2, 	0xe1, 	0xf0, 	
	0xf1, 	0xf2, 	0xf4, 	0xf8, 	0xf9, 	0xfa, 	0xfc, 	0xfe,
	0xff, 	0x7f, 	0xbf, 	0xdf, 	0xef, 	0x6f, 	0xaf, 	0xcf,
	0x4f, 	0x8f, 	0x0f, 	0x07, 	0x0b, 	0x0d, 	0x0e, 	0x06,
	0x0a, 	0x0c, 	0x04, 	0x08, 	0x04, 	0x02, 	0x01, 	0x00
};

int diagpatsize = sizeof(diagpat);

/*
 *	main: perform program control
 */

main(argc, argv)
int	argc;
char	*argv[];
{
	register fid;
	char command[80];
	char cstring[80];

#ifdef CMDEBUG
	fprintf( stderr, "CMDEBUG is set! \n");
#endif

#ifdef DEBUG
	fprintf( stderr, "DEBUG is set! \n");
#endif

	/*
	 *	initialize data structures
	 */

	initall();

	/*
	 *	process command line arguments
	 */

	scanopts(argc, argv);

	/*
	 *	read the dskdefs file, allocate "alt_info" array
	 */

#ifdef DEBUG
fprintf(stderr, "dinit: /* readdef() */\n");
#endif

	readdef(type);

	if( doformat ) {

		/*
		 *	Read bad spot list
		 */

#ifdef DEBUG
fprintf(stderr, "dinit: /* readlist() */\n");
#endif

		if (Ropt) {	/* if set badname to "default" if Ropt */
			sprintf(command,"%s %s", rdprog, rdev );
			system(command);
			badname="/tmp/.xyz123";
		}
		readlist(badname, info_size);
	
		/*
 		*	mkslice: allocate and initialize slice table
 		*/

		mkslice();

		/*
		 *	compute and assign alternates
		 */

#ifdef DEBUG
fprintf(stderr, "dinit: /* assign alternates for known bad tracks */\n");
#endif

		assignalt();	/* assign alternates for known bad tracks */

#ifdef DEBUG
fprintf(stderr, "dinit: /* build disk image of bad track table and hash table */\n");
#endif

		fillbads();	/* build disk image of bad track table and hash table */


		/*
		 *	Do format or configure
		 */

		fmtdrive();

		/*
		 *	Set up config block ptr to slice table, (if any)
		 *
		 *		=== AND === 
		 *
		 *	Initialize diagnostic tracks, (if any)
	 	*/

#ifdef DEBUG
		{
		fprintf(stderr, "dinit: /* initdiag() */\n", hasdiag);
#endif
	
		initdiag();
#ifdef DEBUG
		}
#endif

		/*
		 *	Write data onto cylinder 0
		 */

		wrtcyl0(bdev, rdev);

		/*  
		 *  For old-style hardware badtrack handling set up a second call 
		 *  to the format routine with a -A option.
		 */

		if (( doformat ) && ( ! newformat )
		 && ( VID.vid_strat == DKHARDWARE )) {
			if(( fid = open(bdev, O_WRONLY)) == -1) {
				fprintf(stderr, "dinit: cannot open block device `%s'\n", bdev);
				exit(ENXIO);
			}
			oldformat(" -A");
			close( fid );
		}

		/*
		 *	finally make a filesystem on the disk
		 */

		if ( makefs == YES ) { /* "-m option" is set */
			if (fs0_size)
				mkfs(rdev, '0', fs0_size);
	
			if (fs1_size)
				mkfs(rdev, fs1_slice+'0', fs1_size );
		}

	} else {

	/* disk has been initialized previously, read data on the disk 
	 *	readhead() will NOT return in the following cases:
	 *	1) readhead reads badtrack/spot list from the disk
	 *	2) add new bad spots if -n option
	 */
	readhead(bdev);

	}

exit(0);	/* leave no footprints */
}


/*
 *	scanopts: scan the command line options
 */

scanopts(argc, argv)
int	argc;
char	*argv[];
{
	int	c;
	extern int	optind;
	extern char	*optarg;

	struct stat rstat, bstat;

	/*
	 *	handle command line options
	 */

	while ((c = getopt(argc, argv, "aefTimnorsxRd:b:t:")) != EOF) {
		switch (c) {
		case 'o':
			fprintf( stderr,"Warning: -o is replaced by ddefs(1M)\n");
			break;
		case 'f':
			doformat = YES;
			break;
		case 'e':
			eopt= YES;
			memcpy(motname, "EXORMACS", 8);
			break;
		case 'd':
			dopt= YES;
			strncpy(volname, optarg, 20);
			break;
		case 'b':
			bootloader = optarg;
			hasboot = YES;
			break;
		case 'm':
 			/* if -m, then use fs info in ddefs file to "mkfs" */
			makefs = YES;
			break;
		case 'i':
 			/* if -i, ask for gap size, inodes during mkfs */
			iopt = YES;
			break;
		case 'n':
 			/* if -n, add a new bad track/spot to the disk */
			nopt = YES;
			break;
		case 'r':
 			/* if -r, read bad track/spot from the disk */
			ropt = YES;
			break;
		case 'R':
 			/* if -R, read manufacturers defect list from the disk */
			Ropt = YES;
			break;
		case 's':
 			/* if -s, skip save/restore data when adding new spot */
			sopt = YES;
			break;
		case 'a':
 			/* if -a, (and -r) then print alts too */
			aopt = YES;
			break;
		case 't':
			badname = optarg;
			break;
		case 'x':
 			/* if -x, print list in hex */
			xopt = YES;
			break;
		case 'T':
			topt=YES;
			break;
		case '?':
			fprintf(stderr, USAGE);
			exit(EINVAL);
		}
	}

	/*
	 *	there must be two additional arguments:
	 *
	 *		1. the disk type name
	 *		2. the raw device
	 */

	if (optind + 2 != argc) {
		fprintf(stderr, USAGE);
		exit(EINVAL);
	}

	/*
	 *	get type name
	 */

	type = argv[optind++];

	/*
	 *	check the raw device
	 */

	if ((stat(rdev = argv[optind], &rstat) == -1) 
	|| (access(rdev, S_RW ) == -1)) {
		fprintf(stderr, "dinit: cannot open %s\n", rdev);
#ifndef	DEBUG
		exit(errno);
#endif
	}

	/*
	 *	The device must be a character device
	 */

	if ((rstat.st_mode&S_IFMT) != S_IFCHR) {
		fprintf(stderr, "dinit: must be raw device - %s\n", rdev);
#ifndef	DEBUG
		exit(EACCES);
#endif

	/*
	 *	Check for proper naming convention
	 */

	} else if (strncmp(rdev, "/dev/r", 6) != 0) {
		fprintf(stderr, 
			"dinit: device name must start with `/dev/r'\n");
#ifndef	DEBUG
		exit(EINVAL);
#endif
	}

	/*
	 *	Construct block device name
	 */

	strcpy(bdev, "/dev/");
	strcat(bdev, rdev + 6);

	/*
	 *	Stat the block device
	 */

	if (stat(bdev, &bstat) == -1 || access(bdev, S_RW) == -1) {
		fprintf(stderr, "dinit: cannot open block device `%s'\n", bdev);
#ifndef	DEBUG
		exit(errno);
#endif
	}

	/*
	 *	Check that device is a block device.
	 */

	if (((bstat.st_mode&S_IFMT) != S_IFBLK)
	   ||(minor(rstat.st_rdev) != minor(bstat.st_rdev))) {
		fprintf(stderr, "dinit: no block device for `%s'\n", rdev);
#ifndef	DEBUG
		exit(ENOTBLK);
#endif
	}
}

/*
 *	initialize all major data structures
 */

initall()
{
	memset( &VID, 0, sizeof(struct volumeid) );
	memset( &IOS, 0, sizeof(struct dkconfig) );
	memset( volname, 0, 20 );
	memcpy( motname, "MOTOROLA", 8 );
}
/* compare two "distances" */

comp_dist(p1, p2)
struct 	alt_info *p1, *p2;
{
	return (p1->dk_dist - p2->dk_dist);
}

/* compare routine for "qsort" sorts two dkbadlst structures */
/* if the global variable updown =0 it sorts high to low */
/* if the global variable updown =1 it sorts low to high */

compdk(p1,p2)
struct dkbadlst *p1, *p2;
{
	register d;

	if ((VID.vid_strat == DKHARDSEC) || (VID.vid_strat == DKHARDSPOT)){
		if(updown == 0) {
			if ( d = p2->dk_cyl - p1->dk_cyl )
				return(d);
			if ( d = p2->dk_head - p1->dk_head)
				return(d);
			return(p2->dk_sector - p1->dk_sector);
		}
		else {	/* updown == 1 */
			if ( d = p1->dk_cyl - p2->dk_cyl )
				return(d);
			if ( d = p1->dk_head - p2->dk_head)
				return(d);
			return(p1->dk_sector - p2->dk_sector);
		}
	}
	if(updown == 0)
		return( p2->dk_track - p1->dk_track);
	else
		return( p1->dk_track - p2->dk_track);
}

