/*
 *	Copyright (c) 1987	Motorola Inc.  
 *		All Rights Reserved.
 */
/*		@(#)dinit.h	35.2		*/

#include "stdio.h"
#include "fcntl.h"
#include "filehdr.h"
#include "scnhdr.h"
#include "ldfcn.h"
#include "string.h"
#include "malloc.h"
#include "errno.h"
#include "ctype.h"

#include "sys/types.h"
#include "sys/stat.h"
#include "sys/sysmacros.h"
#include "sys/dk.h"
#include "sys/ino.h"
#include "sys/param.h"


/*
#define	DEBUG
#define	CMDEBUG
*/

/*
 *	general constants
 */
#define NEWDIAG		/* add new bad diagnotic tracks? */
#undef	NEWDIAG

#define YES	1
#define NO	0
#define BIG	((1 << 31) - 1)
#define FLAG	(0xffff)
#define LFLAG	(0xffffffff)
#define S_RW	((S_IREAD|S_IWRITE)>>6)	/* read/write for access(2) */
#define MAXBOOT	(16 * 1024)		/* maximum bootloader size = 16K */
#define LSS	256			/* logical sector size */
#define BLOCKSZ	512			/* largest supported block size */

/* defines for mkfs */

#define FSBSIZE	(BSIZE*2) /* file system block size */

#define	NBINODE	(FSBSIZE/sizeof(struct dinode))

#ifdef DEBUG
char	*diskdefs;
#endif

/*
 *	Usage string
 */

#define USAGE "usage: dinit [-aefimnrRsTx] [-h file] [-d desc] [-b file] [-t file] type rawdev\n"

/* bad track strategies */

#define	DSKDEFS	"/etc/dskdefs"

/*
 *	software bad track support
 */

unsigned char	hash[HASHSIZE];	/* bad track hash table */


/*
 *	bad spot and alternates list
 */

/* dk_cyl	cylinder number OR track #   	---bad spot info. from "user" */
/* dk_head	head number		  	---bad spot info. from "user" */
/* dk_sector	sector number OR offset  	---bad spot info. from "user" */
/* 
/* dk_avalt	available alternate tracks from dskdefs file */
/* 
/* dk_eqv	track equivalent of bad spot 
/*		(may have been specified as cyl/hd/sector) */
/* 
/* dk_alt	computed "assigned" alternate */
/* 
/* dk_dist	computed "distance" to alt */
/* 
/* dk_dups	dup count for sector slip */

 /*
  *	Hardware bad spot replacement uses cylinder, head, sector
  *	or cylinder, head, byte offset.
  */

struct alt_info {	/* info required to compute alternates, etc. */
	unsigned long	dk_cyl;		/* cylinder number OR track # } bad  */
	unsigned long	dk_head;	/* head number		      } spot */
	unsigned long	dk_sector;	/* sector number OR offset    } info */
	unsigned long	dk_avalt;	/* alternate track from dskdefs file */
	unsigned long	dk_eqv;		/* track equivalent of bad spot      */
	unsigned long	dk_alt;		/* assigned alternate */
	long		dk_dist;	/* computed "distance" to alt */
			/* dk_dist: used to determine order of alt allocation */
	long		dk_dups;	/* dup count for sector slip */
	unsigned short	dk_flags;	/* flags for various checks  */
};

#define	dk_track	dk_cyl		/* keep #defines close by..   */
#define	dk_offset	dk_sector	/* keep #defines close by..   */

#define	SSLIP	0x01	/* this bit is set if the bad spot was handled with */
			/* sector slip on the disk before this invocation of */
			/* dinit */

#define	VID	dkblk0.dk_vid
#define	IOS	dkblk0.dk_ios

extern char	*bootloader;		/* bootloader file name 	*/
extern char	*badname;		/* bad spots file name 		*/
extern char	*type;			/* disk type name 		*/
extern char	*rdev;			/* raw device name (slice 7)	*/
extern char	bdev[];			/* block device name (slice 7)	*/
extern char	formatter[];		/* disk formatter name 		*/
extern char	rdprog[];		/* manufacturer defect list program */
extern char	*mfrfile; 		/* name of file to create	*/
extern char	motname[];		/* 'MOTOROLA' or 'EXORMACS' 	*/
extern char	volname[];		/* volume name			*/
extern int	newformat;		/* new format request		*/
extern int	bad_cnt;		/* number of bad spots 	*/
extern int	loc;			/*				*/
extern int	lss;			/*				*/
extern int	updown;			/* tells compdk to sort low to high */

extern struct dkbadlst *bad_lst;/*				*/
extern struct dkbadlst *new_spot;/* cyl/head/logical sector, alt for new bad spot*/
extern unsigned *alt_lst;		/* ptr to list of alternates 	*/

extern struct alt_info *alt_ptr, *sv_ptr; /* ptr to alt_info array 	*/
extern int altcnt ;			/* number of alternates  	*/
extern int info_size;			/* size of "info" array  	*/
extern unsigned char	def_vid_strat;	/* Bad spot strategy (from ddefs file) */
extern unsigned short	def_vid_btl;	/* Max. number of bad spots (from ddefs file) */
extern unsigned char	def_ios_hds;	/* # of heads from ddefs file */
extern unsigned short	def_ios_trk;	/* # of cylinders from ddefs file */
extern int alts_used;			/* number of alternate track used */
extern char **trksave;			/* list of saved sectors for recovery */

/*
 *	option flags
 */

extern char	doformat;		/* format the disk */
extern char	makefs;		/* do NOT build file systems and slice table  */
extern char	iopt;		/* do not ask for gap size, inodes for mkfs   */
extern char	eopt;		/* flag for "-e" option  */
extern char	dopt;		/* flag for "-d" option  */
extern char	nopt;		/* flag for "-n" option  */
extern char	ropt;		/* flag for "-r" option  */
extern char	aopt;		/* flag for "-a" option  */
extern char	xopt;		/* flag for "-x" option  */
extern char	topt;		/* flag for "-T" option  */
extern char	sopt;		/* flag for "-s" option  */
extern char	mandefs;	/* read manufacturers defect list from drive */

/*
 *	configuration information
 */

extern int fs0_size;		/* #blocks in slice 0 for 'mkfs(1M)' */
extern int fs1_size;		/* #blocks in slice 1 for 'mkfs(1M)' */
extern int fs1_slice;		/* slice to use for filesystem 1 (/usr) */
extern int sw_size;		/* #blocks in swap area */
extern int sw_slice;		/* slice to use for swap area */
extern int maxbad;		/* maximum number of bad tracks per disk */
extern struct badent *bttable;		/* bad track table */
extern char hasboot; 		/* copy in the bootloader */
extern int reserved;		/* number of reserved blocks at end */
extern struct slice *slctab;	/* slice table pointer */

/*
 *	diagnostic test area support
 */

extern char hasdiag; 		/* create diagnostic tracks */

extern struct diagent *diagdir;	/* diagnostic test area directory */

extern int diaglo; 	/* the starting track of the first diagnostic area */
extern int diaghi; 	/* the starting track of the second diagnostic area */

/*
 *	volume id and config blocks
 */

extern struct dkblk0 dkblk0;
extern unsigned char diagpat[];	/* diagnostic pattern for vid */
extern int diagpatsize;
