/*		@(#)mkslice.c	35.1			*/	
/*
 *	Copyright (c) 1987 Motorola Inc.  
 *		All Rights Reserved.
 */
#include "dinit.h"

/*
 *	mkslice: allocate and initialize slice table
 */

mkslice()
{	register nblocks;
	register slcbcnt;		/* slice table block count */
	register ssize = sw_size;	/* swap size */

	/* if no slice table, just return */

	if ( IOS.ios_slccnt == 0 )
		return;

	/*
	 *	compute slice table block count
	 */

	slcbcnt = ((sizeof(struct slice)*IOS.ios_slccnt)+BSIZE-1)/BSIZE;

	/*
	 *	allocate slice table
	 */

	if(( slctab = (struct slice *)malloc(slcbcnt*BSIZE)) == NULL ) {
		fprintf( stderr, "dinit: out of space\n" );
		exit(ENOSPC);
	}

	/*
	 *	zero slice table
	 */

	memset( slctab, 0, slcbcnt*BSIZE );

	/*
	 *	compute blocks on disk
	 */

	nblocks =(IOS.ios_spt*IOS.ios_trk*IOS.ios_hds)/(BSIZE/IOS.ios_psm);

	/*
	 *	setup last slice and reinitialize "slctab"
	 */

	slctab[IOS.ios_slccnt-1].nblocks = nblocks;	/* set up last slice */

	/*
	 *	set up slice zero
	 */

	slctab[0].blkoff = VID.vid_rfs*(1024/BSIZE);	/* rfs in K-blocks */
	slctab[0].nblocks = nblocks-slctab[0].blkoff-reserved;

	/*
	 *	root filesystem size 0 or no -m option?
	 */

	if(( fs0_size == 0 ) || ( makefs == NO ))
		return;		/* no filesystem  -done- */

	/*
	 *	set up unused slice 1
	 */

	slctab[1] = slctab[0];
	slctab[0].nblocks = fs0_size;
	if((slctab[1].nblocks -= fs0_size) == 0 )
		slctab[1].blkoff  = 0;
	else
		slctab[1].blkoff += fs0_size;

	/*
	 *	if no /usr and no swap, return
	 */

	if(( ssize == 0 ) && ( fs1_size == 0 ))
		return;

	/*
	 *	compute slice 0 in with root
	 */

	if( ssize && ( sw_slice == 0 )) {
		if((slctab[1].nblocks -= ssize) == 0 )
			slctab[1].blkoff = 0;
		else
			slctab[1].blkoff += ssize;
		slctab[0].nblocks += ssize;
		if( fs1_size == 0 )
			return;	/* done ... no /usr */
		ssize = 0;	/* no more swap */
	}

	/*
	 *	compute /usr filesystem
	 */

	if( fs1_size && ( fs1_slice == 1 )) {
		if((slctab[2].nblocks = slctab[1].nblocks - fs1_size) == 0 )
			slctab[2].blkoff = 0;
		else
			slctab[2].blkoff = slctab[1].blkoff + fs1_size;
		slctab[1].nblocks = fs1_size;

		if( ssize ) {		/* if swap */
			if( sw_slice == 1 ) {	/* swap shares /usr slice */
				if((slctab[2].nblocks -= ssize) == 0 )
					slctab[2].blkoff = 0;
				else
					slctab[2].blkoff += ssize;
				slctab[1].nblocks += ssize;
			} else {	/* swap is private slice 2 */
				if((slctab[3].nblocks = slctab[2].nblocks - ssize) == 0 )
					slctab[3].blkoff = 0;
				else
					slctab[3].blkoff = slctab[2].blkoff + ssize;
				slctab[2].nblocks = ssize;
			}
		}

	/*
	 *	check for swap in slice 1
	 */

	} else if( ssize && ( sw_slice == 1 )) {
		if((slctab[2].nblocks = slctab[1].nblocks - ssize) == 0 )
			slctab[2].blkoff = 0;
		else
			slctab[2].blkoff = slctab[1].blkoff + ssize;
		slctab[1].nblocks = ssize;

		/*
		 *	check for /usr in slice 2
		 */

		if( fs1_size && ( fs1_slice == 2 )) {
			if((slctab[3].nblocks = slctab[2].nblocks - fs1_size) == 0)
				slctab[3].blkoff = 0;
			else
				slctab[3].blkoff = slctab[2].blkoff + fs1_size;
			slctab[2].nblocks = fs1_size;
		}
	}
}

/*
 *	mkfs - call mkfs(1M) to make new filesystems
 */

mkfs(lrdev, sl, sl_size)
register char *lrdev, sl;
register int sl_size;
{
	register char *c;
	register unsigned short gapdef = 7;
	register unsigned short blockdef = 400;
	char cmd[120];
	char dev[120];
	char inodes[10];
	char gap[10];
	char blocks[10];
	static char *mkfs = "/etc/mkfs";

	strcpy(dev, lrdev);

	/* compute "base name" */

	for( c = dev + (strlen(dev) -1 ); isdigit(*c ) ; *c-- ='\0'  );

	*inodes = NULL;
	*gap = NULL;
	*blocks = NULL;
	if(iopt ) {
		fprintf(stdout,"\nMaking filesystem on slice %s%c\n",dev,sl);
		fprintf(stdout,"\tNumber of inodes? (default %d): ",
			NBINODE*(sl_size/
			((FSBSIZE/BSIZE)*(NBINODE*4))));
		fflush(stdout);
		gets(inodes,10,stdin);
		fprintf(stdout,"\tGap size? (default %d): ",gapdef);
		fflush(stdout);
		gets(gap,10,stdin);
		fprintf(stdout,"\tBlocks/cyl? (default %d): ",blockdef);
		fflush(stdout);
		gets(blocks,10,stdin);
	}
	sprintf(cmd, "%s %s%c %d", mkfs,dev,sl,sl_size);
	if(*inodes) {
		strcat(cmd,":");
		strcat(cmd,inodes);
	}
	if(*gap || *blocks) {	/* gaps and blocks must both be present */
		if(*gap == NULL)
			sprintf(gap,"%d",gapdef);
		strcat(cmd," ");
		strcat(cmd,gap);
		if(*blocks == NULL)
			sprintf(blocks,"%d",blockdef);
		strcat(cmd," ");
		strcat(cmd,blocks);
	}
#ifdef DEBUG
	printf("%s\n",cmd);
#endif
	system(cmd);

}
