#ident "@(#)readhead.c	35.4"
/*
 *	Copyright (c) 1987 Motorola Inc.  
 *		All Rights Reserved.
 */
#include "dinit.h"

/*
 *	readdef: read in disk definition file
 */

readdef(name)
register char *name;
{	register FILE *fp;
	char fname[BUFSIZ];
	register i;
	register char *envp;
	char	*getenv();

	/*
	 *	open ddefs file
	 */

	sprintf( fname, "%s/%s", DSKDEFS, name );

	if(( fp = fopen( fname, "r" )) == NULL ) {
		fprintf( stderr, "dinit: %s not found\n", fname );
		exit(errno);
	}

	/*
	 *	read volume id and config block
	 */

	if(fread(&VID, sizeof(VID), 1, fp) != 1) { /* read vid */
		fprintf( stderr, "dinit: read error on file %s\n",fname);
		exit(EIO);
	}

	/* read cfg */
	if(fread(&IOS, sizeof(IOS), 1, fp) != 1) { /* read cnfg */
		fprintf( stderr, "dinit: read error on file %s\n",fname);
		exit(EIO);
	}

	/*
	 *	complete initialization
	 */

	hasdiag = VID.vid_dta;
	VID.vid_dta = 0;

	maxbad = VID.vid_btl;
	if (VID.vid_strat == DKPERFECT )
		maxbad = 1;	/* maxbad is probably = 0 if dkperfect */
	def_vid_btl = maxbad;	/* save (ddefs) maxbad in case of -n option */
	def_vid_strat = VID.vid_strat;	/* save strategy for sanity checks (-n option) */
	VID.vid_btl = 0;

	def_ios_hds = IOS.ios_hds;	/* save (ddefs) # heads in case of -n option */
	def_ios_trk = IOS.ios_trk;	/* save (ddefs) # tracks in case of -n option */

	memcpy(VID.vid_mac, motname, 8);
	memcpy(VID.vid_dtp, diagpat, diagpatsize);
	memcpy(VID.vid_vd, volname, 20);

	/* fs0_size= #blocks in slice0 for 'mkfs(1M)' */
	fs0_size = VID.vid_rsize;
	VID.vid_cas = 1;
	if( envp = getenv("FS0SIZE")) {
		if(( i = atoi(envp)) >= 0)
			fs0_size = i;
	}

	/* fs1_size= #blocks in /usr filesystem for 'mkfs(1M)' */
	fs1_size =VID.vid_usize;
	VID.vid_vol = 0;

	/* fs1_slice = /usr slice */
#ifdef m88k
	fs1_slice = (int) VID.vid_u_uslice;
	fs1_slice = ((fs1_slice << 16) | VID.vid_l_uslice);
	if(fs1_slice > 2)
	{
#endif
#ifdef m68k
	if(( fs1_slice = VID.vid_uslice ) > 2 ) {
#endif
		fs1_slice = 0;
		fs1_size = 0;
	}
#ifdef m88k
	VID.vid_osa_u = 0;
	VID.vid_osa_l = 0;
#endif
#ifdef m68k
	VID.vid_osa = 0;
#endif
	if( envp = getenv("FS1SIZE")) {
		if(( i = atoi(envp)) >= 0)
			fs1_size = i;
	}

	/* sw_size= # blocks in swap */
	sw_size = VID.vid_ssize;
	VID.vid_oss = 0;

	/* sw_slice = swap slice */
	if(( sw_slice = VID.vid_sslice ) > 2 ) {
		sw_slice = 0;
		sw_size = 0;
	}
	VID.vid_osl = 0;

	/* save reserved area */
#ifdef m88k
	reserved = (int) VID.vid_u_reserved;
	reserved = ((reserved << 16) | VID.vid_l_reserved);
#endif
#ifdef m68k
	reserved = VID.vid_reserved;
#endif
#ifdef m88k
	VID.vid_ose_u = 0;
	VID.vid_ose_l = 0;
#endif
#ifdef m68k
	VID.vid_ose = 0;
#endif

	VID.vid_cal = 1;
	altcnt = IOS.ios_alts;		/* read altcnt */

	/* compute size of alt_info array */

	info_size = (maxbad <= altcnt ? altcnt : maxbad) ;

	if ((sv_ptr = (struct alt_info *) malloc( info_size*sizeof(struct alt_info))) == NULL) {
			fprintf( stderr, "dinit: out of space\n" );
			exit(ENOSPC);
	}

	alt_ptr = sv_ptr;	/* save a copy of the alt_info pointer */

	/* clear the alt_info array */

	memset( alt_ptr, 0, info_size*(sizeof(struct alt_info)) );

	/* allocate and make local copy of "alt_lst" */

if(( alt_lst = (unsigned *) malloc( altcnt*sizeof(unsigned))) == NULL ) {
	fprintf( stderr, "dinit: out of space\n" );
	exit(ENOSPC);

	} else if( fread(alt_lst,sizeof(unsigned *),altcnt,fp) != altcnt ) {
		fprintf( stderr, "dinit: alternates read error\n" );
		exit(EIO);
	}

	/* put alternates into the "alt_info" array */
	/* no need to sort entries, they are always in ascending order */

	for(i=0, alt_ptr=sv_ptr; i < altcnt; )
		alt_ptr++->dk_avalt = alt_lst[i++];

	/* read disk formatting program */

	fgets( formatter, BUFSIZ, fp );
	formatter[strlen(formatter) - 1] = '\0';
	if (formatter[0] == '\0') {
		newformat++;
	}

	/* skip over comment string */

	fgets( rdprog, BUFSIZ, fp );

	/* read manufacturer defect list program */

	fgets( rdprog, BUFSIZ, fp );
	rdprog[strlen(rdprog) - 1] = '\0';
}

/*
 *	readhead: read in the old volume ID & config block
 *		then modify (per options specified)
 *		1) bad track/spot table
 *		2) "MOTOROLA" / "EXORMACS" string
 *		3) vol id string
 *		4) add bootloader and update "bootloader fields"
 */

readhead(dev)
char	*dev;
{
	register fid;
	register newspot;

	/*
	 *	open and read volumeid and config block
	 */

	if(( fid = open( dev, O_RDWR )) == -1 ) {
		fprintf( stderr, "dinit: %s not found\n", dev );
		exit(errno);
	}

	/*
	 *	read volume id and config block
	 */


	if(read( fid, &dkblk0, sizeof(struct dkblk0)) == -1) {
		fprintf(stderr, "dinit: cannot read volumeid or config block \n");
		exit(EINVAL);
	}

	if(nopt || ropt) { /* add new bad spot? --OR-- print bad list from disk */
		readlst(fid);	/* read existing bad track/spot table */
		if(nopt) {	/* assign alternate for new bad spot and stop */
			newspot = readnew();	/* get new bad spot from tty */
			if(sopt == NO && (newspot != -1))/* try to save data */
				savedata(newspot);
			if (( VID.vid_strat == (DKSOFTWARE & 0xff))
				|| (VID.vid_strat == DKHARDWARE )) 
				newspot = assignnew(); 	/* assign new alt */
			if ( VID.vid_strat == DKHARDSPOT ) /* MVME321 or MVME360 */
				assignnew(); 	/* restore to bad spot */
			/* NOTE: there is no call to assignnew() for DKPERFECT drives */
			/* don't attempt to assign an alternate for "DKPERFECT" drives */
			fillbads();
			/* don't mess with bootloader or diagnostic directory */
			hasboot = hasdiag = NO;
			initdiag();	/* dummy call to set up pointers */
			wrtcyl0(dev, rdev);	/* update disk info */
			if(sopt == NO && (newspot != -1))/* try to restore data */
				restdata(newspot);
			exit(0);
		} else {	/* print bad spot list and stop */
			printlist();
			exit(0);
		}
	}
	else {		/* nope, modify stuff as required */
		if(eopt)	/* if "-e" then replace MOTOROLA string with */
			memcpy(VID.vid_mac, "EXORMACS", 8);
		
		if(dopt)  /* if "-d" then replace volume name string with */
			memcpy(VID.vid_vd, volname, 20);
	
		if(hasboot) {	/* if "-b" then add bootloader and update vid */
			copyboot(fid);		/* copy boot loader */
		}

		/* if changes, update vid/cnfg blocks */
		if (eopt || dopt || hasboot) {

			if (lseek(fid, 0L, 0) == -1) {/*seek to head 0, cyl 0 */
				fprintf(stderr, "dinit:  seek error (vol id blk) \n");
				exit(1);
			}

			if((write(fid, &dkblk0, sizeof(struct dkblk0)) == -1)) {
				fprintf(stderr,
				"dinit: cannot write vol id or config block\n");
				exit(EINVAL);
			}
		}
	}
	close(fid);
}
