

/*****************************************************************************

	Rev = 1.1		Date of 'get'   = 09:32:08 1/6/89
				Date of 'delta' = 09:30:53 1/6/89

	File = m250copy.c

*****************************************************************************/



/**********************************************************************
 *								      *
 *		Copyright (c) Ferranti plc 1988			      *
 *								      *
 *	This is unpublished proprietory source code of Ferranti plc.  *
 *		The copyright of which vests in Ferranti plc.	      *
 **********************************************************************/


/**********************************************************************
 *								      *
 *	This UNIX command performs backups of disk partitions         *
 *	to streamer tape for the FVME250.			      *
 *								      *
 *	Backups can be performed across several tapes.		      *
 *								      *
 **********************************************************************/


#include <sys/fvme250/f250cmds.h>
#include <sys/fvme250/f250buf.h>
#include <fcntl.h>
#include "tapeheader.h"


#define TDC3620400	90000	/* No. of sectors per 450' tape */
#define TDC3620600	120000	/* No. of sectors per 600' tape */
#define VIPER   	300000	/* No. of sectors per 150Mb tape */

struct tape_header h;
struct tape_header *header = &h;	/* Tape header */
	
char *copy_rel_info = "COPY/RESTORE PROGRAM Ver 1.0 ";


write_header(td, tape_head)
	int td;
	struct tape_header *tape_head;
{
	int id;

	/* Write header to streamer tape */
	printf("Header Info: \n");
	printf("Software Release   : %s\n",tape_head->sof_rel);
	printf("Dump Info          : %s\n", tape_head->dump_info);
	printf("Tape number        : %ld\n", tape_head->tape_no);
	printf("Start address      : %ld\n", tape_head->disk_addr);
	printf("no. sectors to copy: %ld\n", tape_head->no_sectors);

	/* Write header on tape */
	id = ioctl(td, WRITEH_CMD, tape_head);
	if ( id == -1) {
		perror("ioctl WRITEH_CMD");
		return(-1);
	}
	return(0);
}
	

rewind_tape(td)
	int td;
{
	int id;

	printf("Rewinding Tape\n");

	/* Rewind tape */
	id = ioctl(td, REWIND_CMD, 0); 

	if (id == -1) {
		perror("ioctl REWIND_CMD");
		return(-1);
	}
	return(0);
}


copy_data(fd, add, cnt, td)
	int fd;		/* file descriptor for disk device */
	long add;	/* Start of disk to copy */
	long cnt;	/* No. of sectors to copy */
	int td;		/* file descriptor for tape device */
{
	int id;
	struct cp_struct info;

	info.fd = td;
	info.offset = add;
	info.nblocks = cnt;

#ifdef DEBUG
	printf("copy_data: fd %d add %ld cnt %ld td %d\n",fd ,add, cnt, td);
#endif
	id = ioctl(fd, COPY_CMD, &info);

	if (id == -1) {
		perror("ioctl COPY_CMD");
		return(-1);
	}
	return(0);
}


wait_chg()
{
	/* Ask the user to change the tape. 
	   The user replies with 'y' or 'Y' return 0, tape changed.
	   The user replies with any other response return -1, 
		abort backup
	*/
	char str[10];
	int cd;

	printf("Please insert a tape cartridge\n");
	printf("Enter 'y' or 'Y' when tape inserted, any other response will abort the backup\t");
	
	scanf("%s",str);
	cd = str[0];
	if (  cd == 'y' || cd == 'Y')
		return(0);
	return(-1);
}	


main(argc, argv)
	int argc;
	char **argv;
{
	int fd,td,id;
	char cd;
	int i;
	char *tape_dev, *disk_dev;
	long copy_sectors;	/* Number of sectors to copy per tape */
	long sectors;		/* No. sectors remaining be backed-up */
	long start;	/* Address first sector be backed-up on tape */
	long 	tape_length;	/* Length of tape in feet */
	long sectors_per_tape;	/* Length of tape in sectors */
	long sector_address = 0; /* Address of first sector to be backed-up */
		
	int tape_sequence,	/* No. of current tape in sequence */
	       no_tapes;	/* No. of tapes needed to complete back-up */
	int no_tapes_backup;		/* No. of tapes needed for back-up */
	char *malloc();
	char str[5];


	/* only super-user is allowed access to this utility */
	if ( geteuid() != 0 ) {
		printf("Super-user only\n");
		exit(-1);
	}

	/* get name of disk device */
	if (argc > 1)
		disk_dev = argv[1];
	else {
		disk_dev = malloc(30);
		printf("Enter name of disk device...");
		scanf("%s",disk_dev);
		printf("\n");
	}

	/* get name of tape device */
	if (argc > 2)
		tape_dev = argv[2];
	else {
		tape_dev = malloc(30);
		printf("Enter name of tape device...");
		scanf("%s",tape_dev);
		printf("\n");
	}

	if (argc > 3) {
		printf("Usage: m250copy [disk_device] [tape_device]\n");
		exit(-1);
	}

	/* Try and open the disk device */
	if ( (fd = open(disk_dev, O_RDONLY)) == -1) {
		perror("opening disk device");
		exit(-1);
	}

	printf("\nEnter the capacity of tape in the tape cartridge (in MB's)...");
	scanf("%d", &tape_length); 
	printf("\n");
	switch (tape_length) {
		case 45 : 
			sectors_per_tape = TDC3620400;
			break; 
		case 60:
			sectors_per_tape = TDC3620600;
			break; 
		case 150:
			sectors_per_tape = VIPER;
			break; 
		default:
			printf("Unsupported length, defaulting to 45MB tape\n");
			sectors_per_tape = TDC3620400;
	}
	copy_sectors = sectors_per_tape;

	id = ioctl(fd, SIZE_CMD, &sectors);
	if (id == -1) {
		perror("ioctl SIZE_CMD");
		exit(-1);
	}

	printf("\nIs this copy the start of a new dump?\t");
	scanf("%c",&cd);
	while ( cd == '\n')
		scanf("%c", &cd);

	if (  cd == 'N' || cd == 'n') {
		printf("Enter tape sequence number\t");	
		scanf("%d", &tape_sequence);
		tape_sequence--;
		no_tapes_backup = (sectors / sectors_per_tape);
		if ( sectors % sectors_per_tape )
			++no_tapes_backup;
		no_tapes = no_tapes_backup - tape_sequence;
			/* Starting address for curerent tape being used */
		start = (tape_sequence * sectors_per_tape) + sector_address;
		sectors = sectors - (tape_sequence * sectors_per_tape);
	}
	else {
		tape_sequence = 1;
		no_tapes_backup = 
		no_tapes = (sectors / sectors_per_tape) + 1;
		start = sector_address;
	}

	scanf("%c",&cd);

	printf("\nEnter dump release information ( MAX 30 chars)\n");
	i = 0;
	scanf("%c", &cd);
/*	printf("char = %c\n",cd);*/
	while (i < 29 && (cd != '\r') && (cd != '\n')) {
		header->dump_info[i++] = cd;
		scanf("%c", &cd);
		/*printf("char = %c\n",cd);*/
	}
	header->dump_info[i] = 0;
		
	printf("\n");

	strncpy((char *)header->sof_rel, (char *)copy_rel_info, 30);	
	/*blkcopy(copy_rel_info, header->sof_rel, 30);	*/

	printf("This will require %d tape(s)\n",no_tapes);

	while ( no_tapes ) {
		if ( sectors < sectors_per_tape)
			copy_sectors = sectors;
		header->tape_no = tape_sequence;
		header->tape_seq = no_tapes_backup;	
		header->disk_addr = start;
		header->no_sectors = copy_sectors;

		if ( wait_chg () < 0)
			exit(-1);

		/* Try and open the tape device */
		if ( (td = open(tape_dev, O_WRONLY)) == -1) {
			perror("opening tape device");
			exit(-1);
		}

		/* Write header */
		while ( write_header(td, header) < 0){
			rewind_tape (td);
			if ( wait_chg() < 0 )
				return(-1); 
		}

		/* Copy disk to tape */
		while ( copy_data(fd, start, copy_sectors, td) < 0){
			printf("COPY MUST BE DONE AGAIN\n");
			if ( wait_chg () < 0) {
				exit(-1);
			}
			else  {
				rewind_tape(td);
			  	write_header(td, header);
			}
		}

		/* Try and close the tape device */
		if ( (cd = close(td)) == -1) {
			perror("close tape device");
			exit(-1);
		}

		no_tapes--;
		start += copy_sectors;
		sectors -= copy_sectors;
		++tape_sequence;
	}
	
	/* Try and close the disk device */
	if ( (cd = close(fd)) == -1) {
		perror("close disk device");
		exit(-1);
	}

	printf("Finished copy\n");
	return(0);
}

