

/*****************************************************************************

	Rev = 1.1		Date of 'get'   = 09:32:09 1/6/89
				Date of 'delta' = 09:30:55 1/6/89

	File = m250rest.c

*****************************************************************************/



/**********************************************************************
 *								      *
 *		Copyright (c) Ferranti plc 1988			      *
 *								      *
 *	This is unpublished proprietory source code of Ferranti plc.  *
 *		The copyright of which vests in Ferranti plc.	      *
 **********************************************************************/


/**********************************************************************
 *								      *
 *	This UNIX command performs backups of tape to Winchester.     *
 *								      *
 *	Backups can be performed across several tapes.		      *
 *								      *
 **********************************************************************/


#include <sys/fvme250/f250cmds.h>
#include <sys/fvme250/f250buf.h>
#include <fcntl.h>
#include "tapeheader.h"


#define TDC3620400	90000	/* No. of sectors per 450' tape */
#define TDC3620600	120000	/* No. of sectors per 600' tape */
#define VIPER   	300000	/* No. of sectors per 150Mb tape */


struct tape_header h;	
struct tape_header *header = &h;	/* Tape header */
	
char *copy_rel_info = "COPY/RESTORE PROGRAM Ver 1.0 ";


read_header(td, tape_head)
	int td;
	struct tape_header *tape_head;
{
	int id;


	/* Read header from tape */
	id = ioctl(td, READH_CMD, tape_head);
	if ( id == -1) {
		perror("ioctl READH_CMD");
		return(-1);
	}

	/*  print out header information */

	printf("Header Information :\n");
	printf("Software Release   : %s\n",tape_head->sof_rel);
	printf("Dump Info          : %s\n", tape_head->dump_info);
	printf("Tape number        : %ld\n", tape_head->tape_no);
	printf("Start address      : %ld\n", tape_head->disk_addr);
	printf("No. sectors to copy: %ld\n", tape_head->no_sectors);
	return(0);
}
	

rewind_tape(td)
	int td;
{
	int id;

	/* Rewind tape */
	printf("Rewinding Tape\n");
	id = ioctl(td, REWIND_CMD, 0); 

	if (id == -1) {
		perror("ioctl REWIND_CMD");
		return(-1);
	}
	return(0);
}


restore_data(fd, add, cnt, td)
	int fd;		/* file descriptor for disk device */
	long add;	/* Start of disk to copy */
	long cnt;	/* No. of sectors to copy */
	int td;		/* file descriptor for tape device */
{
	int id;
	struct cp_struct info;

	info.fd = td;
	info.offset = add;
	info.nblocks = cnt;

#ifdef DEBUG
	printf("restore_data: fd %d add %ld cnt %ld td %d\n",fd ,add, cnt, td);
#endif
	id = ioctl(fd, RESTORE_CMD, &info);

	if (id == -1) {
		perror("ioctl RESTORE_CMD");
		return(-1);
	}
	return(0);
}


wait_chg()
{
	/* Ask the user to change the tape. 
	   The user replies with 'y' or 'Y' return 0, tape changed.
	   The user replies with any other response return -1, 
		abort backup
	*/
	int cd;
	char str[10];

	printf("Please insert a tape cartridge\n");
	printf("Enter 'y' or 'Y' when tape inserted, any other response will abort the backup\t");
	scanf("%s", str);

	cd = str[0];

	if ( cd == 'y' || cd == 'Y')
		return(0);
	return(-1);
}	


main(argc, argv)
	int argc;
	char **argv;
{
	char cd;
	int fd,td;
	char *tape_dev, *disk_dev;
	long copy_sectors;	/* Number of sectors to copy per tape */
	long start;	/* Address first sector be backed-up on tape */
	char *malloc();

	/* only super-user is allowed access to this utility */
	if ( geteuid() != 0 ) {
		printf("Super-user only\n");
		exit(-1);
	}

	/* get name of disk device */
	if (argc > 1)
		disk_dev = argv[1];
	else {
		disk_dev = malloc(30);
		printf("Enter name of disk device...");
		scanf("%s",disk_dev);
		printf("\n");
	}

	/* get name of tape device */
	if (argc > 2)
		tape_dev = argv[2];
	else {
		tape_dev = malloc(30);
		printf("Enter name of tape device...");
		scanf("%s",tape_dev);
		printf("\n");
	}

	if (argc > 3) {
		printf("Usage: m250restore [disk_device] [tape_device]\n");
		exit(-1);
	}

	/* Try and open the disk device */
	if ( (fd = open(disk_dev, O_WRONLY)) == -1) {
		perror("opening disk device");
		exit(-1);
	}

	do {
		/* Try and open the tape device */
		if ( (td = open(tape_dev, O_RDONLY)) == -1) {
			perror("opening tape device");
			exit(-1);
		}
		if (wait_chg() < 0)
			return(-1);

		while ( read_header(td, header) < 0) {
			rewind_tape(td);
			if (wait_chg() < 0 )
				return(-1);
		}

		if (strcmp (header->sof_rel, copy_rel_info)) {
			printf("No header on tape\n");
			rewind_tape(td);
			return(-1);
		}
		copy_sectors = header->no_sectors;
		start = header->disk_addr;

		printf("\nTHIS WILL OVERWRITE SOME CONTENTS OF THE DISK\n\n");
		printf("start = %ld   count = %ld\n",start, copy_sectors);

		printf("Do you wish to put the data on this tape at the above disk address\n");
		scanf("%c", &cd);
		while ( cd == '\n')
			scanf("%c", &cd);
		
		if ( cd == 'n' || cd == 'N') {
			printf("Enter the new address\t");
			scanf("%d",&start);
		}

		while (restore_data (fd, start, copy_sectors, td) < 0){
			if (wait_chg() < 0)
				return(-1);
			else {
				printf("Unrecoverable error\n");
				exit(-1);
			}
		}

		/* Try and close the tape device */
		if ( (cd = close(td)) == -1) {
			perror("close tape device");
			exit(-1);
		}

		printf("Another tape to restore ?\t");
		scanf("%c", &cd);
		while ( cd == '\n')
			scanf("%c", &cd);
	}
	while ( cd == 'y' || cd == 'Y');
	printf("\n");

	/* Try and close the disk device */
	if ( (cd = close(fd)) == -1) {
		perror("close disk device");
		exit(-1);
	}

	printf("Finished restore\n");
}

