/*
 *                  Copyright (c) 1988 Ferranti Plc.
 *                      All Rights Reserved
 *
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FERRANTI PLC THE COPYRIGHT 
 *                   OF WHICH VESTS IN FERRANTI PLC.
 *
 *   The copyright notice above does not evidence any actual or intended 
 *                   publication of such source code.
 *
 * This source code and the information contained herein is supplied without
 * liability for errors or ommissions and no part may be copied, re-produced, 
 * used or disclosed except as authorised by the Ferranti organisation by 
 *                contract or by other written permission.
 */

/******************************************************************************

	Rev = 5.1		Date of 'get'   = 11:10:57 1/26/89
				Date of 'delta' = 14:58:53 12/8/88

	File = scc_cmd.c

******************************************************************************/

#ident "@(#)scc-cmds:cmds/scc_cmd.c Rev 5.1 14:58:53 12/8/88"

/*
 * Version print string
 */
static char scc_id[]="scc-cmds:cmds/scc_cmd.c Rev 5.1 14:58:53 12/8/88";


#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/termio.h>
#include <local/scc/scclioctl.h>

/*
 * Usage: sccload [-VFflv] [load device] [file name] [card id index]
 *        sccinit [-Vlv] [load device] [card id index]
 *        scc     [-VFflv] [load device] [file name] [card id index]
 *
 * -f 	Takes a filename or "default"
 * -F	Assumes default file name
 * -l 	Specify a load device name (default /dev/scc.ctrl)
 * -v 	Verbose
 * -V	Version message
 *
 * If no card id given then all cards are loaded
 */

/*
 * Global #defines
 */
#define	MAXCARDS	16
#define	TRUE		1
#define	FALSE		0

char *file, *ldfile;
char *cmd;
int fflag, bigfflag, lflag, vflag, loadonly, initonly;
FILE *fp;
int load_fd;

main(argc,argv)
int argc;
char *argv[];
{
	char	*cp;
	FILE 	*fopen();
	int	cards, i;
	int 	scc_addrs[MAXCARDS];
	struct	card_data card_data;
	char **ptr, *strptr;

	/*
	 * Clear all option flags
	 */

	bigfflag= 0; fflag = 0; lflag = 0; vflag = 0;
	file = NULL;
	ldfile = "/dev/scc.ctrl";

	/*
	 * Set the command name into string pointer cmd
	 */

	if (cmd = (char *)strrchr(argv[0],'/'))
		++cmd;
	else
		cmd = argv[0];

	/*
	 * Is it 'init', 'load' or both ?
	 */
	if (strcmp(cmd,"scc_cmd") == 0) {
		fprintf(stderr,"%s: This is the 'generic' utility\n",cmd);
		fprintf(stderr,"%s: Use sccload, sccinit or scc\n",cmd);
		usage(0);
		exit(-1);
	}
	else if (strcmp(cmd,"sccload") == 0)
			loadonly = TRUE;
	else if (strcmp(cmd,"sccinit") == 0)
			initonly = TRUE;
	else if (strcmp(cmd,"scc") != 0) {
		/*
		 * Have a bad command name
		 */
		fprintf(stderr,"%s: Bad command name\n",cmd);
		fprintf(stderr,"%s: Must be sccload, sccinit or scc\n",cmd);
		exit(-1);
	}

	if (argc > 1) {

		/*
		 * We have arguments, but are they options ?
		 */	
		argv++;
		if ((*argv)[0] == '-') {
			/*
			 * Have options, parse them
			 */
			for (cp = *argv++; *cp; cp++) {
			switch (*cp) {
				case 'V': printf("%s: %s\n",
							cmd,scc_id);
					  exit(0);
				case 'F': if (initonly) {
					  	optionerr(cmd,*cp);
						exit(-1);
					  }
						
					  bigfflag++;
					  break;
				case 'h': usage(1);
					  exit(0);
				case 'f': if (initonly) {
					  	optionerr(cmd,*cp);
						exit(-1);
					  }
						
				          file = *argv++;
					  fflag++;
					  if (file == NULL)
						no_file();
						
					  break;
				case 'l': ldfile = *argv++;
					  lflag++;
					  if (ldfile == NULL)
						no_file();
					  break;
				case 'v': vflag++;
					  break;
				case '-': break;
				default : optionerr(cmd,*cp);
					  exit(-1);
			}	/* end of switch */
			}	/* end of for */
		}	/* end of option if statement */

	} 
	else argv++;

	/*
	 * Sort out file name and set file to file name string
	 */

	if ( (fflag) && (bigfflag) ) {
		fprintf(stderr,"%s: Bad option set. \"f\" and \"F\" invalid\n",
					cmd);
		usage(0);
		exit(1);
	}

	if ( (bigfflag) || ((fflag) && (strcmp(file,"default") == 0)) )
		file = "fwscc.dat";

	if (vflag) {
		if (fflag || bigfflag) 
			printf("%s: Load file is %s\n",cmd,file);
		else
			printf("%s: Firmware to be loaded from EPROM\n",cmd);
	}

	/*
	 * Init scc_addrs to -1
	 */
	for (i=0; i< MAXCARDS; i++) scc_addrs[i] = -1;

	/*
	 * If a file load try and open the file
	 */
	if (file != NULL ) {
		if ( (fp = fopen(file, "r")) == NULL ) {
			perror("fopen");
			fprintf(stderr,
				"%s: Failed to open data file %s for reading\n",
					cmd, file);
			exit(1);
		}
	}

	/* 
	 * Open device to get into loader driver 
	 */
	while (TRUE) {
		if ( (load_fd = open(ldfile,O_RDWR)) == -1 ) {
			if (errno == EBUSY) {
				sleep(1);
				continue;
			}
			perror("open");
			fprintf(stderr,
		   "%s: Failed to open load device %s (read/write access)\n",
					cmd, ldfile);
			exit(1);
		}
		else
			break;
	}

	/*
	 * Build card index array table anyway. Cancel any entries that
	 * are not to be loaded later. Thus a card index can be checked to
	 * see if it is valid
	 */
	cards = 0;
	for (i = 0; i< MAXCARDS; i++) {
		card_data.index = i;
		if (ioctl(load_fd, SCCL_ADRS, &card_data) == -1) {
			/*
			 * No more cards !
			 */
			break;
		}
		if (card_data.adrs != (unsigned char *)-1) {
			scc_addrs[i] = 0;
			cards++;
		}
	}

	if (cards == 0) {
		fprintf(stderr,"%s: No SCC cards configured into system\n",
			cmd);
		exit(1);
	}

	if (vflag) {
		if (cards == 1)
			printf("%s: 1 SCC card in system\n", cmd);
		else
			printf("%s: %d SCC cards in system\n", cmd, cards);
	}

	if (*argv == NULL) {
		/*
	 	 * Make all cards present valid to load, e.g. scc_addrs[?] 
		 * set to +1
		 */
		for (i = 0; i < MAXCARDS; i++) {
			if (scc_addrs[i] == 0)
				scc_addrs[i] = 1;
		}
	}
	else {
		/*
		 * Only load cards that are in load list
		 */
		while (*argv != NULL) {
			ptr = &strptr;
			i = strtol(*argv, ptr, 10);

			if (*ptr == *argv++) {
				fprintf(stderr,
				"%s: Card index %s is not a valid integer\n",
					cmd, *ptr);
				exit(1);
			}

			if ( (i < 0) || (i > MAXCARDS) ) {
				fprintf(stderr,
					"%s: Card index %i out of range\n",
								cmd, i);
				exit(1);
			}
			if (scc_addrs[i] == -1) {
				fprintf(stderr,
			"%s: Card index %d not configured in to system\n",
							cmd, i);
				exit(1);
			}
			scc_addrs[i] = 1;
		}
	}	

	if (!initonly) {
		/*
		 * Now the loop for each card in the system
		 */
		for (i = 0; i < MAXCARDS; i++) {
			if (scc_addrs[i] != 1) continue;
	
			if (file == NULL)
				prom_load(i);
			else
				file_load(i);
		}
	}

	if (!loadonly) {
		for (i = 0; i < MAXCARDS; i++) {
			if (scc_addrs[i] != 1) continue;
			scc_init(i);
		}
	}
	
	exit(0);

}

/*
 *
 */
prom_load(index)
int index;
{
	if (vflag) {
		printf("%s: PROM load card %d ...",cmd,index);
		fflush(stdout);
	}

	if (ioctl(load_fd, SCCL_PROMLOAD, &index) == -1 ) {
		fprintf(stderr,"\n");
		perror("ioctl");
		se_error(load_fd,cmd);
		fprintf(stderr,"%s: Card %d failed to complete EPROM load sequence\n",
				cmd, index);
		exit(1);
	}

	if (vflag) printf(" Done\n");

}

/*
 *
 */
file_load(index)
int index;
{			
	char * malloc();
	char bytes_sum[2];
	char data[4096];
	char addr[8];
	int  * bytes;
	int  * checksum;
	int  * address;
	int  * n;
	char ch;
	int    i, c, line_no;
	short  end_record;
	unsigned data_sz, addr_sz;
	struct load_info record;
	int * dummy;
	
	record.index = index;

	if (ioctl(load_fd, SCCL_LDINIT, &record.index) == -1 ) {
		perror("ioctl");
		se_error(load_fd,cmd);
		fprintf(stderr,"%s: Card %d failed to complete SCCL_LDINIT sequence\n",
				cmd, index);
		exit(1);
	}
	

	rewind(fp);

	line_no = 0;
	end_record = FALSE;
	/* get chars while not end of file and 
	 * last record was not an 'end record' */ 

	if (vflag) {
		fprintf(stdout,"%s: Loading card %d .",cmd,index);
		fflush(stdout);
	}
		
	while ( ( c = getc(fp) ) != EOF  && !end_record)
	{
		if ((line_no++ % 100) == 0) {
			if (vflag) {
				fprintf(stdout,".");
				fflush(stdout);
			}
		}

		if ( c != START_CODE )
		{
			fprintf(stderr,
"%s: expect first char on each line of %s to be 'S' and not '%c' (line %d)\n",
cmd,file,c,line_no);
			return(L_ERR);
		}
		switch ( c = getc(fp) )
		{
		case '1':
			end_record = FALSE;
			addr_sz = S9_S1_ADDR_SZ;
			break;
		case '9':
			end_record = TRUE;
			addr_sz = S9_S1_ADDR_SZ;
			break;
		case '2':
			end_record = FALSE;
			addr_sz = S8_S2_ADDR_SZ;
			break;
		case '8':
			end_record = TRUE;
			addr_sz = S8_S2_ADDR_SZ;
			break;
		case '3':
			end_record = FALSE;
			addr_sz = S7_S3_ADDR_SZ;
			break;
		case '7':
			end_record = TRUE;
			addr_sz = S7_S3_ADDR_SZ;
			break;
		case '0':
			end_record = FALSE;
			while ( (c = getc(fp)) != EOR ) ;
			continue;	
		default:
			fprintf(stderr,
"%s: invalid S-record type (%d) at line %d\n",c - '0',cmd,line_no);
			return(L_ERR);
		}
		record.type = c - '0';

		/* 
		 * Get number of bytes in this record 
		 */
		for ( i = 0; i < SUM_SZ; i++)
			bytes_sum[i] = getc(fp);
		sscanf(bytes_sum,"%2x",&bytes);

		/* 
	 	 * Get offset address 
		 */
		for ( i = 0; i < addr_sz; i++ ) 
			addr[i] = getc(fp);
		for (i = addr_sz; i<8; i++)
			addr[i] = 0;

		sscanf(addr, "%x", &address);
		record.offset = (int) address;

		/* get data */
		data_sz = ( (unsigned) bytes * 2 ) - addr_sz - CHKSUM_SZ;
		record.data_sz = (int)data_sz;
		
		if (data_sz > 4096) {
			printf("TOO BIG\n");
			exit(1);
		}

		for ( i = 0; i < data_sz; i++ )
			data[i] = getc(fp);
		record.data = data;

		/*
		 * Skip the checksum
		 */
		getc(fp); getc(fp);

		if ( (c = getc(fp)) != EOR )
		{
			fprintf(stderr,
"%s: expected end_of_record after checksum and not %c (line %d)\n",cmd,c,line_no);
			return(L_ERR);
		}

		
		if ( ioctl(load_fd, SCCL_LOADFW, &record) == -1 ) {
			perror("ioctl");
			se_error(load_fd,cmd);
		fprintf(stderr,"%s: Card %d failed to complete SCCL_LOADFW sequence\n",
				cmd, index);
			return(L_ERR);
		}

		if ((end_record) && (vflag)) {
			printf(". Complete\n");
			printf("%s: Code start address for card %d is 0x%x\n",
						cmd, index, record.offset);
		}

		bytes = address = checksum = n = 0;
	}

	if ( c == EOF && !end_record )
	{
		fprintf(stderr,"%s: last S-record was not an end record!\n",
					cmd);
		return(L_ERR);
	}

}

optionerr(cmd,ch)
char *cmd;
char ch;
{
	fprintf(stderr,"%s: %c - Unknown option\n",cmd,ch);
	usage(0);
}

usage(all)
int all;
{
	printf("%s: Command usage\n",cmd);
printf("usage: sccload [-VFflv] [load device] [file name] [card id index list]\n");
printf("usage: sccinit [-Vlv]   [load device]             [card id index list]\n");
printf("usage: scc     [-VFflv] [load device] [file name] [card id index list]\n");
	if (all) {
		printf("\n");
		printf("options: -V  Print utility version string\n");
		printf("         -F  Assume default S-record file name (fwscc.dat)\n");
		printf("         -f  Specify S-record load file\n");
		printf("         -h  This help message\n");
		printf("         -l  Specify new load special device (default /dev/scc.ctrl)\n");
		printf("         -v  Enable verbose mode\n");
	}
}

no_file()
{
	if (fflag)
		printf("%s: No file name specified for f option\n",cmd);
	else if (lflag)
		printf("%s: No device name specified for l option\n",cmd);
	else {
		fprintf(stderr,"%s: Bad call to no_file()\n",cmd);
	}
	exit(1);
}


scc_init(i)
int i;
{
	struct load_info record;

	if (vflag) {
		printf("%s: Initialise card %d ...",cmd,i);
		fflush(stdout);
	}
	
	record.index = i;

	if ( ioctl(load_fd, SCCL_FWINIT, &record.index) == -1 ) {
		perror("ioctl");
		se_error(load_fd,cmd);
		fprintf(stderr,"%s: Card %d failed to complete SCCL_FWINIT sequence\n",
				cmd, i);
		return(0);
	}

	if (vflag) 
		printf(" Done\n");

}

