/*
 *                  Copyright (c) 1988 Ferranti Plc.
 *                      All Rights Reserved
 *
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FERRANTI PLC THE COPYRIGHT 
 *                   OF WHICH VESTS IN FERRANTI PLC.
 *
 *   The copyright notice above does not evidence any actual or intended 
 *                   publication of such source code.
 *
 * This source code and the information contained herein is supplied without
 * liability for errors or ommissions and no part may be copied, re-produced, 
 * used or disclosed except as authorised by the Ferranti organisation by 
 *                contract or by other written permission.
 */

/******************************************************************************

	Rev = 5.1		Date of 'get'   = 11:10:58 1/26/89
				Date of 'delta' = 14:58:56 12/8/88

	File = sccetab.c

*******************************************************************************

The purpose of this program is to allow :-

1) The SCC input echo tables to be read and output in a text format to 
   the stdout 

2) A text format configuartion file to be used to load the SCC input echo 
   tables

******************************************************************************/

#ident "@(#)scc-cmds:cmds/sccetab.c Rev 5.1 14:58:56 12/8/88"

#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <local/scc/sccioctl.h>

/*
 * Version output string. Not output by what(1)
 */
static char sccetab_id[] = "scc-cmds:cmds/sccetab.c Rev 5.1 14:58:56 12/8/88";

/*
 * #defines
 */

#define	TRUE	1
#define	FALSE	0

#define	IBUFSIZE	80
#define	EOL		'\n'

#define	COMMENT		'*'

#define WHITE_ONLY	1
#define	ANY_CHARS	0

/*
 * Gloabl data elements
 */

char *cmd, *data_file;
char ibuf[IBUFSIZE];
int line;
	 
char head_msg[]   = "SCC-INPUT-ECHO-TABLE-CONFIGUARATION-FILE";
char iblock_msg[] = "SCC-INPUT-ECHO-TABLE-INDEX-BLOCK";
char dblock_msg[] = "SCC-INPUT-ECHO-TABLE-DATA-BLOCK";
char msg_msg[]    = "msg:";

/*
 * externs
 */
char *ctime();

/*
 *
 */
main(argc, argv)
int argc;
char **argv;
{
	FILE *fp;
	long t;
	int i, read_tab, write_tab;
	int index, index1, index2, index3, index4;
	int data1, data2, data3, data4;
	int ret, c, vflag, wcnt, wflag, sflag, mflag, dflag;
	char *cp;
	unsigned char databuf[512];

	if (cmd = (char *)strrchr(argv[0],'/'))
		++cmd;
	else
		cmd = argv[0];

	wcnt = 0;
	dflag = 0;
	sflag = 0;
	mflag = 0;
	vflag = 0;
	wflag = 1;
	write_tab = FALSE;
	read_tab = FALSE;
	
	if (argc == 1) 
		read_tab = TRUE;
	else {

		data_file = argv[1];
		write_tab++;

		argv++;
		if ((*argv)[0] == '-') {

			for (cp = *argv++; *cp; cp++) {
				switch (*cp) {
					case '-' : break;
					case 'h' : usage(1);
						   exit(0);
					case 'd' : dflag++;
						   break;
					case 'w' : wflag = 0;
						   break;
					case 's' : sflag++;
						   break;
					case 'm' : mflag++;
						   break;
					case 'v' : vflag++;
						   break;
					case 'V' : prt_msg();
						   exit(0);
					default  : fprintf(stderr,
				           "%s: %c - Unknown option\n",cmd,*cp);
						   usage(0);
						   exit(1);
				}
			}

			/*
			 * Update data_file as have oprion.
			 */
			data_file = *argv;
		}
	}

	if ((write_tab) && (data_file == NULL)) {
		fprintf(stderr,"%s: No configuration file name specified\n",
						cmd);
		usage(0);
		exit(1);
	}
	
	if (read_tab) {
		/*
		 * Want to read from SCC memory to file.
		 * Load databuf via ioctl call SCCIETR
		 */
		if(ioctl(0, SCCIETR, databuf) == -1) {
			perror("ioctl");
			fprintf(stderr,
		   	"%s: Failed to read input echo table (SCCIETR)\n",
								cmd);
			exit(2);
		}

		printf("%s\n",head_msg);

		t = time((long *)0);	
		printf("%sConfiguration file generate by user id %d, %s",
				msg_msg, getuid(), ctime(&t));
		printf("%s\n",iblock_msg);
		for (i=0; i<256; i++) {
			printf("0x%02x:i=0x%02x", i, databuf[i]);
			if ((i != 0) && (((i+1) % 4) == 0)) 
				printf("\n");
			else 
				printf("\t");
		}
		printf("%s\n",dblock_msg);
		for (i=0; i<256; i++) {
			printf("0x%02x:d=0x%02x", i, databuf[256+i]);
			if ((i != 0) && (((i+1) % 4) == 0)) 
				printf("\n");
			else 
				printf("\t");
		}
		exit(0);
	}
	else if (write_tab) {
		/*
		 * Want to write to SCC memory from file. Open file for reading
		 */
		if ( (fp = fopen(data_file, "r")) == NULL) {
			perror("fopen");
			fprintf(stderr,
		"%s: Failed to open echo table configuration file %s for reading\n",
				cmd, data_file);
			exit(1);
		}
	}
	else {
		fprintf(stderr,"%s: FATAL ERROR #1\n");
		exit(-1);
	}

	/*
	 * Read first line of file
	 */
	clear_ibuf();
	for (index = 0; index < strlen(head_msg); index++) {
		ibuf[index] = getc(fp);

		if (ibuf[index] == EOF) {
			fprintf(stderr,
			"%s: Unexpected EOF in header record, line 1\n",cmd);
			exit(1);
		}

		if (ibuf[index] == EOL) {
			fprintf(stderr,
			"%s: Unexpected EOL in header record, line 1\n",cmd);
			exit(1);
		}

		if (ibuf[index] != head_msg[index]) {
			fprintf(stderr,"%s: Bad character in header record\n",
					cmd);
			fprintf(stderr,"%s: Line 1, column %d, char %c\n",
					cmd, index+1, ibuf[index]);
			exit(1);
		}
	}
	
	/*
	 * Get to end of line. Ignore white space and tabs. Other chars
	 * will cause an error
	 */
	while ( (c = getc(fp)) != EOL) {
		
		if (c == EOF) {
			fprintf(stderr,
			"%s: Unexpected EOF after header record on line 1\n",
									cmd);
			exit(1);
		}
		index++;
		if ((c == ' ') || (c == '\t') )
			continue;
		
		fprintf(stderr,"%s: Bad character at end of header record\n",
					cmd);
		fprintf(stderr,"%s: Line 1, column %d, char %c\n",
					cmd, index, c);
		exit(1);
	}
	
	/*
	 * Have head_msg, and therefore have an echo table config file
	 * Get the "msg:"
	 */
	for (index = 0; index < strlen(msg_msg); index++) {
		c = getc(fp);

		if (c == EOF) {
			fprintf(stderr,
				"%s: Unexpected EOF in MSG string, line 2\n",
						cmd);
			exit(1);
		}

		if (c == EOL) {
			fprintf(stderr,
			"%s: Unexpected EOL in MSG string, line 2\n",
				cmd);
			exit(1);
		}

		if (c != msg_msg[index]) {
			fprintf(stderr,"%s: Bad character in MSG string\n",
					cmd);
			fprintf(stderr,"%s: Line 2, column %d, char %c\n",
					cmd, index+1, c);
			exit(1);
		}
	}

	/*
	 * Print out rest of line if vflag, else skip it
	 */

	if (mflag) vflag++;

	if (vflag)
		printf("%s: ",cmd);
	skip_line(fp,vflag,ANY_CHARS);
	if (vflag)
		printf("\n");

	if (mflag) exit(0);
		

	if ((vflag) && (sflag != 0))
		printf("%s: Syntax check only\n",cmd);

	line = 3;		
	/*
	 * At start of a line, skip any comments
	 */
	if (comment_skip(fp)) {
		fprintf(stderr,
			"%s: Unexpected EOF at line %d\n",cmd,line);
		exit(1);
	}
	
	/*
	 * Check for iblock_msg
	 */
	clear_ibuf();
	for (index = 0; index < strlen(iblock_msg); index++) {
		ibuf[index] = getc(fp);

		if (ibuf[index] == EOF) {
			fprintf(stderr,
			"%s: Unexpected EOF in index block start record, line %d\n",cmd,line);
			exit(1);
		}

		if (ibuf[index] == EOL) {
			fprintf(stderr,
			"%s: Unexpected EOL in index block start record, line %d\n",cmd,line);
			exit(1);
		}

		if (ibuf[index] != iblock_msg[index]) {
			fprintf(stderr,
				"%s: Bad character in index block start record\n",
					cmd);
			fprintf(stderr,"%s: Line %d, column %d, char %c\n",
					cmd, line, index+1, ibuf[index]);
			exit(1);
		}
	}
	
	/*
	 * Get to end of line. Ignore white space and tabs. Other chars
	 * will cause an error
	 */
	while ( (c = getc(fp)) != EOL) {
		
		if (c == EOF) {
			fprintf(stderr,
			"%s: Unexpected EOF after index block start record on line %d\n",
								cmd,line);
			exit(1);
		}
		index++;
		if ((c == ' ') || (c == '\t') )
			continue;
		
		fprintf(stderr,
			"%s: Bad character at end of index block record\n",
					cmd);
		fprintf(stderr,"%s: Line %d, column %d, char %c\n",
					cmd, line, index, c);
		exit(1);
	}

	line++;
	index = 0;

	while (index < 256) {
		/*
		 * At start of a line, skip any comments
		 */
		if (comment_skip(fp)) {
			fprintf(stderr,
				"%s: Unexpected EOF at line %d\n",cmd,line);
			exit(1);
		}

		/*
		 * Get a token line "0x??:i=0x??"
		 */
		ret = fscanf(fp,"0x%x:i=0x%x 0x%x:i=0x%x 0x%x:i=0x%x 0x%x:i=0x%x",
					&index1,&data1,
					&index2,&data2,
					&index3,&data3,
					&index4,&data4);

		if (ret == -1) {
			fprintf(stderr,
			"%s: Undexpected EOF read in index block at line %d\n",
					cmd,line);
			exit(1);
		}
		else if (ret != 8) {
			syntax();
			fprintf(stderr,
			"%s: fscanf of index block failed, check line format\n",
					cmd);
			exit(1);
		}

		if (index != index1) {
			syntax();
			fprintf(stderr,"%s: Expected index element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index] = data1;
		index++;

		if (index != index2) {
			syntax();
			fprintf(stderr,"%s: Expected index element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index] = data2;
		index++;

		if (index != index3) {
			syntax();
			fprintf(stderr,"%s: Expected index element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index] = data3;
		index++;

		if (index != index4) {
			syntax();
			fprintf(stderr,"%s: Expected index element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index] = data4;
		index++;

		/*
		 * Get to start of next line
		 */
		skip_line(fp,0,WHITE_ONLY);

	}	
	
	/*
	 * At start of a line, skip any comments
	 */
	if (comment_skip(fp)) {
		fprintf(stderr,
			"%s: Unexpected EOF at line %d\n",cmd,line);
		exit(1);
	}
	
	/* 
	 * Now at start of data block, check for dblock_msg
	 */
	clear_ibuf();
	for (index = 0; index < strlen(dblock_msg); index++) {
		ibuf[index] = getc(fp);

		if (ibuf[index] == EOF) {
			fprintf(stderr,
			"%s: Unexpected EOF in data block start record, line %d\n",cmd,line);
			exit(1);
		}

		if (ibuf[index] == EOL) {
			fprintf(stderr,
			"%s: Unexpected EOL in data block start record, line %d\n",cmd,line);
			exit(1);
		}

		if (ibuf[index] != dblock_msg[index]) {
			fprintf(stderr,
				"%s: Bad character in data block start record\n",
					cmd);
			fprintf(stderr,"%s: Line %d, column %d, char %c\n",
					cmd, line, index+1, ibuf[index]);
			exit(1);
		}
	}
	
	/*
	 * Get to end of line. Ignore white space and tabs. Other chars
	 * will cause an error
	 */
	while ( (c = getc(fp)) != EOL) {
		
		if (c == EOF) {
			fprintf(stderr,
			"%s: Unexpected EOF after data block start record\n",
									cmd);
			exit(1);
		}
		index++;
		if ((c == ' ') || (c == '\t') )
			continue;
		
		fprintf(stderr,
			"%s: Bad character at end of data block record\n",
					cmd);
		fprintf(stderr,"%s: Line %d, column %d, char %c\n",
					cmd, line, index, c);
		exit(1);
	}
	line++;
	index = 0;

	while (index < 256) {
		/*
		 * At start of a line, skip any comments
		 */
		if (comment_skip(fp)) {
			fprintf(stderr,
				"%s: Unexpected EOF at line %d\n",cmd,line);
			exit(1);
		}

		/*
		 * Get a token line "0x??:d=0x??"
		 */
		ret = fscanf(fp,"0x%x:d=0x%x 0x%x:d=0x%x 0x%x:d=0x%x 0x%x:d=0x%x",
					&index1,&data1,
					&index2,&data2,
					&index3,&data3,
					&index4,&data4);

		if (ret == -1) {
			fprintf(stderr,
			"%s: Undexpected EOF read in data block at line %d\n",
					cmd,line);
			exit(1);
		}
		else if (ret != 8) {
			syntax();
			fprintf(stderr,
			"%s: fscanf of data block failed, check line format\n",
					cmd);
			exit(1);
		}

		if (index != index1) {
			syntax();
			fprintf(stderr,"%s: Expected data element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index+256] = data1;
		index++;

		if (index != index2) {
			syntax();
			fprintf(stderr,"%s: Expected data element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index+256] = data2;
		index++;

		if (index != index3) {
			syntax();
			fprintf(stderr,"%s: Expected data element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index+256] = data3;
		index++;

		if (index != index4) {
			syntax();
			fprintf(stderr,"%s: Expected data element 0x%02x\n",
				cmd,index);
			exit(1);
		}
		databuf[index+256] = data4;

		if (index == 255) {
			if (data4) {
				if (wflag)
				fprintf(stderr,
"%s: WARNING: data element 0xff is none zero. Table overrun possible\n",cmd);
				wcnt++;
			}
		}
		index++;

		/*
		 * Get to start of next line
		 */
		skip_line(fp,0,WHITE_ONLY);

	}	

	if (dflag) {
		printf("%s: Dump of databuf array :-\n",cmd);

		for (i=0; i<256; i++) {
			printf("0x%02x i=0x%02x", i, databuf[i]);
			if ((i != 0) && (((i+1) % 4) == 0)) 
				printf("\n");
			else 
				printf(" | ");
		}
		for (i=0; i<256; i++) {
			printf("0x%02x d=0x%02x", i, databuf[256+i]);
			if ((i != 0) && (((i+1) % 4) == 0)) 
				printf("\n");
			else 
				printf(" | ");
		}
		printf("%s: Array dump complete\n",cmd);
	}

	if (sflag == 0) {
		if(ioctl(0, SCCIETW, databuf) == -1) {
			perror("ioctl");
			fprintf(stderr,
		   	"%s: Failed to write input echo table (SCCIETW)\n",
								cmd);
			exit(2);
		}
	}

	if ((vflag) && (sflag != 0)) {
		printf("%s: Syntax complete, no format errors detected",cmd);
		if (wcnt) 
			printf(", %d warning(s)",wcnt);
		printf("\n");
	}
	exit(0);

}

/*
 * 
 * Support funtions
 */

clear_ibuf()
{
	register i;

	for (i = 0; i < IBUFSIZE; i++) ibuf[i] = (char)0;
}

usage(all)
int all;
{
	printf("usage: %s [-Vhmsvw] [configuration file] [ < /dev/tty-device]\n",
			cmd);
	if (all) {
		printf("\n");
		printf("options: -V  Print utility version string\n");
		printf("         -h  This help message\n");
		printf("         -m  Print message embedded in configuration file\n");
		printf("         -s  Syntax check configuration file only\n");
		printf("         -v  Enable verbose mode\n");
		printf("         -w  Supress warning messages\n");
	}
}

skip_line(fp,prt,white_only)
FILE *fp;
int prt;
{
	register c;

	while ( (c = getc(fp)) != EOF) {
		if (c == EOL) {
			line++;
			return(0);
		}

		if (white_only) {
			if ((c != ' ') && (c != '\t') ) {
				syntax();
				fprintf(stderr,"%s: Expected EOL\n",
					cmd);
				exit(1);
			}
		}

		if (prt) printf("%c",c);
	}

	/*
	 * Have end of file
	 */
	return(1);	
}


comment_skip(fp)
FILE *fp;
{
	register c;

	while ((c = getc(fp)) != EOF) {
		if (c != COMMENT) {
			ungetc(c,fp);
			return(0);
		}
		if (skip_line(fp,0,ANY_CHARS)) break;
	}
	
	/*
	 * Have EOF
	 */
	return(1);
}

syntax()
{
	fprintf(stderr,"%s: Syntax error in file %s at line %d\n",
		cmd,data_file,line);
}

prt_msg()
{
	printf("%s: %s\n",cmd,sccetab_id);
}

