/*
 *                  Copyright (c) 1988 Ferranti Plc.
 *                      All Rights Reserved
 *
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FERRANTI PLC THE COPYRIGHT 
 *                   OF WHICH VESTS IN FERRANTI PLC.
 *
 *   The copyright notice above does not evidence any actual or intended 
 *                   publication of such source code.
 *
 * This source code and the information contained herein is supplied without
 * liability for errors or ommissions and no part may be copied, re-produced, 
 * used or disclosed except as authorised by the Ferranti organisation by 
 *                contract or by other written permission.
 */

/******************************************************************************

	Rev = 5.1		Date of 'get'   = 11:10:59 1/26/89
				Date of 'delta' = 17:04:56 1/20/89

	File = sccfifo.c

******************************************************************************/

#ident "@(#)scc-cmds:cmds/scctty/sccfifo.c Rev 5.1 17:04:56 1/20/89"

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <local/scc/status.h>
#include <local/scc/fifo.h>
#include <local/scc/scclioctl.h>

static char ffinit_id[] = "scc-cmds:cmds/scctty/sccfifo.c Rev 5.1 17:04:56 1/20/89";

#define	TRUE	1
#define	FALSE	0

#define	IBUFSIZE	80
#define	EOL		'\n'

#define	MAXIFCS		32
#define	MAXPORTS	16

#define	COMMENT		'*'
#define	RAW		'r'
#define	HOST		'h'
#define	IFC		'i'

#define WHITE_ONLY	1
#define	ANY_CHARS	0

#define	FFHEAD		28
/* #define	DBUG */

char ibuf[IBUFSIZE];
int line;
char *config_file, *ld_file, *cmd;
	 
char head_msg[] = "SCC-FIFO-CONFIGURATION-FILE";
char msg_msg[]  = "msg:";
char ld_def[]   = "/dev/scc.ctrl";
struct fifo_sizes fifo_defs[MAXPORTS];

int dflag;

#define	MAXCARDS 	16

main(argc, argv)
int argc;
char **argv;
{

	FILE *ld_fp, *fp, *fopen();
	int i, kb_mem, mem, ret, index, c, cnt, p;
	int lflag, kflag, vflag, wflag, sflag, mflag;
	int card_index, port;
	int loadfp;
	char **ptr, *strptr;
	struct card_data card_data;

	int in[MAXPORTS], out[MAXPORTS], pout[MAXPORTS];
	int hout[MAXPORTS], non_canon_hin[MAXPORTS];
	int no_ifc_fifos[MAXPORTS];
	int ifc_size[MAXPORTS*MAXIFCS];

	char *cp;
	struct stat stat_buf;

	/*
	 * A quick assertion test
	 */
	if (sizeof(struct fifo_sizes) != 256) {
		fprintf(stderr,"ASSERTION FAILURE - bad fifo_size size\n");
		exit(1);
	}

	sflag = 0;
	wflag = 1;
	mflag = 0;
	lflag = 0;
	kflag = 0;
	vflag = 0;
	dflag = 0;

	ld_file = ld_def;

	if (cmd = (char *)strrchr(argv[0],'/'))
		++cmd;
	else
		cmd = argv[0];

	if (argc <= 1) {
		fprintf(stderr,
			"%s: No fifo configuration file name specified\n",cmd);
		usage(0);
		exit(1);
	}

	if (argc > 1) {
		argv++;
		if ((*argv)[0] == '-') {
			for (cp = *argv++; *cp; cp++) {
				switch (*cp) {
					case '-' : break;
					case 'l' : lflag++;
						   ld_file = *argv++;
						   break;
					case 'h' : usage(1);
						   exit(0);
					case 'd' : dflag++;
						   break;
					case 'w' : wflag = 0;
						   break;
					case 's' : sflag++;
						   break;
					case 'k' : kflag++;
						   break;
					case 'm' : mflag++;
						   break;
					case 'v' : vflag++;
						   break;
					case 'V' : prt_msg();
						   exit(0);
					default  : fprintf(stderr,
				           "%s: %c - Unknown option\n",cmd,*cp);
						   usage(0);
						   exit(1);
				}
			}
		}
	}

	if ((sflag) && (lflag)) {
		fprintf(stderr,
			"%s: WARNING: -f option ignored as syntax check only\n",
					cmd);
		lflag = 0;
	}

	if ( (lflag) && (ld_file == NULL) ) {
		fprintf(stderr,
			"%s: No special device load file specified\n",cmd);
		usage(0);
		exit(1);
	}

	if (*argv == NULL) {
		fprintf(stderr,
			"%s: No fifo configuration file name specified\n",cmd);
		usage(0);
		exit(1);
	}

	config_file = *argv++;

	if ( (fp = fopen(config_file, "r")) == NULL) {
		perror("fopen");
		fprintf(stderr,
		"%s: Failed to open configuration file %s for reading\n",
				cmd, config_file);
		exit(1);
	}

	if (vflag) 
		printf("%s: Fifo initialisation configuration file is %s\n",
							cmd, config_file);


	if (sflag == 0) {
		/*
		 * Not a syntax check
		 */
		if (*argv == NULL) {
			fprintf(stderr,
				"%s: No card index specified\n",cmd);
			exit(1);
		}
		
		/*
		 * Check exists first
		 */
		if (access(ld_file,00) == -1) {
			perror("access");
			fprintf(stderr,
			"%s: Failed to open load device file %s for writing\n",
								cmd, ld_file);
			exit(1);
		}

		if (stat(ld_file,&stat_buf) == -1) {
			perror("stat");
			fprintf(stderr,
			"%s: Failed to stat load device file %s\n",
								cmd, ld_file);
			exit(1);
		}

		/*
		 * Check is a character device
		 */
		if ((stat_buf.st_mode & S_IFMT) != S_IFCHR) {
			fprintf(stderr,
		"%s: Load device file %s is not a character special device\n",
								cmd, ld_file);
			exit(1);
		}

		while (TRUE) {
			if ( (loadfp = open(ld_file,O_RDWR)) == -1 ) {
				if (errno == EBUSY) {
					sleep(1);
					continue;
				}
				perror("open");
				fprintf(stderr,
		   "%s: Failed to open load device %s (read/write access)\n",
					cmd, ld_file);
				exit(1);
			}
			else
				break;
		}

		if (vflag) 
			printf("%s: Load device file is %s\n",
						cmd, ld_file);

		/*
		 * Get the card index number, extra data after index ignored
		 */
		ptr = &strptr;
		i = strtol(*argv, ptr, 10);
	
		if (*ptr == *argv++) {
			fprintf(stderr,
				"%s: Card index %s in not a valid integer\n",
						cmd,*ptr);
			exit(1);
		}
	
		if ((*argv != NULL) && (wflag)) {
			fprintf(stderr,
				"%s: WARNING - extra data on input ignored\n",
								cmd);
		}
	
		/*
		 * See if card is valid
		 */
		if ( (i < 0) || (i >= MAXCARDS) ) {
			fprintf(stderr,
				"%s: Card index %d out of range\n",
							cmd, i);
			exit(1);
		}
	
		card_data.index = i;
		if (ioctl(loadfp, SCCL_ADRS, &card_data) == -1) {
			/*
			 * No card ! Get 'residual' data
			 */
			if (errno == ENXIO) {
				printf("%d %d\n",
					(int)card_data.adrs,(int)card_data.data);
			}
			else {
				printf("BAD errno\n");
				exit(-1);
			}
			fprintf(stderr,
				"%s: No card at index %d, out of range\n",cmd,i);
			exit(1);
		}
		
		if (card_data.adrs == (unsigned char *)-1) {
			/*
			 * Has a slot, but not present
			 */
			fprintf(stderr,
				"%s: No card at index %d\n",cmd,i);
			exit(1);
		}
	
		/*
		 * Have a card at index i
		 */
	
		if (vflag)
			printf("%s: Index %d, address 0x%08x, status 0x%08x\n",
					cmd,
					i,
					card_data.adrs,
					card_data.data);

		if (card_data.data != (CHK_32|HOST_MDEF)) {
			fprintf(stderr,
				"%s: Bad status for fifo config\n",cmd);;
			exit(1);
		}

		card_index = i;
	}
		
	/*
	 * Read first line. Must be identification line and thus it MUST
	 * have following format :
	 *
	 * SCC-FIFO-CONFIGURATION-FILE
	 * msg: This string or strings can be output by loader
	 */
	clear_ibuf();
#ifdef	DBUG
	printf("DBUG(main): strlen(head_msg) %d\n",strlen(head_msg));
#endif
	for (index = 0; index < strlen(head_msg); index++) {
		ibuf[index] = getc(fp);

		if (ibuf[index] == EOF) {
			fprintf(stderr,
				"%s: Unexpected EOF in header message\n",cmd);
			exit(1);
		}

		if (ibuf[index] == EOL) {
			fprintf(stderr,
				"%s: Unexpected EOL in header message\n",cmd);
			exit(1);
		}

		if (ibuf[index] != head_msg[index]) {
			fprintf(stderr,"%s: Bad character in header record\n",
					cmd);
			fprintf(stderr,"%s: Line 1, column %d, char %c\n",
					cmd, index+1, ibuf[index]);
			exit(1);
		}
	}
#ifdef	DBUG
	printf("DBUG(main): ibuf %s\n",ibuf);
#endif

	/*
	 * Get to end of line. Ignore white space and tabs. Other chars
	 * will cause an error
	 */
	while ( (c = getc(fp)) != EOL) {
		
		if (c == EOF) {
			fprintf(stderr,
				"%s: Unexpected EOF after header message\n",
									cmd);
			exit(1);
		}
		index++;
		if ((c == ' ') || (c == '\t') )
			continue;
		
		fprintf(stderr,"%s: Bad character at end of header record\n",
					cmd);
		fprintf(stderr,"%s: Line 1, column %d, char %c\n",
					cmd, index, c);
		exit(1);
	}

	/*
	 * Have head_msg, and therefore have a fifo init file
	 * Get the "msg:"
	 */
	for (index = 0; index < strlen(msg_msg); index++) {
		c = getc(fp);

		if (c == EOF) {
			fprintf(stderr,
				"%s: Unexpected EOF in MSG string\n",cmd);
		}

		if (c == EOL) {
			fprintf(stderr,
				"%s: Unexpected EOL in MSG string\n",cmd);
		}

		if (c != msg_msg[index]) {
			fprintf(stderr,"%s: Bad character in MSG string\n",
					cmd);
			fprintf(stderr,"%s: Line 1, column %d, char %c\n",
					cmd, index+1, c);
			exit(1);
		}
	}

	/*
	 * Print out rest of line if vflag, else skip it
	 */
#ifdef	DBUG
	printf("DBUG(main): vflag %d\n",vflag);
#endif

	if (mflag) vflag++;

	if (vflag)
		printf("%s: ",cmd);
	skip_line(fp,vflag,ANY_CHARS);
	if (vflag)
		printf("\n");

	if (mflag) exit(0);
		

	if ((vflag) && (sflag != 0))
		printf("%s: Syntax check only\n",cmd);

	/*
	 * Now process each line in file
	 */
	if (vflag) {
		printf("%s: fifo - header size used %d\n",cmd, FFHEAD);
		printf("%s: ifc  - maximum number of fifos is %d\n",
						cmd,MAXIFCS);
	}

	mem = 0;
	line = 3;

	for (p = 0; p < MAXPORTS; p++) {

		if (comment_skip(fp)) {
			fprintf(stderr,
			"%s: Unexpected EOF reading port %d definition\n",
							cmd, p);
			exit(1);
		}

		if (getc(fp) != RAW) {
			syntax();
			fprintf(stderr,"%s: Expected RAW definition, port %d\n",
							cmd,p);
			exit(1);
		}
		ret = fscanf(fp,":port:%d:raw:in=%d,out=%d,pout=%d",
						&port,&in[p],&out[p],&pout[p]);
#ifdef	DBUG
		printf("DBUG(main): ret %d\n",ret);
#endif
		if (ret == -1) {
			fprintf(stderr,
				"%s: Unexpected EOF reading RAW definition\n",
						cmd);
			exit(1);
		}
		else if (ret != 4) {
			syntax();
			fprintf(stderr,"%s: fscanf of RAW definition failed\n",
						cmd);
			exit(1);
		}

		if (p != port) {
			syntax();
			fprintf(stderr,
	"%s: Expected RAW port %d definition, but read port %d definition\n",
							cmd, p, port);
			exit(1);
		}
		skip_line(fp,0,WHITE_ONLY);

		if (vflag) {
			printf("%s: Port %d Fifo Sizes\n",cmd,port);
			printf("%s:    raw : in %d, out %d, priority out %d\n",
						cmd, in[p], out[p], pout[p]);
		}

		/* 
		 * Memory needed ...
		 */
		mem += (FFHEAD * 3);
		mem += (in[p]+1)*2;
		mem += (out[p]+2);
		mem += (pout[p]+2+12);		/* +12 to allow alignment */
	
		if (comment_skip(fp)) {
			fprintf(stderr,"%s: Unexpected EOF reading port %d\n",
							cmd, p);
			exit(1);
		}
		if (getc(fp) != HOST) {
			syntax();
			fprintf(stderr,
				"%s: Expected HOST definition, port %d\n",
							cmd,p);
			exit(1);
		}
		ret = fscanf(fp,":port:%d:host:out=%d,in=%d",
			&port,&hout[p],&non_canon_hin[p]);
#ifdef	DBUG
		printf("DBUG(main): ret %d\n",ret);
#endif
		if (ret == -1) {
			fprintf(stderr,
				"%s: Unexpected EOF reading HOST definition\n",
						cmd);
			exit(1);
		}
		else if (ret != 3) {
			syntax();
			fprintf(stderr,"%s: fscanf of HOST definition failed\n",
						cmd);
			exit(1);
		}
		
		if (p != port) {
			syntax();
			fprintf(stderr,
	"%s: Expected HOST port %d definition, but read port %d definition\n",
							cmd, p, port);
			exit(1);
		}
		skip_line(fp,0,WHITE_ONLY);

		if (vflag) {
			printf("%s:    host: out %d none canonical in %d\n",
						cmd, hout[p], non_canon_hin[p]);
		}
		
		/*
		 * Memeory needed ...
		 */
		mem += (FFHEAD * 2);
		mem += (hout[p]+2);
		mem += (non_canon_hin[p]+2+8);	/* +8 to allow alignment */

		if (comment_skip(fp)) {
			fprintf(stderr,"%s: Unexpected EOF reading port %d\n",
							cmd, p);
			exit(1);
		}
		if (getc(fp) != IFC) {
			syntax();
			fprintf(stderr,
				"%s: Expected IFC definition, port %d\n",
							cmd,p);
			exit(1);
		}
		ret = fscanf(fp,":port:%d:ifc:canon-fifos=%d:",
			&port,&no_ifc_fifos[p]);
#ifdef	DBUG
		printf("DBUG(main): ret %d\n",ret);
#endif
		if (ret == -1) {
			fprintf(stderr,
				"%s: Unexpected EOF reading IFC definition\n",
						cmd);
			exit(1);
		}
		else if (ret != 2) {
			syntax();
			fprintf(stderr,"%s: fscanf of IFC definition failed\n",
						cmd);
			exit(1);
		}
		
		if (p != port) {
			syntax();
			fprintf(stderr,
	"%s: Expected IFC port %d definition, but read port %d definition\n",
							cmd, p, port);
			exit(1);
		}
		
		if (no_ifc_fifos[p] > MAXIFCS) {
			fprintf(stderr,"%s: Too many ifc fifos - Max is %d\n",
				cmd,MAXIFCS);
			exit(1);
		}

		if (vflag) {
			printf("%s:    ifc canonical fifos: number %d\n",
							cmd,no_ifc_fifos[p]);
			printf("%s:       sizes",cmd);
		}

		for (cnt = 0; cnt < no_ifc_fifos[p]; cnt++) {
			ret = fscanf(fp,"%d",&ifc_size[(p*MAXIFCS)+cnt]);
#ifdef	DBUG
		printf("DBUG(main): ret %d\n",ret);
#endif

			if (ret == -1) {
				fflush(stdout); fprintf(stderr,"\n");
				fprintf(stderr,
			"%s: Unexpected EOF reading IFC size definition\n",
							cmd);
				exit(1);
			}
			else if (ret != 1) {
				fflush(stdout); fprintf(stderr,"\n");
				syntax();
		fprintf(stderr,"%s: fscanf of IFC size definitions failed\n",
						cmd);
				exit(1);
			}
			if (vflag) {
				if ( (cnt != 0) && ((cnt % 8) == 0) )
					printf("\n%s:            ",cmd);
				printf("%6d",ifc_size[(p*MAXIFCS)+cnt]);
			}
		
			/*
			 * Memeory needed ...
			 */
			mem += (ifc_size[(p*MAXIFCS)+cnt]+2+4+FFHEAD);
		}
		if (vflag) printf("\n");
		ret = skip_line(fp,0,WHITE_ONLY);

	}

	if (!ret) {
		/*
		 * Not got to end of file 
		 */
		if (comment_skip(fp) == 0) {
			/*
			 * Have skipped any comments at end and not got EOF
		 	 * Next thing MUST be EOF, else we have a line 
		 	 * at the end.
			 */
			if (getc(fp) != EOF) {
				syntax();
		fprintf(stderr,
		"%s: EOF expected, extra data, or blank lines, at file end\n",
						cmd);
				exit(1);
			}
		}
	}

	if (mem == 0) 
		kb_mem = 0;
	else
		kb_mem = ((mem/1024)+1) * 1024;

	if (vflag) {
		printf("%s: Memory required for fifos is %d bytes\n",
								cmd, mem);
		printf("%s: To nearest 1K is %d bytes, %d Kbytes\n",
						cmd, kb_mem, kb_mem/1024);
	}

	if ( (vflag) || (kflag) )
		printf("%s: Clicks of memory required %d\n",cmd,kb_mem/512);

	if (sflag == 0) {
		/*
		 * Generate fifo_defs array
		 */
		for (p = 0; p < MAXPORTS; p++) {
			fifo_defs[p].fs_rawin   = (short)in[p];
			fifo_defs[p].fs_rawout  = (short)out[p];
			fifo_defs[p].fs_rawpout = (short)pout[p];
			
			fifo_defs[p].fs_hostout    = (short)hout[p];
			fifo_defs[p].fs_hostnoncin = (short)non_canon_hin[p];

			fifo_defs[p].fs_ifc_no = no_ifc_fifos[p];

			for (cnt = 0; cnt < no_ifc_fifos[p]; cnt++)
				fifo_defs[p].fs_ifcs[cnt] =
					(short)ifc_size[(p*MAXIFCS)+cnt];
			fifo_defs[p].mem_grab = kb_mem/512;
			fifo_defs[p].card_index = card_index;
		}
		ffinit(loadfp, card_index);

	}

	exit(0);

}

/*
 * 
 * Support funtions
 */

clear_ibuf()
{
	register i;

	for (i = 0; i < IBUFSIZE; i++) ibuf[i] = (char)0;
}
	
skip_line(fp,prt,white_only)
FILE *fp;
int prt;
{
	register c;

	while ( (c = getc(fp)) != EOF) {
		if (c == EOL) {
			line++;
			return(0);
		}

		if (white_only) {
			if ((c != ' ') && (c != '\t') ) {
				syntax();
				fprintf(stderr,"%s: Expected EOL\n",
					cmd);
				exit(1);
			}
		}

		if (prt) printf("%c",c);
	}

	/*
	 * Have end of file
	 */
	return(1);	
}


comment_skip(fp)
FILE *fp;
{
	register c;

	while ((c = getc(fp)) != EOF) {
		if (c != COMMENT) {
			ungetc(c,fp);
			return(0);
		}
		if (skip_line(fp,0,ANY_CHARS)) break;
	}
	
	/*
	 * Have EOF
	 */
	return(1);
}

ffinit(fp,card)
int fp, card;
{
	int cnt, i;

	if (dflag) {

		printf("%s: Load data block for card %d\n",cmd,card);

		for (cnt = 0; cnt < MAXPORTS; cnt ++) 
		{
		printf("\n%s: FIFO defs for channel %d\n",cmd,cnt);
		printf("%s: chan %d rawin %d, rawout %d, rawpout %d\n",
			cmd,
			cnt,
			fifo_defs[cnt].fs_rawin,
			fifo_defs[cnt].fs_rawout,
			fifo_defs[cnt].fs_rawpout);
		printf("%s: chan %d hostout %d, hostnoncin %d, ifc_no %d\n",
			cmd,
			cnt,
			fifo_defs[cnt].fs_hostout,
			fifo_defs[cnt].fs_hostnoncin,
			fifo_defs[cnt].fs_ifc_no);
		printf("%s: chan %d IFCs:-",cmd,cnt);
		for (i = 0; i < 118; i++) {
			if ( (i == 0) || ((i % 4) == 0) )
				printf("\n%s: ",cmd);
			printf("ifc %3d, sz %4d | ",
				i,fifo_defs[cnt].fs_ifcs[i]);
		}
		printf("\n%s: chan %d, mem_grab   0x%x\n",
				cmd, cnt,fifo_defs[cnt].mem_grab);
		printf("%s: chan %d, card_index 0x%x\n",
				cmd, cnt,fifo_defs[cnt].card_index);
		}
	}

	if (ioctl(fp, SCCL_FIFO, fifo_defs) == -1) {
		perror("ioctl");
		se_error(fp,cmd);
		fprintf(stderr,
			"%s: SCCL_FIFO command failed to complete\n",cmd);
		exit(1);
	}
		
}

syntax()
{
	fprintf(stderr,"%s: Syntax in file %s at line %d\n",
		cmd,config_file,line);
}

prt_msg()
{
	printf("%s: %s\n",cmd,ffinit_id);
}

usage(all)
int all;
{
printf("usage: %s [-Vhklmsv] [load device] <config filename> <card-id>\n",
			cmd);

	if (all) {
		printf("\n");
		printf("options: -V  Print utility version string\n");
		printf("         -h  This help message\n");
		printf("         -k  Print number of clicks fifos require\n");
		printf("         -l  Specify new load special device (default /dev/scc.ctrl)\n");
		printf("         -m  Print message embedded in configuration file\n");
		printf("         -s  Syntax check file only\n");
		printf("         -v  Enable verbose mode\n");
	}
}

