/*
 *                  Copyright (c) 1988 Ferranti Plc.
 *                      All Rights Reserved
 *
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FERRANTI PLC THE COPYRIGHT 
 *                   OF WHICH VESTS IN FERRANTI PLC.
 *
 *   The copyright notice above does not evidence any actual or intended 
 *                   publication of such source code.
 *
 * This source code and the information contained herein is supplied without
 * liability for errors or ommissions and no part may be copied, re-produced, 
 * used or disclosed except as authorised by the Ferranti organisation by 
 *                contract or by other written permission.
 */

/******************************************************************************

	Rev = 5.1		Date of 'get'   = 11:11:00 1/26/89
				Date of 'delta' = 14:59:00 12/8/88

	File = sccinfo.c

******************************************************************************/

#ident "@(#)scc-cmds:cmds/sccinfo.c Rev 5.1 14:59:00 12/8/88"

/*
 * Usage: sccinfo [-Vlv] [load device] [card id index]
 *
 * -l 	Specify a load device name (default /dev/loaders)
 * -v 	Verbose
 * -V	Version message
 * -t	terse
 * -s	status only
 * -e 	error only
 * -r 	release only
 *
 * If no card id given then all cards are assumed
 */

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/termio.h>
#include <local/scc/scclioctl.h>

static char sccinfo_id[]="scc-cmds:cmds/sccinfo.c Rev 5.1 14:59:00 12/8/88";

#define	TRUE		1
#define	FALSE		0

#define	MAXCARDS 	16

int tflag;

struct	{
	int present;
	unsigned char *adrs;
	unsigned long status;
	unsigned long sccs_data;
	unsigned long gerror;
} card_info[MAXCARDS];

main(argc,argv)
int argc;
char *argv[];
{
	int sflag, eflag, rflag, vflag, lflag;
	int scc_cnt, scc_max_cards;
	int load_fd, free, i, cards;
	struct card_data card_data;
	char **ptr, *strptr;

	char *cmd, *ldfile, *cp;

	lflag = FALSE;
	vflag = FALSE;
	tflag = FALSE;
	sflag = TRUE;
	rflag = TRUE;
	eflag = TRUE;
	ldfile = "/dev/scc.ctrl";

	/*
	 * Set the command name into string pointer cmd
	 */

	if (cmd = (char *)strrchr(argv[0],'/'))
		++cmd;
	else
		cmd = argv[0];


	if (argc > 1) {

		/*
		 * We have arguments, but are they options ?
		 */	
		argv++;
		if ((*argv)[0] == '-') {
			/*
			 * Have options, parse them
			 */
			for (cp = *argv++; *cp; cp++) {
			switch (*cp) {
				case 'V': printf("%s: %s\n",cmd,sccinfo_id);
					  exit(0);
				case 'l': ldfile = *argv++;
					  lflag = TRUE;
					  break;
				case 'v': vflag = TRUE;
					  break;
				case 'a': sflag = FALSE;
					  rflag = FALSE;
					  eflag = FALSE;
					  break;
				case 'e': rflag = FALSE;
					  sflag = FALSE;
					  break;
				case 'r': sflag = FALSE;
					  eflag = FALSE;
					  break;
				case 's': rflag = FALSE;
					  eflag = FALSE;
					  break;
				case 't': tflag = TRUE;
					  vflag = FALSE;
					  break;
				case 'h': usage(1);
					  exit(0);
				case '-': break;
				default : optionerr(cmd,*cp);
					  exit(127);
			}	/* end of switch */
			}	/* end of for */
		}	/* end of option if statement */

	} 
	else argv++;

	/*
	 * Init card_info to FALSE
	 */
	for (i=0; i< MAXCARDS; i++) card_info[i].present = FALSE;

	if ((lflag) && (ldfile == NULL)) {
		fprintf(stderr,"%s: No device name specified\n",cmd);
		usage(0);
		exit(127);
	}

	/* 
	 * Open device to get into loader driver 
	 */
	while (TRUE) {
		if ( (load_fd = open(ldfile,O_RDWR)) == -1 ) {
			if (errno == EBUSY) {
				sleep(1);
				continue;
			}
			perror("open");
			fprintf(stderr,
		   "%s: Failed to open device %s (read/write access)\n",
					cmd, ldfile);
			exit(1);
		}
		else
			break;
	}

	/*
	 * Build card index array table anyway. Cancel any entries that
	 * are not to be loaded later. Thus a card index can be checked to
	 * see if it is valid
	 */
	cards = 0;
	for (i = 0; i< MAXCARDS; i++) {
		card_data.index = i;
		if (ioctl(load_fd, SCCL_ADRS, &card_data) == -1) {
			/*
			 * No more cards ! Get 'residual' data
			 */
			if (errno == ENXIO) {
				scc_max_cards = (int)card_data.adrs;
				scc_cnt	      = (int)card_data.data;
			}
			else {
				printf("BAD errno\n");
				exit(-1);
			}
			break;
		}
		if (card_data.adrs != (unsigned char *)-1) {
			card_info[i].present = TRUE;
			card_info[i].adrs    = card_data.adrs;
			card_info[i].status  = card_data.data;
			cards++;

			/* 
			 * Get SCCS Issue data
			 */
			if (ioctl(load_fd, SCCL_SCCS, &card_data) == -1) {
				/*
				 * Error, should never get this !
				 */
				printf("CHECK - BAD error\n");
				exit(-1);
			}
			card_info[i].sccs_data = card_data.data;

			/* 
			 * Get error word
			 */
			if (ioctl(load_fd, SCCL_GERROR, &card_data) == -1) {
				/*
				 * Error, should never get this !
				 */
				printf("CHECK - BAD error\n");
				exit(-1);
			}
			card_info[i].gerror = card_data.data;
		}
	}

	if (cards == 0) {
		fprintf(stderr,"%s: No SCC cards configured into system\n",
			cmd);
		exit(1);
	}

	if (vflag) {
		printf("%s: scc_cnt %d, scc_max_cards %d\n",
			cmd, scc_cnt, scc_max_cards);
	}

	if (*argv == NULL) {
		for (i = 0; i < (scc_cnt/16); i++) {
			if (card_info[i].present == FALSE)
			printf("%s: Card %2d - Not present\n",cmd,i);
			else {
				if (tflag)
					printf("%2d ",i);

				if ((sflag) && (tflag))
					printf("0x%08x ",card_info[i].status);
				else if (sflag)
					printf("%s: Card %2d - Address 0x%08x, status 0x%08x\n",
						cmd, i,
						card_info[i].adrs,
						card_info[i].status);

				if ((eflag) && 
					(card_info[i].status == 0x62AE0001)) {
	fprintf(stderr,"%s: Card %2d - No error, awaiting code load\n",cmd,i);
					if (tflag) printf("0x%08x ",0);
				}	
				else if ((eflag) && (tflag))
					printf("0x%08x ",card_info[i].gerror);
				else if (eflag)
					printf("%s: Card %2d - Error word 0x%08x\n",
						cmd,i,card_info[i].gerror);
			
				if ((rflag) && (tflag))
					print_sccs(card_info[i].sccs_data);
				else if (rflag) {
					printf("%s: Card %2d - Firmware ",cmd,i);
					print_sccs(card_info[i].sccs_data);
					printf("\n");
				}
				if (tflag)
					printf("\n");
			}
		}
	}
	else {
		/*
		 * Only cards that are in info list are printed
		 */
	
		while (*argv != NULL) {
			ptr = &strptr;
			i = strtol(*argv, ptr, 10);

			if (*ptr == *argv++) {
				fprintf(stderr,
				"%s: Card index %s in not a valid integer\n",
					cmd,*ptr);
				continue;
			}

			if ( (i < 0) || (i >= MAXCARDS) ) {
				fprintf(stderr,
					"%s: Card index %d out of range\n",
								cmd, i);
				continue;
			}
			else if (i >= (scc_cnt/16)) {
				fprintf(stderr,
			"%s: Card index %d out of range for this build\n",
						cmd,i);
				continue;
			}
			
			if (card_info[i].present == FALSE)
			printf("%s: Card %2d - Not present\n",cmd,i);
			else {
				if (tflag)
					printf("%2d ",i);

				if ((sflag) && (tflag))
					printf("0x%08x ",card_info[i].status);
				else if (sflag)
					printf("%s: Card %2d - Address 0x%08x, status 0x%08x\n",
						cmd, i,
						card_info[i].adrs,
						card_info[i].status);

				if ((eflag) && 
					(card_info[i].status == 0x62AE0001)) {
	fprintf(stderr,"%s: Card %2d - No error, awaiting code load\n",cmd,i);
					if (tflag) printf("0x%08x ",0);
				}	
				else if ((eflag) && (tflag))
					printf("0x%08x ",card_info[i].gerror);
				else if (eflag)
					printf("%s: Card %2d - Error word 0x%08x\n",
						cmd,i,card_info[i].gerror);
			
				if ((rflag) && (tflag))
					print_sccs(card_info[i].sccs_data);
				else if (rflag) {
					printf("%s: Card %2d - Firmware ",cmd,i);
					print_sccs(card_info[i].sccs_data);
					printf("\n");
				}
				if (tflag)
					printf("\n");
			}
		}
	}	
	exit(0);
}

optionerr(cmd,ch)
char *cmd;
char ch;
{
	fprintf(stderr,"%s: %c - Unknown option\n",cmd,ch);
	usage(0);
}

usage(all)
int all;
{
printf("usage: sccinfo [-Vehlrstv] [load device] [card-id [card-id] .. ]\n");
	if (all) {
		printf("\n");
		printf("options: -V  Print utility version string\n");
		printf("         -e  Error word only\n");
		printf("         -h  This help message\n");
		printf("         -l  Specify new load special device (default /dev/scc.ctrl)\n");
		printf("         -r  Release/Issue data only\n");
		printf("         -s  Status word only\n");
		printf("         -t  Enable terse mode\n");
		printf("         -v  Enable verbose mode\n");
	}
}

print_sccs(data)
unsigned long data;
{
	/*
	 * Format of data is 4 printable characters. Print as XX YY
	 */
	if (data == 0x0) {
		if (tflag) printf("NU LL");
		else
		printf("Issue/Release, no release data loaded (null)");
	}
	else if (data == 0x54455354) {	/* 'T'E'S'T */
		if (tflag) printf("TE ST");
		else
		printf("Issue/Release set to TEST, pre-release code loaded");
	}
	else {
		if (!tflag) printf("Issue ");
		if (isprint( (data >> 24)&0xFF ))
			printf("%c", (data >> 24) & 0xFF );
		else 
			printf("?");
		if (isprint( (data >> 16)&0xFF ))
			printf("%c ", (data >> 16) & 0xFF );
		else 
			printf("? ");
		if (!tflag) printf("Release ");
		if (isprint( (data >> 8)&0xFF ))
			printf("%c", (data >> 8) & 0xFF );
		else 
			printf("?");
		if (isprint( (data >> 0)&0xFF ))
			printf("%c", (data >> 0) & 0xFF );
		else 
			printf("?");
	}
}


