/*
 *                  Copyright (c) 1988 Ferranti Plc.
 *                      All Rights Reserved
 *
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FERRANTI PLC THE COPYRIGHT 
 *                   OF WHICH VESTS IN FERRANTI PLC.
 *
 *   The copyright notice above does not evidence any actual or intended 
 *                   publication of such source code.
 *
 * This source code and the information contained herein is supplied without
 * liability for errors or ommissions and no part may be copied, re-produced, 
 * used or disclosed except as authorised by the Ferranti organisation by 
 *                contract or by other written permission.
 */

/******************************************************************************

	Rev = 5.1		Date of 'get'   = 11:11:00 1/26/89
				Date of 'delta' = 14:59:02 12/8/88

	File = sccreset.c

******************************************************************************/

#ident "@(#)scc-cmds:cmds/sccreset.c Rev 5.1 14:59:02 12/8/88"

/*
 * Usage: sccreset [-Vlv] [load device] <card id index>
 *
 * -l 	Specify a control device name (default /dev/scc.ctrl)
 * -v 	Verbose
 * -V	Version message
 *
 * If no card id given then the reset will be aborted
 */

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/termio.h>
#include <local/scc/scclioctl.h>

/*
 * what identification string
 */
static char sccreset_id[] = "scc-cmds:cmds/sccreset.c Rev 5.1 14:59:02 12/8/88";

#define	TRUE		1
#define	FALSE		0

#define	MAXCARDS 	16


struct	{
	int present;
	unsigned char *adrs;
} card_info[MAXCARDS];

int vflag, load_fd;
char *cmd;

main(argc,argv)
int argc;
char *argv[];
{
	int lflag;
	int scc_cnt, scc_max_cards;
	int free, i, cards;
	struct card_data card_data;
	char **ptr, *strptr;

	char *ldfile, *cp;

	lflag = 0; vflag = 0;
	ldfile = "/dev/scc.ctrl";

	/*
	 * Set the command name into string pointer cmd
	 */

	if (cmd = (char *)strrchr(argv[0],'/'))
		++cmd;
	else
		cmd = argv[0];

	if (argc == 1) {
		/*
		 * No card id specified, bad arg count
		 */
		fprintf(stderr,"%s: No card index specified\n",cmd);
		usage(0);
		exit(1);
	}

	/*
	 * We have arguments, but are they options ?
	 */	
	argv++;
	if ((*argv)[0] == '-') {
		/*
		 * Have options, parse them
		 */
		for (cp = *argv++; *cp; cp++) {
		switch (*cp) {
			case 'V': printf("%s: %s\n",cmd,sccreset_id);
				  exit(0);
			case 'l': ldfile = *argv++;
				  lflag++;
				  break;
			case 'v': vflag++;
				  break;
			case 'h': usage(1);
				  exit(0);
			case '-': break;
			default : optionerr(cmd,*cp);
				  exit(127);
		}	/* end of switch */
		}	/* end of for */
	}	/* end of option if statement */

	/*
	 * Do we have any more data, possible card ids ?
   	 */
	if (*argv == NULL) {
		fprintf(stderr,"%s: No card id(s)\n",cmd);
		usage(0);
		exit(1);
	}

	if ( (lflag) && (ldfile == NULL) ) {
		fprintf(stderr,"%s: No device file name specified\n",cmd);
		usage(0);
		exit(1);
	}
	
	/*
	 * Init card_info to FALSE
	 */
	for (i=0; i< MAXCARDS; i++) card_info[i].present = FALSE;

	/* 
	 * Open device to get into loader driver 
	 */
	while (TRUE) {
		if ( (load_fd = open(ldfile,O_RDWR)) == -1 ) {
			if (errno == EBUSY) {
				sleep(1);
				continue;
			}
			perror("open");
			fprintf(stderr,
		   "%s: Failed to open load device %s (read/write access)\n",
					cmd, ldfile);
			exit(1);
		}
		else
			break;
	}

	/*
	 * Build card index array table anyway. Cancel any entries that
	 * are not to be loaded later. Thus a card index can be checked to
	 * see if it is valid
	 */
	cards = 0;
	for (i = 0; i< MAXCARDS; i++) {
		card_data.index = i;
		if (ioctl(load_fd, SCCL_ADRS, &card_data) == -1) {
			/*
			 * No more cards ! Get 'residual' data
			 */
			if (errno == ENXIO) {
				scc_max_cards = (int)card_data.adrs;
				scc_cnt	      = (int)card_data.data;
			}
			else {
				printf("BAD errno\n");
				exit(-1);
			}
			break;
		}
		if (card_data.adrs != (unsigned char *)-1) {
			card_info[i].present = TRUE;
			card_info[i].adrs    = card_data.adrs;
			cards++;
		}
	}

	if (cards == 0) {
		fprintf(stderr,"%s: No SCC cards configured into system\n",
			cmd);
		exit(1);
	}

	/*
	 * Only cards that are in info list are reset 
	 */

	while (*argv != NULL) {
		ptr = &strptr;
		i = strtol(*argv, ptr, 10);

		if (*ptr == *argv++) {
			fprintf(stderr,
				"%s: Card index %s is not a valid integer\n",
								cmd, *ptr);
			continue;
		}

		if ( (i < 0) || (i >= MAXCARDS) ) {
			fprintf(stderr,
				"%s: Card index %d out of range\n",
							cmd, i);
			continue;
		}
		else if (i >= (scc_cnt/16)) {
			fprintf(stderr,
		"%s: Card index %d out of range for this build\n",
					cmd,i);
			continue;
		}
		
		if (card_info[i].present == FALSE)
			printf("%s: Card %d not present\n",cmd,i);
		else {
			reset_card(i);
		}
	}
	
	exit(0);
}

optionerr(cmd,ch)
char *cmd;
char ch;
{
	fprintf(stderr,"%s: %c - Unknown option\n",cmd,ch);
	usage(0);
}

usage(all)
int all;
{
printf("usage: sccreset [-Vhlv] [load device] <card-id> [<card-id> .. ]\n");
	if (all) {
		printf("\n");
		printf("options: -V  Print utility version string\n");
		printf("         -h  This help message\n");
		printf("         -l  Specify new load special device (default /dev/scc.ctrl)\n");
		printf("         -v  Enable verbose mode\n");
	}
}

reset_card(index)
int index;
{
	if (vflag) {
		printf("%s: RESET card %d ...",cmd,index);
		fflush(stdout);
	}

	if ( (index & 0x0000FFFF) != index ) {
		fprintf(stderr,"%s: Funtion reset_card - FATAL ERROR\n");
		fprintf(stderr,
			"%s: Index word (0x%x) has a top 16 bit set !\n",
					cmd, index);
		exit(127);
	}

	/*
	 * SCCL_SWRESET: 
	 */
	if (ioctl(load_fd, SCCL_SWRESET, &index) == -1 ) {
		fprintf(stderr,"\n");
		perror("ioctl");
		se_error(load_fd, cmd);
		fprintf(stderr,
			"%s: Failed to complete primary RESET sequence\n",
				cmd);

		exit(1);
	}

	if (vflag) printf(" Done\n");

}



