/*
 *                  Copyright (c) 1988 Ferranti Plc.
 *                      All Rights Reserved
 *
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FERRANTI PLC THE COPYRIGHT 
 *                   OF WHICH VESTS IN FERRANTI PLC.
 *
 *   The copyright notice above does not evidence any actual or intended 
 *                   publication of such source code.
 *
 * This source code and the information contained herein is supplied without
 * liability for errors or ommissions and no part may be copied, re-produced, 
 * used or disclosed except as authorised by the Ferranti organisation by 
 *                contract or by other written permission.
 */

/******************************************************************************

	Rev = 5.1		Date of 'get'   = 11:11:02 1/26/89
				Date of 'delta' = 14:59:08 12/8/88

	File = srecval.c

******************************************************************************/

#ident "@(#)scc-cmds:cmds/srecval.c Rev 5.1 14:59:08 12/8/88"

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/termio.h>
#include <local/scc/scclioctl.h>
#include <local/scc/boolean.h>

/*
 * -V identification string
 */
static char srecval_id[] = "scc-cmds:cmds/srecval.c Rev 5.1 14:59:08 12/8/88";

/*
 * Usage: srecval [-v] file name
 */

char *cmd;

main(argc,argv)
int argc;
char *argv[];
{
	FILE * fp, * fopen();
	char bytes_sum[2];
	char data[4096];
	char chksum[2];
	char addr[8];
	int  * bytes;
	int  * checksum;
	int  * address;
	int  * n;
	char * ch;
	int    i, c, line_no, total, load_fd;
	int first;
	short  end_record;
	unsigned data_sz, addr_sz;
	struct load_info record;
	int * dummy;
	int dots;

	char *cp;
	int vflag, dflag;

	dflag = 0;
	vflag = 0;

	if (cmd = (char *)strrchr(argv[0],'/'))
		++cmd;
	else
		cmd = argv[0];

	if (argc <= 1) {
		fprintf(stderr,"%s: No S-record file name specified\n",cmd);
		exit(1);
	}


	if (argc > 1) {

		/*
		 * We have arguments ...
		 */	
		argv++;
		if ((*argv)[0] == '-') {
			/*
			 * Have options
			 */
			for (cp = *argv++; *cp; cp++) {
			switch (*cp) {
				case 'V':
					  printf("%s: %s\n",
							cmd,srecval_id);
					  exit(0);
				case 'h': usage(1);
					  exit(0);
				case 'd': dflag++;
					  printf("%s: DEBUG enabled\n",cmd);
					  break;
				case 'v': vflag++;
					  break;
				case '-': break;
				default :
					  fprintf(stderr,
				   "%s: \"%c\" - Unknown option\n",cmd,*cp);
					  usage(0);
					  exit(1);
			}	/* end of switch */
			}	/* end of for */
		}

				
	}
	
	if (*argv == NULL) {
		fprintf(stderr,"%s: No S-record file name specified\n", cmd);
		exit(1);
	}

	if ( ( fp = fopen(*argv, "r") ) == NULL )
	{
		perror("fopen");
		fprintf(stderr,"%s: Failed to open %s for reading\n",
			cmd, *argv);
		exit(1);
	}

	rewind(fp);

	line_no = 0; dots = 0;
	end_record = FALSE;
	/* 
	 * Get chars while not end of file and 
	 * last record was not an 'end record'
	 */

	if (vflag) {
		printf("%s: Validating file %s\n",cmd, *argv);
		printf("%s: Line %5d ",cmd,line_no);
	}

	while ( ( c = getc(fp) ) != EOF  && !end_record)
	{
		if (vflag) {
			if ((line_no % 10) == 0) {
				printf(".");
				if (dots >= 50) {
					dots = 0;
					printf("\n%s: Line %5d .",cmd, line_no);
				}

				dots++;
				fflush(stdout);
			}
		}

		line_no++;

		if ( c != START_CODE ) {
			fprintf(stderr,
		   		"%s: First character line %d is not an 'S', read a '%c'\n",
						cmd,line_no,c);
			exit(1);
		}

		switch ( c = getc(fp) )
		{
		case '1':
			end_record = FALSE;
			addr_sz = S9_S1_ADDR_SZ;
			break;
		case '9':
			end_record = TRUE;
			addr_sz = S9_S1_ADDR_SZ;
			break;
		case '2':
			end_record = FALSE;
			addr_sz = S8_S2_ADDR_SZ;
			break;
		case '8':
			end_record = TRUE;
			addr_sz = S8_S2_ADDR_SZ;
			break;
		case '3':
			end_record = FALSE;
			addr_sz = S7_S3_ADDR_SZ;
			break;
		case '7':
			end_record = TRUE;
			addr_sz = S7_S3_ADDR_SZ;
			break;
		case '0':
			end_record = FALSE;
			while ( (c = getc(fp)) != EOR ) ;
			continue;	
		default:
			fprintf(stderr,
				"%s: Invalid S-record type (%d) at line %d\n",
						cmd, c - '0', line_no);
			exit(1);
		}
		record.type = c - '0';

		if (dflag)
			printf("%s: Line %d, record type is %d\n",
					cmd, line_no, record.type);
		
		/* 
		 * Get number of bytes in this record 
		 */
		for ( i = 0; i < SUM_SZ; i++)
			bytes_sum[i] = getc(fp);
		sscanf(bytes_sum,"%2x",&bytes);

		if (dflag)
			printf("%s: Size = %d\n",cmd,bytes);

		/* 
	 	 * Get offset address 
		 */
		for ( i = 0; i < addr_sz; i++ ) 
			addr[i] = getc(fp);
		for (i = addr_sz; i<8; i++)
			addr[i] = 0;

		if (addr_sz == S9_S1_ADDR_SZ) {
			sscanf(addr, "%4x", &address);
		}
		else if (addr_sz == S8_S2_ADDR_SZ) {
			sscanf(addr, "%6x", &address);
		}
		else if (addr_sz == S7_S3_ADDR_SZ) {
			sscanf(addr, "%8x", &address);
		}
		else {
		}
		record.offset = (int) address;

		if (dflag)
		printf("%s: address = 0x%x\n",cmd,address);

		/* get data */
		data_sz = ( (unsigned) bytes * 2 ) - addr_sz - CHKSUM_SZ;
		record.data_sz = (int) data_sz;

		if (dflag)
			printf("%s: data chars = %d\n",cmd,data_sz);
		
		if (data_sz > 4096) {
			printf("TOO BIG\n");
			exit(1);
		}

		for ( i = 0; i < data_sz; i++ )
			data[i] = getc(fp);
		data[i] = 0;

		record.data = data;
		if (dflag)
			printf("%s: data = %s\n",cmd, data);

		/* 
	 	 * Get checksum 
		 */
		for ( i = 0; i < CHKSUM_SZ; i++ )
			chksum[i] = getc(fp);
		sscanf(chksum, "%2x", &checksum);

		if (dflag)
			printf("%s: checksum = 0x%x\n",cmd,checksum);

#ifdef DBUG
		printf("end_record = %d\n",end_record);
#endif
		if ( (c = getc(fp)) != EOR ) {
			fprintf(stderr,
"%s: Expected end_of_record after checksum and not a '%c', line %d\n",
							cmd, c, line_no);
			exit(1);
		}

		/* check checksum */
		total = 0;
		first = TRUE;

#ifdef	DBUG
		printf("DBUG: bytes_sum = %s\n",bytes_sum);
#endif	
		for ( i = 0; i < SUM_SZ; i++ ) {
			sscanf(&bytes_sum[i], "%1x", &n);
#ifdef	DBUG
			printf("DBUG: %c 0x%x\n",bytes_sum[i],n);
#endif
			calc_total(n, &total, &first);
		}
	
#ifdef	DBUG
		printf("DBUG: addr = %s\n",addr);
#endif	
		for ( i = 0; i < addr_sz; i++ ) {
			sscanf(&addr[i], "%1x", &n);
#ifdef	DBUG
			printf("DBUG: %c 0x%x\n",addr[i],n);
#endif
			calc_total(n, &total, &first);
		}
		
		for ( i = 0; i < data_sz; i++ ) {
			sscanf(&data[i], "%1x", &n);
#ifdef	DBUG
			printf("DBUG: %c 0x%x\n",data[i],n);
#endif
			calc_total(n, &total, &first);
		}
#ifdef DBUG
		printf("total = 0x%x, checksum = 0x%x\n",total,checksum);
#endif
		total = ~total;
#ifdef DBUG
		printf("~total = 0x%x, checksum = 0x%x\n",total,checksum);
#endif
		if ( (total & 0xFF) != (int) checksum ) {
			if (vflag)
				printf("\n");
			fprintf(stderr,
			   "%s: Checksum error at line %d in file %s\n",
					cmd,line_no,*argv);
			exit(255);
		}

		if ( (end_record) && (vflag) )
			printf("\n%s: Code start address is 0x%x\n",
				cmd,record.offset);

		bytes = address = checksum = n = 0;
	}

	if ( fclose(fp) != 0 ) {
		perror("fclose");
		fprintf(stderr,"%s: Failed to close %s\n", cmd, *argv);
		exit(1);
	}

	if ( c == EOF && !end_record ) {
		fprintf(stderr,"%s: Last S-record was not an end record\n",
					cmd);
		exit(1);
	}

	if (vflag)
		printf("%s: S-record file %s validated for loading\n",
								cmd,*argv);
	exit(0);
}

calc_total(n, total, first)
int *total;
int *first;
int n;
{
#ifdef	DBUG
	printf("calc_total: n %d, first %d\n",n,*first);
#endif
	if (*first)
	{
		*total += (int) n * 16;
		*first = FALSE;
#ifdef	DBUG
		printf("calc_total: first was TRUE, now FALSE\n");
#endif
	}
	else
	{
		*total += (int) n;
		*first = TRUE;
#ifdef	DBUG
		printf("calc_total: first was ^TRUE, now TRUE\n");
#endif
	}

#ifdef DBUG
	printf("calc_total: total 0x%x\n",*total);
#endif

}


usage(all)
int all;
{
	printf("usage: srecval [-Vhv] <S-record file name>\n");
	if (all) {
		printf("options: -V  Print utility version string\n");
		printf("         -h  This help message\n");
		printf("         -v  Enable verbose mode\n");
	}
}
