#ident "@(#)ce_bds.c	35.1"

#include	<stdio.h>
#include	<fcntl.h>
#include	<sys/c37x/ci_driver.h>
#include	<sys/c37x/ci_const.h>

#define		CEFMT		"/dev/ci/m37x_%d"
#define		ERROR		-1
#define		BOARDBASE	0xFF000000

/*
 * Different types of printed output
 */
#define		P_HDR		0x0001	/* set=print header */
#define		P_FULL		0x0002	/* set=full format, clear=short */

/*
 * Used for ioctl(CONFIG) to Common Environment (CE) to get board stats.
 */
CPU_TAB	Table[NUM_CPU_ENTRIES];

char	*Name;
char	*Usage = { "Usage: %s [ -sh ]\n" };

extern int	optind;
extern int	errno, sys_nerr;
extern char	*sys_errlist[];

main(argc, argv)
int	argc;
char	*argv[];
{
	int	c;
	int	o_format = (P_HDR|P_FULL);
	int	command_err = FALSE;

	Name = argv[0];

	while ( (c=getopt(argc, argv, "sh")) > 0 ) {
		switch( c ) {
		    case 's':		/* short format output */
			o_format &= ~P_FULL;
			break;
		    case 'h':		/* no headers */
			o_format &= ~P_HDR;
			break;
		    case '?':
			command_err = TRUE;
			break;
		}
		if ( command_err ) {
			break;		/* exit while loop */
		}
	}

	if ( command_err || argc != optind ) {
		fprintf(stderr, Usage, Name);
		exit(1);
	}

	exit( doconfig(o_format) );
}
/*
 * Find a CE device that is openable and get board configuration
 * information form it. This info will be for all CE boards, not just
 * the opened one.
 */
doconfig(outargs)
int	outargs;
{
	int	ce_fd;
	int	i;
	char	devbuf[64];

	/*
	 * We must find one of the minor devices that will open
	 * successfully for the ioctl to succeed.
	 */
	for (i=CPU_OFFSET; i <= NUM_CPU_ENTRIES ; i++) 
	{
		sprintf(devbuf, CEFMT, i);
		if ( (ce_fd=open(devbuf, O_RDONLY)) == ERROR ) {
			continue;
		}
		break;
	}

	if ( ce_fd == ERROR ) 
	{
		/* could not open any CE device */
		my_perror("open of %s failed", devbuf, 0, 0);
		return(1);
	}

	if ( ioctl(ce_fd, CONFIG, Table) == ERROR ) 
	{
		my_perror("CONFIG ioctl failed", 0, 0, 0);
		close(ce_fd);
		return(1);
	}
	close(ce_fd);

	printcfg(outargs);

	return(0);
}

/*
 * Pring out information on boards that are in the system.
 */
printcfg(outargs)
int	outargs;
{
	unsigned long	i;
	char	ctype[28];
	CPU_TAB	*tab = Table;

	/*
	 * Print header if necessary.
	 */
	if ( outargs & P_HDR ) {
		if ( outargs & P_FULL ) {
			printf("cpu    type       addr     board   state\n");
		}
		else {
			printf("cpu    type    board\n");
		}
	}

	/*
	 * Go through cpu table looking for boards that are present.
	 */
	for (i=0; i < NUM_CPU_ENTRIES - 1; i++, tab++) {
		/* ignore slots that have no board associated with them */
		if ( tab->cpu_present == FALSE ) {
			continue;
		}

		sprintf(ctype,"%d", tab->cpu_type);
		if ( outargs & P_FULL ) {
			printf("%2d      %3s    0x%08X    %d     %s\n",
			     tab->cpu_id,
			     tab->cpu_type ? ctype : "",
			     BOARDBASE + (i * 0x100000),
			     i,
			     tab->cpu_running ? "communicating" : "failure");
		}
		else {
			printf("%2d      %3s      %d\n",
			     tab->cpu_id,
			     tab->cpu_type ? ctype : "", i);
		}
	}
}

my_perror(f, a1, a2, a3)
char	*f;
int	a1, a2, a3;
{
	char		*sys_err = "unknown error";

	fprintf(stderr, "%s: ", Name);

	if ( f ) {
		fprintf(stderr, f, a1, a2, a3);
		fprintf(stderr, ": ");
	}

	if ( errno < sys_nerr ) {
		sys_err = sys_errlist[errno];
	}

	fprintf(stderr, "%s\n", sys_err);
	fflush( stderr );
}
