static	char ident[] = "@(#)dl.c	35.1		5/25/88	Copyright MOTOROLA, INC.";
/*MH**************************************************************************
**
**
**      3/17/86         Created (BC)
**	6/15/86		Completely rewritten by Dana Holgate
**	10/1/86		Changed to speed up.  No longer waits for the response
**			for each load data EPB before sending the next one.
**			Instead counts number sent out and number of responses.
**	12/17/86	Changed to read all queued responses before preceeding
**			with load data EPBs.
**	06/10/87	Added multiple board support - R. Robinson.
**
******************************************************************************/
#include <stdio.h>
#include <filehdr.h>
#include <ldfcn.h>
#include <scnhdr.h>
#include <aouthdr.h>
#include <errno.h>
#include <sys/c37x/ci_driver.h>

#define LDFAILURE	0
#define LDSUCCESS	1
#undef FAILURE
#undef SUCCESS

#define FALSE 0
#define TRUE 1


static  unsigned long	dest_cpu;	/* destination CPU number */
static  unsigned long	dst_id;		/* destination id */
static  unsigned long	src_id;		/* source id */
static  int nex_sect = 0;	/* number of sections to be excluded */
static  char    *ex_sect[16];	/* section names to be excluded */


extern  int  ldshread();     /* SYS read section header */
extern  int  ldsseek();      /* SYS seek to section data area */
extern  LDFILE 	*ldopen();       /* SYS open load file */
extern  int  ldclose();      /* SYS close load file */
extern  void     exit();         /* SYS back to OS */
extern  int  ldohseek();     /* SYS seek to the optional file header */
extern	int	 errno;

main(argc, argv)
int   argc;
char    **argv;
{
	char *strtok();
	extern char *optarg;
	extern int optind;
	extern int opterr;
	int ci_fd;	/* Common Interace driver file descriptor */
	LDFILE  *ldfp = (LDFILE *)NULL;	/* a.out ld file pointer */
	int c;
	int err = 0;
	char *cvterr;
	char *note = (char *)NULL;
	unsigned char exec = FALSE;
	unsigned char load = FALSE;
	unsigned char sect = FALSE;
	unsigned char addr = FALSE;
	char    *addr_arg;
	char    *arg_arg;
	unsigned long	exec_arg;
	unsigned long   entrypt;
	AOUTHDR	aouthdr;

	opterr = 1;	/* shut up getopt()! */

	while ((c = getopt(argc,argv,"elx:a:s:")) != EOF)
	{
		switch (c)
		{
		case 'a':	/* execute at given address */
			if (exec || load || sect)
			{
				err++;
				note = "Mutually exclusive options encountered.";
				break;
			}
			addr = TRUE;
			if (optarg == (char *)NULL)
			{
				note = "Missing address argument with -a option.";
				err++;
				break;
			}
			addr_arg = strtok(optarg,",");
			entrypt = strtol(addr_arg,&cvterr, 0);
			if (*cvterr != NULL)
			{
				note = "Unable to convert -a address argument.";
				err++;
				break;
			}
			if ((arg_arg = strtok((char *)NULL,"")) != (char *)NULL)
			{
				exec_arg = strtol(arg_arg,&cvterr, 0);
				if (*cvterr != NULL)
				{
					note = "Unable to convert -a execution argument.";
					err++;
				}
			}
			else
				exec_arg = 0;
			break;
		case 's':
			note = "Execution at symbol, '-s', is not yet implemented.";
			err++;
			break;
		case 'e':	/* execute only */
			if (addr || load || sect)
			{
				err++;
				note = "Mutually exclusive options encountered.";
			}
			exec = TRUE;
			break;
		case 'l':	/* down load only */
			if (addr || exec)
			{
				err++;
				note = "Mutually exclusive options encountered.";
			}
			load = TRUE;
			break;
		case 'x':	/* section to be excluded from download */
			if (addr || exec)
			{
				err++;
				note = "Mutually exclusive options encountered.";
			}
			sect = TRUE;
			if (optarg == (char *)NULL)
			{
				err++;
				note = "Missing section name with -x option.";
				break;
			}
			ex_sect[nex_sect] = optarg;
			nex_sect++;
			break;
		default:			/* all other options */
			err++;
			note = "Unrecognized option.";
			break;
		}
		if (err)
		{
			fprintf(stderr,"%s\n\nUsage: %s { -a<address>[,<arg>] | -s<symbol>[,<arg>] | [ (-x<sect_name>) ... ] [{-e | -l}] <a.out> } <dest_cpu>\n\n",note,argv[0]);
			exit(1);
		}
	}

/*
** check the remaining arguments
*/
	if ((addr && (optind + 1 != argc)) || (!addr && (optind + 2 != argc)))
	{
		note = "Wrong number arguments.";
		fprintf(stderr,"%s\n\nUsage: %s { -a<address>[,<arg>] | -s<symbol>[,<arg>] | [ (-x<sect_name>) ... ] [{-e | -l}] <a.out> } <dest_cpu>\n\n",note,argv[0]);
		exit(1);
	}

/*
** prepare the coff file for use if required
*/
	if (!addr)
	{
		if ((ldfp = ldopen(argv[optind],ldfp)) == (LDFILE *)NULL)
		{
			fprintf(stderr,"%s: unable to open a.out file %s.\n",argv[0],argv[optind]);
			exit(1);
		}
/*
** get the entry point for execution (even though we may not be executing).
*/
		if (ldohseek(ldfp) == LDFAILURE)
		{
			fprintf(stderr,"%s: Unable to seek to aouthdr in a.out file %s.\n",argv[0],argv[optind]);
			exit(1);
		}
		FREAD(&aouthdr,sizeof(AOUTHDR),1,ldfp);
		entrypt = (aouthdr.entry ? aouthdr.entry : aouthdr.text_start);
		exec_arg = 0;
		optind++;	/* increment to next argument */
	}

/*
** get and convert destination cpu and create destination id
*/

	dest_cpu = strtol(argv[optind],&cvterr, 0);

	if (*cvterr != NULL)
	{
		fprintf(stderr,"%s: Unable to convert destination CPU number\n", argv[0]);
		exit(1);
	}

	dst_id = PUT_ID(dest_cpu, CE_IP);


/*
** open and register with the driver
*/
	src_id = 0;

	if ((c = ci_register(dest_cpu,&src_id,&ci_fd)) != 0)
	{
		fprintf(stderr,"%s: unable to open and register with device, errno =%d\n",argv[0],c);
		exit(1);
	}

/*
** if no options specified then set default options
*/
	if (!addr && !load && !exec)  
		load = exec = TRUE;

/*
** Perform the required processing
*/
	if (load)
		do_load(ldfp,ci_fd);

	if (exec || addr)
		do_exec(entrypt, exec_arg, ci_fd);

/*
** close files and deregister
*/
	ldclose(ldfp);

	ci_deregister(src_id,ci_fd);

	exit(0);
}

do_load(ldfp,ci_fd)
LDFILE *ldfp;
int ci_fd;
{
	FILHDR filhdr;
	SCNHDR sechdr;
	int secnum;
	int i;

	if (ldfhread(ldfp,&filhdr) == LDFAILURE)
	{
		fprintf(stderr,"Unable to read a.out file header.\n");
		ci_deregister(src_id,ci_fd);
		exit(1);
	}
	fprintf (stdout," ");
	fflush (stdout);
	for (secnum = 1; secnum <= filhdr.f_nscns; secnum++)
	{
		if (ldshread(ldfp, secnum, &sechdr) == LDFAILURE)
		{
			fprintf(stderr,"Unable to read a.out section header.\n");
			ci_deregister(src_id,ci_fd);
			exit(1);
		}
/*
** Is this section to be excluded
*/
		for (i = 0; i < nex_sect; i++)
		{
			if (strcmp(ex_sect[i],sechdr.s_name) == 0)
				break;
		}

		if (i < nex_sect)
			continue;	/* skip processing of this section */
/*
** section with no data, e.g. .bss or with data e.g. .text
*/
		if (sechdr.s_size && !sechdr.s_scnptr)
			do_sect(FALSE, &sechdr,ldfp,ci_fd);
		else if (sechdr.s_size && sechdr.s_scnptr)
			do_sect(TRUE, &sechdr,ldfp,ci_fd);
	}
/*
** print over same line after all the numbers have been displayed
*/
	fprintf (stdout, "\b ");
	fflush (stdout);
	fprintf (stdout, "\r");
}

do_sect(data, sechdr, ldfp, ci_fd)
unsigned char data;
SCNHDR *sechdr;
LDFILE	*ldfp;
int ci_fd;
{
	static  unsigned char   dbuf[MAX_BUF_SIZE]; /* binary output data */
	unsigned long response = 0;
	BUF 	*buf = (BUF *)dbuf; /* pointer to BUF portion of output data */
	int size;
	int incr;
	unsigned long address;
	int i;
	int rval = TRUE;
	int cnt;

	if (data)
	{
		if (ldnsseek(ldfp,sechdr->s_name) == LDFAILURE)
		{
			fprintf(stderr,"Unable to seek to section data.\n");
			ci_deregister(src_id,ci_fd);
			exit(1);
		}
	}
	else
	{
		for (i = 0; i < MAX_CEDATA_SIZE; i++)
			buf->data[i] = 0;
	}
	
	address = sechdr->s_paddr;
	size = sechdr->s_size;
	buf->next = (BUF *)NULL;

	while (size || response)
	{
		if (rval == TRUE)
		{
			incr = (size < MAX_CEDATA_SIZE ? size : MAX_CEDATA_SIZE );

			if (size)
			{
				if (data)
					FREAD(buf->data, 1, incr, ldfp);
				buf->length = incr;
			}
		}

		while (response)
		{
			cnt++;
			if (get_resp(ci_fd,cnt))
				response--;
			else
				break;
		}

		if (size && ((rval = send(ci_fd, buf, address)) == TRUE))
		{
			response++;
			size -= incr;
			address += incr;
		}
	}
}

send(ci_fd,buf,address)
int ci_fd;
BUF *buf;
unsigned long address;
{
	CE_EPB epb;	/* event parameter block for communication */
	int n;

/*
** format the EPB for download
*/
	epb.hdr.next = 0;
	epb.hdr.mem_id = 0;
	epb.hdr.dst_id = dst_id;
	epb.hdr.src_id = src_id;
	epb.hdr.epb_size = sizeof (CE_EPB);
	epb.hdr.type = CE_LOAD_DATA;
	epb.hdr.status = NO_ERROR;
	epb.hdr.buf = buf;
	epb.hdr.user_def = 0;

	epb.length = buf->length;
	epb.addr = address;

/*
** send the epb
*/
        if ((n = write(ci_fd, &epb, sizeof(CE_EPB))) < 0)
	{
		fprintf(stderr,"Write error, errno = %d\n",errno);
		ci_deregister(src_id, ci_fd);
		exit(1);
	}

	if (n == 0)
		return(FALSE);

	return(TRUE);
}
get_resp(ci_fd,cnt)
int ci_fd;
int cnt;
{
	CE_EPB epb;	/* event parameter block for communication */
	int	n;
	int digit;
	static char wheel[4] ={'|','/','-','\\'};

/*
** get any response waiting to be read
*/
	

	if ((n = read(ci_fd, &epb, sizeof(CE_EPB))) < 0)
	{
		fprintf(stderr,"Read error, errno = %d\n",errno);
		ci_deregister(src_id,ci_fd);
		exit(1);
	}

	if (n == 0)
		return(FALSE);

	if ( epb.hdr.type != CE_LOAD_RESPONSE  || epb.hdr.status != NO_ERROR)
	{
		fprintf(stderr,"Bad response to load data request.\n\ttype = 0x%x, status = 0x%X\n",epb.hdr.type,epb.hdr.status);
		ci_deregister(src_id, ci_fd);
		exit(1);
	}
	else
	{
		if (cnt%10 == 0)
		{
			cnt /= 10;
			fprintf(stdout,"\b%c", wheel[cnt%4]);
			fflush(stdout);
		}
	}
	return(TRUE);
}

do_exec(entrypt, exec_arg, ci_fd)
unsigned long entrypt;
unsigned long exec_arg;
int ci_fd;
{
	CE_EPB epb;
	int n;

	epb.hdr.next = 0;
	epb.hdr.mem_id = 0;
	epb.hdr.dst_id = dst_id;
	epb.hdr.src_id = src_id;
	epb.hdr.epb_size = sizeof(CE_EPB);
	epb.hdr.type = CE_EXECUTE_REQUEST;
	epb.hdr.status = 0;
	epb.hdr.buf = (BUF *)NULL;
	epb.hdr.user_def = 0;

	epb.length = 0;
	epb.addr = entrypt;
	epb.exec_arg = exec_arg;

  	while ((n = write(ci_fd, &epb, sizeof(CE_EPB))) == 0 )
		;

	if (n < 0)
	{
		fprintf(stderr,"Write error, errno = %d\n",errno);
		ci_deregister(src_id,ci_fd);
		exit(1);
	}
/*
** no response is expected from EXECUTE REQUEST
*/
}
