static	char ident[] = "@(#)init.c	35.2		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	init.c
**
**  DESCRIPTION:	Set up the mode the MACSTAT program runs in
**			and sets the signals.
**
**  CONTENTS:		init()		Puts the program in background
**					or leaves it in foreground,
**
**			catchit1()	processes signals,
**
**			catchit2()	processes signals.
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <signal.h>
#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/osi/common/types.h>

extern	BYTE	*getenv();		/* Get environment variables */
extern	DEFAULT	ci_deregister();	/* CE deregistration */

extern	BYTE	*fn;			/* File name */
extern	ULONG	qid;			/* Queue ID */
extern	DEFAULT	dst_cpu;		/* Destination CPU number */
extern	DEFAULT	errno;			/* Error number */
extern	DEFAULT	fd;			/* File descriptor */

	struct 	 termio	 oldterm; /* old terminal status */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	init()
**
**  DESCRIPTION: 	Check the arguments from the command line.
**
**			If the user wants to write stats to a file
**			then fork ourselves and set up to ignore signals
**			else set up terminal for screen output.
**
**  PARAMETERS:		
**
**	Input:		argc	 	- command line argument count
**
**			argv		- command line argument pointer.
**
**	Output:		None.
**
**	Returns:	None.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

VOID
init(argc, argv)

register DEFAULT argc;
register BYTE	 **argv;

{
	VOID	 catchit1();
	VOID	 catchit2();

	struct 	 termio	 newterm; /* new terminal status */

	register DEFAULT p;	  /* kid's process ID */

		 BYTE	 *cvterr; /* conversion error */

	/* 
	** The command line must have at least one argument and possibly
	** an optional file name.  First check for bad args.
	*/

	if ((argc < 2) || (argc > 3))
	{
		printf("\nUsage: %s  [filename] dst_cpu\n", argv[0]);
		exit(1);
	}

	/*
	** Get the destination CPU number. If writing to screen the
	** argument number is 1, else it is 2.
	*/

	if (argc == 2)
		dst_cpu = strtol(argv[1], &cvterr, 0);
	else
	{
		fn = argv[1];

		dst_cpu = strtol(argv[2], &cvterr, 0);
	}

	if ((dst_cpu < CPU_OFFSET) || (dst_cpu > NUM_CPU_ENTRIES) || (*cvterr != NULL))
	{
		fprintf(stdout, "\nerror:  Invalid Destination CPU: dst_cpu = %d\n", dst_cpu);
		exit(1);
	}

	/*
	** If writing to a file, fork ourselves.
	*/

	if (argc == 3)
		if (p = fork())
		{
			if (p < 0)
			{
				perror("can't fork");
				exit(1);
			}

			exit(0);
		}

	/*
	** Set up to catch or ignore signals.
	** If argc = 2, then no file name was present and the user wants
	** to write to the screen.  Otherwise, a file name was included 
	** for the recording of the statistics.
	*/

	if (argc == 2)	
	{
		setupterm(getenv("TERM"), 1, 0);

		/*
		** Set up terminal for single character input.
		*/

		ioctl(0, TCGETA, &newterm);
		ioctl(0, TCGETA, &oldterm);

		newterm.c_iflag &= (~IGNBRK & ~ICRNL & ~INLCR);
		newterm.c_iflag |= BRKINT;
		newterm.c_lflag &= (~ISIG & ~ECHO & ~ECHONL & ~ICANON);
		newterm.c_cc[4]  = 1;
		newterm.c_cc[5]  = 1;

		ioctl(0, TCSETAF, &newterm);

		/*
		** If the user wants the stats on the screen,
		** set up a signal catching function so we can reset
		** the terminal when the user hits the BREAK key.
		*/

		signal(SIGHUP,  catchit1);
		signal(SIGINT,  catchit1);
		signal(SIGQUIT, catchit1);
		signal(SIGTERM, catchit1);
	}
	else
	{
		/*
		** When writing to a file, ignore all termination
		** signals. The only way to kill us is to find out the
		** pid and then do a "kill <pid>".
		*/

		signal(SIGINT,  SIG_IGN);
		signal(SIGQUIT, SIG_IGN);
		signal(SIGHUP,  catchit2);
		signal(SIGTERM, catchit2);
	}
 }

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	catchit1()
**
**  DESCRIPTION:	This function cleans up and exits.
**			Stats are going to the screen so reset the
**			terminal.
**
**  PARAMETERS:		
**
**	Input:		None.
**
**	Output:		None.
**
**	Returns:	None.
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

VOID
catchit1()

{
	register DEFAULT n;

	ioctl (0, TCSETAF, &oldterm);

	if ((n = ci_deregister(qid, fd)) < 0)
	{
		printf("\nMACSTAT deregistration failure, err = %d\n", errno);

		exit(1);
	}

	exit(0);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	catchit2()
**
**  DESCRIPTION:	This function cleans up and exits.
**			Process was forked to catch the termination
**			signal and kill self.
**  PARAMETERS:		
**
**	Input:		None.
**
**	Output:		None.
**
**	Returns:	None.
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

VOID
catchit2()

{
	register DEFAULT n;

	if ((n = ci_deregister(qid, fd)) < 0)
	{
		printf("\nMACSTAT deregistration failure, err = %d\n", errno);

		exit(1);
	}

	exit(0);
}
