static	char ident[] = "@(#)main.c	35.2		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	main.c
**
**  DESCRIPTION:	This is the main procedure to drive the Station
**			Monitor program which periodically gets MAC
**			statistics and prints them to the screen or
**			stores them in a file.
**
**  CONTENTS:		main()		- Main MACSTAT program,
**
**			errexit()	- error exit function.
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/bdb.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/ma_epb.h>
#include <sys/osi/common/sm_epb.h>

#define	RD_SIZE	(MAX_CEEPB_SIZE + MAX_BUF_SIZE)
#define	BLANK	" "

extern	DEFAULT	ci_register();
extern	DEFAULT	ci_deregister();

extern	VOID	init();
extern	VOID	print();

extern	DEFAULT	errno;

/*
**	 Global Variables
*/

BYTE	*fn	= 0;		/* File name address */
ULONG	qid	= 0;		/* Queue ID */
DEFAULT	fd	= 0;		/* File descriptor */
DEFAULT	dst_cpu	= 0;		/* Destination CPU number */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	main()
**
**  DESCRIPTION:	This is the main procedure to drive the Station
**			Monitor program which periodically gets MAC
**			statistics and prints them to the screen or
**			stores them in a file.
**
**  			Check the arguments from the command line.
**
**			If the user wants to write stats to a file
**			then fork ourselves and set up to ignore signals
**			else set up terminal for screen output.
**
**  PARAMETERS:		
**
**	Input:		argc	 	- command line argument count
**
**			argv		- command line argument pointer.
**
**	Output:		None.
**
**	Returns:	None.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
main(argc, argv)

DEFAULT	  argc;
BYTE	**argv;

{
	VOID	 errexit();

	register SM_EPB	 *sm_epb;	/* SM EPB pointer */
	register DEFAULT n;		/* Returned status */
	register FILE	 *f;		/* File pointer */

		 SM_EPB	 stat_epb;	/* SM EPB */

	/*
	** Initialize:
	**	Put the program into the background if it writes to a 
	**	file or the foreground if it writes to the screen.
	**	Set signals according to whether we are writing to a 
	**	file or the screen.
	*/

	init(argc, argv);

	/*
	** If user wants statisics on the screen, clear the screen.
	*/

	if (fn == 0)
		putp(tparm(clear_screen));

	/*
	** Register with the common environment.
	*/

	if ((n = ci_register(dst_cpu, &qid, &fd)) != 0)
	{
		if (fn == 0)
			resetterm();

		printf("\nMACSTAT registration failure, err = %d\n", errno);
		exit(1);
	}

	/*
	** Set us up to WAIT on the response queue.
	*/

	if (ioctl(fd, WAIT, 0) < 0)
		errexit("\nMACSTAT unable to set IOCTL to blocked reads, errno = %d\n", errno);

	/*
	** Begin while loop to get MAC statistics periodically.
	*/

	sm_epb = &stat_epb;

	while (FOREVER)
	{
		/*
		**	Initialize the EPB.
		*/

		sm_epb->hdr.next	= (EPB *)NULL;
		sm_epb->hdr.mem_id	= 0;
		sm_epb->hdr.src_id	= qid;
		sm_epb->hdr.dst_id	= PUT_ID(dst_cpu, MAC_QID);
		sm_epb->hdr.epb_size	= sizeof(SM_EPB);
		sm_epb->hdr.type	= MA_PARAM_REQUEST;
		sm_epb->hdr.status	= 0;
		sm_epb->hdr.buf		= (BUF *)NULL;
		sm_epb->hdr.user_def	= 0;

		sm_epb->mac_sel		= MAC_RD_STATS;

		/*
		**	Write the EPB to the MAC layer.
		*/

		while ((n = write(fd, sm_epb, sizeof(SM_EPB))) <= 0)
			if (n < 0)
				errexit("\nMACSTAT EPB write failure, err = %d\n", errno);

		/*
		** Get response back from MAC.
		*/

		if ((n = read(fd, sm_epb, (ULONG)(RD_SIZE))) < 0)
			errexit("\nMACSTAT EPB read failure err = %d\n", errno);

		/*
		** Check EPB type and status fields.
		*/

 		if (sm_epb->hdr.status != 0)
			errexit("\nMACSTAT bad EPB status returned, status = 0x%x\n", sm_epb->hdr.status);
		else
			if (sm_epb->hdr.type != MA_PARAM_CONFIRM)
				errexit("\nMACSTAT bad EPB type returned, type = 0x%x\n", sm_epb->hdr.type);

		/*
		** Print the MAC statistics to the screen or the file.
		** If no file name was in the command line, then the
		** user wants to write to the screen.
		*/

		if (fn == 0)
		{
			putp(tparm(cursor_home));
			print(stdout, sm_epb);

			/* 
			** Tell the user how to terminate it.
			*/

			fprintf(stdout, "%23sHit BREAK key to exit from display\r", BLANK);
		}
		else
			if ((f = fopen(fn, "w")) != NULL)
			{
				print(f, sm_epb);
				fclose(f);
			}
			else
				errexit("\nMACSTAT cannot open file %s\n", fn);

		/*
		** Take a short nap and then do it all over again.
		*/

		sleep(5);
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	errexit()
**
**  DESCRIPTION:	This is the error exit function. It will
**			display the passed message with the parameter
**			and exit.
**
**  PARAMETERS:		
**
**	Input:		mess	 	- message address,
**
**			arg		- message argument.
**
**	Output:		None.
**
**	Returns:	None.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

VOID
errexit(mess, arg)

register BYTE  *mess;
register ULONG arg;

{
	register DEFAULT n;		/* Returned status */

	if (fn == 0)
		resetterm();

	printf(mess, arg);

	if ((n = ci_deregister(qid, fd)) != 0)
		printf("\nMACSTAT deregistration failure, err = %d\n", errno);

	exit(1);
}
