static	char ident[] = "@(#)cnfg.c	35.3		5/25/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	cnfg.c
**
**  DESCRIPTION:	This module contains the functions that process
**			Configuration Utilities functions.
**
**  CONTENTS:		do_cnfg_sel()	 switch on user's selection
**			do_system_sel()	 switch on user's selection
**			do_xport_sel()	 switch on user's selection
**			do_lance_sel()	 switch on user's selection
**			do_tbc_sel()	 switch on user's selection
**			do_bdb_count()	 display BDB count
**			do_user_limit()  display/modifiy user limit
**			do_mgmt_limit()  display/modifiy management limit
**			do_netw_limit()  display/modifiy network limit
**			do_max_connects() display/modify max # of connections
**			do_min_credit() display/modify minimum credit
**			do_max_credit() display/modify maximum credit
**			do_max_retries() display/modify max # of retries
**			do_retry_ticks() display/modify data retry ticks 
**			do_exp_retry_ticks() display/modify expedited data retry ticks 
**			do_window_timer() display/modify window timer
**			do_ref_wait_timer() display/modify reference wait timer
**			do_ring_size()   display LANCE's ring sizes
**			do_lance_mode()  display/modify LANCE's mode
**			do_station_add() display/modify LANCE's station address
**			do_group_add()   display/modify LANCE's group address
**			do_tbc_pri()	 display/modify TBC's xmit priority
**
**  SYSTEM GLOBALS:	selection	user's menu selection
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/bdb.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>
#include <sys/osi/common/ma_epb.h>
#include "def.h"

extern	VOID	 clr();
extern	VOID	 get_key();
extern	VOID	 ex_shell_cmd();
extern	VOID	 wr_invalid_sel();
extern	DEFAULT	 bld_epb();
extern	DEFAULT	 catchit();
extern	DEFAULT	 str_to_hex();
extern	DEFAULT	 wait_sm_response();

extern	BYTE	 selection[80];
extern	USHORT	 board;
extern	BOOLEAN	 scold_user;

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_cnfg_sel()
**
**  DESCRIPTION:	This function switches on the user's selection
**			from the Configuration Utilities menu and calls
**			the proper routine to implement the selection.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_cnfg_sel()

{
	extern	BOOLEAN	read_selection();

	extern	VOID	wr_xport_menu(),
			wr_banner(),
			wr_tbc_menu(),
			wr_lance_menu(),
			wr_system_menu(),
			do_nvram();

		VOID	do_tbc_sel(),
			do_lance_sel(),
			do_xport_sel(),
			do_system_sel();

	register SM_EPB	 *sm_epb;	/* SM EPB pointer */

		 SM_EPB	 stat_epb;	/* SM EPB */

	switch (selection[0])
	{
		case NVRAM:
		{
			do_nvram();
			break;
		}
	 
		case MAC_PARMS:
		{
			do
			{
				wr_banner();

				fprintf(stdout, "%25sDisplay/Modify MAC Layer Menu\n\n", BLANK);

				switch (board)
				{
					case 372:
					{
						wr_tbc_menu();

						if (read_selection(TBC_PARMS_MENU) != FAILURE)
							do_tbc_sel(&stat_epb);

						break;
					}

					case 147:
					case 374:
					{
						wr_lance_menu();

						if (read_selection(LANCE_PARMS_MENU) != FAILURE)
							do_lance_sel(&stat_epb);

						break;
					}
				}
			}
			while ((selection[0] != CONFIGURATION_UTILITIES) && (selection[0] != MAIN));

			break;
		}
	 
		case SYSTEM_PARMS:
		{
			do
			{
				wr_system_menu();

				if (read_selection(SYSTEM_PARMS_MENU) != FAILURE)
					do_system_sel();
			}
			while ((selection[0] != CONFIGURATION_UTILITIES) && (selection[0] != MAIN));

			break;
		}

		case TRANSPORT_PARMS:
		{
			do
			{
				wr_xport_menu();

				if (read_selection(XPORT_PARMS_MENU) != FAILURE)
					do_xport_sel();
			}
			while ((selection[0] != CONFIGURATION_UTILITIES) && (selection[0] != MAIN));

			break;
		}

		case QUIT:
		{
			catchit();
			break;
		}

		case SHELL:
		{
			ex_shell_cmd();
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_system_sel()
**
**  DESCRIPTION:	This function switches on the user's system selection
**			and calls the proper routine to process the selection.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**
**  METHOD:
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**
**  NOTES:
**
**********************************************************************/

VOID
do_system_sel()

{
	VOID	do_bdb_count(),
		do_user_limit(),
		do_mgmt_limit(),
		do_netw_limit();

	switch (selection[0])
	{
		case BDB_COUNT:
		{
			do_bdb_count();
			break;
		}

		case USER_LIMIT:
		{
			do_user_limit();
			break;
		}

		case MGMT_LIMIT:
		{
			do_mgmt_limit();
			break;
		}

		case NETW_LIMIT:
		{
			do_netw_limit();
			break;
		}

		case QUIT:
		{
			catchit();
			break;
		}

		case SHELL:
		{
			ex_shell_cmd();
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_xport_sel()
**
**  DESCRIPTION:	This function switches on the user's transport selection
**			and calls the proper routine to process the selection.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**
**  METHOD:
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**
**  NOTES:
**
**********************************************************************/

VOID
do_xport_sel()

{
	VOID	do_max_connects();
	VOID	do_min_credit();
	VOID	do_max_credit();
	VOID	do_max_retries();
	VOID	do_retry_ticks();
	VOID	do_exp_retry_ticks();
	VOID	do_window_timer();
	VOID	do_ref_wait_timer();

	switch (selection[0])
	{
		case MAX_CONNECTS:
		{
			do_max_connects();
			break;
		}

		case MIN_CREDIT:
		{
			do_min_credit();
			break;
		}

		case MAX_CREDIT:
		{
			do_max_credit();
			break;
		}

		case MAX_RETRIES:
		{
			do_max_retries();
			break;
		}

		case RETRY_TICKS:
		{
			do_retry_ticks();
			break;
		}

		case RETRY_TICKS_EXP:
		{
			do_exp_retry_ticks();
			break;
		}

		case WINDOW_TICKS:
		{
			do_window_timer();
			break;
		}

		case REF_WAIT_TICKS:
		{
			do_ref_wait_timer();
			break;
		}

		case QUIT:
		{
			catchit();
			break;
		}

		case SHELL:
		{
			ex_shell_cmd();
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_lance_sel()
**
**  DESCRIPTION:	This function switches on the user's LANCE selection
**			and calls the proper routine to process the selection.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**
**  METHOD:
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**
**  NOTES:
**
**********************************************************************/

VOID
do_lance_sel(sm_epb)

register SM_EPB *sm_epb;
{
	VOID	do_ring_size(),
		do_lance_mode(),
		do_station_add(),
		do_group_add();

	switch (selection[0])
	{
		case RING_SIZE:
		{
			do_ring_size(sm_epb);
			break;
		}

		case LANCE_MODE:
		{
			do_lance_mode(sm_epb);
			break;
		}

		case STATION_ADDRESS:
		{
			do_station_add(sm_epb);
			break;
		}

		case GROUP_ADDRESS:
		{
			do_group_add(sm_epb);
			break;
		}

		case QUIT:
		{
			catchit();
			break;
		}

		case SHELL:
		{
			ex_shell_cmd();
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_tbc_sel()
**
**  DESCRIPTION:	This function switches on the user's TBC selection
**			and calls the proper routine to process the selection.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**
**  METHOD:
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**
**  NOTES:
**
**********************************************************************/

VOID
do_tbc_sel(sm_epb)

register SM_EPB	*sm_epb;
{
	VOID	do_tbc_pri();

	switch (selection[0])
	{
		case TBC_PRIORITY:
		{
			do_tbc_pri(sm_epb);
			break;
		}

		case QUIT:
		{
			catchit();
			break;
		}

		case SHELL:
		{
			ex_shell_cmd();
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_bdb_count()
**
**  DESCRIPTION:	This function displays the current BDB count.
**			This value helps the user to select the user,
**			management, and network limits.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_bdb_count()

{
	register SM_EPB  *sm_epb;	/* pointer to EPB */

		 SM_EPB	 stat_epb;	/* Station Manager EPB */

	/*
	**	Get the current BDB count.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_BDB_COUNT_RD_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		**	Wait for Station Manager on board to send
		**	back response.
		*/

		if (wait_sm_response(sm_epb, SM_BDB_COUNT_RD_CONFIRM) == SUCCESS)
		{
			fprintf(stdout, "\n\n%25sCurrent BDB count:  %d\n", BLANK, sm_epb->sys.bdb_count);

			get_key();
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_user_limit()
**
**  DESCRIPTION:	This function displays the current user limit
**			value, gets the new user limit from the user
**			and sends an EPB to the Station Manager on the
**			board to change the user limit value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_user_limit()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register ULONG	user_limit;	/* user limit variable	*/

		 SM_EPB	stat_epb;	/* Station Mangager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current system user limit.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_USER_LIM_RD_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		**	Wait for Station Manager on board to send back response.
		*/

		if (wait_sm_response(sm_epb, SM_USER_LIM_RD_CONFIRM) == SUCCESS)
		{
			fprintf(stdout, "\n\n%25sCurrent system user limit:  %d\n\n", BLANK, sm_epb->sys.o_user_limit);

			/*
			**	Initialize value and input error Boolean, then 
			**	loop until user enters valid response or
			**	carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter user limit:  ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered, assume 
				**	user changed mind.
				*/

				if (buf[0] != NULL)
				{
					user_limit = strtol(buf, &cvterr, 0);

					if (*cvterr != NULL)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put user limit into an EPB and send it to
			**	Station Manager on board to put value into
			**	system user limit variable on board.
			*/

			sm_epb->sys.o_user_limit = user_limit;

			if (bld_epb(sm_epb, SM_USER_LIM_SET_REQUEST, SM_QID) == SUCCESS)
			{
				fprintf(stdout, "\n%25sSystem user limit set to %d.\n", BLANK, user_limit);

				get_key();
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_mgmt_limit()
**
**  DESCRIPTION:	This function displays the current management 
**			limit value, gets the new user limit from the 
**			user and sends an EPB to the Station Manager
**			on the board to change the management limit
**			value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_mgmt_limit()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register ULONG	mgmt_limit;	/* management limit variable */

		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current system management limit.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_MGMT_LIM_RD_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		**	Wait for Station Manager on board to send back response.
		*/

		if (wait_sm_response(sm_epb, SM_MGMT_LIM_RD_CONFIRM) == SUCCESS)
		{
			fprintf(stdout, "\n\n%25sCurrent system management limit:  %d\n\n", BLANK, sm_epb->sys.o_mgmt_limit);

			/*
			**	Initialize value and input error Boolean, then 
			**	loop until user enters valid response or
			**	carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter management limit:  ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered, assume 
				**	user changed mind.
				*/

				if (buf[0] != NULL)
				{
					mgmt_limit = strtol(buf, &cvterr, 0);

					if (*cvterr != NULL)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put management limit into an EPB and send
			**	it to Station Manager on board to put value into
			**	system management limit variable on board.
			*/

			sm_epb->sys.o_mgmt_limit = mgmt_limit;

			if (bld_epb(sm_epb, SM_MGMT_LIM_SET_REQUEST, SM_QID) == SUCCESS)
			{
				fprintf(stdout, "\n%25sSystem management limit set to %d.\n", BLANK, mgmt_limit);

				get_key();
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_netw_limit()
**
**  DESCRIPTION:	This function displays the current network 
**			limit value, gets the new user limit from the 
**			user and sends an EPB to the Station Manager
**			on the board to change the network limit
**			value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_netw_limit()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register ULONG	netw_limit;	/* network limit variable */

		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current system network limit.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_NETW_LIM_RD_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		**	Wait for Station Manager on board to send back response.
		*/

		if (wait_sm_response(sm_epb, SM_NETW_LIM_RD_CONFIRM) == SUCCESS)
		{
			fprintf(stdout,"\n\n%25sCurrent system network limit:  %d\n\n", BLANK, sm_epb->sys.o_netw_limit);

			/*
			**	Initialize value and input error Boolean, then 
			**	loop until user enters valid response or
			**	carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter network limit:  ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered, assume 
				**	user changed mind.
				*/

				if (buf[0] != NULL)
				{
					netw_limit = strtol(buf, &cvterr, 0);

					if (*cvterr != NULL)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put network limit into an EPB and send it 
			**	to Station Manager on board to put value into 
			**	system network limit variable on board.
			*/

			sm_epb->sys.o_netw_limit = netw_limit;

			if (bld_epb(sm_epb, SM_NETW_LIM_SET_REQUEST, SM_QID) == SUCCESS)
			{
				fprintf(stdout, "\n%25sSystem network limit set to %d.\n", BLANK, netw_limit);

				get_key();
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_max_connects()
**
**  DESCRIPTION:	This function displays the current maximum
**			number of connections supported by the node,
**			gets the new maximum number of connections
**			from the user and sends an EPB to the
**			Station Manager on the board to change the
**			transport maximum number of connections value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_max_connects()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register SHORT	max_connects;	/* maximum # of connections */
		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport maximum number of connections.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_MAX_CONNECTS_RD_REQUEST, SM_QID) == FAILURE)
		return;

	/*
	**	Wait for Station Manager on board to send back response.
	*/

	if (wait_sm_response(sm_epb, SM_MAX_CONNECTS_RD_CONFIRM) == FAILURE)
		return;

	fprintf(stdout,"\n\n%25sCurrent maximum number of connections:  %d\n\n",
			BLANK, sm_epb->sys.t_max_connects);

	/*
	**	Initialize value and input error Boolean, then loop
	**	until user enters valid response or carriage return.
	*/

	do
	{
		scold_user = FALSE;

		fprintf(stdout, "%25sEnter maximum number of connections:  ",
				BLANK);

		gets(buf);

		/*
		**	If carriage return is entered, assume user
		**	changed mind.
		*/

		if (buf[0] != NULL)
		{
			max_connects = strtol(buf, &cvterr, 0);

			if ((*cvterr != NULL) || (max_connects <= 0))
				wr_invalid_sel();
		}
		else
			return;
	}
	while (scold_user == TRUE);

	/*
	**	Put maximum credit into into an EPB and send it to 
	**	Station Manager on board to put value into system
	**	maximum credit variable on board.
	*/

	sm_epb->sys.t_max_connects = (USHORT) max_connects;

	if (bld_epb(sm_epb, SM_MAX_CONNECTS_SET_REQUEST, SM_QID) == SUCCESS)
	{
		fprintf(stdout, "\n%25sTransport maximum number of connections set to %d.\n",
				BLANK, max_connects);

		get_key();
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_min_credit()
**
**  DESCRIPTION:	This function displays the current transport
**			minimum credit value, gets the new minimum
**			credit from the user and sends an EPB to the
**			Station Manager on the board to change the
**			transport minimum credit value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_min_credit()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register ULONG	min_credit;	/* minimum credit variable */

		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport minimum credit.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_MIN_CRED_RD_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		**	Wait for Station Manager on board to send back response.
		*/

		if (wait_sm_response(sm_epb, SM_MIN_CRED_RD_CONFIRM) == SUCCESS)
		{
			fprintf(stdout,"\n\n%25sCurrent minimum credit:  %d\n\n", BLANK, sm_epb->sys.t_min_credit);

			/*
			**	Initialize value and input error Boolean, then 
			**	loop until user enters valid response or
			**	carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter minimum credit:  ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered, assume 
				**	user changed mind.
				*/

				if (buf[0] != NULL)
				{
					min_credit = strtol(buf, &cvterr, 0);

					if (*cvterr != NULL)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put minimum credit into an EPB and send it
			**	to Station Manager on board to put value into 
			**	system minimum credit variable on board.
			*/

			sm_epb->sys.t_min_credit = (UBYTE)min_credit;

			if (bld_epb(sm_epb, SM_MIN_CRED_SET_REQUEST, SM_QID) == SUCCESS)
			{
				fprintf(stdout, "\n%25sTransport minimum credit set to %d.\n", BLANK, min_credit);

				get_key();
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_max_credit()
**
**  DESCRIPTION:	This function displays the current transport
**			maximum credit value, gets the new maximum
**			credit from the user and sends an EPB to the
**			Station Manager on the board to change the
**			transport maximum credit value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_max_credit()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register ULONG	max_credit;	/* maximum credit variable */

		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport maximum credit.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_MAX_CRED_RD_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		**	Wait for Station Manager on board to send back response.
		*/

		if (wait_sm_response(sm_epb, SM_MAX_CRED_RD_CONFIRM) == SUCCESS)
		{
			fprintf(stdout,"\n\n%25sCurrent maximum credit:  %d\n\n", BLANK, sm_epb->sys.t_max_credit);

			/*
			**	Initialize value and input error Boolean, then 
			**	loop until user enters valid response or
			**	carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter maximum credit:  ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered, assume 
				**	user changed mind.
				*/

				if (buf[0] != NULL)
				{
					max_credit = strtol(buf, &cvterr, 0);

					if (*cvterr != NULL)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put maximum credit into an EPB and send it
			**	to Station Manager on board to put value into 
			**	system maximum credit variable on board.
			*/

			sm_epb->sys.t_max_credit = (UBYTE)max_credit;

			if (bld_epb(sm_epb, SM_MAX_CRED_SET_REQUEST, SM_QID) == SUCCESS)
			{
				fprintf(stdout, "\n%25sTransport maximum credit set to %d.\n", BLANK, max_credit);

				get_key();
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_max_retries()
**
**  DESCRIPTION:	This function displays the current maximum
**			number of retries done by transport,
**			gets the new maximum number of retries
**			from the user and sends an EPB to the
**			Station Manager on the board to change the
**			transport maximum number of retries value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_max_retries()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register SHORT	max_retries;	/* maximum # of retries */
		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport maximum number of retries.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_MAX_RETRIES_RD_REQUEST, SM_QID) == FAILURE)
		return;

	/*
	**	Wait for Station Manager on board to send back response.
	*/

	if (wait_sm_response(sm_epb, SM_MAX_RETRIES_RD_CONFIRM) == FAILURE)
		return;

	fprintf(stdout,"\n\n%25sCurrent maximum number of retries:  %d\n\n",
			BLANK, sm_epb->sys.t_max_retries);

	/*
	**	Initialize value and input error Boolean, then loop
	**	until user enters valid response or carriage return.
	*/

	do
	{
		scold_user = FALSE;

		fprintf(stdout, "%25sEnter maximum number of retries:  ",
				BLANK);

		gets(buf);

		/*
		**	If carriage return is entered, assume user
		**	changed mind.
		*/

		if (buf[0] != NULL)
		{
			max_retries = strtol(buf, &cvterr, 0);

			if ((*cvterr != NULL) || (max_retries <= 0))
				wr_invalid_sel();
		}
		else
			return;
	}
	while (scold_user == TRUE);

	/*
	**	Put maximum retries into into an EPB and send it to 
	**	Station Manager on board to put value into system
	**	maximum retries variable on board.
	*/

	sm_epb->sys.t_max_retries = (USHORT) max_retries;

	if (bld_epb(sm_epb, SM_MAX_RETRIES_SET_REQUEST, SM_QID) == SUCCESS)
	{
		fprintf(stdout, "\n%25sTransport maximum number of retries set to %d.\n",
				BLANK, max_retries);

		get_key();
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_retry_ticks()
**
**  DESCRIPTION:	This function displays the current value of
**			retries ticks, gets the new value of retries ticks
**			from the user and sends an EPB to the
**			Station Manager on the board to change the
**			transport retry ticks value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_retry_ticks()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register SHORT	retry_timer;	/* number of seconds to retry */
		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport value of retry ticks.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_RETRY_TICKS_RD_REQUEST, SM_QID) == FAILURE)
		return;

	/*
	**	Wait for Station Manager on board to send back response.
	**	Convert response to seconds for display.
	*/

	if (wait_sm_response(sm_epb, SM_RETRY_TICKS_RD_CONFIRM) == FAILURE)
		return;
	retry_timer = (sm_epb->sys.t_retry_ticks) * MSPT / 1000;
	fprintf(stdout,"\n\n%25sCurrent retry timer (in seconds):  %d\n\n",
			BLANK, retry_timer);

	/*
	**	Initialize value and input error Boolean, then loop
	**	until user enters valid response or carriage return.
	*/

	do
	{
		scold_user = FALSE;

		fprintf(stdout, "%25sEnter seconds for retry timer:  ",
				BLANK);

		gets(buf);

		/*
		**	If carriage return is entered, assume user
		**	changed mind.
		*/

		if (buf[0] != NULL)
		{
			retry_timer = strtol(buf, &cvterr, 0);

			if ((*cvterr != NULL) || (retry_timer <= 0))
				wr_invalid_sel();
		}
		else
			return;
	}
	while (scold_user == TRUE);

	/*
	**	Convert the seconds into ticks, and
	**	put retry ticks into an EPB and send it to 
	**	Station Manager on board to put value into system
	**	retry ticks variable on board.
	*/

	sm_epb->sys.t_max_retries = (USHORT) (retry_timer * 1000 / MSPT );

	if (bld_epb(sm_epb, SM_RETRY_TICKS_SET_REQUEST, SM_QID) == SUCCESS)
	{
		fprintf(stdout, "\n%25sTransport retry timer set to %d seconds.\n",
				BLANK, retry_timer);

		get_key();
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_exp_retry_ticks()
**
**  DESCRIPTION:	This function displays the current value of
**			retries ticks for expedited data transfer,
**		 	gets the new value from the user and sends
**			an EPB to the Station Manager on the board to
**			change the transport retry ticks value for expedited
**			data transfer.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_exp_retry_ticks()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register SHORT	exp_retry_timer;/* number of seconds to retry for expediited*/
		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport value of expedited retry ticks.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_EXP_RETRY_TICKS_RD_REQUEST, SM_QID) == FAILURE)
		return;

	/*
	**	Wait for Station Manager on board to send back response.
	**	Convert response to seconds for display.
	*/

	if (wait_sm_response(sm_epb, SM_EXP_RETRY_TICKS_RD_CONFIRM) == FAILURE)
		return;
	exp_retry_timer = (sm_epb->sys.t_exp_retry_ticks) * MSPT / 1000;
	fprintf(stdout,"\n\n%25sCurrent expedited data retry timer (in seconds):  %d\n\n",
			BLANK, exp_retry_timer);

	/*
	**	Initialize value and input error Boolean, then loop
	**	until user enters valid response or carriage return.
	*/

	do
	{
		scold_user = FALSE;

		fprintf(stdout, "%25sEnter seconds for expedited data retry timer:  ",
				BLANK);

		gets(buf);

		/*
		**	If carriage return is entered, assume user
		**	changed mind.
		*/

		if (buf[0] != NULL)
		{
			exp_retry_timer = strtol(buf, &cvterr, 0);

			if ((*cvterr != NULL) || (exp_retry_timer <= 0))
				wr_invalid_sel();
		}
		else
			return;
	}
	while (scold_user == TRUE);

	/*
	**	Convert the seconds into ticks, and
	**	put expedited retry ticks into an EPB and send it to 
	**	Station Manager on board to put value into system
	**	expedited retry ticks variable on board.
	*/

	sm_epb->sys.t_retry_ticks = (USHORT) (exp_retry_timer * 1000 / MSPT);

	if (bld_epb(sm_epb, SM_EXP_RETRY_TICKS_SET_REQUEST, SM_QID) == SUCCESS)
	{
		fprintf(stdout, "\n%25sTransport expedited data retry timer set to %d seconds.\n",
				BLANK, exp_retry_timer);

		get_key();
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_window_timer()
**
**  DESCRIPTION:	This function displays the current value of
**			window timer in seconds, gets the new value
**			from the user and sends an EPB to the Station
**			Manager on the board to change the transport
**			window ticks value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_window_timer()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register SHORT	window_timer;	/* number of seconds for window*/
		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport value of window ticks.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_WINDOW_TICKS_RD_REQUEST, SM_QID) == FAILURE)
		return;

	/*
	**	Wait for Station Manager on board to send back response.
	**	Convert response to seconds for display.
	*/

	if (wait_sm_response(sm_epb, SM_WINDOW_TICKS_RD_CONFIRM) == FAILURE)
		return;
	window_timer = (sm_epb->sys.t_window_ticks) * MSPT / 1000;
	fprintf(stdout,"\n\n%25sCurrent window timer (in seconds):  %d\n\n",
			BLANK, window_timer);

	/*
	**	Initialize value and input error Boolean, then loop
	**	until user enters valid response or carriage return.
	*/

	do
	{
		scold_user = FALSE;

		fprintf(stdout, "%25sEnter seconds for window timer:  ",
				BLANK);

		gets(buf);

		/*
		**	If carriage return is entered, assume user
		**	changed mind.
		*/

		if (buf[0] != NULL)
		{
			window_timer = strtol(buf, &cvterr, 0);

			if ((*cvterr != NULL) || (window_timer <= 0))
				wr_invalid_sel();
		}
		else
			return;
	}
	while (scold_user == TRUE);

	/*
	**	Convert the seconds into ticks, and
	**	put window ticks into an EPB and send it to 
	**	Station Manager on board to put value into system
	**	window ticks variable on board.
	*/

	sm_epb->sys.t_window_ticks = (USHORT) (window_timer * 1000 / MSPT);

	if (bld_epb(sm_epb, SM_WINDOW_TICKS_SET_REQUEST, SM_QID) == SUCCESS)
	{
		fprintf(stdout, "\n%25sTransport window timer set to %d seconds.\n",
				BLANK, window_timer);

		get_key();
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_ref_wait_timer()
**
**  DESCRIPTION:	This function displays the current value of
**			reference wait timer in seconds, gets the new value
**			from the user and sends an EPB to the Station
**			Manager on the board to change the transport
**			reference wait ticks value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_ref_wait_timer()

{
	register SM_EPB	*sm_epb;	/* pointer to EPB */
	register SHORT	ref_wait_timer;	/* number of seconds for reference wait*/
		 SM_EPB	stat_epb;	/* Station Manager EPB */
		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current transport value of reference wait ticks.
	*/

	if (bld_epb((sm_epb = &stat_epb), SM_REF_WAIT_TICKS_RD_REQUEST, SM_QID) == FAILURE)
		return;

	/*
	**	Wait for Station Manager on board to send back response.
	**	Convert response to seconds for display.
	*/

	if (wait_sm_response(sm_epb, SM_REF_WAIT_TICKS_RD_CONFIRM) == FAILURE)
		return;
	ref_wait_timer = (sm_epb->sys.t_ref_wait_ticks) * MSPT / 1000;
	fprintf(stdout,"\n\n%25sCurrent reference wait timer (in seconds):  %d\n\n",
			BLANK, ref_wait_timer);

	/*
	**	Initialize value and input error Boolean, then loop
	**	until user enters valid response or carriage return.
	*/

	do
	{
		scold_user = FALSE;

		fprintf(stdout, "%25sEnter seconds for reference wait timer:  ",
				BLANK);

		gets(buf);

		/*
		**	If carriage return is entered, assume user
		**	changed mind.
		*/

		if (buf[0] != NULL)
		{
			ref_wait_timer = strtol(buf, &cvterr, 0);

			if ((*cvterr != NULL) || (ref_wait_timer <= 0))
				wr_invalid_sel();
		}
		else
			return;
	}
	while (scold_user == TRUE);

	/*
	**	Convert the seconds into ticks, and
	**	put reference wait ticks into an EPB and send it to 
	**	Station Manager on board to put value into system
	**	reference wait ticks variable on board.
	*/

	sm_epb->sys.t_ref_wait_ticks = (USHORT) (ref_wait_timer * 1000 / MSPT);

	if (bld_epb(sm_epb, SM_REF_WAIT_TICKS_SET_REQUEST, SM_QID) == SUCCESS)
	{
		fprintf(stdout, "\n%25sTransport reference wait timer set to %d seconds.\n",
				BLANK, ref_wait_timer);

		get_key();
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_ring_size()
**
**  DESCRIPTION:	This function displays the LANCE's current 
**			transmit and receive ring sizes.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_ring_size(sm_epb)

register SM_EPB  *sm_epb;	/* pointer to EPB */
{
	/*
	**	Get the current ring sizes.
	*/

	sm_epb->mac_sel = MAC_RD_RING;

	if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
	{
		/*
		**	Wait for MAC sublayer on board to send
		**	back response.
		*/

		if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
		{
			fprintf(stdout, "\n\n%25sCurrent Transmit Ring size:  %d\n", BLANK, (sm_epb->mac.ring >> 8));
			fprintf(stdout, "%25sCurrent Receive  Ring size:  %d\n", BLANK, (sm_epb->mac.ring & 0xFF));

			get_key();
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_lance_mode()
**
**  DESCRIPTION:	This function displays the LANCE's current mode
**			word value, gets the new mode word from the
**			user and sends an EPB to the MAC on the board to
**			change the mode word value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_lance_mode(sm_epb)

register SM_EPB	*sm_epb;	/* pointer to EPB */
{
	register ULONG	mode;		/* mode word variable */

		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
	
	/*
	**	Get the current mode word.
	*/

	sm_epb->mac_sel = MAC_RD_MODE;

	if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
	{
		/*
		**	Wait for MAC sublayer on board to send back
		**	response.
		*/

		if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
		{
			fprintf(stdout,"\n\n%25sCurrent mode word:  %04x\n", BLANK, sm_epb->mac.mode);

			/*
			**	Initialize value and input error
			**	Boolean, then loop until user enters
			**	valid response or carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter mode word:    ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered,
				**	assume user changed mind.
				*/

				if (buf[0] != NULL)
				{
					mode = strtol(buf, &cvterr, 0);

					if (*cvterr != NULL)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put mode word into an EPB and send it
			**	to MAC sulayer on board to put value
			**	into LANCE's mode word variable on
			**	board.
			*/

			sm_epb->mac.mode = (USHORT)mode;

			sm_epb->mac_sel  = MAC_WR_MODE;

			if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
			{
				/*
				**	Wait for MAC sublayer on board
				**	to send back response.
				*/

				if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
				{
					fprintf(stdout, "\n%25sMode word set to %02x.\n", BLANK, mode);

					get_key();
				}
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_station_add()
**
**  DESCRIPTION:	This function displays the LANCE's current
**			station address, gets the new station address
**			from the user and sends an EPB to the MAC on
**			the board to change the station address.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_station_add(sm_epb)

register SM_EPB	*sm_epb;	/* pointer to EPB */
{
	register DEFAULT i;		/* just a counter */

		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
		 UBYTE	sta_adr[MACSZ];	/* station address */
	
	/*
	**	Get the station address.
	*/

	sm_epb->mac_sel = MAC_RD_NODE;

	if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
	{
		/*
		**	Wait for MAC sublayer on board to send back
		**	response.
		*/

		if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
		{
			fprintf(stdout,"\n\n%25sCurrent station address:  ", BLANK);

			for (i = 0; i < MACSZ; ++i)
				fprintf(stdout,"%02x", sm_epb->station_addr[i]);
	
			fprintf(stdout,"\n");

			/*
			**	Initialize value and input error
			**	Boolean, then loop until user enters
			**	valid response or carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter station address:    ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered,
				**	assume user changed mind.
				*/

				if (buf[0] != NULL)
				{
					if (str_to_hex(buf, (MACSZ * 2), sta_adr, 0) < 0)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put station address into an EPB and send
			**	it to MAC sulayer on board to put value
			**	into LANCE's station address variable on
			**	board.
			*/

			for (i = 0; i < MACSZ; ++i)
				sm_epb->station_addr[i] = sta_adr[i];

			sm_epb->mac_sel  = MAC_WR_NODE;

			if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
			{
				/*
				**	Wait for MAC sublayer on board
				**	to send back response.
				*/

				if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
				{
					fprintf(stdout,"\n%25sStation address set to ", BLANK);

					for (i = 0; i < MACSZ; ++i)
						fprintf(stdout,"%02x", sta_adr[i]);
	
					fprintf(stdout, "\n\n%8sTo change the address in NVRAM/EEPROM, use the NVRAM selection.\n", BLANK);

					get_key();
				}
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_group_add()
**
**  DESCRIPTION:	This function displays the LANCE's current
**			group address mask, gets the new group address
**			mask from the user and sends an EPB to the MAC 
**			on the board to change the group address.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_group_add(sm_epb)

register SM_EPB	*sm_epb;	/* pointer to EPB */
{
	register DEFAULT i;		/* just a counter */

		 BYTE	*cvterr;	/* conversion error */
		 BYTE	buf[80];	/* input buffer	*/
		 USHORT	grp_adr[4];	/* group address */
	
	/*
	**	Get the group address.
	*/

	sm_epb->mac_sel = MAC_RD_GROUP;

	if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
	{
		/*
		**	Wait for MAC sublayer on board to send back
		**	response.
		*/

		if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
		{
			fprintf(stdout,"\n\n%25sCurrent group address:  ", BLANK);

			for (i = 0; i < 4; ++i)
				fprintf(stdout,"%04x", sm_epb->mac.group[i]);
	
			fprintf(stdout,"\n");

			/*
			**	Initialize value and input error
			**	Boolean, then loop until user enters
			**	valid response or carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%25sEnter group address:    ", BLANK);

				gets(buf);

				/*
				**	If carriage return is entered,
				**	assume user changed mind.
				*/

				if (buf[0] != NULL)
				{
					if (str_to_hex(buf, 16, grp_adr, 0) < 0)
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put group address into an EPB and send
			**	it to MAC sublayer on board to put value
			**	into LANCE's group address variable on
			**	board.
			*/

			for (i = 0; i < 4; ++i)
				sm_epb->mac.group[i] = grp_adr[i];

			sm_epb->mac_sel  = MAC_WR_GROUP;

			if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
			{
				/*
				**	Wait for MAC sublayer on board
				**	to send back response.
				*/

				if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
				{
					fprintf(stdout,"\n%25sGroup address set to ", BLANK);

					for (i = 0; i < 4; ++i)
						fprintf(stdout,"%04x", grp_adr[i]);
	
					fprintf(stdout, "\n\n%8sTo change the address in NVRAM/EEPROM, use the NVRAM selection.\n", BLANK);

					get_key();
				}
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_tbc_pri()
**
**  DESCRIPTION:	This function displays the TBC transmit priority,
**			gets the new TBC prioity from the user and sends
**			an EPB to the MAC layer on the board
**			change the MAC priority.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user's input from Configuration Menu
**			scold_user	Boolean for input error message
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_tbc_pri(sm_epb)

register SM_EPB  *sm_epb;	/* pointer to EPB */
{
	register DEFAULT priority;	/* priority for MAC variable */

		 BYTE	 *cvterr;	/* conversion error */
		 BYTE	 pri_buf[80];	/* priority input buffer */

	/*
	**	Get the current TBC priority from MAC layer.
	*/

	sm_epb->mac_sel = MAC_RD_PRIOR;

	if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
	{
		/*
		**	Wait for MAC layer on board to send back response.
		*/

		if (wait_sm_response(sm_epb, MA_PARAM_CONFIRM) == SUCCESS)
		{
			fprintf(stdout, "\n\n%23sCurrent TBC transmit priority:  %d\n", BLANK, sm_epb->mac.priority);

			/*
			**	Initialize priority and input error Boolean,
			**	then loop until user enters valid priority or
			**	carriage return.
			*/

			do
			{
				scold_user = FALSE;

				fprintf(stdout, "%23sEnter TBC transmit priority - 0 through 7 inclusive:  ", BLANK);

				gets(pri_buf);

				/*
				** If carriage return is entered, assume user
				** changed mind.
				*/

				if (pri_buf[0] != NULL)
				{
					priority = strtol(pri_buf, &cvterr, 0);

					if ((*cvterr != NULL) || ((priority < 0) || (priority > 7)))
						wr_invalid_sel();
				}
				else
					return;
			}
			while (scold_user == TRUE);

			/*
			**	Put TBC priority into into an EPB and send it to
			**	MAC layer on board to put priority into TBC
			**	transmit priority variable on board.
			*/

			sm_epb->mac.priority = (UBYTE)priority;

			sm_epb->mac_sel = MAC_WR_PRIOR;

			if (bld_epb(sm_epb, MA_PARAM_REQUEST, MAC_QID) == SUCCESS)
			{
				fprintf(stdout, "\n%23sPriority in TBC reset to %d.\n\n", BLANK, priority);

				fprintf(stdout, "%8sTo change the priority in NVRAM/EEPROM, use the NVRAM selection.\n", BLANK);

				get_key();
			}
		}
	}
}
