static	char ident[] = "@(#)diag.c	35.1		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	diag.c
**
**  DESCRIPTION:	This module contains the functions that process
**			Diagnostic Utilities functions.
**
**  CONTENTS:		do_diag_sel()	   main driver for diagnostic menu
**			get_log_mask()	   gets log_mask value from board
**			do_trmon_sel()     executes monitor menu selection
**			do_log_selection() toggles log_mask values selected
**			trmon_req()	   does management of traffic ctrl
**  			trmon_banner()	   displays the traffic monitor banner
**  			trmon_trailer()	   holds the screen on the crt or
**					   separates the screens in the file
**			trmon_show()	   displays traffic statistics
**
**  MODULE GLOBALS:	none
**
**  NOTES:		See warning below
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <ctype.h>
#include <time.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>
#include <sys/osi/common/stats.h>
#include "def.h"

#define	RD_SIZE		(MAX_CEEPB_SIZE + MAX_BUF_SIZE)
#define	FN_SIZE		30

extern	VOID	get_key();
extern	VOID	ex_shell_cmd();
extern	VOID	wr_mac_parms_menu();
extern	VOID	wr_log_menu();
extern	VOID	wr_banner();
extern	BOOLEAN	read_selection();
extern	DEFAULT	catchit();
extern	DEFAULT	bld_epb();
extern	DEFAULT	wait_sm_response();
extern	UBYTE	*malloc();


extern	struct	termio	nowterm;	/* new terminal status */
extern	struct	termio	oldterm;	/* old terminal status */
extern	BYTE	selection[80];
extern	USHORT	dst_cpu;
extern	USHORT	board;
extern	BOOLEAN	scold_user;
extern	DEFAULT	errno;

/*
**	W A R N I N G ! ! !	W A R N I N G ! ! !	W A R N I N G ! ! !
**
**	The following definitions are dependend on QID numbers defined
**	in epb.h. If QID's are changed the headings have to be changed
**	accordingly.
*/

BYTE	*tr_hds[] =
{	
	"CEnv",
	"MAC",
	"LLC",
	"Netw",
	"XPort",
	"Sessn",
	"Prsnt",
	"ACSE",
	"FTAM",
	"NMgr",
	"DServ",
};

struct	traffic	*tr_buf	= NULL;
UBYTE	*buf		= (UBYTE *)NULL;
USHORT	tr_board	= 0;		/* MVME37X board number */
ULONG	tr_seq		= 0;		/* sequence number of monitor starts */
BYTE	started[26]	= {0};		/* start time of traffic monitor */
BYTE	stopped[26]	= {0};		/* stop time of traffic monitor */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_diag_sel()
**
**  DESCRIPTION:	This function switches on the user's selection
**			from the Diagnostic Utilities Menu and calls the
**			proper routine to process the selection.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	selection	user input from menu selection
**			dst_cpu		destination cpu
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_diag_sel()

{
	VOID	do_log_selection();
	VOID	do_trmon_sel();
	BOOLEAN	get_log_mask();

	SM_EPB	 stat_epb;
	ULONG	 log_mask;
	BYTE	 sys_buf[20];

	switch (selection[0])
	{
		case MAC_STATS:
		{
			/*
			** Switch off BREAK so when user kills MACSTAT, we
			** still live.
			*/

			ioctl(0, TCGETA, &oldterm);
			ioctl(0, TCGETA, &nowterm);
			nowterm.c_iflag &= IGNBRK;
			ioctl(0, TCSETA, &nowterm);
			sprintf(sys_buf, "macstat %d", dst_cpu);
			system(sys_buf);

			/*
			** We're back.  Switch on BREAK so we can reset
			** terminal when we are terminated by user.
			*/

			sleep (2);
			ioctl(0, TCSETA, &oldterm);
			break;
		}

		case LOGGING:
		{
			if (get_log_mask(&stat_epb, &log_mask) != FAILURE)
			{
				putp(tparm(clear_screen));

				do
				{
					wr_log_menu(log_mask);

					read_selection(LOG_MENU);

					do_log_selection(&stat_epb, &log_mask);
				}
				while ((selection[0] != NO_CHANGE) && (selection[0] != SET_LOG_MASK));
			}

			break;
		}

		case TRMON:
		{
			do
			{
				wr_monitor_menu();

				if (read_selection(TRMON_MENU) != FAILURE)
					do_trmon_sel(&stat_epb);
			}
			while ((selection[0] != DIAGNOSTIC_UTILITIES) && (selection[0] != MAIN));

			break;
		}

		case QUIT:
		{
			catchit();
			break;
		}

		case SHELL:
		{
			ex_shell_cmd();
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	get_log_mask()
**
**  DESCRIPTION:	Gets value of log_mask from board.
**
**  PARAMETERS:		
**
**	Input:		log_mask	- address of logging mask,
**
**			sm_epb		- Station Manager EPB.
**
**	Output:		log_mask
**
**	Returns:	SUCCESS/FAILURE
**
**  SYSTEM GLOBALS:	none
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

BOOLEAN
get_log_mask(sm_epb, log_mask)

register SM_EPB	*sm_epb;
register ULONG	*log_mask;

{
	/*
	** Get an EPB, have it built and sent to Station Manager on the board.
	*/

	if (bld_epb(sm_epb, SM_LOG_MASK_RD_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		** Wait for Station Manager on board to send back response.
		*/

		if (wait_sm_response(sm_epb, SM_LOG_MASK_RD_CONFIRM) == SUCCESS)
		{
			*log_mask = sm_epb->log_mask;

			return (SUCCESS);
		}
	}

	return (FAILURE);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_trmon_sel()
**
**  DESCRIPTION:	Executes the monitor selection.
**
**  PARAMETERS:		
**
**	Input:		sm_epb		- Station Manager EPB.
**
**	Output:		None.
**
**	Returns:	None.
**
**  SYSTEM GLOBALS:	selection	user input buffer
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_trmon_sel(sm_epb)

register SM_EPB	 *sm_epb;

{
	VOID	trmon_show();
	VOID	trmon_req();

	ULONG	clock;

	switch (selection[0])
	{
		case M_STRT:
		{
			++tr_seq;

			sm_epb->sys.clock = time(NULL);

			trmon_req(sm_epb, MON_STRT);

			break;
		}

		case M_READ:
		{
			trmon_req(sm_epb, MON_READ);

			clock = time(NULL);

			strcpy(stopped, ctime(&clock));

			started[24] = stopped[24] = '\0';

			break;
		}

		case M_SHOW:
		case M_WRTE:
		{
			trmon_show(selection[0]);
			break;
		}

		case QUIT:
		{
			catchit();
			break;
		}

		case SHELL:
		{
			ex_shell_cmd();
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_log_selection()
**
**  DESCRIPTION:	Toggles local copy of log_mask and can change
**			log_mask on board.
**
**  PARAMETERS:		
**
**	Input:		log_mask
**
**			sm_epb		- Station Manager EPB.
**
**	Output:		log_mask
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user input buffer
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_log_selection(sm_epb, log_mask)

register SM_EPB	*sm_epb;
register ULONG	*log_mask;

{
	switch (selection[0])
	{
		case HOST_LOGGING:
			*log_mask ^= ~HOST_LOG_MASK;
			break;

		case DEBUG_LOGGING:
			*log_mask ^= ~DEBUG_LOG_MASK;
			break;

		case SYSTEM_LOGGING:
			*log_mask ^= ~SYSTEM_LOG_MASK;
			break;

		case MAC_LOGGING:
			*log_mask ^= ~MAC_LOG_MASK;
			break;

		case LLC_LOGGING:
			*log_mask ^= ~LLC_LOG_MASK;
			break;

		case INTERNET_LOGGING:
			*log_mask ^= ~INTERNET_LOG_MASK;
			break;

		case XPORT_LOGGING:
			*log_mask ^= ~XPORT_LOG_MASK;
			break;

		case SESSION_LOGGING:
			*log_mask ^= ~SESSION_LOG_MASK;
			break;

		case PRESENTATION_LOGGING:
			*log_mask ^= ~PRESENTATION_LOG_MASK;
			break;

		case ACSE_LOGGING:
			*log_mask ^= ~ACSE_LOG_MASK;
			break;

		case FTAM_LOGGING:
			*log_mask ^= ~FTAM_LOG_MASK;
			break;

		case NET_AGENT_LOGGING:
			*log_mask ^= ~NET_AGENT_LOG_MASK;
			break;

		case DIRSVS_LOGGING:
			*log_mask ^= ~DIRSVS_LOG_MASK;
			break;

		case STA_MGR_LOGGING:
			*log_mask ^=~STA_MGR_LOG_MASK;
			break;

		case NO_CHANGE:
			break;

		case SET_LOG_MASK:
			/*
			** Get an epb, build and send it to Station Manager
			** on board to change log_mask there.
			*/

			sm_epb->log_mask = (ULONG)*log_mask;

			bld_epb(sm_epb, SM_LOG_MASK_SET_REQUEST, SM_QID);

			break;

		default:
			/* Ignore previous input if user enters carriage return.*/
			selection[0] = NULL;
			break;
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	trmon_req ()
**
**  DESCRIPTION:	This function calls routines to get an epb, sets
**			the TRMON command and post it to the Station Manager
**			on the board.
**
**  SYSTEM GLOBALS:	none
**
**  PARAMETERS:		
**
**	Input:		sm_epb		- Station Manager EPB,
**
**			cmd		- TRMON command.
**
**	Output:		none
**
**	Returns:	nothing
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

VOID
trmon_req(sm_epb, cmd)

register SM_EPB	*sm_epb;
register ULONG	cmd;
{
	/*
	**  Construct the EPB and send it to Station Manager on the board.
	*/

	sm_epb->log_mask = cmd;	/* !!! use log_mask as command */

	if (bld_epb(sm_epb, SM_TRMON_REQUEST, SM_QID) == SUCCESS)
	{
		/*
		** Get response back from Station Manager on the board.
		*/

		if ((buf != (UBYTE *)NULL) || ((buf = malloc(RD_SIZE)) != (UBYTE *)NULL))
			if (wait_sm_response((sm_epb = (SM_EPB *)buf), SM_TRMON_RESPONSE) == SUCCESS)
			{
				if (sm_epb->log_mask == MON_READ)
				{
					strcpy(started, ctime(&(sm_epb->sys.clock)));

					tr_buf = (struct traffic *)(sm_epb->hdr.buf->data);

					tr_board = sm_epb->board;
				}
			}
	}
}

/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:	trmom_banner()
**
**  DESCRIPTION: 	trmon_banner displays the traffic monitor banner;
**
**  PARRMETERS:           
**
**	Input:		msec		- start/stop time calculation,
**
**			file		- either stdout or file name.
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	none
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**  UPDATE LOG:
**  Date        Name            V/R     Comments
**  --------    -----------     ---     -----------------------------
*************************************************************************/

VOID
trmon_banner(msecs, file, screen_output)

register ULONG	 msecs;
register FILE	 *file;		/* Output file pointer */
register BOOLEAN screen_output; /* Is it file or screen? */

{
	if (screen_output == TRUE)
		putp(tparm(clear_screen));

	switch (board)
	{
		case 372:
		{
			fprintf(file, "\n%24sMicroMAP Performance Monitor\n\n", BLANK);
			break;
		}

		default:
		{
			fprintf(file, "\n%24sMicroTOP Performance Monitor\n\n", BLANK);
			break;
		}
	}

	fprintf(file, "Monitor Started  %s", started + 11);
	fprintf(file, "     Observation Period (msec)  %8d\n", msecs);
	fprintf(file, "Monitor Stopped  %s", stopped + 11);
	fprintf(file, "     Observation Sequence #     %8d\n", tr_seq);
}

/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:	trmon_trailer()
**
**  DESCRIPTION: 	Holds the screen on the crt or separates the
**			screens in the file
**
**  PARAMETERS:           
**
**	Input:		file	either stdout or file name
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	none
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**  UPDATE LOG:
**  Date        Name            V/R     Comments
**  --------    -----------     ---     -----------------------------
*************************************************************************/

VOID
trmon_trailer(screen_output, file)

register BOOLEAN screen_output;	/* Is it for screen or file? */
register FILE	 *file;		/* Output file pointer */

{
	if (screen_output == TRUE)
		get_key();
	else
		fprintf(file, "\n\n\n");
}

/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:	trmon_show()
**
**  DESCRIPTION: 	Displays results of traffic monitoring.
**
**  PARRMETERS:           
**
**	Input:		type
**
**	Output:		traffic statistics
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date        Name            V/R     Comments
**  --------    -----------     ---     -----------------------------
*************************************************************************/

VOID
trmon_show(type)

BYTE	type;
{
	VOID	 trmon_trailer();
	VOID	 trmon_banner();

	register struct  tr_flow *p_flow;
	register ULONG	 msecs;		/* Monitor time in msec */
	register DEFAULT i,j;		/* Loop counters */
	register FILE	 *file;		/* Output file pointer */
	register BYTE	 fname[FN_SIZE];/* Output file name */
		 BOOLEAN screen_output;	/* screen or standard file */


	if (!tr_buf)
	{
		fprintf(stdout, "\n\n%20sNo data available\n", BLANK);

		get_key();

		return;
	}

	/*
	**	Check if display to file is requested
	*/

	if (type == M_WRTE)
	{
		screen_output = FALSE;

		fprintf(stdout, "\n\n%20sEnter file name : ", BLANK);

		errno = 0;

		if ((fgets(fname, FN_SIZE, stdin)) == NULL)
		{
			if (errno != 0)
				fprintf(stdout, "\n\n%20sCan't read file name (errno = %d)\n", BLANK, errno);
			else
				fprintf(stdout, "\n\n%20sNo default name assigned\n", BLANK);

			get_key();

			return;
		}

		fname[strlen(fname) - 1] = '\0';	/* clear trailing LF */

		if ((file = fopen(fname, "a")) == NULL)
		{
			fprintf(stdout, "\n\n%20sCan't open file %s (errno = %d)\n", BLANK, fname, errno);

			get_key();

			return;
		}
	}
	else
	{	screen_output = TRUE;

		file = stdout;
	}

	/*
	**	Do a few calculations before we display everything
	*/

	msecs = (tr_buf->etime - tr_buf->stime) * MSPT;

	trmon_banner(msecs, file, screen_output);

	switch (tr_board)
	{
		case 372:
		{
			fprintf(file, "\nTBC Interface Traffic%12sHost Interface Traffic\n", BLANK);
			fprintf(file, "Frames Transmitted  %8d", tr_buf->txf_cnt);
			fprintf(file, "     Events Transmitted  %8d\n", tr_buf->txe_cnt);
			fprintf(file, "Frames Received     %8d", tr_buf->rxf_cnt);
			fprintf(file, "     Events Received     %8d\n", tr_buf->rxe_cnt);
			fprintf(file, "TBC Interrupts      %8d", tr_buf->ints);

			break;
		}

		default:
		{
			fprintf(file, "\nLANCE Interface Traffic%10sHost Interface Traffic\n", BLANK);
			fprintf(file, "Packets Transmitted %8d", tr_buf->txf_cnt);
			fprintf(file, "     Events Transmitted  %8d\n", tr_buf->txe_cnt);
			fprintf(file, "Packets Received    %8d", tr_buf->rxf_cnt);
			fprintf(file, "     Events Received     %8d\n", tr_buf->rxe_cnt);
			fprintf(file, "LANCE Interrupts    %8d", tr_buf->ints);

			break;
		}
	}

	fprintf(file, "     Host Interrupts     %8d\n", tr_buf->host_int);

	/*
	**	Display EPB flow
	*/

	fprintf(file, "\nLocal Resources       avail. cur_use max_use suc_req err_req\n");
	fprintf(file, "EPB's             %8d%8d%8d%8d%8d\n", tr_buf->max_l_epbs,
		tr_buf->cur_epbs, tr_buf->max_u_epbs, tr_buf->suc_r_epbs,
		tr_buf->err_r_epbs);
	fprintf(file, "BDB's             %8d%8d%8d%8d%8d\n", tr_buf->max_l_bdbs,
		tr_buf->cur_bdbs, tr_buf->max_u_bdbs, tr_buf->suc_r_bdbs,
		tr_buf->err_r_bdbs);
	fprintf(file, "\nShared Resources          shared to local       ");
	fprintf(file, "       local to shared\n");
	fprintf(file, "                      passed blocked max_retry   ");
	fprintf(file, "  passed blocked max_retry\n");
	fprintf(file, "EPB's             %8d%8d%10d  %8d%8d%10d\n",
		tr_buf->cnt_in_epbs, tr_buf->cnt_ib_epbs, tr_buf->max_ir_epbs,
		tr_buf->cnt_out_epbs, tr_buf->cnt_ob_epbs, tr_buf->max_or_epbs);
	fprintf(file, "BDB's             %8d%8d%10d  %8d%8d%10d\n",
		tr_buf->cnt_in_bdbs, tr_buf->cnt_ib_bdbs, tr_buf->max_ir_bdbs,
		tr_buf->cnt_out_bdbs, tr_buf->cnt_ob_bdbs, tr_buf->max_or_bdbs);

	trmon_trailer(screen_output, file);

	trmon_banner(msecs, file, screen_output);

	/*
	**	Display EPB flow
	*/

	fprintf(file, "\nModule Traffic Flow - # of EPB's   ");
	fprintf(file, "posts   %6d", tr_buf->cnt_qpost);
	fprintf(file, "%7spends   %6d\n\n", BLANK, tr_buf->cnt_qpend);

	/*
	**	Print flow matrix and headings
	*/

	fprintf(file, "fm\\to ");

	for (i = 0; i < TR_SZ; ++i)
		fprintf(file, "%5s ", tr_hds[i]);

	fprintf(file, "\n");

	for (i = 0; i < TR_SZ; ++i)
	{
		fprintf(file, "%-5s ", tr_hds[i]);

		for (j = 0; j < TR_SZ; ++j)
		{
			p_flow = &(tr_buf->flow[i][j]);

			fprintf(file, "%5d ", p_flow->epbs);
		}

		fprintf(file, "\n");
	}

	trmon_trailer(screen_output, file);

	trmon_banner(msecs, file, screen_output);

	/*
	**	Display BDB flow
	*/

	fprintf(file, "\nModule Traffic Flow -- # of BDB's\n\n");

	/*
	**	Print flow matrix headings
	*/

	fprintf(file, "\nfm\\to ");

	for (i = 0; i < TR_SZ; ++i)
		fprintf(file, "%5s ", tr_hds[i]);

	fprintf(file, "\n");

	/*
	**	Print flow matrix
	*/

	for (i = 0; i < TR_SZ; ++i)
	{
		fprintf(file, "%-5s ", tr_hds[i]);

		for (j = 0; j < TR_SZ; ++j)
		{
			p_flow = &(tr_buf->flow[i][j]);

			fprintf(file, "%5d ", p_flow->bdbs);
		}

		fprintf(file, "\n");
	}

	trmon_trailer(screen_output, file);

	trmon_banner (msecs, file, screen_output);

	/*
	**	Display data flow
	*/

	fprintf(file, "\nModule Traffic Flow -- Data (KB's)\n\n");

	/*
	**	Print flow matrix headings
	*/

	fprintf(file, "\nfm\\to ");

	for (i = 0; i < (TR_SZ - 1); ++i)
		fprintf(file, "%6s ", tr_hds[i]);

	fprintf(file, "\n");

	/*
	**	Print flow matrix
	*/

	for (i = 0; i < (TR_SZ - 1); ++i)
	{
		fprintf(file, "%-5s ", tr_hds[i]);

		for (j = 0; j < (TR_SZ - 1); ++j)
		{
			p_flow = &(tr_buf->flow[i][j]);

			fprintf(file, "%6.1f ", p_flow->numb / 1024.);
		}

		fprintf(file, "\n");
	}

	if (type == M_SHOW)
		get_key();

	else
		if (fclose(file) == EOF)
			fprintf(stdout, "\n\n%20sCan't close file %s (errno = %d)\n",
				BLANK, fname, errno);
}
