/* @(#)init.c	35.1 */
/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	init.c
**
**  DESCRIPTION:	Process signals.
**
**  CONTENTS:		init()		Set up to catch signals, etc.
**			catchit()	Reset terminal if program terminates
**					abnormally.
**
**  SYSTEM GLOBALS:	ci_fd	file descriptor for common environment calls
*			dst_cpu	destination CPU id
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <signal.h>
#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>

extern	char	*getenv();
	int	catchit();	/* Forward dcl of signal catching func. */

extern	USHORT	dst_cpu;	/* destination CPU id		*/

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	init()
**
**  DESCRIPTION: 	
**			Set up terminal for screen output.
**
**  PARAMETERS:		
**
**	Input:		argc, argv
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	dst_cpu	destination CPU id
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID	init (argc, argv)

register	int	  argc;
register	char	**argv;

{
	LONG	  tmp_cpu;	/* temporary CPU id		*/
	char	 *cvterr;

	/*
	** Check for one argument -- the destination CPU id.
	*/

	if (argc != 2)
	{
		fprintf (stdout, "usage:  %s  destination_cpu\n", argv[0]);
		exit (1);
	}
	tmp_cpu = strtol (argv[1], &cvterr, 0);
	if ((tmp_cpu < CPU_OFFSET) || (tmp_cpu > NUM_CPU_ENTRIES) || (*cvterr != NULL))
	{
		fprintf (stdout, "smc:  Invalid Destination CPU number = %d\n",
				tmp_cpu);
		exit (1);
	}
	dst_cpu = (USHORT)tmp_cpu;
	/*
	** Set up to catch signals so we can reset the terminal.
	*/
	setupterm (getenv("TERM"), 1, 0);
	signal (SIGHUP,  catchit);
	signal (SIGINT,  catchit);
	signal (SIGQUIT, catchit);
	signal (SIGTERM, catchit);

	return;
 }

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	catchit()
**
**  DESCRIPTION:	This function resets the terminal and exits.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	ci_fd	file descriptor for Common Environment calls
**			src_id	our source id used by Common Environment
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

int catchit()
{
extern		int	ci_deregister();

extern		int	errno;
extern		int	ci_fd;
extern		ULONG	src_id;

register	int	n;

	/* 
	** Clear the screen and reset the terminal before terminating.
	*/
	putp (tparm (clear_screen));
	resetterm();
	/*
	** Deregister ourselves before terminating.
	*/
	if ((n = ci_deregister (src_id, ci_fd)) != 0)
		printf ("SM CONSOLE failure in attempt to deregister, err = %d\n",
			 errno);
	exit(0);
}
