/* @(#)main.c	35.1 */
/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	main.c
**
**  DESCRIPTION:	This is the main procedure to drive the Station
**			Monitor Console which displays menus to lead
**			the user through station management.
**
**  CONTENTS:		main ()
**
**  SYSTEM GLOBALS:	ci_fd	file descriptor for common environment calls
**			dst_cpu	destination CPU number
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>
#include "def.h"

#define	RD_SIZE	(MAX_CEEPB_SIZE + MAX_BUF_SIZE)

/*
**			G L O B A L   V A R I A B L E S
*/

DEFAULT	ci_fd;		/* file descriptor for common environment calls	*/
USHORT	dst_cpu;	/* destination CPU number			*/
USHORT	board;		/* board number            			*/
ULONG	src_id;		/* source id 					*/
BOOLEAN	scold_user;	/* flag to control error messages		*/
BYTE	selection[80];	/* user's input buffer				*/

CPU_TAB	cpu_tab[NUM_CPU_ENTRIES];
#define CPU_OFFSET 2	/* offset between CPU number and table entry index */

main (argc, argv)

int	  argc;
char	**argv;

{
extern		int	ci_register();
extern		int	ci_deregister();

extern		int	catchit();
extern		VOID	ex_shell_cmd();
extern		VOID	init();
extern		VOID	wr_main_menu();
extern		BOOLEAN	read_selection();
extern		VOID	node_restart();
extern		VOID	do_cnfg_sel();
extern		VOID	do_diag_sel();

extern		int	  errno;
register	CPU_TAB	  *cpu_ptr;
register	int	  n;

	/*
	** Initialize:
	**	Check for dst_cpu argument and set signals.
	*/
	init (argc, argv);

	/*
	** Register with the common environment.
	*/

	src_id = ID_NUM;

	if ((n = ci_register(dst_cpu, &src_id, &ci_fd)) != 0)
	{
		fprintf (stdout, "SM CONSOLE failed to register, err = %d\n", errno);

		exit (1);
	}

	/*
	** Set us up to WAIT forever on the response queue.
	*/

	if (ioctl(ci_fd, WAIT, 0) < 0)
	{
		printf("SM CONSOLE unable to set ioctl to BLOCKED reads, errno = %d\n", errno);

		if ((n = ci_deregister (src_id, ci_fd)) != 0)
			printf("SM CONSOLE failure in attempt to deregister, err = %d\n", errno);

		exit (1);
	}

	/*
	**	Get CPU table.
	*/

	if (ioctl(ci_fd, CONFIG, cpu_tab) >= 0)
	{
		/*
		**	Get board type.
		*/

		cpu_ptr = &cpu_tab[dst_cpu - CPU_OFFSET];

		if (cpu_ptr->cpu_present && cpu_ptr->cpu_running)
			board = cpu_ptr->cpu_type;
		else
		{
			printf("SM CONSOLE CPU not present/running\n");

			if ((n = ci_deregister(src_id, ci_fd)) != 0)
				printf("SM CONSOLE failure in attempt to deregister, err = %d\n", errno);

			exit (1);
		}
	}
	else
		board = 372;

	/*
	** Display main menu.
	*/

	scold_user = FALSE;
	wr_main_menu ();

	/*
	** Get user menu selection.
	*/

	if (read_selection (MAIN_MENU) == FAILURE);

	/*
	** Loop until user quits.
	*/

	while (selection[0] != QUIT)
	{

		/*
		** Switch on user selection.
		*/

		switch (selection[0])
		{
			case NODE_RESTART:
			{
				node_restart();
				break;
			}

			case CONFIGURATION_UTILITIES:
			{
				while (selection[0] != MAIN)
				{
					wr_cnfg_menu ();
					if (read_selection (CONFIGURATION_MENU)
					    != FAILURE)
						do_cnfg_sel ();
				}
				break;
			}

			case DIAGNOSTIC_UTILITIES:
			{
				while (selection[0] != MAIN)
				{
					wr_diag_menu ();
					if (read_selection (DIAGNOSTIC_MENU)
					    != FAILURE)
						do_diag_sel ();
				}
				break;
			}

			case QUIT:
			{
				/*
				** Reset terminal and exit from program.
				*/
				catchit();
				break;
			}

			case SHELL:
			{
				/*
				** Execute shell command for user.
				*/
				ex_shell_cmd();
				break;
			}

			default:
			{
				scold_user = TRUE;
			}
		} /*switch*/

		/*
		** Display main menu.
		*/
		wr_main_menu ();

		/*
		** Get user selection.
		*/
		if (read_selection (MAIN_MENU) != FAILURE);
	}/*while*/

	/*
	** Reset the terminal before terminating.
	*/

	putp (tparm (clear_screen));
	resetterm ();
}
