static	char ident[] = "@(#)menus.c	35.3		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	menus.c
**
**  DESCRIPTION:	Writes all the menus used by Station Manager.
**
**  CONTENTS:		wr_main_menu()	    writes main menu
**			wr_cnfg_menu()	    writes Configuration Utilities Menu
**			wr_diag_menu()	    writes Diagnostic Utilities Menu
**			wr_monitor_menu()   writes monitor menu
**			wr_system_menu()    writes system menu
**			wr_xport_menu()     writes transport menu
**			wr_lance_menu()     writes LANCE/MAC layer menu
**			wr_tbc_menu()       writes TBC/MAC layer menu
**			wr_log_menu()	    writes menu to toggle logging
**			wr_banner()	    writes logo at top of screen
**			wr_invalid_sel()    writes error msg at bottom of menu
**			wr_incorrect_sel()  writes error msg at bottom of menu
**			wr_error()	    clears screen and writes error msg
**
**  SYSTEM GLOBALS:	none
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/common/types.h>
#include "def.h"

extern	BOOLEAN	scold_user;
extern	DEFAULT	errno;
extern	USHORT	board;

extern	VOID	clr();
extern	VOID	get_key();

BYTE	mot147[] = "\n%24sM O T O R O L A   M V M E 1 4 7\n\n";
BYTE	motmap[] = "\n%23sM O T O R O L A   M i c r o M A P\n\n";
BYTE	mottop[] = "\n%23sM O T O R O L A   M i c r o T O P\n\n";

BYTE	on[]  = "ON ";
BYTE	off[] = "OFF";

BYTE	escape[] = "!.  Escape to Shell\n\n";
BYTE	quit[]   = "q.  Quit\n\n";
BYTE	mainm[]  = "m.  Return to Main Menu\n\n";
BYTE	cnfgm[]  = "c.  Return to Configuration Menu\n\n";

BYTE	cnfgu[]  = "c.  Configuration Utilities ";
BYTE	diagu[]  = "d.  Diagnostic Utilities    ";

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_main_menu()
**
**  DESCRIPTION:	This function prepares the screen and writes
**			the main menu.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_main_menu()

{
	VOID	wr_banner();
	VOID	wr_incorrect_sel();

	wr_banner();

	fprintf(stdout, "%26sStation Manager -- Main Menu\n\n", BLANK);

	switch (board)
	{
		case 147:
		{
			fprintf(stdout, "%10s%s", BLANK, cnfgu);
			fprintf(stdout, "%10s%s", BLANK, quit);
			fprintf(stdout, "%10s%s", BLANK, diagu);
			fprintf(stdout, "%10s%s", BLANK, escape);

			break;
		}

		case 372:
		case 374:
		{
			fprintf(stdout, "%10sr.  Micro%s Node Restart   ", BLANK, (board == 372) ? "MAP" : "TOP");

			fprintf(stdout, "%10s%s", BLANK, quit);
			fprintf(stdout, "%10s%s", BLANK, cnfgu);
			fprintf(stdout, "%10s%s", BLANK, escape);
			fprintf(stdout, "%10s%s\n\n", BLANK, diagu);

			break;
		}
	}

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_cnfg_menu()
**
**  DESCRIPTION:	This function prepares the screen and writes
**			the Configuration Utilities menu.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBAL:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_cnfg_menu()

{
	VOID	wr_banner();
	VOID	wr_incorrect_sel();

	wr_banner();

	fprintf(stdout, "%26sConfiguration Utilities Menu\n\n", BLANK);

	fprintf(stdout, "1. Display/Modify Non-volatile RAM Parameters", BLANK);
	fprintf(stdout, "%6s%s", BLANK, mainm);
	fprintf(stdout, "2. Display/Modify MAC Layer Parameters       ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, quit);

	switch (board)
	{
		case 147:
		{
			fprintf(stdout, "%51s%s", BLANK, escape);

			break;
		}

		case 372:
		case 374:
		{
			fprintf(stdout, "3. Display/Modify System Parameters          ", BLANK);
			fprintf(stdout, "%6s%s", BLANK, escape);
			fprintf(stdout, "4. Display/Modify Transport Layer Parameters\n\n", BLANK);

			break;
		}
	}

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_diag_menu()
**
**  DESCRIPTION:	This function prepares the screen and writes
**			the Diagnostic Utilities menu.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_diag_menu()

{
	VOID	wr_banner();
	VOID	wr_incorrect_sel();

	wr_banner();

	fprintf(stdout, "%27sDiagnostic Utilities Menu\n\n", BLANK);

	fprintf(stdout, "%5s1.  Display MAC Statistics       ", BLANK);
	fprintf(stdout, "%10s%s", BLANK, mainm);
	fprintf(stdout, "%5s2.  Modify Logging               ", BLANK);
	fprintf(stdout, "%10s%s", BLANK, quit);

	switch (board)
	{
		case 147:
		{
			fprintf(stdout, "%48s%s", BLANK, escape);

			break;
		}

		case 372:
		case 374:
		{
			fprintf(stdout, "%5s3.  Monitor Module Traffic       ", BLANK);
			fprintf(stdout, "%10s%s", BLANK, escape);

			break;
		}
	}

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_monitor_menu()
**
**  DESCRIPTION:	This function writes the layer traffic monitor
**			selections and request user input.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_monitor_menu()

{
	VOID	wr_banner();
	VOID	wr_incorrect_sel();

	wr_banner();

	fprintf(stdout, "%27sMonitor Layer Traffic Menu\n\n", BLANK);

	fprintf(stdout, "%5s1.  Start Layer Traffic Monitor", BLANK);
	fprintf(stdout, "%10sd.  Return to Diagnostics Menu\n\n", BLANK);
	fprintf(stdout, "%5s2.  Read Data (Stop Monitor)   ", BLANK);
	fprintf(stdout, "%10s%s", BLANK, mainm);
	fprintf(stdout, "%5s3.  Show Monitor Data          ", BLANK);
	fprintf(stdout, "%10s%s", BLANK, quit);
	fprintf(stdout, "%5s4.  Write Monitor Data to File ", BLANK);
	fprintf(stdout, "%10s%s", BLANK, escape);

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_system_menu()
**
**  DESCRIPTION:	This function writes the system menu.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_system_menu()

{
	VOID	wr_banner();
	VOID	wr_incorrect_sel();

	wr_banner();

	fprintf(stdout, "%21sDisplay/Modify System Parameters Menu\n\n", BLANK);

	fprintf(stdout, "1. Display BDB Count                ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, cnfgm);
	fprintf(stdout, "2. Display/Modify User Limit        ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, mainm);
	fprintf(stdout, "3. Display/Modify Management Limit  ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, quit);
	fprintf(stdout, "4. Display/Modify Network Limit     ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, escape);

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_xport_menu()
**
**  DESCRIPTION:	This function writes the transport menu.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_xport_menu()

{
	VOID	wr_banner();
	VOID	wr_incorrect_sel();

	wr_banner();

	fprintf(stdout, "%19sDisplay/Modify Transport Parameters Menu\n\n", BLANK);

	fprintf(stdout, "1. Maximum Number of Connections    ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, cnfgm);
	fprintf(stdout, "2. Display/Modify Minimum Credit    ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, mainm);
	fprintf(stdout, "3. Display/Modify Maximum Credit    ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, quit);
	fprintf(stdout, "4. Maximum Number of Retries        ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, escape);
	fprintf(stdout, "5. Retry Timer for Data (in seconds)\n\n");
	fprintf(stdout, "6. Retry Timer for Expedited Data (in seconds)\n\n");
	fprintf(stdout, "7. Window Timer (in seconds)\n\n");
	fprintf(stdout, "8. Reference Wait Timer (in seconds)\n\n");

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_lance_menu()
**
**  DESCRIPTION:	This function writes the LANCE/MAC layer menu.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_lance_menu()

{
	VOID	wr_incorrect_sel();

	fprintf(stdout, "%3s1. Display LANCE Ring Sizes         ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, cnfgm);
	fprintf(stdout, "%3s2. Display/Modify LANCE Mode        ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, mainm);
	fprintf(stdout, "%3s3. Display/Modify Station Address   ", BLANK);
	fprintf(stdout, "%6s%s", BLANK, quit);
	fprintf(stdout, "%3s4. Display/Modify Group Address Mask", BLANK);
	fprintf(stdout, "%6s%s", BLANK, escape);

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_tbc_menu()
**
**  DESCRIPTION:	This function writes the TBC/MAC layer menu.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_tbc_menu()

{
	VOID	wr_incorrect_sel();

	fprintf(stdout, "%3s1. Display/Modify TBC TX Priority   ", BLANK);
	fprintf(stdout, "%4s%s", BLANK, cnfgm);
	fprintf(stdout, "%43s%s", BLANK, mainm);
	fprintf(stdout, "%43s%s", BLANK, quit);
	fprintf(stdout, "%43s%s", BLANK, escape);

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_log_menu()
**
**  DESCRIPTION:	This function writes the menu to toggle logging.
**
**  PARAMETERS:	
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_log_menu(log_mask)

register ULONG	log_mask;

{
	VOID	wr_incorrect_sel();

	putp (tparm(cursor_home));

	if (scold_user == TRUE)
	{
		putp(tparm(flash_screen));
		putp(tparm(bell));
	}

	switch (board)
	{
		case 147:
		{
			fprintf(stdout, mot147, BLANK);
			break;
		}

		case 372:
		{
			fprintf(stdout, motmap, BLANK);
			break;
		}

		case 374:
		{
			fprintf(stdout, mottop, BLANK);
			break;
		}
	}

	fprintf(stdout, "%26sDisplay/Modify Logging Menu\n\n", BLANK);

	switch (board)
	{
		case 147:
		{
			fprintf(stdout, "%6sa.  System Logging             %s", BLANK, (log_mask & ~SYSTEM_LOG_MASK) ? on : off);

			fprintf(stdout, "%5sd.  Leave Logging Unchanged\n", BLANK);

			fprintf(stdout, "%6sb.  MAC Logging                %s", BLANK, (log_mask & ~MAC_LOG_MASK) ? on : off);

			fprintf(stdout, "%5se.  Set Logging as Shown\n", BLANK);

			fprintf(stdout, "%6sc.  Station Manager Logging    %s\n\n", BLANK, (log_mask & ~STA_MGR_LOG_MASK) ? on : off);

			break;
		}

		case 372:
		case 374:
		{
			fprintf(stdout, "  a.  Logging to Host            %s", (log_mask & ~HOST_LOG_MASK) ? on : off);

			fprintf(stdout, "%5sh.  Session Logging             %s\n", BLANK, (log_mask & ~SESSION_LOG_MASK) ? on : off);

			fprintf(stdout, "  b.  Logging to Debug Port      %s", (log_mask & ~DEBUG_LOG_MASK) ? on : off);

			fprintf(stdout, "%5si.  Presentation Logging        %s\n", BLANK, (log_mask & ~PRESENTATION_LOG_MASK) ? on : off);

			fprintf(stdout, "  c.  System Logging             %s", (log_mask & ~SYSTEM_LOG_MASK) ? on : off);

			fprintf(stdout, "%5sj.  ACSE Logging                %s\n", BLANK, (log_mask & ~ACSE_LOG_MASK) ? on : off);

			fprintf(stdout, "  d.  MAC Logging                %s", (log_mask & ~MAC_LOG_MASK) ? on : off);

			fprintf(stdout, "%5sk.  FTAM Logging                %s\n", BLANK, (log_mask & ~FTAM_LOG_MASK) ? on : off);

			fprintf(stdout, "  e.  LLC Logging                %s", (log_mask & ~LLC_LOG_MASK) ? on : off);

			fprintf(stdout, "%5sl.  Network Agent Logging       %s\n", BLANK, (log_mask & ~NET_AGENT_LOG_MASK) ? on : off);

			fprintf(stdout, "  f.  Internet Logging           %s", (log_mask & ~INTERNET_LOG_MASK) ? on : off);

			fprintf(stdout, "%5sm.  Directory Services Logging  %s\n", BLANK, (log_mask & ~DIRSVS_LOG_MASK) ? on : off);

			fprintf(stdout, "  g.  Transport Logging          %s", (log_mask & ~XPORT_LOG_MASK) ? on : off);

			fprintf(stdout, "%5sn.  Station Manager Logging     %s\n", BLANK, (log_mask & ~STA_MGR_LOG_MASK) ? on : off);

			fprintf(stdout, "%25so. Leave Logging Unchanged\n", BLANK);
			fprintf(stdout, "%25sp. Set Logging as Shown\n\n", BLANK);

			break;
		}
	}

	wr_incorrect_sel();
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_banner()
**
**  DESCRIPTION:	This function clears the screen.  If there has been
**			an input error, it also rings the bell and flashes
**			the screen.  The Motorola logo is written across
**			the top of the screen.
**
**  PARAMETERS:	
**	Input:		curchar		cursor character
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_banner()

{
	putp (tparm(clear_screen));

	if (scold_user == TRUE)
	{
		putp(tparm(flash_screen));
		putp(tparm(bell));
	}

	switch (board)
	{
		case 147:
		{
			fprintf(stdout, mot147, BLANK);
			break;
		}

		case 372:
		{
			fprintf(stdout, motmap, BLANK);
			break;
		}

		case 374:
		{
			fprintf(stdout, mottop, BLANK);
			break;
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_invalid_sel()
**
**  DESCRIPTION:	This function writes a line under the menu that says the
**			input selection was incorrect.	
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_invalid_sel()

{
	putp(tparm(flash_screen));
	putp(tparm(bell));

	fprintf(stdout, "\n%25sINVALID SELECTION.\n\n", BLANK);

	scold_user = TRUE;
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_incorrect_sel()
**
**  DESCRIPTION:	This function writes a line under the menu that says the
**			input selection was incorrect.	
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_incorrect_sel()

{
	if (scold_user == TRUE)
	{
		scold_user = FALSE;

		fprintf(stdout, "%20sINCORRECT SELECTION -- PLEASE TRY AGAIN.\n", BLANK);
	}
	else
		fprintf(stdout, "\n");

	fprintf(stdout, "%31sEnter Selection:   ", BLANK);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wr_error()
**
**  DESCRIPTION:	This function will clear the screen and write
**			the passed error message then wait for a
**			keystroke.
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	none
**
**  SYSTEM GLOBALS:	scold_user	flag for error messages
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
wr_error(msg)

register BYTE *msg;

{
	scold_user = TRUE;

	clr();

	fprintf(stdout, msg, BLANK, errno);

	get_key();
}
