static	char ident[] = "@(#)nvram.c	35.3		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	nvram.c
**
**  DESCRIPTION:	This module contains code to manipulate NVRAM.
**
**  CONTENTS: 		do_nvram()	menu driver
**			str_to_hex	string to hex value
**			display_field()	displays one field and its value
**			display_all()	displays all NVRAM w/o field names
**			field_prompt()	prompts user for field value input
**			set_csum()	computes checksum for parameter
**			tot_csum()	computes checksum for all NVRAM
**			nvram_write()	posts epb with NVRAM values to board
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/board/nvram.h>
#include <sys/osi/board/mvme372.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>
#include "def.h"

#define	GET_CKSUM	0
#define	SET_CKSUM	1
#define	CMP_CKSUM	2


/*
**	All the strings printed by NVRAM.
*/

extern	BYTE	*exit_menu;
extern	BYTE	*eeprom_done;
extern	BYTE	*nvram_done;
extern	BYTE	*unchanged;
extern	BYTE	*ndx_prompt;
extern	BYTE	*len_prompt;
extern	BYTE	*bad_ndx;
extern	BYTE	*ndx_too_large;
extern	BYTE	*bad_len;
extern	BYTE	*bad_ndx_len;
extern	BYTE	*all_header;
extern	BYTE	*field_hdr;
extern	BYTE	*unspec_hdr;
extern	BYTE	*nvram_hdr;
extern	BYTE	*exit_hdr;
extern	BYTE	*bad_val;
extern	BYTE	*field_val1;
extern	BYTE	*field_val2;
extern	BYTE	*field372[];
extern	BYTE	*field374[];

/*
**	Station Manager Variables
*/

extern	BYTE	selection [80];
extern	BOOLEAN	scold_user;

/*
**	External Functions
*/

extern	DEFAULT	bld_epb();
extern	DEFAULT	wait_sm_response();
extern	VOID	get_key();
extern	VOID	wr_banner();
extern	VOID	nvram372();
extern	VOID	nvram374();
extern	VOID	menu372();
extern	VOID	menu374();
extern	BYTE	cgetchar();


/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	do_nvram ()
**
**  DESCRIPTION:	This is the menu driven function that oversees
**			processing for displaying and modifying NVRAM
**			on the board.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user input buffer
**			scold_user	boolean for error message displays
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
do_nvram ()

{
		VOID	display_all();
		VOID	display_field();
		VOID	field_prompt();
		BYTE	set_csum();
		BYTE	tot_csum();
		BOOLEAN	nvram_write();

	register BYTE	*l = selection;
	register SM_EPB	*epb;

		BYTE	*ptr;
		UBYTE	 csum;
		DEFAULT	 field;
		DEFAULT	 field_ndx;
		DEFAULT	 field_len;
		DEFAULT	 n;
		DEFAULT	 byte_ndx;
		BOOLEAN	 pri_flag;
		BOOLEAN	 bad_ans;

		SM_EPB	 sm_epb;

	/*
	** Build an sm_epb and post it to Station Manager on the board.
	*/

	if (bld_epb((epb = &sm_epb), SM_NVRAM_RD_REQUEST, SM_QID) == FAILURE)
		return;

	/*
	** WAIT for response back from Station Manager on the board.
	*/

	if (wait_sm_response(epb, SM_NVRAM_RD_CONFIRM) == FAILURE)
		return;

	/*
	** Loop until user quits editor.
	*/

	while (FOREVER)
	{
		/*
		**	Display main NVRAM Editor
		*/

		switch (epb->board)
		{
			case 372:
			{
				menu372(epb);
				break;
			}

			case 147:
			case 374:
			{
				menu374(epb);
				break;
			}
		}

		/*
		** get response to editor
		*/

		cgetchar();

		/*
		** If user wants to QUIT, display the Exit Menu, then
		** return to caller when response is correct.
		*/

		if ((*l == 'Q') || (*l == 'q'))
		{
			bad_ans = TRUE;

			while (bad_ans)
			{
				wr_banner();

				fprintf(stdout, exit_hdr);
				fprintf(stdout, exit_menu);

				if (scold_user == TRUE)
				{
					scold_user = FALSE;

					fprintf(stdout, "\n\n%20sINCORRECT SELECTION -- PLEASE TRY AGAIN.\n", BLANK);
					fprintf(stdout, "%25sEnter Selection ==>  ", BLANK);
				}
				else
					fprintf(stdout, "\n\n\n%25sEnter Selection ==>  ", BLANK);

				cgetchar();

				if (*l == NULL)
					break;

				switch (*l)
				{
					/*
					** Change NVRAM values.
					*/

					case 'a':
					case 'A':
					{
						bad_ans = FALSE;

						epb->nvram_write = NVRAM_ONLY;

						if (nvram_write(epb) != FAILURE)
							fprintf(stdout, nvram_done);
						break;
					}

					/*
					** Change NVRAM and EEPROM.
					*/

					case 'b':
					case 'B':
					{
						bad_ans = FALSE;

						epb->nvram_write = NVRAM_EEPROM;

						if (nvram_write(epb) != FAILURE)
							fprintf(stdout, eeprom_done);
						break;
					}

					/*
					** Leave NVRAM and EEPROM unchanged.
					*/

					case 'c':
					case 'C':
					{
						return;
					}

					default:
					{
						scold_user = TRUE;

						continue;
					}
				}
			}

			if (bad_ans)
			{
				fprintf(stdout, "\n\n");
				continue;
			}

			get_key ();

			return;
		}
		else
		{
			fprintf(stdout, "\n");

			pri_flag = FALSE;

			switch (*l)
			{
				/*
				** Display all of NVRAM.
				*/

				case 'u':
				case 'U':
				{
					display_all(epb->nvram);

					bad_ans = TRUE;

					while (bad_ans)
					{
						fprintf(stdout, ndx_prompt);

						gets(l);

						if (*l == NULL)
							break;

						n = strtol(l, &ptr, 0);

						if (*ptr == NULL)
							bad_ans = FALSE;
						else
							fprintf(stdout, bad_ndx);
						/*
						** Leave room for total checksum
						** and last field nibble
						** checksum. Remember the
						** minimum field size is 1 byte
						** (2 nibbles).
						*/

						if (n > (NVRAM_SIZE - 5))
						{
							fprintf(stdout, ndx_too_large);

							bad_ans = TRUE;

							continue;
						}
					}

					if (bad_ans)
					{
						fprintf(stdout, "\n\n");
						continue;
					}

					field_ndx = n;

					bad_ans = TRUE;

					while (bad_ans)
					{
						fprintf(stdout, len_prompt);

						gets(l);

						if (*l == NULL)
							break;

						n = strtol(l, &ptr, 0);

						if (*ptr == NULL)
							bad_ans = FALSE;
						else
							fprintf(stdout, bad_len);
					}

					if (bad_ans)
					{
						fprintf(stdout, "\n\n");
						continue;
					}

					field_len = n;

					if ((field_ndx + (2 * field_len) + 1) > (NVRAM_SIZE - 2))
					{
						fprintf(stdout, bad_ndx_len);
						continue;
					}

					field = UNSPEC;

					break;
				}

				default:
				{
					scold_user = FALSE;

					switch (epb->board)
					{
						case 372:
						{
							nvram372(*l, &field, &field_ndx, &field_len);
							if (field == TX_PRTY)
								pri_flag = TRUE;

							break;
						}

						case 147:
						case 374:
						{
							nvram374(*l, &field, &field_ndx, &field_len);
							break;
						}
					}
				}
			}

			if (scold_user ==  TRUE)
				continue;

			/*
			**	Remove good/bad flag from csum.
			*/

			csum &= 0x0f;

			if (field != UNSPEC)
			{
				wr_banner();
				fprintf(stdout, field_hdr);
			}

			/*
			**	Display current value.
			*/

			display_field(field, field_ndx, field_len, epb);

			bad_ans = TRUE;

			while (bad_ans)
			{
				/*
				**	Get new value.
				*/

				field_prompt(field_len);

				gets(l);

				if (*l == NULL)
					break;

				/*
				**	Convert value.
				*/

				if (pri_flag == TRUE)
				{
					n = strtol (l, &ptr, 0);

					if ((*ptr != NULL) || (n < 0) || (n > 7))
					{
						fprintf(stdout, bad_val);
						continue;
					}

					pri_flag = FALSE;
				}

				byte_ndx = field_ndx >> 1;

				if (str_to_hex(l, (field_len * 2), (epb->nvram + byte_ndx), field_ndx) < 0)
				{
					fprintf(stdout, bad_val);
					continue;
				}

				/*
				**	Compute checksums.
				*/

				set_csum(epb->nvram, field_ndx, field_len, SET_CKSUM);

				tot_csum(epb->nvram, SET_CKSUM);

				bad_ans = FALSE;
			}
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	str_to_hex ()
**
**  DESCRIPTION:	Converts a string to a hex integer for display.
**
**  PARAMETERS:		
**
**	Input:		s	input buffer
**			l	number of characters to convert
**			ndx	offset to first nibble
**
**	Output:		v	destination for converted value
**
**	Returns:	 0	SUCCESS
**			-1	FAILURE
**
**  SYSTEM GLOBALS:	none
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

DEFAULT
str_to_hex (s, l, v, ndx)

register BYTE	 *s;	/* string address			*/
register DEFAULT l;	/* length in nibbles			*/
register UBYTE 	 *v;	/* address of 1st byte to process	*/
register DEFAULT ndx;	/* offset to first nibble		*/

{
	register UBYTE	 n;		/* nibble value */
	register BOOLEAN hi = TRUE;	/* hi or lo nibble indicator */

	if ((ndx & 0x01) == 1)	/* odd index starts at low nibble	*/
		hi = FALSE;

	if ((*s == '0') && ((s[1] == 'x') || (s[1] == 'X')))
		s += 2;
	
	while (*s && l--)
	{
		if ((*s >= '0') && (*s <= '9'))
			n = *s++ - '0';
		else
			if ((*s >= 'a') && (*s <= 'f'))
				n = *s++ - 'a' + 10;
			else
				if ((*s >= 'A') && (*s <= 'F'))
					n = *s++ - 'A' + 10;
				else
					return (-1);

		if (hi)
			*v = n << 4;
		else
		{
			*v = (*v & 0xf0) | (n & 0xf);
			++v;
		}

		hi = (hi ? FALSE : TRUE);
	}

	if (*s || l)
		return (-1);

	return (0);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	display_field ()
**
**  DESCRIPTION:	Displays one field and its value on screen or
**			all of NVRAM.
**
**  PARAMETERS:		
**
**	Input:		field		field name
**			field_ndx	offset into NVRAM
**			field_len	length of field in bytes
**			epb		address of SM_EPB
**
**	Output:		line in NVRAM menu
**
**	Returns:	nothing
**
**  SYSTME GLOBALS:	none
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
display_field(field, field_ndx, field_len, epb)

register DEFAULT field;
register DEFAULT field_ndx;
register DEFAULT field_len;
register SM_EPB	 *epb;

{
	BYTE	set_csum();

	register UBYTE	checksum;	/* valid checksum flag */
	register UBYTE	*nvram;		/* NVRAM addressm */

	/*
	**	Display all of NVRAM.
	*/

	if (field == UNSPEC)
	{
		wr_banner();

		checksum = 0;

		fprintf(stdout, nvram_hdr);
		fprintf(stdout, unspec_hdr);
		fprintf(stdout, "                       %3d     %3d     0x", field_ndx, field_len);
	}
	else
	{
		/*
		**	Display major parameters.
		*/

		switch (epb->board)
		{
			case 372:
			{
				fprintf(stdout, "%s", field372[field]);
				break;
			}

			case 147:
			case 374:
			{
				fprintf(stdout, "%s", field374[field]);
				break;
			}
		}

		if (checksum = set_csum(epb->nvram, field_ndx, field_len, CMP_CKSUM))
			fprintf(stdout, "%3d     %3d        *", field_ndx, field_len);
		else
			fprintf(stdout, "%3d     %3d        0x", field_ndx, field_len);
	}

	/*
	**	NVRAM address plus offset to field.
	*/

	nvram = (UBYTE *)(epb->nvram + (field_ndx >> 1));

	field_len = field_len << 1;

	if ((field_ndx & 0x01) != 0)
	{
		if (checksum == 0)
		{
			fprintf(stdout, "%1x", *(nvram) & 0xf);
			++nvram;
		}

		--field_len;
	}

	/*
	**	Print each byte in the field.
	*/

	while (field_len)
	{
		if (checksum == 0)
			fprintf(stdout, "%1x", (*nvram >> 4) & 0xf);

		if (--field_len)
		{
			if (checksum == 0)
			{
				fprintf(stdout, "%1x", *(nvram) & 0xf);
				++nvram;
			}

			--field_len;
		}
	}

	fprintf(stdout, "\n");
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	display_all ()
**
**  DESCRIPTION:	Display all of NVRAM.
**
**  PARAMETERS:		
**
**	Input:		nvram	address of NVRAM
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
display_all(nvram)

register UBYTE *nvram;

{
	register LONG	  length = (NVRAM_SIZE / 2) - 1;
	register ULONG	  offset = 0;
	register DEFAULT  i;

	wr_banner();

	fprintf(stdout, all_header);

	while (length >= 0)
	{
		fprintf(stdout, "    %3d   ", offset);

		for (i = 0; i < (length < 15 ? length : 15); ++i, ++nvram)
			fprintf(stdout, "%1x %1x ", (*nvram >> 4) & 0x0f, *nvram & 0xf);

		offset += 30;
		length -= 15;

		fprintf(stdout, "\n");
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	field_prompt ()
**
**  DESCRIPTION:	Prompts for user input for field.
**
**  PARAMETERS:		
**
**	Input:		len	length of field in bytes
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	none
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
field_prompt (len)

register DEFAULT len;

{
	fprintf(stdout, field_val1);
	fprintf(stdout, "%d", (len * 2));
	fprintf(stdout, field_val2);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	set_csum()
**
**  DESCRIPTION:	Computes checksum for NVRAM parameter.
**
**  PARAMETERS:		
**
**	Input:		nvram		data
**			index		offset to field in nvram
**			size		size of field in bytes
**			flag		whether or not to set checksum
**
**	Output:		none
**
**	Returns:	csum		checksum
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

BYTE
set_csum (nvram, index, size, flag)

register UBYTE	 *nvram;
register DEFAULT index;
register DEFAULT size;
register UBYTE	 flag;

{
	register BOOLEAN hi;
	register BYTE    csum;

	csum 	= -1;
	nvram 	= nvram + (index >> 1);
	size 	= size << 1;
	hi	= TRUE;

	if ((index & 0x01) == 0x01)
		hi = FALSE;

	/*
	**	XOR adjacent nibbles one nibble at a time.
	*/

	while (size--)
	{
		if (hi)
			csum ^= (*nvram >> 4);
		else
		{
			csum ^= (*nvram & 0x0f);
			++nvram;
		}

		hi = (hi ? FALSE : TRUE);
	}

	/*
	**	Execute the following code if we are
	**	only doing a checksum compare.
	*/

	if (flag == CMP_CKSUM)
	{
		if ((index & 0x01) == 0x00)
			return ((*nvram & 0xf0) ^ (csum << 4));
		else
			return ((*nvram & 0x0f) ^ (csum & 0x0f));
	}

	/*
	**	Execute the following code if we are
	**	are not storing the checksum.
	*/

	if (flag != GET_CKSUM)
	{
		if ((index & 0x01) == 0x00)
			*nvram = (*nvram & 0x0f) | ((csum << 4) & 0xf0);
		else
			*nvram = (*nvram & 0xf0) | (csum & 0x0f);
	}

	return (csum);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	tot_csum()
**
**  DESCRIPTION:	Computes checksum for all of NVRAM.
**
**  PARAMETERS:		
**
**	Input:		nvram		address of NVRAMdata
**			flag		whether or not to set checksum
**
**	Output:		none
**
**	Returns:	csum		checksum
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

BYTE
tot_csum (nvram, flag)

register UBYTE	*nvram;
register UBYTE	 flag;
{
	register DEFAULT i;
	register BYTE	 csum;

	csum = -1;

	for (i = 0; i < (NVRAM_SIZE / 2) - 1; ++i)
	{
		csum ^= *nvram;
		++nvram;
	}

	if (flag != GET_CKSUM)
		*nvram = csum;

	return (csum);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	nvram_write()
**
**  DESCRIPTION:	Call a routine which will build an epb and
**			post it to the Station Manager on the board.
**
**  PARAMETERS:		
**
**	Input:		sm_epb	station manager epb with nvram data in body
**
**	Output:		nothing
**
**	Returns:	SUCCESS/FAILURE
**
**  SYSTEM GLOBALS:	none
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

BOOLEAN
nvram_write (sm_epb)

register SM_EPB	*sm_epb;

{
	/*
	** Build an sm_epb and post it to Station Manager on the board.
	*/

	if (bld_epb(sm_epb, SM_NVRAM_WR_REQUEST, SM_QID) != SUCCESS)
		return (FAILURE);

	return (SUCCESS);
}
