static	char ident[] = "@(#)nvram372.c	35.1		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	nvram372.c
**
**  DESCRIPTION:	This module contains code to manipulate MVME372 NVRAM.
**
**  CONTENTS: 		menu372()	menu driver
**			nvram372()	sets the field variables
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/board/mvme372.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>
#include "def.h"

/*
**	All the strings printed by MVME372 NVRAM
*/

extern	BYTE	*menu_prompt;
extern	BYTE	*field_hdr;

/*
**	Station Manager Variables
*/

extern	BOOLEAN	scold_user;

/*
**	External Functions
*/

extern	VOID	display_field();

/*
**	MVME372 NVRAM Menu
*/

BYTE *field372[] =
{
	"     A. Board Base Address        ",
	"     B. Station Address           ",
	"     C. Group Address Mask        ",
	"     D. Individual Address Mask   ",
	"     E. Slot Time Value           ",
	"     F. Modem Selection Value     ",
	"     G. MAC Transmission Priority ",
	"     H. No Stat Tracking(0=N,1=Y) ",
	"     I. In Ring Desired (0=N,1=Y) ",
	"     J. Prescaler Mode(0->6,1->3) ",
	"     K. High Priority THT         ",
	"     L. TRT(4)                    ",
	"     M. TRT(2)                    ",
	"     N. TRT(0)                    ",
	"     O. TRT Ring Maintenance      ",
	"     P. RM Initial Value          ",
	"     R. Max Inter Solicit Count   ",
	"     S. PTP Register Value        "
};

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	menu372()
**
**  DESCRIPTION:	This is the menu driven function that displays
**			the MVME372 NVRAM fields.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
menu372(epb)

register SM_EPB *epb;
{
	putp(tparm(clear_screen));

	fprintf(stdout, field_hdr);

	display_field(BASE_ADDR, BASE_ADDR_INDEX, BASE_ADDR_SIZE, epb);
	display_field(TS_ADDR, TS_ADDR_INDEX, TS_ADDR_SIZE, epb);
	display_field(GA_MASK, GA_MASK_INDEX, GA_MASK_SIZE, epb);
	display_field(IA_MASK, IA_MASK_INDEX, IA_MASK_SIZE, epb);
	display_field(SLOT_TIME, SLOT_TIME_INDEX, SLOT_TIME_SIZE, epb);
	display_field(MODEM, MODEM_INDEX, MODEM_SIZE, epb);
	display_field(TX_PRTY, TX_PRTY_INDEX, TX_PRTY_SIZE, epb);
	display_field(N_NSTT, N_NSTT_INDEX, N_NSTT_SIZE, epb);
	display_field(N_INRD, N_INRD_INDEX, N_INRD_SIZE, epb);
	display_field(N_PS3, N_PS3_INDEX, N_PS3_SIZE, epb);
	display_field(N_HI_PRI_THT, N_HI_PRI_THT_INDEX, N_HI_PRI_THT_SIZE, epb);
	display_field(N_TAR_ROT_4, N_TAR_ROT_4_INDEX, N_TAR_ROT_4_SIZE, epb);
	display_field(N_TAR_ROT_2, N_TAR_ROT_2_INDEX, N_TAR_ROT_2_SIZE, epb);
	display_field(N_TAR_ROT_0, N_TAR_ROT_0_INDEX, N_TAR_ROT_0_SIZE, epb);
	display_field(N_TAR_ROT_RM, N_TAR_ROT_RM_INDEX, N_TAR_ROT_RM_SIZE, epb);
	display_field(N_RM_INT, N_RM_INT_INDEX, N_RM_INT_SIZE, epb);
	display_field(N_MAX_INT_SOL, N_MAX_INT_SOL_INDEX, N_MAX_INT_SOL_SIZE, epb);
	display_field(N_PTP, N_PTP_INDEX, N_PTP_SIZE, epb);

	fprintf(stdout, menu_prompt);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	nvram372()
**
**  DESCRIPTION:	This function will set the field variables for
**			the passed menu selection value.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
nvram372(l, field, field_ndx, field_len)

register BYTE	 l;
register DEFAULT *field;
register DEFAULT *field_ndx;
register DEFAULT *field_len;
{
	switch (l)
	{
		/*
		**	Board base address.
		*/

		case 'a':
		case 'A':
		{
			*field     = BASE_ADDR;
			*field_ndx = BASE_ADDR_INDEX;
			*field_len = BASE_ADDR_SIZE;

			break;
		}

		/*
		**	Station address.
		*/

		case 'b':
		case 'B':
		{
			*field     = TS_ADDR;
			*field_ndx = TS_ADDR_INDEX;
			*field_len = TS_ADDR_SIZE;

			break;
		}

		/*
		**	Group address mask.
		*/

		case 'c':
		case 'C':
		{
			*field     = GA_MASK;
			*field_ndx = GA_MASK_INDEX;
			*field_len = GA_MASK_SIZE;

			break;
		}

		/*
		**	Individual address mask.
		*/

		case 'd':
		case 'D':
		{
			*field     = IA_MASK;
			*field_ndx = IA_MASK_INDEX;
			*field_len = IA_MASK_SIZE;

			break;
		}

		/*
		**	Slot time value.
		*/

		case 'e':
		case 'E':
		{
			*field     = SLOT_TIME;
			*field_ndx = SLOT_TIME_INDEX;
			*field_len = SLOT_TIME_SIZE;

			break;
		}

		/*
		**	Modem selection.
		*/

		case 'f':
		case 'F':
		{
			*field     = MODEM;
			*field_ndx = MODEM_INDEX;
			*field_len = MODEM_SIZE;

			break;
		}

		/* 
		**	Mac transmit priority.
		*/

		case 'g':
		case 'G':
		{
			*field     = TX_PRTY;
			*field_ndx = TX_PRTY_INDEX;
			*field_len = TX_PRTY_SIZE;

			break;
		}

		/*
		**	No statistics tracking.
		*/

		case 'h':
		case 'H':
		{
			*field     = N_NSTT;
			*field_ndx = N_NSTT_INDEX;
			*field_len = N_NSTT_SIZE;

			break;
		}

		/*
		**	In ring desired.
		*/

		case 'i':
		case 'I':
		{
			*field     = N_INRD;
			*field_ndx = N_INRD_INDEX;
			*field_len = N_INRD_SIZE;

			break;
		}

		/*
		**	Prescaler.
		*/

		case 'j':
		case 'J':
		{
			*field     = N_PS3;
			*field_ndx = N_PS3_INDEX;
			*field_len = N_PS3_SIZE;

			break;
		}

		/*
		**	High Priority token hold time.
		*/

		case 'k':
		case 'K':
		{
			*field     = N_HI_PRI_THT;
			*field_ndx = N_HI_PRI_THT_INDEX;
			*field_len = N_HI_PRI_THT_SIZE;

			break;
		}

		/*
		**	Target Rotation Time (4).
		*/

		case 'l':
		case 'L':
		{
			*field     = N_TAR_ROT_4;
			*field_ndx = N_TAR_ROT_4_INDEX;
			*field_len = N_TAR_ROT_4_SIZE;

			break;
		}

		/*
		**	Target Rotation Time (2).
		*/

		case 'm':
		case 'M':
		{
			*field     = N_TAR_ROT_2;
			*field_ndx = N_TAR_ROT_2_INDEX;
			*field_len = N_TAR_ROT_2_SIZE;

			break;
		}

		/*
		**	Target Rotation Time (0).
		*/

		case 'n':
		case 'N':
		{
			*field     = N_TAR_ROT_0;
			*field_ndx = N_TAR_ROT_0_INDEX;
			*field_len = N_TAR_ROT_0_SIZE;

			break;
		}

		/*
		**	Target rotation time (ring maintenance).
		*/

		case 'o':
		case 'O':
		{
			*field     = N_TAR_ROT_RM;
			*field_ndx = N_TAR_ROT_RM_INDEX;
			*field_len = N_TAR_ROT_RM_SIZE;

			break;
		}

		/*
		**	Ring maintenance initial value.
		*/

		case 'p':
		case 'P':
		{
			*field     = N_RM_INT;
			*field_ndx = N_RM_INT_INDEX;
			*field_len = N_RM_INT_SIZE;

			break;
		}

		/*
		**	Maximum inter solicit count.
		*/

		case 'r':
		case 'R':
		{
			*field     = N_MAX_INT_SOL;
			*field_ndx = N_MAX_INT_SOL_INDEX;
			*field_len = N_MAX_INT_SOL_SIZE;

			break;
		}

		/*
		**	Pad timer preset register.
		*/

		case 's':
		case 'S':
		{
			*field     = N_PTP;
			*field_ndx = N_PTP_INDEX;
			*field_len = N_PTP_SIZE;

			break;
		}

		default:
		{
			scold_user = TRUE;
		}
	}
}
