#ident "@(#)read.c	35.3"

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	read.c
**
**  DESCRIPTION:	This module contains all the functions which
**			read user input from the terminal.
**
**  CONTENTS:		read_selection() reads input from menu selection
**			cgetchar()	 gets 1 character from the screen
**
**  SYSTEM GLOBALS:	scold_user	boolean to control error messages
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <termio.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/common/types.h>
#include "def.h"

/*
**	External Functions
*/

extern	DEFAULT	catchit();

/*
**	External Variables
*/

extern	USHORT	 board;
extern	BOOLEAN	scold_user;
extern	BYTE	selection[80];

struct	termio	oldterm;	/* original terminal status	*/
struct	termio	nowterm;	/* new terminal status		*/

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	read_selection()
**
**  DESCRIPTION:	This function reads the user selection from the menu.
**
**  PARAMETERS:		
**	Input:		menu_name	name of menu	
**
**	Output:		none
**
**	Returns:	SUCCESS/FAILURE
**
**  SYSTEM GLOBAL:	scold_user	flag for error messages
**			selection	user's input buffer
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

BOOLEAN
read_selection(menu_name)

register BYTE	menu_name;

{
	BYTE	cgetchar();

	/*
	** Get the character input from the user and put it in selection[].
	*/

	cgetchar();

	/*
	** Change upper case letter to a lower case letter
	*/

	selection[0] = tolower(selection[0]);

	/* 
	** If user wants to QUIT, call routine to reset terminal and
	** exit from program.
	*/

	if (selection[0] == QUIT)
		catchit();

	/*
	** If user wants to execute a shell command, call system,
	** else check on user selection.
	*/

	if (selection[0] != SHELL)
		switch (menu_name)
		{
			/*
			**	L E V E L   1   M E N U
			*/

			case MAIN_MENU:
			{
				switch (board)
				{
					case 147:
					{
						if ((selection[0] != CONFIGURATION_UTILITIES)
						&&  (selection[0] != DIAGNOSTIC_UTILITIES))
						{
							scold_user   = TRUE;
							selection[0] = NULL;

							return (FAILURE);
						}

						break;
					}

					case 372:
					case 374:
					{
						if ((selection[0] != NODE_RESTART)
						&&  (selection[0] != CONFIGURATION_UTILITIES)
						&&  (selection[0] != DIAGNOSTIC_UTILITIES))
						{
							scold_user   = TRUE;
							selection[0] = NULL;

							return (FAILURE);
						}

						break;
					}
				}

				break;
			}

			/*
			**	L E V E L   2   M E N U S
			*/

			case CONFIGURATION_MENU:
			{
				switch (board)
				{
					case 147:
					{
						if (((selection[0] < NVRAM)
						||   (selection[0] > MAC_PARMS))
						&&  (selection[0] != MAIN))
						{
							scold_user   = TRUE;
							selection[0] = NULL;

							return (FAILURE);
						}

						break;
					}

					case 372:
					case 374:
					{
						if (((selection[0] < NVRAM)
						||   (selection[0] > TRANSPORT_PARMS))
						&&  (selection[0] != MAIN))
						{
							scold_user   = TRUE;
							selection[0] = NULL;

							return (FAILURE);
						}

						break;
					}
				}

				break;
			}

			case DIAGNOSTIC_MENU:
			{
				switch (board)
				{
					case 147:
					{
						if (((selection[0] < MAC_STATS)
						||   (selection[0] > LOGGING))
						&&  (selection[0] != MAIN))
						{
							scold_user   = TRUE;
							selection[0] = NULL;

							return (FAILURE);
						}

						break;
					}

					case 372:
					case 374:
					{
						if (((selection[0] < MAC_STATS)
						||   (selection[0] > TRMON))
						&&  (selection[0] != MAIN))
						{
							scold_user   = TRUE;
							selection[0] = NULL;

							return (FAILURE);
						}

						break;
					}
				}

				break;
			}

			/*
			**	L E V E L   3   M E N U S
			*/

			case LOG_MENU:
			{
				switch (board)
				{
					case 147:
					{
						switch (selection[0])
						{
							case 'a':
							{
								selection[0] = SYSTEM_LOGGING;
								break;
							}

							case 'b':
							{
								selection[0] = MAC_LOGGING;
								break;
							}

							case 'c':
							{
								selection[0] = STA_MGR_LOGGING;
								break;
							}

							case 'd':
							{
								selection[0] = NO_CHANGE;
								break;
							}

							case 'e':
							{
								selection[0] = SET_LOG_MASK;
								break;
							}

							default:
							{
								scold_user   = TRUE;
								selection[0] = NULL;

								return (FAILURE);
							}
						}

						break;
					}

					case 372:
					case 374:
					{
						if ((selection[0] < HOST_LOGGING)
						||  (selection[0] > SET_LOG_MASK))
						{
							scold_user   = TRUE;
							selection[0] = NULL;

							return (FAILURE);
						}

						break;
					}
				}

				break;
			}

			case TRMON_MENU:
			{
				if (((selection[0] < M_STRT)
				||   (selection[0] > M_WRTE))
				&&  (selection[0] != DIAGNOSTIC_UTILITIES)
				&&  (selection[0] != MAIN))
				{
					scold_user   = TRUE;
					selection[0] = NULL;

					return (FAILURE);
				}

				break;
			}

			case SYSTEM_PARMS_MENU:
			{
				if (((selection[0] < BDB_COUNT)
				||   (selection[0] > NETW_LIMIT))
				&&  (selection[0] != CONFIGURATION_UTILITIES)
				&&  (selection[0] != MAIN))
				{
					scold_user   = TRUE;
					selection[0] = NULL;

					return (FAILURE);
				}

				break;
			}

			case XPORT_PARMS_MENU:
			{
				if (((selection[0] < MAX_CONNECTS)
				||   (selection[0] > REF_WAIT_TICKS))
				&&  (selection[0] != CONFIGURATION_UTILITIES)
				&&  (selection[0] != MAIN))
				{
					scold_user   = TRUE;
					selection[0] = NULL;

					return (FAILURE);
				}

				break;
			}

			case LANCE_PARMS_MENU:
			{
				if (((selection[0] < RING_SIZE)
				||   (selection[0] > GROUP_ADDRESS))
				&&  (selection[0] != CONFIGURATION_UTILITIES)
				&&  (selection[0] != MAIN))
				{
					scold_user   = TRUE;
					selection[0] = NULL;

					return (FAILURE);
				}

				break;
			}

			case TBC_PARMS_MENU:
			{
				if ((selection[0] != TBC_PRIORITY)
				&&  (selection[0] != CONFIGURATION_UTILITIES)
				&&  (selection[0] != MAIN))
				{
					scold_user   = TRUE;
					selection[0] = NULL;

					return (FAILURE);
				}

				break;
			}
		}

	return (SUCCESS);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	cgetchar()
**
**  DESCRIPTION:
**	Gets a single character from the terminal and put it in selection[].
**	The character is echoed at the current character cursor position.  The
**	character cursor is moved to the next position.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user input buffer which gets character
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

BYTE
cgetchar()

{
	/*
	** Set up terminal for single character input:
	*/

	ioctl(0, TCGETA, &oldterm);
	ioctl(0, TCGETA, &nowterm);

	nowterm.c_iflag &= (~IGNBRK & ~ICRNL & ~INLCR);
	nowterm.c_iflag |= BRKINT;
	nowterm.c_lflag &= (~ISIG & ~ECHO & ~ECHONL & ~ICANON);
	nowterm.c_cc[4]  = 1;
	nowterm.c_cc[5]  = 1;

	ioctl(0, TCSETAF, &nowterm);

	/*
	** Get a single character:
	*/

	selection[0] = getchar();

	/*
	** If it's a printable character, echo it:
	*/
	
	if ((' ' <= selection[0]) && (selection[0] <= '~'))
		putchar(selection[0]);

	/*
	** Put the terminal back into previous mode:
	*/

	ioctl(0, TCSETA, &oldterm);

	/*
	** Return the acquired character:
	*/

	return (selection[0]);
}
