static	char ident[] = "@(#)sys.c	35.1		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	sys.c
**
**  DESCRIPTION:	This module contains the functions that process
**			the node restart.
**
**  CONTENTS:		node_restart()	resets and downloads uMAP37x
**
**  SYSTEM GLOBALS:	none
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**********************************************************************/

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>
#include "def.h"

extern	VOID	get_key();
extern	VOID	clr();
extern	VOID	wr_error();
extern	DEFAULT	system();
extern	DEFAULT	dl();
extern	DEFAULT	rt();
extern	DEFAULT	bld_epb();
extern	DEFAULT	ci_register();
extern	DEFAULT	ci_deregister();

extern	USHORT	board;
extern	USHORT	dst_cpu;
extern	BOOLEAN	scold_user;
extern	DEFAULT	errno;
extern	DEFAULT	ci_fd;
extern	ULONG	src_id;

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	node_restart()
**
**  DESCRIPTION:	This function resets the MVME37x, kills the
**			demons, resets the Common Environment,
**			downloads/executes MicroMAP, starts the demons
**			and turns off logging.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	ci_fd	file descriptor for common environment
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
node_restart()

{
	register SM_EPB		*sm_epb;

		 SM_EPB		 stat_epb;
		 BYTE		 dl_file[128];
		 BYTE		 sys_buf[128];
		 DEFAULT	 n;

	 /*
	 **	Get file name of object file.
	 */

	fprintf(stdout, "\n\n%25sEnter file name for downloading.\n", BLANK);

	switch (board)
	{	
		case 372:
		{
			fprintf(stdout, "%25s(Carriage return downloads uMAP372.)\n%25s", BLANK, BLANK);
			break;
		}

		default:
		{
			fprintf(stdout, "%25s(Carriage return downloads uTOP374.)\n%25s", BLANK, BLANK);
			break;
		}
	}

	gets(dl_file);

	/*
	**	If carriage return is entered, assume 
	**	default.
	*/

	if (dl_file[0] == NULL)
		switch (board)
		{	
			case 372:
			{
				strcpy(dl_file, "uMAP372");
				break;
			}

			default:
			{
				strcpy(dl_file, "uTOP374");
				break;
			}
		}

	/*
	**	Reset the board.
	*/

	fprintf(stdout, "\n%25sReset in progress, please wait.\n", BLANK);

	/*
	**	Terminate map_demon.
	*/

	fprintf(stdout, "%25sTerminating map_demon.\n", BLANK);

	sprintf(sys_buf, "kill -9 `ps -ef | fgrep \"map_demon / /map/demonlog_%d %d\" | cut -c9-14` 2>/dev/null", dst_cpu, dst_cpu);

	if (system(sys_buf) < 0)
	{
		wr_error("\n\n%25sUnable to terminate map_demon, errno = %d\n");

		return;
	}

	/*
	**	Terminate fvf.
	*/

	fprintf(stdout, "%25sTerminating fvf.\n", BLANK);

	sprintf(sys_buf, "kill -9 `ps -ef | fgrep \"fvf uMAP.fvf /map/public /map/fvflog_%d %d\" | cut -c9-14` 2>/dev/null", dst_cpu, dst_cpu);

	if (system(sys_buf) < 0)
	{
		wr_error("\n\n%25sUnable to terminate fvf, errno = %d\n");

		return;
	}

	/*
	**	Reset the Common Environment.
	*/

	fprintf(stdout, "%25sResetting Common Environment.\n", BLANK);

	if ((n = ioctl(ci_fd, RESET, 0)) < 0)
	{
		wr_error("\n\n%25sUnable to reset Common Environment, errno = %d\n");

		return;
	}

	/*
	**	Register Station Manager Console with the common environment.
	*/

	src_id = ID_NUM;

	if ((n = ci_register(dst_cpu, &src_id, &ci_fd)) != 0)
	{
		wr_error("\n\n%25sFailed to register, errno = %d\n");

		return;
	}

	if ((n = ioctl(ci_fd, WAIT, 0)) < 0)
	{
		wr_error("\n\n%25sUnable to set ioctl to BLOCKED reads, errno = %d\n");

		return;
	}

	/*
	**	Download file to board.
	*/

	fprintf(stdout, "%25sDownloading ", BLANK);

	fflush(stdout);

	sprintf(sys_buf, "dl %s %d", dl_file, dst_cpu);

	if (system(sys_buf) < 0)
	{
		sprintf(sys_buf, "\n\nUnable to download %s, errno = %d\n", dl_file);

		wr_error(sys_buf);

		return;
	}

	/*
	**	Link the directory services and route table files.
	*/

	sprintf(sys_buf, "ln /map/uMAP_%d.dir /map/uMAP.dir 2>/dev/null", dst_cpu);

	if (system(sys_buf) < 0)
	{
		wr_error("\n\n%25sUnable to link uMAP.dir, errno = %d\n");

		return;
	}

	sprintf(sys_buf, "ln /map/uMAP_%d.rte /map/uMAP.rte 2>/dev/null", dst_cpu);

	if (system(sys_buf) < 0)
	{
		wr_error("\n\n%25sUnable to link uMAP.rte, errno = %d\n");

		return;
	}

	sprintf(sys_buf, "ln /map/uMAP_%d.od /map/uMAP.od 2>/dev/null", dst_cpu);

	if (system(sys_buf) < 0)
	{
		wr_error("\n\n%25sUnable to link uMAP.od, errno = %d\n");

		return;
	}

	/*
	**	Start the map demon.
	*/

	sprintf(sys_buf, "map_demon / /map/demonlog_%d %d", dst_cpu, dst_cpu);

	if (system(sys_buf) != 0)
	{
		wr_error("\n\n%25sUnable to start map_demon and fvf_demon, errno = %d\n");

		return;
	}

	/*
	**	Start the fvf demon.
	*/

	sleep (1);

	sprintf(sys_buf, "fvf uMAP.fvf /map/public /map/fvflog_%d %d", dst_cpu, dst_cpu);

	if (system(sys_buf) != 0)
	{
		wr_error("\n\n%25sUnable to start fvf_demon, errno = %d\n");

		return;
	}

	/*
	**	Download the Network Routing Table.
	*/


	sprintf(sys_buf, "rt %d uMAP.rte", dst_cpu); 

	if (system(sys_buf) < 0)
	{
		wr_error("\n\n%25sUnable to download uMAP.rte, errno = %d\n");

		return;
	}

	/*
	**	Fix an EPB to turn off logging (log_mask = 0). Send
	**	it to the board.
	*/

	(sm_epb = &stat_epb)->log_mask = 0;

	if (bld_epb(sm_epb, SM_LOG_MASK_SET_REQUEST, SM_QID) == FAILURE)
		wr_error("\n\n%25sUnable to turn off logging, errno = %d\n");
}
