static	char ident[] = "@(#)util.c	35.1		5/18/88	Copyright MOTOROLA, INC.";

/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	util.c
**
**  DESCRIPTION:	This module contains miscellaneous functions.
**
**  CONTENTS:		clr()	           clears screen, can flash & ring bell
**			get_key()          asks & waits for signal from keyboard
**			ex_shell_cmd()     lets user do shell commands
**			bld_epb()          builds sm_epb and posts it to board
**			wait_sm_response() wait for response from board
**
**  SYSTEM GLOBALS:	none
**
**  MODULE GLOBALS:	none
**
**  NOTES:
**
**********************************************************************/

#include <stdio.h>
#include <curses.h>
#include <term.h>
#include <sys/c37x/ci_driver.h>
#include <sys/osi/common/types.h>
#include <sys/osi/common/selsizes.h>
#include <sys/osi/common/system.h>
#include <sys/osi/common/epb.h>
#include <sys/osi/common/sm_epb.h>
#include "def.h"

#define	RD_SIZE	(MAX_CEEPB_SIZE + MAX_BUF_SIZE)

extern	BYTE	cgetchar();
extern	DEFAULT	ci_deregister();

extern	BYTE	selection[80];
extern	ULONG   src_id;
extern	USHORT  dst_cpu;
extern	DEFAULT errno;
extern	DEFAULT ci_fd;
extern	BOOLEAN	scold_user;

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	clr()
**
**  DESCRIPTION:	Clears the screen and will puts out flash and bell
**			if user made an input error.  The cursor is positioned
**			to line 13.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	scold_user	controls error message output
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
clr()

{
	putp(tparm(clear_screen));

	if (scold_user == TRUE)
	{
		putp(tparm(flash_screen));
		putp(tparm(bell));

		scold_user = FALSE;
	}

	fprintf(stdout, "\n\n\n\n\n\n\n\n\n\n\n");
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	get_key()
**
**  DESCRIPTION:	This function puts a message at the bottom of the
**			screen, waits for user to input any key and sends
**			back a NULL.
**
**  PARAMETERS:		
**
**	Input:		none
**
**	Output:		selection[0] = NULL
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:	selection	user input buffer
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
get_key()

{
	selection[0] = NULL;

	fprintf(stdout, "\n%25sPress any key to continue:  ", BLANK);

	cgetchar();

	selection[0] = NULL;
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ex_shell_cmd()
**
**  DESCRIPTION:	Clears the screen, resets the terminal and
**			allows the user to execute shell commands.
**
**  PARAMETERS:		
**
**	Input:		selection	user input buffer
**
**	Output:		none
**
**	Returns:	nothing
**
**  SYSTEM GLOBALS:
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

VOID
ex_shell_cmd()

{
	putp(tparm (clear_screen));

	resetterm();

	fprintf(stdout, "\n");

	system("sh");
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	bld_epb()
**
**  DESCRIPTION:	Builds an sm_epb inserting the epb_type sent in as a
**			parameter.  Posts the epb to the Station Manager
**			on the board.
**
**  PARAMETERS:		
**
**	Input:		epb	sm_epb with data in body
**			type	event type to insert into epb
**			qid 	destination queue ID
**
**	Output:		none
**
**	Returns:	SUCCESS/FAILURE
**
**  SYSTEM GLOBALS:	scold_user	controls error message
**			ci_fd		file descriptor for common environment
**			src_id		our source id
**			dst_cpu		destination cpu id
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

DEFAULT

bld_epb(epb, type, qid)

register EPB	 *epb;		/* EPB pointer */
register USHORT	 type;		/* event type */
register USHORT	 qid;		/* destination queue ID */

{
	VOID	 clr();
	VOID	 get_key();

	register DEFAULT n;		/* CE return status */
	register ULONG	 stime;		/* time when write started */

	/*
	**	Initialize the sm_epb header.
	*/

	epb->type 	= type;
	epb->status	= 0;
	epb->src_id	= src_id;
	epb->dst_id	= PUT_ID (dst_cpu, qid);
	epb->epb_size	= sizeof (SM_EPB);
	epb->next	= (EPB *)NULL;
	epb->buf	= (BUF *)NULL;

	/*
	**	Post the sm_epb to the Station Manager on the board.
	*/

	errno = 0;

	stime = time(NULL);

	while (((n = write(ci_fd, epb, 1)) == 0) && (time(NULL) - stime) < 60);

	if (n > 0)
		return (SUCCESS);

	scold_user = TRUE;		/* Tell user we got a problem */

	clr ();

	if (n < 0)
		fprintf(stdout, "%25sFailure on write to CE, err = %d\n", BLANK, errno);
	else
		fprintf(stdout, "%25sCouldn't write to CE within 60 secs\n", BLANK);
	get_key();

	return (FAILURE);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	wait_sm_response()
**
**  DESCRIPTION:	Waits for a repsonse from the Staion Manager on the
**			board.
**
**  PARAMETERS:		
**
**	Input:		epb
**
**	Output:		none
**
**	Returns:	pointer to sm_epb returned from the board or NULL
**			for no epb returned
**
**  SYSTEM GLOBALS:	scold_user	controls error message
**			ci_fd		file descriptor for common environment
**			src_id		our source id
**
**  METHOD:
**
**  NOTES:
**
**********************************************************************/

DEFAULT
wait_sm_response(epb, type)

register EPB	*epb;		/* EPB pointer */
register USHORT type;		/* expected response type */

{
	VOID	clr();
	VOID	get_key();

	register DEFAULT n;	/* CE return status */

	/*
	**	Wait on our queue for an epb.
	*/

	if ((n = read(ci_fd, epb, (ULONG)(RD_SIZE))) < 0)
	{
		scold_user = TRUE;

		clr();

		fprintf(stdout, "%25sFailure on read from Common Environment, err = %d\n", BLANK, errno);

		get_key();

		return (FAILURE);
	}

	if (epb->type != type)
	{
		scold_user = TRUE;

		clr();

		fprintf(stdout, "\n\n%23sBad EPB type returned, type = 0x%x, status = 0x%x\n", BLANK, epb->type, epb->status);

		get_key();

		return (FAILURE);
	}

	if (epb->status != NULL)
	{
		scold_user = TRUE;

		clr();

		fprintf(stdout, "\n\n%23sBad EPB status returned, type = 0x%x, status = 0x%x\n", BLANK, epb->type, epb->status);

		get_key();

		return (FAILURE);
	}

	return (SUCCESS);
}
