/******************************************************************************
/
/  @(#)MPSlock.c	35.2
/  @(#)Copyright (C) 1987 by Motorola Computer Systems
/
/	Mechanism to create a 'lock' file to prevent multiple users from
/	trying to access a single printer at the same time.  This routine
/	is an adaptation of the locking routine used by the Quadratron
/	products to lock out printers for their users.  
/
/	The locking mechanism is based upon the existence of a file in the
/	form "/usr/tmp/LCK..dev" where dev equals the device name (e.g., 
/	tty12) of the port that the target printer is attached to.  To 
/	create a lock, call this routine in the following manner:
/
/		/usr/bin/MPSlock get 2 3 4 5 /usr/tmp/LCK..dev 7 8 9 10 11
/
/		Where dev is equal to the device name (e.g., tty12)
/
/	If, at the time that this call is made, a lock file already exists
/	for the target device, MPSlock will continue to check until the
/	lock file is deleted.  
/
/	To clear the lock on the port, call this routine with the following
/	format:
/
/		/usr/bin/MPSlock release 2 3 4 5 /usr/tmp/LCK..dev 7 8 9 10 11
*/

#ident	"@(#)MPSlock.c	35.2"

#define		TIMESPAN	4

#include	<sys/types.h>
#include	<sys/stat.h>
#define	MAXFNAME	256


main(argc,argv)
int	argc;
char	* argv[];

{
	char	master[MAXFNAME];
	int	fd;
	int	i;
	int	rc;

	struct stat	statbuf;
	extern char	*strcpy();
	extern char	*strcat();

	/* Copy the LOCKFILE name into master */

	strcpy(master, argv[6]);

	/* Strip off the file name from the LOCKFILE path name 
	   and put a NULL character at the end of the truncated
	   path name          						*/

	for (i = (strlen(master) -1); i >= 0; i--)
		if (master[i] == '/')
			break;

	i++;
	master[i] = '\0';

	/* Append the value "Qsplock" to the path name			*/

	strcat(master, "Qsplock");

	/* Use stat to determine whether or not a file already exists
	   with the file name that we just created.  If a file already
	   exists, the value returned by stat will be zero.  If a file
	   does not exist, the value returned by stat will be -1.	*/

	rc = stat(master, &statbuf);

	/* Test the result of our stat request to see if a file already
	   exists.  If it dosent, create one.				*/

	if (rc < 0)
	{
		fd = creat(master, 0666);
		close(fd);
	}

	/* Check to see what type of request this is.  If the second 
	   argument in the call is equal to "get" or "tas", this is a
	   request to create a lock file entry for the device associated
	   with this request.  						*/

	if (strcmp(argv[1], "get") == 0 || strcmp(argv[1], "tas") == 0)
	{

		/* This is a request to create a LOCKFILE file for this
		   device.  Use stat to see if a file equal to our LOCKFILE
		   name already exists.  If so, sleep for awhile until the 
		   existing file goes away.				*/

		do
		{
			rc = stat(argv[6], &statbuf);
		
			if (rc >= 0)
				sleep(TIMESPAN);
		}
		while (rc >= 0);

		/* Now, create a new file with a name equal to LOCKFILE.  
		   Note that we do this using a link request, linking our
		   LOCKFILE file to the "Qsplock" file.  This means, of
		   course, that if the "Qsplock" file does not exist, our
		   link request will never complete successfully.	*/

		do
		{
			rc = link(master, argv[6]);

			if (rc < 0)
				sleep(TIMESPAN);
		}
		while (rc < 0);
	}
	else
	{
	/* This is not a "get" or "tas" request.  Therefore, we assume 
	   that the caller wants to delete an existing LOCKFILE entry.

	   Test the file name to make sure we don't blindly unlink "",
	   which might mean ".".                                         */
		
		if (strlen(argv[6]) > 4)
			unlink(argv[6]);
	}
}
