#	@(#)lp.cnfg.sh	35.4
#	@(#)Copyright (C) 1985 by Motorola, Inc.

# modifications to PORT TO VME 131:
# 4/10/86 SBW: /usr/bin/clear doesn't exist; use "tput clear".
# 5/15/86 SBW: add pt37 and pt38 printers.
# 5/15/86 SBW: comment out LANGUAGE checking.
# 4/08/87 SAJ: enable LANGUAGE checking.
#            : add pt40 support.
#            : closed lp.cnfg mutual exclusion lock hole.
#            : Full pathlists to all called utilities (for
#              security reasons). N.B. echo and fexpr are a shell built-ins.
#            : Fixed various bugs and provided some of optimizations 
#              suggested by Rick Smouse's review of this code in 1986.
#            : Tried to modularize this thing somewhat by defining three
#              functions. 
# 5/12/87 SAJ: Changed example device name to correspond to 8000 conventions.
# 5/13/87 SAJ: Eliminated use of namecheck utility. Use fexpr instead.
# 8/06/87 AT:  Modified for OfficeLAN
# 8/12/87 MCM: add PT235 support (pt39)
# 8/17/87 MCM: use the file names in /usr/spool/lp/interface in the "delete
#	       all printers" option.
# 8/17/87 MCM: Add support for the PAC printers.
# 8/19/87 MCM: Add "trap"s to "list all printers" and "display status
#              of all printers" options.  Drop lpstat -a request from
#	       "display status of all printers" option.
# 8/20/87 MCM: Changed logic for USESTTY and CLASS variables.  USESTTY is now
#	       based upon whether or not the user has specified a DEV value
#	       that signifies a parallel port.  CLASS is now based upon
#	       whether or not the printer is a line printer or a character
#	       printer.  As of this moment, the only line printers in the
#	       product line are of printer type pt36.  
# 8/20/87 MCM: Converted printer type entry from use of ptxx values to use
#	       of MCS model numbers.
# 9/15/87 MCM: Prompt for baud rate when configuring pt36s
# 9/16/87 MCM: Ask user whether they want to have banners printed.
# 9/25/87 MCM: When displaying the printers (options 7 & 8) check to see if
#	       OFFICELAN is up and whether user wants all printers
#	       displayed or just the local printers.
# 9/28/87 MCM: When adding new printers, check to see if OFFICELAN is 
#	       configured, and, if it is, ask the user whether they want to
#	       add the printer locally or globally.
# 10/30/87 MCM:	Add function to check for printers for which EMULATION
#		can be specified.
# 11/2/87 MCM:	Add support for PT3607, PT3610, PT3615, PT3620.
# 11/9/87 MCM:	Use device major/minor codes to determine whether device is
#		attached to a serial or parallel port.  
# 11/19/87 MCM:	Add prompts and logic for user entry of the printer "class"
# 12/2/87 MCM:	Corrected logic related to determining whether or not interface
#		file contains LANDEVICE variable.
# 1/11/88 MCM:  Add support for establishing/clearing port_hold.
# 2/4/88  MCM:  Don't delete the default printer assignment when updating a 
#		printer.
# 4/12/88 MCM:	Add support for the 147 parallel port.
#		When creating the interface file, put the "MODEL=" line first
# 4/25/88 MCM:  When changing printer type, call getcset

#
# Define function to get printer type
#
gettype () {
	echo "(Such as PT235, PT2100, PT2200, PT3401, PT3501, PT3604, PT3607, PT3610,"
	echo "         PT3615, PT3620, PT4000, or PT4010): \c"
	read PRINTER
	if [ "$PRINTER" = "" ]
	then
		echo "\n*** You must enter a printer type. ***"
		return 1
	fi
	MCSMODEL=
	case $PRINTER in
		pt30)
			;;
		pt31)
			;;
		pt32)
			;;
		pt34)
			;;
		pt35)
			;;
		pt36)
			;;
		pt37)
			;;
		pt38)
			;;
		pt39)
			;;
		pt40)
			;;
		pt41)
			MCSMODEL=PT4010 ;;
		pt43)
			MCSMODEL=PT3607 ;;
		pt44)
			MCSMODEL=PT3610 ;;
		pt45)
			MCSMODEL=PT3615 ;;
		pt46)
			MCSMODEL=PT3620 ;;
		ptPL)
			;;
		ptLJ)
			;;
		ptLW)	
			;;
		ptTJ)
			;;
		ptFX)
			;;
		[pP][tT]3001)
			PRINTER=pt30
			MCSMODEL=PT3001 ;;
		[pP][tT]235)
			PRINTER=pt39
			MCSMODEL=PT235 ;;
		[pP][tT]2100)
			PRINTER=pt37
			MCSMODEL=PT2100 ;;
		[pP][tT]2200)
			PRINTER=pt38
			MCSMODEL=PT2200 ;;
		[pP][tT]3401)
			PRINTER=pt34
			MCSMODEL=PT3401 ;;
		[pP][tT]3501)
			PRINTER=pt35
			MCSMODEL=PT3501 ;;
		[pP][tT]3604)
			PRINTER=pt36
			MCSMODEL=PT3604 ;;
		[pP][tT]3607)
			PRINTER=pt43
			MCSMODEL=PT3607 ;;
		[pP][tT]3610)
			PRINTER=pt44
			MCSMODEL=PT3610 ;;
		[pP][tT]3615)
			PRINTER=pt45
			MCSMODEL=PT3615 ;;
		[pP][tT]3620)
			PRINTER=pt46
			MCSMODEL=PT3620 ;;
		[pP][tT]4000 | [pP][tT]4005)
			PRINTER=pt40
			MCSMODEL=PT4000 ;;
		[pP][tT]4010)
			PRINTER=pt41
			MCSMODEL=PT4010 ;;
		*)
			echo "\n*** Printer type \"$PRINTER\" is not a supported printer. ***"
			return 1
			;;
		esac
	return 0
}

#
# Define function which sets up language for the printer
#
getlang () {
	LANG=""
        BUILDLANG=""
	BUILDLANGP=""
  	ENGLISUP=""
        NUMLANGS=`/bin/ls /usr/lib/cs.printer/$PRINTER.* 2>/dev/null | /bin/wc -l`
        LANGCNT=0
        if [ "$NUMLANGS" -ne "0" ]
	then
		for i in /usr/lib/cs.printer/$PRINTER.*
		do
              	 	LANGCNT=`fexpr $LANGCNT + 1`
			LANGFILE=`basename $i`
			LANGFILE=`fexpr //$i : '.*\.\(.*\)'`
                       	if [ "$LANGFILE" = "engli" ]
                       	then
				ENGLISUP="$LANGFILE"
                       	fi
			if [ "$BUILDLANGP" ]
			then
				if [ "$LANGCNT" -ne "$NUMLANGS" ]
				then
					BUILDLANGP="$BUILDLANGP, $LANGFILE"
				else
					BUILDLANGP="$BUILDLANGP, or $LANGFILE"
				fi
			else
				BUILDLANGP="$LANGFILE"
			fi
			BUILDLANG="$BUILDLANG\n$LANGFILE"
		done
		echo "\n\nWhat language is this printer going to be printing?\n"
		if [ "$ENGLISUP" ]
		then
			echo "(Default is $ENGLISUP.)\n"
		else
			echo "(Default is no specific language support.)\n"
 		fi
		echo "(Choices are: $BUILDLANGP.): \c"
		read LANG
		if [ "$LANG" ]
		then
			fexpr "$BUILDLANG" : ".*$LANG">/dev/null 2>&1
			if [ $? -ne 0 ] 
			then
				echo "\n*** Printer type \"${MCSMODEL:-$PRINTER}\" doesn't support the \"$LANG\" language. ***"
				/bin/sleep 3
				return 1 #continue
			fi
		else
			if  [ "$ENGLISUP" ]
			then
 				LANG=$ENGLISUP
			fi					
		fi
		return 0
	fi 
	return 2
}

#
# Define function to set up model file and lpadmin and enable the printer
#
setmodel () {
#
#	Set up the temporary model file.
#
	case "$PORTTYPE" in
		p*)
			USESTTY='USESTTY=yes'
			;;
		*)
			USESTTY='USESTTY=yes'
			;;
	esac
	/bin/cat >> /usr/spool/lp/model/tempmodel <<- EOF
	MODEL=$PRINTER
	$USESTTY
	MCSMODEL=$MCSMODEL
	LANG=$LANG
	DEV=/dev/$DEV
	BAUD=$BRATE
	BANNER=$BAN
	LANDEVICE=$GLOBAL
	EMULATION=$EMULATION
	. /usr/spool/lp/model/ptnnx
	EOF

#
#	Before we do any updating, save the name (if any) of the current 
#	default printer for this system
#
	DEFAULTNAME=""
	if [ -f /usr/spool/lp/default ]
	then
		DEFAULTNAME=`/bin/cat /usr/spool/lp/default`
	fi

#
#	Now actually set up lp for the users printer.
#
	set '/bin/who -r'
	if [ "$3" = 2 -o "$3" = 3 ]
	then
		/etc/telinit q
		/bin/sleep 3
	fi
	echo "\n--- Now defining the \"lp\" printer configuration. ---"
#
#	If the user has indicated that this printer is to be added
#	only for local users and OFFICELAN is up, use the lp.orig
#	versions of the various utilities
#
	if [ "$GLOBAL" = no  -a  "$OFFICELANUP" = yes ]
	then
		/usr/lib/lp.orig/lpadmin -x$NAME 2>/dev/null
		/usr/lib/lp.orig/lpadmin -p$NAME -v/dev/$DEV -i/usr/spool/lp/model/tempmodel
		/usr/lib/accept $NAME
		if [ `echo "$CLASS" | wc -w` != 0 ]
		then
			set $CLASS
			while [ "$#" -ne 0 ]
			do
				/usr/lib/lp.orig/lpadmin -p$NAME -c$1
				/usr/lib/accept $1 >/dev/null 2>&1
				shift
			done
		fi
		/usr/bin/lp.orig/enable $NAME
	else
		/usr/lib/lpadmin -x$NAME 2>/dev/null
		/usr/lib/lpadmin -p$NAME -v/dev/$DEV -i/usr/spool/lp/model/tempmodel
		/usr/lib/accept $NAME
		if [ `echo "$CLASS" | wc -w` != 0 ]
		then
			set $CLASS
			while [ "$#" -ne 0 ]
			do
				/usr/lib/lpadmin -p$NAME -c$1
				/usr/lib/accept $1 >/dev/null 2>&1
				shift
			done
		fi
		/usr/bin/enable $NAME
	fi

#
#	Check to see of DEFAULTNAME matchs NAME.  If they do, assume that the
#	/usr/spool/lp/default file is now empty and reload the file with the
#	name of the printer that we just updated.  
#

	if [ "$DEFAULTNAME" = "$NAME" ]
	then
		echo "$NAME" >/usr/spool/lp/default
		/bin/chown lp /usr/spool/lp/default >/dev/null 2>/dev/null
		/bin/chgrp bin /usr/spool/lp/default >/dev/null 2>/dev/null
		/bin/chmod +rw /usr/spool/lp/default
	fi


	/bin/rm -f /usr/spool/lp/model/tempmodel
	/bin/chmod +rw /dev/$DEV >/dev/null 2>&1
	NAME=; export NAME
	/bin/sleep 4
	return 0
}
#
#	Get the baud rate that the pt36/pt43/pt44 is setup for
#
getbaud ()
{
	case "$PORTTYPE" in
	p*)
		BRATE=0
		;;
	*)
		case "$PRINTER" in
		    pt36 | pt43)
			echo "\n\nSelect the baud rate that \"$NAME\" is configured for\n"
			echo "\t1:  9600 BAUD (default)"
			echo "\t2:  19200 BAUD"
			echo "\n\t--Enter the number that indicates the baud rate that \"$NAME\""
			echo "\t      is configured for and <RETURN>:  \c"
			read RATE
			case "$RATE" in
			    2)
				BRATE=19200
				;;
			    *)
				;;
			esac	
			;;
		    pt44)
			echo "\n\nSelect the baud rate that \"$NAME\" is configured for\n"
			echo "\t1:  9600 BAUD"
			echo "\t2:  19200 BAUD (default)"
			echo "\n\t--Enter the number that indicates the baud rate that \"$NAME\""
			echo "\t      is configured for and <RETURN>:  \c"
			read RATE
			case "$RATE" in
			    1)
				BRATE=9600
				;;
			    *)
				BRATE=19200
				;;
			esac	
			;;
		esac
		;;
	esac
}

#
#  Check for EMULATION capability for this printer.
#
getcset() 
{
	case "$MCSMODEL" in 
	    PT4000)
		echo "\nSelect the emulation for \"$NAME\" \n"
		echo "\t1:  Diablo 630  (default)"
		echo "\t2:  4201 IBM Proprinter"
		echo "\n\t--Enter the number that indicates the emulation that you"
		echo "\t      want \"$NAME\" to use and <RETURN>:  \c"
		read CSET
		case "$CSET" in
		    2)
			EMULATION=proprinter
			;;
		    *)
			EMULATION=""
			;;
		esac
		;;
	    PT4010)
		echo "\nSelect the emulation for \"$NAME\" \n"
		echo "\t1:  Postscript  (default)"
		echo "\t2:  HP LaserJet Plus"
		echo "\t3:  Diablo 630"
		echo "\n\t--Enter the number that indicates the emulation that you"
		echo "\t      want \"$NAME\" to use and <RETURN>:  \c"
		read CSET
		case "$CSET" in
		    3)
			EMULATION=diablo
			;;
		    2)
			EMULATION=laserjet
			;;
		    *)
			EMULATION=""
			;;
		esac
		;;
	    *)
		EMULATION=""
		;;
	esac
}

#
#  If this is a PT3615 or PT3620, make sure that the user has assigned this
#  device to a parallel port.
#
checktype()
{
	if [ "$MCSMODEL" = PT3615  -o  "$MCSMODEL" = PT3620 ]
	then
		if [ "$PORTTYPE" != parallel ]
		then
			echo "\n*** $MCSMODEL printers cannot be attached via a serial port. ***"
			return 1
		fi
	fi
	return 0
}

#
#  set PORTTYPE
#
setporttype () {
	PORTTYPE=serial
	set `ls -l /dev/$DEV`
	case $5 in
	    32,)
		PORTTYPE=parallel
		;;
	    34,)
	        case $6 in
		    8 | 24 | 40 | 56 | 72 | 88)
			PORTTYPE=parallel
			;;
		esac
		;;
	    34,104 | 34,120)
		PORTTYPE=parallel
		;;
	    104,)
		PORTTYPE=parallel
		;;
	esac
}

#
#  Get the "class" that this printer is to be associated with
#
getclass () {
	CLASS=""

	case "$PRINTER" in
	    pt36 | pt4[3456])
		CLASS=lpr
		;;
	    *)
		CLASS=cpr
		;;
	esac

	echo "\nThe default class for this printer is \"${CLASS}.\"  If you want to use "
	echo "  \"${CLASS},\" press <RETURN>.  Otherwise, enter the class name that you want to"
	echo "  use and <RETURN>:  \c"

	while :
	do
		read ANS
		if [ -z "$ANS" ]
		then
			return 0
		fi

		if [ `fexpr "$ANS" : '.*'`  -gt  14 ]
		then
			echo "\n*** The size of the name is too large.  It cannot exceed 14 characters."
			echo "\n*** Re-enter the class name. ***"
			continue
		fi

		SAME=`expr $ANS : '\([a-zA-Z0-9_]*\).*' 2>/dev/null`
		if [ $ANS != "$SAME" ]
		then
			echo "\n*** \"$ANS\" is not a legal class name.  You can only"
			echo "    use the characters \"A-Z\", \"a-z\", \"0-9\", and \"_\"."
			echo "\n*** Re-enter the class name. ***"
			continue
		fi

		CLASS=$ANS
		break

	done
	return 0
}

#
#  Establish a port_hold
#

sethold () {
#
#  Make sure that port_hold exists and is executable.
#

	if [ ! -x /local/bin/port_hold ]
	then
		echo "\n*** Can't find/execute \"/local/bin/port_hold\". ***"
		echo "*** port_hold update aborted. ***\n"
		/bin/sleep 3
		return 1
	fi
#
#  Does an entry in /etc/psttys already exist for this port?
#

	if [ -r /etc/psttys ]
	then
		/bin/grep "/dev/$DEV" /etc/psttys >/dev/null 2>/dev/null
		if [ "$?" = 0 ]
		then
			echo "\n*** An entry for \"/dev/$DEV\" already exists in /etc/psttys. ***"
			echo "*** No changes made. ***"
			/bin/sleep 3
			return 0	# Already in /etc/psttys
		fi
	fi
	
	echo "\nWorking.  Please wait."

#
#  Add an entry to the end of /etc/psttys for DEV
#
	
	if [ ! -f /etc/psttys ]
	then
		>/etc/psttys
	fi

	/bin/chmod 0600 /etc/psttys
	echo "stty ixon ixoff -ixany clocal </dev/$DEV" >>/etc/psttys
	/bin/chmod 0444 /etc/psttys
	echo "\nEntry for /dev/$DEV has been added to /etc/psttys."

#
#  If port_hold is running, stop it.
#

	/bin/ps -ef >/usr/tmp/ps
	LINE=`/bin/grep -s '/local/bin/port_hold' /usr/tmp/ps`
	if [ "$?" = 0 ]
	then
		set $LINE
		/bin/kill -9 "$2"
	fi
	/bin/rm /usr/tmp/ps >/dev/null 2>/dev/null

#
#  Start-up port_hold
#

	echo "Starting /local/bin/port_hold using updated /etc/psttys.\n"
	PID=`/local/bin/port_hold </etc/psttys >/usr/tmp/port_hold.log &`
	/bin/sleep 5

}

#
#  Clear a port_hold
#

clearhold () {
#
#  Make sure that an entry exists in /etc/psttys that matches /DEV
#

	/bin/grep -s "/dev/$DEV" /etc/psttys >/dev/null
	if [ "$?" != 0 ]
	then
		echo "\n*** Can't find entry for \"/dev/$DEV\" in /etc/psttys. *** "
		echo "*** No changes made. ***"
		/bin/sleep 4
		return 1
	fi

	echo "\nWorking.  Please wait."

#
#  Delete all lines in /etc/psttys that contain a reference to $DEV
#

	/bin/chmod 0600 /etc/psttys
	/bin/sed "/$DEV/d" /etc/psttys >/tmp/psttys
	cp /tmp/psttys /etc/psttys
	/bin/chmod 0444 /etc/psttys
	/bin/rm /tmp/psttys >/dev/null 2>/dev/null
	echo "\nEntry for /dev/$DEV has been removed from /etc/psttys."

#
#  if port_hold is running, stop it.
#

	/bin/ps -ef >/usr/tmp/ps
	LINE=`/bin/grep -s  '/local/bin/port_hold' /usr/tmp/ps`
	if [ "$?" = 0 ]
	then
		set $LINE
		/bin/kill -9 "$2"
	fi
	/bin/rm /usr/tmp/ps >/dev/null 2>/dev/null

#
#  if /etc/psttys is not an empty file, re-start port_hold
#

	LINE=`/bin/wc -w /etc/psttys`
	set $LINE
	if [ "$1" != 0 ]
	then
		echo "Starting /local/bin/port_hold using updated /etc/psttys\n"
		PID=`/local/bin/port_hold </etc/psttys >/usr/tmp/port_hold.log &`
		/bin/sleep 5
	fi

}
	

#
#	Mainline starts here 
#
#	Determine if the user is allowed to run this utility.
#
ID=`id 2>/dev/null`
fexpr "$ID" : '.*root' >/dev/null 2>&1
if [ $? != 0 ]
then
	echo "lp.cnfg:  You must be \"root\" to use this utility."
	exit 1
fi
/bin/ln /etc/lp.cnfg /tmp/lp.cnfg >/dev/null 2>&1
if [ $? != 0 ]
then
	echo "lp.cnfg:  Only one person may use this utility at a time."
	exit 1
fi
#
#	Initialize the Lp configuration utility
#
trap "" 1 2 3 14 15
/usr/bin/tput clear
echo "--- Now initializing the Lp configuration utility ---"
#
#	Bring down the lp system.
#
/usr/lib/lpshut > /dev/null 2>&1
#
#	Delete any occurences of the schedlock file
#
/bin/rm -f /usr/spool/lp/SCHEDLOCK
#
#	create pstatus and qstatus if they do not exist
#
if [ ! -f /usr/spool/lp/pstatus ]
then
	>/usr/spool/lp/pstatus
	/bin/chown lp /usr/spool/lp/pstatus >/dev/null 2>&1
        /bin/chgrp bin /usr/spool/lp/pstatus >/dev/null 2>&1
	/bin/chmod 644 /usr/spool/lp/pstatus >/dev/null 2>&1
fi
if [ ! -f /usr/spool/lp/qstatus ]
then
	>/usr/spool/lp/qstatus
	/bin/chown lp /usr/spool/lp/qstatus >/dev/null 2>&1
        /bin/chgrp bin /usr/spool/lp/qstatus >/dev/null 2>&1
	/bin/chmod 644 /usr/spool/lp/qstatus >/dev/null 2>&1
fi

# Make sure passwd file in a good state before we edit it.
# (Cannot do this at present because the released passwd fails pwck!

#pwck ||
#{
#	echo "lp.cnfg:  Syntax error in /etc/passwd.  Please correct.
#	/usr/lib/lpsched
#	rm -f /tmp/lp.cnfg
#	exit 1
 
#	Make sure that user "lp" is in the passwd file.
 
/bin/grep '^lp:' /etc/passwd >/dev/null 2>&1
if [ $? != 0 ]
then
	/bin/ed -s /etc/passwd << 'EOF'
H
/^[^:]*:[^:]*:[1-9][0-9][0-9]/i
lp:NONE:71:1:lp Administrator:/usr/spool/lp:
.
p
w
EOF
	/etc/pwck
fi
#
#	Determine whether or not this system has OFFICELAN on it and
#	whether or not OFFICELAN is up.  We determine whether OFFICELAN
#	has been loaded by checking for the existence of ncs.  Then, we
#	exeucte ncs and use the output to determine whether or not the
#	LAN is up
#
	OFFICELANCONF=no
	OFFICELANUP=no
	if [ -f /usr/lan/bin/ncs ]
	then
		OFFICELANCONF=yes
		/usr/lan/bin/ncs >/dev/null 2>&1
		if [ $? = 0 ]
		then
			OFFICELANUP=yes
		fi
	fi
#
#	Now display a menu for the user in order to determine what
#	they want to perform.
#
while :
do
	#
	#	Display the menu.
	#
	/usr/bin/tput clear
	echo "    		Lp System Configuration Utility\n
	 1:  Delete all previous Lp configurations.
	 2:  Add printer configurations into Lp.
	 3:  Change a printer's configuration.
	 4:  Delete one printer from the Lp configuration.
	 5:  Enable a printer.
	 6:  Disable a printer.
	 7:  Display a list of all known printers.
	 8:  Display the status of all printers.
	 9:  Establish the system default printer.
	10:  Assign a class to a printer's configuration.
	11:  Delete a class from a printer's configuration.
	12:  Create a \"port_hold\" for a device.
	13:  Remove a \"port_hold\" for a device.\n\n

	Enter the number of the function you want to perform.
	  Hit the <RETURN> key when you have finished: \c"
	read TASK
	if [ "$TASK" = "" ]
	then
		/usr/lib/lpsched
		/bin/rm -f /tmp/lp.cnfg
		exit 0
	fi
	case "$TASK" in
	1)
		/usr/bin/tput clear
		echo "You are about to delete the system's entire printer configuration"
		echo "\n\t--Press \"C\" and <Return> to continue"
		echo "\n\t--or press just <Return> to abort this function: \c"
		read ANS
		if [ "x$ANS" = "xc" -o "x$ANS" = "xC" ]
		then
		#
		#	Delete all of the printers in the /usr/spool/
		#	lp/interface directory.  If the interface file
		#	indicates that this is a LAN service device
		#	and the LAN is not up, do not delete the
		#	device
		#
			for NAME in `/bin/ls /usr/spool/lp/interface`
			do
				eval `/usr/bin/fgrep LANDEVICE /usr/spool/lp/interface/$NAME`
				if [ $? = 0  -a  "$LANDEVICE" = yes  -a  $OFFICELANUP != yes ]
				then
					echo "Printer \"$NAME\" is a LAN service device and the LAN is not up."
					echo "---Printer \"$NAME\" not deleted---"
				continue
				fi
				/usr/lib/reject $NAME >/dev/null 2>/dev/null
				/usr/lib/lpadmin -x$NAME >/dev/null 2>/dev/null
				if [ "$?" = 0 ]
				then
					echo "---Printer \"$NAME\" has been deleted.---"
				else
					echo "*** \"$NAME\" could not be deleted.  This probably means that"
					echo "***   you currently have files queued to be printed by \"$NAME.\" "
				fi
			done
			/bin/sleep 5
		else
			echo "\n\n\t*** function aborted"
			/bin/sleep 3
		fi
		;;
	2)
		#
		#	This function will add a printer to the lp
		#	configuration.
		#
		/usr/bin/tput clear
		BRATE=""
		BAN=yes
		CLASS=""
		echo "	Add printer configurations into Lp\n"
		while :
		do
			echo "\nPlease enter the printer device name"
			echo "(Such as tty11 for /dev/tty11"
			echo " or tty12 for /dev/tty12): \c"
			read DEV
			if [ "$DEV" = "" ]
			then
				break
			fi
			if [ ! -c /dev/$DEV ]
			then
				echo "\n*** /dev/$DEV isn't a valid device name. ***"
				continue
			fi
			/bin/ps -t $DEV > /dev/null
			if [ $? = 0 ]
			then
				echo "\n*** You must not have anything running on /dev/$DEV. ***"
				continue
			fi
			
			#
			#  Determine whether this is a serial or parallel port
			#

			setporttype

			echo "\n\nWhat name do you want to give this printer? \c"
			read NAME
			if [ "$NAME" = "" ]
			then
				echo "\n*** You must enter a printer name. ***"
				continue
			fi
			if [ "$NAME" = "cpr" -o "$NAME" = "lpr" ]
			then
				echo "\n*** You may not use either \"cpr\", or \"lpr\" as printer names. ***"
				continue
			fi
			if [ -d /usr/spool/lp/request/$NAME ]
			then
				echo "\n*** Printer \"$NAME\" is already defined in lp. ***"
				continue
			fi
			if [ `fexpr "$NAME" : '.*'` -gt 13 ]
			then
				echo "\n*** The size of the name is too large.  Please pick a new printer name. ***"
				continue
			fi
			SAME=`expr $NAME : '\([a-zA-Z0-9_]*\).*' 2>/dev/null`
			if [ $NAME != "$SAME" ]
			then
				echo "\n*** Printer \"$NAME\" is not a legal printer name. ***"
				echo "*** You can only use characters \"A-Z\", \"a-z\", \"0-9\", and \"_\". ***"
				continue
			fi
			echo "\n\nWhat printer model do you have on \"$DEV\"?"
			gettype  	# Get printer type
			if [ $? = 1 ]
			then
				continue
			fi
			checktype	# Check for valid port assignment
			if [ $? = 1 ]
			then
				continue
			fi

			echo "\n\nDo you want \"$NAME\" to be assigned to a \"class\" of"
			echo "  printers? (default is no) [y, n]:  \c"
			read ANS
			case "$ANS" in
			    [yY]*)
				getclass	# Get the "class"
				;;
			    *)
				;;
			esac

			getlang 	# Get printer language
			if [ $? = 1 ]
			then
				continue
			fi
			break
		done
		if [ "$DEV" = "" ]
		then
			continue
		fi
		case "$PRINTER" in
			pt36 | pt4[34])
				getbaud
				;;
		esac
		echo "\n\nDo you want banners printed at the beginning and end"
		echo "  of the output? (default is yes) [y, n]:  \c"
		read STATUS
		case "$STATUS" in
			N* | n*)
				BAN=no
				;;
		esac

		#
		#	Check to see if this printer is capable of emulating
		#	multiple command sets, and, if so, ask the user which
		#	they want to use as the default command set
		#

			getcset

		#
		#	If OFFICELAN is configured on this sytem, ask the user
		#	whether they want to include this printer globally or
		#	just locally.  If they say globally, make sure that 
		#	OFFICELAN is up.
		#
		GLOBAL=no
		if [ "$OFFICELANCONF" = yes ]
		then
			echo "\n\nThis system is configured for OfficeLAN support.  Indicate"
			echo "  whether you wish \"$NAME\" to be available only to users on"
			echo "  this system or to users on other nodes on the LAN as well."
			echo "\n\n\t1:	Users on this system only"
			echo "\t2:	All nodes on the LAN"
			echo "\n\t--Enter the number that indicates your selection (default"
			echo "\t\tis 2) [1, 2]:  \c"
			read ANS
			#	
			#	If the user wants the printer available globally
			#	make sure that OFFICELAN is up.  If it isn't
			#	tell the user that they must bring OFFICELAN
			#	up.  
			#
			case "$ANS" in
				1)
					;;
				*)
					if [ "$OFFICELANUP" != yes ]
					then
						echo "\n\nOfficeLAN must be up to add printers to the LAN network."
						echo "  Your request is aborted."
						/bin/sleep 5
						continue
					fi
					GLOBAL=yes
					;;
			esac
		fi
		setmodel
		;;
	3)
		#
		#	This function will change a printer configuration
		#
		/usr/bin/tput clear
		echo "  Change A Printer Configuration"
		echo "\n\nEnter the printer's name: \c"
		read NAME
		if [ "$NAME" = "" ]
		then
			continue
		fi
		echo "Working.   Please Wait."
		if [ ! -f /usr/spool/lp/interface/$NAME ]
		then
			echo "\n*** Printer \"$NAME\" doesn't exist. ***"
			/bin/sleep 3
			continue
		fi
		eval `/usr/bin/fgrep MODEL /usr/spool/lp/interface/$NAME`
		eval `/usr/bin/fgrep MCSMODEL /usr/spool/lp/interface/$NAME`
		eval `/usr/bin/fgrep LANG /usr/spool/lp/interface/$NAME`
		eval `/usr/bin/fgrep DEV /usr/spool/lp/interface/$NAME`
		eval `/usr/bin/fgrep BAUD /usr/spool/lp/interface/$NAME`
		eval `/usr/bin/fgrep BANNER /usr/spool/lp/interface/$NAME`
		eval `/usr/bin/fgrep EMULATION /usr/spool/lp/interface/$NAME`
	#
	#	Check to see if the interface file contained a LANDEVICE 
	#	entry.  If it didn't we will take our cue as to whether
	#	this printer is to be included on the LAN or not from the
	#	current state of the LAN.  If OFFICELAN is not configured
	#	this can only be a local printer.  If OFFICELAN is configured
	#	we use the current state of OFFICELAN to determine whether
	#	this printer should be configured locally or globally.  If
	#	the LAN is up, we assume that this should be a global 
	#	printer.  Otherwise we assume that it is a local printer.
	#
		eval `/usr/bin/fgrep LANDEVICE /usr/spool/lp/interface/$NAME`
		if [ $? = 0 ]
		then
			GLOBAL=$LANDEVICE
		else
			if [ $OFFICELANCONF = yes  -a  $OFFICELANUP = yes ]
			then
				GLOBAL=yes
			else
				GLOBAL=no
			fi
		fi
		#
		#  Check to see if this printer is part of a class (or classes)
		#
		CLASS=""
		CLASSCNT=`/bin/ls /usr/spool/lp/class 2>/dev/null | /bin/wc -l`
		if [ "$CLASSCNT" -ne 0 ]
		then
			for i in /usr/spool/lp/class/*
			do
				eval /usr/bin/fgrep \'\^"$NAME"\$\' $i >/dev/null 2>/dev/null
				if [ $? = 0 ]
				then
					CLASSFILE=`basename $i`
					if [ "$CLASS" ]
					then
						CLASS="$CLASS $CLASSFILE"
					else
						CLASS="$CLASSFILE"
					fi
				fi
			done
		fi

		DEV=`echo $DEV | /usr/bin/cut -f3 -d"/"`
		PRINTER=$MODEL
		OLDMODEL=$MODEL
		OLDMCSMODEL="$MCSMODEL"
		OLDLANG="$LANG"
		OLDDEV=$DEV
		BRATE="$BAUD"
		BAN="$BANNER"
		/usr/lib/reject $NAME >/dev/null 2>&1
		/usr/bin/disable $NAME >/dev/null 2>&1
		while :
		do
			/usr/bin/tput clear
			echo "\n\n\tPrinter \"$NAME\" is currently attached to device \"$OLDDEV\""
			echo "\t\t\tPrinter type = ${MCSMODEL:-$MODEL}"
			echo "\t\t\tLanguage = $OLDLANG"
			echo "\t\t\tBanner = $BAN"
			echo "\n\n\tSelect the item you wish to change\n"
			echo "\t\t1:  Change device \"$NAME\" is attached to"
			echo "\t\t2:  Change printer type of \"$NAME\""
			echo "\t\t3:  Change language of \"$NAME\""
			echo "\t\t4:  Change banner option of \"$NAME\""
			echo "\n\n\t--Enter the number of the function you want to perform and <Return>"
			echo "\t--or just press <Return> to end changes: \c"
			read TASK
			case "$TASK" in
			1)
				#
				#	Change the printer's device
				#
				while :
				do
					/usr/bin/tput clear
					echo "\n\nCurrent device for printer \"$NAME\" is $OLDDEV"
					echo "\n\nPlease enter the new device name for printer \"$NAME\""
					echo "(Such as tty11 for /dev/tty11"
					echo " or tty12 for /dev/tty12): \c"
					read DEV
					if [ "$DEV" = "" ]
					then
						echo "\n*** You must enter a printer device name. ***"
						/bin/sleep 3
						continue
					fi
					if [ ! -c /dev/$DEV ]
					then
						echo "\n*** \"/dev/$DEV\" isn't a valid device. ***"
						/bin/sleep 3
						continue
					fi
					/bin/ps -t $DEV >/dev/null 2>/dev/null
					if [ $? = 0 ]
					then
						echo "\n*** You must not have anything running on /dev/$DEV. ***"
						/bin/sleep 4
						continue
					fi
				
					#
					#  Set the port type
					#

					setporttype

					#
					#  Make sure that we're not trying to 
					#  a parallel port printer on a serial
					#  port
					#

					checktype	# Validate the port use
					if [ $? = 1 ]
					then
						/bin/sleep 4
						continue
					fi
					break
				done
				echo "\n\nNow changing printer \"$NAME\" device from $OLDDEV to $DEV"
				/bin/sleep 4
				OLDDEV=$DEV
				;;
			2)
				#
				#	Change the printer's type
				#
				while :
				do
					/usr/bin/tput clear
					echo "\n\nPrinter \"$NAME\" is currently type \"${OLDMCSMODEL:-$MODEL}\""
					echo "\nWhat type of printer do you want to change printer \"$NAME\" to?"
					gettype
					if [ $? = 1 ]
					then
						/bin/sleep 5
						continue
					fi
					if [ "$LANG" ]
					then	
						if [ ! -r /usr/lib/cs.printer/$PRINTER.$LANG ]
						then
							echo "\n*** Printer type \"$PRINTER\" doesn't support the \"$LANG\" language. ***"
							echo "\n    Language changed to default."
							LANG=""
							OLDLANG=""
							/bin/sleep 5
						fi
					fi
					if [ $PRINTER = pt36  -a  $OLDMODEL != pt36 ]
					then
						getbaud
					fi
					if [ $PRINTER = pt43  -a  $OLDMODEL != pt43 ]
					then
						getbaud
					fi
					if [ $PRINTER = pt44  -a  $OLDMODEL != pt44 ]
					then
						getbaud
					fi
					getcset		# Get command set emulation
					break
				done
				echo "\n\nNow changing printer \"$NAME\" from type \"${OLDMCSMODEL:-$MODEL}\" to type \"${MCSMODEL:-$PRINTER}\""
				/bin/sleep 3
				MODEL=$PRINTER
				OLDMCSMODEL=$MCSMODEL
				;;
			3)
				#
				#	Change the printer's language
				#
				while :
				do
					/usr/bin/tput clear
					echo "\n\nPrinter \"$NAME\" currently prints in \"$OLDLANG\""
					getlang
					case $? in
						0)	 echo "\nNow changing printer \"$NAME\" from language \"$OLDLANG\" to \"$LANG\"."
							/bin/sleep 3
							OLDLANG=$LANG
							break
							;;
					 	1)	continue
					        	;;
					     	2)	echo "\nThere are no languages supported for printer type \"$MODEL\"."
							/bin/sleep 4
							break
							;;
					esac
				done
				;;
			4)
				#
				#	Change the printer's banner option
				#
				/usr/bin/tput clear
				echo "\n\nPrinter \"$NAME\" currently prints with BANNER=\"$BAN\""
				echo "\n\nDo you want banners printed at the beginning and end"
				echo "  of the output? (default is yes) [y, n]:  \c"
				BAN=yes
				read STATUS
				case "$STATUS" in
					N* | n*)
						BAN=no
						;;
				esac
				;;
			*)
				if [ "$TASK" != "" ]
				then
					echo "\n*** You have entered an invalid function. ***"
					/bin/sleep 3
					continue
				fi
				;;
			esac
			if [ "$TASK" != "" ]
			then
				continue
			fi
			setmodel
			break
		done
		;;
	4)
	
		#	This function will delete one printer from the
		#	Lp configuration.
		#
		/usr/bin/tput clear
		echo "  Delete one printer from the Lp configuration\n"
		while :
		do
			echo "\nWhat printer do you want to delete? \c"
			read NAME
			if [ "$NAME" = "" ]
			then
				break
			fi
			if [ ! -r /usr/spool/lp/interface/$NAME ]
			then
				echo "\n*** Printer \"$NAME\" doesn't exist. ***"
				continue
			fi
			echo "\nDo you really want to delete printer \"$NAME\"? [y, n]:  \c"
			read STATUS
			case "$STATUS" in
				Y* | y*)
					;;
				*)
					break
					;;
			esac
			#
			#	Check to see if this is a LAN printer.  If it 
			#	is, we want to make sure that the LAN is up so
			#	that the entry for this printer will be deleted
			#	from the LCF.  (Although we probably should also
			#	worry about those situations where this is not 
			#	a LAN printer and therefore use lp.orig/lpadmin
			#	if the LAN is not up, we don't.  For this 
			#	situation, we assume that the LAN version of 
			#	lpadmin will sort things out properly if it
			#	dosen't find an entry for this printer in the
			#	LCF.)
			#
			eval `/usr/bin/fgrep LANDEVICE /usr/spool/lp/interface/$NAME`
			if [ $? = 0  -a  "$LANDEVICE" = yes   -a  $OFFICELANUP != yes ]
			then
				echo "\n\nThis printer is a LAN service device.  You must"
				echo "  have the LAN up in order to delete a LAN service device."
				echo "  Your delete request is rejected."

				/bin/sleep 5
				break
			fi
			/usr/bin/disable -c $NAME >/dev/null 2>/dev/null
			/usr/lib/reject $NAME >/dev/null 2>/dev/null
			/usr/lib/lpadmin -x$NAME >/dev/null 2>/dev/null
			if [ "$?" = 0 ]
			then
				echo "--- Printer \"$NAME\" has been deleted. ---"
			else
				echo "\n*** \"$NAME\" could not be deleted.  This probably means that you"
				echo "***   currently have files queued to be printed by \"$NAME.\" "
				/bin/sleep 2
			fi
			/bin/sleep 3
			break
		done
		;;
	5)
		#
		#	This function will enable a printer and cause all
		#	of its requests to be accepted.
		#
		/usr/bin/tput clear
		echo "  Enable a printer\n"
		while :
		do
			echo "\nWhat printer do you want to enable? \c"
			read NAME
			if [ "$NAME" = "" ]
			then
				break
			fi
			if [ ! -d /usr/spool/lp/request/$NAME ]
			then
				echo "\n*** Printer \"$NAME\" doesn't exist. ***"
				continue
			fi
			if [ "$STATUS" != "enabled" ]
			then
				/usr/bin/enable $NAME
			else
				echo "printer \"$NAME\" now enabled"
			fi
			STATUS=`/usr/bin/lpstat -a$NAME 2>/dev/null | cut -f2 -d" "`
			if [ "$STATUS" != "accepting" ]
			then
				/usr/lib/accept $NAME
			else
				echo "destination \"$NAME\" now accepting requests"
			fi
			break
		done
		;;
	6)
		#
		#	This function will disable a printer and cause all
		#	of its requests to be rejected.
		#
		/usr/bin/tput clear
		echo "  Disable a printer\n"
		while :
		do
			echo "\nWhat printer do you want to disable? \c"
			read NAME
			if [ "$NAME" = "" ]
			then
				break
			fi
			if [ ! -d /usr/spool/lp/request/$NAME ]
			then
				echo "\n*** Printer \"$NAME\" doesn't exist. ***"
				continue
			fi
			if [ "$STATUS" = "enabled" ]
			then
				/usr/bin/disable -c $NAME
			else
				echo "printer \"$NAME\" now disabled"
			fi
			STATUS=`/usr/bin/lpstat -a$NAME 2>/dev/null | cut -f2 -d" "`
			if [ "$STATUS" = "accepting" ]
			then
				/usr/lib/reject $NAME
			else
				echo "destination \"$NAME\" is no longer accepting requests"
			fi
			break
		done
		;;
	7)
		#
		#	This function will format and display a list of
		#	all the currently known printers on the system.
		#
		/usr/bin/tput clear
		echo "  Display a list of all known printers\n\n"
		ABORT=no
		trap "ABORT=yes; \
		      break" 2
		while :
		do
			#
			#	Check to see if OFFICELAN is up.  If it
			#	is, ask the user whether they wish to display
			#	all of the printers including those on the
			#	LAN.
			#
			DOALL=no
			if [ $OFFICELANUP = yes ]
			then
				echo "\nDo you wish to display all printers, including those"
				echo "  on other machines on the LAN? [y, n]:  \c"
				read ANS
				case "$ANS" in
					y* | Y*)
						DOALL=yes
						;;
					*)
						;;
				esac
			fi
			if [ "$DOALL" = no  -a  -r /usr/bin/lp.orig/lpstat ]
			then
				echo "\n"
				/usr/bin/lp.orig/lpstat -v
			else
				/usr/bin/lpstat -v
			fi
			break
		done
		trap "" 2
		if [ "$ABORT" = yes ]
		then
			continue
		fi
		echo "\n\nHit the <RETURN> key to return to the main menu.\c"
		read answer
		;;
	8)
		#
		#	This function will display the status of all
		#	known printers on the system.
		#
		/usr/bin/tput clear
		echo "  Display the status of all printers\n\n"
		ABORT=no
		trap "ABORT=yes; \
		      break" 2
		while :
		do
			#
			#	Check to see if OFFICELAN is up.  If it is,
			# 	ask the user whether they wish to display all
			#	of the printers, including those on the lan.
			#
			DOALL=no
			if [ $OFFICELANUP = yes ]
			then
				echo "\nDo you wish to display all printers, including those"
				echo "  on other machines on the LAN? [y, n]:  \c"
				read ANS
				case "$ANS" in
					y* | Y*)
						DOALL=yes
						;;
					*)
						;;
				esac
			fi
			if [ "$DOALL" = no  -a  -r /usr/bin/lp.orig/lpstat ]
			then
				LINE=`/bin/ls /usr/spool/lp/interface`
				echo " "
				if [ "$LINE" = "" ]
				then
					echo "You don't have any printers currently assigned on your system."
				else
					/usr/bin/lp.orig/lpstat -p
				fi
			else
				/usr/bin/lpstat -p
			fi
			break
		done
		trap "" 2
		if [ "$ABORT" = yes ]
		then
			continue
		fi
		echo "\n\nHit the <RETURN> key to return to the main menu.\c"
		read answer
		;;
	9)
		/usr/bin/tput clear
		echo "\n  Establish the system default printer"
		while :
		do
			echo "\nEnter the printer name, or class name, of the new system"
			echo "  default printer:  \c"
			read PRINTER
			if [ "$PRINTER" = "" ]
			then
				break
			fi
			/usr/bin/lpstat -a$PRINTER 2>&1 | /usr/bin/fgrep accepting >/dev/null
			if [ "$?" != "0" ]
			then
				echo "\n*** Printer \"$PRINTER\" does not exist. ***"
				continue
			fi
			echo $PRINTER >/usr/spool/lp/default
			/bin/chown lp /usr/spool/lp/default >/dev/null 2>/dev/null
			/bin/chgrp bin /usr/spool/lp/default >/dev/null 2>/dev/null
			/bin/chmod +rw /usr/spool/lp/default
			echo "\nPrinter \"$PRINTER\" is now the system default printer."
			/bin/sleep 3
			break
		done
		;;
	10)
		/usr/bin/tput clear
		echo "\nAssign a class to a printer's configuration"
		while :
		do
			echo "\nEnter the name of the printer that you want to update:  \c"
			read NAME
			if [ "$NAME" = "" ]
			then
				break
			fi
			if [ ! -f /usr/spool/lp/interface/$NAME ]
			then
				echo "\n*** Printer \"$NAME\" doesn't exist. ***"
				/bin/sleep 3
				continue
			fi

			eval `/usr/bin/fgrep LANDEVICE /usr/spool/lp/interface/$NAME`
			if [ "$LANDEVICE" = yes  -o  "$LANDEVICE" = no ]
			then
				GLOBAL=$LANDEVICE
			else
				if [ $OFFICELANCONF = yes  -a  $OFFICELANUP = yes ]
				then
					GLOBAL=yes
				else
					GLOBAL=no
				fi
			fi

			getclass

			echo "\nNow adding class \"$CLASS\" to printer \"${NAME}.\""
			if [ "$GLOBAL" = no  -a  "$OFFICELANUP" = yes ]
			then
				/usr/lib/lp.orig/lpadmin -p$NAME -c$CLASS
				/usr/lib/accept $CLASS >/dev/null 2>/dev/null
			else
				/usr/lib/lpadmin -p$NAME -c$CLASS
				/usr/lib/accept $CLASS >/dev/null 2>/dev/null
			fi

			/bin/sleep 4
			break
		done
		;;
	11)
		/usr/bin/tput clear
		echo "\nDelete a class from a printer's configuration"
		while :
		do
			echo "\nEnter the name of the printer that you want to delete the"
			echo "  class from:  \c"
			read NAME
			if [ "$NAME" = "" ]
			then
				break
			fi
			if [ ! -f /usr/spool/lp/interface/$NAME ]
			then
				echo "\n*** Printer \"$NAME\" doesn't exist. ***"
				/bin/sleep 3
				continue
			fi
			
			CLASS=""
			CLASSCNT=`/bin/ls /usr/spool/lp/class 2>/dev/null | /bin/wc -l`
			if [ "$CLASSCNT" -ne 0 ]
			then
				for i in /usr/spool/lp/class/*
				do
					eval grep \'\^"$NAME"\$\' $i >/dev/null 2>/dev/null
					if [ $? = 0 ]
					then
						CLASSFILE=`basename $i`
						if [ "$CLASS" ]
						then
							CLASS="$CLASS $CLASSFILE"
						else
							CLASS="$CLASSFILE"
						fi
					fi
				done
			fi

			if [ `echo "$CLASS" | wc -w` = 0 ]
			then
				echo "\n*** Printer \"$NAME\" does not belong to a class. ***"
				/bin/sleep 5
				break
			fi

			if [ `echo "$CLASS" | wc -w` = 1 ]
			then
				echo "\nPrinter \"$NAME\" belongs to class \"$CLASS.\""
				echo "  Do you want to delete class \"$CLASS\" from printer configuration"
				echo "  \"$NAME\"? (default is yes) [y, n]:  \c"
				read ANS
				case "$ANS" in
				    [nN]*)
					echo "\nYour delete request is aborted."
					/bin/sleep 5
					break
					;;
				    *)
					;;
				esac
			else
				while :
				do
					echo "\nPrinter \"$NAME\" belongs to the following classes:"
					I=1
					set $CLASS
					while [ "$#" -ne 0 ]
					do
						echo "\t${I} ->\t$1"
						I=`fexpr $I + 1`
						shift
					done

					echo "\nEnter the number that corresponds to the class that you"
					echo "  want to remove from configuration \"$NAME,\""
					echo "  or, hit <RETURN> to abort this delete request:  \c"
					read ANS
					if [ "$ANS" = "" ]
					then
						echo "\nYour delete request is aborted."
						/bin/sleep 5
						break 2
					fi
					I=1
					set $CLASS
					while [ "$#" -ne 0 ]
					do
						if [ $I = "$ANS" ]
						then
							CLASS=$1
							break 2
						else
							I=`fexpr $I + 1`
							shift
						fi
					done
					echo "\n*** Incorrect response. ***"
					/bin/sleep 5
				done
			fi

			eval `/usr/bin/fgrep LANDEVICE /usr/spool/lp/interface/$NAME`
			if [ $? = 0 ]
			then
				GLOBAL=$LANDEVICE
			else
				if [ $OFFICELANCONF = yes  -a  $OFFICELANUP = yes ]
				then
					GLOBAL=yes
				else
					GLOBAL=no
				fi
			fi
			
			echo "\nNow deleting class \"$CLASS\" from printer \"${NAME}.\""
			if [ "$GLOBAL" = no  -a  "$OFFICELANUP" = yes ]
			then
				/usr/lib/lp.orig/lpadmin -p$NAME -r$CLASS
			else
				/usr/lib/lpadmin -p$NAME -r$CLASS
			fi
		/bin/sleep 4
		break

		done
		;;

	12)
		#
		#	This option will create a "port_hold" on a specified
		#	port (device)
		#
		/usr/bin/tput clear
		echo "Establish a \"port_hold\" for a device\n"
		while :
		do
			echo "\nEnter the printer device name (such as tty11 for /dev/tty11"
			echo " or tty12 for /dev/tty12): \c"
			read DEV
			if [ "$DEV" = "" ]
			then
				break
			fi
			if [ ! -c /dev/$DEV ]
			then
				echo "\n*** /dev/$DEV isn't a valid device name. ***"
				continue
			fi
			/bin/ps -t $DEV >/dev/null
			if [ "$?" = 0 ]
			then
				echo "\n*** You must not have anything running on /dev/$DEV. ***"
				continue
			fi

			sethold
			break
		done
		;;

	13)
		#
		#	This option will delete a "port_hold" on a specified
		#	port (device)
		#
		/usr/bin/tput clear
		echo "Remove a \"port_hold\" for a device\n"
		while :
		do
			echo "\nEnter the printer device name (such as tty11 for /dev/tty11"
			echo " or tty12 for /dev/tty12): \c"
			read DEV
			if [ "$DEV" = "" ]
			then
				break
			fi

			clearhold
			break
		done
		;;
	*)
		echo "\n\n*** You must enter a valid number or <RETURN> ***"
		/bin/sleep 3
		;;
	esac
done
