/*	@(#)m323rd.c	35.4	*/
/*
 *	Copyright (c) 1987	Motorola Inc.  
 *		All Rights Reserved.
 */

/*
 * SYNOPSIS
 *
 *      m323rd  -L  maxcylinder heads maxbads file special 
 *		 results in "file" containing head/cyl
 *      m323rd  -l  maxcylinder heads maxbads file special
 *		 results in "file" containing head/cyl/bfi
 * DESCRIPTION 
 */



#include  "stdio.h"
#include  "fcntl.h"
#include  "sys/dk.h"
#include  "sys/types.h"
#include  "sys/mvme323.h"
#include  "sys/stat.h"
#include  "sys/errno.h"

#define	TRUE	1
#define FALSE	0

#define GRACE	10
#define PHYSBLK	512
#define QK	256
#define BUFSIZE	100 * PHYSBLK
#define	BADFILE	"/etc/badlist"
#define COMMENT	'*'
#define	LINELEN	512

#define MFRLIST1 1
#define MFRLIST2 2



struct m323fmt{
	int  f_trk;	/* Track number */
	int  f_alttrk;	/* Alternate track number */
	int  f_blk;	/* Beginning block number of the track */
	int  f_altblk;	/* Beginning block number of the alternate track */
	int  f_slpsec;	/* Sector to be slipped	*/
};

struct listentry {
	uint	cylinder;
	ushort  head;
};	

/* Element in defect array contained in the first sector of every track on
   the last cylinder of an ESDI disk. */
#define ESDIDSIZE 5	/* Size of ESDI defect */
#define CYLMSB 	0
#define CYLLSB  1
#define BYTEMSB 2
#define BYTELSB 3
#define LENGTH  4

/* ESDI Defect List. The defect list for each surface is contained in sector 0
   of each track on the maximum cylinder. Sector 0 is a 256 byte sector that
   contains the list of defects for that head( surface) only. 
 */
#define MAXESDI 50	/* 250 bytes of the 256 byte sector 0 is dedicated
			for 50  defect descriptions which are 5 bytes each. */  
struct esdilist {
	unsigned char   datasync;	/* data sync byte FEH */
	unsigned char	month;
	unsigned char	day;
	unsigned char	year;
	unsigned char	head;
	unsigned char   pad1;
	unsigned char   pad2;
	unsigned char   defects[ESDIDSIZE * MAXESDI];
/*	struct  esdidefect defects[MAXESDI]; */
};

extern	errno;

struct	volumeid  vol;	/*  Volume id block read off the disk */
struct	dkconfig  dkcf;	/*  Config block read off the disk */
struct badent *bttable; /*  Bad track table read off the disk */
int	slip;		/*  Use sector slip if true */
int	map;		/*  Map to alternate track if true */
int	refmt;		/*  Use reformat if true */
int	verify;		/*  Verify disk if true */
int	fd;		/*  File descriptor of device to be formatted */
char	*device;
char	bdev[80];
char	*cmdname;	/*  Name of command invoked */



main ( argc, argv )
int	argc;
char	**argv;
{
  register int	c;
  register int	track;		/*  Single track to be formatted */
  register int	alt;		/*  Alternate track number */
  register int	sec;		/*  Sector to be slipped */
  extern char	*optarg;
  extern int	optind;
  int fmttype;			/*  Which type of format to execute */
  int fflag = FALSE;		/*  If set to TRUE, then do not format */
  int maxcyl;			/*  cylinder containing manufacturers defect
				    list. */ 	
  int heads;			/*  number of surfaces to read manufacturers 
				    defect lists from. */
  int maxbad;			/*  maximum number of bad tracks */		

  char *mfrfile;		/*  name of bad track file to create after
				    reading manufacturers defect list. */ 

  cmdname = argv[0];

  if (argc == 1) {
	usage ( cmdname );
	exit(2);
  }
  device = argv[argc-1];
  ckdev(  );		/*  Open the device to be formatted */
  
  fmttype = MFRLIST1;	/*  Default to read the bad track list from the disk */
  while ( (c=getopt(argc, argv, "Ll")) != EOF ) {
  	switch (c) {
  
		case 'L': 		/* Read manufacturers defect list */
  			maxcyl = atoi( argv[optind++] );
  			heads  = atoi( argv[optind++] );
			maxbad = atoi( argv[optind++] );
			mfrfile = argv[optind++];
			fmttype = MFRLIST1;
			break;

		case 'l': 		/* Read manufacturers defect list */
  			maxcyl = atoi( argv[optind++] );
  			heads  = atoi( argv[optind++] );
			maxbad = atoi( argv[optind++] );
			mfrfile = argv[optind++];
			fmttype = MFRLIST2;
			break;

		case '?':
			usage ( cmdname );
			exit(2);
			break;
		}
	}

	c = argc - optind - 1;
	if  ( c  !=  0 )  {
	    fprintf( stderr, "%d Extraneous arguement%sstarting with %s\n", 
		   c, c==1?" ":"s ", argv[optind] );
	    exit ( -1 );
	}

	else if ( fmttype  ==  MFRLIST1 ) {
		rdmfrlist(maxcyl, heads, maxbad, mfrfile );
	}
	else if ( fmttype  ==  MFRLIST2 ) {
		rdmfrlist2(maxcyl, heads, maxbad, mfrfile );
	}
	close(fd);
}



/*	Check the last arguement for allowable device */

ckdev( )	
{
  int	minor;
  struct stat	stbuf;
  if ( (fd = open( device, O_RDWR| O_NDELAY )) == -1 ) {	
  	fprintf ( stderr, "%s: cannot open %s\n", cmdname, device );
  	perror ( cmdname );
  	exit(2);
  }
	
  if ( stat( device, &stbuf ) == -1 ) {
  	fprintf ( stderr, "%s: cannot stat %s\n", cmdname, device );
  	perror ( cmdname );
  	exit(2);
  }

  if ( ( stbuf.st_mode & S_IFMT ) != S_IFCHR ) {
  	fprintf ( stderr, "%s: raw devices only\n", cmdname );
  	exit(2);
  }
  
  minor = (int)(stbuf.st_rdev & 0xff);
  if ( (minor & 0x07) != 7 ) {
  	fprintf ( stderr, "%s: slice 7 only\n", cmdname );
  	exit(2);
  }
  strcat( bdev, "/dev/" );
  strcat( bdev, device+6 );

}

/*	print out command usage	    */

usage ( name )
char	*name;
{
  fprintf ( stderr, "usage: %s -lL cyls hds maxbad output_file rdev \n", name );
}


/* read the manufacturers defect list from disk and create the 
specified bad track file.  */
rdmfrlist(maxcyl, heads, maxbad, mfrfile)
int	maxcyl;		/* maximum cylinder */
int     heads;		/* number of heads on drive */
int	maxbad;		/* maximum allowed bad tracks */
char    *mfrfile;	/* name of bad track file to create */
{
	struct dkbadlst *badlist;
	struct dkbadlst *lstptr;
	FILE *fp, *fopen();
	int 	i, listentries;


	listentries = 0;
	/* clear the allocated memory for badlist */
	if((badlist = (struct dkbadlst *) calloc( maxbad, 
		       sizeof(struct dkbadlst))) == NULL){
		 fprintf(stderr,"m323rd: cannot allocate memory.\n");
		 exit(2);
	 }

	if((listentries = convertesdi(badlist, maxcyl, heads)) == -1){ 
		exit(2);
	}
	else if(listentries == 0){
		fprintf(stderr,"m323rd: manufacturers defect list is empty.\n");
		exit(0);
	}

	/* Write the defect list to the specified file */ 
	if((fp = fopen(mfrfile, "w")) == NULL){ 
		fprintf(stderr,"m323rd: can't create the file %s\n", mfrfile);
		exit(2);
	}
	else{
		fprintf(stderr,"The manufacturers defect list on %s contained %d bad track entries which are now in the file %s.\n", device, listentries, mfrfile);
	}
	
	for(lstptr = badlist, i=0; i < listentries ; i++, lstptr++){ 
		fprintf(fp,"%d  %d \n", lstptr->dk_head, lstptr->dk_cyl);
	}
	fclose(fp);

}

/* read the manufacturers defect list from disk and create the 
specified bad track file.  */
rdmfrlist2(maxcyl, heads, maxbad, mfrfile)
int	maxcyl;		/* maximum cylinder */
int     heads;		/* number of heads on drive */
int	maxbad;		/* maximum allowed bad tracks */
char    *mfrfile;	/* name of bad track file to create */
{
	struct dkbadlst *badlist;
	struct dkbadlst *lstptr;
	FILE *fp, *fopen();
	int 	i, listentries;


	listentries = 0;
	/* clear the allocated memory for badlist */
	if((badlist = (struct dkbadlst *) calloc( maxbad, 
		       sizeof(struct dkbadlst))) == NULL){
		 fprintf(stderr,"m323rd: cannot allocate memory.\n");
		 exit(2);
	 }

	if((listentries = convertesdi(badlist, maxcyl, heads)) == -1){ 
		exit(2);
	}
	else if(listentries == 0){
		fprintf(stderr,"m323rd: manufacturers defect list is empty.\n");
		exit(0);
	}

	/* Write the defect list to the specified file */ 
	if((fp = fopen(mfrfile, "w")) == NULL){ 
		fprintf(stderr,"m323rd: can't create the file %s\n", mfrfile);
		exit(2);
	}
	else{
		fprintf(stderr,"The manufacturers defect list on %s contained %d bad track entries which are now in the file %s.\n", device, listentries, mfrfile);
	}
	
	for(lstptr = badlist, i=0; i < listentries ; i++, lstptr++){ 
		fprintf(fp,"%d  %d %d \n", lstptr->dk_head, lstptr->dk_cyl, lstptr->dk_sector);
	}
	fclose(fp);

}

/* Reads manufacturer defect list from an ESDI drive.
   The list is converted to a list of elements which contain only the
   cylinder and head numbers.
 */
convertesdi(badlist, maxcyl, dheads) 
struct dkbadlst *badlist; 
int	maxcyl;
int	dheads;
{
	struct m323mlargs args;
	struct esdilist *elist;  
	char	*mfrlist;
	int	head, i, listentries;
	
	listentries = 0;

	/* Allocate space for one list per head ( surface). */
	if((mfrlist = (char *) malloc(sizeof(struct esdilist)
 	   * dheads)) == NULL){  
		fprintf(stderr,"m323rd: cannot allocate memory.\n");
		exit(2);
	}
	args.buffer = mfrlist;
	args.listsize = sizeof(struct esdilist);
	args.maxcyl = maxcyl;		/* the maximum cylinder */ 
	args.heads = dheads;		/* number of heads on drive */

	/* get manufacturers defect list */
	if(ioctl(fd, (int)RDMFRLIST, &args) == -1){
		fprintf(stderr,"m323rd: ioctl call to read manufacturers defect list from cylinder %d on device %s failed.\n", maxcyl, device);	
	 	exit(2);
	 }

	elist = (struct esdilist *) mfrlist;
	for(head = 0; head < dheads ; head++, elist++){

		/* Validate the defect list data. */
		if(elist->datasync != 0xFE){
			fprintf(stderr,"m323rd: manufacturers defect list not found on %s.\n", device);
			free(mfrlist);
			return(-1);
		}
		if( (elist->head != head) || (elist->month < 1) ||
		    (elist->month > 12) || (elist->day < 1) || 
		    (elist->day > 31) ) {
			fprintf(stderr,"m323rd: manufacturers defect list is erroneous.\n");
			fprintf(stderr," head = %d  month = %d  day = %d.\n", elist->head, elist->month, elist->day);
			free(mfrlist);
			return(-1);
		}

		for(i=0; i< (ESDIDSIZE * MAXESDI); i+= ESDIDSIZE){
			if((elist->defects[i + CYLMSB] == 0xff) &&
			   (elist->defects[i + CYLLSB] == 0xff) &&
			   (elist->defects[i + BYTEMSB] == 0xff) &&
			   (elist->defects[i + BYTELSB] == 0xff) &&
			   (elist->defects[i + LENGTH] == 0xff) ) { 
				badlist->dk_head = 0xffff; /* terminator */
				break;
			}
			badlist->dk_cyl= (elist->defects[i + CYLMSB] << 8) +
				      elist->defects[i + CYLLSB];
			badlist->dk_head = elist->head;	
			badlist->dk_sector = (elist->defects[i + BYTEMSB] << 8) +
				      elist->defects[i + BYTELSB];
			badlist++;
			listentries++;
		}
#if 0
		for(i=0; i<MAXESDI; i++){
			if((elist->defects[i].cylmsb == 0xff) &&
			   (elist->defects[i].cyllsb == 0xff) &&
			   (elist->defects[i].bytemsb == 0xff) &&
			   (elist->defects[i].bytelsb == 0xff) &&
			   (elist->defects[i].length == 0xff) ) { 
				badlist->head = 0xffff; /* terminator */
				break;
			}
			badlist->cylinder = (elist->defects[i].cylmsb << 8) +
				      elist->defects[i].cyllsb;
			badlist->bfi = (elist->defects[i].bytemsb << 8) +
				      elist->defects[i].bytemsb;
			badlist->head = elist->head;	
			badlist++;
		}
#endif
	}
	/* ADD CODE HERE TO SORT THE BAD LIST AND REMOVE DUPLICATES. */
	free(mfrlist);
	return(listentries);
}

