/*		Copyright (c) 1986, 1987, 1988  Motorola		*/
/*		All Rights Reserved				*/
/*		@(#)m332xctl.c	@(#)m332xctl.c	35.2				*/

#include "sys/types.h"
#include <fcntl.h>
#include <stdio.h>
#include <filehdr.h>
#include <ldfcn.h>
#include <scnhdr.h>
#include <aouthdr.h>
#include <syms.h>
#include <errno.h>
#include "sys/termio.h"

#undef	C_STAT
#include "sys/tty.h"
#include "sys/sxt.h"
#include "sys/mvme332xt.h"

#define	NEX_SECT	16	/* maximum number of excluded sections */
#define	BSS		0	/* bss section id */
#define	TEXT_DATA	1	/* data and text section id */
#define	BUF_SIZE	1024	/* download buffer size */

/*
 * Coff file functions
 */
extern  ldshread();     	/* SYS read section header */
extern  ldsseek();      	/* SYS seek to section data area */
extern  LDFILE 	*ldopen();      /* SYS open load file */
extern	char	*ldgetname();	/* SYS get symbol pointer */
extern  ldclose();      	/* SYS close load file */
extern  exit();         	/* SYS back to OS */
extern  ldohseek();     	/* SYS seek to the optional file header */
extern	errno;

/*
 * Error information
 */
extern	int	errno;		/* SYS return error number */
extern	char	*sys_errlist[];	/* SYS error messages */
extern	int	sys_nerr;	/* SYS number of above messages */

LDFILE  *ldptr;			/* a.out ld file pointer */
FILHDR	filhdr;
SCNHDR	scnhdr;
SYMENT	symbol;

/*
 * Global variables
 */
int	nex_sect = 0;		/* number of sections to be excluded */
char    *ex_sect[NEX_SECT];	/* section names to be excluded */
char	*arg0;			/* argument 0 */
int	dev_fd;			/* printer device fd */
char	*dlfile;		/* download file name pointer */
char	*exec_func;		/* function to be excecuted */
int	h_option;		/* option for -h option */
int	debug;			/* debug mode */
struct dl_info dl_info;		/* download information */
char   buf[BUF_SIZE]; 		/* work buffer */


/*
 * Getopt variables
 */
extern char *optarg;
extern int optind;
extern int opterr;


/*
 * m332xctl:
 *	Usage:
 *	    m332xctl {-t |-r | -R | -D | -h <on|off|info> | -g | -s | 
 * 		{-d <dlfile> [-x <sname>]...   -l | -e <fname>}} dev
 *
 *	Options:
 *		-t :	Test the existence of the MVME332xt. Return 0 if it
 *			exists, else return ENXIO.
 *
 *		-r :	Get firmware and driver version and revision numbers
 *
 *		-R :	Get firmware version number in short format.
 *
 *		-D :	Debug mode.
 *
 *		-g :	Get downloadable area information from 	the MVME332xt.
 *			The address and size of download area is displayed.
 *
 *		-s :	Get symbol table of the MVME332xt firmware and display.
 *
 *		-h :	set or get hardware handshake.
 *			
 *	 	-d :	Download a coff file to the MVME332xt.
 *
 *		-x :	Exclude a section from download.
 *			This option must be preceded by -d option in a command.
 *
 *		-l :	Instruct MVME332xt firmware to copy the downloaded 
 *			line switch table to its internal data structure.
 *			This option must be preceded by -d option in a command.
 *
 *		-e :	Instruct MVME332xt firmware to execute a user function
 *			in a downloaded file.
 *			This option must be preceded by -d option in a command.
 *			
 *	dev:
 *			Device name associated with MVME332xt board.
 *			dev must be a printer device for -g, -d, -r, -R
 *			, -s, -l and -x options.",
 */

main(argc, argv)
int	argc;
char	**argv;
{
	char	*note = (char *)NULL;
	char	dflag = 0;		/* download option flag */
	char	gflag = 0;		/* get download info flag */
	char	sflag = 0;		/* get symbol table flag */
	char	rflag = 0;		/* get version and revision flag */
	char	Rflag = 0;		/* get version and revision flag */
	char	lflag = 0;		/* load line switch table */
	char	eflag = 0;		/* execute function flag */
	char	hflag = 0;		/* set/get hardware handshake */
	char	tflag = 0;		/* test existence flag */
	int	err = 0;		/* parse error count */
	int	c;

	arg0 = argv[0];
	optarg = (char *)NULL;

	while ((c = getopt(argc,argv,"trRDh:gsd:x:le:")) != EOF) {
		switch (c) {
		case 't':	/* test for board existence*/
			tflag++;
			break;

		case 'h':	/* get/set hardware handshake */
			if (optarg == (char *)NULL) {
				err++;
				note = "Missing on/off/info for -h option";
				break;
			}
			if (strcmp(optarg, "off") == 0)
				h_option = 0;
			else if (strcmp(optarg, "on") == 0)
				h_option = 1;
			else if (strcmp(optarg, "info") == 0)
				h_option = 2;
			else {
				err++;
				note = "Invalid option for -h";
				break;
			}
			hflag++;
			break;

		case 'r':	/* get firmware version and revision numbers */
			rflag++;
			break;

		case 'R':	/* get firmware version and revision numbers */
			Rflag++;
			break;

		case 'g':	/* get download address and size */
			gflag = 1;
			break;

		case 's':	/* get symbol table */
			sflag = 1;
			break;

		case 'd':	/* download a coff file */
			if (optarg == (char *)NULL) {
				err++;
				note = "Missing file name";
				break;
			}
			dflag = 1;
			dlfile = optarg;
			break;

		case 'x':	/* section to be excluded from download */
			if (optarg == (char *)NULL) {
				err++;
				note = "Missing section name with -x option.";
				break;
			}

			if (nex_sect >= NEX_SECT) {
				err++;
				note = "Number of excluded sections exceed limit.";
				break;
			}
			ex_sect[nex_sect] = optarg;
			nex_sect++;
			break;

		case 'l':	/* download line switch table from dlfile */
			if (dflag == 0) {
				err++;
				note = "-d option is required before -l option.";
				break;
			}
			lflag++;	
			break;

		case 'e':	/* execute the function main in dlfile */
			if (optarg == (char *)NULL) {
				err++;
				note = "Missing function name with -e option.";
				break;
			}

			if (dflag == 0) {
				err++;
				note = "-d option is required before -e option.";
				break;
			}
			eflag++;	
			exec_func = optarg;
			break;

		case 'D':		/* debug mode */
			debug++;
			break;

		default:			/* all other options */
			err++;
			note = "";
			break;
		}

		if (err) {
			fprintf(stderr, "%s\n", note);
			p_menu();
			exit(EINVAL);
		}
	}
	/*
	 * check the remaining arguments
	 */
	if (optind + 1 != argc || strcmp(argv[1], "-") == 0) {
		note = "Wrong number arguments.";
		fprintf(stderr, "%s\n", note);
		p_menu();
		exit(EINVAL);
	}


	/*
	 * open printer device
	 */
	dev_fd = open(argv[optind], O_WRONLY|O_NDELAY);
	if (dev_fd < 0) {
		if (!tflag)
			printerr("unable to open device");
		exit(ENXIO);
	}
	
	if (tflag)
		exit(0);

	if (hflag)
		do_hwhs();

	if (Rflag)
		do_getVR();

	if (rflag)
		do_getvr();

	if (gflag) 
		do_get();

	if (sflag) 
		do_getsym();

	if (dflag) 
		do_load();

	if (lflag)
		do_linesw();

	if (eflag)
		do_exec();

	exit(0);
}

/*
 * Set/Get hardware handshake.
 */
do_hwhs()
{
	int	hw_info;

	switch(h_option) {
	case 0:
	case 1:
		if (ioctl(dev_fd, TCSETHW, h_option) < 0) {
			printerr("unable to set hardware handshake");
			exit(EIO);
		}
		break;
	case 2:
		if (ioctl(dev_fd, TCGETHW, &hw_info) < 0) {
			printerr("unable to get hardware handshake info");
			exit(EIO);
		}
		printf("hardware handshake is %s\n",
					 hw_info ? "enabled" : "disabled");
		break;
	}
}

/*
 * Get firmware and driver version and revision numbers in short format.
 */
do_getVR()
{
	unsigned long	version;

	if (ioctl(dev_fd, TCGETVR, &version) < 0) {
		printerr("unable to get firmware verion number");
		exit(EIO);
	}
	printf("%d%d\n", (version>>8) & 0xff, version & 0xff); 
}

/*
 * Get firmware version and revision numbers and perform what function.
 */
do_getvr()
{
	unsigned long	version;

	if (ioctl(dev_fd, TCGETVR, &version) < 0) {
		printerr("unable to get firmware verion number");
		exit(EIO);
	}
	printf("MVME332xt Firmware Version %d.%d\n",
			 (version>>8) & 0xff, version & 0xff);
	printf("MVME332xt Driver Version %d.%d\n",
			 (version>>24) & 0xff, (version>>16) & 0xff);

	dl_info.ipc_addr = 0;		/* start from beginning of ROM */

	while (dl_info.ipc_addr != SYMEND) {
		dl_info.host_addr = (long)buf;	/* put what result in buf */
		dl_info.count = BUF_SIZE;	/* avail spaces */
		if (ioctl(dev_fd, TCWHAT, &dl_info) < 0) {
			printerr("unable to perform what function");
			exit(EIO);
		}
		/* display on the CRT */
		if (dl_info.count > 0)
			printf("%.*s", dl_info.count, buf);
	}
}

/*
 * Get information about downloadable area on MVME332xt board.
 */
do_get()
{
	if (ioctl(dev_fd, TCGETDL, &dl_info) < 0) {
		printerr("unable to get download information");
		exit(EIO);
	}
	printf("Downloadable area start address = %x, size = %x\n",
			dl_info.ipc_addr, dl_info.count);
}


/*
 * Get symbol table from the MVME332xt.
 */
do_getsym()
{
	dl_info.ipc_addr = 0;		/* start from symbol table entry 0 */

	while (dl_info.ipc_addr != SYMEND) {
		dl_info.host_addr = (long)buf;	/* put symbol table in buf */
		dl_info.count = BUF_SIZE;	/* avail spaces */
		if (ioctl(dev_fd, TCGETSYM, &dl_info) < 0) {
			printerr("unable to get symbol table");
			exit(EIO);
		}
		/* display on the CRT */
		if (dl_info.count > 0)	
			printf("%.*s", dl_info.count, buf);
	}
}


/*
 * Download a coff file to MVME332xt board
 */
do_load()
{
	register int i;
	register int secnum;

	if ((ldptr = ldopen(dlfile, 0)) == NULL) {
		printerr("unable to open download file");
		exit(ENOENT);
	}

	if (ldfhread(ldptr, &filhdr) == FAILURE) {
		printerr("unable to read download file header");
		exit(EBADF);
	}

	if (debug)
		p_filhdr();

	for (secnum = 1; secnum <= filhdr.f_nscns; secnum++) {
		if (ldshread(ldptr, secnum, &scnhdr) == FAILURE) {
			printerr("unable to read section header");
			exit(EBADF);
		}

		if (debug)
			p_scnhdr();

		/*
		 * Ignore section that has size equal to zero
		 */
		if (scnhdr.s_size == 0)
			continue;

		/*
		 * Is this section to be excluded
		 */
		for (i = 0; i < nex_sect; i++) {
			if (strcmp(ex_sect[i], scnhdr.s_name) == 0)
				break;
		}
		if (i < nex_sect)
			continue;	/* skip processing of this section */

		/*
		 * section with no data, e.g. .bss or with data e.g. .text
		 */
		if (scnhdr.s_scnptr == NULL)
			do_sect(BSS);
		else {
			/* seek to this section */
			if (ldsseek(ldptr, secnum) == FAILURE) {
				printerr("unable to seek to data or text section.\n");
				exit(EBADF);
			}
			do_sect(TEXT_DATA);
		}
	}
	/*
	 * print a newline after all the dots have been displayed
	 */
	if (debug)
		putchar('\n');

	ldclose(ldptr);
}

do_sect(sectype)
int sectype;
{
	register int  i, incr;
	register long address, size;

	if (sectype == BSS) {
		for (i = 0; i < BUF_SIZE; i++)
			buf[i] = 0;
	}
	
	address = scnhdr.s_paddr;
	size = scnhdr.s_size;

	while (size)
	{
		if (debug)
			putchar('.');

		incr = (size < BUF_SIZE ? size : BUF_SIZE);
		if (sectype != BSS) {
			FREAD(buf, 1, incr, ldptr);
			if (debug >= 2)
				p_buffer(buf, incr);
		}

		dl_info.host_addr = (long) buf;
		dl_info.ipc_addr = address;
		dl_info.count = incr;

		if (ioctl(dev_fd, TCDLOAD, &dl_info, 0)) {
			printerr("unable to download");
			exit(EIO);
		}
		size -= incr;
		address += incr;
	}

}

/*
 * Load line switch table to MVME332xt Firmware
 */
do_linesw()
{

	if ((ldptr = ldopen(dlfile, 0)) == NULL) {
		printerr("unable to open download file");
		exit(ENOENT);
	}

	if (ldfhread(ldptr, &filhdr) == FAILURE)
	{
		printerr("unable to read download file header");
		exit(EBADF);
	}

	if (search("linetable") == 0) {
		printerr("cannot find symbol name linetable in download file");
		exit(EBADF);
	}
	dl_info.ipc_addr = symbol.n_value;
			
	if (search("linecount") == 0) {
		printerr("cannot find symbol name linecount in download file");
		exit(EBADF);
	}

	if (ldshread( ldptr, symbol.n_scnum, &scnhdr ) == FAILURE) {
		printerr("cannot read section header for symbol name linecnt");
		exit(EBADF);
	}
	
	FSEEK( ldptr, scnhdr.s_scnptr + (symbol.n_value - scnhdr.s_paddr), BEGINNING );
	FREAD( &dl_info.count, 1, sizeof(long), ldptr);
	
	if (ioctl(dev_fd, TCLINE, &dl_info, 0) < 0) {
		printerr("dolwnload linesw table error");
		exit(EIO);
	}
	ldclose(ldptr);
}

/*
 * Execute the downloaded firmware
 */
do_exec()
{
	if ((ldptr = ldopen(dlfile, 0)) == NULL) {
		printerr("unable to open download file");
		exit(ENOENT);
	}

	if (ldfhread(ldptr, &filhdr) == FAILURE)
	{
		printerr("unable to read file header");
		exit(EBADF);
	}

	/*
	 * Search for function name "main"
	 */
	if (search(exec_func) == 0) {
		printerr("cannot find symbol \"%s\" in download file", exec_func);
		exit(EBADF);
	}
			
	if (ioctl(dev_fd, TCEXEC, symbol.n_value, 0) < 0) {
		printerr("target execution error");
		exit(EIO);
	}
	ldclose(ldptr);
}


/*
 * Search for a symbol in a file, return true if found
 */
search(symname)
char 	*symname;
{

	register int symindex;
	register char *name;

	ldtbseek( ldptr );
	for ( symindex = 0; symindex < filhdr.f_nsyms; symindex++ ) {
		if (ldtbread( ldptr, symindex, &symbol ) == FAILURE) {
			printerr("cannot read symbol table");
			exit(EBADF);
		}
		if ((name = ldgetname( ldptr, &symbol )) == (char *) NULL) {
			printerr("ldgetname returned NULL");
			continue;
		}

		if (debug)
			p_symbol();

		if (strcmp(name, symname) == 0) 
			return(1);
			
		symindex += symbol.n_numaux;
	}
	return(0);
}

char *m332emsg[] = {
	"device is busy",
	"invalid parameter",
	"invalid logical unit number",
	"invalid command",
	"invalid channel number",
};

printerr(msg)
char *msg;
{
	char *emsg;

	if (errno > sys_nerr) {			/* is this system error ? */
		if (IS332ERR(errno))
			emsg = m332emsg[errno - ERR_LOW]; 
		else
			emsg = "unknown error number";
	}
	else
		emsg = sys_errlist[errno];

	fprintf(stderr, "%s: %s, error = %d : %s\n", arg0, msg, errno, emsg);
}

/*
 * Debug aid
 */
p_filhdr()
{
	printf("FILE header: nscns: %d, nsyms: %d\n",
				filhdr.f_nscns, filhdr.f_nsyms);
}


p_symbol()
{
	printf("SYMBOL name=%s, value=%x, scnum=%d, tupe=%d, sclass=%d, numaux=%d\n",
		symbol.n_name, symbol.n_value, symbol.n_scnum, symbol.n_type,
		symbol.n_sclass, symbol.n_numaux);
}

p_scnhdr()
{
	printf("SECTION name: %.8s, paddr: %x, vaddr: %x, size: %x, scnptr: %x, relptr: %x,\n\t lnnoptr: %x, nreloc: %x, nlnno: %x, flags: %x\n",
	    scnhdr.s_name, scnhdr.s_paddr, scnhdr.s_vaddr, scnhdr.s_size,
	    scnhdr.s_scnptr,scnhdr.s_relptr, scnhdr.s_lnnoptr, scnhdr.s_nreloc,
	    scnhdr.s_nlnno, scnhdr.s_flags);
}

p_buffer(pbuf, size)
char *pbuf;
{
/*
 * format :
 * 00110022    0000 0000 0000 0000     0000 0000 0000 0000      ................
 */
	char *fmt = "%08x    %04x %04x %04x %04x    %04x %04x %04x %04x    ";

	register unsigned  short *sbuf;
	register char  c, *bbuf, *cp;

	char fbuf[100];

	printf("BUFFER:\n");

	for (sbuf = (unsigned short *) pbuf; sbuf < (unsigned short *) (pbuf + size);) {
		bbuf = (char *) sbuf;
		cp = fbuf + sprintf(fbuf, fmt, sbuf, 
			*sbuf, *(sbuf+1), *(sbuf+2), *(sbuf+3),
			*(sbuf+4), *(sbuf+5), *(sbuf+6), *(sbuf+7));

		sbuf += 8;

		while (bbuf < (char *) sbuf) {
			c = *bbuf++;
			if (c < ' ' || c >= 0x7f)
				c = '.';
			*cp++ = c;
		}
		*cp++ = '\n';
		*cp = '\000';
		printf(fbuf);
	}
}

char	header[] = 
" Usage: \n\t %s {-t | -r | -R | -D | -h <on|off|info> | -g | -s | \n\t {-d <dlfile> [-x <sname>]...   -l | -e <fname>}} dev";

char	*menu[] = {
" Options:",
"	-t :	Test for MVME332xt existence. Return 0 if it exists.",
"	-r :	Get firmware and driver version and revision numbers",
"	-R :	Get firmware version number in short format",
" 	-D :	Debug mode.",
"	-h :	set or get hardware handshake.",
"	-g :	Get downloadable area information from 	the MVME332xt.",
"		The address and size of download area is displayed.",
"	-s :	Get symbol table of the MVME332xt firmware and display.",
" 	-d :	Download a coff file to the MVME332xt.",
" 	-x :	Excude a section from download.",
"		This option must be preceded by -d option in a command.",
" 	-l :	Instruct MVME332xt firmware to copy the downloaded ",
" 		line switch table to its internal data structure.",
"		This option must be preceded by -d option in a command.",
" 	-e :	Instruct MVME332xt firmware to execute a user function",
" 		in the downloaded file.",
"		This option must be preceded by -d option in a command.",
" dev:",
" 		Device name associated with MVME332xt board.",
"		dev must be a printer device for -g, -d, -r, -R, -s, -l ",
"		 and -x options.",
"",
};

p_menu()
{
	register int index;

	fprintf(stderr, header, arg0);
	for (index = 0; index < sizeof(menu)/sizeof(char *); index++)
		fprintf(stderr, "%s\n", menu[index]);
}

