/*	Copyright (c) 1987 Motorola Inc
	All Rights Reserved
		@(#)m355ctl.c	35.1
*/
/* m355ctl: ioctl(2) interface to 9-track tape driver
 *
 */

#include <stdio.h>
#include <errno.h>
#include <sys/mvme355.h>

extern char *optarg;
extern int optind, getopt();

static char error_buf[BUFSIZ];

/* examine request line, do request, and out
 *
 */
int main(argc, argv, envp)
int argc;
char *argv[], *envp[];
{
#	define USAGE "Usage: %s [-re] [-fn] [special]\n"

	register int c, count, request, usage_error, status;
	register char *m355ctl, *type;
	register FILE *stream;

	status = request = usage_error = 0;
	m355ctl = argv[0];
	stream = stdin;
	type = "r";

	while ((c = getopt(argc, argv, "ref:")) != -1) {

		if (request) {
			fprintf(stderr, "%s: too many requests\n", m355ctl);
			return(2);
		}

		switch (c) {
		case 'r':		/* rewind tape			*/
			request = MTREW;
			break;
		case 'e':		/* erase tape			*/
			request = MTERASE;
			break;
		case 'f':		/* move forward n filemarks	*/
			request = MTFSF;
			if ((count = atoi(optarg)) <= 0) {
				errno = EINVAL;
				sprintf(error_buf, "%s: '%s'", m355ctl, optarg);
				perror(error_buf);
				--usage_error;
			}
			break;
		default:		/* usage error			*/
			--usage_error;
			break;
		}
	}

	if (!usage_error) {
		if (!request) {
			fprintf(stderr, "%s: no request found\n", m355ctl);
			--usage_error;
		} else if (optind < argc - 1) {
			fprintf(stderr, "%s: too many arguments\n", m355ctl);
			--usage_error;
		}
	}

	if (usage_error) {
		fprintf(stderr, USAGE, m355ctl);
		return(2);
	}

	if ((optind == argc - 1) && (freopen(argv[optind], type, stream) == NULL)) {
		sprintf(error_buf, "%s: %s", m355ctl, argv[optind]);
		perror(error_buf);
		return(1);
	}

	switch (request) {
	case MTREW:	/* rewind tape		*/
	case MTERASE:	/* erase to end of tape */
		status = ioctl(fileno(stream), request);
		break;
	case MTFSF:	/* move forward n filemarks */
		while ((count--) && (!status))
			status = ioctl(fileno(stream), request);
		break;
	}

	if (status)
		perror(m355ctl);

	return((status) ? 1 : 0);
}
