#ident "@(#)check_sum.c	35.2"

/* CRC processing based on Check/Crc utilities
 * Copyright 1987, Motorola, Inc.
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <filehdr.h>
#include <ar.h>

extern char *argv0;
extern int errcnt;

#define TEENY_BLOCK	128
#define BIG_BLOCK   	( 64 * 1024 )


#ident	"@(#)help:fetch.h	1.4"

/*		OBSOLETE COMMON ARCHIVE FORMAT


	ARCHIVE File Organization:

	_______________________________________________
	|__________ARCHIVE_HEADER_DATA________________|
	|					      |
	|	Archive Header	"o_ar_hdr"	      |
	|.............................................|
	|					      |
	|	Symbol Directory "o_ar_sym"	      |
	|					      |
	|_____________________________________________|
	|________ARCHIVE_FILE_MEMBER_1________________|
	|					      |
	|	Archive File Header "o_arf_hdr"	      |
	|.............................................|
	|					      |
	|	Member Contents (either a.out.h       |
	|			 format or text file) |
	|_____________________________________________|
	|					      |
	|	.		.		.     |
	|	.		.		.     |
	|	.		.		.     |
	|_____________________________________________|
	|________ARCHIVE_FILE_MEMBER_n________________|
	|					      |
	|	Archive File Header "o_arf_hdr"	      |
	|.............................................|
	|					      |
	|	Member Contents (either a.out.h       |
	|			 format or text file) |
	|_____________________________________________|              */


#define	O_ARMAG	"<ar>"
#define	O_SARMAG	4


struct	o_ar_hdr {			/* archive header */
	char	o_ar_magic[O_SARMAG];	/* magic number */
	char	o_ar_name[16];		/* archive name */
	char	o_ar_date[4];		/* date of last archive modification */
	char	o_ar_syms[4];		/* number of ar_sym entries */
};

struct	o_ar_sym {			/* archive symbol table entry */
	char	o_sym_name[8];		/* symbol name, recognized by ld */
	char	o_sym_ptr[4];		/* archive position of symbol */
};

struct	o_arf_hdr {			/* archive file member header */
	char	o_arf_name[16];		/* file member name */
	char	o_arf_date[4];		/* file member date */
	char	o_arf_uid[4];		/* file member user identification */
	char	o_arf_gid[4];		/* file member group identification */
	char	o_arf_mode[4];		/* file member mode */
	char	o_arf_size[4];		/* file member size */
};


#ident	"@(#)lint:comp:common/manifest	1.4"

typedef unsigned int TWORD;


/* The Key:
 *	LDI	defined and initialized: storage set aside
 *	LIB	defined on a library
 *	LDC	defined as a common region on UNIX
 *	LDX	defined by an extern: if ! pflag, same as LDI
 *	LRV	function returns a value
 *	LUV	function used in a value context
 *	LUE	function used in effects context
 *	LUM	mentioned somewhere other than at the declaration
 *	LDS	defined static object (like LDI)
 */
#ident	"@(#)lint:lmanifest	1.6"
# define LINT_LDI 01
# define LINT_LIB 02
# define LINT_LDC 04
# define LINT_LDX 010
# define LINT_LRV 020
# define LINT_LUV 040
# define LINT_LUE 0100
# define LINT_LUM 0200
# define LINT_LDS 0400

# define LINT_LFN 01000  /* filename record */

typedef struct lint_ty {
	TWORD lint_aty;
	short lint_extra;
} LINT_ATYPE;

typedef struct lint_line {
	short lint_line_decflag;
	char * lint_line_name;
	short lint_line_nargs;
	short lint_line_fline;
	LINT_ATYPE lint_line_type;
} LINT_LINE;

union lint_rec {
	struct lint_line lint_rec_l;
	struct {
		short lint_file_decflag;
		char * lint_file_fn;
		int lint_file_mno;
	} lint_rec_f;
};


/* Main Entry-Point
 *
 */
int
check_sum ( pathlist )
char *pathlist;
{
	void init_crc (), file_error (), close ();
	int open (), get_crc (), calc_crc ();

	int fd;
	struct stat statbuf;

	init_crc();

	if ( ( fd = open( pathlist, 0 ) ) < 0 ) {
		file_error ( pathlist );
		errcnt ++;
	} else if ( fstat ( fd, & statbuf ) < 0 ) {
		file_error ( pathlist );
		errcnt ++;
		close ( fd );
	} else {
		if ( calc_crc ( fd, statbuf . st_size ) )
			file_error ( pathlist );
		close ( fd );
	}

	return( get_crc () );
}

/* File Header Buffer Structure
 *
 */
static union {
	char 			c [ BIG_BLOCK ];
	struct filehdr		f;
	struct ar_hdr		a;
	struct o_ar_hdr		o_a;
	struct o_arf_hdr	o_af;
	union lint_rec		lint_rec;
} u_buf;

static char * buf = u_buf.c;

static struct filehdr * filehdr_ptr =		& u_buf . f;
static struct ar_hdr * ar_hdr_ptr =		& u_buf . a;
static struct o_ar_hdr * o_ar_hdr_ptr =		& u_buf . o_a;
static struct o_arf_hdr * o_arf_hdr_ptr =	& u_buf . o_af;
static union lint_rec * lint_rec_ptr =		& u_buf . lint_rec;


/* Calculate CRC
 *
 */
static int
calc_crc ( fd, count )
int             fd;
off_t		count;
{
	void file_error ();
	int gen_ar_crc (), gen_obs_ar_crc (), gen_a_out_crc (),
		gen_lint_lib_crc (), gen_rest_of_file_crc ();

	int nbytes, status;
	
	nbytes = ( count < TEENY_BLOCK ) ? count : TEENY_BLOCK;

	if ( ( nbytes = read ( fd, buf, nbytes ) ) < 0 )
		return ( -1 );

	if ( ( nbytes >= SARMAG ) && ( strncmp ( ARMAG, buf, SARMAG ) == 0 ) )
		return ( gen_ar_crc ( fd, count, nbytes ) );

	if ( ( nbytes >= O_SARMAG ) &&
	     ( strncmp ( O_ARMAG, buf, O_SARMAG ) == 0 ) )
		return ( gen_obs_ar_crc ( fd, count, nbytes ) );

	if ( ( nbytes >= sizeof ( filehdr_ptr -> f_magic ) ) && 
	     ( filehdr_ptr -> f_magic == MC68MAGIC ) )
		return ( gen_a_out_crc ( fd, count, nbytes ) );

	if ( ( nbytes >= sizeof ( lint_rec_ptr -> lint_rec_f ) ) &&
	     ( lint_rec_ptr -> lint_rec_f . lint_file_decflag == LINT_LFN ) &&
	     ( lint_rec_ptr -> lint_rec_f . lint_file_fn == NULL ) )
		return ( gen_lint_lib_crc ( fd, count, nbytes ) );

	if ( nbytes != 0 )
		return ( gen_rest_of_file_crc ( fd, count, nbytes ) );

	/* null length */
	return ( 0 );
}


/* Generate A-Dot-Out File's CRC
 *
 */
static int
gen_a_out_crc ( fd, count, nbytes )
int fd, nbytes;
off_t count;
{
	int gen_rest_of_file_crc ();

	filehdr_ptr -> f_timdat = 0L;
	return ( gen_rest_of_file_crc ( fd, count, nbytes ) );
}


/* Generate Ar(1) File CRC
 *
 */
static int
gen_ar_crc ( fd, count, nbytes )
int fd, nbytes;
off_t count;
{
#	define AR_HDR_SZ	sizeof ( struct ar_hdr )
#	define AR_DATE_SZ	sizeof ( ar_hdr_ptr -> ar_date )


	extern long lseek (), strtol ();
	extern char * memset ();
	void gen_buf_crc (), init_crc ();

	long size;

	if ( lseek ( fd, (long) -( nbytes - SARMAG ), 1 ) < 0L )
		return ( -1 );

	gen_buf_crc ( buf, SARMAG );
	count -= SARMAG;

	while ( ( count >= AR_HDR_SZ ) &&
		( nbytes = read ( fd, buf, AR_HDR_SZ ) ) == AR_HDR_SZ ) {

		memset( ar_hdr_ptr -> ar_date, ' ', AR_DATE_SZ );
		gen_buf_crc ( buf, AR_HDR_SZ );
		count -= AR_HDR_SZ;

		size = ( strtol ( ar_hdr_ptr -> ar_size, NULL, 10 ) + 1L )
				& ( ~ 1L );

		if ( ( count < size ) || ( calc_crc ( fd, size ) ) ) {
			errno = EINVAL;
			init_crc ();
			return ( -1 );
		}
		count -= size;
	}
	if ( count ) {
		errno = EINVAL;
		return ( -1 );
	}
	return ( 0 );
}


/* Generate Obsolete Archive File CRC
 *
 */
static int
gen_obs_ar_crc ( fd, count, nbytes )
int fd, nbytes;
off_t count;
{
#	define O_AR_HDR_SZ	sizeof ( struct o_ar_hdr )
#	define O_AR_DATE_SZ	sizeof ( o_ar_hdr_ptr -> o_ar_date )

#	define ARF_HDR_SZ	sizeof ( struct o_arf_hdr )
#	define ARF_DATE_SZ	sizeof ( o_arf_hdr_ptr -> o_arf_date )

	extern long strtol ();
	char * memset ();
	void gen_buf_crc (), init_crc ();

	long size;

	if ( nbytes < O_AR_HDR_SZ ) {
		errno = EINVAL;
		return ( -1 );
	}

	memset ( o_ar_hdr_ptr -> o_ar_date, '\0', O_AR_DATE_SZ );
	size = sgetl ( o_ar_hdr_ptr -> o_ar_syms ) *
			sizeof ( struct o_ar_sym  ) + O_AR_HDR_SZ;

	if ( ( lseek ( fd, (long) -( nbytes - O_AR_HDR_SZ ), 1 ) < 0L ) ||
	     ( gen_rest_of_file_crc ( fd, size, O_AR_HDR_SZ ) ) )
		return ( -1 );

	count -= O_AR_HDR_SZ;
	while ( ( count > ARF_HDR_SZ ) &&
		( ( nbytes = read ( fd, buf, ARF_HDR_SZ ) ) == ARF_HDR_SZ ) ) {

		memset( o_arf_hdr_ptr -> o_arf_date, '\0', ARF_DATE_SZ );
		gen_buf_crc ( buf, ARF_HDR_SZ );
		count -= ARF_HDR_SZ;

		size = ( sgetl ( o_arf_hdr_ptr -> o_arf_size, NULL, 10 ) + 1L )
				& ( ~ 1L );

		if ( calc_crc ( fd, size ) ) {
			errno = EINVAL;
			init_crc ();
			return ( -1 );
		}
		count -= size;
	}
	if ( count ) {
		errno = EINVAL;
		return ( -1 );
	}
	return ( 0 );
}


/* Generate Lint Library CRC
 *
 */
static int
gen_lint_lib_crc ( fd, count, nbytes )
int fd, nbytes;
long count;
{
#	define LREC_SZ		sizeof ( union lint_rec )
#	define NOT_LINT_FLAGS	( ~ ( LINT_LDI | LINT_LIB | LINT_LDC | \
					LINT_LDX | LINT_LRV | LINT_LUV | \
					LINT_LUE | LINT_LUM | LINT_LDS | \
					LINT_LFN ) )

	int gen_false_lint_lib_crc (), gen_rest_of_file_crc ();
	void gen_buf_crc ();

	extern int errno, read ();
	extern long lseek ();

	int type, nargs;
	char * p;

	long save_count;
	unsigned int save_crchi, save_crclo;
	extern unsigned int crchi, crclo;

	save_count = count;
	save_crchi = crchi;
	save_crclo = crclo;

	if ( lseek ( fd, (long) - nbytes, 1 ) < 0L )
		return ( -1 );

	while ( ( count >= LREC_SZ ) &&
		( ( nbytes = read ( fd, buf, LREC_SZ ) ) == LREC_SZ ) ) {
		/*
		 *	If this is a new file name and module number then
		 *		Set module number to zero
		 *		so that consistent results
		 *		can be achieved;
		 *	Else
		 *		Save the number of arguments description
		 *		structures associated to the given record;
		 *	Endif;
		 */
		type = lint_rec_ptr -> lint_rec_l . lint_line_decflag;

		if ( type & NOT_LINT_FLAGS )
			return ( gen_false_lint_lib_crc ( fd,
					save_count, ( long ) ( count - nbytes ),
					save_crchi, save_crclo ) );

		if ( type & LINT_LFN )
			lint_rec_ptr -> lint_rec_f . lint_file_mno = 0;
		else
			nargs = lint_rec_ptr -> lint_rec_l . lint_line_nargs;

		gen_buf_crc ( buf, LREC_SZ );
		count -= LREC_SZ;

		/*
		 *	Read lint flex-name record.
		 */
		if ( ( nbytes = read ( fd, buf, TEENY_BLOCK ) ) < 0 )
			return ( -1 );
		
		for ( p = buf; ( * p != '\0' ); p ++ )
			if ( ! isascii ( * p ) )
				return ( gen_false_lint_lib_crc ( fd,
						save_count,
						( long ) ( count - nbytes ),
						save_crchi, save_crclo ) );

		++ p;
		if ( lseek ( fd, (long) - ( nbytes - ( p - buf ) ), 1 ) < 0L )
			return ( -1 );

		nbytes = p - buf;
		gen_buf_crc ( buf, nbytes );
		count -= nbytes;

		if ( ! ( type & LINT_LFN ) ) {
			/*
			 *	Number of arguments is negative,
			 *	plus one for VARARGS.
			 */
			if ( nargs < 0 )
				nargs = -nargs - 1;

			nbytes = nargs * sizeof ( LINT_ATYPE );
			if ( gen_rest_of_file_crc ( fd, ( long ) nbytes, 0 ) )
				return ( -1 );

			count -= nbytes;
		}
	}
	if ( nbytes < 0 )
		return ( -1 );
	if ( count )
		return ( gen_false_lint_lib_crc ( fd, save_count, count,
					save_crchi, save_crclo ) );
	return ( 0 );
}


/* Generate CRC For Lint-Library-Like File
 *
 */
static int
gen_false_lint_lib_crc ( fd, save_count, count, save_crchi, save_crclo )
int fd;
long save_count, count;
unsigned int save_crchi, save_crclo;
{
	int gen_rest_file_crc ();
	extern long lseek ();
	extern unsigned int crchi, crclo;

	if ( lseek ( fd, ( long ) ( count - save_count ), 1 ) < 0L )
		return ( -1 );

	crchi = save_crchi;
	crclo = save_crclo;
	return ( gen_rest_of_file_crc ( fd, save_count, 0 ) );
}


/* Generate CRC For File
 *
 */
static int
gen_rest_of_file_crc ( fd, count, nbytes )
int fd, nbytes;
long count;
{
#	define BUF_SZ	( ( count < BIG_BLOCK ) ? count : BIG_BLOCK )

	void gen_buf_crc (), file_error (), init_crc ();

	do {
		gen_buf_crc( buf, nbytes );
		count -= nbytes;
	} while ( ( count > 0 ) &&
		  ( ( nbytes = read ( fd, buf, BUF_SZ ) ) > 0 ) );

	if ( nbytes < 0 ) {
		init_crc ();
		return ( -1 );
	}
	return ( 0 );
}


/* CRC Intermediaries
 *
 */
static unsigned int
crchi = 0,
crclo = 0;


/* Initialize CRC
 *
 */
static void
init_crc()
{
	crchi = crclo = 0;
}


/* LOBYTE & HIBYTE Macros
 *
 */
#define LOBYTE(c)       ( ( c ) & 0x00FF )
#define HIBYTE(c)	( ( c ) & 0xFF00 )


/* Get Generated CRC
 *
 */
static int
get_crc()
{

	return ( ( LOBYTE ( crchi ) << 8 ) | LOBYTE ( crclo ) );
}


/* Generate CRC For Buffer
 *
 */
static void
gen_buf_crc( buf, nbytes )
char * buf;
int nbytes;
{
	extern int Mval [];
	register unsigned d;

	while ( nbytes -- ) {
		d = LOBYTE ( crchi ^= ( * buf ++ ) );
		crchi = LOBYTE ( Mval[ d ] ) ^ crclo;
		crclo = LOBYTE ( HIBYTE ( Mval[ d ] ) >> 8 );
	}
	return;
}


/* Declare File Error
 *
 */
static void
file_error ( pathlist )
char * pathlist;
{
	fprintf( stderr, "%s: %s: ", argv0, pathlist );
	perror ( NULL );
	return;
}

static int Mval[] = {
   0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241,
   0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
   0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40,
   0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
   0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40,
   0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
   0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641,
   0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
   0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240,
   0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
   0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41,
   0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
   0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41,
   0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
   0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640,
   0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
   0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240,
   0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
   0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41,
   0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
   0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41,
   0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
   0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640,
   0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
   0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241,
   0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
   0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40,
   0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
   0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40,
   0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
   0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641,
   0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040
};

