/*
 *		Copyright Motorola, Inc. (c) 1988
 *		All Rights Reserved
 *		@(#)msledit.c	35.5
 *		
 *		mini-sledit 
 *		sledit without all the "curses"
 */

#include	<stdio.h>
#include	<ctype.h>
#include	<sys/signal.h>
#include	<sys/types.h>
#include	<sys/param.h>
#include	<sys/filsys.h>
#include	<sys/stat.h>
#include	<ustat.h>
#include	<fcntl.h>
#include	<sys/ino.h>
#include	<sys/dk.h>

struct sledit {
	unsigned char s_flag;		/* readonly flag */
	short slice;			/* slice */
	struct slice s_tab;		/* slice table entry */
	struct filsys fs;		/* file system superblock */
};

/*	s_flag	*/

#define RDONLY 		0x1		/* readonly */
#define MODIFIED	0x2		/* modified */
#define FSMOD		0x4		/* filesystem modified */
#define NMMOD		0x8		/* filesys name modified */

struct edit {
	short	e_flags;
	char	*e_prompt;
	char	*e_fmt;
	char	*e_help;
} edit[] = {


{/* 0 */	NULL, "offset", "%d ",
"The number of blocks from the beginning of the disk.\n\
This is calculated by adding the previous slice's offset\n\
and the previous slice's size.\n"
},
{/* 1 */	NULL, "slice size", "%d ",
"The number of blocks contained in the slice.\n\
This number plus the offset must not overlap the next slice.\n"
},
{/* 2 */	NULL, "filesystem size", "%d ",
"The number of blocks contained in the filesystem.\n\
This number must be less than or equal to the slice size.\n"
},
{/* 3 */	NULL, "filesystem name", "%.6s ",
"The name of the filesystem (6 characters).\n"
},
{/* 4 */	NULL, "vol-id name", "%.6s ",
"The name of the volumeid (6 characters).\n"
},

{/* end */	NULL, NULL, 0, NULL
},

};
#define C_OFF		0
#define C_SLICE		1
#define	C_FSSIZE	2
#define C_FSNAME	3
#define C_VOLID		4

#define FTYPE(x)	(x & 0170000)	/* file type (from stat(2)) */

char	vflag = 0;		/* ask questions for mkfs */
char	pflag = 0;		/* print flag */
char	mflag = 0;		/* no mkfs flag */
char	sflag = 0;		/* slice copy  flag */
char	*me;			/* name of program */
struct dkblk0  dkbuf;
char	*dev;			/* device name */
char	*slicedev;		/* slice copy device name */
slices = 0;			/* number of slices read from disk */
static struct sledit *slicetab = NULL;
struct slice  slicebuff[BSIZE/sizeof(struct slice)];
char changed = 0;		/* has anything been changed */

/* file system block size */
#if (vax || u3b || u3b5 || m68k) && (FsTYPE == 3)
#define FSBSIZE	(BSIZE*2)
#else
#define	FSBSIZE BSIZE
#endif

#define	NBINODE	(FSBSIZE/sizeof(struct dinode))
main( argc, argv )
int argc;
register char **argv;
{	int optchar;
	register arg;
	register fd;
	register i;
	register n;
	register char *cp;
	extern	char	*optarg;
	extern	int	optind;
	me = argv[0];

	while(( optchar = getopt( argc, argv, "s:mpv" )) != EOF ) {
		switch(optchar) {
		case 'v':	vflag |= RDONLY;
				break;
		case 'p':	pflag++;
				break;
		case 'm':	mflag++;
				break;
		case 's':	sflag++;
				slicedev = optarg;
				break;
		case '?':
		default:	usage();
				break;
		}
	}
	if(( arg = optind ) == argc ) {
		fprintf( stderr,"Missing device file\n");
		usage();
	}

	dev = argv[arg];

	if(sflag) {
		if(( fd = open( slicedev, O_RDONLY )) == -1 ) {
			fprintf(stderr,"Can't open %s\n",dev );
			exit(-1);
		}
		if( readslice(fd,slicedev,0) == -1)
			exit(-1);
		close(fd);
	}
	if(( fd = open( dev, O_RDWR )) == -1 ) {
		fprintf(stderr,"Can't open %s\n",dev );
		exit(-1);
	}
	if( readslice(fd,dev,sflag) == -1)
		exit(-1);
	if( readfilsys(fd) == -1)	/* read the disk filesys info */
		exit(-1);
	prslices();
	if(!pflag)
		if(editslices()){
			writedisk(fd,dev);
			if(!mflag)
				filsys(dev);
			prslices();
		}
	close(fd);
}

/*
 *	readslice - read slicetable  into buffer
 */

readslice(fd,dev,flag)
register fd;
register char *dev;
register flag;
{
	register struct dkblk0 *dp = &dkbuf;
	register struct sledit *sp;
	register struct slice *bp = slicebuff;
	register i;
	static char *mot = "MOTOROLA";
	static char *exor = "EXORMACS";


	/*
	 *	read vid and config info
	 */


	if( read(fd,dp,sizeof(struct dkblk0)) == -1) {
		fprintf(stderr,"Volume id read error %s\n",dev);
		return(-1);
	}

	if( strncmp(mot,dp->dk_vid.vid_mac,8) && strncmp(exor,dp->dk_vid.vid_mac,8)) {
		fprintf(stderr,"Volume id read error %s\n",dev);
		return(-1);
	}

	if((slices = dp->dk_ios.ios_slccnt) == 0) {

		fprintf(stderr,"No slice table %s",dev);
		return(-1);
	}

	if(flag) /* already have slicing info */
		return(slices);

	/*
	 *	read slice table
	 */

	if(lseek(fd,dp->dk_ios.ios_slcblk*BSIZE,0) == -1) {
		fprintf(stderr,"Slice table seek error\n");
		return(-1);
	}

	if( read(fd,bp,BSIZE) == -1) {
		fprintf(stderr,"Slice table read error\n");
		return(-1);
	}
	slicetab = (struct sledit *) malloc(slices*((sizeof(struct sledit))));

	if( slicetab == NULL )	{	/* malloc slice table fails */
		fprintf(stderr,"No space for slice table \n");
		return(-1);
	}

	/* copy slicebuf to slicetab */

	for(i = 0, sp = slicetab; i < slices; i++,sp++,bp++ ) {
		sp->s_tab.nblocks = bp->nblocks;
		sp->s_tab.blkoff = bp->blkoff;
		sp->slice = i;
		sp->s_flag = 0;
		if(i == slices - 1) 	/* last slice, entire disk */
			sp->s_flag =  RDONLY; /* default */
		
	}
	return (slices);
}

/*
 *	read super blocks
 */

readfilsys(fd)
register fd;
{
	register struct sledit *sp;
	register unsigned short i;

	for(i = 0,sp = slicetab; i < slices; i++,sp++ ) {

		if(sp->s_tab.nblocks == 0) { /* no space for filesys */
			sp->fs.s_fsize = 0;
			*sp->fs.s_fname = NULL;
			*sp->fs.s_fpack = NULL;
			continue;
		}

		if(lseek(fd,(sp->s_tab.blkoff + 1)*BSIZE,0) == -1) {	
			fprintf("Super block seek error","",i);
			return(-1);
			}
		if( read(fd,&sp->fs.s_isize,sizeof(struct filsys)) == -1) {
			fprintf("Super block read error","",i);
			return(-1);
		}
		
		if(sp->fs.s_magic != FsMAGIC) {		/* no filesys */
			sp->fs.s_fsize = 0;
			*sp->fs.s_fname = NULL;
			*sp->fs.s_fpack = NULL;
		} else					/* make 1/2K blocks */
			sp->fs.s_fsize *= sp->fs.s_type;


	}
	return(0);
}
prslices()
{
	register struct sledit *sp;
	register i;

	printf("\nslice\toffset\tsl size\tfs size\tfsname\tvol-id\n");
	for(i = 0, sp = slicetab; i < slices; i++,sp++ ) {
		printf("%d\t%d\t%d\t%d\t%.6s\t%.6s\n",sp->slice,sp->s_tab.blkoff,
			sp->s_tab.nblocks,sp->fs.s_fsize,sp->fs.s_fname,
			sp->fs.s_fpack);
	}
}

editslices()
{
	register struct sledit *sp;
	register i;
	register ret;

	while (1) {
		for(i = 0, sp = slicetab; i < slices-1; ) {
			printf("\n");
			switch(editfield(sp)) {
			case 1: return(1); /* ok to write disk */
			case -1: return(0); /* don't write */
			case 2: if(i==0)  /* wrap around */
					sp = &slicetab[(i=slices-2)];
				else{
					sp--;
					i--;
				}
				break;
			case 0:
			default:
				sp++;i++;
			}
		}
	}
}

/*
 *	edit the fields in each slice
 */

char dotbuf[BUFSIZ];   /* buffer for last input string */

editfield(sp)
register struct sledit *sp;
{	register struct edit *ep = edit;
	char line[BUFSIZ];
	
	for( ep = edit; ep->e_prompt != NULL; ep++ ) {
		printf("slice %d>",sp->slice);
		printf( "%30.30s: ", ep->e_prompt );
		prfield(sp,ep,stdout);
		fflush( stdout );

		line[0] = '\0';

		if( fgets( line, BUFSIZ, stdin ) == NULL )
			continue;

		if(( line[0] == '\n' ) || ( line[0] == '\0' )) {
			continue;
		}
		if( strcmp( line, "q\n" ) == 0 ) {
			if( changed ) {
				printf( "\nNo write since last change\n" );
				printf( "(q! overrides)\n\n" );
				continue;
			}
			return(1);
		}

		if( line[0] == '!' ) {	/* shell escape */
			shesc(&line[1]);
			continue;
		}

		if( strcmp( line, "q!\n" ) == 0 )
			return(-1);

		if( line[0] == 'w' ) {	/* write */
			if( line[1] == '\n' ) {	/* write */
				if(checkslice()){
					changed = 0;
					printf("\n`%s' written\n\n", dev );
				}
				continue;
			}
			if( line[1] == '!' ) {	/* write no matter what */

				changed = 0;
				printf("\n`%s' written\n\n", dev );
				continue;
			}
			if( line[1] == ' ' ) {	/* write to file */
				writeother(&line[2]);
				continue;
			}
		}

		if( strcmp( line, "?\n" ) == 0 ) {
			printf("\n%s\n", ep->e_help );
			--ep;
			continue;
		}
		if( strcmp( line, "-\n" ) == 0 ) {
			if( ep == edit )
				return(2);
			--ep;
			--ep;
			continue;
		}
		if( strcmp( line, "t\n" ) == 0 ) {
			prslices();
			--ep;
			continue;
		}
		if( strcmp( line, "help\n" ) == 0 ) {
			printf("\nLegal commands:\n");
			printf("\t<newline>\tstep to next field\n");
			printf("\t-\t\tstep to previous field\n");
			printf("\tO\t\topen space command\n");
			printf("\tC\t\tclose space command\n");
			printf("\tt\t\tprint slice table\n");
			printf("\tw\t\twrite slice table\n");
			printf("\tw file\t\twrite slice table to file\n");
			printf("\tq\t\tquit edit\n");
			printf("\tq!\t\tquit edit (no write)\n");
			printf("\t!\t\texecute interactive shell\n");
			printf("\t!cmd\t\texecute shell command\n");
			printf("\t?\t\tprint information about field\n");
			printf("\t.\t\tuse previous input string\n");
			printf("\thelp\t\tprint this help summary\n\n");
			continue;
		}
		if( line[0] == 'O' ) {	/* Open command  */
			if(doopen(sp))
				changed++;
			continue;
		}

		if( line[0] == 'C' ) {	/* Close command  */
			if(doclose(sp))
				changed++;
			continue;
		}
		if( line[0] == '.' ) {	/* previous string  */
			strcpy(line,dotbuf);
			if( dochange(sp,ep,line) ) 
				changed++;
			ep--;
			continue;
		}
		if( dochange(sp,ep,line) ) 
			changed++;
		strcpy(dotbuf,line);

	}
	return(0);
}

/*
 *	prfield: print field contents
 */

prfield(sp,ep,fp)
register struct sledit *sp;
register struct edit *ep;
register FILE *fp;
{
	if(( ep == NULL ) || ( ep->e_prompt == NULL ))
		return;

	switch(ep-edit) {
	case C_OFF:
		fprintf( fp, ep->e_fmt, sp->s_tab.blkoff );
		break;
	case C_SLICE:
		fprintf( fp, ep->e_fmt, sp->s_tab.nblocks );
		break;
	case C_FSSIZE:
		fprintf( fp, ep->e_fmt, sp->fs.s_fsize );
		break;
	case C_FSNAME:
		fprintf( fp, ep->e_fmt, sp->fs.s_fname );
		break;
	case C_VOLID:
		fprintf( fp, ep->e_fmt, sp->fs.s_fpack );
		break;
	default:
		fprintf( fp, "unknown field");
		break;
	}
}

/*
 *	doclose: close space 
 */

doclose(sp)
register struct sledit *sp;
{
	register struct sledit *psp; /* previous slice */

	if(sp == slicetab) {
		fprintf(stderr,"Can't close slice 0\n");
		return(0);
	}
	psp = sp - 1; /* find previous slice with space */
	while(psp->s_tab.nblocks == 0 && psp != slicetab)
		psp--;

	psp->s_tab.nblocks += sp->s_tab.nblocks;
	/* clean up fields */
	sp->s_tab.nblocks = 0;
	sp->s_tab.blkoff = 0;
	sp->fs.s_fsize = 0;
	sp->fs.s_fname[0] = '\0';
	sp->fs.s_fpack[0] = '\0';
	return(1);
}
/*
 *	doopen: open space for new slice
 */

doopen(sp)
register struct sledit *sp;
{
	register struct sledit *psp; /* previous slice */
	register space,offset;

	if(sp == slicetab) {
		fprintf(stderr,"Can't open slice 0\n");
		return(0);
	}
	if(sp->s_tab.blkoff) {
		fprintf(stderr,"Space already allocated for slice %d\n",sp->slice);
		return(0);
	}
	psp = sp - 1; /* find previous slice with space */
	while(psp->s_tab.nblocks == 0 && psp != slicetab)
		psp--;

	if(psp->s_tab.nblocks <= psp->fs.s_fsize) {
		fprintf(stderr,"No space to allocate for slice %d\n",sp->slice);
		return(0);
	}
	if(psp->fs.s_fsize == 0) {
		fprintf(stderr,"No filesystem on slice %d\n",psp->slice);
		return(0);
	}
	space = psp->s_tab.nblocks - psp->fs.s_fsize;
	psp->s_tab.nblocks = psp->fs.s_fsize;
	offset = psp->s_tab.nblocks +  psp->s_tab.blkoff;
	sp->s_tab.nblocks = space;
	sp->s_tab.blkoff = offset;
	return(1);
}

/*
 *	dochange: execute change in field
 */

dochange(sp,ep,line)
register struct sledit *sp;
register struct edit *ep;
register char *line;
{	register chg = 0;
	register help = 0;
	unsigned n;

	if(( ep == NULL ) || ( ep->e_prompt == NULL ))
		return(0);


	switch( ep->e_fmt[strlen(ep->e_fmt)-2] ) {
	case 'd':	/* decimal */
	case 'x':	/* hex */
		chg = sscanf( line, ep->e_fmt, &n );
		break;
	case 's':	/* string */
		chg = 1;
		line[strlen(line)-1] = '\0';	/* kill \n */
		break;
	default:
		chg = 0;
		break;
	}

	if( chg == 0 ) {	/* scanf failure or blank line */
		printf("\n%s\n\n", ep->e_help );
		return(chg);
	}
	sp->s_flag |= MODIFIED;
	switch(ep-edit) {
	case C_OFF:
		sp->s_tab.blkoff = n;
		if(sp->fs.s_fsize )	/* there is a filesystem,must remake */
			sp->s_flag |= FSMOD|NMMOD;
		break;
	case C_SLICE:
		sp->s_tab.nblocks = n;
		break;
	case C_FSSIZE:
		sp->fs.s_fsize = n;
		sp->s_flag |= FSMOD|NMMOD;
		break;
	case C_FSNAME:
		strncpy(sp->fs.s_fname, line ,6);
		sp->s_flag |= NMMOD;
		break;
	case C_VOLID:
		strncpy(sp->fs.s_fpack, line ,6);
		sp->s_flag |= NMMOD;
		break;
	default:
		fprintf( stderr, "unknown field");
		break;
	}

	return(chg);
}

/*
 *	checkslice - check slice offsets and sizes for overlap 
 * 			except for last slice
 */

checkslice()
{
	register struct sledit *sp;
	register struct sledit *np;
	register i;
	register j;
	register high,low;
	register range;
	register lastslc = slices - 1; 	/* everything but last slice */
	int lastoff;

	/*
	 *	check slice offsets and sizes for overlap
	 */
	
	lastoff = slicetab[lastslc].s_tab.nblocks; /* greatest offset allowed */

	for(i = 0,sp = slicetab; i < lastslc; i++, sp++ ) {
		if(sp->s_tab.nblocks < sp->fs.s_fsize) {
			fprintf(stderr, "Slice %d smaller then filesystem\n",i);
			return(0);
		}
		range = sp->s_tab.nblocks + sp->s_tab.blkoff;

		for(j=0,np = slicetab; j < lastslc; j++ , np++) {
			if(j == i ) /* this slice */
				continue;

			low = np->s_tab.blkoff;
			high = np->s_tab.nblocks + np->s_tab.blkoff;

			if(((sp->s_tab.blkoff < high) && 
			  (sp->s_tab.blkoff >= low)) ||
			  ((sp->s_tab.blkoff < low) && (range > low))) {

				fprintf(stderr,"Slice %d overlaps with slice %d\n",i,j);
				return(0);
			}
		}
		if( range > lastoff) {
			fprintf(stderr,"Slice %d beyond end of disk\n",i);
			return(0);
		}

			
	}
	return(1);
}

/*
 *	writedisk - write buffer into slice table on disk
 */

writedisk(fd,dev)
register fd;
register char *dev;
{
	register struct dkblk0 *dp = &dkbuf;
	register struct sledit *sp;
	register i;
	register struct slice *bp = slicebuff;
	struct stat statarea;
	register struct stat *statp = &statarea;


	/*
	 *	write slice table
	 */

	/* set driver slice table first */

	for(i = 0,sp = slicetab; i < slices; i++, sp++, bp++ ) {
		bp->nblocks = sp->s_tab.nblocks;
		bp->blkoff = sp->s_tab.blkoff;
	}

	if(stat(dev,statp) < 0) {
		fprintf(stderr,"%s: Can't stat %s\n",me,dev);
		return(-1);
	}
	if (FTYPE(statp->st_mode) != S_IFREG)  /* only if a real device */
		if(ioctl(fd,DKSETSLC,slicebuff) == -1) {
			fprintf(stderr,"Slice table ioctl error.\n");
			return(-1);
		}

	/* ok, write to disk */

	if(lseek(fd,dp->dk_ios.ios_slcblk*BSIZE,0) == -1) {
		fprintf(stderr,"Slice table seek error\n.");
		return(-1);
		}
	if( write(fd,slicebuff,BSIZE) == -1) {
		fprintf(stderr,"Slice table write error\n.");
		return(-1);
	}

	return(0);
}

/*
 *	writeother - write buffer into another file
 */

writeother(file)
register char *file;
{
	register struct dkblk0 *dp = &dkbuf;
	register struct sledit *sp;
	register i;
	register struct slice *bp = slicebuff;
	struct stat statarea;
	register struct stat *statp = &statarea;
	register fd;

	file[strlen(file)-1] = '\0';	/* kill \n */
	if(*file == NULL) {
		fprintf(stderr,"File name missing\n");
		return(-1);
	}

	/*
	 *	write slice table
	 */

	/* set driver slice table first */

	for(i = 0,sp = slicetab; i < slices; i++, sp++, bp++ ) {
		bp->nblocks = sp->s_tab.nblocks;
		bp->blkoff = sp->s_tab.blkoff;
	}

	if(( fd = open( file, O_CREAT | O_TRUNC | O_WRONLY,0666)) == -1 ) {
		fprintf(stderr,"Can't open %s\n",file);
		return(-1);
	}

	if(fstat(fd,statp) < 0) {
		fprintf(stderr,"Can't stat %s\n",file);
		close(fd);
		return(-1);
	}
	if (FTYPE(statp->st_mode) != S_IFREG) {
		fprintf(stderr,"Not a regular file: %s\n",file);
		close(fd);
		return(-1);
	}

	/* write volid/config block to disk */

	i = dp->dk_ios.ios_slcblk;	/* save slice block number */
	dp->dk_ios.ios_slcblk = 1;	/* change to 1 to save space */

	if( write(fd,dp,sizeof(struct dkblk0)) == -1) {
		fprintf(stderr,"Volume id write error %s\n",file);
		close(fd);
		return(-1);
	}

	dp->dk_ios.ios_slcblk = i;	/* change back */

	if( write(fd,slicebuff,BSIZE) == -1) {
		fprintf(stderr,"Slice table write error %s\n",file);
		close(fd);
		return(-1);
	}

	close(fd);
	printf("\n`%s' written\n\n", file );
	return(0);
}

/*
 *	shesc: shell escape
 */

shesc(cmd)
register char *cmd;
{	register pid;
	register w;
	int	status;

	cmd[strlen(cmd)-1] = '\0';	/* kill newline */

	if((pid = fork()) == 0) {
		if( *cmd == '\0' )
			execl("/bin/sh", "sh", 0);
		else
			execl("/bin/sh", "sh", "-c", cmd, 0);
		exit(0);
	}

	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);

	while(((w = wait(&status)) != pid) && (w != -1));

	fprintf( stdout, "\nHit return to continue ");
	fflush(stdout);

	while( fgetc( stdin ) != '\n' );
	printf("\n");

	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
}

/*
 *	filsys - make a change to the file systems
 */

filsys(dev)
register char *dev;
{
	register i;
	register struct sledit *sp;

	for(i = 0,sp = slicetab; i < slices - 1; i++,sp++ ) {
		if(sp->fs.s_fsize == 0)		/* no filesys */
			continue;
		if((sp->s_flag & FSMOD) )	/*  filsys change */
			mkfs(dev,i,sp->fs.s_fsize);
		if((sp->s_flag & NMMOD) )	/*  name change */
			label(dev,i,sp);
	}
		
}
/*
 *	mkfs - call mkfs(1M) to make new filesystems
 */

mkfs(lrdev, sl, sl_size)
register char *lrdev, sl;
register int sl_size;
{
	register char *c;
	register unsigned short gapdef = 1;
	register unsigned short blockdef = 400;
	char cmd[120];
	char dev[120];
	char inodes[10];
	char gap[10];
	char blocks[10];
	static char *mkfs = "/etc/mkfs";

	strcpy(dev, lrdev);

	/* compute "base name" */

	for( c = dev + (strlen(dev) -1 ); isdigit(*c ) ; *c-- ='\0'  );

	*inodes = NULL;
	*gap = NULL;
	*blocks = NULL;
	if(vflag ) {
		fprintf(stdout,"\nMaking filesystem on slice %s%c\n",dev,sl);
		fprintf(stdout,"\tNumber of inodes? (default %d): ",
			NBINODE*(sl_size/
			((FSBSIZE/BSIZE)*(NBINODE*4))));
		fflush(stdout);
		gets(inodes,10,stdin);
		fprintf(stdout,"\tGap size? (default %d): ",gapdef);
		fflush(stdout);
		gets(gap,10,stdin);
		fprintf(stdout,"\tBlocks/cyl? (default %d): ",blockdef);
		fflush(stdout);
		gets(blocks,10,stdin);
	}
	sprintf(cmd, "%s %s%d %d", mkfs,dev,sl,sl_size);
	if(*inodes) {
		strcat(cmd,":");
		strcat(cmd,inodes);
	}
	if(*gap || *blocks) {	/* gaps and blocks must both be present */
		if(*gap == NULL)
			sprintf(gap,"%d",gapdef);
		strcat(cmd," ");
		strcat(cmd,gap);
		if(*blocks == NULL)
			sprintf(blocks,"%d",blockdef);
		strcat(cmd," ");
		strcat(cmd,blocks);
	}
#ifdef DEBUG
	printf("%s\n",cmd);
#endif
	system(cmd);

}

/*
 *	label - simulate "labelit(1)" if names given
 */

label(ldev,i,sp)
register char *ldev;
register struct sledit *sp;
register i;
{
	struct filsys fs;
	register fd;
	register struct filsys *fp = &fs;
	register char *c;
	char dev[120];

	strcpy(dev, ldev);

	/* compute "base name" */

	for( c = dev + (strlen(dev) -1 ); isdigit(*c ) ; *c-- ='\0'  );
	sprintf(dev, "%s%d",dev,i);
	if((fd = open(dev,O_RDWR)) == -1) {
		fprintf(stderr,"Can't open %s\n",dev);
		return(-1);
	}
	if(lseek(fd,BSIZE,0) == -1) {	
		fprintf(stderr,"Super block seek error %s\n",dev);
		close(fd);
		return(-1);
	}
	if( read(fd,fp,sizeof(struct filsys)) == -1) {
		fprintf(stderr,"Super block read error %s\n",dev);
		close(fd);
		return(-1);
	}

	strncpy(fp->s_fname,sp->fs.s_fname,6);
	strncpy(fp->s_fpack,sp->fs.s_fpack,6);

	if(lseek(fd,BSIZE,0) == -1) {	
		fprintf(stderr,"Super block seek error %s\n",dev);
		close(fd);
		return(-1);
	}
	if( write(fd,fp,sizeof(struct filsys)) == -1) {
		fprintf(stderr,"Super block write error %s\n",dev);
		close(fd);
		return(-1);
	}
	close(fd);
}
/*
 *	usage: print out usage string
 */

usage()
{
	fprintf( stderr, "Usage: %s [ -s file -m -p -v ] rawdev\n",me);
	exit(1);
}
