/*	Copyright (c) 1987 Motorola Inc
	All Rights Reserved
	@(#)port_hold.c	35.2
*/


#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/termio.h>
#include <sys/signal.h>

/*
			port_hold


   holds tty ports open to prevent strange occurrences

*/



extern int  errno;

#define  NPORTS  34 		/* Depends on system configuration */

#define  LINELN  128
#define  PATHLN  64


struct {

	unsigned flags;      /* what has been done to this port so far */
	int  ofd;            /* file descriptor returned by 'open' of port */
	char  stty[LINELN],  /* stty line for this port saved here */
	path[PATHLN];  /* path list saved here */
	int majr;	/* major number of device */
	int minr;	/* minor number of device */

} port[NPORTS],  /* indexed by virtual line number */
*p = port;


/*      values for "flags" field of above structure      */

#define  INUSE  1
#define  OPENED  2
#define  CLOCALED  4
#define  STTYED  8





main()
{
	char  tline[LINELN],          /* read lines into here for inspection */
	tpath[PATHLN],          /* temporarily store path list here */
	*pos,                   /* points to path list in stty line */
	*tp;                    /* gp */

	struct  stat  stbuf,          /* where we put info from 'stat' */
	*st = &stbuf;

	struct  termio  lineinfo,     /* ioctl work area */
	*li = &lineinfo;

	short  n;                     /* g.p. */

	int  shstat;                  /* status returned by 'system' */

	int majr, minr;
	int i;




	/* STEP 1

	become immune to all signals
	extract 'stty' command lines from stdin
	extract '/dev/ttyxx' path list from each stty line

	*/


	for(n=1; n<18; n++)
		setsig(n);          /* trap or ignore all signals */


	while(1) {

		fflush(stdout);

		errno = 0;

		if( (tp = fgets(tline, LINELN, stdin))  !=  tline ) {  
			/* error or EOF */
			if( errno )   /* error, not just eof */
				printf("\nerror %d reading stdin\n", errno);
			break;    /* terminate loop in either case */
		}


		/*  there are three kinds of input lines that are ignored:

		1. comments, which have a '#' as the first char
		2. lines which are not 'stty' commands
		3. lines which are 'stty's, but don't have redirection
		   of stdin                                             */

		if( (tline[0] & 0x7F) ==  '#' )  continue;

		if( strncmp(tline, "stty", 4) )  continue;

		if( (pos = strchr(tline, '<'))  ==  NULL )  continue;


		/*  at this point we think we have a valid stty command in tline
	let's see if it takes it's stdin from a valid serial device     */

		++pos;      /* advance past '<' */

		if( strlen(pos)  >  PATHLN )
		{
			printf("bad length: %s\n", tline);
			continue;   
		}
			/* unreasonably long path - probably garbage */

		strcpy(tpath, pos);   /* copy path */
		if( (tp = strchr(tpath, '\n'))  ==  NULL )
		{
			printf("bad path: %s\n", tpath);
			continue;    
		}
			/* real strange if don't have '\n', so skip it */
		else
			*tp = '\0';    /* terminate path string correctly */

		if( stat(tpath, st) ) {
			printf("\nerror %d on 'stat(%s)'\n", errno, tpath);
			continue;   /* just disregard this line */
		}

		/* 

		Should really use isatty here.   Check commented out entirely by
		saj 870605 due to sloth. 
		*/

		/* if( (major(st->st_rdev) != 30) ||  (minor(st->st_rdev) >= 32) ) 
      continue;  *//* disregard: redir was not from a serial device! */

		/* all is well, so save the stty line and the path in the port
		array entry corresponding to the minor device              
		*/
		minr = minor(st->st_rdev);
		majr = major(st->st_rdev);

		for(p = port, i = 0; i < NPORTS; i++, p++)
		{
			/* check to see if this is already used */
			if((p->majr == majr) && (p->minr == minr))
				break;
			/* find first unused entry */
			if(!(p->flags & INUSE))
				break;
		}
		if(i >= NPORTS)
		{
			fprintf(stderr, 
			    "port_hold: Exceeded MAXIMUM number of entries: %s\n", tline);
			continue;  /* ignore this line */
		}
		else
		{
			if( p->flags & INUSE ) {
				/* already have a stty for this port! */
				printf("\nmultiple sttys for %s - extra(s) ignored\n", tpath);
				continue;  /* ignore this line */
			}
		}
		p->majr = majr;
		p->minr = minr;
		p->flags |= INUSE;
		
		strcpy(p->stty, tline);     /* save whole stty line */
		strcpy(p->path, tpath);     /* ...and just the path */
	}





	/* STEP 2

	for each port mentioned in a 'stty' line:

	    open the port without waiting for carrier

	    do an ioctl to set 'clocal' mode on the port

	    fork the stty

	  */

	for(n=0; n<NPORTS; n++) {

		fflush(stdout);

		errno = 0;

		p = &port[n];

		if( p->flags & INUSE )  {   /* skip if no stty seen for this port */

			if( (p->ofd = open(p->path, (O_NDELAY|O_RDWR)))  ==  -1 )  {
				printf("\nerror %d in 'open(%s)'\n", errno, p->path);
				continue;   /* abandon further work on this port */
			}

			p->flags |= OPENED;

			printf("\nsuccessful open of %s", p->path);

	/* when we get here, we know we have the port opened, but since we
	 opened in NDELAY mode, we don't know whether or not we have
	 carrier on the port. surprisingly, there is no way to find out.
	 we don't want the 'stty' we will soon fork to hang on its own
	 open of the port, so we now set the port to 'clocal' mode. this
	 will allow the 'stty' to open the port regardless of carrier
	 state. of course, the 'stty' is perfectly free to turn 'clocal off */

			/* get "old" modes */
			if( ioctl(p->ofd, TCGETA, li)  ==  -1 )  {   
				printf("\nerror %d on ioctl(TCGETA) for %s\n", 
					errno, p->path);
				/* abandon further work on this port */
				continue;   
			}

			li->c_cflag |= CLOCAL;

			if( ioctl(p->ofd, TCSETA, li)  ==  -1 )  {   /* force "clocal" */
				printf("\nerror %d on ioctl(TCSETA) for %s\n", 
					errno, p->path);
				/* abandon further work on this port */
				continue;   
			}

			p->flags |= CLOCALED;
			printf("\nclocal set on %s", p->path);


			/*  now we can finally fork the stty  */

			if( shstat = system(p->stty) )  {   
				/* nonzero status returned */
				if( shstat == -1 )    /* fork or exec failed */
					printf("\nerror %d from fork or exec of ", errno);
				else /* it was the child that returned bad status */
					printf("\nerror %d returned by ", shstat);
				printf("'%s'\n", p->stty);
				continue;   /* abandon work on this port */
			}

			p->flags |= STTYED;
			printf("\nsuccessful stty on %s", p->path);

		}
	}


	/* STEP 3:

	all opens and sttys are done

	now we want to pause forever

  */

	printf("\n\nentering pause state\n\n");
	fflush(stdout);

	while(1)
		pause();

}





/*  signal handler - just log signal number  */

sighand(signo)
int  signo;
{
	printf("\ncaught signal %d", signo);
	fflush(stdout);

	setsig(signo);  /* reset signal trap */
}




/*  install signal handler  */

setsig(signo)
int  signo;     /* signal to set trap for */
{
	int  sighand();

	if( signo == SIGCLD )
		signal(signo, SIG_IGN);     /* death of child gets ignored */
	else
		signal(signo, sighand);     /* all others get logged */
}
