#ident  "@(#)sglusr.c	35.3"

/*
 *  Program to allow a "stopover" in single user mode
 *  during system boot.
 *  Copyright, 1988,  Motorola, Inc.
 */


#include <sys/param.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>

#define PAUSE	  5
#define PASSWORDS "/etc/passwd"
#define SHELL	  "/bin/sh"
#define SANE	  "stty sane"
#define SUSER_SHL "su bin -c su"
#define PROMPT    "Enter <CR> to stop in single user mode during system boot."
#define SHL_MSG   "\nSpawning shell [ ^D to continue booting ]...  "
#define SHL_ERR	  "\nERROR: Unable to spawn single user shell."
#define END_MSG   "\nDefault boot sequence continuing...\n"
#define WARNING	  "\nWARNING: Password file not accessible.\n"
#define INITTAB	  "/etc/inittab"
#define LINESIZ	  532

main(argc, argv)
  int argc;
  char **argv;
{

	register int i, j, fd, nchrs;
	char resp[CANBSIZ];
	char tabline[LINESIZ];
	char runlvl = '0';
	FILE *inittab;


	/*
	 * Determine if the system will boot to single user mode;
	 * if so, exit.  Single user shell is unnecessary in this case.
	 */
	if ((inittab = fopen(INITTAB, "r")) != NULL) {

		while (fgets(tabline, LINESIZ, inittab) != NULL) {
			if (strncmp(tabline, "is:", 3) == 0) {
				runlvl = tabline[3];
				break;
			}
		}
		if (runlvl == 's' || runlvl == 'S' || runlvl == '1')
			exit(0);

		fclose(inittab);
	}

	/*  
	 *  After setting "sane" ioctl values, open the control
	 *  terminal with no delay.  Issue single user mode prompt
	 *  and wait PAUSE seconds for a response before exit, 
	 *  allowing the default boot sequence to continue.
	 */

    system(SANE);
    if ((fd = open("/dev/tty",O_RDWR|O_NDELAY)) == -1)  {
	printf("%s: cannot open /dev/tty; errno = %d\n", argv[0], errno);
	exit(-1);
    }

    write(fd, PROMPT, strlen(PROMPT));

    for (i=0; i < PAUSE; i++)  {

	if ((nchrs = read(fd, resp, sizeof(resp))) > 0)  {
	    write(fd, SHL_MSG, strlen(SHL_MSG));
	    /*
	     *  Spawn the single user shell here. 
	     */
	    if (access(PASSWORDS, 4) == 0)  {
		if (system(SUSER_SHL) == -1)
		    write(fd, SHL_ERR, strlen(SHL_ERR));
	    }
	    else {
		write(fd, WARNING, strlen(WARNING));
		if (system(SHELL) == -1)
		    write(fd, SHL_ERR, strlen(SHL_ERR));
	    }
	    break;
	} else sleep (1);

    } /* endfor */

    write(fd, END_MSG, strlen(END_MSG));
    close(fd);
}
