#ident "@(#)symbols.c	35.1     88/08/26     "
#ident "Copyright (c) 1985 AT&T, INC and  MOTOROLA, INC."
#include <stdio.h>
#include <string.h>

#include "symbols.h"
#include "symbols2.h"
#include "dbug.h"

/*
 *	"symbols.c" is a file containing functions for accessing the
 *	symbol table.  The following functions are povided:
 *
 *      newent(string)
 *		Creates a new symbol table entry for the symbol with name
 *		"string".  The type of the symbol is set to undefined.
 *		All other fields of the entry are set to zero.
 *
 *      lookup(string,install_flag,user_name_flag)
 *		Looks up the symbol whose name is "string" in the symbol
 *		table and returns a pointer to the entry for that symbol.
 *		Install_flag is INSTALL if symbol is to be installed,
 *		N_INSTALL if it is only to be looked up.  User_name_flag
 *		is USRNAME if the symbol is user-defined, MNEMON if it is
 *		an instruction mnemonic.
 *
 *
 *	traverse(func)
 *		Goes through the symbol table and calls the function "func"
 *		for each entry.
 *
 *	dmpstb()
 *		Dumps the symbol table out to the intermediate file
 *		"fdstab" after pass one. The file descriptor "fdstab"
 *		should be open for writing on the intermediate file that
 *		is to contain the symbol table before calling this
 *		function.  This procedure is only used in the multiple-
 *		process version of the assembler.
 *
 *	creasyms()
 *		Enters the instruction mnemonics found in instab[] into
 *		the symbol table.
 *
 *	addstr(string)
 *		Enters the "string" into the string table.  Called by
 *		newent().  Space for the string table is initially
 *		malloc()-ed in strtabinit().  If "string" would exceed
 *              the available space, then the string table is realloc()-ed
 *		with a larger size.  This procedure is only used in the
 *		flexnames version of the assembler.
 *
 *	strtabinit()
 *		Sets up the string table, with space malloc()-ed.  This
 *		procedure is only used in the flexnames version of the
 *		assembler.
 */

#if ONEPROC
extern short passnbr;
#endif

symbol
	symtab[NSYMS];
upsymins
	hashtab[ NHASH + 1 ];

#if FLEXNAMES
char    *bufend;

char    *strtab;
char    *strloc;

extern char * realloc();

#endif

unsigned long symcnt = 0;
extern  char    *alloc(), *malloc();
symbol  nullsym, *econoptr = &nullsym, *symqptr = NULL;
symbol *
newent(strptr)
	register char *strptr;
{
	register symbol *symptr;
	extern long addstr();

	DBUG_ENTER ("newent");
	if (symcnt >= NSYMS) {
		symptr = (symbol *) alloc( sizeof( symbol ) );
	}
	else
		symptr = &symtab[ symcnt ];
#if FLEXNAMES
	if( strlen( strptr ) > SYMNMLEN )
	{
		if( !strcmp( strptr, symstr( econoptr ) ) )
		{
			symptr->_name.offset[1] = econoptr->_name.offset[1];
		}
		else
		{
			symptr->_name.offset[1] = addstr( strptr );
		}
		symptr->_name.offset[0] = 0;
		symptr->_name.offset[2] = 0;
	}
	else
	{
#endif
		symptr->_name.offset[0] = 0;
		symptr->_name.offset[1] = 0;
		symptr->_name.offset[2] = 0;
		(void) strcpy( symptr->_name.name, strptr );
#if FLEXNAMES
	}
#endif
	symptr->mneflag = 0;    /* if set, symbol is a mnemonic */
	symptr->tag = 0;
	symptr->styp = UNDEF;
	symptr->sectnum = 0;
	symptr->maxval = 0;
	symptr->value = 0L;
#ifdef M68881
	symptr->value2 = 0;
	symptr->value3 = 0;
#endif /* M68881 */
	symptr->qdex = symcnt++;
	if( symqptr )   /* helps prevent loops */
		symqptr->qnext = symptr;
	symqptr = symptr;
	DBUG_RETURN( symptr );
}

#if PASS1 || ONEPROC
#if DEBUG
unsigned numcalls,
	numids,
	numcoll;
#endif
#endif
instr *insptr;  /* used by creasyms() to point into instab[] */

upsymins
lookup(sptr,install,usrname)
	char *sptr;
	BYTE install;
	BYTE usrname;
{
	register upsymins
		hp;
	register char
		*s, *mark;

	register int
		ihash = 0, jhash = 0;
	register int
		cnt = 9999;
	register upsymins
		ohp;

	DBUG_ENTER ("lookup");
	DBUG_5 ("lk1", "sptr='%s', install=%d, usrname=%d", sptr, install, usrname);
	mark = s = sptr;
#if PASS1 || ONEPROC
#if DEBUG
	if( ! sptr )
	       aerror("Lookup symbol is a null pointer");
	if( ! *sptr )
	       aerror("Lookup symbol is a null string");
	numcalls++;
#endif
#endif
	/***********************************************\
	 *                                             *
	 *  Strictly an ad hock hashing algorithm      *
	 *  based on the empirical results of hashing  *
	 *  6107 labels from the most difficult *.s    *
	 *  files yet encountered: mostly compiler     *
	 *  generated labels (4842) with the others    *
	 *  being worst case (long, similar names):    *
	 *  longest chain: 5                           *
	 *  average non-empty chain: 1.40262           *
	 *  average chain: 0.745574                    *
	 *  table usage: 4354 of 8192 possible         *
	 *                                             *
	\***********************************************/

	while (*s)
	{
		register int i;

		for( i = 12; i && *s; --i )
		{
			jhash = (jhash<<3) + *s++;
		}
		mark = s;
		ihash += jhash;
	}
	ihash += ( sptr - s + *--s ) << 6;
	ihash = ( ihash + ( ihash >> 19 ) ) & NHASH ;
	DBUG_3 ("lkx", "hash value = %d", ihash);
	ohp.stp = (symbol *) &hashtab[ ihash ].stp;
	hp.stp = ohp.stp->snext;
	do
	{
		if( hp.stp == NULL )
		{   /* no symbol ... */
			if( install == INSTALL )
			{
#if PASS1 || ONEPROC
#if DEBUG
				numids++;
#endif
#endif
			    /* make sure we link in new symbol--use ohp */
				if( usrname )
				{
					hp.stp =
					      ohp.stp->snext = newent( sptr );
				}
				else /* if mnemonic */
				{
					hp.itp = ohp.itp->snext = insptr;
				}
			}
			    /* else return a null upsymins */
			break;/********FOUND*******/
		} /* end of hp == NULL */
		else if( strcmp( sptr, symstr( hp.stp ) ) == 0 )
		{   /* if we have found the same symbol in the table */
			DBUG_3 ("lk4", "found match for '%s'", sptr);
			econoptr = hp.stp;
			if( install == INSTALL )
			{
				if( usrname == USRNAME && !hp.stp->mneflag )
					break;/********FOUND*******/
				else if( usrname == MNEMON  && hp.stp->mneflag )
					aerror("Duplicate instruction mnemonic");
			}   /* end of if ( install ) */
			else /* look it up, don't install */
			{
				if( ( usrname && !hp.stp->mneflag ) ||
					     ( !usrname && hp.itp->mneflag ) )
					break;/********FOUND*******/
			}   /* end of if (usrname ...) */
		}   /* end of if we have found the symbol in the table */
		    /* look at next record in the chain */
		ohp = hp;
		hp.stp = hp.stp->snext;
	} while( cnt-- );
	if( !cnt ) aerror("A single hash table entry has > 9999 links");
	DBUG_RETURN( hp );
}

#if !PASS1
traverse(func)
int (*func)();
{
	register short index;
	register symbol *sptr;

	DBUG_ENTER ("traverse");
	sptr = symtab;
	do
	{
		(void) ( *func )( sptr );
		sptr = sptr->qnext;
	}
	while( sptr );

	DBUG_VOID_RETURN;
}
#endif
#if PASS1 && !ONEPROC

FILE *fdstab;
void
dmpstb()
{
	register short index;
	register symbol *sptr;

	DBUG_ENTER ("dmpstb");
	sptr = symtab;
	do
	{
		fwrite( sptr, 1, sizeof( symbol ), fdstab);
		sptr = sptr->qnext;
	}
	while( sptr->qnext );

	DBUG_VOID_RETURN;
}

#endif
#if PASS1 || ONEPROC

extern instr instab[];

creasyms()
{
	DBUG_ENTER ("creasyms");
	for( insptr = instab; insptr->name[0] != '\0'; ++insptr )
	{
		(void) lookup( insptr->name, INSTALL, MNEMON );
	}
	DBUG_VOID_RETURN;
}
#endif

static strsize = 0;

long
addstr( strptr )
	register char	*strptr;
{
	unsigned long offset;
	register int	length;
	register char  *sbptr;

	DBUG_ENTER ("addstr");
	/* round to pointer size */
	length = strlen( strptr );
	if( strloc + length > bufend )
	{
		strsize += 16 * BUFSIZ;
		offset = strloc - strtab;
		if( ( strtab = realloc( strtab, strsize ) ) == NULL )
			aerror("String table out of memory");
		strloc = strtab + offset;
		bufend = strtab + strsize - 1;
	}
	sbptr = strcpy( strloc, strptr );
	strloc += length + 1;
	DBUG_RETURN( sbptr - strtab );
}	/* addstr(strptr) */

char *
strtabinit()
{
	DBUG_ENTER ("strtabinit");

	strsize = 16 * BUFSIZ;
	if( ( strtab = strloc = malloc( strsize ) ) == NULL )
		aerror("String table out of memory");
	strloc += sizeof(long);
	bufend = strtab + strsize - 1;
	DBUG_VOID_RETURN;
}	/* strtabinit() */

