#ident "@(#)addvalue.c	35.1       88/08/26     "
#ident "Copyright (c) 1985 AT&T, INC. and MOTOROLA, INC."

#include "declar.h"

#define INVALID 0		/* operation can not be used in lcse */

/* The structure of the hash table which supports local common subexpression
   elimination consists of the following structures. 

   The hash table is an array of pointers to buckets of ENTRYs which have 
   the same hash value. The hash value is a function of a node's operator
   and the value number of its left and right siblings.

   Each ENTRY points to an LCSE, which contains specific information about
   the possible lcse. Part of the information of an LCSE is a pointer to
   a link list of LCSELINKs. Each LCSELINK points to an occurrence of
   the lcse in the code tree.

   If two or more entries occur in the LCSELINK list, then a pointer to the 
   LCSE is added to the alteration list. This list is processed at the end
   of a basic block.
*/

typedef struct Tlcselink	/* element in list of lcse */
      {
      NODE *lcsenode;		/* pointer to expression in code tree */
      struct Tlcselink *next;	/* pointer to next element in list */
      } LCSELINK;

typedef struct Tlcse		/* local common subexpression info. node */
      {
      int count;		/* the number of expressions in LCSELINK list */
      LCSELINK *lcselink;	/* link list of expressions */
      int var;			/* index into symbol table for the variable 
				** which holds this value number */
      } LCSE;

typedef struct Talt		/* element in LCSE alteration list */
      {
      LCSE *expr;		/* pointer to local common subexpression
				** involved */
      struct Talt *next;	/* next element in list */
      } ALTITEM;

typedef struct Tlcsekey		/* local common subexpression hash table key */
      {
      int op;			/* operator involved in expression */
      TWORD type;		/* type of operation */
      ENTRY *left;		/* value number of left sibling */
      ENTRY *right;		/* value number of right sibling */
      } KEY;

static ALTITEM *altlist;	/* points to the head of the alteration list */
static ENTRY *enterht();
static void makealterations();
static void addalteration();
static void replace();
static int binaryop();
static int unaryop();
static int leafvalue();

/*F************************** FUNCTION HEADER ****************************

FUNCTION NAME:

	addvalueno

ACTION:

	This function controls the assignment of a value number to node "p".
	The type of value number assigned to a node depends on the type of
	the node. BITYPE and UTYPE nodes may obtain their value number
	from the lcse hash table, if "p" is not an INVALID operator.  LTYPE 
	nodes may be unique or get their value number from the symbol table, or 
	the constant hash table. If "allowable" is FALSE then the expression
	can not be entered into the expression hash table.
	Finally, if a possible assignment to an aliased variable is found,
	all former variable value numbers are invalidated.

PARAMETERS:

	Receives:
		
		p: pointer to node which will receive value number
	allowable: boolean flag which indicates if expression rooted by
		   "p" should be added to lcse hash table

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	NODE: see description associated with declaration

COMPLEX LOCALS:
	
	none

FUNCTIONS CALLED:

	optype: classifies the nodes operator as BITYPE, UTYPE, or LTYPE
	binaryop: can binary operator be a part of lcse
	flushstvn: clear all symbol table variable value numbers when
		   an aliased variable is modified
	getvalueno: returns unique value number
	enterht: enter expression into lcse hash table
	unaryop: can unary operator be a part of lcse
	leafvalue: gets value number for a leaf node

CALLED BY:

	recognizebb

METHOD:

	self explanatory

NOTES:

	none

*FE*************************************************************************/

void
addvalueno( p, allowable )

NODE *p;
int allowable;
{
   int op;
   KEY *key;
   NODE *r, *l;
   
   switch( optype(p->opn.op) )
      {
   case BITYPE:
      op = binaryop( p );	/* make the operator unique */
      l = p->in.left;		/* left sibling */
      r = p->in.right;		/* right sibling */
      if (op == INVALID)	/* operator can not be involved in a lcse */
	 p->opn.valueno = (ENTRY *)getvalueno();
      else if (op == ASGFLG)	
	 {	
		/* operator is an assignment node, use the value number of
		** the right side.  If the left side is a simple variable
		** then assign the value number to the simple variable. */

		/* Since an assingment statement has a side effect, it cannot
		** be part of a lcse which is replaced.  The simplest way to
		** do this is to give the assignent node a new value number.
		** this will mess up somewhat in that expressions with 
		** embedded assigments will not be fully optimized.
		*/

	 p->opn.valueno = (ENTRY *)getvalueno();

	 if ( (l->opn.op == STAR)	/* modification of an aliased variable*/
	     || (l->opn.op == FLD ) )	/* modification of a field cnages var */
	    {
	    flushstvn();	/* Invalidate all variable value nos.  */
	    }
	 else if (p->opn.op == ASSIGN)
	    {
	    if (l->opn.flags & SIMPLE)
	       { /* assignment to a simple variable - update symbol table */
	       (symtab + l->opn.sourceno)->valueno = r->opn.valueno;
	       if( lcsedbflg )
	       {
			printf("Assignment to simple variable found\n");
			printf("\tUpdate variable #%d to have valueno %x\n",
				l->opn.sourceno, r->opn.valueno );
	       };
   
               /* save the simple variable's symbol table index 
	       ** in the LCSE struct iff the value number is a pointer */
   
               if ( r->opn.flags & LCSEFLAG )
	          ((LCSE *) r->opn.valueno->data)->var = l->opn.sourceno;
	       }
	    }
	    else		/* i++ or i-- */
	    if( (p->opn.op == INCR) || (p->opn.op == DECR) )
	    {
				/* nodes value is value BEFORE operation */
				/* BUT, node can't be replaced in a CSE */
				/* since it has a side effect (i+=1) */
			/* p->opn.valueno = l->opn.valueno; */
		if (l->opn.flags & SIMPLE)
		   /* assignment to a simple variable - update symbol table */
		   /*		since varialbe is changed, give it a new # */
		   (symtab + l->opn.sourceno)->valueno = (ENTRY *)getvalueno();
		else	/* not a simple variable, but an assign was made */
		    flushstvn();	/* Invalidate all variable value nos. */
		   
	    }
	    else		/* op-equal assingment */
	    {
		if (l->opn.flags & SIMPLE)
		   /* assignment to a simple variable - update symbol table */
		   (symtab + l->opn.sourceno)->valueno = p->opn.valueno;
		else	/* not a simple variable, but an assign was made */
		    flushstvn();	/* Invalidate all variable value nos. */
	       
	    }
	 }
      else 
	 {
	 if (!allowable)
	    p->opn.valueno = (ENTRY *)getvalueno();   
	 else
	 /* make the hash key - the space will be freed later */
	    {
	    key = (KEY *)malloc(sizeof(KEY));
            key->op = op;
	    key->type = p->in.type;
	    key->left = l->opn.valueno;
	    key->right = r->opn.valueno;

	    p->opn.valueno = enterht(key, p);
	    }
	 }
      break;

   case UTYPE:

      op = unaryop(p);
      l = p->in.left;
      if (op == INVALID)	/* operator can not be involved in a lcse */
	 p->opn.valueno = (ENTRY *)getvalueno();
      else
	 {
	 if (!allowable)
	    p->opn.valueno = (ENTRY *)getvalueno();
	 else
	    {
	    /* make the hash key - space will be freed later */
	    key = (KEY *)malloc(sizeof(KEY));
            key->op = op;
	    key->type = p->in.type;
	    key->left = l->opn.valueno;
	    key->right = 0;

	    p->opn.valueno = enterht(key, p);
	    }
	 }
      break;
   
   case LTYPE:
      if( p->stmt.op == ASMNO )
		flushstvn();	/* Invalidate all variable value nos. */
      p->opn.valueno = (ENTRY *)leafvalue( p );
      break;
      }
}


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	binaryop

ACTION:

	Function classifies a binary operator according to its ability to
	be the operator in local common subexpression. All the assignment
	operators are classified as ASGFLG. Operators which can not 
	be operators in a lcse return INVALID. For all the other operators
	the function returns a unique operator which identifies the
	operator (in most cases this is just the operator itself).

PARAMETERS:

	Receives:

		p: pointer to the node being classified

	Returns:

		A unique value for "p"'s operator

EXTERNAL DATA REFERENCES:

	Operator identifiers found in manifest.h

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	addvalueno

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static int
binaryop( p )

NODE *p;
   {
   if (asgop(p->opn.op))	/* is it an assignment operator? */
      return (ASGFLG);

   switch (p->opn.op)
      {
   case CALL:
   case FORTCALL:
   case STCALL:
   case CMP:
   case CBRANCH:
      return(INVALID);
   default:
      return( p->opn.op );
      }
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	unaryop

ACTION:
	Function determines if the operator is permissable in a local
	common subexpression.  If it isn't the function returns an
	INVALID value, otherwise it returns a unique value for the
	operator which is usually just the value of the operator.

	To uniquely classify CONV nodes it is necessary to include their
	type in the computation.  Therefore, CONV nodes return an offset of 
	MANY, the maximum number of operators.  This computation may need 
	to change if other nodes are found which require additional information.

PARAMETERS:

	Receives:

		p: pointer to the node being classified

	Returns:

		A unique value for "p"'s operator

EXTERNAL DATA REFERENCES:

	Operator identifiers found in manifest.h

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	addvalueno

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static int
unaryop ( p )

NODE *p;
{
	NODE *q;
   switch (p->opn.op)
      {
   case STAR:
   case FUNARG:
   case STARG:
   case UNARY CALL:
   case UNARY STCALL:
   case UNARY FORTCALL:
   case GENBR:
   case GENUBR:
   case GENLAB:
   case UNARY AND:
   case FLD:
      return( INVALID );
   case CONV:
      /* MANY is the maximum  number for an operator in manifest. If
	 this method for achieving unique operators is used in other
	 instances than it might be necessary to actually create the
	 operators. */
      q = p->opn.parent;
      if( asgop(q->opn.op) ) return(INVALID);
      return( p->opn.op + p->opn.type + MANY);
   default:
      return( p->opn.op );
      }
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	leafvalue

ACTION:
	Function returns a value number for a leaf node.
	Value numbers for leaves are:
		1) symbol table # for simple locals/parameters and arrays
		2) pointers to LCSE structures in the constant hash table
		   if "p" is an ICON or FCON
		3) all other leaves return a unique number

PARAMETERS:

	Receives:

		p: pointer to the node being classified

	Returns:

		A unique value for leaf "p"

EXTERNAL DATA REFERENCES:

	Operator identifiers found in manifest.h

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	addvalueno

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/


static int
leafvalue( p )

NODE *p;
{
   ENTRY item;
   FLTDBL *key;
   ENTRY *q;

   if (p->opn.flags & SIMPLE) /* simple locals/parameters return symbol table value # */
      return((int)(symtab + p->opn.sourceno)->valueno);

   if (p->opn.flags & ARRAY) /* arrays return symbol table value # */
      return((int)(symtab + p->opn.sourceno)->valueno);

   if ((p->opn.op == ICON || p->opn.op == FCON)
       && p->tn.name == NULL ) /* constants return value # from constant hash table */
      {
      /* make key */
      key = (FLTDBL *) malloc(sizeof(FLTDBL));
      if ( p->opn.op == ICON )
	 {
         key->lg.l = p->tn.lval;
	 key->lg.l2 = 0;
	 }
      else
	 {
	 key->lg.l2 = ((FLTDBL *) &p->fp.dval)->lg.l2;
	 key->lg.l = ((FLTDBL *) &p->fp.dval)->lg.l;
	 }
      item.key = (char *) key;
      item.data = NULL;

      q = hsearch(&constht, consthash, constcmp, item, ENTER);

      if (q->data == NULL)	/* new constant */
	 q->data = (char *) getvalueno();
      else			/* old constant */
	 free(key);
      return((int)q->data);
      }

   return(getvalueno()); /* all other leaves return a unique number */
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	replace

ACTION:

	If expression "p" has already been evaluated and is stored in
	"var", replace "p" with a reference to "var".

PARAMETERS:

	Receives:

		p: pointer to the root of the expression
	      var: symbol table index for the variable which contains
		   the value of expression "p"

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	NODE: see description associated with declaration in manifest.h

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	enterht

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
replace (p, var)
	

NODE *p;
int var;
{
   if( lcsedbflg > 1 )
   {
	printf("-------------------In replace, replacing:\n");
	e22print(p, "T" );
	printf("-------------------with:\n");
   };
   tfree(p);
   /* using p preserves parent link */
   p->tn.op = VAUTO;
   p->tn.lval = (symtab + var)->offset;
   p->tn.rval = 0;
   p->tn.type = (symtab + var)->type;
   if( lcsedbflg > 1 )
   {
	e22print( p, "T" );
	printf("-------------------end replacement\n");
   };
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	addalteration

ACTION:

	This function adds a lcse to the list of alterations. New
	lcses are added to the beginning of list. This guarantees that
	parents will be replaced before their children.

PARAMETERS:

	Receives:

		p: pointer to LCSE

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	altlist: alteration list

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	malloc

CALLED BY:

	enterht

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
addalteration( p )

LCSE *p;
{
   ALTITEM *q;

   q = (ALTITEM *) malloc(sizeof(ALTITEM));
   if (q == NULL)
      cerror("out of space - addalteration");
   q->expr = p;
   q->next = altlist;
   altlist = q;
   lcsestat++;	/* statistics counter */
   if (lcsedbflg)
      {
      printf("\n-------------------------------\nLocal Common Subexpression added to alteration list\n");
      e22print(p->lcselink->lcsenode, "T");
      printf("\n-------------------------------\n");
      }
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	enterht

ACTION:

	This function enters an expression into the LCSE hash table. If
	the expression is a new entry into the table then the LCSE
	data structure is created and initialized. If the expression
	already exists in the table, then add another reference to it, and
	add to the alteration list for later replacement.

	Certain variables may already contain the values for expressions.
	If this circumstance is encountered, replace the expression with
	the variable.

PARAMETERS:

	Receives:

		key: the key which uniquely identifies the expression
		  p: pointer to the expression

	Returns:

		pointer to the entry in the hash table.

EXTERNAL DATA REFERENCES:

	lcseht
	NODE

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	hsearch: enters expresion into the hash table 
	replace: replace an expression with a variable
	classify: identify the variable 
	addalteration: add to alteration list for later replacement


CALLED BY:

	addvalueno

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static ENTRY *
enterht( key, p )

KEY *key;
NODE *p;
{
   LCSE *lcse;
   LCSELINK *link;
   ENTRY *q;
   ENTRY item;

   item.key = (char *) key;
   item.data = NULL;

   /* ENTER the expression in the lcseht */
   q = hsearch(&lcseht, lcsehash, lcsecmp, item, ENTER);
   if( lcsedbflg > 2 )
   {
       printf("Enterht called on the node:\n");
       e222print( 2, p, "T" );
       printf("Value number is %x\n", q );
   }

   if (q->data == NULL)	/* if this is a new entry initialize its information */
      {
      lcse = (LCSE *) malloc(sizeof(LCSE));
      if (lcse == NULL)
         cerror("out of space - addvalueno 1");
      q->data = (char *) lcse;
      lcse->count = 0;
      lcse->var = 0;
      lcse->lcselink = NULL;
      }
   else
      { /* expression already existed in hash table */
      free(key);
      lcse = (LCSE *)q->data;
      }

   if (q == (symtab + lcse->var)->valueno)  
      {
      /* if a variable already contains this expression */
      if( lcsedbflg > 2 )
      {
	printf("This matches the variable: %x \n", lcse->var );
      }
      replace(p, lcse->var);
      classify(p);
      return(q);
      }
   else
      {
      /* add expression to LCSE list */
  
      link = (LCSELINK *) malloc(sizeof(LCSELINK));
      if (link == NULL)
  	  cerror("out of space - addvalueno 2");
      link->next = lcse->lcselink;	/* place at head of list */
      link->lcsenode = p;
      lcse->lcselink = link;
  
      if (++lcse->count == 2) /* add to alteration list for later replacement */
  	 addalteration(lcse);
      }
   p->opn.flags |= LCSEFLAG;
   return( q );
}


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	freelcse

ACTION:

	This function returns the space required for the local common
	subexpression structure. This function is called when the hash
	table supporting lcses is being cleaned up at the beginning of
	each basic block.

PARAMETERS:

	Receives:

		p: pointer to the LCSE in the hash table

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	LCSE hash table

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	free

CALLED BY:

	lcse

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

void
freelcse( p )

LCSE *p;
{
   LCSELINK *q;

   for (q = p->lcselink; q ; q = q->next)
      free(q);
   free( p );
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	delaltlist

ACTION:

	Returns the space required for the alteration list.

PARAMETERS:

	Receives:

		nothing

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	alteration list

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	lcse

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

void
delaltlist()

{
   ALTITEM *p;

   for (p = altlist; p ; p = p->next)
      free(p);
   altlist = NULL;
}


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	lcse

ACTION:

	This function handles the local common subexpressions recognized
	in a single basic block. The alterations are made and
	the space required for the supporting data structures is freed.
	Also, the value numbers for symbol table variables are set to new
	values (i.e., cleared).

PARAMETERS:

	Receives:

		bb: a pointer to a basic block

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	lcseht

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	makealterations
	delaltlist
	cleanupht
	flushstvn

CALLED BY:

	recognizebb

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

void
lcse(bb)

BBPTR bb;
{
   if (bb)
      {
      makealterations(bb);	/* process alteration list */   
      delaltlist();	/* return space required for alteration list */
      cleanupht( &lcseht, freelcse); /* return space for lcse hash table */
      flushstvn();	/* alter value numbers of all symbol table entries */
      }
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	makealterations

ACTION:

	Process the alteration list, if any, and replace the lcse with
	temporaries.  

PARAMETERS:

	Receives:

		bb: pointer to the current basic block

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	altlist: alteration list

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	gettemp: get a temporary for this basic block
	classify: classify the temporary node
	talloc: allocate a node
	tfree: free a node

CALLED BY:

	lcse

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
makealterations(bb)
BBPTR bb;
{
   ALTITEM *p;
   LCSE *lcse;
   LCSELINK *link;
   int offset;
   int type;
   NODE *node;
   NODE *l, *eq, *r;
   int count;
   int exprno;

   for (p = altlist; p ; p = p->next)
      {
      lcse = p->expr;


      /* check to see if the lcse still exists - was it consumed by another
      ** one? */

      count = 0;
      for( link = lcse->lcselink; link; link = link->next)
	 if (link->lcsenode->opn.op != FREE)
	    {
	    type = link->lcsenode->opn.type;
	    node = link->lcsenode;
	    ++count;
	    }

      /* if it exists in sufficient quantity then replace it with a temporary */

      if (count >= 2)
	 {
	 if (lcsedbflg)
	     {
	     printf("\n-------------------------------\nReplacing Local Common Subexpression \n");
	     e22print(node, "T");
	     printf("\n-------------------------------\n");
	     }

	 offset = gettemp(type);
	 if( lcsedbflg > 1 )
	 {
		printf("---------- in makealterations, create new temp,");
		printf("offset is %d\n", offset );
	 }

         /* replace all references to the lcse with the temporary and free
         ** up the space required by the old references - the last instance
	 ** in the list is handled differently as an assignment must be added
	 ** prior to the expression */

	 link = lcse->lcselink;
	 for ( ; count > 1; --count )
	    {
	    for( ; link->lcsenode->opn.op == FREE; link = link->next);
	    node = link->lcsenode;
	    if( lcsedbflg > 1 )
	    {
		printf("In makealterations, delete the tree:\n");
		e22print( node, "T" );
	    };
	    tfree(node);

            /* using node preserves parent link and type */
	    node->tn.op = VAUTO;
	    node->tn.lval = offset;
	    node->tn.rval = 0;
	    if( lcsedbflg > 1 )
	    {
		printf("...And relpace it with the variable:\n");
		e22print( node, "T" );
	    };

	    classify(node);
	    link = link->next;
	    }

         /* replace the first reference to the expression with the temporary and
	 ** create the assignment statement */

	 for( ; link->lcsenode->opn.op == FREE; link = link->next);
	 node = link->lcsenode; 
	 r = talloc();  /* right sibling for the assignment statement */
	 *r = *node;	/* preserve parent's pointer to this node, and 
			** preserve opn.parent and type */

	 /* reset parent pointers for siblings as they now point to node */
	 r->in.left->opn.parent = r;
	 if (optype(r->opn.op) == BITYPE)
	    r->in.right->opn.parent = r;

	 node->tn.op = VAUTO;
	 node->tn.lval = offset;
	 node->tn.rval = 0;
	 classify(node);

	 l = talloc();  /* create left sibling for the assignment statement */
	 *l = *node;

	 eq = talloc();	/* create assignment node */
	 eq->in.op = ASSIGN;
	 eq->in.left = l;
	 eq->in.right = r;
	 eq->in.type = r->tn.type;

	 r->opn.parent = eq;
	 l->opn.parent = eq;

	 /* add assignment statement to codetree - r and l are temps now */

	 /* walk up the tree to the EXPRNO node */
	 for (r = node->opn.parent; r->opn.op != EXPRNO; r = r->opn.parent);
	 exprno = r->stmt.exprno;	/* number of this expression */
				/* find first EXPRNO for this expression */
	 for ( ;
		     r->stmt.pred 			/* there is a pred */
	 	 && (r->stmt.pred->opn.op == EXPRNO)	/* its an EXPRNO node */
		 && (r->stmt.pred->stmt.exprno == exprno);/* its this exprno */
	     r=r->stmt.pred );
			 /* r is now first EXPRNO node for this expression */

	 /* duplicate EXPRNO node and link in the assignment node */
	 /*	new node becomes the first EXPRNO node for this expression */

	 l = talloc();
	 *l = *r;
	 eq->opn.parent = l;
	 l->stmt.left = eq;
	 l->stmt.succ = r;
	 l->stmt.pred = r->stmt.pred;
	 if (r->stmt.pred == NULL)
	    bb->codetree = l;
	 else
	    r->stmt.pred->stmt.succ = l;
	 r->stmt.pred = l;
	 if( lcsedbflg > 1)
	 {
		printf(" First occurnace is now the assignment statement*\n");
		e22print(l,"E");
		printf("------- list of EXPRNO nodes for expr #%d:\n", exprno );
		for(r=l ;
		    r && r->stmt.op== EXPRNO && r->stmt.exprno == exprno; 
		    r = r->stmt.succ )
		{
		    printf("-----\n");
		    e22print(r, "E");
		}
		printf("---------- End of EXPRNO list for expr #%d\n", exprno );
	 };
	 }
	 
      }
}


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	constcmp

ACTION:

	Compares two constants and returns -1, 0 , or 1 depending if
	key1 < key2, key1 == key2, or key1 > key2.
	The double variables are not used for comparison due to byte
	swapping, also the double comparison takes longer.
	The actual value of the number doesn't really matter, as this
	routine is not trying to order them in a binary tree. Just
	make it fast to determine if the number exists in the tree.

PARAMETERS:

	Receives:

		key1, key2: pointers to the two items being compared

	Returns:

		-1, 0, or 1 as explained above

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	leafvalue via hsearch

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

int
constcmp( key1, key2 )

FLTDBL *key1;
FLTDBL *key2;
{
   if (key1->lg.l == key2->lg.l)
      {
      if (key1->lg.l2 == key2->lg.l2)
         return(0);
      else if (key1->lg.l2 < key2->lg.l2)
         return(-1);
      else 
         return(1);
      }
   else if (key1->lg.l < key2->lg.l)
      return(-1);
   else 
      return(1);
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	consthash

ACTION:

	Hashing function for constant hash table.


PARAMETERS:

	Receives:

		ht: pointer to the hash table
		key: pointer to constant to be entered/found

	Returns:

		index into the hash table

EXTERNAL DATA REFERENCES:

	hash table

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	hsearch via leafvalue

METHOD:

	Add the two parts of the constant and find the modulo number of
	the bucket

NOTES:

	none

*FE********************************************************************/

unsigned int
consthash( ht, key )

HASHTAB *ht;
FLTDBL *key;
{
   int result;

   result = (key->lg.l < 0 ? -key->lg.l : key->lg.l) + 
	    (key->lg.l2 < 0 ? -key->lg.l2 : key->lg.l2);
   return( ((result<0)?-result:result) % ht->length);
}


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	lcsehash

ACTION:

	Hashing function for local common subexpression hash table.

PARAMETERS:

	Receives:

		ht: pointer to hash table

	Returns:

		index into the hash table

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	hsearch via enterht

METHOD:

	Add the operators for current entry and its left and right child,
	and divide by the size of the hash table.

NOTES:

	none

*FE********************************************************************/

unsigned int
lcsehash( ht, key )

HASHTAB *ht;
KEY *key;
{
   int result;

   result = (key->op + (int)key->left + (int)key->right);
   return( ((result<0)?-result:result) % ht->length);
}


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	lcsecmp

ACTION:

	Compares two value numbers and returns -1, 0 , or 1 (less than,
	equal, or greater than) if key1 < key2, key1 == key2, or key1 > key2.

PARAMETERS:

	Receives:

		key1, key2: keys to be compared

	Returns:

		-1, 0, or 1 as described above

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	enterht

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

int
lcsecmp( key1, key2 )

KEY *key1;
KEY *key2;
{
   if (key1->op == key2->op 
       && key1->left == key2->left 
       && key1->right == key2->right
       && key1->type == key2->type)
       return(0);
   else if (key1->op < key2->op)
       return(-1);
   else
       return(1);
}

