#ident "@(#)regasg.c	35.1       88/08/26     "
#define Copyright (c) /* 1984 AT&T, INC. and MOTOROLA, INC. */
/*	Copyright 1985 Motorola,1987 Inc.			*/

#include <storclass.h>
#include "declar.h"

static void alterblocks();
static int bbweight;    /* weight of current basic block */
static void alterbb();
static NODE *addparminit();
static void alterexpr();
static void travexpr();
static void travbb();
static void weighlocals();
static void assign();
static void altertrees();
static void initregavail();
static void regannounce();
static int prolog;	/* boolean flag indicating PROLOG node has been seen */
static void chkCONV();
static void addCONV();

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	asgregvars

ACTION:

        This function controls the process which determines which
        local variables may be assigned to registers.

PARAMETERS:

	Receives:

		none

	Returns:

		none

EXTERNAL DATA REFERENCES:

	symflag
	aregstat
	dregstat
	fpregstat

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	initregavail
	weighlocals
	assign
	scopedump
	altertrees

CALLED BY:

	optimize

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

void
asgregvars()
{
   initregavail(scoperoot);	/* traverse the scope tree, bottomup, and 
				** initialize the regavail masks for each 
				** scope */
   weighlocals();       /* traverse basic blocks determining the weight
                        ** of the local variables */
   assign();            /* using the weights of the local variables
                        ** assign registers to the heaviest locals */
   if (symflag)		/* show which variables were assigned to registers */
      scopedump();
   if (aregstat || dregstat || fpregstat) 
      altertrees();     /* if variables have been assigned to registers then 
			** walk the basic blocks and alter the code
                        ** trees to include the new register variables */
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	weighlocals

ACTION:

        This  function  begins  the process of determining the weight of simple 
        local variables for a function.  The code tree for each basic block  is 
        traversed  and  the  weight  of  the  basic block is added to the local 
        variable's weight on each occurrence of the variable.     

PARAMETERS:

	Receives:

		none

	Returns:

		none

EXTERNAL DATA REFERENCES:

	basic block flow graph
	bbweight

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	travbb

CALLED BY:

	asgregvars

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/
   
static void
weighlocals()
{
   BBPTR bb;
   for (bb = fgroot->next; bb; bb = bb->next)
      {
      bbweight = bb->weight; 
      travbb( bb->codetree );
      }
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	travbb

ACTION:

        Traverse code tree adding the weight of the current basic block
        to each simple local variable.

PARAMETERS:

	Receives:

		q: root of code tree

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	travexpr

CALLED BY:

	weighlocals

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
travbb ( q )
      
NODE *q;        /* head of code tree */
{              
   NODE *p;
   for ( p = q; p; p = p->stmt.succ)
      {
      switch( p->tn.op )
         {
         case SBLOCK:
         case EBLOCK: 
#ifdef ASM_COMMENT
	 case ASMNO:
#endif
            break;    
         case EXPRNO:
            travexpr( p );
            break;
         }
      }
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	travexpr

ACTION:

        Traverse expression adding weight of current basic block to each
        simple local variable.

PARAMETERS:

	Receives:

		p: pointer to current node in expression tree

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	travexpr

CALLED BY:

	travbb

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
travexpr( p )

NODE *p;
{
   switch (optype(p->tn.op))
      {
   case BITYPE:
      travexpr( p->in.left );
      travexpr( p->in.right );
      break;
   case UTYPE:
      travexpr( p->in.left );
      break;
   case LTYPE:
      if (p->opn.op == VAUTO || p->opn.op == VPARAM)
	 if ( ISSIMPLE( p ) && 
	     !((symtab + p->opn.sourceno)->flags & ADDRESSTAKEN))
            (symtab + p->opn.sourceno)->weight += bbweight;
      break;
      }
}
                           


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	assign

ACTION:

        Compute which variables (locals and parameters) should be
        assigned to registers. This is done in a two step process:

	1) create an array consisting of the weight of each variable 
	being considered for a register, and an index to the variable's 
	symbol table entry. This array is then sorted according to weight. 

	2) The array is then processed according to weight. For each
	variable, the scope in which it was declared is checked to see if
	a register is free. If a register is free, the register is assigned
	to the variable, and the usage of the register is propogated
	to the surrounding environments.

PARAMETERS:

	Receives:

		nothing

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	symflag

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	createsv
	getreg
	regannounce

CALLED BY:

	asgregvars

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
assign()

{
   int i;
   SORTED *p;
   SYMENT *q;
   SCOPE *r;
   int regmask;

   sortedvars = createsv(); /* creates and sorts the array */

   if (symflag)
      {
      printf("\n\nSORTED VARIABLES ACCORDING TO WEIGHT\n\n");
      for (i = 0; (sortedvars + i)->ndx ; ++i)
	 printf("weight:\t%d\tvariable:\t%s\n", (sortedvars+i)->weight,
		(symtab + (sortedvars+i)->ndx)->name);
      }
   
   /* walk through sortedvars array and assign registers */

   for (i = 0; (sortedvars + i)->ndx; ++i)
      {
      p = sortedvars + i;
      if (p->weight <= 1) break;  /* if used only once then unprofitable */
      q = symtab + p->ndx;
      regmask = getreg(q->scope, q->type, &q->reg);
      if (regmask)
	 {
	 for (r = q->scope->higher; r ; r = r->higher)
	    r->regavail |= regmask;
	 if (q->scope->lower)
	    regannounce(q->scope->lower, regmask);
	 }
      }
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	regannounce

ACTION:

	This function announces the usage of a register to lower scope levels.

PARAMETERS:

	Receives:

		p: current scope
		reg: register being added

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	regannounce

CALLED BY:

	assign
	regannounce

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
regannounce(p, reg)

SCOPE *p;
int reg;
{
   p->regavail |= reg;
   p->regexport |= reg;
   if (p->lower)
      regannounce(p->lower, reg);
   if (p->same)
      regannounce(p->same, reg);
} 



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	altertrees

ACTION:

        Walk the Ctrees and replace the simple local variables and simple
        parameters with their designated registers.

PARAMETERS:

	Receives:

		nothing

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	scoperoot
	prolog
	basic block flow graph

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	alterblocks
	alterbb

CALLED BY:

	asgregvars	

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
altertrees()

{
   BBPTR bb;

   alterblocks( scoperoot );	/* alter SBLOCK and EBLOCK nodes */   
   prolog = FALSE;
   for (bb = fgroot->next; bb; bb = bb->next)
      alterbb(bb);
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	alterblocks

ACTION:

	Alter the SBLOCK and EBLOCK nodes to reflect the new register 
	assignments.

PARAMETERS:

	Receives:

		p: pointer to a scope

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	alterblocks

CALLED BY:

	alterblocks
	altertrees

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void 
alterblocks( p )

SCOPE *p;
{
   p->sblock->tn.rval = p->regexport;
   if (p->higher && p->eblock)
      p->eblock->tn.rval = p->higher->regexport;
   if (p->lower)
      alterblocks(p->lower);
   if (p->same)
      alterblocks(p->same);
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	alterbb

ACTION:

	Traverse the code tree for this basic block, recognizing 
	the expression trees so they can be traversed.

	Identify the beginning of a function for parameter initialization.

PARAMETERS:

	Receives:

		bb: pointer to the current basic block

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	prolog

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	alterexpr
	addparminit

CALLED BY:

	altertrees	

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
alterbb( bb )

BBPTR bb;
{
   NODE *p;

   for (p = bb->codetree; p; p = p->stmt.succ)
      {
      switch( p->tn.op )
         {
         case SBLOCK:
         case EBLOCK:
#ifdef ASM_COMMENT
	 case ASMNO:
#endif
            break;
         case EXPRNO:
            alterexpr(p);
	    if (prolog)
	       {
	       p = addparminit(p); /* add the initialization of parameters */
	       prolog = FALSE;
	       }
            break;
         }
      }
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	addparminit

ACTION:

	If a parameter has been allocated to a register then add the
	expression which initializes the parameter. These expressions
	are added after the PROLOG node.

PARAMETERS:

	Receives:

		q: pointer to the PROLOG node

	Returns:

		returns pointer to the new tree

EXTERNAL DATA REFERENCES:

	symtab
	scoperoot

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	talloc

CALLED BY:

	alterbb

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static NODE*
addparminit( q )

NODE *q;	/* pointer to the PROLOG node */
{
   int i;
   NODE *p, *t, *eq, *l, *r;

   for (i = scoperoot->scopehead; i && (symtab + i)->sclass == C_ARG; 
	i = (symtab + i)->scopendx)
      {
      if ( ! (symtab+i)->reg ) continue;	/* continue if not assigned */

      /* parameter has been assigned to a register; create an assignment
	 node and add it to code tree just after the prolog node "q" */

      /* create EXPRNO */

      t = talloc();
      t->stmt.op = EXPRNO;
      t->stmt.bb = NULL;

      /* link the node into the code tree */

      if (t->stmt.succ = q->stmt.succ) /* this is suppose to be an assignment */
	 t->stmt.succ->stmt.pred = t;
      t->stmt.pred = q;
      q->stmt.succ = t;

      t->stmt.rval = q->stmt.rval;  /* keep the same line number */
      q = t;

      /* create "l" the left side of the assignment, the register node */

      l = talloc();
      l->tn.op = REG;
      l->tn.lval = 0;
      l->tn.rval = (symtab+i)->reg;	/* save register in node */
      l->tn.type = (symtab+i)->type;

      /* create "r" the right side of the assignment, the parameter's node */

      r = talloc();
      r->tn.op = VPARAM;
      r->tn.lval = (symtab+i)->offset;
      r->tn.rval = 0;
      r->tn.type = (symtab+i)->type;

      /* create "=" node */

      eq = talloc();
      eq->in.op = ASSIGN;
      eq->in.left = l;
      eq->in.right = r;
      eq->in.type = (symtab+i)->type;
	 
      t->stmt.left = eq;
      }
   return(q);
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	alterexpr

ACTION:

	Traverse the expression tree and for every VAUTO and VPARAM encountered
        see if it should be replaced by a register.

PARAMETERS:

	Receives:

		p: current node

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	symtab
	prolog

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	alterexpr
	chkCONV

CALLED BY:

	alterbb

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
alterexpr(p)

NODE *p;
{
   SYMENT *q;

   switch( optype(p->tn.op) )
      {
   case BITYPE:
      alterexpr( p->in.left );
      alterexpr( p->in.right );
      break;
   case UTYPE:
      alterexpr( p->in.left );
      break;
   case LTYPE:
      q = (symtab + p->opn.sourceno);
      if (p->opn.flags & SIMPLE && q->reg)
         { /* If it is a variable which should be in a register
	   ** change the reference to the register */
         p->opn.op = REG;
         p->tn.lval = 0;
         p->tn.rval = q->reg;

	 chkCONV(p, q);	/* add conversion node if necessary */

         }
      if (p->tn.op == PROLOG)
	 prolog = TRUE;	/* Flag the beginning of the function, so any
			** registers which have been allocated variables
			** can be initialized */
      }
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	chkCONV

ACTION:

	    When placing a variable in a register, it may be necessary to
	    add a conversion node, and repaint the type of the variable.

	    The following table illustrates what needs to be done when the
	    item on the left is placed in a register and the parent node is
	    the type indicated in the table.

	    key to following action table

	    - change variable to a register and keep the type information in
	      the node

	    1 change the variable to a register and paint the node's type,
	      also add a conversion node

	    x the conversion is semantically illegal and should not occur

          i   pi   ul  pul    f   pf    d   pd    s   ps   us  pus
	  ________________________________________________________
any ptr   1    -    1    -    x    -    x    -    1    -    1    -
i,l,ul,u  -    1    -    1    -    1    -    1    1    1    1    1
    d, f  -    x    -    x    -    x    -    x    -    x    -    x
   s, us  -    -    -    -    -    -    -    -    -    -    -    -
   c, uc  -    -    -    -    -    -    -    -    -    -    -    -

	  c   pc   uc  puc  str pstr    l   pl  uni puni
	  ________________________________________________________
any ptr	  1    -    1    -    x    -    1    -    x    -
i,l,ul,u  1    1    1    1    x    1    -    1    x    1
    d, f  -    x    -    x    x    x    -    x    x    x
   s, us  1    -    1    -    x    -    -    -    x    -   
   c, uc  -    -    -    -    x    -    -    -    x    -   



PARAMETERS:

	Receives:

		p: current variable node

		q: pointer to variable's entry in symbol table

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	addCONV

CALLED BY:

	alterexpr	

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

 static void
 chkCONV(p, q)

 NODE *p;	/* current variable node */
 SYMENT *q;	/* pointer to variable's entry in symbol table */
 {
	 TWORD newtype;
	 TWORD oldtype;

	 newtype = p->tn.type;
	 oldtype = q->type;
	 if (oldtype != newtype) /* CONVersion node may be necessary */
	    {
	    if (ISPTR(oldtype))
	       {
	       if ( ! ISPTR(newtype))
		  {
		  p->tn.type = oldtype;   /* repaint node */
		  addCONV( p, newtype);
		  }
	       }
	    else if (oldtype == INT || oldtype == LONG
		  || oldtype == ULONG || oldtype == UNSIGNED)
	       {
	       if (ISPTR(newtype))
		  {
		  p->tn.type = oldtype;   /* repaint node */
		  addCONV( p, newtype);
		  }
	       else
		  {
		  switch (newtype)
		     {
		  case SHORT:
		  case USHORT:
		  case CHAR:
		  case UCHAR:
		     p->tn.type = oldtype;
		     addCONV(p, newtype);
		     break;
		  default:
		     break;
		     }
		  }
	       }
	    else if (oldtype == SHORT || oldtype == USHORT)
	       {
	       if (newtype == CHAR || newtype ==UCHAR)
		  {
		  p->tn.type = oldtype;
		  addCONV(p, newtype);
		  }
	       }
	 }
}

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	addCONV

ACTION:

	Add a convert node of type "newtype" to node p.

PARAMETERS:

	Receives:

		p: pointer to node receiving CONVersion

		newtype: type of CONVersion noed

	Returns:

		nothing

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	talloc

CALLED BY:

	chkCONV

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
addCONV(p, newtype)

NODE *p;
TWORD newtype;
{
	NODE *t;

	t = talloc(); 	/* create the CONV node */
	t->opn.op = CONV;
	t->tn.rval = 0;
	t->opn.type = newtype;
	t->opn.left = p;
	
	if (p->opn.parent->opn.left == p)	/* alter tree pointers */
	   p->opn.parent->opn.left = t;
	else
	   p->opn.parent->opn.right = t;
	t->opn.parent = p->opn.parent;
	p->opn.parent = t;
}


/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	initregavail

ACTION:

	Traverse the scope tree, bottomup, and initialize the regavail mask
	for each scope.

PARAMETERS:

	Receives:

		p: scope of concern

	Returns:

		nothing


EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	initregavail

CALLED BY:

	initregavail
	asgregvars

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

static void
initregavail(p)

SCOPE *p;
{
   SCOPE *q;

   if (p->lower)	/* go all the way to the bottom */
      initregavail(p->lower);
   else
      for (q = p->higher; q ; q = q->higher)	/* propagate info upwards */
         q->regavail |= p->regavail;
   if (p->same)		/* do the same for siblings */
      initregavail(p->same);
}
