#ident "@(#)storage.c	35.1       88/08/26     "
#ident "Copyright (c) 1984 AT&T, INC. and MOTOROLA, INC."

#include <storclass.h>
#include "declar.h"

/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

addtemp

ACTION:

        This function returns the byte offset for the requested temporary 
        storage location, and updates the block's maximum offset. 
        The alignment of the variable is taken into account.

PARAMETERS:

	Receives:

		level: pointer to the current scope
		type: type of variable to be allocated

	Returns:

		offset for the variable

EXTERNAL DATA REFERENCES:

	symtab
	tempstat
	curscope
	is_m68881

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	addsymbol

CALLED BY:

	gettemp

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

int
addtemp(level, type)
               
SCOPE *level;
int type;
{
   int size;
   int offset;
   int ndx;
   SYMENT *sym;

   size = typesize(type);
   offset = (-level->maxoffset - size)/ 8; /* convert bits to bytes */
   level->maxoffset += size < ALINT ? ALINT : size;

   /* Following line can be deleted, iff front end fixed to output matching
      [] for ALL scopes. The corresponding line in exitscope can then be used */
   level->sblock->opn.csiz = level->maxoffset;

   ndx = getsyment();

   /* Add to the symbol table entry and initialize it */
   sym = (symtab + ndx);
   sym->offset = offset;
   sym->type = type;
   sym->flags = 0;
   sym->sclass = C_AUTO;
   sym->weight = 0;
   sym->reg = 0;
   sym->valueno = (ENTRY *)getvalueno();
   sym->name = "TEMP";           
   sym->scope = level;
   sym->scopendx = 0;
   if (level->scopehead)  /* if symbol list exists */
      addsymbol( level->scopehead, ndx);  /* add to the current symbol list */
   else  /* this is the first symbol for this scope */
      level->scopehead = ndx;
   (symtab + ndx)->scope = curscope;
   ++tempstat;
   return(offset);
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	typesize

ACTION:

        This function returns the size (in bits) of the requested type.

PARAMETERS:

	Receives:

		type: type of concern

	Returns:

		size (in bits) of the requested type

EXTERNAL DATA REFERENCES:

	none

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	addtemp
	lookup

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

int
typesize( type )

int type;
{
   int size;
   if (type != BTYPE(type))  /* if its not simple substitute a long */
      type = LONG;
   switch (type)
      {
      case UCHAR:
      case CHAR: 
         size = SZCHAR;
         break;
      case USHORT:
      case SHORT: 
         size = SZSHORT;
         break;
      case UNSIGNED:
      case INT: 
         size = SZINT;
         break;
      case ENUMTY:
      case ULONG:
      case LONG:
         size = SZLONG;
         break;
      case FLOAT:
         size = SZFLOAT;
         break;
      case DOUBLE:
         size = SZDOUBLE;
         break;
      default:
         size = SZPOINT;
         break;
      }
   return(size);
}



/*F*********************** FUNCTION HEADER *******************************

FUNCTION NAME:

	getreg

ACTION:

        This function returns the next free register 
        for the specified type in the specified scope.

PARAMETERS:

	Receives:

		scope: scope of interes
		t: type of register required

	Returns:

		reg: register found

EXTERNAL DATA REFERENCES:

	scope tree
	aregstat
	fpregstat
	dregstat

COMPLEX LOCALS:

	none

FUNCTIONS CALLED:

	none

CALLED BY:

	assign

METHOD:

	self explanatory

NOTES:

	none

*FE********************************************************************/

int
getreg( scope, t, reg)

SCOPE *scope;
register TWORD t;
int *reg;	/* If function finds a register, the symbol table entry
		** for the variable is updated with this information */
{
        register int i;
        if( ISPTR( t ) )
        {
                /* address register */
                for( i=2; i<=5; ++i )
                {
                        if( scope->regavail & (0x100<<i) ) continue;
			/* register found - now update environment */
			aregstat++;	/* statistics counter */
                        *reg = i;	
			i = 0x100<<i;
                        scope->regexport |= i;
                        scope->regavail |= i;
                        return(i);
                }
                return(0);
        }
	else if ( ( t == DOUBLE || t == FLOAT ) &&
		  ( is_m68881 ) )
	{
		/* floating point register */
		for ( i=2; i<=7; ++i )
		{
                        if( scope->regavail & (0x10000<<i) ) continue;
			/* register found - now update environment */
			fpregstat++;	/* statistics counter */
                        *reg = i;	
			i = 0x10000<<i;
                        scope->regexport |= i;
                        scope->regavail |= i;
                        return(i);
		 }
		 return(0);
	 }
        else 
        {
                /* data register */
		int mask;
		int lastreg;

		if (t == DOUBLE) /* require two registers */
		   {
		   mask = 3;
		   lastreg = 6;
		   }
		else
		   {
		   mask = 1;
		   lastreg = 7;
		   }

                for( i=2; i<=lastreg; ++i )
                {
                        if( scope->regavail & (mask<<i) ) continue;
			/* register found - now update environment */
			dregstat++;	/* statistics counter */
			if(t == DOUBLE) dregstat++;
                        *reg = i;
			mask = mask<<i;
                        scope->regexport |= mask;
                        scope->regavail |= mask;
                        return( mask );
                }
                return(0);
        }
}
