#ident "@(#)scan.c	35.1     88/08/26     "
#ident "Copyright (c) 1984 AT&T, INC."
#ident "Copyright (c) 1985 MOTOROLA, INC."

/*      MC68000/10/20/30 optimizer - File scan.c         */

/* Routines to scan the input and find relevant pieces. */

/* All of these routines assume an input scan pointer, scanp.
** Scanp only moves over characters that have been successfully
** scanned.
*/

#include "optim.h" /* #include "defs" is implicit with optim.h */



extern	char * scanp;		/* scan pointer */

#define	swhite() while (*scanp == '\t' || *scanp == ' ') scanp++
/* sexpr -- scan across expression
**
** This routine scans across an expression, which can consist of
** terms (numbers or names) separated by + or -.  There may be a
** leading -.  Return true if we scan just a single number, false
** if we see a name or any non-leading + or -.
*/

static boolean
sexpr()
{
    boolean onenum = true;	/* true if scanned just one numeric value */
    boolean snum();
    boolean sname();

    if (*scanp == '-')		/* leading - */
	scanp++;		/* skip it */
    
    while (true)		/* loop until we don't find +/- separator */
    {
	if (! snum()) 		/* check for number */
	    if (sname())	/* see name? */
		onenum = false;	/* yes.  expression is not single number */
	    else
		serr("Bad expression");
				/* no.  Neither number nor name */
	
	if (*scanp != '-' && *scanp != '+')
				/* check for operator */
	    return(onenum);	/* none there.  We're done. */
	
	scanp++;		/* skip over operator */
    }
}
/* sline -- scan one input line
**
** This routine scans an input line.  It takes care of labels,
** opcode, and operands.
*/

void
sline()
{
    char * first;		/* remembered beginning pointer */
    void dolab();
    boolean doop();
    boolean sname();
    void dorand();
    RAND * srand();
    RAND * r;

    swhite();			/* skip initial white space */

    /* check for label */

    first = scanp;		/* remember where we are */
    if (sname() && *scanp == ':')
    {
	dolab(first, scanp-first); /* take care of label */
	scanp++;		/* skip : */
	swhite();		/* skip to next non-white space */
    }
    else
	scanp = first;		/* back up in case we saw name */

    /* check for null operation */

    if (*scanp == '\n' || *scanp == CC)
	return;			/* no op code.  Done. */
    
    /* find op mnemonic */

    first = scanp;		/* remember starting point */
    while (islower(*scanp) || *scanp == '.')
	scanp++;		/* scan across likely candidate chars */
    
    if (scanp == first)		/* did we move? */
	serr("Bad op name");
    /* Do operation-specific things */

    if (! doop(first,scanp-first)) /* true if rest of line was processed */
    {
	/* do normal operand processing */
	int randno = 1;		/* current operand number */
	
	do
	{
	    r = srand();	/* scan next operand */
	    dorand(r, randno);	/* do other operand processing */
	    randno++;
	} while (*scanp++ == ','); /* until we don't see a , */
	scanp--;		/* back up over non-, */
	swhite();               /* skip white space */
	/* Must find newline or comment start */

	if (*scanp != '\n' && *scanp != CC)
	    serr("Bad end of instruction");
    }
    return;
}
/* sname -- scan across a "name"
**
** This routine scans across a "name".  In this context, a name is
** a leading ~, or it consists of a leading letter or _,
** followed by letters, digits, _, ~, or %.
** Return a boolean success/failure value.
*/

static boolean
sname()
{
    register char c;		/* current character */
    boolean first = true;	/* whether we're working on the first char */

    if ( *scanp == '~' )	/* ~ as first character is single name */
    {
	scanp++;		/* skip ~ */
	return(true);		/* saw a name */
    }

    while ( (c = *scanp),
	       isalpha(c)	/* letter */
	    || c == '_'		/* underline */
	    || ((! first) && (isdigit(c) || c == '%' || c == '~'))
				/* non-first digit or % or ~ */
	    )
    {
	scanp++;		/* bump scan pointer */
	first = false;		/* no longer on first char */
    }

    return (! first);		/* saw something if not doing first char */
}
/* snum -- scan over number
**
** This routine scans over a number, which can be decimal, octal
** (with preceding 0), or hexadecimal (with preceding 0x or 0X).
** Allow a leading - sign for numbers.
** A success/failure code gets returned.
*/

static boolean
snum()
{
    boolean hex = false;	/* not doing hex value initially */
    register char * p = scanp;	/* use this for scanning */

    if (*p == '-')
	p++;			/* skip over leading sign */

    if (! isdigit(*p))		/* check for digit */
	return(false);		/* return with scanp unaffected */
    
    if (*p++ == '0')		/* check for 0x, 0X */
	if (*p == 'x' || *p == 'X')
	{
	    p++;		/* skip x/X */
	    hex = true;		/* have a hex number */
	}
    
    /* scan digits, being lazy about octal */

    while (isdigit(*p) || (hex && isxdigit(*p)))
	p++;			/* move to next char */
    
    scanp = p;			/* reset scan pointer */
    return(true);		/* say we saw a number */
}


/* sdisp -- scan a displacement in a new M68020 addressing mode 
**
** This routine scans from the beginning of a displacement value
** to the first character afterward.  It returns an estimate of
** the size of the displacement.
** 880823 added capability of estimating immediate values; (RT_IMMNUM)
**        also now used to evaluate absolute memory references is (RT_MEM)
*/

short int
sdisp()
{
    long d_value;
    long strtol();
    char *tmpp = scanp;
    boolean parfound = false;
    boolean singlenum;

    if (*scanp == '(')
	{
	parfound = true;
	scanp++;         /* Scan past "(" */
	}

     /* Expression may have a forced size code */
     /* such as (expr).[bwl] or <number>.[bwl] */

    singlenum = sexpr();  /* read the displacement */

    if (parfound)
	if (*scanp++ != ')' || *scanp != '.')
	    serr("Bad sized displacement in M68020 format");

    if (*scanp == '.')
    {
	scanp++;            /* Pass by "." */
	switch(*scanp++)
	{
	case 'b' : return(1);   /* number fits in 8 bits but may need 16 */
	case 'w' : return(2);
	case 'l' : return(4);
	case 's' : return(4);
	case 'd' : return(8);
	case 'x' : return(12);
	default  : serr("Bad size code for displacement");
	}
    }
    else /* Estimate the size of the expr. explicitly    */
    {
	if (!singlenum)    /* not a single numeric value */
	    return(4);     /* allocate 32 bits, then     */
	else
	{
	    d_value = strtol(tmpp,(char **)NULL,0);
	    if (d_value < -32768 || d_value > 32767)
		return(4); /* number too big for 16 bits */
	    else if (d_value < -128 || d_value > 127)
		return(2); /* number too big for 8 bits */
	    else if (d_value == 0)
		return(0); /* no extension bytes needed */
	    else
		return(1); /* number fits in 8 bits but may need 16 */
	}
    }
}
/* scan_020_fmt -- scan across a new M68020 addressing format
**
** This routine scans from the character past the opening '('
** to the final balanced closing ')'.  The operand is classified
** by type, and the RAND structure is built and returned.
**
** The most general form of operand allowed without double indirection is
**
** (base_disp_expr,Ai,Rj,outer_disp_expr)
**
** Note that any of these four terms may be absent, and that 
** the index register Rj may contain a size code and/or scaling factor.
**
** The forms allowing double indirection are similar to the ones above,
** except that a single [] pair encloses some of the above terms.
**
** RT_020 means that you need a 68020 to execute the addressing mode,
**        and also that a full format extension word is needed.
** RT_DBLIND means a 68010 or a 68020 brief format extension word
**           is needed; note that a scale factor present means that
**           this address mode is 68020 but we don't care because
**           we are not doing error checking, just counting bytes
*/
/* the PC is included as an address register because it behaves the same */
#define ADR_REGS ( REG_A0 | REG_A1 | REG_A2 | REG_A3 | REG_A4 | \
		   REG_A5 | REG_A6 | REG_A7 | REG_PC | REG_FP | REG_SP )

RTYPE
scan_020_fmt(regs_020,d1_size,d2_size) /* scan a 68000/68020 addr format */
REG         * regs_020;   /* regs used in this addressing mode */
short int   * d1_size,    /* first  disp. size in bytes (1,2,4) */
	    * d2_size;    /* second disp. size in bytes (1,2,4) */
{
    register short int cur_dsize;
    extern REG sreg();
    register int regcnt = 0;
    register RTYPE rtype = RT_IREG;                /* operand type */

    while ( true )
    {
	switch(*scanp)
	{
	case ')':
		scanp++; /* next character, please */
		goto afterloop;
	case '[':
	case ']':
		rtype = RT_020;
	case ',':
		scanp++; /* next character, please */
		break;
	case '%':
		*regs_020 |= sreg(); /* save the register value */
		regcnt++;
		if (*scanp == '.')   /* read optional size code */
		    scanp += 2;
		if (*scanp == '*')   /* read optional scale factor */
		    scanp += 2;      /* 68010 brief format can't have one   */
		break;               /* but we don't judge correctness here */
	default : /* It must be a displacement expression, then.  */
		cur_dsize = sdisp(); /* estimate the disp. size */
		if (*d1_size != 0 )
		{
		    rtype = RT_020;
		    /* Outer displacement is always 0, 2 or 4 bytes */
		    if ( cur_dsize == 1 )
			    cur_dsize = 2;
		    *d2_size = cur_dsize;
		}
		else
		{
		    *d1_size = cur_dsize;
		    if ( cur_dsize == 4 )
			rtype = RT_020;
		}
		break;
	}
    }
afterloop:
    /* If it isn't already obviously an 020 mode ... */
    if ( rtype != RT_020 )
    {
	/* If it's a memory indirect or an extended mode with 16-bit offset */
	if ( ( *regs_020 == REG_NONE || ( *d1_size == 2 && regcnt == 2 ) ) )
	{
	    rtype = RT_020;
	}
	/* otherwise if there's no outer displacement ... */
	else if ( *d2_size == 0 )
	{
	    /* If we qualify for a brief format extension word (010 or 020) */
	    if ( regcnt == 2 && *d1_size <= 1 )
	    {
		rtype = RT_DBLIND;
	    }
	    /* else (finally) if it's register indirect with displacement */
	    else if ( regcnt == 1 && *regs_020 & ADR_REGS &&
						   *d1_size && *d1_size <= 2 )
	    {
		rtype = RT_OREG;
	    }
	}
    }
    /* then if not any of above, it's an address register indirect: RT_IREG */
    /* we don't check syntax because the assembler will catch those errors */

    return rtype;
}

/* srand -- scan across complete operand
**
** This routine scans across a complete normal operand.
** The operand is classified by type, and the RAND structure
** is built and returned.  The basic operand types we look for
** are keyed on the first character or two:
**
**      %       Rn                              RT_REG
**      &       number                                  RT_IMMNUM
**      &       <expr>                                  RT_IMMED
**      (       %An)                                    RT_IREG
**      (       %An)+                                   RT_INCR
**      -       (%An)                                   RT_DECR
**      \-      <16-bit_expr>(%An)                      RT_OREG
**      \-      <8-bit_expr>(%An,%Xn.wl)                RT_DBLIND or RT_020
**      (       <16-bit_expr>,%An)                      RT_OREG
**      (       <8-bit-expr>,%An,%Xn.wl)                RT_DBLIND or RT_020
**      (       <16-bit_expr>,%An,%Xn.wl)               RT_020
**      (       <32-bit_expr>,%An,%Xn.wl)               RT_020
**      (       %Dn.wl)                                 RT_020
**      (       <expr>,%Dn.wl)                          RT_020
**      (       <expr>,%An,%Xn.wl,<expr>)               RT_020
**      (       [<expr>,%An],%Xn.wl,<expr>)             RT_020
**      (       [<expr>,%An,%Xn.wl],<expr>)             RT_020
**      (       <expr>,%Xn.wl,<expr>)                   RT_020
**      (       [<expr>,%Xn.wl],<expr>)                 RT_020
**      (       [<expr>],%Xn.wl,<expr>)                 RT_020
**      (       <expr>,<expr>)                          RT_020
**      (       [<expr>],<expr>)                        RT_020
**      (       <expr>)                                 RT_020
**      (       [<expr>])                               RT_020
**      \-     <expr>                                   RT_MEM
**      <expr>                                          RT_MEM
**      <expr> ( %An  )                                 RT_OREG
**      <expr> ( %An, %Xn.wl )                          RT_DBLIND
*/

static RAND *
srand()
{
    REG regs = REG_NONE;	/* registers used */
    RTYPE rtype, rtemp = 0;     /* operand type */
    char * first;		/* starting point for operand */
    RAND * operand;		/* pointer to result structure */

    boolean sexpr();
    REG sreg();

    REG ow_regs = REG_NONE;     /* regs used in the {o:w} field for BFxxx inst*/
    short int displ1  = 0,      /* size in bytes of '020 mode displs. */
	      displ2  = 0;
    char    * tempp;

    RTYPE scan_020_fmt();    /* scan a new 68020 addr format */

    swhite();			/* skip leading white space */
    first = scanp;		/* remember this point */

/* Begin processing based on first non-space character. */

    switch(*scanp)
    {
    case '(':			/* register indirect and auto-increment */
	scanp++;		/* skip ( */
	rtemp = rtype;

	/* interpret all indirect and new M68020 addressing formats */
	rtype = scan_020_fmt(&regs,&displ1,&displ2);

	if( rtemp == RT_DECR )
	{
		if ( rtype != RT_IREG )
			serr("Bad register indirect with pre-decrement");
		rtype = RT_DECR;
	}
	else if (*scanp == '+')  /* check for auto-increment */
	{
	     scanp++;               /* have it.  Skip + */
	     if ( rtype != RT_IREG )
		serr("Bad register indirect with post-increment");
	     rtype = RT_INCR;
	}
	break;
    case '%':			/* register */
	regs = sreg();		/* get register code */
	if (*scanp == ':')	/* read Ri:Rj construct for '020 mul,div */
	{
	    scanp++;
	    regs |= sreg();	/* save Rj register value */
	    rtype = RT_REG_PAIR;
	}
	else
	    rtype = RT_REG;	/* set operand type */
	break;

    case '&':			/* immediate */
	scanp++;		/* skip & */
	displ1 = sdisp();
	if( !displ1 ) ++displ1; /* Immediate must have a non-zero size */
	scanp = first + 1;
	rtype = RT_IMMNUM;	/* assume single number */
	if (! sexpr())		/* single numeric value? */
	    rtype = RT_IMMED;	/* no. */
	break;
    
    case '-':			/* predecrement and others */
	scanp++;		/* skip - */
	if (*scanp == '(')	/* start predecrement */
	{
	    scanp++;		/* skip ( */
	    regs = sreg();	/* get register information */
	    rtype = RT_DECR;	/* set operand type */
	    if (*scanp++ != ')') /* check for closing ) */
		serr("Bad predecrement");
	    break;		/* done */
	}
	scanp--;               /* point to '-' for displacement estimate */
    
	/* fall through to default case */
/* Handle other cases of operand type */

    default:
	displ1 = sdisp();
	scanp = first;
	(void) sexpr();		/* scan across expression */
	if (*scanp != '(')	/* register offset or double index? */
	{
	    rtype = RT_MEM;	/* no, straight memory reference */
	    break;
	}

	scanp++;		/* skip ( */
	rtype = scan_020_fmt( &regs, &displ1, &displ2 );
	break;

    } /* end switch */


#define scan_past(char)    \
while (*scanp != char)     \
    scanp++;               \
scanp++;

    /* check for an appended {o:w} used by M68020 BFxxx instrucs */
    if (*scanp == '{')
    {
	if (*(scanp++) == '%')    /* up to {o    */
	    ow_regs = sreg();     /* save register used */
	scan_past(':');           /* up to {o:   */
	if (*scanp == '%')
	    ow_regs |= sreg();    /* up to {o:w  */
	scan_past('}');           /* up to {o:w} */
    }

    operand = saverand(first, scanp-first, rtype, regs, true);
				/* build RAND structure */
    operand->disp1 = displ1;   /* 0 if not appropriate */
    operand->disp2 = displ2;   /* 0 if not appropriate */
    operand->owuses = ow_regs; /* REG_NONE if not appropriate */
    
    swhite();			/* skip trailing white space */
    return(operand);		/* return new operand */
}
/* Tables for sreg.  They give the correspondence between register
** names and REG value.  There are two parallel tables, one with the
** packed representation for the name, the other with the value.
** Since sreg() uses a linear search, the names are ordered by
** frequency of use, according to several large code samples.
*/

/* Table of packed names */

#define	PACK(a,b)	(((a) << 8 ) + (b))

static unsigned long rname[] =
{
	PACK('d','0'),		/* %d0 */
	PACK('f','p'),		/* %fp */
	PACK('s','p'),		/* %sp */
	PACK('a','0'),		/* %a0 */
	PACK('a','2'),		/* %a2 */
	PACK('a','3'),		/* %a3 */
	PACK('d','1'),		/* %d1 */
	PACK('d','2'),		/* %d2 */
	PACK('a','1'),		/* %a1 */
	PACK('a','4'),		/* %a4 */
	PACK('d','3'),		/* %d3 */
	PACK('a','5'),		/* %a5 */
	PACK('p','c'),		/* %pc */
	PACK('d','4'),		/* %d4 */
	PACK('d','6'),		/* %d6 */
	PACK('d','5'),		/* %d5 */
	PACK('d','7'),		/* %d7 */
	PACK('c','c'),		/* %cc */
	PACK('s','r'),		/* %sr */
	PACK('a','6'),		/* %a6 */
	PACK('a','7'),		/* %a7 */
	PACK(PACK('u','s'),'p'),/* %usp */
	PACK(PACK('f','p'),'0'),/* %fp0 */
	PACK(PACK('f','p'),'1'),/* %fp1 */
	PACK(PACK('f','p'),'2'),/* %fp2 */
	PACK(PACK('f','p'),'3'),/* %fp3 */
	PACK(PACK('f','p'),'4'),/* %fp4 */
	PACK(PACK('f','p'),'5'),/* %fp5 */
	PACK(PACK('f','p'),'6'),/* %fp6 */
	PACK(PACK('f','p'),'7'),/* %fp7 */
	0			/* end of table */
};


/* Table of encode register values.  These must correspond, one-to-one
** with the encoded names above.
*/

static REG rcode[] =
{
	REG_D0,		/* %d0 */
	REG_FP,		/* %fp */
	REG_SP,		/* %sp */
	REG_A0,		/* %a0 */
	REG_A2,		/* %a2 */
	REG_A3,		/* %a3 */
	REG_D1,		/* %d1 */
	REG_D2,		/* %d2 */
	REG_A1,		/* %a1 */
	REG_A4,		/* %a4 */
	REG_D3,		/* %d3 */
	REG_A5,		/* %a5 */
	REG_PC,		/* %pc */
	REG_D4,		/* %d4 */
	REG_D6,		/* %d6 */
	REG_D5,		/* %d5 */
	REG_D7,		/* %d7 */
	REG_CC,		/* %cc */
	REG_SR,		/* %sr */
	REG_A6,		/* %a6 */
	REG_A7,		/* %a7 */
	REG_USP,	/* %usp */
	REG_FP0,	/* %fp0 */
	REG_FP1,	/* %fp1 */
	REG_FP2,	/* %fp2 */
	REG_FP3,	/* %fp3 */
	REG_FP4,	/* %fp4 */
	REG_FP5,	/* %fp5 */
	REG_FP6,	/* %fp6 */
	REG_FP7,	/* %fp7 */
	REG_NONE	/* no register */
};
/* sreg -- scan register name, return encoded value. */

static REG
sreg()
{
    unsigned long name = 0;	/* encoded name */
    unsigned long * lp;		/* pointer into table */

    if (*scanp++ == '%')	/* check for start of register */
    {
	name = PACK(scanp[0],scanp[1]);
	scanp += 2;		/* encode first two chars */

	if (islower(*scanp) || isdigit(*scanp))	/* check for %fp[0-9] or %usp */
	    name = PACK(name,*scanp++);
	
	/* search for register name in list */

	for (lp = rname; *lp != 0; lp++)
	    if (*lp == name)
		return(rcode[lp-rname]);
				/* return encoded value on match */
    }

    /* On any failure, produce error message. */

    serr("Bad register name");
/*NOTREACHED*/
}
