#ident "@(#)screen.c	35.3"
/*
 *		Motorola (c) 1986,1987
 *		All Rights Reserved
 *	screen handler functions for sledit
 */

#include 	<curses.h>
#include	<sys/types.h>
#include	<sys/param.h>
#include	<sys/filsys.h>
#include	<sys/dk.h>
#include	<sys/signal.h>
#include 	"sledit.h"

#define DEBUG

#define HaHa
#undef HaHa

#define CTRL(c) ('c' & 037)
#define ESC	0x1b

static 	int row = 0;		/* current cursor position */
static 	int col = WFIELD2;
static	char ignore = 0;	/* ignore warning */

static char dotbuff[MAXCOL];	/* buffer for . command */

int WLINES;			/* max lines in win */
int WCOLS;			/* max cols in win */

char screenbuf[SLICES*MAXCOL];

/* field table */

static struct field field[FIELDS] = {	WFIELD2, SSIZE,
					WFIELD3, SSIZE,
					WFIELD4, SSIZE,
					WFIELD5, NSIZE,
					WFIELD6, NSIZE,
};

static struct undo undo[2];	/* one for undo, one for unundo */
static char nextundo = 0;	/* which undo buffer to use */

/*
 *	current window
 */


curwin(win)
register WINDOW *win;
{
	WLINES = win->_maxy;
	WCOLS = win->_maxx;
	cbreak();
	nonl();
	noecho();
	idlok(win,TRUE);
	scrollok(win,FALSE);
	keypad(win,TRUE);
	def_prog_mode();
}

/*
 *	screen - screen handler
 */

screen(win,n,rflag)
register WINDOW *win[];
register n;
{
	register i;
	register j;
	register change;

	i = 0;
	curwin(win[i]);	/* this  is current window */
	while((change = wedit(win[i],rflag)) < 0) { /* overlay */
		if( n == 1) {	/* only one screen */
			writemesg("Only one screen",1);
			continue;
		}
		j = i;
		if(change == -1)  /* forward */
			i = (i + 1) % n;
		else 		 /* backward */
			i = ( i == 0 ? n - 1 : i - 1);
		curwin(win[i]);	/* this  is current window */
		clearok(win[i],TRUE);
		wrefresh(win[i]);
	}
			
	return(change);
}

/*
 *	wedit - main screen handling routine
 */

wedit(win,rflag)
register WINDOW *win;
int rflag;
{
	register struct field *fp;
	register char c;
	register char written = 0;	/* written to buffer */
	register unsigned short svrow,svcol;
	char delflag = 0;
	extern char *instructions;
	extern WINDOW *mesgwin;
	extern char *shell;
	extern qflag;
	extern sflag;
	static char *rdonly = "FILE IS READONLY!";
	char shcmd[MAXCOL];
	register char *sp;
	int id;
	WINDOW *readwin();
	

	if(sflag)	/* check if reading in slice table */
		written++;	/* make sure it gets written to disk */

	undo[0].u_row = -1;	/* flag empty undo */
	undo[1].u_row = -1;	/* flag empty undo */
	nextundo = 0;
	row = 0;
	col = WFIELD2;
	for(;;) {
		wmove(win,row,col);
		wrefresh(win);
		c = wgetch(win);
		if( !qflag && ignore && (c != 'I' && c != 'u')) {
		/* hold till decision */
			flash();
			continue;
		}

		switch(c) {

		case '0':		/* goto column 0 */
			col = WFIELD2;
			break;
		case '$':		/* goto last column  */
			col = WFIELD6;
			break;
		case 'w':		/* goto next word */
			fp = &field[where(col)];
			if(fp->f_off == WFIELD6) {
				row = ++row % SLICES;
				col = WFIELD2;
			} else 
				col = (++fp)->f_off;
			break;
		case 'b':		/* goto previous word */
			fp = &field[where(col)];
			if(col > fp->f_off) /* begining of same field*/
				col = fp->f_off;
			else if(fp->f_off == WFIELD2) {
				row = (row + SLICES -1) % SLICES;
				col = WFIELD6;
			} else			  /* previous field */
				col = (--fp)->f_off;
			break;
		case 'e':		/* end of current word */
			fp = &field[where(col)];
			if(col < ((svcol = findend(win,fp,row,col)))) {
				col = svcol;		/* same field */
				break;
			}
			if(fp->f_off == WFIELD6) {	 /* next row */
				row = ++row % SLICES;
				col = WFIELD2;
				fp = &field[where(col)];
			} else if(col == svcol)
				col = (++fp)->f_off; 	/* next field */
			col = findend(win,fp,row,col);
			break;
		case 'd':
			delflag++;
		case 'c':
			if(rflag)
				writemesg(rdonly,1);
			else {
				c = wgetch(win);
				switch(c) {

				case 'w':	/* field from present col */
					fp = &field[where(col)];
					delfield(win,fp,row,col);
					wmove(win,row,col);
					if(delflag == 0)
						winput(win,fp);
					if(writeandcheck(win) == 0)
						written++;
					break;
							
				default:
					flash();
					break;
				}
			}
			delflag = 0;
			break;
		case 'x':	/* delete */
			if(rflag)
				writemesg(rdonly,1);
			else {
				fp = &field[where(col)];
				setundo(win,fp,row,col);
				wmove(win,row,col);
				wdelch(win);
				/* put space at end to compress field */
				wmove(win,row,fp->f_off+fp->f_size);
				winsch(win,' ');
				dotbuff[0] = ' ';
				dotbuff[1] = NULL;
				if(writeandcheck(win) == 0)
					written++;
			}
			break;
		case 'r':	/* replace */
			if(rflag)
				writemesg(rdonly,1);
			else {
				fp = &field[where(col)];
				if((c = wgetch(win)) == ' ' ||
					c == CTRL(H))  { /* don't split field */
					  writemesg("Can't split field",1);
					  break;
				}
				if(fp->f_size == SSIZE && (c < '0' || c > '9')){
					writemesg("Values in decimal",1);
					break;
				}
				setundo(win,fp,row,col);
				wmove(win,row,col);
				wdelch(win);
				dotbuff[0] = c;
				dotbuff[1] = NULL;
				winsch(win,c);
				if(writeandcheck(win) == 0)
					written++;
			}
			break;
		case 'a':	/* append */
			col++;
		case 'i':	/* insert */
			if(rflag)
				writemesg(rdonly,1);
			else {
				fp = &field[where(col)];
				setundo(win,fp,row,col);
				wmove(win,row,col);
				winput(win,fp);
				if(writeandcheck(win) == 0)
					written++;
			}
			break;
		case 'u':	/* undo */
			if(rflag)
				writemesg(rdonly,1);
			else {
				doundo(win);
				written++;
				ignore = 0;
			}
			break;
		case '.':	/* do previous insert */
			put(win,row,col,dotbuff);
			if(writeandcheck(win) == 0)
				written++;
			break;
		case 'I':	/* ignore warning */
			if(writeandcheck(win) == 0) {
				written++;
				ignore = 0;
			}
			break;
		case '\b':
		case 'h':
		case KEY_LEFT:
			fp = &field[where(col)];
			if(col > fp->f_off && (winch(win) != ' '))
				col--;
			else if( col == WFIELD2) 	/* first field */
				flash();
			else				/* left field */
				col = findend(win,--fp,row,fp->f_off);
			break;
		case '\r':
			col = WFIELD2;
		case CTRL(J):
		case 'j':
		case KEY_DOWN:
			if(row < WLINES - 1) {
				fp = &field[where(col)];
				row++;
				if((svcol = findend(win,fp,row,col)) < col)
					col = svcol;
			} else
				flash();
			break;
		case CTRL(K):
		case 'k':
		case KEY_UP:
			if(row > 0) {
				fp = &field[where(col)];
				row--;
				if((svcol = findend(win,fp,row,col)) < col)
					col = svcol;
			} else
				flash();
			break;
		case CTRL(L):
		case ' ':
		case 'l':
		case KEY_RIGHT:
			fp = &field[where(col)];
			if(col < findend(win,fp,row,col))
				col++;
			else if(fp->f_off == WFIELD6) 	/* last field */
				flash();
			else				/* right field */
				col = (++fp)->f_off;
			break;
		case 'z':	/* refresh */
			c = wgetch(win);
			if( c != '.') {
				flash();
				break;
			}
			wclear(mesgwin);	/* clear message window */
			wrefresh(mesgwin);
			break;
		case CTRL(F):		/* next screen forward */
			return(-1);
		case CTRL(B):		/* next screen backward */
			return(-2);
		case ':':
			writemesg(":",0);
			c = wgetch(mesgwin);
			switch(c) {
			case 'q':
				writemesg(":q",0);
				c = wgetch(mesgwin);
				if( c != '\r' && c != '!') {
					flash();
					break;
				}
				if( c == '!') {
					writemesg(":q!",0);
					return(0);
				}
				if(writeandcheck(win) == 0)
					return(written);
				break;

			case '!' : 	/* shell escape */
				writemesg(":!",0);
				sp = shcmd;
				while((*sp = wgetch(mesgwin)) != '\r') {
					waddch(mesgwin,*sp);
					if(*sp == '\b')
						sp--;
					else
						sp++;
					wrefresh(mesgwin);
				}
				waddch(mesgwin,'\r');
				wrefresh(mesgwin);
				*sp = NULL;
				if ((id = fork ()) < 0) {
					break;
				}
				if (id == 0) {
					reset_shell_mode();
					putchar('\n');
					signal(SIGINT,SIG_DFL);
					execl(shell, shell, "-c", shcmd, 0);
					fprintf(stderr,"exec failed\n");
					exit(1);
				}
				wait ((int *) 0);
				fflush(stdout);
				signal(SIGINT,SIG_IGN);
				reset_prog_mode();
				writemesg("<Hit return to continue.>",0);
				while(wgetch(mesgwin) != '\r');
				wclear(mesgwin); /* clear message window */
				wrefresh(mesgwin);
				break;

			case 's' : 	/* save to another file */
				writemesg(":s ",0);
				sp = shcmd;
				while((*sp = wgetch(mesgwin)) != '\r') {
					waddch(mesgwin,*sp);
					if(*sp == '\b')
						sp--;
					else
						sp++;
					wrefresh(mesgwin);
				}
				waddch(mesgwin,'\r');
				wrefresh(mesgwin);
				*sp = NULL;
				if(writeandcheck(win) == 0) {  /* check first */
					writeother(shcmd);
				}
				break;
			default :
				flash();
			}
			break;
		case 'Z':
			c = wgetch(win);
			svrow = row;
			svcol = col;
			if( c != 'Z') { 
				flash();
				break;
			}
			if(rflag) {
				writemesg(rdonly,1);
				break;
			} 
			if(writeandcheck(win) == 0) {
				return(written);
			}
			row = svrow;
			col = svcol;
			break;
		case '?':
			svrow = row;
			svcol = col;
			writemesg("Instructions, hit return to continue",0);
			writewin(win,screenbuf,WLINES,WCOLS); /* save window */
			werase(win);	/* clear window */
			readwin(win,&instructions,WLINES);
			wmove(win,0,0);
			wrefresh(win);
			while(wgetch(win) != '\r');
			werase(win);	/* clear instruction window */
			readwin(win,screenbuf,WLINES); /* put back */
			row = svrow;
			col = svcol;
			break;
		case 'O':	/* open new space */
			if(rflag)
				writemesg(rdonly,1);
			else if(writeandcheck(win) == 0) {
				written++;
				addcmd(win,row);
			}
			break;
		case 'C':	/* compress space */
			if(rflag)
				writemesg(rdonly,1);
			else if(writeandcheck(win) == 0) {
				written++;
				subcmd(win,row);
			}
			break;
#ifdef HaHa
		case CTRL(N):
			writemesg("Sherm, you turkey!!",1);
			break;
		case CTRL(P):
			writemesg("Sherm, you hoser!!",1);
			break;
		case CTRL(^):
			writemesg("Sleditomatic, product of Fly-by-Night Industries"
				,0);
			break;
#endif
		default :
			writemesg("Unknown command, hit '?' for help.",1);
			
		}
	}
}

/*
 *	winput - insert mode in a field
 */

winput(win,fp)
register WINDOW *win;
register struct field *fp;
{
	register c;
	register i;
	register char *dp = dotbuff;
	register end = fp->f_off + fp->f_size - 1;
	register start = fp->f_off;

	if(col > start) {		/* check if splitting field */
		wmove(win,row,col - 1);	/* move cursor to leading char */
		if(winch(win) == ' ' )	/* trying to split field */
			col = start;
		wmove(win,row,col);
	}
	for(;;) {
		wrefresh(win);
		c = wgetch(win);
		if( c == ESC || c == CTRL(D) || c == KEY_EIC || c == '\r') {
			if(--col < fp->f_off) /* move to last char inserted */
				col = fp->f_off;
			break;
		}
		if( c == '\b') {
			if( col == start )  
				flash();
			else {
				wmove(win,row,--col);
				wdelch(win);
				dp--;
				winsch(win,' ');
			}
			continue;
		}
		if( c == ' ' || c == CTRL(H)) {  /* don't split field */
			writemesg("Can't split field",1);
			continue;
		}
		if(fp->f_size == SSIZE && (c < '0' || c > '9')) {
			writemesg("Size and offset values are in decimal",1);
			continue;
		}
			
		if(col >  end ) {
			flash();
			continue;
		}
		wmove(win,row,end);	/* move cursor to end of field */
		wdelch(win);
		wmove(win,row,col);	/* move cursor back */
		winsch(win,c);
		*dp++ = c;	/* save in dotbuff */
		wmove(win,row,col);
		col++;
	}
	*dp = NULL;
}

/*
 *	readwin - read buff into a window
 */

WINDOW *readwin(win,buff,lines)
register  char *buff;
register WINDOW *win;
register lines;
{
	register i;

	for( i = 0; i < lines;i++) {	/* don't forget NULL*/
		mvwaddstr(win,i,0,buff);
		while(*buff++);		/* go past NULL */
		}
	wrefresh(win);
	return (win);
}

/*
 *	writeandcheck - write to sledit buffs and check
 */

writeandcheck(win)
register WINDOW *win;
{
	register ret;

	writewin(win,screenbuf,WLINES,WCOLS);
	ret = stuffslice(screenbuf); 
	if(ret == FATAL) /* undo if error  */
		doundo(win);
	if(ret == WARNING) /* set ignore flag   */
		ignore++;
	return(ret); 
}

/*
 *	writewin - write window into buffer
 */

writewin(win,buff,rows,cols)
register unsigned rows,cols;
register WINDOW *win;
register char *buff;
{
	register i,j;
	register ch;
	register char *bp;

	for( i=0,bp=buff; i < rows;i++) {
		for( j = 0; j < cols;j++) {
			wmove(win,i,j);
			ch = winch(win);
			*bp++ = ch & 0x7f;
		}
		*bp = 0;
		bp = buff = buff + cols + 1; /* next row */
	}
}

/*
 * 	stuffslice - stuff win buff into slice tab
 *		and check
 */

stuffslice(buff)
register char *buff;
{
	register unsigned short i;
	register ret;

	for( i = 0; i < WLINES;i++) {
		if(ret = writeslice(buff,i)) {
			return(ret);
		}
		while(*buff++);
	}
	return(checkslice());
}

/*
 *	writemesg - write to message window
 */

writemesg(mesg,error)
register char *mesg;
{
	extern WINDOW *mesgwin;
	extern qflag;

	if(qflag && error)	/* quiet */
		return;
	wmove(mesgwin,0,0);
	wclrtoeol(mesgwin);
	wprintw(mesgwin,mesg);
	if(error)
		flash();
	wrefresh(mesgwin);
}

/*
 * 	delfield - delete one field starting at col
 */

delfield(win,fp,row,col)
register WINDOW *win;
register struct field *fp;
register col;
register row;
{
	register i;
	register char c;

	setundo(win,fp,row,col);		/* set up for undo */
	wmove(win,row,col);
	for(i = col;i < fp->f_off + fp->f_size;i++) {
		c = winch(win);
		if( c != ' ' ) {
			wdelch(win);
			winsch(win,' ');
			wmove(win,row,++col);
		}
		if( c == ' ' ){
			break;
		}
	}
}

/*
 *	addcmd - add offset and size fields to figure
 *		next offset
 */

addcmd(win,row)
register WINDOW *win;
register row;
{
	register i;
	register off;
	register size;
	register fs;
	register struct field *fp;
	char buff[MAXCOL];	
	register char *bp = buff;
	register struct undo *up;

	if((i = fatslice(row)) == -1) {
		return(-1);	/* no availble space */
	}
	if((fs = fatfs(i)) == 0) {	/* filesys size 0 */
		writemesg("Previous slice has size but no filesystem.",1);
		return(-1);	/* no availble space */
	}
	off = fatoff(i);
	size = fatsize(i);

	/* change size in fat slice */

	sprintf(bp,"%d",fs);
	put(win,i,WFIELD3,bp);

	/* change size in new slice */

	sprintf(bp,"%d",size-fs);
	put(win,row,WFIELD3,bp);

	/* change offset field in new slice */
	
	sprintf(bp,"%d",fs+off);
	put(win,row,WFIELD2,bp);

	col = WFIELD4;	/* position at filesystem */
	up = &undo[(nextundo ? 0 : 1)]; /* last buffer to be filled */
	up->u_row = -1; 		/* kill that last change */

	return(0);
}

/*
 *	subcmd - sub offset and size fields to compress
 */

subcmd(win,roww)
register WINDOW *win;
register roww;
{
	register i;
	register off;
	register fsize;
	register csize;
	register fs;
	register struct field *fp;
	char buff[MAXCOL];	
	register char *bp = buff;
	register struct undo *up;

	if((i = thinslice(roww)) == -1) {
		return(-1);	/* no availble space */
	}
	off = fatoff(i);
	fsize = fatsize(roww);
	csize = fatsize(i);

	/* zero size in compressed slice */

	sprintf(bp,"%d",0);
	put(win,roww,WFIELD3,bp);

	/* change size in fat slice */

	sprintf(bp,"%d",fsize+csize);
	put(win,i,WFIELD3,bp);

	/* zero offset field in compresed slice */
	
	sprintf(bp,"%d",0);
	put(win,roww,WFIELD2,bp);
	
	/* zero filesytem in compressed slice */

	sprintf(bp,"%d",0);
	put(win,roww,WFIELD4,bp);

	/* zero filesytem name in compressed slice */

	sprintf(bp,"%c",' ');
	put(win,roww,WFIELD5,bp);
	put(win,roww,WFIELD6,bp);
	
	up = &undo[(nextundo ? 0 : 1)]; /* last buffer to be filled */
	up->u_row = -1; 		/* kill last change */

	col = WFIELD2;	/* position at previous row, offset col */
	row--;

	return(0);
}

/*
 *	put - do dot or add command
 */

put(win,row,col,buff)
register WINDOW *win;
register char *buff;
register row,col;
{
	register struct field *fp = &field[where(col)];
	register end = fp->f_off + fp->f_size ;
	
	if(*buff == NULL)		/* nothing in buff */
		return;
	setundo(win,fp,row,col);
	wmove(win,row,col);
	while(*buff) {
		if(fp->f_size == SSIZE && (*buff < '0' || *buff > '9')) {
			writemesg("Size and offset values are in decimal",1);
			return;
		}
		if(col >= end ){
			flash();
			col--;
			return;
		}
		wdelch(win);
		winsch(win,*buff++);
		wmove(win,row,++col);
	}
	while(col++ <= end) {		/* clear to end of field */
		wdelch(win);
		winsch(win,' ');
		wmove(win,row,col);
	}
}

/* 
 *	where - in what field am I
 */

where(col)
register col;
{
	register i;
	register struct field *fp = field;

	for(i = 0; i < FIELDS  ; i++) {
		if(col >= fp->f_off)
			fp++;
		if(col < fp->f_off)
			return(i);
	}
	return(i-1);
}
 
/*
 *	findend - find printable end of field
 */

findend(win,fp,row,col)
register WINDOW *win;
register struct field *fp;
register row,col;
{

	wmove(win,row,col);
	while( winch(win) != ' ' && (col < fp->f_off + fp->f_size))
		wmove(win,row,++col);
	while( winch(win) == ' ' && (col > fp->f_off))
		wmove(win,row,--col);
	return(col);
}

/*
 * 	setundo - set up for undo
 */

setundo(win,fp,row,col)
register WINDOW *win;
register struct field *fp;
register row,col;
{
	register struct undo *up = &undo[nextundo];
	register char *cp = &up->u_buff[0];
	register i;

	up->u_row = row;
	col = up->u_col = fp->f_off;
	wmove(win,row,col);
	for(i = 0;i < fp->f_size; i++) { /* copy what is now in field */
		*cp++ = winch(win);
		wmove(win,row,++col);
	}
	*cp = NULL;
	nextundo = (nextundo ? 0 : 1);

}

/*
 *	doundo - do an undo
 */

doundo(win)
register WINDOW *win;
{
	register char *cp;
	register struct undo *up;
	register i;
	register svrow,svcol;

	up = &undo[(nextundo ? 0 : 1)]; /* last buffer to be filled */
	if(up->u_row == -1) 		/* nothing in buffer */
		return;
	svrow = up->u_row;
	svcol = up->u_col;
	cp = &up->u_buff[0];
	setundo(win,&field[where(svcol)],svrow,svcol); /* save */
	wmove(win,svrow,svcol);
	for(;;) {
		if(*cp == NULL)
			break;
		wdelch(win);
		winsch(win,*cp++);
		wmove(win,svrow,++svcol);
	}
	wmove(win,svrow,up->u_col); /* move to undone spot */
	row = svrow;
	col = up->u_col;
}
