#ident "@(#)sledit.c	35.5"
/*
 *		Motorola (c) 1986,1987
 *		All Rights Reserved
 *
 *		sledit - slice editor
 */

#include	<stdio.h>
#include	<curses.h>
#include	<sys/signal.h>
#include	<sys/types.h>
#include	<sys/param.h>
#include	<sys/filsys.h>
#include	<sys/stat.h>
#include	<ustat.h>
#include	<fcntl.h>
#include	<sys/ino.h>
#include	<sys/dk.h>
#include	"sledit.h"

#define DEBUG
#undef	DEBUG

#define FTYPE(x)	(x & 0170000)	/* file type (from stat(2)) */

/* file system block size */
#if (vax || u3b || u3b5 || m68k || m88k) && (FsTYPE == 3)
#define FSBSIZE	(BSIZE*2)
#else
#define	FSBSIZE BSIZE
#endif

#define	NBINODE	(FSBSIZE/sizeof(struct dinode))


struct dk {
	struct volumeid vid;
	struct dkconfig config;
};


char instructions[] = 
{"\
  k = down  j = up      i = insert          u = undo\0\
  h = left  l = right   a = append         ^F = next screen\0\
  arrow keys           cw = change word    ^B = previous screen\0\
  e = end of word      dw = delete word    :q,ZZ = quit\0\
  b = previous word    ESC = end change    :q! = quit(no write)\0\
  w = next word        . = prev insert     :s file = save in file\0\
  r = replace char     I = ignore warning   C = close space\0\
  ! = shell escape     ? = help             O = open space\0\
\0\
\0\
\0\
\0\
\0\
\0\
\0\
\0\
"};

static char header2[] =
{ "Slice Table"};

static char header3[] =
{ "File System Table"};


static char	*me;			/* name of program */
static char	vflag = 0;		/* verbose flag */
char		qflag = 0;		/* quiet flag */
static char	rflag = 0;		/* readonly flag */
char		sflag = 0;		/* slice copy  flag */
static char	lflag = 0;		/* edit last slice */
static char	nflag = 0;		/* non interactive mode */
static 		slices = 0;		/* number of slices read from disk */
static		inwindow = 0;		/* editing inside window */

static struct	dk dkbuf;		/* volid and config  buffer */
static char 	msgbuf[MAXCOL];		/* message buffer */

static char	*dev;			/* device name */
static char	*slicedev;		/* slice copy device name */
static char	devbase[MAXCOL];	/* base device name */

char *shell = NULL;			/* name of shell program */


/* slice table and slice buffer */

static struct sledit *slicetab = NULL;
static struct slice  slicebuff[BSIZE/sizeof(struct slice)];


WINDOW	*editwin[MAXSCREEN];		/* edit window table, SLICES per win */
WINDOW	*mesgwin;			/* message window */


/*
 *	main procedure
 */

main( argc, argv )
int argc;
register char **argv;
{	int optchar;
	register arg;
	register fd;
	register change = 0;
	register i;
	register n;
	register char *cp;
	extern	char	*optarg;
	extern	int	optind;
	WINDOW 	*readwin();
	char buff[MAXCOL];
	char *getenv();

	me = argv[0];

	while(( optchar = getopt( argc, argv, "ls:vrnq" )) != EOF ) {
		switch(optchar) {
		case 'r':	rflag |= RDONLY;
				break;
		case 'q':	qflag++;
				break;
		case 'v':	vflag++;
				break;
		case 's':	sflag++;
				slicedev = optarg;
				break;
		case 'l':	lflag++;
				break;
		case 'n':	nflag++;
				break;
		case '?':
		default:	usage();
				break;
		}
	}
	if(( arg = optind ) == argc ) {
		slederr( "Missing device file","",-1);
		usage();
	}
	dev = argv[arg];
	strcpy(devbase,dev);

	/* chop off slice number from "dev" string */

	cp = &devbase[strlen(devbase) - 1];
	while( *cp >= '0' && *cp <= '9')
		cp--;
	*++cp = NULL;

	if(( fd = open( dev, O_RDONLY )) == -1 ) {
		slederr("Can't open ",dev,-1 );
		exit(-1);
	}
	if( readslice(fd,dev,0) == -1)	/* go read the disk slice info */
		exit(-1);
	close(fd);

	if(sflag) { 	/* slice table from somewhere else */ 
		if(( fd = open( slicedev, O_RDONLY )) == -1 ) {
			slederr("Can't open ",slicedev,-1 );
			exit(-1);
		}
		if( readslice(fd,slicedev,1) == -1) /* again read slice info */
			exit(-1);
		close(fd);
	}
	if(sflag == 0) { 	/* slice table from here */ 
		fd = open( dev, O_RDONLY );
		if( checkmount(fd) == -1)	/* check for mounted slices */
			exit(-1);

		if( readfilsys(fd) == -1)	/* read the disk filesys info */
			exit(-1);
		close(fd);
	}



	if(nflag) { 	/* non-interactive mode */
		if(checkslice())
			exit(-1);
		if(( fd = open( dev, O_WRONLY|O_TRUNC )) == -1 ) {
			slederr( "Can't open ", dev,-1 );
			exit(-1);
		}

		if(writedisk(fd) )  
			exit(-1);
		close(fd);
		exit(0);
	}

	if ((shell = getenv("SHELL")) == NULL)	  /* get name of shell program*/
		shell = "/bin/sh";
	
	/* initialize curses and check TERM */
	if( newterm( getenv("TERM"), stdout, stdin ) == NULL ) {
		slederr( "Unknown TERM, TERMINFO "
		,"or can't find terminfo file.",-1 );
		exit(-1);
	}

	/*
 	 *	going inside curses. From here on, ignore interrupts
	 */

	signal(SIGINT,SIG_IGN);

	/*
	 *	headers
	 */

	sprintf(buff,"%s of:",me);
	readwin(newwin(1,strlen(buff),0,0),buff,1);

	sprintf(buff,"%s, %d slices, 512 byte blocks",dev,slices);
	if(rflag)
		strcat(buff,",READ ONLY");

	readwin(newwin(1,strlen(buff),1,0),buff,1);

	readwin(newwin(1,strlen(header2),HEADER,FIELD2),header2,1);
	readwin(newwin(1,strlen(header3),HEADER,FIELD4),header3,1);
	readwin(newwin(1,5,HEADER+2,FIELD1),"slice",1);
	readwin(newwin(1,6,HEADER+2,FIELD2),"offset",1);
	readwin(newwin(1,4,HEADER+2,FIELD3),"size",1);
	readwin(newwin(1,4,HEADER+2,FIELD4),"size",1);
	readwin(newwin(1,6,HEADER+2,FIELD5),"fsname",1);
	readwin(newwin(1,6,HEADER+2,FIELD6),"vol-id",1);

	/*
	 *	message window
	 */

	mesgwin = newwin(1,MAXCOL,BOTTOM-2,COL1+1);
	readwin(mesgwin,msgbuf,1);

	/*
	 *	allocate windows and fill
	 * 	each window full of SLICES slices
	 */

	for( i = 0,n = slices; n > 0; i++,n -= SLICES) {
		editwin[i] = newwin(slices,WINLEN,LINE1,COL1);
		fillwin(editwin[i],i*SLICES,slices);
	}

	inwindow++;		/* going inside window, print mesgs there */

	change = screen(editwin,i,rflag);
			
	while((rflag == 0) && change) {
		if(( fd = open( dev, O_WRONLY|O_TRUNC )) == -1 ) {
			slederr( "Can't open ", dev,-1 );
			exit(-1);
		}

		if(writedisk(fd) == 0) { /* write sledit buffers to disk  */
			break;
		}
		close(fd);
		change = screen(editwin,i,rflag);
	}
	if( rflag == 0 && change == 0)
		writemesg("No change",0);
	inwindow = 0;

	endwin();		/* exit curses */

	/* reinstate interrupts */
	signal(SIGINT,SIG_DFL);

	if( (rflag == 0) && change)  /* make filsystems if change */
		filsys(devbase);
	putchar('\n');
}

/*
 *	usage: print out usage string
 */

usage()
{
	fprintf( stderr, "Usage: %s [ -s file -v -r -l -n -q] rawdev\n",me);
	exit(ERR);
}

/*
 *	fillwin - fill window with slicetab
 */

fillwin(win,start,end)
register WINDOW *win;
register end;
register start;
{
	register i;
	register struct sledit *sp;
	char buff[MAXCOL];
	char space[SPACE+2];
	char fname[NSIZE+1];
	char fpack[NSIZE+1];
	register char *tp = space;
	register char *bp = buff;

	for( i = 0; i < SPACE; i++)
		*tp++ = ' ';
	*tp = NULL;
	tp = space;
	for( i = start,sp = &slicetab[start]; i < start+end; sp++, i++) {
		strncpy(fname,sp->fs.s_fname,NSIZE);
		strncpy(fpack,sp->fs.s_fpack,NSIZE);
		sprintf(bp,"%-3d%s%-10d%s%-10d%s%-10d%s%-6s%s%-6s",sp->slice,
			tp,sp->s_tab.blkoff,tp,sp->s_tab.nblocks,tp,
			sp->fs.s_fsize,tp,fname,tp,fpack);
		mvwaddstr(win,i-start,0,bp);
	}
}


/*
 *	readslice - read slicetable  into buffer
 */

readslice(fd,dev,again)
register fd;
register char *dev;
register again;
{
	register struct dk *dp = &dkbuf;
	register struct sledit *sp;
	register struct slice *bp = slicebuff;
	register i;
	static char *mot = "MOTOROLA";
	static char *exor = "EXORMACS";

	if ( !again)
		memset(dp,0,sizeof(dkbuf));
	else 		/* reading slice table from second file */
		if(( dp = (struct dk *) malloc(sizeof(struct dk))) == NULL) {
			slederr("No space  ","",-1);
			return(FATAL);
		}

	/*
	 *	read vid and config info
	 */


	if( read(fd,dp,sizeof(struct dk)) == -1) {
		slederr("Volume id read error ",dev,-1);
		return(FATAL);
	}

	if( strncmp(mot,dp->vid.vid_mac,8) && strncmp(exor,dp->vid.vid_mac,8)) {
		slederr("Volume id read error ",dev,-1);
		return(FATAL);
	}

	if((slices = dp->config.ios_slccnt) == 0) {

		slederr("No slice table ",dev,-1);
		return(FATAL);
	}

	/*
	 *	read slice table
	 */

	if(lseek(fd,dp->config.ios_slcblk*BSIZE,0) == -1) {
		slederr("Slice table seek error","",-1);
		return(FATAL);
	}

	if( read(fd,bp,BSIZE) == -1) {
		slederr("Slice table read error","",-1);
		return(FATAL);
	}
	if(slicetab)		/* allocated on last call */
		free(slicetab);
	slicetab = (struct sledit *) malloc( slices*((sizeof(struct sledit))));

	if( slicetab == NULL )	{	/* malloc slice table fails */
		slederr("No space for slice table ","",-1);
		return(FATAL);
	}

	/* copy slicebuf to slicetab */

	for(i = 0, sp = slicetab; i < slices; i++,sp++,bp++ ) {
		sp->s_tab.nblocks = bp->nblocks;
		sp->s_tab.blkoff = bp->blkoff;
		sp->slice = i;
		sp->s_flag = 0;
		if(i == slices - 1) 	/* last slice, entire disk */
			sp->s_flag = (lflag ? 0 : RDONLY); /* default */
		
	}
	if( again)
		free(dp);
	return (slices);
}


/*
 *	stat files, determine if raw dev 
 */

checkmount(fd)
register fd;
{
	register struct sledit *sp;
	register unsigned short len;
	register unsigned short i;
	struct stat statarea;
	struct ustat ustatarea;
	register struct stat *statp = &statarea;
	register struct ustat *ustatp = &ustatarea;
	char bdev[80];
	char num[10];

	if(stat(dev,statp) < 0) {
		fprintf(stderr,"%s: Can't stat %s\n",me,dev);
		return(FATAL);
	}
	if (FTYPE(statp->st_mode) != S_IFCHR) {
		fprintf( stderr,"%s: must be raw device - %s\n",me,dev);
		usage();
	} else if (strncmp(devbase, "/dev/r", 6) != 0) {
		fprintf(stderr, 
			"%s: raw device must be \"/dev/r*\" - %s\n"
				,me,dev);
		usage();
	}
	strcpy(bdev, "/dev/");
	strcat(bdev, devbase+6);
	len = strlen(bdev);

	for(i = 0,sp = slicetab; i < slices; i++,sp++ ) {
		sprintf(num,"%d",i);
		strcat(bdev,num);
		if (stat(bdev, statp) < 0 ) {
			fprintf(stderr, "%s: cannot open block device - %s\n"
				,me,bdev);
			usage();
		}
		if(ustat(statp->st_rdev,ustatp) >= 0) {
		  /* mounted filesystem */
			sp->s_flag |= (MOUNTED | RDONLY);
			printf("MOUNTED FILESYSTEM slice %d\n",i);
		}
		bdev[len] = 0;
	}
	return(0);
}

/*
 *	read super blocks
 */

readfilsys(fd)
register fd;
{
	register struct sledit *sp;
	register unsigned short i;

	for(i = 0,sp = slicetab; i < slices; i++,sp++ ) {

		if(sp->s_tab.nblocks == 0) { /* no space for filesys */
			sp->fs.s_fsize = 0;
			*sp->fs.s_fname = NULL;
			*sp->fs.s_fpack = NULL;
			sp->s_flag |= FSKILL;		/* alright to kill fs */
			continue;
		}

		if(lseek(fd,(sp->s_tab.blkoff + 1)*BSIZE,0) == -1) {	
			slederr("Super block seek error","",i);
			return(FATAL);
			}
		if( read(fd,&sp->fs.s_isize,sizeof(struct filsys)) == -1) {
			slederr("Super block read error","",i);
			return(FATAL);
		}
		
		if(sp->fs.s_magic != FsMAGIC) {		/* no filesys */
			sp->fs.s_fsize = 0;
			*sp->fs.s_fname = NULL;
			*sp->fs.s_fpack = NULL;
			sp->s_flag |= FSKILL;		/* alright to kill fs */
		} else					/* make 1/2K blocks */
			sp->fs.s_fsize *= sp->fs.s_type;


	}
	return(0);
}

/*
 *	thinslice - find the slice that has some size for compression
 *		of the slice
 */

thinslice(slice)
register unsigned short slice;
{
	register short i;
	register struct sledit *sp;
	register struct sledit *np;

	if(slice == 0) {
		slederr("Can't compress slice 0.","",slice);
		return(FATAL);
	}
	np = &slicetab[slice];
	if(ckrdflag(np->s_flag,slice))
		return(FATAL);
	if(np->s_tab.nblocks == 0){	
		slederr("Can't compress, no size.","",slice);
		return(FATAL);
	}
	for(i = slice-1,sp= &slicetab[slice-1]; i >= 0 ;sp--,i--) 
		if(sp->s_tab.nblocks) {
			if(sp->s_tab.nblocks + sp->s_tab.blkoff >
				np->s_tab.blkoff) {
				slederr("Can't compress, overlapped",
					" slices.",slice);
				return(FATAL);
			}
			if(sp->s_tab.nblocks + sp->s_tab.blkoff !=
				np->s_tab.blkoff) {
				slederr("Can't compress, non-continuous"
					," slices.",slice);
				return(FATAL);
			}
			if(ckrdflag(sp->s_flag,i))
				return(FATAL);
			np->s_flag |= FSKILL; /* killing filesys */
			return(i);
		}
	slederr("No slices to compress with.","",slice);
	return(FATAL);
}


/*
 *	fatslice - find the slice that contains the unused portion
 *		of the disk
 */

fatslice(slice)
register unsigned short slice;
{
	register short i;
	register struct sledit *sp;

	if(slice == slices - 1) { 	/* don't touch last slice */
		slederr("Last slice unavailable.","",slice);
		return(FATAL);
	}
	if(slice)
		for(i = slice-1,sp= &slicetab[slice-1]; i >= 0 ;sp--,i--) 
			if(sp->s_tab.nblocks) 
				if(sp->s_tab.nblocks > sp->fs.s_fsize) {
					if(ckrdflag(sp->s_flag,i))
						return(FATAL);
					return(i);
				} else
					break;
	slederr("No available space to allocate.","",slice);
	return(FATAL);
}


/*
 *	fatfs - return size of filesystem on fat slice
 */

fatfs(slice)
register unsigned short slice;
{

	return(slicetab[slice].fs.s_fsize);
}

/*
 *	fatoff - return offset on fat slice
 */

fatoff(slice)
register unsigned short slice;
{

	return(slicetab[slice].s_tab.blkoff);
}

/*
 *	fatsize - return size on fat slice
 */

fatsize(slice)
register unsigned short slice;
{

	return(slicetab[slice].s_tab.nblocks);
}

/*
 *	checkslice - check slice offsets and sizes for overlap 
 * 			except for last slice
 */

checkslice()
{
	register struct sledit *sp;
	register struct sledit *np;
	register i;
	register j;
	register high,low;
	register range;
	register lastslc = slices - 1; 	/* everything but last slice */
	int lastoff;
	char  buff[MAXCOL];
	static char *warn = "Warning, slice";

	/*
	 *	check slice offsets and sizes for overlap
	 */
	
	lastoff = slicetab[lastslc].s_tab.nblocks; /* greatest offset allowed */

	for(i = 0,sp = slicetab; i < lastslc; i++, sp++ ) {
		range = sp->s_tab.nblocks + sp->s_tab.blkoff;

		for(j=0,np = slicetab; j < lastslc; j++ , np++) {
			if(j == i ) /* this slice */
				continue;

			low = np->s_tab.blkoff;
			high = np->s_tab.nblocks + np->s_tab.blkoff;

			if(((sp->s_tab.blkoff < high) && 
			  (sp->s_tab.blkoff >= low)) ||
			  ((sp->s_tab.blkoff < low) && (range > low))) {

				if((sp->s_flag & OVERLAP) == 0) {
					sprintf(buff,"%s %d ",warn,j);
					slederr(buff,
					"overlaps, I to ignore, u to undo.",i);
					sp->s_flag |= OVERLAP;
					np->s_flag |= OVERLAP; /* for both */
					return(WARNING);
				}
			}
		}
		if( range > lastoff) {
			slederr("Slice beyond end of disk","",i);
			return(FATAL);
		}

			
	}
	return(OKAY);
}

/*
 *	writedisk - write buffer into slice table on disk
 */

writedisk(fd)
register fd;
{
	register struct dk *dp = &dkbuf;
	register struct sledit *sp;
	register i;
	register struct slice *bp = slicebuff;



	/*
	 *	write slice table
	 */

	/* set driver slice table first */

	for(i = 0,sp = slicetab; i < slices; i++, sp++, bp++ ) {
		bp->nblocks = sp->s_tab.nblocks;
		bp->blkoff = sp->s_tab.blkoff;
	}

	if(ioctl(fd,DKSETSLC,slicebuff) == -1) {
		slederr("Slice table ioctl error.","",-1);
		return(FATAL);
	}

	/* ok, write to disk */

	if(lseek(fd,dp->config.ios_slcblk*BSIZE,0) == -1) {
		slederr("Slice table seek error","",-1);
		return(FATAL);
		}
	if( write(fd,slicebuff,BSIZE) == -1) {
		slederr("Slice table write error","",i);
		return(FATAL);
	}

	return(0);
}

/*
 *	writeother - write buffer into another file
 */

writeother(file)
register char *file;
{
	register struct dk *dp = &dkbuf;
	register struct sledit *sp;
	register i;
	register struct slice *bp = slicebuff;
	struct stat statarea;
	register struct stat *statp = &statarea;
	register fd;

	if(*file == NULL) {
		slederr("File name missing","",-1);
		return(FATAL);
	}

	/*
	 *	write slice table
	 */

	/* set driver slice table first */

	for(i = 0,sp = slicetab; i < slices; i++, sp++, bp++ ) {
		bp->nblocks = sp->s_tab.nblocks;
		bp->blkoff = sp->s_tab.blkoff;
	}

	if(( fd = open( file, O_CREAT | O_TRUNC | O_WRONLY,0666)) == -1 ) {
		slederr("Can't open ",file,-1 );
		return(FATAL);
	}

	if(fstat(fd,statp) < 0) {
		slederr("Can't stat ",file,-1);
		close(fd);
		return(FATAL);
	}
	if (FTYPE(statp->st_mode) != S_IFREG) {
		slederr("Not a regular file: ",file,-1);
		close(fd);
		return(FATAL);
	}

	/* write volid/config block to disk */

	i = dp->config.ios_slcblk;	/* save slice block number */
	dp->config.ios_slcblk = 1;	/* change to 1 to save space */

	if( write(fd,dp,sizeof(struct dk)) == -1) {
		slederr("Volume id write error ",file,-1);
		close(fd);
		return(FATAL);
	}


	dp->config.ios_slcblk = i;	/* change back */

	if( write(fd,slicebuff,BSIZE) == -1) {
		slederr("Slice table write error",file,i);
		close(fd);
		return(FATAL);
	}

	close(fd);
	return(0);
}
/*
 *	writeslice - write one slice to slicetab
 */

writeslice(buff,n)
register char *buff;
register unsigned short n;
{
	register i;
	register struct sledit *sp = &slicetab[n];
	struct sledit rbuff;	/* read buffer */
	register struct sledit *rp = &rbuff;
	register char *bp;


	if(sscanf(buff,"%hd%d%d%d",&rp->slice, /* no filesys names */
		&rp->s_tab.blkoff,&rp->s_tab.nblocks,
		&rp->fs.s_fsize ) != 4 ) {

		slederr("Unknown or missing field","",n);
		return(FATAL);
	}
	*rp->fs.s_fname = NULL;
	*rp->fs.s_fpack = NULL;
	buff += WFIELD5;		/* filesys name field */
	strncpy(rp->fs.s_fname,buff,NSIZE);
	for(i=0,bp=rp->fs.s_fname;i<NSIZE;bp++,i++) /* kill spaces in name  */
		if(*bp == ' ') {
			*bp = NULL;
			break;
		}
	buff += (WFIELD6 - WFIELD5);	/* filesys pack field */
	strncpy(rp->fs.s_fpack,buff,NSIZE);
	for(i=0,bp=rp->fs.s_fpack;i<NSIZE;bp++,i++) /* kill spaces in name  */
		if(*bp == ' ') {
			*bp = NULL;
			break;
		}

	/*
	 *	check filesystem size
	 */

	if(rp->fs.s_fsize > rp->s_tab.nblocks) {
		slederr("Filesystem size greater than slice size","",n);
		return(FATAL);
	}
	/*
	 *	changing offset
   	 */
	if(rp->s_tab.blkoff != sp->s_tab.blkoff) 
		if(ckrdflag(sp->s_flag,n))
			return(FATAL);
		else if (sp->fs.s_fsize ) { 
			if((sp->s_flag & FSKILL) == 0 ) {
				slederr("Change will destroy filesystem. ",
					"I to ignore, u to undo",n);
				sp->s_flag |= FSKILL;
				return(WARNING);
			}
			sp->s_flag |= (MODIFIED | FSMOD | NMMOD);
		} else {
			sp->s_flag |= MODIFIED;
		}
	/*
	 * 	changing number of blocks
	 */
	if(rp->s_tab.nblocks != sp->s_tab.nblocks) /* number of blks in slice */
		if(ckrdflag(sp->s_flag,n))
			return(FATAL);
		else {
			sp->s_flag |= MODIFIED;
		}
	/*
	 *	changing fileys size
	 */
	if(rp->fs.s_fsize != sp->fs.s_fsize)
		if(ckrdflag(sp->s_flag,n))
			return(FATAL);
		else if (rp->fs.s_fsize > 0) { 
			if((sp->s_flag & FSKILL) == 0 && sp->fs.s_fsize > 0) {
				slederr("Change will destroy filesystem. ",
					"I to ignore, u to undo",n);
				sp->s_flag |= FSKILL;
				return(WARNING);
			}
			sp->s_flag |= (MODIFIED | FSMOD | NMMOD);
		}
	/*
	 *	changing names
	 */
	if(strncmp(sp->fs.s_fname,rp->fs.s_fname,6))
		if(ckrdflag(sp->s_flag,n))
			return(FATAL);
		else {
			sp->s_flag |= (MODIFIED | NMMOD);
		}
	if(strncmp(sp->fs.s_fpack,rp->fs.s_fpack,6))
		if(ckrdflag(sp->s_flag,n))
			return(FATAL);
		else {
			sp->s_flag |= (MODIFIED | NMMOD);
		}

	if(sp->s_flag & MODIFIED) {		/* copy to slicetab */
		sp->fs.s_fsize = rp->fs.s_fsize;
		sp->s_tab.nblocks = rp->s_tab.nblocks;
		sp->s_tab.blkoff = rp->s_tab.blkoff;
		strncpy(sp->fs.s_fname,rp->fs.s_fname,6);
		strncpy(sp->fs.s_fpack,rp->fs.s_fpack,6);
		sp->s_flag &= ~MODIFIED;	/* modification copied */
	}
	return(OKAY);
}

/*
 *	ckrdflag - check if readonly or mounted
 */

static ckrdflag(flag,slice)
register unsigned char flag;
register unsigned short slice;
{
	if(flag & MOUNTED) {
		slederr("Mounted filesystem, Read only","",slice);
		return(FATAL);
	}
	if(flag & RDONLY) {
		slederr("Read only","",slice);
		return(FATAL);
	}
	return(OKAY);
}

/*
 *	slederr - print error message
 */
 
slederr(s1,s2,slice)
register char *s1;
register char *s2;
register slice;
{
	char buff[MAXCOL];
	register char *bp=buff;
	
	if(slice >= 0)
		sprintf(bp,"slice %d: %s%s",slice,s1,s2);
	else
		sprintf(bp,"%s: %s%s",me,s1,s2);
	if(inwindow)
		writemesg(bp,1);
	else
		fprintf(stderr,"%s\n",bp);
}

/*
 *	filsys - make a change to the file systems
 */

filsys(dev)
register char *dev;
{
	register i;
	register struct sledit *sp;
	char path[128];
	char slice[8];

	for(i = 0,sp = slicetab; i < slices - 1; i++,sp++ ) {
		strcpy(path,dev);		/* base pathname */
		sprintf(slice,"%d",i);
		strcat(path,slice);
		if(sp->fs.s_fsize == 0)		/* no filesys */
			continue;
		if((sp->s_flag & FSMOD) )	/*  filsys change */
			makefs(path,sp);
		if((sp->s_flag & NMMOD) )	/*  name change */
			label(path,sp);
	}
		
}

/*
 *	makefs - call mkfs(1M) to make new filesystems
 */

makefs(dev,sp)
register char *dev;
register struct sledit *sp;
{
	register char *cp;
	register unsigned short gapdef = 1;
	register unsigned short blockdef = 400;
	char cmd[120];
	char size[MAXCOL];
	char inodes[MAXCOL];
	char gap[MAXCOL];
	char blocks[MAXCOL];
	static char *mkfs = "/etc/mkfs";


	*inodes = NULL;
	*gap = NULL;
	*blocks = NULL;
	if(vflag ) {
		fprintf(stdout,"\nMaking filesystem on slice %s\n",dev);
		fprintf(stdout,"\tNumber of inodes? (default %d): ",
			NBINODE*(sp->fs.s_fsize/
			((FSBSIZE/BSIZE)*(NBINODE*4))));
		fflush(stdout);
		gets(inodes,MAXCOL,stdin);
		fprintf(stdout,"\tGap size? (default %d): ",gapdef);
		fflush(stdout);
		gets(gap,MAXCOL,stdin);
		fprintf(stdout,"\tBlocks/cyl? (default %d): ",blockdef);
		fflush(stdout);
		gets(blocks,MAXCOL,stdin);
	}
	strcpy(cmd,mkfs);
	strcat(cmd," ");
	strcat(cmd,dev);
	strcat(cmd," ");
	sprintf(size,"%d",sp->fs.s_fsize);
	strcat(cmd,size);
	if(*inodes) {
		strcat(cmd,":");
		strcat(cmd,inodes);
	}
	if(*gap || *blocks) {	/* gaps and blocks must both be present */
		if(*gap == NULL)
			sprintf(gap,"%d",gapdef);
		strcat(cmd," ");
		strcat(cmd,gap);
		if(*blocks == NULL)
			sprintf(blocks,"%d",blockdef);
		strcat(cmd," ");
		strcat(cmd,blocks);
	}
#ifdef DEBUG
	printf("%s\n",cmd);
#endif
	system(cmd);

}

/*
 *	label - simulate "labelit(1)" if names given
 */

label(dev,sp)
register char *dev;
register struct sledit *sp;
{
	struct filsys fs;
	register fd;
	register struct filsys *fp = &fs;

	if((fd = open(dev,O_RDWR)) == -1) {
		slederr("Can't open",dev,-1);
		return(FATAL);
	}
	if(lseek(fd,BSIZE,0) == -1) {	
		slederr("Super block seek error",dev,-1);
		close(fd);
		return(FATAL);
	}
	if( read(fd,fp,sizeof(struct filsys)) == -1) {
		slederr("Super block read error",dev,-1);
		close(fd);
		return(FATAL);
	}

	strncpy(fp->s_fname,sp->fs.s_fname,6);
	strncpy(fp->s_fpack,sp->fs.s_fpack,6);

	if(lseek(fd,BSIZE,0) == -1) {	
		slederr("Super block seek error",dev,-1);
		close(fd);
		return(FATAL);
	}
	if( write(fd,fp,sizeof(struct filsys)) == -1) {
		slederr("Super block write error",dev,-1);
		close(fd);
		return(FATAL);
	}
	close(fd);
	return(0);
}
