/*	Copyright (c) 1986  Motorola Inc.	*/
/*	All Rights Reserved 			*/
/*	@(#)ubug.y	35.2		*/

%term	HE	/* he command */
%term	MM	/* mm command */
%term	MD	/* md command */
%term	MDS	/* mds command */
%term	DC	/* dc command */
%term	ERR	/* error */
%term	QUIT	/* quit or go command */
%term	SYSCMD	/* Unix system command */
%term	DOLLAR /* sign for hex */
%term	AMPER /* sign for decimal */

%term	NUMBER	/* hex number */
%term	SECTNUMBER	/* hex sector number */
%term	BLKNUMBER	/* hex 1K block number */
%term	BNUMBER	/* hex 1/2 K block number */
%term	PLUS	/* binary plus operator */
%term	MINUS	/* binary minus operator */
%term	TIMES	/* binary multiply operator */
%term	DIV	/* binary divide operator */

%term	NL	/* \n */
%term	LONG	/* ;l */
%term	WORD	/* ;w */
%term	BYTE	/* ;b */

%term	LP	/* ( */
%term	RP	/* ) */

%{
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#define TRUE	1
#define FALSE	0
%}
%%

command		: comm
		{	printf("Ubug > ");
		}
		| command comm
		{	printf("Ubug > ");
		}
		;

comm		: MD expr oexpr NL
		{	addr = $2;
			siz = $3;
			display();
		}
		| MD expr NL
		{	addr = $2;
			siz = 0x10;
			display();
		}
		| MD NL
		{
			siz = 0x10;
			display();
		}
		| MDS expr NL
		{	addr = $2;
			siz = 0x10 * 0x10;
			display();
		}
		| MDS NL
		{
			siz = 0x10 * 0x10;
			display();
		}
		| MM expr omod NL
		{	addr = $2;
			msiz = $3;
			modify();
		}
		| MM expr NL
		{	addr = $2;
			msiz = 1;
			modify();
		}
		| MM NL
		{
			modify();
		}
		| QUIT NL
		{	exit(0);
		}
		| DC expr NL
		{	printf("%08x = $%x = &%d\n", $2, $2, $2);	
		}
		| DC NL
		{	printf("Entering dc(1)\n");
			system("dc");
		}
		| HE NL
		{	
			printf("	dc   [expression] hex calculator\n");
			printf("	dc   desk calculator (as in unix)\n");
			printf("	he   help\n");
			printf("	md   [start] [count]   - memory display\n");
			printf("	mds  [start]           - memory display screen\n");
			printf("	mm   [start]           - memory modify\n");
			printf("	q      - quit\n");
			printf("	sh     - escape to shell\n");

			printf("	Numbers are hex\n");
			printf("	  s.number = 256 byte sector number\n");
			printf("	  b.number = 512 byte block number\n");
			printf("	  $number = hex number\n");
			printf("	  &number = decimal number\n");
		}
		| SYSCMD NL
		{	
			system( "sh" );
		}
		| ERR NL
		{	printf("?\n");
		}
		| NL
		{	display( addr, siz );
		}
		| error NL
		{	printf("?\n");
		}
		;

expr		: number
		{	$$ = $1;
#ifdef	DEBUG
		printf("EXPR: %d\n", $$ );
#endif
		}
		| expr PLUS number
		{	$$ = $1+$3;
#ifdef	DEBUG
		printf("EXPR: %d\n", $$ );
#endif
		}
		| expr MINUS number
		{	$$ = $1-$3;
#ifdef	DEBUG
		printf("EXPR: %d\n", $$ );
#endif
		}
		| expr TIMES number
		{	$$ = $1*$3;
#ifdef	DEBUG
		printf("EXPR: %d\n", $$ );
#endif
		}
		| expr DIV number
		{	$$ = $1/$3;
#ifdef	DEBUG
		printf("EXPR: %d\n", $$ );
#endif
		}
		| LP expr RP
		{	$$ = $2;
#ifdef	DEBUG
		printf("EXPR: %d\n", $$ );
#endif
		}
		;

number		: NUMBER
		{	$$ = atox( yytext );
#ifdef	DEBUG
		printf("NUMBER %d\n", $$ );
#endif
		}
		| SECTNUMBER
		{	$$ = 0x100 * atox( &yytext[2] );
#ifdef	DEBUG
		printf("SECTNUMBER %d\n", $$ );
#endif
		}
		| BNUMBER
		{	$$ = 0x200 * atox( &yytext[2] );
#ifdef	DEBUG
		printf("BNUMBER %d\n", $$ );
#endif
		}
		| DOLLAR
		{	$$ = atox( &yytext[1] );
		}
		| AMPER
		{	$$ = atoi( &yytext[1] );
		}
		;

oexpr		: expr
		{	$$ = $1;
		}
		| /* empty */
		{	$$ = siz;
		}
		;

omod		: LONG
		{	$$ = 4;
		}
		| WORD
		{	$$ = 2;
		}
		| BYTE
		{	$$ = 1;
		}
		| /* empty == BYTE */
		{	$$ = msiz;
		}
		;
%%

#include "lex.yy.c"


int	addr	= 0;
int	siz	= 0x10;
int	msiz	= 1;
int	ronly;

int	fd;
FILE	*pf;		/* print file pointer */

char	*printfile = "ubug.out";
int	file_siz;	/* r/w limit for regular files */

main( argc, argv )
int	argc;
char	**argv;

{
	char *target = "/dev/mem";
	struct stat status, *st;

	if( argc > 2 )
	{	fprintf( stderr, "Usage: %s [filename]\n", argv[0] );
		exit(1);
	}
	if( argc == 2 ){
		target = argv[1];
	}

	if(( fd = open( target, 2 )) == -1 )
	{	if(( fd = open( target, 0 )) == -1 )
		{	fprintf(stderr,"%s: can't open %s\n",argv[0],target );
			exit(2);
		}

		fprintf( stderr, "%s opened read only\n" );
		ronly = 1;
	}

	st = &status;
	if( fstat( fd, st) )
	{	fprintf(stderr,"%s: can't stat %s\n",argv[0],target );
		exit(2);
	}
	if( (st->st_mode & S_IFMT) == S_IFREG ) /* set limit on regular files */
	{
		if( ! (file_siz = st->st_size) )
		{
			fprintf(stderr,"%s: %s is an empty file\n",argv[0],target );
			exit(2);
		}
	}

	printf("Target is %s\n", target);
	printf("Ubug > ");
	yyparse();
}


bread( buf, adr, sz )
char	*buf;
int	adr;
int	sz;

{	if( lseek( fd, adr, 0 ) == -1 )
		return(1);
	
	return( read( fd, buf, sz ) != sz );
}

/*
 *	bwrite: write a part of the file
 */

bwrite( buf, adr, sz )
char	*buf;
int	adr;
int	sz;

{	if( ronly )
		return(1);

	if( lseek( fd, adr, 0 ) == -1 )
		return(1);
	
	return( write( fd, buf, sz ) != sz );
}

/*
 *	display: print out part of a file.
 */

display()
{	int	cnt;

	if( siz&0xf )
		siz = (siz&(~0xf))+0x10;

	for( cnt=0; cnt != siz; cnt += 0x10 ) {
		if( dline( addr+cnt ) ) {
			printf("Can't read at %6.6x\n", addr+cnt);
			break;
		}
	}

	addr += cnt;
}

/*
 *	dline: display a line
 */
int
dline( adr )
int	adr;

{	char	buf[0x10];
	static	char	linebuf[100];
	char	*s;
	int	c,i,cnt;

	cnt = 0x10;
	if( bread( buf, adr, cnt ))
	{			/* try partial line */
		if( file_siz > adr && (cnt = file_siz - adr) < 0x10 )
		{
			for( i=0; i<0x10; i++)
				buf[i] = 0;
			if( bread( buf, adr, cnt ))
				return(-1);
		} else
			return(-1);
	}

	s = linebuf;
	s += sprintf(s, "%6.6x    ", adr );
	for( c=0; c<0x10; c += 8 )
	{	for( i=0; i<8; i++ )
		{
			if( c+i < cnt )
				s += sprintf(s, "%2.2x ", buf[c+i]&0xff );
			else
				s += sprintf(s, "   ");
		}
		s += sprintf(s, " ");
	}

	for( c=0; c<cnt; c++ )
		if(( ' ' <= buf[c] ) && ( buf[c] < '\177' ))
			*s++ = buf[c];
		else
			*s++ = '.';
	*s++ = '\n';
	*s++ = 0;

	printf("%s", linebuf);

	return(0);
}

/*
 *	modify: modify the file
 */

modify()

{	char	line[BUFSIZ];
	int	mval;
	union {
		char	bb;
		short	wb;
		long	lb;
	} buf;

	if(( msiz != 1 ) && ( addr&1 ))
		--addr;

	if( smod() )
		return;

	while( fgets( line, BUFSIZ, stdin ) != NULL )
	{	if( strcmp( line, ".\n" ) == 0 )
			break;

		if( strcmp( line, "\n" ) == 0 )
		{	addr += msiz;
			if( smod() )
				break;
			else
				continue;
		}

		if( ! isxdigit( line[0] ) )
			break;

		mval = atox( line );
#ifdef DEBUG
		printf("%s=%lx\n", line, mval );
#endif

		switch( msiz ) {
		case 1:	buf.bb = (char)mval;
			break;
		case 2: buf.wb = (short)mval;
			break;
		case 4: buf.lb = (long)mval;
			break;
		}

		bwrite( &buf, addr, msiz );
		addr += msiz;

		if( strpbrk( line, "." ) || smod() )
			break;
	}
}

/*
 *	smod: set up modification
 */

smod()

{	union {
		char	bb;
		short	wb;
		long	lb;
	} aval;
	int	v;

	if( ckeof( addr ) )
	{
		printf("at EOF\n");
		return( 1 );
	}

	if( ckeof( addr+msiz-1 ) )
	{
		printf("at EOF");
		if( msiz > 1 )
			printf(" try byte");
		if( msiz > 2 )
			printf(" or word");
		printf(" modification\n");
		return( 1 );
	}

#ifdef	DEBUG
	printf("SMOD: %lx at %lx\n", msiz, addr );
#endif

	printf("%8.8x: ", addr );
	bread( &aval, addr, msiz );

	switch( msiz ) {
	case 1:	v = ((int)aval.bb)&0xff;
		printf("%2.2x? ", v );
		break;
	case 2: v = ((int)aval.wb)&0xffff;
		printf("%4.4x? ", v );
		break;
	default:
	case 4:	v = (int)aval.lb;
		printf("%8.8x? ", v );
		break;
	}
	return( 0 );
}


/*
 *	ckeof: check end of file
 */
ckeof( adr )
int adr;
{
	if( file_siz && adr >= file_siz )
		return( 1 );
	return( 0 );
}


yyerror()
{
}

/*
 *	atox: convert hex string
 */

atox( s )
char	*s;

{	int	n;


	for( n=0; *s; s++ )
	{	if(( 'A' <= *s ) && ( *s <= 'F' ))
			*s = (char)((int)*s-0x20);
		if(( *s < '0' ) || ( *s > 'f' )
		|| (( '9' < *s ) && ( *s < 'a' )))
			return(n);

		if( *s <= '9' )
			n = n*16+(int)*s-(int)'0';
		else
			n = n*16+(int)*s-(int)'a'+10;
	}
}
