/*
 
 		Copyright (c) 1986, 1987, 1988 Motorola Inc.
 		All Rights Reserved
 
 		@(#)vmeboot.c	35.4

	boot.c 

	This is the system disk based bootloader.  
	It is passed the drive and controller numbers by the deBUGger

	It returns the boot device drive and controller numbers and
	the inode number of the booted file.

	The Passed in path name is limited to 38 bytes

	It supports the software bad track format used on the VME320
	drives.

*/

#include <sys/types.h>
#include <sys/dir.h>
#include <aouthdr.h>
#include <filehdr.h>
#include <scnhdr.h>
#include <memory.h>

#include <sys/ino.h>
#include <sys/dk.h>

/*
 *	define target MPU board
 */

#ifndef	MVME130
#define	MVME130
#endif


/*
 *	set DEBUG defines
 */

#define	DEBUG		/* general debug printfs */
#undef	DEBUG		/* general debug printfs */

#define	DEBUGR		/* display all disk reads */
#undef	DEBUGR		/* display all disk reads */

#define	DEBUGI		/* display inode and directory searchs */
#undef	DEBUGI		/* display inode and directory searchs */


#define	EXPER		/* experimental version */
#undef EXPER

/*
 *	set NULL
 */

#ifndef	NULL
#define	NULL	0
#endif

/*
 *	BSIZE == logical block size 
 */

#ifndef BSIZE
#define	BSIZE	1024
#endif

/*
	DSKBLK == size of blocks in the bug
 */

#define DSKBLK 256

#define DBPBLK (BSIZE/DSKBLK)	/* Disk Blocks per Block */
#define BPT 8		/* 8 1024 blocks per track on VME320 hard drives */

char	*iplname = {"VME Disk IPL"};
#ifdef	EXPER
char	*iplrel = {"EXPERIMENTAL"};
#else
char	*iplrel = {"35.4"};
#endif

/*
 *	BUG I/O support
 */
#ifdef	MVME130
struct	packet {
	unsigned char	p_clun;		/* cntlr logical unit number	*/
	unsigned char	p_dlun;		/* device logical unit number	*/
	unsigned short	p_status;	/* status word			*/
	unsigned long	p_addr;		/* memory address		*/
	unsigned long	p_blkno;	/* block number			*/
	unsigned short	p_blkcnt;	/* block count to transfer	*/
	unsigned char	p_flags;	/* I/O flags			*/
	unsigned char	p_adrm;		/* address modifier		*/
} packet;
#endif


/*
 *	directory buffer
 */

#define	DIRENT	(BSIZE/sizeof(struct direct))
struct	direct	dbuf[DIRENT];

/* 
 *	VID and badtrack buffer
 */

#define BADTBL 0x300	/* byte offset to bad track table in VID buffer */
#define VBUFSIZ (1024 * 3)

struct volumeid *vid;	/* pointer to VID */
struct badent *btrk;	/* pointer to start of badtrack list */

char vidbuff[VBUFSIZ];	/* 3k buffer for VID and badtrack table */

int	entry;		/* saved entry point from coff header */

/*
 *	iregs contains a copy of the original registers
 */

extern	unsigned iregs[];

#define	DRIVE	0	/* drive number passed in D0 / R2 */
#define	CNTRL	1	/* controller number passed in D1 / R3 */
#define	INODE	2	/* boot file's inode is passed to OS in D2 / R4 */

			/* BUG passes pointer to optional file name */
#ifdef m68k
#define FILENM	13	/* in A5 for 130bug */
#define ENDONM	14	/* pointer to end of name for 130bug in A6 */
#endif
#ifdef m78k
#define FILENM	6	/* and R8 in 78kbug */
#define ENDONM	7	/* pointer to end of name for 78kbug in R9 */
#endif

/*
 *	operating system header information
 *
 *	"os_vec" contains the beginning part of the text/data sections
 *		that immediately follow the "a.out" header information
 *		in the first BSIZE block.
 */

#define OSVEC BSIZE-(sizeof(struct filehdr)+sizeof(struct aouthdr)+sizeof(struct scnhdr))

struct oshdr {
	struct filehdr os_fhdr;
	struct aouthdr os_ahdr;
	struct scnhdr  os_thdr;
	unsigned char  os_vec[OSVEC];
} oshdrblk;

#define MAXDIRSIZ (DIRSIZ*2+10)
/* default path name is copied over if user specifies a path name */
char	namebuf[MAXDIRSIZ] = "stand\0sysV68\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";

char *comp;	/* pointer to path name component */

int osfno;	/* inode of os file */

int halt = 0;	/* halt after load flag cleared initially */
int list = 0;	/* list directory */

main()
{
	printf("%s Version %s\n", iplname, iplrel );

#ifdef	MVME130
	initpkt();		/* initialize packet */
#endif

	getname();		/* copy os path into buff else use default */

	readvid();		/* read disk Volume ID for root offset and 
				bad track table if used */


	osfno = getdir();

	if(list)
		lsdir(osfno);

	loados();		/* load operating system */

	jumpos(entry);
	/* should never reach here */
}

#ifdef	MVME130

/*
 *	initpkt: initialize MVME130 packet
 */

initpkt()
{
	packet.p_clun = iregs[CNTRL];
	packet.p_dlun = iregs[DRIVE];

	packet.p_flags = 0;
	packet.p_adrm = 0;

#ifdef	DEBUG
	printf("%s: controller lun: 0x%x\n", iplname, packet.p_clun );
	printf("%s: device lun: 0x%x\n", iplname, packet.p_dlun );
#endif
}

#endif

readvid()
{
	dread(0, (VBUFSIZ/DSKBLK), (long)vidbuff);/* VID starts at blk zero */

	vid = (struct volumeid *)vidbuff;
	btrk = (struct badent *)(&vidbuff[BADTBL]);
	if(vid->vid_strat != (unsigned char)-1 )
		vid->vid_btl = 0; /* dont do badtrack, done by hardware */
#ifdef	DEBUG
	printf("%s: Bad Track Count: 0x%x\n", iplname, vid->vid_btl);
#endif
}

readblk(blk, buf)
long blk;
long buf;
{
	readnblk(blk, buf, 1);
}

readnblk(blk, buf, blkcnt)
long blk;
long buf;
int blkcnt;
{
	register cnt;
	register unsigned long cblk, svb, strblk;
	register long strbuf = buf;

	cblk = blk + vid->vid_rfs;	/* add in root file system offset */

    strblk = svb = cblk;
    if(vid->vid_btl) {
	register i, j;
	register struct badent *be;
	register unsigned short track; 
	register trkblk; 
	for(cnt = 0, j = 0; j < blkcnt; j++) {
		track = cblk/BPT;		/* convert block to track */
		trkblk = cblk%BPT;	/* get block number on this track */

		for( i = 0, be = btrk; i < vid->vid_btl; i++, be++) {
			if( be->b_bad == track ) {
#ifdef	DEBUG
	printf("%s: Bad Track: 0x%x ,Alternate Track: 0x%x\n", 
	iplname, be->b_bad, be->b_alt);
#endif
				track = be->b_alt;
				/* convert to 1K block number */
				cblk = (track*BPT) + trkblk;
				break;
			}
		}
		if(cblk == svb++) {
			cnt++;
			cblk = svb;
		}
		else {           
	        	if( cnt ) {	/* read the consecutive blocks first */
				dblkread(strblk, cnt, strbuf);
				strbuf += BSIZE * cnt;
				cnt = 0;
	        	}
	    		/* now read the bad block */
			dblkread(cblk, 1, strbuf);
			strbuf += BSIZE;
			strblk = cblk = svb;
			/* start blk and current blk = save b */
		}
	}
    }
    else {
	cnt = blkcnt;
    }

    if ( (buf + (blkcnt * BSIZE)) != strbuf ) { /* read the rest */
	dblkread(strblk, cnt, strbuf);
    }
}

dblkread(blk, cnt, sbuf)	/* Disk block read routine */
{
	dread(blk * DBPBLK, cnt * DBPBLK, sbuf);
}

dread(blk, cnt, sbuf)	/* Disk read routine */
register long blk, cnt, sbuf;
{
#ifdef	DEBUGR
	printf("%s: Block 0x%x, Buffer: 0x%x, Count: 0x%x\n",
	iplname, blk, sbuf, cnt );
#endif
	packet.p_blkno = blk; 
	packet.p_blkcnt = cnt;
	packet.p_addr = sbuf;
	if(dskread(&packet) != cnt) {
	  printf("%s: Disk Read Error, Block 0X%X, Buffer 0X%X, returned \
	   Blkcnt 0X%X, expected Blkcnt 0X%X\n",
		iplname, blk, sbuf, packet.p_blkcnt, cnt );
		gobug();
	}
}

getname()
{	register char *np = namebuf;
	register char *sp = (char *)iregs[FILENM];
	register char *ep = (char *)iregs[ENDONM];

	if( (*sp == '/') && (sp[1] == ';'))
	{
		sp++;	/* skip leading '/' */
		/*
			we know the user has specified something besides 
			the default string, maybe its an 'ls' on the root 
			directory
		*/
		*np = '\0'; /* replace with a null */
	}
	else if (*sp == '/')
		sp++;	/* skip leading '/' */

	while(( sp < ep ) && ( *sp != '\0' ) && 
	( np < namebuf+sizeof(namebuf) )) {
		if(*sp == ';') {
		    if(*(sp+1) == 'h') {
			sp += 2;	/* skip the ';h' */
			halt++;		/* set halt after load flag */
			continue;
		    }
		    if(*(sp+1) == 'l') {
			sp += 2;	/* skip the ';h' */
			list++;		/* set list directory flag */
			continue;
		    }
		
		}
		*np = *sp++;
		if(*np == '/') 
			*np = '\0'; /* replace with a null */
		np++;
	}
	if(np != namebuf) {
	    *np ='\0';
	    if ( np < namebuf+sizeof(namebuf))
		*(np+1) = '\0'; /* put extra null to indicate end of path */
	}
}

/*
 *	getinode : reads passed inode number and fills blkno[]
 *		with block numbers of file.
 *
 *	the inode blocks start 2K after the base of the root file system
 *
 *
 *
 *
 *
 */

#define INOPB (BSIZE/(sizeof(struct dinode)))	/* 16 inodes per 1K block */
#define INOBOFF 2 	/* number of blocks from start of RFS to inodes */
#define BPBLK (BSIZE/sizeof(long))	/* block pointers per block */

	/* these indexes following are after copy from inode to blkno table */
#define SGLIND 10	/* block number index to single indirect block */
#define DBLIND 11	/* block number index to double indirect block */
#define TPLIND 12	/* block number index to triple indirect block */
#define DIOFF (BPBLK+SGLIND)	/* dbl indirect offset starts in blkno table */

struct dinode ibuf[INOPB];	/* inode block buffer */
unsigned long indbuf[BPBLK];	/* buffer for the double indirect block */

/*
 *	Table of 1K block numbers for OS file
 */

long blkno[4096];	/* 4K * 1K leaves enough room for a 4Meg OS file */


getinode(inode)
{
	register int i;
	register unsigned long blk;
	register struct dinode *ip;
	register char *a, *bp;

	blk = ( (i = (inode - 1)) / INOPB) + INOBOFF;

	readblk(blk, ibuf);	/* read block containing disk inode's */

	ip = &ibuf[i % INOPB];

	a = (char *)ip->di_addr;
	bp = (char *)blkno;
		/* copy the single and double block pointers too */
	for( i = 0; i < TPLIND; i++) { 
		*bp++ = 0;	/* copy direct block numbers into blkno[]*/
		*bp++ = *a++;
		*bp++ = *a++;
		*bp++ = *a++;
	}

	/* read the double indirect block first */
	if(blkno[DBLIND]) {
#ifdef DEBUGI
printf("%s: ", iplname );
printf("Double Ptr Blk #%d, Buff 0x%x\n", blkno[11], indbuf );
#endif
		readblk(blkno[DBLIND], indbuf);
		for(i = 0; indbuf[i]; i++) {
#ifdef DEBUGI
printf("%s: ", iplname );
printf("Double Blk #%d, Buff 0x%x\n", indbuf[i], &blkno[DIOFF+(i*BPBLK)] );
#endif
			readblk(indbuf[i], &blkno[DIOFF+(i*BPBLK)]);
		}
		blkno[DIOFF+(i*BPBLK)] = 0; /* null terminate */
	}

	/* read the single indirect block in now */
	if(blkno[SGLIND]) {
		readblk(blkno[SGLIND], &blkno[SGLIND]);
	}
}




/*
 *	lsdir: list directory entrys
 */

lsdir(fno)
{	register struct direct *dp;
	register n, i;
	register int count;
	register char *sp = (char *)iregs[FILENM];
	int fill, size, maxsize;
		
	/* 
		This fixes a problem with trying to boot the default
		kernel after doing an 'ls'. 
		Bug passes the name string ptr in iregs[FILENM]
		This ptr will still contain the original string
		entered for the 'ls'. Nulling the first byte out will
		cause the default kernel to be booted. It isn't a
		problem for the user trying to boot another kernel since
		a new boot string is being entered.	
	*/
	*sp = '\0';

	if(fno)
		getinode(fno);

		size = 0;
		maxsize = 0;

		/* find the largest string  and number of entrys */
		for(i = 0; blkno[i]; i++) {
			readblk(blkno[i], dbuf);
			for( n = 0, dp = dbuf; n < DIRENT; n++, dp++ ) {
				if(!dp->d_ino)
					continue;
				size = strlen(dp->d_name);
				maxsize = (size > maxsize) ? size : maxsize;
			}
		}

#		ifdef DEBUG
		printf("maxsize: %d\n", maxsize);
#		endif

		/* print out a formatted list */
		count = 0;
		for(i = 0; blkno[i]; i++) {
			readblk(blkno[i], dbuf);
			for( n = 0, dp = dbuf; n < DIRENT; n++, dp++ ) {
				if(!dp->d_ino)
					continue;

				size = strlen(dp->d_name);
				fill = (maxsize - size) + 1;

				if((count + size + fill) > 79)
				{
					printf("\n");
					count = 0;
				}

				count += strlen(dp->d_name);
				count += fill; /* sizeof fill */

				if( (strlen(dp->d_name) <= 0) ||
					 (strlen(dp->d_name) >= MAXDIRSIZ)) 
				{
					printf("invalid string: bad directory name\n");
					gobug();
				}
				else
				{
					printf("%s", dp->d_name);
					for (;fill;fill--)
						printf(" ");
				}
				if(count > 70)
				{
					printf("\n");
					count = 0;
				}

			}
		}
	printf("\n");
	gobug();
}
/*
 *	getdir: 
 */

getdir()
{	register struct direct *dp;
	register n, i, fno;

	getinode(2);	/* get the root directory inode */
	fno = 0;
	comp = namebuf;

loop:
	if( *comp == '\0')
	{
		/* 
			if comp is null then the user
			specified the root directory '\'
			'\' gets repleaced by null in getname
		*/
		fno = 2;
	}
	else
	{
		for(i = 0; blkno[i]; i++) {
			readblk(blkno[i], dbuf);
			for( n = 0, dp = dbuf; n < DIRENT; n++, dp++ ) {
				if(!dp->d_ino)
					continue;
				if(!strncmp(dp->d_name, comp, DIRSIZ)) {
					fno = dp->d_ino;
					break;
				}

			}
			if(fno) {
				comp += strlen(comp) +1;
				if(*comp){
					getinode(fno);
					fno = 0;
					goto loop;
				}
				else
					break;
			}
		}
	}
	if(!fno) {
		printf("%s: Can't find file %s!!\n", iplname, comp);
		gobug();
	}
	return(fno);
}

/*
 *	findscn : find specified scn hdr
 */ 

struct scnhdr *
findscn(fp, scn_name)
struct filehdr *fp;
char *scn_name;
{
	register struct scnhdr *sp;
	register i;

	/* get pointer to first hdr */
	sp = (struct scnhdr *)&((struct oshdr *)fp)->os_thdr;
	i = fp->f_nscns;			/* get number of sections */

	while( i--) {
		if(!strncmp(sp->s_name, scn_name, 8)) {
			return(sp);
		}
		sp++;
	}
	return(0);	/* not found */
}

/*
 *	loados: load operating system
 */

loados()
{
	struct oshdr *op = &oshdrblk;
	register struct aouthdr *ap = (struct aouthdr *)&op->os_ahdr;
	struct filehdr *fp = (struct filehdr *)&op->os_fhdr;
	struct scnhdr *tp;
	register long *text;
	register long *data;
	register i, nextblk;

#define MAXBLKS (0xF000 / BSIZE)	/* the 320 can't load more than 64K */
#ifdef DEBUG
	printf("%s: ", iplname );
	printf("loading inode #%d\n", osfno);
#endif
	getinode(osfno);
	iregs[INODE] = osfno;	/* pass os inode to os */

	readblk(blkno[0], op);

	if( fp->f_magic != MC68MAGIC) {	
		printf("%s: operating system bad magic %s\n", iplname, comp);
		gobug();	/* return to bug */
	}
	tp = findscn(fp, ".text");
	if(!tp) {
		printf("%s: No .text section found %s\n", iplname, comp);
		gobug();	/* return to bug */
	}
#ifdef DEBUG
	printf("%s: ", iplname );
	printf("text section header found at 0x%x\n",(long)tp - (long)op);
	printf("%s: ", iplname );
	printf("text section offset 0x%x\n", tp->s_scnptr);
	printf("%s: ", iplname );
	printf("text in 1st buffer at 0x%x\n", (tp->s_scnptr + (long)op));
	printf("%s: ", iplname );
	printf("text_start 0x%x\n",ap->text_start);
	printf("%s: ", iplname );
	printf("tsize  0x%x\n",ap->tsize);
	printf("%s: ", iplname );
	printf("data_start 0x%x\n",ap->data_start);
	printf("%s: ", iplname );
	printf("dsize  0x%x\n",ap->dsize);
	printf("%s: ", iplname );
	printf("entry 0x%x\n",ap->entry);
#endif
	/*
	 *	move first part of vectors already read
	 */

	if(tp->s_scnptr < BSIZE) {
		memcpy(ap->text_start, (tp->s_scnptr + (long)op), 
			(BSIZE - tp->s_scnptr));
		nextblk = 1;
	}
	else
	{
		printf("%s: .text section starts in 2nd blk %s\n", 
		iplname, comp);
		gobug();	/* return to bug */
	}

	/*
	 *	get rest from disk blocks in blkno table
	 */

	{
	    register int *bp;	/* pointer to blkno[] */
	    register int *tbp;	/* temp pointer to blkno[] */
	    register char *osp;	/* pointer to where in the mem os is going */
	    register int blkcnt;
	    bp = &blkno[nextblk];
	    osp = (char *)(ap->text_start + (BSIZE - tp->s_scnptr));
	    while(*bp != 0) {
		for(tbp = bp, blkcnt = 1; blkcnt < MAXBLKS; tbp++, blkcnt++) {
			if(*(tbp+1) != *tbp + 1) {
				break;
			}
		}
		readnblk(*bp, osp, blkcnt);
		bp = tbp + 1;
		osp += BSIZE * blkcnt;
	    }
	}

	/*
	 *	move data in reverse order
	 *		if text_start + tsize != data_start
	 */
	text = (long *)(ap->text_start + ap->tsize);
	if(text < (long *)ap->data_start) {
		text = (long *)((long)text + ap->dsize - sizeof(long));
		data = (long *)(ap->data_start + ap->dsize - sizeof(long));
#ifdef DEBUG
	printf("%s: ", iplname );
	printf("data 0x%x from 0x%x for 0x%x\n", ap->data_start,
	ap->text_start + ap->tsize, ap->dsize);
	printf("%s: source 0x%x, dest: 0x%x, count: 0x%x longs\n", iplname,
		text, data, ap->dsize / sizeof(long));
#endif
		for( i = (ap->dsize / sizeof(long)); i ; i--)
			*data-- = *text--;
	}
	/* save entry point */
	entry = ap->entry;
}
