/*
 *                  Copyright (c) 1988 Ferranti Plc.
 *                      All Rights Reserved
 *
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FERRANTI PLC THE COPYRIGHT 
 *                   OF WHICH VESTS IN FERRANTI PLC.
 *
 *   The copyright notice above does not evidence any actual or intended 
 *                   publication of such source code.
 *
 * This source code and the information contained herein is supplied without
 * liability for errors or ommissions and no part may be copied, re-produced, 
 * used or disclosed except as authorised by the Ferranti organisation by 
 *                contract or by other written permission.
 */

/******************************************************************************

	Rev = 5.1		Date of 'get'   = 11:10:54 1/26/89
				Date of 'delta' = 14:57:46 12/8/88

	File = fifo.h

******************************************************************************/


/******************************************************************************
 *
 * Fifo layout
 * ===========
 *
 * The fifo's used by this module, and all others, have the following
 * physical layout.
 *
 * TAS available:
 * +---------+----+----+----+----+---------+---------+---------+ ...
 * |TAS OK ? |RLCK|RMOV|WLCK|WMOV|    SIZE | RDINDEX | WRINDEX |
 * +---------+----+----+----+----+---------+---------+---------+ ...
 *  0         2    3    4    5    6         8         10 <-- #define indices
 *
 * RMOV and WMOV are used to fix TAS fault on PMMU. A mov.b to the RMOV or WMOV
 * address prior to the access allows the tas to complete ! I have no idea why.
 *
 * 2 flag scheme: [TAS OK set to 0xb0b4 = '~OK(0x4f4b)]
 * +---------+----+----+----+----+---------+---------+---------+ ...
 * |         |READ flgs|WRITE flg|    SIZE | RDINDEX | WRINDEX |
 * +---------+----+----+----+----+---------+---------+---------+ ...
 *  0         2    3    4    5    6         8         10 <-- #define indices
 *            2 = SIORDFLG (locally set by sio card for fifo read)
 *                 3 = REMRDFLG (remote, not sio, set for read)
 *                      4 = SIOWRFLG (locally set by sio card for fifo write)
 *                           5 = REMWRFLG (remote, not sio, set for write)
 *
 * ... +-------------------+-------------------+---------+---------+ ...
 *     |           RLCKCNT |           WLCKCNT | LOWMARK |HIGHMARK |
 * ... +-------------------+-------------------+---------+---------+ ...
 *      12                  16                  20        22
 *
 * ... +---------+----+ .. +----+----+
 *     |SEMAPHORE|DATA| .. |    |    |
 * ... +---------+----+ .. +----+----+
 *      24        26                     <-- #define indices
 *
 * The RLCK/WLCK bytes are used if TAS (RMW) is supported by both processors.
 * If not a two byte flags are used. This scheme is never used for fifo's that
 * are manipulated ONLY locally. However if the remote processor can't handle
 * TAS (eg. any system using current PMMU) them for speed a 2 flag semaphore
 * can be used. [Note that with a modified kernel the TAS does appear to work
 * with the PMMU however the overhead is a hundred plus instructions per TAS !]
 * The two flags should be #defined locally to have meaning...
 *
 * To allocate data space for a fifo use one of the MK?FIFO macros. They 
 * allocate enough space for a fifo with the specified number of elements of 
 * the specified size.  The '?' can be any of the following :-
 * B = byte  8 bits
 * W = word 16 bits
 * L = long 32 bits
 *
 * Example: MKBFIFO(ff, 128) - defines a byte fifo 'ff' for upto a 128 bytes
 *          MKWFIFO(f2, 512) - defines a word fifo 'f2' for upto 512 bytes
 *
 */

#ifndef	ASSEMBLY

#ident "@(#)scc-common:include/local/scc/fifo.h Rev 5.1 14:57:46 12/8/88"

struct fifo
{



	unsigned short	f_tasok;	/* Is TAS available, yes if != b0b4 */

	unsigned char 	ff_rdlock;	/* read 'tas' flag -OR-
					   read sio semaphore flag  */
	unsigned char 	f_remrdflg;	/* read semaphore remote flag       */

	unsigned char 	ff_wrlock;	/* write 'tas' flag -OR-
					   write sio semaphore flag */
	unsigned char 	f_remwrflg;	/* write semaphore remote           */

	short		f_size;		/* size (max number of elements)    */
	short		f_rdindex;	/* read index                       */
	short		f_wrindex;	/* write index                      */
	unsigned long 	f_rdlcnt;	/* number of read flag 'tas' fails  */
	unsigned long 	f_wrlcnt;	/* number of write flag 'tas' fails */
	short		f_lowmark;	/* fifo low water mark		    */
	short		f_highmark;	/* fifo high water mark		    */
	short		f_semaphore;	/* Firmware lock/unlock use only    */
};

#define	f_rdlock	ff_rdlock
#define	f_wrlock	ff_wrlock

#define	HIGHWATER(fp)	(fp->f_highmark)
#define LOWWATER(fp)	(fp->f_lowmark)

#define	f_siordflg	ff_rdlock
#define	f_siowrflg	ff_wrlock

#define	FFEMPTY(fp)	((fp)->f_rdindex==(fp)->f_wrindex)

#define	FFSIZE(fp)	((fp)->f_size-1)

#define	FFINIT(fp,size)	{(fp)->f_size=(size)+1;\
			(fp)->f_tasok=0;\
			(fp)->f_remwrflg=0;\
			(fp)->f_remrdflg=0;\
			(fp)->f_wrlock=0;\
			(fp)->f_rdlock=0;\
			(fp)->f_wrlcnt=0;\
			(fp)->f_rdlcnt=0;\
			(fp)->f_wrindex=0;\
			(fp)->f_rdindex=0;\
			(fp)->f_semaphore=0;}

#define MKBFIFO(name,size) struct { struct fifo head; char fifo[size+2]; } name
/* +2 for byte to word conversions */
#define MKWFIFO(name,size) struct { struct fifo head; short fifo[size+1]; } name
#define MKLFIFO(name,size) struct { struct fifo head; long fifo[size+1]; } name

/*
 * structure MUST be 256 bytes
 */
struct fifo_sizes {
	short fs_rawin, fs_rawout, fs_rawpout;
	short fs_hostout, fs_hostnoncin;
	short fs_ifc_no;
	short fs_ifcs[118];

	/*
	 * The following data values are used by ioctl(2), SCCL_FIFO
	 */
	unsigned long mem_grab;
	unsigned long card_index;
};

#define	OFF_MEM_GRAB	248
#define	OFF_CARD_INDEX	252


#else

ident "@(#)scc-common:include/local/scc/fifo.h Rev 5.1 14:57:46 12/8/88"

#define	RLCK	2
#define RMOV	3
#define	WLCK	4
#define WMOV	5
#define	SIZE	6
#define	RDINDEX	8
#define	WRINDEX	10
#define	RLCKCNT	12
#define	WLCKCNT	16
#define	LOWMRK	20
#define	HIGHMRK	22
#define	SEMA	24
#define	DATA	26

#endif

#define	MAXFIFO		32766		/* max +ve signed short - 1. Must sub 1
					 * as f_size is size supplied in FFINIT
					 * call plus 1
					 */


