#ident	"@(#)install:sadmin/etc/slfsed.c	35.7"

/* Installation Disk Slice/File System Editor
 * Copyright 1988, Motorola, Inc.
 */

#define USAGE "\
Usage: %s [ -D dskdefs_dir ] [ -n name ]    ddefsType\n\
       %s [ -D dskdefs_dir ] [ -n name ] -r ddefsType\n\
       %s [ -D dskdefs_dir ] [ -n name ] -u ddefsType\n\
       %s                                -d specialFile\n\
       %s                                -s specialFile\n\
       %s                                -w specialFile\n\
       %s [ -o outFile ] [ -p sliceNum ] -e sliceFile\n\
\n\
where:\n\
	no switch	output initial slice table for ddefsType\n\
	-r		output root/swap slice table for ddefsType\n\
	-u		output root/swap/usr slice table for ddefsType\n\
		-D		specify dskdefs directory (/etc/dskdefs)\n\
		-n		specify special file name for display\n\
	-d		output swap table for device\n\
	-s		check device's accessibility/swap table existence\n\
	-w		write standard input's swap table to device\n\
	-e		edit standard input's slice table\n\
		-o		output file slice table to outFile\n\
		-p		protect file system name/volume id\n\
\n"


#define OPTSTR "D:rudn:o:wsep:"

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/filsys.h>
#include <sys/stat.h>
#include <ustat.h>
#include <fcntl.h>
#include <sys/ino.h>
#include <sys/dk.h>


#define MinSliceCount 8
#define MaxSliceCount 16


/* Ddefs(1M) data base directory */
#define DSKDEFS "/etc/dskdefs"


/* Constant String Compare Function
 *	C = Constant String;
 *	S = Comparison String Pointer;
 */
extern int strncmp ();
#define STRCONCMP(C,S) strncmp( C, S, sizeof ( C ) - 1 )

/* Constant Strings */
char * MOTOROLA  =  "MOTOROLA";
char * EXORMACS  =  "EXORMACS";


/* File System Block Size
 *
 */
#if (vax || u3b || u3b5 || m68k) && (FsTYPE == 3)

#	define FSBSIZE ( BSIZE * 2 )

#else

#	define FSBSIZE BSIZE

#endif


/* Number Of Inodes Per Block */
#define	NBINODE	( FSBSIZE / sizeof ( struct dinode ) )


/* Utility's Name Pointer */
static char *
myName = "slfsed";

/* Usage error flag */
static int
usageError  =  0;

/* Dskdefs directory for ddefs(1M) utility */
static char *
dskdefs  =  DSKDEFS;

/* Device name */
static char *
deviceName  =  NULL;

/* Output file name */
static char *
outFile  =  NULL;

/* Protect File System Name/Volume ID */
static int
protSlice [ MaxSliceCount ];

static int
protCount  =  0;

/* Internal Slice Description Structure
 *
 */
struct
sliceInfo {
	unsigned int s_flag;	/* readonly flag */
	int s_sliceno;		/* slice number */
	struct slice s_tab;	/* slice table entry */
	struct filsys s_fs;	/* file system superblock */
};

static struct sliceInfo
sliceEg;

#define FsNameSize sizeof ( sliceEg . s_fs . s_fname )
#define FsVolIdSize sizeof ( sliceEg . s_fs . s_fpack )


/* Structure SliceInfo's S_flag Field Masks
 *
 */
#define RDONLY 		0x1		/* readonly */
#define MODIFIED	0x2		/* modified */
#define FSMOD		0x4		/* file system modified */
#define NMMOD		0x8		/* file system name modified */
#define NMPROT		0x10		/* file system name/id protected */


/* Slice Table Structure
 *
 */
struct
sliceTable {
	char * st_spFile;	/* special file pathlist */
	int st_noSlices;	/* number of slices on device */
	int st_noOfK;		/* number of K bytes per block */
	struct sliceInfo * st_slInfo;
				/* slice information list */
};


/* Ddefs(1M) Base Information structure
 *
 */
struct
ddefs {
	char * dd_name;
	struct volumeid dd_vid;
	struct dkconfig dd_ios;
	unsigned * dd_alist;
	char * dd_form;
	char * dd_com;
	char * dd_rdman;
};


/* Main Entry Point
 *
 */
int
main( argc, argv, envp )
int argc;
char * argv [], * envp [];
{
	extern int optind, getopt ();
	extern char * optarg, * getenv ();
	extern long strtol ();

	static int outUsage (), findSliceTable (), writeSliceTable (),
		outInitialSliceTable (), outRootSliceTable (),
		outRootUsrSliceTable (), outDeviceSliceTable (),
		editSliceFile ();

	int c, ( * function ) (), entryNumber;
	char * q;

	myName  =  argv [ 0 ];
	function  =  outInitialSliceTable;

	if ( ( dskdefs  =  getenv ( "DSKDEFS" ) )  ==  NULL )
		dskdefs  =  DSKDEFS;

	while ( ( c = getopt ( argc, argv, OPTSTR ) ) != EOF )
		switch ( c ) {
		case 'd':
			if ( function  ==  outInitialSliceTable ) {
				function  =  outDeviceSliceTable;
				break;
			}
			/* fall through */
		case 'r':
			if ( function  ==  outInitialSliceTable ) {
				function  =  outRootSliceTable;
				break;
			}
			/* fall through */
		case 'u':
			if ( function  ==  outInitialSliceTable ) {
				function  =  outRootUsrSliceTable;
				break;
			}
			/* fall through */
		case 's':
			if ( function  ==  outInitialSliceTable ) {
				function  =  findSliceTable;
				break;
			}
			/* fall through */
		case 'w':
			if ( function  ==  outInitialSliceTable ) {
				function  =  writeSliceTable;
				break;
			}
			/* fall through */
		case 'e':
			if ( function  ==  outInitialSliceTable ) {
				function  =  editSliceFile;
				break;
			}
			/* fall through */
		case '?':
		default:
			function  =  outUsage;
			break;
		case 'D':
			dskdefs  =  optarg;
			break;
		case 'n':
			deviceName  =  optarg;
			break;
		case 'o':
			outFile  =  optarg;
			break;
		case 'p':
			entryNumber  =  ( int ) strtol ( optarg, & q, 0 );

			if ( ( ( ! entryNumber ) && ( optarg  ==  q ) ) ||
			     ( entryNumber  <  0 ) ||
			     ( entryNumber  > MaxSliceCount - 1 ) ) {

				fprintf ( stderr,
					"%s: is an invalid slice number.\n",
					myName, optarg );

				function  =  outUsage;
				break;
			}

			protSlice [ protCount ++ ]  =  entryNumber;
			break;
		}

	return ( ( * function ) ( argc - optind, & argv [ optind ] ) );
}


/* Output Usage Message
 *
 */
static int
outUsage ( argc, argv )
int argc;
char * argv [];
{
	fprintf( stderr, USAGE,
		myName, myName, myName, myName, myName, myName, myName );

	-- usageError;
	return ( 2 );
}


/* Read And Edit Standard Input Slice Table
 *
 */
static int
editSliceFile ( argc, argv )
int argc;
char * argv [];
{
	extern FILE * fopen ();
	extern void perror ();
	extern int fclose ();

	static int extraArguments (), outUsage (), printSliceTable ();
	static struct sliceTable * readSlicingInformation (),
					* editSliceTable ();
	static void freeSliceTable ();

	struct sliceTable * sliceTable, * newSliceTable;
	struct sliceInfo * sliceList;
	FILE * sliceStream, * outStream;
	char * sliceFile;
	int protEntry;

	if ( argc < 1 ) {

		fprintf ( stderr, "%s: sliceFile argument not found.\n",
				myName );
		return ( outUsage ( argc, argv ) );

	} else if ( extraArguments ( argc - 1, & argv [ 1 ] ) )
		return ( outUsage ( argc, argv ) );

	sliceFile  =  argv [ 0 ];

	sliceStream  =  fopen ( sliceFile, "r" );
	if ( sliceStream  ==  NULL ) {
		perror ( sliceFile );
		return ( 1 );
	}
	
	sliceTable  =  readSlicingInformation ( sliceStream );

	if ( fclose ( sliceStream ) ) {
		perror ( sliceFile );
		return ( 1 );
	}

	if ( sliceTable  ==  NULL )
		return ( 1 );

	sliceList  =  sliceTable -> st_slInfo;

	for ( protEntry = 0; protEntry < protCount; protEntry ++ ) {

		if ( protSlice [ protEntry ]  >=  sliceTable -> st_noSlices ) {

			fprintf ( stderr, "%d: protected slice out of range.\n",
					protSlice [ protEntry ] );

			freeSliceTable ( sliceTable );
			return ( 1 );
		}

		sliceList [ protSlice [ protEntry ]  ] . s_flag  |=  NMPROT;
	}

	newSliceTable  =  editSliceTable ( sliceTable );
	if ( newSliceTable  !=  NULL ) {
		freeSliceTable ( sliceTable );
		sliceTable  =  newSliceTable;
	}

	if ( outFile  !=  NULL ) {
		
		outStream  =  ( ! strcmp ( outFile, "-" ) )  ?
				stdout  :  fopen ( outFile, "w" );

		if ( outStream  ==  NULL ) {
			perror ( outFile );
			freeSliceTable ( sliceTable );
			return ( 1 );
		}

		printSliceTable ( outStream, sliceTable );

		if ( fclose ( outStream )  ==  EOF ) {
			perror ( outFile );
			return ( 1 );
		}
	}

	freeSliceTable ( sliceTable );
	return ( 0 );
}


/* Find Slice Table On Disk Drive ( -s )
 *
 */
static int
findSliceTable ( argc, argv )
int argc;
char * argv [];
{
	extern int fprintf (), stat (), fread (), fclose ();
	extern FILE * fopen ();
	extern void perror ();

	static int outUsage (), extraArguments ();

	char * specialFile;
	FILE * devStream;
	struct stat statbuf;
	struct dkblk0 dkblk0;

	if ( argc < 1 ) {

		fprintf ( stderr, "%s: specialFile argument not found.\n",
				myName );
		return ( outUsage ( argc, argv ) );

	} else if ( extraArguments ( argc - 1, & argv [ 1 ] ) )
		return ( outUsage ( argc, argv ) );

	specialFile  =  argv [ 0 ];

	if ( stat ( specialFile, & statbuf )  <  0 ) {
		perror ( specialFile );
		return ( 1 );
	}

	if ( ( statbuf . st_mode & S_IFMT )  !=  S_IFCHR ) {
		fprintf ( stderr,
			"%s: '%s' is not a character special file.\n", 
			myName, specialFile );
		return ( 1 );
	}

	devStream  =  fopen ( specialFile, "r" );
	if ( devStream  ==  NULL ) {
		perror ( specialFile );
		return ( 3 );
	}

	if ( fread ( ( char * ) & dkblk0, sizeof ( struct dkblk0 ),
			1, devStream )  !=  1 ) {
		if ( ferror ( devStream ) )
			perror ( specialFile );
		else
			fprintf ( stderr, "%s: disk vol id read failed.\n",
					myName );

		fclose ( devStream );
		return ( 3 );
	}

	if( STRCONCMP( MOTOROLA, dkblk0 . dk_vid . vid_mac ) &&
	    STRCONCMP( EXORMACS, dkblk0 . dk_vid . vid_mac ) ) {

		fprintf ( stderr, "%s: Not a valid volume ID.\n", specialFile );
		fclose ( devStream );
		return ( 3 );
	}

	if ( ! dkblk0 . dk_ios . ios_slccnt ) {

		fprintf ( stderr, "%s: No slice table found\n", specialFile );
		fclose ( devStream );
		return ( 4 );
	}

	return ( 0 );
}


/* Write Slice Table To Disk Drive ( -w )
 *
 */
static int
writeSliceTable ( argc, argv )
int argc;
char * argv [];
{
	extern int fprintf (), stat (), ioctl (),
			fseek (), fread (), fclose ();
	extern FILE * fopen ();
	extern char * calloc ();
	extern void perror (), free ();

	static int outUsage (), extraArguments ();
	static struct sliceTable * readSlicingInformation ();
	static void freeSliceTable ();

	int entryNumber;
	char * specialFile;
	FILE * devStream;
	struct slice * slice;
	struct stat statbuf;
	struct dkblk0 dkblk0;
	struct sliceTable * sliceTable;
	struct sliceInfo * sliceList;

	if ( argc < 1 ) {

		fprintf ( stderr, "%s: specialFile argument not found.\n",
				myName );
		return ( outUsage ( argc, argv ) );

	} else if ( extraArguments ( argc - 1, & argv [ 1 ] ) )
		return ( outUsage ( argc, argv ) );

	specialFile  =  argv [ 0 ];

	if ( stat ( specialFile, & statbuf )  <  0 ) {
		perror ( specialFile );
		return ( 1 );
	}

	if ( ( statbuf . st_mode & S_IFMT )  !=  S_IFCHR ) {
		fprintf ( stderr,
			"%s: '%s' is not a character special file.\n", 
			myName, specialFile );
		return ( 1 );
	}

	devStream  =  fopen ( specialFile, "r+" );
	if ( devStream  ==  NULL ) {
		perror ( specialFile );
		return ( 1 );
	}

	if ( fread ( ( char * ) & dkblk0, sizeof ( struct dkblk0 ),
			1, devStream )  !=  1 ) {
		if ( ferror ( devStream ) )
			perror ( specialFile );
		else
			fprintf ( stderr, "%s: disk vol id read failed.\n",
					myName );

		fclose ( devStream );
		return ( 1 );
	}

	if( STRCONCMP( MOTOROLA, dkblk0 . dk_vid . vid_mac ) &&
	    STRCONCMP( EXORMACS, dkblk0 . dk_vid . vid_mac ) ) {

		fprintf ( stderr, "%s: Not a valid volume ID.\n", specialFile );
		fclose ( devStream );
		return ( 1 );
	}

	if ( ! dkblk0 . dk_ios . ios_slccnt ) {

		fprintf ( stderr, "%s: No slice table found\n", specialFile );
		fclose ( devStream );
		return ( 1 );
	}

	if ( fseek ( devStream, ( long )
			( dkblk0 . dk_ios . ios_slcblk * BSIZE ), 0 ) ) {
		perror ( specialFile );
		fclose ( devStream );
		return ( 1 );
	}

	sliceTable  =  readSlicingInformation ( stdin );
	if ( sliceTable  ==  NULL ) {
		fclose ( devStream );
		return ( 1 );
	}

	slice  =  ( struct slice * ) calloc ( BSIZE, sizeof ( char ) );

	if ( slice  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( sliceTable );
		fclose ( devStream );
		return ( 1 );
	}

	sliceList  =  sliceTable -> st_slInfo;

	for ( entryNumber  =  0;
	      entryNumber  <  sliceTable -> st_noSlices;
	      entryNumber ++ ) {
		slice [ entryNumber ] . nblocks  =
			sliceList [ entryNumber ] . s_tab . nblocks;
		slice [ entryNumber ] . blkoff  =
			sliceList [ entryNumber ] . s_tab . blkoff;
	}

	freeSliceTable ( sliceTable );

	/* set driver slice table first */

	if ( ioctl ( fileno ( devStream ), DKSETSLC, slice )  ==  -1 ) {

		fprintf ( stderr, "%s: slice table ioctl error.\n", myName );
		free ( ( char * )slice );
		fclose ( devStream );
		return ( 1 );
	}

	if ( fwrite ( ( char * ) slice,
			sizeof ( char ), BSIZE, devStream ) != BSIZE ) {

		perror ( specialFile );
		free ( ( char * ) slice );
		fclose ( devStream );
		return ( 1 );
	}

	free ( ( char * ) slice );
	fclose ( devStream );
	return ( 0 );
}


/* Generate And Output Initial Slice Table ( no switch )
 *
 */
static int
outInitialSliceTable ( argc, argv )
int argc;
char * argv [];
{
	static struct sliceTable * initialSliceTable ();
	static int outSliceTable ();

	return ( outSliceTable ( argc, argv, initialSliceTable ) );
}


/* Generate And Output Root/Swap Slice Table ( -r )
 *
 */
static int
outRootSliceTable ( argc, argv )
int argc;
char * argv [];
{
	static struct sliceTable * rootSliceTable ();
	static int outSliceTable ();

	return ( outSliceTable ( argc, argv, rootSliceTable ) );
}


/* Generate And Output Root/Swap/Usr Slice Table ( -u )
 *
 */
static int
outRootUsrSliceTable ( argc, argv )
int argc;
char * argv [];
{
	static struct sliceTable * rootUsrSliceTable ();
	static int outSliceTable ();

	return ( outSliceTable ( argc, argv, rootUsrSliceTable ) );
}


/* Generate And Output Device's Slice Table ( -d )
 *
 */
static int
outDeviceSliceTable ( argc, argv )
int argc;
char * argv [];
{
	static struct sliceTable * deviceSliceTable ();
	static int outSliceTable ();

	return ( outSliceTable ( argc, argv, deviceSliceTable ) );
}


/* Generate And Output A Slice Table ( no switch, -r, -u, -d )
 *
 */
static int
outSliceTable ( argc, argv, generateSliceTable )
int argc;
char * argv [];
struct sliceTable * ( * generateSliceTable ) ();
{
	extern void perror ();
	extern int fclose ();

	static int printSliceTable ();
	static void freeSliceTable ();

	struct sliceTable * sliceTable;
	FILE * outStream;

	sliceTable  =  ( * generateSliceTable ) ( argc, argv );
	if ( sliceTable  ==  NULL )
		return ( ( usageError )  ?  2  :  1 );

	printSliceTable ( stdout, sliceTable );

	freeSliceTable ( sliceTable );
	return ( 0 );
}


/* Read Device Slice Table ( -d )
 *
 */
static struct sliceTable *
deviceSliceTable ( argc, argv )
int argc;
char * argv [];
{
	extern void perror ();
	extern int strncmp (), fread (), fseek ();
	extern char * calloc (), * strcpy (), * memset ();
	extern FILE * fopen ();

	static int outUsage (), extraArguments ();
	static void freeSliceTable ();

	int entryNumber;
	struct sliceTable * sliceTable;
	struct sliceInfo * sliceList;
	struct slice * slice;
	struct dkblk0 dkblk0;
	FILE * devStream;
	char * specialFile;

	if ( argc < 1 ) {

		fprintf ( stderr, "%s: specialFile argument not found.\n",
				myName );
		outUsage ( argc, argv );
		return ( NULL );

	} else if ( extraArguments ( argc - 1, & argv [ 1 ] ) )
		return ( NULL );

	specialFile  =  argv [ 0 ];

	devStream  =  fopen ( specialFile, "r" );
	if ( devStream  ==  NULL ) {
		perror ( specialFile );
		return ( NULL );
	}

	if ( fread ( ( char * ) & dkblk0, sizeof ( struct dkblk0 ),
			1, devStream )  !=  1 ) {
		perror ( specialFile );
		fclose ( devStream );
		return ( NULL );
	}

	if( STRCONCMP( MOTOROLA, dkblk0 . dk_vid . vid_mac ) &&
	    STRCONCMP( EXORMACS, dkblk0 . dk_vid . vid_mac ) ) {

		fprintf ( stderr, "%s: Not a valid volume ID.\n", specialFile );
		fclose ( devStream );
		return ( NULL );
	}

	if ( ! dkblk0 . dk_ios . ios_slccnt ) {

		fprintf ( stderr, "%s: No slice table found\n", specialFile );
		fclose ( devStream );
		return ( NULL );
	}

	if ( fseek ( devStream,
			(long) ( dkblk0 . dk_ios . ios_slcblk * BSIZE ), 0 ) ) {
		perror ( specialFile );
		fclose ( devStream );
		return ( NULL );
	}

	slice  =  ( struct slice * ) calloc ( dkblk0 . dk_ios . ios_slccnt,
						sizeof ( struct slice ) );

	if ( slice  ==  NULL ) {
		perror ( myName );
		fclose ( devStream );
		return ( NULL );
	}

	if ( fread ( ( char * ) slice, sizeof ( struct slice ),
			dkblk0 . dk_ios . ios_slccnt, devStream )  !=
			dkblk0 . dk_ios . ios_slccnt ) {
		perror ( specialFile );
		free ( ( char * ) slice );
		fclose ( devStream );
		return ( NULL );
	}

	sliceTable  =  ( struct sliceTable * )
				calloc ( 1, sizeof ( struct sliceTable ) );

	if ( sliceTable  ==  NULL ) {
		perror ( myName );
		free ( ( char * ) slice );
		fclose ( devStream );
		return ( NULL );
	}

	sliceTable -> st_spFile    =  NULL;
	sliceTable -> st_slInfo    =  NULL;
	sliceTable -> st_noOfK     =  BSIZE;
	sliceTable -> st_noSlices  =  dkblk0 . dk_ios . ios_slccnt;

	sliceTable -> st_spFile  =
		calloc ( strlen ( specialFile ) + sizeof ( char ),
				sizeof ( char ) );

	if ( sliceTable -> st_spFile  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( sliceTable );
		free ( ( char * ) slice );
		fclose ( devStream );
		return ( NULL );
	}

	strcpy ( sliceTable -> st_spFile, specialFile );

	sliceList  =  ( struct sliceInfo * ) calloc ( sliceTable -> st_noSlices,
					sizeof ( struct sliceInfo ) );

	if ( sliceList  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( sliceTable );
		free ( ( char * ) slice );
		fclose ( devStream );
		return ( NULL );
	}

	sliceTable -> st_slInfo  =  sliceList;

	for ( entryNumber = 0;
	      entryNumber < sliceTable -> st_noSlices;
	      entryNumber ++ ) {
		sliceList [ entryNumber ] . s_sliceno        =  entryNumber;
		sliceList [ entryNumber ] . s_flag           =  0;

		sliceList [ entryNumber ] . s_tab . blkoff   =
					slice [ entryNumber ] . blkoff;
		sliceList [ entryNumber ] . s_tab . nblocks  =
					slice [ entryNumber ] . nblocks;

		if ( ( entryNumber  !=  sliceTable -> st_noSlices - 1 ) &&
		     ( slice [ entryNumber ] . nblocks  !=  0L ) ) {

			if ( ( fseek ( devStream,
					( slice [ entryNumber ] .  blkoff + 1 )
						* BSIZE, 0 ) ) ||
			     ( fread ( & sliceList [ entryNumber ] .  s_fs,
					sizeof ( struct filsys ), 1, devStream )
						!=  1 ) ) {
				perror ( specialFile );
				freeSliceTable ( sliceTable );
				free ( ( char * ) slice );
				fclose ( devStream );
				return ( NULL );
			}

			if ( sliceList [ entryNumber ] . s_fs . s_magic  ==
					FsMAGIC ) {

				/* found a file system */
				/* convert block count to 512-byte blocks */

				sliceList [ entryNumber ] . s_fs . s_fsize  *=
					sliceList [ entryNumber ] .
						s_fs . s_type;
				continue;
			}
		}
		/* did not find a file system */

		sliceList [ entryNumber ] . s_fs . s_fsize   =  0L;

		memset ( sliceList [ entryNumber ] . s_fs . s_fname,
				'\0', FsNameSize );
		memset ( sliceList [ entryNumber ] . s_fs . s_fpack,
				'\0', FsVolIdSize );
	}

	sliceList [ sliceTable -> st_noSlices - 1 ] . s_flag  =  RDONLY;

	free ( ( char * ) slice );
	fclose ( devStream );

	return ( sliceTable );
}


/* Create Initial Slice Table For Command Line Ddefs Type ( no switch )
 *
 */
static struct sliceTable *
initialSliceTable ( argc, argv )
int argc;
char * argv [];
{
	static struct sliceTable * genInitialSliceTable ();
	static struct ddefs * getDdefsData ();
	static void freeDdefsData ();

	struct sliceTable * sliceTable;
	struct ddefs * ddefsData;

	ddefsData  =  getDdefsData ( argc, argv );
	if ( ddefsData  ==  NULL )
		return ( NULL );

	sliceTable  =  genInitialSliceTable ( ddefsData );
	freeDdefsData ( ddefsData );
	return ( sliceTable );
}


/* Create Root/Swap Slice Table For Command Line Ddefs Type ( -r )
 *
 */
static struct sliceTable *
rootSliceTable ( argc, argv )
int argc;
char * argv [];
{
	static struct sliceTable * genRootSliceTable ();
	static struct ddefs * getDdefsData ();
	static void freeDdefsData ();

	struct ddefs * ddefsData;
	struct sliceTable * sliceTable;

	ddefsData  =  getDdefsData ( argc, argv );
	if ( ddefsData  ==  NULL )
		return ( NULL );

	sliceTable  =  genRootSliceTable ( ddefsData );
	freeDdefsData ( ddefsData );
	return ( sliceTable );
}


/* Create Root/Swap/Usr Slice Table For Command Line Ddefs Type ( -u )
 *
 */
static struct sliceTable *
rootUsrSliceTable ( argc, argv )
int argc;
char * argv [];
{
	static struct sliceTable * genRootUsrSliceTable ();
	static struct ddefs * getDdefsData ();
	static void freeDdefsData ();

	struct sliceTable * sliceTable;
	struct ddefs * ddefsData;

	ddefsData  =  getDdefsData ( argc, argv );
	if ( ddefsData  ==  NULL )
		return ( NULL );

	sliceTable  =  genRootUsrSliceTable ( ddefsData );
	freeDdefsData ( ddefsData );
	return ( sliceTable );
}


/* Obtain Command Line DdefsType Information ( no switch, -r, -u )
 *
 */
static struct ddefs *
getDdefsData ( argc, argv )
int argc;
char * argv [];
{
	static struct ddefs * readDdefsData ();
	static int outUsage (), extraArguments ();

	char * ddefsType;

	if ( argc < 1 ) {
		fprintf ( stderr, "%s: ddefsType argument not found.\n",
				myName );
		outUsage ( argc, argv );
		return ( NULL );
	} else if ( extraArguments ( argc - 1, & argv [ 1 ] ) ) {
		outUsage ( argc, argv );
		return ( NULL );
	}

	ddefsType  =  argv [ 0 ];
	return ( readDdefsData ( dskdefs, ddefsType ) );
}


/* Check For Extraneous Arguments On The Command Line
 *
 */
static int
extraArguments ( argc, argv )
int argc;
char * argv [];
{
	if ( argc > 1 ) {

		fprintf ( stderr, "%s: Extra arguments detected.\n", myName );
		outUsage ( argc, argv );
		return ( -1 );
	}
	return ( 0 );
}


/* Generate Initial Slice table
 *
 */
static struct sliceTable *
genInitialSliceTable ( ddefsData )
struct ddefs * ddefsData;
{
	extern void perror (), free ();
	extern char * calloc (), * strcpy (), * memset ();
	extern int strlen ();
	static void freeSliceTable ();

	struct sliceTable * sliceTable;
	struct sliceInfo * sliceList;
	long reserved, nblocks;
	int entryNumber;
	char * theName;


	sliceTable  =  ( struct sliceTable * )
				calloc ( 1, sizeof ( struct sliceTable ) );
	if ( sliceTable  ==  NULL ) {
		perror ( myName );
		return ( NULL );
	}

	sliceTable -> st_spFile    =  NULL;
	sliceTable -> st_slInfo    =  NULL;
	sliceTable -> st_noOfK     =  BSIZE;
	sliceTable -> st_noSlices  =  ddefsData -> dd_ios . ios_slccnt;

	if ( ( theName  =  deviceName )  ==  NULL )
			theName  =  ddefsData -> dd_name;

	sliceTable -> st_spFile  = calloc ( strlen ( theName ) +
					sizeof ( char ), sizeof ( char ) );

	if ( sliceTable -> st_spFile  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	strcpy ( sliceTable -> st_spFile, theName );

	sliceList  =  ( struct sliceInfo * )
				calloc ( sliceTable -> st_noSlices,
						sizeof ( struct sliceInfo ) );
	if ( sliceList  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	sliceTable -> st_slInfo  =  sliceList;

	for ( entryNumber = 0;
	      entryNumber < sliceTable -> st_noSlices;
	      entryNumber ++ ) {
		sliceList [ entryNumber ] . s_sliceno        =  entryNumber;
		sliceList [ entryNumber ] . s_flag	     =  0;
		sliceList [ entryNumber ] . s_tab . blkoff   =  0L;
		sliceList [ entryNumber ] . s_tab . nblocks  =  0L;
		sliceList [ entryNumber ] . s_fs . s_fsize   =  0L;

		memset ( sliceList [ entryNumber ] . s_fs . s_fname,
				'\0', FsNameSize );
		memset ( sliceList [ entryNumber ] . s_fs . s_fpack,
				'\0', FsVolIdSize );
	}

	sliceList [ sliceTable -> st_noSlices - 1 ] . s_flag  =  RDONLY;

	nblocks  =  ( ddefsData -> dd_ios . ios_spt *
		      ddefsData -> dd_ios . ios_trk *
		      ddefsData -> dd_ios . ios_hds ) /
			( BSIZE / ddefsData -> dd_ios . ios_psm );

#ifdef m88k
	reserved = ( int ) ddefsData -> dd_vid . vid_u_reserved;
	reserved = ( reserved << 16 ) | ddefsData -> dd_vid . vid_l_reserved;
#endif

#ifdef m68k
	reserved = ddefsData -> dd_vid . vid_reserved;
#endif

	sliceList [ 0 ] . s_tab . blkoff  =
		ddefsData -> dd_vid . vid_rfs * ( 1024 / BSIZE );

	sliceList [ 0 ] . s_tab . nblocks  = nblocks - reserved -
		sliceTable -> st_slInfo [ 0 ] . s_tab . blkoff;

	sliceList [ ddefsData -> dd_ios . ios_slccnt - 1 ] .  s_tab . nblocks  =
		nblocks;

	return ( sliceTable );
}


/* Generate Root/Swap Slice Table ( -r )
 *
 */
static struct sliceTable *
genRootSliceTable ( ddefsData )
struct ddefs * ddefsData;
{
	extern char * getenv (), * strcpy ();
	extern int atoi ();
	static struct sliceTable * genInitialSliceTable ();

	struct sliceTable * sliceTable;
	struct sliceInfo * sliceList;
	int i, fs0_size, sw_slice, sw_size;
	char * envp;

	if ( ( sliceTable = genInitialSliceTable ( ddefsData ) )  ==  NULL )
		return ( NULL );

	/* fs0_size = number of blocks in slice0 for 'mkfs(1M)' */
	fs0_size  =  ( ( ( envp = getenv ( "FS0SIZE" ) )  !=  NULL ) && 
		     ( ( i = atoi ( envp ) )  >=  0 ) )  ?
			i  :  ddefsData -> dd_vid . vid_rsize;

	if ( ! fs0_size )
		return ( sliceTable );

	sliceList  =  sliceTable -> st_slInfo;

	sliceList [ 1 ] . s_tab . nblocks  =
		sliceList [ 0 ] . s_tab . nblocks - fs0_size;

	sliceList [ 0 ] . s_tab . nblocks  =  ( long ) fs0_size;
	sliceList [ 0 ] . s_fs  . s_fsize  =  ( long ) fs0_size;
	
	sliceList [ 1 ] . s_tab . blkoff  =
		( ! sliceList [ 1 ] . s_tab . nblocks ) ?
			0L  :  sliceList [ 0 ] . s_tab . blkoff + fs0_size;

	strcpy ( sliceList [ 0 ] . s_fs . s_fname, "root" );
	strcpy ( sliceList [ 0 ] . s_fs . s_fpack, "R3" );

	/* sw_size = number of blocks in swap */
	sw_size  =  ddefsData -> dd_vid . vid_ssize;

	/* sw_slice = swap slice */
	if ( ( sw_slice  =  ddefsData -> dd_vid . vid_sslice )  >  2 ) {
		sw_slice = 0;
		sw_size = 0;
	}

	if ( ! sw_size )
		return ( sliceTable );

	if ( sw_slice  ==  0 ) {
		/* swap on root slice */

		sliceList [ 1 ] . s_tab . blkoff  =
( ( sliceList [ 1 ] . s_tab . nblocks -= sw_size ) ==  0L )  ?
			0L  :  sliceList [ 1 ] . s_tab . blkoff + sw_size;

		sliceList [ 0 ] . s_tab . nblocks  +=  sw_size;

	} else {
		sliceList [ 2 ] . s_tab . nblocks  =
			sliceList [ 1 ] . s_tab . nblocks - sw_size;

		sliceList [ 1 ] . s_tab . nblocks  = ( long ) sw_size;
		
		sliceList [ 2 ] . s_tab . blkoff  =
			( ! sliceList [ 2 ] . s_tab . nblocks ) ?
			0L : sliceList [ 1 ] . s_tab . blkoff + sw_size;
	}
	return ( sliceTable );
}


/* Generate Root/Swap/Usr Slice Table ( -u )
 *
 */
static struct sliceTable *
genRootUsrSliceTable ( ddefsData )
struct ddefs * ddefsData;
{
	extern int atoi ();
	extern char * getenv (), * strcpy ();
	static struct sliceTable * genRootSliceTable ();

	struct sliceTable * sliceTable;
	struct sliceInfo * sliceList;
	int i, fs1_size, fs1_slice;
	char * envp;

	sliceTable  =  genRootSliceTable ( ddefsData );
	if ( sliceTable  ==  NULL )
		return ( NULL );

	/* fs1_size= # blocks in /usr filesystem for 'mkfs(1M)' */

	fs1_size  =  ( ( ( envp = getenv ( "FS1SIZE" ) )  !=  NULL ) &&
		       ( ( i = atoi ( envp ) )  >=  0 ) )  ?
			i  :  ddefsData -> dd_vid . vid_usize;

	/* fs1_slice = /usr slice */

#ifdef m88k
	fs1_slice  =  ( int ) ddefsData -> dd_vid . vid_u_uslice;
	fs1_slice  =  ( fs1_slice << 16 ) | ddefsData -> dd_vid . vid_l_uslice;
#endif

#ifdef m68k
	fs1_slice  =  ddefsData -> dd_vid . vid_uslice;
#endif

	if ( fs1_slice  >  2 )
		return ( sliceTable );

	sliceList  =  sliceTable -> st_slInfo;

	sliceList [ fs1_slice + 1 ] . s_tab . nblocks  =
		sliceList [ fs1_slice ] . s_tab . nblocks - fs1_size;

	sliceList [ fs1_slice + 1 ] . s_tab . blkoff = 
		( sliceList [ fs1_slice + 1 ] . s_tab . nblocks  ==  0L ) ?
		0L  :  sliceList [ fs1_slice ] . s_tab . blkoff + fs1_size;

	sliceList [ fs1_slice ] . s_tab . nblocks  =  fs1_size;
	sliceList [ fs1_slice ] . s_fs  . s_fsize  =  fs1_size;

	strcpy ( sliceList [ fs1_slice ] . s_fs . s_fname, "usr" );
	strcpy ( sliceList [ fs1_slice ] . s_fs . s_fpack, "R3" );

	return ( sliceTable );
}


/* Free Slice Table Structure
 *
 */
static void
freeSliceTable ( sliceTable )
struct sliceTable * sliceTable;
{
	extern void free ();

	if ( sliceTable -> st_spFile  !=  NULL )
		free ( sliceTable -> st_spFile );
	if ( sliceTable -> st_slInfo  !=  NULL )
		free ( sliceTable -> st_slInfo );
	free ( sliceTable );
	return;
}


/* Read Slice Table Information From Input Stream ( -e )
 *
 */
static struct sliceTable *
readSlicingInformation ( stream )
FILE * stream;
{
	extern int strcmp (), strncmp ();
	extern void free ();
	extern long strtol ();
	extern char * calloc (), * memset ();

	static char * skipWhiteSpace ();

	static char buffer [ BUFSIZ ];
	static char * spDeviceError  =
"%s: Slice device, slice count, block size format error line.\n";
	static char NoOfKiloBytes []  =  "K byte blocks";

	struct sliceTable * sliceTable;
	struct sliceInfo * sliceList;
	int entryNumber, charCount;
	char * p, * q;

	sliceTable  =  ( struct sliceTable * )
				calloc ( 1, sizeof ( struct sliceTable ) );
	if ( sliceTable  ==  NULL ) {
		perror ( myName );
		return ( NULL );
	}

	sliceTable -> st_spFile  =  NULL;
	sliceTable -> st_slInfo  =  NULL;

	if ( fgets ( buffer, BUFSIZ, stream )  ==  NULL ) {

		if ( feof ( stream ) )
			fprintf ( stderr,
				"%s: No slicing information found.\n", myName );
		else
			perror ( myName );

		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	if ( ( p  =  strchr ( buffer, ',' ) )  ==  NULL ) {

		fprintf ( stderr, spDeviceError, myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}	

	* p ++  =  '\0';

	if ( ! ( charCount  =  strlen ( buffer ) ) ) {

		fprintf ( stderr, "%s: No special file pathlist.\n", myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	sliceTable -> st_spFile  =  calloc ( charCount + 1, sizeof ( char ) );
	if ( sliceTable -> st_spFile  ==  NULL ) {

		perror ( myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	strncpy ( sliceTable -> st_spFile, buffer, charCount );

	sliceTable -> st_noSlices  =  ( int ) strtol ( p, & q, 0 );

	if ( ( ( ! sliceTable -> st_noSlices ) && ( p  ==  q ) ) ||
	     ( sliceTable -> st_noSlices  <  MinSliceCount ) ||
	     ( sliceTable -> st_noSlices  >  MaxSliceCount ) ) {

		fprintf ( stderr, "%s: Invalid number of slices.\n", myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	if ( ( * ( q = skipWhiteSpace ( q ) )  ==  '\0' ) ||
	     ( ( p = strchr ( q, ',' ) )  ==  NULL ) ||
	     ( * p ++  =  '\0', strcmp ( q, "slices" ) ) ) {

		fprintf ( stderr, "%s: 'slices' token not found.\n", myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	sliceList  =  ( struct sliceInfo * )
				calloc ( sliceTable -> st_noSlices,
						sizeof ( struct sliceInfo ) );
	if ( sliceList  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	sliceTable -> st_slInfo  =  sliceList;

	for ( entryNumber = 0;
	      entryNumber < sliceTable -> st_noSlices;
	      entryNumber ++ ) {
		sliceList [ entryNumber ] . s_sliceno        =  entryNumber;
		sliceList [ entryNumber ] . s_flag           =  0;
		sliceList [ entryNumber ] . s_tab . blkoff   =  0L;
		sliceList [ entryNumber ] . s_tab . nblocks  =  0L;
		sliceList [ entryNumber ] . s_fs . s_fsize   =  0L;

		memset ( sliceList [ entryNumber ] . s_fs . s_fname,
				'\0', FsNameSize );
		memset ( sliceList [ entryNumber ] . s_fs . s_fpack,
				'\0', FsVolIdSize );
	}

	sliceList [ sliceTable -> st_noSlices - 1 ] . s_flag  =  RDONLY;

	sliceTable -> st_noOfK  =  ( int ) strtol ( p, & q, 0 );

	if ( ( ! sliceTable -> st_noOfK ) && ( p  ==  q ) ) {

		fprintf ( stderr, "%s: Invalid number of K bytes per block.\n",
				myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	charCount = strlen ( NoOfKiloBytes );

	if ( ( * ( q = skipWhiteSpace ( q ) )  ==  '\0' ) ||
	     ( strncmp ( q, NoOfKiloBytes, charCount ) ) ) {

		fprintf ( stderr, "%s: '%s' token not found.\n",
				myName, NoOfKiloBytes );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	if ( * skipWhiteSpace ( q + charCount )  !=  '\0' ) {

		fprintf ( stderr, "%s: Trailing non-blank characters found.\n",
				myName );
		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	for ( ;; ) {
		if ( fgets ( buffer, BUFSIZ, stream )  ==  NULL ) {

			if ( feof ( stream ) )
				fprintf ( stderr,
					"%s: Slice table header not found.\n",
						myName );
			else
				perror ( myName );

			freeSliceTable ( sliceTable );
			return ( NULL );
		}

		if ( * ( p = skipWhiteSpace ( buffer) )  ==  '\0' )
			continue;

		if ( ! strcmp ( buffer,
"slice\toffset\tsl size\tfs size\tfsname\tvol-id\n" ) )
			break;

		fprintf ( stderr, "%s: Slice table header not recognized.\n",
				myName );

		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	while ( fgets ( buffer, BUFSIZ, stream )  !=  NULL ) {

		if ( * ( p = skipWhiteSpace ( buffer) )  ==  '\0' )
			continue;

		entryNumber  =  ( int ) strtol ( p, & q, 0 );

		if ( ( ( entryNumber  ==  0 ) && ( p  ==  q ) ) ||
		     ( entryNumber < 0 ) ||
		     ( entryNumber >= sliceTable -> st_noSlices ) ) {

			fprintf ( stderr,
"%s: Invalid slice number (field one):\n%s\n", myName, p );
			break;
		}

		sliceList [ entryNumber ] . s_tab . blkoff  = 
			strtol ( p  =  q, & q, 0 );

		if ( ( sliceList [ entryNumber ] . s_tab . blkoff  ==  0L ) &&
		     ( p  ==  q ) ) {

			fprintf ( stderr,
"%s: Invalid slice offset (field two):\n%s\n", myName, buffer );
			break;
		}

		sliceList [ entryNumber ] . s_tab . nblocks  = 
			strtol ( p  =  q, & q, 0 );

		if ( ( sliceList [ entryNumber ] . s_tab . nblocks  ==  0L )
		     && ( p  ==  q ) ) {

			fprintf ( stderr,
"%s: Invalid slice size (field three):\n%s\n", myName, buffer );
			break;
		}

		sliceList [ entryNumber ] . s_fs . s_fsize  = 
			strtol ( p  =  q, & q, 0 );

		if ( ( sliceList [ entryNumber ] . s_fs . s_fsize  ==  0L )
		     && ( p  ==  q ) ) {

			fprintf ( stderr,
"%s: Invalid file system size (field four):\n%s\n", myName, buffer );
			break;
		}

		if ( * ( q = skipWhiteSpace ( q ) )  ==  '\0' ) {

			entryNumber ++;
			continue;
		}

		p  =  sliceList [ entryNumber ] . s_fs . s_fname;

		for ( charCount = 0; charCount < FsNameSize; charCount ++ )

			if ( ( isspace ( * q ) ) || ( ! ( * p ++ = * q ++ ) ) )
				break;

		if ( ( ! isspace ( * q ) ) && ( *q != '\0' ) ) {

			fprintf ( stderr,
"%s: File system name (field five is too long; max. of %d chars):\n%s\n",
				myName, FsNameSize, buffer );
			break;
		}

		if ( * ( q = skipWhiteSpace ( q ) )  ==  '\0' ) {

			entryNumber ++;
			continue;
		}

		p  =  sliceList [ entryNumber ] . s_fs . s_fpack;

		for ( charCount = 0; charCount < FsVolIdSize; charCount ++ )

			if ( ( isspace ( * q ) ) || ( ! ( * p ++ = * q ++ ) ) )
				break;

		if ( ( ! isspace ( * q ) ) && ( *q != '\0' ) ) {

			fprintf ( stderr,
"%s: File system volume ID  (field six is too long; max. of %d chars):\n%s\n",
				myName, FsVolIdSize, buffer );
			break;
		}

		if ( * ( q = skipWhiteSpace ( q ) )  !=  '\0' ) {

			fprintf ( stderr,
"%s: Found extra fields after field six:\n%s\n", myName, buffer );
			break;
		}

		entryNumber ++;
	}

	if ( ! feof ( stream ) ) {

		if ( ferror ( stream ) )
			perror ( myName );

		freeSliceTable ( sliceTable );
		return ( NULL );
	}

	return ( sliceTable );
}


/* Skip White Space
 *
 */
static char *
skipWhiteSpace ( p )
char * p;
{
	while ( isspace ( * p ) )
		p ++;

	return ( p );
}


/* Print Slice Table To Output Stream
 *
 */
static int
printSliceTable ( stream, sliceTable )
FILE * stream;
struct sliceTable * sliceTable;
{
	struct sliceInfo * sliceList;
	int entryNumber;

	fprintf ( stream, "%s, %d slices, %dK byte blocks\n\n",
		sliceTable -> st_spFile,
		sliceTable -> st_noSlices,
		sliceTable -> st_noOfK );

	fputs ( "slice\toffset\tsl size\tfs size\tfsname\tvol-id\n\n",
			stream );

	sliceList  =  sliceTable -> st_slInfo;

	for ( entryNumber  =  0;
	      entryNumber  <  sliceTable -> st_noSlices;
	      entryNumber ++ ) {

		fprintf ( stream, "%d\t%d\t%d\t%d\t%.6s\t%.6s\n",
			sliceList [ entryNumber ] . s_sliceno,
			sliceList [ entryNumber ] . s_tab . blkoff,
			sliceList [ entryNumber ] . s_tab . nblocks,
			sliceList [ entryNumber ] . s_fs . s_fsize,
			sliceList [ entryNumber ] . s_fs . s_fname,
			sliceList [ entryNumber ] . s_fs . s_fpack );
	}
	return ( 0 );
}


/* Free Ddefs(1M) Data Structure
 *
 */
static void
freeDdefsData ( ddefsData )
struct ddefs * ddefsData;
{
	if ( ddefsData -> dd_name  !=  NULL )
		free ( ddefsData -> dd_name );
	if ( ddefsData -> dd_com  !=  NULL )
		free ( ddefsData -> dd_com );
	if ( ddefsData -> dd_form  !=  NULL )
		free ( ddefsData -> dd_form );
	if ( ddefsData -> dd_alist  !=  NULL )
		free ( ddefsData -> dd_alist );
	if ( ddefsData -> dd_rdman  !=  NULL )
		free ( ddefsData -> dd_rdman );
	free ( ddefsData );
	return;
}


/* Read Ddefs(1M) Description File
 *
 */
static struct ddefs *
readDdefsData ( dskdefs, diskType )
char * dskdefs, * diskType;
{
	extern char * getenv(), * calloc ();
	extern int sprintf (), fread ();
	extern FILE * fopen ();

	static void freeDdefsData ();

	struct ddefs * ddefsData;
	FILE * stream;

	static char buffer [ BUFSIZ ];


	ddefsData  =  ( struct ddefs * ) calloc ( 1, sizeof ( struct ddefs ) );

	if ( ddefsData  ==  NULL ) {
		perror ( myName );
		return ( NULL );
	}

	ddefsData -> dd_name   =  NULL;
	ddefsData -> dd_com    =  NULL;
	ddefsData -> dd_form   =  NULL;
	ddefsData -> dd_alist  =  NULL;
	ddefsData -> dd_rdman  =  NULL;

	ddefsData -> dd_name  =  calloc ( strlen ( diskType ) + sizeof ( char ),
					sizeof ( char ) );

	if ( ddefsData -> dd_name  ==  NULL ) {
		perror ( myName );
		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	strcpy ( ddefsData -> dd_name, diskType );

	sprintf( buffer, "%s/%s", dskdefs, diskType );

	if ( ( stream  =  fopen ( buffer, "r" ) ) == NULL ) {

		perror ( buffer );
		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	if ( ( fread ( & ddefsData -> dd_vid,
			sizeof ( struct volumeid ), 1, stream )  !=  1 ) ||
	     ( fread ( & ddefsData -> dd_ios,
			sizeof ( struct dkconfig ), 1, stream )  !=  1 ) ) {

		perror ( diskType );
		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	ddefsData -> dd_alist  =  ( unsigned * )
calloc ( ddefsData -> dd_ios . ios_alts + 1, sizeof ( unsigned ) );

	if ( ddefsData -> dd_alist  ==  NULL ) {

		perror ( myName );
		freeDdefsData ( ddefsData );
		return ( NULL );
	}
	
	if ( fread ( ddefsData -> dd_alist, sizeof ( unsigned ),
				ddefsData -> dd_ios . ios_alts, stream )
			!=  ddefsData -> dd_ios . ios_alts ) {

		fprintf( stderr, "%s: Alternates list read error", diskType );

		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	if ( fgets ( buffer, BUFSIZ, stream )  ==  NULL ) {

		fprintf ( stderr, "%s: Format program read error\n", diskType );

		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	ddefsData -> dd_form  =  calloc ( strlen ( buffer ) + sizeof ( char ),
						sizeof ( char ) );

	if ( ddefsData -> dd_form  ==  NULL ) {

		perror ( diskType );
		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	strcpy ( ddefsData -> dd_form, buffer );

	if ( fgets ( buffer, BUFSIZ, stream )  ==  NULL ) {

		fprintf ( stderr, "%s: Comment string read error\n", diskType );

		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	ddefsData -> dd_com  =  calloc ( strlen ( buffer ) + sizeof ( char ),
						sizeof ( char ) );

	if ( ddefsData -> dd_com  ==  NULL ) {

		perror ( diskType );
		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	strcpy ( ddefsData -> dd_com, buffer );

	buffer [ 0 ]  =  '\0';

	if ( ( fgets ( buffer, BUFSIZ, stream )  ==  NULL ) &&
	     ( ferror ( stream ) ) ) {

		fprintf ( stderr,
"%s: Read manufacturer's defect list command line read error\n", diskType );

		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	ddefsData -> dd_rdman  =  calloc ( strlen ( buffer ) + sizeof ( char ),
						sizeof ( char ) );

	if ( ddefsData -> dd_rdman  ==  NULL ) {

		perror ( diskType );
		freeDdefsData ( ddefsData );
		return ( NULL );
	}

	strcpy ( ddefsData -> dd_rdman, buffer );
	return ( ddefsData );
}


/* Edit Slice Information States
 *
 */
#define ET_ERROR	0	/* Error, quit, don't write */
#define ET_QUITBANG	1	/* Quit, don't write */
#define ET_QUITWRITE	2	/* Quit, and write */
#define ET_NEXTSLICE	3	/* Go on to next slice */
#define ET_BACKSLICE	4	/* Go back to previous slice */


/* Edit Slice Table ( -e )
 *
 */
static struct sliceTable *
editSliceTable ( sliceTable )
struct sliceTable * sliceTable;
{
	extern char * calloc (), * strcpy (), * memcpy ();
	extern int strlen ();
	static int printSliceTable (), editField ();

	struct sliceInfo * sliceList;
	struct sliceTable * newSliceTable;
	int entryNumber;

	printSliceTable ( stdout, sliceTable );

	newSliceTable  =  ( struct sliceTable * )
				calloc ( 1, sizeof ( struct sliceTable ) );

	newSliceTable -> st_spFile    =  NULL;
	newSliceTable -> st_slInfo    =  NULL;
	newSliceTable -> st_noOfK     =  sliceTable -> st_noOfK;
	newSliceTable -> st_noSlices  =  sliceTable -> st_noSlices;

	newSliceTable -> st_spFile  =
		calloc ( strlen ( sliceTable -> st_spFile ) + sizeof ( char ),
				sizeof ( char ) );

	if ( newSliceTable -> st_spFile  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( newSliceTable );
		return ( NULL );
	}

	strcpy ( newSliceTable -> st_spFile, sliceTable -> st_spFile );

	sliceList  =  ( struct sliceInfo * )
				calloc ( sliceTable -> st_noSlices,
					sizeof ( struct sliceInfo ) );

	if ( sliceList  ==  NULL ) {
		perror ( myName );
		freeSliceTable ( newSliceTable );
		return ( NULL );
	}

	newSliceTable -> st_slInfo  =  sliceList;

	memcpy ( ( char * ) sliceList, ( char * ) sliceTable -> st_slInfo,
		sliceTable -> st_noSlices * sizeof ( struct sliceInfo ) );

	entryNumber  =  0;

	for (;;)
		switch ( editField ( newSliceTable, entryNumber ) ) {

		default:
			fprintf ( stderr,
				"%s: Illegal editField () return value\n",
				myName );
			/* fall through */

		case ET_ERROR:
		case ET_QUITBANG:
			freeSliceTable ( newSliceTable );
			return ( NULL );

		case ET_QUITWRITE:
			return ( newSliceTable );

		case ET_NEXTSLICE:
			entryNumber  =
				( entryNumber < newSliceTable ->
							st_noSlices - 1 ) ?
				entryNumber + 1  :  0;
			break;

		case ET_BACKSLICE:
			entryNumber  =  ( entryNumber  ==  0 ) ?
		newSliceTable -> st_noSlices - 1 : entryNumber - 1;
			break;
		}

	/*NEVERREACHED*/
}

/* Internal Edit Field Data Base
 *
 */
struct
edit {
	int e_flags;
	int e_type;
	char * e_prompt;
	char * e_fmt;
	char * e_help;
};

/* Edit Field Types (e_type)
 *
 */
#define EF_NUMBER	1
#define EF_STRING	2

static struct edit
edit [] = {
	{
		/* Slice Offset Field */
		0,
		EF_NUMBER,
		"offset",
		"%d ",
		"\nThe number of blocks from the begining of the disk.\n\
This is calculated by adding the previous slice's offset\n\
and the previous slice's size.\n\n"
	}, {
		/* Slice Size Field */
		0,
		EF_NUMBER,
		"slice size",
		"%d ",
		"\nThe number of blocks contained in the slice.\n\
This number plus the offset must not overlap the next slice.\n\n"
	}, {
		/* File System Size Field */
		0,
		EF_NUMBER,
		"filesystem size",
		"%d ",
		"\nThe number of blocks contained in the filesystem.\n\
This number must be less than or equal to the slice size.\n\n"
	}, {
		/* File System Name Field */
		0,
		EF_STRING,
		"filesystem name",
		"%.6s ",
		"\nThe name of the filesystem (6 characters).\n\n"
	}, {
		/* File System Volume ID Field */
		0,
		EF_STRING,
		"vol-id name",
		"%.6s ",
		"\nThe name of the volumeid (6 characters).\n\n"
	}, {
		/* End Field List */
		0,
		0,
		NULL,
		NULL,
		NULL
	}
};

/* Field Subscripts
 *
 */
#define C_OFF		0
#define C_SLICE		1
#define	C_FSSIZE	2
#define C_FSNAME	3
#define C_VOLID		4
#define C_END		5


static char dotbuf [ BUFSIZ ];
static int hasChanged  =  0;

static int
editField ( sliceTable, entryNumber )
struct sliceTable * sliceTable;
int entryNumber;
{
#	define NoWriteWarning "\
\n\
No write since last change\n\
(q! overrides)\n\n"

#	define HelpMessage "\
\n\
Legal commands:\n\
\n\
	<newline>	step to next field\n\
	-		step to previous field\n\
	O		open space command\n\
	C		close space command\n\
	t		print slice table\n\
	w		write slice table\n\
	w file		write slice table to file\n\
	q		quit edit\n\
	q!		quit edit (no write)\n\
	!		execute interactive shell\n\
	!cmd		execute shell command\n\
	?		print information about field\n\
	.		use previous input string\n\
	help		print this help summary\n\
\n"

	extern int printf (), fflush (), strcmp ();
	extern char * strcpy ();
	static int printField (), shellEscape (), checkSliceTable (),
		doOpen (), doClose (), doChange ();

	struct sliceInfo * sliceInfo;
	char buffer [ BUFSIZ ];
	int fieldNumber;
	
	fieldNumber  =  0;
	sliceInfo  =  & sliceTable -> st_slInfo [ entryNumber ];
	putchar ( '\n' );

	while ( edit [ fieldNumber ] . e_prompt  !=  NULL ) {

		printf ( "slice %d>%30.30s: ", sliceInfo -> s_sliceno,
			edit [ fieldNumber ] . e_prompt );

		printField ( sliceTable, entryNumber, fieldNumber, stdout );

		if ( fflush ( stdout ) ) {
			perror ( "STANDARD OUTPUT" );
			return ( ET_ERROR );
		}

		if ( fgets ( buffer, BUFSIZ, stdin )  ==  NULL ) {
			if ( ferror ( stdin ) )
				perror ( "STANDARD INPUT" );
			return ( ET_ERROR );
		}

		if ( buffer [ 0 ]  ==  '!' ) {
			/* shell escape */
			shellEscape ( & buffer [ 1 ] );
			continue;
		}

		if ( ! strcmp ( buffer, "\n" ) ) {
			/* next field */
			fieldNumber ++;
			continue;
		}

		if ( ! strcmp ( buffer, "q\n" ) ) {
			if ( hasChanged ) {
				fputs ( NoWriteWarning, stdout );
				continue;
			}
			return ( ET_QUITWRITE );
		}

		if ( ! strcmp ( buffer, "q!\n" ) )
			return ( ET_QUITBANG );

		if ( ! strcmp ( buffer, "w\n" ) ) {
			/* write */
			if ( ! checkSliceTable ( sliceTable ) )
				continue;

			hasChanged  =  0;
			printf ( "\n`%s' written\n\n",
					sliceTable -> st_spFile );
			continue;
		}

		if ( ! strcmp ( buffer, "w!\n" ) ) {
			/* write no matter what */
			hasChanged  =  0;
			printf ( "\n`%s' written\n\n",
					sliceTable -> st_spFile );
			continue;
		}

		if ( ! strncmp ( buffer, "w ", sizeof ( "w " ) ) ) {
			/* write to file */
			puts ( "\nThat command is not supported.\n" );
			continue;
		}

		if ( ! strcmp ( buffer, "?\n" ) ) {
			fputs ( edit [ fieldNumber ] . e_help, stdout );
			continue;
		}

		if ( ! strcmp ( buffer, "-\n" ) ) {
			/* back up command */
			if ( fieldNumber  ==  0 )
				return ( ET_BACKSLICE );

			-- fieldNumber;
			continue;
		}

		if ( ! strcmp ( buffer, "O\n" ) ) {
			/* Open command  */
			if ( doOpen ( sliceTable, entryNumber ) )
				-- hasChanged;
			continue;
		}

		if ( ! strcmp ( buffer, "C\n" ) ) {
			/* Close command  */
			if ( doClose ( sliceTable, entryNumber ) )
				-- hasChanged;
			continue;
		}

		if ( ! strcmp ( buffer, ".\n" ) ) {
			/* previous string */
			strcpy ( buffer, dotbuf );
			if ( doChange ( sliceTable, entryNumber,
						fieldNumber, buffer ) ) {
				-- hasChanged;
				fieldNumber ++;
			}
			continue;
		}

		if ( ! strcmp ( buffer, "t\n" ) ) {
			printSliceTable ( stdout, sliceTable );
			continue;
		}

		if ( ! strcmp ( buffer, "help\n" ) ) {
			fputs ( HelpMessage, stdout );
			continue;
		}

		if ( doChange ( sliceTable, entryNumber,
						fieldNumber, buffer ) ) {
			-- hasChanged;
			fieldNumber ++;
		}

		strcpy ( dotbuf, buffer );
	}

	return ( ET_NEXTSLICE );
}


/* Print Field Contents
 *
 */
static int
printField ( sliceTable, entryNumber, fieldNumber, stream )
struct sliceTable * sliceTable;
int entryNumber, fieldNumber;
FILE * stream;
{
	struct sliceInfo * sliceList;
	long value;
	char * valuePointer;

	sliceList  =  sliceTable -> st_slInfo;

	switch ( fieldNumber ) {
	case C_OFF:
		value  =  sliceList [ entryNumber ] . s_tab . blkoff;
		break;
	case C_SLICE:
		value  =  sliceList [ entryNumber ] . s_tab . nblocks;
		break;
	case C_FSSIZE:
		value  =  sliceList [ entryNumber ] . s_fs . s_fsize;
		break;
	case C_FSNAME:
		valuePointer  =  sliceList [ entryNumber ] .  s_fs . s_fname;
		break;
	case C_VOLID:
		valuePointer  =  sliceList [ entryNumber ] .  s_fs . s_fpack;
		break;
	}

	fprintf ( stream, edit [ fieldNumber ] . e_fmt,
			( edit [ fieldNumber ] . e_type  ==  EF_NUMBER )  ?
				value  :  ( long ) valuePointer );
	return ( 0 );
}


/* Escape To Shell
 *
 */
static int
shellEscape ( buffer )
char * buffer;
{
	extern int strlen (), fork (), execl (), wait (),
		fputs (), fflush ();
	extern int ( * signal () ) ();

	int pid, w, status;

	buffer [ strlen ( buffer ) - 1 ]  =  '\0';

	if ( ( pid = fork () )  ==  0 ) {
		/* child */

		if ( buffer [ 0 ]  ==  '\0' )
			execl ( "/bin/sh", "sh", 0 );
		else
			execl ( "/bin/sh", "sh", "-c", buffer, 0 );

		exit ( 1 );
	}

	signal ( SIGINT,  SIG_IGN );
	signal ( SIGQUIT, SIG_IGN );

	while ( ( ( w = wait ( & status ) )  !=  pid )  &&  ( w  !=  -1 ) )
		;

	fputs ( "\nPress the RETURN key to continue:  ", stdout );
	fflush ( stdout );

	getchar ();
	putchar ( '\n' );

	signal(SIGINT,  SIG_DFL);
	signal(SIGQUIT, SIG_DFL);

	return ( 0 );
}


/* Check Slice Offsets And Sizes For Overlap (Except For Last Slice)
 *
 */
static int
checkSliceTable ( sliceTable )
struct sliceTable * sliceTable;
{
#	define checkOverlap(e,o) \
	( ( sliceList [ e ] . s_tab . blkoff  <			\
			sliceList [ o ] . s_tab . blkoff +	\
			sliceList [ o ] . s_tab . nblocks ) &&	\
	  ( sliceList [ e ] . s_tab . blkoff  >=		\
			sliceList [ o ] . s_tab . blkoff ) ) ||	\
	( ( sliceList [ e ] . s_tab . blkoff  <			\
			sliceList [ o ] . s_tab . blkoff ) &&	\
	  ( sliceList [ e ] . s_tab . nblocks +			\
			sliceList [ e ] . s_tab . blkoff  >	\
			sliceList [ o ] . s_tab . blkoff ) )

	struct sliceInfo * sliceList;
	int entryNumber, otherEntry;

	sliceList  =  sliceTable -> st_slInfo;

	for ( entryNumber  =  0;
	      entryNumber  <  sliceTable -> st_noSlices - 1;
	      entryNumber ++ ) {

		if ( sliceList [ entryNumber ] . s_tab . nblocks  <
				sliceList [ entryNumber ] . s_fs . s_fsize ) {
			
			fprintf ( stderr,
		"\nSlice %d's size is smaller than its file system's size\n\n",
				sliceList [ entryNumber ] . s_sliceno );

			return ( 0 );
		}

		for ( otherEntry  =  0;
		      otherEntry  <  sliceTable -> st_noSlices - 1;
		      otherEntry ++ ) {

			if ( entryNumber  ==  otherEntry )
				continue;

			if ( checkOverlap ( entryNumber, otherEntry ) ) {

				fprintf ( stderr,
					"\nSlice %d overlaps with slice %d\n\n",
					entryNumber, otherEntry );

				return ( 0 );
			}
		}

		if ( sliceList [ entryNumber ] . s_tab . nblocks +
				sliceList [ entryNumber ] . s_tab . blkoff  >
		     sliceList [ sliceTable -> st_noSlices - 1 ] .
				s_tab . nblocks ) {

			fprintf ( stderr, "\nSlice %d beyond end of disk\n\n",
				entryNumber );

			return ( 0 );
		}
	}
	return ( -1 );
}


/* Open Slice Space Command
 *
 */
static int
doOpen ( sliceTable, entryNumber )
struct sliceTable * sliceTable;
int entryNumber;
{
	struct sliceInfo * sliceList;
	int prevEntryNumber;

	if ( entryNumber  ==  0 ) {
		fprintf ( stderr, "\nCan't open slice 0\n" );
		return ( 0 );
	}

	sliceList  =  sliceTable -> st_slInfo;

	if ( sliceList [ entryNumber ] . s_tab . blkoff ) {
		fprintf ( stderr, "\nSpace already allocated for slice %d\n",
				sliceList [ entryNumber ] . s_sliceno );
		return ( 0 );
	}

	prevEntryNumber  =  entryNumber - 1;
	while ( ( ! sliceList [ prevEntryNumber ] . s_tab . nblocks )  &&
		( prevEntryNumber  >  0 ) )
			-- prevEntryNumber;

	if ( sliceList [ prevEntryNumber ] . s_tab . nblocks  <=
			sliceList [ prevEntryNumber ] . s_fs . s_fsize ) {
		fprintf ( stderr,
			"No space to allocate for slice %d\n",
			sliceList [ entryNumber ] . s_sliceno );

		return ( 0 );
	}

	if ( sliceList [ prevEntryNumber ] . s_fs . s_fsize  ==  0L ) {
		fprintf ( stderr, "No filesystem on slice %d\n",
		sliceList [ prevEntryNumber ] . s_sliceno );
		return ( 0 );
	}

	sliceList [ entryNumber ] . s_tab . nblocks  =
		sliceList [ prevEntryNumber ] . s_tab . nblocks -
		sliceList [ prevEntryNumber ] . s_fs  . s_fsize;

	sliceList [ entryNumber ] . s_tab . blkoff  =
		sliceList [ prevEntryNumber ] . s_fs  . s_fsize +
		sliceList [ prevEntryNumber ] . s_tab . blkoff;

	sliceList [ prevEntryNumber ] . s_tab . nblocks  =
		sliceList [ prevEntryNumber ] . s_fs . s_fsize;

	return ( -1 );
}


/* Close Space Command
 *
 */
static int
doClose( sliceTable, entryNumber )
struct sliceTable * sliceTable;
int entryNumber;
{
	extern char * memset ();

	struct sliceInfo * sliceList;
	int prevEntryNumber;

	if ( entryNumber  ==  0 ) {
		fprintf ( stderr, "\nCan't close slice 0\n\n" );
		return ( 0 );
	}

	prevEntryNumber  =  entryNumber - 1;
	sliceList  =  sliceTable -> st_slInfo;

	while ( ( ! sliceList [ prevEntryNumber ] . s_tab . nblocks ) &&
		( prevEntryNumber > 0 ) )
			-- prevEntryNumber;

	sliceList [ prevEntryNumber ] . s_tab . nblocks  +=
		sliceList [ entryNumber ] . s_tab . nblocks;

	sliceList [ entryNumber ] . s_tab . nblocks  =  0L;
	sliceList [ entryNumber ] . s_tab . blkoff   =  0L;
	sliceList [ entryNumber ] . s_fs  . s_fsize  =  0L;

	if ( ! sliceList [ entryNumber ] . s_flag & NMPROT ) {
		memset ( sliceList [ entryNumber ] . s_fs . s_fname,
				'\0', FsNameSize );
		memset ( sliceList [ entryNumber ] . s_fs . s_fpack,
				'\0', FsVolIdSize );
	}
	return ( -1 );
}


/* Change Field Value
 *
 */
static int
doChange ( sliceTable, entryNumber, fieldNumber, buffer )
struct sliceTable * sliceTable;
int entryNumber;
int fieldNumber;
char * buffer;
{
	extern int strlen (), sscanf ();
	
	struct sliceInfo * sliceList;
	int makeChange;
	long n;

	sliceList  =  sliceTable -> st_slInfo;

	if ( ( sliceList [ entryNumber ] . s_flag & RDONLY ) ||
	     ( ( sliceList [ entryNumber ] . s_flag & NMPROT ) &&
	       ( ( fieldNumber  ==  C_FSNAME ) ||
	         ( fieldNumber  ==  C_VOLID  ) ) ) ) {

			fputs ( "\nYou may not modify this field.\n\n",
					stderr );
			return ( 0 );
	}

	makeChange  =  0;

	switch ( edit [ fieldNumber ] . e_type ) {

	case EF_NUMBER:
		makeChange  =  sscanf ( buffer, edit [ fieldNumber ] . e_fmt,
						( char * ) & n );
		break;

	case EF_STRING:
		-- makeChange;
		buffer [ strlen ( buffer ) - 1 ]  =  '\0';
		break;

	default:
		break;
	}

	if ( ! makeChange ) {	/* scanf failure or blank line */
		printf ( "\n%s\n\n", edit [ fieldNumber ] . e_help );
		return ( 0 );
	}

	switch ( fieldNumber ) {
	case C_OFF:
		sliceList [ entryNumber ] . s_tab . blkoff  =  n;

		if ( sliceList [ entryNumber ] . s_fs . s_fsize )
			sliceList [ entryNumber ] . s_flag  |=  FSMOD | NMMOD;

		break;

	case C_SLICE:
		sliceList [ entryNumber ] . s_tab . nblocks  =  n;
		break;

	case C_FSSIZE:
		sliceList [ entryNumber ] . s_fs . s_fsize  =  n;
		sliceList [ entryNumber ] . s_flag  |=  FSMOD | NMMOD;
		break;

	case C_FSNAME:
		strncpy ( sliceList [ entryNumber ] . s_fs . s_fname,
				buffer, FsNameSize );
		sliceList [ entryNumber ] . s_flag  |=  NMMOD;
		break;

	case C_VOLID:
		strncpy ( sliceList [ entryNumber ] . s_fs . s_fpack,
				buffer, FsVolIdSize );
		sliceList [ entryNumber ] . s_flag  |=  NMMOD;
		break;
	}

	sliceList [ entryNumber ] . s_flag  |=  MODIFIED;

	return ( -1 );
}
