#ident	"@(#)install:libmerge/Dialers/rdDialers.c	35.1"
/* Read Dialers Stream
 * Copyright 1988, Motorola, Inc.
 */

#include <stdio.h>
#include <Dialers.h>

void bsfix();

/*
 *	dialers_p -> an open dialer stream structure pointer
 *	arps	 -> an array of argument pointers 
 *	narps	 -> number of elements in array
 * Return value:
 *	-1	 -> Error occurred
 *	0	 -> End of file
 *	>0	 -> success - number of fields filled in
 */

int
readDialers ( dialers_p, arps, narps )
DIALERS_T *dialers_p;
char *arps[];
int narps;
{
	char *ptr;
	int arg_count;

	ptr = fgets ( dialers_p -> dial_bp, DIALERSLINE, dialers_p -> dial_fp );
	if ( ptr == NULL )
		return ( feof ( dialers_p -> dial_fp ) ? 0 : -1 );

	if ((dialers_p -> dial_bp[0] == '#') || (dialers_p -> dial_bp[0] == ' ') ||
	    (dialers_p -> dial_bp[0] == '\t') || (dialers_p -> dial_bp[0] == '\n')) {
		if ( narps > 1 ) {
			arps[0] = dialers_p -> dial_bp;
			arps[1] = NULL;
			return ( 1 );
		}
		arps[0] = NULL;
		return ( 0 );
	}

	arg_count = getargs( dialers_p -> dial_bp, arps, narps );
	bsfix( arps );
	return( arg_count );
}

#ident	"@(#)uucp:getargs.c	2.2"

/*
 * generate a vector of pointers (arps) to the
 * substrings in string "s".
 * Each substring is separated by blanks and/or tabs.
 *	s	-> string to analyze -- s GETS MODIFIED
 *	arps	-> array of pointers -- count + 1 pointers
 *	count	-> max number of fields
 * returns:
 *	i	-> # of subfields
 *	arps[i] = NULL
 */

static int
getargs(s, arps, count)
register char *s, *arps[];
{
	register int i;

	for (i = 0; i < count; i++) {
		while (*s == ' ' || *s == '\t')
			*s++ = '\0';
		if (*s == '\n')
			*s = '\0';
		if (*s == '\0')
			break;
		arps[i] = s++;
		while (*s != '\0' && *s != ' '
			&& *s != '\t' && *s != '\n')
				s++;
	}
	arps[i] = NULL;
	return(i);
}

/***
 *      bsfix(args) - remove backslashes from args
 *
 *      \123 style strings are collapsed into a single character
 *	\000 gets mapped into \N for further processing downline.
 *      \ at end of string is removed
 *	\t gets replaced by a tab
 *	\n gets replaced by a newline
 *	\r gets replaced by a carriage return
 *	\b gets replaced by a backspace
 *	\s gets replaced by a blank 
 *	any other unknown \ sequence is left intact for further processing
 *	downline.
 */

void
bsfix (args)
char **args;
{
	register char *str, *to, *cp;
	register int num;

	for (; *args; args++) {
		str = *args;
		for (to = str; *str; str++) {
			if (*str == '\\') {
				if (str[1] == '\0')
					break;
				switch (*++str) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
					for ( num = 0, cp = str
					    ; cp - str < 3
					    ; cp++
					    ) {
						if ('0' <= *cp && *cp <= '7') {
							num <<= 3;
							num += *cp - '0';
						}
						else
						    break;
					}
					if (num == 0) {
						*to++ = '\\';
						*to++ = 'N';
					} else
						*to++ = num;
					str = cp-1;
					break;

				case 't':
					*to++ = '\t';
					break;

				case 's':	
					*to++ = ' ';
					break;

				case 'n':
					*to++ = '\n';
					break;

				case 'r':
					*to++ = '\r';
					break;

				case 'b':
					*to++ = '\b';
					break;

				default:
					*to++ = '\\';
					*to++ = *str;
					break;
				}
			}
			else
				*to++ = *str;
		}
		*to = '\0';
	}
}
