#ident	"@(#)install:libmerge/gettydefs/rdgttydefs.c	35.1"
/* Read Gettydefs Stream
 * Copyright 1988, Motorola, Inc.
 */

#include <stdio.h>
#include <sys/types.h>
#include <termio.h>
#include <gettydefs.h>
#include <errno.h>
#include <ctype.h>

#define		FAILURE		-1
#define		SUCCESS		0

#define		ID		1
#define		IFLAGS		2
#define		FFLAGS		3
#define		MESSAGE		4
#define		NEXTID		5

GETTYDEFS_P *
readgettydefs ( gettydefs_f, gettydefs_p )
GETTYDEFS_F * gettydefs_f;
GETTYDEFS_P * gettydefs_p;
{
	char * getword (), * fields (), quoted ();
	GETTYDEFS_T * gptr;
	char * ptr, * line, * optr;
	int c, oldc, rawc, state, size, field;

/* Search for two \n's in a row. */

	oldc = '\0';
	line = gettydefs_f -> getty_bp;

	for ( ptr = line; ptr < ( line + GETTY_MAXLINE + 1 ); ptr++ ) {
		if ( ( rawc = getc( gettydefs_f -> getty_fp ) ) == EOF )
			break;
		c = * ptr = rawc;
		if ( ( c == '\n' ) && ( oldc == '\n' ) )
			break;
		oldc = c;
	}

/* If we didn't end with a '\n' or EOF, then the line is too long. */
/* Skip over the remainder of the stuff in the line so that we */
/* start correctly on next line. */

	if ( ( rawc != EOF ) && ( c != '\n' ) ) {
		for ( oldc = '\0'; ( rawc = getc( gettydefs_f -> getty_fp ) ) != EOF; oldc = c ) {
			c = rawc;
			if ( ( c == '\n' ) && ( oldc == '\n' ) )
				break;
		}
		errno = E2BIG;
		return ( NULL );
	}

/* If we ended at the end of the file, then if there is no */
/* input, break out immediately */

	if ( ( rawc == EOF ) && ( ptr == line ) )
		return ( NULL );

/* If the last character stored was an EOF or '\n', replace it */
/* with a '\0'. */

/* If the buffer is full, then make sure there is a null after the */
/* last character stored. */

	if ( ( * ptr == ( EOF & 0377 ) ) || ( * ptr == '\n' ) )
		* ptr = '\0';
	else
		*++ptr == '\0';

/* If line starts with #, treat as comment */

	if ( * line == '#' ) {
		gettydefs_p -> gfo_type = GFO_COMMENT;
		gettydefs_p -> gfo_comment = line;
		return ( gettydefs_p );
	}

/* Initialize "def" and "gptr". */

	gptr = gettydefs_f -> getty_gp;

	gptr -> g_id = NULL;
	gptr -> g_iflags . c_iflag = 0;
	gptr -> g_iflags . c_oflag = 0;
	gptr -> g_iflags . c_cflag = 0;
	gptr -> g_iflags . c_lflag = 0;
	gptr -> g_fflags . c_iflag = 0;
	gptr -> g_fflags . c_oflag = 0;
	gptr -> g_fflags . c_cflag = 0;
	gptr -> g_fflags . c_lflag = 0;
	gptr -> g_message = NULL;
	gptr -> g_nextid = NULL;

/* Now that we have the complete line, scan if for the various */
/* fields.  Advance to new field at each unquoted '#'. */

	state = ID;
	ptr = line;
	while ( ( state != FAILURE ) && ( state != SUCCESS ) ) {
		switch (state) {

/* Find word in ID field and move it to "d_id" array. */

		case ID:
			strncpy( gettydefs_f -> getty_d_id, getword( ptr, &size ), MAXIDLENGTH );
			gptr->g_id = gettydefs_f -> getty_d_id;

/* Move to the next field.  If there is anything but white space */
/* following the id up until the '#', then set state to FAILURE. */

			ptr += size;
			while ( isspace(*ptr) )
				ptr++;

			if ( *ptr != '#' ) {
				field = state;
				state = FAILURE;
			} else {
				ptr++;	/* Skip the '#' */
				state = IFLAGS;
			}
			break;

/* Extract the "g_iflags" */

		case IFLAGS:
			if ( ( ptr = fields ( ptr, & gptr->g_iflags ) ) == NULL ) {
				field = state;
				state = FAILURE;
			} else {
				ptr++;
				state = FFLAGS;
			}
			break;

/* Extract the "g_fflags". */

		case FFLAGS:
			if ( ( ptr = fields ( ptr , & gptr -> g_fflags ) ) == NULL ) {
				field = state;
				state = FAILURE;
			} else {

/* Force the CREAD mode in regardless of what the user specified. */

				ptr++;
				state = MESSAGE;
			}
			break;

/* Take the entire next field as the "login" message. */
/* Follow usual quoting procedures for control characters. */

		case MESSAGE:
			for ( optr = gettydefs_f -> getty_d_message;
			      ( ( ( c = * ptr ) != '\0' ) && ( c != '#' ) );
			      ptr ++ ) {

/* If the next character is a backslash, then get the quoted */
/* character as one item. */

				if ( c == '\\' ) {
					c = quoted( ptr, & size );

/* -1 accounts for ++ that takes place later. */

					ptr += size - 1;
				}

/* If there is room, store the next character in d_message. */

				if ( optr < & gettydefs_f -> getty_d_message[ MAXMESSAGE ] )
					* optr ++ = c;
			}

/* If we ended on a '#', then all is okay.  Move state to NEXTID. */
/* If we didn't, then set state to FAILURE. */

			if ( c == '#' ) {
				gptr -> g_message = gettydefs_f -> getty_d_message;
				state = NEXTID;

/* Make sure message is null terminated. */

				* optr ++ = '\0';
				ptr ++;
			} else {
				field = state;
				state = FAILURE;
			}
			break;

/* Finally get the "g_nextid" field.  If this is successful, then */
/* the line parsed okay. */

		case NEXTID:

/* Find the first word in the field and save it as the next id. */

			strncpy ( gettydefs_f -> getty_d_nextid, getword ( ptr, & size ), MAXIDLENGTH );
			gptr->g_nextid = gettydefs_f -> getty_d_nextid;

/* There should be nothing else on the line.  Starting after the */
/* word found, scan to end of line.  If anything beside white */
/* space, set state to FAILURE. */

			ptr += size;
			while ( isspace( * ptr ) )
				ptr++;

			if (*ptr != '\0') {
				field = state;
				state = FAILURE;
			} else
				state = SUCCESS;
			break;
		}
	}

/* If a line was successfully picked up and parsed, compare the */
/* "g_id" field with the "id" we are looking for. */

	if (state == SUCCESS) {

		gettydefs_p -> gfo_type = GFO_ENTRY;
		gettydefs_p -> gfo_entry = gptr;

		return ( gettydefs_p );

	}

	errno = EINVAL;
	return ( NULL );
}

static char *
getword( ptr, size )
char * ptr;
int * size;
{
	char * optr,c;
	char quoted ();
	static char word[ MAXIDLENGTH + 1 ];
	int qsize;

/* Skip over all white spaces including quoted spaces and tabs. */

	* size = 0;
	while ( ( isspace ( * ptr ) ) || ( * ptr == '\\' ) ) {
		if ( * ptr == '\\' ) {
			c = quoted ( ptr, & qsize );
			( * size ) += qsize;
			ptr += qsize + 1;

/* If this quoted character is not a space or a tab or a newline */
/* then break. */
			if ( isspace ( c ) == 0 )
				break;
		} else {
			( * size ) ++;
			ptr ++;
		}
	}

/* Put all characters from here to next white space or '#' or '\0' */
/* into the word, up to the size of the word. */

	for ( optr = word, * optr = '\0';
	      ( isspace ( * ptr ) == 0 )  && ( * ptr != '\0' ) && ( * ptr != '#' );
	      ptr++, ( * size ) ++ ) {

/* If the character is quoted, analyze it. */

		if ( * ptr == '\\' ) {
			c = quoted ( ptr, & qsize );
			( * size ) += qsize;
			ptr += qsize;
		} else
			c = * ptr;

/* If there is room, add this character to the word. */

		if ( optr < & word[ MAXIDLENGTH + 1 ] )
			* optr ++ = c;
	}

/* Make sure the line is null terminated. */

	* optr ++ = '\0';
	return ( word );
}

/*	"quoted" takes a quoted character, starting at the quote	*/
/*	character, and returns a single character plus the size of	*/
/*	the quote string.  "quoted" recognizes the following as		*/
/*	special, \n,\r,\v,\t,\b,\f as well as the \nnn notation.	*/

static char
quoted( ptr, qsize )
char * ptr;
int * qsize;
{
	char c, * rptr;
	int i;

	rptr = ptr;
	switch ( * ++ rptr ) {
	case 'n':
		c = '\n';
		break;
	case 'r':
		c = '\r';
		break;
	case 'v':
		c = '\013';
		break;
	case 'b':
		c = '\b';
		break;
	case 't':
		c = '\t';
		break;
	case 'f':
		c = '\f';
		break;
	default:

/* If this is a numeric string, take up to three characters of */
/* it as the value of the quoted character. */

		if ( ( * rptr >= '0' ) && ( * rptr <= '7' ) ) {
			for ( i = 0, c = 0; ( i < 3 ); i ++) {
				c = c * 8 + ( * rptr - '0' );
				if ( * ++ rptr < '0' || * rptr > '7')
					break;
			}
			rptr--;

/* If the character following the '\\' is a NULL, back up the */
/* ptr so that the NULL won't be missed.  The sequence */
/* backslash null is essentually illegal. */

		} else if (*rptr == '\0') {
			c = '\0';
			rptr--;

/* In all other cases the quoting does nothing. */

		} else
			c = *rptr;
		break;
	}

/* Compute the size of the quoted character. */

	( * qsize ) = rptr - ptr + 1;
	return( c );
}

/*	"fields" picks up the words in the next field and converts all	*/
/*	recognized words into the proper mask and puts it in the target	*/
/*	field.								*/

static char *
fields ( ptr, termio )
char * ptr;
struct termio * termio;
{
	extern struct GETTY_Symbols imodes[], omodes[], cmodes[], lmodes[];

	struct GETTY_Symbols * symbol, * search();
	char * word, * getword();
	int size;

	termio->c_iflag = 0;
	termio->c_oflag = 0;
	termio->c_cflag = 0;
	termio->c_lflag = 0;

	while ( ( * ptr != '#' ) && ( * ptr != '\0' ) ) {

/* Pick up the next word in the sequence. */

		word = getword ( ptr, & size );

/* If there is a word, scan the two mode tables for it. */

		if ( * word != '\0' ) {

/* If the word is the special word "SANE", put in all the flags */
/* that are needed for SANE tty behavior. */

			if ( strcmp ( word, "SANE" ) == 0 ) {
				termio -> c_iflag |= ISANE;
				termio -> c_oflag |= OSANE;
				termio -> c_cflag |= CSANE;
				termio -> c_lflag |= LSANE;
			} else if ( ( symbol = search ( word, imodes ) ) != NULL )
				termio -> c_iflag |= symbol -> s_value;
			else if ( ( symbol = search ( word, omodes ) ) != NULL )
				termio -> c_oflag |= symbol -> s_value;
			else if ( ( symbol = search ( word, cmodes ) ) != NULL )
				termio -> c_cflag |= symbol -> s_value;
			else if ( ( symbol = search ( word, lmodes ) ) != NULL )
				termio -> c_lflag |= symbol -> s_value;
			else
				return ( NULL );
		}

/* Advance pointer to after the word. */

		ptr += size;
	}

/* If we didn't end on a '#', return NULL, otherwise return the */
/* updated pointer. */

	return ( ( * ptr != '#' ) ? NULL : ptr );
}

static struct GETTY_Symbols *
search( target, symbols )
char * target;
struct GETTY_Symbols * symbols;
{

/* Each symbol array terminates with a null pointer for an */
/* "s_symbol".  Scan until a match is found, or the null pointer */
/* is reached. */

	while ( symbols -> s_symbol != NULL )
		if ( strcmp ( target, symbols -> s_symbol ) == 0 )
			return ( symbols );
		else
			symbols ++;
	return(NULL);
}

/*	Modes set with the TCSETAW ioctl command.			*/

static struct GETTY_Symbols imodes[] = {
	"IGNBRK",	IGNBRK,
	"BRKINT",	BRKINT,
	"IGNPAR",	IGNPAR,
	"PARMRK",	PARMRK,
	"INPCK",	INPCK,
	"ISTRIP",	ISTRIP,
	"INLCR",	INLCR,
	"IGNCR",	IGNCR,
	"ICRNL",	ICRNL,
	"IUCLC",	IUCLC,
	"IXON",		IXON,
	"IXANY",	IXANY,
	"IXOFF",	IXOFF,
	NULL,	0
};

static struct GETTY_Symbols omodes[] = {
	"OPOST",	OPOST,
	"OLCUC",	OLCUC,
	"ONLCR",	ONLCR,
	"OCRNL",	OCRNL,
	"ONOCR",	ONOCR,
	"ONLRET",	ONLRET,
	"OFILL",	OFILL,
	"OFDEL",	OFDEL,
	"NLDLY",	NLDLY,
	"NL0",		NL0,
	"NL1",		NL1,
	"CRDLY",	CRDLY,
	"CR0",		CR0,
	"CR1",		CR1,
	"CR2",		CR2,
	"CR3",		CR3,
	"TABDLY",	TABDLY,
	"TAB0",		TAB0,
	"TAB1",		TAB1,
	"TAB2",		TAB2,
	"TAB3",		TAB3,
	"BSDLY",	BSDLY,
	"BS0",		BS0,
	"BS1",		BS1,
	"VTDLY",	VTDLY,
	"VT0",		VT0,
	"VT1",		VT1,
	"FFDLY",	FFDLY,
	"FF0",		FF0,
	"FF1",		FF1,
	NULL,		0
};

static struct GETTY_Symbols cmodes[] = {
	"B0",		B0,
	"B50",		B50,
	"B75", 		B75,
	"B110",		B110,
	"B134",		B134,
	"B150",		B150,
	"B200",		B200,
	"B300",		B300,
	"B600",		B600,
	"B1200",	B1200,
	"B1800",	B1800,
	"B2400",	B2400,
	"B4800",	B4800,
	"B9600",	B9600,
	"B19200",	B19200,
#ifdef m68k
	"B38400",	B38400,
#endif
				/* the following aliases should be deleted  */
				/* when cleaned out of the source */
	"EXTA",		EXTA,		/* same as B19200 */
	"EXTB",		EXTB,		/* same as B38400 */
					/**/
	"CS5",		CS5,
	"CS6",		CS6,
	"CS7",		CS7,
	"CS8",		CS8,
	"CSTOPB",	CSTOPB,
	"CREAD",	CREAD,
	"PARENB",	PARENB,
	"PARODD",	PARODD,
	"HUPCL",	HUPCL,
	"CLOCAL",	CLOCAL,
	NULL,		0
};

static struct GETTY_Symbols lmodes[] = {
	"ISIG",		ISIG,
	"ICANON",	ICANON,
	"XCASE",	XCASE,
	"ECHO",		ECHO,
	"ECHOE",	ECHOE,
	"ECHOK",	ECHOK,
	"ECHONL",	ECHONL,
	"NOFLSH",	NOFLSH,
	NULL,		0
};

/*	Terminal types set with the LDSETT ioctl command.		*/

static struct GETTY_Symbols terminals[] = {
	"none",		TERM_NONE,
#ifdef	TERM_V10
	"vt100",	TERM_V10,
#endif
#ifdef	TERM_H45
	"hp45",		TERM_H45,
#endif
#ifdef	TERM_C10
	"c100",		TERM_C100,
#endif
#ifdef	TERM_TEX
	"tektronix",	TERM_TEX,
	"tek",		TERM_TEX,
#endif
#ifdef	TERM_D40
	"ds40-1",	TERM_D40,
#endif
#ifdef	TERM_V61
	"vt61",		TERM_V61,
#endif
#ifdef	TERM_TEC
	"tec",		TERM_TEC,
#endif
	NULL,		0
};

/*	Line disciplines set by the TIOCSETD ioctl command.		*/

#ifndef	LDISC0
#define	LDISC0	0
#endif

static struct GETTY_Symbols linedisc[] = {
	"LDISC0",		LDISC0,
	NULL,		0
};

