#ident	"@(#)install:libmerge/inittab/rdinittab.c	35.1"
/* Read Inittab Stream
 * Copyright 1988, Motorola, Inc.
 */

#ident	"@(#)init:init.c	1.20"

#include <stdio.h>
#include <inittab.h>
#include <memory.h>
#include <errno.h>
#include <ctype.h>

#define	TRUE	1
#define	FALSE	0
#define	FAILURE	-1

#define	ID	1
#define	LEVELS	2
#define	ACTION	3
#define	COMMAND	4

static char *actions[] = {
	"off","respawn","ondemand","once","wait","boot",
	"bootwait","powerfail","powerwait","initdefault",
	"sysinit",
};

static short act_masks[] = {
	M_OFF,M_RESPAWN,M_ONDEMAND,M_ONCE,M_WAIT,M_BOOT,M_BOOTWAIT,
	M_PF,M_PWAIT,M_INITDEFAULT,M_SYSINIT,
};

INITTAB_T *
readinittab ( inittab_f )
INITTAB_F * inittab_f;
{
	FILE * fp_inittab;
	char * shcmd;
	INITTAB_T * cmd;
	int i, proceed;
	char * ptr;
	int c;
	int state;
	char lastc, * ptr1;
	int errnum;
	extern int errno;

/* Zero out the cmd itself before trying next line. */
	fp_inittab = inittab_f -> init_fp;
	cmd = ( INITTAB_T * ) memset ( ( char * ) inittab_f -> init_cmd, '\0', sizeof ( INITTAB_T ) );
	shcmd = inittab_f -> init_shcmd;

/* Read in lines of /etc/inittab, parsing at colons, until a line */
/* is read in which doesn't end with a backslash.  Do not start */
/* if the first character read is an EOF.  Note that this means */
/* that should a line fail to end in a newline, it will still */
/* be processed, since the "for" will terminate normally once */
/* started, regardless of whether line terminates with a newline */
/* or an EOF. */
	if ( ( c = fgetc ( fp_inittab ) ) == EOF )
		return ( NULL );

	ptr = shcmd;
	state = ID;
	lastc = '\0';
	for ( proceed = TRUE; ( proceed ) && ( c != EOF ); lastc = c , c = fgetc ( fp_inittab ) ) {

/* If we are not in the FAILURE state and haven't yet reached */
/* the shell command field, process the line, otherwise just */
/* look for a real end of line. */
		if ( ( state != FAILURE ) && ( state != COMMAND ) ) {

/* Squeeze out spaces and tabs. */
			if ( ( c == ' ' ) || ( c == '\t' ) )
				continue;

/* If the character is a ':', then check the previous field for */
/* correctness and advance to the next field. */
			if ( c == ':' ) {
				switch ( state ) {
				case ID:

/* Check to see that there are only 1 to 4 characters for the id. */
					if ( ( ( i = ptr - shcmd ) < 1 ) || ( i > 4 ) )
						state = FAILURE;
					else {
						memcpy ( cmd -> c_id, shcmd, i );
						ptr = shcmd;	/* Reset pointer */
						state = LEVELS;
					}
					break;

				case LEVELS :

/* Build a mask for all the levels that this command will be */
/* legal */
					cmd -> c_levels = 0;
					for ( ptr1 = shcmd; ptr1 < ptr; ptr1 ++ ) {
						if ( ( * ptr1 >= '0' ) && ( * ptr1 <= '6' ) )
							cmd -> c_levels |= ( MASK0 << ( * ptr1 - '0' ) );
						else if ( ( * ptr1 >= 'a' ) && ( *ptr1 <= 'c' ) )
							cmd -> c_levels |= ( MASKa << ( * ptr1 - 'a' ) );
						else if ( ( * ptr1 == 's' ) || ( * ptr1 == 'S' ) )
							cmd -> c_levels |= MASKSU;
						else {
							state = FAILURE;
							break;
						}
					}
					if ( state != FAILURE ) {
						state = ACTION;
						ptr = shcmd;	/* Reset the buffer */
					}
					break;

				case ACTION :

/* Null terminate the string in shcmd buffer and then try to match */
/* against legal actions.  If the field is of length 0, then the */
/* default of "RESPAWN" is used if the id is numeric, otherwise */
/* the default is "OFF". */
					if ( ptr == shcmd ) {
						if ( ( isdigit ( cmd -> c_id [ 0 ] ) )
						   && ( ( cmd -> c_id [ 1 ] == '\0' ) || ( isdigit( cmd -> c_id [ 1 ] ) ) )
						   && ( ( cmd -> c_id [ 2 ] == '\0' ) || ( isdigit( cmd -> c_id [ 2 ] ) ) )
						   && ( ( cmd -> c_id [ 3 ] == '\0' ) || ( isdigit( cmd -> c_id [ 3 ] ) ) ) )
							cmd -> c_action = M_RESPAWN;
						else
							cmd -> c_action = M_OFF;
					} else {
						cmd -> c_action = 0;
						*ptr = '\0';
						for ( i= 0; i < sizeof ( actions ) / sizeof ( char * ); i ++ ) {
							if ( strcmp ( shcmd, actions [ i ] ) == 0 ) {
								if ( ( cmd -> c_levels & MASKSU )
								   && ! ( act_masks [i] & (M_INITDEFAULT|M_PF|M_PWAIT) ) )
									cmd -> c_action = 0;
								else
									cmd -> c_action = act_masks [ i ];
								break;
							}
						}
					}

/* If the action didn't match any legal action, set state to */
/* FAILURE. */
					if ( cmd -> c_action == 0 )
						state = FAILURE;
					else {
						state = COMMAND;
					}
					ptr = shcmd;
					break;
				}
				continue;
			}
		}

/* If the character is a '\n', then this is the end of a line. */
/* If the '\n' wasn't preceded by a backslash, it is also the end */
/* of an /etc/inittab command.  If it was preceded by a backslash */
/* then the next line is a continuation.  Note that the */
/* continuation '\n' falls through and is treated like other */
/* characters and is stored in the shell command line. */
		if ( c == '\n' ) {
			if ( lastc != '\\' ) {
				proceed = FALSE;
				* ptr = '\0';
				break;
			}
		}

/* For all other characters just stuff them into the command */
/* as long as there aren't too many of them. Make sure there is */
/* room for a terminating '\0' also. */
		if ( ptr >= ( shcmd + INITTAB_MAXCMDL-1 ) )
			state = FAILURE;
		else
			* ptr ++ = c;

/* If the character we just stored was a quoted backslash, then */
/* change "c" to '\0', so that this backslash will not cause a */
/* subsequent '\n' to appear quoted.  In otherwords '\' '\' '\n' */
/* is the real end of a command, while '\' '\n' is a continuation. */
		if ( ( c == '\\' ) && ( lastc == '\\' ) )
			c = '\0';
	}

/* Make sure all the fields are properly specified for a good */
/* command line. */
	if ( state == COMMAND ) {
		cmd -> c_command = shcmd;
		return ( cmd );
	}
	errno = EINVAL;
	return ( NULL );
}
