# @(#)install:sadmin/tools/badspots.sh	35.1
# Define Bad Spots List Methodologies Utility
# Copyright 1988, Motorola, Inc.

# Functions:
#
#	o BadSpotsList
#	  Output supported bad spots list methods;
#
#	o NameFor: badSpotsFile
#	  Output the name for the bad spots definition method;
#
#	o LetterFor: badSpotsFile
#	  Output the letter associated to the bad spots definition method;
#
#	o TitleFor: badSpotsFile
#	  Output the title of the define bad spots list method specified;
#
#	o DescriptionFor: badSpotsFile
#	  Output the description for the specified bad spots method file;
#
#	o ScriptFor: badSpotsFile
#	  Output the shell script fragment associated to this method file;
#

beVerbose=false
usageError=false

while [ ${#} -ne 0 ]
do
	argument="${1}"

	case "${argument}" in
	-* )
		: continue
		;;
	* )
		break
		;;
	esac

	shift

	while :
	do
		case "${argument}" in
		-B )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing badspots_dir argument.
!!!
				usageError=true
				break 2
			fi
			BadSpotsDirectory=${1}
			shift
			;;
		-B* )
			BadSpotsDirectory=`expr "${argument}" : '-B\(.*\)'`
			break
			;;
		-- )
			break 2
			;;
		-* )
			0<<!!! 1>&2 cat -
${0}: '`expr "${argument}" : '\(-.\).*`' is not a supported switch.
!!!
			usageError=true
			break 2
			;;
		esac

		case "${argument}" in
		-? )
			break
			;;
		esac

		argument=-`expr "${argument}" : '-.\(.*\)'`
	done
done

if ${usageError}
then
	: skip this part
else
	BadSpotsOptionsDirectory=${BadSpotsOptionsDirectory:-../../badspots}

	while :
	do
		case ${#} in
		0 )
			0<<!!! 1>&2 cat -
${0}: Function missing from command line.
!!!
			usageError=true
			break
			;;
		* )
			Function=${1}
			shift

			case ${Function} in
			BadSpotsList )
				: OK.
				;;
			NameFor: | \
			LetterFor: | \
			TitleFor: | \
			DescriptionFor: | \
			ScriptFor: )
				case ${#} in
				0 )
					0<<!!! 1>&2 cat -
${0}: Method type missing on command line.
!!!
					usageError=true
					break
					;;
				esac

				BadSpotsFile=${1}
				shift

				if [ ! -f \
			${BadSpotsOptionsDirectory}/${BadSpotsFile} ]
				then
					0<<!!! 1>&2 cat -
${0}: Bad spots option '${BadSpotsFile}' is not supported.
!!!
					usageError=true
					break
				fi

				. ${BadSpotsOptionsDirectory}/${BadSpotsFile}
				;;
			X* )
				0<<!!! 1>&2 cat -
${0}: Unrecognized function on command line.
!!!
				usageError=true
				break
				;;
			esac
		esac

		if ${usageError}
		then
			break
		fi

		case ${#} in
		0 )
			: OK.
			;;
		* )
			0<<!!! 1>&2 cat -
${0}: Too many arguments.
!!!
			usageError=true
			break
			;;
		esac

		break
	done
fi

if ${usageError}
then
	0<<!!! 1>&2 cat -
Usage:
	${0} [ -B badspots_dir ] token...
	where token... may be
		BadSpotsList
		NameFor: badSpotsFile
		LetterFor: badSpotsFile
		TitleFor: badSpotsFile
		DescriptionFor: badSpotsFile
		ScriptFor: badSpotsFile
!!!
	exit 2
fi

case ${Function} in
BadSpotsList )
	ls "${BadSpotsOptionsDirectory}"
	;;
NameFor: )
	echo "${BadSpotsName}"
	;;
LetterFor: )
	echo "${BadSpotsLetter}"
	;;
TitleFor: )
	echo "${BadSpotsTitle}"
	;;
DescriptionFor: )
	echo "${BadSpotsDescription}"
	;;
ScriptFor: )
	0<<!!!! cat -
${BadSpotsScript}
!!!!
	;;
esac
exit 0
