# @(#)install:sadmin/tools/diskctlrs.sh	35.1
# Look-Up Disk Controller/Drive Information Script
# Copyright 1988, Motorola, Inc.

# Functions:
#
#	o ControllersList
#	  Output supported disk controllers file list;
#
#	o DesignationFor: controllerTypeFile
#	  Output the designation for the specified controller;
#
#	o TypeFor: controllerTypeFile
#	  Output the name for the specified controller;
#
#	o TitleFor: controllerTypeFile
#	  Output the controller title for the specified controller;
#
#	o DescriptionFor: controllerTypeFile
#	  Output the description for specified controller;
#
#	o NumbersFor: controllerTypeFile
#	  Output the controller numbers supported by SYSTEM V/68.
#
#	o UpgradeScriptFor: controllerTypeFile
#	  Output the code fragment to upgrade the format for a disk
#	  drive associated to this controller (used exclusively with
#	  the MVME323 controller);
#
#	o DriveNumbersFor: controllerTypeFile ControllerNumber: controllerNumber
#	  Output drive numbers for specific controller;
#
#	o DiskTypesFor: controllerTypeFile ControllerNumber: controllerNumber
#		DriveNumber: driveNumber
#	  Output disk type information for controller and
#	  driver number;
#

beVerbose=false
usageError=false

while [ ${#} -ne 0 ]
do
	argument="${1}"

	case "${argument}" in
	-* )
		: continue
		;;
	* )
		break
		;;
	esac

	shift

	while :
	do
		case "${argument}" in
		-C )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing diskctlrs_dir argument.
!!!
				usageError=true
				break 2
			fi
			ControllersDirectory=${1}
			shift
			;;
		-C* )
			ControllersDirectory=`expr "${argument}" : '-C\(.*\)'`
			break
			;;
		-- )
			break 2
			;;
		-* )
			0<<!!! 1>&2 cat -
${0}: '`expr "${argument}" : '\(-.\).*`' is not a supported switch.
!!!
			usageError=true
			break 2
			;;
		esac

		case "${argument}" in
		-? )
			break
			;;
		esac

		argument=-`expr "${argument}" : '-.\(.*\)'`
	done
done

if ${usageError}
then
	: skip this part
else
	ControllersDirectory=${ControllersDirectory:-../../diskctlrs}

	while :
	do
		case ${#} in
		0 )
			0<<!!! 1>&2 cat -
${0}: Function missing from command line.
!!!
			usageError=true
			break
			;;
		* )
			Function="${1}"
			shift

			case ${Function} in
			ControllersList )
				: OK.
				;;
			DesignationFor: | \
			TypeFor: | \
			TitleFor: | \
			DescriptionFor: | \
			NumbersFor: | \
			UpgradeScriptFor: | \
			DriveNumbersFor: | \
			DiskTypesFor: )
				case ${#} in
				0 )
					0<<!!! 1>&2 cat -
${0}: Controller file name missing on command line.
!!!
					usageError=true
					break
					;;
				esac

				ControllerTypeFile=${1}
				shift

				if [ ! -f \
			${ControllersDirectory}/${ControllerTypeFile} ]
				then
					0<<!!! 1>&2 cat -
${0}: Controller '${ControllerTypeFile}' is not supported.
!!!
					usageError=true
					break
				fi

				. ${ControllersDirectory}/${ControllerTypeFile}

				case ${Function} in
				DriveNumbersFor: | \
				DiskTypesFor: )
					case ${#} in
					0 )
						0<<!!! 1>&2 cat -
${0}: 'ControllerNumber:' token missing from command line.
!!!
						usageError=true
						break
						;;
					esac

					ControllerNumberToken=${1}
					shift

					case ${ControllerNumberToken} in
					ControllerNumber: )
						: OK.
						;;
					* )
						0<<!!! 1>&2 cat -
${0}: '${ControllerNumberToken}' should be 'ControllerNumber:'.
!!!
						usageError=true
						break
						;;
					esac

					case ${#} in
					0 )
						0<<!!! 1>&2 cat -
${0}: Controller number missing from command line.
!!!
						usageError=true
						break
						;;
					esac

					ControllerNumber=${1}
					shift

					if [ X"${ControllerNumbers}" != X"" ] &&
					   echo "${ControllerNumbers}" |
						1>/dev/null grep \
				"^${ControllerNumber}	"
					then
						: success
					else
						0<<!!! 1>&2 cat -
${0}: Controller number '${ControllerNumber}'\
 not valid for controller '${ControllerTypeFile}'.
!!!
						exit 1
					fi 
					;;
				esac

				case ${Function} in
				DiskTypesFor: )
					case ${#} in
					0 )
						0<<!!! 1>&2 cat -
${0}: 'DriveNumber:' token missing from command line.
!!!
						usageError=true
						break
						;;
					esac
					DriveNumberToken=${1}
					shift

					case ${DriveNumberToken} in
					DriveNumber: )
						: OK.
						;;
					* )
						0<<!!! 1>&2 cat -
${0}: '${DriveNumberToken}' should be 'DriveNumber:'.
!!!
						usageError=true
						break
						;;
					esac

					case ${#} in
					0 )
						0<<!!! 1>&2 cat -
${0}: Drive number missing from command line.
!!!
						usageError=true
						break
						;;
					esac

					DriveNumber=${1}
					shift

					if [ X"${ControllerNumbers}" != X"" ] &&
					   echo "${ControllerNumbers}" | 
						sed -n -e \
				"s/^${ControllerNumber}	//p" |
						tr ';/' '\012	' |
						1>/dev/null grep \
				"^${DriveNumber}	"
					then
						: success
					else
						0<<!!! 1>&2 cat -
${0}: Disk number '${DriveNumber}' not valid for
	controller '${ControllerTypeFile}',
	controller number '${ControllerNumber}'.
!!!
						exit 1
					fi
					;;
				esac
				;;
			* )
				0<<!!! 1>&2 cat -
${0}: Unrecognized function on command line.
!!!
				usageError=true
				break
				;;
			esac
		esac

		if ${usageError}
		then
			break
		fi

		case ${#} in
		0 )
			: OK.
			;;
		* )
			0<<!!! 1>&2 cat -
${0}: Too many arguments.
!!!
			usageError=true
			break
			;;
		esac

		break
	done
fi

if ${usageError}
then
	0<<!!! 1>&2 cat -
Usage:
	${0} [ -C diskctlrs_dir ] token...
	where token... may be
		ControllersList
		DesignationFor: controllerTypeFile
		TypeFor: controllerTypeFile
		TitleFor: controllerTypeFile
		DescriptionFor: controllerTypeFile
		NumbersFor: controllerTypeFile
		UpgradeScriptFor: controllerTypeFile
		DriveNumbersFor: controllerTypeFile \\
			ControllerNumber: controllerNumber
		DiskTypesFor: controllerTypeFile \\
			ControllerNumber: controllerNumber \\
			DriveNumber: driveNumber
!!!
	exit 2
fi

case ${Function} in
ControllersList )
	ls "${ControllersDirectory}"
	;;
DesignationFor: )
	echo "${ControllerDesignation}"
	;;
TypeFor: )
	echo "${ControllerName}"
	;;
TitleFor: )
	echo "${ControllerTitle}"
	;;
DescriptionFor: )
	echo "${ControllerDescription}"
	;;
NumbersFor: )
	echo "${ControllerNumbers}" |
		sed -e 's/	.*//'
	;;
UpgradeScriptFor: )
	if [ X"${ControllerUpgradeScript}" != X"" ]
	then
		echo "${ControllerUpgradeScript}"
	fi
	;;
DriveNumbersFor: )
	echo "${ControllerNumbers}" |
	sed -n -e "s/^${ControllerNumber}	//p" |
	tr ';' '\012' |
	sed -e 's:/.*::'
	;;
DiskTypesFor: )
	echo "${ControllerNumbers}" |
	sed -n -e "s/^${ControllerNumber}	//p" |
	tr ';/' '\012	' |
	sed -n -e "s/^${DriveNumber}	//p" |
	tr ',' '\012'
	;;
esac

exit 0
