# @(#)install:sadmin/tools/disktypes.sh	35.1
# Disk Drives Information Utility
# Copyright 1988, Motorola, Inc.

# Functions:
#
#	o DiskTypesList
#	  Output supported disk types file list;
#
#	o TypeFor: diskTypeFile
#	  Output the disk name for the specified disk file;
#
#	o TitleFor: diskTypeFile
#	  Output the disk title for the specified disk file;
#
#	o DescriptionFor: diskTypeFile
#	  Output the description for the specified disk file;
#
#	o DdefsNameFor: diskTypeFile
#	  Output the ddefs(1M) type file name for the specified disk file;
#
#	o ReadDefectListCommandFor: diskTypeFile
#	  Output the ddefs(1M) command line which outputs the manufacturer's
#	  bad tracks list;
#
#	o BadSpotsOptionsFor: diskTypeFile
#	  Output bad track options for the specified disk type file;
#	  
#	o AternateTracksFor: diskTypeFile
#	  Output the /etc/diskalts list of FE-cylinders which constitute
#	  the initial list of bad spots for most MVME320 disk drives;
#

beVerbose=false
usageError=false

while [ ${#} -ne 0 ]
do
	argument="${1}"

	case "${argument}" in
	-* )
		: continue
		;;
	* )
		break
		;;
	esac

	shift

	while :
	do
		case "${argument}" in
		-D )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing disktypes_dir argument.
!!!
				usageError=true
				break 2
			fi
			DiskTypesDirectory=${1}
			shift
			;;
		-D* )
			DiskTypesDirectory=`expr "${argument}" : '-D\(.*\)'`
			break
			;;
		-- )
			break 2
			;;
		-* )
			0<<!!! 1>&2 cat -
${0}: '`expr "${argument}" : '\(-.\).*`' is not a supported switch.
!!!
			usageError=true
			break 2
			;;
		esac

		case "${argument}" in
		-? )
			break
			;;
		esac

		argument=-`expr "${argument}" : '-.\(.*\)'`
	done
done

if ${usageError}
then
	: skip this part
else
	DiskTypesDirectory=${DiskTypesDirectory:-../../disktypes}

	while :
	do
		case ${#} in
		0 )
			0<<!!! 1>&2 cat -
${0}: Function missing from command line.
!!!
			usageError=true
			break
			;;
		* )
			Function=${1}
			shift

			case ${Function} in
			DiskTypesList )
				: OK.
				;;
			TypeFor: | \
			TitleFor: | \
			DescriptionFor: | \
			DdefsNameFor: | \
			ReadDefectListCommandFor: | \
			BadSpotsOptionsFor: | \
			AlternateTracksFor: )
				case ${#} in
				0 )
					0<<!!! 1>&2 cat -
${0}: Disk type missing on command line.
!!!
					usageError=true
					break
					;;
				esac

				DiskTypeFile="${1}"
				shift

				if [ ! -f \
			${DiskTypesDirectory}/${DiskTypeFile} ]
				then
					0<<!!! 1>&2 cat -
${0}: Disk type '${DiskTypeFile}' is not supported.
!!!
					usageError=true
					break
				fi

				. ${DiskTypesDirectory}/${DiskTypeFile}
				;;
			X* )
				0<<!!! 1>&2 cat -
${0}: Unrecognized function on command line.
!!!
				usageError=true
				break
				;;
			esac
		esac

		if ${usageError}
		then
			break
		fi

		case ${#} in
		0 )
			: OK.
			;;
		* )
			0<<!!! 1>&2 cat -
${0}: Too many arguments.
!!!
			usageError=true
			break
			;;
		esac

		break
	done
fi

if ${usageError}
then
	0<<!!! 1>&2 cat -
Usage:
	${0} [ -D disktypes_dir ] token...
	where token... may be
		DiskTypesList
		TypeFor: diskTypeFile
		TitleFor: diskTypeFile
		DescriptionFor: diskTypeFile
		DdefsNameFor: diskTypeFile
		ReadDefectListCommandFor: diskTypeFile
		BadSpotsOptionsFor: diskTypeFile
		AlternateTracksFor: diskTypeFile
!!!
	exit 2
fi

case ${Function} in
DiskTypesList )
	ls "${DiskTypesDirectory}"
	;;
TypeFor: )
	echo ${DiskTypeName}
	;;
TitleFor: )
	echo "${DiskTypeTitle}"
	;;
DescriptionFor: )
	echo "${DiskTypeDescription}"
	;;
DdefsNameFor: )
	echo ${DiskTypeDdefsName}
	;;
ReadDefectListCommandFor: )
	if [ X"${DiskTypeReadDefectListCommand}" != X"" ]
	then
		echo "${DiskTypeReadDefectListCommand}"
	fi
	;;
BadSpotsOptionsFor: )
	echo "${DiskTypeBadSpotsOptions}"
	;;
AlternateTracksFor: )
	if [ X"${DiskTypeAlternateTracks}" != X"" ]
	then
		echo "${DiskTypeAlternateTracks}"
	fi
	;;
esac
exit 0
