# @(#)install:sadmin/tools/upgradein.sh	35.1
# Upgrade Release Information Utility
# Copyright 1988, Motorola, Inc.

FirstUpgradeReleaseFile=FE83.10

# Functions:
#
#	o UpgradeReleaseList
#	  Output upgradable release list;
#
#	o NameFor: upgradeReleaseFile
#	  Output the release name for the specified upgrade release file;
#
#	o DescriptionFor: upgradeReleaseFile
#	  Output the description for the specified upgrade release file;
#
#	o TapePACK_For: upgradeReleaseFile
#	  Output the PACK field found in the igf(1M) header for the tape
#	  associated to the specified upgrade release file;
#
#	o TapeDESC_For: upgradeReleaseFile
#	  Output the DESC field found in the igf(1M) header for the tape
#	  associated to the specified upgrade release file;
#
#	o TapeVER_For: upgradeReleaseFile
#	  Output the VER field found in the igf(1M) header for the tape
#	  associated to the specified upgrade release file;
#
#	o SaveFileListFor: upgradeReleaseFile
#	  Output the list of files which are to be preserved prior to
#	  the upgrade for the specified upgrade release file;
#
#	o ObsoleteFileListFor: upgradeReleaseFile
#	  Output the list of files which are to be deleted prior to
#	  the upgrade for the specified upgrade release file;
#
#	o NextReleaseNameFor: upgradeReleaseFile
#	  Output the name of the release following this release
#	  for the specified upgrade release file;
#
#	o ArchiveListFor: upgradeReleaseFile
#	  Output the archive list for the new product tape for the
#	  specified upgrade release file;
#
#	o ScriptFor: upgradeReleaseFile
#	  Output the upgrade script which upgrades the given release
#	  to the next release for the specified upgrade release file;
#

beVerbose=false
usageError=false

while [ ${#} -ne 0 ]
do
	argument="${1}"

	case "${argument}" in
	-* )
		: continue
		;;
	* )
		break
		;;
	esac

	shift

	while :
	do
		case "${argument}" in
		-U )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing upgrades_dir argument.
!!!
				usageError=true
				break 2
			fi
			UpgradesDirectory=${1}
			shift
			;;
		-U* )
			UpgradesDirectory=`expr "${argument}" : '-U\(.*\)'`
			break
			;;
		-- )
			break 2
			;;
		-* )
			0<<!!! 1>&2 cat -
${0}: '`expr "${argument}" : '\(-.\).*`' is not a supported switch.
!!!
			usageError=true
			break 2
			;;
		esac

		case "${argument}" in
		-? )
			break
			;;
		esac

		argument=-`expr "${argument}" : '-.\(.*\)'`
	done
done

if ${usageError}
then
	: skip this part
else
	UpgradesDirectory=${UpgradesDirectory:-../../upgrades}

	while :
	do
		case ${#} in
		0 )
			0<<!!! 1>&2 cat -
${0}: Function missing from command line.
!!!
			usageError=true
			break
			;;
		* )
			Function=${1}
			shift

			case ${Function} in
			UpgradeReleaseList )
				: OK.
				;;
			NameFor: | \
			DescriptionFor: | \
			TapePACK_For: | \
			TapeDESC_For: | \
			TapeVER_For: | \
			SaveFileListFor: | \
			ObsoleteFileListFor: | \
			NextReleaseNameFor: | \
			ArchiveListFor: | \
			ScriptFor: )
				case ${#} in
				0 )
					0<<!!! 1>&2 cat -
${0}: Upgrade release file argument missing on command line.
!!!
					usageError=true
					break
					;;
				esac

				UpgradeReleaseFile="${1}"
				shift

				if [ ! -d \
		${UpgradesDirectory}/${UpgradeReleaseFile} ] ||
				   [ ! -f \
	${UpgradesDirectory}/${UpgradeReleaseFile}/${UpgradeReleaseFile} ]
				then
					0<<!!! 1>&2 cat -
${0}: Upgrade release '${UpgradeReleaseFile}' is not supported.
!!!
					usageError=true
					break
				fi

	. ${UpgradesDirectory}/${UpgradeReleaseFile}/${UpgradeReleaseFile}
				;;
			* )
				0<<!!! 1>&2 cat -
${0}: Unrecognized function on command line.
!!!
				usageError=true
				break
				;;
			esac
		esac

		if ${usageError}
		then
			break
		fi

		case ${#} in
		0 )
			: OK.
			;;
		* )
			0<<!!! 1>&2 cat -
${0}: Too many arguments.
!!!
			usageError=true
			break
			;;
		esac

		break
	done
fi

if ${usageError}
then
	0<<!!! 1>&2 cat -
Usage:
	${0} [ -U upgrades_dir ] token...
	where token... may be
		UpgradeReleaseList
		NameFor: upgradeReleaseFile
		DescriptionFor: upgradeReleaseFile
		TapePACK_For: upgradeReleaseFile
		TapeDESC_For: upgradeReleaseFile
		TapeVER_For: upgradeReleaseFile
		SaveFileListFor: upgradeReleaseFile
		ObsoleteFileListFor: upgradeReleaseFile
		NextReleaseNameFor: upgradeReleaseFile
		ArchiveListFor: upgradeReleaseFile
		ScriptFor: upgradeReleaseFile
!!!
	exit 2
fi

case ${Function} in
UpgradeReleaseList )
	upgradeReleaseFile=${FirstUpgradeReleaseFile}

	UpgradeReleaseFiles=`
		ls ${UpgradesDirectory} |
		while read upgradesName
		do
			if [ -d ${UpgradesDirectory}/${upgradesName} ] &&
		[ -f ${UpgradesDirectory}/${upgradesName}/${upgradesName} ]
			then
				echo ${upgradesName}
			fi
		done`

	while :
	do
		echo ${upgradeReleaseFile}

		nextReleaseName=`
			upgradeinfo NextReleaseNameFor: ${upgradeReleaseFile}`

		if [ X${nextReleaseName} = X ]
		then
			break
		fi

		lastUpgradeReleaseFile=${upgradeReleaseFile}

		upgradeReleaseFile=`
			0<<!!! cat - |
${UpgradeReleaseFiles}
!!!
			while read upgradeReleaseFile
			do
				upgradeReleaseName=\`
		upgradeinfo NameFor: ${upgradeReleaseFile}\`

				if [ X"${upgradeReleaseName}" = \
					X"${nextReleaseName}" ]
				then
					echo ${upgradeReleaseFile}
					exit 0
				fi
			done
			exit ${?}`

		if [ ${?} -ne 0 ]
		then
			0<<!!! 1>&2 cat -
${0}: Cannot determine successor to ${lastUpgradeReleaseFile}.
!!!
			exit 1
		fi
	done
	;;
NameFor: )
	echo ${UpgradeReleaseName}
	;;
DescriptionFor: )
	if [ X"${UpgradeReleaseDescription}" != X"" ]
	then
		0<<!!! cat -
${UpgradeReleaseDescription}
!!!
	fi
	;;
TapePACK_For: )
	echo ${UpgradeReleasePACK}
	;;
TapeDESC_For: )
	echo "${UpgradeReleaseDESC}"
	;;
TapeVER_For: )
	echo "${UpgradeReleaseVER}"
	;;
SaveFileListFor: )
	if [ X"${UpgradeReleaseSaveList}" != X"" ]
	then
		0<<!!! cat -
${UpgradeReleaseSaveList}
!!!
	fi
	;;
ObsoleteFileListFor: )
	if [ X"${UpgradeReleaseObsoleteList}" != X"" ]
	then
		0<<!!! cat -
${UpgradeReleaseObsoleteList}
!!!
	fi
	;;
NextReleaseNameFor: )
	echo ${UpgradeReleaseNextReleaseName}
	;;
ArchiveListFor: )
	echo ${UpgradeReleaseArchiveList}
	;;
ScriptFor: )
	if [ X"${UpgradeReleaseScript}" != X"" ]
	then
		0<<!!! cat -
${UpgradeReleaseScript}
!!!
	fi
	;;
esac
exit 0
