#ident	"@(#)install:upgrades/BS81.05/gettydefs.c	35.4"
/* Gettydefs Merge Program
 * Copyright 1988, Motorola, Inc.
 */

/*   This is a very easy update for gettydefs:
 *   simply append the following entry to the
 *   end of the gettydefs file:
 *
 * pc220# B9600 CLOCAL IGNBRK IGNPAR ISTRIP IXON IXOFF ECHO OPOST ONLCR # \
 * IGNBRK IGNPAR ISTRIP ICRNL IXON IXANY OPOST ONLCR B9600 CS8 CREAD CLOCAL ISIG ICANON ECHO ECHOE ECHOK #\nPC Login: #pc220
 *
 */

#include <sys/types.h>
#include <time.h>
#include <stdio.h>
#include <termio.h>
#include <gettydefs.h>

#define USAGE "Usage: %s [-v] [gettydefs...]\n"
#define OPTSTRING "v"

#define DEF_ROOT_DIR "/root"
#define DEF_GETTY_DEFS "/etc/gettydefs"

static char * my_name;
static int verbose = 0;
static char gettydefs [ 128 ];

int
main( argc, argv, envp )
int argc;
char *argv[], *envp[];
{
	extern char * optarg, * getenv (), * strcpy (), * strcat ();
	extern int optind;
	int update();

	int count;
	char * root;

	verbose = count = 0;
	my_name = argv[0];

	if ( ( root = getenv ( "RootDirectory" ) ) == NULL )
		root = DEF_ROOT_DIR;

	strcpy ( gettydefs, root );
	strcat ( gettydefs, DEF_GETTY_DEFS );

	for ( ;; ) {
		switch ( getopt ( argc, argv, OPTSTRING ) ) {
		case -1:
			break;
		case 'v':
			--verbose;
			continue;
		case '?':
			count ++;
			continue;
		}
		break;
	}

	if ( count ) {
		fprintf ( stderr, USAGE, my_name );
		return ( 2 );
	}

	if ( optind == argc )
		count = update ( gettydefs );
	else
		while ( optind < argc )
			count += update ( argv[ optind ++ ] );

	return ( count );
}

/*   Define the new gettydefs entry
 */

static GETTYDEFS_T new_gettydefs_entry = {
	"pc220",
	{
		IGNBRK | IGNPAR | ISTRIP | IXON | IXOFF,
		OPOST | ONLCR,
		B9600 | CLOCAL,
		ECHO,
		'\0',
		{
			'\0',
			'\0',
			'\0',
			'\0',
			'\0',
			'\0',
			'\0',
			'\0'
		}
	},
	{
		IGNBRK | IGNPAR | ISTRIP | ICRNL | IXON | IXANY,
		OPOST | ONLCR,
		B9600 | CS8 | CREAD | CLOCAL,
		ISIG | ICANON | ECHO | ECHOE | ECHOK,
		'\0',
		{
			'\0',
			'\0',
			'\0',
			'\0',
			'\0',
			'\0',
			'\0',
			'\0'
		}
	},
	"\nPC Login: ",
	"pc220"
};

int
update ( filename )
char * filename;
{
	GETTYDEFS_F * in_gettydefs_f, * out_gettydefs_f;
	GETTYDEFS_P gettydefs_p;
	char * tempname, * buffer;
	int count, already_done, write_now, any_change;
	time_t clock;

	extern char * malloc ();
	void update_error ();
	int termiocmp();

	if ( verbose )
		printf ( "%s:\n", filename );

	in_gettydefs_f = opengettydefs ( filename, "r" );
	if ( in_gettydefs_f == NULL ) {
		update_error ( filename );
		return ( 1 );
	}
	tempname = tempnam ( NULL, "gtyd" );
	if ( tempname == NULL ) {
		update_error ( NULL );
		closegettydefs ( in_gettydefs_f );
		return ( 1 );
	}
	out_gettydefs_f = opengettydefs ( tempname, "w" );
	if ( out_gettydefs_f == NULL ) {
		update_error ( NULL );
		free ( tempname );
		closegettydefs ( in_gettydefs_f );
		return ( 1 );
	}
	write_now = already_done = 0;
	clock = time ( NULL );

	fprintf ( out_gettydefs_f -> getty_fp, "# %s: updated: %s\n", my_name, ctime ( &clock ) );

	if ( ferror ( out_gettydefs_f -> getty_fp ) ) {
		update_error ( tempname );
		closegettydefs ( out_gettydefs_f );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		closegettydefs ( in_gettydefs_f );
		return ( 1 );
	}

	any_change = 0;
	while ( readgettydefs ( in_gettydefs_f, & gettydefs_p ) != NULL ) {
		if ( ( ! already_done ) && ( gettydefs_p . gfo_type == GFO_ENTRY ) &&
		     ( strcmp ( gettydefs_p . gfo_entry -> g_id, new_gettydefs_entry . g_id ) == 0 ) ) {
			if ( ( termiocmp ( & gettydefs_p . gfo_entry -> g_iflags, & new_gettydefs_entry . g_iflags ) == 0 ) &&
			     ( termiocmp ( & gettydefs_p . gfo_entry -> g_fflags, & new_gettydefs_entry . g_fflags ) == 0 ) &&
			     ( strcmp ( gettydefs_p . gfo_entry -> g_nextid, new_gettydefs_entry . g_nextid ) == 0 ) ) {
				if ( verbose )
					printf ( "\t%s entry already properly implemented.\n", new_gettydefs_entry . g_id );
			} else {
				if ( verbose )
					printf ( "\t%s entry found and updated.\n", new_gettydefs_entry . g_id );
				fputs ( "# ", out_gettydefs_f -> getty_fp );
				--write_now;
			}
			--already_done;
		}
		if ( writegettydefs ( out_gettydefs_f, & gettydefs_p ) ) {
			update_error ( tempname );
			closegettydefs ( out_gettydefs_f );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			closegettydefs ( in_gettydefs_f );
			return ( 1 );
		}
		if ( write_now ) {
			write_now = 0;

			gettydefs_p . gfo_type = GFO_ENTRY;
			gettydefs_p . gfo_entry = & new_gettydefs_entry;

			if ( writegettydefs ( out_gettydefs_f, & gettydefs_p ) ) {
				update_error ( tempname );
				closegettydefs ( out_gettydefs_f );
				if ( unlink ( tempname ) )
					update_error ( tempname );
				free ( tempname );
				closegettydefs ( in_gettydefs_f );
				return ( 1 );
			}
			any_change ++;
		}
	}
	if ( ferror ( in_gettydefs_f -> getty_fp ) ) {
		update_error ( filename );
		closegettydefs ( out_gettydefs_f );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		closegettydefs ( in_gettydefs_f );
		return ( 1 );
	}
	if ( ! already_done ) {
		if ( verbose )
			printf ( "\t%s entry appended to file.\n", new_gettydefs_entry . g_id );

		gettydefs_p . gfo_type = GFO_ENTRY;
		gettydefs_p . gfo_entry = & new_gettydefs_entry;

		if ( writegettydefs ( out_gettydefs_f, & gettydefs_p ) ) {
			update_error ( tempname );
			closegettydefs ( out_gettydefs_f );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			closegettydefs ( in_gettydefs_f );
			return ( 1 );
		}
		any_change ++;
	}
	if ( closegettydefs ( out_gettydefs_f ) ) {
		update_error ( tempname );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		closegettydefs ( in_gettydefs_f );
		return ( 1 );
	}
	if ( closegettydefs ( in_gettydefs_f ) ) {
		update_error ( filename );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		return ( 1 );
	}
	if ( ! any_change ) {
		if ( verbose )
			puts ( "\tNo changes required - no action taken." );

		if ( unlink ( tempname ) ) {
			update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
	} else {
		buffer = malloc( BUFSIZ );
		if ( buffer == NULL ) {
			update_error ( NULL );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
		sprintf ( buffer, "mv %s %s", tempname, filename );
		if ( system ( buffer ) ) {
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			free ( buffer );
			return ( 1 );
		}
		free ( buffer );
	}
	free ( tempname );
	return ( 0 );
}

static void
update_error ( string )
char * string;
{
	fprintf ( stderr, ( string == NULL ) ? "%s: " : "%s: %s: ", my_name, string );
	perror ( NULL );
	return;
}

static int
termiocmp ( termio_p, termio_q )
struct termio * termio_p, * termio_q;
{
	if ( ( termio_p -> c_iflag == termio_q -> c_iflag ) &&
	     ( termio_p -> c_oflag == termio_q -> c_oflag ) &&
	     ( termio_p -> c_cflag == termio_q -> c_cflag ) &&
	     ( termio_p -> c_lflag == termio_q -> c_lflag ) &&
	     ( termio_p -> c_line  == termio_q -> c_line  ) &&
	     ( strncmp ( termio_p -> c_cc, termio_q -> c_cc, NCC ) == 0 ) )
		return ( 0 );
	return ( 1 );
}
