# @(#)install:upgrades/BS83.08/perms.sh	35.3
# Copyright 1988, Motorola, Inc.
#
# This script will upgrade the /etc/perms file to /etc/filesys in support
# of new versions of MCD value added utilities: chk, dcpy, fmt, fs, mnt, 
# real, tt, and umnt.

PATH=$PATH:${RootDirectory}/usr/bin
OLDPERMS=$1
NEWPERMS=${RootDirectory}/etc/filesys
REL_FILESYS=${RootDirectory}/usr/BOS/rel/etc/filesys

#
# shell procedure to translate the /etc/perms file to /etc/filesys format.
#
xlateperms ()
{
  OPERMS=$1

  cat $OPERMS |\
   awk '	$1 !~ /^$|#|#.*|<real_dev>|\/dev\/02s7|\/dev\/03s7/ && \
		$2 !~ /default/ 				    \
					{printf "%-12s%-12s%-12s%-12s%-12s%s", $1, $2, $5, $3, $4, $6
					 for (i=7; i<=NF; i++) printf " %s", $i
					 printf "\n"}' > tmp.perms

  ls -l ${RootDirectory}/dev/dsk > tmp.DSKdir
  devs=`awk ' {print $1}' tmp.perms`
  for d in $devs
    do
  #
  #   Check for the existence of each device in the OPERMS file.  If found
  #   procede to find the link in /dev/dsk for the device.  If either the
  #   link or the device itself cannot be found, discard the entry.
  #   Note that the pattern matching of Major/minor device numbers is done 
  #   directly from the "ls -l" listing to ensure matching column positions.
  #
      ls -l ${RootDirectory}/${d} > /dev/null 2>&1
      if [ $? -ne 0 ]
        then
	  echo "WARNING: perms -> filesys translation failure; $d not found."
	  echo "         Entry for device $d discarded."
	  sed -e "\^$d^d" tmp.perms > tmp.perms2
        else
          pat=`ls -l ${RootDirectory}/${d} | cut -c34-40`
    	  dsk_entry=`grep "$pat" tmp.DSKdir`

    	  if [ "x$dsk_entry" = "x" ]
      	    then
	      echo "WARNING: perms -> filesys translation failure; device $d has"
	      echo "         no link in /dev/dsk.  Entry for device $d discarded."
	      sed -e "\^$d^d" tmp.perms > tmp.perms2
      	    else
    	      slice=`echo $dsk_entry | cut -d' ' -f10`
    	      sed -e "s^$d^$slice^" tmp.perms > tmp.perms2
    	  fi
      fi
      mv tmp.perms2 tmp.perms
  done

  # Rename and reformat the permissions for readibility.
  cat tmp.perms | \
   awk ' {printf "%-12s%-12s%-12s%-12s%-12s%s", $1, $2, $3, $4, $5, $6
	  for (i=7; i<=NF; i++) printf " %s", $i
	  printf "\n"}' > filesys.int
  rm -f tmp.DSKdir tmp.perms
}

cp ${REL_FILESYS} ${NEWPERMS}
#
# ASSUMPTION:  The file, NEWPERMS, is in place (i.e., the "template") before
#              this merge script is executed.

if [ -r ${OLDPERMS} ]
  then
    if [ -s ${OLDPERMS} ]
      then
	xlateperms ${OLDPERMS}
	if [ -s filesys.int ]
	  then
	    echo "#\n# Non-obsolete entries translated from the /etc/perms file during upgrade.\n#" >> ${NEWPERMS}
	    cat filesys.int >> ${NEWPERMS}
	fi
	rm -f filesys.int
    fi
    rm -f ${OLDPERMS}
fi
exit 0
