# @(#)install:upgrades/BS83.08/standard.sh	35.4
# Sysgen(1M) Standard Configuration File Upgrade Script
# Copyright 1988, Motorola, Inc.

OBSOLETE_COLLECTIONS=`0<<!!! cat -
du
vme330.tcp
!!!`

OBSOLETE_ITEMS=`0<<!!! cat -
kernel	NBLK1024,NBLK128,NBLK16,NBLK2048,NBLK256,NBLK4,NBLK4096,NBLK512,\
NBLK64,NLOG,NMUXLINK,NQUEUE,NSTREAM,NSTREVENT,NSTRPUSH,NUMTIMOD,SCKSANITY,\
SPCNT,STRCTLSZ,STRLOFRAC,STRMEDFRAC,STRMSGSZ
misc	sxt,sxt #1,sxtcnt,sxtcnt #1
!!!`

NEW_COLLECTIONS=`0<<!!! cat -
strbufs
streams
vme143
vme327
vme327c2
!!!`

NEW_ITEMS=`0<<!!! cat -
kernel		PHYSBSIZE,PHYSCNT,PHYSSPACE,TODCSYNC
vme147devs	d147maxblk
!!!`

beVerbose=false
usageError=false

while [ ${#} -ne 0 ]
do
	argument="${1}"

	case "${argument}" in
	-* )
		: continue
		;;
	* )
		break
		;;
	esac

	shift

	while :
	do
		case "${argument}" in
		-v* )
			beVerbose=true
			;;
		-C )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing config_dir argument.
!!!
				usageError=true
				break 2
			fi
			ConfigurationDirectory=${1}
			shift
			;;
		-C* )
			ConfigurationDirectory=`expr "${argument}" : '-C\(.*\)'`
			break
			;;
		-D )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing desc_dir argument.
!!!
				usageError=true
				break 2
			fi
			DescriptionDirectory=${1}
			shift
			;;
		-D* )
			DescriptionDirectory=`expr "${argument}" : '-D\(.*\)'`
			break
			;;
		-R )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing root_dir argument.
!!!
				usageError=true
				break 2
			fi
			RootDirectory=${1}
			shift
			;;
		-R* )
			RootDirectory=`expr "${argument}" : '-R\(.*\)'`
			break
			;;
		-S )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing sysgen_dir argument.
!!!
				usageError=true
				break 2
			fi
			SysgenDirectory=${1}
			shift
			;;
		-S* )
			SysgenDirectory=`expr "${argument}" : '-S\(.*\)'`
			shift
			;;
		-T )
			if [ ${#} -lt 1 ]
			then
				0<<!!! 1>&2 cat -
${0}: Missing tmpdir argument.
!!!
				usageError=true
				break 2
			fi
			TMPDIR=${1}
			shift
			;;
		-T* )
			TMPDIR=`expr "${argument}" : '-T\(.*\)'`
			break
			;;
		-- )
			break 2
			;;
		-* )
			0<<!!! 1>&2 cat -
${0}: '`expr "${argument}" : '\(-.\).*`' is not a supported switch.
!!!
			usageError=true
			break 2
			;;
		esac

		case "${argument}" in
		-? )
			break
			;;
		esac

		argument=-`expr "${argument}" : '-.\(.*\)'`
	done
done

if ${usageError}
then
	0<<!!! 1>&2 cat -
Usage: ${0} [-v] [-C config_dir] [-D desc_dir] [configuration]...
!!!
	exit 2
fi

RootDirectory=${RootDirectory:-/root}
TMPDIR=${TMPDIR:-${RootDirectory}/usr/tmp}

SysgenDirectory=\
${RootDirectory}/${SysgenDirectory:-usr/src/uts/m68k/sysgen}
ConfigurationDirectory=\
${RootDirectory}/${ConfigurationDirectory:-usr/src/uts/m68k/sysgen/machines}
DescriptionDirectory=\
${RootDirectory}/${DescriptionDirectory:-usr/src/uts/m68k/sysgen/descriptions}

if [ X"${NEW_COLLECTIONS}" != X"" ]
then
	NEW_COLLECTIONS=`
		echo "${NEW_COLLECTIONS}" |
		while read collection
		do
			if [ -f ${DescriptionDirectory}/${collection} ]
			then
				: Success
			else
				0<<!!! 1>&2 cat -
${0}: New collection '${collection}' not found; not enabled during merge.
!!!
				continue
			fi
			echo "${collection}"
		done`
fi

if [ X"${NEW_ITEMS}" != X"" ]
then
	NEW_ITEMS=`
		echo "${NEW_ITEMS}" |
		while read collection items
		do
			if [ -f ${DescriptionDirectory}/${collection} ]
			then
				: Success
			else
				0<<!!! 1>&2 cat -
${0}: Collection '${collection}' not found;
${0}: New items in collection '${collection}' not enabled during merge.
!!!
				continue
			fi

			OLDIFS=${IFS}
			IFS=,

			hasOneOrMoreItems=false

			for item in ${items}
			do
				1>/dev/null fgrep -x -l \
"Description Keyword:	\"${item}\"" ${DescriptionDirectory}/${collection}
				case ${?} in
				0 )
					: found it
					;;
				1 )
					0<<!!! 1>&2 cat -
${0}: ${collection}: New item '${item}' not found; not enabled during merge.
!!!
					continue
					;;
				2 | * )
					0<<!!! 1>&2 cat -
${0}: ${collection}: Unexpected fgrep(1) error occurred.
${0}: ${collection}: New item '${item}' not found; not enabled during merge.
!!!
					continue
					;;
				esac

				if ${hasOneOrMoreItems}
				then
					echo ",${item}\c"
				else
					echo "${collection}\t${item}\c"
					hasOneOrMoreItems=true
				fi
			done

			if ${hasOneOrMoreItems}
			then
				echo
			fi

			IFS=${OLDIFS}
		done`
fi


source=${TMPDIR}/sg0$$
destination=${TMPDIR}/sg1$$

# Merge all configurations unless otherwise instructed.

if [ ${#} -eq 0 ]
then
	set -- `ls ${ConfigurationDirectory}`
fi

trap "rm -f ${source} ${destination}; exit" 0
trap "trap '' 0; rm -f ${source} ${destination}; exit 1" 1 2 3 15

fatal_error ()
{
	error_number=${1}
	shift

	0<<!!! 1>&2 cat -
${0}: Unexpected ${*} error occurred (#${error_number}).
${0}: Configuration '${configuration}' has not been merged.
${0}: Terminating execution.
!!!
	kill $$
	exit 1
}

for configuration
do
	if ${beVerbose}
	then
		0<<!!! cat -
${0}: Merging configuration '${configuration}'.
!!!
	fi

	if 1>${source} cat ${ConfigurationDirectory}/${configuration}
	then
		: Success
	else
		fatal_error 0100 'cat(1)'
	fi

	if ${beVerbose}
	then
		0<<!!! cat -
${0}: ${configuration}: Disabling obsolete collections.
!!!
	fi

	if [ X"${OBSOLETE_COLLECTIONS}" != X"" ]
	then
		echo "${OBSOLETE_COLLECTIONS}"
	fi |
	while read collection
	do
		0<${source} 1>/dev/null fgrep -x \
			"Select Description File:	\"${collection}\""
		case ${?} in
		0 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: Disable obsolete collection '${collection}'.
!!!
			fi
			;;
		1 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: Obsolete collection '${collection}' not enabled.
!!!
			fi
			continue
			;;
		2 | * )
			fatal_error 0200 'fgrep(1)'
			;;
		esac

		if 0<<!!! 1>/dev/null ed - ${source}
\$a
Select Description File:
.
1
/^Select Description File:	"${collection}"
.,/^Select Description File/-1d
\$d
w
q
!!!
		then
			: Success
		else
			fatal_error 0300 'ed(1)'
		fi
	done

	if ${beVerbose}
	then
		0<<!!! cat -
${0}: ${configuration}: Disabling obsolete collection items.
!!!
	fi

	if [ X"${OBSOLETE_ITEMS}" != X"" ]
	then
		echo "${OBSOLETE_ITEMS}"
	fi |
	while read collection items
	do
		1>/dev/null fgrep -x -l \
"Select Description File:	\"${collection}\"" ${source}
		case ${?} in
		0 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: Removing obsolete items in collection '${collection}'.
!!!
			fi
			;;
		1 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: Cannot remove obsolete items \
in collection '${collection}';
${0}: ${configuration}: '${collection}' not enabled.
!!!
			fi
			continue
			;;
		2 | * )
			fatal_error 0400 'fgrep(1)'
			;;
		esac

		if 0<<!!! 1>/dev/null ed -
r ${source}
\$a
Select Description File:
.
/^Select Description File:	"${collection}"$/kx
'x
/^Select Description File:/ky
'x,'yw ${destination}
Q
!!!
		then
			: Success
		else
			fatal_error 0500 'ed(1)'
		fi

		OLDIFS=${IFS}
		IFS=,

		for item in ${items}
		do
			if ${beVerbose}
			then
				1>/dev/null fgrep -x -l \
"Select Description:	\"${item}\"" ${destination}
				case ${?} in
				0 )
					: found one
					;;
				1 )
					0<<!!! cat -
${0}: ${configuration}: ${collection}: Obsolete item '${item}' was not enabled.
!!!
					;;
				2 | * )
					fatal_error 0600 'fgrep(1)'
					;;
				esac
			fi

			while :
			do
				1>/dev/null fgrep -x -l \
"Select Description:	\"${item}\"" ${destination}
				case ${?} in
				0 )
					if ${beVerbose}
					then
						0<<!!! cat -
${0}: ${configuration}: ${collection}: Disabling obsolete item '${item}'.
!!!
					fi
					;;
				1 )
					: No more, break out.
					break
					;;
				2 | * )
					fatal_error 0700 'fgrep(1)'
					;;
				esac

				if 0<<!!! 1>/dev/null \
					ed - ${destination}
/^Select Description:	"${item}"$/kx
'x+1g/^Value:/d
'xd
w
q
!!!
				then
					: Success
				else
					fatal_error 0800 'ed(1)'
				fi
			done
		done

		IFS=${OLDIFS}

		if 0<<!!! 1>/dev/null ed - ${destination}
\$d
w
q
!!!
		then
			: Success
		else
			fatal_error 0850 'ed(1)'
		fi

		if 0<<!!! 1>/dev/null ed - ${source}
\$a
Select Description File:
.
/^Select Description File:	"${collection}"$/kx
'x-1r ${destination}
'x
'x,/^Select Description File:/-1d
\$d
w
q
!!!
		then
			: Success
		else
			fatal_error 0900 'ed(1)'
		fi
	done

	if ${beVerbose}
	then
		0<<!!! cat -
${0}: ${configuration}: Enabling new collections.
!!!
	fi

	if [ X"${NEW_COLLECTIONS}" != X"" ]
	then
		echo "${NEW_COLLECTIONS}"
	fi |
	while read collection
	do
		1>/dev/null fgrep -x -l \
"Select Description File:	\"${collection}\"" ${source} 
		case ${?} in
		0 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: New collection '${collection}' was already enabled.
!!!
			fi
			continue
			;;
		1 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: Enabling new collection '${collection}'.
!!!
			fi
			;;
		2 | * )
			fatal_error 1000 'fgrep(1)'
			;;
		esac

		0<${DescriptionDirectory}/${collection} 1>${destination} \
				grep '^Description Keyword:'
		case ${?} in
		0 | 1 )
			: got zero or more.
			;;
		2 | * )
			fatal_error 1100 'grep(1)'
			;;
		esac

		if 0<<!!! 1>/dev/null ed - ${source}
\$a
Select Description File:	"${collection}"
Select File's Version:	""
.
\$kx
\$r ${destination}
'x+1,\$s/^Description Keyword:/Select Description:/
w
q
!!!
		then
			: Success
		else
			fatal_error 1200 'ed(1)'
		fi
	done

	if ${beVerbose}
	then
		0<<!!! cat -
${0}: ${configuration}: Enabling new items found in existing collections.
!!!
	fi

	if [ X"${NEW_ITEMS}" != X"" ]
	then
		echo "${NEW_ITEMS}"
	fi |
	while read collection items
	do
		1>/dev/null fgrep -x -l \
"Select Description File:	\"${collection}\"" ${source}
		case ${?} in
		0 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: Enabling new items in collection '${collection}'.
!!!
			fi
			;;
		1 )
			if ${beVerbose}
			then
				0<<!!! cat -
${0}: ${configuration}: New items in collection '${collection}' not enabled.
${0}: ${configuration}: Collection '${collection}' is not enabled.
!!!
			fi
			continue
			;;
		2 | * )
			fatal_error 1300 'fgrep(1)'
			;;
		esac

		if 0<<!!! 1>/dev/null ed -
r ${source}
\$a
Select Description File:
.
/^Select Description File:	"${collection}"$/kx
'x
/^Select Description File:/-1ky
'x,'yw ${destination}
Q
!!!
		then
			: Success
		else
			fatal_error 1400 'ed(1)'
		fi

		OLDIFS="${IFS}"
		IFS=,

		for item in ${items}
		do
			1>/dev/null fgrep -x -l \
"Select Description:	\"${item}\"" ${destination}
			case ${?} in
			0 )
				if ${beVerbose}
				then
					0<<!!! cat -
${0}: ${configuration}: ${collection}: New item '${item}' was already enabled.
!!!
				fi
				continue
				;;
			1 )
				if ${beVerbose}
				then
					0<<!!! cat -
${0}: ${configuration}: ${collection}: Enabling new item '${item}'.
!!!
				fi
				;;
			2 | * )
				fatal_error 1500 'fgrep(1)'
				;;
			esac

			if 0<<!!! 1>/dev/null ed - ${destination}
\$a
Select Description:	"${item}"
.
w
q
!!!
			then
				: Success
			else
				fatal_error 1600 'ed(1)'
			fi

		done

		IFS="${OLDIFS}"

	   	if 0<<!!! 1>/dev/null ed - ${source}
\$a
Select Description File:
.
/^Select Description File:	"${collection}"$/kx
'x-1r ${destination}
'x
'x,/^Select Description File:/-1d
\$d
w
q
!!!
		then
			: Success
		else
			fatal_error 1700 'ed(1)'
		fi
	done

	if cp ${source} ${ConfigurationDirectory}/${configuration}
	then
		: Success
	else
		0<<!!! 1>&2 cat -
${0}: Unexpected cp(1) error occurred (#1800).
${0}: Configuration '${configuration}' may have been corrupted.
${0}: Terminating execution.
!!!
		exit 1
		#NOTREACHED
	fi
done
