#ident "@(#)install:upgrades/BS83.08/utmp.c	35.1"

/* BS83.08 Utmp/Wtmp Upgrade Program
 * Copyright 1988, Motorola, Inc.
 */

#include <sys/types.h>
#include <stdio.h>

#define DEF_ROOT_DIR "/root"

#ifndef TARGET
#define TARGET "/etc/utmp"
#endif /* TARGET */

#define	UTMP TARGET

#define USAGE "Usage: %s [-v] [utmp...]\n"
#define OPTSTRING "v"

static int verbose = 0;
static char * my_name;
static char utmp [ 128 ];

#define	ut_name	ut_user

struct utmp
{
	char ut_user[8];		/* User login name */
	char ut_id[4]; 			/* /etc/lines id(usually line #) */
	char ut_line[12];		/* device name (console, lnxx) */
	short ut_pid;			/* process id */
	short ut_type; 			/* type of entry */
	struct exit_status
	{
		short e_termination;	/* Process termination status */
		short e_exit;		/* Process exit status */
	}
	ut_exit;			/* The exit status of a process
					 * marked as DEAD_PROCESS.
					 */
	time_t ut_time;			/* time entry was made */
	char ut_host[16];		/* host name, if remote */
};

struct outmp
{
	char ut_user[8];		/* User login name */
	char ut_id[4]; 			/* /etc/lines id(usually line #) */
	char ut_line[12];		/* device name (console, lnxx) */
	short ut_pid;			/* process id */
	short ut_type; 			/* type of entry */
	struct oexit_status
	{
		short e_termination;	/* Process termination status */
		short e_exit;		/* Process exit status */
	}
	ut_exit;			/* The exit status of a process
					 * marked as DEAD_PROCESS.
					 */
	time_t ut_time;			/* time entry was made */
};

/*	Definitions for ut_type						*/

#define	EMPTY		0
#define	RUN_LVL		1
#define	BOOT_TIME	2
#define	OLD_TIME	3
#define	NEW_TIME	4
#define	INIT_PROCESS	5	/* Process spawned by "init" */
#define	LOGIN_PROCESS	6	/* A "getty" process waiting for login */
#define	USER_PROCESS	7	/* A user process */
#define	DEAD_PROCESS	8
#define	ACCOUNTING	9

#define	UTMAXTYPE	ACCOUNTING	/* Largest legal value of ut_type */

/*	Special strings or formats used in the "ut_line" field when	*/
/*	accounting for something other than a process.			*/
/*	No string for the ut_line field can be more than 11 chars +	*/
/*	a NULL in length.						*/

#define	RUNLVL_MSG	"run-level %c"
#define	BOOT_MSG	"system boot"
#define	OTIME_MSG	"old time"
#define	NTIME_MSG	"new time"


/* Reads in records from standard in.  Records are presumed to
** all be of type struct outmp.  Writes them back out in struct
** utmp format (w/extra 16 bytes of null in new ut_host field)
** Thus, this works for utmp and wtmp files, and any others 
** which are simple concatenations of outmp structs....
*/

int
main( argc, argv, envp )
int argc;
char *argv[], *envp[];
{
	extern char * optarg, * getenv (), * strcpy (), * strcat ();
	extern int optind;

	int update();

	int count;
	char * root;

	verbose = count = 0;
	my_name = argv[0];

	if ( ( root = getenv ( "RootDirectory" ) ) == NULL )
		root = DEF_ROOT_DIR;

	strcpy ( utmp, root );
	strcat ( utmp, UTMP );

	for ( ; ; ) {
		switch ( getopt ( argc, argv, OPTSTRING ) ) {
		case -1:
			break;
		case 'v':
			--verbose;
			continue;
		case '?':
			count ++;
			continue;
		}
		break;
	}

	if ( count ) {
		fprintf ( stderr, USAGE, my_name );
		return ( 2 );
	}

	if ( optind == argc )
		count = update ( utmp );
	else
		while ( optind < argc )
			count += update ( argv[ optind ++ ] );

	return ( count );
}

static int
update ( filename )
char * filename;
{
	extern FILE * fopen ();
	extern char * tempnam (), * memset ();
	extern char * malloc ();
	extern void free ();
	extern int unlink (), sprintf (), system (), fclose ();
	void update_error ();


	static union {
		struct outmp outmp;
		struct utmp utmp;
	} buffer;

#	define old_utmp buffer.outmp
#	define new_utmp buffer.utmp

	int count;
	char * tempname, * cmd_buffer;
	FILE * in_utmp_file, * out_utmp_file;

	count = 0;

	if ( verbose )
		printf ( "%s:\n", filename );

	if ( ( in_utmp_file = fopen ( filename, "r" ) ) == NULL ) {
		update_error ( filename );
		return ( 1 );
	}

	if ( ( tempname = tempnam ( NULL, "utmp" ) ) == NULL ) {
		update_error ( NULL );
		fclose ( in_utmp_file );
		return ( 1 );
	}

	if ( ( out_utmp_file = fopen ( tempname, "w" ) ) == NULL ) {
		update_error ( tempname );
		free ( tempname );
		fclose ( in_utmp_file );
		return ( 1 );
	}

	memset ( ( char * ) & buffer, '\0', sizeof ( buffer ) );

	while ( fread ( & old_utmp, sizeof ( old_utmp ),
						1, in_utmp_file ) == 1 ) {
		count ++;
		if ( fwrite ( & new_utmp, sizeof ( new_utmp ),
						1, out_utmp_file ) != 1 ) {
			update_error ( tempname );
			fclose ( out_utmp_file );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			fclose ( in_utmp_file );
			return ( 1 );
		}
	}

	if ( ! feof ( in_utmp_file ) ) {
		update_error ( filename );
		fclose ( out_utmp_file );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		fclose ( in_utmp_file );
		return ( 1 );
	}

	if ( fclose ( out_utmp_file ) ) {
		update_error ( tempname );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		fclose ( in_utmp_file );
		return ( 1 );
	}

	if ( fclose ( in_utmp_file ) ) {
		update_error ( tempname );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		return ( 1 );
	}

	if ( ( cmd_buffer = malloc( BUFSIZ ) ) == NULL ) {
		update_error ( NULL );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		return ( 1 );
	}

	sprintf ( cmd_buffer, "mv %s %s", tempname, filename );

	if ( system ( cmd_buffer ) ) {
		fprintf ( stderr, "%s: mv(1) failed.\n", my_name );
		free ( cmd_buffer );
		unlink ( tempname );
		free ( tempname );
		return ( 1 );
	}

	free ( cmd_buffer );
	free ( tempname );

	if ( verbose )
		fprintf ( stderr, "\tUpgraded %d records.\n", count);

	return ( 0 );
}

static void
update_error ( string )
char * string;
{
	fprintf ( stderr, ( string == NULL ) ? "%s: " : "%s: %s: ",
			my_name, string );
	perror ( NULL );
	return;
}
