#ident	"@(#)install:upgrades/FE83.10/Dialers.c	35.4"
/* Dialers Merge Program
 * Copyright 1988, Motorola, Inc.
 */

/*   The following is the diff(1)'s of old ( < ) and
 *   new ( > ) Dialers files
 *
 *   < nls "" "" NLPS:000:001:1\N\c
 *   ---
 *   > nls "" "" NLPS:000:001:101\N\c
 *
 *   < acu5212 "" "" \d\cEN\c : D\D\r\c COMPLETE
 *   ---
 *   > acu5212 =W-D "" \dQOG0\dEN\c : D\T\r\c PLETE
 *
 *   < intacu "" "" \dQ\c "" \dEN\c + AT\r\c : D\D\r\c DIALING "" COMPLETE
 *   ---
 *   > intacu ==-, "" \K\dEN\dAT\r\c : Q\c + \021\c "" \pEN\c + \0041 + AT\r\c : D\T\r\c COMPLETE
 *
 */
#include <sys/types.h>
#include <time.h>
#include <stdio.h>
#include <Dialers.h>

#define USAGE "Usage: %s [-v] [Dialers...]\n"
#define OPTSTRING "v"

#define DEF_ROOT_DIR "/root"
#define DEF_DIALERS "/usr/lib/uucp/Dialers"

static char * my_name;
static int verbose = 0;
static char dialers [ 128 ];

int
main( argc, argv, envp )
int argc;
char *argv[], *envp[];
{
	extern char * optarg, * getenv (), * strcpy (), * strcat ();
	extern int optind;

	int update();

	int count;
	char * root;

	verbose = count = 0;
	my_name = argv[0];

	if ( ( root = getenv ( "RootDirectory" ) ) == NULL )
		root = DEF_ROOT_DIR;

	strcpy ( dialers, root );
	strcat ( dialers, DEF_DIALERS );

	for ( ;; ) {
		switch ( getopt ( argc, argv, OPTSTRING ) ) {
		case -1:
			break;
		case 'v':
			--verbose;
			continue;
		case '?':
			count ++;
			continue;
		}
		break;
	}

	if ( count ) {
		fprintf ( stderr, USAGE, my_name );
		return ( 2 );
	}

	if ( optind == argc )
		count = update ( dialers );
	else
		while ( optind < argc )
			count += update ( argv[ optind ++ ] );

	return ( count );
}

static char *nls_arps[] = {
	"nls",
	"\"\"",
	"\"\"",
	"NLPS:000:001:101\\N\\c",
	NULL
};

static char *acu5212_arps[] = {
	"acu5212",
	"=W-D",
	"\"\"",
	"\\dQOG0\\dEN\\c",
	":",
	"D\\T\r\\c",
	"PLETE",
	NULL
};

static char *intacu_arps[] = {
	"intacu",
	"==-,",
	"\"\"",
	"\\K\\dEN\\dAT\r\\c",
	":",
	"Q\\c",
	"+",
	"\021\\c",
	"\"\"",
	"\\pEN\\c",
	"+",
	"\0041",
	"+",
	"AT\r\\c",
	":",
	"D\\T\r\\c",
	"COMPLETE",
	NULL
};

#define narpof(arps) ( ( sizeof ( arps ) / sizeof ( char * ) ) - 1 )

static struct entry_struct {
	char **entry_arps;
	int entry_narp;
} entries[] = {
	{ nls_arps,	narpof ( nls_arps )	},
	{ acu5212_arps,	narpof ( acu5212_arps )	},
	{ intacu_arps,	narpof ( intacu_arps )	},
	{ NULL,		0			}
};

int
update ( filename )
char * filename;
{
	DIALERS_T * in_dialers_p, * out_dialers_p;
	char * tempname, * buffer, ** arp_a, ** arp_b;
	int count, any_changes;
	struct entry_struct *entry_p;
	time_t clock;

	void update_error ();
	static char *arps[ D_MAX ];
	extern char * malloc ();

	if ( verbose )
		printf ( "%s:\n", filename );

	in_dialers_p = openDialers ( filename, "r" );
	if ( in_dialers_p == NULL ) {
		update_error ( filename );
		return ( 1 );
	}
	tempname = tempnam ( NULL, "r3v3D" );
	if ( tempname == NULL ) {
		update_error ( NULL );
		closeDialers ( in_dialers_p );
		return ( 1 );
	}
	out_dialers_p = openDialers ( tempname, "w" );
	if ( out_dialers_p == NULL ) {
		update_error ( NULL );
		free ( tempname );
		closeDialers ( in_dialers_p );
		return ( 1 );
	}
	clock = time ( NULL );
	fprintf ( out_dialers_p -> dial_fp, "# %s: updated: %s", my_name, ctime ( &clock ) );

	any_changes = 0;
	while ( ( count = readDialers ( in_dialers_p, arps, D_MAX ) ) > 0 ) {
		entry_p = entries;
		while ( entry_p -> entry_arps != NULL )
			if ( strcmp ( entry_p -> entry_arps[ 0 ], arps[ D_TYPE ] ) == 0 ) {
				arp_a = & entry_p -> entry_arps [ 1 ];
				arp_b = & arps [ 1 ];
				while ( ( * arp_a != NULL ) && ( * arp_b != NULL ) &&
					( strcmp ( * arp_a, * arp_b ) == 0 ) ) {
						arp_a ++;
						arp_b ++;
				}
				if ( ( * arp_a == NULL ) && ( * arp_b == NULL ) ) {
					entry_p ++;
					continue;
				}
				if ( verbose )
					printf ( "\tupdating '%s' entry.\n", arps[ D_TYPE ] );

				fputs ( "# ", out_dialers_p -> dial_fp );
				any_changes ++;
				break;
			} else
				entry_p ++;

		writeDialers ( out_dialers_p, arps, count );

		if ( entry_p -> entry_arps != NULL )
			writeDialers ( out_dialers_p, entry_p -> entry_arps, entry_p -> entry_narp );
	}
	if ( count < 0 ) {
		update_error ( filename );
		closeDialers ( out_dialers_p );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		closeDialers ( in_dialers_p );
		return ( 1 );
	}
	if ( closeDialers ( out_dialers_p ) ) {
		update_error ( tempname );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		closeDialers ( in_dialers_p );
		return ( 1 );
	}
	if ( closeDialers ( in_dialers_p ) ) {
		update_error ( filename );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		return ( 1 );
	}
	if ( any_changes ) {
		buffer = malloc( BUFSIZ );
		if ( buffer == NULL ) {
			update_error ( NULL );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
		sprintf ( buffer, "mv %s %s", tempname, filename );
		if ( system ( buffer ) ) {
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			free ( buffer );
			return ( 1 );
		}
		free ( buffer );
	} else {
		if ( verbose )
			puts ( "\tNo changes required.  No action taken." );

		if ( unlink ( tempname ) ) {
			update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
	}
	free ( tempname );
	return ( 0 );
}

static void
update_error ( string )
char * string;
{
	fprintf ( stderr, ( string == NULL ) ? "%s: " : "%s: %s: ", my_name, string );
	perror ( NULL );
	return;
}
