#ident	"@(#)install:upgrades/FE83.10/group.c	35.3"
/* Group Merge Program
 * Copyright 1988, Motorola, Inc.
 */

#include <sys/types.h>
#include <stdio.h>
#include <group.h>
#include <memory.h>
#include <pwd.h>
#include <passwd.h>
#include <string.h>

#define G_ARRAY_INC 16

#define DEF_ROOT_DIR "/root"
#define USER_GROUP "/etc/group"
#define USER_PASSWD "/etc/passwd"

#define USAGE "Usage: %s [-v] [-N newgroup] [-P passwd] [group...]\n"
#define OPTSTRING "vN:P:"

static int verbose = 0;
static char * my_name;
static char group [ 128 ];
static char passwd [ 128 ];

static GROUP_T
	rootgroup = { "root", "", 0, "root", NULL },
	othergroup = { "other", "", 1, NULL },
	bingroup = { "bin", "", 2, "root", "bin", "daemon", NULL },
	sysgroup = { "sys", "", 3, "root", "bin", "sys", "adm", NULL },
	admgroup = { "adm", "", 4, "root", "adm", "daemon", "listen", NULL },
	uucpgroup = { "uucp", "", 5, "uucp", NULL },
	mailgroup = { "mail", "", 6, "root", NULL },
	nuucpgroup = { "nuucp", "", 10, "nuucp", NULL },
	daemongroup = { "daemon", "", 12, "root", "daemon", NULL },
	servicegroup = { "service", "", 99, "service", NULL };

static GROUP_T * new_group_entries [] = {
	& rootgroup, & othergroup, & bingroup, & sysgroup,
	& admgroup, & uucpgroup, & mailgroup, & nuucpgroup,
	& daemongroup, & servicegroup, NULL
};

static GROUP_T ** new_group_array = new_group_entries;

int
main( argc, argv, envp )
int argc;
char *argv[], *envp[];
{
	extern char * optarg, * getenv (), * strcpy (), * strcat ();
	extern int optind;

	GROUP_T ** read_group_array ();
	int check_user_group_8 (), update();
	void free_group_array ();

	int count;
	char * root;

	verbose = count = 0;
	my_name = argv[0];

	if ( ( root = getenv ( "RootDirectory" ) ) == NULL )
		root = DEF_ROOT_DIR;

	strcpy ( group, root );
	strcat ( group, USER_GROUP );

	strcpy ( passwd, root );
	strcat ( passwd, USER_PASSWD );

	for ( ; ; ) {
		switch ( getopt ( argc, argv, OPTSTRING ) ) {
		case -1:
			break;
		case 'N':
			new_group_array = read_group_array ( optarg );
			if ( new_group_array == NULL )
				return ( 1 );
			continue;
		case 'P':
			strcpy ( passwd, optarg );
			continue;
		case 'v':
			--verbose;
			continue;
		case '?':
			count ++;
			continue;
		}
		break;
	}

	if ( count ) {
		fprintf ( stderr, USAGE, my_name );
		return ( 2 );
	}

	if ( check_user_group_8 () )
		return ( 1 );

	if ( optind == argc )
		count = update ( group );
	else
		while ( optind < argc )
			count += update ( argv[ optind ++ ] );

	return ( count );
}

static int
no_user_group_8;

static int
check_user_group_8 ()
{
	PASSWD_F * passwd_f;
	PASSWD_T * passwd_t;
	void update_error ();

	passwd_f = openpasswd ( passwd, "r" );
	if ( passwd_f == NULL ) {
		update_error ( passwd );
		return ( 1 );
	}
	no_user_group_8 = 1;
	while ( ( passwd_t = readpasswd ( passwd_f ) ) != NULL )
		if ( passwd_t -> pw_gid == 8 )
			no_user_group_8 = 0;

	if ( ! feof ( passwd_f -> passwd_fp ) ) {
		update_error ( passwd );
		if ( closepasswd ( passwd_f ) )
			update_error ( passwd );
		return ( 1 );
	}
	if ( closepasswd ( passwd_f ) ) {
		update_error ( passwd );
		return ( 1 );
	}
	return ( 0 );
}

static int
update ( filename )
char * filename;
{
	GROUP_F * in_group_f, * out_group_f;
	GROUP_T ** new_group_a, ** group_a, ** group_array, ** old_group_array;
	char * tempname, * buffer;
	int count, any_changes;

	extern char * malloc ();
	extern void free (), qsort ();
	GROUP_T ** read_group_array (), ** expand_group_array (), * dup_group ();
	void free_group_array (), update_error ();
	int group_p_cmp ();

	if ( verbose )
		printf ( "%s:\n", filename );

	if ( ( group_array = read_group_array ( filename ) ) == NULL )
		return ( 1 );

	count = 0;
	for ( group_a = group_array; * group_a != NULL; group_a ++ )
		count ++;

	/*   Combine the two group collections together, sort them, and then out.
	 *   Remove group 'rje' if possible.  Remove new groups which are already defined
	 *   on the user's system.
	 */
	any_changes = 0;
	for ( new_group_a = new_group_array; * new_group_a != NULL; new_group_a ++ ) {
		for ( group_a = group_array; * group_a != NULL; group_a ++ )
			if ( ( strcmp ( ( * new_group_a ) -> group_name, ( * group_a ) -> group_name ) == 0 ) ||
			     ( ( * new_group_a ) -> group_id == ( * group_a ) -> group_id ) )
				break;
		if ( * group_a == NULL ) {
			group_array = expand_group_array ( old_group_array = group_array );
			if ( group_array == NULL ) {
				update_error ( NULL );
				free_group_array ( old_group_array );
				return ( 1 );
			}
			if ( verbose )
				printf ( "\tadding group '%s' (id = %d).\n",
					( * new_group_a ) -> group_name, ( * new_group_a ) -> group_id );
			if ( ( group_array [ count ++ ] = dup_group ( * new_group_a ) ) == NULL ) {
				update_error ( NULL );
				free_group_array ( group_array );
				return ( 1 );
			}
			any_changes ++;
		}
	}

	qsort ( group_array, count, sizeof ( GROUP_T * ), group_p_cmp );

	tempname = tempnam ( NULL, "grp" );
	if ( tempname == NULL ) {
		update_error ( NULL );
		free_group_array ( group_array );
		return ( 1 );
	}
	out_group_f = opengroup ( tempname, "w" );
	if ( out_group_f == NULL ) {
		update_error ( tempname );
		free ( tempname );
		free_group_array ( group_array );
		return ( 1 );
	}
	for ( group_a = group_array; * group_a != NULL ; group_a ++ ) {
		if ( ( no_user_group_8 ) &&
		     ( strcmp ( "rje", ( * group_a ) -> group_name ) == 0 ) &&
		     ( ( * group_a ) -> group_id == 8 ) &&
		     ( strcmp ( ( * group_a ) -> group_members [ 0 ], "rje" ) == 0 ) &&
		     ( strcmp ( ( * group_a ) -> group_members [ 1 ], "shqer" ) == 0 ) &&
		     ( ( * group_a ) -> group_members [ 2 ] == NULL ) ) {
			if ( verbose )
				puts ( "\tremoving 'rje' group." );
			any_changes ++;
			continue; /* skip this entry */
		}
		if ( writegroup ( out_group_f, * group_a ) ) {
			update_error ( tempname );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			free_group_array ( group_array );
			return ( 1 );
		}
	}
	free_group_array ( group_array );

	if ( closegroup ( out_group_f ) ) {
		update_error ( tempname );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		return ( 1 );
	}
	if ( ! any_changes ) {
		if ( verbose )
			puts ( "\tNo changes required - no action taken." );
		if ( unlink ( tempname ) ) {
			update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
	} else {
		buffer = malloc( BUFSIZ );
		if ( buffer == NULL ) {
			update_error ( NULL );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
		sprintf ( buffer, "mv %s %s", tempname, filename );
		if ( system ( buffer ) ) {
			free ( buffer );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
		free ( buffer );
	}
	free ( tempname );
	return ( 0 );
}

static GROUP_T **
read_group_array ( filename )
{
	GROUP_T * group_t, ** group_array, ** old_group_array;
	GROUP_F * group_f;
	int count;

	GROUP_T * dup_group (), ** create_group_array (), ** expand_group_array ();
	void free_group_array (), update_error ();

	group_f = opengroup ( filename, "r" );
	if ( group_f == NULL ) {
		update_error ( filename );
		return ( NULL );
	}
	group_array = create_group_array ();
	if ( group_array == NULL ) {
		update_error ( NULL );
		if ( closegroup ( group_f ) )
			update_error ( filename );
		return ( NULL );
	}
	count = 0;
	while ( ( group_t = readgroup ( group_f ) ) != NULL ) {
		group_array = expand_group_array ( old_group_array = group_array );
		if ( group_array == NULL ) {
			update_error ( NULL );
			free_group_array ( old_group_array );
			if ( closegroup ( group_f ) )
				update_error ( filename );
			return ( NULL );
		}
		if ( ( group_array [ count ++ ] = dup_group ( group_t ) ) == NULL ) {
			update_error ( NULL );
			if ( closegroup ( group_f ) )
				update_error ( filename );
			free_group_array ( group_array );
			return ( NULL );
		}
	}
	if ( ! feof ( group_f -> grp_fp ) ) {
		update_error ( filename );
		closegroup ( group_f );
		free_group_array ( group_array );
		return ( NULL );
	}
	if ( closegroup ( group_f ) ) {
		update_error ( filename );
		free_group_array ( group_array );
		return ( NULL );
	}
	return ( group_array );
}

typedef struct {
	int	ga_size;	/* total size in bytes */
	int	ga_avail;	/* entries available */
	GROUP_T * ga_first;	/* first entry */
} _GROUP_ARRAY;

static GROUP_T **
create_group_array ()
{
	_GROUP_ARRAY  * _group_array;
	extern char * malloc ();

	_group_array = ( _GROUP_ARRAY * ) malloc ( sizeof ( _GROUP_ARRAY ) + G_ARRAY_INC * sizeof ( GROUP_T * ) );
	if ( _group_array == NULL )
		return ( NULL );
	
	_group_array -> ga_size = sizeof ( _GROUP_ARRAY ) + G_ARRAY_INC * sizeof ( GROUP_T * );
	_group_array -> ga_avail = G_ARRAY_INC;

	memset ( & _group_array -> ga_first, '\0', ( G_ARRAY_INC + 1 ) * sizeof ( GROUP_T * ) );

	return ( & _group_array -> ga_first );
}

static GROUP_T **
expand_group_array ( group_array )
GROUP_T ** group_array;
{
	_GROUP_ARRAY * _group_array;
	int old_size;
	extern char * realloc ();

	_group_array = ( _GROUP_ARRAY * ) ( ( ( char * ) group_array ) - ( sizeof ( int ) * 2 ) );
	if ( ! _group_array -> ga_avail ) {
		old_size = _group_array -> ga_size;
		_group_array -> ga_size += ( G_ARRAY_INC * sizeof ( GROUP_T * ) );

		_group_array = ( _GROUP_ARRAY * ) realloc ( _group_array, _group_array -> ga_size );
		if ( _group_array == NULL )
			return ( NULL );

		memset ( ( ( char * ) _group_array ) + old_size, '\0', G_ARRAY_INC * sizeof ( GROUP_T * ) );
		_group_array -> ga_avail = G_ARRAY_INC;
	}
	( _group_array -> ga_avail ) -- ;
	return ( & _group_array -> ga_first );
}

static void
free_group_array ( group_array )
GROUP_T ** group_array;
{
	extern void free ();
	void free_group ();

	GROUP_T ** group_a;

	for ( group_a = group_array; * group_a != NULL; group_a ++ )
		free_group ( * group_a );
	free ( group_array );
}

static int
group_p_cmp ( group_a, group_b )
GROUP_T ** group_a, ** group_b ;
{
	return ( ( * group_a ) -> group_id - ( * group_b ) -> group_id );
}

static GROUP_T *
dup_group ( group_t )
GROUP_T * group_t;
{
	GROUP_T * n_group_t;
	char ** member_p, ** member_q, * buffer_p;
	int members_length;

	members_length = 0;
	for ( member_p = group_t -> group_members; * member_p != NULL; member_p ++ )
		members_length += strlen ( * member_p ) + 1;

	buffer_p = malloc ( sizeof ( GROUP_T ) + strlen ( group_t -> group_name ) + 1 +
			strlen ( group_t -> group_passwd ) + 1 + members_length );

	if ( buffer_p == NULL ) {
		update_error ( NULL );
		return ( NULL );
	}
	n_group_t = ( GROUP_T * ) buffer_p;
	buffer_p += sizeof ( GROUP_T );

	n_group_t -> group_name = strcpy ( buffer_p, group_t -> group_name );
	buffer_p += strlen ( group_t -> group_name ) + 1;
	n_group_t -> group_passwd = strcpy ( buffer_p, group_t -> group_passwd );
	buffer_p += strlen ( group_t -> group_passwd ) + 1;
	n_group_t -> group_id = group_t -> group_id;

	member_q = n_group_t -> group_members;
	for (member_p = group_t -> group_members; * member_p != NULL; member_p ++ ) {
		* member_q ++ = strcpy ( buffer_p, * member_p );
		buffer_p += strlen ( * member_p ) + 1;
	}
	* member_q = NULL;
	return ( n_group_t );
}

static void
free_group ( group_t )
GROUP_T * group_t;
{
	extern void free () ;

	free ( ( char * ) group_t );
}

static void
update_error ( string )
char * string;
{
	fprintf ( stderr, ( string == NULL ) ? "%s: " : "%s: %s: ", my_name, string );
	perror ( NULL );
	return;
}
