#ident	"@(#)install:upgrades/FE83.10/passwd.c	35.2"
/* Passwd Merge Program
 * Copyright 1988, Motorola, Inc.
 */

#include <sys/types.h>
#include <stdio.h>
#include <passwd.h>
#include <memory.h>
#include <pwd.h>
#include <string.h>

#define P_ARRAY_INC 16
#define DEF_ROOT_DIR "/root"
#define USER_PASSWD "/etc/passwd"
#define USAGE "Usage: %s [-v] [-N newpasswd] [passwd...]\n"
#define OPTSTRING "vN:"

static int verbose = 0;
static char * my_name;
static char passwd [ 128 ];

static PASSWD_T
	entry_root = {
		"root",
		"",
		0, 1,
		"",
		"0000-Admin(0000)",
		"",
		"/",
		""
	},
	entry_daemon = {
		"daemon",
		"",
		1, 12,
		"",
		"0000-Admin(0000)",
		"",
		"/",
		""
	},
	entry_bin = {
		"bin",
		"",
		2, 2,
		"",
		"0000-Admin(0000)",
		"",
		"/bin",
		""
	},
	entry_sys = {
		"sys",
		"",
		3, 3,
		"",
		"0000-Admin(0000)",
		"",
		"/usr/src",
		""
	},
	entry_adm = {
		"adm",
		"",
		4, 4,
		"",
		"0000-Admin(0000)",
		"",
		"/usr/adm",
		""
	},
	entry_uucp = {
		"uucp",
		"",
		5, 5,
		"",
		"0000-uucp(0000)",
		"",
		"/usr/lib/uucp",
		""
	},
	entry_listen = {
		"listen",
		"",
		7, 4,
		"",
		"listen process for rfs",
		"",
		"/usr/net/nls",
		""
	},
	entry_nuucp = {
		"nuucp",
		"",
		10, 10,
		"",
		"0000-uucp(0000)",
		"",
		"/usr/spool/uucppublic",
		"/usr/lib/uucp/uucico"
	},
	entry_trouble = {
		"trouble",
		"",
		70, 1,
		"",
		"trouble(0000)",
		"",
		"/usr/lib/trouble",
		""
	},
	entry_lp = {
		"lp",
		"",
		71, 2,
		"",
		"0000-lp(0000)",
		"",
		"/usr/spool/lp",
		""
	},
	entry_service = {
		"service",
		"9zm6A5zHYaOPM",
		99, 99,
		"",
		"CSO login",
		"",
		"/u/service",
		""
	},
	entry_setup = {
		"setup",
		"",
		0, 0,
		"",
		"general system administration",
		"",
		"/usr/admin",
		"/bin/rsh"
	},
	entry_powerdown = {
		"powerdown",
		"",
		0, 0,
		"",
		"general system administration",
		"",
		"/usr/admin",
		"/bin/rsh"
	},
	entry_sysadm = {
		"sysadm",
		"",
		0, 0,
		"",
		"general system administration",
		"",
		"/usr/admin",
		"/bin/rsh"
	},
	entry_checkfsys = {
		"checkfsys",
		"",
		0, 0,
		"",
		"check diskette file system",
		"",
		"/usr/admin",
		"/bin/rsh"
	},
	entry_makefsys = {
		"makefsys",
		"",
		0, 0,
		"",
		"make diskette file system",
		"",
		"/usr/admin",
		"/bin/rsh"
	},
	entry_mountfsys = {
		"mountfsys",
		"",
		0, 0,
		"",
		"mount diskette file system",
		"",
		"/usr/admin",
		"/bin/rsh"
	},
	entry_umountfsys = {
		"umountfsys",
		"",
		0, 0,
		"",
		"unmount diskette file system",
		"",
		"/usr/admin",
		"/bin/rsh"
	},
	entry_br = {
		"br",
		"",
		0, 0,
		"",
		"backup & restore",
		"",
		"/usr/admin",
		"/bin/rsh"
	};

static PASSWD_T * new_passwd_entries [] = {
	& entry_root, & entry_daemon, & entry_bin, & entry_sys,
	& entry_adm, & entry_uucp, & entry_listen, & entry_nuucp,
	& entry_trouble, & entry_lp, & entry_service, & entry_setup,
	& entry_powerdown, & entry_sysadm, & entry_checkfsys, & entry_makefsys,
	& entry_mountfsys, & entry_umountfsys, & entry_br,
	NULL
};

static PASSWD_T ** new_passwd_array = new_passwd_entries;

int
main( argc, argv, envp )
int argc;
char *argv[], *envp[];
{
	extern char * optarg, * getenv (), * strcpy (), * strcat ();
	extern int optind;

	PASSWD_T ** read_passwd_array ();
	int update();
	void free_passwd_array ();

	int count;
	char * root;

	verbose = count = 0;
	my_name = argv[0];

	if ( ( root = getenv ( "RootDirectory" ) ) == NULL )
		root = DEF_ROOT_DIR;

	strcpy ( passwd, root );
	strcat ( passwd, USER_PASSWD );

	for ( ; ; ) {
		switch ( getopt ( argc, argv, OPTSTRING ) ) {
		case -1:
			break;
		case 'N':
			if ( ( new_passwd_array = read_passwd_array ( optarg ) ) == NULL )
				return ( 1 );
			continue;
		case 'v':
			--verbose;
			continue;
		case '?':
			count ++;
			continue;
		}
		break;
	}

	if ( count ) {
		fprintf ( stderr, USAGE, my_name );
		return ( 2 );
	}

	if ( optind == argc )
		count = update ( passwd );
	else
		while ( optind < argc )
			count += update ( argv[ optind ++ ] );

	return ( count );
}

static int
update ( filename )
char * filename;
{
	PASSWD_F * in_passwd_f, * out_passwd_f;
	PASSWD_T ** new_passwd_a, ** passwd_a, ** passwd_array, ** old_passwd_array;
	char * tempname, * buffer;
	int count, any_change;

	extern char * malloc ();
	extern void free (), qsort ();
	PASSWD_T ** read_passwd_array (), ** expand_passwd_array (), * dup_passwd ();
	void free_passwd_array (), update_error ();
	int passwd_p_cmp ();

	if ( verbose )
		printf ( "%s:\n", filename );

	if ( ( passwd_array = read_passwd_array ( filename ) ) == NULL )
		return ( 1 );

	count = 0;
	for ( passwd_a = passwd_array; * passwd_a != NULL; passwd_a ++ )
		count ++;

	/*   Combine the two user collections together, sort them, and then out.
	 *   Remove user 'rje' if possible.  Remove new user which are already defined
	 *   on the user's system.  User 'daemon' requires special attention.
	 */
	any_change = 0;
	for ( new_passwd_a = new_passwd_array; * new_passwd_a != NULL; new_passwd_a ++ ) {
		for ( passwd_a = passwd_array; * passwd_a != NULL; passwd_a ++ ) {
			if ( strcmp ( ( * new_passwd_a ) -> pw_name, ( * passwd_a ) -> pw_name ) == 0 ) {
				if ( ( strcmp ( "daemon", ( * passwd_a ) -> pw_name ) == 0 ) &&
				     ( ( * new_passwd_a ) -> pw_uid == ( * passwd_a ) -> pw_uid ) &&
				     ( ( * passwd_a ) -> pw_gid != ( * new_passwd_a ) -> pw_gid ) ) {
					if ( verbose )
						printf ( "\tUpdating user 'daemon's group id from %d to %d.\n",
							( * passwd_a ) -> pw_gid, ( * new_passwd_a ) -> pw_gid );
					( * passwd_a ) -> pw_gid = ( * new_passwd_a ) -> pw_gid;
					any_change ++;
				}
				break;
			}
			if ( ( * new_passwd_a ) -> pw_uid == ( * passwd_a ) -> pw_uid )
				break;
		}
		if ( * passwd_a == NULL ) {
			passwd_array = expand_passwd_array ( old_passwd_array = passwd_array );
			if ( passwd_array == NULL ) {
				update_error ( NULL );
				free_passwd_array ( old_passwd_array );
				return ( 1 );
			}
			if ( verbose )
				printf ( "\tadding new user, '%s' (uid = %d).\n",
					( * new_passwd_a ) -> pw_name, ( * new_passwd_a ) -> pw_uid );
			if ( ( passwd_array [ count ++ ] = dup_passwd ( * new_passwd_a ) ) == NULL ) {
				update_error ( NULL );
				free_passwd_array ( passwd_array );
				return ( 1 );
			}
			any_change ++;
		}
	}
	tempname = tempnam ( NULL, "pw" );
	if ( tempname == NULL ) {
		update_error ( NULL );
		free_passwd_array ( passwd_array );
		return ( 1 );
	}
	out_passwd_f = openpasswd ( tempname, "w" );
	if ( out_passwd_f == NULL ) {
		update_error ( tempname );
		free ( tempname );
		free_passwd_array ( passwd_array );
		return ( 1 );
	}
	for ( passwd_a = passwd_array; * passwd_a != NULL ; passwd_a ++ ) {
		if ( ( strcmp ( "rje", ( * passwd_a ) -> pw_name ) == 0 ) &&
		     ( ( * passwd_a ) -> pw_uid == 18 ) &&
		     ( ( * passwd_a ) -> pw_gid == 18 ) ) {
			if ( verbose )
				puts ( "\tRemoving user 'rje'." );
			any_change ++;
			continue; /* skip this entry */
		}
		if ( writepasswd ( out_passwd_f, * passwd_a ) ) {
			update_error ( tempname );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			free_passwd_array ( passwd_array );
			return ( 1 );
		}
	}
	free_passwd_array ( passwd_array );

	if ( closepasswd ( out_passwd_f ) ) {
		update_error ( tempname );
		if ( unlink ( tempname ) )
			update_error ( tempname );
		free ( tempname );
		return ( 1 );
	}
	if ( ! any_change ) {
		if ( verbose )
			puts ( "\tNo changes required - no action taken." );
		if ( unlink ( tempname ) ) {
			update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
	} else {
		buffer = malloc( BUFSIZ );
		if ( buffer == NULL ) {
			update_error ( NULL );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
		sprintf ( buffer, "mv %s %s", tempname, filename );
		if ( system ( buffer ) ) {
			free ( buffer );
			if ( unlink ( tempname ) )
				update_error ( tempname );
			free ( tempname );
			return ( 1 );
		}
		free ( buffer );
	}
	free ( tempname );
	return ( 0 );
}

static PASSWD_T **
read_passwd_array ( filename )
{
	PASSWD_T * passwd_t, ** passwd_array, ** old_passwd_array;
	PASSWD_F * passwd_f;
	static char buffer [ 256 ];
	int count;

	PASSWD_T * dup_passwd (), ** create_passwd_array (), ** expand_passwd_array ();
	void free_passwd_array (), update_error ();

	passwd_f = openpasswd ( filename, "r" );
	if ( passwd_f == NULL ) {
		update_error ( filename );
		return ( NULL );
	}
	passwd_array = create_passwd_array ();
	if ( passwd_array == NULL ) {
		update_error ( NULL );
		if ( closepasswd ( passwd_f ) )
			update_error ( filename );
		return ( NULL );
	}
	count = 0;
	while ( ( passwd_t = readpasswd ( passwd_f ) ) != NULL ) {

		passwd_array = expand_passwd_array ( old_passwd_array = passwd_array );
		if ( passwd_array == NULL ) {
			update_error ( NULL );
			free_passwd_array ( old_passwd_array );
			if ( closepasswd ( passwd_f ) )
				update_error ( filename );
			return ( NULL );
		}
		if ( ( passwd_array [ count ++ ] = dup_passwd ( passwd_t ) ) == NULL ) {
			update_error ( NULL );
			if ( closepasswd ( passwd_f ) )
				update_error ( filename );
			free_passwd_array ( passwd_array );
			return ( NULL );
		}
	}
	if ( ! feof ( passwd_f -> passwd_fp ) ) {
		sprintf ( buffer, "%s, line %d", filename, count + 1 );
		update_error ( buffer );
		closepasswd ( passwd_f );
		free_passwd_array ( passwd_array );
		return ( NULL );
	}
	if ( closepasswd ( passwd_f ) ) {
		update_error ( filename );
		free_passwd_array ( passwd_array );
		return ( NULL );
	}
	return ( passwd_array );
}

typedef struct {
	int	ga_size;	/* total size in bytes */
	int	ga_avail;	/* entries available */
	PASSWD_T * ga_first;	/* first entry */
} _PASSWD_ARRAY;

static PASSWD_T **
create_passwd_array ()
{
	_PASSWD_ARRAY  * _passwd_array;
	extern char * malloc ();

	_passwd_array = ( _PASSWD_ARRAY * ) malloc ( sizeof ( _PASSWD_ARRAY ) + P_ARRAY_INC * sizeof ( PASSWD_T * ) );
	if ( _passwd_array == NULL )
		return ( NULL );
	
	_passwd_array -> ga_size = sizeof ( _PASSWD_ARRAY ) + P_ARRAY_INC * sizeof ( PASSWD_T * );
	_passwd_array -> ga_avail = P_ARRAY_INC;

	memset ( & _passwd_array -> ga_first, '\0', ( P_ARRAY_INC + 1 ) * sizeof ( PASSWD_T * ) );

	return ( & _passwd_array -> ga_first );
}

static PASSWD_T **
expand_passwd_array ( passwd_array )
PASSWD_T ** passwd_array;
{
	_PASSWD_ARRAY * _passwd_array;
	int old_size;
	extern char * realloc ();

	_passwd_array = ( _PASSWD_ARRAY * ) ( ( ( char * ) passwd_array ) - ( sizeof ( int ) * 2 ) );
	if ( ! _passwd_array -> ga_avail ) {
		old_size = _passwd_array -> ga_size;
		_passwd_array -> ga_size += ( P_ARRAY_INC * sizeof ( PASSWD_T * ) );

		_passwd_array = ( _PASSWD_ARRAY * ) realloc ( _passwd_array, _passwd_array -> ga_size );
		if ( _passwd_array == NULL )
			return ( NULL );

		memset ( ( ( char * ) _passwd_array ) + old_size, '\0', P_ARRAY_INC * sizeof ( PASSWD_T * ) );
		_passwd_array -> ga_avail = P_ARRAY_INC;
	}
	( _passwd_array -> ga_avail ) -- ;
	return ( & _passwd_array -> ga_first );
}

static void
free_passwd_array ( passwd_array )
PASSWD_T ** passwd_array;
{
	extern void free ();
	void free_passwd ();

	PASSWD_T ** passwd_a;

	for ( passwd_a = passwd_array; * passwd_a != NULL; passwd_a ++ )
		free_passwd ( * passwd_a );
	free ( passwd_array );
}

static int
passwd_p_cmp ( passwd_a, passwd_b )
PASSWD_T ** passwd_a, ** passwd_b ;
{
	return ( ( * passwd_a ) -> pw_uid - ( * passwd_b ) -> pw_uid );
}

static PASSWD_T *
dup_passwd ( passwd_t )
PASSWD_T * passwd_t;
{
	PASSWD_T * n_passwd_t;
	char ** member_p, ** member_q, * buffer_p;

	buffer_p = malloc ( sizeof ( PASSWD_T ) + strlen ( passwd_t -> pw_name ) + 1 +
			strlen ( passwd_t -> pw_passwd ) + 1 + strlen ( passwd_t -> pw_age ) + 1 +
			strlen ( passwd_t -> pw_comment ) + 1 + strlen ( passwd_t -> pw_gecos ) + 1 +
			strlen ( passwd_t -> pw_dir ) + 1 + strlen ( passwd_t -> pw_shell ) + 1 );

	if ( buffer_p == NULL ) {
		update_error ( NULL );
		return ( NULL );
	}
	n_passwd_t = ( PASSWD_T * ) buffer_p;
	buffer_p += sizeof ( PASSWD_T );

	n_passwd_t -> pw_name = strcpy ( buffer_p, passwd_t -> pw_name );
	buffer_p += strlen ( passwd_t -> pw_name ) + 1;
	n_passwd_t -> pw_passwd = strcpy ( buffer_p, passwd_t -> pw_passwd );
	buffer_p += strlen ( passwd_t -> pw_passwd ) + 1;
	n_passwd_t -> pw_uid = passwd_t -> pw_uid;
	n_passwd_t -> pw_gid = passwd_t -> pw_gid;
	n_passwd_t -> pw_age = strcpy ( buffer_p, passwd_t -> pw_age );
	buffer_p += strlen ( passwd_t -> pw_age ) + 1;
	n_passwd_t -> pw_comment = strcpy ( buffer_p, passwd_t -> pw_comment );
	buffer_p += strlen ( passwd_t -> pw_comment ) + 1;
	n_passwd_t -> pw_gecos = strcpy ( buffer_p, passwd_t -> pw_gecos );
	buffer_p += strlen ( passwd_t -> pw_gecos ) + 1;
	n_passwd_t -> pw_dir = strcpy ( buffer_p, passwd_t -> pw_dir );
	buffer_p += strlen ( passwd_t -> pw_dir ) + 1;
	n_passwd_t -> pw_shell = strcpy ( buffer_p, passwd_t -> pw_shell );
	buffer_p += strlen ( passwd_t -> pw_shell ) + 1;

	return ( n_passwd_t );
}

static void
free_passwd ( passwd_t )
PASSWD_T * passwd_t;
{
	extern void free () ;

	free ( ( char * ) passwd_t );
}

static void
update_error ( string )
char * string;
{
	fprintf ( stderr, ( string == NULL ) ? "%s: " : "%s: %s: ", my_name, string );
	perror ( NULL );
	return;
}
