#ident "@(#)netdb.c	35.3"

/*
 * Copyright (c) 1987, Motorola Microcomputer Division
 * Author    (a) 1987, Damon Anton Permezel
 *
 * Net Database Routines
 *
 * These are supposed to emulate the 4.2 routines.
 */

#include <sys/types.h>
#include <netinet/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdio.h>

#ifndef ntohl
#	define	ntohl
#	define	htonl
#	define	ntohs
#	define	htons
#endif

struct ndb {
	FILE *fp;		/* file pointer		*/
	char *filename;		/* file name		*/
	struct mem *mem;	/* ^ to malloc'd mem	*/
	char stayopen;		/* flag: stay open	*/
};

struct mem {
	struct mem *link;	/* ^ to next piece of mem	*/
};

static struct ndb host  = { NULL, NDB_HOSTS     };
static struct ndb net   = { NULL, NDB_NETWORKS  };
static struct ndb proto = { NULL, NDB_PROTOCOLS };
static struct ndb serv  = { NULL, NDB_SERVICES  };

static char *savstr(), *get_tok();
static char **_aliases();

/*
 * _setent -	common code for the setXXXent() routines
 */
static
_setent(db)
register struct ndb *db; {
	int ret;

	if (db->fp == NULL) {
		db->fp = fopen(db->filename, "r");
		if (db->fp == NULL)
			return 0;
	}

	rewind(db->fp);

	return 1;
}

/*
 * _endent -	common code for the endXXXent() functions
 */
static
_endent(db)
register struct ndb *db; {

	if (db->fp) {
		fclose(db->fp);
		db->fp = 0;
	}
	db->stayopen = 0;
}

/*
 * savstr -	save the string, record memory used
 *
 * Input:
 *	str	-	^ to the string
 *	db	-	^ to the database control struct
 */
static char *
savstr(str, db)
register char *str;
register struct ndb *db; {
	register struct mem *mp;
	register len;

	len = strlen(str) + 1;

	mp = (struct mem *) malloc(len + sizeof (*mp));

	memcpy(mp + 1, str, len);

	mp->link = db->mem;
	db->mem  = mp;

	return (char *) (mp + 1);
}

/*
 * _getent -	move to the next valid looking entry
 */
static
_getent(db)
register struct ndb *db; {
	char c;
	register skip;
	register struct mem *mp;

	/*
	 * free any memory from last time
	 */
	while (mp = db->mem) {
		db->mem = mp->link;
		free(mp);
	}

	skip = 0;
	while (!ferror(db->fp)) {
		switch (c = getc(db->fp)) {
		    case EOF:
			return 0;

		    case '\n':
			skip = 0;
			continue;

		    case ' ':
		    case '\t':
			continue;

		    case '#':
			++skip;
			continue;

		    default:
			if (skip)
				continue;
			ungetc(c, db->fp);
			return 1;
		}
	}
}

/*
 * get_tok -	get a token
 */
static char *
get_tok(db)
register struct ndb *db; {
	char buf[256];		/* Watch it! we limit tokens to 255 chars! */
	register char *p;
	register c, cnt;
	register enum state { START, GET_CHARS, ERROR, FINISH } state;

	state = START;
	cnt = 0;
	p = buf;

	while (state != ERROR && state != FINISH) {
		c = getc(db->fp);
		switch (c) {
		    case ' ':
		    case '\t':
			if (state == START)
				continue;	/* discard leading white */
			state = FINISH;
			continue;

		    case '#':
			ungetc(c, db->fp);
			if (state == START)  {
				return 0;
			}
			state = FINISH;
			continue;
		    case '\n':
			if (state == START) {
				return 0;
			}
			state = FINISH;
			/*
			 * Put new-line back so it can be caught in state
			 * START and discarded.
			 */
			ungetc(c, db->fp);
			continue;

		    default:
			if (state == START)
				state = GET_CHARS;
			if (++cnt > sizeof (buf) -1) {
				state = ERROR;
				continue;
			}
			*p++ = c;
		}
	}

	return (state == ERROR) ? 0 : savstr(buf, db, *p++ = 0);
}

/*
 * _aliases -	parse the list of aliases
 */
static char **
_aliases(db)
register struct ndb *db; {
	char *list[30 + 1];
	struct mem *mp;
	register cnt;

	for (cnt = 0; cnt < 30; ++cnt) {
		if (list[cnt] = get_tok(db))
			continue;

		break;
	}

	if (!cnt)
		return 0;

	list[cnt++] = 0;

	mp = (struct mem *) malloc (sizeof (*mp) + cnt * sizeof (char *));
	mp->link = db->mem;
	db->mem  = mp;

	memcpy(mp + 1, list, cnt * sizeof (char *));

	return (char **) (mp + 1);
}

/*
 * aliascmp -	compare a name against a list of aliases
 *
 * Returns:
 *	 0	-	match  found (a la strcmp)
 *	!0	-	match !found
 */
static
aliascmp(aliases, name)
register char **aliases, *name; {

	if (aliases == 0)
		return !0;

	while (*aliases && strcmp(name, *aliases))
		++aliases;

	return *aliases == 0;
}

/*
 * sethostent -	open the HOSTS database file
 */
sethostent(stayopen) {

	host.stayopen = stayopen;

	return (_setent(&host));
}

/*
 * endhostent -	close the HOSTS database
 */
endhostent() {

	return _endent(&host);
}
/*
 * gethostent -	return the next entry from the HOSTS database
 */
struct hostent *
gethostent() {
	int ret;
	static struct hostent ent;
	static struct sockaddr_in addr;
	char *tok;

	try_again:
	if (_getent(&host) == 0)
		return 0;

	ent.h_addrtype = AF_INET;
	ent.h_length   = sizeof (addr.sin_addr);
	ent.h_addr     = (char *) &addr.sin_addr;

	/*
	 * get the address
	 */
	tok = get_tok(&host);
	if (!tok) 	/* shouldn't happen unless 1st token > 255 chars */
	    goto try_again;

	addr.sin_addr.s_addr  = htonl(inet_addr(tok));

	/*
	 * get the host name
	 */
	tok = get_tok(&host);
	if (!tok)
		goto try_again;

	ent.h_name = tok;

	ent.h_aliases = _aliases(&host);

	/*
	 * if addr was bogus, skip entry
	 */
	if (addr.sin_addr.s_addr == htonl(-1))
		goto try_again;

	return &ent;
}

/*
 * gethostbyname -	look for HOSTS entry matching name
 */
struct hostent *
gethostbyname(name)
char *name; {
	register struct hostent *hp = 0;

	if ( sethostent(1) ) {
		while (hp = gethostent())
			if (!strcmp(hp->h_name, name)
			||  (hp->h_aliases && !aliascmp(hp->h_aliases, name)))
				break;
	
		endhostent();
	}

	return (hp);
}

/*
 * gethostbyaddr -	look for HOSTS entry matching addr
 */
struct hostent *
gethostbyaddr(addr, len, type)
char *addr; {
	register struct hostent *hp = 0;

	if( sethostent(1) ) {
		while (hp = gethostent())
			if (hp->h_addrtype == type
			&& bcmp(hp->h_addr, addr, len) == 0)
				break;

		endhostent();
	}

	return (hp);
}

/*
 * setservent -	open the SERVICES database file
 */
setservent(stayopen) {

	serv.stayopen = stayopen;

	return (_setent(&serv));
}

/*
 * endservent -	close the SERVICES database
 */
endservent() {

	return _endent(&serv);
}
/*
 * getservent -	return the next entry from the SERVICES database
 */
struct servent *
getservent() {
	int ret;
	static struct servent ent;
	char *tok, *p;

	if (_getent(&serv) == 0)
		return 0;

	/*
	 * get the official service name
	 */
	tok = get_tok(&serv);
	if (!tok) {
	    error:
		if (!serv.stayopen)
			_endent(&serv);
		return 0;
	}

	ent.s_name = tok;
	/*
	 * get the port #/protocol name
	 */
	tok = get_tok(&serv);
	if (!tok)
		goto error;

	for (p = tok; *p && *p != '/'; ++p)
		;

	if (!p)
		goto error;

	*p++ = 0;

	ent.s_port  = atoi(tok);
	ent.s_proto = p;

	ent.s_aliases = _aliases(&serv);

	return &ent;
}

/*
 * getservbyname -	look for SERVICES entry matching name
 */
struct servent *
getservbyname(name, proto)
char *name, *proto; {
	register struct servent *sp = 0;

	if ( setservent(1) ) {
		while (sp = getservent()) {
			if (proto && strcmp(sp->s_proto, proto))
				continue;
			if (!strcmp(sp->s_name, name)
			|| !aliascmp(sp->s_aliases, name))
				break;
		}
		
		endservent();
	}

	return (sp);
}

/*
 * getservbyport -	look for SERVICES entry matching port
 */
struct servent *
getservbyport(port, proto)
char *proto; {
	register struct servent *sp = 0;

	if ( setservent(1) ) {
		while (sp = getservent()) {
			if (proto && strcmp(sp->s_proto, proto))
				continue;
			if (sp->s_port == port)
				break;
		}

		endservent();
	}

	return (sp);
}

/*
 * setnetent -	open the NETWORKS database file
 */
setnetent(stayopen) {

	net.stayopen = stayopen;

	return (_setent(&net));
}

/*
 * endnetent -	close the NETWORKS database
 */
endnetent() {

	return _endent(&net);
}

/*
 * getnetent -	return the next entry from the NETWORKS database
 */
struct netent *
getnetent() {
	int ret;
	static struct netent ent;
	char *tok;

	try_again:
	if (_getent(&net) == 0)
		return 0;

	ent.n_addrtype = AF_INET;

	/*
	 * get the network name
	 */
	tok = get_tok(&net);
	if (!tok)	    /* shouldn't happen unless 1st token > 255 chars */
		goto try_again;

	ent.n_name = tok;

	/*
	 * get the network number
	 */
	tok = get_tok(&net);
	if (!tok) 	
	    goto try_again;

	ent.n_net = inet_network (tok);

	ent.n_aliases = _aliases(&net);

	/*
	 * if network was bogus, skip entry
	 */
	if (ent.n_net == (ulong)-1)
		goto try_again;

	return &ent;
}


/*
 * getnetbyname -	look for NETWORKS entry matching name
 */
struct netent *
getnetbyname(name)
char *name; {
	register struct netent *np = 0;

	if ( setnetent(1) ) {
		while (np = getnetent())
			if (!strcmp(np->n_name, name)
			||  (np->n_aliases && !aliascmp(np->n_aliases, name)))
				break;
		
		endnetent();
	}

	return (np);
}

/*
 * getnetbyaddr -	look for NETWORKS entry matching addr
 */
struct netent *
getnetbyaddr(netnum, type)
int netnum; {
	register struct netent *np = 0;

	if ( setnetent(1) ) {
		while (np = getnetent())
			if (np->n_addrtype == type && np->n_net == netnum)
				break;

		endnetent();
	}

	return (np);
}

