/*		Copyright (c) 1985 AT&T		*/
/*		All Rights Reserved		*/
/* @(#)atan.c	6.1	 */
/*	Added to Ferranti System V by msw 21/3/89 */

/*	@(#)atan.c	6.1		*/
/*	atan.c	1.10	*/
/*LINTLIBRARY*/
/*
 *	atan returns the arctangent of its double-precision argument,
 *	in the range [-pi/2, pi/2].
 *	There are no error returns.
 *
 *	atan2(y, x) returns the arctangent of y/x,
 *	in the range [-pi, pi].
 *	atan2 discovers what quadrant the angle is in and calls atan.
 *	atan2 returns EDOM error and value 0 if both arguments are zero.
 *
 */

#include <math.h>
#include <values.h>
#include <errno.h>
#include <sys/fpu.h>

double
atan(x)
register double x;
{
	double y;
	struct exception exc;
	register unsigned int status = atan881(x,&y);

	if (!(status & FPCC_NAN))
		return(y); /* No important error conditions found */

	/* If input was a NaN, report a domain error! */

	exc.arg1 = x;
	exc.name = "atan";

	exc.type = DOMAIN;
	exc.retval = 0.0;
	if (!matherr(&exc)) 
		{
		(void) write(2, exc.name, 4);
		(void) write(2, ": DOMAIN error\n", 15);
		errno = EDOM;
		}
	return (exc.retval);
}

double
atan2(y, x)
register double y, x;
{
	register int neg_y = 0;
	struct exception exc;

	exc.arg1 = y;
	if (!x && !y) {
		exc.type = DOMAIN;
		exc.name = "atan2";
		exc.arg2 = x;
		exc.retval = 0.0;
		if (!matherr(&exc)) {
			(void) write(2, "atan2: DOMAIN error\n", 20);
			errno = EDOM;
		}
		return (exc.retval);
	}
	/*
	 * The next lines determine if |x| is negligible compared to |y|,
	 * without dividing, and without adding values of the same sign.
	 */
	if (exc.arg1 < 0) {
		exc.arg1 = -exc.arg1;
		neg_y++;
	}
	if (exc.arg1 - _ABS(x) == exc.arg1)
		return (neg_y ? -M_PI_2 : M_PI_2);
	/*
	 * The next line assumes that if y/x underflows the result
	 * is zero with no error indication, so it's safe to divide.
	 */
	return ((y = atan(y/x)) == 0 || x > 0 ? y :
	    neg_y ? y - M_PI : y + M_PI);
}
