/*		Copyright (c) 1985 AT&T		*/
/*		All Rights Reserved		*/
/* @(#)hypot.c	6.1	 */
/*	Added to Ferranti System V by msw 21/3/89 */

/*	@(#)hypot.c	6.1		*/
/*	hypot.c	1.10	*/
/*LINTLIBRARY*/
/*
 *	hypot(a, b) returns sqrt(a^2 + b^2), avoiding unnecessary overflows.
 *	Returns ERANGE error and value HUGE if the correct value would overflow.
 *	Returns EDOM   error and value 0 if either input is a NaN.
 */

#include <math.h>
#include <values.h>
#include <errno.h>
#include <sys/fpu.h>

extern double hypoterr();

double
hypot(a, b)
register double a, b;
{
	register double asv = a,bsv = b,t;
	double y;

	if (nan881(a,b))			/* an input was a NaN */
		return(hypoterr(EDOM,a,b));

	if (a < 0)
		a = -a;
	if (b < 0)
		b = -b;
	if (a > b) {				/* make sure |a| <= |b| */
		t = a;
		a = b;				/* swap a and b */
		b = t;
	}

	if (b == 0)
		return(b);

	/* We now have b>0, and b>=a>=0. */

	if (a==b)
		(void) sqrt881(2.0,&y);	/* covers the case a=b=infinity */
	else {
		t = a/b;		/* quiet underflow is possible here */
		(void) sqrt881(1+t*t,&y);
	}

	if (mul881(b,&y) & (EXC_OVFL|FPCC_I))
		/* b was infinity or overflow occurred on the multiply */
		return(hypoterr(ERANGE,asv,bsv));
	else
		return(y);
}


double hypoterr(err,a,b)
register int err;
register double a, b;
{
	struct exception exc;

	exc.arg1 = a;
	exc.arg2 = b;
	exc.name = "hypot";
	exc.type = (err == ERANGE) ? OVERFLOW : DOMAIN;
	exc.retval = (err == ERANGE) ? HUGE : 0;
	if (!matherr(&exc)) {
		if ((errno = err) == EDOM) {
			(void) write(2, exc.name, 5);
			(void) write(2, ": DOMAIN error\n", 15);
		}
	}
	return (exc.retval);
}
