/*	@(#)ckmode.c	35.1		created:8/24/88 
 *
 *	Copyright (c)  Motorola Inc.  1986,1987
 *	All rights Reserved
 */

#include	<ctype.h>
#include	<mnt.h>
#include	<stdio.h>
#include	<string.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<mnttab.h>



/*	ckmode():	Determines if the given device has a specific
 *			permission flag set.  If the permission asked
 *			for does not exist then the routine will return
 *			a ( 1 ).  This will cause a 
 *			if ( ckmode() ) {
 *			   print error; 
 *			   exit
 *			}
 *			to work properly.
 *
 *	This is accomplished by searching for the upper case letter,
 *	If this is true then any user has that permission.  
 *
 *	If super user then search for the lower case letter.  If
 *	true then super user ONLY has this permission.
 *	
 *	If the permission asked for was \0, then return an error 2.
 */

ckmode ( dev, ask )
register struct usrdev *dev;
register int    ask;
{
	int	toupper();
	int	tolower();

	if  ( ! ask )   {		/*  If the permission asked for was null then return an error. */
	    fprintf( stderr,"WARNING:  %s Checking for null permissions\n", __FILE__ );
	    return( 1 );
	}

	if ( strchr( dev->modes, toupper(ask) )  !=  NULL  )	/*  If the desired permision exists, for ALL */
		return( 0 );

	if ( getuid() != (unsigned short)0 )	/*  If not super user then it is all over */
		return( 1 );

	if ( strchr( dev->modes, tolower(ask) )  !=  NULL  )	/*  If the desired permision exists, for ROOT */
		return( 0 );

	return( 1 );	/*  If no permissions for super user return 1 */
}



/*
 *	Open the permissions file and determine 
 *	all of the devices with a given access permission
 *
 *	This function starts with:
 *		perm	-	A given permission to search for.
 *
 *	This function builds a linked list of mnt_que structures.
 *
 */



gdevs( perm )
register int	perm;
{
	register FILE	*map;
	struct	usrdev	tmp;		/*  temp spot to read perms file into */
	register struct	usrdev	*dsk;		/*  Pointer to the temp spot above    */
	static char line[BUFSIZ];
	struct	mnt_que *svstruct();
	char	*strrchr();
	register n = 0;			/*  Number of devices with requested perm */

	if ( (map = fopen( PERMS_FILE, "r" ))  ==  NULL ) {
		fprintf( stderr, "Could not open the permissions file.\n" );
		return( 0 );
	}

	dsk = &tmp;	
	while (  fgets( line, BUFSIZ, map) != NULL )
	if ( (sscanf( line, "%s%s%s%s%s%*[^']'%[^']'", dsk->slice, dsk->alias, dsk->fsize, dsk->modes, dsk->mnt_pt, dsk->pgm )  !=  EOF )  ) {
		if (  strncmp(dsk->slice, "#", 1) == 0 ) {
			continue;
		}
		if ( ckmode( dsk, perm ) ) {  /* then this dev does not have this permission */
			continue;
		}
		n = n + addq( dsk->alias, dsk->mnt_pt );
	}
	return( n );
}



/*	
 *	Add an device to the mount que.  
 *	
 *	WARNING:	The return value is the number added to the que
 *			Therefore an error code is 0.
 */

addq( dev, dir )
register char	*dev;
register char	*dir;
{
	register struct	mnt_que	*a;
	extern	struct mnt_que *mnt_head;
	extern	struct mnt_que	*mnt_tail;

#ifdef	EBUG
	fprintf( stderr, "adding %s to the que\n", dev );
#endif
	if  ( ( a = svstruct( dev, dir )) == NULL )   {
	    fprintf( stderr, "WARNING:  Could not add %s to the que\n", dev );
	    return( 0 );	/*  Error condition */
	}
	a->m_next = NULL;
	if  ( !mnt_head ) {
		mnt_head = a;
	}
	else {
		mnt_tail->m_next = a;
	}
	mnt_tail = a;
	return( 1 );		/* This works */
}



/*	Save a string and return a pointer to the saved copy	*/

char *svstr( s )
register char	*s;
{
    register char	*a;
    char		*malloc();

    if  ( s == NULL )	/* if saving a NULL pointer, return a null pointer */
	return( NULL );
    if  ( (a = malloc((unsigned)(strlen(s)+1))) == NULL )   {
	fprintf( stderr, "Attempted svstr and could not malloc any space\n" );
	return(NULL);
    }
    
    strcpy( a, s );
    return(a);
}

/*	Save a structure and return a pointer to that structure */

struct mnt_que *svstruct( dev, dir )
register char	*dev;
register char	*dir;
{
	
	register struct mnt_que	*a;
	char	*malloc();

	if  ( (a = (struct mnt_que *)malloc((unsigned)sizeof(struct mnt_que))) == NULL )   {
		fprintf( stderr, "Attempted svstr and could not malloc any space\n" );
		return(NULL);
	}

	a->m_dev = svstr( dev );
	a->m_dir = svstr( dir );
	a->m_next = NULL;

	return( a );

}



/*	print that sturcture */

prstruct( s )
register struct mnt_que	*s;
{
	printf( "mnt_que structure at %x\n", s );
	if  ( s->m_dev  ) 
		printf( "	device %s\n", s->m_dev );
	else
		printf( "	device   NULL\n" );
	if  ( s->m_dir  ) 
		printf( "	mntdir %s\n", s->m_dir );
	else
		printf( "	mntdir   NULL\n" );
	if  ( s->m_next  ) 
		printf( "	next   %x\n", s->m_next );
	else
		printf( "	next   NULL\n" );
}



/*
 *	Get the mount directory listed in the PERMS_FILE
 */

char *gdir( dev, dir )
register char	*dev;
register char	*dir;
{
	struct	usrdev  disk;		/*  Info on the requested disk.      */
	int	getperms();		/*  Function to read the PERMS_FILE  */

	strcpy( disk.alias, dev );

	if  ( getperms( &disk )  ) {
		fprintf( stderr, "dir():  Unknown device searched for\n" );
		return( NULL );
	}

	strcpy( dir, disk.mnt_pt );
	return( dir );

}

/* 
 *	Get the last slice if given with format
 */

gslice( dev ,ask )
register struct usrdev *dev;
register char ask;
{
	register char *sp;

	if ((sp = strchr( dev->modes, toupper(ask))) == NULL )
		return(-1);
	sp++;
	if(*sp >= '0' && *sp <= '9')
		return(atoi(sp));
	return(0);
}
