static char SysVNFSID[] = "@(#)pmap_clnt.c	35.1";
#ifndef lint
static char sccsid[] = "@(#)pmap_clnt.c	35.1";
#endif

/*
 * pmap_clnt.c
 * Client interface to pmap rpc service.
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 */

#include <rpc/types.h>
#include <netinet/in.h>
#include <rpc/xdr.h>
#include <rpc/auth.h>
#include <rpc/clnt.h>
#include <rpc/rpc_msg.h>
#include <rpc/pmap_prot.h> 
#include <rpc/pmap_clnt.h>
#include <sys/socket.h>
#include <sys/fs/nfs/time.h>
#include <stdio.h>
/*#include <net/if.h>*/
/*#include <sys/ioctl.h>*/
/*#include <arpa/inet.h>*/
#define NAMELEN 255

static struct timeval timeout = { 5, 0 };
static struct timeval tottimeout = { 60, 0 };
static struct sockaddr_in myaddress;

	/* no printing */
#define	DEBUG0		0
	/* print bad news */
#define	DEBUG1		1
	/* print inportant good news */
#define	DEBUG2		2
	/* commentary */
#define	DEBUG3		3
	/* gossip */
#define	DEBUG4		4
#define	V5		1

void clnt_perror();


/*
 * Set a mapping between program,version and port.
 * Calls the pmap service remotely to do the mapping.
 */
bool_t
pmap_set(program, version, protocol, port)
	ulong program;
	ulong version;
	ulong protocol;
	ushort port;
{
	struct sockaddr_in myaddress;
	int socket = -1;
	register CLIENT *client;
	struct pmap parms;
	bool_t rslt;

#if	DEBUG > DEBUG4
	fprintf (stderr,
		 "pmap_set (program 0x%x, version %d, protocol %d, port %d)\n",
			program, version, protocol, port);
#endif

	get_myaddress(&myaddress);

#if	DEBUG > DEBUG2
	dump_in_addr (stderr, "", myaddress.sin_addr);
#endif

	client = clntudp_bufcreate(&myaddress, PMAPPROG, PMAPVERS,
	    timeout, &socket, PMAPMSGSZ, PMAPMSGSZ);
	if (client == (CLIENT *)NULL)
		return (FALSE);
	parms.pm_prog = program;
	parms.pm_vers = version;
	parms.pm_prot = protocol;
	parms.pm_port = port;

#if	DEBUG > DEBUG4
	fprintf (stderr,
		"pmap_set calling CLNT_CALL PMAPPROC_SET\n");
#endif

	if (CLNT_CALL(client, PMAPPROC_SET, xdr_pmap, &parms, xdr_bool, &rslt,
	    tottimeout) != RPC_SUCCESS) {

#if	DEBUG > DEBUG2
	fprintf (stderr,
		"pmap_set bad CLNT_CALL\n");
#endif

		clnt_perror(client, "Cannot register service");
		return (FALSE);
	}

#if	DEBUG > DEBUG4
	fprintf (stderr,
		"pmap_set good CLNT_CALL\n");
	fflush (stderr);
#endif

	CLNT_DESTROY(client);

#if	DEBUG > DEBUG4
	fprintf (stderr,
		"pmap_set good CLNT_DESTROY\n");
	fflush (stderr);
#endif

	/*
	 *	  I can hardly believe this.
	 *
	 *	(void)close(socket);
	 */

#if	DEBUG > DEBUG4
	fprintf (stderr,
		"pmap_set returning %d\n", rslt);
	fflush (stderr);
#endif

	return (rslt);
}

/*
 * Remove the mapping between program,version and port.
 * Calls the pmap service remotely to do the un-mapping.
 */
bool_t
pmap_unset(program, version)
	ulong program;
	ulong version;
{
	struct sockaddr_in myaddress;
	int socket = -1;
	register CLIENT *client;
	struct pmap parms;
	bool_t rslt;

#if	DEBUG > DEBUG4
	fprintf (stderr,
		 "pmap_set (program 0x%x, version %d)\n",
			program, version);
#endif

	get_myaddress(&myaddress);
	client = clntudp_bufcreate(&myaddress, PMAPPROG, PMAPVERS,
	    timeout, &socket, PMAPMSGSZ, PMAPMSGSZ);
	if (client == (CLIENT *)NULL)
		return (FALSE);
	parms.pm_prog = program;
	parms.pm_vers = version;
	parms.pm_port = parms.pm_prot = 0;
	CLNT_CALL(client, PMAPPROC_UNSET, xdr_pmap, &parms, xdr_bool, &rslt,
	    tottimeout);
	CLNT_DESTROY(client);
	(void)close(socket);
	return (rslt);
}

/* 
 * don't use gethostbyname, which would invoke yellow pages
 */
get_myaddress(addr)
	struct sockaddr_in *addr;
{
#if	SYSTEM == V5
	int	my_addr = MY_IN_ADDR;

	/*
	 *	  Now honestly, how is a 3b2 supposed to know
	 *	its own name?    
	 */

	addr -> sin_family = AF_INET;
	addr -> sin_port = htons(PMAPPORT);
	addr -> sin_addr = *((struct in_addr *)&my_addr);

	return;
#endif

#if	SYSTEM != V5
	int s;
	char buf[BUFSIZ];
	struct ifconf ifc;
	struct ifreq ifreq, *ifr;
	int len;

#if	DEBUG > DEBUG2
	fprintf (stderr,
		"get_myaddress: socket\n");
#endif

	if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
	    perror("get_myaddress: socket");
	    exit(1);
	}
	ifc.ifc_len = sizeof (buf);
	ifc.ifc_buf = buf;
	/*	if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0) {
		perror("get_myaddress: ioctl (get interface configuration)");
		exit(1);
	}	*/
	ifr = ifc.ifc_req;
	for (len = ifc.ifc_len; len; len -= sizeof ifreq) {
		ifreq = *ifr;
		/*	if (ioctl(s, SIOCGIFFLAGS, (char *)&ifreq) < 0) {
			perror("get_myaddress: ioctl");
			exit(1);
		}	*/
		if ((ifreq.ifr_flags & IFF_UP) &&
		    ifr->ifr_addr.sa_family == AF_INET) {
			*addr = *((struct sockaddr_in *)&ifr->ifr_addr);
			addr->sin_port = htons(PMAPPORT);
			break;
		}
		ifr++;
	}
	close(s);
#endif
}

#if	DEBUG > DEBUG2
dump_in_addr (win, string, hdr)
FILE    *win;
struct in_addr	hdr;
{

        fprintf (win, "in_addr: %s: %s\n", string, inet_ntoa (hdr));

}
#endif
