static char SysVNFSID[] = "@(#)mountxdr.c	35.1";
#ifndef lint
static	char sccsid[] = "@(#)mountxdr.c	35.1";
#endif

#include <stdio.h>
#include <rpc/rpc.h>
#include <sys/fs/nfs/time.h>
#include <sys/errno.h>
#include <sys/fs/nfs/nfs.h>
#include <rpcsvc/mount.h>

#define xdr_dev_t xdr_short

xdr_fhstatus(xdrs, fhsp)
	XDR *xdrs;
	struct fhstatus *fhsp;
{
#ifdef	DEBUG
	fprintf (stderr,
		"xdr_fhstatus(xdrs 0x%x, fhsp 0x%x)\n",
		xdrs,
		fhsp);
	dump_xdr (xdrs);
#endif
	if (!xdr_int(xdrs, &fhsp->fhs_status))	{
		fprintf (stderr,
			"xdr_fhstatus: xdr_int fails\n");
		return FALSE;
	}
#ifdef	DEBUG
	fprintf (stderr,
		"xdr_fhstatus: fhsp->fhs_status 0x%x\n",
		fhsp->fhs_status);
	dump_xdr (xdrs);
#endif
	if (fhsp->fhs_status == 0) {
		if (!xdr_fhandle(xdrs, &fhsp->fhs_fh))	{
			fprintf (stderr,
				"xdr_fhstatus: xdr_fhandle fails\n");
			return FALSE;
		}
	}
#ifdef	DEBUG
	fprintf (stderr,
		"xdr_fhstatus: undefined return\n");
#endif
}

xdr_fhandle(xdrs, fhp)
	XDR *xdrs;
	fhandle_t *fhp;
{
#ifdef	DEBUG
	fprintf (stderr,
		"xdr_fhandle(xdrs 0x%x, fhp 0x%x)\n",
		xdrs,
		fhp);
	dump_xdr (xdrs);
#endif
	if (xdr_opaque(xdrs, fhp, NFS_FHSIZE)) {
		return (TRUE);
	}
	fprintf (stderr,
		"xdr_fhandle: xdr_opaque fails xdrs 0x%x fhp 0x%x\n",
		xdrs,
		fhp);
	return (FALSE);
}

bool_t
xdr_path(xdrs, pathp)
	XDR *xdrs;
	char **pathp;
{
	if (xdr_string(xdrs, pathp, 1024)) {
		return(TRUE);
	}
	fprintf (stderr,
		"xdr_path: xdr_string fails\n");
	return(FALSE);
}

/* 
 * body of a mountlist
 */
bool_t
xdr_mountbody(xdrs, mlp)
	XDR *xdrs;
	struct mountlist *mlp;
{
	if (!xdr_path(xdrs, &mlp->ml_name))	{
		fprintf (stderr,
			"xdr_mountbody: xdr_path(1) fails\n");
		return FALSE;
	}
	if (!xdr_path(xdrs, &mlp->ml_path))	{
		fprintf (stderr,
			"xdr_mountbody: xdr_path(2) fails\n");
		return FALSE;
	}
	return(TRUE);
}

bool_t
xdr_mountlist(xdrs, mlp)
	register XDR *xdrs;
	register struct mountlist **mlp;
{
	/*
	 * more_elements is pre-computed in case the direction is
	 * XDR_ENCODE or XDR_FREE.  more_elements is overwritten by
	 * xdr_bool when the direction is XDR_DECODE.
	 */
	int more_elements;
	register int freeing = (xdrs->x_op == XDR_FREE);
	register struct mountlist **nxt;

	while (TRUE) {
		more_elements = (*mlp != NULL);
		if (! xdr_bool(xdrs, &more_elements))	{
			fprintf (stderr,
				"xdr_mountlist: xdr_bool fails\n");
			return (FALSE);
		}
		if (! more_elements)
			return (TRUE);  /* we are done */
		/*
		 * the unfortunate side effect of non-recursion is that in
		 * the case of freeing we must remember the nxt object
		 * before we free the current object ...
		 */
		if (freeing)
			nxt = &((*mlp)->ml_nxt); 
		if (! xdr_reference(xdrs, mlp, sizeof(struct mountlist),
		    xdr_mountbody))	{
			fprintf (stderr,
				"xdr_mountlist: xdr_reference fails\n");
			return (FALSE);
		}
		mlp = (freeing) ? nxt : &((*mlp)->ml_nxt);
	}
#ifdef	DEBUG
	fprintf (stderr,
		"xdr_mountlist: undefined return\n");
#endif
}

/*
 * Strange but true: the boolean that tells if another element
 * in the list is present has already been checked.  We handle the
 * body of this element then check on the next element.  YUK.
 */
bool_t
xdr_groups(xdrs, gr)
	register XDR *xdrs;
	register struct groups *gr;
{
	/*
	 * more_elements is pre-computed in case the direction is
	 * XDR_ENCODE or XDR_FREE.  more_elements is overwritten by
	 * xdr_bool when the direction is XDR_DECODE.
	 */
	int more_elements;
	register int freeing = (xdrs->x_op == XDR_FREE);
	register struct groups *nxt;

	while (TRUE) {
		/*
		 * the unfortunate side effect of non-recursion is that in
		 * the case of freeing we must remember the nxt object
		 * before we free the current object ...
		 */
		if (freeing)
			nxt = gr->g_next; 
		if (! xdr_path(xdrs, &(gr->g_name)))	{
			fprintf (stderr,
				"xdr_groups: xdr_path fails\n");
			return (FALSE);
		}
		more_elements = (gr->g_next != NULL);
		if (! xdr_bool(xdrs, &more_elements))	{
			fprintf (stderr,
				"xdr_groups: xdr_bool fails\n");
			return (FALSE);
		}
		if (! more_elements) {
			gr->g_next = NULL;
			return (TRUE);  /* we are done */
		}
		if (! xdr_reference(xdrs, &(gr->g_next), sizeof(struct groups),
		    xdr_groups))	{
			fprintf (stderr,
				"xdr_groups: xdr_reference fails\n");
			return (FALSE);
		}
		gr = (freeing) ? nxt : gr->g_next;
	}
#ifdef	DEBUUG
	fprintf (stderr,
		"xdr_groups: undefined return\n");
#endif
}

/* 
 * body of a exportlist
 */
bool_t
xdr_exportbody(xdrs, ex)
	XDR *xdrs;
	struct exports *ex;
{
	int more_elements;

	if (!xdr_path(xdrs, &ex->ex_name))	{
		fprintf (stderr,
			"xdr_exportbody: xdr_path fails\n");
		return FALSE;
	}
	more_elements = (ex->ex_groups != NULL);
	if (! xdr_bool(xdrs, &more_elements))	{
		fprintf (stderr,
			"xdr_exportbody: xdr_bool fails\n");
		return (FALSE);
	}
	if (! more_elements) {
		ex->ex_groups = NULL;
		return (TRUE);  /* we are done */
	}
	if (! xdr_reference(xdrs, &(ex->ex_groups), sizeof(struct groups),
	    xdr_groups))	{
		fprintf (stderr,
			"xdr_exportbody: xdr_reference fails\n");
		return (FALSE);
	}
	return(TRUE);
}


/*
 * Encodes the export list structure "exports" on the
 * wire as:
 * bool_t eol;
 * if (!eol) {
 * 	char *name;
 *	struct groups *groups;
 * }
 * where groups look like:
 * if (!eog) {
 *	char *gname;
 * }
 */
bool_t
xdr_exports(xdrs, exp)
	register XDR *xdrs;
	register struct exports **exp;
{
	/*
	 * more_elements is pre-computed in case the direction is
	 * XDR_ENCODE or XDR_FREE.  more_elements is overwritten by
	 * xdr_bool when the direction is XDR_DECODE.
	 */
	int more_elements;
	register int freeing = (xdrs->x_op == XDR_FREE);
	register struct exports **nxt;

	while (TRUE) {
		more_elements = (*exp != NULL);
		if (! xdr_bool(xdrs, &more_elements))	{
			fprintf (stderr,
				"xdr_exports: xdr_bool fails\n");
			return (FALSE);
		}
		if (! more_elements)
			return (TRUE);  /* we are done */
		/*
		 * the unfortunate side effect of non-recursion is that in
		 * the case of freeing we must remember the nxt object
		 * before we free the current object ...
		 */
		if (freeing)
			nxt = &((*exp)->ex_next); 
		if (! xdr_reference(xdrs, exp, sizeof(struct exports),
		    xdr_exportbody))	{
			fprintf (stderr,
				"xdr_exports: xdr_reference fails\n");
			return (FALSE);
		}
		exp = (freeing) ? nxt : &((*exp)->ex_next);
	}
}

static
dump_xdr (xdrs)
	XDR	*xdrs;
{
	fprintf (stderr,
		"xdrs 0x%x x_public 0x%x x_private 0x%x x_base 0x%x x_handly 0x%x\n",
		xdrs,
		xdrs->x_public,
		xdrs->x_private,
		xdrs->x_base,
		xdrs->x_handy);
	dump_buffer ("x_base: ", xdrs->x_base, 64);
}

static
dump_buffer (string, from, length)
char    *string;
char    *from;
int     length;
{
        int     i;
        char    *p;

        fprintf (stderr, "%s", string);
        for (p = from, i = 0; i < length; i++, p++)     {
                if ((i & 0x3) == 0) fprintf (stderr, " ");
                if ((*p & 0xf0) == 0) fprintf (stderr, "0");
                fprintf (stderr, "%x", (*p & 0xff));
        }
        fprintf (stderr, "\n");

}
 

