/*@(#)bp_intr.c	35.1*/
/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	bp_intr()	
**
**  DESCRIPTION:	Interrupt service routine for Unix host BPP.
**
**  CONTENTS:		This function is the interrupt service routine
**
**  PARAMETERS:		
**
**	Input:		cpu		- number of the cpu causing the interrupt
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/ccb.h>
#include <sys/bppv2/bm.h>

#include <sys/bppv2/mvme147/bpp_local.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

DEFAULT bp_intr(cpu)
	register	CPU_ID	cpu;		/* CPU number */
{
	extern void		bp_wakeup();
	extern void		bp_rel_wind();
	extern PCT_T	*bp_local_pipe_tbl[];
	extern PCT_T	*bp_remote_pipe_tbl[];
	extern BM_PCB_T	*bp_pcb_array[];

	register	PCT_T		*pct;		/* CPU table entry address */
	register	BP_MPB_T  	*mpb;		/* MPB pointer */
	register	BM_PCB_T	*pcb;		/* pointer to pool control block */
#ifdef VME147
	register	EPB			*epb;		/* pointer to EPB message */
#endif
	register 	int			i;

	LOG("BPP: interrupt on cpu %d\n", cpu);

	/*
	**	Generate CPU table address.
	*/

	pct = bp_local_pipe_tbl[cpu];

	/*
	**	Do pipe read until no more messages.
	*/

	while ((mpb = (BP_MPB_T *)bp_pread(pct)) != (BP_MPB_T *)NULL)
	{
		/*
		**	Send the message. If it's destination is registered,
		**  the message will be routed locally.  If the message is for
		**	a different CPU, it is put on the pipe for that CPU.
		**  Otherwise, bp_send() will fail and the message must
		**  either be returned or discarded.
		*/

		if (bp_send(mpb->dst_id, mpb->src_id, mpb, mpb->data_size, mpb->data) 
			!= BPP_SUCCESSFUL)
		{ /* could not route -- return or discard */
			if (mpb->msg_status != BPP_SUCCESSFUL)
			{ /* this must be a return message with bad status from
				another board -- discard */
				bp_ret_mpb(mpb);
			}
			else
			{ /* we've failed to deliver a message which must now 
				be returned */
				mpb->msg_status = BP_DST_NOT_REGISTERED;
				bp_swap_id(mpb);
				bp_send(mpb->dst_id, mpb->src_id, mpb, mpb->data_size, mpb->data);
			}
		}
	}

/*  If resource requests are pending a remote CPU releasing resources
	the PENDING_BUF flag is set in the local pipe control structure.  The
	remote processor checks this flag when releasing shared resources and
	generates an interrupt when it's set.  If the flag is currently set
	we process all the released buffers and then reset it.
*/
	pct = bp_local_pipe_tbl[cpu];
	if (pct->pipe_status & PENDING_BUF)
	{
		pct->pipe_status &= ~ PENDING_BUF;
		for (i = 0, pcb = bp_pcb_array[0]; i <= BP_MAX_POOLS; i++, pcb++)
		{
			if (pcb && pcb->pool.cpu == cpu)
			{
				bp_rel_wind(pcb->pool.pool_id);
				bp_wakeup(pcb->pool.pool_id);
			}
		}
	}
	return (BPP_SUCCESSFUL);
}
