/*@(#)bp_pipe.c	35.1*/
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	bp_pipe.c
**
**  DESCRIPTION:	This module contains the BPP pipe read and write
**					functions.
**
**  FUNCTIONS:		bp_write, bp_read
**
**  NOTES:
**
**  UPDATE LOG:
**  Date		Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**	01/01/88	J. Holbrook		1.0	Initial coding.
**
***********************************************************************/

#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/bm.h>
#include <sys/bppv2/ccb.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>

#include <sys/bppv2/mvme147/bpp_local.h>

BPE_T	*bp_get_bpe();		/* Get BPE block */
void	bp_rel_bpe();		/* Release BPE block */
extern	DEFAULT	splx();			/* Set priority level */
ULONG	bp_intr_remote();

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	bp_pwrite()	
**
**  DESCRIPTION:	Write a message to the outgoing pipe.
**
**  CONTENTS:		This function will place the passed MPB on the pipe
**					to the CPU addressed.
**
**  PARAMETERS:		
**
**	Input:		
**			
**
**	Output:		NONE
**
**	Returns:	Write status.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

DEFAULT bp_pwrite(pct, mpb)
	register	PCT_T 	*pct;		/* pipe control table pointer */
	register BP_MPB_T	*mpb;		/* MPB pointer */
{
	extern ULONG		bpp_disable_level;
	extern BP_MPB_T		*bp_kludge_in();
	extern BP_MPB_T		*bp_kludge_out();

	register	BPE_T	*bpe;		/* Outgoing BPE */
	register	DEFAULT pri;		/* Current priority level */
	register	BPE_T	*bp;		/* Local channel BPP tail pointer */

	/*
	**	Inhibit interrupts.
	*/
	DISABLE(pri, bpp_disable_level);

	/*
	**	First, determine if a pipe exists. If so, get a BPE
	**	from the BPE pool, fill it, and send it to the MVME37x
	**	device.
	*/

	if( ( bp = pct->bpp_write ) != ( BPE_T * ) NULL )
	{
		if( ( bpe = bp_get_bpe( pct ) ) != ( BPE_T * ) NULL )
		{
/***********************************************************
* kludge: The ci driver to new BPP interface requires mapping
*	of EPB and BUF pointers to and from MPB and BDB pointers.
*	This kludge tests for the ci driver as destination or
*	source of the message and maps accordingly.
************************************************************/
		if (pct->cpu_id == CPU_NO)
		{ /* this message is coming from ci driver -- map to mpb and BDB */
			mpb = bp_kludge_in(mpb);
		}
		else if (mpb->dst_id.cpu == CPU_NO)
		{ /* destination is ci driver -- map to EPB and BUF */
			mpb = bp_kludge_out(mpb);
		}

			/*
			**	Fill in the BPE and mark it as the
			**	terminator.  Marking the least-significant bit
			**  with one accomplishes this.
			*/

			bpe->next = (BPE_T *) 0x0001;

			/*
			**	Link in the new terminator, fill in
			**	current tail BPE, unmark old terminator
			**	envelope, and move the pipe tail pointer
			**	to the terminator.
			*/
			bp = LOCAL_ADDRESS(bp);
			bp->m.message = VME_ADDRESS(mpb);
			bpe = VME_ADDRESS(bpe);
			bp->next   = bpe;

			pct->bpp_write	= bpe;

			/*
			**	Enable interrupts.
			*/

			ENABLE( pri );

			/*
			**	Tell the device.
			*/
			(*pct->rmt_intr)(pct->rmt_arg);
			return( BPP_SUCCESSFUL );
		}
		else
		{
			ENABLE( pri );
			return( BPP_NO_BPE );
		}
	}
	else
	{
		ENABLE( pri );
		return( BPP_NO_PIPE );
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	bp_pread()	
**
**  DESCRIPTION:	Read message from the incoming pipe.
**
**  CONTENTS:		This function will read an MPB from the
**			local channel table's incoming pipe.
**
**  PARAMETERS:		
**
**	Input:		
**
**	Output:		NONE
**
**	Returns:	MPB pointer.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

BP_MPB_T *bp_pread(pct)
	register	PCT_T 	*pct;		/* pipe control table pointer */
{
	extern ULONG		bpp_disable_level;

	register	BPE_T		*bpe;		/* BPE pointer */
	register	BP_MPB_T	*mpb;		/* MPB pointer */
	register	DEFAULT 	pri;		/* Current priority level */

	/*
	**	Inhibit interrupts.
	*/

	DISABLE(pri, bpp_disable_level);

	/*
	**	Check for anything in the pipe. If nothing just return.
	*/
	/* Get the read head BPE and map it to local view */

	bpe = LOCAL_ADDRESS(pct->bpp_read);

	if( (ULONG) bpe->next & 0x0001 )
	{ /* read queue is empty */
		ENABLE( pri );
		return( ( BP_MPB_T * ) NULL );
	}

	/*
	**	Move the pipe head.
	*/

	pct->bpp_read = (BPE_T *) bpe->next;

	/*
	**	Enable interrupts.
	*/

	ENABLE( pri );

	/*
	**	Get the message from the envelope and return the
	**	BPE to the remote.
	*/

	mpb = bpe->m.message;

	bp_rel_bpe( pct, VME_ADDRESS(bpe) );

	return( LOCAL_ADDRESS(mpb) );
}
/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_intr_remote(pct)
**
**  DESCRIPTION:		This function uses the description in the PCT to
**						generate and interrupt to the remote processor.
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/

ULONG	bp_intr_remote(pct)

	PCT_T		*pct;
{
	
}

/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_pipe_status()
**
**  DESCRIPTION:	
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/

ULONG	bp_get_pipe_status(cpu_tag)

	ULONG		cpu_tag;
{
}

ULONG	bp_set_pipe_status(cpu_id, status_mask, status_value)

	USHORT		cpu_id;
	ULONG		status_mask;
	ULONG		status_value;
{
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	bp_get_bpe()	
**
**  DESCRIPTION:	Get a shared BPE from the shared BPE pool.
**
**  CONTENTS:		This function returns the address of a
**			BPE from the shared BPE pool. If a BPE is not
**			available, it returns a NULL pointer.  The pool is
**			managed between a pair of CPU's;  one CPU removes BPEs
**			from the head of a linked list, and the other adds BPEs
**			to the tail of the list.  The pool must always contain
**			at least one BPE since the tail's next pointer is used
**			to synchronize the CPUs.
**
**  PARAMETERS:		
**
**	Input:		pct			pointer to pipe control table
**
**	Output:		NONE
**
**	Returns:	Pointer to BPE.
**
**  METHOD:
**
**  NOTES:		In configurations in which the system address view of
**				shared memory differs from the local address view, this
**				routine always returns a local address.
**
**  UPDATE LOG:
**  Date		Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

BPE_T *bp_get_bpe( pct )
	register	PCT_T 	*pct;		/* pipe control table pointer */
{
	extern ULONG		bpp_disable_level;

	register	BPE_T	*bpe;		/* BPE pointer */
	register DEFAULT	pri;		/* Current priority level */
	
	/*
	**	Set default pointer to NULL.
	*/

	DISABLE(pri, bpp_disable_level);

/* If this is not the last BPE, remove it from the head and return
	it's address */

	bpe = LOCAL_ADDRESS(pct->get_bpe);

	if ( ((ULONG) bpe->next & 0x0001) == 0 )
	{
		/*
		**	Update list pointer.
		*/

		pct->get_bpe = bpe->next;

		bpe->next   = ( BPE_T * ) NULL;
		bpe->m.message = ( BP_MPB_T * ) NULL;
	}
	else
	{ /* no BPE available */
		ENABLE( pri );
		return((BPE_T *) NULL);
	}

	ENABLE( pri );

	return( bpe );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	bp_rel_bpe()	
**
**  DESCRIPTION:	Return a shared BPE_T to the shared BPE pool.  The
**					BPE_T is added to the tail of a linked list.
**
**  PARAMETERS:		
**
**	Input:		pct			pointerr to pct for remote
**				bpe			Pointer to BPE to be released.  This pointer
**							is always in the local address view for
**							implementations in which the local and system
**							addresses of shared memory differ.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date		Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

void bp_rel_bpe( pct, bpe )
	register	PCT_T 	*pct;		/* pipe control table pointer */
	register BPE_T	*bpe;		/* BPE address */
{
	extern ULONG		bpp_disable_level;

	register 	BPE_T	*rpt;		/* BPE pool tail address */
	register	DEFAULT pri;		/* Current priority level */

	/* mark BPE to be released with terminator */
	bpe->next   = (struct bpe_t *) 0x0001;

	/*
	**	Inhibit interrupts.
	*/

	DISABLE(pri, bpp_disable_level);

	rpt = LOCAL_ADDRESS(pct->ret_bpe);
	rpt->next = (struct bpe_t *) VME_ADDRESS(bpe);
	pct->ret_bpe = (BPE_T *) VME_ADDRESS(bpe);

	/*
	**	Enable interrupts.
	*/

	ENABLE( pri );
}

