/*@(#)bp_regdreg.c	35.2*/
/*MH***************************** MODULE HEADER *******************************
**
**  MODULE NAME:	bp_regdreg.c
**
**  DESCRIPTION:    	This module contains BPP registration and routing
**			functions for BPP release 2.
**
**  NOTES:
**
**  CHANGES:     DESCRIPTION                       	DATE     ENGINEER
**  ___________________________________________________________________
**  Original coding.								12/14/87     jeh
**
*******************************************************************************/

#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/ccb.h>
#include <sys/bppv2/bm.h>

#ifdef VME147
#include "ci_driver.h"
#include "ci_const.h"
#include "ci_struct.h"
#include "ci_global.h"
#endif

#include <sys/bppv2/mvme147/bpp_local.h>

extern REG_ENT_T	bp_reg_table[];


/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_send(dest_ip, src_ip, mpb, length, data)
**
**  DESCRIPTION:		This function performs message routing for BPP.
**						The message is first passed through the source
**						IP's output routing function if it is non-null,
**						If the destination IP is local and the routing
**						function exists, the destination IP's input 
**						routing function is called.  If the function is null,
**						the mpb is linke to the pending message list in the
**						registration table entry for the IP.
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/

LONG	bp_send(dest_ip, src_ip, mpb, length, data)
   register BP_IP_T		dest_ip;			/* destination IP */
   register BP_IP_T		src_ip;				/* source IP */
   register BP_MPB_T	*mpb;				/* pointer to message */
   USHORT				length;				/* length of message in bytes */
   UBYTE				*data;				/* pointer to message data */

{
	extern	CPU_ID		bpp_local_cpu;
	extern	ULONG		bpp_disable_level;
	extern	PCT_T		*bp_local_pipe_tbl[];

	register REG_ENT_T	*regptr;
	ULONG				err;
	int					lev;			/* save interrupt level */

/* If the source is local, fill in local cpu in the src_id field.
	The cpu identification is forced so that any messages sent off-
	board have their origin identified.
*/
	if (src_ip.cpu == 0)
		src_ip.cpu = bpp_local_cpu;

/* fill in system fields of mpb */
	mpb->src_id = src_ip;
	mpb->dst_id = dest_ip;
	mpb->data_size = length;
	mpb->data = data;
	
	/* See if this message is from the local cpu and not just bouncing
		off on its way to another board (router kludge) */
	
	if (src_ip.cpu == bpp_local_cpu)
	{ /* perform output routing function for IP */
		if (src_ip.ip >= BP_MAX_REGISTERED || 
			bp_reg_table[src_ip.ip].status != IP_REGISTERED)
		{	/* invalid send request -- reject here since we index by ip */
			return(BPP_INVALID_IP);
		}
	   	regptr = &bp_reg_table[src_ip.ip];	/* get sender's registration entry */
	/* perform any output filtering specified by the sender */
		if (regptr->route_out)
		{
	      	(*regptr->route_out)(mpb, regptr->arg_out);
		}
	}

	if (dest_ip.cpu == 0 || dest_ip.cpu == bpp_local_cpu)
	{ /* this message is destined for local cpu */
	   	if (dest_ip.ip > BP_MAX_REGISTERED ||
			bp_reg_table[dest_ip.ip].status != IP_REGISTERED)
	   	{	/* invalid destination ip -- reject this send */
			return(BPP_INVALID_IP);
	   	}

/* use the routing function for local destination */
		regptr = &bp_reg_table[dest_ip.ip];	/* get dest's registration entry */
		if (regptr->route_in)
		{
      		(*regptr->route_in)(mpb, regptr->arg_in);
		}
		else
		{  /* no routing function -- queue to registration entry */
			DISABLE(lev, bpp_disable_level);
			if (regptr->que_tail)
			{
				regptr->que_tail->sys_hdr.bmh_link = (BP_SYSHDR_T *) mpb;
				regptr->que_tail = mpb;
			}
			else
			{
				regptr->que_tail = regptr->que_head = mpb;
			}
			ENABLE(lev);
		}
	}
	else
	{ /* route to remote cpu */
		if (bp_pwrite(bp_local_pipe_tbl[dest_ip.cpu], mpb) != BPP_SUCCESSFUL)
			return(BPP_NO_PIPE);
		return(BPP_SUCCESSFUL);
	}
	return(BPP_SUCCESSFUL);
}


/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_open_ip(ip, input_func, input_arg, output_func,
**									output_arg)
**
**  DESCRIPTION:	
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/
LONG	bp_open_ip(ip, input_func, input_arg, output_func, output_arg)
   USHORT		ip;				/* interface point requested for registration */
   PTF		input_func;			/* routing function for input */
   ULONG		input_arg;		/* argument for input routing function */
   PTF		output_func;		/* routing function for output */
   ULONG		output_arg;		/* argument for output routing function */
{
	int				lev;
	REG_ENT_T		*regptr;

   DISABLE(lev, bpp_disable_level);
	if (ip == 0)
	{ /* no specific IP requested -- find any */
	}
	else
	{ /* specific IP requested */
		if (ip >= BP_MAX_REGISTERED)
		{ /* registration IP is invalid */
			ENABLE(lev);
			return(BPP_INVALID_IP);
		}
		regptr = &bp_reg_table[ip];
		if (regptr->status & IP_REGISTERED)
		{ /* Ip already in use -- reject request */
			ENABLE(lev);
			return(BPP_IP_IN_USE);
		}
	}
	regptr->status |= IP_REGISTERED;
	regptr->route_in = input_func;
	regptr->arg_in = input_arg;
	regptr->route_out = output_func;
	regptr->arg_out = output_arg;
	regptr->que_tail = (BP_MPB_T *) NULL;
	regptr->que_head = (BP_MPB_T *) NULL;

	ENABLE(lev);
	return(0);
}


/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_close_ip(ip)
**
**  DESCRIPTION:	
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/

LONG	bp_close_ip(ip)
   USHORT	ip;		/* interface point to be closed */
{
	int			lev;
	register REG_ENT_T	*regptr;

	DISABLE(lev, bpp_disable_level);
	if (ip > BP_MAX_REGISTERED)
	{ /* registration IP is invalid */
		ENABLE(lev);
		return(BPP_INVALID_IP);
	}

	regptr = &bp_reg_table[ip];
	if (regptr->status & IP_REGISTERED)
	{ /* deregister and clean up table */
		regptr->route_in = (PTF) 0;
		regptr->arg_in = 0;
		regptr->route_out = (PTF) 0;
		regptr->arg_out = 0;
		regptr->que_tail = (BP_MPB_T *) NULL;
		regptr->que_head = (BP_MPB_T *) NULL;
		regptr->status |= IP_REGISTERED;
	}
	else
	{ /* IP is not open */
		return(BPP_IP_NOT_OPEN);
	}

	return(0);
}

/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_receive(ip, mpb, timeout)
**
**  DESCRIPTION:	
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/

LONG	bp_receive(ip, mpbptr, timeout)
	USHORT		ip;
	BP_MPB_T	**mpbptr;			/* return variable for received mpb */
	ULONG		timeout;
{
	int					lev;			/* interrupt level save variable */
	register REG_ENT_T	*regptr;		/* pointer to reg table entry */

	DISABLE(lev, bpp_disable_level);
	if (ip > BP_MAX_REGISTERED)
	{ /* registration IP is invalid */
		ENABLE(lev);
		return(BPP_INVALID_IP);
	}

	regptr = &bp_reg_table[ip];
	if (regptr->status & IP_REGISTERED)
	{
		if (*mpbptr = regptr->que_head)
		{ /* queue has something on it */
			regptr->que_head = (BP_MPB_T *) regptr->que_head->sys_hdr.bmh_link;
			if (! regptr->que_head)
				regptr->que_tail = (BP_MPB_T *) NULL;
		}
	}
	else
	{ /* can't read from unregistered IP */
		ENABLE(lev);
		return(BPP_IP_NOT_OPEN);
	}
	if (*mpbptr)
	{ /* there was a message to read */
		ENABLE(lev);
		return(BPP_SUCCESSFUL);
	}
/* no message to read -- see if we're to wait on a timer */
	if (timeout)
	{ /* do something to wait here */
		ENABLE(lev);
		return(BPP_NO_MESSAGE);
	}
}
