/*@(#)bp_windows.c	35.1*/
/*MH************************ MODULE HEADER ****************************
**
**  MODULE NAME:	bp_windows.c
**
**  DESCRIPTION:	Window management functions.
**
**  CONTENTS:		bp_rel_wind()
**					bp_wakeup()
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:	
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/mvme147/bpp_local.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/bm.h>

void	*bp_rel_wind(pool_id)
	POOL_ID		pool_id;
{

	extern	 BM_PCB_T	*bp_pcb_array[];
	extern	 BM_WINDOW	bm_windows[];
	extern	LONG		bpp_disable_level;

	int						pri;
	register BM_PCB_T		*pcb;
	register BP_SYSHDR_T	*mpb;
	register BP_SYSHDR_T	*check_ptr;
	register BM_WINDOW		*wp;
	register BM_WID_T		rel_wid;

	pcb = bp_pcb_array[pool_id];

    DISABLE(pri,bpp_disable_level);

	/*
	**  NOTE:  This module separates the release processing from the 
	**			allocation processing to limit the total amount of time
	**		    that interrupts must be disabled.  
	*/

	check_ptr = (BP_SYSHDR_T *) LOCAL_ADDRESS(pcb->last_proc);

	/* 
	**  Update PCB available buffer count and window limit 
	**   information from any released buffers
	*/

	while ( check_ptr->bmh_link != NULL )
	{
		check_ptr = (BP_SYSHDR_T *) LOCAL_ADDRESS(check_ptr->bmh_link);
		rel_wid	  = check_ptr->window_id;
		if ((rel_wid > 0) && (rel_wid < BP_MAX_WINDOWS+1))
		{
			wp = &bm_windows[rel_wid];
			if (wp->current)
			{
				wp->current--;
			}
		/* NOTE:  If requests are queued for this resource window, we
			we move the request control structure to the pcb ready list
			so that the request will satisfied.
		*/
			if (wp->next_pend)
			{ /* a request is pending */
				if (pcb->pend_list.next_pend)
				{
					pcb->pend_tail->next_pend = wp->next_pend;
					pcb->pend_tail = wp->next_pend;
				}
				else
				{
					pcb->pend_list.next_pend = wp->next_pend;
					pcb->pend_tail = wp->next_pend;
				}
				wp->next_pend = wp->next_pend->next_pend;
				pcb->pend_tail->next_pend = (BM_PEND_T *) NULL;
			}
		}
		/*
		**  Update available buffer count for the PCB 
		*/
		pcb->free_count++;
	}
	pcb->last_proc = (BP_SYSHDR_T *) VME_ADDRESS(check_ptr);
	ENABLE(pri);
}


/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_wakeup(pool_id)
**
**  DESCRIPTION:	While there are buffers in the specified pool, this
**					function services any requests in the PCB's pend_list.
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/
void bp_wakeup(pool_id)
	BM_POOL_T		pool_id;
{

	extern	 BM_PCB_T	*bp_pcb_array[];
	extern	 BM_WINDOW	bm_windows[];
	void				bp_rel_pend();

	register BM_PCB_T	*pcb;
	register BP_SYSHDR_T	*mpb;
	register BM_WINDOW	*wp;
	register BM_PEND_T	*p, *tp;
	int					pri;

	pcb = bp_pcb_array[pool_id];
	mpb = (BP_SYSHDR_T *) LOCAL_ADDRESS(pcb->get_ptr);
	p = &pcb->pend_list;

	while (p->next_pend)
	{ /* something pending for resource */
		DISABLE(pri,bpp_disable_level);
		mpb = (BP_SYSHDR_T *) LOCAL_ADDRESS(pcb->get_ptr);
		if (mpb->bmh_link == (BP_SYSHDR_T *)NULL)
		{ /* no more buffers to distribute */
			ENABLE(pri);
			return;
		}
		else
		{
			mpb = (BP_SYSHDR_T *) LOCAL_ADDRESS(mpb);
			pcb->free_count--;
			pcb->get_ptr = mpb->bmh_link;
			tp = p->next_pend;
			wp = &bm_windows[tp->wind_id];
			wp->current++;
			p->next_pend = p->next_pend->next_pend;
			mpb->buf_status &= ~ BUF_IN_POOL;
			mpb->window_id = tp->wind_id;

			ENABLE(pri);
			switch ((ULONG) tp->func_code)
			{
				case STD_PEND:
					break;
				case NO_PEND:
					break;
				default:
					(*tp->func_code)(mpb, tp->func_arg);
					break;
			}
			bp_rel_pend(tp);
		}
	}
}


BM_PEND_T *bp_get_pend()

{
	extern BM_PEND_T	*bp_pend_head, *bp_pend_tail;

	register BM_PEND_T	*p;

	if (p = bp_pend_head)
	{
		bp_pend_head = p->next_pend;
		return (p);
	}
	else
	{
		return ((BM_PEND_T *) 0);
	}
}

void	bp_rel_pend(p)
	register BM_PEND_T	*p;

{
	extern BM_PEND_T	*bp_pend_head, *bp_pend_tail;

	if (p)
	{
		p->next_pend = (struct bm_pend_t *) NULL;
		bp_pend_tail->next_pend = p;
		bp_pend_tail = p;
	}
}


/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_open_window()
**
**  DESCRIPTION:		Opens a window or semaphore pool and returns the
**						window ID which is actually an index into the
**						BM_WINDOW structure array.
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**				window_size		Number of elements in the window.
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date		Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**	01/05/87	jeh			2.0	Initial coding.
************************************************************************/

ULONG	bp_open_window(window_size)
	USHORT		window_size;
{
	extern BM_WINDOW	bm_windows[];

	register int	i;
	register BM_WINDOW	*bptr;

	bptr = bm_windows;
	for (i = 0; i < BP_MAX_WINDOWS; i++, bptr++)
	{
		if (bptr->maximum == 0)
		{ /* window structure is not in use */
			break;
		}
	}

	if (i < BP_MAX_WINDOWS)
	{ /* we've found an available window */
		bptr->maximum = window_size;
		bptr->current = 0;
		bptr->next_pend = (BM_PEND_T *) NULL;
		bptr->pend_tail = (BM_PEND_T *) NULL;
		return ((ULONG) i);
	}
	else
	{
		return(BM_NO_WINDOW_AVAIL);
	}
}

ULONG	bp_close_window(window_id)
	USHORT		window_id;
{
	extern BM_WINDOW	bm_windows[];

	register BM_WINDOW	*bptr;

	if (window_id >= BP_MAX_WINDOWS)
	{ /* bad window id */
		return ((ULONG) BM_WINDOWS_PARM);
	}

	bptr = bm_windows;
	bptr->maximum = 0;
	bptr->current = 0;
	bptr->next_pend = (BM_PEND_T *) NULL;
	bptr->pend_tail = (BM_PEND_T *) NULL;
	return ((ULONG) BM_SUCCESS);
}

ULONG	bp_get_window(window_id)
	USHORT	window_id;
{
	if (window_id >= BP_MAX_WINDOWS)
	{ /* bad window id */
		return ((ULONG) BM_WINDOWS_PARM);
	}
	if (bm_windows[window_id].maximum == 0)
	{ /* window id not open */
		return ((ULONG) BM_WINDOWS_PARM);
	}

	return((ULONG) bm_windows[window_id].current);
}

ULONG	bp_set_window(window_id, window_size)
	USHORT		window_id;
	USHORT		window_size;
{
	if (window_id >= BP_MAX_WINDOWS)
	{ /* bad window id */
		return ((ULONG) BM_WINDOWS_PARM);
	}
	if (bm_windows[window_id].maximum == 0)
	{ /* window id not open */
		return ((ULONG) BM_WINDOWS_PARM);
	}
	bm_windows[window_id].maximum = window_size;
	return((ULONG) BM_SUCCESS);
}

ULONG	bp_alloc_window(window_id)
	USHORT		window_id;
{
	extern BM_WINDOW	bm_windows[];

	register BM_WINDOW	*bptr;

	if (window_id >= BP_MAX_WINDOWS)
	{ /* bad window id */
		return ((ULONG) BM_WINDOWS_PARM);
	}
	if (bm_windows[window_id].maximum == 0)
	{ /* window id not open */
		return ((ULONG) BM_WINDOWS_PARM);
	}

	bptr = bm_windows;
	if (bptr->current < bptr->maximum)
	{ /* OK -- allocate a token */
		bptr->current++;
		return((ULONG) BM_SUCCESS);
	}
	else
	{
		return((ULONG) BM_NO_WINDOW_AVAIL);
	}
}


ULONG	bp_rel_window(window_id)
	USHORT		window_id;
{
	extern BM_WINDOW	bm_windows[];

	register BM_WINDOW		*bptr;
	register BM_PCB_T		*pcb;

	int						pri;

	if (window_id >= BP_MAX_WINDOWS)
	{ /* bad window id */
		return ((ULONG) BM_WINDOWS_PARM);
	}
	if (bm_windows[window_id].maximum == 0)
	{ /* window id not open */
		return ((ULONG) BM_WINDOWS_PARM);
	}

	bptr = bm_windows;
	if (bptr->current)
	{ /* OK -- give back a token */
		bptr->current--;
		/* NOTE:  If requests are queued for this resource window, we
			we move the request control structure to the pcb ready list
			so that the request will satisfied.
		*/
		if (bptr->next_pend)
		{ /* a request is pending */
			pcb = bp_pcb_array[bptr->next_pend->resource_id];

    		DISABLE(pri,bpp_disable_level);
			if (pcb->pend_list.next_pend)
			{
				pcb->pend_tail->next_pend = bptr->next_pend;
				pcb->pend_tail = bptr->next_pend;
			}
			else
			{
				pcb->pend_list.next_pend = bptr->next_pend;
				pcb->pend_tail = bptr->next_pend;
			}
			bptr->next_pend = bptr->next_pend->next_pend;
			pcb->pend_tail->next_pend = (BM_PEND_T *) NULL;
			ENABLE(pri);
		}
		return((ULONG) BM_SUCCESS);
	}
	else
	{
		return((ULONG) BM_WINDOWS_PARM);
	}
}
