/*MH************************ MODULE HEADER ****************************
**
** @(#)bpp_local.c	35.6
**
**  MODULE NAME:	bp_local.c
**
**  DESCRIPTION:	This module contains the local board-specific
**					function required for the BPP/BPPIR implemenataion
**					on the MVME147.  This functions consist of service
**					routines specified in the BPP specification as well
**					as local environment initialization.
**
**  CONTENTS:		bp_get_memory()
**					bp_free_memory()
**					bp_local_init()
**
**
**  SYSTEM GLOBALS:
**
**  MODULE GLOBALS:
**
**  NOTES:	
**
**  UPDATE LOG:
**  Date		Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**	12/29/87	jeh			1.0	Initial coding.
**
**********************************************************************/

#include "sys/errno.h"
#include "sys/param.h"
#include "sys/sysmacros.h"
#include "sys/types.h"
#include "sys/signal.h"
#include "sys/dir.h"
#include "sys/immu.h"
#include "sys/region.h"
#include "sys/proc.h"
#include "sys/user.h"
#include "sys/mvmecpu.h"

#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/bm.h>
#include <sys/bppv2/ccb.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

#include <sys/bppv2/mvme147/bpp_local.h>
#include <sys/osi/common/bdb.h>
#include <sys/bppv2/mvme147/bp_ci.h>

/******************************************************************
   Misc. variables for local pipe to c37x driver.  We need a cpu_ent
	for the new BPP to get buffers from the c37x pools, and we need
	local and remote PCT_T tables for the c37x driver to use the
	pipe read and write routines from the new BPP driver.
*******************************************************************/

extern ULONG		log_mask;
extern PCT_T		*bp_local_pipe_tbl[];
extern PCT_T		*bp_remote_pipe_tbl[];
extern CCB_T		*bp_ccb_tbl[];
static CCB_T		bp_ccb[1];
CPU_ENT				bp_cpu_ent;			/* ci driver CPU_ENT for this side */
static CH_TBL		bp_rc;				/* remote channel table for CPU_ENT */

/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_local_init()
**
**  DESCRIPTION:	MVME147-specific initialization.
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**	12/29/87	jeh			1.0	Initial coding.
**
************************************************************************/

static UBYTE *bp_memory_base;
static ULONG bpp_memory_space = 1024;

LONG	bp_local_init()

{
	extern	void	ma_main();
	extern	void	sm_main();
	extern	void	sm_init();
	extern	ULONG	bp_intr();
	extern	ULONG	c37xintr();
	extern	BYTE	*bp_malloc();
	extern	LONG	bpp_disable_level;
	extern	USHORT	bp_default_wid;
	extern	CPU_ENT ci_cpu[];
	extern	CPU_ID	bpp_local_cpu;	/* should be CPU_ID, but compiled to
										byte access rather than word that way */

	extern	ULONG	bp_pwrite();
	extern	ULONG	bp_pread();
	extern	UBYTE	*bppv2_local_cpu;
	extern	LONG	**bppv2_remote_pipe_tbl;
	extern	ULONG	(*bppv2_pread)();
	extern	ULONG	(*bppv2_pwrite)();
	extern       int     cpuboard;

	register CPU_ENT *c;		/* CPU table entry address -- old BPP */
	register PCT_T	 *pct;
	register int	 i;
	ULONG			err_return;

	/* sanity check - is this really a 147? */
	if(cpuboard != VME147)
		return;

	bpp_local_cpu = VME147_CPU_NO;		/* kludge for now -- cpu number
											assignment is given to responders
											and must be negotiated among
											initiators */

	bp_memory_base = (UBYTE *) bp_malloc(bpp_memory_space);
	if (bp_memory_base == NULL)
	{
		LOG("BPP:  Initialization, memory allocation error\n");
		return (BPP_INIT_ERR);
	}
	
	bppv2_local_cpu = &bpp_local_cpu;
	bppv2_remote_pipe_tbl = (LONG **) &bp_remote_pipe_tbl[0];
	bppv2_pread = bp_pread;
	bppv2_pwrite = bp_pwrite;
	bpp_disable_level = 5;

	/* Adjust bpp disable level to sr format */
	bpp_disable_level = (bpp_disable_level << 8) + 0x2000;	

	/* Initialize pipe and driver tables.  Open a channel to the old
		BPP.  Initialize the CPU_ENT for MVME147 used by the c37x driver
		and also the pipe control tables used by the new BPP. */

	c = &ci_cpu[VME147_CPU_NO - CPU_OFFSET];
	c->lc.local_cpu = CPU_NO;		/* set lc cpu numbers */
	c->lc.remote_cpu = VME147_CPU_NO;
	c->no_ipc = FALSE;				/* Assume an active state */
	c->no_com = FALSE;
	c->regbase = MVME147_OFFSET;	/* set offset to define cpu type */

	c->memadr = (LONG) bp_malloc(BPP_C37X_MEM_SZ, &err_return);
	if (c->memadr == NULL)
	{
		c->no_ipc = TRUE;				/* Set channel inactive */
		c->no_com = TRUE;
		LOG("BPP:  Initialization, memory allocation error\n");
		return (BPP_INIT_ERR);
	}
	
	cispinit(VME147_CPU_NO - CPU_OFFSET);
/*	
**	This is a bogus test since VME147_CPU_NO = 8 and c37x_add[i] contains
**  addresses like ffff0000 or such.  Hence, i always gets incremented to
**  "c37x_cnt" and thus is an index into the 7th element in the CPU table
**  array.  Also, there is only one argument expected on the subsequent
**  call to ci_rc_init():  a cpu number.  wlv 8/31/88
**  
**  for (i = 0; i < c37x_cnt; i++)
**	{ 
*/
		/* find the cpu_tab entry index for VME147 */
/*
**		if ( (ULONG) c37x_add[i] == VME147_CPU_NO)
**			break;
**	}
*/
	ci_rc_init(VME147_CPU_NO - CPU_OFFSET);		/* wlv 8/31/88 */
	ci_reg_init(VME147_CPU_NO - CPU_OFFSET);

	c->rc->intr_level = 5;  /* wlv 8/31/88 */
	c->lc.intr_level  = 5;  /* wlv 8/31/88 */
	bp_cpu_ent.lc = *(c->rc);
	bp_cpu_ent.rc = &bp_rc;
	*bp_cpu_ent.rc = c->lc;
	bp_cpu_ent.rr = c->lr;
	bp_cpu_ent.lr = c->rr;

/* swap local and remote cpu numbers for this side's CPU_ENT */
	bp_cpu_ent.lc.local_cpu = bp_cpu_ent.rc->remote_cpu;
	bp_cpu_ent.lc.remote_cpu = bp_cpu_ent.rc->remote_cpu;
	bp_cpu_ent.rc->local_cpu = bp_cpu_ent.lc.local_cpu;
	bp_cpu_ent.rc->remote_cpu = bp_cpu_ent.lc.remote_cpu;

	/* initialize pct's for pipe management */
	bp_local_pipe_tbl[CPU_NO] = &bp_ccb[0].pit;
	bp_remote_pipe_tbl[CPU_NO] = &bp_ccb[0].prt;
	bp_ccb_tbl[CPU_NO] = bp_ccb;

/* Initilaize the local Pipe Control Table */
	pct = bp_local_pipe_tbl[CPU_NO];
	pct->bpp_type = INIT_RESP;
	pct->bpp_mode = INITIATOR;
	pct->cpu_id = VME147_CPU_NO;
	pct->cpu_type = VME147;
	pct->get_bpe = (BPE_T *) c->lc.get_bpe;
	pct->ret_bpe = (BPE_T *) c->lc.ret_bpe;
	pct->bpp_read = (BPE_T *) c->lc.bpp_read;
	pct->bpp_write = (BPE_T *) c->lc.bpp_write;
	pct->rmt_arg = VME147_CPU_NO;
	pct->rmt_intr = (PTF) c37xintr;

/* Initialize the remote Pipe Control Table */	
	pct = bp_remote_pipe_tbl[CPU_NO];
	pct->bpp_type = RESPONDER;
	pct->bpp_mode = RESPONDER;
	pct->cpu_id = CPU_NO;
	pct->cpu_type = VME147;
	pct->get_bpe = (BPE_T *) c->rc->get_bpe;
	pct->ret_bpe = (BPE_T *) c->rc->ret_bpe;
	pct->bpp_read = (BPE_T *) c->rc->bpp_read;
	pct->bpp_write = (BPE_T *) c->rc->bpp_write;
	pct->rmt_arg = CPU_NO;
	pct->rmt_intr = (PTF) bp_intr;
	
	/* Create the default MPB buffer pool */
/****************************************************
	if (bp_create_mpb_pool((BYTE *) NULL, (ULONG) 0, (ULONG) BP_DEF_MPB_SIZE,
							(USHORT) BP_DEF_POOL_ID, (USHORT) 0,
							(USHORT) BP_DEF_LOCAL_MPBS, &err_return))
	{
		LOG("BPP:  Init unable to create default MPB pool, err = %x\n", err_return);
	}
****************************************************/
	/* Open default window to default MPB pool */
/****************************************************
	bp_default_wid = bp_open_window(BP_DEF_LOCAL_MPBS/2);
****************************************************/

	/* register SMA and MAC and perform initialization */

	if (bp_open_ip((USHORT) SMA_IP, sm_main, (ULONG) 0, (PTF) 0, 0))
	{
		LOG("MAC:  Cannot open SM IP\n");
	}
	sm_init();

	if (bp_open_ip((USHORT) MAC_IP, ma_main, (ULONG) 0, (PTF) 0, 0))
	{
		LOG("MAC:  Cannot open MAC IP\n");
	}
	ma_init();
}
/* **************************************************
* 
* Module: 	bp_get_memory()
*
* Purpose: 	This function allocates a block of shared memory
*		of the size specified and updates the bp_memory_base pointer
*		by that amount.
*
* Data used:	global bp_memory_base pointer
*
* Input Parms:	
*			type		type of memory -- SHARED or PRIVATE
*			base_addr	address of a pointer to return address of block
*			size 		size (in bytes) of requested memory chunk
*
* Returns:	
*			BPP_SUCCESSFUL		Allocation was successful.
*			INSUFF_MEMORY		Allocation was unsuccessful.
*
*/

bp_get_memory(type, base_addr, size)
	LONG	type;
	UBYTE	**base_addr;
    LONG   size;
{
	register LONG	temp_size;

	if ((temp_size = bpp_memory_space - size) < 0 )
		return(INSUFF_MEMORY);

	bpp_memory_space = temp_size;
	bp_memory_base += (((size + 1)/ 2) * 2);/* force to even byte boundary */
	*base_addr = (UBYTE *) bp_memory_base;
    return(BPP_SUCCESSFUL);
}

/* **************************************************
* 
* Module: 	bp_free_memory()
*
* Purpose: 	This function frees a block of shared memory of the size 
*			specified.
*
* Data used:	bp_memory_base 	pointer
*
* Input Parms:	
*			type		type of memory -- SHARED or PRIVATE
*			base_addr	address of a pointer to return address of block
*			size 		size (in bytes) of requested memory chunk
*
* Returns:	
*			BPP_SUCCESSFUL		Allocation was successful.
*
*/

ULONG	bp_free_memory(type, base_addr, size)
	LONG	type;
	UBYTE	*base_addr;
	LONG	size;
{
}

UBYTE	*bp_local_addr(ptr)
	UBYTE	*ptr;
{
	return(ptr);
}

UBYTE	*bp_system_addr(ptr)
	UBYTE	*ptr;
{
	return(ptr);
}

/*FH***************************** FUNCTION HEADER *************************
**
**  FUNCTION NAME:	bp_malloc.c
**
**  DESCRIPTION:    	bp_malloc() attempts to get a contiguous
**			block of memory pools that is not split across a four (4)
**			megabyte boundary.
**
**  Global Data:	
**
**  External Refs:
**
**  Input Parameters:	chunksz - size of needed memory block
**
**  Return Parameters: 	SUCCESS		base address of allocated memory
**						FAILURE		zero
**
**  NOTES:
**
**  CHANGES:     DESCRIPTION                       DATE     ENGINEER
**  ___________________________________________________________________
**
*******************************************************************************/

#define FOUR_MEG_SPLIT(a, s)	((a & 0xc00000) != ((a+s-1) & 0xc00000))


BYTE *bp_malloc(chunksz)
LONG    chunksz;
{
        LONG    i, n;
        LONG    clicks;
        BYTE    *addr_a, *addr_b;
        LONG    size_a, size_b;
		LONG	memsz; 
		BYTE	*memadr;
        BYTE    errflg;


        clicks = btoc(chunksz);
        memsz  = ctob(clicks);


	memadr = (BYTE *) getcpages(clicks, 3);
	if (memadr == 0)
		goto memerror;

	addr_a = memadr;

	if (FOUR_MEG_SPLIT((LONG) memadr, memsz)) 
	{
		/*
		 * WE are straddling a 4 Meg boundary.  Release memory
		 * past the boundary, try to alloc again, and if succesful,
		 * then release the stuff we were holding on to below the
		 * boundary.
		 */
		addr_a = memadr;
		addr_b = memadr + 0x0400000;
		size_a = addr_b - addr_a;
		size_b = memsz - size_a;

		LOG("alloc_mem: straddle: addr_a = %x size_a = %x\n",
			addr_a, memsz);
		LOG("                     addr_b = %x size_b = %x\n",
			addr_b, size_b);

		freecpages(btoc(addr_b), btoc(size_b));

		memadr = (BYTE *) getcpages(clicks, 3);
		if (memadr == 0) 
		{
			freecpages(addr_a, btoc(size_a));
			goto memerror;
		}

		if (FOUR_MEG_SPLIT((LONG) memadr, memsz))
			goto memerror;

		freecpages(addr_a, btoc(size_a));
	}

	LOG("BPPinit: SUCCESS! memadr = %x, memsz = %x\n", memadr, memsz);

	return(memadr);


memerror:
        LOG("BPPinit:  memory allocation error!\n");

        return(0);
}

bp_is_private(addr)
	BYTE	*addr;
{
	return(TRUE);
}
/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_kludge_in()
**
**  DESCRIPTION:	Map from BUF to BDB structure and place the adjusted
**					pointer in the epb header for the local environment.
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/

BP_MPB_T	*bp_kludge_in(psepb)
	register PSEPB	*psepb;
{
	register PSBUF	*psbuf;
	register BDB	*bdb;

	psepb = (PSEPB *) ((BYTE *)psepb - MPB_HDR_SIZE);

	if (psepb->epb.hdr.buf != NULL)
	{  /* map BUF to BDB */
		psbuf = (PSBUF *) psepb->epb.hdr.buf;
		psbuf = (PSBUF *) ((BYTE *) psbuf - BUF_FILL - BDB_HDR_SIZE 
							-sizeof(BP_MPB_T));

		bdb = (BDB *) psbuf->fill;
		psepb->epb.hdr.buf = (BUF *) bdb;
	
		bdb->pid	 = 2;
		bdb->next_bdb	 = NULL;
		bdb->buff_size	 = PDU_SIZE;
		bdb->buff_offset = BUF_HDR_SIZE + BUF_FILL;;
		bdb->buff_length = psbuf->buf.length;
		bdb->data	 = bdb->dummy;
	}
	else
	{ /* no BDB attached */
		psepb->epb.hdr.buf = (BUF *) NULL;
	}

	psepb->mpb.dst_id.cpu = (USHORT) GET_CPU(psepb->epb.hdr.dst_id);
	psepb->mpb.dst_id.ip = (USHORT) GET_IP(psepb->epb.hdr.dst_id);
	psepb->mpb.src_id.cpu = (USHORT) GET_CPU(psepb->epb.hdr.src_id);
	psepb->mpb.src_id.ip = (USHORT) GET_IP(psepb->epb.hdr.src_id);
	psepb->mpb.data_size = (USHORT) psepb->epb.hdr.epb_size;
	psepb->mpb.data = (UBYTE *) &psepb->epb;

	return((BP_MPB_T *) psepb);
}
/*FH************************ FUNCTION HEADER ****************************
**
**  FUNCTION NAME:      bp_kludge_out()
**
**  DESCRIPTION:	Map from BDB to BUF structure.  If MAC or LLC have
**					removed header, we will have to copy to move data
**					up next to the BUF header.  We can avoid copying for only
**					one of two different header lengths.  Ethernet
**					has zero bytes of header removed, and OSI has MAC and
**					LLC header removed.
**
**  SYSTEM GLOBALS:	
**
**  INPUT:		
**
**  OUTPUT:		
**
**  RETURNS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
************************************************************************/

BP_MPB_T	*bp_kludge_out(psepb)
	register PSEPB	*psepb;

{
	register PSBUF	*psbuf;
	register BDB	*bdb;

	if (psepb->epb.hdr.buf != NULL)
	{
		bdb = (BDB *) psepb->epb.hdr.buf;
		psbuf = (PSBUF *) ((BYTE *) bdb - sizeof(BP_MPB_T));

		if (bdb->buff_offset)
		{ /* Oh shit! Move data to start of buf */
			o_bytecopy(bdb->data+bdb->buff_offset, psbuf->buf.data, bdb->buff_length);
		}
		psbuf->buf.length = bdb->buff_length;
		psbuf->buf.next = (BUF *) NULL;
	
		psepb->epb.hdr.src_id = PUT_ID(bpp_local_cpu, 
									GET_IP(psepb->epb.hdr.src_id));
		psepb->epb.hdr.buf = (BUF *) &psbuf->buf;
	}
	else
	{ /* no BDB attached */
		psepb->epb.hdr.buf = (BUF *) NULL;
	}

	return((BP_MPB_T *) &psepb->epb);
}
