/*	@(#)cispinit.c	35.1	*/

/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	cispinit.c
**
**  DESCRIPTION:	This module contains the special initialization
**			of the pipe between the c37x driver and the version 2
**			BPP driver.
**
**  FUNCTIONS:		
**
**  NOTES:
**
**  UPDATE LOG:
**  Date		Name		V/R	Comments
**  --------	----------- ---	-----------------------------
**
***********************************************************************/

#include <sys/types.h>
#include <sys/param.h>
#include <sys/immu.h>
#include <sys/region.h>
#include <sys/proc.h>

#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/bm.h>
#include <sys/bppv2/ccb.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

#include <sys/bppv2/mvme147/bpp_local.h>
#include <sys/osi/common/bdb.h>
#include <sys/bppv2/mvme147/bp_ci.h>

extern	CPU_ID	bpp_local_cpu;

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	cispinit()	
**
**  DESCRIPTION:	Initializes the Common Interface's memory
**			pool.
**
**  CONTENTS:		First, the function will allocate memory for
**			the remote channel table. Next, it will 
**			initialize the memory pools for the local/remote
**			BPE, EPB, and BUF pools.
**
**  PARAMETERS:		
**
**	Input:		cpu		- CPU number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**
**********************************************************************/

	
VOID
cispinit(cpu)

register USHORT	 cpu;				/* CPU number */
{
	register BPE	 *bpe;			/* Current BPE address */
	register PSBUF	 *buf;			/* Current BUF address */
	register DEFAULT i;			/* Just a counter */
	register PSEPB *max_epb;		/* Current EPB address */
	register CPU_ENT *c;			/* CPU entry address */

	/*
	**	Initialize the local's BPE pool.
	*/

	c = &ci_cpu[cpu];

	bpe = (BPE *)((UBYTE *)(c->rc = (CH_TBL *)c->memadr) + sizeof(CH_TBL));

	c->lc.bpp_write = c->rc->bpp_read = bpe; /* Outgoing pipe terminator */
	c->lc.bpp_write->next = (struct bp_env *) 0xffff;	/* Mark as terminator */
	c->lc.bpp_write->flags = 0x0001;	/* Mark as terminator */

	c->lc.get_bpe = ++bpe;			/* Head of local's pool */

	for (i = 2; i < N_BPES; ++i)		/* Pool member */
	{
		bpe->next = bpe + 1;
		(bpe++)->flags = 0x0000;
	}

	bpe->next = (BPE *) (struct bp_env *) 0xffff;		/* New BPP terminator */
	bpe->flags = 0x0001;			/* Terminator */

	c->rc->ret_bpe = bpe;			/* Tail of local's pool */

	/*
	**	Initialize the remote's BPE pool.
	*/

	c->lc.bpp_read = c->rc->bpp_write = ++bpe; /* Incoming pipe terminator */
	c->lc.bpp_read->next = ( struct bp_env *) 0x0001;		/* Mark as terminator */
	c->lc.bpp_read->flags = 0x0001;		/* Mark as terminator */

	c->rc->get_bpe = ++bpe;			/* Head of remote's pool */

	for (i = 2; i < N_BPES; ++i)		/* Pool member */
	{
		bpe->next = bpe + 1;
		(bpe++)->flags = 0x0000;
	}

	bpe->next = ( struct bp_env *) 0x0001;			/* Terminator */
	bpe->flags = 0x0001;			/* Terminator */

	c->lc.ret_bpe = bpe;			/* Tail of remote's pool */

	/*
	**	Initialize the local's EPB pool.
	*/

	c->res_epb = (CE_EPB *)&(max_epb = (PSEPB *)(++bpe))->epb;

	c->lr.epb_head = (EPB *) &(++max_epb)->epb;	/* Local's head tail */

	for (i = 2; i < N_R_PSEPBS; ++i)		/* Pool member */
	{
		max_epb->epb.hdr.next = (EPB *) &(max_epb + 1)->epb;
		max_epb->epb.hdr.mem_id = PUT_PID(c->lc.local_cpu, c->lc.remote_cpu);
		(max_epb++)->epb.hdr.user_def = MEMBER;
		max_epb->mpb.sys_hdr.creator.cpu 	 = bpp_local_cpu;
		max_epb->mpb.sys_hdr.creator.pool_id = BP_LOC_EPB_ID;
		max_epb->mpb.sys_hdr.manager.cpu 	 = bpp_local_cpu;
		max_epb->mpb.sys_hdr.manager.pool_id = BP_LOC_EPB_ID;
		max_epb->mpb.sys_hdr.partnercpu		 = CPU_NO;
		max_epb->mpb.sys_hdr.buf_status	 	= SHARED_BUF | BUF_IN_POOL;
		max_epb->mpb.sys_hdr.mpb_size 		 = sizeof(PSEPB);
		max_epb->mpb.sys_hdr.window_id  	 = NULL;
	}	

	max_epb->epb.hdr.next     = (EPB *)NULL;
	max_epb->epb.hdr.mem_id = PUT_PID(c->lc.local_cpu, c->lc.remote_cpu);
	max_epb->epb.hdr.user_def = TERMINATOR;

	c->rr.epb_tail = (EPB *) &max_epb->epb;	/* Local's pool tail */

	/*
	**	Initialize the remote's EPB pool.
	*/

	c->rr.epb_head = (EPB *) &(++max_epb)->epb;	/* Remote's pool head */

	for (i = 1; i < N_R_PSEPBS; ++i)		/* Pool member */
	{
		max_epb->epb.hdr.next = (EPB *) &(max_epb + 1)->epb;
		max_epb->epb.hdr.mem_id = PUT_PID(c->lc.remote_cpu, c->lc.local_cpu);
		(max_epb++)->epb.hdr.user_def = MEMBER;
		max_epb->mpb.sys_hdr.creator.cpu 	 = bpp_local_cpu;
		max_epb->mpb.sys_hdr.creator.pool_id = BP_RMT_EPB_ID;
		max_epb->mpb.sys_hdr.manager.cpu 	 = bpp_local_cpu;
		max_epb->mpb.sys_hdr.manager.pool_id = BP_RMT_EPB_ID;
		max_epb->mpb.sys_hdr.partnercpu		 = CPU_NO;
		max_epb->mpb.sys_hdr.buf_status	 	= SHARED_BUF | BUF_IN_POOL;
		max_epb->mpb.sys_hdr.mpb_size 		 = sizeof(PSEPB);
		max_epb->mpb.sys_hdr.window_id  	 = NULL;
	}	

	max_epb->epb.hdr.next     = (EPB *)NULL;
	max_epb->epb.hdr.mem_id = PUT_PID(c->lc.remote_cpu, c->lc.local_cpu);
	max_epb->epb.hdr.user_def = TERMINATOR;

	c->lr.epb_tail = (EPB *) &max_epb->epb;	/* Remote's pool tail */

	/*
	**	Initialize the local's BUF pool.
	*/

	buf = (PSBUF *) (PSBUF *) ++max_epb;
	c->lr.buf_head = (BUF *) &buf->buf;

	for (i = 1; i < N_L_PSBUFS+N_R_PSBUFS; ++i)
	{
		buf->buf.next = (BUF *) &(buf + 1)->buf;
		++buf;
		buf->mpb.sys_hdr.creator.cpu 	 = bpp_local_cpu;
		buf->mpb.sys_hdr.creator.pool_id = BP_LOC_BUF_ID;
		buf->mpb.sys_hdr.manager.cpu 	 = bpp_local_cpu;
		buf->mpb.sys_hdr.manager.pool_id = BP_LOC_BUF_ID;
		buf->mpb.sys_hdr.partnercpu		 = CPU_NO;
		buf->mpb.sys_hdr.buf_status	 	= SHARED_BUF | BUF_IN_POOL;
		buf->mpb.sys_hdr.mpb_size 		 = sizeof(PSEPB);
		buf->mpb.sys_hdr.window_id  	 = NULL;
	}

	buf->buf.next = (BUF *)NULL;

	c->lr.buf_tail = &buf->buf;

	/*
	**	Initialize the remote's PSBUF pool.
	*/
/***************************************************************
**  Only one combined pool!!
	c->rr.buf_head = &(++buf)->buf;

	for (i = 1; i < N_R_PSBUFS; ++i)
	{
		buf->buf.next = (BUF *) &(buf + 1)->buf;
		++buf;
		buf->mpb.sys_hdr.creator.cpu 	 = bpp_local_cpu;
		buf->mpb.sys_hdr.creator.pool_id = BP_LOC_BUF_ID;
		buf->mpb.sys_hdr.manager.cpu 	 = bpp_local_cpu;
		buf->mpb.sys_hdr.manager.pool_id = BP_LOC_BUF_ID;
		buf->mpb.sys_hdr.partnercpu		 = CPU_NO;
		buf->mpb.sys_hdr.buf_status	 	= SHARED_BUF | BUF_IN_POOL;
		buf->mpb.sys_hdr.mpb_size 		 = sizeof(PSEPB);
		buf->mpb.sys_hdr.window_id  	 = NULL;
	}

	buf->buf.next = (BUF *)NULL;

	c->rr.buf_tail = &buf->buf;
**************************************************************/
}
