/*	@(#)ci_debug.c	35.1	*/
#ifdef DEBUG
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci_debug.c
**
**  DESCRIPTION:	This module contains the debug dump
**			functions.
**
**  FUNCTIONS:		ci_dump_epb, ci_dump_bpe
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  03/20/86	Dana Holgate	1.1	Initial Coding.
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x.
**  09/10/87	dtk             2.1     Changed location of header files
**
***********************************************************************/


#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_dump_epb()	
**
**  DESCRIPTION:	Display/dump the passed EPB contents.
**
**  CONTENTS:		This function will display the passed message
**			along with the passed EPB's contents.
**
**  PARAMETERS:		
**
**	Input:		note		- Message address,
**			epb		- EPB pointer.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_dump_epb(note, epb)

register BYTE *note;		/* Message */
register EPB  *epb;		/* EPB pointer */
{
	printf("\n%s: epb = 0x%x\n", note, epb);
	printf("mem_id = 0x%x\tsrc_id = 0x%x\n", epb->mem_id, epb->src_id);
	printf("dst_id = 0x%x\tsize   = 0x%x\n", epb->dst_id, epb->epb_size);
	printf("type   = 0x%x\tstatus = 0x%x\n", epb->type, epb->status);
	printf("buf    = 0x%x\tu_def  = 0x%x\n", epb->buf, epb->user_def);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_dump_bpe()	
**
**  DESCRIPTION:	Display/dump the passed BPE contents.
**
**  CONTENTS:		This function will display the passed message
**			along with the passed BPE's contents.
**
**  PARAMETERS:		
**
**	Input:		note		- Message address,
**			bpe		- BPE pointer.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_dump_bpe(note, bpe)

register BYTE *note;		/* Message */
register BPE  *bpe;		/* BPE pointer */
{
	printf("\n%s: bpe = 0x%x\n", note, bpe);
	printf("next   = 0x%x\tmsg/cmd = 0x%x\n", bpe->next, bpe->msg);
	printf("dst_id = 0x%x\tsrc_id  = 0x%x\n", bpe->dst_id, bpe->src_id);
	printf("flags  = 0x%x\n", bpe->flags);
}
#endif
