/*	@(#)ci_init.c	35.1	*/
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci_init.c
**
**  DESCRIPTION:	This module contains the Common Interface
**			initialization functions.
**
**  FUNCTIONS:		ci_mem_init, ci_pipe_init, ci_reg_init,
**			ci_rc_init,  ci_bpp_wait
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Coding.
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x.
**  06/10/87	R. Robinson	1.2	Fixed bug in ci_bpp_wait.
**  05/30/88	jeh             2.1     Changed location of header files
**  05/30/88	jeh		2.1	Removed cache-disable for VME130.
**					All shared memory is now cache-
**					disabled at time of allocation.
**					Alsowrite initial pointer as 2
**					shorts to eliminate VME130 hardware
**					dependency.
***********************************************************************/

#include <sys/types.h>
#include <sys/param.h>
#include <sys/immu.h>
#include <sys/region.h>
#include <sys/proc.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

extern	EPB	*ci_pread();		/* Read pipe input */
extern	DEFAULT	ci_pwrite();		/* Write to pipe */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_mem_init()	
**
**  DESCRIPTION:	Initializes the Common Interface's memory
**			pool.
**
**  CONTENTS:		First, the function will allocate memory for
**			the remote channel table. Next, it will 
**			initialize the memory pools for the local/remote
**			BPE, EPB, and BUF pools.
**
**  PARAMETERS:		
**
**	Input:		cpu		- CPU number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_mem_init(cpu)

register USHORT	 cpu;				/* CPU number */
{
	register BPE	 *bpe;			/* Current BPE address */
	register BUF	 *buf;			/* Current BUF address */
	register DEFAULT i;			/* Just a counter */
	register MAX_EPB *max_epb;		/* Current EPB address */
	register CPU_ENT *c;			/* CPU entry address */

	/*
	**	Initialize the local's BPE pool.
	*/

	c = &ci_cpu[cpu];

	bpe = (BPE *)((UBYTE *)(c->rc = (CH_TBL *)c->memadr) + sizeof(CH_TBL));

	c->lc.bpp_write = c->rc->bpp_read = bpe; /* Outgoing pipe terminator */
	c->lc.bpp_write->flags = 0x0001;	/* Mark as terminator */

	c->lc.get_bpe = ++bpe;			/* Head of local's pool */

	for (i = 2; i < N_BPES; ++i)		/* Pool member */
	{
		bpe->next = bpe + 1;
		(bpe++)->flags = 0x0000;
	}

	bpe->flags = 0x0001;			/* Terminator */

	c->rc->ret_bpe = bpe;			/* Tail of local's pool */

	/*
	**	Initialize the remote's BPE pool.
	*/

	c->lc.bpp_read = c->rc->bpp_write = ++bpe; /* Incoming pipe terminator */
	c->lc.bpp_read->flags = 0x0001;		/* Mark as terminator */

	c->rc->get_bpe = ++bpe;			/* Head of remote's pool */

	for (i = 2; i < N_BPES; ++i)		/* Pool member */
	{
		bpe->next = bpe + 1;
		(bpe++)->flags = 0x0000;
	}

	bpe->flags = 0x0001;			/* Terminator */

	c->lc.ret_bpe = bpe;			/* Tail of remote's pool */

	/*
	**	Initialize the local's EPB pool.
	*/

	c->res_epb = (CE_EPB *)(max_epb = (MAX_EPB *)(++bpe));

	c->lr.epb_head = (EPB *)(++max_epb);	/* Local's head tail */

	for (i = 2; i < N_EPBS; ++i)		/* Pool member */
	{
		max_epb->hdr.next = (EPB *)(max_epb + 1);
		(max_epb++)->hdr.user_def = MEMBER;
	}	

	max_epb->hdr.next     = (EPB *)NULL;
	max_epb->hdr.user_def = TERMINATOR;

	c->rr.epb_tail = (EPB *)max_epb;	/* Local's pool tail */

	/*
	**	Initialize the remote's EPB pool.
	*/

	c->rr.epb_head = (EPB *)(++max_epb);	/* Remote's pool head */

	for (i = 1; i < N_EPBS; ++i)		/* Pool member */
	{
		max_epb->hdr.next = (EPB *)(max_epb + 1);
		(max_epb++)->hdr.user_def = MEMBER;
	}	

	max_epb->hdr.next     = (EPB *)NULL;
	max_epb->hdr.user_def = TERMINATOR;

	c->lr.epb_tail = (EPB *)max_epb;	/* Remote's pool tail */

	/*
	**	Initialize the local's BUF pool.
	*/

	c->lr.buf_head = buf = (BUF *)(++max_epb);

	for (i = 1; i < N_L_BUFS; ++i)
	{
		buf->next = buf + 1;
		++buf;
	}

	buf->next = (BUF *)NULL;

	c->lr.buf_tail = buf;

	/*
	**	Initialize the remote's BUF pool.
	*/

	c->rr.buf_head = ++buf;

	for (i = 1; i < N_R_BUFS; ++i)
	{
		buf->next = buf + 1;
		++buf;
	}

	buf->next = (BUF *)NULL;

	c->rr.buf_tail = buf;
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_pipe_init()	
**
**  DESCRIPTION:	Establish pipe communication with MVME37x
**			device.
**
**  CONTENTS:		This function will establish a pipe communication
**			with the remote MVME37x device by first sending
**			it the remote channel table address and a
**			shared resource EPB.
**
**  PARAMETERS:		
**
**	Input:		cpu		- CPU number.
**
**	Output:		NONE
**
**	Returns:	SUCCESS 	- Pipe communication is
**					  established,
**
**			FAILURE		- Cannot establish pipe.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  08/28/86	Dana Holgate	1.1	Added Cache Accelerator control
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**	05/30/88	jeh			2.1	Changed write of rc address from long
**								to two shorts.
**********************************************************************/

BOOLEAN
ci_pipe_init(cpu)

register USHORT	 cpu;			/* CPU number */
{
	 BOOLEAN ci_bpp_wait();   	/* Wait for pipe status */

	register ULONG	 i;		/* Just a counter */
	register EPB	 *epb;		/* EPB pointer */
	register CE_EPB	 *ce_epb;	/* CE_EPB pointer */
	register CPU_ENT *c;		/* CPU table entry address */

	/*
	**	Wait for status register to contain BPP_RUN.
	*/

	if (ci_bpp_wait(cpu) == FALSE)
	{
		printf("%sinit: Unable to create pipe, CPU %d not ready\n", ci_dev, (cpu + CPU_OFFSET));
		return (FAILURE);
	}


	/*
	**	Next, put the remote channel tables address into
	**	the MVME37x data register.  The address must be written as two
	**	short words, since at least some 37x boards respond only to D16
	**	accesses.
	*/

	c = &ci_cpu[cpu];
	*((USHORT *) (c->boardadr + c->regbase)) =
		((ULONG) c->rc >> 16) & 0xffff;
	*((USHORT *) (c->boardadr + c->regbase + 2)) =
		((ULONG) c->rc & 0xffff);

	/*
	**	Next, set the control register to BPP_PIPE and
	**	wait for the MVME37x response. If either an
	**	invalid or no response occurs, return with an
	**	error
	*/

	BPPCR(cpu, BPP_PIPE);

	for (i = BPP_WAIT_TIME; i; --i)
	{
		if ((epb = ci_pread(c)) != (EPB *)NULL)
		{
			if (epb != (EPB *)c->rc)
			{
				printf("%sinit: Unable to create pipe, invalid response from CPU %d\n", ci_dev, (cpu + CPU_OFFSET));

				return (FAILURE);
			}
			else
				break;
		}
	}

	if (i == 0)
	{
		printf("%sinit: Unable to create pipe, no response from CPU %d\n", ci_dev, (cpu + CPU_OFFSET));

		return (FAILURE);
	}
	else
		if (ci_debug)
		{
			printf("Remote version: %s\tCPU number: %d\n", c->rc->version, c->rc->remote_cpu);
			printf("Local  version: %s\tCPU number: %d\n", c->lc.version, c->lc.local_cpu);
		}

	/*
	**	Next, get the CPU number from the remote channel table
	**	and place it in the local channel table. Then initialize
	**	the shared resource EPB and send it to the MVME37x.
	*/

	c->lc.remote_cpu = c->rc->remote_cpu;

	ce_epb = c->res_epb;

	ce_epb->hdr.next      = (EPB *)NULL;
	ce_epb->hdr.mem_id    = PUT_PID(c->lc.local_cpu, c->lc.remote_cpu);
	ce_epb->hdr.dst_id    = PUT_ID(c->lc.remote_cpu, CE_IP);
	ce_epb->hdr.src_id    = PUT_ID(c->lc.local_cpu, CE_IP);
	ce_epb->hdr.epb_size  = sizeof(CE_EPB);
	ce_epb->hdr.type      = CE_PIPE_INIT_REQUEST;
	ce_epb->hdr.status    = NO_ERROR;
	ce_epb->hdr.buf       = c->rr.buf_head;
	ce_epb->hdr.user_def  = 0;

	ce_epb->sh_epb_size   = MAX_CEEPB_SIZE;
	ce_epb->sh_buf_size   = MAX_BUF_SIZE;
	ce_epb->minimum_bufs  = MIN_R_BUFS;
	ce_epb->maximum_bufs  = MAX_R_BUFS;
	ce_epb->epbbuf_count  = N_R_BUFS;
	ce_epb->epb_pool_head = c->rr.epb_head;
	ce_epb->epb_pool_tail = c->rr.epb_tail;

	ce_epb->qid           = 0;
	ce_epb->addr          = 0;
	ce_epb->length        = 0;
	ce_epb->eot           = 0;

	if (ci_pwrite(c, ce_epb) != NO_ERROR)
	{
		printf("%sinit: Unable to create CPU %d pipe, can't send resources\n", ci_dev, (cpu + CPU_OFFSET));

		return (FAILURE);
	}

	/*
	**	Finally, mark the mem_id field of each EPB.
	*/

	epb = c->lr.epb_head;

	while (epb->user_def == MEMBER)
	{
		epb->mem_id = PUT_PID(c->lc.local_cpu, c->lc.remote_cpu);
		epb = epb->next;
	}

	epb->mem_id = PUT_PID(c->lc.local_cpu, c->lc.remote_cpu);

	return (SUCCESS);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_reg_init()	
**
**  DESCRIPTION:	Initialize the registration table.
**
**  CONTENTS:		This function will initialize (i.e. clear out)
**			the registration table.
**
**  PARAMETERS:		
**
**	Input:		cpu		- CPU number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/29/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_reg_init( cpu )
	REG	USHORT	cpu;		/* CPU number */
{
	REG	DEFAULT i;		/* Just a counter */
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	REG_ENT	*r;		/* Registration table entry address */

	c = &ci_cpu[cpu];

	for( i = 0; i < NUM_REG_ENTRIES; ++i )
	{
		r		= &( c->rt[i] );
		r->process_id	= EMPTY;
		r->block	= 0;
		r->proc		= ( struct proc * ) NULL;
		r->rq_head	= ( EPB * ) NULL;
		r->rq_tail	= ( EPB * ) NULL;
		r->rx_int	= NULL;
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_rc_init()	
**
**  DESCRIPTION:	Initialize the remaining elements of the
**			remote channel table.
**
**  CONTENTS:		This function will set the remote channel
**			table's version ID, CPU number, and interrupt
**			level/vector.
**
**  PARAMETERS:		
**
**	Input:		cpu		- CPU number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/29/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_rc_init(cpu)
	REG	USHORT	cpu;		/* CPU number */
{
	REG	DEFAULT i;		/* Just a counter */
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	CH_TBL	*rcp;		/* Remote channel table addres */

	/*
	**	Initialize.
	*/

	c = &ci_cpu[cpu];

	rcp = c->rc;

	/*
	**	Copy current version ID to remote channel table.
	*/

	for( i = 0; i < CI_VER_SZE; ++i )
		rcp->version[i] = c->lc.version[i] = CI_VER[i];

	/*
	**	Assign remote channel table CPU numbers.
	*/

	rcp->local_cpu		= c->lc.local_cpu	= CPU_NO;
	rcp->remote_cpu		= cpu + CPU_OFFSET;

	/*
	**	Set interrupt level and vector number in remote channel
	**	table.
	*/

	rcp->intr_level		= ( UBYTE ) c37x_ile[cpu];
	rcp->intr_vector	= ( UBYTE ) ( c37x_ive[cpu] / 4 );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_bpp_wait()	
**
**  DESCRIPTION:	Waits a limited amount of time for a BPP_RUN
**			in the BPP status register.
**
**  CONTENTS:		This function will returns a TRUE when the
**			BPP status is reached within the time limit.
**			Otherwise it returns FALSE.
**
**  PARAMETERS:		
**
**	Input:		cpu		- CPU/Minor device number.
**
**	Output:		NONE
**
**	Returns:	TRUE		- Status register = BPP_RUN,
**			FALSE		- Status register = ?.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  04/01/86	Dana Holgate	1.1	Initial Release
**  08/28/86	Dana Holgate	1.1	Added Cache Accelerator control
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**  06/10/87	R. Robinson	1.2	Changed OFF and ON to FALSE and
**					TRUE respectively.
**
**********************************************************************/

BOOLEAN
ci_bpp_wait(cpu)

register USHORT	 cpu;			/* CPU/Minor device number */
{
	register DEFAULT timer;		/* Timer count */
	register BOOLEAN result;	/* Function result */

	/*
	**	Initialize variables
	*/

	timer  = BPP_WAIT_TIME;
	result = FALSE;

	/*
	**	Wait for status register.
	*/

	while (timer--)
		if (BPPSR(cpu) == BPP_RUN)
		{
			result = TRUE;
			break;
		}

	return (result);
}
