/*	@(#)ci_malloc.c	35.2	*/
/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_malloc()	
**
**  DESCRIPTION:	Allocate memory for pools.
**
**  CONTENTS:		This function will allocate memory for the BPE,
**			EPB, BUF pools and the remote channel table.
**
**  PARAMETERS:		
**
**	Input:		cpu		- CPU number,
**			chunksz		- Memory size needed.
**
**	Output:		NONE
**
**	Returns:	SUCCESS		- Memory allocated,
**			FAILURE		- No memory.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:		Only R3V4, BS83.03 and later Unix versions are
**			supported.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Release
**  05/01/87	E. Skinner	1.1	Port to System V/68 Release 3.
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**	05/30/88	jeh			2.1	Adapted to new getcpages() interface for
**								cache-disabled allocation, removed #ifdef's,
**								and changed location of header files.
**	09/15/88	jeh			2.2	Removed code to prven allocated memory
**								from crossing 4 meg boundary.
**********************************************************************/

#include <sys/param.h>
#include <sys/sysmacros.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

BOOLEAN
ci_malloc(cpu, chunksz)

USHORT	cpu;				/* CPU number */
ULONG	chunksz;			/* Memory size needed */
{

	register CPU_ENT *c;		/* CPU table entry address */

		 ULONG	 memsz;		/* Block size */
		 ULONG	 clicks;	/* Click count */

	/*
	**	Generate block and click counts.
	*/

	clicks = btoc(chunksz);

	memsz  = ctob(clicks);

	/*
	**	Try to allocate memory.
	*/
	/* The second argument of getcpages() is 3 to indicate that
		cache-disabled memory should be allocated. This technique
		works on R3V4 BS83.03 or later. */

	if (((c = &ci_cpu[cpu])->memadr = getcpages(clicks, 3)) != 0)
		return (SUCCESS);

	return (FAILURE);
}
