/*	@(#)ci_pipe.c	35.1	*/
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci_pipe.c
**
**  DESCRIPTION:	This module contains the Common Interface
**			pipe read and write functions.
**
**  FUNCTIONS:		ci_pwrite, ci_pread
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  04/01/86	Dana Holgate	1.1	Initial Coding.
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x.
**  09/10/87	dtk             2.1     Changed location of header files
**
***********************************************************************/


#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

extern	UBYTE	*bppv2_local_cpu;
extern	LONG	(*bppv2_remote_pipe_tbl)[];
extern	ULONG	(*bppv2_pread)();
extern	ULONG	(*bppv2_pwrite)();

extern	BPE	*ci_get_bpe();		/* Get BPE block */
extern	VOID	ci_rel_bpe();		/* Release BPE block */
extern	DEFAULT	splx();			/* Set priority level */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_pwrite()	
**
**  DESCRIPTION:	Write an EPB to the outgoing pipe.
**
**  CONTENTS:		This function will place the passed EPB
**			on the local channel table's outgoing
**			pipe.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			epb		- EPB pointer.
**
**	Output:		NONE
**
**	Returns:	Write status.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  04/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

DEFAULT
ci_pwrite(c, epb)
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	EPB	*epb;		/* EPB pointer */
{
	REG	BPE	*bpe;		/* Outgoing BPE */
	REG	DEFAULT pri;		/* Current priority level */
	REG	BPE	*bp;		/* Local channel BPP tail pointer */

	/*
	** If this is a request for the VME147 new BPP environment, call
	**	the bp_pwrite routine instead.
	*/

	if (bppv2_local_cpu)
		if (c->rc->remote_cpu == *bppv2_local_cpu)
		{
			return((DEFAULT) (*bppv2_pwrite)((*bppv2_remote_pipe_tbl)[CPU_NO], epb));
		}

	/*
	**	First, determine if a pipe exists. If so, get a BPE
	**	from the BPE pool, fill it, and send it to the MVMe37x
	**	device.
	*/
	pri = splx( SPLEV( c->rc->intr_level ) );

	if( ( bp = c->lc.bpp_write ) != ( BPE * ) NULL )
	{
		if( ( bpe = ci_get_bpe( c ) ) != ( BPE * ) NULL )
		{
			/*
			**	Fill in the BPE and mark it as the
			**	terminator.
			*/

			bpe->flags |= 0x0001;

			/*
			**	Link in the new terminator, fill in
			**	current tail BPE, unmark old terminator
			**	envelope, and move the pipe tail pointer
			**	to the terminator.
			*/

			bp->next   = bpe;
			bp->msg    = epb;
			bp->dst_id = epb->dst_id;
			bp->src_id = epb->src_id;
			bp->flags &= 0xFFFE;

			c->lc.bpp_write = bpe;

			/*
			**	Enable interrupts.
			*/

			splx( pri );

			/*
			**	Tell the device.
			*/

			BPPCR( ( c->lc.remote_cpu - CPU_OFFSET ), BPP_BPE );

			return( NO_ERROR );
		}
		else
		{
			splx(pri);
			return( NO_BPE );
		}
	}
	else
	{
		splx(pri);
		return( NO_PIPE );
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_pread()	
**
**  DESCRIPTION:	Read an EPB from the incoming pipe.
**
**  CONTENTS:		This function will reade an EPB from the
**			local channel table's incoming pipe.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry.
**
**	Output:		NONE
**
**	Returns:	EPB pointer.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  04/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**  01/19/88	J. Holbrook		Added new BPP pipe interface for
**
**********************************************************************/

EPB *
ci_pread(c)
	REG	CPU_ENT	*c;		/* CPU table entry address */
{
	REG	BPE	*bpe;		/* BPE pointer */
	REG	EPB	*epb;		/* EPB pointer */
	REG	DEFAULT pri;		/* Current priority level */

	if (bppv2_local_cpu)
		if (c->rc->remote_cpu == *bppv2_local_cpu)
		{
			return((EPB *) (*bppv2_pread)((*bppv2_remote_pipe_tbl)[CPU_NO]));
		}

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	Check for anything in the pipe. If nothing just return.
	*/

	if( c->lc.bpp_read->flags & 0x0001 )
	{
		splx( pri );
		return( ( EPB * ) NULL );
	}

	/*
	**	Get the first BPE in the pipe and move the pipe head.
	*/

	bpe = c->lc.bpp_read;

	c->lc.bpp_read = bpe->next;

	/*
	**	Enable interrupts.
	*/

	splx( pri );

	/*
	**	Get the message from the envelope and return the
	**	BPE to the remote.
	*/

	epb = bpe->msg;

	ci_rel_bpe( c, bpe );

	return( epb );
}
