/*	@(#)ci_regdreg.c	35.1	*/
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci_regdreg.c
**
**  DESCRIPTION:	This module contains the Common Interface
**			registration/dregistration functions.
**
**  FUNCTIONS:		ci_reg,    ci_ret_rel, ci_ret_send,
**			ci_dst_ok, ci_is_reg,  ci_find_empty
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Coding.
**  05/01/87	E. Skinner	1.1	Port to System V/68 Release 3.
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x.
**  06/29/87	R. Robinson	1.2	Modified ci_reg().
**  09/10/87	dtk             2.1     Changed location of header files
**
***********************************************************************/

#include <sys/errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/signal.h>
#include <sys/dir.h>
#include <sys/immu.h>
#include <sys/region.h>
#include <sys/proc.h>
#include <sys/user.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

extern	EPB	*ci_uread();		/* Read user's queue */
extern	VOID	ci_uwrite();		/* Write user's queue */
extern	VOID	ci_add_retry();		/* Something is wrong */
extern	VOID	ci_swap_id();		/* Swap IDs */
extern	VOID	ci_rel_buf();		/* Return BUF */
extern	VOID	ci_rel_epb();		/* Return EPB */
extern	VOID	ci_ret_rel();		/* Return/release EPB/BUFs */
extern	DEFAULT ci_pwrite();		/* Write to remote */
extern	BOOLEAN	ci_find_empty();	/* Find empty table entry */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_reg()	
**
**  DESCRIPTION:	Perform the registration and dregistration
**			functions.
**
**  CONTENTS:		Based upon the passed EPB's type, this function
**			will either register or deregister with the
**			37x's Common Environment.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			epb		- EPB pointer,
**			pid		- process ID.
**
**	Output:		NONE
**
**	Returns:	An error is returned if the operation could
**			not be performed.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  04/23/87	Martin Dorr	1.1	Fixed superfluous CE reg response
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**  06/29/87	R. Robinson	1.2	Removed call to ci_rel_epb().
**
**********************************************************************/

DEFAULT
ci_reg( c, epb, pid )
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	CE_EPB	*epb;		/* CE EPB pointer */
	REG	USHORT	pid;		/* Process ID */
{
	extern	VOID	c37_rxint();	/* Receive handler */
	REG	DEFAULT	err;		/* Return error code */
		USHORT	ip;		/* Interface point */
	REG	DEFAULT	pri;		/* Current Priority level */
	REG	REG_ENT	*r;		/* Registration table entry address */

	/*
	**	Initialize return status.
	*/

	err = NO_ERROR;

	/*
	**	Check for an attached buffer.  If present, return an error.
	*/

	if( epb->hdr.buf != ( BUF * ) NULL )
		return( ENOTBLK );
	
	/*
	**	Process the desired operation.
	*/

	ip = GET_IP( epb->hdr.src_id );

	switch( epb->hdr.type )
	{
		case CE_DEREGISTRATION_REQUEST:
		{
			if( ip < NUM_REG_ENTRIES )
			{
				r = &( c->rt[ip] );

				if( r->process_id == pid )
				{
					pri = splx( SPLEV( c->rc->intr_level ) );

					r->process_id	= EMPTY;
					r->proc		= ( struct proc * )NULL;
					r->block	= 0;
					r->reg_notified	= FALSE;
					r->rx_int	= NULL;

					splx( pri );

					ci_ret_rel(c, ip);
				}
				else
					err = ECHILD;
			}
			else
				err = ERANGE;
		}
		break;

		case CE_REGISTRATION_REQUEST:
		{
			if( ip == 0 )
			{
				pri = splx( SPLEV( c->rc->intr_level ) );

				if( ci_find_empty( c, &ip ) == TRUE )
				{
					r		= &( c->rt[ip] );
					r->process_id   = pid;
					r->proc         = u.u_procp;
					r->block	= 0;
					r->reg_notified	= FALSE;
					r->rx_int	= c37_rxint;
					/* registration this way not permitted
					   for kernel level users */
				}
				else
					err = ENOSPC;

				splx( pri );
			}
			else if( ip < NUM_REG_ENTRIES )
			{
				pri = splx( SPLEV( c->rc->intr_level ) );

				r = &( c->rt[ip] );

				if( r->process_id == EMPTY )
				{
					r->process_id   = pid;
					r->proc         = u.u_procp;
					r->block	= 0;
					r->reg_notified	= FALSE;
					r->rx_int	= c37_rxint;
					/* registration this way not permitted
					   for kernel level users */
				}
				else
					err = EMLINK;

				splx( pri );
			}
			else	
				err = ERANGE;
		}
		break;
	}

	return( err );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_ret_rel()	
**
**  DESCRIPTION:	Empty out the read queue for the indicated IP.
**
**  CONTENTS:		This function will examine each EPB in the
**			queue and either return it to it's source or
**			if it is already being returned to it's source
**			(this IP is the EPB's original source) release
**			the EPB and BUF to the pools.
**
**			The EPBs must be "ci_uread" from the queue
**			without blocking (waiting for another EPB to
**			arrive).
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			ip		- Interface point.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_ret_rel(c, ip)

register CPU_ENT *c;				/* CPU table entry address */
register USHORT	 ip;				/* Interface point */
{
	extern	 VOID	ci_ret_send();		/* Return EPB to sender */

	register EPB	*epb;			/* EPB pointer */

	/*
	**	Return all EPBs.
	*/

	while ((epb = ci_uread(c, ip, 0)) != (EPB *)NULL)
	{
		if (epb->status < CE_STATUS)
			ci_ret_send(c, epb, UNDELIVERABLE);
		else
		{
			if (epb->buf != (BUF *)NULL)
				ci_rel_buf(c, epb->buf);

			ci_rel_epb(c, epb);
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_ret_send()	
**
**  DESCRIPTION:	Return EPB with status to sender of EPB.
**
**  CONTENTS:		This function will place the passed status into
**			the EPB and return the EPB to the originator
**			(i.e. based on source ID ) of the EPB.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			epb		- EPB pointer,
**			status		- return status.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_ret_send(c, epb, status)

register CPU_ENT *c;				/* CPU table entry address */
register EPB	 *epb;				/* EPB pointer */
register USHORT	 status;			/* EPB status */
{
	 extern	BOOLEAN	ci_dst_ok();		/* Check destination process */

	/*
	**	Swap source and destination IDs.
	*/

	ci_swap_id(epb);

	/*
	**	If CPU is local, then place the EPB on the
	**	user queue, else return it to the remote CPU.
	*/

	if (GET_CPU(epb->dst_id) == c->lc.local_cpu)
	{
		if (ci_dst_ok(c, GET_IP(epb->dst_id)) == TRUE)
		{
			epb->status = status;

			ci_uwrite(c, GET_IP(epb->dst_id), epb);
		}
		else
		{
			if (epb->buf != (BUF *)NULL)
				ci_rel_buf(c, epb->buf);

			ci_rel_epb(c, epb);
		}
	}
	else
	{
		if (epb->status < CE_STATUS)
		{
			epb->status = status;

			if (ci_pwrite(c, epb) != NO_ERROR)
				ci_add_retry(c, epb);
		}
		else
		{
			if (epb->buf != (BUF *)NULL)
				ci_rel_buf(c, epb->buf);

			ci_rel_epb(c, epb);
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_dst_ok()	
**
**  DESCRIPTION:	Determine if destination process is still
**			alive.
**
**  CONTENTS:		This function will determine if the destination
**			process can still receive EPBs. (i.e. is the
**			process still alive).
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			ip		- Interface point.
**
**	Output:		NONE
**
**	Returns:	TRUE		- process is still alive,
**			FALSE		- process is gone.
**
**  NOTES:		1) This code is also executed from interrupt level.
**			2) Kernel level users have negative process IDs.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  05/01/87	E. Skinner	1.1	Port to System V/68 Release 3.
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

BOOLEAN
ci_dst_ok( c, ip )
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	USHORT	ip;		/* Interface point */
{
	REG	DEFAULT	pri;		/* Current priority level */
	REG	REG_ENT	*r;		/* Registration table entry address */

	/*
	**	First, determine if the Interface Point and 
	**	corresponding process ID are valid.
	*/

	r = &( c->rt[ip] );

	if( ( ip <= NUM_REG_ENTRIES ) && ( r->process_id != EMPTY ) )
	{
		/*
		**	Kernel level users return now.
		*/

		if( ( ( short ) r->process_id ) < 0 )
			return( TRUE );

		/*
		**	Check the process ID in the registration
		**	table.
		*/

		if( r->process_id == r->proc->p_pid )
		{
			/*
			**	Check the process' status.
			*/

			switch( r->proc->p_stat )
			{
				case SSLEEP:		/* Awaiting an event */
				case SRUN:		/* Running */
				case SSTOP:		/* Traced */
				case SXBRK:		/* Xswapped */
					return( TRUE );
			}
		}

		/*
		**	Either process IDs dont' match or the state is funny.
		**	Clear process from the registration table.
		*/

		pri = splx( c->rc->intr_level );

		r->process_id	= EMPTY;
		r->proc		= ( struct proc * ) NULL;

		splx(pri);

		/*
		**	Release anything pending.
		*/

		ci_ret_rel( c, ip );
	}

	return( FALSE );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_is_reg()	
**
**  DESCRIPTION:	Determine if the process is registered.
**
**  CONTENTS:		This function will determine if the passed
**			process ID is in the Common Interface's
**			registration table.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			pid		- Process ID,
**			ip		- Interface point pointer.
**
**	Output:		NONE
**
**	Returns:	TRUE		- Process is registered,
**			FALSE		- Process is not registered.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

BOOLEAN
ci_is_reg(c, pid, ip)

register CPU_ENT *c;			/* CPU table entry address */
register USHORT	 pid;			/* Process ID */
register USHORT	 *ip;			/* Interface point pointer */
{
	register DEFAULT i;		/* Just a counter */

	for (i = 0; i < NUM_REG_ENTRIES; ++i)
		if (c->rt[i].process_id == pid)
		{
			*ip = i;

			return (TRUE);
		}

	return (FALSE);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_find_empty()	
**
**  DESCRIPTION:	Find an empty entry in the registration table.
**
**  CONTENTS:		This function will find an empty place in the
**			registration table.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table  entry,
**			ip		- Interface point.
**
**	Output:		NONE
**
**	Returns:	TRUE		- Registration table is not full,
**			FALSE		- Registration table is full.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:		Find_empty does not check element 0 of the
**			registration table so that it is not given to
**			someone who does not specifically request that
**			IP. This allows the bootloader to use that IP
**			with out forcing him to be the first process
**			to open the driver.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  06/09/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

BOOLEAN
ci_find_empty(c, ip)

register CPU_ENT *c;			/* CPU table entry address */
register USHORT	 *ip;			/* Interface point */
{
	register DEFAULT i;		/* Just a counter */

	for (i = 1; i < NUM_REG_ENTRIES; ++i)
		if (c->rt[i].process_id == EMPTY)
		{
			*ip = i;

			return (TRUE);
		}

	return (FALSE);
}
