/*	@(#)ci_utils.c	35.1	*/
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci_utils.c
**
**  DESCRIPTION:	This module contains the Common Interface
**			utility functions.
**
**  FUNCTIONS:		ci_swap_id, ci_add_retry
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Coding.
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x.
**  09/10/87	dtk             2.1     Changed location of header files
**
***********************************************************************/


#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>

extern VOID	ci_rel_buf();	/* Return BUF */
extern VOID	ci_rel_epb();	/* Return EPB */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_swap_id()	
**
**  DESCRIPTION:	Swap EPB destination and source IDs.
**
**  CONTENTS:		This function will swap the passed EPB's
**			destination and source ID fields.
**
**  PARAMETERS:		
**
**	Input:		epb		- EPB pointer.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_swap_id(epb)

register EPB	*epb;			/* EPB address */
{
	register ULONG 	tmp_id;		/* Temporary holder */

	tmp_id      = epb->src_id;
	epb->src_id = epb->dst_id;
	epb->dst_id = tmp_id;
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_add_retry()	
**
**  DESCRIPTION:	Release undelivered EPB/BUF.
**
**  CONTENTS:		This function will print an error message
**			and release the passed undelivered EPB and
**			BUF blocks.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			epb		- EPB pointer.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  05/15/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_add_retry(c, epb)

register CPU_ENT *c;			/* CPU table entry address */
register EPB	 *epb;			/* EPB pointer */
{
	printf("Releasing undelivered message - please fix me!!\n");

	if (epb->buf != (BUF *)NULL)
		ci_rel_buf(c, epb->buf);

	ci_rel_epb(c, epb);
}
