/*		Copyright (c) 1986,1987 Motorola		
 *		All Rights Reserved		
 *
 *		bppmgr.c - manager for buffered pipe protocol
 */

#ident	"@(#)bppmgr.c	35.2"

#include "sys/types.h"
#include "sys/bpp.h"

#define TOUT	2000 	/* length for timeout */
#define NULL	0

/*
 *	bppcrqueue - create envelope and command packet queue
 */

bppcrque(que,cnt,pp,ep,eye,am,busw)
register struct b_envelope **que;
register cnt;
register struct b_packet *pp;
register struct b_envelope *ep;
register unsigned eye;
register unsigned char am;
register unsigned char busw;
{
	register unsigned short i;

	/*
	 *	place packets in envelopes
	 */

	*que = NULL;
	for(i = 0; i < cnt; i++) {

		ep->e_link = *que;
		*que = ep;

		/* initialize packets */
		pp->p_eye = eye;
		pp->p_am = am;
		pp->p_busw = busw;
		pp->p_statp = 0xff; 	/* always return status on same pipe */

		(ep++)->e_pkt = pp++;
	}
}

/*
 *	bppcrchan - create channel
 */

bppcrchan(cp,rp,que,ilev,ivec,am)
register struct b_device *rp;
register struct b_channel *cp;
register struct b_envelope **que;
register unsigned char ilev;
register unsigned char ivec;
register unsigned char am;
{
	register tout;		/* timeout */
	register struct b_envelope *ep;
	struct b_envelope *bppget();

	/* set up channel */

	/* put NULL envelopes on channel pipes */

	if((ep = bppget(que)) == NULL)
		return(-1);
	ep->e_valid = INVALID;
	cp->c_cmd.p_head = cp->c_cmd.p_tail = ep;

	if((ep = bppget(que)) == NULL)
		return(-1);
	ep->e_valid = INVALID;
	cp->c_stat.p_head = cp->c_stat.p_tail = ep;

	/* fill in rest of fields in channel */

	cp->c_level = ilev;
	cp->c_vector = ivec;
	cp->c_am = am;
	cp->c_valid = INVALID;

	/*
	 *	gain exclusive access to device
	 */

	for( tout = TOUT; tout && (rp->m_cr&B_BSY); tout-- );
	if( tout <= 0 ) {
		return(-1);	/* error */
	}

	for( tout = TOUT; tout && (rp->m_tas&TASBIT); tout-- );
	if( tout <= 0 ) {
		return(-1);	/* error */
	}

	rp->m_tas |= TASBIT;	/* not safe like a real TAS */

	/*
	 *	set up register file
	 */

	rp->m_am = am;
#ifdef	m88k
	rp->m_addrhi = (unsigned)cp >> 16;
	rp->m_addrlo = (unsigned)cp;
#else
	rp->m_addr = (unsigned)cp;
#endif
	rp->m_tas = (rp->m_tas&TASMASK) | VCMD | CREATE;

	/*
	 *	tell controller that command is ready
	 */

#ifdef ANGEL
	dcacheflush();
#endif
	rp->m_cr |= B_ATTN;

	/*
	 *	wait for VSTAT bit
	 */

	for( tout=TOUT; tout && ((rp->m_tas&VSTAT) == 0); tout-- );

	if( tout <= 0 ) {
		return(-1);	/* error */
	}

	if( rp->m_sr ) {
		return(-1);	/* error */
	}

	/*
	 *	indicate command complete
	 */

	rp->m_tas |= CCOMP;

	/*
	 *	now tell controller
	 */

#ifdef ANGEL
	dcacheflush();
#endif
	rp->m_cr |= B_ATTN;

	return(cp->c_chan);		/* return channel number */
}

/*
 *	bppget - get free env from que
 */

struct b_envelope *bppget(que)
register struct b_envelope **que;
{
	register struct b_envelope *ep;

	if(*que == NULL)
		return(NULL);
	ep = *que;
	*que = ep->e_link;
	ep->e_link = NULL;
	return(ep);
}

/*
 *	bppfree - free envelope and put on queue
 */

bppfree(que,ep)
register struct b_envelope **que;
register struct b_envelope *ep;
{
	ep->e_link = *que;
	*que = ep;
}

/*
 *	bppsget -   safe get of free env from que
 */

struct b_envelope *bppsget(que)
register struct b_envelope **que;
{
	register struct b_envelope *ep;
	register s;

	s = splhi();
	if(*que == NULL)
		return(NULL);
	ep = *que;
	*que = ep->e_link;
	ep->e_link = NULL;
	splx(s);
	return(ep);
}

/*
 *	bppsfree - safe free envelope and put on queue
 */

bppsfree(que,ep)
register struct b_envelope **que;
register struct b_envelope *ep;
{
	register s;

	s = splhi();
	ep->e_link = *que;
	*que = ep;
	splx(s);
}

/*
 *	bppsend - send a packet to controller
 */

bppsend(cp,dp,ep)
register struct b_channel *cp;
register struct b_envelope *ep;
register struct b_device *dp;
{
	register struct b_packet *pp;
	register struct b_envelope *tp = cp->c_cmd.p_tail;

	/* switch packet from envelope and on null envelope */

	ep->e_valid = INVALID;
	pp = ep->e_pkt;
	pp->p_cmdp = cp->c_chan; 	/* channel number */
	ep->e_pkt = tp->e_pkt;
	tp->e_pkt = pp;

	/* put envelope on tail as null envelope */

	tp->e_link = ep;
	cp->c_cmd.p_tail = ep;
	tp->e_valid = VALID;

	/* tell controller */

#ifdef ANGEL
	dcacheflush();
#endif
	dp->m_cr |= B_ATTN;
}

/*
 *	bpprcv - receive an envelope from a channel
 */

struct b_envelope *bpprcv(cp)
register struct b_channel *cp;
{
	register struct b_envelope *ep = cp->c_stat.p_head;

	if( ep->e_valid == INVALID ) { /* spurious interrupt */
		return(NULL);
	}

	ep->e_valid = INVALID;

	/* set status head to next envelope */

	cp->c_stat.p_head = ep->e_link;
	return(ep);
}
