/*		Copyright (c) 1986,1987,1988 Motorola,Inc.	*/
/*		All Rights Reserved				*/
/*		@(#)dsk.c	35.4	*/
/*
 *	disk discipline routines
 */
#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/elog.h"
#include "sys/file.h"
#include "sys/buf.h"
#include "sys/iobuf.h"
#include "sys/systm.h"
#include "sys/dk.h"
#include "sys/dsk.h"
#include "sys/cmn_err.h"

#ifndef REL2
#include "sys/open.h"
#endif

#define DEBUG

#ifdef	DEBUG		/* debugging code */
#define STATIC
#define	DEBUGPR(x,y)	{ if (x & dskDebug) printf("%s\n", y); }

/* int dskDebug = 0x3ff; */
int dskDebug = 0;

#define	DBINIT		(1<<0)
#define	DBOPEN		(1<<1)
#define	DBCLOSE		(1<<2)
#define	DBSTRAT0	(1<<3)
#define	DB0DONE		(1<<4)
#define	DBRDWR		(1<<5)
#define	DBIOCTL		(1<<6)
#define	DBCLR		(1<<7)
#define	DB1DONE		(1<<8)
#define	DBSTRAT1	(1<<9)
#define	DBTIMER0	(1<<10)
#define	DBTIMER1	(1<<11)
#define	DBTIMER		(1<<12)

static char *ASSERTCTL = "ASSERT(%s) failed in %s, line %d\n";
#define ASSERT(X) { if (!(X)) printf(ASSERTCTL, "X", __FILE__, __LINE__);}

#else			/* no debugging code */

#define	STATIC	static
#define	DEBUGPR(x,y)
#define ASSERT(X)

#endif

#define	SLICE(d)	(d&((1<<dd->d_slcbits)-1))

#define	acts	io_s1		/* used for sorting I/O queue */
/*
 *	Globals
 */

struct dsk *ddq = NULL;		/* queue of dsk structures */

STATIC	char	motstring[10] = "MOTOROLA";
STATIC	char	exorstring[10] = "EXORMACS";

STATIC dsktst(vid_mac)
register char *vid_mac;
{	register char *mot = motstring;
	register char *vd = vid_mac;

	if( *mot++ == *vd++ )
		if( *mot == *vd )
			return(1);
	
	mot = exorstring;
	if( *mot++ != *vid_mac++ )
		return(0);

	return( *mot == *vid_mac );
}

/*
 *	dskinit: disk discipline initialize routine
 */

dskinit(dd)
register struct dsk *dd;
{
	int dsktimer();

	DEBUGPR(DBINIT,"dskinit");

	if (ddq == NULL)	/* start timer going on first call to dskinit */
		timeout(dsktimer, 0, DSKTIMEOUT*HZ);
	dd->d_link = ddq;	/* add dsk struct to the queue */
	ddq = dd;
}

/*
 *	dskopen: disk discipline open routine
 */
#ifdef REL2
dskopen(dp, modes, omask)
#else
dskopen(dp, modes, otbl, otyp, omask)
register *otbl;
register otyp;
#endif
register struct iobuf *dp;
register modes;
int omask;
{
	register struct dsk *dd;
	register struct buf *bp;
	register struct dkblk0 *blk0;
	register struct volumeid *vid;
	register struct dkconfig *dkcfg;
	register *bptr;
	struct buf *geteblk();
	register slcblk, slccnt;
	int badcnt;

	DEBUGPR(DBOPEN,"dskopen");

	while (dp->b_flags&B_OPEN)
		/* drive is being initialized */
		sleep(dp, PRIBIO);

	if (dp->b_flags&B_ONCE) {
		if (dp->b_flags&B_WCLOSE && dp->b_flags&B_REINIT) {
			register s;

			s = spl6();	
			while (dp->b_flags&B_WCLOSE)
				sleep(dp, PRIBIO);
			splx(s);
#ifdef	SWBADTRK
			if (dd->d_flags&D_SWBAD) { /* if S/W badtrack support */
				baddel(dp->io_badptr);
				dp->io_badptr = 0;
			}
#endif
		}
		else {
#ifdef REL2
			/* the drive is already initialized */
			dp->io_sopen |= omask;	/* save open slice */
#else
			/*
			 * the drive is already initialized
			 * mark the logical device open
			 */

			if(otyp == OTYP_LYR)
				++otbl[OTYP_LYR];
			else if( otyp < OTYPCNT)
				otbl[otyp] |= omask;
#endif
			return;
		}
	}
	dp->b_flags |= B_OPEN;			/* in progress */
	dd = (struct dsk *) dp->io_dskptr;
	dp->io_slccnt = 1<<dd->d_slcbits;	/* default slice count */
	bp = geteblk();
	bp->b_proc = u.u_procp;
	bp->b_dev = dp->b_dev;
	bp->b_blkno = 0;

	/* do default configuration, if controller handles configs */
	if (dd->d_flags&(D_CONFIG|D_DFL_CONF) ) {
		bp->b_error = 0;
		bp->b_flags |= B_DFL_CONF|B_CONF|B_ABS;
		(*dd->d_strategy)(bp);
		iowait(bp);
		bp->b_flags &= ~(B_DFL_CONF|B_CONF|B_DONE|B_ABS);
		if (dd->d_flags&(D_DFL_ERR) && u.u_error) 
			goto done;
		u.u_error = 0;			/* don't worry about errors */
	}

	/* read volume ID & config block (first 512 bytes of disk) */
	/* only if not previously read with first config 	  */
	if ((dd->d_flags&(D_NOVID)) == 0 ) {
		bp->b_error = 0;
		bp->b_flags |= B_READ|B_ABS;
		(*dd->d_strategy)(bp);
		iowait(bp);
		bp->b_flags &= ~(B_READ|B_DONE|B_ABS);

		if (u.u_error)			/* if read error */
			goto done;
	}

	blk0 = (struct dkblk0 *)bp->b_un.b_addr;
	vid = &(blk0->dk_vid);

	if (!dsktst(vid->vid_mac)) {	/* foreign disks? */
		if (dd->d_flags&D_MOTTST) 
			u.u_error = 1;
		goto done;		/* don't do reconfig	*/
	}

	/* save some fields for the bad block & slice table handling */
	badcnt = vid->vid_btl;
	dkcfg =  &(blk0->dk_ios);
	slcblk = dkcfg->ios_slcblk;
	slccnt = dkcfg->ios_slccnt;

	/* if controller handles configs, re-configure drive */
	if (dd->d_flags&D_CONFIG) {
		bp->b_error = 0;
		bp->b_flags |= B_CONF|B_ABS;		/* using config blk */
		(*dd->d_strategy)(bp);
		iowait(bp);
		bp->b_flags &= ~(B_CONF|B_DONE|B_ABS);
		if (u.u_error)
			goto done;
	}

#ifdef	SWBADTRK
	/* if S/W bad track support is needed */
	if (dd->d_flags&D_SWBAD) {
		bp->b_blkno = BADBLK;
		bp->b_error = 0;
		bp->b_flags |= B_READ|B_ABS;
		(*dd->d_strategy)(bp);
		iowait(bp);
		bp->b_flags &= ~(B_READ|B_DONE|B_ABS);

		if (u.u_error)
			goto done;
 
		bptr = &dp->io_badptr;
		*bptr = 0;
		if ((*bptr = badinit(bp->b_un.b_addr, badcnt)) == 0)
			u.u_error = ENXIO;
	}
#endif

	/* set up the slice table */
	if (slcblk) {
		/* if slice table > 1024 bytes its not good */
		if (slccnt*sizeof(struct slice) > 1024) {
			u.u_error = ENXIO;
			goto done;
		}
		bp->b_blkno = slcblk;
		bp->b_error = 0;
		bp->b_flags |= B_READ|B_ABS;
		(*dd->d_strategy)(bp);
		iowait(bp);
		bp->b_flags &= ~(B_READ|B_DONE|B_ABS);

		if (u.u_error)
			goto done;

		if (slccnt) {
			bzero(dp->io_slctbl, (1<<dd->d_slcbits) *
				sizeof(struct slice));
			bcopy((bp->b_un.b_addr),dp->io_slctbl,
				slccnt*sizeof(struct slice));
		}
	}

done:	if (u.u_error == 0 || modes&FNDELAY) {
		dp->b_flags |= B_ONCE;		/* set init done */
#ifdef REL2
		dp->io_sopen |= omask;		/* save open slice */
#else
		if(otyp == OTYP_LYR)
			++otbl[OTYP_LYR];
		else if( otyp < OTYPCNT)
			otbl[otyp] |= omask;
#endif
	}
	if (modes&FNDELAY)
		u.u_error = 0;
	brelse(bp);
	dp->b_flags &= ~B_OPEN;			/* open done */
	wakeup(dp);
}

/*
 *	dskclose: disk discipline close routine
 */

#ifdef REL2
dskclose(dp, modes, mask)
#else
dskclose(dp, modes, otbl, otyp, mask)
register *otbl;
register otyp;
#endif
register mask;
register struct iobuf *dp;
int modes;
{
	register struct dsk *dd;

	DEBUGPR(DBCLOSE,"dskclose");
 
#ifdef REL2
	if (((dp->io_sopen &= ~mask) == 0) && ((dp->b_flags&B_SPECL) == 0)) {
		if (dp->b_active == 0) {
			if (dp->b_flags&B_REINIT)
				dp->b_flags &= ~B_ONCE;	/* init on re-open */
#else
	/* mark logical device closed */
	if(otyp == OTYP_LYR)
		--otbl[OTYP_LYR];
	else if (otyp < OTYPCNT)
		otbl[otyp] &= ~mask;
	if ((otbl[otyp] == 0) && ((dp->b_flags&B_SPECL) == 0)) {
		if (dp->b_active == 0) {
			if (dp->b_flags&B_REINIT)
				dp->b_flags &= ~B_ONCE;	/* init on re-open */
#endif
#ifdef	SWBADTRK

			/* close phys dev when all logical devs are closed */
#ifndef REL2
			for( i = 0; i < OTYPCNT && otbl[i] == 0; ++i);

			if( i == OTYPCNT){
#endif
				dd = (struct dsk *) dp->io_dskptr;
				if (dd->d_flags&D_SWBAD) { /* if S/W badtrack */
					baddel(dp->io_badptr);
					dp->io_badptr = 0;
#ifndef REL2
				}
#endif
			}
#endif
		}
		else				/* closed but still active */
			dp->b_flags |= B_WCLOSE;	/* set close wanted */
	}
}

/*
 *	dsk0strategy: disk discipline strategy routine for controllers
 *		      that do the strategy function.
 */

dsk0strategy(bp, dp)
register struct buf *bp;
register struct iobuf *dp;
{
	register struct dsk *dd;
	register dev;
	register last;
	register s;
	register struct iotime *ip;

	DEBUGPR(DBSTRAT0,"dskstrat0");
 

	dev = bp->b_dev;
	dd = (struct dsk *) dp->io_dskptr;

	if (dp->b_flags&B_DEVDEAD) {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}	
	if (bp->b_bcount&(BSIZE-1)) {		/* no partial block I/O */
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}

	/*
	 * If B_ABS flag is on, the block # is absolute.
	 * There is no need to check slice limits.
	 */
	if ((bp->b_flags&B_ABS) == 0) {

		register struct slice *sp;

		sp = (struct slice *) dp->io_slctbl + SLICE(dev);
		last = sp->nblocks;

		/* request within limits of slice? */
		if (bp->b_blkno < 0 || bp->b_blkno >= last) {
			if (bp->b_blkno == last && bp->b_flags&B_READ)
				bp->b_resid = bp->b_bcount;
			else {
				bp->b_flags |= B_ERROR;
				bp->b_error = ENXIO;
			}
			iodone(bp);
			return;
		}
	}

	/* do some I/O statistics accounting */
	bp->b_start = lbolt;
	dp->io_qcnt++;
	ip = (struct iotime *)dp->io_stp;
	if (bp->b_flags&B_READ || bp->b_flags&B_WRITE) {
		ip->io_cnt++;
		ip->io_bcnt += btoc(bp->b_bcount);
	}
	else
		ip->io_other++;

	/* queue the I/O request on the end of the iobuf queue */

	s = spl6();
	bp->av_back = dp->b_actl;
	dp->b_actl->av_forw = bp;
	bp->av_forw = NULL;
	dp->b_actl = bp;
	splx(s);

	/* Always call start routine, let it decide whether to start or not */
	dp->b_active++;
	(*dd->d_start)(dp);
}

/*
 *	dsk0done: disk discipline "I/O done" routine for controllers
 *		  that do the strategy function.
 */

dsk0done(dp)
register struct iobuf *dp;
{
	register struct buf *bp;
	register struct dsk *dd;
	register struct buf *newbp;
	int s;

	DEBUGPR(DB0DONE,"dsk0done");
 

	/*
	 * dequeue all finished buffers, finished buffers are ones
	 * that have the B_STSRTN (status packet returned) flag set
	 */
	s = spl6();
	for (bp = dp->b_actf; bp != NULL; bp = bp->av_forw) {
		if (bp->b_flags&B_STSRTN) {	/* if status packet returned */
			dsk0deque(bp,dp);

			/* log any errors */
			logberr(dp,bp->b_flags&B_ERROR);

			/* Point to the previous element in the iobuf queue. */
			newbp = bp->av_back;
			bp->b_flags &= ~B_STSRTN;	/* clean up flags */
			iodone(bp);			/* return buffer */
			bp = newbp;
		}
		/*
		 * Finish search when buf pointed to by b_acts is hit.
		 */
		if (bp == (struct buf *)dp->b_acts)
			break;
	}
	splx(s);

	dd = (struct dsk *) dp->io_dskptr;
	/* if any unprocessed bufs queued, call drvstart() */
	if (((struct buf *)dp->b_acts)->av_forw)
		(*dd->d_start)(dp);
	else {
		dp->b_active = 0;
		if (dp->b_flags&B_WCLOSE) {
			dp->b_flags &= ~B_WCLOSE;	/* finish close */
			if (dp->b_flags&B_REINIT) {
				dp->b_flags &= ~B_ONCE;	/* re-init on re-open */
				wakeup(dp);
			}
#ifdef	SWBADTRK
			if (dd->d_flags&D_SWBAD) { /* if S/W badtrack support */
				baddel(dp->io_badptr);
				dp->io_badptr = 0;
			}
#endif
		}
	}
}

/*
 *	dsk1strategy: disk discipline strategy routine for controllers
 *		      that don't do the strategy function.
 */

dsk1strategy(bp, dp)
register struct buf *bp;
register struct iobuf *dp;
{
	register struct dsk *dd;
	register dev;
	register last;
	register s;
	register struct iotime *ip;

	DEBUGPR(DBSTRAT1,"dskstrat1");
 

	dev = bp->b_dev;
	dd = (struct dsk *) dp->io_dskptr;

	if (dp->b_flags&B_DEVDEAD) {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}	
	if (bp->b_bcount&(BSIZE-1)) {		/* no partial block I/O */
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}

	/*
	 * If B_ABS flag is on, the block # is absolute.
	 * There is no need to check slice limits.
	 */
	if ((bp->b_flags&B_ABS) == 0) {

		register struct slice *sp;

		sp = (struct slice *) dp->io_slctbl + SLICE(dev);
		last = sp->nblocks;

		/* request within limits of slice? */
		if (bp->b_blkno < 0 || bp->b_blkno >= last) {
			if (bp->b_blkno == last && bp->b_flags&B_READ)
				bp->b_resid = bp->b_bcount;
			else {
				bp->b_flags |= B_ERROR;
				bp->b_error = ENXIO;
			}
			iodone(bp);
			return;
		}
	}

	/* do some I/O statistics accounting */
	bp->b_start = lbolt;
	bp->av_forw = NULL;
	dp->io_qcnt++;
	ip = (struct iotime *)dp->io_stp;
	if (bp->b_flags&B_READ || bp->b_flags&B_WRITE) {
		ip->io_cnt++;
		ip->io_bcnt += btoc(bp->b_bcount);
	}
	else
		ip->io_other++;


	/* put request in buffer queue */
	s = spl6();		/* spl() high enough to block timers */
	if( dp->b_actf == NULL ) {
		dp->b_actf = bp;
		dp->b_actl = bp;
		dp->acts = (int)bp;
	} else {
		register struct buf *ap;
		register struct buf *cp;

		if(((ip->io_cnt)&(STEPSIZE-1)) == 0 )
			dp->acts = (int)dp->b_actl;
		for( ap=(struct buf *)dp->acts; cp=ap->av_forw; ap=cp ) {
			register s1,s2;

			if ((s1 = ap->b_resid - bp->b_resid)<0) s1 = -s1;
			if ((s2 = ap->b_resid - cp->b_resid)<0) s2 = -s2;
			if (s1 < s2)
				break;
		}
		ap->av_forw = bp;
		if ((bp->av_forw = cp) == NULL)
			dp->b_actl = bp;
	}

	splx(s);
	/* Always call start routine, let it decide whether to start or not */
	dp->b_active++;
	(*dd->d_start)(dp);
}
/*
 *	dsk1done: disk discipline "I/O done" routine for controllers
 *		  that don't do the strategy function.
 */

dsk1done(dp)
register struct iobuf *dp;
{
	register struct buf *bp;
	register struct dsk *dd;

	DEBUGPR(DB0DONE,"dsk1done");
 

	bp = dp->b_actf;
	dsk1deque(bp,dp); 		/* dequeue finished buffer */
	logberr(dp,bp->b_flags&B_ERROR); /* log any errors */
	iodone(bp);			/* return buffer */

	dd = (struct dsk *) dp->io_dskptr;

	/* if any unprocessed bufs queued, call drvstart() */
	if ((struct buf *)dp->b_actf)
		(*dd->d_start)(dp);
	else {
		dp->b_active = 0;
		if (dp->b_flags&B_WCLOSE) {
			dp->b_flags &= ~B_WCLOSE;	/* finish close */
			if (dp->b_flags&B_REINIT) {
				dp->b_flags &= ~B_ONCE;	/* re-init on re-open */
				wakeup(dp);
			}
#ifdef	SWBADTRK
			if (dd->d_flags&D_SWBAD) { /* if S/W badtrack support */
				baddel(dp->io_badptr);
				dp->io_badptr = 0;
			}
#endif
		}
	}
}

#ifdef REL2

/*
 *	dskrdwr: disk discipline read/write routine
 */

dskrdwr(dp, dev, mode)
register struct iobuf *dp;
register dev;
register mode;
{
	register struct dsk *dd;
	register struct slice *slc;

	DEBUGPR(DBRDWR,"dskrdwr");
 
	dd = (struct dsk *) dp->io_dskptr;

	slc = (struct slice *) dp->io_slctbl + SLICE(dev);
	if (physck(slc->nblocks,mode))
		physio(dd->d_strategy, 0, dev, mode);
}
#endif

/*
 *	dskioctl: disk discipline "ioctl" routine
 */

dskioctl(dp, cmd, args, mode, bp)
register struct iobuf *dp;
register cmd;
register char *args;
register mode;
register struct buf *bp;
{

	DEBUGPR(DBIOCTL,"dskioctl");

	switch(cmd) {

	case DKGETSLC:	if (copyout(dp->io_slctbl, args,
			    dp->io_slccnt * sizeof(struct slice)) == -1) {
				u.u_error = EFAULT;
				return;
			}
			return(1);

	case DKSETSLC:	if (copyin(args,bp->b_un.b_addr,
			    dp->io_slccnt * sizeof(struct slice)) == -1) {
				u.u_error = EFAULT;
				return;
			}
			bcopy(bp->b_un.b_addr,dp->io_slctbl,
				dp->io_slccnt * sizeof(struct slice));
			return(1);
	default:
			/* return false - request not satisfied */
			return(0);
	};
}

/*
 *	dskclr: clear the driver after powerfail
 */

dskclr(dp,count)
register struct iobuf *dp;
register count;
{
	register struct dsk *dd;
	register pos;

	DEBUGPR(DBCLR,"dskclr");

	/* check all of the iobufs of each driver for outstanding I/Os */
	for (pos=0; pos < count; pos++, dp++) {
		dp->b_forw = NULL;		/* dequeue */
		dp->b_active = 0;
		if (dp->b_actf) {		/* buffers enqueued? */
			dp->b_active++;
			dd = (struct dsk *) dp->io_dskptr;
			(*dd->d_start)(dp);
		}
	}
}

/*
 *	dsktimer: check on the state of each driver periodically
 */

dsktimer()
{
	register struct dsk *dd;

	DEBUGPR(DBTIMER,"dsktimer");


	for (dd = ddq; dd != NULL; dd = dd->d_link)
		(*dd->d_timer)();	/* call each driver timer routine */

	timeout(dsktimer, 0, DSKTIMEOUT*HZ);
}

/*
 *	dsk0timer: check on status of a driver periodically for controllers
 *		   that do the strategy function.
 */

dsk0timer(dp,count)
register struct iobuf *dp;
register count;
{
	register struct buf *bp;
	register struct dsk *dd;
	register pos;
	register age;
	register oldage;

	DEBUGPR(DBTIMER0,"dsktimer0");

	for (pos=0; pos < count; pos++, dp++) {
		/* if queue is empty, look at next dp */
		if ((bp = dp->b_actf) == NULL)
			continue;
		dd = (struct dsk *)dp->io_dskptr;

		/*
		 *	Only age buffers that have been sent to the IPC.
		 */
		for (bp = dp->b_actf; bp != ((struct buf *)dp->b_acts)->av_forw;
			bp = bp->av_forw) {

			/*
			 * If status packet has already been returned by IPC
			 * the timing of the associated buf can be stopped.
			 */
			if (bp->b_flags&B_STSRTN)
				continue;

			oldage = (bp->b_flags&B_LONGCMD) ? dd->d_longtime :
					dd->d_shorttime;
			age = lbolt - bp->b_start;
			if (age > oldage) {
#ifdef	DEBUG
				printf("Timeout on buf %x\n", bp);
#endif
				dd->d_print(dp->b_dev,"Timeout");
				dp->b_flags |= B_DEVDEAD;
				break;
			}
		}
		if (dp->b_flags&B_DEVDEAD) {
			for (bp = dp->b_actf; bp != NULL; bp = dp->b_actf) {
				dsk0deque(bp,dp);
				logberr(dp, B_ERROR);
				bp->b_flags |= B_ERROR;
				iodone(bp);
			}
			dp->b_active = 0;
		}

		/* if a buffer is waiting & driver is inactive, restart it */
		if ((((struct buf *)(dp->b_acts))->av_forw != NULL)
		    && (dp->b_active == 0)) {
			dp->b_active++;
			dd->d_start(dp);
		}
	}
}

/*
 *	dsk1timer: check on status of a driver periodically for controllers
 *		   that don't do the strategy function.
 */

dsk1timer(dp,count,dskd)
register struct iobuf *dp;
register count;
register struct dsk *dskd;
{
	register struct buf *bp;
	register struct dsk *dd;
	register struct pktcnt *pcp;
	register pos;
	register age;
	register oldage;
	register unsigned short i;

	DEBUGPR(DBTIMER1,"dsktimer1");

	for (pos=0; pos < count; pos++, dp++) {
		/* if queue is empty, look at next dp */
		if ((bp = dp->b_actf) == NULL)
			continue;
		dd = (struct dsk *)dp->io_dskptr;
		if (dd != dskd) /* check to see if this is indeed a disk */
			continue;
		pcp = (struct pktcnt *) &dp->io_pktcnt;	

		/*
		 *	Only age buffers that have been sent to the IPC.
		 */

		for (bp = dp->b_actf,i=0;i < pcp->cnt ;i++,bp=bp->av_forw) {

			if(bp == NULL)
				break;
			oldage = (bp->b_flags&B_LONGCMD) ? dd->d_longtime :
					dd->d_shorttime;
			age = lbolt - bp->b_start;
			if (age > oldage) {
#ifdef	DEBUG
				printf("Timeout on buf %x\n", bp);
#endif
				dd->d_print(dp->b_dev,"Timeout");
				dp->b_flags |= B_DEVDEAD;
				break;
			}
		}
		if (dp->b_flags&B_DEVDEAD) {
			for (bp = dp->b_actf; bp ; bp = dp->b_actf) {
				dsk1deque(bp,dp);
				logberr(dp, B_ERROR);
				bp->b_flags |= B_ERROR;
				iodone(bp);
			}
			dp->b_active = 0;
		}

		/* if a buffer is waiting & driver is inactive, restart it */
		if (((struct buf *)(dp->b_actf)) && (dp->b_active == 0)) {
			dp->b_active++;
			dd->d_start(dp);
		}
	}
}

/*
 *	dsk0deque: dequeue a buffer from the iobuf queue
 *		   Used by routines that use the "dsk0" routines.
 */

STATIC dsk0deque(bp,dp)
register struct buf *bp;
register struct iobuf *dp;
{
	register s;

	s = spl6();
	bp->b_pkt = (int)NULL;
	bp->av_back->av_forw = bp->av_forw;
	if (bp->av_forw != NULL)		/* if not last on queue */
		bp->av_forw->av_back = bp->av_back;
	else					/* if last on q */
		dp->b_actl = bp->av_back;
	if (dp->b_acts == (int)bp)		/* if last processed */
		dp->b_acts = (int)bp->av_back;	/* fix up last processed ptr */
	dp->io_qcnt--;
	splx(s);
}

/*
 *	dsk1deque: dequeue a buffer from the iobuf queue
 *		   Used by routines that use the "dsk1" routines.
 */

STATIC dsk1deque(bp,dp)
register struct buf *bp;
register struct iobuf *dp;
{
	register s;

	s = spl6();
	dp->b_actf = bp->av_forw;
	dp->io_qcnt--;
	if (bp == (struct buf *)dp->acts)
		dp->acts = (int)dp->b_actf;
	bp->b_pkt = (int)NULL;
	bp->b_resid = 0;
	splx(s);
}

/*
 *	dsk2strategy: disk discipline strategy routine for controllers
 *		      that don't do the strategy function.
 *		      coalescing handler
 */

dsk2strategy(bp, dp, coalesce,maxcoal)
register struct buf *bp;
register struct iobuf *dp;
register int coalesce;
int maxcoal;
{
	register dev;
	register last;
	register s;
	struct iotime *ip;
	struct slice *sp;
	struct dsk *dd;

	DEBUGPR(DBSTRAT1,"dskstrat2");
 

	dev = bp->b_dev;
	dd = (struct dsk *) dp->io_dskptr;
	sp = (struct slice *) dp->io_slctbl + SLICE(dev);

	if (dp->b_flags&B_DEVDEAD) {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}	
	if (bp->b_bcount&(BSIZE-1)) {		/* no partial block I/O */
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}

	/*
	 * If B_ABS flag is on, the block # is absolute.
	 * There is no need to check slice limits.
	 */
	if ((bp->b_flags&B_ABS) == 0) {

		last = sp->nblocks;

		/* request within limits of slice? */
		if (bp->b_blkno < 0 || bp->b_blkno >= last) {
			if (bp->b_blkno == last && bp->b_flags&B_READ)
				bp->b_resid = bp->b_bcount;
			else {
				bp->b_flags |= B_ERROR;
				bp->b_error = ENXIO;
			}
			iodone(bp);
			return;
		}
		bp->coal_soff = sp->blkoff;
	}
	else
		bp->coal_soff = 0;

	/* do some I/O statistics accounting */
	bp->b_start = lbolt;
	bp->av_forw = NULL;
	bp->av_back = NULL;

	ip = (struct iotime *)dp->io_stp;
	if (bp->b_flags&B_READ || bp->b_flags&B_WRITE) {
		ip->io_cnt++;
		ip->io_bcnt += btoc(bp->b_bcount);
	}
	else
		ip->io_other++;

	/* put request in buffer queue */
	s = spl6();		/* spl() high enough to block timers */
	if( dp->b_actf == NULL ) {
		dp->b_actf = bp;
		dp->b_actl = bp;
		dp->acts = (int)bp;
	} else {
		register struct buf *ap;
		register struct buf *cp;

		if(((ip->io_cnt)&(STEPSIZE-1)) == 0 )
			dp->acts = (int)dp->b_actl;
		for( ap=(struct buf *)dp->acts; cp=ap->av_forw; ap=cp ) {
			register s1,s2;

			if ((s1 = ap->b_resid - bp->b_resid)<0) s1 = -s1;
			if ((s2 = ap->b_resid - cp->b_resid)<0) s2 = -s2;
			if (s1 < s2)
				break;
		}

		if((coalesce) && (ap != dp->b_actf) 
		&& (trydskcoales(ap, bp, maxcoal-1,
		 ap->b_bcount/BSIZE ) == 0))
		
			;		/* bp was linked for coalescing */
		else{
					/* bp was not linked for coalescing */
			ap->av_forw = bp;
			if ((bp->av_forw = cp) == NULL)
				dp->b_actl = bp;
		}
	}
	dp->io_qcnt++;

	splx(s);
	/* Always call start routine, let it decide whether to start or not */
	dp->b_active++;
	(*dd->d_start)(dp);
}

/*
 *	dsk2deque: dequeue a buffer from the iobuf queue
 *		   Used by routines that use the "dsk2" routines.
 */

STATIC dsk2deque(bp,dp)
register struct buf *bp;
register struct iobuf *dp;
{
	register s;

	s = spl6();
	dp->b_actf = bp->av_forw;
	dp->io_qcnt--;
	if (bp == (struct buf *)dp->acts)
		dp->acts = (int)dp->b_actf;
	bp->b_resid = 0;
	splx(s);
}

/*
 *	dsk2done: disk discipline "I/O done" routine for controllers
 *		  that don't do the strategy function.
 */

dsk2done(dp)
register struct iobuf *dp;
{
	register struct buf *bp;
	register struct dsk *dd;
	register struct buf *coalbp, *oldbp;

	DEBUGPR(DB0DONE,"dsk2done");
 

	bp = dp->b_actf;

	logberr(dp,bp->b_flags&B_ERROR); /* log any errors */

	dsk2deque(bp,dp); 		/* dequeue finished buffer */

	coalbp = bp->av_back;
	while(coalbp){ 
		coalbp->b_resid = bp->b_resid;
		coalbp->b_flags |= bp->b_flags & B_ERROR;
		oldbp = coalbp;
		coalbp = coalbp->av_back;
		iodone(oldbp);
	}

	iodone(bp);			/* return buffer */

	dd = (struct dsk *) dp->io_dskptr;

	/* if any unprocessed bufs queued, call drvstart() */
	if ((struct buf *)dp->b_actf)
		(*dd->d_start)(dp);
	else {
		dp->b_active = 0;
		if (dp->b_flags&B_WCLOSE) {
			dp->b_flags &= ~B_WCLOSE;	/* finish close */
			if (dp->b_flags&B_REINIT) {
				dp->b_flags &= ~B_ONCE;	/* re-init on re-open */
				wakeup(dp);
			}
#ifdef	SWBADTRK
			if (dd->d_flags&D_SWBAD) { /* if S/W badtrack support */
				baddel(dp->io_badptr);
				dp->io_badptr = 0;
			}
#endif
		}
	}
}

/*
 *	dsk2timer: check on status of a driver periodically for controllers
 *		   that don't do the strategy function.
 */

dsk2timer(dp,count,dskd)
register struct iobuf *dp;
register count;
register struct dsk *dskd;
{
	register struct buf *bp;
	register struct dsk *dd;
	register struct pktcnt *pcp;
	int i;	
	struct buf *coalbp;
	struct buf *oldbp;
	register pos;
	register age;
	register oldage;

	DEBUGPR(DBTIMER1,"dsktimer2");

	for (pos=0; pos < count; pos++, dp++) {
		/* if queue is empty, look at next dp */
		if ((bp = dp->b_actf) == NULL)
			continue;
		dd = (struct dsk *)dp->io_dskptr;
		if (dd != dskd) /* check to see if this is indeed a disk */
			continue;
		/*
		 *	Only age buffers that have been sent to the IPC.
		 */

		pcp = (struct pktcnt *) &dp->io_pktcnt;	


		/* 
			the scatter/gather buffers have the same age 
			as the head of their buffer list. 
		*/
		for (bp = dp->b_actf,i=0;i < pcp->cnt ;i++,bp=bp->av_forw) {

			if(bp == NULL)
				break;
			oldage = (bp->b_flags&B_LONGCMD) ? dd->d_longtime :
				dd->d_shorttime;
			age = lbolt - bp->b_start;
			if (age > oldage) {
#ifdef	DEBUG
				printf("Timeout on buf %x\n", bp);
#endif
				dd->d_print(dp->b_dev,"Timeout");
				dp->b_flags |= B_DEVDEAD;
				break;
			}
		}

		if (dp->b_flags&B_DEVDEAD) {
			for (bp = dp->b_actf; bp ; bp = dp->b_actf) {
				coalbp = bp->av_back;
				while(coalbp){ 
					coalbp->b_resid = bp->b_resid;
					coalbp->b_flags |= bp->b_flags&B_ERROR;
					logberr(dp, B_ERROR);
					oldbp = coalbp;
					coalbp = coalbp->av_back;
					iodone(oldbp);
				}
				dsk2deque(bp,dp);
				logberr(dp, B_ERROR);
				bp->b_flags |= B_ERROR;
				iodone(bp);
			}
			dp->b_active = 0;
		}

		/* if a buffer is waiting & driver is inactive, restart it */
		if (((struct buf *)(dp->b_actf)) && (dp->b_active == 0)) {
			dp->b_active++;
			dd->d_start(dp);
		}
	}
}

/* 	trydskcoales: try to combine individual block I/O requests. 
*	Call under interrupt protection.
*	Caller must assure that "cbp" is NOT the head of the I/O request queue
*    	which may be an I/O request already in progress.
*
*	Requests to be combined are linked together using the "av_back"
*	pointer. 
*
*/
trydskcoales(cbp, newbp, maxentries, pblocks)
register struct buf *cbp;	/* first bp in list */ 
register struct buf *newbp;	/* try to link this new bp to end of list  */
register int	maxentries;	/* maximum number of bps permitted in list */
register short	pblocks;	/* physical blocks per file system block */
{
  register struct buf *bp;	/* last bp in list */
  register int  entries; 

  if(((cbp->b_flags & B_PHYS) == 0)
  &&((newbp->b_flags & B_PHYS) == 0)
  &&((cbp->b_flags & B_READ) == (newbp->b_flags & B_READ))){

	bp = cbp;
	for( entries=0 ; bp->av_back ; (bp = bp->av_back), entries++)
		;

	if(( entries < maxentries)
	&&(newbp->coal_soff == bp->coal_soff)
	&&(newbp->b_blkno == bp->b_blkno + pblocks)){

		bp->av_back = newbp;

		return(0);	/* Successful */
	}
   }
	return(-1);	/* Unsuccessful */
}
