/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)kern-port:io/gentty.c	35.5"
/*
 *	indirect driver for controlling tty.
 */
#include "sys/types.h"
#include "sys/param.h"
#include "sys/immu.h"
#include "sys/sysmacros.h"
#include "sys/fs/s5dir.h"
#include "sys/errno.h"
#include "sys/signal.h"
#include "sys/region.h"
#include "sys/proc.h"
#include "sys/psw.h"
#include "sys/user.h"
#include "sys/conf.h"
#include "sys/tty.h"
#include "sys/stream.h"
#include "sys/mvmecpu.h"
#include "sys/open.h"
#include "sys/gentty.h"
#include "sys/m68681.h"

syopen(dev, flag)
{

	if (sycheck()) {
		if (cdevsw[major(u.u_ttyd)].d_str) {
			if (!u.u_ttyip) {
				u.u_error = ENXIO;
				return;
			}
			stropen(u.u_ttyip, flag);
		} else
		   (*cdevsw[major(u.u_ttyd)].d_open)(u.u_ttyd,flag,OTYP_CHR);
	}
}

syread(dev)
{

	if (sycheck()) {
		if (cdevsw[major(u.u_ttyd)].d_str)
			strread(u.u_ttyip);
		else
			(*cdevsw[major(u.u_ttyd)].d_read)(minor(u.u_ttyd));
	}
}

sywrite(dev)
{

	if (sycheck()) {
		if (cdevsw[major(u.u_ttyd)].d_str)
			strwrite(u.u_ttyip);
		else
			(*cdevsw[major(u.u_ttyd)].d_write)(minor(u.u_ttyd));
	}
}

syioctl(dev, cmd, arg, mode)
{

	if (sycheck()) {
		if (cdevsw[major(u.u_ttyd)].d_str)
			strioctl(u.u_ttyip, cmd, arg, mode);
		else
			(*cdevsw[major(u.u_ttyd)].d_ioctl)(minor(u.u_ttyd), cmd, arg, mode);
	}
}

syselect(dev, flag)
	dev_t dev;
	int flag;
{

	if (sycheck())
		return ((*cdevsw[major(u.u_ttyd)].d_select)(u.u_ttyd, flag));
	else
		return (0);
}

sycheck()
{
	if (u.u_ttyp == NULL) {
		u.u_error = ENXIO;
		return(0);
	}
	if (*u.u_ttyp != u.u_procp->p_pgrp) {
		u.u_error = EIO;
		return(0);
	}
	return(1);
}

/*
 * The following routines comprise the generic console driver
 * that calls the board specific drivers.
 */

extern consdev;
extern cpuboard;
extern int dualcons;
extern int Dconcurrent;

consopen(dev, flag)
	unsigned int dev,flag;
{
	(*cdevsw[major(consdev)].d_open)(consdev, flag);
}

consclose(dev)
	dev_t dev;
{
	(*cdevsw[major(consdev)].d_close)(consdev);
}

consread(dev)
	dev_t dev;
{
	(*cdevsw[major(consdev)].d_read)(consdev);
}

conswrite(dev)
	dev_t dev;
{
	(*cdevsw[major(consdev)].d_write)(consdev);
}

/*
 * consselect - select on the generic console 
 */
consselect(dev, rw)
dev_t dev;
{
	(*cdevsw[major(consdev)].d_select)(consdev, rw);
}

consioctl(dev, cmd, arg, mode)
	dev_t  dev;
	unsigned int cmd, arg, mode;
{

	switch(cmd)
	{
		case CONC_ON:
			if(Dconcurrent){ /* already on */
				return;
			}
				
			Dconcurrent = ON; /* dual console mode on */
			(*cdevsw[major(dualcons)].d_ioctl)(dualcons, cmd, arg, mode);
			if(u.u_error)	
				Dconcurrent = OFF; /* dual console mode off */

		        return;

		case CONC_OFF:
			if(! Dconcurrent){ /* already off */
				return;
			}

			(*cdevsw[major(dualcons)].d_ioctl)(dualcons, cmd, arg, mode);

			/* turn dual console mode off regardless of result of
			   the ioctl call. */
			Dconcurrent = OFF;

			return;

		default:
			(*cdevsw[major(consdev)].d_ioctl)(consdev, cmd, arg, mode);
	}

}

consputc(c)
unsigned char c;
{
	(*cdevsw[major(consdev)].d_ioctl)(consdev,CONC_PUTC, c, 0);
}

consinputc(c, status)
unsigned char c;
unsigned char status;
{
	(*cdevsw[major(consdev)].d_ioctl)(consdev,CONC_IPUTC,
 	 ( c | (status << 8)), 0);
}


extern cpu_termdriv();


consinit()
{
#if defined (ANGEL)
	m681init();
#else
		/* Moved to io/mvmecpu.c. MSW */
	cpu_termdriv();
#endif
}

dconsputc(c)
unsigned char c;
{
	(*cdevsw[major(dualcons)].d_ioctl)(dualcons,CONC_PUTC, c, 0);
}

/*
 * Call the appropriate board dependent console putchar routine.
 * If dual console mode is enabled then also call the alternate
 * console driver putc routine.  The MK68564 Terminal Driver's putc
 * routine takes care of sending the character to the alternate
 * console port itself.
 */

#if ! defined (ANGEL)
putchar(ccc)
	char ccc;
{

	consputc(ccc);

	if((Dconcurrent) && (cpuboard != VME131))
		dconsputc(ccc);
}
#endif

/*
	Called from device driver open routine to prevent an open 
	on a port which is currently being used as a dual console. 
	The process sleeps until the dual console mode is turned off.
*/
dc_osleep(tp)
register struct tty *tp;
{
	if(Dconcurrent)
		/* sleep if port has previously been opened as a dual console */
		while(tp->t_state & DCON){
			sleep(( caddr_t)&tp->t_canq,TTIPRI);
		}
}

