/*
 *	Copyright (c) 1984 AT&T
 *	Copyright (c) 1987 Motorola Inc
 *	All Rights Reserved
 *	@(#)gtty.m4	35.4
 */

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)gtty.m4	35.4"

/*
 *            General tty driver macros
 *
 *	This module contains a set of macros which have been put
 *	together to help in writing terminal drivers.  Some of the 
 * 	macros are complete driver routines while others are parts
 * 	of routines.  Macro calls to GTTY_open, GTTY_close, GTTY_read,
 *	GTTY_write, GTTY_ioctl, GTTY_proc and GTTY_scan, can be made 
 *	to generate the open, close, read, write, ioctl, proc and scan
 * 	routines in their entirety.  The other macros contain general
 * 	portions of the	param and interrupt routines.  
 *
 *	The following symbols should be defined before using this 
 *	macro package:
 *
 *		Name			Example
 *		GTTYopen		m5open
 *		GTTYclose		m5close
 *		GTTYread		m5read
 *		GTTYwrite		m5write
 *		GTTYioctl		m5ioctl
 *		GTTYproc		m5proc
 *		GTTYmodem		m5modem
 *		GTTYparam		m5param
 *		GTTYscan		m5scan
 *		GTTY_cnt		m5_cnt
 *		GTTY_delay		m5_delay
 *		GTTY_tty		m5_tty
 *		GTTY_addr		m5_addr
 *		GTTY_scan_flg		m5_scan_flg
 * 		GTTY_work		m5_work
 *		GTTYMSCAN		M5MSCAN
 *   		GTTY_CHAR		R6_CHAR
 *		GTTY_ACT		R6_ACT
 *		GTTYCSCAN		R6CSCAN
 *		DEVICE			m5_io
 *		SPLSCAN			spl6
 *		SPLINT			spl5
 *		
 *
 *	The following device specific macros should be defined:
 *
 *		Name			Description
 *		GTTYxie		- enable transmitter interrupt 
 *		GTTYxid		- disable transmitter interrupt 
 *		GTTYrie		- enable receiver interrupt 
 *		GTTYrid		- disable receiver interrupt 
 *		GTTYoutch	- output character to tx data register
 *		GTTYoutc_xie	- enable transmitter interrupt and output
 *				  character to tx data register, in whatever
 *				  order is required by the device.		
 *		GTTYtxen	- enable the transmitter  
 *		GTTYtxdis	- disable the transmitter
 *		GTTYbrkon	- transmit break
 *		GTTYbrkoff	- stop transmitting break
 *		GTTYtdra	- return 1 if tx data register available,
 *					 0 if not
 *		GTTYrda		- return 1 if receiver data available,
 *					 0 if not
 *		GTTYinitvars	- for any special initialization during the 
 *				  open routine. This should be defined to be
 *				  null for drivers that do not need it.
 *		GTTYresint	- reset interrupt at the end of interrupt 
 *				  processing. For devices that require no
 *				  such resetting, define this to be null.
 *		GTTYresesint	- reset external status interrupt. Define
 *				  as null if not needed.
 *		DEVTOADDR	- needs to br defined for drivers for devices
 *				  having one channel per chip. The default 
 *				  definition is for two channels per serial
 *				  i/o chip.
 *		INCADDR		- should be defined as null for devices that
 *				  have one channel per chip.
 *
 *		cdelay		- if the device has a large transmitter fifo
 *				  and needs to do a delay within the close
 *				  before hanging up the line, this should
 *				  be defined as `delay($1)'.  Otherwise,
 *				  `cdelay' should be defined as null (`').
 *
 *		dcdscn		- for devices that generate an interrupt
 *				  when dcd changes from inactive to active
 *				  this should be defined as null. For other 
 *				  devices, "dcdscn" should not be defined.
 *				  If this is not defined, the scan macro will
 *				  also expand the part that checks for DCD 
 *				  turning on and takes the appropriate 
 *				  actions.  
 */


/********** DEFINES FOR DEBUG PRINTING *************************/
#define	DEBUG		/* turns on debug messages */
#undef DEBUG

#ifdef DEBUG

/* DEBUG levels */
#define LOG_START	10	/* start a subroutine */
#define LOG_RETURN	10	/* return from a subroutine */
#define LOG_WARN	5	/* print a warning message */
#define LOG_INTERN	20	/* debug messages in a function */
#define LOG_MAX		20	/* highest debug level */

#define LOG(level, message)			\
	if ((GTTY_debug & 0xFF) >= level) {	\
		printf message;			\
	}
#else
#define LOG(level, message) {}	/* do nothing */
#endif

/****************************************************************/

/* declarations for debug printing */
#ifdef DEBUG
int GTTY_debug = LOG_MAX;	/* debug flag */
#endif



/****************************************************************/

define(`G_OFF',`0')dnl
define(`G_ON',`1')dnl
define(`G_NULL',`2')dnl
define(`G_HI_MARK',`18')dnl

/*
 *	The macros DEVTOADDR and DEVTOTP should be used to get 
 *	the address of the device registers and the tty structure.
 */

ifdef(`DEVTOADDR',,`define(`DEVTOADDR',` GTTY_addr[dev/2]')dnl ')dnl
define(`DEVTOTP',` &GTTY_tty[dev]')dnl 
ifdef(`INCADDR',,`define(`INCADDR', `if (dev & 1) addr++')dnl ')dnl

define(`MISSING',`((struct DEVICE *)0)')dnl	/* board not there */



/* 	The "open" macro should be called without any arguments.  Its  
 * 	expansion results in the complete open routine.
 *
 * 	The "open" routine initializes device parameters and waits for
 * 	the device to go on-line.  The parameters in the routine are:
 *
 *		Name			Description
 *		dev			minor device number
 *		flag			mode of the device being opened
 */

define(`GTTY_open',						
`GTTYopen (device, flag)						
unsigned int	device, flag;						
{								
	register struct DEVICE *addr;
	int	GTTYproc ();						
	register struct tty *tp; 
	register unsigned int	dev;						
								
	dev = minor(device);

	addr = DEVTOADDR();				
	tp = DEVTOTP();

	LOG(LOG_START, ("GTTYopen: started dev = %d, flag=0x%x\n", 
	    dev, flag));

	/* check for valid dev, and missing device */		
	if (dev >= GTTY_cnt || (addr == MISSING)) {		
		u.u_error = ENXIO;				
		LOG(LOG_RETURN, ("GTTYopen: invalid or missing dev=%d\n", dev));
		return;						
	}							

	/* Put process to sleep if the port is currently being used as a 
	   dual console. */ 

	if(Dconcurrent)
		/* sleep if port has previously been opened as a dual console */
		while (GTTY_state[dev] & DCON) {
			sleep(( caddr_t)&tp->t_canq,TTIPRI);
		}

	INCADDR();

 	/* 
	 * 	If the line has not previously been opened, and is
  	 *	not currently being opened, call the ttinit routine  
	 *	to initialize the tty structure.  Set the pointer
	 *	of the command processing routine in the tty structure
	 *	to the address of the "proc" routine. Call the "modem" 
 	 *	routine to turn on the data terminal ready (DTR) signal,
	 * 	and call the "param" routine to initialize the device
	 * 	registers and start the "scan" routine.
	 */
 
	if (! (tp->t_state & (ISOPEN | WOPEN))) {			
		ttinit(tp);		/* init tty structure */

		if(device == consdev)
			GTTY_state[dev] |= SCON; 	/* system console */

		tp->t_proc = GTTYproc;			
		GTTYinitvars();
		GTTYrie();		/* enable receiver interrupt */	
		GTTYparam (dev);					
	}							

	/*
	 *	If the line is local or a carrier has been detected on
	 *	the line, then set t_state to CARR_ON, signifying that there
	 * 	is a connection on the line.  Otherwise, turn the CARR_ON 
	 *	bit off.
	 */

	if (tp->t_cflag & CLOCAL || GTTYmodem (dev,G_ON))		
	{
		tp->t_state |= CARR_ON;				
		wakeup(&tp->t_canq);		
	}
	else							
		tp->t_state &= ~CARR_ON;			
								

	/*
	 * 	If the device has been opened with "O_NDELAY" as part of 
	 *	its mode setting, then return without waiting for carrier. 
	 *	Otherwise, if the carrier is not on, set t_state to WOPEN
	 * 	(waiting to open) and sleep on the canonical queue for the 
	 * 	tty. The wakeup is done in either of the following routines: 
	 *	(a) in the interrupt routine on getting a dcd-on interrupt 
	 *	for a device that does not have the CARR_ON bit set in 
	 * 	t_state or (b) in the scan routine, when a device is found 
	 *	which has a local carrier, or whose dcd bit in the status 
	 *	register is set.
	 */


	if (!(flag & FNDELAY))					
		while (!(tp->t_state & CARR_ON)) {		
			tp->t_state |= WOPEN;			
			sleep(( caddr_t)&tp->t_canq, TTIPRI);	
		}						
	(*linesw[tp->t_line].l_open)(tp);			
	LOG(LOG_RETURN, ("GTTYopen: return okay\n"));
}')dnl								




/* 	The "close" macro should be called without any arguments, to
 *	generate the entire close routine.
 *
 *	The "close" routine calls the line discipline close to flush the i/o
 *	queues and turns off the data terminal ready (DTR) signal for phone 
 *	line connections.
 *	
 *		Parameter    		Description
 *		dev			minor device number
 */

define(`GTTY_close',						
`GTTYclose (dev)							
unsigned int	dev;						
{								
	register struct tty *tp; 
	int	cur_lvl;						
									
	dev = minor(dev);

	tp = DEVTOTP();

	LOG(LOG_START, ("GTTYclose: started dev = %d\n", dev));

	(*linesw[tp->t_line].l_close)(tp);			
	cur_lvl = SPLINT();						

	/*
	 *	If the terminal is set to hang up on the last close,
	 *	and the t_state bit is not ISOPEN, call the modem 
	 * 	routine to turn off DTR for this device.  For devices
	 *	that have large transmitter fifos, do a delay call.
	 *
	 */
 
	if ((tp->t_cflag & HUPCL) && (!(tp->t_state & ISOPEN))) {	
		cdelay(GTTY_delay [tp->t_cflag&CBAUD]); /* delay for fifo empty */
		GTTYmodem (dev, G_OFF);				
	}							
	GTTY_state[dev] &= ~SCON;
	splx(cur_lvl);						

	LOG(LOG_RETURN, ("GTTYclose: return okay\n"));

}')dnl								




/*
 *	The read macro takes no arguments and contains the complete
 *	read routine.
 *
 * 	The read routine is called via cdevsw to read character data 
 *	from the device. Calls the line discipline read routine.
 */

define(`GTTY_read',						
`GTTYread (dev)							
unsigned int	dev;						
{								
	register struct tty *tp;

	dev = minor(dev);

	LOG(LOG_START, ("GTTYread: started dev = %d\n", dev));

	tp = DEVTOTP();						
	(*linesw[tp->t_line].l_read)(tp);			

	LOG(LOG_RETURN, ("GTTYread: return okay\n"));

}')dnl								




/*
 *	The write macro takes no arguments and contains the complete
 *	write routine.
 *
 * 	The write routine is called via cdevsw to write character data 
 *	to the device. Calls the line discipline write routine.
 */

define(`GTTY_write',						
`GTTYwrite (dev)							
unsigned int	dev;						
{								
	register struct tty *tp;

	dev = minor(dev);

	LOG(LOG_START, ("GTTYwrite: started dev = %d\n", dev));

	tp = DEVTOTP();						
	(*linesw[tp->t_line].l_write)(tp);			

	LOG(LOG_RETURN, ("GTTYwrite: return okay\n"));

}')dnl




/*
 *	The ioctl macro does not take any arguments.  It 
 *	should be called to generate the entire ioctl routine.
 *
 *	The ioctl routine is called via cdevsw to implement 
 *	ioctl system calls. Calls ttiocom to either get or 
 *	change tty structure settings.  If the settings are 
 *	changed, calls param to change the corresponding 
 *	settings on the device.
 */

define(`GTTY_ioctl',						
`GTTYioctl (device, cmd, arg, mode)					
unsigned int	device, cmd, arg, mode;				
{								
	register struct tty *tp;
	register cflag;
	register iflag;
	register char cflgtest, iflgtest;
	register unsigned int dev;
	register struct DEVICE *addr;

	dev = minor(device);
	addr = DEVTOADDR();
	INCADDR();

	LOG(LOG_START, 
	     ("GTTYioctl: started dev=%d, cmd=0x%x, arg=0x%x, mode=0x%x\n",
	     dev, cmd, arg, mode));

	tp = DEVTOTP();						
	cflgtest = G_OFF;
	iflgtest = G_OFF;

#ifdef CFLAGMASK
	cflag = tp->t_cflag&CFLAGMASK;
#endif

#ifdef IFLAGMASK
	iflag = tp->t_iflag&IFLAGMASK;
#endif

	switch(cmd){

		case CONC_PUTC:
			GTTYputc (dev, arg);
			break;

		case CONC_IPUTC:
			GTTYinputc (dev, (arg&0xff), (arg >> 8));
			break;

		case CONC_BOOT:
			/* The system was booted in dual console mode.
			   The port is already open. Do not disturb the current
			   baud rate and protocol settings, but do enable 
			   interrupts. */
			GTTYrie();	/* enable receiver interrupts */
			GTTY_state[dev] |= DCON; 
			break;

		case CONC_ON:
			if( GTTY_state[dev] & DCON){ /* dual console mode is on */
				break;
			}
			if( ! (tp->t_state & (ISOPEN | WOPEN))){
				GTTYopen (device, 0);  	

				/* Remember to close when dual console is turned				   off. */
				if(!u.u_error)
					GTTY_state[dev] |= (DCON | DCCLOSE); 
				break;
			}
			/* Port is already open due to a previous open. Dont
			   disturb the current baud rate and protocol settings
			   on the port. */

			GTTY_state[dev] |= DCON; 
			
			break;
		


		case CONC_OFF:
			if(!(GTTY_state[dev] & DCON)){ /*dual console mode is off */
				break;
			}
			if(GTTY_state[dev] & DCCLOSE){
				GTTYclose (device);
			}
			GTTY_state[dev] &= ~(DCON | DCCLOSE);
			wakeup(&tp->t_canq);		

			break;

		default:
			if (ttiocom(tp, cmd, arg, mode)){
#ifdef CFLAGMASK  
				cflgtest = (cflag != (tp->t_cflag & CFLAGMASK));
#endif

#ifdef IFLAGMASK
				iflgtest = (iflag != (tp->t_iflag & IFLAGMASK));
#endif

			}
			if ( cflgtest || iflgtest)
				GTTYparam (dev);

			break;

	}  /* end of switch */
	LOG(LOG_RETURN, ("GTTYioctl: return okay\n"));

}')dnl




/*
 *	A macro call to GTTY_proc expands to the entire proc routine.
 *	This macro takes no arguments. 
 *
 * 	The proc routine is called by various driver and line discipline
 *	routines to perform various commands.
 *
 *		Parameter	Description
 *		tp		pointer to tty structure for device
 *		cmd		command to be executed
 *
 *	The various commands are:
 *		T_TIME		turn off break signal
 *		T_WFLUSH	reset transmit buffer
 *		T_RESUME	resume output
 *		T_OUTPUT	call line discipline routine to refill
 *				transmit buffer
 *		T_SUSPEND	stop output
 *		T_BLOCK		block input
 *		T_RFLUSH	flush input queue
 *		T_UNBLOCK	restart input
 *		T_BREAK		send break for 0.25 second
 *		T_PARM		call the param routine
 */

define(`GTTY_proc',							
`GTTYproc (tp, cmd)	
register struct tty *tp;						
unsigned int	cmd;							
{									
	register struct DEVICE *addr; 					
	int	dev;					
	register int	piplev;						

	piplev = SPLSCAN();						
	dev = tp - GTTY_tty;						
	addr = DEVTOADDR();
	INCADDR();
									
	LOG(LOG_START, 
	     ("GTTYproc: started tp=0x%x, dev=%d, cmd=%d, t_state=0x%x\n", 
	     tp, dev, cmd, tp->t_state));
	
	switch (cmd) {							
									
	case T_TIME:		/* resuming output after delay */	
		tp->t_state &= ~TIMEOUT;				
		GTTYbrkoff();	/* turn off break */
		LOG(LOG_INTERN, ("GTTYproc: T_TIME\n"));
		goto start;						
	 
	/*
	 * 	For the T_WFLUSH command, remove the characters from
	 *	the transmit buffer, and reset the count of the number
	 *	of characters to be transmitted to zero.
	 */

	case T_WFLUSH:		
		tp->t_tbuf.c_size -= tp->t_tbuf.c_count;		
		tp->t_tbuf.c_count = 0;					
		LOG(LOG_INTERN, ("GTTYproc: T_WFLUSH\n"));
		/* fall thru */						
									
	case T_RESUME:		/* resume output, XON has been received */
		LOG(LOG_INTERN, ("GTTYproc: T_RESUME started\n"));
		tp->t_state &= ~TTSTOP;					
		GTTYtxen();		/* re-enable transmitter */	
		if (!GTTYtdra())	/* transmitter buffer available ? */	
			break;						
		LOG(LOG_INTERN, ("GTTYproc: T_RESUME\n"));
		/* fall thru */						
									
	case T_OUTPUT:		
	start:  
		GTTY_toutput();	
		break;
									
	case T_SUSPEND:	   /* stop output, XOFF has been received */	

		/* There may be pending characters in the transmitter
		 *  ( in FIFO + 1 in shift register), therefore disabling the
		 *  transmitter is the appropriate way to stop the transmission
		 *  as soon as possible to prevent overrun condition on the 
		 *  other side.
		 */

		GTTYtxdis();    		/* disable transmitter */	
		tp->t_state |= TTSTOP; /* output is not done if TTSTOP is set */
		LOG(LOG_INTERN, ("GTTYproc: T_SUSPEND\n"));
		break;						
									
	case T_BLOCK:	

		/*
		 *	Used to suspend input to the device.  The proc routine
		 * 	is called from the receive interrupt routine with
		 *	T_BLOCK as the argument when the raw input queue 
		 *	exceeds the "high-water" mark.  
		 */

		tp->t_state &= ~TTXON;					
		tp->t_state |= TBLOCK;					
		
		/*	
		 *	If BUSY is set, set TTXOFF in t_state.  This will tell
		 *	the transmit interrupt routine to send a control-s 
		 *	ie CSTOP on the line to halt input.  Otherwise, send
		 *	the character CSTOP.
		 */

		if (tp->t_state & BUSY) 
			tp->t_state |= TTXOFF;				
		else 							
			GTTYoutch(CSTOP);				
		LOG(LOG_INTERN, ("GTTYproc: T_BLOCK\n"));
		break;							
									
	case T_RFLUSH:	
		/*
		 *	T_RFLUSH is used to flush the input queues.
		 *	Do nothing if input is not blocked. Otherwise,
		 * 	fall through to T_UNBLOCK which resets TTXOFF
		 *	and TBLOCK and sets TTXON in t_state.
		 */

		LOG(LOG_INTERN, ("GTTYproc: T_RFLUSH started\n"));

		if (!(tp->t_state & TBLOCK))				
			break;						
		LOG(LOG_INTERN, ("GTTYproc: T_RFLUSH\n"));
		/* fall thru */						
									
	case T_UNBLOCK:					
		GTTY_tunblock();
		break;							
									
	case T_BREAK:		
		GTTY_tbreak();
		break;							
									
	case T_PARM:	
		LOG(LOG_INTERN, ("GTTYproc: T_PARM\n"));
		GTTYparam (dev);						
		break;							
	}								
	splx(piplev);							

	LOG(LOG_RETURN, ("GTTYproc: return okay\n"));
}')dnl

/* 
 *	The following macro is the T_OUTPUT case of the GTTYproc macro
 *	defined above.
 */

define(`GTTY_toutput',
`{							
		register struct ccblock *tbuf;				

		LOG(LOG_INTERN, ("GTTYproc: T_OUTPUT\n"));

		/*
 		*	If the state of the tty is BUSY or stopped (TTSTOP),
 		*	do nothing.
 		*/

		if (tp->t_state & (BUSY | TTSTOP))			
			break;						

		/*
		 *	Check to see if a transmit buffer has been allocated
 		 *	or the transmit buffer is empty.  If a transmit buffer
	 	 *	has been allocated, reset the data pointer in it to the
 		 *	beginning of the block.  Call the line discipline 
 		 *	output routine to move characters from the output 
 		 *	queue to the transmit buffer.  If no characters are
 		 *	placed in the transmit buffer, do nothing.  Otherwise,
 		 *	output a character and set t_state to BUSY.
 		 */ 	

		tbuf = &tp->t_tbuf;					
		if (tbuf->c_ptr == NULL || tbuf->c_count == 0) {	
		 	if (tbuf->c_ptr)				
		      		tbuf->c_ptr -= tbuf->c_size - tbuf->c_count;
		 	if (!(CPRES & (*linesw[tp->t_line].l_output)(tp))) 
			 	break;					
	  	}							
		tp->t_state |= BUSY;			/* send char */ 

		GTTYoutc_xie(*tp->t_tbuf.c_ptr);			

		/* Also send character to dual console if enabled. */
		if(Dconcurrent && (GTTY_state[dev] & SCON))
			dconsputc(*tp->t_tbuf.c_ptr);

		LOG(LOG_INTERN, ("GTTYproc: dev=%d, put char=0x%x\n",
				  dev, *tp->t_tbuf.c_ptr));

		++tp->t_tbuf.c_ptr;					
		--tp->t_tbuf.c_count;	/* update count */		
	}')dnl								



/* 
 *	The following macro is the T_UNBLOCK case of the GTTYproc macro
 *	defined above.
 */

define(`GTTY_tunblock',
	`{	
		/*
		 *	The proc routine is called with T_UNBLOCK as the
		 * 	argument by the interrupt routine when it finds 
		 *	a tty whose t_state is T_BLOCK and whose raw queue
		 *	has gone below a "low-water" mark.
		 *
		 */	

		tp->t_state &= ~(TTXOFF | TBLOCK);			

		/*	
		 *	If BUSY is set, set TTXON in t_state.  This will tell
		 *	the transmit interrupt routine to send a control-q 
		 *	(CSTART) on the line to restart input. Otherwise, 
		 *	send the character CSTART.
		 */

		if (tp->t_state & BUSY)
			tp->t_state |= TTXON;				
		else 							
			GTTYoutch(CSTART);				
		LOG(LOG_INTERN, ("GTTYproc: T_UNBLOCK\n"));
		}')dnl

/* 
 *	The following macro is the T_BREAK case of the GTTYproc macro
 *	defined above.
 */

define(`GTTY_tbreak',
	`{
		/*
		 *	Send a break for 1/4 seconds.
		 *	Start transmitting a break. The timeout routine
		 *	will call ttrstrt after HZ/4 clock ticks (.25 secs).
		 *	The ttrstrt routine will call GTTYproc with T_TIME
		 *	which stops the break condition.
		 */

		GTTYbrkon();		
		tp->t_state |= TIMEOUT;					
	  	timeout(ttrstrt, tp, HZ/4);				
		LOG(LOG_INTERN, ("GTTYproc: T_BREAK\n"));
	}')dnl


/*
 *	The macro GTTY_scan is to be used to generate the "scan" 
 *	routine of all tty drivers. 
 *
 *	The scan routine is called from the open routine on the first
 *	open of the device, to ensure that the characters that have  
 *	been placed in the receive buffer by the receive interrupt 
 *	routine, are stored in the raw input queue, for each device.
 *	It calls the line-discipline l_input routine to store the 
 *	characters read.
 *
 *	Each element of the GTTY_work array corresponds to one minor
 *	device number.  Each element has one of three values.  When
 *	a character is placed in the receive buffer, by the interrupt
 *	routine, it also sets GTTY_work for that device to GTTY_ACT.
 *	When the scan routine is entered, if the value of GTTY_work
 *	for a device is GTTY_ACT, it is changed to GTTY_CHAR.  The 
 *	next time the scan routine is entered, if the value is still
 *	GTTY_CHAR, it is set to zero and the receive buffer is moved
 *	to the raw input queue.
 *	 
 */	
define(`GTTY_scan',						
`GTTYscan ()
{
	register struct tty	*tp;
	register int		dev;
	register char		*workp;
	register struct m5_io	*addr;
	static	scancnt;



	for (dev = 0, tp = DEVTOTP(), workp = GTTY_work; dev < GTTY_cnt; dev++, workp++,tp++) {
		if (*workp & GTTY_CHAR) {				
			*workp = 0;					
			if (tp->t_rbuf.c_ptr) {				
				tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size	
						- tp->t_rbuf.c_count;	
				(*linesw[tp->t_line].l_input)(tp, L_BUF);
			}						
		}							
		if (*workp & GTTY_ACT)					
			*workp |= GTTY_CHAR;				
	}								

	GTTY_dcdscn();
	
	timeout(GTTYscan, (struct tty *) 0, GTTYCSCAN);
}')dnl									



/*
 *	This part of the scan routine is called in the case of devices 
 *	that do not generate interrupts when DCD changes from inactive
 *	to active.  In that case, scanning is needed to wakeup the
 *	task that is sleeping on the line, waiting for DCD to become
 *	active before opening the line. 
 */

ifdef(`GTTY_dcdscn',,`define(`GTTY_dcdscn',
`{									
									
	/*
	 * 	If this is the GTTYMSCANth call to the scan routine,
	 * 	then check the status of the carrier, otherwise 
 	 *	do nothing.
	 */
 
	if (!(scancnt++ % GTTYMSCAN)) {					

		for (dev = 0, tp = DEVTOTP(); dev < GTTY_cnt; dev++, tp++) {

			/*
			 * 	If the line is not open and is not waiting 
			 * 	to be opened, do nothing.
 			 */
	
			if (!(tp->t_state & (ISOPEN | WOPEN)))		
				continue;				

			/*
			 *	If the carrier is not local or the carrier 
			 *	has been detected, and t_state does not
			 *	have CARR_ON set, all processes sleeping
			 * 	on the line are awaken and t_state is set
			 *	to CARR_ON.
			 */
	
			if (tp->t_cflag & CLOCAL || GTTYmodem (dev, G_NULL)) {
				if ((tp->t_state & CARR_ON) == 0) {	
					wakeup(&tp->t_canq);		
					tp->t_state |= CARR_ON;		
				}					
			}						
		}							
	}
}')dnl
')dnl

/*
 *	This macro starts off the scan routine, if it has not
 *	already been started. This should be called at the 
 *	beginning of the param routine.
 */

define(`GTTY_startscan',						
	`if (GTTY_scan_flg == 0) {						
		GTTY_scan_flg++;						
		GTTYscan ();						
	}')dnl



/*	The macro GTTY_dcdoffint should be used as the part of the
 *	interrupt routine that handles the type of interrupt
 *	generated when DCD changes from active to inactive.
 *
 */

define(`GTTY_dcdoffint',						
	`{								
		/*	If the CARR_ON is set in t_state, and the tty
		 *	is open, a hangup signal is sent to the process
		 *	group leader for that tty, the data terminal ready
		 *	signal is turned off, and ttyflush is called to
		 *	empty the input and output queues.
		 */

		if (!(tp->t_cflag & CLOCAL)) { 
			if (tp->t_state & CARR_ON) {				
				if (tp->t_state & ISOPEN) {			
			    		signal(tp->t_pgrp, SIGHUP);
			    		GTTYmodem (dev, G_OFF);			
			    		ttyflush(tp, (FREAD | FWRITE));		
				}						
				tp->t_state &= ~CARR_ON;			
			}
		}							
	}')dnl	



define(`GTTY_dcdonint',
		`{
			/*
			 * 	If the line is not open and is not waiting 
			 * 	to be opened, do nothing.
			 *	has been detected, and t_state does not
			 *	have CARR_ON set, all processes sleeping
			 * 	on the line are awaken and t_state is set
			 *	to CARR_ON.
			 */

			if (tp->t_state & (ISOPEN | WOPEN)) {
				if ((tp->t_state & CARR_ON) == 0) {	
					wakeup(&tp->t_canq);		
					tp->t_state |= CARR_ON;		
				}					
			}
		}')dnl





/*
 *	This macro should be called to handle the transmit
 * 	interrupt case of the interrupt routine.  
 */

define(`GTTY_xint',
`{
	sysinfo.xmtint++;  /* Increment counter of transmitter interrupts */

	/*
	 * 	If the state of the tty is TTXOFF (input is to be stopped),
	 *	send a control-s (CSTART) character; if the state is TTXON,
	 *	then send a control-q (CSTART) character.  
	 */	

	if (tp->t_state & TTXOFF){
		GTTYoutch(CSTOP);
		tp->t_state &= ~TTXOFF;
	} else if (tp->t_state & TTXON){
		GTTYoutch(CSTART);
		tp->t_state &= ~TTXON;
	} else {						
		/*
	 	 *    	Otherwise, disable  transmitter interrupts, reset 
		 * 	the BUSY bit in t_state, and call the proc routine
		 *	to move characters from the output queue to the 
		 *	transmit buffer.  The proc routine will re-enable       
		 *	transmitter interrupts if there are characters to
		 *	be output.
		 */	

		GTTYxid();						
		tp->t_state &= ~BUSY;				
		GTTYproc (tp, T_OUTPUT); 				
	}							
}')dnl


/*
 *	The following macro handles receiver interrupts.
 *
 *	Before this macro is called, the following variables should
 *	be defined as and assigned to the values shown:
 *
 *		Variable	Defined as	Value
 *		c		unsigned char	character received
 *		flg		int		tp->t_iflag
 *		lcnt		int		1
 *		lbuf[3]		array of char		
 *
 */

define(`GTTY_rint',
	`{								
		unsigned char	ctmp;					

		sysinfo.rcvint++; /* Increment counter of receiver interrupts */

		/*	If start/stop output control is enabled in t_iflag,
		 *	the state of the tty is TTSTOP and the character 
		 *	read from the receive buffer is a CSTART, then call
		 *	the command processing routine to resume output.
		 *	If the character is a CSTOP, call the proc routine
		 * 	to suspend output.
		 */
		
		if (flg & IXON) {						
			ctmp= c & 0x7f;				
			if (tp->t_state & TTSTOP){			
				if ((ctmp == CSTART) || (flg & IXANY))		
   					GTTYproc (tp, T_RESUME);		
			} else {					
				if (ctmp == CSTOP)			
					GTTYproc (tp, T_SUSPEND);		
			  }						
			if (ctmp == CSTART || ctmp == CSTOP) {
				GTTYresint();		
				return; /* ignore these characters */	
			}
		}							
		/*
		 *	If ISTRIP is set, strip input character to 7 bits,
		 *	or if less than CS8, strip possible parity bit,
		 *	instead of letting it pass thru.  This conforms with
		 *	the intent of ATT of not having parity in the data byte.
		 */

		if (flg & ISTRIP || (tp->t_cflag & CSIZE) < CS8)
			c &= 0x7f;					
		
			/*
			 *	Otherwise, to avoid ambiguity, a valid
			 *	character of 0xFF is read as 0xFF, 0xFF.
			 */

		else if (c == 0xff && flg & PARMRK) {			
				lbuf[1]= 0xff;				
				lcnt= 2;				
		}		   						
	}')dnl								



/*	Should be called for break interrupt processing */

define(`GTTY_bint',							
	`{								
		if (!(flg&IGNBRK) && (flg&BRKINT))				
			(*linesw[tp->t_line].l_input)			
				        (tp, L_BREAK);			
		GTTYresesint();
		GTTYresint();
		return;	/* ignore the current character */		
	}')dnl	



/*	Receiver error interrupt processing */
	
define(`GTTY_eint',							
	`{								
		/*
		 *	If PARMRK is set, a character with a framing
		 *	or parity error is read as the three-character
		 *	sequence 0xFF, 0, X, where X is the data of the
		 *	character received in error. 
		 */

		if (flg & PARMRK) {					
			lbuf[2] = 0xFF;						
			lbuf[1] = 0;					
			lcnt = 3;					
			sysinfo.rawch += 2;				
		} 
		/*
		 *	If PARMRK (mark parity errors) is not set,
		 *	a framing or parity error which is not 
		 *	ignored is read as the character NUL (0).
		 */

		else c = 0;						
	}')dnl




/*
 *	This macro is to be called at the end of the received character
 *	processing, in the interrupt routine.  It moves the characters
 *	from the temporary buffer "lbuf" into the receiver buffer.
 */

define(`GTTY_rbuf',							
	`{								
		lbuf[0] = c;						
		if (tp->t_rbuf.c_ptr != NULL) {				
			while (lcnt) {					
				cur_lvl = SPLSCAN();			
				*tp->t_rbuf.c_ptr++ = lbuf[--lcnt];
				/*
				 * If the number of characters has crossed
				 * the high water mark, and the state is not
				 * T_BLOCK, then call GTTYproc to stop the
				 * input.
				 */
 
				if ((count = --tp->t_rbuf.c_count) < G_HI_MARK) {
					if (!(tp->t_state & TBLOCK) && tp->t_iflag & IXOFF)
					    GTTYproc (tp, T_BLOCK);	
					/*
					 * If there is no receiver data 
					 * available,  or if there is 
					 * is no space left in the receive
					 * buffer, call the line discipline
					 * l_input routine to reinitialize
					 * the receive buffer and move the
					 * characters from the receive buffer
					 * to the raw queue.
					 */

					if(!(GTTYrda) || count <= 1) {	
			    		GTTY_work[dev] = 0;	
			    		tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size - count;
			    		(*linesw[tp->t_line].l_input)(tp,L_BUF);
						if(tp->t_rawq.c_cc <= TTXOHI && tp->t_iflag & IXOFF)
							GTTYproc (tp, T_UNBLOCK);
					}				
				}					
			    GTTY_work[dev] = GTTY_ACT;		
			    /* note GTTY_CHAR bit cleared */		
				splx(cur_lvl);				
			} 						
		}
	}')dnl 	



/*	The macro GTTY_b0 should be called from within the
 *	param routine to check for a baud rate of zero,
 *	and drop the carrier if the baud rate is zero.
 * 
 *	If the tty is open, a hangup signal is sent to the 
 *	process group leader for that tty, ttyflush is called 
 *	to empty the input and output queues and the line is
 *	disconnected.
 */

define(`GTTY_b0',
`{
	if ((flags & CBAUD) == B0)	/* hang up on zero baud rate */
		if (tp->t_state & ISOPEN){
			signal(tp->t_pgrp, SIGHUP);
			ttyflush(tp, (FREAD | FWRITE));
			GTTYmodem (dev, G_OFF);
		}
}')dnl


/*
 *	This macro should be called from the param routine to
 *	check for a local, direct connection with no modem 
 *	control.
 *
 *	If the line is local and the CARR_ON bit in t_state is
 *	not set, then this piece of code wakes up the processes
 *	waiting for a carrier, and updates the DCD status in 
 *	t_state.
 */

define(`GTTY_clocl',
`{
	if (flags & CLOCAL)		
		if ((tp->t_state & CARR_ON) == 0) {
			wakeup(&tp->t_canq);
			tp->t_state |= CARR_ON;
		}
}')dnl
