/*		Copyright (c) 1985 Motorola Inc.		*/
/*		All Rights Reserved				*/
/*  @(#)kern-port:io/lp.c	35.3	*/
/*
 *	line printer driver
 */

#ident	"@(#)kern-port:io/lp.c	35.3"

#undef DEBUG
#include "sys/param.h"
#include "sys/types.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/tty.h"
#include "sys/lprio.h"
#include "sys/lp0.h"
#include "sys/mc68153.h"
#include "sys/sysmacros.h"


lp_open(lp)
register struct lp *lp;
{
	if (lp->col == 0) {		/* first time only */
		lp->ind = 4;		/* set indentation */
		lp->col = 132;		/* set number of columns */
		lp->line = 66;		/* set number of lines/page */
	}
}

lp_output(lp,c)
register struct lp *lp;
register c;
{
	if (lp->flag & L_RAW) {		/* No canonical processing */
		putc (c, &lp->l_outq);
		return;
	}

	/* ccc = current character count (logical printer head placement) */
	/* mcc = maximum character count (actual physical prtr head placement */
	/* mlc = maximum line count */

	switch(c) {
	case '\t':
		lp->ccc = ((lp->ccc+8-lp->ind) & ~7) + lp->ind;
		return;
	case '\n':
		lp->mlc++;
		if(lp->mlc >= lp->line )
			c = '\f';
	case '\f':
		putc('\r', &lp->l_outq);
		lp->mcc = 0;
		if (lp->mlc) {
			putc(c, &lp->l_outq);
			if(c == '\f')
				lp->mlc = 0;
		}
		goto strtlp;
	case '\r':
		putc('\r', &lp->l_outq);
	   strtlp:
		lp->ccc = lp->ind;
		splx(lp->lev);
		(*lp->lp_routine)(lp);
		spl0();
		return;
	case '\b':
		if(lp->ccc > lp->ind)
			lp->ccc--;
		return;
	case ' ':
		lp->ccc++;
		return;
	default:
		if(lp->ccc < lp->mcc) {
			if (lp->flag&L_NOCR) {
				lp->ccc++;
				return;
			}
			putc('\r', &lp->l_outq);
			lp->mcc = 0;
		}
		if(lp->ccc < lp->col) {
			while(lp->ccc > lp->mcc) {
				putc(' ', &lp->l_outq);
				lp->mcc++;
			}
			putc(c, &lp->l_outq);
			lp->mcc++;
		}
		lp->ccc++;
	}
}

lp_write(lp)
register struct lp *lp;
{	register c;

	if (u.u_error)
		return;

	while (u.u_count) {		/* while bytes remaining for output */

		/* while charctr count in outpt queue is already large enough */
		/* "sleep" for a while so as not to flood the character buffr */
		/* pool.                                                      */

		while (lp->l_outq.c_cc > LPHIWAT) {
			splx(lp->lev);	/* inhibit interrupts from lp */
			/* check for interrupts in progress */
			if (u.u_error = (*lp->lp_routine)(lp))
				return;
			if (lp->l_outq.c_cc > LPHIWAT) {
#ifdef DEBUG
				printf("lp0: going to sleep \n");
#endif
				lp->flag |= L_ASLP;
				sleep(lp, LPPRI);
			}
			spl0();  /* start interrupts from lp */
		}
		c = fubyte(u.u_base++);  /* fetch a byte from user space */
		if (c < 0) {  /* if bad return error */
			u.u_error = EFAULT;
			break;
		}
		u.u_count--;
		lp_output(lp,c);  /* store character in queue */
	}
	splx(lp->lev);
	u.u_error = (*lp->lp_routine)(lp);
	spl0();
}

lp_ioctl(cmd, arg, mode, lp)
register struct lp *lp;
{
	struct lprio lpr;

	switch (cmd) {
	default:
		u.u_error = EINVAL;
		return;

	case LPRGET:
		lpr.ind = lp->ind;
		lpr.col = lp->col;
		lpr.line = lp->line;
		if (copyout(&lpr, arg, sizeof lpr))
			u.u_error = EFAULT;
		return;

	case LPRSET:
		if (copyin(arg, &lpr, sizeof lpr)) {
			u.u_error = EFAULT;
			return;
		}
		lp->ind = lpr.ind;
		lp->col = lpr.col;
		lp->line = lpr.line;
		return;

	case LPRRAW:		/* canonical processing disabled */
		lp->flag |= L_RAW;
		lp->ccc = 0;
		lp->mcc = 0;
		lp->mlc = 0;
		return;

	case LPRCAN:		/* canonical processing enabled */
		lp->flag &= ~L_RAW;
		return;
	}
}
