
/*		Copyright (c) 1985 Motorola,1988 Inc.		*/
/*		All Rights Reserved				*/
/*  		@(#)lp147.c	35.1					*/
/*
 *	MVME147 line printer driver
 */

#include "sys/param.h"
#include "sys/types.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/tty.h"
#include "sys/lprio.h"
#include "sys/lp0.h"
#include "sys/pcc.h"
#include "sys/sysmacros.h"

#define DEBUG		/* turn on entry printfs */
#undef DEBUG
#define DEBUGINT	/* turn on "unexpected interrupt" printf */
#undef DEBUGINT

extern struct pcc *pcc_addr[];
extern int l147_cnt;
extern int l147_ilev[];
extern int l147_ivec[];

#define DSR_ADDR  (char *)0xfffe2800	/* Printer Data/Status register address */
#define LPMAX	1	

struct lp lp_dt[LPMAX];

/* initialize the board */

l147open(dev, mode)
{
	register struct lp *lp;
	register struct pcc *lpdev;
	extern l147start();
	unsigned char stat;
#ifdef REL2
	register unsigned short port = dev;
#else
	register unsigned short port = minor(dev);
#endif

	/* if not a printer we use or another lp file is already open */
	/* generate I/O error      */
#ifdef DEBUG
	printf("LP147: open\n");
#endif

	if (port >= l147_cnt || ((lp = &lp_dt[port])->flag&L_OPEN)) { 
		u.u_error = EIO;
		return;
	}

	lpdev = pcc_addr[port];

	/* check for device, use data register to avoid generating
	   a strobe */
	if ( bprobe (&lpdev->picr, -1) ) {	/* is device missing */
		u.u_error = ENXIO;
		return;
	}
	
	if (lp->lp_routine == 0) {
		lp->lp_routine = l147start;

		if (l147_cnt > LPMAX) {
			l147_cnt = LPMAX;
			if (port >= LPMAX) {
				u.u_error = EIO;
				return;
			}
		}

		/* initialize lp data structures */
		if (l147_ilev[port] > 7) {
			u.u_error = EIO;
			return;
		}
		lp->lev = 0x2000 | (l147_ilev[port] << 8);
		lp->dev = port;
		lp_open(lp);

		lpdev->pcr  = 0;
		lpdev->picr  = l147_ilev[port] | PCC_PIENAB;
		l147timer();		/* start timer */

	}

	/* if printer selected and not faulted then OK otherwise generate */
	/* I/O error. */

	stat = (*DSR_ADDR) & PCC_STATMSK;
	if ((stat & (PCC_SELECT | PCC_FAULT | PCC_PE )) != PCC_SELECT ) {
		u.u_error = EIO;
		return;
	}

	lp->flag |= L_OPEN;
	lp_output(lp,'\f');	/* send form feed character to prtr to ensure */
    			        /* that new characters start on new page      */
}

l147close(dev)
{
#ifdef REL2
	register unsigned short port = dev;
#else
	register unsigned short port = minor(dev);
#endif
	register struct lp *lp = &lp_dt[port];

#ifdef DEBUG
	printf("LP147: close\n");
#endif
	lp_output(lp,'\f');
	/* now turn off "fault" indicator   */
	lp->flag &= ~L_FAULT;
	lp->flag &= ~L_OPEN;
}

l147intr(dev)
{	register struct lp *lp = &lp_dt[dev];
	register x;

	/* check interrupt */
#ifdef DEBUG
	printf("LP147: interrupt\n");
#endif
#ifdef DEBUGINT
	if((lp->flag&L_BUSY) == 0) {
		printf("LP147: unexpected interrupt\n");
		x = *DSR_ADDR;
		return;
	}
#endif
	x = spl6();			/* lock out timer */
	lp->flag &= ~(L_BUSY|L_TOUT);
	splx(x);

	/* process interrupt */
	l147start(lp);

	/* character count in queue is low and asleep, wakeup process */
	if (lp->l_outq.c_cc <= LPLOWAT && lp->flag&L_ASLP) {
		lp->flag &= ~L_ASLP;
		wakeup(lp);
	}
}

l147write(dev)
{
#ifdef REL2
	register unsigned short port = dev;
#else
	register unsigned short port = minor(dev);
#endif
	lp_write(&lp_dt[port]);
}

l147start(lp)
register struct lp *lp;
{
	register struct pcc *lpdev;
	register int c;
	register tmp;
	register cnt;
	struct cblock *cp;
	char stat;

#ifdef DEBUG
	printf("LP147 start\n");
#endif
	if( lp->flag&L_BUSY )
		return(0);

	lpdev = pcc_addr[lp->dev];


	stat = (*DSR_ADDR) & PCC_STATMSK;
	if ((stat & (PCC_SELECT | PCC_FAULT | PCC_PE )) != PCC_SELECT ) {
		lpdev->picr |= PCC_FALTINT;    /* clear Fault Interrupt flag */
		if (lp->flag&L_FAULT)
			return(0);
		else {
			lp->flag |= L_FAULT;		/* flag the fault */
			printf("LP147: printer offline\n");
			return(0);
		};
	};

	lp->flag &= ~L_FAULT;

#ifdef DEBUG
	printf("LP147 status: %x\n", stat);
#endif

	/* check interrupt register to see if have ACK  */

	if (lpdev->picr & PCC_ACKINT) {
		lp->flag &= ~L_WACK;
	}

	/* if printer ready and character present strobe character out */

	if (!(lp->flag & L_WACK)) {
	    lpdev->picr |= PCC_ACKINT;    /* clear ACK Interrupt flag */
	    if((c = getc(&lp->l_outq)) >= 0) {
		*DSR_ADDR = c;
#ifdef DEBUG
		printf("LP147: \\%x", c );
#endif
		lp->flag |= L_WACK;	
		lp->flag |= L_BUSY;	
	    }
	}

	return(0);
}

l147ioctl(dev, cmd, arg, mode)
{
#ifdef REL2
	register unsigned short port = dev;
#else
	register unsigned short port = minor(dev);
#endif
	lp_ioctl(cmd,arg,mode,&lp_dt[port]);
}

l147timer()
{	register struct lp *lp;
	register struct pcc *lpdev;
	register dev;

	for (lp = &lp_dt[0], dev=0; dev<l147_cnt; dev++, lp++) {
#ifdef DEBUG
		printf("LP147: timer flag = %x\n",lp->flag);
#endif
		/* L_BUSY on iff waiting for an interrupt */
		if ((lp->flag&L_BUSY) == 0)
			continue;
		/* L_TOUT on iff no interrupt occurred in last 60 secs */
		if ((lp->flag&L_TOUT) == 0) {
			lp->flag |= L_TOUT;
			continue;
		}
		/* no interrupt in last 60 seconds */
		/* fake an interrupt condition */
		lp->flag &= ~(L_BUSY|L_TOUT|L_WACK);
		l147start(lp);		/* continue */
		/* wakeup any waiting processes */
		if (lp->l_outq.c_cc <= LPLOWAT && lp->flag&L_ASLP) {
#ifdef DEBUG
			printf("LP147: wakeup\n");
#endif
			lp->flag &= ~L_ASLP;
			wakeup(lp);
		}
	}

	timeout(l147timer,0,60*HZ);
}
