/*		Copyright (c) 1985 Motorola,1987 Inc.
 *		All Rights Reserved
 *
 *	@(#)kern-port:io/m68230.c	35.2
 *
 *	mc68230 line printer driver
 */

#ident	"@(#)kern-port:io/m68230.c	35.2"
#undef DEBUG

/************** Compiler Switches ***************************************/

/* HOWFAR	:	to see which routine has been called */
/* ERRMSG	:	to get additional error messages on system console */
/* STATISTIC	:	to get statistic information after close */

/************* Statistic values *****************************************/

#ifdef STATISTIC
int cnt_intr[LPMAX];	/* number of interupt service routine entries */
int cnt_open[LPMAX];	/* number of m230open routine entries */
int cnt_close[LPMAX];	/* number of m230close routine entries */
int cnt_write[LPMAX];	/* number of m230write routine entries */
int cnt_ioctl[LPMAX];	/* number of m230ioctl routine entries */
int cnt_start[LPMAX];	/* number of m230start routine entries */
int cnt_transfer[LPMAX];/* number of char sent to printer */
int cnt_sleep[LPMAX];	/* number of calls to sleep */
int cnt_wakeup[LPMAX];	/* number of Calls to wakeup */
int cnt_fubyte[LPMAX];	/* number of Calls to fubyte */
#endif

/************* INCLUDES *************************************************/

#include	"sys/param.h"
#include	"sys/types.h"
#include	"sys/dir.h"
#include	"sys/signal.h"
#include	"sys/user.h"
#include	"sys/errno.h"
#include	"sys/tty.h"
#include	"sys/lprio.h"
#include	"sys/lp0.h"			/* version 7.4 or later */
#include	"sys/sysmacros.h"

/************** Defines specific for mc68230 ***************************/

#define	LPMAX	2	/* up to two mc68230 are supported */
#define	LPMAXILEV 5	/* interrupt level should not exceed 5 */

/* status of printer recognized at port C data register (offset 0x99) */
#define	LPSEL	0x04	/* printer selected */
#define	LPFAULT	0x02	/* printer in fault status or paper out */
#define	LPBUSY	0x01	/* printer busy */
#define	LPOK	0x04	/* printer status: select, no fault, not busy */

/* further definition for mc68230 initialisation */
#define GCRDHL	0x03	/* port general control reg. disable handshake lines */
#define ADDRSO	0xff	/* port A data direction register set lines output */
#define CDDRSPO	0x10	/* port C data direc. reg. set pc4 prime* to output */
#define CDRINIT	0x10	/* port C data register init value */
#define SRRVECI	0x18	/* port servive request register vectored interrupts */
#define ACRIBO	0x79	/* port A control reg interrupt if both buffer empty */
#define GCRH1E	0x13	/* port general control reg. mode 0 h1 enable */
#define ENBL	0x0	/* one write to enable printer port enable c1 */

/* enable/disable interrupt generation after buffer are empty */
#define ACRDH1I 0x02	/* port A control reg. disable/enable H1 interrupts */

/************************************************************************/
/*									*/
/* The following is a description of the device registers in use here	*/
/*									*/
/************************************************************************/

struct m230_device {
	unsigned char	pgcr;	/* port general control register 81 */
	char fill0;
	unsigned char	psrr;	/* port service request register 83 */
	char fill1;
	unsigned char	pAddr;	/* port a data direction register 85 */
	char fill2;
	unsigned char	pBddr;	/* port b data direction register 87 */
	char fill3;
	unsigned char	pCddr;	/* port c data direction register 89 */
	char fill4;
	unsigned char	pivr;	/* port interrupt vector register 8b */
	char fill5;
	unsigned char	pAcr;	/* port a control register 8d */
	char fill6;
	unsigned char	pBcr;	/* port b control register 8f */
	char fill7;
	unsigned char	pAdr;	/* port a data register 91 */
	char fill8;
	unsigned char	pBdr;	/* port b data register 93 */
	char fill9;
	unsigned char	pAar;	/* port a alternate register 95 */
	char fill10;
	unsigned char	pBar;	/* port b alternate register 97 */
	char fill11;
	unsigned char	pCdr;	/* port c data register 99 */
	char fill12;
	unsigned char	psr;	/* port status register 9b */
	char fill13[37];	/* centronics not used by hardware */
	/* timer interface not used here */
	unsigned char	enbl;	/* printer port enable  c1 */
	char fill14;
	/* remaining 62 bytes not used */
};

/*************** EXTERNALS **********************************************/
/*						 Declarations in conf.c */
extern int m230_ivec[];
extern int m230_cnt;
extern int m230_ilev[];

/************* DECLARATIONS ***************************************/

struct lp m230lp_dt[LPMAX]; 	/* comes from lp0.h */
struct m230_device *m230_addr[];
int	m230_present;		/* how many mc68230 modules are found */
short   m230_timflg;	    /* permits timer to start in first open */


m230init()
{
	register struct lp *lp;
	register struct m230_device *lpdev;
	register i;
	extern	m230start();

#ifdef HOWFAR
	printf ("m230init: entering\n");
#endif

	if (m230_cnt > LPMAX) {
		printf ("m230init: Number of devices ");
		printf ("%d reduced to %d\n", m230_cnt, LPMAX);
		m230_cnt = LPMAX;
	}

	/* number of m68230 modules found */
	m230_present = 0;

	/* initialize MC68230 on all m68230 */
	for (i = 0; i < m230_cnt; i++) {

		lpdev = m230_addr[i];
		lp = &m230lp_dt[i];

		if ( !bprobe (&lpdev->pivr, 0) ) {	/* is device missing */
			/* number of m68230 modules found */
			m230_present++;

			lp->lp_routine = m230start;
			if (m230_ilev[i] > LPMAXILEV) {
				printf("m230init: warning, lowered ilev to %d\n",LPMAXILEV );
				m230_ilev[i] = LPMAXILEV;
			}
			lp->lev = 0x2000 | (m230_ilev[i] << 8);
			lp->dev = i;

			lpdev->pgcr  = GCRDHL;	/* disable handshake lines */
			lpdev->pAddr = ADDRSO;	/* set port data lines output */
			lpdev->pCddr = CDDRSPO;	/* set pc4 (prime*) to output */
			lpdev->pCdr  = CDRINIT;	/* set to initial value */
			lpdev->psrr  = SRRVECI;	/* set to vectored interrupts */
			lpdev->pAcr  = ACRIBO;	/* intr if both buf are empty */
			lpdev->pgcr  = GCRH1E;	/* mode 0, h1 enable */
			lpdev->pivr  = (unsigned) m230_ivec[i]>>2;
			lpdev->enbl  = ENBL;	/* access printer enable */
		}	/* end if */
	}	/*end for */
}	/* end of m230init() */


m230open (dev, mode)
{
	register struct lp	*lp;
	register struct m230_device	*lpdev;

	dev = minor(dev);

#ifdef HOWFAR
	printf ("m230open: dev=%d\n", dev);
#endif
	if (dev >= LPMAX ) {	/* device number not in range */
		u.u_error = ENODEV;
		return;
	}

	if  ((lp = &m230lp_dt[dev])->flag&L_OPEN) {
#ifdef ERRMSG 
		printf ("m230open: printer already open\n");
#endif 
		u.u_error = EMFILE;
		return;
	}
#ifdef STATISTIC
	printf("\nm230open: old statistics:\n");
	m230stat(dev);
	m230clrstat(dev);
	printf("m230open: initialized statistics:\n");
	m230stat(dev);
#endif
#ifdef STATISTIC
	cnt_open[dev]++;
#endif

	if ( !lp->lp_routine ) {
#ifdef ERRMSG 
		printf ("m230open: lp_routine has not been set\n");
#endif 
		u.u_error = EIO;
		return;
	}

	/* start timer, if open is called first time */
	if ( m230_timflg == 0 ) {
		m230_timflg++;
		m230timer();
	}

	lpdev = m230_addr[dev];

	/* the mc68230 register pCdr (offset 99) should look like
	 * pCdr = FC (hex) = 1111 1100 (bin)
	 *			   |||
	 *			   ||^--- 1 : printer busy
	 *			   |^---- 1 : printer fault or paper out
	 *			   ^----- 0 : printer not selected
	 * unfortunatly different printer are handling these centronic lines
	 * different. So we cannot do a good analysis what happened to the
	 * printer.
	 */
	/*
	 * Don't do this check here, wait for start.
	 * Otherwise printer will be busy if two requests are done close
	 * to each other and the second request will fail.
	 *
	 * if ( ((lpdev->pCdr)&(LPSEL|LPFAULT|LPBUSY)) != LPOK) {
	 *	printf("m230open: printer powered off, not selected or busy\n");
	 *	u.u_error = EIO;
 	 *	return;
	 *}
	 */
	lp->flag |= L_OPEN;
	lp_open (lp);
	lp_output (lp,'\f');		/* send the usual formfeed */

}	/* end of m230open() */


m230close (dev)
{
	register struct lp	*lp;
	register struct m230_device	*lpdev;

	dev = minor(dev);
	lp = &m230lp_dt[dev];
#ifdef HOWFAR
	printf ("m230close: dev=%d\n", dev);
#endif
#ifdef STATISTIC
	cnt_close[dev]++;
#endif
	lp_output (lp,'\f');

#ifdef STATISTIC
	printf("m230open: result statistics \n");
	m230stat(dev);
#endif
	lp->flag &= ~L_FAULT;	/* Clear the faullt flag */
	lp->flag &= ~L_OPEN;
}


m230write (dev)
{
	register struct lp	*lp;
	register struct m230_device	*lpdev;

	dev = minor(dev);

#ifdef HOWFAR
	printf ("m230write: dev=%d\n", dev);
#endif
#ifdef STATISTIC
	cnt_write[dev]++;
#endif
	lp = &m230lp_dt[dev];
	lpdev = m230_addr[dev];
	lp_write (lp);
	lpdev->pAcr |= ACRDH1I;	/* enable h1 interrupts */
}


m230intr (dev)
{
	register struct lp *lp = &m230lp_dt[dev];
	register x;

#ifdef HOWFAR 
	printf ("m230intr: dev = %d \n",dev);
#endif
#ifdef STATISTIC
	cnt_intr[dev]++;
#endif

	/* if data register is empty, we are getting an interrupt,
		so the printer is now not more busy */

	x = spl6();		/* lock out timer */

	lp->flag &= ~(L_BUSY | L_TOUT);
	splx(x);

	/* process interrupt */
	m230start(lp);

	/* character count in queue is low and asleep, wakeup process */
	if (lp->l_outq.c_cc <= LPLOWAT && lp->flag & L_ASLP) {

		lp->flag &= ~L_ASLP;
#ifdef STATISTIC
		cnt_wakeup[dev]++;
#endif
		wakeup (lp);

	}
}	/* end of m230intr() */


m230ioctl (dev, cmd, arg, mode)
{
	register struct lp	*lp;

	dev = minor(dev);

#ifdef HOWFAR 
	printf ("m230ioctl: dev=%d, cmd=%d, arg=0x%x, mode=0x%x\n",
	    dev, cmd, arg, mode);
#endif
#ifdef STATISTIC
	cnt_ioctl[dev]++;
#endif
	lp = &m230lp_dt[dev];
	lp_ioctl (cmd, arg, mode, lp);
}


m230start (lp)
register struct lp	*lp;
{
	register struct m230_device	*lpdev;
	register char c;

#ifdef HOWFAR
	printf ("m230start: entering\n");
#endif
#ifdef STATISTIC
	cnt_start[lp->dev]++;
#endif

	/* if start is not called from m230intr and character transfer
		is in process, return (software check) */
	if(lp->flag & L_BUSY) {
		return(0);
	}

	lpdev = m230_addr[lp->dev];

	/* if status register says character transfer in process 
		(hardware check). this should never happen */
	if((lpdev->psr & 1) == 0) {
		printf("m230start: psr = 0, unexpected status\n");
		return(0);
	}
	/* the mc68230 register pCdr (offset 99) should look like
	 * pCdr = FC (hex) = 1111 1100 (bin)
	 *			   |||
	 *			   ||^--- 1 : printer busy
	 *			   |^---- 1 : printer fault or paper out
	 *			   ^----- 0 : printer not selected
	 * unfortunatly different printer are handling these centronic lines
	 * different. So we cannot do a good analysis what happened to the
	 * printer.
	 * 
	 * Observed the following on Dataproducts B600-9:
	 * 
	 *	pCdr		printer status
	 * 
	 *	FC		cable attached to 335, switch on, on line
	 *	F9		cable attached to 335, switch on, off line
	 *	FB		cable attached to 335, switch on, out of paper 
	 *	FA		cable attached to 335, switch off
	 *	FF		cable detached from 335 
	 *
	 * The problem is, once you return an error to the calling routine,
         * the printer gets disabled by the system, and must be re-enabled
   	 * in order to print again.
	 * We certainly don't want this to happen if we run out of paper,
	 * for example.
	 * Since I can't tell for sure why the printer can't print, I
	 * will let the timer routine wake me up periodically forever.
	 */
	if ( ((lpdev->pCdr)&(LPSEL|LPFAULT|LPBUSY)) != LPOK) {
		if ( lp->flag & L_FAULT) {
			lpdev->pAcr &= ~ACRDH1I; /* disable interrupts */
			lp->flag |= L_BUSY;	/* wait for return to online */
			return (0);
		}
		else {
			lp->flag |= L_FAULT;

#ifdef ERRMSG
			printf("m230start: printer powered off, not selected or busy\n");
#endif
			return(0);
		}
	}
	lp->flag &= ~L_FAULT;

	if ((c = getc (&lp->l_outq)) < 0) {

		lpdev->pAcr &= ~ACRDH1I;	/* disable h1 interrupt */

	} else {	 /* write the character if present */

		lp->flag = lp->flag | L_BUSY;
		lpdev->pAdr = c;	/* write c to data register */

#ifdef STATISTIC
		cnt_transfer[lp->dev]++;
#endif
		lpdev->pAcr |= ACRDH1I;	/* enable h1 interrupts */
	}

	return(0);
}



m230timer()
{
	register struct lp *lp;
	register struct m230_device *lpdev;
	register dev;

	for (lp = &m230lp_dt[0], dev=0; dev<m230_cnt; dev++, lp++) {
#ifdef DEBUG
		printf("m230timer: timer flag = %x\n",lp->flag);
#endif
		/* L_BUSY on iff waiting for an interrupt */
		if ((lp->flag&L_BUSY) == 0)
			continue;
		/* L_TOUT on iff no interrupt occurred in last 10 secs */
		if ((lp->flag&L_TOUT) == 0) {
			lp->flag |= L_TOUT;
			continue;
		}
		/* no interrupt in last 60 seconds */
		lpdev = m230_addr[dev];

		/* fake an interrupt condition */
		lp->flag &= ~(L_BUSY|L_TOUT);
		m230start(lp);		/* continue */
		/* wakeup any waiting processes */
		if (lp->l_outq.c_cc <= LPLOWAT && lp->flag&L_ASLP) {
#ifdef DEBUG
			printf("m230timer: wakeup\n");
#endif
			lp->flag &= ~L_ASLP;
			wakeup(lp);
		}
	}

	timeout(m230timer,0,10*HZ);
}


#ifdef STATISTIC
m230stat(dev)
int dev;
{

	printf("230#  start fubyte sent intr open close sleep wake\n");
	printf("%d "    ,dev);
	printf("%d "    ,cnt_start[dev]);
	printf("%d "    ,cnt_fubyte[dev]);
	printf("%d "    ,cnt_transfer[dev]);
	printf("%d "    ,cnt_intr[dev]);
	printf("%d "    ,cnt_open[dev]);
	printf("%d "    ,cnt_close[dev]);
	printf("%d "    ,cnt_sleep[dev]);
	printf("%d \n\n",cnt_wakeup[dev]);
}

m230clrstat(dev)	/* initialise statistic variables */
int dev;
{
	cnt_intr [dev]     = 0;
	cnt_open [dev]     = 0;
	cnt_close [dev]    = 0;
	cnt_write [dev]    = 0;
	cnt_ioctl [dev]    = 0;
	cnt_start [dev]    = 0;
	cnt_transfer [dev] = 0;
	cnt_sleep [dev]    = 0;
	cnt_wakeup [dev]   = 0;
	cnt_fubyte [dev]   = 0;
}
#endif
