/*
 *	Copyright (c) 1984 AT&T
 *	Copyright (c) 1987 Motorola Inc
 *	All Rights Reserved
 */

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)kern-port:io/m68681.c	35.8"

/*********************************************************************
 *
 *  Title:	  MC68681  Terminal Driver
 *
 *  Description:  This is a UNIX SYSTEM V/68 serial character I/O 
 *	          driver. 
 *
 *		  The driver communicates with the MC68681 DUARTs 
 *
 *
 *
 *********************************************************************/

/*********************** COMPILER SWITCHES ***************************/

/* #define HOWFAR		printing debug messages  */
/* #define STATISTICS		do some statistics */
/* #define M136DETECT		when 136 board is different from 335 */
/* M136DETECT was an attempt to find out which board we were on by */
/* reading a register that is not defined on the 335. Now that we are */
/* using two drivers instead of one, we don't need it. */



/*********************** INCLUDES ************************************/

#include "sys/param.h"
#include "sys/types.h"
#include "sys/dir.h"
#include "sys/file.h"
#include "sys/signal.h"
#include "sys/tty.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/sysinfo.h"
#include "sys/sysmacros.h"
#include "sys/gentty.h"
#include "sys/mvmecpu.h"

/*
    XREGS is defined for the 135, 136, 140 cpu boards and
    NOT defined for the 335 card.
 */

#ifdef	XREGS
#define	m681ckdcd	x681ckdcd
#define	m681_sv_reg	x681_sv_reg
#define	m681_openduarts	x681_openduarts
#define	m681_speeds	x681_speeds
#define	m681_scanflg	x681_scanflg
#define	m681_work	x681_work
#define	m681init	x681init
#define	m681_cnt	x681_cnt
#define	m681_addr	x681_addr
#define	m681_ivec	x681_ivec
#define	m681open	x681open
#define	m681scan	x681scan
#define	m681_tty	x681_tty
#define	m681proc	x681proc
#define	m681param	x681param
#define	m681modem	x681modem
#define	m681close	x681close
#define	m681read	x681read
#define	m681write	x681write
#define	m681ioctl	x681ioctl
#define	m681putc	x681putc
#define	m681intr	x681intr
#define	m681sendchar	x681sendchar
#define	m681limit	x681limit
#define m681wclose	x681wclose
#define m681break	x681break
#define m681_state 	x681_state
#define m681inputc	x681inputc
#endif

/*********************** CONFIGURATION *******************************/
#define MAXBOARDS	6		/* max count of BOARDS */
#define MAXDUARTS	(MAXBOARDS*2)	/* max count of DUARTS */
#define MAXPORTS	(MAXDUARTS*2)	/* max count of PORTS  */

/*********************** DEFINITIONS *********************************/
#define	MISSING	0	/* flag, missing a configured DUART */
#define ON	1	/* flag for modem() */
#define OFF	0	/* flag for modem() */
#define A_PORT  0	/* flag for port a */
#define B_PORT  1	/* flag for port b */


/*********************** TYP DEFINTIONS *****************************/
typedef enum {
	FALSE,
	TRUE
} 
BOOL;


/*********************** MACROS ***********************************/
#define SPLCLOCK   spl6		/* set clock priority level */
#define SPLM681    spl5		/* set m681 driver priority level */

/* the access to the same command register must be delayed,
   because the chip has some hardware problems in this case */
#define DELAY_CR   { register i; for ( i=0; i<250; ++i ); }


/*********************** MC68681 DEFINITIONS ************************/
#define PORTOFFSET	0x10	/* offset for port b address space */

/* mode register 1: MR1x operations */
#define	RXRTS	0x80	/* enable receiver RTS */
#define PAREN	0x00	/* with parity */
#define PARDIS	0x10	/* no parity */
#define EVENPAR	0x00	/* even parity */
#define ODDPAR	0x04	/* odd parity */
#define CL5	0x00	/* 5 bits per char */
#define CL6	0x01	/* 6 bits per char */
#define CL7	0x02	/* 7 bits per char */
#define CL8	0x03	/* 8 bits per char */
#define PARMODEMASK 0x18  /* parity mode mask */
#define PARTYPEMASK 0x04  /* parity type mask */
#define CLMASK	    0x03  /* character length mask */

/* mode register 2: MR2x operations */
#define TXRTS	0x20	/* enable transmitter RTS */
#define TXCTS	0x10	/* enable transmitter CTS */
#define SB2	0x0f	/* 2 stop bits */
#define SB1	0x07	/* 1 stop bit */
#define SB1L5	0x00	/* 1 stop bit at 5 bits per character */

#define SBMASK	0x0f	/* stop bit mask */

/* clock-select register: CSRx operations */
#define NOBAUD	-1	/* 50 and 200 baud are not possible */
/* they are not in Baud register set 2 */
#define BD75	0x00	/* 75 baud */
#define BD110	0x11	/* 110 baud */
#define BD134	0x22	/* 134.5 baud */
#define BD150	0x33	/* 150 baud */
#define BD300	0x44	/* 300 baud */
#define BD600	0x55	/* 600 baud */
#define BD1200	0x66	/* 1200 baud */
#define BD1800	0xaa	/* 1800 baud */
#define BD2400	0x88	/* 2400 baud */
#define BD4800	0x99	/* 4800 baud */
#define BD9600	0xbb	/* 9600 baud */
#define BD19200	0xcc	/* 19200 baud */

#define DEFBAUD BD9600	/* default value if baudrate is not possible */


/* channel command register: CRx operations */
#define MRRESET 0x10	/* reset mr pointer to mr1 */
#define RXRESET 0x20	/* reset receiver */
#define TXRESET 0x30	/* reset transmitter */
#define ERRRESET 0x40	/* reset error status */
#define BRKINTRESET 0x50 /* reset channel's break interrupt */
#define BRKSTART 0x60	/* start break */
#define BRKSTOP 0x70	/* stop break */
#define	TXDIS	0x08	/* disable transmitter */
#define	TXEN	0x04	/* enable transmitter */
#define	RXDIS	0x02	/* disable receiver */
#define	RXEN	0x01	/* enable receiver */

/* status register: SRx status */
#define RBRK	0x80	/* received break */
#define FRERR	0x40	/* frame error */
#define PERR	0x20	/* parity error */
#define ROVRN	0x10	/* receiver overrun error */
#define TXEMT	0x08	/* transmitter empty */
#define TXRDY	0x04	/* transmitter ready */
#define FFULL	0x02	/* receiver FIFO full */
#define RXRDY	0x01	/* receiver ready */

/* output port configuration  register: OPCR operations */
#define	OPSET	0x00	/* set all op lines to op function */

/* output port register: OP operations */
#define OPDTRB	0x08	/* DTR line output b */
#define OPDTRA	0x04	/* DTR line output a */
#define OPRTSB	0x02	/* RTS line output b */
#define OPRTSA	0x01	/* RTS line output a */

/* auxiliary control register: ACR operations */
#define BDSET1	0x00	/* baudrate generator set 1 */
#define BDSET2	0x80	/* baudrate generator set 2 */
#define	CCLK1	0x60	/* timer clock: external rate.  TA */
#define CCLK16	0x30	/* counter clock: x1 clk divided by 16 */
#define	SLCTIM	30720	/* Timer count to get 60 Hz time slice */
/*#define	SLCTIM	15360	/* Timer count to get 60 Hz time slice */

/* input port register: IP operations */
#define IPDCDIB 0x08	/* IP3 change == DCD input on port B of mvme136 */
#define IPDCDIA 0x04	/* IP2 change == DCD input on port A of mvme136 */

/* Defines for mvme335 */
#define IPDCDB	0x20 	/* DCD line input b */
#define IPDCDA	0x10  	/* DCD line input a */


#define IPDSRB  0x08	/* DSR line input b */
#define IPDSRA  0x04	/* DSR line input a */
#define IPCTSB  0x02	/* CTS line input b */
#define IPCTSA  0x01	/* CTS line input a */

/* interrupt status and mask register: ISR status and IMR mask */
#define IIPCHG	0x80	/* input port change */
#define IBRKB	0x40	/* delta break b */
#define IRXRDYB	0x20	/* receiver ready b */
#define ITXRDYB	0x10	/* transmitter ready b */
#define	ITIMER	0x08	/* Enable timer interrupts. */
#define IBRKA	0x04	/* delta break a */
#define IRXRDYA	0x02	/* receiver ready a */
#define ITXRDYA	0x01	/* transmitter ready a */


/* interrupts from port a or b */
#define AINTPORT	( IRXRDYA | ITXRDYA )
#define BINTPORT	( IRXRDYB | ITXRDYB )


/* HW write register index for m681_wr_regs[] */
#define MR1A	0	/* mode register 1 a */
#define CSRA    1	/* clock-select register a*/
#define CRA	2	/* command register a */
#define TBA	3	/* transmitter buffer a */
#define ACR	4	/* auxialiary control register*/
#define IMR	5	/* interrupt mask register */
#define CTUR	6	/* counter/timer upper reg */
#define	CTLR	7 	/* counter/timer lower reg */
#define MR1B	8	/* mode register 1 b */
#define CSRB	9	/* clock-select register b*/
#define CRB     10	/* command register b */
#define TBB	11	/* transmitter buffer b */
#define IVR	12	/* interrupt vector register */
#define OPCR	13	/* output port config reg */
#define OPRSET  14	/* output port: bit set cmd */
#define OPRRESET 15	/* output port: bit reset cmd */
#define MR2A	16	/* mode register 2 a */
#define MR2B	17	/* mode register 2 b */
#define MAXREG	18	/* max count of registers */

#ifdef	XREGS

/*
 *	Structure of MC68681 hardware read registers.
 */
struct m681_rd_reg
{
	unsigned char	rd_mra;		/* mode register a */
	unsigned char	rd_sra;		/* status register a*/
	unsigned char	dummy5;		/* do not access */
	unsigned char	rd_rba;		/* receiver buffer a */
	unsigned char	rd_ipcr;	/* input port change register */
	unsigned char	rd_isr;		/* interrupt status register */
	unsigned char	rd_cur;		/* current MSB of counter */
	unsigned char	rd_clr;		/* current LSB of counter */
	unsigned char	rd_mrb;		/* mode register b */
	unsigned char	rd_srb;		/* status register b*/
	unsigned char	dummyxb;	/* do not access */
	unsigned char	rd_rbb;		/* receiver buffer b */
	unsigned char	rd_ivr;		/* interrupt vector register */
	unsigned char	rd_ip;		/* input port (unlatched) */
	unsigned char	rd_ctstart;	/* start-counter command */
	unsigned char	rd_ctstop;	/* stop-counter command */
};

/*
 *	Structure of MC68681 hardware write registers.
 */
struct m681_wr_reg
{
	unsigned char	wr_mra;		/* mode register a */
	unsigned char	wr_csra;	/* clock-select register a*/
	unsigned char	wr_cra;		/* command register a */
	unsigned char	wr_tba;		/* transmitter buffer a */
	unsigned char	wr_acr;		/* auxialiary control register*/
	unsigned char	wr_imr;		/* interrupt mask register */
	unsigned char	wr_ctur;	/* counter/timer upper reg */
	unsigned char	wr_ctlr;	/* counter/timer lower reg */
	unsigned char	wr_mrb;		/* mode register b */
	unsigned char	wr_csrb;	/* clock-select register b*/
	unsigned char	wr_crb;		/* command register b */
	unsigned char	wr_tbb;		/* transmitter buffer b */
	unsigned char	wr_ivr;		/* interrupt vector register */
	unsigned char	wr_opcr;	/* output port config reg */
	unsigned char	wr_oprset;	/* output port: bit set cmd */
	unsigned char	wr_oprreset;	/* output port: bit reset cmd */
};

#else

#if defined (ASPEN) || defined (ANGEL)


/*
 *	Structure of MC68681 hardware read registers.
 */
struct m681_rd_reg
{
	unsigned int	rd_mra;		/* mode register a */
	unsigned int	rd_sra;		/* status register a*/
	unsigned int	dummy5;		/* do not access */
	unsigned int	rd_rba;		/* receiver buffer a */
	unsigned int	rd_ipcr;	/* input port change register */
	unsigned int	rd_isr;		/* interrupt status register */
	unsigned int	rd_cur;		/* current MSB of counter */
	unsigned int	rd_clr;		/* current LSB of counter */
	unsigned int	rd_mrb;		/* mode register b */
	unsigned int	rd_srb;		/* status register b*/
	unsigned int	dummyxb;	/* do not access */
	unsigned int	rd_rbb;		/* receiver buffer b */
	unsigned int	rd_ivr;		/* interrupt vector register */
	unsigned int	rd_ip;		/* input port (unlatched) */
	unsigned int	rd_ctstart;	/* start-counter command */
	unsigned int	rd_ctstop;	/* stop-counter command */
};

/*
 *	Structure of MC68681 hardware write registers.
 */
struct m681_wr_reg
{
	unsigned int	wr_mra;		/* mode register a */
	unsigned int	wr_csra;	/* clock-select register a*/
	unsigned int	wr_cra;		/* command register a */
	unsigned int	wr_tba;		/* transmitter buffer a */
	unsigned int	wr_acr;		/* auxialiary control register*/
	unsigned int	wr_imr;		/* interrupt mask register */
	unsigned int	wr_ctur;	/* counter/timer upper reg */
	unsigned int	wr_ctlr;	/* counter/timer lower reg */
	unsigned int	wr_mrb;		/* mode register b */
	unsigned int	wr_csrb;	/* clock-select register b*/
	unsigned int	wr_crb;		/* command register b */
	unsigned int	wr_tbb;		/* transmitter buffer b */
	unsigned int	wr_ivr;		/* interrupt vector register */
	unsigned int	wr_opcr;	/* output port config reg */
	unsigned int	wr_oprset;	/* output port: bit set cmd */
	unsigned int	wr_oprreset;	/* output port: bit reset cmd */
};
#else
/*
 *	Structure of MC68681 hardware read registers.
 */
struct m681_rd_reg
{
	unsigned char	dummy0;		/* start at odd address */
	unsigned char	rd_mra;		/* mode register a */
	unsigned char	dummy2;
	unsigned char	rd_sra;		/* status register a*/
	unsigned char	dummy4;
	unsigned char	dummy5;		/* do not access */
	unsigned char	dummy6;
	unsigned char	rd_rba;		/* receiver buffer a */
	unsigned char	dummy8;
	unsigned char	rd_ipcr;	/* input port change register */
	unsigned char	dummy10;
	unsigned char	rd_isr;		/* interrupt status register */
	unsigned char	dummy12;
	unsigned char	rd_cur;		/* current MSB of counter */
	unsigned char	dummy14;
	unsigned char	rd_clr;		/* current LSB of counter */
	unsigned char	dummy16;
	unsigned char	rd_mrb;		/* mode register b */
	unsigned char	dummy18;
	unsigned char	rd_srb;		/* status register b*/
	unsigned char	dummy20;
	unsigned char	dummyxb;	/* do not access */
	unsigned char	dummy22;
	unsigned char	rd_rbb;		/* receiver buffer b */
	unsigned char	dummy24;
	unsigned char	rd_ivr;		/* interrupt vector register */
	unsigned char	dummy26;
	unsigned char	rd_ip;		/* input port (unlatched) */
	unsigned char	dummy28;
	unsigned char	rd_ctstart;	/* start-counter command */
	unsigned char	dummy30;
	unsigned char	rd_ctstop;	/* stop-counter command */
};

/*
 *	Structure of MC68681 hardware write registers.
 */
struct m681_wr_reg
{
	unsigned char	dummy0;		/* start at odd address */
	unsigned char	wr_mra;		/* mode register a */
	unsigned char	dummy2;
	unsigned char	wr_csra;	/* clock-select register a*/
	unsigned char	dummy4;
	unsigned char	wr_cra;		/* command register a */
	unsigned char	dummy6;
	unsigned char	wr_tba;		/* transmitter buffer a */
	unsigned char	dummy8;
	unsigned char	wr_acr;		/* auxialiary control register*/
	unsigned char	dummy10;
	unsigned char	wr_imr;		/* interrupt mask register */
	unsigned char	dummy12;
	unsigned char	wr_ctur;	/* counter/timer upper reg */
	unsigned char	dummy14;
	unsigned char	wr_ctlr;	/* counter/timer lower reg */
	unsigned char	dummy16;
	unsigned char	wr_mrb;		/* mode register b */
	unsigned char	dummy18;
	unsigned char	wr_csrb;	/* clock-select register b*/
	unsigned char	dummy20;
	unsigned char	wr_crb;		/* command register b */
	unsigned char	dummy22;
	unsigned char	wr_tbb;		/* transmitter buffer b */
	unsigned char	dummy24;
	unsigned char	wr_ivr;		/* interrupt vector register */
	unsigned char	dummy26;
	unsigned char	wr_opcr;	/* output port config reg */
	unsigned char	dummy28;
	unsigned char	wr_oprset;	/* output port: bit set cmd */
	unsigned char	dummy30;
	unsigned char	wr_oprreset;	/* output port: bit reset cmd */
};

#endif
#endif

union m681_io {
	struct m681_rd_reg	read;		/* read registers */
	struct m681_wr_reg	write;		/* write registers */
};

#ifdef	XREGS


/*
 *	Structure of MC68681 hardware port dependent read registers.
 */
struct m681_prd_reg
{
	unsigned char	rd_mr;		/* mode register */
	unsigned char	rd_sr;		/* status register */
	unsigned char	dummy5;
	unsigned char	rd_rb;		/* receiver buffer */
	unsigned char	dummy9;
	unsigned char	dummy11;
	unsigned char	dummy13;
	unsigned char	dummy15;
};

/*
 *	Structure of MC68681 hardware port dependent write registers.
 */
struct m681_pwr_reg
{
	unsigned char	wr_mr;		/* mode register */
	unsigned char	wr_csr;		/* clock-select register */
	unsigned char	wr_cr;		/* command register */
	unsigned char	wr_tb;		/* transmitter buffer */
	unsigned char	dummy9;
	unsigned char	dummy11;
	unsigned char	dummy13;
	unsigned char	dummy15;
};

#else

#if defined (ASPEN) || defined (ANGEL)


/*
 *	Structure of MC68681 hardware port dependent read registers.
 */
struct m681_prd_reg
{
	unsigned int	rd_mr;		/* mode register */
	unsigned int	rd_sr;		/* status register */
	unsigned int	dummy5;
	unsigned int	rd_rb;		/* receiver buffer */
	unsigned int	dummy9;
	unsigned int	dummy11;
	unsigned int	dummy13;
	unsigned int	dummy15;
};

/*
 *	Structure of MC68681 hardware port dependent write registers.
 */
struct m681_pwr_reg
{
	unsigned int	wr_mr;		/* mode register */
	unsigned int	wr_csr;		/* clock-select register */
	unsigned int	wr_cr;		/* command register */
	unsigned int	wr_tb;		/* transmitter buffer */
	unsigned int	dummy9;
	unsigned int	dummy11;
	unsigned int	dummy13;
	unsigned int	dummy15;
};

#else
/*
 *	Structure of MC68681 hardware port dependent read registers.
 */
struct m681_prd_reg
{
	unsigned char	dummy0;		/* start at odd address */
	unsigned char	rd_mr;		/* mode register */
	unsigned char	dummy2;
	unsigned char	rd_sr;		/* status register */
	unsigned char	dummy4;
	unsigned char	dummy5;
	unsigned char	dummy6;
	unsigned char	rd_rb;		/* receiver buffer */
	unsigned char	dummy8;
	unsigned char	dummy9;
	unsigned char	dummy10;
	unsigned char	dummy11;
	unsigned char	dummy12;
	unsigned char	dummy13;
	unsigned char	dummy14;
	unsigned char	dummy15;
};

/*
 *	Structure of MC68681 hardware port dependent write registers.
 */
struct m681_pwr_reg
{
	unsigned char	dummy0;		/* start at odd address */
	unsigned char	wr_mr;		/* mode register */
	unsigned char	dummy2;
	unsigned char	wr_csr;		/* clock-select register */
	unsigned char	dummy4;
	unsigned char	wr_cr;		/* command register */
	unsigned char	dummy6;
	unsigned char	wr_tb;		/* transmitter buffer */
	unsigned char	dummy8;
	unsigned char	dummy9;
	unsigned char	dummy10;
	unsigned char	dummy11;
	unsigned char	dummy12;
	unsigned char	dummy13;
	unsigned char	dummy14;
	unsigned char	dummy15;
};

#endif
#endif

union m681_pt_io {
	struct m681_prd_reg	read;		/* read registers */
	struct m681_pwr_reg	write;		/* write registers */
};

/*
 *	Structure of saved registers
 */
struct m681_sv_reg
{
	unsigned char	sv_mr1[2];	/* mode register 1 a */
	unsigned char	sv_mr2[2];	/* mode register 2 a */
	unsigned char	sv_csr[2];	/* clock-select register a*/
	unsigned char	sv_cr[2];		/* command register a */
	unsigned char	sv_acr;		/* auxialiary control register*/
	unsigned char	sv_imr;		/* interrupt mask register */
	unsigned char	sv_ivr;		/* interrupt vector register */
};

/* defines for receiver and DCD scanner */
#define M681CSCAN 2	/* clock ticks for character scan */
#define M681_CHAR 01	/* character arrived flag (m681_work) */
#define M681_ACT  02	/* character in buffer flag (m681_work) */
#define M681_DCD  04	/* DCD wait flag (m681_work) */



/********************* SPECIAL IOCTL COMMAND DEFINITIONS ******************/
/* HWHAND ioctls same as in MVME331 driver */
#define MSETHWHAND	(('m'<<8)+20)	/* set hardware handshake */
#define MCLEARHWHAND	(('m'<<8)+21)	/* clear hardware handshake */
#define MGETHWHAND	(('m'<<8)+22)	/* get hardware handshake mode */

#define M681STATUS	(('m'<<8)+40 )  /* get status of duarts */
#define M681GETSTAT	(('m'<<8)+42 )  /* get statistics */

struct debugioctl {
	char *buf;	/* debug buffer start */
	int  size;	/* size of debug buffer */
	char *pt;	/* current pointer in debug buffer */
};

#ifdef STATISTICS
struct m681_stat {
	int open_cnt;	/* count of open calls */
	int intr_cnt;	/* count of interrupts */
	int tx_cnt;	/* count of transmitted characters */
	int rx_cnt;	/* count of received characters */
	int brk_cnt;	/* count of break events */
	int rovrn_cnt;	/* count of overruns */
	int frerr_cnt;	/* count of frame errors */
	int perr_cnt;	/* count of parity errors */
	int rxoff_cnt;	/* count of received XOFF chars */
	int rxon_cnt;	/* count of received XON  chars */
	int txoff_cnt;	/* count of transmitted XOFF chars */
	int txon_cnt;	/* count of transmitted XON  chars */
};
#endif


/****************** DEFINITIONS FOR DEBUG PRINTING *****************/
#ifdef HOWFAR

/* DEBUG levels */
#define LOG_START	10	/* start a subroutine */
#define LOG_RETURN	10	/* return form a subroutine */
#define	LOG_WARN	 5	/* print a warning message */
#define LOG_INTERN	20	/* debug messages in a function */
#define LOG_MAX		20	/* highest debug level */

#define LOG( level, message ) 			\
    if ( (m681_debug&0xff) >= level ) {		\
	    printf message ;	 		\
    }
#else
#define LOG( level, message ) {} /* do nothing */
#endif


/*********************** EXTERNALS ***********************************/
/* Declarations in conf.c */
extern int cpuboard;		/* current processor board running */
extern int m681_cnt;		/* count of serial ports */
extern int m681_ivec[];		/* interrupt vectors */
extern union m681_io *m681_addr[MAXDUARTS];	/* address table */
extern struct tty m681_tty[];	/* tty structures */
extern  unsigned int  consdev;		/* console device */
extern  unsigned int  dualcons;		/* alternate console device */
extern 	int	Dconcurrent;		/* dual console mode enable */

/*********************** DECLARATIONS ********************************/

/* flags which should be in tty structure (t_state) but are not because
   the tty structure must not be enlarged. */
short	m681_state[MAXPORTS];

/* saved write register to initialize a port */
struct m681_sv_reg m681_sv_reg[MAXDUARTS];

BOOL   m681_openduarts[MAXDUARTS];	/* opened DUART flags */

#ifdef M136DETECT
/* saved board type -- either 335 or 136 */
int m681_btype[MAXDUARTS];
#define M136 136
#define M335 335
#endif

/* wait for close to complete before opening */
int m681wclose[MAXPORTS] = {0};

/* do a stop break only if this flag is set */
/* otherwise don't mess with the 681 */
int m681break[MAXPORTS] = {0};

/* speed tables */
int m681_speeds[] = {
	0,		/* 0 baud, special HUP condition */
	NOBAUD,		/* 50 baud, not implemented */
	BD75,		/* 75 baud */
	BD110,		/* 110 baud */
	BD134,		/* 134.5 baud */
	BD150,		/* 150 baud */
	NOBAUD,		/* 200 baud, not implemented */
	BD300,		/* 300 baud */
	BD600,		/* 600 baud */
	BD1200,		/* 1200 baud */
	BD1800,		/* 1800 baud */
	BD2400,		/* 2400 baud */
	BD4800,		/* 4800 baud */
	BD9600,		/* 9600 baud */
	BD19200,	/* 19200 baud */
	BD19200 	/* 19200 baud */
};

/* declarations for receiver and DCD scanner */
short   m681_scanflg;	    /* permits scan to start in first open */
char	m681_work[MAXPORTS];	/* activity info for scanner */

int	m681limit = CLSIZE - 0;	/* dump buffer more often in systems which */
				/* cannot do canonical processing of 64 chars */
				/* without overrun			*/


#ifdef HOWFAR
/* declarations for debug prints */
int m681_debug = LOG_WARN;	/* debug flag */
#endif


#ifdef STATISTICS
/* statistic data */
struct m681_stat m681_stats[MAXPORTS];
#endif


/*********************************************************************
 *
 *  Function:	  m681init
 *
 *  Description:  Initialization of the driver
 *
 *
 *  Syntax:	  m681init()
 *
 *  Input:	  -
 *
 *  Output:       -
 *
 *
 *  Called by:	  initializer routine of system
 *
 *  Calls to:     bprobe()
 *		  SPLM681()
 *		  splx()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681init()
 *
 *  BEGIN
 *
 *	block interrupts for this driver [ SPLM681() ];
 *
 *	FOR  all configured DUARTs
 *	DO
 *	    IF  status register is not readable [ bprobe() ];
 *	    THEN
 *		set missing state of this DUART;
 *	    ELSE
 *		save standard initialization of registers;
 *		reset port a & b;
 *		initialize port a & b;
 *		initialize common registers of the ports;
 *	    ENDIF
 *	ENDFOR
 *		
 *	free interrupts for this driver [ splx() ];
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/

m681init()
{
	register union m681_io *addr;       /* pointer to DUART regs */
	register union m681_pt_io *ptaddr;  /* pointer to port regs */
	int i, j;				/* loop counter */
	int port;				/* port index */
	int cur_lvl;			/* current interrupt level */

	LOG( LOG_WARN, ( "M681_DRIVER: m681_debug=%d, addr=0x%x\n",
	    m681_debug, &m681_debug ) );

	LOG( LOG_START, ( "m681init: started\n" ) );

	/* block interrupts of this driver */
	cur_lvl = SPLM681();

	/* probe the DUARTs */
	for ( i=0; i<(m681_cnt/2); i++ ) {

		/* try to read the status register a */
		if ( bprobe( (char *) &(m681_addr[i]->read.rd_mra), -1 ) ) {

			/* DUART not found */
			LOG( LOG_WARN,
			    ( "m681_init: Warning: missing DUART %d  on board %d, addr=0x%x\n",
			    (i%2)+1, i/2,
			    (char *) &(m681_addr[i]->read.rd_mra) ) );
			m681_addr[i] = MISSING;

			continue;
		}
		LOG( LOG_INTERN,
		    ( "m681_init: DUART %d  on board %d at addr=0x%x found\n",
		    (i%2)+1, i/2,
		    (char *) &(m681_addr[i]->read.rd_mra) ) );

		/* DUART exists, init both ports */

		addr = m681_addr[i];
		ptaddr = ( union m681_pt_io *)addr;

#ifdef M136DETECT
		/* test to see if device is on mvme335 or mvme136 */
		if ( bprobe( (char *) &(m681_addr[i]->read.dummy4), -1 ) ) 
			m681_btype[i] = M335;
		else 
			m681_btype[i] = M136;
		LOG( LOG_WARN, ( "M681_DRIVER: board type=%d, addr=0x%x\n",
	    	m681_btype[i], addr ) );
#endif   

		/* save standard initialization */

		m681_sv_reg[i].sv_mr1[A_PORT] = PARDIS | CL8;
		m681_sv_reg[i].sv_mr2[A_PORT] = SB1;
		m681_sv_reg[i].sv_csr[A_PORT] = BD9600;
		m681_sv_reg[i].sv_cr[A_PORT]  = TXEN | RXEN;

		m681_sv_reg[i].sv_mr1[B_PORT] = PARDIS | CL8;
		m681_sv_reg[i].sv_mr2[B_PORT] = SB1;
		m681_sv_reg[i].sv_csr[B_PORT] = BD9600;
		m681_sv_reg[i].sv_cr[B_PORT]  = TXEN | RXEN;

#if defined (ASPEN) || defined (ANGEL)
		m681_sv_reg[i].sv_acr  = BDSET2 | CCLK1;
#else
		m681_sv_reg[i].sv_acr  = BDSET2 | CCLK16 | IPDCDIB | IPDCDIA;
#endif
		/* Start out with Tx and RX interrupts disabled */
		/* Enable input port change interrupt */
		m681_sv_reg[i].sv_imr  = IIPCHG;

		/* reset port a */
		addr->write.wr_cra  = RXRESET     | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_cra  = TXRESET     | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_cra  = ERRRESET    | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_cra  = BRKINTRESET | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_cra  = MRRESET     | TXDIS | RXDIS;

		/* reset port b */
		addr->write.wr_crb  = RXRESET     | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_crb  = TXRESET     | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_crb  = ERRRESET    | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_crb  = BRKINTRESET | TXDIS | RXDIS;
		DELAY_CR;
		addr->write.wr_crb  = MRRESET     | TXDIS | RXDIS;
		DELAY_CR;


		/* initialize ports */
		port = A_PORT;		/* first loop is init A_PORT */
		for ( j=0; j<2; ++j ) {
			ptaddr->write.wr_mr  = m681_sv_reg[i].sv_mr1[port];
			ptaddr->write.wr_mr  = m681_sv_reg[i].sv_mr2[port];
			if (!Dconcurrent)
			    ptaddr->write.wr_csr = m681_sv_reg[i].sv_csr[port];
			ptaddr->write.wr_cr  = m681_sv_reg[i].sv_cr [port];
			port = B_PORT;	/* second loop is init B_PORT */
			ptaddr = ptaddr + 1;
		}


		/* initialize common register of a DUART */
		addr->write.wr_ctur  = SLCTIM>>8;
		addr->write.wr_ctlr  = SLCTIM & 0xFF;
		addr->write.wr_acr  = m681_sv_reg[i].sv_acr;
		addr->write.wr_imr  = m681_sv_reg[i].sv_imr;
		addr->write.wr_opcr = OPSET;
		addr->write.wr_ivr = (unsigned char)((m681_ivec[i])>>2);

		LOG( LOG_INTERN,
		    ( "m681init: initialized DUART %d  on board %d\n",
		    (i%2)+1, i/2 ) );
	}

	splx( cur_lvl );


	LOG( LOG_RETURN, ( "m681init: return okay\n" ) );
	return;
}


/*********************************************************************
 *
 *  Function:	  m681open
 *
 *  Description:  Open of the driver
 *
 *
 *  Syntax:	  m681open( dev, flag )
 *
 *  Input:	  int dev;	minor device number
 *		  int flag;	open flag
 *
 *  Output:       u.u_error = ENXIO	if invalid device
 *
 *
 *  Called by:	  Character device  open  switch
 *
 *  Calls to:     m681scan()
 *		  m681modem()
 *		  m681param()
 *		  ttinit()
 *		  line switch  l_open()
 *		  sleep()
 *		  SPLCLOCK()
 *		  splx()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681open()
 *
 *  BEGIN
 *
 *	IF  open function is called first time
 *	THEN
 *	    start scanner;
 *	ENDIF
 *
 *	IF  invalid device number
 *	THEN
 *	    set user error  to  "No such device or address" (ENXIO);
 *	    RETURN;
 *	ENDIF
 *
 *	IF  this line is not opened or not waiting for opened
 *	THEN
 *	    init tty structure [ ttinit() ];
 *	    set hardware parameters [ m681param() ];
 *	ENDIF
 *	
 *	IF  CLOCAL mode  OR  carrier available [ m681modem() ]
 *	THEN
 *	    set carrier-on state in tty structure;
 *	ELSE
 *	    clear carrier-on state in tty structure;
 *	    IF  open flag  is not FNDELAY
 *	    THEN
 *		set DCD check flag in scan list;
 *		sleep until carrier is available;
 *	    ENDIF
 *	ENDIF
 *
 *	call open procedure in upper tty driver [ l_open() ];
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/


m681open( device, flag )
unsigned int device, flag;
{
	register struct tty *tp;	  /* pt to tty structure of this port */
	register union m681_io *addr; /* pointer to DUART regs */
	register unsigned int dev;
	int cur_lvl;		  /* current interrupt level */
	int m681proc();

	dev = minor(device);

	LOG( LOG_START, ( "m681open: started dev=%d, flag=0x%x\n",
	    dev, flag ) );

	/* start scanner, if open is called first time */
	if ( m681_scanflg == 0 ) {
		m681_scanflg++;
		m681scan();
	}


	if (dev >= m681_cnt) {		/* check for valid dev */
		u.u_error = ENXIO;
		LOG( LOG_RETURN, ( "m681open: return error, invalid dev=%d\n",
		    dev ) );
		return;
	}
	addr = m681_addr[ dev/2 ];
	if ( (int)addr == MISSING ) { /* dev is missing */
		u.u_error = ENXIO;
		LOG( LOG_RETURN, ( "m681open: return error, missing dev=%d\n",
		    dev ) );
		return;
	}


	tp = &m681_tty[dev];		/* get tty struct */

	if(Dconcurrent)
		/* sleep if port has previously been opened as a dual console */
		while(m681_state[dev] & DCON){
			sleep(( caddr_t)&tp->t_canq,TTIPRI);
		}

	cur_lvl = SPLCLOCK();
	while(m681wclose[dev])
	{
		if(sleep((caddr_t)&tp->t_canq, TTIPRI|PCATCH))
		{
			u.u_error = ENXIO;
			splx(cur_lvl);
			return;
		}
	}	
	splx(cur_lvl);

	if (! (tp->t_state & (ISOPEN|WOPEN))) {
		ttinit( tp );			/* init tty structure */

		if(device == consdev)
			m681_state[dev] |= SCON; 	/* system console */

		tp->t_proc = m681proc;
		/* set up DTR and RTS */
		if (dev & 1)  /* PORT B */
			addr->write.wr_oprset = OPDTRB | OPRTSB;
		else	/* PORT A */
			addr->write.wr_oprset = OPDTRA | OPRTSA;
		m681param( dev );
	}

	/* check carrier */
	if(tp->t_cflag&CLOCAL || m681modem(dev,ON)) {
		tp->t_state |= CARR_ON;
		wakeup((caddr_t)&tp->t_canq);
	}
	else {
		tp->t_state &= ~CARR_ON;
		if ( !( flag & FNDELAY ) ) {

			/* wait on DCD active */
			/* DCD changes don't produce an interrupt,
	       so we will scan for changes */
			m681_work[dev] = M681_DCD;
			LOG( LOG_INTERN,
			    ( "m681open: dev=%d, waiting on DCD\n", dev ) );

			cur_lvl = SPLCLOCK();
			while (! (tp->t_state & CARR_ON)) {
				tp->t_state |= WOPEN;
				sleep(( caddr_t)&tp->t_canq,TTIPRI);
			}
			splx( cur_lvl );
		}
	}

	(*linesw[tp->t_line].l_open)(tp);

	/* set duart open flag, for interrupt check loop */
	m681_openduarts[dev/2] = TRUE;

#ifdef STATISTICS
	m681_stats[dev].open_cnt++;
#endif

	LOG( LOG_RETURN, ( "m681open: return okay\n" ) );
	return;
}


/*********************************************************************
 *
 *  Function:	  m681close
 *
 *  Description:  Close of the driver
 *
 *
 *  Syntax:	  m681close( dev )
 *
 *  Input:	  int dev;	minor device number
 *
 *  Output:       
 *
 *
 *  Called by:	  Character device  close  switch
 *
 *  Calls to:     m681modem()
 *		  line switch  l_close()
 *		  SPLCLOCK()
 *		  splx()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681close()
 *
 *  BEGIN
 *
 *	call close procedure in upper tty driver [ l_close() ];
 *
 *	block all interrupts for this driver [ SPLCLOCK() ];
 *
 *	IF  hangup on close mode  AND  not opened by another process
 *	THEN
 *	    hangup the line [ m681modem()];
 *	ENDIF
 *
 *	free interrupts for this driver [ splx() ];
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/


m681close( dev )
unsigned int dev;
{
	register struct tty *tp;
	int cur_lvl;

	dev = minor(dev);
   
	LOG( LOG_START, ( "m681close: started dev=%d\n", dev ) );

	tp = &m681_tty[dev];

	cur_lvl = SPLCLOCK();
	m681wclose[dev] = 1;
	splx( cur_lvl );

	( *linesw[tp->t_line].l_close )( tp );

	if ((tp->t_cflag & HUPCL) && !(tp->t_state & ISOPEN))
		m681modem( dev, OFF );

	/* check if any port of this duart is opened */
	if ( !( (&m681_tty[(dev/2)*2  ])->t_state & ISOPEN)
	    && !( (&m681_tty[(dev/2)*2+1])->t_state & ISOPEN) ) {
		/* set duart open flag, for interrupt check loop */
		m681_openduarts[dev/2] = FALSE;
	}

	cur_lvl = SPLCLOCK();
	m681wclose[dev] = 0;
	m681_state[dev] &= ~SCON;
	splx( cur_lvl );
	wakeup((caddr_t)&tp->t_canq);

	LOG( LOG_RETURN, ( "m681close: return okay\n" ) );

	return;
}


/*********************************************************************
 *
 *  Function:	  m681read
 *
 *  Description:  Read of the driver
 *
 *
 *  Syntax:	  m681read( dev )
 *
 *  Input:	  int dev;	minor device number
 *
 *  Output:       
 *
 *
 *  Called by:	  Character device  read  switch
 *
 *  Calls to:     line switch  l_read()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681read()
 *
 *  BEGIN
 *
 *	call read procedure in upper tty driver [ l_read() ];
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/


m681read(dev)
unsigned int dev;
{
	register struct tty *tp;

	dev = minor(dev);

	LOG( LOG_START, ( "m681read: started dev=%d\n", dev ) );

	tp = &m681_tty[ dev ];
	( *linesw[tp->t_line].l_read )( tp );

	LOG( LOG_RETURN, ( "m681read: return okay\n" ) );

	return;
}



/*********************************************************************
 *
 *  Function:	  m681write
 *
 *  Description:  Write of the driver
 *
 *
 *  Syntax:	  m681write( dev )
 *
 *  Input:	  int dev;	minor device number
 *
 *  Output:       
 *
 *
 *  Called by:	  Character device  write  switch
 *
 *  Calls to:     line switch  l_write()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681write()
 *
 *  BEGIN
 *
 *	call write procedure in upper tty driver [ l_write() ];
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/


m681write(dev)
unsigned int dev;
{
	register struct tty *tp;

	dev = minor(dev);

	LOG( LOG_START, ( "m681write: started dev=%d\n", dev ) );

	tp = &m681_tty[ dev ];
	( *linesw[tp->t_line].l_write )( tp );

	LOG( LOG_RETURN, ( "m681write: return okay\n" ) );

	return;
}



/*********************************************************************
 *
 *  Function:	  m681ioctl
 *
 *  Description:  Ioctl of the driver
 *
 *
 *  Syntax:	  m681ioctl( dev, cmd, arg, mode )
 *
 *  Input:	  unsigned int dev;	minor device number
 *		  unsigned int cmd;	command
 *		  unsigned int arg;	argument pointer
 *		  unsigned int mode;	ioctl mode
 *
 *  Output:       
 *
 *
 *  Called by:	  Character device  ioctl  switch
 *
 *  Calls to:     copyout()
 *		  ttiocom()
 *		  m681param()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681ioctl()
 *
 *  BEGIN
 *
 *	SWITCH  command
 *	    
 *	    CASE  set hardware handshake:
 *		enable RTS control;
 *		enable CTS control;
 *		BREAK;
 *		
 *	    CASE  clear hardware handshake:
 *		disable RTS control;
 *		disable CTS control;
 *		BREAK;
 *		
 *	    CASE  get hardware handshake status:
 *		IF  RTS control is set
 *		THEN
 *		    status = TRUE;
 *		ELSE
 *		    status = FALSE;
 *		ENDIF
 *		BREAK;
 *		
 *	    CASE  get statistics:
 *		copy statistic data to user space [ copyout() ];
 *		BREAK;
 *
 *	    DEFAULT
 *		call ioctl procedure in upper tty driver [ ttiocom() ];
 *		IF  hardware dependent i/o control
 *		THEN
 *		    set hardware parameters [ m681param() ];
 *		ENDIF
 *
 *	ENDSWITCH
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/



#define	CFLAGMASK	(CBAUD|CSIZE|CSTOPB|PARENB|PARODD|CREAD)

m681ioctl( device, cmd, arg, mode )
unsigned int device, cmd, arg, mode;
{
	register struct tty *tp;
	struct debugioctl debstr;

	register union m681_pt_io	*ptaddr;
	register union m681_io	*addr;
	int duart, port;
	unsigned char mr1, mr2;
	int hwhandstat;
	register cflag;
	register unsigned int dev;

	dev = minor(device);

	LOG( LOG_START,
	    ( "m681ioctl: started dev=%d, cmd=0x%x, arg=0x%x, mode=0x%x\n",
	    dev, cmd, arg, mode ) );

	duart = dev/2;

	addr = m681_addr[duart];

	if ( dev & 1 ) {
		port = B_PORT;
		ptaddr = ( union m681_pt_io *) addr + 1;
	}
	else {
		port = A_PORT;
		ptaddr = ( union m681_pt_io *) addr;
	}

	tp = &m681_tty[dev];

	switch ( cmd ) {

	case MSETHWHAND:
		/* enable RTS control */
		mr1 = m681_sv_reg[duart].sv_mr1[port];
		mr1 |= RXRTS;

		/* enable CTS control */
		mr2 = m681_sv_reg[duart].sv_mr2[port];
		mr2 |= TXCTS;

		/* write mode registers to duart */
		ptaddr->write.wr_cr = MRRESET;
		ptaddr->write.wr_mr = mr1;
		ptaddr->write.wr_mr = mr2;

		LOG( LOG_INTERN,
		    ( "m681ioctl: set HWHAND mr1 to 0x%x, mr2 to 0x%x\n",
		    mr1, mr2 ) );

		/* save changed mode registers */
		m681_sv_reg[duart].sv_mr1[port] = mr1;
		m681_sv_reg[duart].sv_mr2[port] = mr2;
		break;

	case MCLEARHWHAND:
		/* disable RTS control */
		mr1 = m681_sv_reg[duart].sv_mr1[port];
		mr1 &= ~RXRTS;

		/* disable CTS control */
		mr2 = m681_sv_reg[duart].sv_mr2[port];
		mr2 &= ~TXCTS;

		/* write mode registers to duart */
		ptaddr->write.wr_cr = MRRESET;
		ptaddr->write.wr_mr = mr1;
		ptaddr->write.wr_mr = mr2;

		LOG( LOG_INTERN,
		    ( "m681ioctl: clear HWHAND mr1 to 0x%x, mr2 to 0x%x\n",
		    mr1, mr2 ) );

		/* save changed mode registers */
		m681_sv_reg[duart].sv_mr1[port] = mr1;
		m681_sv_reg[duart].sv_mr2[port] = mr2;
		break;

	case MGETHWHAND:
		if ( m681_sv_reg[duart].sv_mr1[port] & RXRTS ) {
			hwhandstat = 1;
		}
		else {
			hwhandstat = 0;
		}
		if( copyout( &hwhandstat, arg, sizeof( hwhandstat ) )) {
			u.u_error = EFAULT;
		}
		break;

#ifdef STATISTICS
	case M681GETSTAT:
		if ( copyout( m681_stats, arg, sizeof( m681_stats ) ) )
			u.u_error = EFAULT;
		break;
#endif

	case CONC_PUTC:
		m681putc (dev, arg);
		break;

	case CONC_IPUTC:
		m681inputc (dev, (arg&0xff), (arg >> 8));
		break;

	case CONC_BOOT:
		/* The system was booted in dual console mode.
		   The port is already open. Do not disturb the current
		   baud rate and protocol settings, but do enable 
		   interrupts. */
		if (port == A_PORT)
			m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | IRXRDYA;
		else
			m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | IRXRDYB;
		addr -> write.wr_imr = m681_sv_reg[duart].sv_imr;
		          			/* enable receiver interrupts */

		m681_state[dev] |= DCON; 
		break;

	case CONC_ON:
		if( m681_state[dev] & DCON){ /* dual console mode is on */
			break;
		}
		if( ! (tp->t_state & (ISOPEN | WOPEN))){
			m681open (device, 0);  	

			/* Remember to close when dual console is turned				   off. */
			if(!u.u_error)
				m681_state[dev] |= (DCON | DCCLOSE); 
			break;
		}
		/* Port is already open due to a previous open. Dont
		   disturb the current baud rate and protocol settings
		   on the port. */

		m681_state[dev] |= DCON; 
		
		break;
		


	case CONC_OFF:
		if(!(m681_state[dev] & DCON)){ /*dual console mode is off */
			break;
		}
		if(m681_state[dev] & DCCLOSE){
			m681close (device);
		}
		m681_state[dev] &= ~(DCON | DCCLOSE);
		wakeup(&tp->t_canq);		
		break;

	default:
		cflag = tp->t_cflag&CFLAGMASK;
		if ( ttiocom( tp, cmd, arg, mode ) )
			if( cflag != (tp->t_cflag&CFLAGMASK) )
				m681param( dev );
	}

	LOG( LOG_RETURN, ( "m681ioctl: return okay\n" ) );

	return;
}


/*********************************************************************
 *
 *  Function:	  m681intr
 *
 *  Description:  Interrupt service routine of the driver
 *
 *
 *  Syntax:	  m681intr( dev )
 *
 *  Input:	  int dev;	minor device number of the 
 *				interrupting DUART
 *				Interrupt from Port A or B has the same
 *				device number, because there is only
 *				one interupt source for both ports.
 *				
 *  Output:       
 *
 *
 *  Called by:	  Interrupt handler
 *
 *  Calls to:     m681proc()
 *		  line switch  l_input()
 *		  SPLCLOCK()
 *		  splx()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681intr()
 *
 *  BEGIN
 *
 *  WHILE  all opened duarts are checked
 *  DO
 *	set the duart number for checking:
 *	    IF  first entry in loop
 *	    THEN
 *		duart number which forced the interrupt;
 *	    ELSE
 *		duart number is the next in the queue of opened duarts;
 *	    ENDIF
 *
 *	check this duart for all interrupt sources from both ports:
 *	    check which port interrupts, start with A port;
 *	    read status register of the port;
 *
 *	    IF  transmitter is ready
 *	    THEN
 *		IF  XOFF should be sent
 *		THEN
 *		    send XOFF character;
 *		ELSE
 *		    IF  XON should be sent
 *		    THEN
 *			send XON character;
 *		    ELSE
 *			disable transmitter;
 *			check for characters to send [ m681proc( T_OUTPUT ) ];
 *		    ENDIF
 *		ENDIF
 *	    ENDIF
 *
 *	    IF  receiver is ready
 *	    THEN
 *		read character from port;
 *
 *		check for special characters:
 *		    IF  Break-Character received
 *		    THEN
 *			call input procedure in upper tty driver [ l_input( L_BREAK ) ];
 *			clear break state;
 *		    ELSE
 *			IF  output was stopped and XON character received
 *			THEN
 *			    resume output [ m681proc( T_RESUME ) ];
 *			ENDIF
 *		    ENDIF
 *		
 *		process character:
 *		    IF  strip mode 
 *		    THEN
 *			strip character to 7 bits;
 *		    ENDIF
 *
 *		    IF  error at receiving character
 *		    THEN
 *			clear error state;
 *			IF  parity checking is on
 *			THEN
 *			    IF  parity mark is on
 *			    THEN
 *				put 0xff and 0x00 on input queue;
 *			    ELSE
 *				character is NULL character;
 *			    ENDIF
 *			ENDIF
 *		    ENDIF
 *
 *		stash character in receive buffer;
 *		IF  receive buffer is full
 *		THEN
 *		    call input procedure in upper tty driver [ l_input( L_BUF ) ];
 *		ELSE
 *		    activate scan routine;
 *		ENDIF
 *
 *	    
 *	    IF  BREAK state in ISR register ( perhaps from end of BREAK )
 *	    THEN
 *		clear it;
 *	    ENDIF
 *
 *	    END of checking one port
 *
 *	END of checking duart loop
 *
 *	RETURN;
 *
 *  END
 *
 *********************************************************************/


m681intr( dev )
unsigned int dev;
{
	register union m681_io	*addr;	    /* pointer to DUART regs */
	register union m681_pt_io	*ptaddr;    /* pointer to port regs */

	register unsigned char	isr;	    /* interrupt status reg */
	register unsigned char	sr;	    /* status reg of port a/b */
	unsigned char		ipcr;	    /* input port change reg */
	unsigned char		ip;	    /* input port */

	register struct tty	*tp;	    /* tty structure */

	register unsigned char	c, ctmp;    /* received character */
	BOOL			rcvflag;    /* receive flag */
	unsigned int		duart;	    /* duart index */
	unsigned int		intrduart;  /* interrupt duart index */
	int			port;	    /* flag for a or b port */

	char			lbuf[3];    /* temp receive buffer */
	int			lcnt;	    /* count of chars in lbuf */
	int			cur_lvl;    /* current int level */
	int			lvl;    /* current int level */
	int			tmpflag;    /* temp flag */
	unsigned 	char dcstatus;

	BOOL	firstloop = TRUE;	    /* first loop flag */
#ifdef STATISTICS
	BOOL	firstintr = TRUE;	    /* first intr flag */
#endif

	LOG( LOG_START, ( "m681intr: started dev=%d\n", dev ) );

	lvl = SPLCLOCK();

	while( TRUE ) {
		if ( firstloop == TRUE ) {
			firstloop = FALSE;
			/* check the duart which forced the interrupt */
			duart = dev/2;
			intrduart = duart;
		}
		else {
			/* check all opened duarts */
			if ( ++duart == MAXDUARTS ) {
				duart = 0;
			}
			if ( duart == intrduart ) {
				break;	/* return from the interrupt routine */
			}
			if ( duart == MAXDUARTS ) {
				duart = 0;
			}
			if ( m681_openduarts[duart] == FALSE ) {
				/* check next duart */
				continue;
			}
		}

		LOG( LOG_INTERN, ( "m681intr: check duart=%d\n", duart ) );

		if ( (int) ( addr = m681_addr[duart] ) == MISSING ) {
			LOG( LOG_WARN,
			    ( "m681intr: Warning check interrupt by missing duart=%d\n",
			    duart ) );
			continue;
		}

		while( TRUE ) {	/* check for all interrupt sources */

			/* interrupted by which port ? */
			isr = addr->read.rd_isr;
			/* must mask with saved copy of imr */
			isr = isr & m681_sv_reg[duart].sv_imr;
			LOG( LOG_INTERN, ( "m681intr: isr=0x%x\n", isr ) );
			if( isr & AINTPORT ) {
				port = A_PORT;
				dev = duart*2;
			}
			else {
				if( isr & BINTPORT ) {
					port = B_PORT;
					dev = duart*2 + 1;
				}
				else {
					if( isr & IIPCHG ) { /* inputport chg */
						ipcr = addr->read.rd_ipcr;
						ip = addr->read.rd_ip;
						m681ckdcd(duart,ipcr,ip);
					}
					break;
				}
			}

#ifdef STATISTICS
			if ( firstintr == TRUE ) {
				firstintr = FALSE;
				m681_stats[dev].intr_cnt++;
			}
#endif
			tp = &m681_tty[dev];

			/* set port address */
			if ( port == A_PORT ) {
				ptaddr = ( union m681_pt_io *) addr;
			}
			else {
				ptaddr = ( union m681_pt_io *) addr + 1;
			}

			/* read status reg */
			sr = ptaddr->read.rd_sr;

			LOG( LOG_INTERN,
			    ( "m681intr: intr by duart=%d port=%d, sr=0x%x, tp=0x%x\n", 
			    duart, port, sr, tp ) );

			/* transmit interrupt handler */
			if ( sr & TXRDY ) {
				LOG( LOG_INTERN, ( "m681intr: TXRDY interrupt\n" ) );
				sysinfo.xmtint++;
#ifdef STATISTICS
				m681_stats[dev].tx_cnt++;
#endif
				if ( tp->t_state & TTXOFF ) {
					/* send XOFF char */
					ptaddr->write.wr_tb = CSTOP;
					tp->t_state &= ~TTXOFF;
#ifdef STATISTICS
					m681_stats[dev].txoff_cnt++;
#endif
				}
				else {
					if ( tp->t_state & TTXON ) {
						/* send XON char */
						ptaddr->write.wr_tb = CSTART;
						tp->t_state &= ~TTXON;
#ifdef STATISTICS
						m681_stats[dev].txon_cnt++;
#endif
					}
					else {
						tp->t_state &= ~BUSY;

						/* disable Tx */
						/* ptaddr->write.wr_cr = TXDIS; */
						if (port == A_PORT)
							m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr & ~ITXRDYA;
						else
							m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr & ~ITXRDYB;
						addr -> write.wr_imr = m681_sv_reg[duart].sv_imr;

						/* check for chars to send */
						m681proc( tp, T_OUTPUT );
					}
				}
			}

			/* receiver interrupt handler*/
			if ( sr & RXRDY ) {

				sysinfo.rcvint++;
#ifdef STATISTICS
				m681_stats[dev].rx_cnt++;
#endif
				/* read character from port */
				c = ptaddr->read.rd_rb;
				LOG( LOG_INTERN, 
				    ( "m681intr: receive on dev=%d char=0x%x\n", dev, c));
				/* check break condition */
				if ( sr & RBRK ) {
					LOG( LOG_INTERN, ( "m681intr: BREAK interrupt\n" ) );
					tmpflag= tp->t_iflag;
					if ( !(tmpflag&IGNBRK) && (tmpflag&BRKINT) ) {
						(*linesw[tp->t_line].l_input) (tp, L_BREAK);
					}

					/* clear break state */
					ptaddr->write.wr_cr = BRKINTRESET;
					DELAY_CR;
					ptaddr->write.wr_cr = ERRRESET;

					rcvflag = FALSE;

#ifdef STATISTICS
					m681_stats[dev].brk_cnt++;
#endif

					if (Dconcurrent && (m681_state[dev] & DCON)) {
						dcstatus = DC_BREAK;
						c = 0;
						consinputc(c, dcstatus);
						continue;
					}

					LOG( LOG_INTERN,
					    ( "m681intr: break detected for dev=%d\n", dev ) );
				}
				else {
					lcnt= 1;
					rcvflag = TRUE;
					if(Dconcurrent)
						if(m681_state[dev] & DCON){
							dcstatus = 0;
							consinputc(c, dcstatus);
							continue;
						}
				}

				/* check for software handshake */
				if ( tp->t_iflag & IXON ) {
					ctmp = c & 0x7f;
					if ( tp->t_state & TTSTOP ) {
						if ( (ctmp==CSTART) || (tp->t_iflag & IXANY) ) {
							LOG( LOG_INTERN,
							    ( "m681intr: dev=%d: XON received\n",
							    dev ) );
							(*tp->t_proc)( tp, T_RESUME );
#ifdef STATISTICS
							m681_stats[dev].rxon_cnt++;
#endif
						}
					}
					else {
						if ( ctmp == CSTOP ) {
							(*tp->t_proc)( tp, T_SUSPEND );
							LOG( LOG_INTERN,
							    ( "m681intr: dev=%d: XOFF received\n", dev ) );
#ifdef STATISTICS
							m681_stats[dev].rxoff_cnt++;
#endif
						}
					}
					if ( ctmp == CSTART || ctmp == CSTOP ) {
						rcvflag = FALSE;  /* do not process character */
					}
				}

				if ( rcvflag == TRUE ) {
		/*
		 *	If ISTRIP is set, strip input character to 7 bits,
		 *	or if less than CS8, strip possible parity bit,
		 *	instead of letting it pass thru.  This conforms with
		 *	the intent of ATT of not having parity in the data byte.
		 */

					if ( tp->t_iflag & ISTRIP
					     || (tp->t_cflag & CSIZE) < CS8) {
						c &= 0x7f;		/* stripped to 7 bits */
					}
					else {
						if ( c == 0xff && tp->t_iflag & PARMRK ) {
							/* prepare a second 0xff for the special case
			       of receiving a 0xff, if PARMRK mode is set */
							lbuf[1]= 0xff;
							lcnt= 2;
						}
					}

					if ( sr & (FRERR|PERR|ROVRN) ) { /*errors*/
						if(Dconcurrent)
							if(m681_state[dev] & DCON){
								dcstatus = 0;
								if (sr & PERR)
									dcstatus |= DC_PARERR;
								if (sr & ROVRN)
									dcstatus |= DC_ROVRN;
								if (sr & FRERR)
									dcstatus |= DC_FRERR;
								consinputc(c, dcstatus);
								continue;
							}

#ifdef STATISTICS
						if ( sr & FRERR ) {
							m681_stats[dev].frerr_cnt++;
							LOG( LOG_INTERN,
							    ( "FRERR: duart=%d, port=%d, sr=0x%x\n", 
							    duart, port, sr ) );

						}
						else {
							if ( sr & PERR ) {
								m681_stats[dev].perr_cnt++;
								LOG( LOG_INTERN,
								    ( "PERR: duart=%d, port=%d, sr=0x%x\n", 
								    duart, port, sr ) );
							}
							else {
								if ( sr & ROVRN ) {
/* code folded from here */
	m681_stats[dev].rovrn_cnt++;
	LOG( LOG_INTERN,
	    ( "ROVRN: duart=%d, port=%d, sr=0x%x\n", 
	    duart, port, sr ) );
/* unfolding */
								}
							}
						}
#endif
						/* clear error state */
						ptaddr->write.wr_cr = ERRRESET;

						/* parity error checking */
						tmpflag=tp->t_iflag;
						if (!(tmpflag&IGNPAR)) {	/* no checking at all */
							lcnt = 1;
							if (!(tmpflag&INPCK))	/* no parity check */
								sr &= ~PERR;
							if (sr & (PERR|ROVRN)) {

								if (tmpflag & PARMRK) {
/* code folded from here */
	lbuf[2] = 0xFF;
	lbuf[1] = 0;
	lcnt = 3;
	sysinfo.rawch += 2;
/* unfolding */
								} else
/* code folded from here */
	c = 0;
/* unfolding */
							}
						}
					}

					if ( sr & ROVRN ) {
						LOG( LOG_WARN, ( "m681intr: Warning: rx overrun on dev %d\n", dev ) );
					}

					/* stash character in r_buf */
					lbuf[0] = c;
					if ( tp->t_rbuf.c_ptr != NULL ) {
						while ( lcnt ) {
							cur_lvl = SPLCLOCK(); /* to block m681scan() from
													 running at clock priority*/
							*tp->t_rbuf.c_ptr++ = lbuf[--lcnt];
							LOG( LOG_INTERN, 
							    ( "m681intr: put 0x%x on receive buffer = 0x%x, c_count=%d\n",
							    lbuf[lcnt], tp->t_rbuf.c_ptr-1, tp->t_rbuf.c_count-1 ) );
							if ( --tp->t_rbuf.c_count <= m681limit) {
								m681_work[dev] = 0;
								tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size
								    - tp->t_rbuf.c_count;
								(*linesw[tp->t_line].l_input)(tp,L_BUF);
							}
							else {
								m681_work[dev] = M681_ACT;
							}
							/* note M681_CHAR bit cleared */
							splx(cur_lvl);
						}
					}
					else {
						LOG( LOG_INTERN, 
						    ( "m681intr: receive: dev=%d, rbuf.c_ptr=0x%x, c=%x\n",
						    dev, tp->t_rbuf.c_ptr, c ) );
					}
				}
			}

			/* if BREAK state in ISR register ( perhaps from end of BREAK ),
	       then clear it */
			if ( ( (port == A_PORT ) && ( isr & IBRKA ) )
			    || ( (port == B_PORT ) && ( isr & IBRKB ) ) ) {
				/* clear break state */
				ptaddr->write.wr_cr = BRKINTRESET;
			}

		}    /* end of interrupt check of one duart */

	}	/* end of loop thru all duarts */

	LOG( LOG_RETURN, ( "m681intr: return okay\n" ) );

	splx(lvl);

	return;
}

/******************************************************************/

m681inputc(dev, c, rxstat)
register unsigned int	dev;
register unsigned char	c;
register unsigned char	rxstat;
{
	register struct tty	*tp;
	register int		flg, count;
	char			lbuf[3];
	int			lcnt;
	int			cur_lvl;
	unsigned char		ctmp;					
	
	tp = &m681_tty[dev];
	flg= tp->t_iflag;
	
	switch (rxstat) {

	case DC_BREAK:
					if ( !(flg & IGNBRK) && (flg & BRKINT) ) {
						(*linesw[tp->t_line].l_input) (tp, L_BREAK);
					}
					return;

	case DC_PARERR:
	case DC_ROVRN:
	case DC_FRERR:
		
		lcnt = 1;
		flg = tp->t_iflag;
		
		if (!(flg & IGNPAR)) {
			if (!(flg & INPCK))
				rxstat &= ~DC_PARERR;
			if (rxstat & (DC_PARERR | DC_ROVRN | DC_FRERR)) {
				/*
		 		 *	If PARMRK is set, a character with a framing
		 		 *	or parity error is read as the three-character
		 		 *	sequence 0xFF, 0, X, where X is the data of the
		 		 *	character received in error. 
		 		 */

				if (flg & PARMRK) {					
					lbuf[2] = 0xFF;						
					lbuf[1] = 0;					
					lcnt = 3;					
					sysinfo.rawch += 2;				
				} 
				/*
		 		 *	If PARMRK (mark parity errors) is not set,
		 		 *	a framing or parity error which is not 
		 		 *	ignored is read as the character NUL (0).
		 		 */

				else c = 0;						
			}
		}
		break;
		
	case DC_DCDON:
	
		if (m681_work[dev] & M681_DCD) {
			wakeup ( &tp->t_canq );
			tp->t_state |= CARR_ON;
			m681_work[dev] &= ~M681_DCD;
		} 
		return;

	case DC_DCDOFF:
		if ( !( tp->t_state & CLOCAL ) ) {
			if ( tp->t_state & CARR_ON ) {
				if ( tp->t_state & ISOPEN ) {
					signal (tp->t_pgrp,SIGHUP);
					ttyflush (tp, (FREAD | FWRITE) );
				}
				tp->t_state &= ~CARR_ON;
			}
		}
		return;

	default:
		lcnt = 1;
		flg = tp->t_iflag;

		sysinfo.rcvint++; /* Increment counter of receiver interrupts */

		/*	If start/stop output control is enabled in t_iflag,
		 *	the state of the tty is TTSTOP and the character 
		 *	read from the receive buffer is a CSTART, then call
		 *	the command processing routine to resume output.
		 *	If the character is a CSTOP, call the proc routine
		 * 	to suspend output.
		 */
		
		if (flg & IXON) {						
			ctmp= c & 0x7f;				
			if (tp->t_state & TTSTOP){			
				if ((ctmp == CSTART) || (flg & IXANY))		
   					m681proc (tp, T_RESUME);		
			} else {					
				if (ctmp == CSTOP)			
					m681proc (tp, T_SUSPEND);		
			  }						
			if (ctmp == CSTART || ctmp == CSTOP) {
				return; /* ignore these characters */	
			}
		}							
		/*
		 *	If ISTRIP is set, strip input character to 7 bits,
		 *	or if less than CS8, strip possible parity bit,
		 *	instead of letting it pass thru.  This conforms with
		 *	the intent of ATT of not having parity in the data byte.
		 */

		if (flg & ISTRIP || (tp->t_cflag & CSIZE) < CS8)
			c &= 0x7f;		/* stripped to 7 bits */
		
			/*
			 *	Otherwise, to avoid ambiguity, a valid
			 *	character of 0xFF is read as 0xFF, 0xFF.
			 */

		else if (c == 0xff && flg & PARMRK) {			
				lbuf[1]= 0xff;				
				lcnt= 2;				
		}		   						
		break;
		
	}
	/* stash character in r_buf */
	lbuf[0] = c;
	if ( tp->t_rbuf.c_ptr != NULL ) {
		while ( lcnt ) {
			cur_lvl = SPLCLOCK(); /* to block m681scan() from
										running at clock priority*/
			*tp->t_rbuf.c_ptr++ = lbuf[--lcnt];
			LOG( LOG_INTERN, 
							( "m681intr: put 0x%x on receive buffer = 0x%x, c_count=%d\n",
							lbuf[lcnt], tp->t_rbuf.c_ptr-1, tp->t_rbuf.c_count-1 ) );
			if ( --tp->t_rbuf.c_count <= m681limit) {
				m681_work[dev] = 0;
				tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size
								- tp->t_rbuf.c_count;
				(*linesw[tp->t_line].l_input)(tp,L_BUF);
			}
			else {
				m681_work[dev] = M681_ACT;
			}
			/* note M681_CHAR bit cleared */
			splx(cur_lvl);
		}
	}
}

/*********************************************************************
 *
 *  Function:	  m681proc
 *
 *  Description:  Perform output of the driver
 *
 *
 *  Syntax:	  m681proc( tp, cmd )
 *
 *  Input:	  struct tty *tp;	TTY structure
 *		  int cmd;		command
 *
 *  Output:       
 *
 *
 *  Called by:	  upper level TTY driver and
 *		  interrupt service routine
 *
 *  Calls to:     m681param()
 *		  m681sendchar()
 *		  line switch  l_output()
 *		  timeout()
 *		  SPLCLOCK()
 *		  splx()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681proc()
 *
 *  BEGIN
 *
 *	SWITCH  command
 *
 *	    CASE T_TIME:
 *		resuming output after delay:
 *		    stop a break output
 *		    goto T_OUTPUT case;
 *
 *	    CASE T_WFLUSH:
 *		flush output queue:
 *		    clear transmit queue;
 *		fall thru;
 *
 *	    CASE T_RESUME:
 *		resume output after a holdup due to an XOFF;
 *		fall thru;
 *
 *	    CASE OUTPUT:
 *		handle first or next char:
 *		    IF  port is BUSY or STOPPED
 *		    THEN
 *			BREAK;
 *		    ENDIF
 *
 *		    IF  no character in transmit queue
 *		    THEN
 *			call output procedure in upper tty driver [ l_output() ];
 *			IF  no characters to output
 *			THEN
 *			    BREAK;
 *			ENDIF
 *		    ENDIF
 *
 *		    send character [ m681sendchar() ];
 *		    update count;
 *		    BREAK;
 *
 *	    CASE T_SUSPEND:
 *		stop output, XOFF has been received:
 *		    set STOP flag in status;
 *		    BREAK;
 *
 *	    CASE T_BLOCK:
 *		block input:
 *		    set BLOCK flag in status;
 *		    IF  hardware handshake is on
 *		    THEN
 *			reset RTS line;
 *		    ELSE
 *			send XOFF character:
 *			    IF  transmitter is busy
 *			    THEN
 *				set flag to transmit XOFF;
 *			    ELSE
 *				send XOFF character [ m681sendchar() ];
 *			    ENDIF
 *		    ENDIF
 *		    BREAK;
 *
 *	    CASE T_RFLUSH:
 *		flush input queue:
 *		    IF  input not blocked
 *		    THEN
 *			BREAK;
 *		    ENDIF
 *		    fall thru;
 *
 *	    CASE T_UNBLOCK:
 *		unblock input:
 *		    clear BLOCK and XOFF flag in status;
 *		    IF  hardware handshake is on
 *		    THEN
 *			set RTS line;
 *		    ELSE
 *			send XON character:
 *			    IF  transmitter is busy
 *			    THEN
 *				set flag to transmit XON;
 *			    ELSE
 *				send XON character [ m681sendchar() ];
 *			    ENDIF
 *		    ENDIF
 *		    BREAK;
 *
 *	    CASE T_BREAK:
 *		send break condition for 1/4 second:
 *		    start the break in the duart and enable transmitter;
 *		    sent a null character to the transmitter to
 *			prevent teh transmitter empty interrupt;
 *		    start the timer for 1/4 second [ timeout() ];
 *		    BREAK;
 *
 *	    CASE T_PARM:
 *		set hardware parameters:
 *		    call hardware parameter setting [ m681param() ];
 *		    BREAK;
 *	ENDSWITCH
 *
 *	RETURN;
 *
 *  END
 *
 *********************************************************************/


m681proc(tp, cmd)
register struct tty *tp;
unsigned int cmd;
{
	register union m681_io	*addr;
	register union m681_pt_io	*ptaddr;
	int	dev, duart, port;
	int cur_lvl;

	extern ttrstrt();

	cur_lvl = SPLCLOCK();

	dev = tp - m681_tty;

	LOG( LOG_START, ( "m681proc: started tp=0x%x, dev=%d, cmd=%d, t_state=0x%x\n",
	    tp, dev, cmd, tp->t_state ));

	/* set addresses and index of DUART */
	duart = dev/2;

	addr = m681_addr[duart];

	if ( dev & 1 ) {
		port   = B_PORT;
		ptaddr = ( union m681_pt_io *) addr + 1;
	}
	else {
		port   = A_PORT;
		ptaddr = ( union m681_pt_io *) addr;
	}

	switch (cmd) {

	case T_TIME:		/* resuming output after delay */
		tp->t_state &= ~TIMEOUT;
		if(m681break[dev])
		{
			ptaddr->write.wr_cr = BRKSTOP; 	/* stop a break*/
			m681break[dev] = 0;
		}
		LOG( LOG_INTERN, ( "m681proc: T_TIME\n" ) );
		goto start;

	case T_WFLUSH:		/* flush output queue */
		tp->t_tbuf.c_size -= tp->t_tbuf.c_count;
		tp->t_tbuf.c_count = 0;
		LOG( LOG_INTERN, ( "m681proc: T_WFLUSH\n" ) );
		/* fall thru */

	case T_RESUME:		/* resume output, XON has been received */
		tp->t_state &= ~TTSTOP;
		LOG( LOG_INTERN, ( "m681proc: T_RESUME\n" ) );
		/* fall thru */

	case T_OUTPUT:		/* handle first or next char */
start:	    
		{
			register struct ccblock *tbuf;
			LOG( LOG_INTERN, ("m681proc: T_OUTPUT started\n") );

			if (tp->t_state & (BUSY|TTSTOP))
				break;
			tbuf = &tp->t_tbuf;
			if (tbuf->c_ptr == NULL || tbuf->c_count == 0) {
				/* tbuf is empty */
				if (tbuf->c_ptr)
					tbuf->c_ptr -= tbuf->c_size;
				if (!(CPRES & (*linesw[tp->t_line].l_output)(tp))) {

					LOG( LOG_INTERN,
					    ( "m681proc: T_OUTPUT  TX disabled\n") );
					break;
				}
			}

			LOG( LOG_INTERN, ( "m681proc: dev=%d, put char=0x%x\n", 
			    dev, *tp->t_tbuf.c_ptr ) );
			if ( m681sendchar( tp, ptaddr, *tp->t_tbuf.c_ptr,addr,port, duart ) == 0 ) {
				LOG( LOG_WARN,
				    ( "m681proc: Warning: T_OUTPUT: Send char not okay\n" ) );
			}
			else {
				/* Also send character to dual console if enabled. */
				if(Dconcurrent && (m681_state[dev] & SCON))
					dconsputc(*tp->t_tbuf.c_ptr);
				tp->t_tbuf.c_ptr++;
				tp->t_tbuf.c_count--;	/* update count */
			}

		}
		break;

	case T_SUSPEND:	    /* stop output, XOFF has been received */
		tp->t_state |= TTSTOP;
		LOG( LOG_INTERN, ( "m681proc: T_SUSPEND\n" ) );
		break;

	case T_BLOCK:	    /* block input */
		tp->t_state &= ~TTXON;
		tp->t_state |= TBLOCK;

		if ( m681_sv_reg[duart].sv_mr1[port] & RXRTS ) {
			/* if HW handshake is enabled, clear RTS */
			if ( port == A_PORT ) {
				addr->write.wr_oprreset = OPRTSA;
			}
			else {	/* Port B */
				addr->write.wr_oprreset = OPRTSB;
			}
		}
		else {
			/* send XOFF */
			if ( tp->t_state & BUSY ) {
				tp->t_state |= TTXOFF;
			}
			else {
				if (  m681sendchar( tp, ptaddr, CSTOP,addr,port,duart ) == 0 ) {
					tp->t_state |= TTXOFF;
					/* enable the transmitter to force an interrupt */
					ptaddr->write.wr_cr = TXEN;
				}
#ifdef STATISTICS
				else {
					/* okay, XOFF was sent */
					m681_stats[dev].txoff_cnt++;
				}
#endif
			}
		}
		LOG( LOG_INTERN, ( "m681proc: T_BLOCK\n" ) );
		break;

	case T_RFLUSH:	    /* flush input queue */
		/* if input blocked, send XON */
		LOG( LOG_INTERN, ( "m681proc: T_RFLUSH\n" ) );
		if (!(tp->t_state & TBLOCK))
			break;
		/* fall thru */

	case T_UNBLOCK:	    /* unblock input */
		tp->t_state &= ~(TTXOFF | TBLOCK);

		if ( m681_sv_reg[duart].sv_mr1[port] & RXRTS ) {
			/* if HW handshake is enabled, enable RTS */
			if ( port == A_PORT ) {
				addr->write.wr_oprset = OPRTSA;
			}
			else {	/* Port B */
				addr->write.wr_oprset = OPRTSB;
			}
		}
		else {
			/* send XON */
			if ( tp->t_state & BUSY ) {
				tp->t_state |= TTXON;
			}
			else {
				if (  m681sendchar( tp, ptaddr, CSTART,addr,port,duart) == 0 ) {
					tp->t_state |= TTXON;
					/* enable the transmitter to force an interrupt */
					ptaddr->write.wr_cr = TXEN;
				}
#ifdef STATISTICS
				else {
					m681_stats[dev].txon_cnt++;
				}
#endif
			}
		}
		LOG( LOG_INTERN, ( "m681proc: T_UNBLOCK\n" ) );
		break;

	case T_BREAK:	    /* send break condition for 1/4 second */
		/* the duart must be enabled with a dummy byte,
	       to prevent the transmitter empty interrupt */
		ptaddr->write.wr_cr = BRKSTART|TXEN;
		ptaddr->write.wr_tb = 0;
		tp->t_state |= TIMEOUT;
		m681break[dev] = 1;
		timeout(ttrstrt, tp, HZ/4);
		LOG( LOG_INTERN, ( "m681proc: T_BREAK\n" ) );
		break;

	case T_PARM:	/* used by sxt driver in R5.2 */
		LOG( LOG_INTERN, ( "m681proc: T_PARAM\n" ) );
		m681param(dev);
		break;
	}

	LOG( LOG_RETURN, ( "m681proc: return okay\n" ) );

	splx(cur_lvl);

	return;
}



/*********************************************************************
 *
 *  Function:	  m681param
 *
 *  Description:  Setting hardware parameters
 *
 *
 *  Syntax:	  m681param( dev )
 *
 *  Input:	  int dev;	minor device number
 *
 *  Output:       u.u_error = EINVAL	if baudrate is invalid
 *
 *
 *  Called by:	  Upper tty device driver
 *		  or m681init()
 *		  or m681ioctl()
 *		  or m681proc()
 *
 *  Calls to:     ttyflush()
 *		  signal()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681param()
 *
 *  BEGIN
 *
 *	disable transmitter and receiver
 *
 *	IF  baudrate is zero baud rate
 *	THEN
 *	    IF  device is opened
 *	    THEN
 *		send hangup signal to process group [ signal() ];
 *		flush read and write queue [ ttyflush() ];
 *	    ENDIF
 *	ELSE
 *	    IF  baudrate is not implemented
 *	    THEN
 *		set EINVAL error code;
 *		set baudrate to default value;
 *	    ENDIF
 *	    set baudrate;
 *	ENDIF
 *
 *	set up character size;
 *
 *	set up stop bits;
 *
 *	set up parity;
 *
 *	IF  receiver enable flag is TRUE
 *	THEN
 *	    enable receiver;
 *	ENDIF
 *
 *	re-enable transmitter;
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/


m681param(dev)
unsigned int dev;
{
	register struct tty	*tp;
	register union m681_io	*addr;
	register union m681_pt_io	*ptaddr;
	register flags;
	register int indx;
	int duart, port;
	int speeds;
	unsigned char mr1, mr2;


	LOG( LOG_START, ( "m681param: started dev=%d\n", dev ));

	duart = dev/2;

	addr = m681_addr[duart];

	if ( dev & 1 ) {
		port = B_PORT;
		ptaddr = ( union m681_pt_io *) addr + 1;
	}
	else {
		port = A_PORT;
		ptaddr = ( union m681_pt_io *) addr;
	}

	tp = &m681_tty[dev];
	flags = tp->t_cflag;

	LOG( LOG_INTERN, ( "m681param: tp=0x%x, flags=0x%x\n", tp, flags) );

	/* Reset to make global changes*/
	/* disable Tx and Rx */
	/* ptaddr->write.wr_cr = TXDIS | RXDIS; */
	if (port == A_PORT)
		m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr & ~(ITXRDYA | IRXRDYA);
	else
		m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr & ~ITXRDYB;
	addr -> write.wr_imr = m681_sv_reg[duart].sv_imr;

	/* hang up on zero baud rate */
	if ( ( flags & CBAUD ) == B0 ) {
		m681modem(dev,OFF);
		return;
	}
	else {
		/* set baudrate */
		indx=(flags&CBAUD);
		speeds = m681_speeds[indx];
		if ( speeds == NOBAUD ) {
			u.u_error = EINVAL;
			speeds = DEFBAUD;
		}
		ptaddr->write.wr_csr = speeds;
		m681_sv_reg[duart].sv_csr[port] = speeds;
	}


	/* get saved mode registers and clear set up parameters */
	mr1 = m681_sv_reg[duart].sv_mr1[port];
	mr1 &= ~( CLMASK | PARTYPEMASK | PARMODEMASK );

	mr2 = m681_sv_reg[duart].sv_mr2[port];
	mr2 &= ~SBMASK;


	/* set up character size */
	if ((flags & CS8) == CS8)
	{
		mr1 |= CL8;
	}
	else if (flags & CS7)
	{
		mr1 |= CL7;
	}
	else if (flags & CS6)
	{
		mr1 |= CL6;
	}
	else	/* CS5 */
	{
		mr1 |= CL5;
	}


	/* set up stop bits */
	if (flags & CSTOPB)
	{
		mr2 |= SB2;
	}
	else
	{
		if ( flags & CS5 ) {
			/* special set up if 5 bits per character */
			mr2 |= SB1L5;
		}
		else {
			mr2 |= SB1;
		}
	}


	/* set up parity */
	if ( flags & PARENB ) {
		mr1 |= PAREN;
		if ( flags & PARODD ) {
			mr1 |= ODDPAR;
		}
		else {
			mr1 |= EVENPAR;
		}
	}
	else {
		mr1 |= PARDIS;
	}


	if ( ( m681_sv_reg[duart].sv_mr1[port] != mr1 )
	    || ( m681_sv_reg[duart].sv_mr2[port] != mr2 ) ) {
		/* write mode registers to duart */
		ptaddr->write.wr_cr = MRRESET;
		ptaddr->write.wr_mr = mr1;
		ptaddr->write.wr_mr = mr2;

		LOG( LOG_INTERN, ( "m681param: set mr1 to 0x%x, mr2 to 0x%x\n",
		    mr1, mr2 ) );

		/* save changed mode registers */
		m681_sv_reg[duart].sv_mr1[port] = mr1;
		m681_sv_reg[duart].sv_mr2[port] = mr2;
	}


	/* enable receiver? */
	if ( flags & CREAD ) {
		/* ptaddr->write.wr_cr = RXEN; */
		if (port == A_PORT)
			m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | IRXRDYA;
		else
			m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | IRXRDYB;
		addr -> write.wr_imr = m681_sv_reg[duart].sv_imr;
	}

	/* re-enable the transmitter */
	DELAY_CR;
	/* ptaddr->write.wr_cr = TXEN; */
	if (port == A_PORT)
		m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | ITXRDYA;
	else
		m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | ITXRDYB;
	addr -> write.wr_imr = m681_sv_reg[duart].sv_imr;


	LOG( LOG_RETURN, ( "m681param: return okay\n" ) );

	return;
}



/*********************************************************************
 *
 *  Function:	  m681scan
 *
 *  Description:  Scanner of the driver
 *
 *
 *  Syntax:	  m681scan()
 *
 *  Input:	  
 *
 *  Output:       
 *
 *
 *  Called by:	  m681init() and from timeout scheduler
 *
 *  Calls to:     wakeup()
 *		  timeout()
 *		  line switch  l_input()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681scan()
 *
 *  BEGIN
 *
 *	FOR  all devices 
 *	DO
 *	    IF  DCD scan is active
 *	    THEN
 *		read DCD input;
 *		IF  DCD is active
 *		THEN
 *		    wakeup waiting routine [ wakeup() ];
 *		    set carrier flag;
 *		    clear DCD scan flag;
 *		ENDIF
 *	    ENDIF
 *
 *	    IF  Input Character Flag is active
 *	    THEN
 *		move characters form receive buffer to input queue
 *					[ l_input() ];
 *	    ENDIF
 *
 *	ENDFOR
 *
 *	RETURN;
 *		
 *  END
 *
 *********************************************************************/


m681scan()
{
	register struct tty	*tp;
	register int	dev;
	register char	*workp;
	int dcdstate;		/* status of DCD input */

	for ( dev = 0, workp = m681_work; dev < m681_cnt; dev++, workp++ ) {

			/* read DCD input */
			if ( dev & 1 ) {	/* Port B */
				dcdstate = m681_addr[dev/2]->read.rd_ip & IPDCDB;
			}
			else {	        /* Port A */
				dcdstate = m681_addr[dev/2]->read.rd_ip & IPDCDA;
			}
#if defined (ASPEN) || defined (ANGEL)
			dcdstate = 0;   /* No dcd line on aspen */
#endif
			tp = &m681_tty[dev];

			if (*workp & M681_DCD) {	/* DCD checking */
	
				if ( dcdstate == 0 ) {
					wakeup( &tp->t_canq );
					tp->t_state |= CARR_ON;
					m681_work[dev] &= ~M681_DCD;
					LOG( LOG_INTERN,
					    ("m681scan: dev=%d, DCD found\n", dev ) );
				}
			}
			else { /* if we lost DCD, and the device was open, signal */
	
				if ( !(dcdstate == 0) ) {
					
					if ( !( tp->t_cflag & CLOCAL ) ) {
						
						if ( tp->t_state & CARR_ON ) {
	
							if ( tp->t_state & ISOPEN ) {
	
								signal (tp->t_pgrp,SIGHUP);
								ttyflush (tp, (FREAD | FWRITE) );
							}
							tp->t_state &= ~CARR_ON;
						}
					}
				}
			}

		if (*workp & M681_CHAR) {

			/* move characters from receive buffer to input queue */
			tp = &m681_tty[dev];
			*workp = 0;
			if (tp->t_rbuf.c_ptr) {
				tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size
				    - tp->t_rbuf.c_count;
				(*linesw[tp->t_line].l_input)(tp, L_BUF);
			}
		}
		if (*workp & M681_ACT)
			*workp |= M681_CHAR;

	}
	/* restart scanner */
	timeout( m681scan, (struct tty *) 0, M681CSCAN);

	return;
}


/*********************************************************************
 *
 *  Function:	  m681modem
 *
 *  Description:  set or read modem lines
 *
 *
 *  Syntax:	  m681modem( dev, flag )
 *
 *  Input:	  int dev;	minor device number
 *		  int flag;	flag
 *
 *  Output:       0		if  DCD not found
 *  		  1		if  DCD found
 *
 *
 *  Called by:	  m681open() or m681close()
 *
 *  Calls to:     m681scan()
 *		  m681modem()
 *		  m681param()
 *		  ttinit()
 *		  line switch  l_open()
 *		  sleep()
 *		  SPLCLOCK()
 *		  splx()
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681modem()
 *
 *  BEGIN
 *
 *	IF  flag is OFF  ( hang up  )
 *	THEN
 *	    reset DTR and RTS outputs;
 *	    IF  not CLOCAL mode
 *	    THEN
 *		send hangup signal to process group
 *	    ELSE
 *		IF  no carrier
 *		THEN
 *		    wakeup waiting processi, who is waiting on carrier
 *							[ wakeup() ];
 *		ENDIF
 *	    ENDIF
 *	ELSE
 *	    set DTR and RTS outputs;
 *	ENDIF
 *
 *	read DCD input;
 *
 *	IF  DCD found
 *	THEN 
 *	    set carrier flag;
 *	    RETURN  1;
 *	ELSE
 *	    RETURN  0;
 *	ENDIF
 *
 *  END
 *
 *********************************************************************/


m681modem(dev,flag)
register int dev;
int flag;
{

	register union m681_io	*addr;
	register union m681_pt_io	*ptaddr;
	int duart, port;

	register unsigned char *csr;
	int cur_lvl;
	register struct tty	*tp;
	int dcdstate, dcdflag;

	LOG( LOG_START, ( "m681modem: started dev=%d, flag=0x%x\n", 
	    dev, flag ));

	duart = dev/2;

	addr = m681_addr[duart];

	if ( dev & 1 ) {
		port = B_PORT;
		ptaddr = ( union m681_pt_io *) addr + 1;
	}
	else {
		port = A_PORT;
		ptaddr = ( union m681_pt_io *) addr;
	}

	tp = &m681_tty[dev];
	cur_lvl=SPLCLOCK();

	if ( flag==OFF ) {			/* hang up */

		/* reset DTR and RTS outputs */
		/* reset port A and port B for mvme335 */
		/*  BUT */
		/* only port B for the 135/136 */
		/* don't do it if you want software abort to work on the 135 */
#ifndef XREGS
		if ( port == A_PORT ) {
			addr->write.wr_oprreset = OPDTRA | OPRTSA;
		}
		else {	/* Port B */
			addr->write.wr_oprreset = OPDTRB | OPRTSB;
		}
#else
		if ( port == B_PORT )
			addr->write.wr_oprreset = OPDTRB | OPRTSB;
#endif

		if ( !(tp->t_cflag & CLOCAL) ) {
			/* send HANG UP signal to process group */
			if ( tp->t_state & CARR_ON ) {
				if ( tp->t_state & ISOPEN ) {
					signal(tp->t_pgrp,SIGHUP);
					ttyflush(tp,(FREAD|FWRITE));
				}
				tp->t_state &= ~CARR_ON;
			}
		}
		else {
			/* CLOCAL mode: carrier is not dependent of hardware */
			if ( !(tp->t_state & CARR_ON) ) {
				wakeup( &tp->t_canq );
				tp->t_state |= CARR_ON;
			}
		}
	}
	else {	/* ON MODE */
		/* set DTR and RTS outputs */
		if ( port == A_PORT ) {
			addr->write.wr_oprset = OPDTRA | OPRTSA;
		}
		else {	/* Port B */
			addr->write.wr_oprset = OPDTRB | OPRTSB;
		}
	}


	/* read DCD input */
	if ( port == A_PORT ) {
		dcdstate = addr->read.rd_ip & IPDCDA;
	}
	else {	/* Port B */
		dcdstate = addr->read.rd_ip & IPDCDB;
	}

#if defined (ASPEN) || defined (ANGEL)
	dcdstate = 0;   /* No dcd line on aspen */
#endif
	if ( dcdstate == 0 ) {
		/* DCD found */
		tp->t_state |= CARR_ON;
		dcdflag = 1;
	}
	else {
		/* DCD not found */
		dcdflag = 0;
	}

	splx(cur_lvl);

	LOG( LOG_RETURN, ( "m681modem: return %d\n", dcdflag ) );

	return( dcdflag );

}


/*********************************************************************
 *
 *  Function:	  m681sendchar
 *
 *  Description:  send a character
 *
 *
 *  Syntax:	  m681sendchar( tp, ptaddr, c,addr,port )
 *
 *  Input:	  struct tty *tp;	    TTY structure
 *		  union m681_pt_io *ptaddr; Port address
 *		  char  c;		    character
 *     	   	  union m681_io	*addr;
 *		  int	port;
 *
 *  Output:       1	if character was sent
 *		  0     if character was not sent
 *
 *
 *  Called by:	  m681proc()
 *
 *  Calls to:     
 *
 *********************************************************************/


/*********************************************************************
 *
 *  PROCEDURE:  m681sendchar()
 *
 *  BEGIN
 *
 *	set BUSY flag in tty structure;
 *	enable transmitter;
 *	
 *	IF  transmitter is ready
 *	THEN
 *	    transmit character;
 *	    RETURN  1;
 *	ELSE
 *	    RETURN  0;
 *	ENDIF
 *
 *  END
 *
 *********************************************************************/


m681sendchar( tp, ptaddr, c, addr, port, duart )
struct tty	  *tp;
union m681_pt_io  *ptaddr;
char c;
union m681_io	*addr;
int	port, duart;
{
	int sr;

	LOG( LOG_START, ( "m681sendchar: started t_state=0%o, addr=0x%x, char=0x%x\n",
	    tp->t_state, ptaddr, c ) );


	tp->t_state = BUSY | tp->t_state ;	/* send char */

	/* enable transmitter */
	/* ptaddr->write.wr_cr = TXEN; */
	if (port == A_PORT)
		m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | ITXRDYA;
	else
		m681_sv_reg[duart].sv_imr = m681_sv_reg[duart].sv_imr | ITXRDYB;
	addr -> write.wr_imr = m681_sv_reg[duart].sv_imr;

	sr = ptaddr->read.rd_sr;
	if ( ptaddr->read.rd_sr & TXRDY ) {
		ptaddr->write.wr_tb = c;
		LOG( LOG_RETURN, ( "m681sendchar: return okay\n" ) );

		return( 1 );
	}
	else {
		LOG( LOG_WARN,
		    ( "m681sendchar: Warning: T_OUTPUT: TXRDY not set\n" ) );
		LOG( LOG_WARN, ( "m681sendchar: t_state=0%o, addr=0x%x, char=0x%x, sr=0x%x\n",
		    tp->t_state, ptaddr, c, sr ) );
		return(0);
	}
}

/*********************************************************************
 *
 *  Function:	  m681ckdcd
 *
 *  Description:  check for dcd interrupt
 *
 *
 *  Syntax:	  m681ckdcd( duart, ipcr, ip )
 *
 *  Input:	  unsigned int duart;	    duart index
 *		  unsigned char  ipcr;		image of input change register
 *		  unsigned char  ip;		image of input register
 *
 *  Output:      
 *
 *
 *********************************************************************/
m681ckdcd (duart,ipcr,ip)
unsigned int duart;
unsigned char ipcr;
unsigned char ip;
{
	register struct tty 	*tp;
	int dcdstate;		/* status of DCD input */
	unsigned int dev;

	if (ipcr & IPDCDIA ) { /* port A */
		dev = duart * 2;
		tp = &m681_tty[dev];
		dcdstate = ip & IPDCDA;
		if (m681_work[dev] & M681_DCD) {
			if ( dcdstate == 0) {
				if (Dconcurrent)
					if (m681_state[dev] & DCON) {
						consinputc(0, DC_DCDON);
					}	
				wakeup ( &tp->t_canq );
				tp->t_state |= CARR_ON;
				m681_work[dev] &= ~M681_DCD;
			}
		} 

		else { /* if we lost DCD, and the device was open, signal */

			if ( !(dcdstate == 0) ) {
				
				if (Dconcurrent)
					if (m681_state[dev] & DCON) {
						consinputc(0, DC_DCDOFF);
					}	

				if ( !( tp->t_state & CLOCAL ) ) {
					
					if ( tp->t_state & CARR_ON ) {

						if ( tp->t_state & ISOPEN ) {

							signal (tp->t_pgrp,SIGHUP);
							ttyflush (tp, (FREAD | FWRITE) );
						}
						tp->t_state &= ~CARR_ON;
					}
				}
			}
		}

	}
	if (ipcr & IPDCDIB ) { /* port B */
		dev = duart * 2 + 1;
		tp = &m681_tty[dev];
		dcdstate = ip & IPDCDB;
		if (m681_work[dev] & M681_DCD) {
			if ( dcdstate == 0) {
				if (Dconcurrent)
					if (m681_state[dev] & DCON) {
						consinputc(0, DC_DCDON);
					}	
				wakeup ( &tp->t_canq );
				tp->t_state |= CARR_ON;
				m681_work[dev] &= ~M681_DCD;
			}
		} 

		else { /* if we lost DCD, and the device was open, signal */

			if ( !(dcdstate == 0) ) {

				if (Dconcurrent)
					if (m681_state[dev] & DCON) {
						consinputc(0, DC_DCDOFF);
					}	
				
				if ( !( tp->t_state & CLOCAL ) ) {
					
					if ( tp->t_state & CARR_ON ) {

						if ( tp->t_state & ISOPEN ) {

							signal (tp->t_pgrp,SIGHUP);
							ttyflush (tp, (FREAD | FWRITE) );
						}
						tp->t_state &= ~CARR_ON;
					}
				}
			}
		}
	}
	
}

/*********************************************************************
 *
 *  Title:	  MC68681  Putchar Module
 *
 *  Description:  This is a UNIX SYSTEM V/68 serial character I/O 
 *	          driver. 
 *
 *		  The driver communicates with the MC68681 DUARTs 
 *	          This module will be linked in if the "force M681con"
 *		  statement appears in master.	
 *
 *
 *
 *********************************************************************/
char X681con;	/* unique identifier so putchar can be linked */
extern struct m681_sv_reg m681_sv_reg[];
m681putc(dev, c)
register unsigned int dev;
register char c;
{
	register union m681_io	*addr;
	register union m681_pt_io  *ptaddr;
	int cur_lvl;


	cur_lvl = SPLM681();		/* inhibit interrupts */
	addr = m681_addr[dev/2];
	if ( dev & 1 ) {
		ptaddr = ( union m681_pt_io *) addr + 1;
	}
	else {
		ptaddr = ( union m681_pt_io *) addr;
	}

   	/* Assume first port initialized if we get here. */
	/* Assume the bug initializes the port */

	/* inhibit interrupts on the chip */
	addr->write.wr_imr = m681_sv_reg[0].sv_imr & ~ITXRDYA;
	/* make sure transmtter is enabled */
	DELAY_CR;
	ptaddr->write.wr_cr = TXEN;

	/* If the character is a line feed(\n) */
	/* then follow it with carriage return (\r) */
	for ( ;; )
	{
		while ( !(ptaddr->read.rd_sr & TXRDY) );
		ptaddr->write.wr_tb = c;
		if ( c != '\n' )
			break;
		c = '\r';
	}
	/* wait for transmitter to empty */
	while ( !(ptaddr->read.rd_sr & TXEMT) );

	/* restore the previous state */
	addr->write.wr_imr = m681_sv_reg[0].sv_imr;
	DELAY_CR;
	ptaddr->write.wr_cr = m681_sv_reg[0].sv_cr[A_PORT];

	splx( cur_lvl);
	return;

}
