/*	@(#)mem143.c	35.6	*/
/*
 * MVME143 DRAM related routines.  
 *
 * COPYRIGHT 1988 BY MOTOROLA INC.
 */

#include "sys/param.h"
#include "sys/types.h"
#include "sys/utsname.h"
#include "sys/elog.h"
#include "sys/erec.h"
#include "sys/cmn_err.h"
#include "sys/mvmecpu.h"
#include "sys/mvme143.h"

extern int parityon, paritymode;
extern int cpuboard;

void
m143init()
{
	register char *pcdr_addr, *pbcr_addr, *stat_addr, status;

	if(cpuboard != VME143)	/* Be sure on vme143 board */
		return;

	pbcr_addr = V143_PBCR;
	pcdr_addr = V143_PCDR;
	  /* disable parity interrupt */
	*pbcr_addr &= ~V143_PARENB;
	  /* clear previous enables */
	*pcdr_addr |= V143nVPAREN | V143nPAREN;
	  /*
	   * read and then write back status register
	   * to clear any interrupts pending
	   */
	stat_addr = V143_STAT;		/* parity status register address */
	status = *stat_addr;
	*stat_addr = status;
	if (parityon) {
		if (paritymode == 2) {	/* enable parity interrupt */
			/*
			 * fast mode (no extra cycles)
			 * but nongraceful when we catch one
			 */
			  /* enable parity error interrupt */
			*pbcr_addr |= V143_PARENB;
		}
		else if (paritymode == 3) {
			/*
			 * slower mode (1 extra cycle on reads only)
			 * and graceful when we catch one
			 */
			  /* set both enables (bus error interrupt) */
			*pcdr_addr &= ~(V143nVPAREN | V143nPAREN);
		}
	}
}

/* m143parity()
 *
 * Check for parity error conditions.
 * If there is one then:
 *	- If the parity error is not a repeat of the last parity condition,
 *	  i.e., the error location is different from the last parity error
 *		* save this error's location
 *		* reset parity error instance counter to zero.
 *	- Inform the console of the error condition if the parity error
 *	  instance counter is equivalent to zero modulo V143_MINCNT.
 *	- Increment parity error instance counter.
 *	- Log the error.
 *	- if this was a user mode parity error, then, attempt to fix the error:
 *		* Disable bus interrupt on parity error.
 *		* Read, then, write the 4 bytes starting at divisible by 4 loc.
 *		* Re-enable bus interrupt on parity error.
 * Clear the parity error status bit.
 * Return non-zero if parity error condition was indicated.
 */
/* mode argument values */
#define KERNEL 0
#define USER 1

int
m143parity(loc, mode)
char *loc;
int mode;
{
	extern void logmemory();

	register char *p;
	register int i, j, ch, status;

	static char *modename[] = { "Kernel", "User" };
	static char *parity_msg =
		"MVME143 %s Mode Parity Error; location = 0x%x\n";

	static char *lloc = (char *)-1;	/* last err location */
	static int count = 0;		/* similiar counter */
	register char *stat_addr, *enb_addr, svenb;

	if(cpuboard != VME143)	/* Be sure on vme143 board */
		return(0);

	stat_addr = V143_STAT;		/* parity status register address */
	enb_addr = V143_PCDR;		/* parity error enable address */
	status = 0;
	if (*stat_addr & V143_PARERR) {
		  /* write 1's to clear */
		*stat_addr |= V143_PARERR;
		if (lloc != loc) {
			lloc = loc;
			count = 0;
		}
		if ((count & V143_MINCNT) == 0)
			cmn_err(CE_WARN, parity_msg,
				modename[mode], loc);
		count++;

		logmemory(0, *stat_addr, E_UME);

		if (mode == USER) {
			  /* save both parity error enable values */
			svenb = *enb_addr & (V143nVPAREN | V143nPAREN);
			  /* disable both parity error enable values */
			*enb_addr |= V143nVPAREN | V143nPAREN;
			  /* make sure that we start on divisible by 4 */
			p = (char *)((int)loc & ~3);
			for (j = 0; j < 4; j++) {
				ch = fubyte(p);
				subyte(p, ch);
				p++;
			}
			  /* restore both parity error enable values */
			*enb_addr = (*enb_addr & ~(V143nVPAREN | V143nPAREN))
			    | svenb;
		}
		status--;
	}
	return(status);
}
