/*	@(#)mem147.c	35.4	*/
/*
 * MVME147 DRAM related routines.  
 * Also see "sys/pcc.h" for further information.
 *
 * COPYRIGHT 1988 BY MOTOROLA INC.
 */

#include "sys/param.h"
#include "sys/types.h"
#include "sys/utsname.h"
#include "sys/elog.h"
#include "sys/erec.h"
#include "sys/cmn_err.h"
#include "sys/pcc.h"
#include "sys/mvmecpu.h"

/* minimum number of repeated corrected errors
 * before display of message on system console device
 */
#define V147_MINCNT	15		/* Minimum count for parity errors */

extern 	struct	pcc	*pcc_addr[];
extern int parityon, paritymode;
extern int cpuboard;

void m147init()
{
	register struct pcc *addr;

	if(cpuboard != VME147)		/* only on 147 board */
		return;

	addr = pcc_addr[0];
	addr->gpcr &= ~PCC_PARMSK;	/* clear previous mode setting */
	if (parityon) {
		if (paritymode == 2)
			addr->gpcr |= PCC_PMODE2;
		else if (paritymode == 3)
			addr->gpcr |= PCC_PMODE3;
	}
}

/* m147parity()
 *
 * Check for parity error conditions.
 * If there is one then:
 *	- If the parity error is not a repeat of the last parity condition,
 *	  i.e., the error location is different from the last parity error
 *		* save this error's location
 *		* reset parity error instance counter to zero.
 *	- Inform the console of the error condition if the parity error
 *	  instance counter is equivalent to zero modulo V147_MINCNT.
 *	- Increment parity error instance counter.
 *	- Log the error.
 *	- if this was a user mode parity error, then, attempt to fix the error:
 *		* Disable bus interrupt on parity error.
 *		* Read, then, write the 4 bytes starting at divisible by 4 loc.
 *		* Re-enable bus interrupt on parity error.
 * Clear the parity error status bit.
 * Return non-zero if parity error condition was indicated.
 */
/* mode argument values */
#define KERNEL 0
#define USER 1

int m147parity(loc, mode)
char *loc;
int mode;
{
	extern void logmemory();

	register unsigned char svmode;
	register char *p;
	register int i, j, ch, status;

	static char *modename[] = { "Kernel", "User" };
	static char *parity_msg =
		"MVME147 %s Mode Parity Error; location = 0x%x\n";

	static char *lloc = (char *)-1;			/* last err location */
	static int count = 0;				/* similiar counter */
	register struct pcc *addr;

	if(cpuboard != VME147)		/* only on 147 board */
		return(0);

	addr = pcc_addr[0];
	status = 0;
	if (addr->gpsr & PCC_PARERR) {
		addr->gpsr |= PCC_PARERR;	/* clear parity error status */
		if (lloc != loc) {
			lloc = loc;
			count = 0;
		}
		if ((count & V147_MINCNT) == 0)
			cmn_err(CE_WARN, parity_msg,
				modename[mode], loc);
		count++;

		logmemory(0, addr->gpsr, E_UME);

		if (mode == USER) {
			svmode = addr->gpcr & PCC_PARMSK; /* save parity mode */
			addr->gpcr &= ~PCC_PARMSK; /* disable parity checking */
			  /* make sure that we start on divisible by 4 */
			p = (char *)((int)loc & ~3);
			for (j = 0; j < 4; j++) {
				ch = fubyte(p);
				subyte(p, ch);
				p++;
			}
			addr->gpcr |= svmode;	/* restore parity mode */
		}
		status--;
	}
	return(status);
}
