/*	Copyright (c) 1987 Motorola Inc
	All Rights Reserved
	@(#)kern-port:io/mvme050.c	35.3
*/
#ident	"@(#)kern-port:io/mvme050.c	35.3"

/*
 *	This file initializes the Motorola mvme050 system
 *	controller, time of day clock functionality, and
 *	handles I/O to the front panel display and switches.
 */
#include "sys/mc68153.h"
#include "sys/types.h"
#include "sys/signal.h"
#include "sys/dir.h"
#include "sys/param.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/todc.h"

#define  BIM1		0xc0		/* bim 1 offset */
#define  BIM2		0xe0		/* bim 2 offset */

#define  DISPLAY_ADDR   0xa0		/* display address offset */
#define  SWITCH_ADDR    0xa0		/* switch address offset */
#define  BLANK_ADDR	0x8a		/* blank address offset */
#define  BLANK	 	0x0		/* blank out display register */
#define  UNBLANK	0x8		/* unblank display register */

#define  BLANK_CMD	1		/* turn off display */
#define	 TODC_OFFSET	0x100-1 	/* -1 as m050_addr is an odd address */


m050init ()
{
	extern int	m050_addr[];
	extern int	m050_ivec[];
	extern int	m5_ivec[];

	extern unsigned char  tdcinfo;		/* todc exist/valid flag */
	extern unsigned char  *tdcaddr;		/* address of todc */
	extern unsigned char  have_050; 	/* existence flag for MVME050 */

	register vec = m050_ivec[0]/4;		/* get base vector # */
	register struct bim *bim_ptr;


	/*
	 *  If MVME050 is not accessible, zero it's address, otherwise
	 *  check for need to setup the 050 time of day clock.  If the
	 *  system contains a cpu-board resident ("local") TODC, then
	 *  the 050 TODC is not used.
	 */
	if( bprobe( m050_addr[0], -1 ) ) {
		m050_addr[0] = 0;		/* Mark board not present */
		return;
	} else {
		have_050 = 1;			/* Mark board accessible */
		if (!IS_TODC(tdcinfo)) {	/* No local TODC; setup 050's */
			tdcinfo |= TODC_IS;
			tdcaddr = (unsigned char *) m050_addr[0] + TODC_OFFSET;
		}
	}

	bim_ptr = (struct bim *) (m050_addr[0] + BIM1);

	/* special case for serial ports */
	bim_ptr->vctreg0 = m5_ivec[0]/4;
	bim_ptr->vctreg1 = m5_ivec[1]/4;
	bim_ptr->crlreg0 = (EXT | IRE | 5);
	bim_ptr->crlreg1 = (EXT | IRE | 5);

	/* general case */
	vec++;
	vec++;
	bim_ptr->vctreg2 = vec++;
	bim_ptr->vctreg3 = vec++;

	bim_ptr->crlreg2 = 0;
	bim_ptr->crlreg3 = 0;

	bim_ptr = (struct bim *) (m050_addr[0] + BIM2);

	bim_ptr->vctreg0 = vec++;
	bim_ptr->vctreg1 = vec++;
	bim_ptr->vctreg2 = vec++;
	bim_ptr->vctreg3 = vec;

	bim_ptr->crlreg0 = 0;		/* disabled */
	bim_ptr->crlreg1 = 0;		/* disabled */
	bim_ptr->crlreg2 = 0;		/* disabled */
	bim_ptr->crlreg3 = 0;		/* disabled */
}


m050read (dev)
int	dev;
{
	extern int	m050_addr[];

	if (m050_addr[0] == 0) {	/* Is board present? */
		u.u_error = EIO;
		return;
	}

	passc (*(char *) (m050_addr[0] + SWITCH_ADDR));
}


m050write (dev)
int	dev;
{
	extern int	m050_addr[];
	register int	c;

	if (m050_addr[0] == 0) {	/* Is board present? */
		u.u_error = EIO;
		return;
	}

	if ( (c = cpass()) != -1 )
		*(char *) (m050_addr[0] + DISPLAY_ADDR) = c;
}


m050ioctl (dev, cmd, arg, mode)
int	dev, cmd, mode;
char	*arg;
{
	extern int	m050_addr[];
	register char	*cp;

	if (m050_addr[0] == 0) {	/* Is board present? */
		u.u_error = EIO;
		return;
	}

	cp = (char *) (m050_addr[0] + BLANK_ADDR);

	if (cmd == BLANK_CMD)
		*cp = BLANK;
	else
		*cp = UNBLANK;
}

m050intr(r0,r1,r2,r3,r4,r5,r6,r7,r8,r9,r10,r11,r12,r13,r14,ps,pc,sform)
short	sform;
{	register unsigned vec = sform&0xfff;
	register unsigned ivec;
	register struct bim *bim_ptr;
	register char *bp;
	extern m050_ivec[];

	ivec = vec;
	ivec /= 4;

	vec -= m050_ivec[0];
	vec /= 4;
	/* vec == interrupt # */

	if( vec >= 8 )
		return;
	
	if( vec < 4 )
		bim_ptr = (struct bim *) (m050_addr[0] + BIM1);
	else {
		vec -= 4;
		bim_ptr = (struct bim *) (m050_addr[0] + BIM2);
	}
	switch(vec) {
	case 0:	bp = &bim_ptr->crlreg0;	break;
	case 1:	bp = &bim_ptr->crlreg1;	break;
	case 2:	bp = &bim_ptr->crlreg2;	break;
	case 3:	bp = &bim_ptr->crlreg3;	break;
	}

	if(( *bp&IRAC ) == 0 ) {	/* not setup by m050prime? */
		printf("MVME050: unexpected interrupt on vector %x\n", ivec );
		*bp = 0;	/* disable further interrupts */
	}
}

m050prime(vec,vector,control)
register unsigned vec;
register unsigned vector;
register unsigned control;
{	register struct bim *bim_ptr;
	register char *bp;

	if (m050_addr[0] == 0)		/* Is board present? */
		return;

	if( vec >= 8 )
		return;

	if( vec < 4 )
		bim_ptr = (struct bim *) (m050_addr[0] + BIM1);
	else {
		vec -= 4;
		bim_ptr = (struct bim *) (m050_addr[0] + BIM2);
	}

	if( vector ) {
		switch(vec) {
		case 0:	bp = &bim_ptr->vctreg0;	break;
		case 1:	bp = &bim_ptr->vctreg1;	break;
		case 2:	bp = &bim_ptr->vctreg2;	break;
		case 3:	bp = &bim_ptr->vctreg3;	break;
		}
		*bp = vector;
	}

	switch(vec) {
	case 0:	bp = &bim_ptr->crlreg0;	break;
	case 1:	bp = &bim_ptr->crlreg1;	break;
	case 2:	bp = &bim_ptr->crlreg2;	break;
	case 3:	bp = &bim_ptr->crlreg3;	break;
	}
	*bp = control;
}
