
/*		All Rights Reserved			*/
/*		@(#)mvme147dk.c	35.7				*/
/****************************
 *			    *
 *	mvme147 Driver:     *
 *     			    *
 ***************************/

/*******************************************************************************
 *
 *	Include files
 *
 ******************************************************************************/

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/elog.h"
#include "sys/iobuf.h"
#include "sys/systm.h"
#include "sys/dk.h"
#include "sys/dsk.h"
#include "sys/scsi.h"
#include "sys/mvme147s.h"
#include "sys/open.h"
#include "sys/sysmacros.h"
#include "sys/immu.h"
#include "sys/cmn_err.h"
#include "sys/mvmecpu.h"
#include "sys/disjointio.h"


#undef	DEBUG	
#undef	DEBUGC
#undef SANITY
#undef	FWDEBUG	
#undef CLNPKTS
#undef JUNK
#undef REALRAWIO

#define ULONG unsigned long
#define UCHAR unsigned char

#undef TRACE
#include "sys/mvme336t.h"
	
#ifdef FWDEBUG
#define FWSIZE 8192
unsigned char fwtrc[FWSIZE];
int sent;
#endif

/* coalescing enable flag */
extern int m147coal;

/* coalescing count print */
int m147scgacnt = 65;			

#define M147 10
#define DISKATTR	DISK1.disk_attr

/*  maximum number of outstanding packets per device */
int m147lim = M147LIM; 

#ifdef	DEBUG		/* debugging code */
#define STATIC
#define	DEBUGPR(x,d,s)	{ if (x & d147Debug) d147print(d,s); }

int d147Debug = 0x0000;

#define	DBINIT		(1<<0)
#define	DBBOPEN		(1<<1)
#define	DBCOPEN		(1<<2)
#define	DBBCLOSE	(1<<3)
#define	DBCCLOSE	(1<<4)
#define	DBSTRAT		(1<<5)
#define	DBSTART		(1<<6)
#define	DBCONF		(1<<7)
#define	DBFORMAT	(1<<8)
#define	DBICMD		(1<<9)
#define	DBINTR		(1<<10)
#define	DBERR		(1<<11)
#define	DBREAD		(1<<12)
#define	DBWRITE		(1<<13)
#define	DBIOCTL		(1<<14)
#define	DBCLR		(1<<15)
#define	DBTIMER		(1<<16)
#define	DBDUMP		(1<<17)
#define	DBDOIO		(1<<18)

static char *ASSERTCTL = "ASSERT(%s) failed in %s, line %d\n";
#define ASSERT(X) { if (!(X)) printf(ASSERTCTL, "X", __FILE__, __LINE__);}

#else			/* no debugging code */

#define	STATIC	static
#define	DEBUGPR(x,d,s)
#define ASSERT(X)

#endif

#ifdef	DEBUGC		/* command debugging code */

#define	PRCMD(p)	{ if (d147Cmdbug) d147prcmd(p); }
int d147Cmdbug = 1;

#else

#define	PRCMD(p)	/* no command debugging code */

#endif

#ifdef TRACE

#define DTRAIL(x) (('D'<<24)|('>'<<16)|('>'<<8)|('x'))


struct trace m147trcmem[MAX_TRC]; 

int d147_trace = 0;

extern chktrc();

#define DTRC(id,a,b,c,d,e,f,g) if(d147_trace){chktrc(id,a,b,c,d,e,f,g);}

struct trace *m147trace;
struct trace *m147trcbase;

#else

#define DTRC(id,a,b,c,d,e,f,g) 
#define DTRAIL(x)

#endif

/* RTS on */

/*******************************************************************************
 *
 *	Each MVME147 can support up to 7 SCSI controllers.
 *	Each SCSI contoller can control up to 8 drives simultaneously:
 *
 ******************************************************************************
 *
 *	MVME147 Disk Controller Driver Major/Minor Number Assignment Coding
 *
 *               Major                 Minor
 *      <-  5  -> <- 3 -> <- 3 -> <  -1  -> <-   4    ->
 *      ----------------- ------------------------------
 *      |        | SCSI  | perph |          |          |
 *      |        | addr  | drive | floppy   |   slice  |
 *      ----------------- ------------------------------
 *
 *
 *	SCSI addr = logical address of SCSI controller
 *	perph drive = logical drive number
 *	floppy bit = this device is a floppy


	The major/minor number is used to access a 
table built at configuration time (see m147space.h).
This table contains the default configuration paramaters for
each SCSI drive.

There will be 7 major numbers with 8 devices per major number.
Major number 0 will be defined as a disk, major number 4 will be
defined as tape...


These are the default devices:

0	disk
1	disk
2	disk
3	disk
4	tape
5	tape
6	floppy
7	unassigned

 ******************************************************************************/


/*******************************************************************************
 *
 *	Major/Minor Number Manipulating Macros
 *
 ******************************************************************************/

#define	SLCMSK		((1<<M147SBTS)-1)
#define	SLICE(d)	(d&SLCMSK)		/* slice number */
#define	MKDEV(c,p)	((c<<8)|(p<<5)) /* scsi adr| scsi device */
#define SCBUS(d)	((d & 0x0800) >> 11)
#define SCADR(d)	((d & 0x0700) >> 8)
#define SCDEV(d)	((d & 0x00e0) >> 5)
/* MKPUNIT takes scsi ctl, scsi perf nibbles and makes punit */
#define MKPUNIT(d)	(((d & 0x0700) >> 4) | ((d & 0x00e0) >> 5))
#define	POS(d)	((SCADR(d)*m147maxscdev) + SCDEV(d))
#define HIGHSLICE(x)	((SLICE(x) == 3) || (SLICE(x) == 4))
#define LOWSLICE(x)	((SLICE(x) == 0) || (SLICE(x) == 7))
#define FLOPBIT		0x10
#define M147FLOP(x)	(x & FLOPBIT)




/*******************************************************************************
 *	MVME147 Winchester & Floppy Default Configuration Values
 ******************************************************************************/

/* Winchester Values */

#define W_ECCFLG	1		/* ECC flag; ECC in use */
#define W_INITCHR	0x4e		/* initialization character */



/*******************************************************************************
 *
 *	Global variables
 *
 ******************************************************************************/

STATIC	unsigned d147_ctls;

#define	MAXPOS	(d147_ctls*m147maxtab)	/* maximum number of disk slots */

#define	D147SHRT	30		/* timeout limit for short commands */
#define	D147LONG	3000		/* timeout limit for long commands */


STATIC	struct dsk d147dsk;		/* disk descriptor for the mvme147 */
STATIC	struct dsk f147dsk;		/* floppy descriptor for the mvme147 */

#define	D147IOCTL	('M'<<8)



/*******************************************************************************
 *	External Variables
 ******************************************************************************/

extern unsigned char  *d147_addr[];
extern int d147_bmaj;			/* block major */
extern int d147_cnt;			/* count */
extern int d147_ilev[];			/* interrupt level */
extern int d147_ivec[];			/* interrupt level */

extern int d147otbl[];			/* open table */

/*******************************************************************************
 *	Driver Variables - declared in conf.c (space.h or io.h)
 ******************************************************************************/

extern struct iobuf		m147utab[];	/* one per drive */
extern struct iotime		m147stat[];	/* one per drive */
extern struct slice		d147slices[];	/* slice tables */
extern struct slice		d147flop;	/* default floppy slice table */
extern struct scs_pkt		m147pkt[];	/* command packets */
extern struct scs_pkt	*m147freeq;	/* list of free packets */
extern struct scs_disk 	*m147params[];	/* disk descriptor params */
extern struct scs_disk 	m147ddefs[];	/* disk open descriptor params */
extern struct work m147_work[]; /* firm-ware work area */
extern int m147maxscbus;		/* max number of configured buses */
extern int m147maxscadr;		/* max number of configured addrs */
extern int m147maxscdev;		/* max number of configured devs */
extern int m147maxtab;			/* max table size */
extern struct scs_disk m147dsdd5;	/* normal 5 1/4 inch floppy config */
extern struct scs_disk m147pcat;	/* normal 5 1/4 inch floppy config */
extern struct scga scgalist[];		/* scatter/gather list */
extern struct djntio d147djlist[];		/* disjointio list */
extern int d147maxblk;	/* maximum number of entrys in scatter/gather list */

/*******************************************************************************
 *	MVME147 iobuf usage
 *
 *	iobuf.h		mvme147.c	Usage
 *	-------		---------	-----
 *	b_flags		b_flags		Drive state
 *	b_forw
 *	b_back
 *	b_actf		b_actf		head (first) of I/O queue
 *	b_actl		b_actl		tail (last) of I/O queue
 *	b_dev		b_dev		logical major/minor device number
 *	b_active	b_active	busy flag
 *	b_errcnt	b_errcnt	error count
 *	io_erec		io_erec		error record
 *	io_nreg		io_nreg		number of registers to log upon error
 *	io_addr		io_eregs	address of registers to log on error
 *	io_mba		io_csr		address of device CSR
 *	io_stp		io_stp		I/O statistics
 *	io_start	io_start	I/O statistics
 *	io_s1		io_acts		used for sorting i/o queue      (dsk.c)
 *	io_s2		io_qcnt		number of bufs on iobuf queue	(dsk.h)
 *	io_s3		io_slccnt	number of allowable slices	(dsk.h)
 *	io_s4		io_badptr	pointer to bad block table	(dsk.h)
 *	io_s5		io_desc		disk descriptor (configuration struct)
 *	io_s6		io_pktcnt	packet count (limit & count)	(dsk.h)
 *	io_s7		io_sopen	open slices flag		(dsk.h)
 *	io_s8		io_slctbl	pointer to the slice table	(dsk.h)
 *	io_s9		io_dskptr	pointer to dsk structure	(dsk.h)
 ******************************************************************************/

#define	io_eregs	io_addr
#define	io_csr		io_mba
#define	io_desc		io_s5


/*
 *	d147init: initialize driver
 */



d147init()
{
	register unsigned char sc_adr, sc_dev;
	register n;
	register struct scs_pkt *ep;
	register struct scs_disk **dd = (struct scs_disk **)m147params;
	register struct scs_disk *tmp;
	register struct scs_disk *ddefs = &m147ddefs[0];
	struct slice *slc = &d147slices[0];
	register int device;

	int d147strategy();
	int d147start();
	int d147timer();
	int f147timer();
	int d147print();
	extern int cpuboard;

#	ifdef FWDEBUG
	char *cfwdebug = (char *)0xfffe07a6;
	int *ifwdebug = (int *)0xfffe07aa;
	

	cfwdebug++;
	*cfwdebug = (char)0x44;	

	*ifwdebug++  = (int)&fwtrc[0];
	*ifwdebug = (int)&fwtrc[FWSIZE-1];
#	endif


	DEBUGPR(DBINIT,0,"init");


	d147_ctls = d147_cnt;


	/* set up free queue of envelope/packet pairs */

	/* create the queue  if not already made */
	if(m147freeq == (struct scs_pkt *)NULL)
	{
		m147freeq = &m147pkt[0];
		m147crque(m147freeq,M147PKTS);
	}

	/* check to see if we're on the correct processor */
	if(cpuboard != VME147)
	{
		d147_addr[0] = NULL;
		return;
	}

	for( sc_adr = 0; sc_adr < m147maxscadr; sc_adr++ ) {
		for(sc_dev = 0; sc_dev < m147maxscdev; ddefs++,dd++,sc_dev++,
				slc += M147SLCS) {
	
			/* Tape intialized in mvme147tp.c */
			tmp = *dd;
			device = tmp->DISK1.device_type;
			if ( (device == DISK) || (device == FLOP) )
			{
				ddefs->DISK1 = tmp->DISK1;
				ddefs->DISK2 = tmp->DISK2;
				d147dinit(MKDEV(sc_adr,sc_dev),ddefs,slc);
			}
		}
	}
	/* convert vector address to vector number */
	d147_ivec[0] /= 4;

	/* initialize floppy structure */
	f147dsk.d_flags = (D_DFL_CONF|D_FMTU);
	f147dsk.d_strategy = d147strategy;
	f147dsk.d_start = d147start;
	f147dsk.d_timer = f147timer;
	f147dsk.d_print = d147print;
	f147dsk.d_shorttime = D147SHRT * HZ;
	f147dsk.d_longtime = D147LONG * HZ;
	f147dsk.d_slcbits = M147SBTS;

	/* initialize dsk structure */
	d147dsk.d_flags = (D_DFL_CONF|D_CONFIG|D_FMTU);
	d147dsk.d_strategy = d147strategy;
	d147dsk.d_start = d147start;
	d147dsk.d_timer = d147timer;
	d147dsk.d_print = d147print;
	d147dsk.d_shorttime = D147SHRT * HZ;
	d147dsk.d_longtime = D147LONG * HZ;
	d147dsk.d_slcbits = M147SBTS;


	dskinit(&d147dsk);
	dskinit(&f147dsk);


#	ifdef TRACE
	m147trcbase = (struct trace *) &m147trcmem[0];
	for(m147trace = m147trcbase, n = 0; n < MAX_TRC;
		n++, m147trace++)
	m147trace = m147trcbase;
	{
		m147trace->id = 0;
		m147trace->i0 = 0;
		m147trace->i1 = 0;
		m147trace->i2 = 0;
		m147trace->i3 = 0;
		m147trace->i4 = 0;
		m147trace->i5 = 0;
		m147trace->i6 = 0;
	}
#	endif


}


/*
 *	initialize disks
 */

STATIC d147dinit(dev,dd,slctbl)
register unsigned short dev;
register struct scs_disk *dd;
register struct slice *slctbl;
{
	register struct iobuf *dp;
	register struct pktcnt *pcp;
	int device;
	int index;
	struct slice *flpslc = &d147flop;

#	ifdef DEBUG
	printf("d147dinit: device: %x\n", dev);
#	endif

	dp = &m147utab[POS(dev)];
	pcp = (struct pktcnt *)&dp->io_pktcnt;
	pcp->cnt = 0;
	pcp->limit = m147lim;	
	dp->b_dev = dev; /* logical device */
	dp->io_nreg = sizeof(struct scs_pkt); /* length of eregs */
	device = dd->DISK1.device_type;
	dd->DISK1.device_type = SCS_DISK;
	dp->io_desc = (int)dd;
	dp->io_slctbl = (int)slctbl;

	if(device == FLOP)
	{
		/* this device is now a floppy, copy all the slice info */
		dp->b_dev = (dev | FLOPBIT); /* logical device */
		for( index = 0; index < M147SLCS; index++, slctbl++, flpslc++)
			*slctbl = *flpslc;
		/* always force the floppy to re-initialize */
		dp->b_flags |= B_REINIT;
		dp->io_dskptr = (int)&f147dsk;
	}
	else
	{
	dp->io_dskptr = (int)&d147dsk;
	}

	if( major(rootdev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}

	if( major(swapdev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}

	if( major(pipedev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}
}


/*
 *	d147bopen: perform open on a block device
 */

d147bopen(dev, modes, otyp)
register otyp;
register dev;
register modes;
{

	DEBUGPR(DBBOPEN,dev,"bopen");

	d147open(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d147copen: perform open on a character device
 */

d147copen(dev, modes, otyp)
register otyp;
register modes;
register dev;
{

	DEBUGPR(DBCOPEN,dev,"copen");

	d147open(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d147open: perform open for all devices
 */

STATIC d147open(dev,modes,omask,otyp)
register dev;
register otyp;
register modes;
register omask;
{
	register struct iobuf *dp;
	register struct scs_disk *wd;
	register *otbl;
	register i;

	/*
	 *	if the device or processor is nonexistent, then error
	 */
	if( (SCADR(dev) > (m147maxscadr-1)) ||
		 (SCDEV(dev) > (m147maxscdev-1) ) ||
		(d147_addr[0] == NULL)
	  )
	{
		u.u_error = ENXIO;
		return;
	}
	i = POS(dev);
	dp = &m147utab[i];
	otbl = &d147otbl[i];

	if(M147FLOP(dev))
	{
		if(dp->io_dskptr != (int)&f147dsk) {  /* no dev */
		u.u_error = ENXIO;
		return;
	}
		dp->b_dev = dev;
		if(HIGHSLICE(dev))
			*(struct scs_disk *)dp->io_desc = m147pcat;
		else
			*(struct scs_disk *)dp->io_desc = m147dsdd5;
		wd = (struct scs_disk *)dp->io_desc;
		wd->DISK1.device_type = SCS_DISK;
		
	}
	else
	{
		if(dp->io_dskptr != (int)&d147dsk) {  /* no dev */
			u.u_error = ENXIO;
			return;
		}
	}

	dskopen(dp, modes, otbl, otyp, omask);
}

/*
 *	d147bclose: perform close on a block device
 */

d147bclose(dev, modes, otyp)
register otyp;
register dev;
register modes;
{

	DEBUGPR(DBBCLOSE,dev,"bclose");

	d147close(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d147cclose: perform close on a character device
 */

d147cclose(dev, modes, otyp)
register otyp;
register dev;
register modes;
{

	DEBUGPR(DBCCLOSE,dev,"cclose");

	d147close(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d147close: perform close for all devices
 */

d147close(dev, modes, mask, otyp)
register dev;
register otyp;
register modes;
register mask;
{
	register *otbl;
	register struct iobuf *dp;
	register i;

	i = POS(dev);
	dp = &m147utab[i];
	otbl = &d147otbl[i];

	dskclose(dp, modes, otbl, otyp, mask);

}

/*
 *	d147strategy: perform the disk I/O strategy algortithm
 */

d147strategy(bp)
register struct buf *bp;
{
	register struct iobuf *dp;
	register struct scs_disk *wd;
	register struct slice *slc;
	register unsigned short dev = bp->b_dev;
	register int coalesce = m147coal;

	DEBUGPR(DBSTRAT,dev,"strategy");

	dp = &m147utab[POS(dev)];

#	ifdef SANITY
	if( (SCADR(dp->b_dev) != SCADR(bp->b_dev))
	 || (SCDEV(dp->b_dev) != SCDEV(bp->b_dev)) )
	{
		spl7();
		printf("d147strategy: bp: %x, dp: %x\n", bp, dp);
		printf("dp->b_dev: %x, bp->b_dev\n",
			 dp->b_dev, bp->b_dev);
		cmn_err(CE_PANIC,"d147strategy: bad b_dev");
	}
#	endif

	/*
	 * Do system accounting calculations
	 */
	if ((bp->b_flags&B_ABS) == 0) {
		/* save cylinder number in b_resid for sadp(1) */
		slc = (struct slice *)dp->io_slctbl + SLICE(bp->b_dev);
		bp->b_resid = bp->b_blkno;	/* get block number */
		bp->b_resid += slc->blkoff;	/* add slice start */

		wd = (struct scs_disk *)dp->io_desc;
		/*
		 *	Calculate cylinder number for profiler and 
		 *	coalescing
		 */
		bp->b_resid *= (BSIZE / wd->DISK2.bytes_per_block);
		bp->b_resid /= (wd->DISK1.sect_per_track * 
			wd->DISK1.number_heads);
	}

	/* 
		don't allow byte count to exceed the maximumm number
		of entrys for the scatter/gather array necessary for
		a 256k byte transfer.

		62 - full 4096 byte blocks
		 2 - partial blocks - usually ce4 & 31c
		 1 - full 4096 block
	     -------
		400000
	*/
	if(bp->b_bcount > ((d147maxblk-1)*NBPP))
	{
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}

	/* don't coalesce a floppy, too much work */
	if(M147FLOP(bp->b_dev))
		coalesce = 0;

	/* call the disk discipline strategy (type 2) routine */
	dsk2strategy(bp, dp, coalesce, d147maxblk);
}

/*
 *	d147start: start the I/O
 */


d147start(dp)
register struct iobuf *dp;
{
	register struct buf *bp;
	register struct scs_pkt *ep;
	register s;
	register unsigned short i;
	register struct pktcnt *pcp;

	int d147ihand();
	struct scs_pkt *m147get();

	DEBUGPR(DBSTART,dp->b_dev,"start");

	s = spl6();

	if((bp = dp->b_actf) == NULL)
	{
		splx(s);
		return;
	}
	
	pcp = (struct pktcnt *)&dp->io_pktcnt;
	if(pcp->cnt >= pcp->limit)
	{
		splx(s);
		return;	
	}

	pcp->cnt++;

	

#	ifdef SANITY
	if( (SCADR(dp->b_dev) != SCADR(bp->b_dev))
	 || (SCDEV(dp->b_dev) != SCDEV(bp->b_dev)) )
	{
		printf("d147start: bp: %x, dp: %x\n", bp, dp);
		printf("dp->b_dev: %x, bp->b_dev: %x\n",
			 dp->b_dev, bp->b_dev);
		cmn_err(CE_PANIC,"d147start: bad b_dev");
	}
#	endif


	/*
		 * Get a command packet.
	 */

	if ((ep = m147get()) == NULL) {
#		ifdef DEBUG
		printf("147dk: out of packets\n");
#		endif
		splx(s);
		return;
	}

	blkacty |= (1<<M147);	/* indicate mvme147 activity */

	/*
		 * Set up packet
	 */

	ep->scs_ctrlr = SCADR(bp->b_dev);
	ep->scs_dev = SCDEV(bp->b_dev);
	/* save the pointer to the utab */
	ep->utab = dp;
	ep->inthand = d147ihand;

	ep->scs_vnum = d147_ivec[0];
	/* set the interrupt level */
	ep->scs_level = d147_ilev[0];
	ep->scs_stat0 = ep->scs_stat1 = ep->scs_stat2 = ep->scs_stat3 = 0;
	ep->EDISK2.work_area  = (unsigned char *)&m147_work[POS(dp->b_dev)];
	/*
		 * Process the command
	 */
	if (bp->b_flags&B_CONF)			/* config request */
		d147conf(dp,bp,ep);
	else if (bp->b_flags&(B_FORMAT|B_FMTU|B_FIXBS))	/* format */
		d147format(dp,bp,ep);
	else {					/* read/write request */
		d147rw(dp, bp, ep);
	}
	splx(s);
}
/*
 *
 *
 *
*/
d147rw(dp, bp, ep)
struct iobuf *dp;
register struct buf *bp;
register struct scs_pkt *ep;
{
	struct scs_disk *wd = (struct scs_disk *)dp->io_desc;
	struct slice *slc;
	register struct scga *list = &scgalist[(POS(bp->b_dev)*d147maxblk)];
	register struct djntio *ndj = &d147djlist[(POS(bp->b_dev)*d147maxblk)];
	register struct djntio *tmpdj =  ndj;
	register int djsize;
	register struct buf *nbp;
	int lsn, nblk;
	int trk0blk;
	int bpbshift;
	int count, i;
	int s;


	ep->scs_funcc = (bp->b_flags&B_READ) ? SCS_READ : SCS_WRITE;
	slc = (struct slice *)dp->io_slctbl + SLICE(bp->b_dev);

	lsn = bp->b_blkno;

	/* don't add slice start block for absolute requests */
	if ((bp->b_flags&B_ABS) == 0)
		lsn += slc->blkoff;	/* slice beg. */

	ep->RDWR.address = (unsigned char *)bp->b_un.b_addr;

	nblk = bp->b_bcount/ wd->DISK2.bytes_per_block;


	if(M147FLOP(bp->b_dev))
	{
		bpbshift = (wd->DISK2.bytes_per_block == 256) ? 1 : 0;
		if( (wd->DISK1.disk_attr & NONSTD) == NULL)
		{
			/* 
				This is a low density IBM format 
				floppy disk 
			*/
			/* make lsn relative to 256 byte blocks */	
			lsn <<= 1;

			if( (lsn+nblk) < 8)
			{
				/* all blocks are within trk 0 */
				nblk <<= 1;
				lsn  <<= 1;
			}
			else  if(lsn < 8)
			{
				/* blocks are in track 0 and cross */
				/* track 0 boundary */
				trk0blk = 8  - lsn;
				nblk -= trk0blk;
				nblk  += (trk0blk << 1);
				lsn  <<= 1;
			}
			else
			{
				lsn = lsn + 8;
			}
		}
		else
		{
			lsn = lsn << bpbshift;
		}
	}
	ep->RDWR.block = lsn;
	ep->RDWR.block_count = nblk;


	
	s = spl6();

	if(bp->av_back){

		splx(s);

#		ifdef JUNK
		junk();
#		endif
		ep->RDWR.address = (unsigned char *)list;
		for(nbp = bp, i=0, count = 0; nbp;
		    (nbp = nbp->av_back), i++, list++){
			nbp->b_start = lbolt;
			list->address = (unsigned char *)nbp->b_un.b_addr;
			list->link = 1;
			list->func_code = 5;
			list->count = nbp->b_bcount;
			count += nbp->b_bcount;
		}
		(--list)->link = 0;
		ep->RDWR.block_count = (count / wd->DISK2.bytes_per_block);
		ep->RDWR.scga_count  = i;


		if(i > m147scgacnt){
		    if(bp->b_flags&B_READ)
			printf("m147rw: combined %d read requests\n",i);
		    else
			printf("m147rw: combined %d write requests\n",i);
		}
	}
	else
	{
#		ifdef REALRAWIO
		if((bp->b_flags & B_PHYS) && (bp->b_dev != swapdev))
		{
			djsize = disjointio(bp->b_proc, bp->b_un.b_addr,
				 bp->b_bcount, bp->b_flags, ndj, d147maxblk);
			if(djsize < 0)
				printf("disjointio error\n");
#			ifdef DEBUG
			printf("djsize: %d\n", djsize);
			printf("bpaddr: %x, bpcount: %x\n",
				bp->b_un.b_addr, bp->b_bcount);
			for(i = 0; i < djsize; i++, tmpdj++)
				printf("addr: %x, count: %x\n",tmpdj->addr, 
					tmpdj->count); 
#			endif
			ep->RDWR.address = (unsigned char *)list;

			for(count = 0,i = 0; i < djsize; i++, ndj++, list++)
			{
				list->address = (unsigned char *)ndj->addr;
				list->count = ndj->count;
				list->link = 1;
				list->func_code = 5;
				count += ndj->count;
			}
			(--list)->link = 0;
			ep->RDWR.block_count = (count / wd->DISK2.bytes_per_block);
			ep->RDWR.scga_count  = djsize;
		}
		else
#endif
			ep->RDWR.scga_count  = 0;
	}
done:
	splx(s);
	ep->cmd_ctrl = wd->DISK1.cache;


	/*
	 * Reset the time in b_start for safety timer purposes.
	 */
	{	
		register struct iotime *ip;
		ip = &m147stat[POS(bp->b_dev)];
		ip->io_resp += lbolt - bp->b_start;
	}
	bp->b_start = lbolt;

	/*
	 * Set the time in the iobuf struct for the
	 * system activity profiler.
	 */

	if (dp->io_start == 0)
		dp->io_start = lbolt;

	DTRC(DTRAIL(1),dp,ep->scs_funcc,ep->EDISK2.work_area,ep->RDWR.block,ep->RDWR.block_count, ep->RDWR.address,ep);

	s = spl6();

#	ifdef FWTRC
	sent++;
	if(sent >= 2) printf("sent: %d\n", sent);
#	endif

	M147SEND(ep);
	splx(s);
}
/* dummy routine for easier break points */
junk() { }


/*
 *	d147conf: perform configuration command
 */

d147conf(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct scs_pkt *ep;
{
	register struct dkconfig *dkcfg;
	register struct scs_disk *wd = (struct scs_disk *)dp->io_desc;
	register dft;
	register s;

	DEBUGPR(DBCONF,bp->b_dev,"conf");

	/* set up the command packet */
	ep->scs_funcc = SCS_DKATTACH;

	/*
	 * Winchester/Floppy configure (write disk descriptor)
	 */

	/* Is this a default or a re-config? */
	dft = (bp->b_flags & B_DFL_CONF) ? 1 : 0;

	if(!dft) {	/* This is a re-config */
		register struct volumeid *vid;

		vid = (struct volumeid *) (int)(bp->b_un.b_addr);
		dkcfg = (struct dkconfig *)((int)bp->b_un.b_addr + 256);
		/* Get number of slices and put it in iobuf struct */
		dp->io_slccnt = dkcfg->ios_slccnt;

		/* controller type - high order byte */
		wd->DISK1.controller_type =  (vid->vid_typ >> 8);
		/* periperal type - low order byte */
		wd->DISK1.device_type =  0;
		wd->DISK1.number_heads = dkcfg->ios_hds; /* number of heads */
		wd->DISK1.number_cyls =  dkcfg->ios_trk; /* number of cyl*/
		wd->DISK2.bytes_per_block =   dkcfg->ios_psm; /* bytes per sec*/
		/* sectors per track*/
		wd->DISK1.sect_per_track =  dkcfg->ios_spt; 
		wd->DISK1.step_rate =  dkcfg->ios_sr; /* step rate */
		/* pre-compensation cyl */
		wd->DISK1.precomp_cyl = dkcfg->ios_pcom; 
		wd->DISK1.interleave = dkcfg->ios_ilv;

		/* set up floppy attribute word descriptor */
		if (M147FLOP(dp->b_dev)) {
			register unsigned short flop;
	
			if(HIGHSLICE(dp->b_dev))
				wd->DISK1.disk_attr = M147PCAT;
			else
				wd->DISK1.disk_attr = M147DSDD5;
		}
		else
			wd->DISK1.disk_attr = (dkcfg->ios_catr & SCS_ATW);
		/* set the cache to use default */
		wd->DISK1.cache = ((dkcfg->ios_catr >> 16) & SCS_CACHE);
		wd->DISK1.fmt_list = ((dkcfg->ios_catr >> 12) & SCS_FMTLST);
		
		wd->DISK2.retry = 10;

		wd->DISK2.alt_resv =  dkcfg->ios_alts;	/* #  of alternates */
		wd->DISK2.sector_slip_count = dkcfg->ios_ssc;

	}

	ep->EDISK1 = wd->DISK1;
	ep->EDISK2 = wd->DISK2;
	ep->EDISK2.work_area  = wd->DISK2.work_area = (unsigned char *)&m147_work[POS(dp->b_dev)];


	DTRC(DTRAIL(2),dp,ep->scs_funcc,ep->EDISK2.work_area,ep->RDWR.block,ep->RDWR.block_count, ep->RDWR.address,ep);
	s = spl6();
	M147SEND(ep);
	splx(s);
}

/*
 *	d147format: perform format command
 */

d147format(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct scs_pkt *ep;
{
	struct scs_disk *wd = (struct scs_disk *)dp->io_desc;
	register s;

	DEBUGPR(DBFORMAT,bp->b_dev,"format");

	if(bp->b_flags & B_FIXBS)
	{
		/* set up packet */
		ep->scs_funcc = SCS_AAS;
		
		/* bad spot address */
		ep->FMT.defect_list = (struct scs_dfctblk *)bp->b_un.b_addr;
		/* bad spot count */
		ep->FMT.number_defects = 1;	
		ep->cmd_ctrl = 0;	/* bad spot type  = physical sector */
		ep->FMT.interleave = wd->DISK1.interleave;
	}
	else
	{
		register struct scs_dfcthdr *dkf = 
			(struct scs_dfcthdr *)bp->b_un.b_addr;

		/* set up packet */
		ep->scs_funcc = SCS_FORMAT;
		
		/* bad spot address */
		ep->FMT.defect_list = (struct scs_dfctblk *)dkf;
		/* bad spot count */
		ep->FMT.number_defects = dkf->count;	
		/* 
			if there are defects then it must be 
			the complete user list 
		*/
		if(dkf->count)
			ep->cmd_ctrl = SCS_DEFECT;	
		else
			ep->cmd_ctrl = 0;
		/* count is only used for temporary storage */
		dkf->count = NULL;
		ep->FMT.interleave = wd->DISK1.interleave;
	}

	DTRC(DTRAIL(3),dp,ep->scs_funcc,ep->EDISK2.work_area,ep->RDWR.block,ep->RDWR.block_count, ep->RDWR.address,ep);

	s = spl6();
	M147SEND(ep);
	splx(s);
}


/*
 * d147intr: handle the mvme147 interrupts
 */

d147ihand(ep)
register struct scs_pkt *ep;
{
	register struct iobuf *dp;
	register struct buf *bp;
	register struct pktcnt *pcp;
	struct cmd_tbl *cc;
	register rdev;	
	register int s;
	struct scs_pkt	*m147rcv();
	register struct iotime *ip;

	DEBUGPR(DBINTR,0,"interrupt");

	s = spl6();

	if (d147_addr[0] == NULL) {
		d147print(0, "Undefined interrupt");
		splx(s);
		return;
	}

	dp = (struct iobuf *)ep->utab;
	rdev = dp->b_dev;
	cc = (struct cmd_tbl *) &m147_work[POS(rdev)];
	pcp = (struct pktcnt *) &dp->io_pktcnt;
	ip = &m147stat[POS(rdev)];

	if(pcp->cnt)
		pcp->cnt--;

#	ifdef FWTRC
	if(sent)
		sent--;
#	endif

	/* if this is the last packet, reset */
	if(pcp->cnt == NULL);
	{
		ip->io_act += lbolt - dp->io_start;
		dp->io_start = 0;
	}

	/*
	 * Find the associated buffer structure.
	 */ 

	if( (bp = dp->b_actf) == NULL ) { /* drive active? */
		/* assert this must never be true */
#ifdef DEBUG
		d147print( rdev, "lost request, cannot be" );
#endif
		splx(s);
		return;
	}
	if (bp->b_flags&B_READ || bp->b_flags&B_WRITE) {
		register struct iotime *ip;

		ip = &m147stat[POS(rdev)];

		ip->io_resp += lbolt - bp->b_start;
	}
	/* don't report recoverable errors */
	if ((ep->scs_stat1)  && (ep->scs_stat1 != RECOV_ERR))
		d147err(dp,bp,ep,cc);
	DTRC(DTRAIL(4),dp,ep->scs_funcc,ep->EDISK2.work_area,ep->RDWR.block,ep->RDWR.block_count, ep->RDWR.address,ep);
	m147free(ep);

	/* if last packet */
	if(pcp->cnt == NULL);
		blkacty &= ~(1<<M147);	/* indicate activity done */


	dsk2done(dp);		/* finish up */

	splx(s);

}

/*
 *	d147err:  process mvme147 return status
 */

d147err(dp,bp,pp,cc)
register struct iobuf *dp;
register struct buf *bp;
register struct scs_pkt *pp;
register struct cmd_tbl *cc;
{
	register trk;
	register sec;
	register d147eblk;		/* block # where error occurred */
	register struct scs_disk *wd = (struct scs_disk *)dp->io_desc;
	struct buf *sav_actf;
	int svdev;
	int s;

        DEBUGPR(DBERR,bp->b_dev,"error");



	/* set up pointer to "registers" for fmtberr() */
	/* This changes the b_dev field so must save */
	svdev = dp->b_dev;
	dp->b_dev = makedev(M147,minor(dp->b_dev));
	dp->io_addr = (paddr_t)pp;

	/*
	 * calculate block that error was on for fmtberr()
	 * and save it for d147prerr
	 */

	d147eblk =  cc->sector_error;

	s = spl6();
	sav_actf = dp->b_actf;		/* trick fmtberr() */
	dp->b_actf = bp;
	fmtberr(dp,d147eblk);		/* error block */
	dp->b_actf = sav_actf;		/* fix queue back */
	dp->b_dev = svdev;		/* restore device */
	splx(s);
	/* N.B. what about recovered errors ? */
	if(pp->scs_stat1) {	/* fatal error */
		d147prerr(bp->b_dev,pp,d147eblk,cc);	/* display error */
		bp->b_resid = bp->b_bcount;
		bp->b_flags |= B_ERROR;
	}
}

/*
 *	d147prerr: print out error information
 */

STATIC d147prerr(dev,ep,d147eblk,cc)
register dev,d147eblk;
register struct scs_pkt *ep;
register struct cmd_tbl *cc;
{
	register unsigned char cs;
	unsigned long txfer_count;

	cs = ep->scs_stat1;

	d147print(dev,"\nERROR");

	PRCMD(ep);				/* print command data */

#ifdef	DEBUG
	/* read/write */
	if (ep->scs_funcc == SCS_READ || ep->scs_funcc == SCS_WRITE)
		printf("\nPhysical Block: 0x%x(%d)",d147eblk,d147eblk);
	txfer_count = cc->transfer_address - ep->RDWR.address;	
	printf("\n\t\t\t\tTermination Transfer Count: %x(%d)", txfer_count,
		txfer_count);
	printf("\n\t\t\t\tStatus Parameter 0: %x", ep->scs_stat0);
	printf("\n\t\t\t\tStatus Parameter 1: %x", ep->scs_stat1);
	printf("\n\t\t\t\tStatus Parameter 2: %x", ep->scs_stat2);
	printf("\n\t\t\t\tStatus Parameter 3: %x", ep->scs_stat3);
	printf("\n");
#else
	if (ep->scs_funcc == SCS_READ || ep->scs_funcc == SCS_WRITE)
		printf("\nPhysical Block: 0x%x(%d)",d147eblk,d147eblk);
	printf("\tCommand Byte: %x", ep->scs_funcc);
	printf("\tStatus Byte 1: %x", ep->scs_stat1);
	printf("\n");
#endif
}

/*
 *	d147prcmd: print command debug information
 */

STATIC d147prcmd(pp)
register struct scs_pkt *pp;
{
	printf("Command: ");

	switch (pp->scs_funcc) {
		case SCS_DKATTACH: printf("Write Descriptor, ");		break;
		case SCS_FORMAT: printf("Format , ");			break;
		case SCS_READ:	printf("Read Blocks, ");		break;
		case SCS_WRITE:	printf("Write Blocks, ");		break;

		default:	printf("Unknown: %x, ", pp->scs_funcc);	break;
	}

	printf("Device type: %x, Unit number: %x\n", pp->scs_ctrlr, pp->scs_dev);
}

/*
 *	d147print: print mvme147 driver error messages
 */

d147print(dev,str)
register dev;
register char *str;
{

	printf("%s on MVME147 ctl %d, drive %d, slice %d\n", str,
		SCADR(dev), SCDEV(dev), SLICE(dev));
}

/*
 *	d147breakup: physical I/O breakup routine
 */
#ifndef REALRAWIO
d147breakup(bp)
register struct buf *bp;
{
	dma_breakup(d147strategy, bp, BSIZE);
}
#endif

/*
 *	d147read: raw I/O read
 */

d147read(dev)
register dev;
{
	register struct iobuf *dp;
	register struct slice *slc;

	DEBUGPR(DBREAD,dev,"read");

	dp = &m147utab[POS(dev)];
	slc = (struct slice *) dp->io_slctbl + SLICE(dev);
	if (physck(slc->nblocks,B_READ))
#ifdef REALRAWIO
		physio(d147strategy, 0, dev, B_READ);
#else
		physio(d147breakup, 0, dev, B_READ);
#endif
}

/*
 *	d147write: raw I/O write
 */

d147write(dev)
register dev;
{
	register struct iobuf *dp;
	register struct slice *slc;

	DEBUGPR(DBWRITE,dev,"write");

	dp = &m147utab[POS(dev)];
	slc = (struct slice *) dp->io_slctbl + SLICE(dev);
	if (physck(slc->nblocks,B_WRITE))
#ifdef REALRAWIO
		physio(d147strategy, 0, dev, B_WRITE);
#else
		physio(d147breakup, 0, dev, B_WRITE);
#endif
}

/*
 *	d147ioctl: perform ioctl functions
 */

d147ioctl(dev,cmd,args,mode)
register dev;
register cmd;
register char *args;
register mode;
{
	register struct buf *bp;
	register struct iobuf *dp;
	register char *addr;
	register struct buf *bb;
	register struct scs_disk *wd;
	register size;
	register i;
	struct buf *geteblk();


	DEBUGPR(DBIOCTL,dev,"ioctl");

	dp = &m147utab[POS(dev)];
	wd = (struct scs_disk *)dp->io_desc;

	bp = geteblk();

	if (dskioctl(dp,cmd,args,mode,bp)) {
		brelse(bp);
		return;	/* command was handled by disk discipline ioctl */
	}

	switch (cmd) {

	case DKGETCFG:
	case DKGETINFO:	{
			register struct dkconfig *dc;

			clrbuf(bp);
			/* copy needed fields from descriptor structure */
			dc = (struct dkconfig *)(bp->b_un.b_addr +
				sizeof(struct volumeid));

			dc->ios_sr = wd->DISK1.step_rate;
			dc->ios_spt = wd->DISK1.sect_per_track;
			dc->ios_hds = wd->DISK1.number_heads;
			dc->ios_trk = wd->DISK1.number_cyls;
			dc->ios_pcom = wd->DISK1.precomp_cyl;
			dc->ios_psm = wd->DISK2.bytes_per_block;

			if (M147FLOP(dev)) {
				dc->ios_atw = (wd->DISKATTR&DOUBLETRK) ? ATWDD : 0;
				dc->ios_atw |= (wd->DISKATTR&MFM) ? ATWEN : 0;
				dc->ios_atw |= (wd->DISKATTR&EIGHT) ? ATWFS : 0;
				dc->ios_atw |= (wd->DISKATTR&DOUBLESIDE) ? ATWDS : 0;
			}

			if (cmd == DKGETCFG) {
				addr = bp->b_un.b_addr +
					sizeof(struct volumeid);
				size = sizeof(struct dkconfig);
			}
			else {
				addr = bp->b_un.b_addr;
				size = sizeof(struct dkblk0);
			}

			copyout(addr,args,size);
			brelse(bp);
			return;
			}

	case DKSETCFG:
	case DKSETINFO: if (cmd == DKSETCFG) {
				addr = bp->b_un.b_addr +
					sizeof(struct volumeid);
				size = sizeof(struct dkconfig);
			}
			else {
				addr = bp->b_un.b_addr;
				size = sizeof(struct dkblk0);
			}
			if (copyin(args,addr,size) == -1) {
				brelse(bp);
				return;
			}
			/*
			 * If the volumeid is available, check to see that
			 * the bad spots to be given to the driver for a
			 * format command are going to be in the form of
			 * "physical sector".
			 */
			if (cmd == DKSETINFO) {
				register struct volumeid *vid;

				vid = (struct volumeid *)bp->b_un.b_addr;
				if ( (vid->vid_strat != DKHARDSEC) &&
					(vid->vid_strat != DKHARDSPOT) &&
					(vid->vid_strat != DKPERFECT) ) {
					brelse(bp);
					u.u_error = ENXIO;
					return;
				}
			}
			bp->b_flags |= B_CONF|B_ABS;
			break;

	case DKFIXBADSPOT:
		{
			/* only read in 1 defect and re-format it */
			register struct dkbadlst dkb;
			register struct scs_dfctblk *dfb = 
				(struct scs_dfctblk *)bp->b_un.b_addr;

			if (copyin(args,&dkb,sizeof(struct dkbadlst))
			    == -1) {
				brelse(bp);
				return;
			}
			dfb->lsn = dkb.dk_sector;
			/* set the defect type to block number */
			dfb->list_code = 0;
		}
			bp->b_flags |= B_ABS|B_FIXBS;
			break;

	case DKFORMAT: {
				struct dkfmt *dkf;
				register struct dkbadlst *bs;
				register struct dkbadlst *defect;
				register cnt;
				register struct scs_dfcthdr *b1hdr;
				register struct scs_defect *b1;
				register struct scs_blockdfct *b2;
				int fmt_list = wd->DISK1.fmt_list;

				bb = geteblk();		/* for bad sec list */

				if (copyin(args,bb->b_un.b_addr,
					sizeof(struct dkfmt)) == -1) {
					brelse(bb);
					brelse(bp);
					return;
				}

				dkf = (struct dkfmt *)bb->b_un.b_addr;
				if (dkf->dk_badcnt * sizeof(struct scs_defect
) >
				    SBUFSIZE) {
					brelse(bb);
					brelse(bp);
					u.u_error = ENXIO;
					return;
				}


				b1hdr = (struct scs_dfcthdr *)bp->b_un.b_addr;
				b1hdr->list_code = fmt_list;
				b1hdr->count = cnt = dkf->dk_badcnt;
				b1hdr++;
				b1 = (struct scs_defect *)b1hdr;
				b2 = (struct scs_blockdfct *)b1hdr;
				defect = dkf->dk_badlst;
				bs = (struct dkbadlst *)bb->b_un.b_addr;

				/* build the real bad sec list */
				while (cnt-- > 0) {
					if((copyin(defect, bs, sizeof(struct dkbadlst))) == -1)
					{
						brelse(bp);
						brelse(bb);
						u.u_error = ENXIO;
						return;
					}
						
					if(fmt_list == SCS_BLOCK)
					{
						b2->offset = bs->dk_sector;
					}
					else if( (fmt_list == SCS_SECTOR) ||
					         (fmt_list == SCS_BFI) )
					{
						b1->cylinder = bs->dk_cyl;
						b1->head = bs->dk_head;
						b1->offset = bs->dk_sector;
					}
					else
					{
						brelse(bp);
						brelse(bb);
						u.u_error = ENXIO;
						return;
					}
					b1++;
					b2++;
					bs++;
					defect++;
				}

				bp->b_flags |= B_FORMAT|B_ABS|B_LONGCMD;
				break;
			}

	default:	u.u_error = EINVAL;
			brelse(bp);
			return;
	}

	bp->b_proc = u.u_procp;
	bp->b_dev = dev;
	bp->b_bcount = 0;
	bp->b_blkno = 0;
	d147strategy(bp);
	iowait(bp);
	if (bp->b_flags&B_FORMAT) {
		brelse(bb);
	}
	bp->b_flags &= ~(B_CONF|B_FORMAT|B_ABS|B_LONGCMD|
			 B_ISTRAT|B_ICACHE|B_IPC_CONF|B_FIXBS);
	brelse(bp);
}

/*
 *	d147timer: check on status of MVME147 driver periodically
 */

f147timer()
{

	DEBUGPR(DBTIMER,0,"timer");

	/* call the disk discipline timer routine */
	dsk2timer(&m147utab[0],MAXPOS,&f147dsk);
}

/*
 *	d147timer: check on status of MVME147 driver periodically
 */

d147timer()
{

	DEBUGPR(DBTIMER,0,"timer");

	/* call the disk discipline timer routine */
	dsk2timer(&m147utab[0],MAXPOS,&d147dsk);
}

/*
 *	d147clr: clear the driver after powerfail
 */

d147clr()
{
	DEBUGPR(DBCLR,0,"clr");

	dskclr(&m147utab[0],MAXPOS);
}

#ifdef DUMP

/*
 *	mvme147 dump support
 */

STATIC	int	d147offset;

/*
 *	d147dump: dump memory to mvme147 swap area
 */

d147dump()
{
	int d147doio();
	extern	dev_t	dumpdev;
	register dev = dumpdev;

	spl7();				/* disable interrupts */

	DEBUGPR(DBDUMP,dev,"dump");

	if ( d147_addr[0] == NULL) {
		printf("Dump device controller not present\n");
		return;
	}

	{	register struct slice *sp;

		sp = (struct slice *)m147utab[POS(dev)].io_slctbl + SLICE(dev);
		d147offset = sp->blkoff;
	}

}

/*
 *	d147doio: dump out one block of memory
 */

STATIC d147doio(dev,blkno,maddr,rdwr)
register dev;
register blkno;
char *maddr;
int rdwr;
{
	register struct scs_pkt *pp;
	register struct scs_disk *wd;
	struct scs_pkt *m147get();

	DEBUGPR(DBDOIO,dev,"doio");

	if ((pp = m147get()) == NULL) {
		for (;;);
	}

	wd = &m147params[POS(dev)];
	pp->scs_funcc = (rdwr == B_READ) ? SCS_READ : SCS_WRITE;
	pp->scs_ctrlr = SCADR(dev);
	pp->scs_dev = SCDEV(dev);
	pp->RDWR.block = blkno + d147offset;
	pp->RDWR.address = (unsigned long)maddr;
	pp->RDWR.block_count = SBUFSIZE / BSIZE;
	M147SEND(ep);

	do {
	} while (ep == NULL );	/* wait until done */

	/*
	 *	will hang forever on the above "do-while" statement
	 *	when the controller fails to respond.
	 */

	if (pp->stat1 )				/* an error occurred */
		d147prerr(dev,pp);

	m147free(ep);
}
#else

d147dump()
{
}
#endif 


m147crque( queue, pktcount)
struct scs_pkt *queue;
int pktcount;
{
	int count;

	for(count = 0; count <= pktcount - 1; count++, queue++)
	{
		queue->next = queue+1;
	}
	(--queue)->next = NULL;
}

struct scs_pkt *m147get()
{
	struct scs_pkt *nq;
	register unsigned char *data;
	register int i;

	/* get packet at top of queue */
	nq = m147freeq;
	/* set top of queue to next in queue */
	m147freeq = m147freeq->next;
	/* clean up packets next pointer */
	nq->next = NULL;

#	ifdef CLNPKTS
	for(i = 0, data = (unsigned char *)nq; i < sizeof(struct scs_pkt); 
		i++, data++)
			*data = 0;
#	endif

	return(nq);
}

m147free(ep)
struct scs_pkt *ep;
{
	/* set packet next pointer to head of queue */
	ep->next = m147freeq;
	/* set head of queue to packet */	
	m147freeq = ep;
}


#ifdef TRACE

chktrc(id, a,b,c,d,e,f,g)
int id,a,b,c,d,e,f,g;
{
	register struct trace *xpt = m147trace;
	int lvl;

	lvl = spl7();

	xpt->id = id;
	xpt->i0 = a;
	xpt->i1 = b;
	xpt->i2 = c; 
	xpt->i3 = d;
	xpt->i4 = e;
	xpt->i5 = f;
	xpt->i6 = g;
	(xpt+1 > &m147trcbase[MAX_TRC-1])?xpt=m147trcbase:xpt++;
	xpt->id = END_TRACE;
	xpt->i0 = 0;
	xpt->i1 = 0;
	xpt->i2 = 0;
	xpt->i3 = 0;
	xpt->i4 = 0;
	xpt->i5 = 0;
	xpt->i6 = 0;
	m147trace = xpt;

	splx(lvl);

}

#endif

intr_trc(ep)
struct scs_pkt *ep;
{
	DTRC(DTRAIL(5),ep,ep->scs_stat0,ep->scs_stat1,ep->scs_stat2,ep->scs_stat3,0,0);
}
