/*		Copyright (c) 1988 Motorola Inc.		*/
/*		All Rights Reserved				*/
/*
 *	mvme147tp.c: Motorola MVME147tp VMEmodule Streamer Tape SCSI driver
 *
 *	@(#)mvme147tp.c	35.3 
 *
 *	The MVME147tp VMEmodule Streamer Tape SCSI Controller is an intelligent
 *	peripheral controller (IPC) VMEmodule designed to interface
 *	with the extended address line VMEmodules of the VMEbus family.
 *	The MVME147tp supports eight streamer tape drives per controller
 *	using the QIC120, QIC150 and QIC-24 Streaming Tape Interface.
 *
 *	This driver will support forward seeks on reads.
 */

/*	INCLUDE files 	*/

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/systm.h"
#include "sys/buf.h"
#include "sys/signal.h"

#include "sys/fs/s5param.h"
#include "sys/fs/s5dir.h"
#include "sys/immu.h"
#include "sys/region.h"
#include "sys/cmn_err.h"

#include "sys/proc.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/elog.h"
#include "sys/iobuf.h"
#include "sys/dbuf.h"
#include "sys/scsi.h"
#include "sys/mvme147s.h"
#include "sys/mvmecpu.h"



#define ULONG unsigned long
#define UCHAR unsigned char

#undef TRACE

#include "sys/mvme336t.h"

#define	MULTITAPE	/* multiple tape support */
#define	STATCMD		/* do stat on open */
#undef	BYTESWAP	/* include byteswap code */

#define READWRITE
#undef  READWRITE


#define DEBUG
#undef DEBUG

#ifndef	M147
#define	M147	10
#endif

#define SCS_INIT SCS_REW

#undef p_badcnt
#undef p_badtrk


extern int t147maxbsize;		/* maximum DMA buffer size in conf.c */
extern int m147maxscadr;
extern int m147maxscdev;

/*
 *	MVME147tp Driver Minor Number Assignments
 *
 *	/dev/mt and	Minor	Contr-	Reten-	Rewind on	Append or
 *	/dev/rmt names	Number	oller 	sion?	close ?		Truncate ?
 *	--------------	------	------	----	---------	----------
 *	m147_0tr	0	0	Yes	Yes		Truncate
 *	m147_0tnr	1	0	Yes	No		Truncate
 *	m147_0ar	2	0	Yes	Yes		Append
 *	m147_0anr	3	0	Yes	No		Append
 *	m147_0t		4	0	No	Yes		Truncate
 *	m147_0tn	5	0	No	No		Truncate
 *	m147_0a		6	0	No	Yes		Append
 *	m147_0an	7	0	No	No		Append
 *
 *	m147_1tr	0	1	Yes	Yes		Truncate
 *	m147_1tnr	1	1	Yes	No		Truncate
 *	m147_1ar	2	1	Yes	Yes		Append
 *	m147_1anr	3	1	Yes	No		Append
 *	m147_1t		4	1	No	Yes		Truncate
 *	m147_1tn	5	1	No	No		Truncate
 *	m147_1a		6	1	No	Yes		Append
 *	m147_1an	7	1	No	No		Append
 *
 *	m147_Xt		X*4	X	No	Yes		Truncate
 *	m147_Xtn	X*4+1	X	No	No		Truncate
 *	m147_Xa		X*4+2	X	No	Yes		Append
 *	m147_Xan	X*4+3	X	No	No		Append
 */

/*
 *	MVME147tp Defines
 */

#define NORETENS(d)	(d & 4) 	/* dev -> no retension on open */
#define	RWD(d)		(!(d&1))	/* dev -> rewind flag */
#define	TDEV(d)		(!(d&2))	/* dev -> default to truncate */
#define	ADEV(d)		(d&2)		/* dev -> default to append */
#define	MKDEV(c,p)	((c<<8)|(p<<5)) /* bus|cadr|perf -> minor */
#define SCADR(d)	((d & 0x0700) >> 8)
#define SCDEV(d)	((d & 0x00e0) >> 5)
#define MKINDEX(d)	(((d & 0x07) << 5) | ((d & 0x70) << 4))
#define SCD(d)          ((d & 0x0fe0) >> 5)
#define MKPUNIT(d)	(((d & 0x0700) >> 4) | ((d & 0x00e0) >> 5))

/* dev -> table position */
#define	POS(d) ( (SCADR(d)*m147maxscdev) + SCDEV(d) )

/*
 *	MVME147tp Variables
 */

/*
 *	t147space.h
 */

extern	struct	iotime	m147stat[];
extern	struct	iobuf	m147utab[];
extern struct 	scs_tape *m147params[];	/* tape descriptor default params */
extern struct 	scs_tape m147ddefs[];	/* tape open descriptor default params */
extern struct  work m147_work[]; 	/* firm-ware work area */
extern int cpuboard;			/* type of cpu */

#ifndef REL2
extern	struct iostat	t147iostp[];	/* one per controller */
#endif

/*
 *	MVME147tp iobuf usage
 *
 *	iobuf.h		mvme147tp.c	Usage
 *	-------		---------	-----
 *	b_flags		b_flags		Drive state
 *	b_forw
 *	b_back
 *	b_actf		b_actf		head (first) of I/O queue
 *	b_actl		b_actl		tail (last) of I/O queue
 *	b_dev		b_dev		major/minor device number
 *	b_active	b_active	busy flag
 *	b_errcnt	b_errcnt	error count
 *	io_erec		io_erec		error record
 *	io_nreg		io_nreg		number of registers to log upon error
 *	io_addr		io_addr		csr address
 *	io_mba		io_chan		channel structure address
 *	io_stp		io_stp		I/O statistics
 *	io_start	io_start	I/O statistics
 *	io_s1		io_bufp		buffer pointer for double buffering
 *	io_s2		io_bsize	buffer size
 *	io_s3		io_blkno	offset block number
 *	io_s4		io_nxrec	next record
 *	io_s5		io_freeq	free queue pointer
 *	io_s6		io_fileno	file number on tape
 *	io_s7		io_desc		tape descriptor (configuration struct)
#ifdef	MULTITAPE
 *	io_s8		io_proc		process id
#else
 *	io_s8
#endif
 *	io_s9(loword)	io_stat		status from last command
#ifdef	READWRITE
*       io_s9(hiword)	io_moreflgs	more flags 
#endif
 */

/*
 *	MVME147 Tape IOCTL support
 */

#define	M147IOCTL	('M'<<8)
#define	M147REWIND	(M147IOCTL|1)	/* rewind tape */
#define	M147ERASE	(M147IOCTL|2)	/* erase tape */
#define	M147RETENSION	(M147IOCTL|3)	/* retension tape */
#define	M147WRTFM	(M147IOCTL|4)	/* write filemark */
#define	M147RDFM	(M147IOCTL|5)	/* read filemark */
#define	M147SETDMA	(M147IOCTL|6)	/* set DMA buffer size */
#define	M147GETDMA	(M147IOCTL|7)	/* get DMA buffer size */
#define	M147BYTESWAP	(M147IOCTL|8)	/* set/reset byteswapping */


#ifdef TRACE

int t147_trace = 0;

extern chktrc();

#define TTRAIL(x) (('T'<<24)|('>'<<16)|('>'<<8)|('x'))

#define TTRC(id,a,b,c,d,e,f,g) if(t147_trace){chktrc(id,a,b,c,d,e,f,g);}

#else

#define TTRC(id,a,b,c,d,e,f,g) 

#endif

/* RTS on */



#define	INVALID	0
#define	VALID	1

#ifdef READWRITE
/* io_moreflgs definitions (upper word), loword is io_proc */
#define RDWROPEN 0x00010000		/* opened for read or write, not both*/ 
#define READONLY 0x00020000		/* reads only are permitted. */
#define WRITEONLY 0x00040000		/* writes only are permitted. */
#endif

#define	INF	0x7fffffff		/* a very large number */
#define	B_CMD	B_FORMAT
#define	B_INIT	B_VERIFY
#define	B_ERASE	B_DONE
#define	B_OFF	B_ASYNC			/* drive offline */
#ifdef	BYTESWAP
#define	B_BSWAP	B_STALE			/* byte swapping on */
#endif


#ifdef	DEBUG		/* debugging code */

#define	STATIC
#define	DEBUGPR(x,d,s)	{ if( x & m147tpdebug ) t147print(d,s); }
#define PRINT1(a)	{ if(t147tpdebug) printf(a); }
#define PRINT2(a,b)	{ if(t147tpdebug) printf(a,b); }
#define PRINT3(a,b,c)	{ if(t147tpdebug) printf(a,b,c); }
#define PRINT4(a,b,c,d)	{ if(t147tpdebug) printf(a,b,c,d); }
#define PRINT5(a,b,c,d,e)	{ if(t147tpdebug) printf(a,b,c,d,e); }
#define PRINT6(a,b,c,d,e,f)	{ if(t147tpdebug) printf(a,b,c,d,e,f); }
#define PRINT(a)	{ if(m147tpdebug) printf(a); }

#define	DBINIT		(1<<0)
#define	DBOPEN		(1<<1)
#define	DBBOPEN		(1<<2)
#define	DBCOPEN		(1<<3)
#define	DBCLOSE		(1<<4)
#define	DBBCLOSE	(1<<5)
#define	DBCCLOSE	(1<<6)
#define	DBSTRAT		(1<<7)
#define	DBSTART		(1<<8)
#define	DBINTR		(1<<9)
#define	DBREAD		(1<<10)
#define	DBWRITE		(1<<11)
#define	DBIOCTL		(1<<12)
#define	DBCLR		(1<<13)
#define	DBDUMP		(1<<14)
#define	DBDOIO		(1<<15)

#else			/* no debugging code */

#define	STATIC	static
#define	DEBUGPR(x,d,s)
#define PRINT1(a)
#define PRINT2(a,b)
#define PRINT3(a,b,c)
#define PRINT4(a,b,c,d)
#define PRINT5(a,b,c,d,e)
#define PRINT6(a,b,c,d,e,f)
#define PRINT(a)

#endif

#define	io_chan		io_mba
#define	io_bufp		io_s1
#define	io_bsize	io_s2
#define	io_blkno	io_s3
#define	io_nxrec	io_s4
#define	io_freeq	io_s5
#define	io_fileno	io_s6
#define	io_desc		io_s7
#ifdef	MULTITAPE
#define	io_proc		io_s8
#endif
#define	io_stat		io_s9
#ifdef READWRITE
#define io_moreflgs	io_s9
#endif





extern struct scs_pkt	m147pkt[];	/* packets - one per controller */
extern struct scs_pkt *m147freeq;	/* free queue - one  */


extern	int	t147packets;		/* number of packets */
extern	int	t147envelopes;		/* number of envelopes */

/*
 *	conf.h
 */

extern	unsigned char *d147_addr[];
extern	int		t147_ilev[];
extern	int		t147_ivec[];
extern	int		t147_cnt;
extern	int		t147_bmaj;
extern	int		t147_cmaj;

/*
 *	local variables
 */




#ifdef DEBUG
/* tape debug mask, see mvme147.h for bit encoding */
int	m147tpdebug = 1;
int	t147tpdebug = 1;
#endif


/*
 *	MVME147tp Procedures
 *
 *	t147init: initialize the mvme147tp and the driver
 */

t147init()
{
	register struct iobuf *dp = &m147utab[0];
	struct scs_tape **tt = (struct scs_tape **)m147params;
	struct scs_tape *tmp;
	struct scs_tape *tdefs = (struct scs_tape *)&m147ddefs[0];
	int sc_adr, sc_dev;
	
	DEBUGPR(DBINIT,0,"init");


	if((cpuboard != VME147) || bprobe( d147_addr[0], -1 )) {
		d147_addr[0] = NULL;
		return;
	}

	t147_ivec[0] /= 4;

	t147_chinit();

	for( sc_adr = 0; sc_adr < m147maxscadr; sc_adr++ ) {
		for(sc_dev = 0; sc_dev < m147maxscdev; sc_dev++,tt++,tdefs++) {
	
			/* for disk init see mvme147dk.c */
			tmp = *tt;
			/* init tape entry */
			if (tmp->TAPE1.device_type == TAPE) {
				tdefs->TAPE1 = tmp->TAPE1;
				tdefs->TAPE2 = tmp->TAPE2;
				t147devinit(MKDEV(sc_adr,sc_dev) ,tdefs);
				/* increment defined range of devices */	
			}
		}
	}
}

STATIC t147devinit(dev,tt)
register unsigned short dev;
struct scs_tape *tt;
{
	register struct iobuf *dp;

	dp = &m147utab[POS(dev)];
	if ((dp->io_addr = (paddr_t)d147_addr[0]) == NULL)
		return;

	PRINT4("addr: %x, dp: %x, tt: %x\n", dp->io_addr,dp,tt);

	dp->io_nreg = sizeof(struct scs_tape);


#ifdef READWRITE
	dp->io_stat &= 0xffff0000;
#endif
	dp->io_stat = 0;
	dp->io_bsize = t147maxbsize;
	dp->io_fileno = 0;
	tt->TAPE1.device_type = SCS_TAPE;
	dp->io_desc = (int)tt;
	dp->b_flags = B_INIT;		/* init 147tp first time */
}

STATIC t147_chinit()
{
	/*
	 *	initialize channel and envelopes
	 */

	/* create the queue  if not already made */
	if(m147freeq == (struct scs_pkt *)NULL)
	{
		m147freeq = &m147pkt[0];
		m147crque(m147freeq,M147PKTS,m147pkt);
	}

}

/*
 *	t147open: open an mvme147tp device
 */

STATIC t147open(dev,flag,raw)
register dev;
register flag;
register raw;
{	
	register struct iobuf *dp;
	register unsigned short cstat;
	int stat;
	int	t147strategy();

	DEBUGPR(DBOPEN,dev,"open");

	/*
	 *	if the device is nonexistent, then error
	 */
	if( (SCADR(dev) > (m147maxscadr-1)) ||
		 (SCDEV(dev) > (m147maxscdev-1) ) )
	{
		u.u_error = ENXIO;
		return;
	}

	dp = &m147utab[POS(dev)];

	if((dp->b_flags&B_OPEN)||(dp->io_addr == NULL)) {
		u.u_error = ENXIO;
		return;
	}


#ifdef READWRITE
	/*
	 *	allow device to be opened for both read and write, but 
	 *	depending on first access, only allow read only or write only.
	 */

	if((flag&FREAD) && (flag&FWRITE)){
		dp->io_moreflgs &= ~( READONLY | WRITEONLY);
		dp->io_moreflgs |= RDWROPEN;
	}
#else
	/*
	 *	will not permit read-and-write opens
	 */

	if( (flag&FREAD) && (flag&FWRITE) ) {
		u.u_error = ENXIO;
		return;
	}
#endif
	/*
	 *	set up the device table
	 */

	dp->b_flags &= ~B_OFF;
	dp->b_flags |= B_OPEN;
	dp->b_dev = dev;
	dp->io_blkno = 0;
	dp->io_nxrec = INF;

#ifdef	STATCMD
	if(t147command(dp,SCS_STAT,0))
		u.u_error = 0;
	cstat = dp->io_stat;
	/*
	 *	if the unit is offline and the "O_NDELAY" flag
	 *	isn't used, or if the unit is write protected and
	 *	the file is being opened for writing, then error.
	 */

	if ( ( ((flag&FNDELAY)==0) && ( cstat == S_OFFLINE) )
	|| ( dp->b_flags & B_OFF)) {
		dp->b_flags &= ~B_OPEN;
		u.u_error = ENXIO;
		return;
	}
	if ( (flag&FWRITE) && (cstat == S_WP) ) {
		dp->b_flags &= ~B_OPEN;
		u.u_error = ENXIO;
		t147print(dev," Write protected"); 
		return;
	}
#endif
	/* 	Retension tape for all write opens unless "NORETENS" bit 
	 *	is set in minor device number.
	 *
	 */
	if( (flag&FWRITE) && (! NORETENS(dev))) {
		if(t147command(dp, SCS_RET, 0)) {
			u.u_error = ENXIO;
			dp->b_flags &= ~B_OPEN;
			return;
		}
	}
	/*
	 *	if being opened for a write, check for truncate or append
	 */

	if( flag&FWRITE ) {
		
		/* set default action */
		if( (flag&(FTRUNC|FAPPEND)) == 0 )
			flag |= ADEV(dev)? FAPPEND : FTRUNC;

		if( flag&FTRUNC ) {	/* O_TRUNC == truncate ... */

			/*
			 *	force a rewind ...
			 *	can't overwrite on streamer tape
			 */

			if( t147command(dp,SCS_REW,0) ) {
				u.u_error = ENXIO;
				dp->b_flags &= ~B_OPEN;
				return;
			}


		} else {
			/* append, so position at end of data */

			if( t147command(dp,SCS_SPACE,SKEOD) ) {
				u.u_error = ENXIO;
				dp->b_flags &= ~B_OPEN;
				return;
			}
		}
	}

	/*
	 *	If the tape is opened as a character device (raw),
	 *	and if the O_NDELAY flag is not used, then allocate
	 *	a set of buffers to implement double buffering.
	 */

	if( raw && dp->io_bsize && ((flag&FNDELAY) == 0) ) {
#ifdef	MULTITAPE
		if( dp->io_bufp != NULL ) {
			if( u.u_procp->p_pid != dp->io_proc ) {
				dbufclose(dp->io_bufp);
				dp->io_bufp = NULL;
				printf("MVME147tp: DMA buffers discarded\n");
			} else {
				/* set up retry */
				dbufioctl( dp->io_bufp, DBUFRETRY, 0 );
			}
		} 

		if( dp->io_bufp == NULL ) {
#endif
			dp->io_bufp =
				stat = dbufopen(dp->io_bsize,
					dp->b_dev,t147strategy);
				if(stat == NULL)
				{
					PRINT("Error from dbufopen\n");
				}
			if( dp->io_bufp == NULL ) {
				dp->b_flags &= ~B_OPEN;
				return;
			}
#ifdef	MULTITAPE
		}
#endif
	}
#ifdef	MULTITAPE
	else if( dp->io_bufp ) {
		dbufclose(dp->io_bufp);
		dp->io_bufp = NULL;
		printf("MVME147tp: DMA buffers discarded\n");
	}

	dp->io_proc = u.u_procp->p_pid;
#endif


	DEBUGPR(DBOPEN,dev,"open successful");
	return;
}

/*
 *	t147bopen: open a block device
 */

t147bopen(dev,flag)
{
	DEBUGPR(DBBOPEN,dev,"bopen");
	t147open(dev,flag,0);
}

/*
 *	t147copen: open a character (raw) device
 */

t147copen(dev,flag)
{
	DEBUGPR(DBCOPEN,dev,"copen");
	t147open(dev,flag,1);
}

/*
 *	t147close: close an mvme147tp device
 */

STATIC t147close(dev,flag,raw)
register dev;
{	register struct iobuf *dp = &m147utab[POS(dev)];

	DEBUGPR(DBCLOSE,dev,"close");

#ifdef READWRITE
	if( (flag&FREAD) && (flag&FWRITE))
		dp->io_moreflgs &= ~(RDWROPEN | READONLY | WRITEONLY);

#endif
	/*
	 *	if the io_bufp is non-null, then flush and
	 *	delete the DMA buffers used for double buffering.
	 */

	if( dp->io_bufp ) {
#ifdef	DEBUG
		if( ! raw )
			t147print(dev,"illegal close");
#endif
#ifdef	MULTITAPE
		if( (flag&FREAD) || ((dp->b_flags&B_ERROR) == 0) ) {
#endif
			dbufclose(dp->io_bufp);
			dp->io_bufp = NULL;
#ifdef	MULTITAPE
		} else {
			dp->b_flags &= ~B_ERROR;
			printf("MVME147tp: DMA buffers still active.\n");
		}
#endif
	}

	/*
	 *	if the unit was opened for writing, then write
	 *	a filemark to indicate end of file
	 */

	if( dp->b_flags&B_ERASE )
		dp->b_flags &= ~B_ERASE;

	else if( flag&FWRITE )
		t147command(dp,SCS_WEOF,0);

	/*
	 *	if the unit is a "rewind-on-close", then rewind the device
	 */


	if( RWD(dev) )
		t147command(dp,SCS_REW,0);
	else if(dp->io_blkno && (flag&FREAD)) /* no FM, seek to next file */
		t147command(dp,SCS_SPACE,SKNEXT);

	dp->b_flags &= ~B_OPEN;

	DEBUGPR(DBCLOSE,dev,"close successful");
}

/*
 *	t147bclose: close a block device
 */

t147bclose(dev,flag)
{
	DEBUGPR(DBBCLOSE,dev,"bclose");
	t147close(dev,flag,0);
}

/*
 *	t147cclose: close a character device
 */

t147cclose(dev,flag)
{
	DEBUGPR(DBCCLOSE,dev,"cclose");
	t147close(dev,flag,1);
}

/*
 *	t147command: issue a command and call strategy
 */

STATIC t147command(dp,cmd,arg)
register struct iobuf *dp;
register cmd;
register arg;
{	register struct buf *bp;

	bp = geteblk();			/* get empty buffer */
	bp->b_proc = u.u_procp;
	bp->b_dev = dp->b_dev;
	bp->b_bcount = cmd;
	bp->b_blkno = arg;
	bp->b_error = 0;
	bp->b_flags |= B_CMD|B_PHYS;
	t147strategy(bp);
	iowait(bp);			/* wait until done */
	bp->b_flags &= ~(B_CMD|B_PHYS);
	brelse(bp);
	return(u.u_error);
}

/*
 *	t147strategy: perform an I/O
 */

t147strategy(bp)
register struct buf *bp;
{	register struct iobuf *dp = &m147utab[POS(bp->b_dev)];
	register lvl;

	DEBUGPR(DBSTRAT,bp->b_dev,"strategy");

	if(((bp->b_flags&B_CMD ) == 0) && (bp->b_bcount&(BSIZE-1))) {
		/* no partial block I/O */
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}

#ifdef READWRITE
	 /* if device was opened for read/write then depending on first access,
	    permit reads only or writes only. 
	 */
	if(((bp->b_flags&B_CMD) == 0) && (dp->io_moreflgs & RDWROPEN)){
		if(!(dp->io_moreflgs & (READONLY | WRITEONLY))){
			if(bp->b_flags & B_READ)
				dp->io_moreflgs |= READONLY;
			else
				dp->io_moreflgs |= WRITEONLY;
		}
		else{
	     	      if(!(((bp->b_flags&B_READ) && (dp->io_moreflgs&READONLY))
		    || (((bp->b_flags&B_READ) == 0) && (dp->io_moreflgs&WRITEONLY))))
		        {	 /* cant mix reads and writes */
				bp->b_flags |= B_ERROR;
				bp->b_error = ENXIO;
				iodone(bp);
				return;
			}
		}
	}
#endif

	bp->b_start = lbolt;
	if(( bp->b_flags&B_CMD ) == 0 )
		m147stat[POS(bp->b_dev)].io_bcnt += btoc(bp->b_bcount);
	bp->av_forw = NULL;

	lvl = spl6();

	if( dp->b_actf == NULL )
		dp->b_actf = bp;
	else
		dp->b_actl->av_forw = bp;
	dp->b_actl = bp;
	if( dp->b_active == 0 )
		t147start(dp);

	splx(lvl);
}

/*
 *	t147start: start an I/O
 */

STATIC t147start(dp)
register struct iobuf *dp;
{	register struct buf *bp;
	register struct scs_pkt *ep;
	struct scs_tape *tp_desc = (struct scs_tape *)dp->io_desc;
	register cmd;
	register bpblkno;
	struct scs_pkt *m147get();
	int t147ihand();

	DEBUGPR(DBSTART,dp->b_dev,"start");

	/*
	 *	get next buffer ... if null return
	 */

	while(( bp = dp->b_actf ) != NULL ) {

		dp->b_active++;

		if(( bp->b_flags&B_CMD ) == 0 ) {

			/*
			 *	if the current block number is greater
			 *	than the next record number on the tape,
			 *	then it's an error (lseek(2) too far).
			 */

			if( (bpblkno = bp->b_blkno) > dp->io_nxrec ) {
				bp->b_error = EIO;
				goto abort;
			}

			/*
			 *	if the current block is equal to the
			 *	next record number, then we're at the
			 *	end of the tape file (on a filemark).
			 */

			if((bpblkno==dp->io_nxrec)&&(bp->b_flags&B_READ)) {
				clrbuf(bp);
				bp->b_resid = bp->b_bcount;
				goto next;
			}

			/*
			 *	set the next record number to one past the
			 *	buffer block number (on writes).
			 */
			if(( bp->b_flags&B_READ ) == 0 )  {
				PRINT1("start_write: ");
			}
			else {
				PRINT1("start_read: ");
			}

			PRINT5("bp: %x, dp: %x, bpb: %x, dpb: %x\n", 
				bp, dp, bp->b_blkno, dp->io_blkno);
			if(( bp->b_flags&B_READ ) == 0 ) {
				dp->io_nxrec = bpblkno + bp->b_bcount/BSIZE;

				/*
				 *	no seeks on write
				 */

				if( dp->io_blkno != bp->b_blkno ) {
					PRINT1("t147start: unexpected write\n");
					bp->b_error = EIO;
					goto abort;
				}
			} else if( dp->io_blkno > bpblkno ) {
				bp->b_error = ENXIO;
				goto abort;
			}

		}

		if( dp->b_flags&B_OFF )
			goto abort;

		/*
		 *	if command, then set command
		 */

		if( dp->b_flags&B_INIT )
			cmd = SCS_TPATTACH;
		else if( (dp->b_flags&B_ONCE) == 0 )
			cmd = SCS_INIT;
		else if( bp->b_flags&B_CMD )
			cmd = bp->b_bcount;	/* see t147command */
		else
			cmd = (bp->b_flags&B_READ)? SCS_READ : SCS_WRITE;

		/*
		 *	set up packet
		 */

		if((ep = m147get()) == NULL) {  /* no packets */
				PRINT2("t147*out of packets: dev: %x\n", 
					dp->b_dev);
				bp->b_error = EIO;
				goto abort;
			}
		ep->scs_funcc = cmd;
		ep->scs_ctrlr = SCADR(dp->b_dev);
		ep->scs_dev = SCDEV(dp->b_dev);
		ep->utab = dp; /* save pointer to struct iobuf */
		/* save pointer to interrupt handler */
		ep->inthand = t147ihand; 
		/* set the vector number */
		ep->scs_vnum = t147_ivec[0];
		/* set the interrupt level */
		ep->scs_level = t147_ilev[0];
		ep->scs_stat0 = ep->scs_stat1 = ep->scs_stat2 = ep->scs_stat3 = 0;
		ep->ETAPE2.work_area = (unsigned long *)&m147_work[POS(dp->b_dev)];

		switch(cmd) {

		case SCS_READ:	/* read to eof */
			ep->RDWR.block = bp->b_blkno;

		case SCS_WRITE:	/* write logical */

			if( bp->b_flags&B_PHYS ) {
				dp->io_stp->io_ops++;

			} else if( dp->io_blkno == bpblkno ) {
				dp->io_stp->io_ops++;

			} else	/* do block read and throw away later */
				dp->io_stp->io_misc++;
#ifdef DEBUG
			if(cmd == SCS_READ) {
				PRINT1("read: ");
			}
			else {
				PRINT1("write: ");
			}
#endif
			PRINT5("bp: %x, dp: %x, bpb: %x, dpb: %x\n", 
				bp, dp, bp->b_blkno, dp->io_blkno);
			ep->RDWR.address = (unsigned char *)paddr(bp);
			ep->RDWR.block_count = bp->b_bcount/BSIZE;
			ep->RDWR.scga_count = 0;

			if( dp->io_blkno < bp->b_blkno ) {
				/* seek forward ... but not too far */
				if(ep->RDWR.block_count > (bp->b_blkno-dp->io_blkno))
					ep->RDWR.block_count = bp->b_blkno-dp->io_blkno;
			}
			break;

		case SCS_TPATTACH:
			ep->ETAPE1 = tp_desc->TAPE1;
			ep->ETAPE2 = tp_desc->TAPE2;
			ep->ETAPE2.work_area = tp_desc->TAPE2.work_area = 
				(unsigned long *)&m147_work[POS(dp->b_dev)];
			break;


		case SCS_REW:	/* rewind */
			/* allow seek to end of data after write eof */
			dp->io_fileno = 0;
			dp->io_stp->io_misc++;
			/* only respond when complete */
			ep->REWIND.immediate = 0;
			break;
		case SCS_WEOF:	/* write end-of-file */
			ep->WRTFM.number_of_fm = 1;
			bp->b_blkno = 0;	/* for rewind on close */
		case SCS_ERASE:	/* erase */
		case SCS_STAT:	/* get status */
			dp->io_fileno = 0;
			dp->io_stp->io_misc++;
			break;

		case SCS_RET:	/* retension */
			ep->LDUNL.mode = RETENSION;
			dp->io_fileno = 0;
			dp->io_stp->io_misc++;
			break;

		case SCS_SPACE:	/* seek */
			dp->io_stp->io_misc++;
			dp->io_stp->io_misc++;
			/* the mode is saved in b_blkno */
			ep->SPACE.mode = bp->b_blkno;
			if(bp->b_blkno == SKNEXT)	/* next file */
				ep->SPACE.number = 1;
			break;
		default:
			PRINT2("MVME147tp bad command: %x\n", cmd );
			break;
		}

		blkacty |= (1<<M147);

		/* set up I/O start time */
		if( dp->io_start == 0 )
			dp->io_start = lbolt;
		 
		/*
		 *	everything ok ... send
		 */
		TTRC(TTRAIL(1),dp,ep,ep->ETAPE2.work_area,ep->RDWR.block,ep->RDWR.block_count, ep->RDWR.address,0);
		M147SEND(ep);
		return;

		/*
		 *	come here on error
		 */

	abort:
		bp->b_flags |= B_ERROR;
	next:
		dp->b_actf = bp->av_forw;
		iodone(bp);
	}

	dp->b_active = 0;
}

/*
 *	t147intr: interrupt handler
 */

t147ihand(ep)
register struct scs_pkt *ep;
{	
	register struct iobuf *dp;
	register struct buf *bp;
	register dev;
	register resid;
	register unsigned short stat;
	register fatal;
	register termcnt;
	register cmd;
	unsigned long txfer_count;
	struct cmd_tbl *cc;

	DEBUGPR(DBINTR,0,"intr");


	/*
	 *	get status envelope
	 */

#ifdef	DEBUG
	if(d147_addr[0] == NULL)	
	{
		t147print(dev,"null address in interrupt");
		return;
	}
#endif


	dp = ep->utab;
	dev = dp->b_dev;
	cc = (struct cmd_tbl *)&m147_work[POS(dp->b_dev)];

	bp = dp->b_actf;

	if( bp == NULL ) {
		logstray(dp->io_addr);
		return;
	}

	blkacty &= ~(1<<M147);
	dp->b_active = 0;

	
	cmd = ep->scs_funcc;
	termcnt = ((cc->transfer_address - ep->RDWR.address)/BSIZE);
	dp->io_stat = ep->scs_stat1;
	stat = ep->scs_stat3;
	fatal = ep->scs_stat1;

	PRINT6("intr: dev: %x, cmd: %x, cnt: %x, fatal: %x, stat: %x\n",
		dev, cmd, termcnt, fatal, stat);

	TTRC(TTRAIL(2),dp,ep,ep->ETAPE2.work_area,ep->RDWR.block,ep->RDWR.block_count, ep->RDWR.address,0);

	/* file mark or end of media, shouldn't have an error */
	if((stat&S_FM)  || (stat & S_EOM))
	{
#		ifdef DEBUG
		printf("fatal: %x, status: %x\n", fatal, stat);
		printf("termcnt: %d\n", termcnt);
#		endif
		fatal = 0;
	}
	else if( fatal ) {
		if( fatal == S_WP )  /* for STAT command */
			dp->io_stat |= S_WP;
		else {
			/* don't allow any more commands, force init
				and rewind on next open */
			dp->b_flags |= ( B_INIT | B_OFF);
			dp->b_flags &=  ~(B_ONCE);
			bp->b_flags |= B_ERROR;
			t147err(dp,bp,ep);
			m147free(ep);	/* give back envelope */
			goto complete;
		}
	}
	m147free(ep);	/* give back envelope */

	if((cmd == SCS_READ ) || (cmd == SCS_WRITE )) {
		/* termcnt is already in blocks */
		resid = termcnt;
		dp->io_blkno += resid;
		if( (stat&S_FM )  || (stat&S_EOM)){
			dp->io_blkno = 0;
			dp->io_nxrec = bp->b_blkno + resid;
		} else {
			if( dp->io_blkno <= bp->b_blkno ) {
				/* ignore I/O ... continue */
				t147start(dp);
				return;
			}
		}

	/*
	 *	
	 */

	} else if( cmd == SCS_INIT ) {
		if(( dp->b_flags&B_ONCE) == 0) {	/*  */
			dp->b_flags |= B_ONCE;		/* */
			t147start(dp);
			return;
		}
	
	
	/*
	 *	if an init command has been issued
	 */

	} else if( dp->b_flags&(B_INIT) ) {
		dp->b_flags &= ~(B_INIT);
		t147start(dp);		/* do command */
		return;
	}


	/*
	 *	I/O is complete ... continue
	 */

	complete:

	logberr(dp,stat);
	dp->b_actf = bp->av_forw;

	if(((bp->b_flags&B_CMD ) == 0) && 
	(!(fatal) && ((cmd == SCS_WRITE ) || (cmd == SCS_READ)))) {
		/* termcnt is in blocks, so convert to bytes as is b_bcount */
		resid = (bp->b_bcount - (termcnt*BSIZE));
		bp->b_resid = (resid <= 0)? 0 : resid;

	} else if((bp->b_flags&B_CMD ) == 0) 
		bp->b_resid = bp->b_bcount;


	{	register struct iotime *ip = &m147stat[0];

		ip->io_resp += lbolt - bp->b_start;
		ip->io_act += lbolt - dp->io_start;
		dp->io_start = 0;
	}


	iodone(bp);		/* complete the I/O */
	t147start(dp);		/* start next I/O */
}

#ifndef REL2

t147breakup(bp)
{
	dma_breakup(t147strategy,bp,BSIZE);
}

#endif

/*
 *	t147read: character device read
 */

t147read(dev)
register dev;
{	register struct iobuf *dp = &m147utab[POS(dev)];

	DEBUGPR(DBREAD,dev,"read");

#ifdef READWRITE
	if(dp->io_moreflgs&RDWROPEN){
		if(dp->io_moreflgs&WRITEONLY){
			u.u_error = ENXIO;
			return;
		}
		else
			dp->io_moreflgs |= READONLY;
	}
#endif
	
	if( dp->io_bufp )
		dbufread( dp->io_bufp );
	else
 		physio( t147breakup, 0, dev, B_READ );
	return;
}

/*
 *	t147write: character device write
 */

t147write(dev)
{	register struct iobuf *dp = &m147utab[POS(dev)];

	DEBUGPR(DBWRITE,dev,"write");

#ifdef READWRITE
	if(dp->io_moreflgs&RDWROPEN){
		if(dp->io_moreflgs&READONLY){
			u.u_error = ENXIO;
			return;
		}
		else
			dp->io_moreflgs |= WRITEONLY;
	}
#endif
	
	if( dp->io_bufp ) {
		dbufwrite( dp->io_bufp );
#ifdef	MULTITAPE
		if( u.u_error || u.u_count != 0 )
		{
			dp->b_flags |= B_ERROR;	/* remember error */
		}
#endif
	} else
 		physio( t147breakup, 0, dev, B_WRITE );
	return;
}

/*
 *	t147ioctl: ioctl support for the MVME147tp
 */

t147ioctl( dev, cmd, arg, mode )
register dev;
{	register struct iobuf *dp = &m147utab[POS(dev)];

	DEBUGPR(DBIOCTL,dev,"ioctl");

	switch(cmd) {
	case M147REWIND:	/* rewind tape */
		t147command(dp,SCS_REW,0);
		break;
	case M147ERASE:		/* erase tape */
		if( t147command(dp,SCS_ERASE,0) == 0 )
			dp->b_flags |= B_ERASE;
		break;
	case M147RETENSION:	/* retension tape */
		t147command(dp,SCS_RET,0);
		break;
	case M147WRTFM:		/* write filemark */
		t147command(dp,SCS_WEOF,0);
		break;
	case M147RDFM:		/* read filemark */
		t147command(dp,SCS_SPACE,SKNEXT);
		break;
#ifdef	BYTESWAP
	case	M147BYTESWAP:	/* set/reset byte swapping */
		if( arg != 0 )
			dp->b_flags |= B_BSWAP;
		else
			dp->b_flags &= ~B_BSWAP;
		t147command(dp,SCS_TPATTACH,0);
		break;
#endif
	case M147SETDMA:	/* set DMA buffer size */
		if( ! suser() ) {
			u.u_error = EPERM;
			return;
		}

		if( dp->io_bufp == NULL )
			dp->io_bsize = arg;

		else {
			dbufioctl(dp->io_bufp,DBUFSET,arg);
			if( u.u_error )
				dp->io_bsize = 0;
			else
				dp->io_bsize = arg;
		}
		break;
	case M147GETDMA:	/* get DMA buffer size */
		if(( dp->io_bufp == NULL ) || ( dp->io_bsize == 0 )) {
			copyout( &dp->io_bsize, arg, sizeof(dp->io_bsize) );
		} else
			dbufioctl(dp->io_bufp,DBUFGET,arg);
		break;


	default:
		u.u_error = EINVAL;
		break;
	}
}

/*
 *	t147clr: clear the mvme147tp after a power fail
 */

t147clr()
{
	DEBUGPR(DBCLR,0,"clr");
}

/*
 *	t147err:  process mvme147 return status
 */

STATIC t147err(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct scs_pkt *ep;
{
	register t147eblk;		/* block # where error occurred */
	register unsigned short svdev;
	register s;

	/* set up pointer to "registers" for fmtberr() */
	/* This changes the b_dev field so must save */
	svdev = dp->b_dev;
	dp->b_dev = makedev(M147,minor(dp->b_dev));
	dp->io_addr = (paddr_t)ep;

	/*
	 * calculate block that error was on for fmtberr()
	 * and save it for t147prerr
	 */

	t147eblk = ep->RDWR.block;
	fmtberr(dp,t147eblk);		/* error block */
	dp->b_dev = svdev;		/* restore device */

	t147prerr(bp->b_dev,ep,t147eblk);	/* display error */
}


/*
 *	t147prerr: print out error information
 */

STATIC t147prerr(dev,ep,t147eblk)
register dev,t147eblk;
register struct scs_pkt *ep;
{
	register unsigned char fat;
	register unsigned char rec;
	register cmd;

	rec = ep->scs_stat3;
	fat = ep->scs_stat1;

	if (fat)
		t147print(dev,"\nFATAL ERROR");
#ifdef	DEBUG
	else if (rec)
		t147print(dev,"\nRECOVERED ERROR");

	t147prcmd((cmd=ep->scs_funcc)); 

	printf("Fatal and Recovered Status: %x %x\n", fat, rec);
#endif

	if( rec & S_EOM)
		printf("\t End of media");
	else switch(fat) {
			case S_WP:	printf("\t Write protected"); break;
			case S_EOD:	printf("\t End of data"); break;
			case S_OFFLINE: printf("\t Tape not ready"); break;
			case S_ILLEGAL: printf("\t Illegal request"); break;
			case S_TOUT: printf("\t Controller timeout"); break;
			default: printf("\t Fatal error status: 0x%x",fat);
		}
	if (fat && (cmd == SCS_READ || cmd == SCS_WRITE))
		printf("\n\t Logical Block: 0x%x(%d)",t147eblk,t147eblk);
	printf("\n");
}

/*
 *	t147print: driver print procedure
 */

t147print(dev,str)
register dev;
register char *str;
{

	printf("%s on MVME147 SCSI ctl %d, Tape drive %d\n", str,
			SCADR(dev),SCDEV(dev));
}


/*
 *	t147prcmd: print MVME147tp command
 */

STATIC t147prcmd(cmd)
register unsigned short cmd;

{	switch(cmd) {
	case SCS_READ:	t147prm("Read logical");		break;
	case SCS_STAT:	t147prm("Read status");			break;
	case SCS_WRITE:	t147prm("Write blocks");		break;
	case SCS_TPATTACH:	t147prm("Write ipc descriptor");	break;
	case SCS_WEOF:	t147prm("Write eof");			break;
	case SCS_REW:	t147prm("Rewind tape");			break;
	case SCS_ERASE:	t147prm("Erase tape");			break;
	case SCS_SPACE:	t147prm("Seek");			break;
	default:
		PRINT2("MVME147tp: Unknown command %x.\n", cmd );
		break;
	}
}

/*
 *	t147prm: print message with MVME: at front
 */

STATIC t147prm(s)
{	PRINT2("MVME147tp: %s.\n", s );
}

/*
 *	dummy routine
 */

t147dump()
{
}
