/*		Copyright (c) 1986,1987 Motorola		*/
/*		All Rights Reserved				*/
/*		@(#)mvme321.c	35.2					*/
/****************************
 *			    *
 *	mvme321 Driver:     *
 *     			    *
 ***************************/
#undef DEBUG

/*******************************************************************************
 *
 *	Include files
 *
 ******************************************************************************/

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/elog.h"
#include "sys/iobuf.h"
#include "sys/systm.h"
#include "sys/dk.h"
#include "sys/dsk.h"
#include "sys/bpp.h"
#include "sys/mvme321.h"

#ifndef REL2
#include "sys/open.h"
#endif

/*
#define	DEBUG	1
#define	DEBUGC	1
*/

#ifdef	DEBUG		/* debugging code */
#define STATIC
#define	DEBUGPR(x,d,s)	{ if (x & m321Debug) m321print(d,s); }

int m321Debug = 0;

#define	DBINIT		(1<<0)
#define	DBBOPEN		(1<<1)
#define	DBCOPEN		(1<<2)
#define	DBBCLOSE	(1<<3)
#define	DBCCLOSE	(1<<4)
#define	DBSTRAT		(1<<5)
#define	DBSTART		(1<<6)
#define	DBCONF		(1<<7)
#define	DBFORMAT	(1<<8)
#define	DBICMD		(1<<9)
#define	DBINTR		(1<<10)
#define	DBERR		(1<<11)
#define	DBREAD		(1<<12)
#define	DBWRITE		(1<<13)
#define	DBIOCTL		(1<<14)
#define	DBCLR		(1<<15)
#define	DBTIMER		(1<<16)
#define	DBDUMP		(1<<17)
#define	DBDOIO		(1<<18)

static char *ASSERTCTL = "ASSERT(%s) failed in %s, line %d\n";
#define ASSERT(X) { if (!(X)) printf(ASSERTCTL, "X", __FILE__, __LINE__);}

#else			/* no debugging code */

#define	STATIC	static
#define	DEBUGPR(x,d,s)
#define ASSERT(X)

#endif

#ifdef	DEBUGC		/* command debugging code */

#define	PRCMD(p)	{ if (m321Cmdbug) m321prcmd(p); }
int m321Cmdbug = 1;

#else

#define	PRCMD(p)	/* no command debugging code */

#endif

/*******************************************************************************
 *
 *	This driver will support up to 4 controllers.
 *	Each MVME321 can control up to 8 drives simultaneously:
 *
 ******************************************************************************
 *
 *	MVME321 Disk Controller Driver Minor Number Assignment Coding
 *
 *		 <-   1    -> <- 1 -> <- 2 -> <- 4 ->
 *		 ------------------------------------
 *		| controller |  type | drive | slice |
 *		 ------------------------------------
 *
 *							Standard Names
 *       	    Unit		 		 Controller
 *   Minor Number  Type/Drv 	  Device               0 Name  1 Name
 *
 *	000X	     0      Winchester drive 0     	 0sX     8sX
 *	001X	     1      Winchester drive 1     	 1sX     9sX
 *	002X	     2      Winchester drive 2     	 2sX    10sX
 *	003X	     3      Winchester drive 3     	 3sX    11sX
 *	004X	     4      Floppy drive 0	   	 4sX    12sX
 *	005X	     5      Floppy drive 1	   	 5sX    13sX
 *	006X	     6      Floppy drive 2	   	 6sX    14sX
 *	007X	     7      Floppy drive 3	   	 7sX    15sX
 *
 ******************************************************************************/


/*******************************************************************************
 *
 *	Minor Number Manipulating Macros
 *
 ******************************************************************************/

#define	SLCMSK		((1<<M321SBTS)-1)
#define	SLICE(d)	(d&SLCMSK)		/* slice number */
#define	POS(d)		((d>>M321SBTS)&0xf)	/* position in utab */
#define	CTL(d)		((d>>(M321SBTS+3))&0x1)	/* controller number */
#define DRV(d)		((d>>M321SBTS)&0x3)	/* drive number */
#define UNIT(d)		((d>>M321SBTS)&0x7)	/* unit (type + drive) */
#define	FLOPPY(d)	(d&(4<<M321SBTS))	/* floppy unit */
#define	MKDEV(c,d)	((c<<(M321SBTS+3))|(d<<M321SBTS)) /* ctl/dev -> minor */

/*******************************************************************************
 *
 *	Global variables
 *
 ******************************************************************************/

#define	MAXDEV	8			/* up to 8 devices per controller */
STATIC	unsigned m321_ctls;
#define	MAXPOS	(m321_ctls*MAXDEV)	/* maximum number of slots */

#define	M321SHRT	15		/* timeout limit for short commands */
#define	M321LONG	1200		/* timeout limit for long commands */

STATIC unsigned char m321adrm;		/* address modifier */
STATIC	int	m321busw;		/* bus width */

STATIC	struct b_envelope *m321frpkts;	/* list of free command packets */
STATIC	struct dsk	m321dsk;	/* disk descriptor for the mvme321 */
STATIC	int	m321eblk;		/* block # where error occurred */
STATIC	ushort	m321wopn;		/* count of open winchesters */

#define	BYCACHE	0x8000			/* bypass cache on read */
#define	NOCACHE	0x4000			/* don't reallocate cache on read */
#define	CACHE	0x0000			/* cache the read */

#define	M321IOCTL	('M'<<8)
#define	INITSTRAT	(M321IOCTL|0x0)	/* initialize strategy function     */
#define	INITCACHE	(M321IOCTL|0x1)	/* initialize cache buffers	    */
#define	INITIPC		(M321IOCTL|0x2)	/* initialize IPC descriptor	    */

/* controller attribute word bit assignments */
#define	CATBSF		0x00000001	/* Buffered seek flag bit	    */

/*******************************************************************************
 *	MVME321 Format Structure
 ******************************************************************************/

struct m321fmt {
	struct dkfmt f_dkf;
	unsigned long *f_badtrks;	/* returned bad trks from mvme321 */
};

#define f_trkcnt	f_dkf.dk_trkcnt
#define	f_trkno		f_dkf.dk_trkno
#define	f_badcnt	f_dkf.dk_badcnt
#define	f_badlst	f_dkf.dk_badlst

/*******************************************************************************
 *	MVME321 Bad Spot Structure
 ******************************************************************************/

struct m321badspot {
	ushort		bs_cyl;		/* cylinder number of bad spot */
	unsigned char	bs_hd;		/* head number of bad spot */
	unsigned char	bs_sec;		/* bad spot sector number (not used) */
	ushort		bs_offset;	/* byte offset of bad spot */
};

/*******************************************************************************
 *	External Variables
 ******************************************************************************/

extern struct b_device *m321_addr[];
extern int m321_bmaj;			/* block major */
extern int m321_cnt;			/* count */
extern int m321_ilev[];			/* interrupt level */
extern int m321_ivec[];			/* interrupt level */

#ifndef REL2
extern int m321otbl[];			/* open table */
#endif

/*******************************************************************************
 *	Driver Variables - declared in conf.c (space.h or io.h)
 ******************************************************************************/

extern struct iobuf	m321utab[];	/* one per drive */
extern struct iotime	m321stat[];	/* one per drive */
extern struct slice	m321slices[];	/* slice tables */
extern struct b_packet	m321pkt[];	/* command packets */
extern struct b_envelope m321env[];	/* envelopes */
extern struct m321wdesc m321wd[];	/* winchester descriptor structures */
extern struct m321fdesc m321fd[];	/* floppy descriptor structures */
extern struct b_channel	m321chan[];	/* channel for MVME321 communication */
extern int 		m321ctlflg[];	/* flag word; one per controller */
extern ushort		m321lim[];	/* winchester packet limits */
extern struct m321cache	m321cache[];	/* default cache configuration */
extern ushort		m321maxbuf[];	/* default max_buf_sz for IPC */

/*******************************************************************************
 *	MVME321 iobuf usage
 *
 *	iobuf.h		mvme321.c	Usage
 *	-------		---------	-----
 *	b_flags		b_flags		Drive state
 *	b_forw
 *	b_back
 *	b_actf		b_actf		head (first) of I/O queue
 *	b_actl		b_actl		tail (last) of I/O queue
 *	b_dev		b_dev		major/minor device number
 *	b_active	b_active	busy flag
 *	b_errcnt	b_errcnt	error count
 *	io_erec		io_erec		error record
 *	io_nreg		io_nreg		number of registers to log upon error
 *	io_addr		io_eregs	address of registers to log on error
 *	io_mba		io_csr		address of device CSR
 *	io_stp		io_stp		I/O statistics
 *	io_start	io_start	I/O statistics
 *	io_s1		io_desc		disk descriptor (configuration struct)
 *	io_s2		io_qcnt		number of bufs on iobuf queue	(dsk.h)
 *	io_s3		io_slccnt	number of allowable slices	(dsk.h)
 *	io_s4		io_badptr	pointer to bad block table	(dsk.h)
 *	io_s5		b_acts		ptr to last pkt processed by IPC(dsk.h)
 *	io_s6		io_pktcnt	packet count (limit & count)	(dsk.h)
 *	io_s7		io_sopen	open slices flag		(dsk.h)
 *	io_s8		io_slctbl	pointer to the slice table	(dsk.h)
 *	io_s9		io_dskptr	pointer to dsk structure	(dsk.h)
 ******************************************************************************/

#define	io_eregs	io_addr
#define	io_csr		io_mba
#define	io_desc		io_s1

/*******************************************************************************
 *	MVME321 Winchester Status Codes
 ******************************************************************************/

#define	S_NOMARK	1<<0	/* uPD7261: no address mark found	      */
#define	S_WRTPROT	1<<1	/* uPD7261: write to write protected disk     */
#define	S_SECERR	1<<2	/* uPD7261: sector not found		      */
#define	S_DRVERR	1<<3	/* uPD7261: drive is not in ready state	      */
#define	S_HWERR		1<<4	/* uPD7261: equipment check condition	      */
#define	S_DATAERR	1<<5	/* uPD7261: error occurred in data field      */
#define	S_FIFOERR	1<<6	/* uPD7261: fifo over/under run		      */
#define	S_CYLEND	1<<7	/* end of cylinder (possible F/W problem)     */
#define	S_DRVILL	1<<8	/* illegal drive number			      */
#define	S_LOCILL	1<<9	/* illegal location requested		      */
#define	S_FMTERR	1<<10	/* format track unsuccessful		      */
#define	S_ILLDESC	1<<11	/* illegal disk descriptor		      */
#define	S_ALTERR	1<<12	/* format: not enough alternate tracks	      */
#define	S_SEEKERR	1<<13	/* error in seek operation		      */
#define	S_BUSY		1<<14	/* uPD7261: busy condition (F/W problem?)     */
#define	S_BADVER	1<<15	/* data does't coincide w/ verify operation   */
#define	S_CRCID		1<<16	/* CRC error				      */
#define	S_RESET		1<<17	/* uPD7261: cmd terminated in a reset	      */
#define	S_ECCERR	1<<18	/* ECC correctable error		      */
#define	S_ABNORM	1<<19	/* uPD7261: abnormal cmd completion indicated */

#define	S_RANGE		1<<8	/* block out of range (alternate area)	      */
#define	S_ALT		1<<9	/* mvme321 had to access alternate track      */
#define	S_MEMERR	1<<10	/* memory error (not enough, etc.)	      */
#define	S_CACHBSY	1<<11	/* cache busy, can't free buffers	      */
#define	S_CACHINVL	1<<12	/* error caused cache to be invalidated	      */
#define	S_NOCACH	1<<13	/* no data cache used for command	      */
#define	S_DCACHMS	1<<14	/* data cache miss			      */
#define	S_ICACHMS	1<<15	/* instruction cache miss		      */

/*******************************************************************************
 *	MVME321 Floppy Status Codes
 ******************************************************************************/

#define	S_BADSIZ	1<<0	/* Bad disk size in descriptor		      */
#define	S_EOTSM		1<<1	/* End-of-transfer size mismatch	      */
#define	S_BADTPI	1<<2	/* Bad tpi combination specified	      */
#define	S_DRVMTR	1<<3	/* Drive motot not coming on		      */
#define	S_DOOR		1<<4	/* Disk door open			      */
#define	S_NCMLPT	1<<5	/* Command not completing		      */
#define	S_BADRSTR	1<<6	/* Bad restore operation		      */
#define	S_ILLSIDE	1<<7	/* Illegal side reference on device	      */
#define	S_ILLTRK	1<<8	/* Illegal track reference on device	      */
#define	S_ILLSEC	1<<9	/* Illegal sector reference on device	      */
#define	S_ILLSTP	1<<10	/* Illegal step rate specified		      */
#define	S_BADDENS	1<<11	/* Bad density specified		      */
#define	S_FWRTPROT	1<<12	/* Write protected disk			      */
#define	S_FFMTERR	1<<13	/* Format error				      */
#define	S_BADLOC	1<<14	/* Can't find side, track, or sector	      */
#define	S_ICRCERR	1<<15	/* CRC error in ID field(s)		      */
#define	S_DCRCERR	1<<16	/* CRC error in data field		      */
#define	S_DMAUR		1<<17	/* MC68440 DMA underrun			      */

/*******************************************************************************
 *	MVME321 Winchester & Floppy Default Configuration Values
 ******************************************************************************/

/* Winchester Values */

#define W_RETRY		3		/* retry count */
#define W_RRTRY		1		/* recalibrate retry count */
#define W_ECCFLG	1		/* ECC flag; ECC in use */
#define W_CRCECC	0		/* CRC/ECC use; use ECC */
#define W_STEP		0x0e		/* stepping rate index */
#define W_BSF		0		/* buffered seek flag */
#define W_INITCHR	0x4e		/* initialization character */
#define	W_BRF		0		/* buffered recalibrate flag; unused */
#define W_SLIPS		1		/* sectors to slip per track */
#define W_ALTCNT	0		/* number of alternate tracks */
#define W_LSECTRK	32		/* sectors per track */
#define W_TRKCYL	1		/* tracks/cylinder (heads) */
#define W_CYLS		2		/* number of cylinders */
#define W_REDCYL	1		/* reduced write cylinder */
#define W_STRCYL	0		/* starting cylinder */
#define W_ENDCYL	1		/* ending cylinder */
#define W_SECSIZ	256		/* sector size */

/* Floppy Values */

#define F_RETRY		3		/* retry count */
#define F_STEP		0		/* stepping rate index */
#define F_DR_TPI	48		/* drive tpi */
#define F_MED_TPI	48		/* medium tpi */
#define F_SIDES		1		/* number of sides */
#define F_TRKSIDE	1		/* tracks per side */
#define F_DENS0		1		/* density of track 0 */
#define F_SCTR0		16		/* sectors/track on track 0 */
#define F_BYTE0		128		/* sector size on track 0 */
#define F_DENS		2		/* density everywhere but track 0 */
#define F_SECTRK	16		/* sectors/track everywhere but trk 0 */
#define F_SECSIZ	256		/* sector size everywhere but track 0 */
#define F_INITCHR	0x4e		/* initialization character */
#define F_FORMAT	1		/* format - IBM */
#define F_PRECOMP	2		/* precompization cylinder */
#define F_CACHE		2		/* cache enable */

/*
 *	m321init: initialize driver
 */

m321init()
{
	register struct dsk *dd;
	register i;
	register struct iobuf *dp;
	register struct b_envelope *ep;
	register struct b_packet *pp;
	int m321strategy();
	int m321start();
	int m321timer();
	int m321print();

	DEBUGPR(DBINIT,0,"init");

#if defined(M68020) && defined(REL2) 	/* because of 330 board */
	m321adrm = DATA32;
	m321busw = 2;
#else
	m321adrm = DATA24;
	m321busw = 1;
#endif

	m321_ctls = m321_cnt/(MAXDEV*M321SLCS);

	/* set up free queue of envelope/packet pairs */
	bppcrque(&m321frpkts,M321PKTS,m321pkt,m321env,0x4d333231,m321adrm,
		m321busw);

	for (i = 0; i < m321_ctls; i++) {
		register struct b_device *rp;
		register unit;

		rp = m321_addr[i];
		if (bprobe(rp, -1)) {		/* read only */
			m321_addr[i] = NULL;	/* controller not present */
			continue;
		}

		/* convert vector address to vector number */
		m321_ivec[i] /= 4;

		/* set up IPC channel */
		if (bppcrchan(&m321chan[i],rp,&m321frpkts,m321_ilev[i],
			m321_ivec[i],m321adrm) == -1) {
			/* create channel error, disable controller */
			m321_addr[i] = NULL;
			printf("MVME321 Create channel error\n");
			printf("MVME321 Controller %x disabled\n", i);
			continue;
		}

		{	register struct pktcnt *pcp;
			register pos;

			/* controller present */
			pos = MAXDEV*i;
			dp = &m321utab[pos];

			/* initialize disks */
			for (unit=0; unit < MAXDEV; dp++, pos++, unit++) {
				dp->b_dev = MKDEV(i,unit);
				dp->io_csr = (physadr)rp;
				dp->io_nreg = 15;	/* length of eregs */
				pcp = (struct pktcnt *)&dp->io_pktcnt;
				pcp->cnt = 0;
				if (FLOPPY(unit<<M321SBTS)) {
					dp->io_desc = (int)&m321fd[pos-4*(i+1)];
					pcp->limit = FLOP_LIM;
					dp->b_flags |= B_REINIT;
				}
				else {
					dp->io_desc = (int)&m321wd[pos-4*i];
					pcp->limit = m321lim[0];
				}
				/*
				 * initialize dp pointers for double linked list
				 */
				dp->b_actl = (struct buf *)dp;
				dp->b_acts = (int)dp;
				dp->io_slctbl = (int)&m321slices[pos*M321SLCS];
				dp->io_dskptr = (int)&m321dsk;
			}
		}
	}
	/* initialize dsk structure */
	m321dsk.d_flags = D_CONFIG|D_FMTU;
	m321dsk.d_strategy = m321strategy;
	m321dsk.d_start = m321start;
	m321dsk.d_timer = m321timer;
	m321dsk.d_print = m321print;
	m321dsk.d_shorttime = M321SHRT * HZ;
	m321dsk.d_longtime = M321LONG * HZ;
	m321dsk.d_slcbits = M321SBTS;

	if( major(rootdev) == m321_bmaj ) {
		m321utab[POS(rootdev)].b_flags |= B_SPECL;
	}

	if( major(swapdev) == m321_bmaj ) {
		m321utab[POS(swapdev)].b_flags |= B_SPECL;
	}

	if( major(pipedev) == m321_bmaj ) {
		m321utab[POS(pipedev)].b_flags |= B_SPECL;
	}

	dskinit(&m321dsk);
}

/*
 *	m321bopen: perform open on a block device
 */

#ifdef REL2
m321bopen(dev, modes)
#else
m321bopen(dev, modes, otyp)
register otyp;
#endif
register dev;
register modes;
{
	DEBUGPR(DBBOPEN,dev,"bopen");

#ifdef REL2
	m321open(&m321utab[POS(dev)], modes, 1<<SLICE(dev));
#else
	m321open(&m321utab[POS(dev)], modes, 1<<SLICE(dev), otyp);
#endif
}

/*
 *	m321copen: perform open on a character device
 */

#ifdef REL2
m321copen(dev, modes)
#else
m321copen(dev, modes, otyp)
register otyp;
#endif
register modes;
{
	DEBUGPR(DBCOPEN,dev,"copen");

#ifdef REL2
	m321open(&m321utab[POS(dev)], modes, 1<<SLICE(dev)+8);
#else
	m321open(&m321utab[POS(dev)], modes, 1<<SLICE(dev), otyp);
#endif
}

/*
 *	m321open: perform open for all devices
 */

#ifdef REL2
STATIC m321open(dp,modes,omask)
#else
STATIC m321open(dp,modes,omask,otyp)
register otyp;
#endif
register struct iobuf *dp;
register modes;
register omask;
{
	register struct buf *bp;
	register struct b_ipc *ip;
	ushort firsttime = 0;
#ifndef REL2
	register *otbl;
	register ushort i;
#endif
	struct buf *geteblk();

	if (m321_addr[CTL(dp->b_dev)] == NULL) {
		u.u_error = ENXIO;
		return;
	}

#ifdef REL2
	if (dp->io_sopen == 0)
		firsttime = 1;
	dskopen(dp, modes, omask);
#else
	otbl = &m321otbl[((CTL(dp->b_dev) * MAXDEV) + UNIT(dp->b_dev))
		* OTYPCNT];

	for( i = 0; i < OTYPCNT && otbl[i] == 0; ++i);
	if( i == OTYPCNT)
		firsttime= 1;

	dskopen(dp, modes, otbl, otyp, omask);
#endif

	if (u.u_error)
		return;
	if (!(FLOPPY(dp->b_dev)) && firsttime) {
		m321wopn++;
		m321setlim(dp->b_dev);
	}

	if ((m321ctlflg[CTL(dp->b_dev)] & B_ONCE) == 0) {
		bp = geteblk();
		clrbuf(bp);

		/* initialize the cache */
		bcopy(&m321cache[CTL(dp->b_dev)], bp->b_un.b_addr,
			sizeof(struct m321cache));
	
		bp->b_proc = u.u_procp;
		bp->b_dev = dp->b_dev;
		bp->b_blkno = 0;
		bp->b_error = 0;
		bp->b_flags |= B_ABS|B_ICACHE;
		m321strategy(bp);
		iowait(bp);
		bp->b_flags &= ~(B_ABS|B_ICACHE|B_DONE);
		if (u.u_error)
			return;

		clrbuf(bp);

		/* set the max_buf_sz */
		ip = (struct b_ipc *) bp->b_un.b_addr;
		ip->i_max_buf_sz = m321maxbuf[CTL(dp->b_dev)];

		bp->b_proc = u.u_procp;
		bp->b_dev = dp->b_dev;
		bp->b_blkno = 0;
		bp->b_error = 0;
		bp->b_flags |= B_IPC_CONF|B_CONF|B_ABS;
		m321strategy(bp);
		iowait(bp);
		bp->b_flags &= ~(B_IPC_CONF|B_CONF|B_ABS|B_DONE);
		brelse(bp);

		if (u.u_error)
			return;
		m321ctlflg[CTL(dp->b_dev)] |= B_ONCE;
	}
}

/*
 *	m321bclose: perform close on a block device
 */

#ifdef REL2
m321bclose(dev, modes)
#else
m321bclose(dev, modes, otyp)
register otyp;
#endif
register dev;
register modes;
{
	DEBUGPR(DBBCLOSE,dev,"bclose");

#ifdef REL2
	m321close(&m321utab[POS(dev)], modes, 1<<SLICE(dev));
#else
	m321close(&m321utab[POS(dev)], modes, 1<<SLICE(dev), otyp);
#endif
}

/*
 *	m321cclose: perform close on a character device
 */

#ifdef REL2
m321cclose(dev, modes)
#else
m321cclose(dev, modes, otyp)
register otyp;
#endif
register dev;
register modes;
{
	DEBUGPR(DBCCLOSE,dev,"cclose");

#ifdef REL2
	m321close(&m321utab[POS(dev)], modes, 1<<SLICE(dev)+8);
#else
	m321close(&m321utab[POS(dev)], modes, 1<<SLICE(dev), otyp);
#endif
}

/*
 *	m321close: perform close for all devices
 */

#ifdef REL2
m321close(dp, modes, mask)
#else
m321close(dp, modes, mask, otyp)
register otyp;
#endif
register struct iobuf *dp;
register modes;
register mask;
{
#ifdef REL2
	dskclose(dp, modes, mask);

	if (!(FLOPPY(dp->b_dev)) && (dp->io_sopen == 0)) {
		m321wopn--;
		m321setlim(dp->b_dev);
	}
#else
	int *otbl;
	ushort i;

	otbl = &m321otbl[((CTL(dp->b_dev) * MAXDEV) + UNIT(dp->b_dev))
		* OTYPCNT];

	dskclose(dp, modes, mask, otyp);

	for( i = 0; i < OTYPCNT && otbl[i] == 0; ++i);
	if( i == OTYPCNT){
		m321wopn--;
		m321setlim(dp->b_dev);
	}
#endif
}

/*
 *	m321setlim: set limits for winchesters when one is opened or closed
 */

m321setlim(dev)
register dev;
{
	register struct iobuf *dp;
	register struct pktcnt *pcp;
	register i;

	for (i = 0; i < 4; i++) {
		dp = &m321utab[POS(MKDEV(CTL(dev),i))];
		pcp = (struct pktcnt *)&dp->io_pktcnt;
		if (dp->io_sopen != 0)		/* if it's open, change limit */
			pcp->limit = m321lim[m321wopn];
		else				/* if closed, set to default */
			pcp->limit = m321lim[0];
	}
}

/*
 *	m321strategy: perform the disk I/O strategy algortithm
 */

m321strategy(bp)
register struct buf *bp;
{
	register struct iobuf *dp;
	register struct m321wdesc *wd;
	register struct m321fdesc *fd;
	register struct slice *slc;

	DEBUGPR(DBSTRAT,bp->b_dev,"strategy");

	dp = &m321utab[POS(bp->b_dev)];

	/*
	 * Do system accounting calculations
	 */
	if ((bp->b_flags&B_ABS) == 0) {
		/* save cylinder number in b_resid for sadp(1) */
		slc = (struct slice *)dp->io_slctbl + SLICE(bp->b_dev);
		bp->b_resid = bp->b_blkno;		/* get block number */
		bp->b_resid += slc->blkoff;	/* add slice start */
		if (FLOPPY(bp->b_dev)) {		/* divide by blk/cyl */
			fd = (struct m321fdesc *)dp->io_desc;
			bp->b_resid *= (BSIZE / fd->f_secsiz);
			bp->b_resid /= (fd->f_sectrk * fd->f_sides);
		}
		else {					/* divide by blk/cyl */
			wd = (struct m321wdesc *)dp->io_desc;
			bp->b_resid *= (BSIZE / wd->w_secsiz);
			bp->b_resid /= (wd->w_sectrk * wd->w_trkcyl);
		}
	}
	/* call the disk discipline strategy (type 0) routine */
	dsk0strategy(bp, dp);
}

/*
 *	m321start: start the I/O
 */

m321start(dp)
register struct iobuf *dp;
{
	register struct buf *bp;
	register struct buf *ap;
	register struct pktcnt *pcp;
	register struct b_envelope *ep;
	struct b_envelope *bppget();
	register s;

	DEBUGPR(DBSTART,bp->b_dev,"start");

	pcp = (struct pktcnt *) &dp->io_pktcnt;	

	/*
	 * Process I/O requests
	 */
	while (1) {
		s = spl6();
		if ((bp = ((struct buf *)dp->b_acts)->av_forw) == NULL) {
			splx(s);
			break;
		}

		/*
		 * Determine how many packets have been sent.  If the
		 * limit has been reached don't send any more now.
		 */
		if (pcp->cnt >= pcp->limit) {
			splx(s);
			break;		/* too many outstanding requests */
		}

		/*
		 * Get a command packet.
		 */
		if ((ep = bppget(&m321frpkts)) == NULL) {
			splx(s);
			break;
		}
		dp->b_acts = (int)bp;	/* indicate that bp was processed */
		pcp->cnt++;		/* one more packet sent */
		blkacty |= (1<<M321);	/* indicate mvme321 activity */
		splx(s);

		/*
		 * Process the command
		 */
		if (bp->b_flags&B_CONF)			/* config request */
			m321conf(dp,bp,ep);
		else if (bp->b_flags&(B_FORMAT|B_FMTU|B_FIXBS))	/* format */
			m321format(dp,bp,ep);
		else if (bp->b_flags&(B_ISTRAT|B_ICACHE))
			m321icmd(dp,bp,ep);
		else {					/* read/write request */
			register struct b_packet *pp;
			register struct slice *slc;

			/*
			 * Set up packet
			 */
			pp = ep->e_pkt;
			pp->p_cmdp = m321chan[CTL(bp->b_dev)].c_chan;
			pp->p_num = (unsigned long)bp;
			pp->p_cmd = (bp->b_flags&B_READ) ? C_READS : C_WRTS;
			pp->p_dtype = FLOPPY(bp->b_dev) ? D_FLOP : D_HARD;
			pp->p_unit = UNIT(bp->b_dev);	/* logical device */
			if (bp->b_bcount > SBUFSIZE)
				pp->p_cp1 = NOCACHE;
			else
				pp->p_cp1 = CACHE;
			pp->p_cp2 = 0;
			pp->p_cp3 = 0;

			slc = (struct slice *)dp->io_slctbl + SLICE(bp->b_dev);
			pp->p_blkno = bp->b_blkno;
			/* don't add slice start block for absolute requests */
			if ((bp->b_flags&B_ABS) == 0)
				pp->p_blkno += slc->blkoff;	/* slice beg. */
			pp->p_sa = (unsigned long)bp->b_un.b_addr;
			pp->p_count = bp->b_bcount/BSIZE;
	
			/* save the packet number in the buf struct */
			bp->b_pkt = (int)pp;

			/*
			 * Reset the time in b_start for safety timer purposes.
			 */
			{	struct iotime *ip = &m321stat[POS(bp->b_dev)];
				ip->io_resp += lbolt - bp->b_start;
			}
			bp->b_start = lbolt;

			/*
			 * Set the time in the iobuf struct for the
			 * system activity profiler.
			 */
			if (dp->io_start == 0)
				dp->io_start = lbolt;

			s = spl6();
			bppsend(&m321chan[CTL(bp->b_dev)],dp->io_csr,ep);
			splx(s);
		}
	}
}

/*
 *	m321conf: perform configuration command
 */

m321conf(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct b_envelope *ep;
{
	register struct b_packet *pp;
	register struct m321wdesc *wd;
	register struct m321fdesc *fd;
	register struct volumeid *vid;
	register struct dkconfig *dkcfg;
	register dft;
	register s;

	DEBUGPR(DBCONF,bp->b_dev,"conf");

	/* set up the command packet */
	pp = ep->e_pkt;
	pp->p_cmdp = m321chan[CTL(bp->b_dev)].c_chan;
	pp->p_num = (unsigned long)bp;
	pp->p_cmd = C_WIPC;
	pp->p_unit = DRV(bp->b_dev);
	pp->p_cp1 = 0;
	pp->p_cp2 = 0;
	pp->p_cp3 = 0;

	/* IPC configure (IPC write descriptor) */
	if (bp->b_flags&B_IPC_CONF) {
		pp->p_dtype = D_IPC;
		pp->p_sa = (unsigned long)bp->b_un.b_addr;
	}
	else {
		/*
		 * Winchester/Floppy configure (write disk descriptor)
		 */

		/* Is this a default or a re-config? */
		dft = (bp->b_flags & B_DFL_CONF) ? 1 : 0;

		/* This is a re-config */
		if (!dft) {
			vid = (struct volumeid *) (int)(bp->b_un.b_addr);
			dkcfg = (struct dkconfig *)((int)bp->b_un.b_addr + 256);
			/* Get number of slices and put it in iobuf struct */
			dp->io_slccnt = dkcfg->ios_slccnt;
		}

		/* set up floppy descriptor */
		if (FLOPPY(bp->b_dev)) {
			int ds, dd, md, dens, fmt;
	
			fd = (struct m321fdesc *)dp->io_desc;

			/* If a re-config, get values from attribute word */
			if (!dft) {
				ds = (dkcfg->ios_atw&ATWDS) ? 2 : 1;
				dd = (dkcfg->ios_atw&ATWDD) ? 96 : 48;
				md = (dkcfg->ios_atw&ATWMF) ? 96 : 48;
				dens = (dkcfg->ios_atw&ATWTD) ? 2 : 1;
				fmt = (dkcfg->ios_atw&ATWSN) ? 1 : 2;
			}
	
			fd->f_device = UNIT(bp->b_dev);   /* logical device */
			fd->f_drive = DRV(bp->b_dev);
			fd->f_retry = F_RETRY;
			fd->f_step = (dft) ? F_STEP : dkcfg->ios_sr;
			fd->f_dr_tpi = (dft) ? F_DR_TPI : dd;
			fd->f_med_tpi = (dft) ? F_MED_TPI : md;
			fd->f_sides = (dft) ? F_SIDES : ds;
			fd->f_trkside = (dft) ? F_TRKSIDE : dkcfg->ios_trk;
			fd->f_dens0 = F_DENS0; 
			fd->f_sctr0 = (dft) ? F_SCTR0 : dkcfg->ios_spt;
			fd->f_byte0 = F_BYTE0;
			fd->f_dens = (dft) ? F_DENS : dens;
			fd->f_sectrk = (dft) ? F_SECTRK : dkcfg->ios_spt;
			fd->f_secsiz = (dft) ? F_SECSIZ : dkcfg->ios_psm;
			fd->f_initchr = F_INITCHR;
			fd->f_format = (dft) ? F_FORMAT : fmt;
			fd->f_precomp = (dft) ? F_PRECOMP : dkcfg->ios_pcom;
			fd->f_cache = F_CACHE;
			fd->f_lblksz = BSIZE;
	
			pp->p_dtype = D_FLOP;
			pp->p_sa = (unsigned long)fd;
		}
		/* set up winchester descriptor */
		else {
			wd = (struct m321wdesc *)dp->io_desc;
	
			wd->w_valid = VALID;
			wd->w_drive = DRV(bp->b_dev);
			wd->w_device = UNIT(bp->b_dev);   /* logical device */
			wd->w_retry = W_RETRY;
			wd->w_rec_rtry = W_RRTRY;
			wd->w_eccflag = W_ECCFLG;
			wd->w_crcecc = W_CRCECC;
			wd->w_step = (dft) ? W_STEP : dkcfg->ios_sr;
			wd->w_bsf = (dft) ? W_BSF : (dkcfg->ios_catr)&CATBSF;
			wd->w_initchr = W_INITCHR;
			wd->w_brf = W_BRF;
			wd->w_slips = (dft) ? W_SLIPS : dkcfg->ios_ssc;
			wd->w_altcnt = (dft) ? W_ALTCNT : dkcfg->ios_alts;
			wd->w_sectrk = (dft) ? W_LSECTRK : dkcfg->ios_spt;
			wd->w_trkcyl = (dft) ? W_TRKCYL : dkcfg->ios_hds;
			wd->w_cyls = (dft) ? W_CYLS : dkcfg->ios_trk;
			wd->w_redcyl = (dft) ? W_REDCYL : dkcfg->ios_pcom;
			wd->w_strcyl = W_STRCYL;
			wd->w_endcyl = (dft) ? W_ENDCYL : dkcfg->ios_trk - 1;
			wd->w_lblksz = BSIZE;
			wd->w_secsiz = (dft) ? W_SECSIZ : dkcfg->ios_psm;
	
			pp->p_dtype = D_HARD;
			pp->p_sa = (unsigned long)wd;
		}
	}

	/* save the packet address in the buf struct */
	bp->b_pkt = (int)pp;

	s = spl6();
	bppsend(&m321chan[CTL(bp->b_dev)],dp->io_csr,ep);
	splx(s);
}

/*
 *	m321format: perform format command
 */

m321format(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct b_envelope *ep;
{
	register struct b_packet *pp;
	register struct m321wdesc *wd;
	register struct m321fdesc *fd;
	register struct m321fmt *dkf;
	register s;
	register struct dkbadlst *bsp;

	DEBUGPR(DBFORMAT,bp->b_dev,"format");

	if (bp->b_flags & B_FIXBS) {		/* Fix Bad Spot Command */

		/* get pointer to bad spot info struct */
		bsp = (struct dkbadlst *)bp->b_un.b_addr;

		/* set up packet */
		pp = ep->e_pkt;
		pp->p_cmdp = m321chan[CTL(bp->b_dev)].c_chan;
		pp->p_num = (unsigned long)bp;
		pp->p_cmd = C_FIXBS;
		pp->p_dtype = D_HARD;
		pp->p_unit = UNIT(bp->b_dev);	/* logical device */
		pp->p_cyl = (ushort)bsp->dk_cyl;
		pp->p_hd =  (unsigned char)bsp->dk_head;
		pp->p_sec = (unsigned char)bsp->dk_sector;
	}
	else {					/* Format Command */

		/* get pointer to disk format struct */
		dkf = (struct m321fmt *)bp->b_un.b_addr;

		/* set up packet */
		pp = ep->e_pkt;
		pp->p_cmdp = m321chan[CTL(bp->b_dev)].c_chan;
		pp->p_num = (unsigned long)bp;
		pp->p_cmd = FLOPPY(bp->b_dev) ? C_FMTD : C_FMTV;
		pp->p_dtype = FLOPPY(bp->b_dev) ? D_FLOP : D_HARD;
		pp->p_unit = UNIT(bp->b_dev);	/* logical device */
		if (FLOPPY(bp->b_dev)) {
			fd = (struct m321fdesc *)dp->io_desc;
		}
		else {
			unsigned long *badtrk;

			/* fill in format parameters from dkfmt struct */
			wd = (struct m321wdesc *)dp->io_desc;
			pp->p_hd = dkf->f_trkno % wd->w_trkcyl;
			pp->p_cyl = dkf->f_trkno / wd->w_trkcyl;
			pp->p_count = dkf->f_trkcnt;
			pp->p_sa = (unsigned long)dkf->f_badlst;
			pp->p_badcnt = dkf->f_badcnt;
			badtrk = (unsigned long *)&pp->p_badtrk;
			*badtrk = (unsigned long)dkf->f_badtrks;/* bad trks */
		}
	}

	/* save the packet address in the buf struct */
	bp->b_pkt = (int)pp;

	s = spl6();
	bppsend(&m321chan[CTL(bp->b_dev)],dp->io_csr,ep);
	splx(s);
}

/*
 * m321icmd: execute the mvme321 "init" commands
 */

m321icmd(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct b_envelope *ep;
{
	register struct b_packet *pp;
	register s;

	DEBUGPR(DBICMD,bp->b_dev,"init cmd");

	/* set up packet */
	pp = ep->e_pkt;
	pp->p_cmdp = m321chan[CTL(bp->b_dev)].c_chan;
	pp->p_num = (unsigned long)bp;
	pp->p_dtype = D_HARD;
	pp->p_unit = UNIT(bp->b_dev);	/* logical device */

	if (bp->b_flags&B_ISTRAT) {		/* initialize strategy */
		register struct m321strat *sp;

		sp = (struct m321strat *)bp->b_un.b_addr;
		pp->p_cp1 = sp->s_sfunc;
		pp->p_cp2 = sp->s_blk;

		pp->p_cmd = C_SINIT;
	}
	else {					/* initialize cache buffers */
		pp->p_sa = (unsigned long)bp->b_un.b_addr;
		pp->p_cmd = C_CINIT;
	}

	/* save the packet address in the buf struct */
	bp->b_pkt = (int)pp;

	s = spl6();
	bppsend(&m321chan[CTL(bp->b_dev)],dp->io_csr,ep);
	splx(s);
}

/*
 * m321intr: handle the mvme321 interrupts
 */

m321intr(dev)
{
	register struct iobuf *dp;
	register struct buf *bp;
	register rdev;		/* dev is only 1st device on interrupting ctl */
	register struct pktcnt *pcp;
	register struct b_envelope *ep;
	register struct b_packet *pp;
	struct b_envelope	*bpprcv();
	short i;
	int s;

	DEBUGPR(DBINTR,dev,"interrupt");


	if (m321_addr[CTL(dev)] == NULL) {
		m321print(dev, "Undefined interrupt");
		return;
	}

	/* find packet that interrupted */

	while (1) {			/* the "break" will get out of loop */
		s = spl6();
		if ((ep = bpprcv(&m321chan[CTL(dev)])) == NULL) {
			splx(s);
			break;
		}
		pp = ep->e_pkt;

		/*
		 * Get the real device number from the packet.
		 */
		if (pp->p_cmd == C_WIPC)	/* if IPC write descriptor */
			if (pp->p_dtype == D_IPC)
				rdev = MKDEV(CTL(dev),pp->p_unit);
			else			/* if disk write descriptor */
				rdev = MKDEV(CTL(dev),
					  ((pp->p_dtype<<2)|(pp->p_unit)));
		else				/* if read, write, format */
			rdev = MKDEV(CTL(dev),pp->p_unit);

		dp = &m321utab[POS(rdev)];
		dp->b_flags |= B_INTPROC;	/* flag interrupt processed */
		pcp = (struct pktcnt *)&dp->io_pktcnt;
		if (pcp->cnt)			/* decrement the packet count */
			pcp->cnt--;
		splx(s);

		/* if last packet back from the mvme321 */
		if (pcp->cnt == 0) {
			struct iotime *ip = &m321stat[POS(rdev)];

			ip->io_act += lbolt - dp->io_start;
			dp->io_start = 0;
		}

		/*
		 * Find the associated buffer structure.
		 */ 
		bp = (struct buf *)pp->p_num;
		if (bp->b_pkt == (int)pp) {
			/* mark buf finished (status packet returned) */
			bp->b_flags |= B_STSRTN;
			if (bp->b_flags&B_READ || bp->b_flags&B_WRITE) {
				struct iotime *ip = &m321stat[POS(rdev)];

				ip->io_resp += lbolt - bp->b_start;
			}
			if (pp->p_cstat&S_ERROR)
				m321err(dp,bp,pp);
			else
				bp->b_resid = 0;
		}
		else
			m321print(rdev,"No buf associated with status packet");

		s = spl6();
		bppfree(&m321frpkts,ep);
		splx(s);
	}

	/*
	 * Call dsk0done() to complete the ISR processing.
	 */
	for (i = 0; i < MAXDEV; i++) {
		dp = &m321utab[POS(dev) + i];
		if (dp->b_flags&B_INTPROC) {
			dp->b_flags &= ~B_INTPROC;
			dsk0done(dp);				/* finish up */
			blkacty &= ~(1<<M321);	/* indicate activity done */
		}
	}
}

/*
 *	m321err:  process mvme321 return status
 */

m321err(dp,bp,pp)
register struct iobuf *dp;
register struct buf *bp;
register struct b_packet *pp;
{
	register trk;
	register sec;
	register struct m321wdesc *wd;
	register struct m321fdesc *fd;
	register unsigned long cstat;
	register unsigned short estat;
	struct buf *sav_actf;
	int s;

        DEBUGPR(DBERR,bp->b_dev,"error");

	cstat = pp->p_cstat;
	estat = pp->p_estat;

	/*
	 * If any of the meaningless bits are on in the extended status word
	 * and no fatal error, don't do anything.
	 */
	if ((estat&S_RANGE || estat&S_ALT || estat&S_CACHBSY || estat&S_CACHINVL
	    || estat&S_NOCACH || estat&S_DCACHMS || estat&S_ICACHMS)
	    && !(cstat&S_FATAL) && cstat&~(S_ERROR|S_FATAL)) {
		bp->b_resid = 0;
		return;
	}

	/* set up pointer to "registers" for fmtberr() */
#ifdef REL2
	dp->io_eregs = (physadr)&pp->p_cp3;
#else
	dp->io_eregs = (paddr_t)&pp->p_cp3;
#endif
	pp->p_cp3 = pp->p_cmd;
	dp->b_dev = makedev(M321,minor(dp->b_dev));

	/*
	 * calculate block that error was on for fmtberr()
	 * and save it for m321prerr
	 */
	if (FLOPPY(bp->b_dev)) {
		fd = (struct m321fdesc *)dp->io_desc;
		trk = (pp->p_lcyl * fd->f_sides) + pp->p_lhd;
		sec = (trk * fd->f_sectrk) + pp->p_lsec;
		/* check to see if block is on track 0 */
		if (sec < fd->f_sctr0)
			m321eblk = sec / (BSIZE / fd->f_byte0);
		else
			m321eblk = sec / (BSIZE / fd->f_secsiz);
	}
	else {
		wd = (struct m321wdesc *)dp->io_desc;
		if (wd->w_valid != VALID)
			m321eblk = 0;
		else {
			trk = (pp->p_lcyl * wd->w_trkcyl) + pp->p_lhd;
			sec = (trk * wd->w_sectrk) +
				pp->p_lsec;
			m321eblk = sec / (BSIZE / wd->w_secsiz);
		}
	}

	s = spl6();
	sav_actf = dp->b_actf;		/* trick fmtberr() */
	dp->b_actf = bp;
	fmtberr(dp,m321eblk);		/* error block */
	dp->b_actf = sav_actf;		/* fix queue back */
	splx(s);
	m321prerr(bp->b_dev,pp);	/* display error */
	if(pp->p_cstat&S_FATAL) {	/* fatal error */
		bp->b_resid = bp->b_bcount;
		bp->b_flags |= B_ERROR;
	}
	else
		bp->b_resid = 0;
}

/*
 *	m321prerr: print out error information
 */

STATIC m321prerr(dev,pp)
register dev;
register struct b_packet *pp;
{
	register unsigned long cs;
	register unsigned short es;

	cs = pp->p_cstat;
	es = pp->p_estat;

	if (cs&S_FATAL)
		m321print(dev,"\nFATAL ERROR");
	else
		m321print(dev,"\nRECOVERABLE ERROR");

	PRCMD(pp);				/* print command data */

#ifdef	DEBUG
	printf("Completion & Extended Status: %x %x\n", cs, es);
#endif

#ifdef	DEBUG
	if (cs&S_STOI)		m321print(dev,"System bus error");
	if (cs&S_ITOC)		m321print(dev,"Local bus error");
	if (cs&S_CTOD)		m321print(dev,"IPC/device transfer error");
	if (cs&S_CPUABORT)	m321print(dev,"CPU abort");
	if (cs&S_ETERM)		m321print(dev,"Early termination");
	if (cs&S_ABORT)		m321print(dev,"IPC abort");
	if (cs&S_BADCMD)	m321print(dev,"Invalid command code");
	if (cs&S_BADDEV)	m321print(dev,"Bad device");
	if (cs&S_BADUNIT)	m321print(dev,"Bad unit");
	if (cs&S_TIMEOUT)	m321print(dev,"MVME321 controller timeout");
	
	if(pp->p_dtype == D_HARD) {
		if (cs&S_ABNORM)	m321print(dev,"Abnornal command completion");
		if (cs&S_ECCERR)	m321print(dev,"ECC correctable error");
		if (cs&S_RESET)		m321print(dev,"Reset by uPD7261");
		if (cs&S_CRCID)		m321print(dev,"CRC error in ID field");
		if (cs&S_BADVER)	m321print(dev,"Verify does not check");
		if (cs&S_BUSY)		m321print(dev,"Busy condition");
		if (cs&S_SEEKERR)	m321print(dev,"Error in seek operation");
		if (cs&S_ALTERR)	m321print(dev,"Format: not enough alternates");
		if (cs&S_ILLDESC)	m321print(dev,"Illegal disk descriptor");
		if (cs&S_FMTERR)	m321print(dev,"Error formatting track");
		if (cs&S_LOCILL)	m321print(dev,"Illegal location requested");
		if (cs&S_DRVILL)	m321print(dev,"Illegal Drive");
		if (cs&S_CYLEND)	m321print(dev,"End of cylinder indicated");
		if (cs&S_FIFOERR)	m321print(dev,"Fifo over/under run");
		if (cs&S_DATAERR)	m321print(dev,"Data field error");
		if (cs&S_HWERR)		m321print(dev,"Hardware error");
		if (cs&S_DRVERR)	m321print(dev,"Drive not ready");
		if (cs&S_SECERR)	m321print(dev,"Sector not on track");
		if (cs&S_WRTPROT)	m321print(dev,"Write protected disk");
		if (cs&S_NOMARK)	m321print(dev,"No address mark");

		if (es&S_RANGE)		m321print(dev,"Block out of range");
		if (es&S_ALT)		m321print(dev,"mvme321 accessed alternate");
		if (es&S_MEMERR)	m321print(dev,"mvme321 memroy error");
		if (es&S_CACHBSY)	m321print(dev,"Cache busy");
		if (es&S_CACHINVL)	m321print(dev,"Cache invalidated");
		if (es&S_NOCACH)	m321print(dev,"No data cached");
		if (es&S_DCACHMS)	m321print(dev,"Data cache miss");
		if (es&S_ICACHMS)	m321print(dev,"Instruction cache miss");
	}
	else {
		if (cs&S_BADSIZ)	m321print(dev,"Bad disk size in descriptor");
		if (cs&S_EOTSM)		m321print(dev,"End-of-transfer size mismatch");
		if (cs&S_BADTPI)	m321print(dev,"Bad tpi combination specified");
		if (cs&S_DRVMTR)	m321print(dev,"Drive motor not coming on");
		if (cs&S_DOOR)		m321print(dev,"Disk door open");
		if (cs&S_NCMLPT)	m321print(dev,"Command not completing");
		if (cs&S_BADRSTR)	m321print(dev,"Bad restore operation");
		if (cs&S_ILLSIDE)	m321print(dev,"Illegal side reference");
		if (cs&S_ILLTRK)	m321print(dev,"Illegal track reference");
		if (cs&S_ILLSEC)	m321print(dev,"Illegal sector reference");
		if (cs&S_ILLSTP)	m321print(dev,"Illegal step rate specified");
		if (cs&S_BADDENS)	m321print(dev,"Bad density specified");
		if (cs&S_FWRTPROT)	m321print(dev,"Write protected disk");
		if (cs&S_FFMTERR)	m321print(dev,"Format error");
		if (cs&S_BADLOC)	m321print(dev,"Sector not found");
		if (cs&S_ICRCERR)	m321print(dev,"CRC error in ID field(s)");
		if (cs&S_DCRCERR)	m321print(dev,"CRC error in data field");
		if (cs&S_DMAUR)		m321print(dev,"MC68440 DMA underrun");
	}
#endif
	if(pp->p_dtype == D_HARD) {
		if (S_ITOC || cs&S_CTOD || cs&S_CPUABORT || cs&S_ETERM
			|| cs&S_ABORT || cs&S_BADCMD || cs&S_STOI
			|| cs&S_BADDEV || cs&S_BADUNIT || cs&S_ABNORM
			|| cs&S_RESET || cs&S_BUSY || cs&S_SEEKERR
			|| cs&S_FIFOERR || cs&S_ILLDESC || cs&S_LOCILL
			|| cs&S_DRVILL || cs&S_CYLEND || cs&S_ECCERR
			|| cs&S_NOMARK)	m321print(dev,"MVME321 internal error");

		if (cs&S_TIMEOUT)	m321print(dev,"Controller timeout");
		if (cs&S_CRCID)		m321print(dev,"CRC error in ID field");
		if (cs&S_BADVER)	m321print(dev,"Verify does not check");
		if (cs&S_ALTERR)	m321print(dev,"Format: not enough alternates");
		if (cs&S_FMTERR)	m321print(dev,"Error formatting track");
		if (cs&S_DATAERR)	m321print(dev,"Data field error");
		if (cs&S_HWERR)		m321print(dev,"Hardware error");
		if (cs&S_DRVERR)	m321print(dev,"Drive not ready");
		if (cs&S_SECERR)	m321print(dev,"Sector not on track");
		if (cs&S_WRTPROT)	m321print(dev,"Write protected disk");
		if (es&S_MEMERR)	m321print(dev,"MVME321 internal error");
	}
	else {
		if (cs&S_ITOC || cs&S_CTOD || cs&S_CPUABORT
			|| cs&S_ETERM || cs&S_ABORT || cs&S_BADCMD
			|| cs&S_STOI || cs&S_BADDEV || cs&S_BADUNIT
			|| cs&S_EOTSM || cs&S_NCMLPT || cs&S_BADRSTR
			|| cs&S_ILLSIDE || cs&S_ILLTRK || cs&S_ILLSEC
			|| cs&S_BADSIZ || cs&S_ILLSTP
			|| cs&S_DMAUR)	m321print(dev,"MVME321 internal error");

		if (cs&S_TIMEOUT)	m321print(dev,"Controller timeout");
		if (cs&S_BADTPI)	m321print(dev,"Bad tpi combination specified");
		if (cs&S_DRVMTR)	m321print(dev,"Drive motor not coming on");
		if (cs&S_DOOR)		m321print(dev,"Disk door open");
		if (cs&S_BADDENS)	m321print(dev,"Bad density specified");
		if (cs&S_FWRTPROT)	m321print(dev,"Write protected disk");
		if (cs&S_FFMTERR)	m321print(dev,"Format error");
		if (cs&S_BADLOC)	m321print(dev,"Sector not found");
		if (cs&S_ICRCERR)	m321print(dev,"CRC error in ID field(s)");
		if (cs&S_DCRCERR)	m321print(dev,"CRC error in data field");
	}

	printf("\nLogical Head: %x(%d)", pp->p_lhd,pp->p_lhd);
#ifdef	DEBUG
	printf("\t\t\tPhysical Head: %x(%d)", pp->p_phd,pp->p_phd);
#endif
	printf("\nLogical Cylinder: %x(%d)", pp->p_lcyl,pp->p_lcyl);
#ifdef	DEBUG
	printf("\t\tPhysical Cylinder: %x(%d)", pp->p_pcyl,pp->p_pcyl);
#endif
	printf("\nLogical Sector: %x(%d)",pp->p_lsec,pp->p_lsec);
#ifdef	DEBUG
	printf("\t\tSystem Address: %x", pp->p_sstat);
#endif
	if (pp->p_cmd == C_READS || pp->p_cmd == C_WRTS)	/* read/write */
		printf("\nPhysical Block: %x(%d)",m321eblk,m321eblk);
#ifdef	DEBUG
	printf("\n\t\t\t\tTermination Transfer Count: %x(%d)", pp->p_termcnt,
		pp->p_termcnt);
	printf("\n\t\t\t\tSystem Bus Status: %x", pp->p_sberror);
	printf("\n\t\t\t\tLocal Bus Status: %x", pp->p_lberror);
	printf("\n\t\t\t\tRetry Count: %x(%d)\n", pp->p_retry, pp->p_retry);
#else
	printf("\n");
#endif
}

/*
 *	m321prcmd: print command debug information
 */

STATIC m321prcmd(pp)
register struct b_packet *pp;
{
	printf("Command: ");

	switch (pp->p_cmd) {
		case C_WIPC:	printf("Write Descriptor, ");		break;
		case C_FMTV:	printf("Format Verify, ");		break;
		case C_FMTT:	printf("Format Track, ");		break;
		case C_FMTD:	printf("Format Floppy, ");		break;
		case C_READS:	printf("Read Blocks, ");		break;
		case C_WRTS:	printf("Write Blocks, ");		break;

		default:	printf("Unknown: %x, ", pp->p_cmd);	break;
	}

	printf("Device type: %x, Unit number: %x\n", pp->p_dtype, pp->p_unit);
}

/*
 *	m321print: print mvme321 driver error messages
 */

m321print(dev,str)
register dev;
register char *str;
{
	printf("%s on MVME321 drive %d, ctl %d, slice %d\n", str,
		UNIT(dev), CTL(dev), SLICE(dev));
}

#ifndef REL2
/*
 *	m321breakup: physical I/O breakup routine
 */

m321breakup(bp)
register struct buf *bp;
{
	dma_breakup(m321strategy, bp, BSIZE);
}
#endif

/*
 *	m321read: raw I/O read
 */

m321read(dev)
register dev;
{
#ifndef REL2
	register struct iobuf *dp;
	register struct slice *slc;

#endif
	DEBUGPR(DBREAD,dev,"read");

#ifdef REL2
	dskrdwr(&m321utab[POS(dev)], dev, B_READ);
#else
	dp = &m321utab[POS(dev)];
	slc = (struct slice *) dp->io_slctbl + SLICE(dev);
	if (physck(slc->nblocks,B_READ))
		physio(m321breakup, 0, dev, B_READ);
#endif
}

/*
 *	m321write: raw I/O write
 */

m321write(dev)
register dev;
{
#ifndef REL2
	register struct iobuf *dp;
	register struct slice *slc;
#endif

	DEBUGPR(DBWRITE,dev,"write");

#ifdef REL2
	dskrdwr(&m321utab[POS(dev)], dev, B_WRITE);
#else
	dp = &m321utab[POS(dev)];
	slc = (struct slice *) dp->io_slctbl + SLICE(dev);
	if (physck(slc->nblocks,B_WRITE))
		physio(m321breakup, 0, dev, B_WRITE);
#endif
}

/*
 *	m321ioctl: perform ioctl functions
 */

m321ioctl(dev,cmd,args,mode)
register dev;
register cmd;
register char *args;
register mode;
{
	register struct buf *bp;
	register struct buf *bb;
	register struct buf *al;
	register struct iobuf *dp;
	register struct dkconfig *dc;
	register char *addr;
	register size;
	register struct dkbadlst *b1;
	register struct m321badspot *b2;
	register i;
	struct buf *geteblk();
	struct m321strat *sp;


	DEBUGPR(DBIOCTL,dev,"ioctl");

	dp = &m321utab[POS(dev)];

	bp = geteblk();

	if (dskioctl(dp,cmd,args,mode,bp)) {
		brelse(bp);
		return;	/* command was handled by disk discipline ioctl */
	}

	switch (cmd) {

	case DKGETCFG:
	case DKGETINFO:	clrbuf(bp);
			/* copy needed fields from descriptor structure */
			dc = (struct dkconfig *)(bp->b_un.b_addr +
				sizeof(struct volumeid));
			if (FLOPPY(dev)) {
				register struct m321fdesc *fd;

				fd = (struct m321fdesc *)dp->io_desc;

				dc->ios_sr = fd->f_step;
				dc->ios_trk = fd->f_trkside;
				dc->ios_spt = fd->f_sectrk;
				dc->ios_psm = fd->f_secsiz;
				dc->ios_pcom = fd->f_precomp;
				dc->ios_atw |= (fd->f_sides == 2)? ATWDS : 0;
				dc->ios_atw |= (fd->f_dr_tpi == 96)? ATWDD : 0;
				dc->ios_atw |= (fd->f_med_tpi == 96)? ATWMF : 0;
				dc->ios_atw |= (fd->f_dens == 2)? ATWEN : 0;
				dc->ios_atw |= (fd->f_format == 1)? ATWSN : 0;
			}
			else {
				register struct m321wdesc *wd;

				wd = (struct m321wdesc *)dp->io_desc;

				dc->ios_sr = wd->w_step;
				dc->ios_ssc = wd->w_slips;
				dc->ios_spt = wd->w_sectrk;
				dc->ios_hds = wd->w_trkcyl;
				dc->ios_trk = wd->w_cyls;
				dc->ios_pcom = wd->w_redcyl;
				dc->ios_psm = wd->w_secsiz;
			}

			if (cmd == DKGETCFG) {
				addr = bp->b_un.b_addr +
					sizeof(struct volumeid);
				size = sizeof(struct dkconfig);
			}
			else {
				addr = bp->b_un.b_addr;
				size = sizeof(struct dkblk0);
			}

			copyout(addr,args,size);
			brelse(bp);
			return;

	case DKSETCFG:
	case DKSETINFO: if (cmd == DKSETCFG) {
				addr = bp->b_un.b_addr +
					sizeof(struct volumeid);
				size = sizeof(struct dkconfig);
			}
			else {
				addr = bp->b_un.b_addr;
				size = sizeof(struct dkblk0);
			}
			if (copyin(args,addr,size) == -1) {
				brelse(bp);
				return;
			}
			/*
			 * If the volumeid is available, check to see that
			 * the bad spots to be given to the driver for a
			 * format command are going to be in the form of
			 * "byte offset".
			 */
			if (cmd == DKSETINFO) {
				register struct volumeid *vid;

				vid = (struct volumeid *)bp->b_un.b_addr;
				if (vid->vid_strat != DKHARDSPOT &&
					vid->vid_strat != DKPERFECT) {
					brelse(bp);
					u.u_error = ENXIO;
					return;
				}
			}
			bp->b_flags |= B_CONF|B_ABS;
			break;

	case DKFORMAT:	if (copyin(args,bp->b_un.b_addr,sizeof(struct dkfmt))
			    == -1) {
				brelse(bp);
				return;
			}
			if (FLOPPY(dev) == 0) {		/* if winchester */
				register struct m321fmt *dkf;
				register struct dkbadlst *bs;
				register cnt;

				dkf = (struct m321fmt *)bp->b_un.b_addr;
				al = geteblk();		/* for returned alts */
				/*
				 * Borrow the alt tracks list for a moment
				 * to copy the bad spot list into before
				 * building the real mvme321 bad spot list.
				 * Can only handle as many bad spots as
				 * m321badspot structs will fit into 1K bytes.
				 */
				if (dkf->f_badcnt * sizeof(struct m321badspot) >
				    SBUFSIZE) {
					brelse(bp);
					brelse(al);
					u.u_error = ENXIO;
					return;
				}

				bb = geteblk();		/* for bad sec list */

				b1 = (struct dkbadlst *)al->b_un.b_addr;
				b2 = (struct m321badspot *)bb->b_un.b_addr;
				cnt = dkf->f_badcnt;
				bs = dkf->f_badlst;

				/*
				 * build the real mvme321 bad spot list
				 * from the dinit(1M) bad spot list.
				 */
				while (cnt-- > 0) {
					if (copyin(bs++,b1,
				    	    sizeof(struct dkbadlst)) == -1) {
						brelse(bp);
						brelse(al);
						brelse(bb);
						return;
					}

					b2->bs_cyl = (ushort)b1->dk_cyl;
					b2->bs_hd = (unsigned char)b1->dk_head;
					b2->bs_sec = 0;
					b2->bs_offset = (ushort)b1->dk_offset;
					b2++;
				}

				clrbuf(al);	/* freshen up alt list */

				/* point dkfmt to mvme321 bad spot list */
				dkf->f_badlst = (struct dkbadlst *)
						bb->b_un.b_addr;
				/* point dfmt to area for returned alt list */
				dkf->f_badtrks = (unsigned long *)
						 al->b_un.b_addr;
			}
			bp->b_flags |= B_FORMAT|B_ABS|B_LONGCMD;
			break;

	case DKFIXBADSPOT:
			if (copyin(args, bp->b_un.b_addr,
				sizeof(struct dkbadlst)) == -1) {
				brelse(bp);
				return;
			}
			bp->b_flags |= B_ABS|B_FIXBS;
			break;

	case INITSTRAT:	if (copyin(args, bp->b_un.b_addr,
				sizeof(struct m321strat)) == -1) {
				brelse(bp);
				return;
			}
			sp = (struct m321strat *)bp->b_un.b_addr;
			if (sp->s_sfunc < FIFO || sp->s_sfunc > ELEV) {
				u.u_error = EINVAL;
				brelse(bp);
				return;
			}
			bp->b_flags |= B_ABS|B_ISTRAT;
			break;

	case INITCACHE:	if (copyin(args, bp->b_un.b_addr,
				sizeof(struct m321cache)) == -1) {
				brelse(bp);
				return;
			}
			bp->b_flags |= B_ABS|B_ICACHE;
			break;

	case INITIPC:
			if (copyin(args, bp->b_un.b_addr,
				sizeof(struct b_ipc)) == -1) {
				brelse(bp);
				return;
			}
			bp->b_flags |= B_IPC_CONF|B_CONF|B_ABS;
			break;

	default:	u.u_error = EINVAL;
			brelse(bp);
			return;
	}

	bp->b_proc = u.u_procp;
	bp->b_dev = dev;
	bp->b_bcount = 0;
	bp->b_blkno = 0;
	m321strategy(bp);
	iowait(bp);
	if ((bp->b_flags&B_FORMAT) && (!FLOPPY(bp->b_dev))) {
		brelse(bb);
		brelse(al);
	}
	bp->b_flags &= ~(B_CONF|B_FORMAT|B_ABS|B_LONGCMD|
			 B_ISTRAT|B_ICACHE|B_IPC_CONF|B_FIXBS);
	brelse(bp);
}

/*
 *	m321timer: check on status of MVME321 driver periodically
 */

m321timer()
{

	DEBUGPR(DBTIMER,0,"timer");

	/* call the disk discipline timer routine */
	dsk0timer(&m321utab[0],MAXPOS);
}

/*
 *	m321clr: clear the driver after powerfail
 */

m321clr()
{
	DEBUGPR(DBCLR,0,"clr");

	dskclr(&m321utab[0],MAXPOS);
}

/*
 *	mvme321 dump support
 */

STATIC	int	m321offset;

/*
 *	m321dump: dump memory to mvme321 swap area
 */

m321dump()
{
	int m321doio();
	extern	dev_t	dumpdev;
	register dev = dumpdev;

	spl7();				/* disable interrupts */

	DEBUGPR(DBDUMP,dev,"dump");

	if (CTL(dev) >= m321_ctls || m321_addr[CTL(dev)] == NULL) {
		printf("Dump device controller not present\n");
		return;
	}

	{	register struct slice *sp;

		sp = (struct slice *)m321utab[POS(dev)].io_slctbl + SLICE(dev);
		m321offset = sp->blkoff;
	}

#ifdef REL2
	memdump(m321doio);
#endif
}

/*
 *	m321doio: dump out one block of memory
 */

STATIC m321doio(dev,blkno,maddr,rdwr)
register dev;
register blkno;
char *maddr;
int rdwr;
{
	register struct b_device *rp;
	register struct b_envelope *ep;
	register struct b_packet *pp;
	register long dmp_eye = 0x44554d50;
	struct b_envelope *bppget();

	DEBUGPR(DBDOIO,dev,"doio");

	if ((ep = bppget(&m321frpkts)) == NULL) {
		for (;;);
	}

	pp = ep->e_pkt;
	pp->p_eye = dmp_eye;
	pp->p_cmd = (rdwr == B_READ) ? C_READS : C_WRTS;
	pp->p_dtype = D_HARD;
	pp->p_unit = DRV(dev);
	pp->p_sunit = 0;
	pp->p_blkno = blkno + m321offset;
	pp->p_sa = (unsigned long)maddr;
	pp->p_count = SBUFSIZE / BSIZE;
	pp->p_cp1 = 0;
	pp->p_cp2 = 0;
	pp->p_cp3 = 0;
	rp = m321_addr[CTL(dev)];

	bppsend(&m321chan[CTL(dev)],rp,ep);

	do {
		ep = bpprcv(&m321chan[CTL(dev)]);
		pp = ep->e_pkt;
	} while (ep == NULL || pp->p_eye != dmp_eye);	/* wait until done */

	/*
	 *	will hang forever on the above "do-while" statement
	 *	when the controller fails to respond.
	 */

	if (pp->p_cstat != 0)				/* an error occurred */
		m321prerr(dev,pp);

	bppfree(&m321frpkts,ep);
}
