/*
 *	Copyright (c) 1984 AT&T
 *	Copyright (c) 1987 Motorola Inc
 *	All Rights Reserved
 */

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)kern-port:io/mvme323.c	35.4"
/*
 *	MVME323 Driver: This driver is for Motorola's MVME323 ESDI
 *		Drive Disk controller. The MVME323 is a VMEbus compatible
 *		board with multiple drive capabilities. The MVME323 has an
 *		MC68000 on board to perform intelligent disk control.
 *		The MVME323 supports 4 ESDI-compatible disk drives and
 *		has no floppy disk support.
 */

/*
 *	DEBUG support
 */


#define COALES 
short	m323coal = 1;		/* enable/disable block I/O coalescing */

short 	m323mxenb = 1;		/* allow attempt to go to MACSI mode at
				   init time. */

#define DEBUG 9
#undef DEBUG

#define	DEBUGC			/* command debugging */

#ifdef	DEBUG

#define	STATIC

short	cpr323thresh = 15;	/* threshold which suppresses coalesce message*/

int	DebugM323 = DEBUG;	/*	0 -> no debug print outs */
				/*	1 -> init/open/close/read/write/dump */
				/*	2 -> 1 plus strategy/config */
				/*	3 -> 2 plus start/clr/cmd */
				/*	4 -> 3 plus intr */
				/*	5 -> 4 plus timeout */
				/*	6 -> 5 and verbose */

#else

#define	STATIC	static

#endif


/*
 *	This driver will support up to 2 controllers
 *	Each MVME323 can control up to 4 drives simultaneously:
 *
 *	Logical Unit Number   Drive	Device
 *		(LUN)
 *
 *		 0		0	Drive 0 - fixed
 *		 1		1	Drive 1 - fixed
 *		 2		2	Drive 2 - fixed
 *		 3		3	Drive 3 - fixed
 */

#define SLCTSIZE M323MAXSLC*sizeof(struct size)
 
STATIC	unsigned m323_ctls;

/* maximum number of slots */
#define	MAXPOS	(m323_ctls*MAXDEV)

/* maximum number of logical slots */
#define	MAXLPOS	(m323_ctls*MAXDEV)

/* address modifier */
STATIC unsigned char m323adrm;

/* global memory type */
STATIC unsigned char m323mtyp;

#ifdef DEBUG
	int m323io16b = 0;		/* counter for 16 bit data xfers */
	int m323io32b = 0;		/* counter for 32 bit data xfers */
#endif

/*
 *	MVME323 Disk Controller Driver Minor Number Assignment Coding
 *
 * field width	 <- 1 bit  -> <- 1 bit -> <- 2 bits -> <- 4 bits->
 *		 ------------------------------------------------
 * field	| controller |  not used |  drive     |   slice  |
 *		 ------------------------------------------------
 *
 *							Standard Names
 *   Minor Number    Logical				 Controller
 *   (in octal)       Unit 	Drive			0 Name	   1 Name
 *	000X		0	Drive 0 Fixed		m323_0sX  m323_c1d0sX
 *	001X		1	Drive 1 Fixed		m323_1sX  m323_c1d1sX
 *	002X		2	Drive 2 Fixed		m323_2sX  m323_c1d2sX
 *	003X		3	Drive 3 Fixed		m323_3sX  m323_c1d3sX
 */

#define	SLICE(d)	(d&0xf)		/* minor -> slice */
#define	UNIT(d)		((d>>4)&0x3)	/* minor -> unit (phys drive) */
#define	MKDEV(c,u)	((c<<7)|(u<<4))	/* ctl/unit -> minor number */
#define	CTL(d)		((d&0x80)>>7)	/* controller number */
#define	POS(d)		(((d&0x80)>>5)|UNIT(d)) /* position in m323utab */
#define DRV_NUM(d)	(UNIT(d)<<4)
#define	RMVBL(d)	(d&0x00)		/* removable media */
#define	FIXED(d)	(!RMVBL(d))	/* fixed media */
#define	VOL(d)		(0x00)	/* volume number */
#define	MKOPDEV(d)	((d&0x00)<<3)	/* make command option device */
#define	MKOPUN(d)	((d&0x00)<<3)	/* make command option unit number */
#define	SCDRV(s)	((s>>3)&1)	/* status change drive */
#define	COPTOU(x)	((x>>7)&1)	/* command option unit */
#define	COPTOV(x)	((x>>6)&1)	/* command option volume */

#define SGLIST(d)	(&m323sglist[POS(d) * M323MXCOAL])
#define	ILLEGAL(d)	((POS(d)>m323_cnt)||((d)&0x40)) /* illegal minor dev */

/* macros */
/*
 * TRK2BLK lets format operations pass the bounds test in strategy 
 *  and provides a block offset for use with error reporting.
 */
#define TRK2BLK(ibp,trk) (ibp->ib_bps*ibp->ib_spt/BSIZE * trk )

/* Macros to generate block offset of I/O for error logging */
#define SECPBLK		(BSIZE/((struct m323uib *)dp->uib)->ib_bps) /* sectors/block */
#define IO_BOFF		(bp->b_blkno+(int)bp->soff/SECPBLK) /* I/O blk offset */

/* If the configuration sector or its length changes, this driver must change */
#define CFGSS 1		/* Configuration sector number in 1/4K sectors */
#define CFGLEN 1	/* Length of configuration sector in 1/4K chunks */

/*
 *	MVME323 Disk Controller Constants
 */

/*	Drive Status Register	*/
#define DSTAT(d,s)	((d)?(s>>8&0xff):(s&0xff))	/* drive status */
/*	Signals from the drive	*/
#define DRV_RDY 	0x01	/* Drive Ready - up to speed & ready to go */
#define WRT_PROT	0x02	/* Write Protected - at current cylinder */
#define DRV_BUSY	0x04	/* Drive Busy - dual ported drive only */
#define FAULT   	0x08	/* Drive Fault - controller violated spec */
#define ON_CYL   	0x10	/* On Cylinder */
#define SEEK_ERR	0x20	/* Seek Error */
/*	Calculated or generated by the controller */
#define UNIT_PRS	0x40	/* Unit Present - responded to select */
#define UNIT_RDY	0x80	/* Unit Ready - on cylinder & no faults */

/*	Command/Status Register	*/
#define SCHG_SRC	0x08	/* Status Change Source - interrupting drive */
#define ERR_LCMD	0x10	/* Error Last Command */
#define SCHG_INT	0x20	/* Status Change Interrupt */
#define DONE_INT	0x40	/* Command Completion Interrupt */
#define GO_BUSY		0x80	/* Go/Busy - start command execution */
#define M323BERR	0x0100	/* Bus error during controller memory access */
#define M323BDCL	0x1000	/* Board clear */
#define M323NOSF	0x2000	/* No sysfail on board fail */
#define M323BDOK	0x4000	/* Board passes power up diagnostics */
#define M323SLED	0x8000	/* Status LED static mode */

 /* MVME323 memory type codes */
#define MEMT08		0	/* byte wide (8 bit ) data transfers */
#define MEMT16		2	/* word wide (16 bit ) data transfers */
#define MEMT32		3	/* long wide (32 bit ) data transfers */

/*
 *	MVME323 Command Codes
 */

/*   command name	code		comment 	*/
#define	DIAGNOS		0x70	/* Execute power up diagnostics */
#define	READ_LNG	0x71	/* Reads data & ECC for diagnostics */
#define	WRIT_LNG	0x72	/* Writes data & ECC for diagnostics */
#define	READ_HDR	0x74	/* Reads random header on current cyl */
#define READ_FLAW       0x76	/* Reads manufacturer defect list */
#define	GET_CNFG	0x77	/* Returns effective driver parameters (UIB) */

#define	WRIT_SBF	0x78	/* Write diagnostic sector to controller */
#define	READ_SBF	0x79	/* Read sector written with WRIT_SBF */
#define	READ_SEC	0x81	/* Read sector(s) */
#define	WRIT_SEC	0x82	/* Write sector(s) */
#define	VRFY_SEC	0x83	/* Verify sector(s) readable without error */

#define	FMT_TRK		0x84	/* Format track */
#define	MAP_TRAK	0x85	/* Format out bad track on drive */
#define	HNDSHAKE	0x86	/* Handshake controller */
#define	INIT_DRV	0x87	/* Initialize drive parameters */
#define	RESTORE 	0x89	/* Attempts to clear drive fault(s) */
#define	SEEK	 	0x8a	/* Seek to specified position */

#define	REFORMAT	0x8b	/* Format track saving bad spot mapping */
#define	FMT_TRKD	0x8C	/* Format track with data */
#define	MAP_SECT	0x90	/* Format out bad sector on track */
#define	READ_NOC	0x94	/* Read sectors ignoring cache buffer */
#define	VRFY_TRK	0x99	/* Verify track readable without error */
#define	TRACK_ID	0x9a	/* Diagnostic returns track header info */
#define	FETCH_EX	0x9b	/* Fetch IOPB and execute */

#define SCATTER		0xa1	/* Read to non-contiguous memory */
#define GATHER 		0xa2	/* Write from non-contiguous memory */
/*
 *	MVME323 Command Options
 */

#define	ECC_EN		0x01	/* ECC enable */
#define	INT_EN		0x02	/* Interrupt enable */
#define	NOERRCHK	0x04	/* Disables error check on read */
#define	RES_DUAL	0x08	/* Reserves dual ported disk */
#define	LOGICAL		0x10	/* Expect and use logical sector values */
#define	LINK_IOPB	0x20	/* Next IOPB is linked to this one */
#define	VOLUME		0x40	/* Logical volume on specified drive */
#define	DRIVE		0x80	/* Physical drive for command */

#define FMT_PATT	0x1234	/* Format pattern used for:  Format track */
#define FMT_PATM	0xABCD	/* Format pattern used for:  Map alt track */
#define FMT_PATS	0xF00D	/* Format pattern used for:  Sector slip */
#define FMT_PATR	0xF00D	/* Format pattern used for:  Reformat */
/*
 *	MVME323 UIB Options
 */

#define FOUR_DRVS	0x40	/* Select 4 drive operation. */


/*
 *	MVME323 Status Codes
 */

#define	DONE_OK		0x80	/* Command completed with no error */
#define	BUSY_OK		0x81	/* Command in progress */
#define	DONE_ERR	0x82	/* Command completed with error */
#define	DONE_EXC	0x83	/* Command completed with exception */
#define	UNSTARTED	0	/* Command unstarted, no ctlr status code */

/*
 *	MVME323 Error Codes For Commands Completed With Exception
 */

#define	RETRIES(s)	(s&0xf) /* Retries field from error code where */
/* value 0xf indicates 15 or more depending on count specified in UIB */

#define	RES_USED	0x10	/* Restore and reseek used */
#define	BAD_DATA	0x40	/* Bad data moved to system memory */
#define	ECC_USED	0x80	/* ECC used to correct bad data */

/*
 *	MVME323 Error Codes For Commands Completed With Error
 */

#define	DRV_NRDY	0x10	/* Drive not ready */
#define	SEEK_CYL	0x12	/* Seek error - wrong cylinder */
#define	ECC_CODE	0x13	/* ECC code discrpancy */
#define	BAD_CMD		0x14	/* Invalid command code (byte 0 of IOPB) */
#define	BAD_FEX		0x15	/* Illegal fetch and execute attempted */
#define	BIG_SNUM	0x16	/* Sector number too large for drive */
#define	MEM_TYPE	0x17	/* Illegal memory type specified */

#define	BUS_TOUT	0x18	/* Bus not available in < 1 ms */
#define	HDR_CHKS	0x19	/* Header checksum error */
#define	WRT_PERR	0x1a	/* Disk write protected */
#define	NOT_SEL		0x1b	/* No response from unit on select */
#define	SERR_TOUT	0x1c	/* No correct seek in 3 secs after RESTORE */
#define	FLT_TOUT	0x1d	/* Fault not cleared 3 secs after RESTORE */
#define	DRV_FLT		0x1e	/* Drive fault occured, RESTORE required */
#define	RDY_TOUT	0x1f	/* Drive not ready 3 secs after RESTORE */

#define	PAST_END	0x20	/* Multi sector I/O exceeded end of media */
#define	UIB_ZERO	0x21	/* Uib for specified volume contains zeros */
#define	HDR_PAD		0x22	/* Bad post header pad byte found */
#define	FIX_FAIL	0x23	/* Failed to fix uncorrectable error */
#define	BAD_CYLX	0x24	/* Logical sector num -> bad cylinder num */
#define	BAD_HDX		0x25	/* Logical sector num -> bad head num */
#define	BAD_SECX	0x26	/* Logical sector num -> bad phys sector num */
#define	OVERRUN		0x27	/* Data overrun - suspect clock from drive */


#define	NO_INDEX	0x28	/* No index pulse after 65ms on format cmd */
#define	SEC_NFND	0x29	/* Sector not found during read or write */
#define	BAD_HDNO	0x2a	/* Head number in header wrong */
#define	DATA_SYNC	0x2b	/* Sync word in data field not valid */
#define	SEEK_TOUT	0x2d	/* Seek not complete in 500ms */
#define	BUSY_TOUT	0x2e	/* Drive held by dual port over 500ms */
#define	OFF_CYL		0x2f	/* Not on cylinder in 3 secs after select */

#define	RTZ_TOUT	0x30	/* Not normal complete 3 secs after RESTORE */
#define	HDR_SYNC	0x31	/* Sync word in header not valid */

#define	NOT_INIT	0x40	/* Write or format attempted before INIT_DRV */
#define	SMALL_GAP	0x42	/* Specified gap too small - minimun gap is 5 */
#define	SEEK_DRV	0x4b	/* Seek error reported by drive */
#define	MAP_NOP		0x4c	/* No sector pulse on track to be mapped */

#define	BAD_SPT		0x50	/* Sectors per track in UIB invalid */
#define	BIG_SEC		0x51	/* Bytes per sector in UIB > buffer size */
#define	BAD_ILV		0x52	/* Invalid interleave in UIB > SPT or 0 */
#define	HEAD_OR		0x53	/* Head number in IOPB out of range per UIB */
#define	CYL_OR		0x54	/* Cylinder no. in IOPB out of range per UIB */

#define	IOPB_FAIL	0x60	/* Bus error on external IOPB transfer */
#define	DMA_FAIL	0x61	/* Bus error on DMA transfer */
#define	ALLIGN		0x62	/* VME buffer address not alligned */
				/* Word allign 8 & 16 bit transfers */
				/* Long word allign 32 bit transfers */
#define	NOSPARE		0x6e	/* Spare sector already in use */
#define	NOSPARES	0x6f	/* UIB indicates no spares enabled */

#define	NOT_YET		0xff	/* Command not implemented */

/*
 * 	The following error codes are not used or require support
 * 	directly from Motorola.
 *
 *	Codes not used: 11, 2c, 32-3f, 41, 43-4a, 4e, 4f, 55-5c, 70-ef.
 *
 *	Call Motorola with IOPB and UIB info for error codes f0-fe.
 */

/*
 *	Diagnostic Message Constants and Strings
 */

#define	WORDMSB	15
#define	BYTEMSB	7
#define CSRMSK	0xf1e0
#define DRVSMSK	0xff
#define RTRYMSK	0x0f
#define EXCPMSK	0xf0
#define COPMSK	0x3f
#define ATRMSK	0x7f

static char *csrdis[16] = {	/* control and status register */
	0,0,0,0, 	" err_last", 
	0,		" done",
	" go_busy",	" berr",
	0,0,0, 		" bdclr",
	" sfdis",	" bok",
	" sled"
	};

static char *dstatdis[8] = {	/* drive status */
	" drv_rdy",	" wrt_prot",
	" drv_busy", 	" fault",
	" on_cyl", 	" seek_err",
	" unit_pres", 	" unit_rdy"
	};

static char *excode[8] = {	/* exception codes */
	0,0,0,0,
	" restore_&_reseek",
	" reserve_bit_set",
	" data_may_be_erroneous",
	" ECC_used",
	};

static char *copdis[8] = {	/* command options */
	" ECC_on",		" ints_on",
	" no_error_chk",	" hold_port",
	" logical",		" linked_iopb",
	0,0
	};

static char *atrdis[8] = {	/* attributes */
	" reseek_on",		" bad_data_on",
	" inc_by_head",		" dual_port",
	" SC_int_on",		" CE",
	" SSE",			0
	};

/*
 *	Include files
 */

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/fs/s5dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/elog.h"
#include "sys/iobuf.h"
#include "sys/systm.h"
#include "sys/mvmecpu.h"
#include "sys/file.h"
#include "sys/dk.h"
#include "sys/open.h"

#define	MVME323
#include "sys/mvme323.h"

typedef unsigned char u_char;

/*
 *		MVME323 Addresses and vectors
 *
 *	Switch settings 	Board Number	Base Address
 *	8 7 6 5 4 3 2 1
 *	o x o x o o o o		     0		0xffffa000
 *	o x o x o o o x		     1		0xffffa200
 *
 *	o = on  x = off 
 *
 *	Although the MVME323 controller can use different interrupt
 *	vectors for normal and error interrupts, the Unix driver is
 *	no less efficient using only one per controller and will use
 *	the vector assigned to normal completion.
 *
 *	Controller	Vector	Address		Comment
 *	    0		0xf9	0x324	Normal or only vector
 *	    0		0xf8	0x3e0	Error vector (when used)
 *	    1		0xf7	0x3dc	Normal or only vector
 *	    1		0xf6	0x3d8	Error vector (when used)
 */


/*
 *	External Variables
 */

extern int m323MACSI;
extern struct device *m323_addr[];
extern int m323_bmaj;			/* block major */
extern int m323_cnt;			/* count */
extern int m323_ilev[];			/* interrupt level */
extern int m323_ivec[];			/* interrupt vector */
extern struct m323list m323sglist[];	/* scatter/gather list array */

struct size
{	daddr_t	nblocks;
	int	blkoff;
};

/*
 *	Driver Variables
 *
 *	Declared in conf.c (space.h or io.h)
 */

extern struct size m323_sizes[][M323MAXSLC];
 
extern struct m323ctlr	m323ctlr[];	/* per controller control structure */
extern struct iotime	m323stat[];	/* one per drive */
extern struct iobuf 	m323utab[];	/* one per drive */
extern struct m323uib	m323uib[];	/* unit init blocks, one per drive */
extern struct m323config m323config[];	/* per volume configuration info */
extern struct m323config m323dfcf[];	/* default configurations */
extern unsigned char m323types[];	/* disk types */
extern struct m323eregs	m323eregs[];	/* per ctlr error log "registers" */
extern int	m323otbl[];		/* open table */


/*
 *	Constants for m323timer
 */

#define	wtime	io_s9
#define	WOK	0
#define	WABORT	1
#define	WERR	20
#define	WSECS	5

#define	WSTART	0	/* start timer */
#define	WCONT	1	/* continue timer */

/*
 *	defines for buf and iobuf structures
 */

#define	acts	io_s1		/* used for sorting I/O queue */
#define	qcnt	io_s2		/* length of I/O queue */
#define	slccnt	io_s3		/* slice count from disk */
#define	uib	io_s4		/* pointer to the unit init block (uib) */
#define	dstat	io_s5		/* drive status */
#define	cmdbuf	io_s6		/* pointer to I/O parameter block (iopb) */
#define altqf   io_s7		/* alternate queue forward pointer */

#define	cylin	b_umd		/* holds cylinder number */
#define	soff	b_s1		/* holds sector offset of slice for I/O */

/* b_flags constants */
#define	B_CONF	0x040000	/* need to perform configuration on volume 0 */
#define	B_CONFR	0x080000	/* need to perform configuration on volume 1 */
#define	B_FMTBS	0x100000	/* format bad spot */
#define	B_SPECL	0x200000	/* rootdev, swapdev, or pipedev */
#define	B_FMTA	0x400000	/* format alternate track */
#define B_FMTS  0x800000	/* format with sector slip */
#define B_RDFLAW 0x2000000	/* read manufacturers defect list */
#define B_RESTORE 0x4000000	/* restore the drive */
#define B_ABS   0x8000000	/* absolute read or write(ignore slice offset)*/

/* identify all non read/write I/O commands for correct I/O time accounting */  
#define NON_RWIO  (B_CONF|B_CONFR|B_FMTBS|B_FMTA|B_FMTS) 


/*
 *	Other constants and defines
 */

#define TRUE	1		/* Generic Booleans */
#define FALSE	0
#define ON	1
#define OFF	0

#define OKAY	0		/* non-error return value */
#define ERRET	-1		/* error return value */

#define	RETRY	10		/* hardware retry count */
#define NORETRY 1		/* only one try */
#define TRYCNT	5		/* default software try count */


#define	STEPSIZE 32		/* I/O queue step */
#define	READ_ONLY -1

#define DMA_CNT 8		/* DMA holds bus for 8 transfers: see note */
/*
 * DMA Note: an unwritten rule of thumb for the VMEbus requires every bus master
 * to give up the bus once every 8 microseconds.  Another unwritten rule
 * says any bus master should be able to withstand 64us of bus inaccess.
 * Eight (8) transfers at an estimated 1.5us per transfer is 12us.
 * This also allows for several bus masters to take a turn at getting the
 * bus for 8us before any one times out.
 */

/* address modifier code factors */
#define DATA_ACC	1	/* data access */
#define SUPERVSR	4	/* supervisor mode */
#define STD_24		0x38	/* standard 24-bit address */
#define EXT_32		0x08	/* extended 32-bit address */

#define LONGMSK		3	/* mask to identify long allignment */

/*
 *	Assumptions, conventions and operational notes.
 */

	/* NOTE:
	 * the following variables are currently set once at initialization 
	 * time and are expected to remain untouched throughout operation.
	 * This is reasonable for currently supported commands and is subject
	 * to change.
	 *	In the IOPB:
	 *		pb_ilev		{ interrupt level }
	 *		pb_eivec	{ error interrupt vector }
	 *		pb_nivec	{ normal interrupt vector }
	 *		pb_dmac		{ dma burst length }
	 *		pb_skew		{ uib skew override }
	 *		pb_iopbp	{ linked iopb pointer }
	 *		pb_imtyp	{ linked iopb memory type }
	 *		pb_iamod	{ linked iopb address modifier }
	 *	In the UIB
	 *		ib_silev	{ status change interrupt level }
	 *		ib_sivec	{ status change interrupt vector }
	 *		ib_attrib	{ attributes }
	 */		

	 /*
	  *	For cpytype 682, 68020 systems a 32 bit data bus is assumed
	  *	else 16 bit bus assumed.
	  */

struct device {
	unsigned char	ds[2];	/* drive status for two drives */
	ushort		csr;	/* control status register */
	struct m323iopb iopb1;	/* I/O parameter block */
};

/*#define drv0stat	ds[1]*/
/*#define drv1stat	ds[0]*/
#define	DRV0STAT	0x1fd
#define	DRV1STAT	0x1fc
#define	DRV2STAT	0x1fb
#define	DRV3STAT	0x1fa

#define	drv0stat(addr)	*((char *)((char *)(addr) + DRV0STAT))
#define	drv1stat(addr)	*((char *)((char *)(addr) + DRV1STAT))
#define	drv2stat(addr)	*((char *)((char *)(addr) + DRV2STAT))
#define	drv3stat(addr)	*((char *)((char *)(addr) + DRV3STAT))


/*
 *	Driver procedures
 *
 *	m323init: initialize driver
 */

m323init()
{
	if(m323mxenb && (x323init() == OKAY))
		x323install();
	else
		m323oinit();
}

m323oinit()
{	register struct device *rp;
	register struct iobuf *dp;
	register struct m323uib *ibp;
	register struct m323iopb *pbp;
	register unit;
	register i;
	register pos;

#ifdef	DEBUG
	if( DebugM323 )
		m323print( 0, "init" );
#endif

	if( cputype == PROC020 || cputype == PROC030 ) {
		m323adrm = EXT_32|SUPERVSR|DATA_ACC;
		m323mtyp = MEMT32;		/* assumption */ 
	} else {
		m323adrm = STD_24|SUPERVSR|DATA_ACC;
		m323mtyp = MEMT16;		/* assumption */ 
	}

	/*
	 *	the number of controllers is equal to
	 *	m323_cnt divided by:
	 *
	 *	- MAXDEV to divide out the drive field of the minor number
	 *	- MAXDEV is 4.
	 *	- M323MAXSLC to divide out the slice field of the minor number
	 *	- M323MAXSLC is currently 16.
	 */

	m323_ctls = m323_cnt/(MAXDEV*M323MAXSLC);

	for( i = 0; i < m323_ctls; i++ ) {
		rp = m323_addr[i];
		if( bprobe( rp, READ_ONLY )) {
			/* controller not present */
#ifdef	DEBUG
			if( DebugM323 > 0 )
				m323print(MKDEV(i,0),"Controller not present");
#endif
			unit = m323_ctls;
			m323_ctls = i;
			while( i < unit )
				m323_addr[i++] = NULL;
			break;

		} else {
			/* controller present */
			pos=MAXDEV*i;
			dp = &m323utab[pos];

			/* convert vector address to vector number */
			m323_ivec[i] /= 4;

			/* initialize per physical disk */
			for(unit=0; unit<MAXDEV; dp++, pos++, unit++ ) {

				dp->b_flags = B_CONFR|B_CONF; /* force config */

				/* set up for error logging */
				dp->io_stp = (struct iostat *)&(m323stat[pos].ios);
				dp->io_addr = (paddr_t)&m323eregs[i]; 
				dp->io_nreg = M323EREG;

				/* point to device registers and uib */
				pbp = &rp->iopb1;
				dp->cmdbuf = (int)pbp;
				ibp = &m323uib[pos];
				dp->uib = (int)ibp;
				/* set and forget variables in iopb and uib */
				pbp->pb_ilev = ibp->ib_silev = m323_ilev[i];
				pbp->pb_ilev |= DRV_NUM(unit);
				pbp->pb_eivec = pbp->pb_nivec = ibp->ib_sivec 
					= m323_ivec[i];
				pbp->pb_dmac = DMA_CNT;
				pbp->pb_skew = 0;	/* skew in uib used */
				ibp->ib_attrib = M323DATR;

				/* no linked iopb's needed */
#ifdef	m88k
				pbp->pb_iopbphi = NULL;
				pbp->pb_iopbplo = NULL;
#else
				pbp->pb_iopbp = (struct m323iopb *) NULL;
#endif
				pbp->pb_imtyp = 0;	/* N/U */
				pbp->pb_iamod = 0; 	/* N/U */
			 }
		}
	}
	if( major(rootdev) == m323_bmaj ) {
		m323utab[POS(rootdev)].b_flags |= B_SPECL;
	}
	if( major(swapdev) == m323_bmaj ) {
		m323utab[POS(swapdev)].b_flags |= B_SPECL;
	}
	if( major(pipedev) == m323_bmaj ) {
		m323utab[POS(pipedev)].b_flags |= B_SPECL;
	}
}

/*
 *	m323open: open a drive (called from m323bopen() and m323copen())
 */

STATIC m323open(dev, flag, otyp, mask)
register dev, flag, otyp, mask;
{
	register struct iobuf *dp;
	struct volumeid *vp;
	struct dkconfig *cfp;
	register long	slcblk;
	struct size	*sp;
	struct size	*slcp;
	int	*otbl;		/* address of open table */
	int i;

#ifdef	DEBUG
	if( DebugM323 )
		m323print(dev, "open");
#endif

	/* validate */
	dp = &m323utab[POS(dev)];
	otbl=&m323otbl[POS(dev) * OTYPCNT];

	if (CTL (dev) >= m323_ctls || ILLEGAL(dev) ) {
		u.u_error = ENXIO;
		return;
	}

	while( dp->b_flags & B_OPEN) 	/* open in progress? */
		sleep(dp, PRIBIO);

	if(dp->b_flags & B_ONCE) {
		/*
		 * the drive is already initialized
		 * mark the logical device open
		 */

		if(otyp == OTYP_LYR)
			++otbl[OTYP_LYR];
		else if( otyp < OTYPCNT)
			otbl[otyp] |= mask;
		return;
	}

	/* set open in progress */
	dp->b_flags |= B_OPEN;

	/*
	 * do default config, read vid & cfg blk, do re-config,
	 * & read slice table.
	 */
	if(m323cf(dev, dp)) {
		u.u_error = EIO;
		goto done;
	}

done:	if (u.u_error == 0 || flag&FNDELAY) {
		dp->b_flags |= B_ONCE;		/* set init done */
		if(otyp == OTYP_LYR)
			++otbl[OTYP_LYR];
		else if( otyp < OTYPCNT)
			otbl[otyp] |= mask;
	}
	if (flag&FNDELAY)
		u.u_error = 0;
	dp->b_flags &= ~B_OPEN;			/* open done */
	wakeup(dp);
}
	


/*
 *	m323bopen: perform open on a block device.
 */

m323bopen(dev, flag, otyp)
register dev,flag, otyp;

{

	dev = minor(dev);
	m323open(dev, flag, otyp, (1 << SLICE(dev)));
}


/*
 *	m323copen: perform open on a character device.
 */

m323copen(dev, flag, otyp)
register dev;

{

	dev = minor(dev);
	m323open(dev, flag, otyp, (1 << SLICE(dev) ));
}


/*
 *	m323tst: test of ID string in volume ID
 */

STATIC	char	motstring[10] = "MOTOROLA";
STATIC	char	exorstring[10] = "EXORMACS";

STATIC m323tst(vid_mac)
register *vid_mac;
{	register *mot = (int *)motstring;
	register *vd = vid_mac;

	if( *mot++ == *vd++ )
		if( *mot == *vd )
			return(1);
	
	mot = (int *)exorstring;
	if( *mot++ != *vid_mac++ )
		return(0);

	return( *mot == *vid_mac );
}

/*
 *	m323cf: performs a drive configuration.
 *		Returns OKAY if default configure 
 *		was successful else returns ERRET.
 */

STATIC int m323cf(dev,dp)	/* configure device */
register dev;
register struct iobuf *dp;
{	register struct buf *bp;
	register struct m323config *cf;
	struct buf *geteblk();
	register struct volumeid *vp; 
	register long	slcblk;
	register short	slccnt; 
	struct dkconfig *cfp;
	struct size *slcp;
	struct size *sp;
	static int m323kludge = 0;

#ifdef	DEBUG
	if( DebugM323 > 1 )
		m323print( dev, "cf" );
#endif

	/*
	 * KLUDGE: Have to initialize drive zero before initializing
	 * any other drives.
	 */

	if (m323kludge == 0) {
		register struct iobuf *dp0;

		m323kludge = 1;		/* only do this once */

		/*
		 *	configure the disk with the default info
		 */

		cf = &m323dfcf[m323types[POS(MKDEV(CTL(dev),0))]];
		dp0 = &m323utab[POS(MKDEV(CTL(dev),0))];
		if( m323setcf( MKDEV(CTL(dev),0), dp0, cf)) {	/* set dft cfg*/
			return( ERRET );		/* permission denied */
		}

		bp = geteblk();
		bp->b_proc = u.u_procp;
		bp->b_dev = (MKDEV(CTL(dev),0) | 7);

		/* do default configure */
		bp->b_flags |= B_CONF;		/* configure */
		bp->b_error = 0;
		bp->b_blkno = 0;
		bp->b_bcount = 0;
		m323strategy(bp);
		iowait(bp);
		bp->b_flags &= ~(B_CONF | B_DONE);

		if ( u.u_error ) {
			/* bad news, couldn't even get default config to go */
			brelse(bp);
			return( ERRET );
		}

	}

	/*
	 *	configure the disk with the default info
	 */

	cf = &m323dfcf[m323types[POS(dev)]];
	if( m323setcf( dev, dp, cf)) {	/* set default configuration */
		return( ERRET );		/* permission denied */
	}

	dev |= 7;			/* change to slice 7 */

	bp = geteblk();
	bp->b_proc = u.u_procp;
	bp->b_dev = dev;


#ifdef	DEBUG
	if( DebugM323 > 1 )
		m323print( dev, "default config" );
#endif

	/* do default configure */
	bp->b_flags |= B_CONF;		/* configure */
	bp->b_error = 0;
	bp->b_blkno = 0;
	bp->b_bcount = 0;
	m323strategy(bp);
	iowait(bp);
	bp->b_flags &= ~(B_CONF | B_DONE);

	if ( u.u_error ) {
		/* bad news, couldn't even get default config to go */
		brelse(bp);
		return( ERRET );
	}

#ifdef	DEBUG
	if( DebugM323 > 1 )
		
		m323print( dev, "read vid & cnfig" );
#endif


	/* set up buffer for the config data */
	bp->b_flags  |= B_PHYS | B_READ | B_ABS;
	bp->b_error = 0;
	bp->b_proc = u.u_procp;
	bp->b_dev = dev;
	bp->b_blkno = 0;	/* reading absolute block zero */
	bp->b_bcount = BSIZE;

	/* setup ptrs assuming vid is sector 0,	config area is sector 1	*/
	vp = (struct volumeid *)(bp->b_un.b_addr);
	cfp = (struct dkconfig *)((int)bp->b_un.b_addr + 256);

	/* send	to strategy */
	m323strategy (bp);
	iowait (bp);
	bp->b_flags &= ~(B_PHYS | B_READ | B_ABS | B_DONE);

	/* error check,	signature string */
	if (bp->b_flags & B_ERROR || !m323tst(vp->vid_mac)) {
		brelse(bp);
		u.u_error = ENXIO;
		return( ERRET );
	}

	slccnt = cfp->ios_slccnt;	/* save number of slices */
	slcblk = cfp->ios_slcblk;	/* save block number of slice table */
	
	/* reinitialize using the actual config information read from the disk */

	m323scnfg(dev, cfp);

	dp = &m323utab[POS(dev)];
	dp->slccnt = slccnt;		/* save actual slice count */

	dev |= 7;			/* change to slice 7 */

	bp->b_proc = u.u_procp;
	bp->b_dev = dev;


#ifdef	DEBUG
	if( DebugM323 > 1 )
		m323print( dev, "actual config" );
#endif

	/* do actual configure */
	bp->b_flags |= B_CONF;		/* configure */
	bp->b_error = 0;
	bp->b_blkno = 0;
	bp->b_bcount = 0;
	m323strategy(bp);
	iowait(bp);
	bp->b_flags &= ~(B_CONF | B_DONE);

	if ( u.u_error ) {
		/* bad news, couldn't get the actual config to go */
		brelse(bp);
		return( ERRET );
	}

	/* read slice table and put info in m323_sizes[] */

	if ( (slccnt <= M323MAXSLC)  && (slcblk != 0) ) {

		/* read block that has the slice table */
		bp->b_flags |= B_PHYS|B_ABS|B_READ;
		bp->b_error = 0;
		bp->b_proc = u.u_procp;
		bp->b_dev = dev;
		bp->b_blkno = slcblk;
		bp->b_bcount = BSIZE;

		m323strategy(bp);
		iowait(bp);
		bp->b_flags &= ~(B_PHYS | B_READ | B_ABS | B_DONE);

		if ( !(bp->b_flags & B_ERROR) ) {

			/* copy from the slice table to m323_sizes[] */
			sp = &m323_sizes[POS(dev)][0];
			slcp = (struct size *)((int)bp->b_un.b_addr);
			for (; slccnt > 0 ; slccnt-- ,sp++,slcp++) {

				sp->nblocks = slcp->nblocks;
				sp->blkoff = slcp->blkoff;
			}
		}
	}
	brelse(bp);
	u.u_error = 0;		/* guarantee no error returned here */
				/* at least defaults are in effect */
	return( OKAY );
}

/*
 *   m323setcf: sets configuration parameters for the physical drive 
 *		associated with dev according to the values in the
 *		structure pointed to by cfp. If the request is accepted,
 *		the uib structure pointed to by dp->dconf is modified.
 *		Returns OKAY if change is accepted, else returns ERRET.
 *
 *		If the structure pointed to by cfp is not the
 *		configuration M323NULL and if the change is 
 *		considered compatible with what is already set,
 *		it is accepted. If the change is not compatible
 *		and only the devices indicated by devmsk are open,
 *		the change is accepted and the head values
 *		for the other volume are set to zero, disabling the
 *		other volume.
 */

STATIC int m323setcf(dev, dp, cf) /* set configuration parameters */
register dev;
register struct iobuf *dp;
register struct m323config *cf;
{	
	register struct m323uib *ibp = (struct m323uib *)dp->uib;
	register vol = VOL(dev);

#ifdef	DEBUG
	if( DebugM323 > 1 )
		m323print( dev, "setcf" );
#endif

	if( cf->cf_vsh == M323NLCF )	/* if the config is M323NULL */
		return( ERRET );	/*    reject request	*/

#if 0
	if( cf->cf_ubps == 0 ) 	/* if the unformatted bytes/sec is 0 */
		return( ERRET );	/*    reject request	*/
#endif


	/* if change is compatible */
		if ( cf->cf_spt == ibp->ib_spt &&
			cf->cf_skew == ibp->ib_skew &&
			cf->cf_bps == ibp->ib_bps &&
			cf->cf_ubps == ibp->ib_ubps &&
			cf->cf_gap1 == ibp->ib_gap1 &&
			cf->cf_gap2 == ibp->ib_gap2 &&
			cf->cf_intlv == ibp->ib_intlv &&
			cf->cf_cyls == ibp->ib_cyls &&
			(cf->cf_attrib & M323SSE) == (ibp->ib_attrib & M323SSE)
			) {

		/* set "harmless" and volume specific values */

			ibp->ib_retry = cf->cf_retry;	/* retry count */
			if( cf->cf_attrib & M323CEN )	/* cache enable */
				ibp->ib_attrib |= M323CEN;
			else
				ibp->ib_attrib &= ~M323CEN;

			ibp->ib_sh(vol) = cf->cf_vsh; /* starting head */
			ibp->ib_nh(vol) = cf->cf_vnh; /* head count */
			return( OKAY );
		};

	/* otherwise set all user configurable values in uib */
	ibp->ib_spt = cf->cf_spt;
	ibp->ib_skew = cf->cf_skew;
	ibp->ib_bps = cf->cf_bps;
	ibp->ib_gap1 = cf->cf_gap1;
	ibp->ib_gap2 = cf->cf_gap2;
	ibp->ib_intlv = cf->cf_intlv;
	ibp->ib_retry = cf->cf_retry;
	ibp->ib_cyls = cf->cf_cyls;
	ibp->ib_attrib &= ~M323UATR;	/* set user setable attributes */
	ibp->ib_attrib |= M323UATR & cf->cf_attrib;
	ibp->ib_ubps = cf->cf_ubps;
	ibp->ib_r0 |= FOUR_DRVS;

	/* set heads for this volume */
	ibp->ib_sh(vol) = cf->cf_vsh;	/* starting head */
	ibp->ib_nh(vol) = cf->cf_vnh;	/* head count */

	return( OKAY );

}

/*
 *	m323close: close a device ... only called on the last open.
 *		Watch out for rootdev, swapdev, and pipedev
 *		Called from m323bclose() and m323cclose().
 */

STATIC m323close(dev, flag, otyp, mask)	/* last close on device */
register dev, flag, otyp, mask;
{	register struct iobuf *dp;
	register pos = POS(dev);
	register int *otbl;

#ifdef	DEBUG
	if( DebugM323 )
		m323print(dev, "close");
#endif

	dp = &m323utab[pos];
	otbl=&m323otbl[((CTL(dev) * MAXDEV) + UNIT(dev)) * OTYPCNT];

	if (CTL (dev) >= m323_ctls || ILLEGAL(dev) ) {
		u.u_error = ENXIO;
		return;
	}

	/* mark logical device closed */

	if(otyp == OTYP_LYR)
		--otbl[OTYP_LYR];
	else if( otyp < OTYPCNT)
		otbl[otyp] &= ~mask;

}

/*
 *	m323bclose: close block device
 */

m323bclose(dev, flag, otyp)
register dev, flag, otyp;

{
	dev = minor(dev);
	m323close(dev, flag, otyp, (1 << SLICE(dev)));
}

/*
 *	m323cclose: close character device
 */

m323cclose(dev, flag, otyp)
register dev, flag, otyp;

{
	dev = minor(dev);
	m323close(dev, flag, otyp, (1 << SLICE(dev)));
}

/*
 *	m323strategy: perform disk I/O strategy algorithm.
 *		Call m323start to get the controller started if necessary.
 */

m323strategy(bp)
register struct buf *bp;
{
	register struct iobuf *dp;
	register last, co, dev, seccyl;
	register struct m323ctlr *ctlr;
	register struct iotime *ip;
	static   x;

	register struct m323uib *ibp; 
	register struct size *sp;

	dev = minor(bp->b_dev);
#ifdef	DEBUG
	if( DebugM323 > 1 )
		printf("%s on MVME323 ctl %d, drive %d, vol %d, slice %d blk %d\n",
		"strategy", CTL(dev), UNIT(dev), VOL(dev), SLICE(dev), 
		bp->b_blkno);
#endif

	dp = &m323utab[POS(dev)];
	ibp = (struct m323uib *)dp->uib;
	sp = &m323_sizes[POS(dev)][SLICE(dev)];

	seccyl = ibp->ib_spt*ibp->ib_nh(VOL(dev));
	co = (sp->blkoff)/seccyl;
	last = sp->nblocks;
	

	if (bp->b_bcount&(BSIZE-1)) {		/* no partial block I/O */
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}


	if (bp->b_blkno < 0 || bp->b_blkno >= last ) {
		if (bp->b_blkno == last && bp->b_flags&B_READ)
			bp->b_resid = bp->b_bcount;
		else {
			bp->b_flags |= B_ERROR;
			bp->b_error = ENXIO;
		}
		iodone(bp);
		return;
	}

	bp->av_forw = NULL;
	bp->av_back = NULL;
	bp->b_start = lbolt;


	bp->cylin = bp->b_blkno*(BSIZE/ibp->ib_bps)/seccyl + co;
	if(bp->b_flags & B_ABS)
		bp->soff = 0;
	else
		 bp->soff = sp->blkoff;

	/* log I/O statistics for this drive */
	ip = &m323stat[POS(dev)];

	if( ( bp->b_flags & NON_RWIO) )
		++(ip->io_other);
	else {
		++(ip->io_cnt);	
		ip->io_bcnt += bp->b_bcount/BSIZE;
	}


	/* put request in buffer queue */
	x = spl6();		/* spl() high enough to block m323timer() */
	if( dp->b_actf == NULL ) {
		dp->b_actf = bp;
		dp->b_actl = bp;
		dp->acts = (int)bp;
	} else {
		register struct buf *ap;
		register struct buf *cp;

		if(((ip->io_cnt)&(STEPSIZE-1)) == 0 )
			dp->acts = (int)dp->b_actl;
		for( ap=(struct buf *)dp->acts; cp=ap->av_forw; ap=cp ) {
			int	s1,s2;

			if ((s1 = ap->cylin - bp->cylin)<0) s1 = -s1;
			if ((s2 = ap->cylin - cp->cylin)<0) s2 = -s2;
			if (s1 < s2)
				break;
		}
#ifdef COALES
		if((m323coal) && (ap != dp->b_actf) 
		&& (try323coales(ap, bp, M323MXCOAL-1,
		 ap->b_bcount/BSIZE ) == 0))
		
			;		/* bp was linked for coalescing */
		else{
					/* bp was not linked for coalescing */
			ap->av_forw = bp;
			if ((bp->av_forw = cp) == NULL)
				dp->b_actl = bp;
		}
#else
		ap->av_forw = bp;
		if ((bp->av_forw = cp) == NULL)
			dp->b_actl = bp;
#endif
	}
	dp->qcnt++;		/* another one in queue */

	ctlr = &m323ctlr[CTL(dev)];
	if( ctlr->cc_actp == NULL ) {	/* controller inactive ? */
		m323start(CTL(dev));
		m323timer(WSTART);	/* start timer */
	}
	splx(x);
}

/* 	try323coales: try to combine individual block I/O requests. 
*	Call under interrupt protection.
*	Caller must assure that "cbp" is NOT the head of the I/O request queue
*    	which may be an I/O request already in progress.
*
*	Requests to be combined are linked together using the "av_back"
*	pointer. 
*
*/
#ifdef COALES
try323coales(cbp, newbp, maxentries, pblocks)
register struct buf *cbp;	/* first bp in list */ 
register struct buf *newbp;	/* try to link this new bp to end of list  */
register int	maxentries;	/* maximum number of bps permitted in list */
register short	pblocks;	/* physical blocks per file system block */
{
  register struct buf *bp;	/* last bp in list */
  register int  entries; 

  if(((cbp->b_flags & B_PHYS) == 0)
  &&((newbp->b_flags & B_PHYS) == 0)
  &&((cbp->b_flags & B_READ) == (newbp->b_flags & B_READ))){

	bp = cbp;
	for( entries=0 ; bp->av_back ; (bp = bp->av_back), entries++)
		;

	if(( entries < maxentries)
	&&(newbp->soff == bp->soff)
	&&(newbp->b_blkno == bp->b_blkno + pblocks)){

		bp->av_back = newbp;

		return(0);	/* Successful */
	}
   }
	return(-1);	/* Unsuccessful */
}
#endif
 		


/*
 *	m323start: start controller
 */

m323start( ctl )
register ctl;

{	register struct buf *bp;
	register struct iobuf *dp, *nextdp;
	register d;
	static unsigned int drobin[2];

#ifdef	DEBUG
	if( DebugM323 > 2 )
		m323print( MKDEV( ctl, 0), "start");
#endif

 	/* select next device to be serviced */
	dp = (struct iobuf *)NULL; 

	for( d=0; d < MAXDEV; d++ ) {
		nextdp = &m323utab[((MAXDEV*ctl)+drobin[ctl])];
		drobin[ctl] = (drobin[ctl]+1) % MAXDEV;
		if( nextdp->b_actf != NULL ) {
			dp = nextdp;
			break;
		}
	}
	if(( m323ctlr[ctl].cc_actp = dp ) == NULL ) {
		return;		/* if no requests, cc_actp marked idle */
	}
	dp->b_errcnt = TRYCNT;
	m323dstart( dp );
}



/*
 *	m323dstart: start device
 */

m323dstart( dp )
register struct iobuf *dp;
{	 register struct buf *bp;

	bp = dp->b_actf;			/* next request */

#ifdef	DEBUG
	if( DebugM323 > 2 )
		m323print( bp->b_dev, "dstart");
#endif
	dp->b_dev = bp->b_dev;			/* current device */

	if( bp->b_flags&B_CONF ) {		/* B_CONF */
		m323docf(dp,bp);

	} else if( bp->b_flags&(B_RESTORE) ) {	/* B_RESTORE */
		m323restore(dp,bp);
	} else if( bp->b_flags&(B_FORMAT) ) {	/* B_FORMAT */
		dp->wtime = WERR*4;		/* increase timeout */
		m323format(dp,bp);
	} else if( bp->b_flags&(B_RDFLAW) ) {	/* read mfr defect list */
		m323rdmfr(dp,bp);
	} else {				/* B_READ or B_WRITE */
		/* start I/O segment */
		dp->wtime = WERR;
		bp->b_resid = bp->b_bcount;	/* remaining bytes */
		m323ustart(dp,bp);
	}
}

/*
 *	m323docf: perform actual configuration command
 */

STATIC m323docf(dp,bp)
register struct iobuf *dp;
register struct buf *bp;

{	register struct m323iopb *pbp;
	register dev = bp->b_dev;

#ifdef	DEBUG
	if( DebugM323 > 2 )
		m323print( dev, "docf" );
#endif

	pbp = (struct m323iopb *)dp->cmdbuf;

	/* set up iopb for initialize command */
	pbp->pb_cmd = INIT_DRV;
	pbp->pb_opt = MKOPDEV(dev)|INT_EN;
	pbp->pb_ilev = (DRV_NUM(dev) | m323_ilev[CTL(dev)]);
	pbp->pb_stat = 0;
	pbp->pb_err = 0;
#ifdef	m88k
	pbp->pb_lsect.hi = pbp->pb_lsect.lo = 0;
	pbp->pb_un1.pb_typ2.pb_scnt = 0;
	pbp->pb_badr.hi = (short)((int)(&m323uib[POS(dev)])>>16);
	pbp->pb_badr.lo = (short)&m323uib[POS(dev)];
	pbp->pb_un1.pb_typ2.pb_bmtyp = MEMT16;
	pbp->pb_un1.pb_typ2.pb_bamod = m323adrm;
#else
	pbp->pb_lsect = 0;
	pbp->pb_scnt = 0;
	pbp->pb_badr = (char *)&m323uib[POS(dev)];
	pbp->pb_bmtyp = MEMT16;
	pbp->pb_bamod = m323adrm;
#endif

	dp->b_errcnt = NORETRY;
	dp->io_start = lbolt;
	m323doit(CTL(dev));
}

/*
 *	m323rdmfr: read manufacturers defect list 
 */
m323rdmfr(dp,bp)
register struct iobuf *dp;
register struct buf *bp;
{
	register struct m323iopb    *pbp;
	register struct m323uib     *ibp;
	register dev = bp->b_dev;

	pbp = (struct m323iopb *)dp->cmdbuf;
	ibp = (struct m323uib *)dp->uib;

	pbp->pb_cmd = READ_FLAW;
	pbp->pb_cyl = bp->b_resid;
	pbp->pb_trk = bp->b_error;
	pbp->pb_psect = 0; 
	pbp->pb_dmac = 0;
#ifdef	m88k
	pbp->pb_un1.pb_typ2.pb_scnt = 1;
	pbp->pb_badr.hi = (short)((int)(bp->b_un.b_addr) >> 16);
	pbp->pb_badr.lo = (short)bp->b_un.b_addr;
#else
	pbp->pb_scnt = 1;
	pbp->pb_badr = bp->b_un.b_addr;
#endif
	pbp->pb_opt = MKOPUN(dev) | INT_EN ;  /* since no logical translation don't need MKODEV(dev) to give VOL to map sector */
	pbp->pb_ilev = (DRV_NUM(dev) | m323_ilev[CTL(dev)]);
	pbp->pb_stat = 0;
	pbp->pb_err = 0;

	/* set memory type and address modifiers */
#ifdef	m88k
	if( (int)pbp->pb_badr.lo & LONGMSK ) {	/* if not long alligned */
		pbp->pb_un1.pb_typ2.pb_bmtyp = MEMT16;	/* force 16 bit trans */
	} else {
		pbp->pb_un1.pb_typ2.pb_bmtyp = m323mtyp;/* use specified type */
	}
	pbp->pb_un1.pb_typ2.pb_bamod = m323adrm;
#else
	if( (int)pbp->pb_badr & LONGMSK ) {	/* if not long alligned */
		pbp->pb_bmtyp = MEMT16;		/* force 16 bit transfer */
	} else {
		pbp->pb_bmtyp = m323mtyp;	/* use specified type */
	}
	pbp->pb_bamod = m323adrm;
#endif

	bp->b_error = 0;
	dp->io_start = lbolt;

	m323doit( CTL(dev) );
}

/*
 *	m323restore: restore the drive to clear a drive fault
 */
m323restore(dp,bp)
register struct iobuf *dp;
register struct buf *bp;
{
	register struct m323iopb    *pbp;
	register struct m323uib     *ibp;
	register dev = bp->b_dev;

	pbp = (struct m323iopb *)dp->cmdbuf;
	ibp = (struct m323uib *)dp->uib;

	pbp->pb_cmd = RESTORE;
	pbp->pb_dmac = 0;
#ifdef	m88k
	pbp->pb_badr.hi = (short)((int)(bp->b_un.b_addr) >>16);
	pbp->pb_badr.lo = (short)bp->b_un.b_addr;
#else
	pbp->pb_badr = bp->b_un.b_addr;
#endif
	pbp->pb_opt = MKOPUN(dev) | INT_EN ;  /* since no logical translation don't need MKODEV(dev) to give VOL to map sector */
	pbp->pb_ilev = (DRV_NUM(dev) | m323_ilev[CTL(dev)]);
	pbp->pb_stat = 0;
	pbp->pb_err = 0;

	/* set memory type and address modifiers */
#ifdef	m88k
	if( (int)pbp->pb_badr.lo & LONGMSK ) {	/* if not long alligned */
		pbp->pb_un1.pb_typ2.pb_bmtyp = MEMT16;	/* force 16 bit trans */
	} else {
		pbp->pb_un1.pb_typ2.pb_bmtyp = m323mtyp;/* use specified type */
	}
	pbp->pb_un1.pb_typ2.pb_bamod = m323adrm;
#else
	if( (int)pbp->pb_badr & LONGMSK ) {	/* if not long alligned */
		pbp->pb_bmtyp = MEMT16;		/* force 16 bit transfer */
	} else {
		pbp->pb_bmtyp = m323mtyp;	/* use specified type */
	}
	pbp->pb_bamod = m323adrm;
#endif

	bp->b_error = 0;
	dp->io_start = lbolt;

	m323doit( CTL(dev) );
}

/*
 *	m323format: perform actual format command
 */

m323format(dp,bp)
register struct iobuf *dp;
register struct buf *bp;

{
	register struct m323iopb    *pbp;
	register struct m323uib     *ibp;
	register m323fmt *fmtp;
	register dev = bp->b_dev;
	register long seccyl;		/* sectors per cylinder */
	register long secno;		/* sector number */

#ifdef DEBUG
	if  ( DebugM323 > 2 )
	    m323print( dev, "format" );
#endif

	pbp = (struct m323iopb *)dp->cmdbuf;
	ibp = (struct m323uib *)dp->uib;
	fmtp = (m323fmt *)(bp->b_un.b_addr);

	if  ( bp->b_flags & (B_FMTA) ) {  	/* Map alternate track */
		pbp->pb_cmd = MAP_TRAK;
		pbp->pb_altcyl = fmtp->dk_alt / ibp->ib_nh( VOL(dev) );
		/*  Actual head =  track offset + starting head number */
		pbp->pb_alttrk = fmtp->dk_alt % ibp->ib_nh( VOL(dev) ); 
		pbp->pb_alttrk += ibp->ib_sh( VOL(dev) );
		pbp->pb_pat = FMT_PATM;
	}
	else if  ( bp->b_flags & (B_FMTS) ) {	/* Format with sector slip */
		pbp->pb_cmd = MAP_SECT;
		pbp->pb_psect = fmtp->dk_sector;
		pbp->pb_pat = FMT_PATS;
	}
	else {					/* Simple Format */
		pbp->pb_cmd = FMT_TRK;
		pbp->pb_pat = FMT_PATT;
	}

	/*	Set up Common code	*/
	pbp->pb_opt = MKOPUN(dev) | INT_EN ;  /* since no logical translation don't need MKODEV(dev) to give VOL to map sector */
	pbp->pb_ilev = (DRV_NUM(dev) | m323_ilev[CTL(dev)]);
	pbp->pb_stat = 0;
	pbp->pb_err = 0;
	pbp->pb_cyl = fmtp->dk_cyl;
	pbp->pb_trk = fmtp->dk_head;
	dp->b_errcnt = NORETRY;
	dp->io_start = lbolt;

	m323doit( CTL(dev) );
}

/*
 *	m323ustart: perform I/O on specified unit
 */

m323ustart(dp,bp)
register struct iobuf *dp;
register struct buf *bp;

{	register struct m323iopb *pbp;
	register struct m323uib *ibp;
	register dev = bp->b_dev;
	register char *  bufaddr;
	register struct buf *nbp;
	register struct m323list *list;
	register int i;

#ifdef	DEBUG
	if( DebugM323 > 2 )
		m323print( dev, "ustart" );
#endif

	pbp = (struct m323iopb *)dp->cmdbuf;
	ibp = (struct m323uib *)dp->uib;


#ifdef COALES
	if(bp->av_back){
		list = (struct m323list *)SGLIST(dev);
#ifdef	m88k
		pbp->pb_badr.hi = (short)((int)(list) >> 16);
		pbp->pb_badr.lo = (short)list;
#else
		pbp->pb_badr = (char *)list;
#endif
		for(nbp = bp, i=0; nbp;
		    (nbp = nbp->av_back), i++, list++){
			list->bcount = nbp->b_bcount;
#ifdef	m88k
			list->addr_hi = (ushort)((int)(nbp->b_un.b_addr)>>16);
			list->addr_lo = (ushort)nbp->b_un.b_addr;
#else
			list->addr = nbp->b_un.b_addr;
#endif
			list->memtype = m323mtyp;  
			list->addrmod = m323adrm;
   		}
		pbp->pb_entries = i;
#ifdef DEBUG
		if(i >cpr323thresh){
		    if(bp->b_flags&B_READ)
			printf("m323ustart: combined %d read requests\n",i);
		    else
			printf("m323ustart: combined %d write requests\n",i);
		}
#endif
#ifdef	m88k
		pbp->pb_un1.pb_typ2.pb_scnt = i * (bp->b_resid/ibp->ib_bps);
#else
		pbp->pb_scnt = i * (bp->b_resid/ibp->ib_bps);
#endif
		pbp->pb_cmd = (bp->b_flags&B_READ)? SCATTER : GATHER;
   	}
	else
	{
#endif
#ifdef	m88k
		pbp->pb_badr.hi = (short)((int)(bp->b_un.b_addr) >> 16);
		pbp->pb_badr.lo = (short)bp->b_un.b_addr;
#else
		pbp->pb_badr = bp->b_un.b_addr;
#endif
		pbp->pb_cmd = (bp->b_flags&B_READ)? READ_SEC: WRIT_SEC;
#ifdef	m88k
		pbp->pb_un1.pb_typ2.pb_scnt = bp->b_resid/ibp->ib_bps;
#else
		pbp->pb_scnt = bp->b_resid/ibp->ib_bps;
#endif
#ifdef COALES
	}
#endif


	pbp->pb_opt = MKOPDEV(dev)|LOGICAL|INT_EN|ECC_EN;
	pbp->pb_ilev = (DRV_NUM(dev) | m323_ilev[CTL(dev)]);
	pbp->pb_stat = 0;
	pbp->pb_err = 0;
#ifdef	m88k
	pbp->pb_lsect.hi = ((bp->b_blkno*(BSIZE/ibp->ib_bps))+(int)bp->soff >> 16);
	pbp->pb_lsect.lo = (bp->b_blkno*(BSIZE/ibp->ib_bps))+(int)bp->soff;
#else
	pbp->pb_lsect = (bp->b_blkno*(BSIZE/ibp->ib_bps))+(int)bp->soff;
#endif

	/* set memory type and address modifiers */
#ifdef	m88k
	if( (int)pbp->pb_badr.lo & LONGMSK ) {	/* if not long alligned */
		pbp->pb_un1.pb_typ2.pb_bmtyp = MEMT16;		/* force 16 bit transfer */
#else
	if( (int)pbp->pb_badr & LONGMSK ) {	/* if not long alligned */
		pbp->pb_bmtyp = MEMT16;		/* force 16 bit transfer */
#endif
#ifdef DEBUG
		m323io16b++;
#endif
	} else {
#ifdef DEBUG
		m323io32b++;
#endif
#ifdef	m88k
		pbp->pb_un1.pb_typ2.pb_bmtyp = m323mtyp;	/* use specified type */
	}
	pbp->pb_un1.pb_typ2.pb_bamod = m323adrm;
#else
		pbp->pb_bmtyp = m323mtyp;	/* use specified type */
	}
	pbp->pb_bamod = m323adrm;
#endif

	dp->io_start = lbolt;	/* start the clock, this one's ready to go */

	m323doit(CTL(dev));
}

/*
 *	m323doit: initiate command to controller
 */

m323doit(ctl)
register ctl;

{	register csr, dev;
	register struct device *rp;

	dev = MKDEV(ctl,0);
	rp = (struct device *)m323_addr[ctl];

#ifdef	DEBUG
	if( DebugM323 > 2 ) {
		m323print( dev , "doit" );
#ifdef	DEBUGC
		m323prall(rp);
#endif
	}
#endif


	csr = rp->csr;		/* get controller status */

	/* if the board is not ok or the controller is busy */
	if( !(csr & M323BDOK) || (csr & GO_BUSY) ) {
#ifdef	DEBUG
		if( DebugM323 > 2 ) {
			printf("Status = %x\n", csr&0xffff );
		}
#endif
		/* go no further */
		m323print(dev, "not ready" );
		return;
	}
#ifdef ANGEL
	dcacheflush();
#endif

	/* otherwise, set the GO bit among others */

	rp->csr = M323SLED|M323BDOK|M323NOSF|GO_BUSY;
}

/*
 *	m323intr: handle mvme323 interrupts
 */

m323intr(base_dev)
{
	if(m323MACSI)
		x323intr(base_dev);
	else
		m323ointr(base_dev);
}

m323ointr(base_dev)
{	register struct device *rp;
	register struct iobuf *dp;
	register struct buf *bp;
	register struct buf *coalbp;
	register struct m323iopb *pbp;
	register ctl, dev, erf=0;
	struct buf	*oldbp;
	int csr;

	ctl = base_dev / (MAXDEV*M323MAXSLC);
	dev = MKDEV(ctl,0);

#ifdef	DEBUG
	if( DebugM323 > 3 )
		m323print(dev, "intr");
#endif

	if( m323_addr[ctl] == NULL ) {
#ifdef	DEBUG
		m323print( dev, "Unexpected interrupt" );
#endif
		return;
	}

	rp = (struct device *) m323_addr[ctl];

	/*
	 *	Command completion interrupt
	 */

	csr = rp->csr;			/* save original for error logging */
	if( rp->csr & DONE_INT ) {

		rp->csr &= ~DONE_INT;	/* clear interrupt */

		if( (dp = m323ctlr[ctl].cc_actp) == NULL ) { /* ctlr active? */
			/* log illegal interrupt */
			return;
		}

		if( (bp = dp->b_actf) == NULL ) { /* drive active? */
			/* assert this must never be true */
#ifdef DEBUG
			m323print( dp->b_dev, "lost request, cannot be" );
#endif
			return;
		}


		dp->wtime = WOK;	/* cancel m323timer */
		bp->b_resid = 0;	/* default no error indication */
		bp->cylin = 0;		/* clear cylinder number */
		dev = bp->b_dev;	/* logical dev being serviced */
		pbp = (struct m323iopb *)dp->cmdbuf;

		/* if a RESTORE just completed , then retry original command */
		if( bp->b_flags & B_RESTORE){
			bp->b_flags &= ~B_RESTORE;
			m323dstart(dp);
			return;
		}

		/* if command completed ok with exception */
		if( pbp->pb_stat == DONE_EXC ) {

			/* log soft error, with block offset to the I/O */
			m323elog(rp, dp, IO_BOFF, csr);

		/* else if status indicates a serious error */
		} else if( pbp->pb_stat != DONE_OK ) {

			/* log error, with block offset to the I/O */
			m323elog(rp, dp, IO_BOFF, csr);

			/* if try count not exhausted, retry */
			if( --dp->b_errcnt > 0 ) {
#ifdef DEBUG
				if(DebugM323) {
					m323print(dev, "M323: Soft Error\n");
					printf(" Block Offset %d\n",(IO_BOFF));
				}
#endif
				if(pbp->pb_err == DRV_FLT)
					bp->b_flags |= B_RESTORE;

				m323dstart( dp );
				return;
			}

			/* hard error */
			erf++;
			m323print(dev, "M323:Error");
			m323prall(rp);

			/* indicate error in request buffer */
			bp->b_resid = bp->b_bcount;
			bp->b_flags |= B_ERROR;
		}

		logberr( dp, erf);	/* close error log on this request */


		/* dequeue buffer */
		dp->b_actf = bp->av_forw;
		dp->qcnt--;
		if (bp == (struct buf *)dp->acts)
			dp->acts = (int)dp->b_actf;

		/* log read/write I/O times */
		if( !( bp->b_flags & NON_RWIO) )
		{	register struct iotime *ip = &m323stat[POS(dev)];

			ip->io_resp += lbolt - bp->b_start;
			ip->io_act += lbolt - dp->io_start;
		}

		/*
		 *	notify requestor I/O completed and free buffer
		 */

#ifdef COALES

		coalbp = bp->av_back;
		while(coalbp){ 
			coalbp->b_resid = bp->b_resid;
			coalbp->b_flags |= bp->b_flags & B_ERROR;
			oldbp = coalbp;
	    	        coalbp = coalbp->av_back;
			iodone(oldbp);
		}
#endif

		iodone(bp);
			
		/* start next I/O */
		m323start(ctl);
		return;
	}


	/*
	 *	Status change interrupt
	 */

	if( rp->csr & SCHG_INT ) {
		register drive;

		/* read new status for specified drive */
		drive = SCDRV( rp->csr );	/* identify drive */
		m323utab[ POS( MKDEV(ctl,drive) ) ].dstat = rp->ds[drive];

		rp->csr &= ~SCHG_INT;	/* clear interrupt */
	}
}


/*
 *	m323breakup: Physical I/O Breakup routine
 */

m323breakup(bp)
register struct buf *bp;
{
	dma_breakup(m323strategy, bp, BSIZE);
}


/*
 *	m323read: raw I/O read entry point
 */

m323read(dev)
register dev;
{	
#ifdef	DEBUG
	if( DebugM323 )
		m323print(dev, "read");
#endif

	if(physck(m323_sizes[POS(dev)][SLICE(dev)].nblocks,B_READ))
		physio(m323breakup, 0, dev, B_READ);
}



/*
 *	m323write: raw I/O write entry point
 */

m323write(dev)
register dev;
{
#ifdef	DEBUG
	if( DebugM323 )
		m323print(dev, "write");
#endif

	if(physck(m323_sizes[POS(dev)][SLICE(dev)].nblocks,B_WRITE))
		physio(m323breakup, 0, dev, B_WRITE);
}

/*
 *	m323ioctl: ioctl(2) support for the m323
 *
 *	Supported functions:
 *
 *		Format Track
 *		Format Unit
 *		Get Config
 *		Set Config (must supply cf_gpl3 for formatting)
 *		Format Alternate
 *              Read Manufacturers Defect List 
 */

m323ioctl(dev, cmd, arg, mode)
register dev;		/* device */
register cmd;		/* ioctl command */
register char *arg;	/* ctl args */
register mode;		/* mode */

{	
	register struct iobuf *dp;
	register struct buf *bp;
	register struct m323uib *ibp;
	struct m323config cf;
	struct buf *geteblk();

#ifdef DEBUG
	if  ( DebugM323 )
	    m323print( dev, "ioctl" );
#endif

	dev = minor(dev);
	dp = &m323utab[POS(dev)];
	ibp = (struct m323uib *)dp->uib;

	/*  linesw table guarantees char device, So check for slice 7 */
	if  (  SLICE(dev)  !=  7 ) {
	       u.u_error = EINVAL;
	       return;
	}
	bp = geteblk();


	switch( cmd )  {

	case M323SET:	/*  Set configuration */
			if ( (copyin( (struct m323config *)arg, &cf, sizeof(struct m323config) ) ) == -1 ) {
				u.u_error = EINVAL;
				brelse(bp);
				return;
			}
			if ( m323setcf( dev, dp, &cf)  ) {
				u.u_error = EINVAL;
				brelse(bp);
				return;
			}
			bp->b_flags |= B_CONF;	/* dev points to Vol */
			bp->b_blkno = 0;
			m323opst(dev,bp);
			break;

	case M323GET:	/*  Get configuration */
			cf.cf_vsh = ibp->ib_sh( VOL(dev) );
			cf.cf_vnh = ibp->ib_nh( VOL(dev) );
			cf.cf_spt = ibp->ib_spt;
			cf.cf_skew = ibp->ib_skew;
			cf.cf_bps = ibp->ib_bps;
			cf.cf_gap1 = ibp->ib_gap1;
			cf.cf_gap2 = ibp->ib_gap2;
			cf.cf_intlv = ibp->ib_intlv;
			cf.cf_retry = ibp->ib_retry;
			cf.cf_cyls = ibp->ib_cyls;
			cf.cf_attrib = ibp->ib_attrib;
			if  ( (copyout( &cf, (struct m323config *)arg, sizeof(struct m323config) ) ) == -1) {
	       			u.u_error = EINVAL;
				brelse(bp);
				return;
			}
			brelse(bp);
			return;
			
			

	case RDMFRLIST: /* Read Manufacturers Defect List from disk */   

			rdmfrlist(dev, arg, dp, bp);
			brelse(bp);	
			return;



	case DKSETSLC:
			/* set the fields in m323_sizes */
			if ( copyin (arg, &m323_sizes[POS(dev)][0],SLCTSIZE ) )
				u.u_error = EFAULT;
			return;
		

	case DKGETSLC:

			/* write out the fields in m323_sizes */
			if ( copyout ( &m323_sizes[POS(dev)][0], arg,SLCTSIZE ) )
				u.u_error = EFAULT;
			return;

	case DKSETVID:	/* set vid info */
			if( copyin(arg,bp->b_un.b_addr,sizeof(struct volumeid)) == -1)
				u.u_error = EINVAL;
			else
				if (m323svid(dev,bp->b_un.b_addr) == ERRET )
					u.u_error = EINVAL;
			break;

				
	case DKGETVID:	/* get vid info */
			if( copyin(arg,bp->b_un.b_addr,sizeof(struct volumeid)) == -1)
				u.u_error = EINVAL;
			else {
				m323gvid(dev,bp->b_un.b_addr);
				if (copyout (bp->b_un.b_addr,arg,sizeof(struct volumeid) ) == -1)
					u.u_error = EINVAL;
			}
			break;

	case DKSETCFG:	/* set config info */
			if( copyin(arg,bp->b_un.b_addr,sizeof(struct dkconfig)) == -1)
				u.u_error = EINVAL;
			else{
				if (m323scnfg(dev,bp->b_un.b_addr)== ERRET )
					u.u_error = EINVAL;
				else {
					bp->b_flags |= B_CONF;
					bp->b_blkno = 0;
					m323opst(dev,bp);
				}
			}
			break;

	case DKGETCFG:	/* get config info */
			if( copyin(arg,bp->b_un.b_addr,sizeof(struct dkconfig)) == -1)
				u.u_error = EINVAL;
			else {
				m323gcnfg(dev,bp->b_un.b_addr);
				if (copyout (bp->b_un.b_addr,arg,sizeof(struct volumeid) ) == -1)
					u.u_error = EINVAL;
			}
			break;

	case DKSETINFO:	/* set vid and config info */
			if( copyin(arg,bp->b_un.b_addr,sizeof(struct dkblk0)) == -1)
				u.u_error = EINVAL;
			else{
				if (m323svid(dev,bp->b_un.b_addr)== ERRET )
					u.u_error = EINVAL;
				else {
					if (m323scnfg(dev,bp->b_un.b_addr + sizeof(struct volumeid)) == ERRET )
						u.u_error = EINVAL;
					else {
							/*
			 				* The volumeid is available, check to see that
			 				* the bad spots to be given to the driver for a
			 				* format command are going to be in the form of
			 				* "byte offset".
			 				*/
							register struct volumeid *vid;

							vid = (struct volumeid *)bp->b_un.b_addr;
							if (vid->vid_strat != DKHARDSPOT) {
								brelse(bp);
								u.u_error = ENXIO;
								return;
							}
						bp->b_flags |= B_CONF;
						bp->b_blkno = 0;
						m323opst(dev,bp);
	
					}
				}
			}
			break;
	case DKGETINFO:	/* get vid and config info */
			if( copyin(arg,bp->b_un.b_addr,sizeof(struct dkblk0)) == -1)
				u.u_error = EINVAL;
			else {
				m323gvid(dev,bp->b_un.b_addr);

				m323gcnfg(dev,bp->b_un.b_addr+sizeof(struct volumeid));
				if (copyout (bp->b_un.b_addr,arg,sizeof(struct dkblk0) ) == -1)
					u.u_error = EINVAL;
			}

			break;

	case DKFORMAT:	/* format specified range of tracks */
			m323dkfmt( dev, arg, bp, ibp );
			break;

	case DKFIXBADSPOT:	/* format bad spot */
			bp->b_flags |= B_FMTBS;
			m323dkfmt(dev, arg, bp, ibp);
			break;

	default:	u.u_error = EINVAL;
			brelse(bp);
			return;
	}

	/* clear the flags THIS routine wrote */
	bp->b_flags &= ~( B_FORMAT | B_FMTA | B_FMTS | B_CONF | B_FMTBS);
	brelse(bp);
	return;
}

rdmfrlist(dev, arg, dp, bp)
register dev;		/* device */
register char *arg;	/* ioctl args */
register struct iobuf *dp;
register struct buf *bp;
{
	struct m323mlargs larg;	/* ioctl arguments */
	char  *bufptr;
	int	head;			/* head */
	uint  listsize;			/* sector size */


			if  ( (copyin( (struct m323mlargs *)arg, &larg, 
				sizeof(struct m323mlargs))) == -1) {
				u.u_error = EINVAL; 
				return;
			}
			listsize = larg.listsize;
			
			/* read first sector on all tracks of last cylinder */
			bp->b_proc = u.u_procp;
			bp->b_dev = dev;
			bp->b_bcount = BSIZE;
			for( head=0, bufptr=larg.buffer; head <larg.heads;
			     head++, bufptr += listsize){
	
				bp->b_resid = larg.maxcyl; 	
				bp->b_error = head;
				bp->b_flags |= B_RDFLAW;
				bp->b_flags &= ~(B_ERROR | B_DONE);
				m323strategy(bp);
				iowait(bp);
				if(bp->b_flags & B_ERROR){
					u.u_error = EIO;
					break;
				}
				copyout(bp->b_un.b_addr, bufptr, listsize);
			}

}

/*
 *	m323opst: start operation set up in buf struct
 *		Returns non-zero on error.
 */

m323opst(dev,bp)
int dev;
struct buf *bp;
{	
	int rval;

	bp->b_proc = u.u_procp;
	bp->b_dev = dev;
	bp->b_bcount = 0;
	bp->b_error = 0;
	m323strategy(bp);
	iowait(bp);
	if( (rval = bp->b_flags & B_ERROR) && !u.u_error )
		u.u_error = EIO;
	return( rval );
}



/*
 *	m323dkfmt: disk format
 *		Returns with u.u_error set on error.
 */

m323dkfmt( dev, arg, bp, ibp )
int dev;
struct dkfmt *arg;
struct buf *bp;
struct m323uib *ibp;
{
	struct dkfmt dkfmt, *dkfp;
	struct dkbadlst *flp, *uflp, bs;		/* flaw & user_flaw ptr */
	m323fmt *fmtp;
	int trk, sector, f, trkno, trkcnt, badcnt, maxtrks;
	int hds, shd;


	/* fetch dkfmt struct */

	dkfp = &dkfmt;
	if((bp->b_flags & B_FMTBS) == 0 ) {
		if ( copyin( arg, dkfp, sizeof( struct dkfmt) ) ) {
			u.u_error = EFAULT;
			return;
		}
	}
	else {
		/* fix bad spot -- fetch dkbadspot structure */
		if ( copyin( arg, &bs, sizeof( struct dkbadlst) ) ) {
			u.u_error = EFAULT;
			return;
		}
		/* fix bad spot command must have an alternate track */
		if(bs.dk_alt == 0) {
			u.u_error = EINVAL;
			return;
		}
		/*
		 * Set up the dkfmt struct to look like a format command,
		 * then proceed like a format.
		 */
		dkfp->dk_trkno = bs.dk_cyl * ibp->ib_nh(VOL(dev)) + bs.dk_head;
		dkfp->dk_trkcnt = 1;
		dkfp->dk_badcnt = 1;
		dkfp->dk_badlst = &bs;
	}

#ifdef DEBUG
	if(DebugM323) {
		printf(" dkfmt: trkcnt %d, trkno %d, badcnt %d, badlst %d, altcnt %d, altlst %d\n",
		dkfp->dk_trkcnt, dkfp->dk_trkno,
		dkfp->dk_badcnt, dkfp->dk_badlst,
		dkfp->dk_altcnt, dkfp->dk_altlst );
	}
#endif
	/* check args */
	if( ! (trkcnt = dkfp->dk_trkcnt) )
		return;				/* no work to do */

	/* maxtrks = cyls * hds */
	hds = ibp->ib_nh(VOL(dev));
	maxtrks = ibp->ib_cyls * hds;
	trkno = dkfp->dk_trkno;
	badcnt = dkfp->dk_badcnt;

	if( trkno < 0 || trkcnt < 0 || badcnt < 0
	    || trkno + trkcnt > maxtrks ) {
		u.u_error = EINVAL;
		return;
	}

	/* format range of tracks */
	bp->b_flags |= B_FORMAT;
	shd = ibp->ib_sh( VOL(dev) );		/* starting head for volume */
	fmtp = (m323fmt *)bp->b_un.b_addr;
	for( trk=trkno; trk < trkno + trkcnt; trk++) {
		fmtp->dk_cyl = trk/hds;
		fmtp->dk_head = trk % hds + shd;
		bp->b_blkno = TRK2BLK(ibp,trk);
		bp->b_flags &= ~B_DONE;
		if( m323opst(dev,bp) )
			return;		/* exit on error */
		if( issig() ) {	
			u.u_error = EINTR;	/* exit on signal */
			return;
		}
	}

	/* format out each flaw */
	if( badcnt && !(uflp = dkfp->dk_badlst)) {
		u.u_error = EINVAL;
		return;			/* must have badlst to proceed */
	}
	for( f=1; f <= badcnt; f++, uflp++) {
		/* fetch flaw data */
		if((bp->b_flags & B_FMTBS) == 0 ) {
			/* format command */
			if ( copyin( uflp, fmtp, sizeof(struct dkbadlst) ) ) {
				u.u_error = EFAULT;
				return;		/* exit on error */
			}
		}
		else {
			/* fix bad spot command */
			fmtp->dk_alt = bs.dk_alt;	/* point to bad spot structure */
		}

#ifdef DEBUG
	if(DebugM323 == 1) {
		printf(" Flaw data: cyl %d, head %d, offset %d, alt %d\n",
		fmtp->dk_cyl, fmtp->dk_head, fmtp->dk_offset, fmtp->dk_alt);
	}
#endif
		/* verify bad spot is within the range */
		trk = fmtp->dk_cyl * hds + fmtp->dk_head;
		if( trk < trkno || trk >= trkno + trkcnt || fmtp->dk_alt < 0) {
			m323print(dev, "M323: Ignored, Out of Range \n");
			printf(" Flaw data: cyl %d, head %d, offset %d, alt %d\n",
				fmtp->dk_cyl, fmtp->dk_head, fmtp->dk_offset,
				fmtp->dk_alt);
			continue;
		}

		if( fmtp->dk_alt ) {
		/* map track */
			fmtp->dk_head += shd;	/* adjust for volume */
			bp->b_blkno = TRK2BLK(ibp,trk);
			bp->b_flags &= ~(B_DONE|B_FMTA|B_FMTS);
			bp->b_flags |= B_FMTA;
			if( m323opst(dev,bp) )
				return;		/* exit on error */
		} else {
		/* slip sector */
			if( ! ibp->ib_ubps ) {
				m323print(dev, "M323: unformatted sector size 0\n");
				printf(" cannot calculate slip sector\n");
				u.u_error = EINVAL;
				return;			/* exit on error */
			}
			fmtp->dk_head += shd;	/* adjust for volume */
			{
			/* calculate physical fault sector */
#			define MARGIN 1	/* adjustment for margin of error */
			int uspt;	/* useable sect/trk (spt +1) */
			int pfs;	/* physical fault sector */
			int lfs;	/* logical fault sector */
			int askew;	/* actual skew */

			fmtp->dk_offset += MARGIN; /* push fault into sector */
			uspt = ibp->ib_spt + 1;	   /* slip sector is usable */
			pfs = fmtp->dk_offset  / ibp->ib_ubps;
			if( pfs >= uspt ) {
				/* bad offset or fault in runt - ignore */
				m323print(dev, "M323: Ignored, Flaw in Runt or Out of Range \n");
				printf(" Flaw data: cyl %d, head %d, offset %d\n",
				fmtp->dk_cyl, fmtp->dk_head, fmtp->dk_offset - MARGIN);
				continue;
			}
			
			/* NOTE: Interleave other than 1 not supported */
			/* calculate logical fault sector */

			/* adjust for spiral offset (skew) */
			askew = ibp->ib_skew ? uspt - ibp->ib_skew : 0;
			lfs = pfs - ((askew * fmtp->dk_head) % uspt);
			if( lfs < 0 )
				lfs += uspt; 	/* keep it on the track */
			if( lfs == uspt - 1 )	/* if flaw is in spare */
				continue;	/* no need to slip */
			
			fmtp->dk_sector = lfs;	/* use lfs in slip command */
			}
#ifdef DEBUG
	if(DebugM323 == 2) {
		printf(" Slip data: cyl %d, head %d, sector %d, skew %d, spt %d\n",
			fmtp->dk_cyl, fmtp->dk_head, fmtp->dk_sector,
			ibp->ib_skew, ibp->ib_spt );
	}
#endif
			bp->b_blkno = TRK2BLK(ibp,trk);
			bp->b_flags &= ~(B_DONE|B_FMTA|B_FMTS);
			bp->b_flags |= B_FMTS;
			if( m323opst(dev,bp) )
				return;		/* exit on error */
		}
		if( issig() ) {	
			u.u_error = EINTR;	/* exit on signal */
			return;
		}
	} 
}

/*
 *	m323svid: set volume id info
 */
int m323svid(dev,vidptr)
register dev;
register struct volumeid *vidptr;
{
	if( vidptr->vid_cas == CFGSS && vidptr->vid_cal == CFGLEN ) 
		return(OKAY);
	else
		return(ERRET);
}
/*
 *	m323gvid: get volume id info
 */
m323gvid(dev,vidptr)
register dev;
register struct volumeid *vidptr;
{
	vidptr->vid_cas = CFGSS;
	vidptr->vid_cal = CFGLEN;
}

/*
 *	m323scnfg: set configuration 
 */
int m323scnfg(dev,cfp)
register dev;
register struct dkconfig *cfp;
{
	register struct m323config cf;
	register struct iobuf *dp;

	if (cfp->ios_slccnt > M323MAXSLC)
		return(ERRET);
	dp = &m323utab[POS(dev)];
	cf.cf_vsh = cfp->ios_shn;
	cf.cf_vnh = cfp->ios_hds;
	cf.cf_spt = cfp->ios_spt;
	cf.cf_skew =  cfp->ios_sof;
	cf.cf_bps = cfp->ios_psm;
	cf.cf_gap1 = cfp->ios_gpb1;
	cf.cf_gap2 = cfp->ios_gpb2;
	cf.cf_intlv = cfp->ios_ilv;
	cf.cf_cyls = cfp->ios_trk;
#ifdef m88k
	cf.cf_attrib = (unsigned short)cfp->ios_catr_l;
#else
	cf.cf_attrib = (unsigned short)cfp->ios_catr;
#endif
	cf.cf_ubps = cfp->ios_uss;
	return (m323setcf(dev,dp,&cf));
	 
}
/*
 *	m323gcnfg: get configuration 
 */
m323gcnfg(dev,cfp)
register dev;
register struct dkconfig *cfp;
{
	register struct m323uib *ibp;
	register struct iobuf *dp;

	dp = &m323utab[POS(dev)];
	ibp = (struct m323uib *)dp->uib;
	cfp->ios_shn = ibp->ib_sh(VOL(dev));
  	cfp->ios_hds = ibp->ib_nh(VOL(dev));
	cfp->ios_spt = ibp->ib_spt;
 	cfp->ios_sof = ibp->ib_skew;
	cfp->ios_psm = ibp->ib_bps;
	cfp->ios_gpb1 = ibp->ib_gap1;
	cfp->ios_gpb2 = ibp->ib_gap2;
 	cfp->ios_ilv = ibp->ib_intlv;
 	cfp->ios_trk = ibp->ib_cyls;
#ifdef m88k
 	cfp->ios_catr_l = (unsigned short)ibp->ib_attrib;
 	cfp->ios_catr_u = (unsigned short)(ibp->ib_attrib >> 16);
#else
 	cfp->ios_catr = (unsigned long)ibp->ib_attrib;
#endif
	cfp->ios_uss = ibp->ib_ubps;
	cfp->ios_slccnt = dp->slccnt;	/* get actual slice count */
}	 



/*
 *	m323timer: periodically check on the state of the driver
 */

m323timer(cmd)
register cmd;
{
	static timeron;		/* timer on flag */
	register queued = 0;
	register ctl, pos;
	register struct iobuf *dp;
	register struct buf *bp;

#ifdef	DEBUG
	if( DebugM323 > 4 )
		m323print( 0, "timer");
#endif

	if( cmd == WSTART ) {
		if( timeron )
			return;
		timeron++;
	} else if( ! timeron )
			return;

	/* check each physical drive to insure no request is hung */
	for( pos=0, dp = &m323utab[0]; pos < MAXPOS; pos++, dp++ ) {

		ctl = pos/MAXDEV;
		if( dp->b_actf != NULL ) {
			queued++;
			if( m323ctlr[ctl].cc_actp == NULL ) { /* ctl active? */
#ifdef DEBUG
				printf("MVME323 controller %d not active but should be\n", ctl);
				printf("drive %d has I/O queued\n", pos%MAXDEV);
#endif
				m323start(ctl);
			}
		}

		if( dp->wtime == WOK )	/* nothing to do */
			continue;


		if( dp->wtime-- == WABORT ) {
			logberr(dp, B_ERROR);
			while (bp = dp->b_actf) {
				bp->b_flags |= B_ERROR;
				dp->b_actf = bp->av_forw;
				bp->cylin = 0;
				iodone(bp);
			}
			dp->b_active = 0;
			dp->b_errcnt = 0;
			dp->qcnt = 0;
			printf("MVME323 controller %d, drive %d not available\n",
					pos/MAXDEV, pos%MAXDEV );
		}
	}

	if( queued )	/* some I/O queued? */
		/* wake up every WSECS seconds */
		timeout( m323timer, WCONT, WSECS*HZ );
	else
		timeron = 0;
}


/*
 *	m323elog: performs error logging
 */

#ifdef DEBUG
	int m323ecnt;		/* (soft) error counter */
#endif

m323elog(rp, dp, blkoff, csr)
register struct device *rp;
register struct iobuf *dp;
int csr, blkoff;
{	
	register struct m323iopb *pbp;
	register struct m323eregs *erp;

	pbp = (struct m323iopb *)dp->cmdbuf;
	erp = (struct m323eregs *)dp->io_addr;

	erp->er_blkoff = blkoff;	/* block offset of error */
	/*erp->d0erstat = rp->drv0stat;	/* drive status */
	/*erp->d1erstat = rp->drv1stat;*/
	erp->d0erstat = drv0stat(rp);
	erp->d1erstat = drv1stat(rp);
	erp->d2erstat = drv2stat(rp);
	erp->d3erstat = drv3stat(rp);

	/* csr passed this way because it changes on interrupt clear */
	erp->er_csr = csr;		/* control status register */
	erp->er_cmd = pbp->pb_cmd;	/* command */
	erp->er_opt = pbp->pb_opt;	/* command option */
	erp->er_stat = pbp->pb_stat;	/* status code */
	erp->er_err = pbp->pb_err;	/* error code */
#ifdef DEBUG
	m323ecnt++;			/* error counter */
#endif
	/* get major number of block device for error logging */
	dp->b_dev = makedev( M323, minor(dp->b_dev));

	fmtberr( dp, 0 ); 		/* post to error log */
}



/*
 *	m323print: perform mvme323 driver standard format printing
 */

m323print(dev,str)
register dev;
char *str;
{	
	printf("%s on MVME323 ctl %d, drive %d, vol %d, slice %d\n", str, 
			CTL(dev), UNIT(dev), VOL(dev), SLICE(dev));
}



/*
 *	m323prall: display all command and status info
 *		for indicated controller
 */

m323prall(rp)
register struct device *rp;
{	
	register struct m323iopb *pbp;
	register done = FALSE;

	pbp = &rp->iopb1;

	while( ! done ) {			/* for each iopb */

		m323prpb(pbp);			/* display iopb */

		/* done if link bit not set or next iopb address is NULL */
		if( (pbp->pb_opt & LINK_IOPB) || 
#ifdef	m88k
			( (pbp = (struct m323iopb *)(pbp->pb_iopbphi << 16 + pbp->pb_iopbplo)) == NULL) ) 
#else
			( (pbp = pbp->pb_iopbp) == NULL) ) 
#endif
			done = TRUE;
	}

	m323prcsr(rp);		/* display control and status registers */
}

/*
 *	m323prpb: display command and status info
 *		for iopb pointed to by pbp argument
 */

m323prpb(pbp)
register struct m323iopb *pbp;
{	
	register unsigned char c, s, e;
	int goodstat;

	/* command */
	printf(" CMD:     ");
	switch( c = pbp->pb_cmd ) {
	case DIAGNOS:	printf(" Execute power up diagnostics ");	break;
	case READ_LNG:	printf(" Read data & ECC for diagnostics ");	break;
	case WRIT_LNG:	printf(" Write data & ECC for diagnostics ");	break;
	case READ_HDR:	printf(" Read random header on current cyl ");	break;
	case GET_CNFG:	printf(" Return current drive params (UIB) ");	break;
	case WRIT_SBF:	printf(" Write diagnostic sector to ctlr ");	break;
	case READ_SBF:	printf(" Read sector written with WRIT_SBF ");	break;
	case READ_SEC:	printf(" Read sector(s) ");			break;
	case WRIT_SEC:	printf(" Write sector(s) ");			break;
	case VRFY_SEC:	printf(" Verify sector(s) ");			break;
	case FMT_TRK:	printf(" Format track ");			break;
	case MAP_TRAK:	printf(" Format out bad track ");		break;
	case HNDSHAKE:	printf(" Handshake controller ");		break;
	case INIT_DRV:	printf(" Initialize drive params ");		break;
	case RESTORE: 	printf(" Attempt to clear drive fault(s) ");	break;
	case SEEK: 	printf(" Seek to specified position ");		break;
	case REFORMAT:	printf(" Format trk saving bad spot mapping ");	break;
	case FMT_TRKD:	printf(" Format track with data ");		break;
	case MAP_SECT:	printf(" Format out bad sector on track ");	break;
	case READ_NOC:	printf(" Read sectors no cache ");		break;
	case VRFY_TRK:	printf(" Verify track ");			break;
	case TRACK_ID:	printf(" Return track header info ");		break;
	case FETCH_EX:	printf(" Fetch IOPB and execute ");		break;
	case SCATTER:	printf(" Scatter read data");			break;
	case GATHER:	printf(" Gather write data");			break;
	case READ_FLAW: printf(" Read ESDI flaw map");			break;
	default:	printf(" Unrecognized command 0x%x ", c );
			break;
	}

	/* packet status */
	goodstat = TRUE;
	switch( s = pbp->pb_stat ) {
	case DONE_OK:	printf(" --  completed with no error ");	break;
	case BUSY_OK:	printf(" --  in progress ");			break;
	case DONE_ERR:	printf(" --  completed with error ");		break;
	case DONE_EXC:	printf(" --  completed with exception ");	break;
	case UNSTARTED:	printf(" --  unstarted ");			break;
	default:	printf(" Unrecognized status 0x%x ", s );
			goodstat = FALSE;
			break;
	}	
	printf("\n");

	/* error codes */
	if( s == DONE_ERR ) {
		printf(" ERROR:   ");
		switch( e = pbp->pb_err ) {
		case DRV_NRDY:	printf(" Drive not ready ");			break;
		case SEEK_CYL:	printf(" Seek error - wrong cylinder ");	break;
		case ECC_CODE:	printf(" ECC code discrpancy ");		break;
		case BAD_CMD:	printf(" Invalid command code (byte 0 of IOPB) ");	break;
		case BAD_FEX:	printf(" Illegal fetch and execute attempted ");	break;
		case BIG_SNUM:	printf(" Sector number too large for drive ");	break;
		case MEM_TYPE:	printf(" Illegal memory type specified ");	break;
		case BUS_TOUT:	printf(" Bus not available in < 1 ms ");	break;
		case HDR_CHKS:	printf(" Header checksum error ");		break;
		case WRT_PERR:	printf(" Disk write protected ");		break;
		case NOT_SEL:	printf(" No response from unit on select ");	break;
		case SERR_TOUT:	printf(" No correct seek 3 secs after RESTORE ");	break;
		case FLT_TOUT:	printf(" Fault not cleared 3 secs after RESTORE ");	break;
		case DRV_FLT:	printf(" Drive fault occured, RESTORE required ");	break;
		case RDY_TOUT:	printf(" Drive not ready 3 secs after RESTORE ");	break;
		case PAST_END:	printf(" Multi sector I/O exceeded end of media ");	break;
		case UIB_ZERO:	printf(" UIB for specified volume contains zeros ");	break;
		case HDR_PAD:	printf(" Bad post header pad byte found ");	break;
		case FIX_FAIL:	printf(" Failed to fix uncorrectable error ");	break;
		case BAD_CYLX:	printf(" Logical sector num -> bad cylinder num ");	break;
		case BAD_HDX:	printf(" Logical sector num -> bad head num ");	break;
		case BAD_SECX:	printf(" Logical sector num -> bad phys sector num ");	break;
		case OVERRUN:	printf(" Data overrun - suspect clock from drive ");	break;
		case NO_INDEX:	printf(" No index pulse after 65ms on format cmd ");	break;
		case SEC_NFND:	printf(" Sector not found during read or write ");	break;
		case BAD_HDNO:	printf(" Head number in header wrong ");	break;
		case DATA_SYNC:	printf(" Sync word in data field not valid ");	break;
		case SEEK_TOUT:	printf(" Seek not complete in 500ms ");	break;
		case BUSY_TOUT:	printf(" Drive held by dual port over 500ms ");	break;
		case OFF_CYL:	printf(" Not on cylinder 3 secs after select ");	break;
		case RTZ_TOUT:	printf(" Not normal complete 3 secs after RESTORE ");	break;
		case HDR_SYNC:	printf(" Sync word in header not valid ");	break;
		case NOT_INIT:	printf(" Write or format attempted before INIT_DRV ");	break;
		case SMALL_GAP:	printf(" Specified gap too small - minimun gap is 5 ");	break;
		case SEEK_DRV:	printf(" Seek error reported by drive ");	break;
		case MAP_NOP:	printf(" No sector pulse on track to be mapped ");	break;
		case BAD_SPT:	printf(" Sectors per track in UIB invalid ");	break;
		case BIG_SEC:	printf(" Bytes per sector in UIB > buffer size ");	break;
		case BAD_ILV:	printf(" Invalid interleave in UIB > SPT or 0 ");	break;
		case HEAD_OR:	printf(" Head number in IOPB out of range per UIB ");	break;
		case CYL_OR:	printf(" Cylinder no. in IOPB out of range per UIB ");	break;
		case IOPB_FAIL:	printf(" Bus error on external IOPB transfer ");	break;
		case DMA_FAIL:	printf(" Bus error on DMA transfer ");	break;
		case ALLIGN:	printf(" VME buffer address not alligned ");	break;
		case NOSPARE:	printf(" Spare sector already in use ");	break;
		case NOSPARES:	printf(" UIB indicates no spares enabled ");	break;
		case NOT_YET:	printf(" Command not implemented ");		break;
		default:	printf(" Unexpected error code 0x%x ", e );	break;
		}	
		printf("\n");

		/*	Call Motorola Corp for error codes f0-fe. */
		if(e > 0xf0  &&  e < 0xff)
			printf("Call manufacturer with IOPB and UIB info.\n");
	}

	/* exception codes */
	if( s == DONE_EXC ) {

		printf(" EXCEPTION:");
		e = pbp->pb_err;

		if( e &= RTRYMSK ) {		/* if retries used */
			printf(" %d", e);
			if( e == RTRYMSK ) printf(" or more");
			printf(" retries ");
		}
		prbits( pbp->pb_err, BYTEMSB, EXCPMSK, excode );
		printf("\n");
	}

	/* command options */
	{	register unsigned char co;

		co = pbp->pb_opt;
		printf(" CMD_OPTS: Unit %d  Vol %d", COPTOU(co), COPTOV(co));
		prbits( co, BYTEMSB, COPMSK, copdis );
		printf("\n");
	}

	/* if status is recognized display details*/
	if( goodstat ) {
		/* REFORMAT is the same but would also return a fmt_pattern */
		if(  (c == FMT_TRK) || (c == MAP_TRAK) || (c == MAP_SECT) )  {
			printf("  Cylinder 0x%x  head 0x%x   sector 0x%x \n",
				pbp->pb_cyl, pbp->pb_trk, pbp->pb_psect);
#ifdef	m88k
			printf("  Sectors formatted  0x%x \n", pbp->pb_un1.pb_typ2.pb_scnt);
#else
			printf("  Sectors formatted  0x%x \n", pbp->pb_scnt);
#endif
		}
		else {
#ifdef	m88k
			printf("          Logical_sector 0x%x \n", pbp->pb_lsect.hi <<16 + pbp->pb_lsect.lo);
#else
			printf("          Logical_sector 0x%x \n", pbp->pb_lsect);
#endif

#ifdef	m88k
			printf("          Sector_count 0x%x  buffer address 0x%x\n",
				pbp->pb_un1.pb_typ2.pb_scnt, pbp->pb_badr.hi <<16 + pbp->pb_badr.lo);
#else
			printf("          Sector_count 0x%x  buffer address 0x%x\n",
				pbp->pb_scnt, pbp->pb_badr);
#endif
		}

#ifdef	m88k
#else
		if( pbp->pb_iopbp )
			printf("  Next IOPB address 0x%x\n", pbp->pb_iopbp);
#endif

	}

	/* if uib used */
	if( c == INIT_DRV || c == GET_CNFG )
#ifdef	m88k
		m323pruib((struct m323uib *)(pbp->pb_badr.hi << 16 + pbp->pb_badr.lo)); /* display uib */
#else
		m323pruib((struct m323uib *)pbp->pb_badr); /* display uib */
#endif

}

/*
 *	m323pruib: display unit initialization block info
 *		for uib pointed to by ibp argument
 */

m323pruib(ibp)
register struct m323uib *ibp;
{	
	printf(" UIB:");
	printf("  Vol_0 starting_head 0x%x  head_count 0x%x\n",
		ibp->ib_sh(0), ibp->ib_nh(0));
	printf("  Vol_1 starting_head 0x%x  head_count 0x%x\n",
		ibp->ib_sh(1), ibp->ib_nh(1));
	printf("  Sectors_per_trk 0x%x  spiral_skew 0x%x  bytes_per_sect 0x%x\n",
		ibp->ib_spt, ibp->ib_skew, ibp->ib_bps);
	printf("  Gap1 0x%x  gap2 0x%x  sector_interleave 0x%x\n",
		ibp->ib_gap1, ibp->ib_gap2, ibp->ib_intlv);
	printf("  Retry_cnt %d  cylinders %d\n",
		ibp->ib_retry, ibp->ib_cyls);
	prbits( ibp->ib_attrib, BYTEMSB, ATRMSK, atrdis );
	printf("  Reserved 0x%x  status_chg_int_lvl 0x%x  vector 0x%x\n",
		ibp->ib_r0, ibp->ib_silev, ibp->ib_sivec);
}

/*
 *	m323prcsr: display control and status register 
 *		and drive status info for indicated controller
 */

m323prcsr(rp)
register struct device *rp;
{	
	register unsigned csr,d;
	register i;

	csr = rp->csr;
	printf(" CSR:     ");
	prbits( csr, WORDMSB, CSRMSK, csrdis );
	if(csr & SCHG_INT) {
		printf(" stat_chg_drv_");
		if(csr & SCHG_SRC)
			printf("1");
		else
			printf("0");
		}
	printf("\n");

	/* if drive status is non-zero, display it */
	for( i=1; i>=0; --i) {
		if( d = rp->ds[i] ) {
			printf(" DRIVE_%d: ",1-i);
			prbits( d, BYTEMSB, DRVSMSK, dstatdis );
			printf("\n");
		}
	}

#ifdef DEBUG
	/* csr dump */
	if( DebugM323 > 5 ) {
		printf(" CSR_DUMP: d0_stat=0x%x  d1_stat=0x%x d2_stat=0x%x  d3_stat=0x%x  csr=0x%x\n",
			drv0stat(rp), drv1stat(rp), drv2stat(rp), drv3stat(rp), csr);
	}
#endif
}

/*
 *	prbits: prints strings from msgs[] array
 *		for each bit set in both reg and mask
 *		starting at most significant bit msb
 */

STATIC prbits( reg, msb, mask, msgs )
register unsigned int reg, msb, mask;
register char *msgs[];
{
	register i,m;

	for(i=msb, m=1<<msb; i>=0; --i, m>>=1) {
		if((m&mask)&reg)
			printf("%s", msgs[i]);
	}
}



/*
 *	m323clr: clear the driver after powerfail
 */

m323clr()

{	
	/* no hardware to support this function... no warm start */
		/* reboot required after power fail */
}



/*
 *	mvme323 dump support.
 */

m323dump()
{	
}
