/*		Copyright (c) 1988 Motorola, Inc.	*/
/*		All Rights Reserved			*/
/*		@(#)mvme327dk.c	35.15 */
/****************************
 *			    *
 *	mvme327 Driver:     *
 *     			    *
 ***************************/

/*******************************************************************************
 *
 *	Include files
 *
 ******************************************************************************/

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/buf.h"
#include "sys/elog.h"
#include "sys/iobuf.h"
#include "sys/systm.h"
#include "sys/dk.h"
#include "sys/dsk.h"
#include "sys/sobpp.h"
#include "sys/mvme327.h"
#include "sys/open.h"


#define	DEBUG	
#undef DEBUG
#define	DEBUGC
#undef DEBUGC
	

#ifndef	M327
#define	M327	9
#endif


#ifdef	DEBUG		/* debugging code */
#define STATIC
#define	DEBUGPR(x,d,s)	{ if (x & d327Debug) d327print(d,s); }

int d327Debug = 0xffff;

#define	DBINIT		(1<<0)
#define	DBBOPEN		(1<<1)
#define	DBCOPEN		(1<<2)
#define	DBBCLOSE	(1<<3)
#define	DBCCLOSE	(1<<4)
#define	DBSTRAT		(1<<5)
#define	DBSTART		(1<<6)
#define	DBCONF		(1<<7)
#define	DBFORMAT	(1<<8)
#define	DBICMD		(1<<9)
#define	DBINTR		(1<<10)
#define	DBERR		(1<<11)
#define	DBREAD		(1<<12)
#define	DBWRITE		(1<<13)
#define	DBIOCTL		(1<<14)
#define	DBCLR		(1<<15)
#define	DBTIMER		(1<<16)
#define	DBDUMP		(1<<17)
#define	DBDOIO		(1<<18)

static char *ASSERTCTL = "ASSERT(%s) failed in %s, line %d\n";
#define ASSERT(X) { if (!(X)) printf(ASSERTCTL, "X", __FILE__, __LINE__);}

#else			/* no debugging code */

#define	STATIC	static
#define	DEBUGPR(x,d,s)
#define ASSERT(X)

#endif

#ifdef	DEBUGC		/* command debugging code */

#define	PRCMD(p)	{ if (d327Cmdbug) d327prcmd(p); }
int d327Cmdbug = 1;
#else
#define	PRCMD(p)	/* no command debugging code */
#endif

/*******************************************************************************
 *
 *	This driver will support up 2 MVME327A or 1 MVME327B.
 *	Each MVME327A can support up to 7 SCSI controllers.
 *	Each MVME327B can support up to 15 SCSI controllers.
 *	Each SCSI contoller can control up to 8 drives simultaneously:
 *
 ******************************************************************************
 *
 *	MVME327 Disk Controller Driver Major/Minor Number Assignment Coding
 *
 *               Major                 Minor
 *      <-  4  -> <-1-> <- 3 -> <- 3 -> <-1-> <-  4  ->
 *      ----------------------- -------------------------
 *      |        | 327 | SCSI  | perph |     |          |
 *      |        | BUS | addr  | drive | flop|   slice  |
 *      ----------------------- -------------------------
 *
 *
 *	327 BUS = clear if 0 , set if 1 
 *	SCSI addr = logical address of SCSI controller
 *	perph drive = logical drive number
 *	flop = floppy bit
 *


	The major/minor number is used to access a 
table built at configuration time (see m327space.h).
This table contains the default configuration paramaters for
each SCSI drive.

There will be 16 major numbers with 8 devices per major number.
Major number 0 will be defined as a disk, major number 4 will be
defined as tape...

The floppy will be determined by a bit set in the minor number.

These are the default devices:

0	disk
1	disk
2	disk
3	disk
4	tape
5	tape
6	unassigned
7	floppy

 ******************************************************************************/


/*******************************************************************************
 *
 *	Major/Minor Number Manipulating Macros
 *
 ******************************************************************************/

#define	SLCMSK		((1<<M327SBTS)-1)
#define	SLICE(d)	(d&SLCMSK)		/* slice number */
#define	MKDEV(b,c,p)	((b<<11)|(c<<8)|(p<<5)) /* bus|scsiadr|perf -> minor */
#define SCBUS(d)	((d & 0x0800) >> 11)
#define SCADR(d)	((d & 0x0700) >> 8)
#define SCDEV(d)	((d & 0x00e0) >> 5)
/* MKPUNIT takes scsi ctl, scsi perf nibbles and makes punit */
#define MKPUNIT(d)	(((d & 0x0700) >> 4) | ((d & 0x00e0) >> 5))
#define	POS(d)	((SCBUS(d)*m327maxscadr*m327maxscdev) + (SCADR(d)*m327maxscdev) + SCDEV(d))
#define CTL(d)		((m327a) ? SCBUS(d) : 0)


/*******************************************************************************
 *
 *	MVME327 Local Floppy - 2 per MVME327
 *
 ******************************************************************************/


#define FLOPS		2
#define	FPOS(d)		((SCBUS(d)*FLOPS) + SCDEV(d)) /* floppy position */
#define FLOPBIT		0x10
#define FLOP327(x)	(x & FLOPBIT)  
#define HIGHSLICE(x)	(SLICE(x) == 3 || SLICE(x) == 4)  /* high speed flop */


/*******************************************************************************
 *	MVME327 Winchester & Floppy Default Configuration Values
 ******************************************************************************/

/* Winchester Values */

#define W_ECCFLG	1		/* ECC flag; ECC in use */
#define W_INITCHR	0x4e		/* initialization character */
#define W_RETRY		10		/* number of retrys */
#define W_CACHE		(1<<23)		/* cache entry bit 24 = 0 is cache on */
#define W_ZONE		(1<<11)		/* zone type bit 12 = 0 is track */



/*******************************************************************************
 *
 *	Global variables
 *
 ******************************************************************************/

#define	MAXPOS	(d327_cnt*m327maxtab)	/* maximum number of disk slots */
#define	MAXFPOS	(d327_cnt*FLOPS)	/* maximum number of floppy slots */

#define	D327SHRT	30		/* timeout limit for short commands */
#define	D327LONG	3600		/* timeout limit for long commands */
#define	F327SHRT	100		/* flop timeout limit short commands */
#define	F327LONG	1200		/* flop timeout limit long commands */


STATIC	struct dsk d327dsk;		/* disk descriptor for the mvme327 */

STATIC	struct dsk f327dsk;		/* floppy descriptor for the mvme327 */

#define SCAT				/* enable scatter/gather code */

#ifdef SCAT
/* coalescing enable */
extern int m327coal;

/* coalescing count print */
int m327scatcnt = 65;			
#endif



/*******************************************************************************
 *	External Variables
 ******************************************************************************/

extern struct sb_device *d327_addr[];
extern int d327_cnt;			/* count */
extern int d327_ilev[];			/* interrupt level */
extern int d327_ivec[];			/* interrupt level */

extern int d327otbl[];			/* open table */
extern int d327fotbl[];			/* floppy open table */

extern int m327a;			/* MVME327A controller */

/*******************************************************************************
 *	Driver Variables - declared in conf.c (space.h or io.h)
 ******************************************************************************/

extern struct iobuf		m327utab[];	/* one per drive */
extern struct iobuf		m327futab[];	/* one per floppy drive */
extern struct iotime		m327stat[];	/* one per drive */
extern struct iotime		m327fstat[];	/* one per floppy drive */
extern struct slice		d327slices[];	/* slice tables */
extern struct slice		d327fslices[];	/* slice tables */
extern struct sb_packet		m327pkt[];	/* command packets */
extern struct sb_envelope 	m327env[];	/* envelopes */
extern struct sb_envelope	*m327freeq;	/* list of free packets */
extern struct d_disk	 	*m327params[];	/* disk descriptor params */
extern struct d_disk	 	*f327params[];	/* disk descriptor params */
extern struct d_disk	 	m327fdesc[];	/* floppy descriptor params */
extern struct d_disk	 	m327desc[];	/* descriptor params */
extern struct sb_channel 	d327chan[];	/* channel for MVME327 communication */
extern int m327maxscbus;		/* max number of configured buses */
extern int m327maxscadr;		/* max number of configured addrs */
extern int m327maxscdev;		/* max number of configured devs */
extern int m327maxtab;			/* max table size */
extern int m327maxcoal;			/* max number of coalesced buffers */
#ifdef SCAT
extern struct sb_scatlist 	m327scatlist[];	/* scatter/gather list */
#endif

/*******************************************************************************
 *	MVME327 iobuf usage
 *
 *	iobuf.h		mvme327.c	Usage
 *	-------		---------	-----
 *	b_flags		b_flags		Drive state
 *	b_forw
 *	b_back
 *	b_actf		b_actf		head (first) of I/O queue
 *	b_actl		b_actl		tail (last) of I/O queue
 *	b_dev		b_dev		logical major/minor device number
 *	b_active	b_active	busy flag
 *	b_errcnt	b_errcnt	error count
 *	io_erec		io_erec		error record
 *	io_nreg		io_nreg		number of registers to log upon error
 *	io_addr		io_eregs	address of registers to log on error
 *	io_mba		io_csr		address of device CSR
 *	io_stp		io_stp		I/O statistics
 *	io_start	io_start	I/O statistics
 *	io_s1		io_acts		used for sorting i/o queue      (dsk.c)
 *	io_s2		io_qcnt		number of bufs on iobuf queue	(dsk.h)
 *	io_s3		io_slccnt	number of allowable slices	(dsk.h)
 *	io_s4		io_badptr	pointer to bad block table	(dsk.h)
 *	io_s5		io_desc		disk descriptor (configuration struct)
 *	io_s6		io_pktcnt	packet count (limit & count)	(dsk.h)
 *	io_s7		io_sopen	open slices flag		(dsk.h)
 *	io_s8		io_slctbl	pointer to the slice table	(dsk.h)
 *	io_s9		io_dskptr	pointer to dsk structure	(dsk.h)
 ******************************************************************************/

#define	io_eregs	io_addr
#define	io_csr		io_mba
#define	io_desc		io_s5


/*
 *	d327init: initialize driver
 */

d327init()
{
	register unsigned char sc_bus, sc_adr, sc_dev;
	register n;
	register struct sb_envelope *ep;
	register struct d_disk **dd = &m327params[0];
	register struct sb_packet *pp;
	register struct sb_device *rp;
	struct slice *slc = &d327slices[0];
	int d327strategy();
	int d327start();
	int d327timer();
	int f327timer();
	int d327print();

	DEBUGPR(DBINIT,0,"init");



	/* set up free queue of envelope/packet pairs */

	/* create the queue  if not already made */
	if(m327freeq == (struct sb_envelope *)NULL)
		sobppcrque(&m327freeq,m327maxscbus*M327PKTS,m327pkt,
			m327env,DATA32,BUSW);

	for( n=0; n < d327_cnt; n++ ) { /* check if board is there */
		if( bprobe( d327_addr[n], -1 ) ) {
			d327_addr[n] = NULL;
			continue;
		}
	}

	/* if MVME327a then there are d327_cnt number of busses */
	n = (m327a) ? d327_cnt : m327maxscbus;

	for( sc_bus = 0; sc_bus < n; sc_bus++ ) {
		if((rp = d327_addr[sc_bus]) == NULL)
			continue;
		for( sc_adr = 0; sc_adr < m327maxscadr; sc_adr++ ) {
			for(sc_dev = 0; sc_dev < m327maxscdev; dd++,sc_dev++) {
		
				/* Tape intialized in mvme327tp.c */

				if ((*dd)->d_ptype == M327DISK) {
					d327dinit(MKDEV(sc_bus,sc_adr,sc_dev),rp
						,*dd,slc);
					slc += M327SLCS;
				}
			}
		}
		/* convert vector address to vector number */
		d327_ivec[sc_bus] /= 4;
		if (sobppcrchan(&d327chan[sc_bus],rp,&m327freeq,
			d327_ilev[sc_bus], d327_ivec[sc_bus],DATA32) == -1) {
			/* create channel error */
			d327_addr[sc_bus] = NULL;
			printf("MVME327 Create channel error\n");
			printf("MVME327 Channel %x disabled\n", sc_bus);
			continue;
		}
	}

	/* initialize local floppys */

	slc = &d327fslices[0];
	for( n=0; n < d327_cnt * FLOPS; dd++,n++ ) {
		if((rp = d327_addr[n/FLOPS]) == NULL)
			continue;
		d327finit(n,rp,&m327fdesc[n],slc);
		slc += M327SLCS;
	}

	/* initialize floppy dsk structure */
	f327dsk.d_flags = D_DFL_CONF|D_FMTU;
	f327dsk.d_strategy = d327strategy;
	f327dsk.d_start = d327start;
	f327dsk.d_timer = f327timer;
	f327dsk.d_print = d327print;
	f327dsk.d_shorttime = F327SHRT * HZ;
	f327dsk.d_longtime = F327LONG * HZ;
	f327dsk.d_slcbits = M327SBTS;

	/* initialize dsk structure */
 	d327dsk.d_flags = D_MOTTST|D_DFL_ERR|D_CONFIG|D_FMTU|D_NOVID;
	d327dsk.d_strategy = d327strategy;
	d327dsk.d_start = d327start;
	d327dsk.d_timer = d327timer;
	d327dsk.d_print = d327print;
	d327dsk.d_shorttime = D327SHRT * HZ;
	d327dsk.d_longtime = D327LONG * HZ;
	d327dsk.d_slcbits = M327SBTS;


	dskinit(&d327dsk);
	dskinit(&f327dsk);
}


/*
 *	initialize disks
 */

STATIC d327dinit(dev,rp,dd,slctbl)
register struct sb_device *rp;
register struct d_disk *dd;
register unsigned short dev;
{
	register struct iobuf *dp;
	register struct pktcnt *pcp;
	register struct d_disk *ddd;
	register unsigned short i;

	/* if the board isn't there, return */
	if (d327_addr[CTL(dev)] == NULL)
		return;

	i = POS(dev);
	dp = &m327utab[i];
	/* copy default parameters */
	ddd = &m327desc[i];
	*ddd = *dd;

	dp->b_dev = dev;
	dp->io_csr = (physadr)rp;
	dp->io_nreg = M327NREGS; /* length of eregs */
	pcp = (struct pktcnt *) &dp->io_pktcnt;
	pcp->cnt = 0;
	pcp->limit = M327LIM; 
	dp->io_desc = (int)ddd;
	dp->io_slctbl = slctbl;
	dp->io_dskptr = (int)&d327dsk;
	if( major(rootdev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}

	if( major(swapdev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}

	if( major(pipedev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}
}


/*
 *	initialize floppys
 */

STATIC d327finit(dev,rp,dd,slctbl)
register struct sb_device *rp;
register struct d_disk *dd;
register unsigned short dev;
{
	register struct iobuf *dp;
	register struct pktcnt *pcp;
	register unsigned short rdev;

	dp = &m327futab[dev];
	dp->b_flags |= B_REINIT;
	rdev = MKDEV(dev/FLOPS,0,(dev%FLOPS)) | FLOPBIT;
	dp->b_dev = rdev;
	dp->io_csr = (physadr)rp;
	dp->io_nreg = M327NREGS; /* length of eregs */
	pcp = (struct pktcnt *) &dp->io_pktcnt;
	pcp->cnt = 0;
	pcp->limit = M327LIM; 
	dp->io_desc = (int)dd;
	dp->io_slctbl = slctbl;
	dp->io_dskptr = (int)&f327dsk;
	if( major(rootdev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}

	if( major(swapdev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}

	if( major(pipedev) == major(dev) ) {
		dp->b_flags |= B_SPECL;
	}
}

/*
 *	d327bopen: perform open on a block device
 */

d327bopen(dev, modes, otyp)
register otyp;
register dev;
register modes;
{

	DEBUGPR(DBBOPEN,dev,"bopen");

	d327open(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d327copen: perform open on a character device
 */

d327copen(dev, modes, otyp)
register otyp;
register modes;
register dev;
{

	DEBUGPR(DBCOPEN,dev,"copen");

	d327open(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d327open: perform open for all devices
 */

STATIC d327open(dev,modes,omask,otyp)
register dev;
register otyp;
register modes;
register omask;
{
	register struct iobuf *dp;
	register *otbl;
	register i;

	/*
	 *	if the device is nonexistent, then error
	 */
	if((d327_addr[CTL(dev)] == NULL) || (SCADR(dev) > (m327maxscadr-1)) ||
		 (!(FLOP327(dev)) && SCDEV(dev) > (m327maxscdev-1) ))
	{
		u.u_error = ENXIO;
		return;
	}
	if(FLOP327(dev)) {
		i = FPOS(dev);
		dp = &m327futab[i];
		otbl = &d327fotbl[i];
		if(dp->io_dskptr != (int)&f327dsk) {  /* no dev */
			u.u_error = ENXIO;
			return;
		}
		/* check for high speed slice, copy parameters */
		*(struct d_disk *)dp->io_desc = *f327params[HIGHSLICE(dev)];
	} else {
		i = POS(dev);
		dp = &m327utab[i];
		otbl = &d327otbl[i];
		if(dp->io_dskptr != (int)&d327dsk) {  /* no dev */
			u.u_error = ENXIO;
			return;
		}
	}

	dskopen(dp, modes, otbl, otyp, omask);
}

/*
 *	d327bclose: perform close on a block device
 */

d327bclose(dev, modes, otyp)
register otyp;
register dev;
register modes;
{

	DEBUGPR(DBBCLOSE,dev,"bclose");

	d327close(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d327cclose: perform close on a character device
 */

d327cclose(dev, modes, otyp)
register otyp;
register dev;
register modes;
{

	DEBUGPR(DBCCLOSE,dev,"cclose");

	d327close(dev, modes, 1<<SLICE(dev), otyp);
}

/*
 *	d327close: perform close for all devices
 */

d327close(dev, modes, mask, otyp)
register dev;
register otyp;
register modes;
register mask;
{
	register *otbl;
	register struct iobuf *dp;
	register i;

	if(FLOP327(dev)) {
		i = FPOS(dev);
		dp = &m327futab[i];
		otbl = &d327fotbl[i];
	} else {
		i = POS(dev);
		dp = &m327utab[i];
		otbl = &d327otbl[i];
	}
	dskclose(dp, modes, otbl, otyp, mask);

}

/*
 *	d327strategy: perform the disk I/O strategy algortithm
 */

d327strategy(bp)
register struct buf *bp;
{
	register struct iobuf *dp;
	register struct d_disk *wd;
	register struct slice *slc;
	register unsigned short dev = bp->b_dev;
#ifdef SCAT
	register int coalesce = m327coal;
#endif

	DEBUGPR(DBSTRAT,dev,"strategy");

	if(FLOP327(dev))
		dp = &m327futab[FPOS(dev)];
	else
		dp = &m327utab[POS(dev)];

	/*
	 * Do system accounting calculations
	 */
	if ((bp->b_flags&B_ABS) == 0) {
		/* save cylinder number in b_resid for sadp(1) */
		slc = (struct slice *)dp->io_slctbl + SLICE(bp->b_dev);
		bp->b_resid = bp->b_blkno;	/* get block number */
		bp->b_resid += slc->blkoff;	/* add slice start */

		wd = (struct d_disk *)dp->io_desc;
	/*
	 *	Calculate cylinder number for profiler and strategy
	 */
		bp->b_resid *= (BSIZE / wd->d_bps);
		bp->b_resid /= (wd->d_spt * wd->d_heads);
	}
#ifdef SCAT
	/* call the disk discipline strategy (type 2) routine */

	if(FLOP327(bp->b_dev))
		coalesce = 0;
	dsk2strategy(bp, dp, coalesce,m327maxcoal);
#else
	/* call the disk discipline strategy (type 1) routine */
	dsk1strategy(bp, dp);
#endif
}

/*
 *	d327start: start the I/O
 */

d327start(dp)
register struct iobuf *dp;
{
	register struct buf *bp;
	register struct sb_envelope *ep;
	register struct pktcnt *pcp;
	register struct sb_packet *pp;
	struct sb_envelope *sobppget();
	register unsigned short seetooth = 0;
	register s;
	register unsigned short i;

	DEBUGPR(DBSTART,dp->b_dev,"start");

	s = spl6();

	pcp = (struct pktcnt *) &dp->io_pktcnt;	

	for(bp=(struct buf *)dp->b_actf,i=0;bp;bp=bp->av_forw,i++){
		
		/*
		 *	This "for" loop is intended for use when
		 *	multiple packets are being sent.
		 *	However with coalesing, only one packet 
		 * 	is being sent to the controller and you
		 *	need this check to see that the packet
		 * 	just sent didn't get serviced.
		 *	If it did you can end up on the free
		 *	list.
		 */

		if(dp->b_actf == NULL)
			break;

		/*
		 * Determine how many packets have been sent.  If the
		 * limit has been reached don't send any more now.
		 */
		if (pcp->cnt >= pcp->limit) {
			break;		/* too many outstanding requests */
		}

		if(bp == (struct buf *)dp->acts) 	/* buffer at sawtooth */
			seetooth = 1;

		if( i < pcp->cnt )  	/* already sent */
			continue;

		/*
		 * Get a command packet.
		 */

		if ((ep = sobppget(&m327freeq)) == NULL) {
			break;
		}
		if(seetooth)		/* change sawtooth to this one */
			dp->acts = (int)bp;

		pcp->cnt++;		/* one more packet sent */
		blkacty |= (1<<M327);	/* indicate mvme327 activity */
		splx(s);

		/*
		 * Set up packet
		 */

		pp = ep->e_pkt;
		bzero(&pp->p_paddr,0x24); /* clear packet */
		/* high order nib is scsi addr,low order unit */
		if(!FLOP327(bp->b_dev)) {
			pp->p_punit = MKPUNIT(bp->b_dev);
			if((!m327a) && SCBUS(bp->b_dev)) /* 2nd SCSI bus */
				pp->p_pdtype = SD_SCSI2;
			else
				pp->p_pdtype = SD_SCSI1;

		} else { /* not on SCSI bus,must have own unit and type */
			pp->p_punit = SCDEV(bp->b_dev);
			pp->p_pdtype = SD_FLOP;
		}
		/*
		 * Process the command
		 */
		if (bp->b_flags&B_CONF)			/* config request */
			d327conf(dp,bp,ep);
		else if (bp->b_flags&(B_FORMAT|B_FMTU|B_FIXBS))		/* format */
			d327format(dp,bp,ep);
		else {					/* read/write request */
			d327rdwr(dp,bp,ep);
		}
		s = spl6();
	}
	splx(s);
}


/*
 *	d327rdwr: perform read/write request 
 */

STATIC d327rdwr(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct sb_envelope *ep;
{
	register s;
	register unsigned short i;
	register struct slice *slc;
	register struct sb_packet *pp;
	register struct d_disk *wd = (struct d_disk *)dp->io_desc;

	pp = ep->e_pkt;
	pp->p_cmd = (bp->b_flags&B_READ) ? SC_READ : SC_WRITE;
	slc = (struct slice *)dp->io_slctbl + SLICE(bp->b_dev);
	pp->p_paddr = bp->b_blkno;

	/* don't add slice start block for absolute requests */

	if ((bp->b_flags&B_ABS) == 0)
		pp->p_paddr += slc->blkoff;	/* slice beg. */
	pp->p_saddr = (unsigned long)bp->b_un.b_addr;
	pp->p_tcount = bp->b_bcount/wd->d_bsize;
	/* save the pointer to the utab */
	pp->p_utab = (int)dp;

#ifdef DEBUG
	/* save the packet number in the buf struct */
	bp->b_pkt = (int)pp;
#endif
#ifdef SCAT

	s = spl6();
	if((bp->av_back) && (!FLOP327(bp->b_dev))) {
		register unsigned short count;
		register struct buf *nbp;
		register struct sb_scatlist *list =
			&m327scatlist[(POS(bp->b_dev)*m327maxcoal)];

		pp->p_saddr = (unsigned)list;
		for(nbp = bp, i=0, count = 0; nbp;
		    (nbp = nbp->av_back), i++, list++){
			nbp->b_start = lbolt;
			list->address = (unsigned char *)nbp->b_un.b_addr;
			list->count = nbp->b_bcount;
			count += nbp->b_bcount;
		}
		pp->p_tcount = count/wd->d_bsize;
		pp->p_sgcount = i;	/* scatter/gather count	*/

#ifdef DEBUG
		if(i > m327scatcnt){
		    if(bp->b_flags&B_READ)
			printf("m327rw: combined %d read requests\n",i);
		    else
			printf("m327rw: combined %d write requests\n",i);
		}
#endif
	} else
		pp->p_sgcount = 0;	/* scatter/gather count	*/
	splx(s);

#endif
	/*
	 * Reset the time in b_start for safety timer purposes.
	 */
	{	
		register struct iotime *ip;
		if(FLOP327(bp->b_dev))
			ip = &m327fstat[FPOS(bp->b_dev)];
		else
			ip = &m327stat[POS(bp->b_dev)];
		ip->io_resp += lbolt - bp->b_start;
	}
	bp->b_start = lbolt;

	/*
	 * Set the time in the iobuf struct for the
	 * system activity profiler.
	 */
	if (dp->io_start == 0)
		dp->io_start = lbolt;

	s = spl6();
	sobppsend(&d327chan[CTL(bp->b_dev)],dp->io_csr,ep);
	splx(s);
}

/*
 *	d327conf: perform configuration command
 */

d327conf(dp,bp,ep)
struct iobuf *dp;
struct buf *bp;
struct sb_envelope *ep;
{
	register struct sb_packet *pp;
	register struct dkconfig *dkcfg;
	register struct d_disk *wd = (struct d_disk *)dp->io_desc;
	register dft;
	register s;

	DEBUGPR(DBCONF,bp->b_dev,"conf");

	/* set up the command packet */
	pp = ep->e_pkt;
	pp->p_cparm1 = 0;
	pp->p_cparm2 = 0;
	pp->p_cparm3 = 0;

	/*
	 * Winchester/Floppy configure (write disk descriptor)
	 */

	/* Is this a default or a re-config? */
	dft = (bp->b_flags & B_DFL_CONF) ? 1 : 0;

	if(!dft) {	/* This is a re-config */
		register struct volumeid *vid;

		pp->p_cmd = SC_WIPC;
		vid = (struct volumeid *) (int)(bp->b_un.b_addr);
		dkcfg = (struct dkconfig *)((int)bp->b_un.b_addr + 256);
		/* Get number of slices and put it in iobuf struct */
		dp->io_slccnt = dkcfg->ios_slccnt;

		/* controller type - high order byte */
		wd->d_ctype =  (vid->vid_typ >> 8);
		/* periperal type - low order byte */
		wd->d_ptype =  (vid->vid_typ & 0x00ff);
		wd->d_heads = dkcfg->ios_hds;	/* number of heads */
		wd->d_fixed = SD_FIXED;	/* fixed or removeable	*/	
		wd->d_cyls =  dkcfg->ios_trk;	/* number of cyl*/
		wd->d_bps =   dkcfg->ios_psm;	/* bytes per sec*/
		wd->d_bsize = BSIZE;		/* logical sector size */
		wd->d_spt =  dkcfg->ios_spt;	/* sectors per track*/
		wd->d_rspt =  dkcfg->ios_ssc;	/* reserved secs per trk */
		wd->d_hrd = 0;			/* hard/soft sector flag*/
		wd->d_ileave = dkcfg->ios_ilv;	/* Interleave factor of media */
		wd->d_fchar = W_INITCHR;	/* format init char */
		wd->d_retry = W_RETRY;		/* retry count		*/
		wd->d_step =  dkcfg->ios_sr;	/* step rate */
		wd->d_precomp = dkcfg->ios_pcom; /* pre-compensation cyl */
		wd->d_redwrite =  0;		 /* reduce write cyl */
		/* zone type (upper nibble) 0 = track, 1 = cylinder 
		   alternate units (upper nibble) 0 = track, 1 = cylinder */ 
		wd->d_alttype = ((dkcfg->ios_catr & W_ZONE) ? 0x11 : 0);
		wd->d_altnum =  dkcfg->ios_alts;/*  number of alternates */
		wd->d_ecc = W_ECCFLG;		/* ECC flag	 */
		wd->d_spiral = dkcfg->ios_sof;	/* Spiral offset of media */
		wd->d_gap1 = dkcfg->ios_gpb1;	/* gap1 size 	*/
		wd->d_gap2 = dkcfg->ios_gpb2;	/* gap2 size 	*/
		wd->d_gap3 = dkcfg->ios_gpb3;	/* gap3 size 	*/
		wd->d_gap4 = dkcfg->ios_gpb4;	/* gap4 size 	*/
		wd->d_strat = 0;	/* disk strategy select	*/
		wd->d_block = 0;	/* disk blocking factor	*/
		wd->d_skew = 0;		/* cylinder skew	*/
		/* cache entry	bit 24 = 0 is cache on */
		wd->d_cache = ((dkcfg->ios_catr & W_CACHE) ? 0 : 1);
		/* set up floppy attribute word descriptor */
		if (FLOP327(dp->b_dev)) {
			register unsigned char flop;
	
			flop = (dkcfg->ios_atw&ATWSN) ? IBMFMT : 0; /* IBM ? */
			flop |= (dkcfg->ios_atw&ATWMF) ? T0DENS : 0; /* track 0 */
			flop |= (dkcfg->ios_atw&ATWMF) ? M96TPI : 0; /* media dens */
			flop |= (dkcfg->ios_atw&ATWDD) ? D96TPI : 0; /* drive dens */
			flop |= (dkcfg->ios_atw&ATWFS) ? DRATE : 0; /* 8" ? */
			flop |= (dkcfg->ios_atw&ATWEN) ? MFMENC : 0;
			wd->d_flop = flop;
			wd->d_fixed = SD_REMOVE; /* fixed or removeable	*/	
		}
		pp->p_saddr = (unsigned long)wd;
	} else if (FLOP327(dp->b_dev)) {
		pp->p_cmd = SC_WIPC;
		pp->p_saddr = (unsigned long)wd;
	} else { /* OPEN command to read VOLID */
		pp->p_cmd = SC_OPEN;
		pp->p_saddr = (unsigned long)bp->b_un.b_addr;
		pp->p_tcount = 2; /* max block count assuming 256 byte blocks */
		pp->p_sgcount = 0;	/* scatter/gather count	*/
		pp->p_cparm3 = sizeof(struct dkblk0); /* number of bytes */
	}


	/* save the pointer to the utab */
	pp->p_utab = (int)dp;

#ifdef DEBUG
	/* save the packet address in the buf struct */
	bp->b_pkt = (int)pp;
#endif

	s = spl6();
	sobppsend(&d327chan[CTL(bp->b_dev)],dp->io_csr,ep);
	splx(s);
}

/*
 *	d327format: perform format command
 */

d327format(dp,bp,ep)
register struct iobuf *dp;
register struct buf *bp;
register struct sb_envelope *ep;
{
	register struct sb_packet *pp;
	register s;

	DEBUGPR(DBFORMAT,bp->b_dev,"format");

	pp = ep->e_pkt;

	if (bp->b_flags & B_FIXBS) {		/* Fix Bad Spot Command */
		register struct dkbadlst *dkb=(struct dkbadlst *)bp->b_un.b_addr;
		pp->p_cmd = SC_FIXBS;
		pp->p_saddr = (unsigned long)&dkb->dk_sector;
		pp->p_327badcnt = 1; /* bad spot count */
	} else {
		register struct dkfmt *dkf=(struct dkfmt *)bp->b_un.b_addr;

		/* set up packet */
		pp->p_cmd = SC_FMT;
		if(!FLOP327(dp->b_dev)) {
			pp->p_saddr = (unsigned long)dkf->dk_badlst;
			pp->p_327badcnt = dkf->dk_badcnt; /* bad spot count */
			pp->p_327badtype = 1; /* bad spot type = physical sector */
		}

		pp->p_paddr = 0;	/* starting cyl and head */
		pp->p_tcount = 0xffffffff; /* entire unit */
		pp->p_utab = (int)dp;	/* save the pointer to the utab */

	}
#ifdef DEBUG
	/* save the packet address in the buf struct */
	bp->b_pkt = (int)pp;
#endif
	s = spl6();
	sobppsend(&d327chan[CTL(bp->b_dev)],dp->io_csr,ep);
	splx(s);
}


/*
 * d327intr: handle the mvme327 interrupts
 */

d327intr(ctl)
{
	register struct iobuf *dp;
	register struct buf *bp;
	register rdev;		/* dev is only 1st device on interrupting ctl */
	register struct pktcnt *pcp;
	register struct sb_packet *pp;
	register struct sb_envelope *ep;
	struct sb_envelope	*sobpprcv();
	int s;

	DEBUGPR(DBINTR,ctl,"interrupt");

#ifdef DEBUG
	if (d327_addr[ctl] == NULL) {
		d327print(ctl, "Undefined interrupt");
		return;
	}
#endif


	while (1) {			/* the "break" will get out of loop */
		s = spl6();
		/* find packet that interrupted */
		if ((ep = sobpprcv(&d327chan[ctl])) == NULL) {
			splx(s);
			break;
		}
		pp = ep->e_pkt;

		dp = (struct iobuf *)pp->p_utab;

		rdev = dp->b_dev;
		pcp = (struct pktcnt *)&dp->io_pktcnt;
		if (pcp->cnt)			/* decrement the packet count */
			pcp->cnt--;
		splx(s);

		/* if last packet back from the mvme327 */
		if (pcp->cnt == 0) {
			register struct iotime *ip;
			if(FLOP327(rdev))
				ip = &m327fstat[FPOS(rdev)];
			else
				ip = &m327stat[POS(rdev)];

			ip->io_act += lbolt - dp->io_start;
			dp->io_start = 0;
		}

		/*
		 * Find the associated buffer structure.
		 */ 
		if( (bp = dp->b_actf) == NULL ) { /* drive active? */
			/* assert this must never be true */
#ifdef DEBUG
			d327print( rdev, "lost request, cannot be" );
#endif
			return;
		}
#ifdef DEBUG
		if (bp->b_pkt == (int)pp) {
#endif
		if (bp->b_flags&B_READ || bp->b_flags&B_WRITE) {
			register struct iotime *ip;
			if(FLOP327(rdev))
				ip = &m327fstat[FPOS(rdev)];
			else
				ip = &m327stat[POS(rdev)];

			ip->io_resp += lbolt - bp->b_start;
		}
		/* ignore errors from Write Descriptor */
		if (pp->p_recover || (pp->p_faterr && (pp->p_cmd != SC_WIPC)))
				d327err(dp,bp,pp);

#ifdef DEBUG
		} else
			d327print(rdev,"No buf associated with status packet");
#endif
		s = spl6();
		sobppfree(&m327freeq,ep);
		splx(s);
		if (pcp->cnt == 0) 
			blkacty &= ~(1<<M327);	/* indicate activity done */
#ifdef SCAT
		dsk2done(dp);		/* finish up */
#else
		dsk1done(dp);		/* finish up */
#endif
	}

}

/*
 *	d327err:  process mvme327 return status
 */

STATIC d327err(dp,bp,pp)
register struct iobuf *dp;
register struct buf *bp;
register struct sb_packet *pp;
{
	register d327eblk;		/* sector # where error occurred */
	register unsigned short svdev;
	register s;

        DEBUGPR(DBERR,bp->b_dev,"error");

	/* set up pointer to "registers" for fmtberr() */
	pp->p_cparm2 = dp->b_dev;  /* save device */
	pp->p_cparm3 = pp->p_cmd;  /* save command */
	/* This changes the b_dev field so must save */
	svdev = dp->b_dev;
	dp->b_dev = makedev(M327,minor(dp->b_dev));
	dp->io_eregs = (paddr_t)&pp->p_cparm2;

	/*
	 * calculate block that error was on for fmtberr()
	 * and save it for d327prerr
	 */

	d327eblk = pp->p_paddr;
	s = spl6();			/* lock out timer */
	fmtberr(dp,d327eblk);		/* error block */
	splx(s);
	dp->b_dev = svdev;		/* restore device */
	if(pp->p_faterr) {	/* fatal error */
		d327prerr(bp->b_dev,pp,d327eblk);	/* display error */
		bp->b_resid = bp->b_bcount;
		bp->b_flags |= B_ERROR;
	} else if(pp->p_recover) {	/* recovered error */
		d327prerr(bp->b_dev,pp,d327eblk);	/* display error */
	}
	pp->p_faterr = 0;
	pp->p_recover = 0;
	pp->p_deverr = 0;
	pp->p_cparm2 = 0;
	pp->p_cparm3 = 0;
}

/*
 *	d327prerr: print out error information
 */

STATIC d327prerr(dev,pp,d327eblk)
register dev,d327eblk;
register struct sb_packet *pp;
{
	register unsigned char fat;
	register unsigned char rec;
	register unsigned short deverr;

	fat = pp->p_faterr;
	rec = pp->p_recover;
	deverr = pp->p_deverr;

	if (fat)
		d327print(dev,"\nFATAL ERROR");
#ifdef	DEBUG
	else if (rec)
		d327print(dev,"\nRECOVERED ERROR");

	PRCMD(pp);				/* print command data */
	printf("Fatal and Recovered Status: %x %x\n", fat, rec);
	printf("Device Status: %x \n", deverr);

	switch(fat) {

		case E_BADD:	d327print(dev," bad descriptor"); break;
		case E_BADC:	d327print(dev," bad command"); break;
		case E_UNCMD:	d327print(dev," unimplemented comm"); break;		
		case E_BADDRV:	d327print(dev," bad drive"); break;
		case E_BADADDR:	d327print(dev," bad logical address"); break;
		case E_BADSCAT:	d327print(dev," bad scatter/gather table"); break;
		case E_UNDEV:	d327print(dev," unimplemented device"); break;
		case E_NOINIT:	d327print(dev," unit not initialized"); break;


		case E_NOID:	d327print(dev," no id found on track"); break;
		case E_SEEK:	d327print(dev," seek error"); break;
		case E_RELCAT:	d327print(dev," relocate track error"); break;
		case E_RNFID:	d327print(dev," record not found,bad ID"); break;
		case E_DSF:	d327print(dev," data sync fault"); break;
		case E_ECC:	d327print(dev," ECC error"); break;
		case E_RNF:	d327print(dev," record not found"); break;


		case E_DRVF:	d327print(dev," drive fault"); break;
		case E_WRTP:	d327print(dev," write protected"); break;
		case E_NOMOT:	d327print(dev," motor not on"); break;
		case E_DOOR:	d327print(dev," door open"); break;
		case E_NRDY:	d327print(dev," disk not ready"); break;


		case E_BUS:	d327print(dev," VME bus error"); break;
		case E_BADAL:	d327print(dev," bad address alignment"); break;
		case E_BTIME:	d327print(dev," bus timeout"); break;
		case E_INVAL:	d327print(dev," invalid DMA xfer count"); break;


		case E_ALT:	d327print(dev," not enough alternates"); break;
		case E_FMT:	d327print(dev," format failed"); break;
		case E_VER:	d327print(dev," verify error"); break;
		case E_BPARAM:	d327print(dev," bad format params"); break;
		case E_NOFIX:	d327print(dev," can't fix bad spot"); break;
		default:	printf(" Fatal error 0x%x",fat);
	}

	switch (deverr) {
		default:	printf(" Device error 0x%x",deverr);
	}
#else
	if(fat) {
		switch(fat) {
			case E_NOINIT:	printf("\t Unit not initialized");break;
			case E_DRVF:	printf("\t Drive fault"); break;
			case E_WRTP:	printf("\t Write protected"); break;
			case E_NRDY:	printf("\t Drive not ready"); break;
			case E_FMT:	printf("\t Format failed"); break;
			case E_NOFIX:	printf("\t Can't fix bad spot"); break;
			default: printf("\t Fatal error status: 0x%x",fat);
		}
		if (pp->p_cmd == SC_READ || pp->p_cmd == SC_WRITE)
			printf("\n\t Logical Sector: 0x%x(%d)",d327eblk,d327eblk);
	}
#endif
#ifdef	DEBUG
	printf("\t\tPrimary Address: %x", pp->p_paddr);
	printf("\t\tSecondary Address: %x", pp->p_saddr);
	printf("\n\t\t\t\tTermination Transfer Count: 0x%x(%d)", pp->p_termcnt,
		pp->p_termcnt);
	printf("\n\t\t\t\tStatus Parameter 1: %x", pp->p_sparm1);
	printf("\n\t\t\t\tStatus Parameter 2: %x", pp->p_sparm2);
	printf("\n\t\t\t\tStatus Parameter 3: %x", pp->p_sparm3);
	printf("\n\t\t\t\tRetry Count: %x(%d)\n", pp->p_retry, pp->p_retry);
#else
	printf("\n");
#endif
}

/*
 *	d327prcmd: print command debug information
 */

STATIC d327prcmd(pp)
register struct sb_packet *pp;
{
	printf("Command: ");

	switch (pp->p_cmd) {
		case SC_WIPC:	printf("Write Descriptor, ");		break;
		case SC_FMT:	printf("Format , ");			break;
		case SC_FIXBS:	printf("Fix badspot , ");		break;
		case SC_READ:	printf("Read Blocks, ");		break;
		case SC_WRITE:	printf("Write Blocks, ");		break;

		default:	printf("Unknown: %x, ", pp->p_cmd);	break;
	}

	printf("Device type: %x, Unit number: %x\n", pp->p_pdtype, pp->p_punit);
}

/*
 *	d327print: print mvme327 driver error messages
 */

d327print(dev,str)
register dev;
register char *str;
{

	if(FLOP327(dev))
		printf("%s on MVME327 ctl %d, floppy drive %d, slice %d\n", str,
			 CTL(dev),SCDEV(dev), SLICE(dev));
	else
		printf("%s on MVME327 ctl %d, SCSI ctl %d, drive %d, slice %d\n", str,
			CTL(dev),SCADR(dev),SCDEV(dev),SLICE(dev));
}

/*
 *	d327breakup: physical I/O breakup routine
 */

d327breakup(bp)
register struct buf *bp;
{
	dma_breakup(d327strategy, bp, BSIZE);
}

/*
 *	d327read: raw I/O read
 */

d327read(dev)
register dev;
{
	register struct iobuf *dp;
	register struct slice *slc;

	DEBUGPR(DBREAD,dev,"read");

	if(FLOP327(dev)) 
		dp = &m327futab[FPOS(dev)];
	else
		dp = &m327utab[POS(dev)];
	slc = (struct slice *) dp->io_slctbl + SLICE(dev);
	if (physck(slc->nblocks,B_READ))
		physio(d327breakup, 0, dev, B_READ);
}

/*
 *	d327write: raw I/O write
 */

d327write(dev)
register dev;
{
	register struct iobuf *dp;
	register struct slice *slc;

	DEBUGPR(DBWRITE,dev,"write");

	if(FLOP327(dev))
		dp = &m327futab[FPOS(dev)];
	else
		dp = &m327utab[POS(dev)];
	slc = (struct slice *) dp->io_slctbl + SLICE(dev);
	if (physck(slc->nblocks,B_WRITE))
		physio(d327breakup, 0, dev, B_WRITE);
}

/*
 *	d327ioctl: perform ioctl functions
 */

d327ioctl(dev,cmd,args,mode)
int dev;
register cmd;
register char *args;
int mode;
{
	register struct buf *bp;
	register struct iobuf *dp;
	register char *addr;
	register struct buf *bb;
	register struct d_disk *wd;
	int size;
	int i;
	struct buf *geteblk();


	DEBUGPR(DBIOCTL,dev,"ioctl");

	if(FLOP327(dev))
		dp = &m327futab[FPOS(dev)];
	else
		dp = &m327utab[POS(dev)];
	wd = (struct d_disk *)dp->io_desc;

	bp = geteblk();

	if (dskioctl(dp,cmd,args,mode,bp)) {
		brelse(bp);
		return;	/* command was handled by disk discipline ioctl */
	}

	switch (cmd) {

	case DKGETCFG:
	case DKGETINFO:	{
			register struct dkconfig *dc;

			clrbuf(bp);
			/* copy needed fields from descriptor structure */
			dc = (struct dkconfig *)(bp->b_un.b_addr +
				sizeof(struct volumeid));
			if (FLOP327(dev)) {
				dc->ios_atw = (wd->d_flop&IBMFMT) ? ATWSN : 0;
				dc->ios_atw |= (wd->d_flop&M96TPI) ? ATWMF : 0;
				dc->ios_atw |= (wd->d_flop&D96TPI) ? ATWDD : 0;
				dc->ios_atw |= (wd->d_flop&MFMENC) ? ATWEN : 0;
				dc->ios_atw |= (wd->d_flop&DRATE) ? ATWFS : 0;
			}


			dc->ios_sr = wd->d_step;
			dc->ios_ssc = wd->d_rspt;
			dc->ios_spt = wd->d_spt;
			dc->ios_hds = wd->d_heads;
			dc->ios_trk = wd->d_cyls;
			dc->ios_pcom = wd->d_redwrite;
			dc->ios_psm = wd->d_bps;

			if (cmd == DKGETCFG) {
				addr = bp->b_un.b_addr +
					sizeof(struct volumeid);
				size = sizeof(struct dkconfig);
			}
			else {
				addr = bp->b_un.b_addr;
				size = sizeof(struct dkblk0);
			}

			copyout(addr,args,size);
			brelse(bp);
			return;
			}

	case DKSETCFG:
	case DKSETINFO: if (cmd == DKSETCFG) {
				addr = bp->b_un.b_addr +
					sizeof(struct volumeid);
				size = sizeof(struct dkconfig);
			}
			else {
				addr = bp->b_un.b_addr;
				size = sizeof(struct dkblk0);
			}
			if (copyin(args,addr,size) == -1) {
				brelse(bp);
				return;
			}
			/*
			 * If the volumeid is available, check to see that
			 * the bad spots to be given to the driver for a
			 * format command are going to be in the form of
			 * "physical sector".
			 */
			if (cmd == DKSETINFO) {
				register struct volumeid *vid;

				vid = (struct volumeid *)bp->b_un.b_addr;
				if (vid->vid_strat != DKHARDSEC &&
					vid->vid_strat != DKPERFECT) {
					brelse(bp);
					u.u_error = ENXIO;
					return;
				}
			}
			bp->b_flags |= B_CONF|B_ABS;
			break;

	case DKFIXBADSPOT:
			if (copyin(args, bp->b_un.b_addr,
				sizeof(struct dkbadlst)) == -1) {
				brelse(bp);
				return;
			}
			bp->b_flags |= B_ABS|B_FIXBS;
			break;

	case DKFORMAT: {
			register struct sb_badsec *b1;

			if (copyin(args,bp->b_un.b_addr,sizeof(struct dkfmt))
			    == -1) {
				brelse(bp);
				return;
			}
			if (!FLOP327(dev) ) { /* if winchester */
				register struct dkfmt *dkf;
				register struct dkbadlst *bs;
				register cnt;
				struct dkbadlst bd;

				dkf = (struct dkfmt *)bp->b_un.b_addr;

				if (dkf->dk_badcnt * sizeof(struct sb_badsec) >
				    SBUFSIZE) {
					brelse(bp);
					u.u_error = ENXIO;
					return;
				}

				bb = geteblk();		/* for bad sec list */

				b1 = (struct sb_badsec *)bb->b_un.b_addr;
				cnt = dkf->dk_badcnt;
				bs = dkf->dk_badlst;

				/* build the real bad sec list */
				while (cnt-- > 0) {
					if (copyin(bs++,&bd,
				    	    sizeof(struct dkbadlst)) == -1) {
						brelse(bp);
						brelse(bb);
						u.u_error = ENXIO;
						return;
					}
					b1->b_cyl = bd.dk_cyl;
					b1->b_head = bd.dk_head;
					b1->b_sec = bd.dk_sector;
					b1++;
				}

				/* point bp at real bad sec list */

				dkf->dk_badlst = (struct dkbadlst *)
						bb->b_un.b_addr;
			}
			bp->b_flags |= B_FORMAT|B_ABS|B_LONGCMD;
			break;
			}

	default:	u.u_error = EINVAL;
			brelse(bp);
			return;
	}

	bp->b_proc = u.u_procp;
	bp->b_dev = dev;
	bp->b_bcount = 0;
	bp->b_blkno = 0;
	d327strategy(bp);
	iowait(bp);
	if ((bp->b_flags&B_FORMAT) && (!FLOP327(dev))) {
		brelse(bb);
	}
	bp->b_flags &= ~(B_CONF|B_FORMAT|B_FIXBS|B_ABS|B_LONGCMD);
	brelse(bp);
}

/*
 *	d327timer: check on status of MVME327 driver periodically
 */

d327timer()
{

	DEBUGPR(DBTIMER,0,"timer");

	/* call the disk discipline timer routine */
#ifdef SCAT
	dsk2timer(&m327utab[0],MAXPOS,&d327dsk);
#else
	dsk1timer(&m327utab[0],MAXPOS,&d327dsk);
#endif
}

/*
 *	f327timer: check on status of MVME327 floppy driver periodically
 */

f327timer()
{

	DEBUGPR(DBTIMER,0,"floppy timer");

	/* call the disk discipline timer routine */
	dsk1timer(&m327futab[0],MAXFPOS,&f327dsk);
}

/*
 *	d327clr: clear the driver after powerfail
 */

d327clr()
{
	DEBUGPR(DBCLR,0,"clr");

	dskclr(&m327utab[0],MAXPOS);
	dskclr(&m327futab[0],MAXFPOS);
}

#ifdef DUMP

/*
 *	mvme327 dump support
 */

STATIC	int	d327offset;

/*
 *	d327dump: dump memory to mvme327 swap area
 */

d327dump()
{
	int d327doio();
	extern	dev_t	dumpdev;
	register dev = dumpdev;

	spl7();				/* disable interrupts */

	DEBUGPR(DBDUMP,dev,"dump");

	if (CTL(dev) >= d327_cnt || d327_addr[CTL(dev)] == NULL) {
		printf("Dump device controller not present\n");
		return;
	}

	{	register struct slice *sp;

		sp = (struct slice *)m327utab[POS(dev)].io_slctbl + SLICE(dev);
		d327offset = sp->blkoff;
	}

}

/*
 *	d327doio: dump out one block of memory
 */

STATIC d327doio(dev,blkno,maddr,rdwr)
register dev;
register blkno;
char *maddr;
int rdwr;
{
	register struct sb_device *rp;
	register struct sb_envelope *ep;
	register struct sb_packet *pp;
	register struct d_disk *wd;
	struct sb_envelope *sobppget();

	DEBUGPR(DBDOIO,dev,"doio");

	if ((ep = sobppget(&m327freeq)) == NULL) {
		for (;;);
	}

	wd = &m327desc[POS(dev)];
	pp = ep->e_pkt;
	pp->p_cmd = (rdwr == B_READ) ? SC_READ : SC_WRITE;
	pp->p_pdtype = SD_SCADR1;
	pp->p_punit = MKPUNIT(dev);
	pp->p_sunit = 0;
	pp->p_paddr = blkno + d327offset;
	pp->p_saddr = (unsigned long)maddr;
	pp->p_tcount = SBUFSIZE / BSIZE;
	pp->p_cparm1 = 0;
	pp->p_cparm2 = 0;
	pp->p_cparm3 = 0;
	rp = d327_addr[CTL(dev)];

	sobppsend(&d327chan[CTL(dev)],rp,ep);

	do {
		ep = sobpprcv(&d327chan[CTL(dev)]);
		pp = ep->e_pkt;
	} while (ep == NULL );	/* wait until done */

	/*
	 *	will hang forever on the above "do-while" statement
	 *	when the controller fails to respond.
	 */

	if (pp->p_faterr )				/* an error occurred */
		d327prerr(dev,pp);

	sobppfree(&m327freeq,ep);
}
#else
d327dump()
{
}
#endif

