/*		Copyright (c) 1988 Motorola Inc.		*/
/*		All Rights Reserved				*/
/*
 *	mvme327tp.c: Motorola MVME327tp VMEmodule Streamer Tape SCSI driver
 *
 *		@(#)mvme327tp.c	35.4
 *
 *	The MVME327tp VMEmodule Streamer Tape SCSI Controller is an intelligent
 *	peripheral controller (IPC) VMEmodule designed to interface
 *	with the extended address line VMEmodules of the VMEbus family.
 *	The MVME327tp supports eight streamer tape drives per controller
 *	using the QIC120, QIC150 and QIC-24 Streaming Tape Interface.
 *
 *	This driver will support forward seeks on reads.
 */

/*	INCLUDE files 	*/

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/systm.h"
#include "sys/buf.h"
#include "sys/signal.h"

#include "sys/fs/s5param.h"
#include "sys/fs/s5dir.h"
#include "sys/immu.h"
#include "sys/region.h"
#include "sys/cmn_err.h"

#include "sys/proc.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/elog.h"
#include "sys/iobuf.h"
#include "sys/sobpp.h"
#include "sys/dbuf.h"
#include "sys/mvme327.h"

#define	MULTITAPE	/* multiple tape support */
#define	STATCMD		/* do stat on open */
#define	BYTESWAP	/* include byteswap code */

#define READWRITE
#undef  READWRITE


#undef DEBUG

#ifndef	M327
#define	M327	9
#endif

#define SC_INIT SC_REW

#undef p_badcnt
#undef p_badtrk


extern int t327maxbsize;		/* maximum DMA buffer size in conf.c */
extern int m327maxscbus;
extern int m327maxscadr;
extern int m327maxscdev;

/*
 *	MVME327tp Driver Minor Number Assignments
 *
 *	/dev/mt and	Minor	Contr-	Reten-	Rewind on	Append or
 *	/dev/rmt names	Number	oller 	sion?	close ?		Truncate ?
 *	--------------	------	------	----	---------	----------
 *	m327_0t		0	0	No	Yes		Truncate
 *	m327_0tn	1	0	No	No		Truncate
 *	m327_0a		2	0	No	Yes		Append
 *	m327_0an	3	0	No	No		Append
 *	m327_1t		4	1	No	Yes		Truncate
 *	m327_1tn	5	1	No	No		Truncate
 *	m327_1a		6	1	No	Yes		Append
 *	m327_1an	7	1	No	No		Append
 *
 *	m327_Xt		X*4	X	No	Yes		Truncate
 *	m327_Xtn	X*4+1	X	No	No		Truncate
 *	m327_Xa		X*4+2	X	No	Yes		Append
 *	m327_Xan	X*4+3	X	No	No		Append
 */

/*
 *	MVME327tp Defines
 */

#define NORETENS(d)	(d & 4) 	/* dev -> no retension on open */
#define	RWD(d)		(!(d&1))	/* dev -> rewind flag */
#define	TDEV(d)		(!(d&2))	/* dev -> default to truncate */
#define	ADEV(d)		(d&2)		/* dev -> default to append */
#define	MKDEV(b,c,p)	((b<<11)|(c<<8)|(p<<5)) /* bus|cadr|perf -> minor */
#define SCBUS(d)	((d & 0x0800) >> 11)
#define SCADR(d)	((d & 0x0700) >> 8)
#define SCDEV(d)	((d & 0x00e0) >> 5)
#define MKINDEX(d)	(((d & 0x07) << 5) | ((d & 0x70) << 4))
#define SCD(d)          ((d & 0x0fe0) >> 5)
#define MKPUNIT(d)	(((d & 0x0700) >> 4) | ((d & 0x00e0) >> 5))

/* dev -> table position */
#define	POS(d) ( (SCBUS(d)*(m327maxscadr*m327maxscdev)) + (SCADR(d)*m327maxscdev) + SCDEV(d) )
#define CTL(d)		((m327a) ? SCBUS(d) : 0)

/*
 *	MVME327tp Variables
 */

/*
 *	t327space.h
 */

extern	struct	iotime	m327stat[];
extern	struct	iobuf	m327utab[];
extern struct 	t_tape 	*m327params[];	/* tape descriptor default params */
extern struct 	t_tape 	m327desc[];	/* tape descriptor default params */
extern	struct iostat	t327iostp[];	/* one per controller */

/*
 *	MVME327tp iobuf usage
 *
 *	iobuf.h		mvme327tp.c	Usage
 *	-------		---------	-----
 *	b_flags		b_flags		Drive state
 *	b_forw
 *	b_back
 *	b_actf		b_actf		head (first) of I/O queue
 *	b_actl		b_actl		tail (last) of I/O queue
 *	b_dev		b_dev		major/minor device number
 *	b_active	b_active	busy flag
 *	b_errcnt	b_errcnt	error count
 *	io_erec		io_erec		error record
 *	io_nreg		io_nreg		number of registers to log upon error
 *	io_addr		io_addr		csr address
 *	io_mba		io_chan		channel structure address
 *	io_stp		io_stp		I/O statistics
 *	io_start	io_start	I/O statistics
 *	io_s1		io_bufp		buffer pointer for double buffering
 *	io_s2		io_bsize	buffer size
 *	io_s3		io_blkno	offset block number
 *	io_s4		io_nxrec	next record
 *	io_s5		io_freeq	free queue pointer
 *	io_s6		io_fileno	file number on tape
 *	io_s7		io_desc		tape descriptor (configuration struct)
#ifdef	MULTITAPE
 *	io_s8		io_proc		process id
#else
 *	io_s8
#endif
 *	io_s9(loword)	io_stat		status from last command
#ifdef	READWRITE
*       io_s9(hiword)	io_moreflgs	more flags 
#endif
 */

/*
 *	MVME327 Tape IOCTL support
 */

#define	M327IOCTL	('M'<<8)
#define	M327REWIND	(M327IOCTL|1)	/* rewind tape */
#define	M327ERASE	(M327IOCTL|2)	/* erase tape */
#define	M327RETENSION	(M327IOCTL|3)	/* retension tape */
#define	M327WRTFM	(M327IOCTL|4)	/* write filemark */
#define	M327RDFM	(M327IOCTL|5)	/* read filemark */
#define	M327SETDMA	(M327IOCTL|6)	/* set DMA buffer size */
#define	M327GETDMA	(M327IOCTL|7)	/* get DMA buffer size */
#define	M327BYTESWAP	(M327IOCTL|8)	/* set/reset byteswapping */



/*
 *	MVME327tp Tape Composite Status Bit Assignments
 */

#define S_EOM		(1<<14)	/* end of media */
#define	S_FM		(1<<15)	/* Filemark detected. */
#define	S_OFFLINE	(0x08)	/* Drive not online. */
#define	S_WP		(0x09)	/* Write protected. */

#define SENSE(x)	(x & 0x0F)

#define E_EOD		0x83	/* end of data */

#define	INVALID	0
#define	VALID	1

#ifdef READWRITE
/* io_moreflgs definitions (upper word), loword is io_proc */
#define RDWROPEN 0x00010000		/* opened for read or write, not both*/ 
#define READONLY 0x00020000		/* reads only are permitted. */
#define WRITEONLY 0x00040000		/* writes only are permitted. */
#endif

#define	INF	0x7fffffff		/* a very large number */
#define	B_CMD	B_FORMAT
#define	B_INIT	B_VERIFY
#define	B_ERASE	B_DONE
#define	B_OFF	B_ASYNC			/* drive offline */
#ifdef	BYTESWAP
#define	B_BSWAP	B_STALE			/* byte swapping on */
#endif


#ifdef	DEBUG		/* debugging code */

#define	STATIC
#define	DEBUGPR(x,d,s)	{ if( x & m327tpdebug ) t327print(d,s); }
#define PRINT1(a)	{ if(t327tpdebug) printf(a); }
#define PRINT2(a,b)	{ if(t327tpdebug) printf(a,b); }
#define PRINT3(a,b,c)	{ if(t327tpdebug) printf(a,b,c); }
#define PRINT4(a,b,c,d)	{ if(t327tpdebug) printf(a,b,c,d); }
#define PRINT5(a,b,c,d,e)	{ if(t327tpdebug) printf(a,b,c,d,e); }
#define PRINT6(a,b,c,d,e,f)	{ if(t327tpdebug) printf(a,b,c,d,e,f); }
#define PRINT(a)	{ if(m327tpdebug) printf(a); }

#define	DBINIT		(1<<0)
#define	DBOPEN		(1<<1)
#define	DBBOPEN		(1<<2)
#define	DBCOPEN		(1<<3)
#define	DBCLOSE		(1<<4)
#define	DBBCLOSE	(1<<5)
#define	DBCCLOSE	(1<<6)
#define	DBSTRAT		(1<<7)
#define	DBSTART		(1<<8)
#define	DBINTR		(1<<9)
#define	DBREAD		(1<<10)
#define	DBWRITE		(1<<11)
#define	DBIOCTL		(1<<12)
#define	DBCLR		(1<<13)
#define	DBDUMP		(1<<14)
#define	DBDOIO		(1<<15)

#else			/* no debugging code */

#define	STATIC	static
#define	DEBUGPR(x,d,s)
#define PRINT1(a)
#define PRINT2(a,b)
#define PRINT3(a,b,c)
#define PRINT4(a,b,c,d)
#define PRINT5(a,b,c,d,e)
#define PRINT6(a,b,c,d,e,f)
#define PRINT(a)

#endif


#define	io_chan		io_mba
#define	io_bufp		io_s1
#define	io_bsize	io_s2
#define	io_blkno	io_s3
#define	io_nxrec	io_s4
#define	io_freeq	io_s5
#define	io_fileno	io_s6
#define	io_desc		io_s7
#ifdef	MULTITAPE
#define	io_proc		io_s8
#endif
#define	io_stat		io_s9
#ifdef READWRITE
#define io_moreflgs	io_s9
#endif
#ifdef	DEBUG
#endif





extern struct sb_packet	m327pkt[];	/* packets - one per controller */
extern struct sb_envelope m327env[];	/* envelopes - three per controller */
extern struct sb_channel t327chan[];	/* channels - one per bus */
extern struct sb_envelope *m327freeq;	/* free queue - one  */


extern	int	t327packets;		/* number of packets */
extern	int	t327envelopes;		/* number of envelopes */

extern int m327a;

/*
 *	conf.h
 */

extern	struct sb_device *d327_addr[];
extern	int		t327_ilev[];
extern	int		t327_ivec[];
extern	int		t327_cnt;
extern	int		t327_bmaj;
extern	int		t327_cmaj;

/*
 *	local variables
 */



STATIC  initted = 0;

#ifdef DEBUG
/* tape debug mask, see mvme327.h for bit encoding */
int	m327tpdebug = 0;
int	t327tpdebug = 0;
#endif


/*
 *	MVME327tp Procedures
 *
 *	t327init: initialize the mvme327tp and the driver
 */

t327init()
{
	register struct iobuf *dp = &m327utab[0];
	struct t_tape **tt = &m327params[0];
	register n;
	int sc_bus,sc_adr, sc_dev;
	
	DEBUGPR(DBINIT,0,"init");


	for( n=0; n < t327_cnt; n++ ) {
		if( bprobe( d327_addr[n], -1 ) )
			d327_addr[n] = NULL;
	}

	/* if MVME327a then there are t327_cnt number of busses */
	n = (m327a) ? t327_cnt : m327maxscbus;
	for( sc_bus = 0; sc_bus < n; sc_bus++ ) {
		for( sc_adr = 0; sc_adr < m327maxscadr; sc_adr++ ) {
			for(sc_dev = 0; sc_dev < m327maxscdev; sc_dev++, tt++) {
		
				/* for disk init see mvme327dk.c */

				/* init tape entry */
				if ((*tt)->t_dtype == M327TAPE) {
					t327devinit(MKDEV(sc_bus,sc_adr,sc_dev)
						,*tt);
				}
			}
		}
		t327_chinit(sc_bus, DATA32);
	}
}

STATIC t327devinit(dev,tt)
register unsigned short dev;
register struct t_tape *tt;
{
	register struct iobuf *dp;
	register sc_bus = SCBUS(dev);
	register struct t_tape *ttt;
	register unsigned short i;

	i = POS(dev);
	dp = &m327utab[i];
	/* copy default parameters */
	ttt = &m327desc[i];
	*ttt = *tt;

	/* if the board isn't there, return */
	if ((dp->io_addr = (paddr_t)d327_addr[CTL(dev)]) == NULL)
		return;

	PRINT4("addr: %x, dp: %x, tt: %x\n", dp->io_addr,dp,tt);

	dp->io_nreg = M327NREGS;

	dp->io_chan = (physadr)&t327chan[sc_bus];
	dp->io_freeq = (int)&m327freeq;
#ifdef READWRITE
	dp->io_stat &= 0xffff0000;
#endif
	dp->io_stat = 0;
	dp->io_bsize = t327maxbsize;
	dp->io_fileno = 0;
	dp->io_desc = (int)ttt;
	dp->b_flags = B_INIT;		/* init 327tp first time */
}

STATIC t327_chinit(sc_bus,am)
register sc_bus;
register am;
{
	register struct iobuf *dp;
	register struct sb_channel *cp;
	register sc_dev;
	/*
	 *	initialize channel and envelopes
	 */

	if((d327_addr[sc_bus] == NULL) && m327a)
		return;
	 
	cp = (struct sb_channel *)&t327chan[sc_bus];

	/* if a channel has been allocated go to next */
	if(cp->c_chan)
		return;

	cp->c_prio = 0;

	/* create the queue  if not already made */
	if(m327freeq == (struct sb_envelope *)NULL)
		sobppcrque(&m327freeq,m327maxscbus*M327PKTS,m327pkt,
			m327env,am,2);


	/* create the channel */
	if( sobppcrchan(cp,d327_addr[sc_bus],&m327freeq,
		t327_ilev[sc_bus],t327_ivec[sc_bus]/sizeof(long),am) == -1 )
	{
		printf("t327: cannot create channel: %d\n", sc_bus);
	}
}

/*
 *	t327open: open an mvme327tp device
 */

STATIC t327open(dev,flag,raw)
register dev;
register flag;
register raw;
{	
	register struct iobuf *dp;
	register unsigned short cstat;
	register stat;
	int	t327strategy();

	DEBUGPR(DBOPEN,dev,"open");

	/*
	 *	if the device is nonexistent, then error
	 */
	if((d327_addr[CTL(dev)] == NULL) || (SCADR(dev) > (m327maxscadr-1)) ||
		 ( SCDEV(dev) > (m327maxscdev-1) ))
	{
		u.u_error = ENXIO;
		return;
	}

	dp = &m327utab[POS(dev)];

	/* only one open at a time */
	if(dp->b_flags&B_OPEN) {
		u.u_error = ENXIO;
		return;
	}


#ifdef READWRITE
	/*
	 *	allow device to be opened for both read and write, but 
	 *	depending on first access, only allow read only or write only.
	 */

	if((flag&FREAD) && (flag&FWRITE)){
		dp->io_moreflgs &= ~( READONLY | WRITEONLY);
		dp->io_moreflgs |= RDWROPEN;
	}
#else
	/*
	 *	will not permit read-and-write opens
	 */

	if( (flag&FREAD) && (flag&FWRITE) ) {
		u.u_error = ENXIO;
		return;
	}
#endif
	/*
	 *	set up the device table
	 */

	dp->b_flags &= ~B_OFF;
	dp->b_flags |= B_OPEN;
	dp->b_dev = dev;
	dp->io_blkno = 0;
	dp->io_nxrec = INF;

#ifdef	STATCMD
	if(t327command(dp,SC_STAT,0))
		u.u_error = 0;
	cstat = dp->io_stat;
	/*
	 *	if the unit is offline and the "O_NDELAY" flag
	 *	isn't used, or if the unit is write protected and
	 *	the file is being opened for writing or
	 *	there is a fatal error on STAT, then error.
	 */

	if ( ( ((flag&FNDELAY)==0) && ( SENSE(cstat) == S_OFFLINE) )
	|| ( dp->b_flags & B_OFF)) {
		dp->b_flags &= ~B_OPEN;
		u.u_error = ENXIO;
		return;
	}
	if ( (flag&FWRITE) && (SENSE(cstat) == S_WP) ) {
		dp->b_flags &= ~B_OPEN;
		u.u_error = ENXIO;
		t327print(dev," Write protected"); 
		return;
	}
#endif
	/* 	Retension tape for all write opens unless "NORETENS" bit 
	 *	is set in minor device number.
	 *
	 */
	if( (flag&FWRITE) && (! NORETENS(dev))) {
		if(t327command(dp, SC_RET, 0)) {
			u.u_error = ENXIO;
			dp->b_flags &= ~B_OPEN;
			return;
		}
	}
	/*
	 *	if being opened for a write, check for truncate or append
	 */

	if( flag&FWRITE ) {
		
		/* set default action */
		if( (flag&(FTRUNC|FAPPEND)) == 0 )
			flag |= ADEV(dev)? FAPPEND : FTRUNC;

		if( flag&FTRUNC ) {	/* O_TRUNC == truncate ... */

			/*
			 *	force a rewind ...
			 *	can't overwrite on streamer tape
			 */

			if( t327command(dp,SC_REW,0) ) {
				u.u_error = ENXIO;
				dp->b_flags &= ~B_OPEN;
				return;
			}


		} else {
			/* append, so position at end of data */

			if( t327command(dp,SC_SPACE,SKEOD) ) {
				u.u_error = ENXIO;
				dp->b_flags &= ~B_OPEN;
				return;
			}
		}
	}

	/*
	 *	If the tape is opened as a character device (raw),
	 *	and if the O_NDELAY flag is not used, then allocate
	 *	a set of buffers to implement double buffering.
	 */

	if( raw && dp->io_bsize && ((flag&FNDELAY) == 0) ) {
#ifdef	MULTITAPE
		if( dp->io_bufp != NULL ) {
			if( u.u_procp->p_pid != dp->io_proc ) {
				dbufclose(dp->io_bufp);
				dp->io_bufp = NULL;
				printf("MVME327: DMA buffers discarded\n");
			} else {
				/* set up retry */
				dbufioctl( dp->io_bufp, DBUFRETRY, 0 );
			}
		} 

		if( dp->io_bufp == NULL ) {
#endif
			dp->io_bufp =
				stat = dbufopen(dp->io_bsize,
					dp->b_dev,t327strategy);
				if(stat == NULL)
				{
					PRINT("Error from dbufopen\n");
				}
			if( dp->io_bufp == NULL ) {
				dp->b_flags &= ~B_OPEN;
				return;
			}
#ifdef	MULTITAPE
		}
#endif
	}
#ifdef	MULTITAPE
	else if( dp->io_bufp ) {
		dbufclose(dp->io_bufp);
		dp->io_bufp = NULL;
		printf("MVME327: DMA buffers discarded\n");
	}

	dp->io_proc = u.u_procp->p_pid;
#endif


	DEBUGPR(DBOPEN,dev,"open successful");
	return;
}

/*
 *	t327bopen: open a block device
 */

t327bopen(dev,flag)
{
	DEBUGPR(DBBOPEN,dev,"bopen");
	t327open(dev,flag,0);
}

/*
 *	t327copen: open a character (raw) device
 */

t327copen(dev,flag)
{
	DEBUGPR(DBCOPEN,dev,"copen");
	t327open(dev,flag,1);
}

/*
 *	t327close: close an mvme327tp device
 */

STATIC t327close(dev,flag,raw)
register dev;
{	register struct iobuf *dp = &m327utab[POS(dev)];

	DEBUGPR(DBCLOSE,dev,"close");

#ifdef READWRITE
	if( (flag&FREAD) && (flag&FWRITE))
		dp->io_moreflgs &= ~(RDWROPEN | READONLY | WRITEONLY);

#endif
	/*
	 *	if the io_bufp is non-null, then flush and
	 *	delete the DMA buffers used for double buffering.
	 */

	if( dp->io_bufp ) {
#ifdef	DEBUG
		if( ! raw )
			t327print(dev,"illegal close");
#endif
#ifdef	MULTITAPE
		if( (flag&FREAD) || ((dp->b_flags&B_ERROR) == 0) ) {
#endif
			dbufclose(dp->io_bufp);
			dp->io_bufp = NULL;
#ifdef	MULTITAPE
		} else {
			dp->b_flags &= ~B_ERROR;
			printf("MVME327: DMA buffers still active.\n");
		}
#endif
	}

	/*
	 *	if the unit was opened for writing, then write
	 *	a filemark to indicate end of file
	 */

	if( dp->b_flags&B_ERASE )
		dp->b_flags &= ~B_ERASE;

	else if( flag&FWRITE )
		t327command(dp,SC_WEOF,0);

	/*
	 *	if the unit is a "rewind-on-close", then rewind the device
	 */


	if( RWD(dev) )
		t327command(dp,SC_REW,0);
	else if(dp->io_blkno && (flag&FREAD)) /* no FM, seek to next file */
		t327command(dp,SC_SPACE,SKNEXT);

	dp->b_flags &= ~B_OPEN;

	DEBUGPR(DBCLOSE,dev,"close successful");
}

/*
 *	t327bclose: close a block device
 */

t327bclose(dev,flag)
{
	DEBUGPR(DBBCLOSE,dev,"bclose");
	t327close(dev,flag,0);
}

/*
 *	t327cclose: close a character device
 */

t327cclose(dev,flag)
{
	DEBUGPR(DBCCLOSE,dev,"cclose");
	t327close(dev,flag,1);
}

/*
 *	t327command: issue a command and call strategy
 */

STATIC t327command(dp,cmd,arg)
register struct iobuf *dp;
register cmd;
register arg;
{	register struct buf *bp;

	bp = geteblk();			/* get empty buffer */
	bp->b_proc = u.u_procp;
	bp->b_dev = dp->b_dev;
	bp->b_bcount = cmd;
	bp->b_blkno = arg;
	bp->b_error = 0;
	bp->b_flags |= B_CMD|B_PHYS;
	t327strategy(bp);
	iowait(bp);			/* wait until done */
	bp->b_flags &= ~(B_CMD|B_PHYS);
	brelse(bp);
	return(u.u_error);
}

/*
 *	t327strategy: perform an I/O
 */

t327strategy(bp)
register struct buf *bp;
{	register struct iobuf *dp = &m327utab[POS(bp->b_dev)];
	register lvl;

	DEBUGPR(DBSTRAT,bp->b_dev,"strategy");

	if(((bp->b_flags&B_CMD ) == 0) && (bp->b_bcount&(BSIZE-1))) {
		/* no partial block I/O */
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}

#ifdef READWRITE
	 /* if device was opened for read/write then depending on first access,
	    permit reads only or writes only. 
	 */
	if(((bp->b_flags&B_CMD) == 0) && (dp->io_moreflgs & RDWROPEN)){
		if(!(dp->io_moreflgs & (READONLY | WRITEONLY))){
			if(bp->b_flags & B_READ)
				dp->io_moreflgs |= READONLY;
			else
				dp->io_moreflgs |= WRITEONLY;
		}
		else{
	     	      if(!(((bp->b_flags&B_READ) && (dp->io_moreflgs&READONLY))
		    || (((bp->b_flags&B_READ) == 0) && (dp->io_moreflgs&WRITEONLY))))
		        {	 /* can't mix reads and writes */
				bp->b_flags |= B_ERROR;
				bp->b_error = ENXIO;
				iodone(bp);
				return;
			}
		}
	}
#endif

	bp->b_start = lbolt;
	if(( bp->b_flags&B_CMD ) == 0 )
		m327stat[POS(bp->b_dev)].io_bcnt += btoc(bp->b_bcount);
	bp->av_forw = NULL;

	lvl = spl6();

	if( dp->b_actf == NULL )
		dp->b_actf = bp;
	else
		dp->b_actl->av_forw = bp;
	dp->b_actl = bp;
	if( dp->b_active == 0 )
		t327start(dp);

	splx(lvl);
}

/*
 *	t327start: start an I/O
 */

STATIC t327start(dp)
register struct iobuf *dp;
{	register struct buf *bp;
	register struct sb_envelope *ep;
	register struct sb_channel *cp = (struct sb_channel *)dp->io_chan;
	struct t_tape *tp_desc = (struct t_tape *)dp->io_desc;
	register cmd;
	register bpblkno;
	struct sb_envelope *sobppget();

	DEBUGPR(DBSTART,dp->b_dev,"start");

	/*
	 *	get next buffer ... if null return
	 */

	while(( bp = dp->b_actf ) != NULL ) {

		dp->b_active++;

		if(( bp->b_flags&B_CMD ) == 0 ) {

			/*
			 *	if the current block number is greater
			 *	than the next record number on the tape,
			 *	then it's an error (lseek(2) too far).
			 */

			if( (bpblkno = bp->b_blkno) > dp->io_nxrec ) {
				bp->b_error = EIO;
				goto abort;
			}

			/*
			 *	if the current block is equal to the
			 *	next record number, then we're at the
			 *	end of the tape file (on a filemark).
			 */

			if((bpblkno==dp->io_nxrec)&&(bp->b_flags&B_READ)) {
				clrbuf(bp);
				bp->b_resid = bp->b_bcount;
				goto next;
			}

			/*
			 *	set the next record number to one past the
			 *	buffer block number (on writes).
			 */
			if(( bp->b_flags&B_READ ) == 0 )  {
				PRINT1("start_write: ");
			}
			else {
				PRINT1("start_read: ");
			}

			PRINT5("bp: %x, dp: %x, bpb: %x, dpb: %x\n", 
				bp, dp, bp->b_blkno, dp->io_blkno);
			if(( bp->b_flags&B_READ ) == 0 ) {
				dp->io_nxrec = bpblkno + bp->b_bcount/BSIZE;

				/*
				 *	no seeks on write
				 */

				if( dp->io_blkno != bp->b_blkno ) {
					PRINT1("t327start: unexpected write\n");
					bp->b_error = EIO;
					goto abort;
				}
			} else if( dp->io_blkno > bpblkno ) {
				bp->b_error = ENXIO;
				goto abort;
			}

		}

		if( dp->b_flags&B_OFF )
			goto abort;

		/*
		 *	if command, then set command
		 */

		if( (dp->b_flags&B_ONCE) == 0 )
			cmd = SC_WIPC;
		else if( dp->b_flags&B_INIT )
			cmd = SC_INIT;
		else if( bp->b_flags&B_CMD )
			cmd = bp->b_bcount;	/* see t327command */
		else
			cmd = (bp->b_flags&B_READ)? SC_READ : SC_WRITE;

		/*
		 *	set up packet
		 */

	{	register struct sb_packet *pp;

		if((ep = sobppget(dp->io_freeq)) == NULL) {  /* no packets */
				PRINT2("t327*out of packets: dev: %x\n", 
					dp->b_dev);
				bp->b_error = EIO;
				goto abort;
			}
		pp = ep->e_pkt;
		bzero(&pp->p_paddr,0x24); /* clear packet */
		pp->p_cmd = cmd;
		if((!m327a) && SCBUS(dp->b_dev)) /* 2nd SCSI bus */
			pp->p_pdtype = SD_SCSI2;
		else
			pp->p_pdtype = SD_SCSI1;
		pp->p_punit = MKPUNIT(dp->b_dev);
		pp->p_utab = (int)dp; /* save pointer to struct iobuf */
		pp->p_faterr = pp->p_deverr = 0;

		switch(cmd) {

		case SC_READ:	/* read to eof */
			pp->p_paddr = bp->b_blkno;

		case SC_WRITE:	/* write logical */

			if( bp->b_flags&B_PHYS ) {
				dp->io_stp->io_ops++;

			} else if( dp->io_blkno == bpblkno ) {
				dp->io_stp->io_ops++;

			} else	/* do block read and throw away later */
				dp->io_stp->io_misc++;
#ifdef DEBUG
			if(cmd == SC_READ) {
				PRINT1("read: ");
			}
			else {
				PRINT1("write: ");
			}
#endif
			PRINT5("bp: %x, dp: %x, bpb: %x, dpb: %x\n", 
				bp, dp, bp->b_blkno, dp->io_blkno);
			pp->p_saddr = (unsigned)paddr(bp);
			pp->p_tcount = bp->b_bcount/BSIZE;
			if( dp->io_blkno < bp->b_blkno ) {
				/* seek forward ... but not too far */
				if(pp->p_tcount > (bp->b_blkno-dp->io_blkno))
					pp->p_tcount = bp->b_blkno-dp->io_blkno;
			}
			break;

		case SC_WIPC:
			if(dp->b_flags & B_BSWAP)
				tp_desc->t_byteswap = 1;
			else
				tp_desc->t_byteswap = 0;
			pp->p_saddr = (int)tp_desc;
			pp->p_am = DATA32;
			pp->p_busw = BUSW;
			break;

		case SC_WEOF:	/* filemark */
			pp->p_tcount = 1;
			bp->b_blkno = 0;	/* for rewind on close */
		case SC_REW:	/* rewind */
		case SC_ERASE:	/* erase */
		case SC_STAT:	/* get status */
			dp->io_fileno = 0;
			dp->io_stp->io_misc++;
			break;

		case SC_RET:	/* retension */
			pp->p_cmdctrl = RETENSION;
			dp->io_fileno = 0;
			dp->io_stp->io_misc++;
			break;

		case SC_SPACE:	/* seek */
			dp->io_stp->io_misc++;
			dp->io_stp->io_misc++;
			pp->p_cmdctrl = bp->b_blkno;
			if(bp->b_blkno == SKNEXT)	/* next file */
				pp->p_tcount = 1;
			break;
		default:
			PRINT2("MVME327tp bad command: %x\n", cmd );
			break;
		}

		blkacty |= (1<<M327);

		/* set up I/O start time */
		if( dp->io_start == 0 )
			dp->io_start = lbolt;
		 

	}
		/*
		 *	everything ok ... send
		 */

	{	register struct sb_device *rp = (struct sb_device *)dp->io_addr;

		sobppsend(cp, rp, ep);
		return;
	}

		/*
		 *	come here on error
		 */

	abort:
		bp->b_flags |= B_ERROR;
	next:
		dp->b_actf = bp->av_forw;
		iodone(bp);
	}

	dp->b_active = 0;
}

/*
 *	t327intr: interrupt handler
 */

t327intr(ctl)
int ctl;
{	
	register struct sb_channel *cp;
	register struct iobuf *dp;
	register struct buf *bp;
	register dev;
	register struct sb_packet *pp;
	register resid;
	register unsigned short stat;
	register fatal;
	register termcnt;
	register cmd;
	struct sb_envelope *ep;
	struct sb_envelope *sobpprcv();

	DEBUGPR(DBINTR,ctl,"intr");


	/*
	 *	get status envelope
	 */

#ifdef	DEBUG
	if(d327_addr[ctl] == NULL)	
	{
		t327print(dev,"null address in interrupt");
		return;
	}
#endif


	cp = &t327chan[ctl];
	while((ep = sobpprcv(cp)) != NULL ) {

		pp = ep->e_pkt;
		dp = (struct iobuf *)pp->p_utab;
		dev = dp->b_dev;

		bp = dp->b_actf;

		if( bp == NULL ) {
			logstray(dp->io_addr);
			return;
		}

		blkacty &= ~(1<<M327);
		dp->b_active = 0;

		
		cmd = pp->p_cmd;
		termcnt = pp->p_termcnt;
		dp->io_stat = stat = pp->p_deverr;
		fatal = pp->p_faterr;
		PRINT6("intr: dev: %x, cmd: %x, cnt: %x, fatal: %x, stat: %x\n",
			dev, cmd, termcnt, fatal, stat);

		/*
		 *	if fatal error
		 */

		if( fatal ) {
			if( fatal == E_WRTP )  /* for STAT command */
				dp->io_stat |= S_WP;
			else {
				dp->b_flags |= ( B_INIT | B_OFF);
				bp->b_flags |= B_ERROR;
				t327err(dp,bp,pp);
				sobppfree(dp->io_freeq,ep);
				goto complete;
			}
		}
		sobppfree(dp->io_freeq,ep);	/* give back envelope */

		/*
		 *	if the command is a read or write
		 */

		if(((cmd == SC_WRITE ) || (cmd == SC_READ))) {
			resid = termcnt/BSIZE;
			dp->io_blkno += resid;
			if( stat&S_FM ) {
				dp->io_blkno = 0;
				dp->io_nxrec = bp->b_blkno + resid;
			} else {
				if( dp->io_blkno <= bp->b_blkno ) {
					/* ignore I/O ... continue */
					t327start(dp);
					return;
				}
			}

		/*
		 *	if command is a write IPC
		 */

		} else if( cmd == SC_WIPC ) {
			if(( dp->b_flags&B_ONCE) == 0) { /* initial WIPC */
				dp->b_flags |= B_ONCE;	
				t327start(dp);
				initted = 1;	
				return;
			}
		
		
		/*
		 *	if an init command has been issued
		 */

		} else if( dp->b_flags&(B_INIT) ) {
			dp->b_flags &= ~(B_INIT);
			t327start(dp);		/* do command */
			return;
		}


		/*
		 *	I/O is complete ... continue
		 */

	complete:
		logberr(dp,bp->b_flags & B_ERROR);
		dp->b_actf = bp->av_forw;

		if(((bp->b_flags&B_CMD ) == 0) && 
		(!(fatal) && ((cmd == SC_WRITE ) || (cmd == SC_READ)))) {
			resid = bp->b_bcount - termcnt;
			bp->b_resid = (resid <= 0)? 0 : resid;

		} else if((bp->b_flags&B_CMD ) == 0) 
			bp->b_resid = bp->b_bcount;


		{	register struct iotime *ip = &m327stat[ctl];

			ip->io_resp += lbolt - bp->b_start;
			ip->io_act += lbolt - dp->io_start;
			dp->io_start = 0;
		}


		iodone(bp);		/* complete the I/O */
		t327start(dp);		/* start next I/O */
	}
}


t327breakup(bp)
{
	dma_breakup(t327strategy,bp,BSIZE);
}


/*
 *	t327read: character device read
 */

t327read(dev)
register dev;
{	register struct iobuf *dp = &m327utab[POS(dev)];

	DEBUGPR(DBREAD,dev,"read");

#ifdef READWRITE
	if(dp->io_moreflgs&RDWROPEN){
		if(dp->io_moreflgs&WRITEONLY){
			u.u_error = ENXIO;
			return;
		}
		else
			dp->io_moreflgs |= READONLY;
	}
#endif
	
	if( dp->io_bufp )
		dbufread( dp->io_bufp );
	else
 		physio( t327breakup, 0, dev, B_READ );
	return;
}

/*
 *	t327write: character device write
 */

t327write(dev)
{	register struct iobuf *dp = &m327utab[POS(dev)];

	DEBUGPR(DBWRITE,dev,"write");

#ifdef READWRITE
	if(dp->io_moreflgs&RDWROPEN){
		if(dp->io_moreflgs&READONLY){
			u.u_error = ENXIO;
			return;
		}
		else
			dp->io_moreflgs |= WRITEONLY;
	}
#endif
	
	if( dp->io_bufp ) {
		dbufwrite( dp->io_bufp );
#ifdef	MULTITAPE
		if( u.u_error || u.u_count != 0 )
		{
			dp->b_flags |= B_ERROR;	/* remember error */
		}
#endif
	} else
 		physio( t327breakup, 0, dev, B_WRITE );
	return;
}

/*
 *	t327ioctl: ioctl support for the MVME327tp
 */

t327ioctl( dev, cmd, arg, mode )
register dev;
{	register struct iobuf *dp = &m327utab[POS(dev)];

	DEBUGPR(DBIOCTL,dev,"ioctl");

	switch(cmd) {
	case M327REWIND:	/* rewind tape */
		t327command(dp,SC_REW,0);
		break;
	case M327ERASE:		/* erase tape */
		if( t327command(dp,SC_ERASE,0) == 0 )
			dp->b_flags |= B_ERASE;
		break;
	case M327RETENSION:	/* retension tape */
		t327command(dp,SC_RET,0);
		break;
	case M327WRTFM:		/* write filemark */
		t327command(dp,SC_WEOF,0);
		break;
	case M327RDFM:		/* read filemark */
		t327command(dp,SC_SPACE,SKNEXT);
		break;
#ifdef	BYTESWAP
	case	M327BYTESWAP:	/* set/reset byte swapping */
		if( arg != 0 )
			dp->b_flags |= B_BSWAP;
		else
			dp->b_flags &= ~B_BSWAP;
		t327command(dp,SC_WIPC,0);
		break;
#endif
	case M327SETDMA:	/* set DMA buffer size */
		if( ! suser() ) {
			u.u_error = EPERM;
			return;
		}

		if( dp->io_bufp == NULL )
			dp->io_bsize = arg;

		else {
			dbufioctl(dp->io_bufp,DBUFSET,arg);
			if( u.u_error )
				dp->io_bsize = 0;
			else
				dp->io_bsize = arg;
		}
		break;
	case M327GETDMA:	/* get DMA buffer size */
		if(( dp->io_bufp == NULL ) || ( dp->io_bsize == 0 )) {
			copyout( &dp->io_bsize, arg, sizeof(dp->io_bsize) );
		} else
			dbufioctl(dp->io_bufp,DBUFGET,arg);
		break;


	default:
		u.u_error = EINVAL;
		break;
	}
}

/*
 *	t327clr: clear the mvme327tp after a power fail
 */

t327clr()
{
	DEBUGPR(DBCLR,0,"clr");
}

/*
 *	t327print: driver print procedure
 */

t327print(dev,str)
register dev;
register char *str;
{

	printf("%s on MVME327 ctl %d, SCSI ctl %d, Tape drive %d\n", str,
			CTL(dev),SCADR(dev),SCDEV(dev));
}


/*
 *	t327err:  process mvme327 return status
 */

STATIC t327err(dp,bp,pp)
register struct iobuf *dp;
register struct buf *bp;
register struct sb_packet *pp;
{
	register t327eblk;		/* block # where error occurred */
	register unsigned short svdev;
	register s;


	/* set up pointer to "registers" for fmtberr() */
	pp->p_cparm2 = dp->b_dev;  /* save device */
	pp->p_cparm3 = pp->p_cmd;  /* save command */
	/* This changes the b_dev field so must save */
	svdev = dp->b_dev;
	dp->b_dev = makedev(M327,minor(dp->b_dev));
	dp->io_addr = (paddr_t)&pp->p_cparm2;

	/*
	 * calculate block that error was on for fmtberr()
	 * and save it for t327prerr
	 */

	t327eblk = pp->p_paddr;
	s = spl6();			/* lock out timer */
	fmtberr(dp,t327eblk);		/* error block */
	splx(s);
	dp->b_dev = svdev;		/* restore device */

	t327prerr(bp->b_dev,pp,t327eblk);	/* display error */
	pp->p_cparm2 = 0;
	pp->p_cparm3 = 0;
}


/*
 *	t327prerr: print out error information
 */

STATIC t327prerr(dev,pp,t327eblk)
register dev,t327eblk;
register struct sb_packet *pp;
{
	register unsigned char fat;
	register unsigned char rec;
	register unsigned short deverr;

	fat = pp->p_faterr;
	rec = pp->p_recover;
	deverr = pp->p_deverr;

	if (fat)
		t327print(dev,"\nFATAL ERROR");
#ifdef	DEBUG
	else if (rec)
		t327print(dev,"\nRECOVERED ERROR");

	t327prcmd(pp->p_cmd);	 /* print command data */
	printf("Fatal and Recovered Status: %x %x\n", fat, rec);
	printf("Device Status: %x \n", deverr);

	switch(fat) {

		case E_BADD:	t327print(dev," bad descriptor"); break;
		case E_BADC:	t327print(dev," bad command"); break;
		case E_UNCMD:	t327print(dev," unimplemented comm"); break;		
		case E_BADDRV:	t327print(dev," bad drive"); break;
		case E_BADADDR:	t327print(dev," bad logical address"); break;
		case E_BADSCAT:	t327print(dev," bad scatter/gather table"); break;
		case E_UNDEV:	t327print(dev," unimplemented device"); break;
		case E_NOINIT:	t327print(dev," unit not initialized"); break;


		case E_NOID:	t327print(dev," no id found on track"); break;
		case E_SEEK:	t327print(dev," seek error"); break;
		case E_RELCAT:	t327print(dev," relocate track error"); break;
		case E_RNFID:	t327print(dev," record not found,bad ID"); break;
		case E_DSF:	t327print(dev," data sync fault"); break;
		case E_ECC:	t327print(dev," ECC error"); break;
		case E_RNF:	t327print(dev," record not found"); break;


		case E_DRVF:	t327print(dev," drive fault"); break;
		case E_WRTP:	t327print(dev," write protected"); break;
		case E_NOMOT:	t327print(dev," motor not on"); break;
		case E_DOOR:	t327print(dev," door open"); break;
		case E_NRDY:	t327print(dev," tape not ready"); break;


		case E_BUS:	t327print(dev," VME bus error"); break;
		case E_BADAL:	t327print(dev," bad address alignment"); break;
		case E_BTIME:	t327print(dev," bus timeout"); break;
		case E_INVAL:	t327print(dev," invalid DMA xfer count"); break;


		case E_ALT:	t327print(dev," not enough alternates"); break;
		case E_FMT:	t327print(dev," format failed"); break;
		case E_VER:	t327print(dev," verify error"); break;
		case E_BPARAM:	t327print(dev," bad format params"); break;
		case E_NOFIX:	t327print(dev," can't fix bad spot"); break;
		default:	printf(" Fatal error 0x%x",fat);
	}

	switch (deverr) {
		default:	printf(" Device error 0x%x",deverr);
	}
#else

	if( deverr & S_EOM ) 
		printf("\n\t End of media"); 
	else {
		switch(fat) {
			case E_DRVF:	printf("\t Drive fault"); break;
			case E_WRTP:	printf("\t Write protected"); break;
			case E_NRDY:	printf("\t Tape not ready"); break;
			case E_EOD:	printf("\t End of data"); break;
			default: printf("\t Fatal error status: 0x%x",fat);
		}
		if (fat && (pp->p_cmd == SC_READ || pp->p_cmd == SC_WRITE))
			printf("\n\t Logical Block: 0x%x(%d)",t327eblk,t327eblk);
	}
#endif
#ifdef	DEBUG
	printf("\t\tPrimary Address: %x", pp->p_paddr);
	printf("\t\tSecondary Address: %x", pp->p_saddr);
	printf("\n\t\t\t\tTermination Transfer Count: 0x%x(%d)", pp->p_termcnt,
		pp->p_termcnt);
	printf("\n\t\t\t\tStatus Parameter 1: %x", pp->p_sparm1);
	printf("\n\t\t\t\tStatus Parameter 2: %x", pp->p_sparm2);
	printf("\n\t\t\t\tStatus Parameter 3: %x", pp->p_sparm3);
	printf("\n\t\t\t\tRetry Count: %x(%d)\n", pp->p_retry, pp->p_retry);
#else
	printf("\n");
#endif
}

/*
 *	t327prcmd: print MVME327tp command
 */

STATIC t327prcmd(cmd)
register unsigned short cmd;

{	switch(cmd) {
	case SC_READ:	t327prm("Read logical");		break;
	case SC_STAT:	t327prm("Read status");			break;
	case SC_WRITE:	t327prm("Write blocks");		break;
	case SC_WIPC:	t327prm("Write ipc descriptor");	break;
	case SC_WEOF:	t327prm("Write eof");			break;
	case SC_REW:	t327prm("Rewind tape");			break;
	case SC_ERASE:	t327prm("Erase tape");			break;
	case SC_SPACE:	t327prm("Seek");			break;
	default:
		PRINT2("MVME327tp: Unknown command %x.\n", cmd );
		break;
	}
}


/*
 *	t327prm: print message with MVME: at front
 */

STATIC t327prm(s)
{	PRINT2("MVME327tp: %s.\n", s );
}

/*
 *
 *		Dump routines
 *
 */

/*
 *	t327dump: dump memory to tape
 */

t327dump()
{	
	int t327doio();
	extern	dev_t dumpdev;
	register dev = dumpdev;
	register struct iobuf *dp = &m327utab[POS(dev)];
	struct sb_packet packet;
	register struct sb_packet *pp = &packet;

	DEBUGPR(DBDUMP,dev,"dump");

	if(dp->io_addr == NULL)
	{
		printf("Dump device controller not present\n");
		return;
	}
	spl7();
	pp->p_cmd = SC_INIT;
	if(t327sendcmd(pp,dev,SD_TAPE))
		return(-1);
	if(initted == 0) {
		pp->p_cmd = SC_WIPC;
		pp->p_saddr = dp->io_desc;
		if(t327sendcmd(pp,dev,SD_IPC))
			return(-1);
	} 

	dp->io_blkno = INF - 1;
	/*	until memdump.c  is ported
	memdump(t327doio);
	*/

	pp->p_cmd = SC_WEOF;
	if(t327sendcmd(pp,dev,SD_TAPE))
		return(-1);

	pp->p_cmd = SC_REW;
	if(t327sendcmd(pp,dev,SD_TAPE))
		return(-1);
}

/*
 *	t327doio: write one block of memory
 */

STATIC t327doio(dev,blkno,maddr,rdwr)
register dev;		/* device # */
register blkno;		/* block number on tape (512-byte) */
register char *maddr;	/* memory address */
register rdwr;		/* B_READ or B_WRITE */
{	register struct iobuf *dp = &m327utab[POS(dev)];
	struct sb_packet packet;
	register struct sb_packet *pp = &packet;
	register i;

	DEBUGPR(DBDOIO,dev,"doio");

	/* set up packet */
	pp->p_cmd = (rdwr == B_READ) ? SC_READ : SC_WRITE;
	pp->p_paddr = 0;
	pp->p_saddr = (unsigned) maddr;
	pp->p_tcount = 1;		/* one block */

	if(t327sendcmd(pp,dev,SD_TAPE))
		return(-1);

	if(dp->io_blkno + 1 < blkno)	/* skip hole */
		for(i=dp->io_blkno+1;i<blkno;i++)
			t327sendcmd(pp,dev,SD_TAPE);
	dp->io_blkno = blkno;
}



/*
 *	t327sendcmd - send command to vme327tp
 */

STATIC t327sendcmd(pp,dev,type)
register struct sb_packet *pp;
register dev;
register type;
{	register struct iobuf *dp = &m327utab[POS(dev)];
	register struct sb_envelope *ep;
	register struct sb_packet *cpp;
	register tout = INF;
	struct sb_envelope *sobppget();
	struct sb_envelope *sobpprcv();

	if((ep = sobppget(dp->io_freeq)) == NULL) {  /* no packets */
		t327print(dev,"no free envelopes");
		return(-1);
	}

	cpp = ep->e_pkt;

	/* copy packet */

	cpp->p_cmd = pp->p_cmd;
	cpp->p_paddr = pp->p_paddr;
	cpp->p_saddr = pp->p_saddr;
	cpp->p_tcount = pp->p_tcount;
	cpp->p_pdtype = type;	/* device type */
	
	sobppsend(dp->io_chan,dp->io_addr,ep);

	/*
	 * wait till packet put on status tail 
	 */

	while ((ep = sobpprcv(dp->io_chan)) == NULL && --tout > 0)
		;

	if(tout <= 0) {
		t327print(dev,"timeout while waiting for completion");
		return(-1);
	}

	pp = ep->e_pkt;


	if( pp->p_deverr ) 		/* an error occurred */
		/* print error message(s) */
		t327prerr(dev,pp);

	sobppfree(dp->io_freeq,ep);
	return(0);
}
