/*		Copyright (c) 1986, 1987, 1988 Motorola			*/
/*		All Rights Reserved				*/
/*		@(#)mvme332xt.c	35.5	*/
/*
 *	mvme332xt.c: Motorola MVME332xt VMEmodule Communication Controller 
 *
 *	The MVME332xt VMEmodule Communication Controller is an intelligent
 *	peripheral controller (IPC) VMEmodule designed to interface
 *	with the extended address line VMEmodules of the VMEbus family.
 *	The MVME332xt supports eight serial communication lines and one
 *	parallel printer port.
 *
 */

#ident	"@(#)mvme332xt.c	35.5"

/*	INCLUDE files 	*/

#include "sys/types.h"
#include "sys/param.h"
#include "sys/sysmacros.h"

#ifndef REL2
#include "sys/immu.h"
#include "sys/region.h"
#define	 ISP
#define	 KERNEL		1
#define	 defined_io	1
#else
#include "sys/text.h"	
#endif

#include "sys/proc.h"
#include "sys/dir.h"
#include "sys/signal.h"
#include "sys/user.h"
#include "sys/inode.h"
#include "sys/elog.h"
#include "sys/file.h"
#include "sys/errno.h"
#include "sys/termio.h"
#include "sys/ttold.h"
#include "sys/tty.h"
#include "sys/sxt.h"
#include "sys/mvme332xt.h" 

#ifdef	ISP
#include "sys/cstty.h"
#endif

#define	TAS		/* do tas instruction in getchan */
#undef	TAS

#define	DEBUG
#undef	DEBUG

#ifdef	DEBUG		/* debugging code */
#define	DEBUGPR(x,d,s)	{ if( x & m332Debug ) m332print(d,s); }
#define	DBINIT		(1<<0)
#define	DBOPEN		(1<<1)
#define	DBCLOSE		(1<<2)
#define	DBINTR		(1<<3)
#define	DBREAD		(1<<4)
#define	DBWRITE		(1<<5)
#define	DBIOCTL		(1<<6)

int	m332Debug = DBINIT|DBOPEN|DBCLOSE|DBINTR|DBREAD|DBWRITE|DBIOCTL;

#define	ASSERT(EX)	if (!(EX)) {printf("Assertion failed: %s, in file %s, at line %d\n", "EX", __FILE__, __LINE__); while(1);}

#else			/* no debugging code */

#define	DEBUGPR(x,d,s)
#define	ASSERT(EX)

#endif

/* copy from user space to system space */
#define	COPYIN(usr, sys, cnt)\
	if (copyin(usr, sys, cnt)) {\
		u.u_error = EFAULT;\
		break;\
	}

/* copy from sys space to user space */
#define	COPYOUT(sys, usr, cnt)\
	if (copyout(sys, usr, cnt)) {\
		u.u_error = EFAULT;\
		break;\
	}

/*
 *	Global variables
 */
int	m332major;			/* 332XT major number, used by SXT */

/* local copy of m332_addr[] */
static	struct	m332_dev *m332_Addr[CTL(0xff) + 1];
static	int 	m332adrm;		/* address modifier */
static	int	m332busw;		/* bus width */

/*
 * Serial device open's default termio.
 */
#ifndef REL2
extern	int	greek;			/* dual language mode */

struct termio  m332deftio = {		/* release 3 defaults      */
	0,						/* c_iflag */
	0,						/* c_oflag */
	B9600|CS8|CREAD|HUPCL,				/* c_cflag */
	0,						/* c_lflag */
	0,						/* c_line  */
	{						/* c_cc[]  */
		CINTR, CQUIT, CERASE, CKILL, CEOF, 0, 0, 0,
	},
};
#else
struct termio  m332deftio = {		/* release 2 defaults      */
	BRKINT|IGNPAR|ISTRIP|ICRNL|IXON|IXOFF|IXANY,	/* c_iflag */
	ONLCR|TAB3|OPOST,				/* c_oflag */
	B9600|CS8|CREAD|HUPCL,				/* c_cflag */
	ISIG|ICANON|ECHOE|ECHO|ECHOK,			/* c_lflag */
	DEF_LINE,					/* c_line  */
	{						/* c_cc[]  */
		CINTR, CQUIT, CERASE, CKILL, CEOF, 0, 0, 0,
	},
};
#endif

/*
 * Printer device open's default termio.
 */
struct termio  m332defprt = {
	0,						/* c_iflag */
	ONLCR|TAB3|OPOST,				/* c_oflag */
	HUPCL,						/* c_cflag */
	0,						/* c_lflag */
	DEF_LINE,					/* c_line  */
	{						/* c_cc[]  */
		0, 0, 0, 0, 0, 0, 0, 0,
	},
};

/*
 *	conf.h
 */
extern	struct	m332_dev	*m332_addr[];
extern	int			m332_ilev[];
extern	int			m332_ivec[];
extern	int			m332_cnt;
extern int 			m332_cmaj;


/*
 *	MVME332xt Procedures
 *
 *	m332init: initialize the mvme332xt and the driver
 */
m332init()		/* called by system initialization */
{
	m332major = m332_cmaj; /* set major number for SXT */
}

m332_init()		/* called by the first open */
{
	register struct m332_dev *dp;
	register struct cb *cb;
	register struct b_envelope *ep;
	register struct b_packet *pkt;
	register struct dev_info *iptr;
	register unsigned ctl, dev;

	unsigned long	timeout;
	unsigned long	bufsize = sizeof(struct m332_dev);
	unsigned char	dual_language;

	DEBUGPR(DBINIT,0,"init");
#ifdef	M68020
	m332adrm = DATA32;
	m332busw = 2;
#else
	m332adrm = DATA24;
	m332busw = 1;
#endif

#ifndef REL2
	dual_language = greek;
#else
	dual_language = 0;
#endif
	for (ctl = 0; ctl < (m332_cnt/MAXDEV); ctl++) {
		/*
		 * Make sure everything fits in the shared ram
		 */
		if (bufsize > SHARE_SIZE-VEC_SIZE) {
			DEBUGPR(DBINIT, DEV(ctl, 0),
			    "ring buffer allocation exceeds 64k");
			m332_addr[ctl]= NULL;
			continue;
		}

		if (bprobe(dp = (m332_addr[ctl]), -1)) {	/* read only */
			DEBUGPR(DBINIT, DEV(ctl, 0), "Controller not present");
			m332_addr[ctl]= NULL;
			continue;
		}

		/* wait for controller to be ready */
		for (timeout = IPC_TIMEOUT; timeout && (dp->ipc.m_cr&IPC_BUSY);
			timeout--);
		if (timeout == 0) {
			DEBUGPR(DBINIT, DEV(ctl, 0), "Controller never ready");
			m332_addr[ctl]= NULL;
			continue;
		}

		/* set up free queue of envelope/packet pairs */
		bppcrque(&dp->freepkt, NPKTS, dp->pkt, dp->env,
			0x4d333332, m332adrm, m332busw);

		/* set up IPC channel */
		if (bppcrchan(&dp->channel, &dp->ipc, &dp->freepkt,
			m332_ilev[ctl], m332_ivec[ctl]/4, m332adrm) == -1) {
			/* create channel error, disable controller */
			m332_addr[ctl]= NULL;
			m332print(DEV(ctl, 0), "Create channel error");
			printf("MVME332xt: Controller %x disabled\n", ctl);
			continue;
		}

		/* set to indicate a valid controller */
		m332_Addr[ctl] = dp;
		/*
		 *	send an INIT packet	
		 */
		ep = m332bppget(&dp->freepkt);
		pkt = ep->e_pkt;
		pkt->p_cmd = INIT;
#ifdef m88k
		pkt->pa_pp_blknohi = ((int) dp->misc.init_info) >> 16;
		pkt->pa_pp_blknolo = (int) dp->misc.init_info;
#else
		pkt->p_init_info = (int) dp->misc.init_info;
#endif

		/* initialize information for port 0 */
		cb   = &dp->cb[0];
		iptr = &dp->misc.init_info[0];
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring0;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring0;
		iptr->w_size = WR_SIZE0;
		cb->w_size   = WR_SIZE0 - 1;
		iptr->r_size = RD_SIZE0;
		cb->r_size   = RD_SIZE0 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 1 */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring1;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring1;
		iptr->w_size = WR_SIZE1;
		cb->w_size   = WR_SIZE1 - 1;
		iptr->r_size = RD_SIZE1;
		cb->r_size   = RD_SIZE1 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 2 */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring2;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring2;
		iptr->w_size = WR_SIZE2;
		cb->w_size   = WR_SIZE2 - 1;
		iptr->r_size = RD_SIZE2;
		cb->r_size   = RD_SIZE2 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 3 */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring3;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring3;
		iptr->w_size = WR_SIZE3;
		cb->w_size   = WR_SIZE3 - 1;
		iptr->r_size = RD_SIZE3;
		cb->r_size   = RD_SIZE3 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 4 */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring4;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring4;
		iptr->w_size = WR_SIZE4;
		cb->w_size   = WR_SIZE4 - 1;
		iptr->r_size = RD_SIZE4;
		cb->r_size   = RD_SIZE4 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 5 */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring5;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring5;
		iptr->w_size = WR_SIZE5;
		cb->w_size   = WR_SIZE5 - 1;
		iptr->r_size = RD_SIZE5;
		cb->r_size   = RD_SIZE5 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 6 */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring6;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring6;
		iptr->w_size = WR_SIZE6;
		cb->w_size   = WR_SIZE6 - 1;
		iptr->r_size = RD_SIZE6;
		cb->r_size   = RD_SIZE6 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 7 */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring7;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring7;
		iptr->w_size = WR_SIZE7;
		cb->w_size   = WR_SIZE7 - 1;
		iptr->r_size = RD_SIZE7;
		cb->r_size   = RD_SIZE7 - 1;
		bcopy(&m332deftio, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* initialize information for port 8 : printer device */
		cb++;
		iptr++;
		iptr->w_ring = cb->w_ring = (struct ring *) &dp->w_ring8;
		iptr->r_ring = cb->r_ring = (struct ring *) &dp->r_ring8;
		iptr->w_size = WR_SIZE8;
		cb->w_size   = WR_SIZE8 - 1;
		iptr->r_size = RD_SIZE8;
		cb->r_size   = RD_SIZE8 - 1;
		bcopy(&m332defprt, &iptr->def_tio, sizeof(struct termio));
		iptr->dual = dual_language;

		/* send init packet to the MVME332xt */
		pkt->p_ilev = m332_ilev[ctl];
#ifdef m88k
		pkt->p_counthi = (m332_ivec[ctl]/4) >> 16;
		pkt->p_countlo = m332_ivec[ctl]/4;
#else
		pkt->p_ivec = m332_ivec[ctl]/4;
#endif
		cb = &dp->cb[0];
#ifdef m88k
		pkt->p_numhi = ((unsigned) cb) >> 16;
		pkt->p_numlo = (unsigned) cb;
#else
		pkt->p_cb = (unsigned) cb;
#endif

		m332send(dp, ep);

		if (cb->stat != 0) {
			m332print(DEV(ctl, 0), "Initialization error");
			printf("MVME332xt: Controller %x disabled\n", ctl);
			/* init error, disable controller */
			m332_addr[ctl] = NULL;
			m332_Addr[ctl] = NULL;
			continue;
		}
		/*
		 *	send  EVENT packet to each port 
		 */
		for(dev = 0; dev< NDEV_PER_CTL; dev++, cb++) {
			/* create a write_wakeup envelope */
			ep = m332bppget(&dp->freepkt);
			cb->write_env = ep;
			pkt = ep->e_pkt;
#ifdef m88k
			pkt->p_numhi = ((unsigned) cb) >> 16;
			pkt->p_numlo = (unsigned) cb;
#else
			pkt->p_cb = (unsigned)cb;
#endif
			pkt->p_cmd = W_WAKEUP;
			pkt->p_unit = dev;

			/* create a read_wakeup envelope */
			ep = m332bppget(&dp->freepkt);
			cb->read_env = ep;
			pkt = ep->e_pkt;
#ifdef m88k
			pkt->p_numhi = ((unsigned) cb) >> 16;
			pkt->p_numlo = (unsigned) cb;
#else
			pkt->p_cb = (unsigned)cb;
#endif
			pkt->p_cmd = R_WAKEUP;
			pkt->p_unit = dev;

			/* send an event envelope to MVME332xt */
			ep = m332bppget(&dp->freepkt);
			pkt = ep->e_pkt;
			pkt->p_cmd = EVENT;
			pkt->p_unit = dev;
#ifdef m88k
			pkt->p_numhi = ((unsigned) cb) >> 16;
			pkt->p_numlo = (unsigned) cb;
#else
			pkt->p_cb = (unsigned)cb;
#endif
			bppsend(&dp->channel, &dp->ipc, ep);
		}
	}
}


/*
 *	m332open: open an mvme332xt device
 */
m332open( dev,flag)
register unsigned dev;
register flag;
{
#define	NEVER_INITED	0
#define	IN_INIT		1
#define	INIT_DONE	2

	register struct b_envelope *ep;
	register struct b_packet *pkt;
	register struct proc *pp;
	register struct cb *cb;
	struct	m332_dev *dp;
	unsigned short   unit;

	static	short	m332_first_init;

	DEBUGPR(DBOPEN, dev, "Open");
	if (m332_first_init == NEVER_INITED) {
		m332_first_init = IN_INIT;
		m332_init();
		m332_first_init = INIT_DONE;
		wakeup(&m332_first_init);
	}

	while (m332_first_init == IN_INIT)
		sleep(&m332_first_init, TTIPRI);

	/*
	 *	if the device is nonexistent, then error
	 */
	unit = UNIT(dev);
	if(((dp = m332_Addr[CTL(dev)]) == NULL) || (unit > NDEV_PER_CTL)) {
		u.u_error = ENXIO;
		return;
	}
	/*
	 *	send open packet to MVME332xt
	 */
	cb = &dp->cb[unit];
	ep = m332bppget(&dp->freepkt);
	pkt = ep->e_pkt;
	pkt->p_cmd = OPEN;
	pkt->p_unit = unit;
	pkt->p_iomode = flag;
#ifdef m88k
	pkt->p_numhi = ((unsigned) cb) >> 16;
	pkt->p_numlo = (unsigned) cb;
#else
	pkt->p_cb = (unsigned)cb;
#endif

	m332send(dp, ep);

	if (!(flag & FNDELAY)) {
		splhi();
		while (!(cb->stat & DEV_READY))
			sleep(&cb->stat, TTIPRI);
		spl0();
	}

	pp = u.u_procp;
	if (pp->p_pid == pp->p_pgrp && u.u_ttyp == NULL &&
	    cb->pgrp[cb->sxtcon] == 0) {
		u.u_ttyp = &cb->pgrp[cb->sxtcon];
		cb->pgrp[cb->sxtcon] = pp->p_pgrp;
	}
	DEBUGPR(DBOPEN, dev, "Open successful");
}


/*
 *	m332close: close an mvme332xt device
 */
m332close(dev, flag)
register unsigned dev;
int	flag;
{
	register struct b_envelope *ep;
	register struct b_packet *pkt;
	register struct cb *cb;
	struct	m332_dev *dp;
	unsigned short   unit;

	DEBUGPR(DBCLOSE, dev, "Close");
	unit = UNIT(dev);
	dp = m332_Addr[CTL(dev)];
	/*
	 *	send CLOSE packet to MVME332xt
	 */
	cb = &dp->cb[unit];
	ep = m332bppget(&dp->freepkt);
	pkt = ep->e_pkt;
	pkt->p_cmd = CLOSE;
#ifdef m88k
	pkt->p_counthi = flag) >> 16;
	pkt->p_countlo = flag;
#else
	pkt->p_iomode = flag;
#endif
	pkt->p_unit = unit;
#ifdef m88k
	pkt->p_numhi = ((unsigned) cb) >> 16;
	pkt->p_numlo = (unsigned) cb;
#else
	pkt->p_cb = (unsigned)cb;
#endif

	splhi();
	bppsend(&dp->channel, &dp->ipc, ep);
	sleep(pkt, PZERO);
	spl0();
	DEBUGPR(DBCLOSE, dev, "Close successful");
}


/*
 *	m332read: read
 */
m332read( dev )
unsigned dev;
{
	register struct cb *cb;
	register struct ring *ring;
	register int	put, get;
	register int	framecnt, copycnt;
	struct	m332_dev *dp;
	int 	 s;

	DEBUGPR(DBREAD, dev, "Read");
	dp = m332_Addr[CTL(dev)];
	cb = &dp->cb[UNIT(dev)];
	ring = cb->r_ring;		/* get read ring pointer */

	while (u.u_count) {
		put = ring->put;
		get = ring->get;
		if (put == get) {	/* ring is empty */
			s = splhi();
			/* if no read wakeup packet pending, send one */
			if (!(cb->stat & R_PENDING)) {
				cb->stat |= R_PENDING;
				bppsend(&dp->channel, &dp->ipc, cb->read_env);
			}
			if (u.u_fmode & FNDELAY) {	/* return if NO_DELAY */
				splx(s);
				break;
			}
			sleep(&cb->rsleep[cb->sxtcon], TTIPRI);
			splx(s);

			if (cb->read_env->e_pkt->p_error) {
				cb->stat &= ~DEV_READY;
				return;
			}
			continue;
		}
		framecnt = ring->buf[get++];
		get &= cb->r_size;

		copycnt = minimum(u.u_count, get > put ?
			  minimum(cb->r_size+1-get, framecnt) : framecnt);
		if (copycnt) {
			/* call copyout to read in one contiguous data block */
			COPYOUT(get + ring->buf, u.u_base, copycnt);
			u.u_count -= copycnt;
			u.u_base += copycnt;
			framecnt -= copycnt;
		}
		if (framecnt == 0) {   /* all available char are read */
			get += copycnt;
			get &= cb->r_size;
			ring->get = (get + 1) & cb->r_size;
			if (ring->buf[get] & RD_DEL)
				break;	/* return if there is a delimiter */
		}
		else {	/* adjust the record framecnt for future read */
			ring->get = (get + copycnt-1) & cb->r_size;
			ring->buf[ring->get] = framecnt;
		}
	}
	DEBUGPR(DBREAD, dev, "Read successful");
}


/*
 *	m332write: write
 */
m332write( dev )
unsigned dev;
{
	register struct cb *cb;
	register struct ring *ring;
	register int 	put, get, copycnt;
	struct	m332_dev *dp;
	int    s;

	DEBUGPR(DBWRITE, dev, "Write");
	dp = m332_Addr[CTL(dev)];
	cb = &dp->cb[UNIT(dev)];
	ring = cb->w_ring;		/* get write ring pointer */

	while (u.u_count) {
		if (!(cb->stat & DEV_READY))
			return;
		put = ring->put;
		get = ring->get;
		if (((put+1) & cb->w_size) == get) {
			s = splhi();
			/* if no write wakeup packet pending, send one */
			if (!(cb->stat & W_PENDING)) {
				cb->stat |= W_PENDING;
				bppsend(&dp->channel, &dp->ipc, cb->write_env);
			}
			sleep(&cb->wsleep[0], TTOPRI);
			splx(s);

			if (cb->write_env->e_pkt->p_error) {
				cb->stat &= ~DEV_READY;
				return;
			}
			continue;
		}
		if (put < get)
			copycnt = minimum(u.u_count, get - put - 1);
		else
			copycnt = (get ? minimum(u.u_count, cb->w_size+1 - put)
				       : minimum(u.u_count, cb->w_size - put));

		/* call copyin to write out one contiguous data block */
		COPYIN(u.u_base, put + ring->buf, copycnt);
		u.u_count -= copycnt;
		u.u_base += copycnt;
		ring->put = (put+copycnt) & cb->w_size;
	}
	DEBUGPR(DBWRITE, dev, "Write successful");
}


/*
 *	m332intr: interrupt handler
 */
m332intr(dev)
unsigned dev;
{
	register struct b_envelope *ep;
	register struct b_packet *pkt;
	register struct cb *cb;
	register struct	m332_dev *dp;
	register int	rcode;

	DEBUGPR(DBINTR, dev, "Intr");
	if ((dp = m332_Addr[CTL(dev)]) == NULL) {
		m332print(dev, "Unknown interrupt");
		return;
	}

	for(;;) {
		/* get an envelope from channel status pipe */
		ep = dp->channel.c_stat.p_head;
		if( ep->e_valid == INVALID )	/* end of pipe */
			break;

		/* invalidate and dequeue the envelope */
		ep->e_valid = INVALID;
		dp->channel.c_stat.p_head = ep->e_link;

		pkt = ep->e_pkt;
		if (pkt->p_state == P_STALE) {
			pkt->p_state = P_DONE;
			m332bppfree(&dp->freepkt, ep);
			continue;
		}
		pkt->p_state = P_DONE;
		cb = (struct cb *) pkt->p_cb;

		switch(pkt->p_cmd) {

		case INIT:
			cb->stat = pkt->p_error;
			m332bppfree(&dp->freepkt, ep);
			wakeup(pkt);
			break;

		case R_WAKEUP:
			cb->stat &= ~R_PENDING; /* clear wakeup pkt pending */
			cb->read_env = ep;
			wakeup(&cb->rsleep[cb->sxtcon]);
#ifdef CB_SELECT
			if (cb->t_rsel) {
				selwakeup(cb->t_rsel, cb->stat & CB_RCOLL);
				cb->stat &= ~CB_RCOLL;
				cb->t_rsel = 0;
			}
#endif
			break;

		case W_WAKEUP:
			cb->stat &= ~W_PENDING; /* clear wakeup pkt pending */
			cb->write_env = ep;
			wakeup(&cb->wsleep[0]);
#ifdef CB_SELECT
			if (cb->t_wsel) {
				selwakeup(cb->t_wsel, cb->stat & CB_WCOLL);
				cb->stat &= ~CB_WCOLL;
				cb->t_wsel = 0;
			}
#endif
			break;

		case CLOSE:
			if (pkt->p_retcode & (E_DCD|E_PR_SELECT)) {
				if (!(cb->stat & DEV_READY)) {
					cb->stat |= DEV_READY;
					wakeup(&cb->stat);
				}
			}
			else
				cb->stat &= ~DEV_READY;
			m332bppfree(&dp->freepkt, ep);
			cb->pgrp[cb->sxtcon] = 0;
			wakeup(pkt);
			break;

		case OPEN:
			if (pkt->p_retcode & (E_DCD|E_PR_SELECT)) {
				if (!(cb->stat & DEV_READY)) {
					cb->stat |= DEV_READY;
					wakeup(&cb->stat);
				}
			}
			else
				cb->stat &= ~DEV_READY;
			m332bppfree(&dp->freepkt, ep);
			wakeup(pkt);
			break;

		case IOCTL:
#ifdef	m88k
			pkt->p_numhi = (long) ep >> 16; /* save envelope */
			pkt->p_numlo = (long) ep; /* save returned envelope */
#else
			pkt->p_rtn_env = (long) ep; /* save returned envelope */
#endif
			wakeup(pkt);
			break;

		case EVENT:
			rcode = pkt->p_retcode;

			if (rcode & E_INTR)
				signal(cb->pgrp[cb->sxtcon], SIGINT);

			if (rcode & E_QUIT)
				signal(cb->pgrp[cb->sxtcon], SIGQUIT);

			if (rcode & E_HUP)
				signal(cb->pgrp[0], SIGHUP);

			if (rcode & (E_DCD|E_PR_SELECT)) {
				cb->stat |= DEV_READY;
				wakeup(&cb->stat);
			}
			else
				if (rcode & E_NO_DCD)
					cb->stat &= ~DEV_READY;

			if (rcode & E_SWTCH) {
				cb->sxtcon = 0; /* set control to channel 0 */
				wakeup(&cb->pgrp[0]);
			}

			m332pr_chk(dev | pkt->p_unit, rcode);
			bppsend(&dp->channel, &dp->ipc, ep);
			break;
		}
	}
}


/*
 *	m332ioctl: ioctl support for the MVME332xt
 */
m332ioctl( dev, cmd, arg, mode )
register unsigned dev, cmd, arg, mode;
{
	register struct b_envelope *ep;
	register struct b_packet *pkt;
	register struct cb *cb;
	int	count, size;
	unsigned int	from, dest;
	struct dl_info	*dl_info;
	struct	m332_dev *dp;
	unsigned short  unit;

	DEBUGPR(DBIOCTL, dev, "Ioctl");
	unit = UNIT(dev);
	dp = m332_Addr[CTL(dev)];

	cb = &dp->cb[unit];
	ep = m332bppget(&dp->freepkt);
	pkt = ep->e_pkt;
	pkt->p_cmd = IOCTL;
	pkt->p_unit = unit;
#ifdef	m88k
	pkt->pa_pp_blknohi = cmd >> 16;	/* IOCTL command */
	pkt->pa_pp_blknolo = cmd;	/* IOCTL command */
#else
	pkt->p_iocmd = cmd;	/* IOCTL command */
#endif
#ifdef	m88k
	pkt->p_sahi = arg >> 16;	/* IOCTL arg */
	pkt->p_salo = arg;	/* IOCTL arg */
#else
	pkt->p_ioarg = arg;	/* IOCTL arg */
#endif
#ifdef	m88k
	pkt->p_counthi = mode >> 16;
	pkt->p_countlo = mode;
#else
	pkt->p_iomode = mode;	/* IOCTL mode */
#endif
#ifdef m88k
	pkt->p_numhi = ((unsigned) cb) >> 16;
	pkt->p_numlo = (unsigned) cb;
#else
	pkt->p_cb = (unsigned)cb;
#endif

	switch(cmd) {

	case TCSETA:	/* set to new termio */
	case TCSETAW:
	case TCSETAF:
	case TCSETDF:	/* set default open termio */
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct termio));
#else
		COPYIN(arg, &pkt->p_termio, sizeof(struct termio));
#endif
		break;

	case TIOCSETP:
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct termio));
#else
		COPYIN(arg, &pkt->p_sgttyb, sizeof(struct sgttyb));
#endif
		break;

	case LDSETT:
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct termio));
#else
		COPYIN(arg, &pkt->p_termcb, sizeof(struct termcb));
#endif
		break;

	case TCDLOAD:	/* download object code to MVME332xt	*/
		if (!suser())
		 	break;
		/* IOCTL arg points to dl_info structure */
#ifdef m88k
		dl_info = (struct dl_info *) &pkt->pp_psahi;
#else
		dl_info = (struct dl_info *) &pkt->p_param;
#endif
		COPYIN(arg, dl_info, sizeof(struct dl_info));
		from = dl_info->host_addr;
		count = dl_info->count;
		lock_dlbuf(dp);		/* get access to download buffer */
		dl_info->host_addr = (unsigned long)(dp->misc.dl_buf);
		while (count > 0) {
			size = minimum(count, DL_BUF_SIZE);
			COPYIN(from, dp->misc.dl_buf, size);
			dl_info->count = size;	/* transfer size */
			m332send(dp, ep);
#ifdef m88k
			ep = (struct b_envelope *) ((pkt->p_numhi << 16) + pkt->p_numlo);
#else
			ep = (struct b_envelope *) pkt->p_rtn_env;
#endif
			if (pkt->p_error)
				break;
			count -= size;
			from += size;
			dl_info->ipc_addr += size;
		}
		unlock_dlbuf(dp);		/* release download buffer */
		m332bppfree(&dp->freepkt, ep);
		return;

	case TCGETSYM:	/* get symbol table from the MVME332xt	*/
	case TCWHAT:	/* perform what function */
		if (!suser())
			break;
		/* IOCTL arg points to dl_info structure */
#ifdef m88k
		dl_info = (struct dl_info *) &pkt->pp_psahi;
#else
		dl_info = (struct dl_info *) &pkt->p_param;
#endif
		COPYIN(arg, dl_info, sizeof(struct dl_info));
		dest = dl_info->host_addr;	/* user buf address */
		count = dl_info->count;		/* user buffer size */	
		lock_dlbuf(dp);		/* get access to download buffer */
		dl_info->host_addr = (unsigned long)(dp->misc.dl_buf);
		dl_info->count = minimum(dl_info->count, DL_BUF_SIZE);
		break;

	case TCLINE:	/* load line discipline to MVME332xt	*/
		if (!suser())
			break;
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYIN(arg, &pkt->p_param, sizeof(struct dl_info));
#endif
		break;

	case TCEXEC:
		if (!suser())
			break;
#ifdef m88k
		dl_info = (struct dl_info *) &pkt->pp_psahi;
#else
		dl_info = (struct dl_info *) &pkt->p_param;
#endif
		dl_info->ipc_addr = arg;
		break;

	case TCGETA:	/* get current termio */
	case TIOCGETP:	/* get current sgttyb */
	case LDGETT:	/* get current termcb */
	case TCSBRK:	/* send break */
	case TCXONC:	/* control xon/xoff */
	case TCFLSH:	/* flush queues */
	case TCSETHW:	/* set hardware flow control option */
	case TCGETHW:	/* get hardware flow control option */
	case TCGETVR:	/* get firware version and revision numbers */
	case TCGETDL:	/* get start addr and size of download area */
	case TCGETDF:	/* get default open termio data */
		break;
#ifdef	ISP
	case CSGETTT:	/* request translate table data */
	{
		register int		to;
		struct	csttiohdr	iohdr;

		COPYIN(arg, &iohdr, sizeof(iohdr));
		to = arg + (int)(&iohdr.cs_h) - (int)(&iohdr);
		count = iohdr.cs_tmax;

		/* IOCTL arg points to dl_info structure */
#ifdef m88k
		dl_info = (struct dl_info *) &pkt->pp_psahi;
#else
		dl_info = (struct dl_info *) &pkt->p_param;
#endif
		dl_info->id = iohdr.cs_h.cs_tnum;
		lock_dlbuf(dp);		/* get access to download buffer */
		dl_info->host_addr = (unsigned long)(dp->misc.dl_buf);
		dl_info->ipc_addr = 0;
		while (count > 0) {
			size = minimum(count, DL_BUF_SIZE);
			dl_info->count = size;	/* transfer size */
			m332send(dp, ep);
#ifdef m88k
			ep = (struct b_envelope *) ((pkt->p_numhi << 16) + pkt->p_numlo);
#else
			ep = (struct b_envelope *) pkt->p_rtn_env;
#endif
			if (pkt->p_error)
				break;
			size = dl_info->count;
			if (size == 0) {
				if (dl_info->ipc_addr == 0) {
					iohdr.cs_h.cs_tlen = 0;
					COPYOUT(&iohdr, arg, sizeof(iohdr));
				}
				break;
			}
			COPYOUT(dl_info->host_addr, to, size);
			count -= size;
			to += size;
		}
		unlock_dlbuf(dp);		/* release download buffer */
		m332bppfree(&dp->freepkt, ep);
		return;
	}

	case CSSETTT:	/* request translate table data */
	{
		struct	csttiohdr	iohdr;

		if (!suser()) 		/* only super user can do this */
			break;
		COPYIN(arg, &iohdr, sizeof(iohdr));
#ifdef m88k
		dl_info = (struct dl_info *) &pkt->pp_psahi;
#else
		dl_info = (struct dl_info *) &pkt->p_param;
#endif
		dl_info->id = iohdr.cs_h.cs_tnum;
		dl_info->ucount = iohdr.cs_h.cs_tlen;

		if (iohdr.cs_h.cs_tlen == 0) 	/* remove the table */
			break;			/* send packet as normal */
		from = arg + (int)(&iohdr.cs_h) - (int)(&iohdr);
		count = iohdr.cs_h.cs_tlen;

		/* IOCTL arg points to dl_info structure */
		lock_dlbuf(dp);		/* get access to download buffer */
		dl_info->host_addr = (unsigned long)(dp->misc.dl_buf);
		dl_info->ipc_addr = 0;
		while (count > 0) {
			size = minimum(count, DL_BUF_SIZE);
			dl_info->count = size;	/* transfer size */
			COPYIN(from, dl_info->host_addr, size);
			m332send(dp, ep);
#ifdef m88k
			ep = (struct b_envelope *) ((pkt->p_numhi << 16) + pkt->p_numlo);
#else
			ep = (struct b_envelope *) pkt->p_rtn_env;
#endif
			if (pkt->p_error)
				break;
			count -= size;
			from += size;
		}
		unlock_dlbuf(dp);		/* release download buffer */
		m332bppfree(&dp->freepkt, ep);
		return;
	}

	case CSGETO:	/* request character set option */
		break;

	case CSSETO:	/* change character set option */
	case CSSETOW:	/* change character set option after wait */
	case CSSETOF:	/* change character set option after wait, then flush */
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYIN(arg, &pkt->p_csopt, sizeof(struct csopt));
#endif
		break;
#endif
	case TCSETAS:	/* set termio with arg from the system space */
		pkt->p_iocmd = TCSETA;	/* IOCTL command */
		bcopy(arg, &pkt->p_termio, sizeof(struct termio));
		break;

	default:
		DEBUGPR(DBIOCTL, dev, "Unknown ioctl command");
		u.u_error = ENXIO;
		break;
	}

	if (u.u_error != 0) {	/* if error, free packet then return */
		m332bppfree(&dp->freepkt, ep);
		return;
	}

	m332send(dp, ep);	/* send ioctl packet to controller */

	switch(cmd) {

	case TCGETA:
	case TCGETDF:	/* get default open termio data */
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYOUT(&pkt->p_termio, arg, sizeof(struct termio));
#endif
		break;

	case TIOCGETP:
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYOUT(&pkt->p_sgttyb, arg, sizeof(struct sgttyb));
#endif
		break;

	case LDGETT:
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYOUT(&pkt->p_termcb, arg, sizeof(struct termcb));
#endif
		break;

	case TCGETVR:	/* get firmware version and revision numbers */
		pkt->p_param |= (7 << 24) + (5 << 16);
		/* fall through */

	case TCGETHW:	/* get hardware flow control option */
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYOUT(&pkt->p_param, arg, sizeof(long));
#endif
		break;

	case TCGETDL:	/* get download area address and size */
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYOUT(&pkt->p_param, arg, sizeof(struct dl_info));
#endif
		break;

	case TCGETSYM:	/* get symbol table from the MVME332xt	*/
	case TCWHAT:	/* perform what function */
		for (;;) {
			COPYOUT(dl_info, arg, sizeof(struct dl_info));
			COPYOUT(dp->misc.dl_buf, dest, dl_info->count);
			break;
		}
		unlock_dlbuf(dp);	/* release download buffer */
		break;
#ifdef	ISP
	case CSGETO:	/* request character set option */
#ifdef m88k
		COPYIN(arg, &pkt->pp_psahi, sizeof(struct dl_info));
#else
		COPYOUT(&pkt->p_csopt, arg, sizeof(struct csopt));
#endif
		break;
#endif
	}
#ifdef m88k
	m332bppfree(&dp->freepkt, (pkt->p_numhi << 16) + pkt->p_numlo);
#else
	m332bppfree(&dp->freepkt, pkt->p_rtn_env);
#endif
	DEBUGPR(DBIOCTL, dev, "Ioctl successful");
}


/*
 * m332send: 	send packet to MVME332xt, wait for completion.
 *		longjump to u.u_qsav if it catches any signal.
 */
m332send(dp, ep)
register struct m332_dev *dp;
register struct b_envelope *ep;
{
	register struct b_packet *pkt;
	register struct b_envelope *tp;
	int  s;

	pkt = ep->e_pkt;
	pkt->p_state = P_BUSY;		/* mark the packet busy */
	ep->e_valid = INVALID;		/* make this env become a null env */

	pkt->p_cmdp = dp->channel.c_chan; /* plug channel number into packet */
	s = splhi();			/* mask intr since we access channel */
	tp = dp->channel.c_cmd.p_tail;	/* pointer to the old null envelope */
	tp->e_pkt = pkt;		/* plug the packet into old null env */
	tp->e_link = ep;		/* link the new null env */
	dp->channel.c_cmd.p_tail = ep;	/* update the command pipe tail */
	tp->e_valid = VALID;		/* valid the old null envelope */
	dp->ipc.m_cr |= B_ATTN;		/* tell controller */

	while (pkt->p_state != P_DONE) {
		if (sleep(pkt, TTIPRI | PCATCH)) {
			if (pkt->p_state == P_DONE)
#ifdef m88k
				m332bppfree(&dp->freepkt, (pkt->p_numhi << 16) + pkt->p_numlo);
#else
				m332bppfree(&dp->freepkt, pkt->p_rtn_env);
#endif
			else
				pkt->p_state = P_STALE;
			u.u_error = EINTR;	/* return error */
			longjmp(u.u_qsav);	/* never return here */
		}
	}
	u.u_error = pkt->p_error;
	splx(s);
}

/*
 *	m332bppget - get free env from que
 */
struct b_envelope *
m332bppget(freepkt)
struct b_envelope **freepkt;
{
	register struct b_envelope *ep;
	int  s;

	s = splhi();
	while (*freepkt == NULL) {
		if (sleep(freepkt, TTIPRI | PCATCH)) {
			m332prm("Free packet pool is empty");
			u.u_error = EINTR;
			longjmp(u.u_qsav);	/* never return here */
		}
	}
	ep = *freepkt;
	*freepkt = ep->e_link;
	ep->e_link = NULL;
	splx(s);
	return(ep);
}

/*
 *	m332bppfree - free envelope and put on queue
 */
m332bppfree(freepkt, ep)
register struct b_envelope *ep;
struct b_envelope **freepkt;
{
	int  s;

	s = splhi();
	ep->e_link = *freepkt;
	*freepkt = ep;
	if (ep->e_link == 0)
		wakeup(freepkt);
	splx(s);
}

/*
 *	m332print: driver print procedure
 */
m332print(d,s)
int d, s;
{
	printf("MVME332xt: %s on controller %d, unit %d\n", s,CTL(d),UNIT(d));
}

/*
 *	m332prm: print message with MVME: at front
 */
m332prm(s)
char *s;
{
	printf("MVME332xt: %s.\n", s );
}

m332pr_chk(dev, rcode)
int	dev;
unsigned short rcode;
{
	switch(rcode & (E_PR_SELECT|E_PR_FAULT|E_PR_POUT)) {

	case E_PR_FAULT:
		m332print(dev, "PRINTER is deselected");
		break;

	case E_PR_FAULT|E_PR_POUT:
		m332print(dev, "PRINTER is out of paper");
		break;

	case E_PR_FAULT|E_PR_SELECT:
		m332print(dev, "PRINTER fault for unknown reason");
		break;
	}
}

/* console putchar routines */
m332putc(dev, c)
unsigned dev;
unsigned char c;
{
	register struct m332_dev *dp;
	int s;

	if ((dp = m332_Addr[CTL(dev)]) == NULL)
		return(-1);
	s = splhi();
	while(dp->ipc.m_tas&TASBIT);	/* gain exclusive access to device */
	dp->ipc.m_tas |= TASBIT;	/* not safe like a real TAS */

	/* set up register file */
#ifdef m88k
	dp->ipc.m_addrhi = (unsigned)((UNIT(dev) << 8) | c) >> 16;
	dp->ipc.m_addrlo = (unsigned)((UNIT(dev) << 8) | c);
#else
	dp->ipc.m_addr = (unsigned)((UNIT(dev) << 8) | c);
#endif
	dp->ipc.m_tas = (dp->ipc.m_tas&TASMASK) | VCMD | CPUTC;
	dp->ipc.m_cr |= B_ATTN;	/* tell controller that command is ready */

	while((dp->ipc.m_tas&VSTAT) == 0);	/* wait for valis status */
	dp->ipc.m_tas |= CCOMP;			/* indicate command complete */
	dp->ipc.m_cr |= B_ATTN;			/* notify controller */
	splx(s);
	return(0);
}

/*
 * lock_dlbuf: get access to the down load buffer.
 */
lock_dlbuf(dp)
register struct m332_dev *dp;
{
	while (dp->buf_busy) {	/* get access to download buffer */
		if (sleep(&dp->buf_busy, TTOPRI | PCATCH)) {
			dp->buf_busy = 0;
			u.u_error = EINTR;
			longjmp(u.u_qsav);
		}
	}
	dp->buf_busy = 1;
}

/*
 * unlock_dlbuf: release the down load buffer.
 */
unlock_dlbuf(dp)
register struct m332_dev *dp;
{
	dp->buf_busy = 0;
	wakeup(&dp->buf_busy);
}

#ifdef CB_SELECT
m332select(dev, rw)
	unsigned dev;
	int rw;
{
	int s = spltty();
	extern selwait;
	register struct cb *cb;
	register struct ring *ring;
	register int	put, get;
	struct	m332_dev *dp;

	dp = m332_addr[CTL(dev)];
	ASSERT(dp);

	cb = &dp->cb[UNIT(dev)];

	switch (rw) {
	    case FREAD:
		ring = cb->r_ring;	/* read ring pointer */
		put = ring->put;
		get = ring->get;
		if (put != get) 	/* ring buffer has something */
			goto win;

		/* if no read wakeup packet pending, send one */
		if(!(cb->stat & R_PENDING)) {
			cb->stat |= R_PENDING;
			bppsend(&dp->channel, &dp->ipc, cb->read_env);
		}

		if (cb->t_rsel && cb->t_rsel->p_wchan == (caddr_t)&selwait)
			cb->stat |= CB_RCOLL;
		else
			cb->t_rsel = u.u_procp;
		break;

	    case FWRITE:
		ring = cb->w_ring;	/* write ring pointer */
		put = ring->put;
		get = ring->get;
		if (((put+1) & cb->w_size) != get)
			goto win;

		/* if no write wakeup packet pending, send one */
		if (!(cb->stat & W_PENDING)) {
			cb->stat |= W_PENDING;
			bppsend(&dp->channel, &dp->ipc, cb->write_env);
		}

		if (cb->t_wsel && cb->t_wsel->p_wchan == (caddr_t)&selwait)
			cb->stat |= CB_WCOLL;
		else
			cb->t_wsel = u.u_procp;
		break;
	}
	splx(s);
	return (0);
win:
	splx(s);
	return (1);
}
#endif
