/*		Copyright (c) 1987, 1988 Motorola			*/
/*		All Rights Reserved				*/
/*		@(#)kern-port:io/mvme336sio.c	35.6	
/*
 */
/*
 *      Driver for MVME336 SIO Cluster controller
 *
 *	Features:
 *	
 * 		handles 96 terminals in one swell foop!
 *		all data transmitted and received via xpc hyper link
 *
 */

#ident "@(#)kern-port:io/mvme336sio.c	35.6"
#undef DEBUG
#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/fs/s5dir.h"
#include "sys/file.h"
#include "sys/signal.h"
#include "sys/tty.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/sysinfo.h"
#include "sys/mvme336s.h"
#include "sys/mvme336c.h"
#include "sys/immu.h"
#include "sys/mvme336t.h"

/* redefine spl6 to be used by the s336 */
#define spls336 spl6

/* the s336 driver is supposed to handle 96 ports across 6 clusters */
#define SIO_MAX 96

/* number of devices - set in conf.c */
extern int s336_cnt;

/* s336 tty structure, set in conf.c */
extern struct tty s336_tty[];            

/* an s336 control structure that is allocated on a per s336 port basis */
struct sio_ctrl sio_ctrl[SIO_MAX];

/* number of characters to wait for before doing input processing */
int max_intrx = (CLSIZE);

/* number of characters before using bcopy */
int max_rxcnt = 64;

/* number of characters to process before a rx ack is sent */
/* this count must be 1 less that the s336xolo count */
int max_rxack = 29;

/* cluster high water mark */
int s336xohi = 50;

/* cluster low water mark */
int s336xolo = 30;

/* number od ticks to wait on outstanding packet gross timeout */
int maxclose = 208;


#ifdef TRACE

int s336_trace = 0;

extern chktrc();

#define STRC(id,a,b,c,d,e,f,g) if(s336_trace){chktrc(id,a,b,c,d,e,f,g);}

#else

#define STRC(id,a,b,c,d,e,f,g) 

#endif

/* RTS on */
UCHAR rts_on = RTS_ON;

/* RTS off */
UCHAR rts_off = RTS_OFF;

/* DTR on */
UCHAR dtr_on = DTR_ON;

/* DTR off */
UCHAR dtr_off = DTR_OFF;

/* the SCC-2698 Octal UART command to start a break */
UCHAR start_brk = BRK_START;

/* the SCC-2698 Octal UART command to stop a break */
UCHAR stop_brk = BRK_STOP;


/* the number of timer ticks until the s336scan routine is called */
int s336tick = 2;

/* supported baud rates */
UCHAR s336speeds[] =
{
	/* default is 9600 baud, unsupported bauds are 9600 */
	0,	/* baud 0 */
	0,	/* baud 50 */
	BD75,
	BD110,
	0,	/* baud 134.5 */
	BD150,
	0,	/* baud 200 */
	BD300,
	BD600,
	BD1200,
	BD1800,
	BD2400,
	BD4800,
	BD9600,
	BD19200,
	BD38400
};

int s336debug = 0;

#ifdef DEBUG
#define PR1(str)		if(s336debug) printf(str);
#define PR2(str, parm1)		if(s336debug) printf(str, parm1);

#else
#define PR1(str)		
#define PR2(str, parm1)		
#endif





/*
*	Routine: s336open
*	Purpose: to initialize a port for reading for writing
*	Input:
*		dev - device major and minor number
*		flag - read or wrote flag - not used
*	Output:
*	Data:
*		s336tty - tty control structure
*		sio_ctrl - s336 control structure
*/

s336open(dev, flag)
register unsigned int dev, flag;
{
        register struct tty *tp;
        register struct sio_ctrl *sc;
	UINT lvl;

        extern int s336proc();
	extern int s336scan();
        extern unsigned sleep();
	extern s336rx_intr();
	extern s336tx_intr();
	extern s336reset();

	/* only minors allowed */
        dev = minor(dev);

	STRC(SIOOPEN,0,dev,0,0,0,0,0);


        tp = &s336_tty[dev];             /* get tty struct */
        sc = &sio_ctrl[dev];   /* get s336 state struct */

	/* check for valid dev && good device */
        if ((dev >= s336_cnt) || (sc->condition == DEAD) )
	{ 
                u.u_error = ENXIO;
                return;
        }

	/*
		catch all those wanting to open 
		when we are trying to close
		or trying to finish the first open
	*/

	lvl = spls336();
	while(sc->t_state & WCLOSE)
	{
		STRC(SIOOPEN,1,0,dev,0,0,0,0);
		if(sleep(( caddr_t)&sc->t_state,(TTOPRI|PCATCH)))
		{
			u.u_error = EIO;
			splx(lvl);
			return;
		}
	}
	splx(lvl);

	/* if we're not already open and not waiting for open then ... */
        if (! (tp->t_state & (ISOPEN|WOPEN))) 
	{
		tp->t_state |= WOPEN;
		STRC(SIOOPEN,2,0,dev,0,0,0,0);
		setup_entry(dev, s336rx_intr, s336tx_intr, s336reset);
                ttinit(tp);             /* init tty structure */
                sc->tp = tp;
		sc->device = dev;
                tp->t_proc = s336proc;
		sc->max_outstanding = MAXOUTPKT;
		sc->outstanding = 0;
		sc->pkt_id = 1;
		sc->sent = sc->tx_fifo;
		sc->ack = sc->tx_fifo;
		sc->high_rxcnt = max_rxack;
		sc->brkflg = 0;
		sc->rflush = 0;
		sc->wflush = 0;
		sc->hwflow = 0;
                s336param(dev);
        }
	/*
		if we don't care about the state of 
		the RS-232 lines or the control
		lines are already in the proper
		state then turn the carrier state on
		otherwise turn it off
	*/
		
        if(tp->t_cflag&CLOCAL || s336modem(dev,ON))
	{
                tp->t_state |= CARR_ON;
		wakeup((caddr_t)&tp->t_canq);
	}
        else
                tp->t_state &= ~CARR_ON;


	/*
		if we need to delay the open until the control
		lines are in the proper state then wait for carrier
	*/
        if (!(flag & FNDELAY))
	{
		lvl = spls336();
                while (! (tp->t_state & CARR_ON)) 
		{
			STRC(SIOOPEN,3,0,dev,0,0,0,0);
			tp->t_state |= WOPEN;
                        sleep(( caddr_t)&tp->t_canq, TTIPRI);
                }
		splx(lvl);
	}
	/* make this port a member of the live wire set */
        (*linesw[tp->t_line].l_open)(tp);
	STRC(SIOOPEN,4,0,dev,0,0,0,0);
}

/*
*	Routine: s336close
*	Purpose: close the port
*	Input:
*		dev - device major and minor number
*	Output:
*	Data:
*		s336tty - tty control structure
*		sio_ctrl - s336 control structure
*/

s336close(dev)
unsigned int dev;
{
        register struct tty *tp;
	register struct sio_ctrl *sc;
        int cur_lvl;

	/* only minors allowed */
        dev = minor(dev);

	STRC(SIOCLOSE,0,dev,0,0,0,0,0);

	sc = &sio_ctrl[dev];
        tp = sc->tp;


	/*
		set the waiting for close  flag
		this prevents anyone from opening the port
		while we're trying to close it
	*/
        cur_lvl=spls336();

	sc->t_state |= WCLOSE;

	while(sc->outstanding)
	{
		/* wait enough time for packets to complete then finish close */
		/* sc->wclose = maxclose; */
		/* wait for output to complete */
		/*
			sleep here will not return a 1 with PCATCH
		*/
		sleep((caddr_t)&tp->t_canq, (TTOPRI|PCATCH));
		{
			/*
				the close must be hung so
				force a good close
			*/
			if(sc->outstanding)
			{
				ttyflush(tp,(FREAD|FWRITE));
				sc->outstanding = 0;
			}
			break;	
		}
	}


	/* reset the open flag */
        (*linesw[tp->t_line].l_close)(tp);

	/* hang up the modem */
        if (tp->t_cflag & HUPCL)
	{
                s336modem(dev,OFF);
		s336send_cmd(sc, RESET_UART, 0, 0, WAIT);
	}

	/* indicate that it's O.K. to open the port now */
	sc->t_state &= ~WCLOSE;

	/*
		If a getty is waiting for a port to open and
		is then turned off the WOPEN bit is set and
		is never turned off. This prevents the port from
		doing the proper open initialization. So mask out 
		this bit
	*/
	tp->t_state &= ~WOPEN;

        splx(cur_lvl);

	/* wakeup any one waiting for this port */
	wakeup(( caddr_t)&sc->t_state);
	wakeup(( caddr_t)&tp->t_state);
}

/*
*	Routine: s336read
*	Purpose: Initiate a read on this port and get any data waiting 
*	Input:
*		dev - device major and minor number
*	Output:
*	Data:
*		s336tty - tty control structure
*		sio_ctrl - s336 control structure
*/

s336read(dev)
unsigned int dev;
{
        register struct tty *tp;

        dev = minor(dev);
        tp = &s336_tty[dev];

	STRC(SIOREAD,0,dev,0,0,0,0,0);

	/* call the line disipline */
        (*linesw[tp->t_line].l_read)(tp);
}


/*
*	Routine: s336write
*	Purpose: Initiate a write on this port
*	Input:
*		dev - device major and minor number
*	Output:
*	Data:
*		s336tty - tty control structure
*		sio_ctrl - s336 control structure
*/

s336write(dev)
unsigned int dev;
{
        register struct tty *tp;

        dev = minor(dev);
        tp = &s336_tty[dev];

	STRC(SIOWRITE,0,dev,0,0,0,0,0);

	/* call the line disipline */
        (*linesw[tp->t_line].l_write)(tp);
}

/*
*
*       ioctl entry
*
*/

#define CFLAG (CBAUD|CSIZE|CSTOPB|PARENB|PARODD|CREAD)
#define IFLAG (IXON|IXOFF|IXANY)

s336ioctl(dev, cmd, arg, mode)
register unsigned int dev, cmd, arg, mode;
{
        register struct tty *tp;
	register int cflag, iflag;
	register struct sio_ctrl *sc;
	UCHAR hwflow;
	UCHAR rev;

        dev = minor(dev);

	sc = &sio_ctrl[dev];
        tp = sc->tp;

	STRC(SIOIOCTL,0,dev,0,0,0,0,0);

	switch(cmd)
	{
	case MSETHWHAND:
		sc->hwflow = 1;
		s336param(dev);
		break;
	case MRESETHWHAND:
		sc->hwflow = 0;
		s336param(dev);
		break;
	case MGETHWHAND:
		hwflow = sc->hwflow;
		copyout(&hwflow, arg, sizeof(hwflow));
		break;
	case MGETROMREV:
		s336send_cmd(sc, ROMREV, &sc->romrev, sizeof(sc->romrev), WAIT);
		rev = sc->romrev;
		copyout(&rev, arg, sizeof(rev));
		break;
		
	default:
		cflag = (tp->t_cflag & CFLAG);
		iflag = (tp->t_iflag & IFLAG);
		if(ttiocom(tp, cmd, arg, mode))
		{
			if( (cflag != (tp->t_cflag & CFLAG)) ||
				(iflag != (tp->t_iflag & IFLAG)) )
			{
				s336param(dev);
			}
		}
	}
}

/*
*
*       param entry
*
*/

s336param(dev)
unsigned int dev;
{
        register struct tty     *tp;
        register struct sio_ctrl *sc;
	register struct port_init *pi;
        register ushort         flags;
        register int indx;
        register unsigned char  charsize;

        sc = &sio_ctrl[dev];
        tp = sc->tp;
	pi = &sc->port_init;

        flags = tp->t_cflag;

	STRC(SIOPARAM,0,flags,0,dev,0,0,0);

        if ((flags & CBAUD) == B0)      /* hang up on zero baud rate */
	{
		STRC(SIOPARAM,1,tp->t_state,0,dev,0,0,0);
		s336modem(dev,OFF);
		return;
	}

        indx=(flags&CBAUD);
	pi->clock_sel = 0;
	pi->clock_sel = ((s336speeds[indx] << 4) | (s336speeds[indx]));

	STRC(SIOPARAM,2,pi->clock_sel,0,dev,0,0,0);

	/*
		This section modifies the parity checking, number of stop bits
		and receiver enable if necessary.
	*/

	pi->mode_reg2 &= ~STOP_MASK;

        if (flags & CSTOPB)                     /* set # stop bits        */
		pi->mode_reg2 |= STOPB2;
        else
		pi->mode_reg2 |= STOPB1;

	pi->mode_reg1 &= ~PRTY_MASK;

        if (flags & PARENB)                   /* set up parity */
	{
		pi->mode_reg1 &= ~PRTY_OEM; /* parity odd/even mask */
                if (flags & PARODD)
			pi->mode_reg1 |= ODD_PRTY;	
        }
	else
		pi->mode_reg1 |= PRTY_OFF;

        /* RX bit patterns and TX bit patterns are the same */

        switch( flags & CSIZE ) {
        case CS5:
		charsize = BPC5;
		break;
        case CS6:       
		charsize = BPC6;      
		break;
        case CS7:       
		charsize = BPC7;      
		break;
        case CS8:       
		charsize = BPC8;      
		break;
        }

	pi->mode_reg1 &= ~BPC_MASK;
	pi->mode_reg1 |= charsize;

	pi->cmd_reg &= ~RXMSK;
        /* enable receiver? */
        if (flags & CREAD)
		pi->cmd_reg |= RX_ENABLE;
        else
		pi->cmd_reg |= RX_DISABLE;

	pi->cmd_reg |= TX_ENABLE;


	/* don't touch RTS and DTR */
	pi->rts = NOT_USED;
	pi->dtr = NOT_USED;

	/* this clears any of the following flow control parameters */
	pi->xoff_any_enable = 0;

	if(tp->t_iflag & IXON)
	{
		pi->xoff_any_enable |= EN_XON;

		if(tp->t_iflag & IXANY)
			pi->xoff_any_enable |= EN_XANY;
	}

	sc->rx_cnt = 0;
	if(tp->t_iflag & IXOFF)
	{
		pi->xoff_any_enable |= EN_XOFF;
		pi->xoff_high = s336xohi;
		pi->xoff_low = s336xolo;
		if(sc->hwflow)
			pi->xoff_any_enable |= EN_RTS;
	}
		
	if(sc->hwflow)
	{
		/* CTS asserted enables transmitter */
		pi->mode_reg2 |= AUTO_CTS;	
	}
	else
	{
		/* CTS asserted does nothing */
		pi->mode_reg2 &= ~AUTO_CTS;	
	}

	/* xon/xoff characters */
	pi->xon = XON;
	pi->xoff = XOFF;

	STRC(SIOPARAM,3,pi->mode_reg1,pi->mode_reg2,pi->cmd_reg,0,dev,0);

	s336send_cmd(sc, INIT_PORT, pi, sizeof(struct port_init), WAIT);

	STRC(SIOPARAM,4,0,dev,0,0,0,0);
}

/*
*
*       proc entry
*
*/

s336proc(tp, cmd)
register struct tty *tp;
register unsigned int cmd;
{
        struct sio_ctrl *sc;
        int dev;
        register int cur_lvl;

	extern ttrstrt();


        dev = tp - s336_tty;
        sc = &sio_ctrl[dev];

	STRC(SIOPROC,0,dev,0,0,0,0,0);

        switch (cmd) {

        case T_TIME:            /* resuming output after delay */
		STRC(SIOPROC,1,0,dev,0,0,0,0);
                tp->t_state &= ~TIMEOUT;
		if(sc->brkflg)
		{
			s336send_cmd(sc,CMD_REG,&stop_brk,sizeof(UCHAR),NOWAIT);
			sc->brkflg = 0;
		}
                goto start;

        case T_WFLUSH:          /* flush output queue */
		STRC(SIOPROC,2,0,dev,0,0,0,0);
                tp->t_tbuf.c_size -= tp->t_tbuf.c_count;
                tp->t_tbuf.c_count = 0;
		if(!sc->wflush)
		{
			sc->wflush++;
			s336send_cmd(sc, FLUSH_XMIT, 0, 0, NOWAIT);
		}
                /* fall thru */


        case T_RESUME:          /* resume output, XON has been received */
		STRC(SIOPROC,4,0,dev,0,0,0,0);

		/* fall through */


        case T_OUTPUT:          /* handle first or next char */
        start: {
                register struct ccblock *tbuf;

		STRC(SIOPROC,5,0,dev,0,0,0,0);

                if (tp->t_state & BUSY)
			break;

		for(;sc->outstanding < sc->max_outstanding;)
		{
			tbuf = &tp->t_tbuf;
			if (tbuf->c_ptr == NULL || tbuf->c_count == 0) 
			{
				if (tbuf->c_ptr)
					tbuf->c_ptr -= tbuf->c_size - 
						tbuf->c_count;
				if (!(CPRES &
					(*linesw[tp->t_line].l_output)(tp)))  
				{
						return;
				}
			}

			cur_lvl = spls336();
			/* send data */
			tp->t_state |= BUSY;

			splx(cur_lvl);

			STRC(SIOPROC,6,0,dev,sc->outstanding,0,0,0);

			s336send_data(sc,XMIT_DATA, tbuf->c_ptr, tbuf->c_count);

			tbuf->c_ptr += tbuf->c_count;
			tbuf->c_count = 0;

		}

                break;
        }

        case T_SUSPEND:
		STRC(SIOPROC,8,0,dev,0,0,0,0);
                break;

        case T_BLOCK:           /* send XOFF */
		STRC(SIOPROC,9,tp->t_state,0,dev,0,0,0);
		tp->t_state |= TBLOCK;
                break;

        case T_RFLUSH:  /* flush input queue.  if input blocked, send XON */

		STRC(SIOPROC,10,dev,0,0,0,0,0);
		if(!sc->rflush)
		{
			sc->rflush++;
			s336send_cmd(sc,ACK_RXCNT,&sc->rx_cnt,sizeof(UCHAR),
				NOWAIT);
			sc->rx_cnt = 0;
		}
                if (!(tp->t_state & TBLOCK))
                        break;
                /* fall thru */

        case T_UNBLOCK:         /* send XON */
		STRC(SIOPROC,11,0,dev,0,0,0,0);
		tp->t_state &= ~TBLOCK;
                break;

        case T_BREAK:           /* send break condition for 1/4 second */
		STRC(SIOPROC,12,0,dev,0,0,0,0);
		sc->brkflg = 1;
		s336send_cmd(sc, CMD_REG, &start_brk, sizeof(int), WAIT);
		timeout(ttrstrt, tp, HZ/4);
                break;

        case T_PARM:
		STRC(SIOPROC,13,0,dev,0,0,0,0);
                s336param(dev);
                break;
        }
	STRC(SIOPROC,14,0,dev,0,0,0,0);
}



s336init()
{
	register UINT port;
	register struct sio_ctrl *sc;
        register struct tty *tp;
	extern int x336_addr[];

	for(port = 0,
		tp = s336_tty,
		sc = sio_ctrl;
		port < s336_cnt;
		port++, sc++, tp)
	{
		sc->condition = DEAD;
		sc->tp = tp;
	}
	STRC(SIOINIT,0,0,0,0,0,0,0);
	/* if the board is here, start the scan */
	/* check the xpc ram not the xpc control registers */
	if(!bprobe((x336_addr[0]+0x20000), -1))
		timeout(s336scan, 0, s336tick);

}


s336tx_intr(xpc)
UINT xpc;
{
	STRC(SIOTXINTR,0,xpc,0,0,0,0,0);

}





/*
*
*       interrupt entry
*
*/

UCHAR data[MAX_XRX_DATA];

s336rx_intr(xpc_data, xpc, count)
UCHAR *xpc_data;
int xpc;
register int count;
{
        register struct tty *tp;
	register UCHAR device;
	register struct sio_ctrl *sc;
	register struct sio_ctrl *scc;
	register struct port_data *rp;

        /*
        */
	STRC(SIORXINTR,0,data,xpc,count,0,0,0);
	if(ODD(count))
	{
		printf("odd number of bytes in rx pkt: %d, pkt: %x\n",
			count, rp);
	}

	if(count > max_rxcnt)
	{
		bcopy(xpc_data, data, count);
		rp = (struct port_data *)data;
	}
	else
		rp = (struct port_data *)xpc_data;
	/*
		the port number range is only for a short range
		of s336 devices 0-1f
		so it is only necessary to add from the start of
		an sio_ctrl block to the end of that block
	*/

	device = rp->port;
	scc = sio_ctrl + SIO(xpc);
	sc = scc + device;
	tp = sc->tp;

	for( ; count > 0; count -= 2, rp++)
        {
		if(device != rp->port)
		{ 
			device = rp->port;
			sc = scc + device;
		}
                switch(rp->type)
                {
		case RX_DATA:
			s336rx(sc, rp->data, 0);
			break;

		case CHAR_STAT:
			/* character status returned for next character */
			if((rp[1].type == RX_DATA) &&
				(rp[1].port == device))
			{
				s336rx(sc, rp[1].data, rp->data);
				rp++;
				count -= 2;
			}
			else
			{
				/*
					this will only happen if the data that
					had an error was an xoff/xon
				*/
				s336rx(sc, 0xff, rp->data);
			}
			break;

		case RS232_STAT:
			/* RS-232 status returned from a request for status */
			sc->rs232 = rp->data;
			break;

		case IPCR:
			/* RS-232 status returned because of a change of state*/
			/* we just had a class 3 event, process */	
			chk_status(rp->data, sc);
			break;

		case TX_ACK:

			tp = sc->tp;
			if(*sc->ack != rp->data)
			{
				if(s336debug)
				{
					printf("port:%x,bad id:%d,data: %d\n", 
						device, *sc->ack, rp->data);
				}
			}
			else
				*sc->ack = 0;

			if(sc->outstanding)
			{
				(sc->ack<&sc->tx_fifo[MAXOUTPKT-1])?(sc->ack++):(sc->ack=sc->tx_fifo);
				sc->outstanding--;
			}
			tp->t_state &= ~BUSY;
			s336proc(tp, T_OUTPUT);
			if((sc->outstanding == NULL) && 
				(sc->t_state & WCLOSE))
					wakeup((caddr_t)&tp->t_canq);

			break;

		case CMD_ACK:
			sc->t_state &= ~BUSY;
			/* allow more of the same command */
			switch(rp->data)
			{
			case ACK_RXCNT:
				sc->rflush = 0;
				break;
			case FLUSH_XMIT:
				sc->wflush = 0;
				break;
			}

			wakeup((caddr_t)&sc->t_state);
			break;

		case DATA_LOST:
			/*
				we could not process the data
				from the cluster in enough time
				and some of it was lost.
				This is an indication that
				something is seriously wrong.
				Disconnect and clean up
			*/
			printf("Data lost on cluster: %d\n", xpc);
			restart_x336(xpc);
			break;
		case RXSTOPPED:
			printf("Cluster ran out of rx packets: %d\n", xpc);
			break;
		case ROMREV:
			sc->romrev = rp->data;
			break;
		default:
			printf("error: port: %d,cmd %x, data: %x\n",
				rp->port, rp->type, rp->data);	
                }
        }
}

/*
*
*       scan entry
*
*/

s336scan()
{
        register struct tty     *tp;
        register int            dev;
        register char           *workp;
	register struct ccblock *rbuf;
	register struct sio_ctrl *sc;
	int junk;

	sc = sio_ctrl;
        for (dev = 0 ; dev < s336_cnt; dev++, workp++, sc++) 
	{
		tp = sc->tp;
		/* gross timeout waiting for transmit packets to complete */
		/*	
		if((sc->t_state & WCLOSE) && (--sc->wclose == NULL))
		{
			ttyflush(tp,(FREAD|FWRITE));
			sc->outstanding = NULL;
			wakeup((caddr_t)&sc->t_state);
		}
		*/	
		/* the xpc link died, scan for it to come alive */
		if(sc->condition == DEAD)
			if(x336condition(dev) == ALIVE)
			{
				sc->condition = ALIVE;
				sc->t_state &= ~BUSY;
				tp->t_state &= ~BUSY;
				wakeup((caddr_t)&sc->t_state);
				wakeup((caddr_t)&tp->t_state);
				if(tp->t_state & ISOPEN)
				{
					if(tp->t_cflag & CLOCAL)
					{
					   s336send_cmd(sc, INIT_PORT, 
					      &sc->port_init, 
					      sizeof(struct port_init),NOWAIT); 
					}
					else
					{
						signal(tp->t_pgrp,SIGHUP);
					}
					wakeup((caddr_t)&tp->t_state);
					ttyflush(tp,(FREAD|FWRITE));
				}
			}

		workp = (char *)&sc->sio_work;
                if (*workp & R6_CHAR) 
		{
                        *workp = 0;
			rbuf = &tp->t_rbuf;
                        if (rbuf->c_ptr) 
			{
                                rbuf->c_ptr -= rbuf->c_size - rbuf->c_count;
                                (*linesw[tp->t_line].l_input)(tp, L_BUF);
				STRC(SIOSCAN,0,dev,0,0,0,0,0);
                        }
                }
                if (*workp & R6_ACT)
                        *workp |= R6_CHAR;
		/*
			rx count acks only go back if
			rx input flow control
			is enabled (IXOFF)
			and the count is high enough
			and we aren't blocked
		*/
		if( (tp->t_iflag & IXOFF) &&
			(sc->rx_cnt > sc->high_rxcnt) &&
			!(tp->t_state & TBLOCK))
		{
			if(!sc->rflush)
			{
				sc->rflush++;
				s336send_cmd(sc, ACK_RXCNT, &sc->rx_cnt,
					sizeof(UCHAR), NOWAIT);
				sc->rx_cnt = 0;
			}
		}
        }
        timeout(s336scan, 0, s336tick);
}
/*
*
*       modem entry
*
*/

s336modem(dev,flag)
int dev;
int flag;
{
        register struct sio_ctrl *sc;
        register struct tty     *tp;

        sc = &sio_ctrl[dev];
        tp = sc->tp;
        
	STRC(SIOMODEM,0,dev,0,0,0,0,0);
        if(flag==OFF)
	{
		STRC(SIOMODEM,1,0,dev,0,0,0,0);

		s336send_cmd(sc, CMD_REG, &rts_off, sizeof(UCHAR), WAIT);
		s336send_cmd(sc, DTR_CHG, &dtr_off, sizeof(UCHAR), WAIT);

                if(!(tp->t_cflag & CLOCAL))
                {
			STRC(SIOMODEM,2,0,dev,0,0,0,0);
                        if(tp->t_state & CARR_ON){
                                if(tp->t_state & ISOPEN)
				{
                                        signal(tp->t_pgrp,SIGHUP);
                                        ttyflush(tp,(FREAD|FWRITE));
                                }
                                tp->t_state &= ~CARR_ON;
                        }
                }
                else if(!(tp->t_state & CARR_ON))
		{
			STRC(SIOMODEM,3,0,dev,0,0,0,0);
			/* we're clocal and no carrier */
                        wakeup((caddr_t)&tp->t_canq);
                        tp->t_state |= CARR_ON;
                }
        } 
	else
	{
		/*
			assert RTS and DTR, the interrupt for
			the change in line state should come
			along later
		*/
		STRC(SIOMODEM,4,0,dev,0,0,0,0);
		s336send_cmd(sc, CMD_REG, &rts_on, sizeof(UCHAR), WAIT);
		s336send_cmd(sc, DTR_CHG, &dtr_on, sizeof(UCHAR), WAIT);
        }

	STRC(SIOMODEM,5,0,dev,0,0,0,0);
	/* this command will cause sc->rs232 to be set to the current status */
	s336send_cmd(sc, RS232_STAT, 0, 0, WAIT);

	STRC(SIOMODEM,6,sc->rs232,0,dev,0,0,0);

        return( (sc->rs232 & DCD));
}


/*
	this routine assumes that tp->t_state is not busy and
	that this packet can go out immediately

	The data ack interrupt will reset the
	busy bit thus allowing the next data block to go out
*/
s336send_data(sc, cmd, data,count)
register struct sio_ctrl *sc;
register int cmd;
UCHAR *data;
int count;
{
	register struct xmit_pkt *pkt;
	register struct xmit_packet *tp;
	register int dev = sc->device;
	register UINT lvl;

	extern struct xmit_pkt *get_pkt();

	STRC(SIOSENDDATA,sc,cmd,data,count,dev,0,0);

	/* no sense sending commands to a dead link */
	if(sc->condition == DEAD)
		return;

	/* 
		from here to the splx must be uninterruptable, so
		 packets aren't lost
	*/

	lvl = spls336();

	if(( pkt = get_pkt(dev)) == NULL)
	{
		u.u_error = EIO;
		return(NULL);
	}

	tp = (struct xmit_packet *)pkt->txp;

	tp->tx_hdr.port_no = (dev & 0x0F);
	tp->tx_hdr.function = cmd;
	tp->tx_hdr.count = count;
	tp->tx_hdr.packet_id = sc->pkt_id;

	*(sc->sent) = sc->pkt_id;	

	(sc->sent < &sc->tx_fifo[MAXOUTPKT-1])?(sc->sent++):(sc->sent=sc->tx_fifo);
	(sc->pkt_id > MAXOUTPKT)?(sc->pkt_id=1):(sc->pkt_id++);
	sc->outstanding++;

	if(count)
		bcopy(data, tp->data, count);

	send_pkt(pkt, dev, count + sizeof(struct tx_hdr));

	splx(lvl);

}

s336send_cmd(sc, cmd, data, count, zzzzs)
register struct sio_ctrl *sc;
register int cmd;
UCHAR *data;
register int count;
register int zzzzs; /* guess what this one is */
{
	register struct xmit_pkt *pkt;
	register struct xmit_packet *tp;
	register int lvl;
	register int dev = sc->device;

	extern struct xmit_pkt *get_pkt();

	STRC(SIOSENDCMD,sc,cmd,data,count,zzzzs,0,dev);
	/* 
		wait for a previous commmand to finish 
		before starting another
	*/

	/* no sense sending commands to a dead link */
	if(sc->condition == DEAD)
		return;

	/* 
		from here to the splx must be uninterruptable, so
		 packets aren't lost
	*/

	lvl = spls336();

	if(( pkt = get_pkt(dev)) == NULL)
	{
		u.u_error = EIO;
		return(NULL);
	}
	tp = (struct xmit_packet *)pkt->txp;

	tp->tx_hdr.port_no = (dev & 0x0F);
	tp->tx_hdr.function = cmd;
	tp->tx_hdr.count = count;
	tp->tx_hdr.packet_id = cmd;

	if(count)
		bcopy(data, tp->data, count);


	if(zzzzs)
		sc->t_state |= BUSY;

	send_pkt(pkt, dev, (count + sizeof(struct tx_hdr)));

	splx(lvl);

	if(zzzzs)
	{
		lvl = spls336();
		while(sc->t_state & BUSY)
		{
			if(sleep((caddr_t)&sc->t_state, (TTOPRI|PCATCH)))
			{
				sc->t_state &= ~BUSY;
				break;
			}
		}
		splx(lvl);
	}

}


s336rx(sc, c, status)
register struct sio_ctrl *sc;
register UCHAR c;
register UCHAR status;
{
	register int lcnt;
	register UCHAR lbuf[3];
	register int flg;
	register int cur_lvl;
	register struct tty *tp;
	register int count;


	tp = sc->tp;
	sysinfo.rcvint++;

	/* we don't accept data if the port is closed */
	if(!(tp->t_state & ISOPEN))
		return;

	flg=tp->t_iflag;

	/* keep track of all characters received */
	sc->rx_cnt++;

	lcnt = 1;

	/*
	 *	If ISTRIP is set, strip input character to 7 bits,
	 *	or if less than CS8, strip possible parity bit,
	 *	instead of letting it pass thru.  This conforms with
	 *	the intent of ATT of not having parity in the data byte.
	 */
	if (flg & ISTRIP || (tp->t_cflag & CSIZE) < CS8)
		c &= 0x7f;
	else if(c == 0xff && tp->t_iflag & PARMRK){
		lbuf[1]= 0xff;
		lcnt= 2;
	}




	if (!(flg&IGNPAR)) {	/* no checking at all */
		if (!(flg&INPCK))	/* no parity check */
			status &= ~PARERR;
		if (status & (PARERR|OVRRUN)) {
			if (flg & PARMRK) {
				lbuf[2] = 0xFF;
				lbuf[1] = 0;
				lcnt = 3;
				sysinfo.rawch += 2;
			} else
				c = 0;
		}
	}
	/* break processing */
	if ( status & RCVBRK ) 
	{
		if ( !(flg & IGNBRK) && (flg & BRKINT) )
		{
			(*linesw[tp->t_line].l_input)(tp, L_BREAK );
		}
	}

	/* stash character in r_buf */
	lbuf[0] = c;
	if (tp->t_rbuf.c_ptr != NULL) 
	{
		while (lcnt) 
		{
			*tp->t_rbuf.c_ptr++ = lbuf[--lcnt];
			sc->sio_work = 0;
			tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size -
				--tp->t_rbuf.c_count;
			(*linesw[tp->t_line].l_input)(tp,L_BUF);
			sc->sio_work = R6_ACT;
			/* note R6_CHAR bit cleared */
		} 
	} 
}



chk_status(ipcr, sc)
register UCHAR ipcr;
register struct sio_ctrl *sc;
{
	/* ipcr stands for Input Port Change Register */

	register struct tty *tp = sc->tp;

	STRC(SIOCHKSTATUS,sc,ipcr,0,sc->device,0,0,0);

	/* carrier processing */
	if(ipcr & DDCD)
	{
		if ( ipcr & DCD ) 
		{
			sc->dcdflg++;
			if ((tp->t_state & (CARR_ON|WOPEN)) == WOPEN) 
			{
				wakeup((caddr_t)&tp->t_canq);
				tp->t_state |= CARR_ON;
			}
		}
		else
		{
			sc->dcdflg = 0;
			if ( !(tp->t_cflag & CLOCAL) ) 
			{
				if(tp->t_state & CARR_ON ) 
				{
					if (tp->t_state & ISOPEN) 
					{
						signal(tp->t_pgrp,SIGHUP);
						ttyflush(tp,(FREAD|FWRITE));
					}
					tp->t_state &= ~CARR_ON;
				}
			}
		}
	}
}



s336reset(xpc, condition)
UINT xpc;
UINT condition;
{
	register struct sio_ctrl *sc;
	register int port;
	register int maxport;

	STRC(SIORESET,0,xpc,condition,0,0,0,0);
	for(port = SIO(xpc), maxport = SIO(xpc+1),
		sc = &sio_ctrl[SIO(xpc)];
		port < maxport; 
		port++, sc++)
	{
		/* set the dead flag for all the ports associated with an xpc */

		if(condition == SHUTDOWN)
		{
			sc->condition = DEAD;
		}
	}
}

