/*		Copyright (c) 1987 Motorola			*/
/*		All Rights Reserved				*/
/*		@(#)kern-port:io/mvme336xpc.c	35.3	
/*
 */
#ident "@(#)kern-port:io/mvme336xpc.c	35.3"
/*



	X.25 Protocol Controller (XPC)

	This file consists of the following sub-routines:

	setup_entry
	x336init
	init_x336mem
	get_tx_pkt
	send_pkt
	x336tx
	return_tx_pkt
	x336intr
	x336rx_intr
	x336tx_intr
	x336catchall
	restart_x336

*/

#define XPCMACRO

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/fs/s5dir.h"
#include "sys/file.h"
#include "sys/signal.h"
#include "sys/tty.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/sysinfo.h"
#include "sys/mvme336x.h"
#include "sys/mvme336c.h"
#include "sys/immu.h"
#include "sys/mvme336t.h"



/*
	This structure is used to control the I/O on each x336.
*/
struct xpc_ctrl x336ctrl[MAX_XPC];

int xpcout = 0;

int maxcnt = 0;


#ifdef TRACE

int x336_trace = 0;
struct trace *x336trace;
struct trace *x336trcbase;

#define XTRC(id,a,b,c,d,e,f,g) if(x336_trace){chktrc(id,a,b,c,d,e,f,g);}

#else

#define XTRC(id,a,b,c,d,e,f,g) 

#endif


#define TRUE (-1)
#define FALSE (0)

#define splx336 spl6

int x336debug = 0;



/*
	routine: setup_entry
	purpose: assign the service routines for read, write and error
	input:
		device - the real device number
		rx - receive service function
		tx - transmit service function
		err - error service function
	data:

	called by:

*/

setup_entry(device, rx, tx, err)
int device;
PFI rx, tx, err;
{
	register struct xpc_ctrl *xp = &x336ctrl[XPC(device)];

	XTRC(REGISTER,0,xp,device,xp->rx_func,xp->tx_func,xp->error_func,0)

	if(xp->rx_func || xp->tx_func || xp->error_func)
		return(REGIS_ERROR);

	XTRC(REGISTER, 1, xp, device, rx, tx, err,0);

	xp->rx_func = rx;
	xp->tx_func = tx;
	xp->error_func = err;

	return(REGIS_OK);
}

clr_entry(device)
int device;
{
	register struct xpc_ctrl *xp = &x336ctrl[XPC(device)];


	XTRC(DEREGISTER, 0, xp, device,0,0,0,0);

	xp->rx_func = (PFI ) NULL;
	xp->tx_func = (PFI ) NULL;
	xp->error_func = (PFI ) NULL;

}


/*
	routine: x336init
	purpose: initialize all of the available x336's
	input: none
	data: x336_addr, x336_ivec, x336_ilev, x336_cnt, x336_ctrl
	called by: io_init
*/


x336init()
{
	extern int x336_ivec[];
	extern int x336_addr[];
	extern int x336_ilev[];
	extern int x336_cnt;

	UCHAR *x336mem = (UCHAR *)(x336_addr[0] + XPC_MEM);
	register ALL_XPC *all_x336 = (ALL_XPC *)x336_addr[0];

	register struct xpc_ctrl *xp;
	register XPC_CMD *x336;
	register int i;
	register UCHAR *mem;
	int size;
	int lvl;
	int vect;

	lvl = splx336();

	/* check to see if the board is available */
	if(!bprobe(x336mem, -1))
	{
#		ifdef TRACE
		size = btoc(sizeof(struct trace) * MAX_TRC); 
		if(mem = (UCHAR *)sptalloc(size, PG_VALID, 0))
		{
			x336trcbase = (struct trace *) mem;
			for(x336trace = x336trcbase, i = 0; i < MAX_TRC;
				i++, x336trace++)
			x336trace = x336trcbase;
			{
				x336trace->id = 0;
				x336trace->i0 = 0;
				x336trace->i1 = 0;
				x336trace->i2 = 0;
				x336trace->i3 = 0;
				x336trace->i4 = 0;
				x336trace->i5 = 0;
				x336trace->i6 = 0;
			}
		}
		else
			printf("XPC: No memory available for trace\n");

#		endif

		bzero(x336mem, XPC_MEM);

		XTRC(INIT, 0,0,0,0,0,0,0);

		for( i = 0; i < x336_cnt; i++)
		{
			/* get to the x336 specific parameters */
			xp = &x336ctrl[i];

			/* ptr to the x336 registers */
			x336 = &all_x336->xpc_cmd[i];

			/*
				the x336 is set to alive when
				an interrupt is generated to 
				indicate that it is indeed
				connected, that is, when
				xp->alive is set to ALIVE;
			*/
			/* init memory for this x336 */
			SEND_CMD(x336, RESET);

			vect  = (x336_ivec[i]/4);
			(int) x336mem = init_x336mem(i, x336mem, xp, x336,vect);

			/* mark the state as coming down */
			xp->alive = DEAD;

			SEND_CMD(x336, STOP_LK);
			/*
				now await the new mode status of 
				disconnected to do a start link.
			*/
		}
		/* set the interrupt level and enable interrupts */
		all_x336->intr_ctrl = (x336_ilev[0] | INT_ENABLE);
	}
	splx(lvl);

}

/*
	Routine:
	Purpose: Initialize packets on a per XPC basis.
	Input:
		x336index - id number of x336
		x336mem	- area to set aside for this x336
		xp - ptr to the x336 control structure for this x336
		x336 - ptr to the hardware registers associated with this x336
		iv - interrupt vector to be used with x336
	Data:
	Called by: x336init

	Notes:

		There shouldn't be more than 6 rx packets in use at
		any one time on all of the XPC's at once. Even so there
		are MAX_RX packets available.

		A rx packet contains information for all ports (16)
		on a cluster.

*/



init_x336mem(x336n, x336mem, xp, x336, iv)
UINT x336n;
register char *x336mem;
register struct xpc_ctrl *xp;
register XPC_CMD *x336;
register int iv;
{
	register struct station *st;
	register struct rx_packet *rp;
	register struct rx_frame *rf;
	register struct xmit_pkt *tp;
	register int rx_cnt;
	register int tx_cnt;

	XTRC(MEMINIT, 0, x336n, x336mem, xp, x336, iv,0);

	xp->head_txq = (struct xmit_pkt *)NULL;
	xp->tail_txq = (struct xmit_pkt *)NULL;
	xp->device = x336n;

#	if DEBUG1
	/* only used to force x336's tp talk to each other */
	if(ODD(x336n))
	{
		xp->local_adr = REMOTE;
		xp->rmt_adr = LOCAL;
	}
	else
#	endif
	{
		xp->local_adr = LOCAL;
		xp->rmt_adr = REMOTE;
	}

	xp->xpc_cmd = x336;
	xp->rx_reset = 0;
	xp->tx_reset = 0;
	xp->tx_err = 0;
	xp->rx_err = 0;
	xp->tx_connect = 0;

	/* this is the start of the station table */
	st = xp->station = (struct station *)x336mem;

	/* setup basic station table entries */

	st->option = CRC_REJECT;
	st->time_out = MAX_TIMEOUT;
	st->time_scale = TS512;
	st->pad_time = PAD_TIME;
	st->out_frame = MAX_OSFL;
	st->retries = MAX_RETRY;
	st->rx_mask = (UDEF|ILL|RX_AERR|RX_BERR|OVRN|RXI|RXFRMR|RTE);
	st->tx_mask = (NEWMD|RTRYL|TX_AERR|TX_BERR|URUN|IFAK);
	st->rx_stat_clr = NULL;
	st->tx_stat_clr = NULL;
	st->local_adr = xp->local_adr;
	st->remote_adr = xp->rmt_adr;

	x336mem += sizeof(struct station);


	/* setup transmit frame table pointers */

	st->ttfc = NULL;
	st->tx_tbl = xp->tx_frame = (struct xmit_frame *)x336mem;

	x336mem += (sizeof(struct xmit_frame) * MAX_TX);

	/* set up receive frame table pointers */

	st->rtfc = NULL;
	st->rx_tbl = xp->rx_frame = (struct rx_frame *)x336mem;

	x336mem += (sizeof(struct rx_frame) * MAX_RX);

	/* setup area for receive data packets */

	xp->rx_pkt = (struct rx_packet *)x336mem;
	x336mem += (sizeof(struct rx_packet) * MAX_RX);

	for( rf = xp->rx_frame, rx_cnt = 0, rp = xp->rx_pkt; 
		rx_cnt < MAX_RX; 
		rx_cnt++, rf++, rp++)
	{
		rf->rx_tbl_stat = 0;
		rf->rx_buf_adr = (struct rx_packet *)rp->data;
		rf->rx_buf_len = MAX_XRX_DATA;
	}
	(--rf)->rx_tbl_stat = EOT;
	xp->cur_rx_ptr = xp->rx_frame;

	/* setup area for transmit data packets */

	xp->tx_pkt = (struct xmit_pkt *)x336mem;

	x336mem += (MAX_XTX_DATA * MAX_TX);

	for(tx_cnt = 0, tp = xp->tx_pkt; tx_cnt < MAX_TX; tp++, tx_cnt++)
	{
		tp->next = tp+1;
	}
	tp--;
	tp->next = NULL;

	xp->iv = iv;

	/* load the x336 station table parameters */
	set_stp(xp);

	return((int)x336mem);
}       

/*
	function: get_pkt
	purpose: get a packet for transmission from the packet list
	input: device - device on which the operation is to be performed
	data:
	called by:
*/

struct xmit_pkt *get_pkt(device)
register int device;
{
	register struct xpc_ctrl *xp = &x336ctrl[XPC(device)];
	register struct xmit_pkt *pkt;
	register int lvl;

	XTRC(GETPKT, 0, device, xp, 0,0,0,0);

	if(xp->tx_pkt == NULL)
	{
		XTRC(GETPKT, 1, device, xp, 0,0,0,0);
		printf("x336:%d out of transmit packets.\n", XPC(device));
		return(NULL);
	}

	XTRC(GETPKT, 2, device, xp, 0,0,0,0);

	lvl = splx336();

	pkt = xp->tx_pkt;
	xp->tx_pkt = pkt->next;
	pkt->next = NULL;

	xpcout++;

	splx(lvl);


	return(pkt);
}
/*
	function: send_pkt
	purpose: enqueue data to the x336 or queue data to the tx queue
	input:
		pkt - transmit data packet which is to be queued
		device - device on which to perform the operation

	data:
	called by:
*/

send_pkt(pkt, device,count)
register struct xmit_pkt *pkt;
register int device;
register int count;
{
	register struct xpc_ctrl *xp = &x336ctrl[XPC(device)];
	register int lvl;

	XTRC(SENDPKT, 0, device, xp, pkt,count,0,0);

	pkt->count = count;
	pkt->device = device;

	lvl = splx336();
	if(xp->state & TX_BUSY)
	{
		XTRC(SENDPKT, 1, device, xp, pkt,xp->head_txq,0,0);
		if(xp->head_txq)
		{
			XTRC(SENDPKT, 2, device, xp, pkt,0,0,0);
			/*
				were busy tranmitting and we have gotten
				buffers to transmit, so just add it
				to the list
			*/
			xp->tail_txq->next = pkt;
			xp->tail_txq = pkt;
			pkt->next = NULL;
		}
		else
		{
			XTRC(SENDPKT, 3, device, xp, pkt,0,0,0);
			/*
				we're busy but haven't q'ed
				any buffers since we went to transmit
				the last batch
			*/
			xp->head_txq = pkt;
			xp->tail_txq = pkt;
			pkt->next = NULL;
		}
	}
	else
	{
		XTRC(SENDPKT, 4, device, xp, pkt,0,0,0);
		/*
			we're not busy, so we can send the buffer out
			immediately.
		*/
		xp->state |= TX_BUSY;
		xp->head_txq = pkt;
		xp->tail_txq = pkt;
		pkt->next = NULL;
		x336tx(xp);
	}
	splx(lvl);
	XTRC(SENDPKT, 5, device, xp, pkt,0,0,0);
}

/*
	routine:
	purpose:
	input:
	data:
	called by:
*/

x336tx(xp)
register struct xpc_ctrl *xp;
{
	register struct xmit_frame *tf;
	register struct xmit_pkt *tq;
	register int count;
	UINT lvl;


	XTRC(XPCTX, 0, xp, xp->head_txq,0,0,0,0);

	if(xp->head_txq) /* if we have buffers to send then ... */
	{
		/*
			process all of the buffers in the Q
			and attach each of them to a transmit
			frame
		*/

		xp->cur_tx_ptr = xp->head_txq;
		xp->cur_tx_tail = xp->tail_txq;
		for(count = 0,tq = xp->head_txq,tf = xp->tx_frame;
			 tq; tf++,tq = tq->next, count++)
		{
			tf->tx_tbl_stat = NULL;
			tf->tx_buf_adr = tq->txp;
			tf->tx_buf_len = tq->count;
		}
		(--tf)->tx_tbl_stat = EOT; 

		if(x336debug) 
			printf("count: %d\n", count);

		xp->head_txq = (struct xmit_pkt *)NULL;
		xp->tail_txq = (struct xmit_pkt *)NULL;

		SEND_CMD(xp->xpc_cmd, LD_TXTP);

	}
	else
	{
		/* there is nothing else to do */
		xp->state &= ~TX_BUSY;
	}
}

/*
	routine:
	purpose:
	input:
	data:
	called by:
*/


x336tx_intr(xp)
register struct xpc_ctrl *xp;
{
	UINT device;
	PFI func;
	register struct xmit_pkt *cur;
	int count = 0;

	XTRC(XPCTXINT, 0, xp, 0,0,0,0,0);

	/* keep track of the used xmit packets */
/*
	cur = xp->cur_tx_ptr;
	while (cur)
	{
		cur = cur->next;
		xpcout--;
		count++;
	}
	if(count > maxcnt)
	{
		maxcnt = count;
		printf("xpc: sent:%d\n", count);
	}
*/

	xp->cur_tx_tail->next = xp->tx_pkt;
	xp->tx_pkt = xp->cur_tx_ptr;
	
	device = (UINT)xp->device;
	if(func = xp->tx_func)
	{
		XTRC(XPCTXINT, 1, func,device,0,0,0,0);
		(*func)(XPC(device));
	}

	x336tx(xp);	

}

/*
	routine:
	purpose:
	input:
	data:
	called by:
*/

x336rx_intr(xp)
register struct xpc_ctrl *xp;
{

	struct station *st = xp->station;
	register int x336 = xp - x336ctrl;

	register struct rx_frame *crp; /* current receive pointer */
	register int count;
	register PFI func;
	register struct rx_frame *last;
	register struct rx_packet *pkt;

	XTRC(XPCRXINT, 0, xp, x336,0,0,0,0);

	SEND_CMD(xp->xpc_cmd, DMP_PAR);
	WAIT_SEMA(xp->xpc_cmd);

	last = st->next_rx;

	crp = xp->cur_rx_ptr;

	XTRC(XPCRXINT, 0, xp, x336,last,crp,0,0);

	for(;crp != last; crp++)
	{
		/* the address of data is set when pkt is setup in init */
		count = crp->rx_buf_len - crp->rx_final_cnt;
		pkt = crp->rx_buf_adr; 
		if(func = xp->rx_func)
		{
			XTRC(XPCRXINT, 3, func,pkt,x336,count,0,0);
			(*func)( pkt, x336, count);
		}
		if (crp->rx_tbl_stat == EOT)
		{
			crp = xp->rx_frame;
			break;
		}
	}

	if(st->rx_status & RTE )
	{
		st->rx_stat_clr = RTE;

		SEND_CMD(xp->xpc_cmd, CL_STAT); /* clear the rx and tx status */

		SEND_CMD(xp->xpc_cmd, LD_RXTP);

	}
	xp->cur_rx_ptr = crp;

	XTRC(XPCRXINT, 2, xp, 0,crp,0,count,0);

}

/*
	routine:
	purpose:
	input:
	data:
	called by:
*/

x336intr(x336index)
int x336index;
{

	register struct xpc_ctrl *xp;
	register struct station *st;
	register struct xpc_cmd *x336;
	register int tx_status;
	register int rx_status;
	register int mode;
	int lvl;
	
	lvl = splx336();

	xp = &x336ctrl[x336index];
	st = xp->station;
	x336 = xp->xpc_cmd;

	tx_status = st->tx_status;
	rx_status = st->rx_status;
	mode = st->mode_desc;

	SEND_CMD(xp->xpc_cmd, DMP_PAR);
	WAIT_SEMA(xp->xpc_cmd);

	tx_status |= st->tx_status;
	rx_status |= st->rx_status;
	mode |= st->mode_desc;

	XTRC(XPCINT, 0, xp, st, tx_status,rx_status,0,0);

	st->tx_stat_clr = tx_status;
	st->rx_stat_clr = (rx_status & ~RTE);

	SEND_CMD(x336, CL_STAT); /* clear the rx and tx status */

	if(rx_status & RXI)
	{
		x336rx_intr(xp);
	}
	if(tx_status & IFAK)
	{
		x336tx_intr(xp);
	}
	/* if it's not receive or transmit complete */
	rx_status &= ~RXI;
	tx_status &= ~IFAK;
	if((rx_status) || (tx_status))
		x336catchall(xp, rx_status, tx_status, mode); 
	
	splx(lvl);
}

/*
	for the most part this is used to catch catastrophic 
	errors and is a gross means of disconnect processing

	Most errors should be recoverable but some arn't
*/
/*
	routine:
	purpose:
	input:
	data:
	called by:
*/

x336catchall(xp, rx_status, tx_status, saved_mode)
register struct xpc_ctrl *xp;
register int rx_status;
register int tx_status;
register int saved_mode;
{
	register struct station *st = xp->station;
	register PFI func = xp->error_func;
	register int x336 = xp - x336ctrl;
	register USHORT old_mode;
	register USHORT mode;
	register USHORT new_mode;
	
	XTRC(CATCH, 0, xp, st, tx_status,rx_status,0,0);

	if(rx_status)
	{
		if(rx_status & (RX_AERR|RX_BERR|ILL|OVRN|RXFRMR))
		{
			xp->rx_reset++;

			/* reset the xpc */
			SEND_CMD(xp->xpc_cmd, RESET);

			if( func )
			{
				XTRC(CATCH, 1,func,x336,0,0,0,0);
				(*func)(x336, SHUTDOWN);
			}

			restart_x336(x336);

		}
		if (rx_status & UDEF)
		{
			xp->rx_err++;
			rx_logerr(x336,rx_status & UDEF);
		}
	}

	if(tx_status)
	{
		if(tx_status & (RTRYL|TX_AERR|TX_BERR))
		{
			xp->tx_reset++;

			/* reset the xpc */
			SEND_CMD(xp->xpc_cmd, RESET);

			if( func )
			{
				XTRC(CATCH, 2,func,x336,0,0,0,0);
				(*func)(x336, SHUTDOWN);
			}
			restart_x336(x336);
		}
		else if(tx_status & (NEWMD|URUN))
		{
			if(tx_status & URUN)
			{
				xp->tx_err++;
				tx_logerr(x336,tx_status & URUN);
			}
			else	/* it's a new mode */
			{
				old_mode  = xp->mode_desc;
				new_mode  = saved_mode;

				xp->mode_desc = new_mode;
				mode = saved_mode ^ old_mode;

				XTRC(CATCH, 3, new_mode,0,0,0,0,0);
				XTRC(CATCH, 4, old_mode,0,0,0,0,0);
				if(mode & CONN)
				{
					if(new_mode & CONN)
					{
						XTRC(CATCH, 5, 0,0,0,0,0,0);
						xp->alive = ALIVE;
						xp->tx_connect++;
						if( func )
						{
							XTRC(CATCH,6,0,0,0,0,0,0);
							(*func)(x336, STARTUP);
						}
					}
					else
					{
						/* reset the xpc */
						SEND_CMD(xp->xpc_cmd, RESET);

						xp->alive = DEAD;

						XTRC(CATCH,7,func,x336,0,0,0,0);
						if( func )
						{
							XTRC(CATCH,8,func,x336,0,0,0,0);
							(*func)(x336, SHUTDOWN);
						}
						restart_x336(x336);
					}
				}
			}
		}
	}
}


/*
	routine:
	purpose:
	input:
	data:
	called by:
*/

#ifndef XPCMACRO

x336send_cmd(x336, cmd)
register struct xpc_cmd *x336;
UINT cmd;
{
	/* XTRC(SENDCMD, 0, x336, cmd,0,0,0,0); */
	while(x336->cmd_sem != READY);
	x336->cmd_sem = cmd;
}

x336wait_sema(x336)
register struct xpc_cmd *x336;
{
	/* XTRC(WAITSEMA, 0, x336,0,0,0,0,0); */
	while(x336->cmd_sem != READY); 
}

/*
	routine:
	purpose:
	input:
	data:
	called by:
*/

x336send_data(data, x336, cmd)
register UINT data;
register struct xpc_cmd *x336;
UINT cmd;
{
	/* XTRC(SENDDATA, 0, x336,cmd, data,0,0,0); */

	while(x336->cmd_sem != READY);
	x336->data_hi = (data >> 16);
	x336->data_lo = (data & 0xffff);
	x336->cmd_sem = cmd;
}

/*
	routine:
	purpose:
	input:
	data:
	called by:
*/


x336send_iv(iv, x336, cmd)
UINT iv;
register struct xpc_cmd *x336;
UINT cmd;
{
	/* XTRC(SENDIV, 0, x336,cmd, iv,0,0,0); */
	while(x336->cmd_sem != READY);
	x336->ivect = iv;
	x336->cmd_sem = cmd;
}


#endif


restart_x336(x336)
int x336;
{
	register struct xpc_ctrl *xp = &x336ctrl[x336];
	
	XTRC(RESTART, 0, xp,0,0,0,0,0);

	if(xp->state & TX_BUSY)
	{
		xp->cur_tx_tail->next = xp->tx_pkt;
		xp->tx_pkt = xp->cur_tx_ptr;
		xp->state &= ~TX_BUSY;
		xp->cur_tx_ptr = (struct xmit_pkt *)NULL;
		xp->cur_tx_tail = (struct xmit_pkt *)NULL;
	}
	if(xp->head_txq)
	{
		xp->tail_txq->next = xp->tx_pkt;
		xp->tx_pkt = xp->head_txq;
		xp->head_txq = (struct xmit_pkt *)NULL;
		xp->tail_txq = (struct xmit_pkt *)NULL;
		
	}

	/* re-initialize the x336 with the station table parameters */
	set_stp(xp);

	/* mark the living state as coming down */
	xp->alive = DEAD;

	/* make sure the link is brought down */
	SEND_CMD(xp->xpc_cmd, STOP_LK);

}

set_stp(xp)
register struct xpc_ctrl *xp;
{
	register struct xpc_cmd *x336 = xp->xpc_cmd;

	/* setup the hardware characteristics */
	SEND_CMD(x336, (SET_HDW|NRZ|BSRC|NOT_DORGM|BUSW));

	/* initialize the station configuration */
	SEND_CMD(x336, (SET_STA|ECRC|ST_NECNT));

	/* initialize the station table pointer */
	SEND_DATA(xp->station, x336, LD_STPT);

	/* load the interrupt vector */
	/* and load the station table parameters */
	SEND_IV(xp->iv, x336, LD_STPR);

	/* load the receive table pointer */
	SEND_CMD(x336, LD_RXTP);

}

rx_logerr(x336,status)
int x336,status;
{
}

tx_logerr(x336,status)
int x336,status;
{
}

x336condition(dev)
UINT dev;
{
	register struct xpc_ctrl *xp = &x336ctrl[XPC(dev)];
	
	return(xp->alive);

}

#ifdef TRACE

chktrc(id, a,b,c,d,e,f,g)
int id,a,b,c,d,e,f,g;
{
	register struct trace *xpt = x336trace;
	int lvl;

	lvl = spl7();

	xpt->id = id;
	xpt->i0 = a;
	xpt->i1 = b;
	xpt->i2 = c; 
	xpt->i3 = d;
	xpt->i4 = e;
	xpt->i5 = f;
	xpt->i6 = g;
	(xpt+1 > &x336trcbase[MAX_TRC-1])?xpt=x336trcbase:xpt++;
	xpt->id = END_TRACE;
	xpt->i0 = 0;
	xpt->i1 = 0;
	xpt->i2 = 0;
	xpt->i3 = 0;
	xpt->i4 = 0;
	xpt->i5 = 0;
	xpt->i6 = 0;
	x336trace = xpt;

	splx(lvl);

}

#endif


#ifdef XPC_OPEN
#else
x336open () {}
x336close() {}
x336read() {}
x336write() {}
x336ioctl() {}
#endif
