/*
 *	Copyright (c) 1984 AT&T
 *	Copyright (c) 1987 Motorola Inc
 *	All Rights Reserved
 */

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

#ident	"@(#)kern-port:io/mvme350.c	35.4"
/*
 *	mvme350.c: Motorola MVME350 VMEmodule Streamer Tape Controller driver
 *
 *
 *	The MVME350 VMEmodule Streamer Tape Controller is an intelligent
 *	peripheral controller (IPC) VMEmodule designed to interface
 *	with the extended address line VMEmodules of the VMEbus family.
 *	The MVME350 supports only one streamer tape drive per controller
 *	using the QIC-02 Streaming Tape Interface.
 *
 *	This driver will support forward seeks on reads.
 */

/*	INCLUDE files 	*/

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/systm.h"
#include "sys/mvmecpu.h"
#include "sys/buf.h"
#include "sys/signal.h"

#ifdef REL2
#include "sys/dir.h"
#else
#include "sys/fs/s5param.h"
#include "sys/fs/s5dir.h"
#include "sys/immu.h"
#include "sys/region.h"
#include "sys/cmn_err.h"
#endif

#include "sys/proc.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/file.h"
#include "sys/elog.h"
#include "sys/iobuf.h"
#include "sys/bpp.h"
#include "sys/dbuf.h"
#include "sys/mvme350.h"

#define	MULTITAPE	/* multiple tape support */
#define	STATCMD		/* do stat on open */
#define	BYTESWAP	/* include byteswap code */
#define	TAS		/* do tas instruction in getchan */
#undef	TAS

/*
 *	The MVME350 controller does not support Reads and Writes
 * 	intermixed. It only allows Reads or Writes.
 *	The READWRITE conditional compilation flags allows users
 * 	to open the device Read/Write but only allows Reads OR Writes.
 *	Having servers allow clients to think there is functionality
 *	when there isn't is a VERY bad practice. 
 *	The clients should be changed to allow for what the driver provides,
 *	not the other way around.
 */
#define READWRITE
#undef  READWRITE

#define RETRY		/* retry stat command */

#undef DEBUG

#ifndef	M350
#define	M350	0
#endif

#ifdef	DEBUG		/* debugging code */

#define	STATIC
#define	DEBUGPR(x,d,s)	{ if( x & m350Debug ) m350print(d,s); }

int	m350Debug = 0;

#define	DBINIT		(1<<0)
#define	DBOPEN		(1<<1)
#define	DBBOPEN		(1<<2)
#define	DBCOPEN		(1<<3)
#define	DBCLOSE		(1<<4)
#define	DBBCLOSE	(1<<5)
#define	DBCCLOSE	(1<<6)
#define	DBSTRAT		(1<<7)
#define	DBSTART		(1<<8)
#define	DBINTR		(1<<9)
#define	DBREAD		(1<<10)
#define	DBWRITE		(1<<11)
#define	DBIOCTL		(1<<12)
#define	DBCLR		(1<<13)
#define	DBDUMP		(1<<14)
#define	DBDOIO		(1<<15)

#else			/* no debugging code */

#define	STATIC	static
#define	DEBUGPR(x,d,s)

#endif

#ifdef	m88k
#define SPLEV(X)  ( X )  /* convert interrupt level to splx()  
					     compatible arg */
#else
#define SPLEV(X)  ( 0x2000 | ( X << 8 ))  /* convert interrupt level to splx()  
					     compatible arg */
#endif
#define	TOUT	2000			/* timeout */

#ifdef REL2
int m350maxbsize = DBUFMAX/2;
#else
extern int m350maxbsize;		/* maximum DMA buffer size in conf.c */
#endif

/*
 *	MVME350 Driver Minor Number Assignments
 *
 *	/dev/mt and	Minor	Contr-		Rewind on	Append or
 *	/dev/rmt names	Number	oller 	Unit	close ?		Truncate ?
 *	--------------	------	------	----	---------	----------
 *	m350_0t		0	0	0	Yes		Truncate
 *	m350_0tn	1	0	0	No		Truncate
 *	m350_0a		2	0	0	Yes		Append
 *	m350_0an	3	0	0	No		Append
 *	m350_1t		4	1	0	Yes		Truncate
 *	m350_1tn	5	1	0	No		Truncate
 *	m350_1a		6	1	0	Yes		Append
 *	m350_1an	7	1	0	No		Append
 *
 *	m350_Xt		X*4	X	0	Yes		Truncate
 *	m350_Xtn	X*4+1	X	0	No		Truncate
 *	m350_Xa		X*4+2	X	0	Yes		Append
 *	m350_Xan	X*4+3	X	0	No		Append
 */

/*
 *	MVME350 Defines
 */

#define NORETENS(d)	(d & 0x80) 	/* dev -> no retension on open */
#define	CTL(d)		((d>>2)&0x1f)	/* dev -> controller */
#define	UNIT(d)		(0)		/* dev -> unit */
#define	POS(d)		((d>>2)&0x1f)	/* dev -> table position */
#define	RWD(d)		(!(d&1))	/* dev -> rewind flag */
#define	TDEV(d)		(!(d&2))	/* dev -> default to truncate */
#define	ADEV(d)		(d&2)		/* dev -> default to append */

#define	MAXUNIT	1

/*
 *	MVME350 Tape Composite Status Bit Assignments
 */

#define	S_FM		(1<<0)	/* Filemark detected. */
#define	S_BLOCK		(1<<1)	/* Block in error not located. */
#define	S_UNREC		(1<<2)	/* Unrecoverable data error. */
#define	S_EOM		(1<<3)	/* End of Media. */
#define	S_WP		(1<<4)	/* Write protected. */
#define	S_OFFLINE	(1<<5)	/* Drive not online. */
#define	S_NOCART	(1<<6)	/* Cartridge not in place. */
#define	S_BYTE0		(1<<7)	/* Exception occurred in byte 0. */
#define	S_RESET		(1<<8)	/* Reset/power up occurred. */
#define	S_BOM		(1<<11)	/* Beginning of Media. */
#define	S_RETRIES	(1<<12)	/* Eight or more retries. */
#define	S_NODATA	(1<<13)	/* No data detected. */
#define	S_ILLEGAL	(1<<14)	/* Illegal command. */
#define	S_BYTE1		(1<<15)	/* Exception occurred in byte 1. */
#define	S_NOFM		(1<<16)	/* No file mark encountered. */
#define	S_NOTBOT	(1<<17)	/* Not at beginning of tape. */
#define	S_TRESET	(1<<18)	/* Tape reset did not occur. */

/*
 *	MVME350 Variables
 */

/*
 *	m350space.h
 */

extern	struct	iotime	m350stat[];
extern	struct	iobuf	m350tab[];
#ifndef REL2
extern	struct iostat	m350iostp[];	/* one per controller */
#endif

/*
 *	MVME350 iobuf usage
 *
 *	iobuf.h		mvme350.c	Usage
 *	-------		---------	-----
 *	b_flags		b_flags		Drive state
 *	b_forw
 *	b_back
 *	b_actf		b_actf		head (first) of I/O queue
 *	b_actl		b_actl		tail (last) of I/O queue
 *	b_dev		b_dev		major/minor device number
 *	b_active	b_active	busy flag
 *	b_errcnt	b_errcnt	error count
 *	io_erec		io_erec		error record
 *	io_nreg		io_nreg		number of registers to log upon error
 *	io_addr		io_addr		csr address
 *	io_mba		io_chan		channel structure address
 *	io_stp		io_stp		I/O statistics
 *	io_start	io_start	I/O statistics
 *	io_s1		io_bufp		buffer pointer for double buffering
 *	io_s2		io_bsize	buffer size
 *	io_s3		io_blkno	offset block number
 *	io_s4		io_nxrec	next record
 *	io_s5		io_freeq	free queue pointer
 *	io_s6		io_fileno	file number on tape
 *	io_s7		io_stat		status from last command
#ifdef	MULTITAPE
 *	io_s8		io_proc		process id
#else
 *	io_s8
#endif
#ifdef	READWRITE
*       io_s9(hiword)	io_moreflgs	more flags 
#endif
#ifdef	UNDERRUNS
*	io_s9		io_under	composite underrun count
#endif
 */

#define	io_chan		io_mba
#define	io_bufp		io_s1
#define	io_bsize	io_s2
#define	io_blkno	io_s3
#define	io_nxrec	io_s4
#define	io_freeq	io_s5
#define	io_fileno	io_s6
#define	io_stat		io_s7
#ifdef	MULTITAPE
#define	io_proc		io_s8
#endif
#ifdef READWRITE
#define io_moreflgs	io_s9
#endif
#ifdef	DEBUG
#define	UNDERRUNS
#define	io_under	io_s9
#endif

#ifdef READWRITE
/* io_moreflgs definitions (upper word), loword is io_proc */
#define RDWROPEN 0x00010000		/* opened for read or write, not both*/ 
#define READONLY 0x00020000		/* reads only are permitted. */
#define WRITEONLY 0x00040000		/* writes only are permitted. */
#endif

#define	INF	0x7fffffff		/* a very large number */
#define	B_CMD	B_FORMAT
#define	B_INIT	B_FORMAT
#define	B_ERASE	B_DONE
#define	B_FM	B_AGE			/* filemark encountered */
#define	B_OFF	B_ASYNC			/* drive offline */
#define	B_READOP B_DELWRI		/* read operation done */
#ifdef RETRY
#define B_STAT	B_VERIFY  		/* retry STAT command */
#endif
#ifdef	BYTESWAP
#define	B_BSWAP	B_STALE			/* byte swapping on */
#endif

#define	ENVPERCHAN	3		/* envelopes per channel */


extern struct b_packet	m350pkt[];	/* packets - one per controller */
extern struct b_envelope m350env[];	/* envelopes - three per controller */
extern struct b_channel m350chan[];	/* channels - one per controller */
extern struct b_envelope *m350freeq[];	/* free queue - one per controller */

#define	INVALID	0
#define	VALID	1
#define	S32DATA	0x0d			/* address modifier - 32 bit */
#define	S24DATA	0x3d			/* address modifier - 24 bit */


extern	int	m350packets;		/* number of packets */
extern	int	m350envelopes;		/* number of envelopes */


/*
 *	conf.h
 */

extern	struct b_device	*m350_addr[];
extern	int		m350_ilev[];
extern	int		m350_ivec[];
extern	int		m350_cnt;
extern	int		m350_bmaj;
extern	int		m350_cmaj;

/*
 *	local variables
 */

/* ipc command buffer */
STATIC	struct b_ipc	m350ipc = { "","",0,0,0,0,0x200,0,0,0,0,0,0,0,0 };
static  initted = 0;

/*
 *	MVME350 Procedures
 *
 *	m350init: initialize the mvme350 and the driver
 */

m350init()
{	register struct iobuf *dp;
	register struct b_channel *cp;
	register c = m350_cnt/MAXUNIT;
	extern short cputype;		/* identifies processor */
	register am;
	register n;
	
	DEBUGPR(DBINIT,0,"init");

	am = (cputype == PROC020 || cputype == PROC030)? S32DATA : S24DATA;

	for( n=0; n < c; n++ ) {
		if( bprobe( m350_addr[n], -1 ) )
			m350_addr[n] = NULL;
	}

	for( dp = &m350tab[0], n = 0; n < m350_cnt; n++, dp++ ) {
#ifdef REL2
		if(( dp->io_addr = (physadr)m350_addr[n/MAXUNIT] ) == NULL )
#else
		if(( dp->io_addr = (paddr_t)m350_addr[n/MAXUNIT] ) == NULL )
#endif
			continue;
		dp->io_nreg = sizeof(struct b_device);
#ifndef REL2
 		dp->io_stp = &m350iostp[n];
#endif
		dp->io_chan = (physadr)&m350chan[n];
		dp->io_freeq = (int)&m350freeq[n];
		dp->io_stat = 0;
		dp->io_bsize = m350maxbsize;
		dp->io_fileno = 0;
		dp->b_flags = B_INIT;		/* init 350 first time */

		/*
		 *	initialize channel and envelopes
		 */

		cp = (struct b_channel *)dp->io_chan;
		cp->c_prio = 0;

		{	register struct b_envelope *ep;
			register struct b_packet *pp;

			pp = &m350pkt[n];
			ep = &m350env[n*ENVPERCHAN];

			bppcrque(&m350freeq[n],ENVPERCHAN,pp,ep,0x6d333530,
				am,((cp->c_am == S24DATA)? 1 : 2));
		}

		if( bppcrchan(cp,dp->io_addr,&m350freeq[n],
			m350_ilev[n],m350_ivec[n]/sizeof(long),am) == -1 )
			dp->io_addr = NULL;

	}
}

/*
 *	m350open: open an mvme350 device
 */

STATIC m350open(dev,flag,raw)
register dev;
register flag;
register raw;
{	register struct iobuf *dp = &m350tab[POS(dev)];
	register cstat;
	int	m350strategy();

	DEBUGPR(DBOPEN,dev,"open");

	/*
	 *	if the device is nonexistent, then error
	 */

	if((CTL(dev) > m350_cnt)||(dp->b_flags&B_OPEN)||(dp->io_addr == NULL)) {
		u.u_error = ENXIO;
		return;
	}

	/*
	 *	will not permit read-and-write opens
	 */

	if( (flag&FREAD) && (flag&FWRITE) ) {
		u.u_error = ENXIO;
		return;
	}

#ifdef READWRITE
	/*
	 *	allow device to be opened for both read and write, but 
	 *	depending on first access, only allow read only or write only.
	 */

	if((flag&FREAD) && (flag&FWRITE)){
		dp->io_moreflgs &= ~( READONLY | WRITEONLY);
		dp->io_moreflgs |= RDWROPEN;
	}

#endif
	/*
	 *	set up the device table
	 */

	dp->b_flags &= ~B_OFF;
	dp->b_flags |= B_OPEN;
	dp->b_dev = makedev( m350_bmaj, dev );
	dp->io_blkno = 0;
	dp->io_nxrec = INF;

#ifdef	STATCMD
	/* The 150Mb Archive drive returns a bad status after
	a new tape has been placed in the drive.  The temporary solution
	is to issue 2 C_STAT commands and to ignore any error from
	the first.
	*/

#ifdef RETRY
	dp->b_flags |= B_STAT;
	if(m350command(dp,C_STAT,0)) 
#endif
		if(m350command(dp,C_STAT,0))
			u.u_error = 0;
	cstat = dp->io_stat;
	/*
	 *	if the unit is offline and the "O_NDELAY" flag
	 *	isn't used, or if the unit is write protected and
	 *	the file is being opened for writing, then error.
	 */

	if(( ((flag&FNDELAY)==0) && ( cstat&(S_OFFLINE|S_NOCART) ))
	||(( (flag&FWRITE) && (cstat&S_WP) )) ) {
		dp->b_flags &= ~B_OPEN;
		u.u_error = ENXIO;
		return;
	}
#endif
	/* 	Retension tape for all write opens unless "NORETENS" bit 
	 *	is set in minor device number.
	 *
	 */
	if( (flag&FWRITE) && (! NORETENS(dev))) {
		if(m350command(dp, C_RET, 0)) {
			u.u_error = ENXIO;
			dp->b_flags &= ~B_OPEN;
			return;
		}
	}
	/*
	 *	if being opened for a write, check for truncate or append
	 */

	if( flag&FWRITE ) {
		
		/* set default action */
		if( (flag&(FTRUNC|FAPPEND)) == 0 )
			flag |= ADEV(dev)? FAPPEND : FTRUNC;

		if( flag&FTRUNC ) {	/* O_TRUNC == truncate ... */

			/*
			 *	force a rewind ...
			 *	can't overwrite on streamer tape
			 */

			if( m350command(dp,C_REW,0) ) {
				u.u_error = ENXIO;
				dp->b_flags &= ~B_OPEN;
				return;
			}

		/*
		 *	otherwise append ...
		 *	if the unit is being opened with the "O_APPEND" flag,
		 *	then position the tape at the end of data.
		 */

		} else if( m350command(dp,C_SKEOD,-1) ) {
			u.u_error = ENXIO;
			dp->b_flags &= ~B_OPEN;
			return;
		}
	}

	/*
	 *	If the tape is opened as a character device (raw),
	 *	and if the O_NDELAY flag is not used, then allocate
	 *	a set of buffers to implement double buffering.
	 */

	if( raw && dp->io_bsize && ((flag&FNDELAY) == 0) ) {
#ifdef	MULTITAPE
		if( dp->io_bufp != NULL ) {
			if( u.u_procp->p_pid != dp->io_proc ) {
				dbufclose(dp->io_bufp);
				dp->io_bufp = NULL;
				printf("MVME350: DMA buffers discarded\n");
			} else {
				/* set up retry */
				dbufioctl( dp->io_bufp, DBUFRETRY, 0 );
			}
		} 

		if( dp->io_bufp == NULL ) {
#endif
			dp->io_bufp =
				dbufopen(dp->io_bsize,dp->b_dev,m350strategy);
			if( dp->io_bufp == NULL ) {
				dp->b_flags &= ~B_OPEN;
				return;
			}
#ifdef	MULTITAPE
		}
#endif
	}
#ifdef	MULTITAPE
	else if( dp->io_bufp ) {
		dbufclose(dp->io_bufp);
		dp->io_bufp = NULL;
		printf("MVME350: DMA buffers discarded\n");
	}

	dp->io_proc = u.u_procp->p_pid;
#endif


	DEBUGPR(DBOPEN,dev,"open successful");
}

/*
 *	m350bopen: open a block device
 */

m350bopen(dev,flag)
{
	DEBUGPR(DBBOPEN,dev,"bopen");
	m350open(dev,flag,0);
}

/*
 *	m350copen: open a character (raw) device
 */

m350copen(dev,flag)
{
	DEBUGPR(DBCOPEN,dev,"copen");
	m350open(dev,flag,1);
}

/*
 *	m350close: close an mvme350 device
 */

STATIC m350close(dev,flag,raw)
register dev;
{	register struct iobuf *dp = &m350tab[POS(dev)];

	DEBUGPR(DBCLOSE,dev,"close");

#ifdef READWRITE
	if( (flag&FREAD) && (flag&FWRITE))
		dp->io_moreflgs &= ~(RDWROPEN | READONLY | WRITEONLY);

#endif
	/*
	 *	if the io_bufp is non-null, then flush and
	 *	delete the DMA buffers used for double buffering.
	 */

	if( dp->io_bufp ) {
#ifdef	DEBUG
		if( ! raw )
			m350print(dev,"illegal close");
#endif
#ifdef	MULTITAPE
		if( (flag&FREAD) || ((dp->b_flags&B_ERROR) == 0) ) {
#endif
			dbufclose(dp->io_bufp);
			dp->io_bufp = NULL;
#ifdef	MULTITAPE
		} else {
			dp->b_flags &= ~B_ERROR;
			printf("MVME350: DMA buffers still active.\n");
		}
#endif
	}

	/*
	 *	if the unit was opened for writing, then write
	 *	a filemark to indicate end of file
	 */

	if( dp->b_flags&B_ERASE )
		dp->b_flags &= ~B_ERASE;

	else if( flag&FWRITE )
		m350command(dp,C_WEOF,0);

	/*
	 *	if the unit is a "rewind-on-close", then rewind the device
	 */

	/* if read and not at filemark terminate read */
	if((flag&FREAD) && ((dp->b_flags&B_FM) == 0)) {
		if( RWD(dev) )
			m350command(dp,C_DINIT,0);	/* do init for rewind */
		else if (dp->b_flags&B_READOP) {	/* if last op a read  */
			m350command(dp,C_TREAD,0);
			dp->b_flags &= ~B_READOP;
			}
	}

	if( RWD(dev) )
		m350command(dp,C_REW,0);

	dp->b_flags &= ~B_OPEN;

	DEBUGPR(DBCLOSE,dev,"close successful");
}

/*
 *	m350bclose: close a block device
 */

m350bclose(dev,flag)
{
	DEBUGPR(DBBCLOSE,dev,"bclose");
	m350close(dev,flag,0);
}

/*
 *	m350cclose: close a character device
 */

m350cclose(dev,flag)
{
	DEBUGPR(DBCCLOSE,dev,"cclose");
	m350close(dev,flag,1);
}

/*
 *	m350command: issue a command and call strategy
 */

STATIC m350command(dp,cmd,arg)
register struct iobuf *dp;
register cmd;
register arg;
{	register struct buf *bp;

	bp = geteblk();			/* get empty buffer */
	bp->b_proc = u.u_procp;
	bp->b_dev = dp->b_dev;
	bp->b_bcount = cmd;
	bp->b_blkno = arg;
	bp->b_error = 0;
	bp->b_flags |= B_CMD|B_PHYS;
	m350strategy(bp);
	iowait(bp);			/* wait until done */
	bp->b_flags &= ~(B_CMD|B_PHYS);
	brelse(bp);
	return(u.u_error);
}

/*
 *	m350strategy: perform an I/O
 */

m350strategy(bp)
register struct buf *bp;
{	register struct iobuf *dp = &m350tab[POS(bp->b_dev)];

	DEBUGPR(DBSTRAT,bp->b_dev,"strategy");

	if(((bp->b_flags&B_CMD ) == 0) && (bp->b_bcount&(BSIZE-1))) {
		/* no partial block I/O */
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		iodone(bp);
		return;
	}

#ifdef READWRITE
	 /* if device was opened for read/write then depending on first access,
	    permit reads only or writes only. 
	 */
	if(((bp->b_flags&B_CMD) == 0) && (dp->io_moreflgs & RDWROPEN)){
		if(!(dp->io_moreflgs & (READONLY | WRITEONLY))){
			if(bp->b_flags & B_READ)
				dp->io_moreflgs |= READONLY;
			else
				dp->io_moreflgs |= WRITEONLY;
		}
		else{
	     	      if(!(((bp->b_flags&B_READ) && (dp->io_moreflgs&READONLY))
		    || (((bp->b_flags&B_READ) == 0) && (dp->io_moreflgs&WRITEONLY))))
		        {	 /* cant mix reads and writes */
				bp->b_flags |= B_ERROR;
				bp->b_error = ENXIO;
				iodone(bp);
				return;
			}
		}
	}
#endif

	bp->b_start = lbolt;
	if(( bp->b_flags&B_CMD ) == 0 )
		m350stat[POS(bp->b_dev)].io_bcnt += btoc(bp->b_bcount);
	bp->av_forw = NULL;

	splx(SPLEV(m350_ilev[POS(bp->b_dev)]));

	if( dp->b_actf == NULL )
		dp->b_actf = bp;
	else
		dp->b_actl->av_forw = bp;
	dp->b_actl = bp;
	if( dp->b_active == 0 )
		m350start(dp);

	spl0();
}

/*
 *	m350start: start an I/O
 */

STATIC m350start(dp)
register struct iobuf *dp;
{	register struct buf *bp;
	register struct b_envelope *ep;
	struct b_channel *cp = (struct b_channel *)dp->io_chan;
	register cmd;
	register bpblkno;
	struct b_envelope *bppget();

	DEBUGPR(DBSTART,dp->b_dev,"start");

	/*
	 *	get next buffer ... if null return
	 */

	while(( bp = dp->b_actf ) != NULL ) {

		dp->b_active++;

		if(( bp->b_flags&B_CMD ) == 0 ) {

			/*
			 *	if the current block number is greater
			 *	than the next record number on the tape,
			 *	then it's an error (lseek(2) too far).
			 */

			if( (bpblkno = bp->b_blkno) > dp->io_nxrec ) {
				bp->b_error = EIO;
				goto abort;
			}

			/*
			 *	if the current block is equal to the
			 *	next record number, then we're at the
			 *	end of the tape file (on a filemark).
			 */

			if((bpblkno==dp->io_nxrec)&&(bp->b_flags&B_READ)) {
				clrbuf(bp);
				bp->b_resid = bp->b_bcount;
				goto next;
			}

			/*
			 *	set the next record number to one past the
			 *	buffer block number (on writes).
			 */

			if(( bp->b_flags&B_READ ) == 0 ) {
				dp->io_nxrec = bpblkno + bp->b_bcount/BSIZE;

				/*
				 *	no seeks on write
				 */

				if( dp->io_blkno != bp->b_blkno ) {
#ifdef	DEBUG
					printf("m350start: unexpected write\n");
#endif
					bp->b_error = EIO;
					goto abort;
				}
			} else if( dp->io_blkno > bpblkno ) {
				bp->b_error = ENXIO;
				goto abort;
			}

		}

		if( dp->b_flags&B_OFF )
			goto abort;

		/*
		 *	if command, then set command
		 */

		if( dp->b_flags&B_INIT )
			cmd = C_DINIT;
		else if( (dp->b_flags&B_ONCE) == 0 )
			cmd = C_WIPC;
		else if( bp->b_flags&B_CMD )
			cmd = bp->b_bcount;	/* see m350command */
		else
			cmd = (bp->b_flags&B_READ)? C_REOF : C_WRTL;

		/*
		 *	set up packet
		 */

	{	register struct b_packet *pp;

		if((ep = bppget(dp->io_freeq)) == NULL) {  /* no packets */
				bp->b_error = EIO;
				goto abort;
			}
		pp = ep->e_pkt;
		pp->p_cmd = cmd;
		pp->p_dtype = D_TAPE;	/* tape device type */

		switch(cmd) {

		case C_REOF:	/* read to eof */
#ifdef	m88k
			pp->pa_pp_blknohi = dp->io_fileno>>16;
			pp->pa_pp_blknolo = dp->io_fileno;
#else
			pp->p_blkno = dp->io_fileno;
#endif
			dp->b_flags |= B_READOP;	/* read done */

		case C_WRTL:	/* write logical */

			if( bp->b_flags&B_PHYS ) {
				dp->io_stp->io_ops++;

			} else if( dp->io_blkno == bpblkno ) {
				dp->io_stp->io_ops++;

			} else	/* do block read and throw away later */
				dp->io_stp->io_misc++;

#ifdef	m88k
			pp->p_sahi = (unsigned)paddr(bp)>>16;
			pp->p_salo = (unsigned)paddr(bp);
#else
			pp->p_sa = (unsigned)paddr(bp);
#endif
#ifdef	m88k
			pp->p_counthi = (bp->b_bcount/BSIZE)>>16;
			pp->p_countlo = bp->b_bcount/BSIZE;
#else
			pp->p_count = bp->b_bcount/BSIZE;
#endif
			if( dp->io_blkno < bp->b_blkno ) {
#ifdef	m88k
				/* seek forward ... but not too far */
				if(((pp->p_counthi<<16) + pp->p_countlo) > (bp->b_blkno-dp->io_blkno))
					pp->p_counthi = bp->b_blkno-dp->io_blkno>>16;
					pp->p_countlo = bp->b_blkno-dp->io_blkno;
#else
				/* seek forward ... but not too far */
				if(pp->p_count > (bp->b_blkno-dp->io_blkno))
					pp->p_count = bp->b_blkno-dp->io_blkno;
#endif
			}
			break;

		case C_WIPC:
#ifdef	BYTESWAP
			if( dp->b_flags&B_ONCE ) 
				pp->p_dtype = D_TAPE;	/* tape device type */
			else
#endif
				pp->p_dtype = D_IPC;	/* IPC device type */
#ifdef	m88k
			pp->p_sahi = (int)&m350ipc>>16;
			pp->p_salo = (int)&m350ipc;
#else
			pp->p_sa = (int)&m350ipc;
#endif
#ifdef	BYTESWAP
#ifdef	m88k
			pp->op_pp_cp1 = ((dp->b_flags&B_BSWAP) == B_BSWAP);
#else
			pp->p_badcnt = ((dp->b_flags&B_BSWAP) == B_BSWAP);
#endif
#endif
			break;

		case C_WEOF:	/* write eof */
		case C_TREAD:	/* terminate read */
#ifdef	m88k
			pp->p_counthi = 0;
			pp->p_countlo = 0;
#else
			pp->p_count = 0;
#endif
			dp->io_stp->io_misc++;
			break;

#ifdef	STATCMD
		case C_STAT:	/* get status */
#endif
		case C_DINIT:	/* init */
		case C_REW:	/* rewind */
		case C_RET:	/* retension */
		case C_ERASE:	/* erase */
			dp->io_fileno = 0;
			dp->io_stp->io_misc++;
			break;

		case C_SKEOD:	/* seek to eod */
			dp->io_stp->io_misc++;
#ifdef	m88k
			pp->pa_pp_blknohi = bp->b_blkno>>16;
			pp->pa_pp_blknolo = bp->b_blkno;
#else
			pp->p_blkno = bp->b_blkno;
#endif
			break;
#ifdef	DEBUG
		default:
			printf("MVME350 bad command: %x\n", cmd );
			break;
#endif
		}

		blkacty |= (1<<M350);

		/* set up I/O start time */
		if( dp->io_start == 0 )
			dp->io_start = lbolt;
		 

	}
		/*
		 *	everything ok ... send
		 */

	{	register struct b_device *rp = (struct b_device *)dp->io_addr;

		bppsend(cp,rp,ep);
		return;
	}

		/*
		 *	come here on error
		 */

	abort:
		bp->b_flags |= B_ERROR;
	next:
		dp->b_actf = bp->av_forw;
		iodone(bp);
	}

	dp->b_active = 0;
}

/*
 *	m350intr: interrupt handler
 */

m350intr(ctl)
{	register struct iobuf *dp = &m350tab[ctl];
	register struct buf *bp = dp->b_actf;
	register dev = bp->b_dev;
	register struct b_packet *pp;
	register resid;
	register stat;
	register termcnt;
	register cmd;
	struct b_envelope *ep;
#ifdef	UNDERRUNS
	int under;
#endif
	struct b_envelope *bpprcv();

	DEBUGPR(DBINTR,dev,"intr");

	{	
		struct b_channel *cp = (struct b_channel *)dp->io_chan;

		/*
		 *	get status envelope
		 */

#ifdef	DEBUG
		if(m350_addr[ctl] == NULL)	
			return;
#endif


		if((ep = bpprcv(cp)) == NULL ) {
#ifdef	DEBUG
			m350print(dev,"spurious interrupt");
#endif
			return;
		}


#ifdef	DEBUG
		if( ep->e_pkt == NULL )
#ifdef REL2
			panic("MVME350: null status");
#else
 			cmn_err(CE_PANIC, "MVME350: null status");
#endif
#endif

		pp = ep->e_pkt;


	}

	if( bp == NULL ) {
		logstray(dp->io_addr);
		return;
	}

	blkacty &= ~(1<<M350);
	dp->b_active = 0;

#ifdef	m88k
	dp->io_fileno = pp->ops_opstat[2];
#else
	dp->io_fileno = pp->p_opstat[2];
#endif
	cmd = pp->p_cmd;
#ifdef	m88k
	termcnt = (pp->p_termcnthi << 16) + pp->p_termcntlo;
#else
	termcnt = pp->p_termcnt;
#endif
	dp->io_stat = stat = pp->p_cstat;
	pp->p_cstat = 0;

#ifdef UNDERRUNS
#ifdef	m88k
	under = pp->ops_opstat[0];
#else
	under = pp->p_opstat[0];
#endif
#endif
	bppfree(dp->io_freeq,ep);	/* give back envelope */

#ifdef RETRY
	if( dp->b_flags&B_STAT ) { 	/* Ignore the first error from C_STAT */
		stat &= ~S_ERROR;
		dp->b_flags &= ~B_STAT;
	}
#endif
	if(stat&S_ERROR) {

		if(( stat&S_EOM ) && (( cmd == C_WRTL )) || ( cmd == C_WEOF ))
			stat &= ~S_FATAL;

		else if((stat&S_FM) == 0) {
			/* print error message(s) */
			m350prerr(dev,cmd,stat);
#ifndef REL2
			dp->b_dev = makedev(M350, minor(bp->b_dev));
#endif
			fmtberr(dp,stat);
			if( stat&S_OFFLINE )
				dp->b_flags |= B_OFF;
		}
	}

	/*
	 *	if fatal error
	 */

	if( stat&S_FATAL ) {
		dp->b_flags |= B_INIT;
		bp->b_flags |= B_ERROR;

	/*
	 *	if command is a write IPC
	 */

	} else if( cmd == C_WIPC ) {
		if(( dp->b_flags&B_ONCE) == 0) {	/* initial WIPC */
			dp->b_flags |= B_ONCE;		/* changing blocksize */
			m350start(dp);
			initted = 1;			/* blocksize changed */
			return;
		}
	
	
	/*
	 *	if an init command has been issued
	 */

	} else if( dp->b_flags&(B_INIT) ) {
		dp->b_flags &= ~(B_INIT);
		m350start(dp);		/* do command */
		return;

	/*
	 *	if the command is a read or write
	 */

	} else if((cmd == C_REOF ) || (cmd == C_WRTL )) {
		resid = termcnt/BSIZE;
#ifdef	UNDERRUNS
		dp->io_under += under;
#endif
		dp->io_blkno += resid;
		if( dp->io_blkno <= bp->b_blkno ) {
			m350start(dp);	/* ignore I/O ... continue */
			return;
		}
		if( stat&S_FM ) {
			dp->io_blkno = 0;
			dp->io_nxrec = bp->b_blkno + resid;
		}
	}

	/*
	 *	I/O is complete ... continue
	 */

	logberr(dp,stat);
	dp->b_actf = bp->av_forw;

	if(((bp->b_flags&B_CMD ) == 0) && 
	(!(stat&S_FATAL) && ((cmd == C_WRTL ) || (cmd == C_REOF)))) {
		resid = bp->b_bcount - termcnt;
		bp->b_resid = (resid <= 0)? 0 : resid;
		if( stat & S_FM )
			dp->b_flags |= B_FM;
		else
			dp->b_flags &= ~B_FM;

	} else if((bp->b_flags&B_CMD ) == 0) 
		bp->b_resid = bp->b_bcount;


	{	register struct iotime *ip = &m350stat[ctl];

		ip->io_resp += lbolt - bp->b_start;
		ip->io_act += lbolt - dp->io_start;
		dp->io_start = 0;
	}

	iodone(bp);		/* complete the I/O */
	m350start(dp);		/* start next I/O */
}

#ifndef REL2

m350breakup(bp)
{
	dma_breakup(m350strategy,bp,BSIZE);
}

#endif

/*
 *	m350read: character device read
 */

m350read(dev)
register dev;
{	register struct iobuf *dp = &m350tab[POS(dev)];

	DEBUGPR(DBREAD,dev,"read");

#ifdef READWRITE
	if(dp->io_moreflgs&RDWROPEN){
		if(dp->io_moreflgs&WRITEONLY){
			u.u_error = ENXIO;
			return;
		}
		else
			dp->io_moreflgs |= READONLY;
	}
#endif
	
	if( dp->io_bufp )
		dbufread( dp->io_bufp );
	else
#ifdef REL2
		physio( m350strategy, 0, dev, B_READ );
#else
 		physio( m350breakup, 0, dev, B_READ );
#endif
}

/*
 *	m350write: character device write
 */

m350write(dev)
{	register struct iobuf *dp = &m350tab[POS(dev)];

	DEBUGPR(DBWRITE,dev,"write");

#ifdef READWRITE
	if(dp->io_moreflgs&RDWROPEN){
		if(dp->io_moreflgs&READONLY){
			u.u_error = ENXIO;
			return;
		}
		else
			dp->io_moreflgs |= WRITEONLY;
	}
#endif
	
	if( dp->io_bufp ) {
		dbufwrite( dp->io_bufp );
#ifdef	MULTITAPE
		if( u.u_error || u.u_count != 0 )
			dp->b_flags |= B_ERROR;	/* remember error */
#endif
	} else
#ifdef REL2
		physio( m350strategy, 0, dev, B_WRITE );
#else
 		physio( m350breakup, 0, dev, B_WRITE );
#endif
}

/*
 *	m350ioctl: ioctl support for the MVME350
 */

m350ioctl( dev, cmd, arg, mode )
register dev;
{	register struct iobuf *dp = &m350tab[POS(dev)];

	DEBUGPR(DBIOCTL,dev,"ioctl");

	switch(cmd) {
	case M350REWIND:	/* rewind tape */
		m350command(dp,C_REW,0);
		break;
	case M350ERASE:		/* erase tape */
		if( m350command(dp,C_ERASE,0) == 0 )
			dp->b_flags |= B_ERASE;
		break;
	case M350RETENSION:	/* retension tape */
		m350command(dp,C_RET,0);
		break;
	case M350WRTFM:		/* write filemark */
		m350command(dp,C_WEOF,0);
		break;
	case M350RDFM:		/* read filemark */
		m350command(dp,C_SKEOD,dp->io_fileno);
		break;
#ifdef	BYTESWAP
	case	M350BYTESWAP:	/* set/reset byte swapping */
		if( arg != 0 )
			dp->b_flags |= B_BSWAP;
		else
			dp->b_flags &= ~B_BSWAP;
		m350command(dp,C_WIPC,0);
		break;
#endif
	case M350SETDMA:	/* set DMA buffer size */
		if( ! suser() ) {
			u.u_error = EPERM;
			return;
		}

		if( dp->io_bufp == NULL )
			dp->io_bsize = arg;

		else {
			dbufioctl(dp->io_bufp,DBUFSET,arg);
			if( u.u_error )
				dp->io_bsize = 0;
			else
				dp->io_bsize = arg;
		}
		break;
	case M350GETDMA:	/* get DMA buffer size */
		if(( dp->io_bufp == NULL ) || ( dp->io_bsize == 0 )) {
			copyout( &dp->io_bsize, arg, sizeof(dp->io_bsize) );
		} else
			dbufioctl(dp->io_bufp,DBUFGET,arg);
		break;

#ifdef	UNDERRUNS
	case (M350IOCTL|'U'):	/* get underrun count */
		copyout( &dp->io_under, arg, sizeof( dp->io_under ) );
		break;
#endif

	default:
		u.u_error = EINVAL;
		break;
	}
}

/*
 *	m350clr: clear the mvme350 after a power fail
 */

m350clr()
{
	DEBUGPR(DBCLR,0,"clr");
}

/*
 *	m350print: driver print procedure
 */

m350print(d,s)
{	printf("MVME350: %s on controller %d, drive %d\n", s, CTL(d), UNIT(d) );
}

/*
 *	m350prerr: print an MVME350 error message
 */

STATIC m350prerr(dev,cmd,stat)
{
	if( (stat&S_ERROR) == 0 )
		return;
	m350print(dev,"Error");
	m350prcmd(cmd);
	m350prstat(stat);
}

/*
 *	m350prcmd: print MVME350 command
 */

m350prcmd(cmd)
register unsigned short cmd;

{	switch(cmd) {
	case C_DINIT:	m350prm("Initialize device");	break;
	case C_READL:	m350prm("Read logical");		break;
	case C_STAT:	m350prm("Read status");		break;
	case C_REOF:	m350prm("Read blocks");		break;
	case C_WRTL:	m350prm("Write blocks");		break;
	case C_WIPC:	m350prm("Write ipc descriptor");	break;
	case C_WEOF:	m350prm("Write eof");		break;
	case C_REW:	m350prm("Rewind tape");		break;
	case C_RET:	m350prm("Retension tape");		break;
	case C_ERASE:	m350prm("Erase tape");		break;
	case C_SKEOD:	m350prm("Seek to end-of-data");	break;
	case C_TREAD:	m350prm("Terminate read");		break;
	default:
		printf("MVME350: Unknown command %x.\n", cmd );
		break;
	}
}

/*
 *	m350prstat: print MVME350 composite status
 */

STATIC m350prstat(stat)
register unsigned stat;

{
	if( stat == 0 ) {
		m350prm("No error");
		return;
	}

	if (stat&S_FM)		m350prm("Filemark detected");
#ifdef	DEBUG
	if (stat&S_BLOCK)	m350prm("Block in error not located");
#endif
	if (stat&S_UNREC)	m350prm("Unrecoverable data error");
	if (stat&S_EOM)		m350prm("End of Media");
	if (stat&S_WP)		m350prm("Write protected");
	if (stat&S_OFFLINE)	m350prm("Drive not online");
	if (stat&S_NOCART)	m350prm("Cartridge not in place");
#ifdef	DEBUG
	if (stat&S_BYTE0)	m350prm("Exception occurred in byte 0");
	if (stat&S_RESET)	m350prm("Reset/power up occurred");
#endif
	if (stat&S_BOM)		m350prm("Beginning of Media");
#ifdef	DEBUG
	if (stat&S_RETRIES)	m350prm("Eight or more retries");
#endif
	if (stat&S_NODATA)	m350prm("No data detected");
#ifdef	DEBUG
	if (stat&S_ILLEGAL)	m350prm("Illegal command");
	if (stat&S_BYTE1)	m350prm("Exception occurred in byte 1");
#endif
	if (stat&S_NOFM)	m350prm("No file mark encountered");
	if (stat&S_NOTBOT)	m350prm("Not at beginning of tape");
	if (stat&S_TRESET)	m350prm("Tape reset did not occur");
#ifdef	DEBUG
	if (stat&S_CTOD)	m350prm("Controller-To-Device");
	if (stat&S_ITOC)	m350prm("IPC-To-Controller");
	if (stat&S_STOI)	m350prm("System-To-IPC");
#endif
	if (stat&S_TIMEOUT)	m350prm("Timeout");
	if (stat&S_BADUNIT)	m350prm("Bad Unit");
	if (stat&S_BADDEV)	m350prm("Bad Device");
#ifdef	DEBUG
	if (stat&S_BADCMD)	m350prm("Bad Command");
	if (stat&S_ABORT)	m350prm("IPC Abort");
	if (stat&S_ETERM)	m350prm("Early Termination");
	if (stat&S_CPUABORT)	m350prm("CPU Abort");
	if (stat&S_FATAL)	m350prm("Fatal Error");
	if (stat&S_ERROR)	m350prm("Error");
#endif
}

/*
 *	m350prm: print message with MVME: at front
 */

STATIC m350prm(s)
{	printf("MVME350: %s.\n", s );
}


/*
 *
 *		Dump routines
 *
 */

/*
 *	m350dump: dump memory to tape
 */

m350dump()
{	
	int m350doio();
	extern	dev_t dumpdev;
	register dev = dumpdev;
	register struct iobuf *dp = &m350tab[POS(dev)];
	struct b_packet packet;
	register struct b_packet *pp = &packet;

	DEBUGPR(DBDUMP,dev,"dump");

	if((CTL(dev) > m350_cnt)) {
		printf("Dump device controller not present\n");
		return;
	}
	spl7();
	pp->p_cmd = C_DINIT;
	if(m350sendcmd(pp,dev,D_TAPE))
		return(-1);
	if(initted == 0) {
		pp->p_cmd = C_WIPC;
#ifdef	m88k
		pp->p_sahi = (int)&m350ipc>>16;
		pp->p_salo = (int)&m350ipc;
#else
		pp->p_sa = (int)&m350ipc;
#endif
		if(m350sendcmd(pp,dev,D_IPC))
			return(-1);
	} 

	dp->io_blkno = INF - 1;
	/*	until memdump.c  is ported
	memdump(m350doio);
	*/

	pp->p_cmd = C_WEOF;
	if(m350sendcmd(pp,dev,D_TAPE))
		return(-1);

	pp->p_cmd = C_REW;
	if(m350sendcmd(pp,dev,D_TAPE))
		return(-1);
}

/*
 *	m350doio: write one block of memory
 */

STATIC m350doio(dev,blkno,maddr,rdwr)
register dev;		/* device # */
register blkno;		/* block number on disk (512-byte) */
register char *maddr;	/* memory address */
register rdwr;		/* B_READ or B_WRITE */
{	register struct iobuf *dp = &m350tab[POS(dev)];
	struct b_packet packet;
	register struct b_packet *pp = &packet;
	register i;

	DEBUGPR(DBDOIO,dev,"doio");

	/* set up packet */
	pp->p_cmd = (rdwr == B_READ) ? C_REOF : C_WRTL;
#ifdef	m88k
	pp->pa_pp_blknohi = 0;
	pp->pa_pp_blknolo = 0;
#else
	pp->p_blkno = 0;
#endif
#ifdef	m88k
	pp->p_sahi = (unsigned) maddr>>16;
	pp->p_salo = (unsigned) maddr;
#else
	pp->p_sa = (unsigned) maddr;
#endif
#ifdef	m88k
	pp->p_counthi = 0;
	pp->p_countlo = 1;
#else
	pp->p_count = 1;		/* one block */
#endif

	if(m350sendcmd(pp,dev,D_TAPE))
		return(-1);

	if(dp->io_blkno + 1 < blkno)	/* skip hole */
		for(i=dp->io_blkno+1;i<blkno;i++)
			m350sendcmd(pp,dev,D_TAPE);
	dp->io_blkno = blkno;
}



/*
 *	m350sendcmd - send command to vme350
 */

STATIC m350sendcmd(pp,dev,type)
register struct b_packet *pp;
register dev;
register type;
{	register struct iobuf *dp = &m350tab[POS(dev)];
	register struct b_envelope *ep;
	register struct b_packet *cpp;
	register tout = INF;
	struct b_envelope *bppget();
	struct b_envelope *bpprcv();

	if((ep = bppget(dp->io_freeq)) == NULL) {  /* no packets */
		m350print(dev,"no free envelopes");
		return(-1);
	}

	cpp = ep->e_pkt;

	/* copy packet */

	cpp->p_cmd = pp->p_cmd;
#ifdef	m88k
	cpp->pa_pp_blknohi = pp->pa_pp_blknohi;
	cpp->pa_pp_blknolo = pp->pa_pp_blknolo;
#else
	cpp->p_blkno = pp->p_blkno;
#endif
#ifdef	m88k
	cpp->p_sahi = pp->p_sahi;
	cpp->p_salo = pp->p_salo;
#else
	cpp->p_sa = pp->p_sa;
#endif
#ifdef	m88k
	cpp->p_counthi = pp->p_counthi;
	cpp->p_countlo = pp->p_countlo;
#else
	cpp->p_count = pp->p_count;
#endif
	cpp->p_dtype = type;	/* device type */
	
	bppsend(dp->io_chan,dp->io_addr,ep);

	/*
	 * wait till packet put on status tail 
	 */

	while ((ep = bpprcv(dp->io_chan)) == NULL && --tout > 0)
		;

	if(tout <= 0) {
		m350print(dev,"timeout while waiting for completion");
		return(-1);
	}

	pp = ep->e_pkt;


	if( pp->p_cstat&S_ERROR ) 		/* an error occurred */
		/* print error message(s) */
		m350prerr(dev,pp->p_cmd,pp->p_cstat);

	bppfree(dp->io_freeq,ep);
	return(0);
}
